package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2020
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.Trade; // (time,price,shares) container
import com.polytechnik.realtimebook.TraderInterface; // for TIME_SCALE=1000000000L
import com.polytechnik.trading.SimpleSupportPriceStateSaver;
import com.polytechnik.trading.SimpleSupportPrice;
import com.polytechnik.trading.SimpleSupportPriceLegendreShifted;
import com.polytechnik.trading.SimpleSupportPriceLaguerre;
import com.polytechnik.trading.SimpleSupportPriceMonomials;


/** A reference implementation of dynamic equation non--stationary solution, support price version.
 * @version $Id: CallAMuseOfCashFlowAndLiquidityDeficitWithSimpleSupportPrice.java,v 1.10 2022/09/14 00:57:03 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class CallAMuseOfCashFlowAndLiquidityDeficitWithSimpleSupportPrice {

static int SimpleSupportPrice_outputDataHeader(final java.io.PrintWriter out)
{
    out.println(
	"|1:T\t2:shares\t3:P_read"
	+"\t4:p_last"
	+PrintFields.getEVXDataHeader("I",5)
	+"\t11:"+"sst.P_of_last_spike_in_I"
	+"\t12:"+"sst.getSumDp()"
	+"\t13:"+"x_at_last_spike_in_I"
	+"\t14:"+"w_at_last_spike_in_I"
	+"\t15:"+"dPsdt_at_last_spike_in_I"
	+"\t16:"+"DIRdPs"
	+"\t17:"+"PsChange_at_last_spike_in_I"
		);
    return 1;
}

static int SimpleSupportPrice_outputData(final java.io.PrintWriter out,final long currenttime,final double price,final int shares,final SimpleSupportPriceStateSaver r)
{
    out.println(String.valueOf(currenttime)+"\t"+shares+"\t"+price
		+"\t"+(r.sst.P)
		+PrintFields.getEVXDataData(r.sst.I)
		+"\t"+r.sst.P_of_last_spike_in_I
		+"\t"+r.sst.getSumDp() // different from P_of_last_spike_in_I in a constant
		+"\t"+r.x_at_last_spike_in_I
		+"\t"+r.w_at_last_spike_in_I
		+"\t"+r.dPsdt_at_last_spike_in_I
		+"\t"+r.DIRdPs
		+"\t"+r.PsChange_at_last_spike_in_I
		);
    return 1;
}


public static class TradeDataProcessorSimpleSupportPrice extends TradeDataProcessor<SimpleSupportPriceStateSaver> {
    final SimpleSupportPrice p;
    public TradeDataProcessorSimpleSupportPrice(final MuseConfig c,final SimpleSupportPrice p){super(c);this.p=p;}
    SimpleSupportPriceStateSaver ref=null;
    public int outputDataHeader(final java.io.PrintWriter out) throws java.io.IOException {return SimpleSupportPrice_outputDataHeader(out);}
    public int outputDataFooter(final java.io.PrintWriter out) throws java.io.IOException {
	out.println("| calculations done");
	return 1;
    }

    public int processOneReadLine(final java.io.PrintWriter out,final long currenttime,final Trade t) throws java.io.IOException{
	final var r=calculateData(currenttime,t);
	if(r==null) return 0; // skip N/A data
	return SimpleSupportPrice_outputData(out,currenttime,t.price,t.shares,r);
    }

    SimpleSupportPriceStateSaver calculateData(final long currenttime,final Trade t){
	// No buffer is used, pure recurrence:
	// only "actual" trades carry last price, otherwise price field is has limit order book meaning.
	if(t.shares>0){
	    // Simplistic model: run it on execution events only.
	    return this.ref=p.getNext(currenttime,t.price,t.shares,this.ref);
	} else {
	    //  limit order book events. t.price is order price meaning, skip for now
	}
	return null;
    }
} // END OF class TradeDataProcessorSimpleSupportPrice


public static void main(final String [] args) throws java.io.IOException
{
    final MuseConfig c=new MuseConfig(args);
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }
    else if(c.measure.equals("SimpleSupportPriceLegendreShifted"))
    {
	new TradeDataProcessorSimpleSupportPrice(c,new SimpleSupportPriceLegendreShifted(c.n,c.tau*TraderInterface.TIME_SCALE)).processFiles();
    }
    else if(c.measure.equals("SimpleSupportPriceLaguerre"))
    {
	new TradeDataProcessorSimpleSupportPrice(c,new SimpleSupportPriceLaguerre(c.n,c.tau*TraderInterface.TIME_SCALE)).processFiles();
    }
    else if(c.measure.equals("SimpleSupportPriceMonomials"))
    {
	new TradeDataProcessorSimpleSupportPrice(c,new SimpleSupportPriceMonomials(c.n,c.tau*TraderInterface.TIME_SCALE)).processFiles();
    }
    else throw new RuntimeException("The measure "+c.measure+" is not supported, allowed values:{SimpleSupportPriceLegendreShifted|SimpleSupportPriceLaguerre|SimpleSupportPriceMonomials}");
}

} // END OF class CallAMuseOfCashFlowAndLiquidityDeficitWithSimpleSupportPrice
