package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2020
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.Trade; // (time,price,shares) container
import com.polytechnik.realtimebook.TraderInterface; // for TIME_SCALE=1000000000L
import com.polytechnik.trading.SupportPrice;
import com.polytechnik.trading.SupportPriceLegendreShifted;
import com.polytechnik.trading.SupportPriceLaguerre;
import com.polytechnik.trading.SupportPriceMonomials;


/** A reference implementation of dynamic equation non--stationary solution.
 * @version $Id: CallAMuseOfCashFlowAndLiquidityDeficitWithSupportPrice.java,v 1.54 2022/09/14 00:57:03 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class CallAMuseOfCashFlowAndLiquidityDeficitWithSupportPrice {

static int SupportPrice_outputDataHeader(final java.io.PrintWriter out)
{
    out.println(
	"|1:T\t2:shares\t3:P_read"
	+"\t4:p_last"
	+"\t5:"+"pi_average"
	+"\t6:"+"pt_average"
	+PrintFields.getEVXDataHeader("I",7)
	+"\t13:"+"dpdt_0"
	+"\t14:"+"p_IH"
	+"\t15:"+"pt_IH"
	+"\t16:"+"dpdt_IH"
	+"\t17:"+"var1pI_IH"
	+"\t18:"+"var1pI_x"
	+"\t19:"+"sst.P_of_last_spike_in_I"
	+"\t20:"+"sst.getSumDp()"
	+"\t21:"+"w_psiH"
	+"\t22:"+"x_at_last_spike_in_I"
	+"\t23:"+"w_at_last_spike_in_I"
	+"\t24:"+"dpdt_at_last_spike_in_I"
	+"\t25:"+"dPsdt_at_last_spike_in_I"
	+"\t26:"+"DIRdPs"
	+"\t27:"+"PsChange_at_last_spike_in_I"
		);
    return 1;
}

static <RES extends SupportPrice<?>> int SupportPrice_outputData(final java.io.PrintWriter out,final long currenttime,final Trade t,final RES r)
{
    out.println(String.valueOf(currenttime)+"\t"+t.shares+"\t"+t.price
		+"\t"+(r.p_last+r.p_offset)
		+"\t"+(r.pi_average+r.p_offset)
		+"\t"+(r.pt_average+r.p_offset)
		+PrintFields.getEVXDataData(r.I)
		+"\t"+r.dpdt_0
		+"\t"+(r.p_IH+r.p_offset)
		+"\t"+(r.pt_IH+r.p_offset)
		+"\t"+r.dpdt_IH
		+"\t"+r.var1pI_IH
		+"\t"+r.var1pI_x
		+"\t"+r.sst.P_of_last_spike_in_I
		+"\t"+r.sst.getSumDp() // different from P_of_last_spike_in_I in a constant
		+"\t"+r.w_psiH
		+"\t"+r.x_at_last_spike_in_I
		+"\t"+r.w_at_last_spike_in_I
		+"\t"+r.dpdt_at_last_spike_in_I
		+"\t"+r.dPsdt_at_last_spike_in_I
		+"\t"+r.DIRdPs
		+"\t"+r.PsChange_at_last_spike_in_I
		);
    return 1;
}


public static abstract class TradeDataProcessorSupportPrice<RES extends SupportPrice<?>> extends TradeDataProcessor<RES> {
    public TradeDataProcessorSupportPrice(final MuseConfig c){super(c);}
    RES ref=null;
    public int outputDataHeader(final java.io.PrintWriter out) throws java.io.IOException {return SupportPrice_outputDataHeader(out);}
    public int outputDataFooter(final java.io.PrintWriter out) throws java.io.IOException {
	out.println("| calculations done");
	return 1;
    }

    @Override public int processOneReadLine(final java.io.PrintWriter out,final long currenttime,final Trade t) throws java.io.IOException{
	return SupportPrice_outputData(out,currenttime,t,calculateData(currenttime,t));
    }

    abstract RES calculateData(final long currenttime,final Trade t);

} // END OF class TradeDataProcessorSupportPrice


public static void main(final String [] args) throws java.io.IOException
{
    final MuseConfig c=new MuseConfig(args);
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }
    else if(c.measure.equals("SupportPriceLegendreShifted"))
    {
	new TradeDataProcessorSupportPrice<SupportPriceLegendreShifted>(c){
	    public SupportPriceLegendreShifted calculateData(final long currenttime,final Trade t){
		// actual trades has >0 shares
		if(t.shares>0) addTradeToBuffer(t);
		return this.ref=
		    SupportPriceLegendreShifted.calculate(c.n,currenttime,c.tau*TraderInterface.TIME_SCALE,buf,n_trades_in_buffer,this.ref);
	    }
	}.processFiles();
    }
    else if(c.measure.equals("SupportPriceLaguerre"))
    {
	new TradeDataProcessorSupportPrice<SupportPriceLaguerre>(c){
	    public SupportPriceLaguerre calculateData(final long currenttime,final Trade t){
		// actual trades has >0 shares
		if(t.shares>0) addTradeToBuffer(t);
		return this.ref=
		    SupportPriceLaguerre.calculate(c.n,currenttime,c.tau*TraderInterface.TIME_SCALE,buf,n_trades_in_buffer,this.ref);
	    }
	}.processFiles();
    }
    else if(c.measure.equals("SupportPriceMonomials"))
    {
	new TradeDataProcessorSupportPrice<SupportPriceMonomials>(c){
	    public SupportPriceMonomials calculateData(final long currenttime,final Trade t){
		// actual trades has >0 shares
		if(t.shares>0) addTradeToBuffer(t);
		return this.ref=
		    SupportPriceMonomials.calculate(c.n,currenttime,c.tau*TraderInterface.TIME_SCALE,buf,n_trades_in_buffer,this.ref);
	    }
	}.processFiles();
    }
    else throw new RuntimeException("The measure "+c.measure+" is not supported, allowed values:{SupportPriceLegendreShifted|SupportPriceLaguerre|SupportPriceMonomials}");
}

} // END OF class CallAMuseOfCashFlowAndLiquidityDeficitWithSupportPrice
