package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.*;

/** A simple image interpolation. Two multiindex constraints are used.
 * 1. each x,y power is <= n_x; n_y
 * 2. power(x)+power(y)=D.
 * Similar to ExampleImageInterpolation.scala of https://arxiv.org/abs/1511.01887
 * @version $Id: ExampleImageInterpolation2.java,v 1.23 2024/04/21 19:51:20 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
class ExampleImageInterpolation2 {
    final String imageurl;
    final int dimx,dimy;
    final double [] origraster;

    ExampleImageInterpolation2(final String imageurl) throws java.io.IOException,java.net.URISyntaxException {
	this.imageurl=imageurl;
	final java.awt.image.BufferedImage image=
	    javax.imageio.ImageIO.read(new java.net.URI(imageurl).toURL());
	this.dimx=image.getWidth();
	this.dimy=image.getHeight();
	this.origraster=image.getData().getPixels(0, 0, dimx, dimy, new double [dimx*dimy]);
    }

    @FunctionalInterface
    public interface PixelFunction<T> {
	T apply(int px,int py);
    }

    double fromPixelToX(final int px){ return 2.0*px/(dimx-1.0)-1;}
    double fromPixelToY(final int py){ return 2.0*py/(dimy-1.0)-1;}
    double getF(final int px,final int py){return origraster[px+dimx*py];}
    double getWeight(final int px,final int py){return 1;}

    public DataReadObservationVectorXF getImageAsSample(PixelFunction<double []> getXinBasis,
							java.util.List<String> Xheader){
	final java.util.ArrayList<ObservationVectorXF> data=new java.util.ArrayList<>();
	final double weight=1;
	for(int py=0;py<dimy;py++){
	    for(int px=0;px<dimx;px++){
		data.add(new ObservationVectorXF(String.valueOf(py)+":"+px,getXinBasis.apply(px,py),getF(px,py),getWeight(px,py)));
	    }
	}
	return new DataReadObservationVectorXF(data,Xheader);
    }

    static final BasisFunctionsCalculatable B=true?new Chebyshev():new Legendre();
    public DataReadObservationVectorXF getImageNxNySample(final int nx,final int ny){
	final int nA=nx*ny;
	final java.util.ArrayList<String> Xheader=new java.util.ArrayList<>(nA);
	for(int jy=0;jy<ny;jy++) for(int jx=0;jx<nx;jx++) Xheader.add("["+jx+"]["+jy+"]");
	return getImageAsSample((px,py)->{
		final double [] A=new double [nA],
		    Tx=B.getX0Moments(nx,fromPixelToX(px)),
		    Ty=B.getX0Moments(ny,fromPixelToY(py));
		for(int jy=0,o=0;jy<ny;jy++) for(int jx=0;jx<nx;jx++) A[o++]=Tx[jx]*Ty[jy];
		return A;
	    },Xheader);
    }
    public DataReadObservationVectorXF getImageMultiIndexDSample(final int D){
	return
	    AttributesProductsMultiIndexed.getAttribsProductsMultiIndexed(
		D,
		// create n=3 (px,py,1) sample, then multi-index it.
		getImageAsSample((px,py)->new double []{fromPixelToX(px),fromPixelToX(py),1},java.util.List.of("X","Y","1")),
		(x)->x, // assume fromPixelToX & fromPixelToY are properly regularized
		(order,x)->B.calculate(order,x));
    }

    /** Construct an image as interpolation. The DataReadObservationVectorXF data is
     * assumed to be in raster order, this is why "private".
     */
    private java.awt.image.BufferedImage getImageInterpolated(final RadonNikodymSpectralModel rn,final DataReadObservationVectorXF data,java.util.function.ToDoubleFunction<RadonNikodymSpectralModel.RNPointEvaluation> feval){
	if(data.data.size()!=dimx*dimy) throw new IllegalArgumentException();
	final java.awt.image.BufferedImage rimg = new java.awt.image.BufferedImage(dimx, dimy, java.awt.image.BufferedImage.TYPE_BYTE_GRAY /*TYPE_USHORT_GRAY*/);
	final java.awt.image.WritableRaster wr=rimg.getRaster();
	for(int py=0,o=0;py<dimy;py++){
	    for(int px=0;px<dimx;px++){
		wr.setPixel(px, py, new double []{feval.applyAsDouble(rn.getRNatXoriginal(data.data.get(o++).x))});
	    }
	}
	return rimg;
    }

    java.awt.image.BufferedImage getImageFromRaster(double [] r){
	if(r.length!=(dimx*dimy)) throw new IllegalArgumentException("r.length="+r.length+" dimx="+dimx+" dimy="+dimy);
	final java.awt.image.BufferedImage rimg = new java.awt.image.BufferedImage(dimx, dimy, java.awt.image.BufferedImage.TYPE_BYTE_GRAY /*TYPE_USHORT_GRAY*/);
	final java.awt.image.WritableRaster  wr=rimg.getRaster();
	for(int py=0,o=0;py<dimy;py++){
	    for(int px=0;px<dimx;px++){
		wr.setPixel(px, py, new double []{r[o++]});
	    }
	}
	return rimg;
    }

    void saveImage(java.awt.image.BufferedImage img,
		   final java.util.function.Function<String,String> getoutputfilename) throws java.io.IOException,java.net.URISyntaxException {
	final String fout=getoutputfilename.apply(new java.io.File(new java.net.URI(imageurl).getPath()).getName())+".png";
	System.err.println("Saving to "+fout);
	javax.imageio.ImageIO.write(img, "png", new java.io.File(fout));
    }

    public static void main(String [] args) throws java.io.IOException,java.net.URISyntaxException {
	if(args.length!=4)
	{
	    System.err.println("usage: ExampleImageInterpolation2 {file:path/|http:host/}file.png nX nY D");
	    System.exit(1);
	}
	final int nX=java.lang.Integer.valueOf(args[1]),
	    nY=java.lang.Integer.valueOf(args[2]),
	    D=java.lang.Integer.valueOf(args[3]);
	final String imageurl=args[0];
	final ExampleImageInterpolation2 E=new ExampleImageInterpolation2(imageurl);
	E.saveImage(E.getImageFromRaster(E.origraster),(fn)->fn+".same");

	if(true){ // The result of https://arxiv.org/abs/1511.01887 . The basis x^jx * y^jy , 0<=jx<nx and 0<=jy<ny
	    System.err.println("Creating nX="+nX+" nY="+nY+" sample");
	    final DataReadObservationVectorXF sample=E.getImageNxNySample(nX,nY);
	    System.err.println("Created basis size="+sample.xHeader.size());
	    System.err.println("Creating RN");
	    final RadonNikodymSpectralModel rn_sample=sample.getRadonNikodymSpectralModel_WithoutRegularization();

	    System.err.println("Output LS for nX="+nX+" x nY="+nY+" basis size="+rn_sample.nX);
	    E.saveImage(E.getImageInterpolated(rn_sample,sample,(p)->p.getLS()),(fn)->fn+".LS."+nX+"."+nY+".bs"+rn_sample.nX);
	    System.err.println("Output RN for nX="+nX+" x nY="+nY+" basis size="+rn_sample.nX);
	    E.saveImage(E.getImageInterpolated(rn_sample,sample,(p)->p.getRN()),(fn)->fn+".RN."+nX+"."+nY+".bs"+rn_sample.nX);
	}

	if(true){ // The result of https://arxiv.org/abs/1906.00460 . The basis x^jx * y^jy , 0<=jx+jy<=D
	    System.err.println("Creating D="+D+" sample");
	    final DataReadObservationVectorXF sample=E.getImageMultiIndexDSample(D);
	    System.err.println("Created basis size="+sample.xHeader.size());
	    System.err.println("Creating RN");
	    final RadonNikodymSpectralModel rn_sample=sample.getRadonNikodymSpectralModel_WithoutRegularization();

	    System.err.println("Output LS for D="+D+" basis size="+rn_sample.nX);
	    E.saveImage(E.getImageInterpolated(rn_sample,sample,(p)->p.getLS()),(fn)->fn+".LS.D."+D+".bs"+rn_sample.nX);
	    System.err.println("Output RN for D="+D+" basis size="+rn_sample.nX);
	    E.saveImage(E.getImageInterpolated(rn_sample,sample,(p)->p.getRN()),(fn)->fn+".RN.D."+D+".bs"+rn_sample.nX);
	}
    }
}
