package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.*;
import com.polytechnik.kgo.*;

/** An stub for image intepolation.
 * @version $Id: ImageInterpolationStub.java,v 1.17 2024/04/21 19:51:20 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
class ImageInterpolationStub {
    final String imageurl;
    final int dimx,dimy;
    final double [] origraster;


    static final int IMGTYPE=java.awt.image.BufferedImage.TYPE_BYTE_GRAY;
    static final double IMGTYPE_MAX=255;

    ImageInterpolationStub(final String imageurl) throws java.io.IOException,java.net.URISyntaxException {
	this.imageurl=imageurl;
	final java.awt.image.BufferedImage image=
	    javax.imageio.ImageIO.read(new java.net.URI(imageurl).toURL());
	this.dimx=image.getWidth();
	this.dimy=image.getHeight();
	this.origraster=image.getData().getPixels(0, 0, dimx, dimy, new double [dimx*dimy]);
    }

    @FunctionalInterface
    public interface PixelFunction<T> {
	T apply(int px,int py);
    }

    double fromPixelToX(final int px){ return 2.0*px/(dimx-1.0)-1;}
    double fromPixelToY(final int py){ return 2.0*py/(dimy-1.0)-1;}
    double getF(final int px,final int py){
	//System.err.println("forig="+origraster[px+dimx*py]);
	return origraster[px+dimx*py];
    }
    double getWeight(final int px,final int py){return 1;}

    public DataReadObservationVectorXVectorF getImageAsSample(PixelFunction<double []> getXinBasis,
							      java.util.List<String> Xheader,
							      PixelFunction<double []> getFinBasis,
							      java.util.List<String> Fheader){
	final var data=new java.util.ArrayList<ObservationVectorXVectorF>();
	final double weight=1;
	for(int py=0;py<dimy;py++){
	    for(int px=0;px<dimx;px++){
		data.add(new ObservationVectorXVectorF(
			     String.valueOf(py)+":"+px,
			     getXinBasis.apply(px,py),
			     getFinBasis.apply(px,py),
			     getWeight(px,py)));
	    }
	}
	return new DataReadObservationVectorXVectorF(data,Xheader,Fheader);
    }

    static final BasisFunctionsCalculatable B=true?new Chebyshev():new Legendre(),
	FB=B;
    public DataReadObservationVectorXVectorF getImageNxNySample(final int nx,final int ny,final int nF){
	final int nA=nx*ny;
	final var Xheader=new java.util.ArrayList<String>(nA);
	final var Fheader=new java.util.ArrayList<String>(nF);
	for(int jy=0;jy<ny;jy++) for(int jx=0;jx<nx;jx++) Xheader.add("["+jx+"]["+jy+"]");
	for(int jf=0;jf<nF;jf++) Fheader.add("F["+jf+"]");
	return getImageAsSample((px,py)->{
		final double [] A=new double [nA],
		    Tx=B.getX0Moments(nx,fromPixelToX(px)),
		    Ty=B.getX0Moments(ny,fromPixelToY(py));
		for(int jy=0,o=0;jy<ny;jy++) for(int jx=0;jx<nx;jx++) A[o++]=Tx[jx]*Ty[jy];
		return A;
	    },Xheader,
	    (px,py)->FB.getX0Moments(nF,-1+2*getF(px,py)/IMGTYPE_MAX/* scale from [0:1] to [-1:1] */),Fheader);
    }

    /** Construct an image, KGO processeed. The DataReadObservationVectorXF data is
     * assumed to be in raster order, this is why "private".
     */
    /*private*/ java.awt.image.BufferedImage getImageKGOProcessed(final KGOSolutionVectorXVectorF kgo,final DataReadObservationVectorXVectorF data,java.util.function.ToDoubleBiFunction<ObservationVectorXVectorF,VectorXVectorFPointAccess> feval){
	if(data.data.size()!=dimx*dimy) throw new IllegalArgumentException();
	final java.awt.image.BufferedImage rimg = new java.awt.image.BufferedImage(dimx, dimy, IMGTYPE /*TYPE_USHORT_GRAY*/);
	final java.awt.image.WritableRaster wr=rimg.getRaster();
	for(int py=0,of=0;py<dimy;py++){
	    for(int px=0;px<dimx;px++){
		final var o=data.data.get(of++);
		wr.setPixel(px, py, new double []{feval.applyAsDouble(o,kgo.evaluateAt(o.x))});
	    }
	}
	return rimg;
    }

    java.awt.image.BufferedImage getImageFromRaster(double [] r){
	if(r.length!=(dimx*dimy)) throw new IllegalArgumentException("r.length="+r.length+" dimx="+dimx+" dimy="+dimy);
	final java.awt.image.BufferedImage rimg = new java.awt.image.BufferedImage(dimx, dimy, IMGTYPE /*TYPE_USHORT_GRAY*/);
	final java.awt.image.WritableRaster  wr=rimg.getRaster();
	for(int py=0,o=0;py<dimy;py++){
	    for(int px=0;px<dimx;px++){
		wr.setPixel(px, py, new double []{r[o++]});
	    }
	}
	return rimg;
    }

    void saveImage(java.awt.image.BufferedImage img,
		   final java.util.function.Function<String,String> getoutputfilename) throws java.io.IOException,java.net.URISyntaxException {
	final String fout=getoutputfilename.apply(new java.io.File(new java.net.URI(imageurl).getPath()).getName())+".png";
	System.err.println("Saving to "+fout);
	javax.imageio.ImageIO.write(img, "png", new java.io.File(fout));
    }

    static double getFforEvaluation(final ObservationVectorXVectorF o,final VectorXVectorFPointAccess p){
	final var r=p.getFofMaximalProbability();
	final double f=r[1];
	return 0.5*(f+1) /* back from [-1:1] to [0:1] */
	    *IMGTYPE_MAX;
    }
    static double getPforEvaluation(final ObservationVectorXVectorF o,final VectorXVectorFPointAccess p){
	final var r=p.getFProbabilityAtX(o.f);
	return r*IMGTYPE_MAX; // scale probability to 0:255 to present as an image
    }

}
