package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.DataReadObservationVectorXVectorF;
import com.polytechnik.utils.UnitTests;
import java.io.File;
import com.polytechnik.kgo.DemoRecoverUnitaryMatrixFromSeq;
// large matrices with data
import static com.polytechnik.algorithms.PrintOrthogonalSeq_Matrices.*;

/** Print a sequence of unitary transforms.
  * @version $Id: PrintOrthogonalSeq.java,v 1.17 2024/05/15 15:28:34 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
  */
public class PrintOrthogonalSeq {

    static final double eps=1e-7;

    final int dim;
    final double [] matr;
    final double [] initialV;
    final int nSteps;

    PrintOrthogonalSeq(final double [] matr,
		       final double [] initialV,
		       final int nSteps){
	this.dim=initialV.length;
	this.nSteps=nSteps;
	this.matr=matr.clone();
	this.initialV=initialV.clone();
	if(matr.length!=dim*dim||initialV.length!=dim) throw new IllegalArgumentException();

	if(eps>0){
	    final double [] mm=TMatr.mul_MA_x_MBTrans(dim,matr,dim,matr,dim);
	    UnitTests.checkDoubleArrayRel1(mm,"mm",TMatr.getUnitMatrix(dim),"I",eps);
	    UnitTests.checkDoubleRel1(TMatr.mulVV(initialV,initialV),"<initV|initV>",1,"1",eps);
	}
    }

    /** add unit as 0-th element. We need this as our regularization method requires const in basis
     */
    private static double [] addOne0thElement(final int d,final double [] m){
	final int nd=d+1;
	final double [] r=new double [nd*nd];
	r[0]=1;
	for(int i=0;i<d;i++){
	    for(int j=0;j<d;j++){
		final int ni=1+i,nj=1+j;
		r[ni*nd+nj]=m[i*d+j];
	    }
	}
	return r;
    }

    public void printSeq(final String filename,
			 final java.util.Random r) throws java.io.IOException {
	final var out=new java.io.PrintWriter(filename);
	final var s=DataReadObservationVectorXVectorF.SEPARATOR;
	System.err.println("Writing "+nSteps+" points to "+filename);

	out.println("| Output M^l X of "+nSteps+" steps");
	out.println("| MATRIX:");
	for(int i=0;i<dim;i++){
	    final var v=EVXData.getPsiFromBasis(i,matr,dim,0);
	    out.println("|"+java.util.Arrays.toString(v));
	}
	out.println("| INITIAL VECTOR:");
	out.println("|"+java.util.Arrays.toString(initialV));
	out.println("| Header:");
	int o=0;
	out.print("|N:"+(o++));
	for(int i=0;i<dim;i++){
	    out.print(s+"x:"+(o++));
	}
	out.println(s+"random_sign:"+(o++)+s+"L2:"+(o++));
	double [] x=initialV;
	// starts with 0, one more to have exactly nSteps mapping x_l -> x_{l+1}
	for(int l=0;l<=nSteps;l++){
	    final double [] xNext=TMatr.getMatr_x_A(dim,matr,x);
	    out.print(String.valueOf(l));
	    final int signX=r==null?1:(r.nextBoolean()?1:-1);
	    for(int i=0;i<dim;i++){
		final double xV=signX*x[i];
		out.print(s+xV);
	    }
	    out.println(s+signX+s+TMatr.mulVV(x,x));
	    x=xNext;
	}
	out.close();
    }

    public static void main(String [] args) throws java.io.IOException{

	// for random signs
	final var r=new java.util.Random(123);


	new PrintOrthogonalSeq(
	    M3,
	    TMatr.getVectorScaled(V3,1/Math.sqrt(TMatr.mulVV(V3,V3))),
	    1000).printSeq("/tmp/SO3.csv",r);

	// generate random orthogonal matrix https://math.stackexchange.com/questions/2098897/generating-random-orthogonal-matrices
	/*
	import numpy as np
	from scipy.stats import ortho_group

         m = ortho_group.rvs(dim=40)
	 mm = np.matmul(m.T, m)
	 print(m)
	 print(mm)
         np.random.rand(40)
         np.savetxt("/tmp/some_i.txt",m,delimiter=',')
	*/

	new PrintOrthogonalSeq(
	    M5,
	    TMatr.getVectorScaled(V5,1/Math.sqrt(TMatr.mulVV(V5,V5))),
	    1000).printSeq("/tmp/O5.csv",r);


	new PrintOrthogonalSeq(
	    M7,
	    TMatr.getVectorScaled(V7,1/Math.sqrt(TMatr.mulVV(V7,V7))),
	    1000).printSeq("/tmp/O7.csv",r);

	new PrintOrthogonalSeq(
	    M17,
	    TMatr.getVectorScaled(V17,1/Math.sqrt(TMatr.mulVV(V17,V17))),
	    1000).printSeq("/tmp/O17.csv",r);

	new PrintOrthogonalSeq(
	    M40,
	    TMatr.getVectorScaled(V40,1/Math.sqrt(TMatr.mulVV(V40,V40))),
	    1000).printSeq("/tmp/O40.csv",r);


    }


/** Auto test unitary matrix recovery. */
static class TestAuto {

    static double maxElementDiff(final double [] m1,final double [] m2){
	if(m1.length!=m2.length) throw new IllegalArgumentException();
	double dMax=0;
	for(int j=0;j<m1.length;j++){
	    final double d=Math.abs(m1[j]-m2[j]);
	    if(d>dMax) dMax=d;
	}
	return dMax;
    }
    /** Check for +- sign. */
    static double uMatrixDiff(final double [] m1,final double [] m2){
	final double dp=maxElementDiff(m1,m2),dm=maxElementDiff(m1,TMatr.getVectorScaled(m2,-1));
	return Math.min(dp,dm);
    }

    static void printMatrix(final String txt,
			    final int n,
			    final double [] uM,
			    final double [] M) throws java.io.IOException {
	System.out.println(txt+" output for dim="+n);
	// print matrices
	for(int i=0;i<n;i++){
	    final var v=EVXData.getPsiFromBasis(i,uM,n,0);
	    System.out.println(java.util.Arrays.toString(v));
	}
	final double d=uMatrixDiff(M,uM);
	System.out.println(txt+" DIFF for dim="+n+" is "+d);
	System.out.flush();
    }

    static void testOne(final double [] M,final double [] V,final java.util.Random r) throws java.io.IOException {
	final int n=V.length;
	final String sf=File.createTempFile("O_"+n+"_f",".csv").toString();
	new PrintOrthogonalSeq(
	    M,
	    V,
	    1000).printSeq(sf,r);
	final var args=new String[]{
		"--data_file_to_build_unitarymodel_from="+sf,
		"--data_cols="+(3+n)+":1,"+n+":0"
	};

	final double []
	    uGram=DemoRecoverUnitaryMatrixFromSeq.getUFromArgsGramMatrixChannel(args),
	    uUnit=DemoRecoverUnitaryMatrixFromSeq.getUFromArgsUnitMatrixChannel(args);
	printMatrix("GRAM",n,uGram,M);
	printMatrix("UNIT",n,uUnit,M);
    }


    public static void main(String [] args) throws java.io.IOException {
	// for random signs
	final var r=new java.util.Random(123);

	testOne(
	    M3,
	    TMatr.getVectorScaled(V3,1/Math.sqrt(TMatr.mulVV(V3,V3))),
	    r);

	testOne(
	    M5,
	    TMatr.getVectorScaled(V5,1/Math.sqrt(TMatr.mulVV(V5,V5))),
	    r);

	testOne(
	    M7,
	    TMatr.getVectorScaled(V7,1/Math.sqrt(TMatr.mulVV(V7,V7))),
	    r);

	testOne(
	    M17,
	    TMatr.getVectorScaled(V17,1/Math.sqrt(TMatr.mulVV(V17,V17))),
	    r);

	testOne(
	    M40,
	    TMatr.getVectorScaled(V40,1/Math.sqrt(TMatr.mulVV(V40,V40))),
	    r);

    }

}// END OF class TestAuto

}
