package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.itch.*;

/** Test Alexander Rybalov model.
 *  @version $Id: Rybalov.java,v 1.36 2020/02/06 01:28:21 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Rybalov extends DummyTrader {

    static final long T_START=(9*3600+30*60)*TraderInterface.TIME_SCALE, // 9:30
	T_END=(15*3600+55*60)*TraderInterface.TIME_SCALE, // 15:55
	T_GLOBAL_END=T_END+20*TraderInterface.TIME_SCALE; // 15:55 + 20 seconds

    static final int SHARES=100; // for now
    static final int N_KEEP_TICKS=3000;

    static class Transaction {
	final Trade tick;
	final int shares;
	final String text;
	final RMessage message;
	Transaction(final Trade tick,
		    final int shares,
		    final String text,
		    final RMessage message)
	{
	    this.tick=tick;
	    this.shares=shares;
	    this.text=text;
	    this.message=message;
	}
    }

    static boolean isGlobalTimeEnd(long time)
    {
	return time>T_GLOBAL_END;
    }

    static class TestTrader {
	final String ticker;
	final String source;
        final OrdersAndTrades ot;
	final java.util.ArrayList<Transaction> transactions=
	    new java.util.ArrayList<Transaction>();
	int total_shares=0;
	TestTrader(final String ticker,
		   final String source,
		   final OrdersAndTrades ot)
	{
	    this.ticker=ticker;
	    this.source=source;
	    this.ot=ot;
	}

	boolean isTimeOK(final Trade tick)
	{
	    if(!isGlobalTimeEnd(tick.time) && T_START<tick.time && tick.time<T_END)
		return true;
	    return false;
	}
	boolean isStopLossTriggered(final Trade tick)
	{
	    return false;
	}

	void enterBuy(final Trade tick,int shares,final RMessage message)
	{
	    if(total_shares==0 && isTimeOK(tick))
	    {
		final Transaction t=new Transaction(tick,shares,"ENTER BUY",message);
		transactions.add(t);
		total_shares+=t.shares;
	    }
	}
	void exitBuy(final Trade tick,final RMessage message,final String ee)
	{
	    if(total_shares>0)
	    {
		final Transaction t=new Transaction(tick,-total_shares,ee!=null?ee:"EXIT BUY",message);
		transactions.add(t);
		total_shares+=t.shares;
	    }
	}
	void enterSell(final Trade tick,int shares,final RMessage message)
	{
	    if(total_shares==0 && isTimeOK(tick))
	    {
		final Transaction t=new Transaction(tick,-shares,"ENTER SELL",message);
		transactions.add(t);
		total_shares+=t.shares;
	    }
	}
	void exitSell(final Trade tick,final RMessage message,final String ee)
	{
	    if(total_shares<0)
	    {
		final Transaction t=new Transaction(tick,-total_shares,ee!=null?ee:"EXIT SELL",message);
		transactions.add(t);
		total_shares+=t.shares;
	    }
	}
	void forceClose(final Trade tick,final RMessage message)
	{
	    exitBuy(tick,message,"FORCE EXIT BUY");
	    exitSell(tick,message,"FORCE EXIT SELL");
	}

	void processEvent(final Trade tick,final RMessage message)
	{
	    n_processevent_called++;
	    volume+=tick.shares;

	    if(message==null) return;

	    if(Double.isNaN(firstsum))
	    {
		if(isTimeOK(tick))
		    firstsum=message.stats;
	    }
	    if(message.enter_buy) enterBuy(tick,SHARES,message);
	    if(message.enter_sell) enterSell(tick,SHARES,message);
	    if(message.exit_buy) exitBuy(tick,message,null);
	    if(message.exit_sell) exitSell(tick,message,null);
	    if(!isTimeOK(tick) || isStopLossTriggered(tick)) forceClose(tick,message);
	}

        String gettTradesAndPNL()
	{
	    final java.io.StringWriter sw=new java.io.StringWriter();
	    final java.io.PrintWriter out=new java.io.PrintWriter(sw);
	    out.println("Ticker\t"+this.ticker);
	    out.println("Source\t"+this.source);
	    out.println("N_trades\t"+this.transactions.size());
	    out.println("N_events\t"+this.n_processevent_called);
	    out.println("Volume\t"+this.volume);
	    out.println("Version\t$Id: Rybalov.java,v 1.36 2020/02/06 01:28:21 mal Exp $");


	    // print P&L
	    int s=0;
	    double p=0;
	    out.println("TIME\tSHARES\tPRICE\tTEXT\tCOMMENTS\t");
	    for(Transaction x: transactions)
	    {
		out.println(BookItem.getTimeHHMMSS(x.tick.time)+"\t"+
			    x.shares+"\t"+
			    x.tick.price+"\t"+
			    x.text+"\t"+
			    (x.message==null?"":x.message.dumpData())+"\t");
		s+=x.shares;
		p=p-x.tick.price*x.shares;
	    }
	    if(s!=0||total_shares!=0)
	    {
		throw new RuntimeException("ticker="+ticker+" s="+s+" total_shares="+total_shares);
	    }
	    out.println("P&L\t"+p);
	    out.close();
	    return sw.toString();
	}

	int n_processevent_called=0;
	long volume=0;
	Trade last_processed=null;
	double firstsum=Double.NaN;
    } // END of class TestTrader



    private static class RMessage {
	final boolean enter_buy,enter_sell,exit_buy,exit_sell;
	final double stats;
	RMessage(final boolean enter_buy,
		 final boolean enter_sell,
		 final boolean exit_buy,
		 final boolean exit_sell,
		 final double stats)
	{
	    this.enter_buy=enter_buy;
	    this.enter_sell=enter_sell;
	    this.exit_buy=exit_buy;
	    this.exit_sell=exit_sell;
	    this.stats=stats;
	}
	String dumpData()
	{
	    return "enter_buy="+enter_buy+" enter_sell="+enter_sell+" exit_buy="+exit_buy+" exit_sell="+exit_sell;
	}
    }


    final Distribution<Trade> savedtrades=new Distribution<Trade>();
    static void copyToArray(Distribution<Trade> st,final OrdersAndTrades ot)
    {
	final int ntrades=ot.tradebook.copyTrades(st);
	if(ntrades<=N_KEEP_TICKS)
	{
	    return;
	}
	final Trade [] buf=st.getBuffer();
	System.arraycopy(buf,ntrades-N_KEEP_TICKS,buf,0,N_KEEP_TICKS);
	java.util.Arrays.fill(buf,N_KEEP_TICKS,buf.length,null);
	st.setSize(N_KEEP_TICKS);
    }


    private boolean flag_finished=false;
    private final java.util.Map<String,TestTrader> sbs=
	new java.util.HashMap<String,TestTrader>();
    public void update(final OrdersAndTrades ot,
		       final long currenttime,
		       final int eventid){
	TestTrader bs=sbs.get(ot.ticker);
	if(bs==null)
	{
	    bs=new TestTrader(ot.ticker,source,ot);
	    sbs.put(ot.ticker,bs);
	}

	if(!flag_finished && !isGlobalTimeEnd(currenttime))
	{
	    updateByTicker(bs,currenttime,eventid);
	}
	else
	{ // force close if there were np ticks an the EOD
	    if(!flag_finished)
	    { // force close event if there is no recent ticks on the stock
		flag_finished=true;
		for(TestTrader x:sbs.values())
		{
		    x.forceClose(x.ot.tradebook.getLastAddedElement(),null);
		}
	    }
	    ot.clearBefore(currenttime);
	}
    }

    void updateByTicker(final TestTrader bs,
			final long currenttime,
			final int eventid){

	if(false && !bs.ot.ticker.equals("AAPL")) // for debugging one may enable just one stock
	{
	    bs.ot.clearBefore(currenttime);
	    return;
	}

	Trade tl=bs.ot.tradebook.getLastAddedElement();
	if(tl!=null && tl!=bs.last_processed)
	{
	    bs.last_processed=tl;
	    copyToArray(this.savedtrades,bs.ot);
	    final Trade [] buf=this.savedtrades.getBuffer();
	    final int ntrades=this.savedtrades.getSize();
	    if(ntrades>0 && buf[ntrades-1]!=tl)
	    {
		throw new RuntimeException("copyToArray is broken");
	    }
	    final RMessage rm=get_event(ntrades,buf,bs.firstsum);
	    bs.processEvent(tl,rm);

	    if(ntrades>=N_KEEP_TICKS)
	    {
		bs.ot.clearBefore(buf[0].time);
	    }
	    this.savedtrades.clearAll();
	}
    }



    final String source;
    Rybalov(final String source)
    {
	this.source=source;
    }



void dumpResults(final String out_dir) throws java.io.IOException
{
    for(TestTrader t:sbs.values())
    {
	final String fname=out_dir+
	    com.polytechnik.outils.EncodeDecode.urlencode(t.ticker)+
		".csv";
	System.err.print("Saving "+t.ticker+" trading data to file="+fname);
	final java.io.PrintWriter out=new java.io.PrintWriter(fname);
	out.print(t.gettTradesAndPNL());
	out.close();
	System.err.println(" DONE");
    }
}


public static void main(String [] args) throws java.io.IOException
{
    if(args.length!=2)
    {
	System.err.println("Rybalov.main itch_file_1 output_dir");
	System.exit(1);
    }

    final String fname=args[0],out_dir=args[1];
    System.err.println("Reading data from file="+fname+" and saving results to dir="+out_dir);
    final Rybalov t=new Rybalov(fname);
    t.runParser(fname);

    t.dumpResults(out_dir);
}



// k is one based
// N is data.Length
static double dct(double[] yt, int k, int N, int fromIdx)
{
    double res = 0;
    int i;
    int last = fromIdx + N;
    int n = 1;

    for (i = fromIdx; i < last; i++)
    {
	res = res + yt[i] * Math.cos(Math.PI * (2 * n - 1) * (k - 1) / (2 * N));
	n++;
    }

    // This line can be optimized by moving out of the function to
    // the calling function
    res = res / Math.sqrt(N);
    if (k > 1)
    {
	res = Math.sqrt(2) * res;
    }

    return res;
}

static final int RYBALOV_SIZE=480;

static final boolean PRINT_DEBUG=false;


RMessage get_event(final int n_t,final Trade [] x,final double mnn)
{

if(n_t<RYBALOV_SIZE)
    return null;

if(PRINT_DEBUG)
{
    System.err.println("=========== DEBUG; orig prices for n_t="+n_t);
    for(int k=0;k<n_t;k++)
    {
	System.err.println("x["+k+"]="+x[k].price);
    }
}

final double [] yt=new double [RYBALOV_SIZE/4];
// =========== copy every 4-th element to yt[]
int l=yt.length;
for(int k=n_t;k>n_t-RYBALOV_SIZE;k-=4)
{
    final int i=k-1,ll=(--l);
    yt[ll]=x[i].price;
}
if(l!=0) throw new RuntimeException("every 4-th element copying is broken. l="+l);
if(PRINT_DEBUG)
{
    System.err.println("Every 4-th element="+java.util.Arrays.toString(yt));
}

// =========== calculate new_mnn.
double new_mnn=0;
int ns_mnn=0;
for(int k=1;k<=yt.length;k++)
{
    final int i=k-1;
    if(20<=k && k<=120)
    {
	ns_mnn++;
	new_mnn+=yt[i];
    }
}
new_mnn=(new_mnn/ns_mnn)*0.993;
if(PRINT_DEBUG)
{
    System.err.println("new_mnn calculated="+new_mnn+" elements in sum="+ns_mnn+" argument mnn="+mnn);
}

// =========== normalize yt using argument mnn, not new_mnn
// =========== the result is saved back to yt.
final double nr=1000;
for(int i=0;i<yt.length;i++)
{
    yt[i]=(yt[i]-mnn)/mnn*nr;
}
if(PRINT_DEBUG)
{
    System.err.println("Normalized yt="+java.util.Arrays.toString(yt));
}

// swap yt elements, so the latest one would be at 0 and the earliest one at yt.length-1.
for(int k=0;k<yt.length/2;k++)
{
    final int k2=yt.length-1-k;
    final double s=yt[k2];
    yt[k2]=yt[k];
    yt[k]=s;
}
if(PRINT_DEBUG)
{
    System.err.println("Inversed, latest is at 0 yt="+java.util.Arrays.toString(yt));
}

final double [] // temporary arrays
    c=new double [4],
    d=new double [4],
    e=new double [4],
    ed=new double [4],
    f=new double [4],
    fd=new double [4];

/* // FIXME
for(int i=0;i<4;i++)
{
    c[i]=max(dct(yt(i-100:i-50)));
    d[i]=min(dct(yt(i-100:i-50)));
}
*/
// FIXME
for(int i=0;i<4;i++)
{
    c[i]=-1+2*Math.random();
    d[i]=-1+2*Math.random();
}


boolean enter_buy=false,exit_buy=false,enter_sell=false,exit_sell=false;
for(int i=4;--i>=0;) // from earliest to latest(=0)
{
    if(yt[i]>0)
    {
        e[i]=yt[i]-0.1*c[i];
        f[i]=0;
        fd[i]=0;
        if( (e[i]-3)>0 && (c[i]/yt[i]>2) )
	    ed[i]=1;
        else if(e[i]<0)
	    ed[i]=-1;
        else ed[i]=0;
    }
    else
    {
        f[i]=0.1*d[i]-yt[i];
        e[i]=0;
        ed[i]=0;
        if( (f[i]-3>0) && (d[i]/yt[i]>2) )
           fd[i]=1;
	else if(f[i]<0)
           fd[i]=-1;
        else fd[i]=0;
    }

    double se=0,sf=0;
    for(int m=0;m<4;m++)
    {
	se+=ed[m];
	sf+=fd[m];
    }
    se=Math.signum(se-5);
    sf=Math.signum(sf-5);

    if(se>=0) enter_sell=true;
    if(sf>=0) enter_buy=true;
    if(ed[i]<0) exit_sell=true;
    if(fd[i]<0) exit_buy=true;
}

// FIXME
if(true)
{ // force arbitrary trading for debugging
    enter_buy=Math.random()<0.02;
enter_sell=Math.random()<0.02;
exit_buy=Math.random()<0.015;
exit_sell=Math.random()<0.015;
}

return new RMessage(enter_buy,enter_sell,exit_buy,exit_sell,new_mnn);

}

}
