package com.polytechnik.algorithms;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.Trade; // (time,price,shares) container
import com.polytechnik.realtimebook.TraderInterface; // for TIME_SCALE=1000000000L
import com.polytechnik.freemoney.FreeMoneyForAll;
import com.polytechnik.freemoney.CommonlyUsedMoments;
import com.polytechnik.freemoney.CommonlyUsedMomentsLegendreShifted;
import com.polytechnik.freemoney.CommonlyUsedMomentsLaguerre;
import com.polytechnik.freemoney.CommonlyUsedMomentsMonomials;
import com.polytechnik.freemoney.PFuture;

/** A test of {@link com.polytechnik.freemoney.FreeMoneyForAll}.
 * @version $Id: TestCall_FreeMoneyForAll.java,v 1.48 2022/09/30 10:26:13 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class TestCall_FreeMoneyForAll {

private static int _outputDataHeader(final java.io.PrintWriter out)
{
    int n=0;
    out.println(
	"|1:T\t2:shares\t3:P_last"
	+"\t4:"+"pv_average"
	+"\t5:"+"stdev_pv_average"
	+"\t6:"+"pv_IH"
	+"\t7:"+"Tv_IH"
	// other fields
	+PrintFields.getPFutureHeader("pFV.",(n=8+42)-42)
	+PrintFields.getPFutureHeader("pFA.",(n+=42)-42)
		);
    return 1;
}

private static int _outputData(final java.io.PrintWriter out,final long currenttime,final Trade t,final FreeMoneyForAll r)
{
    out.println(String.valueOf(currenttime)+"\t"+t.shares+"\t"+t.price
		+"\t"+r.pv_average
		+"\t"+r.stdev_pv_average
		+"\t"+r.pv_IH
		+"\t"+r.Tv_IH
		// other fields
		+PrintFields.getPFutureData(r.pFV)
		+PrintFields.getPFutureData(r.pFA)
		);
    return 1;
}

@FunctionalInterface interface StorageInterface<RES extends CommonlyUsedMoments> {
    /** Create initial instance ({@code RES} constructor call).
     */
    RES init(int n, double atau);

    /** Update internal storage.
     * One can use either the same object mutation or create a new one, this version mutates the object
     * using {@link com.polytechnik.freemoney.CommonlyUsedMoments#updateWithSingleObservation}.
     */
    default RES updateStorage(long currenttime,RES prev, long t_last, double P_last, long dv_last){
	// a simple bundle with state change
	prev.updateWithSingleObservation(currenttime,t_last,P_last,dv_last);
	// simplistic version: same object mutation
	return prev;
    }
}



public static class TradeDataProcessorTPV<RES extends CommonlyUsedMoments> extends TradeDataProcessor<RES> {
    private RES savedState;
    private final StorageInterface<RES> updater;

    TradeDataProcessorTPV(
	final MuseConfig c,
	final StorageInterface<RES> updater){
	super(c);
	this.updater=updater;
	this.savedState=updater.init(
	    2*c.n-1, // required moments number
	    c.tau*TraderInterface.TIME_SCALE // tau
				     );
    }
    @Override public int outputDataHeader(final java.io.PrintWriter out) throws java.io.IOException {return _outputDataHeader(out);}
    @Override public int outputDataFooter(final java.io.PrintWriter out) throws java.io.IOException {
	out.println("| calculations done");
	return 1;
    }

    @Override public int processOneReadLine(final java.io.PrintWriter out,final long currenttime,final Trade t) throws java.io.IOException{
	return _outputData(out,currenttime,t,calculateData(currenttime,t));
    }

    FreeMoneyForAll calculateData(final long currenttime,final Trade t){
	final RES r=updater.updateStorage(currenttime,savedState,t.time,t.price,t.shares);
	this.savedState=r;
	// Spaghetti code: secondary sampling can be performed
	// only after direct sampling and solving |I|psi>=lambda|psi>
	// set a flag for FreeMoneyForAll to update internal integrator with maximal eigenvalue.
	final boolean flagUseSecondarySamplingFordIHdt=true;
	return new FreeMoneyForAll(c.n,r,flagUseSecondarySamplingFordIHdt);
    }
} // END OF class TradeDataProcessorTPV


public static void main(final String [] args) throws java.io.IOException
{
    final MuseConfig c=new MuseConfig(args);
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }
    else if(c.measure.equals("CommonlyUsedMomentsLegendreShifted"))
    {
	new TradeDataProcessorTPV<CommonlyUsedMomentsLegendreShifted>(c,CommonlyUsedMomentsLegendreShifted::new).processFiles();
    }
    else if(c.measure.equals("CommonlyUsedMomentsLaguerre"))
    {
	new TradeDataProcessorTPV<CommonlyUsedMomentsLaguerre>(c,CommonlyUsedMomentsLaguerre::new).processFiles();
    }
    else if(c.measure.equals("CommonlyUsedMomentsMonomials"))
    {
	new TradeDataProcessorTPV<CommonlyUsedMomentsMonomials>(c,CommonlyUsedMomentsMonomials::new).processFiles();
    }
    else throw new RuntimeException("The measure "+c.measure+" is not supported, allowed values:{CommonlyUsedMomentsLegendreShifted,CommonlyUsedMomentsLaguerre,CommonlyUsedMomentsMonomials}");
}

} // END OF class TestCall_FreeMoneyForAll
