package com.polytechnik.algorithms;

import com.polytechnik.utils.*;
import java.util.*;
import java.io.*;

/** Test distribution parameters.
 * $Id: TestDistributionParameters.java,v 1.21 2020/09/27 20:14:05 mal Exp $
 */
class TestDistributionParameters {
    static double [] readObservationsFromStrings(java.io.BufferedReader br) throws Exception
    {
	final ArrayList<Double> l=new ArrayList<Double>();
	String input;
	while((input=br.readLine())!=null){
	    l.add(Double.valueOf(input));
	}
	final double [] x=new double [l.size()];
	for(int k=0;k<x.length;k++)
	{
	    x[k]=l.get(k).doubleValue();
	}
	return x;
    }
    static double [] getMomentsFromObservations(final int n,final double [] obs)
    {
	final double [] m=new double [n];
	for(int k=0;k<obs.length;k++)
	{
	    double x=obs[k],s=1;
	    for(int j=0;j<m.length;j++,s*=x)
	    {
		m[j]+=s;
	    }
	}
	return m;
    }
    static double [] getChi2DistributedObservations(final int n_points,final int n_chi2_degree,final java.util.Random r)
    {

	final double [] x=new double [n_points];
	for(int k=0;k<x.length;k++)
	{
	    for(int j=0;j<n_chi2_degree;j++)
	    {
		x[k]+=Math.pow(r.nextGaussian(),2);
	    }
	}
	return x;
    }
    static double [] getChi2ExactMoments(final int n,final int n_degree)
    {
	final double [] m=new double [n];
	double s=1;
	for(int k=0;k<m.length;k++)
	{
	    m[k]=s;
	    s*=(n_degree+2*k);
	}
	return m;
    }

    static final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> BP=new com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis();

    static class Stats {
	final double aver,sigma2,xi3,excess_xi4;
	Stats(final double aver,final double sigma2,final double xi3,final double excess_xi4)
	{
	    this.aver=aver;
	    this.sigma2=sigma2;
	    this.xi3=xi3;
	    this.excess_xi4=excess_xi4;
	}
	static Stats getFromMom(final double [] m)
	{
	    final double [] F=new double []{m[4]/m[0],-4*m[3]/m[0],6*m[2]/m[0],-4*m[1]/m[0],1};
	    final double aver=m[1]/m[0],sigma2=BP.B.getD2(F,aver)/(4*3),
		nonnorm_xi3=BP.B.getD1(F,aver)/(-4),nonnorm_xi4=BP.B.getD0(F,aver);
	    return new Stats(aver,
			     sigma2,
			     nonnorm_xi3/Math.pow(sigma2,3.0/2),
			     nonnorm_xi4/Math.pow(sigma2,2)-3);
	}
	public String toString()
	{
	    return "Stats:aver="+aver+" sigma2="+sigma2+" xi3="+xi3+" excess_xi4="+excess_xi4;
	}
    }
    static class InvarXi {
	final double xlow,xhigh,wlow,whigh,i3;
	InvarXi(final double xlow,
		final double xhigh,
		final double wlow,final double whigh)
	{
	    this.xlow=xlow;
	    this.xhigh=xhigh;
	    this.wlow=wlow;
	    this.whigh=whigh;
	    this.i3=(wlow-whigh)/(wlow+whigh);
	}
	static InvarXi getFromMom(final double [] m)
	{
	     final Quadrature QUAD=BP.getGaussQuadratureFromMoments(2,m);
	     final double [] xq=QUAD.x,wq=QUAD.w;
	     //System.err.println("P="+java.util.Arrays.toString(BP.getSingleOrthogonalPolynomial(2,m)));
	     //System.err.println("xq="+java.util.Arrays.toString(xq));
	     //System.err.println("wq="+java.util.Arrays.toString(wq));
	     return new InvarXi(xq[0],xq[1],wq[0],wq[1]);
	}
	public String toString()
	{
	    return "InvarXi:xlow="+xlow+" xhigh="+xhigh+" wlow="+wlow+" whigh="+whigh+" i3="+i3;
	}
    }
    static void printExactXi3(final int kmax)
    {
	System.out.println("#k\txi3\ti3");
	for(int k=1;k<=kmax;k++)
	{
	    final double [] mom_chi=getChi2ExactMoments(5,k);
	    System.out.println(String.valueOf(k)+"\t"+Stats.getFromMom(mom_chi).xi3+
			       "\t"+InvarXi.getFromMom(mom_chi).i3);
	}
    }

    public static void main(final String [] args) throws Exception
    {
	if(args.length!=0 && args.length!=3)
	{
	    System.err.println("main [n_points n_gedree initrandom]");
	    System.exit(1);
	}
	final double [] x;
	if(args.length==0){
	    x=readObservationsFromStrings(new BufferedReader(new InputStreamReader(System.in)));
	    System.out.println("points read from stdin: "+x.length);
	}
	else {
	    printExactXi3(150);
	    final int n_points=Integer.valueOf(args[0]),
		n_degree=Integer.valueOf(args[1]),
		initrandom=Integer.valueOf(args[2]);
	    final java.util.Random r=new java.util.Random(initrandom);
	    System.out.println("n_degree="+n_degree+"\tinitrandom="+initrandom);
	    x=getChi2DistributedObservations(n_points,n_degree,r);
	    System.out.println("Chi2 points generated: "+x.length);
	    final Stats distr_chi2_exact=new Stats(n_degree,n_degree*2,Math.sqrt(8.0/n_degree),(12.0/n_degree));
	    System.out.println("exact:"+distr_chi2_exact.toString());

	    final double [] mom_chi=getChi2ExactMoments(5,n_degree);
	    System.out.println("exact:"+Stats.getFromMom(mom_chi).toString());
	    System.out.println("exact:"+InvarXi.getFromMom(mom_chi).toString());
	}
	final double [] m=getMomentsFromObservations(5,x);
	System.out.println(Stats.getFromMom(m).toString());
	System.out.println(InvarXi.getFromMom(m).toString());
    }

}
