package com.polytechnik.broker;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */

import com.polytechnik.realtimebook.*;

/** Dump Data.
 *  The code is to be called from a single-threaded context.
 *  This will allow us to have a 100% reproduction of the results.
 *  @version $Id: DumpDataBroker.java,v 1.77 2012/01/24 10:57:53 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class DumpDataBroker extends GBroker {

    private static final boolean FLAG_FORCE_EXECUTION_AT_ANY_PRICE=false,
	FLAG_PRICE_EQUAL_ISOK=false;

    /** A kludgy way to run an event later on, without multi-thread code.
     */
    private abstract class RunEvent {
	final long origtime,t_after;
	private boolean flag_processed=false;
	RunEvent(final long origtime,final long t_after)
	{
	    this.origtime=origtime;
	    this.t_after=t_after;
	    if(this.t_after<0)
	    {
		throw new RuntimeException("this.t_after="+this.t_after);
	    }
	}
	void setProcessed()
	{
	    errIfProcessed();
	    this.flag_processed=true;
	}

	void errIfProcessed()
	{
	    if(this.flag_processed)
	    {
		throw new IllegalStateException();
	    }
	}
	void errIfNotProcessed()
	{
	    if(!this.flag_processed)
	    {
		throw new IllegalStateException();
	    }
	}
	abstract void execute(final String txt,final long brokertime);
    }


    private class ExecuteOrder extends RunEvent {
	final String ticker;
	final BOrder order;
	final boolean emulate_failure;
	final long orderId;
	ExecuteOrder(final String ticker,
		     final long origtime,
		     final long t_after,
		     final BOrder order,
		     final boolean emulate_failure)
	{
	    super(origtime,t_after);
	    this.ticker=ticker;
	    this.order=order;
	    this.emulate_failure=emulate_failure;
	    this.orderId=getNextID();

	}
	void execute(final String txt,final long brokertime)
	{
	    setProcessed();
	    if(this.emulate_failure)
	    {
		System.err.println("BROKER: emulate EXECUTIONFAILED2 (order added event missed)");
	    }
	    else
	    {
		orderAdded_FromBroker(order,brokertime);
	    }
	    final double price=order.getSharesSign()>0?order.price+broker_per_share_comission:order.price-broker_per_share_comission;
	    orderExecuted_FromBroker(new ATrade("T_"+String.valueOf(this.orderId),
						brokertime,
						orderId,
						price,
						order.getShares(),
						order.side,
						txt,null),-1,brokertime);
	}
    }

    private class CancelOrder extends RunEvent {
	final long ordertocancel;
	CancelOrder(final long origtime,final long t_after,long ordertocancel)
	{
	    super(origtime,t_after);
	    this.ordertocancel=ordertocancel;
	}
	void execute(final String txt,final long brokertime)
	{
	    setProcessed();
	    orderCancelled_FromBroker(ordertocancel,txt,brokertime);
	}
    }


    private final java.util.ArrayList<ExecuteOrder> toexecute=new java.util.ArrayList<ExecuteOrder>();
    private final java.util.ArrayList<CancelOrder> tocancel=new java.util.ArrayList<CancelOrder>();
    private final boolean EMULATE_FAILURES;
    private final double broker_per_share_comission;
    private final long min_delay;
    public DumpDataBroker(final boolean EMULATE_FAILURES,
			  final double broker_per_share_comission,
			  final long min_delay)
    {
	this.EMULATE_FAILURES=EMULATE_FAILURES;
	this.broker_per_share_comission=broker_per_share_comission;
	this.min_delay=min_delay;
    }

    /** Simple constructor without delays & breakages. */
    public DumpDataBroker()
    {
	this(false,0,0);
    }


    private final java.util.Random r=new java.util.Random(-991L);
    private final double
	PROBABLITY_EXECUTIONFAILED1=0.02,
	PROBABLITY_EXECUTIONFAILED2=PROBABLITY_EXECUTIONFAILED1+0.02,
	PROBABLILITY_CANCELMISSED=0.05,
	PROBABLILITY_CANCELFAILED=0.04+PROBABLILITY_CANCELMISSED;


    public void init() {}

    public void readInitialPosition(final TraderContainer s)
    {
	// no initial position
    }

    private void processCancel(final long tradertime,final String ticker)
    {
	final int n=tocancel.size();
	int n_left=0;
	for(int i=0;i<n;i++)
	{
	    final CancelOrder e=tocancel.get(i);
	    if(tradertime>=e.t_after)
	    {
		// remove an order from execution list
		for(int j=0,nt=toexecute.size();j<nt;j++)
		{
		    final ExecuteOrder eo=toexecute.get(j);
		    eo.errIfProcessed();
		    if(eo.orderId==e.ordertocancel)
		    {
			toexecute.remove(j);
			break;
		    }
		}
		e.execute("cancel at "+tradertime,tradertime);
	    }
	    else
	    {
		System.err.println("CANCEL DELAY: tradertime="+tradertime+" e.t_after="+e.t_after);
		e.errIfProcessed();
		if(i>n_left)
		{
		    // swap elements
		    tocancel.set(i,tocancel.get(n_left));
		    tocancel.set(n_left,e).errIfNotProcessed();
		}
		else if(i<n_left)
		{
		    throw new IllegalStateException();
		}
		n_left++;
	    }
	}
	for(int i=n;i-->n_left;)
	{
	    tocancel.remove(i).errIfNotProcessed();
	}
    }

    private void processExecute(final long tradertime,final String ticker,final Distribution<Trade> trades)
    {
	// process pending cancel first
	processCancel(tradertime,ticker);

	// pending execute events
	final int n=toexecute.size();
	int n_left=0;
	long prevt=-1;
	for(int i=0;i<n;i++)
	{
	    final ExecuteOrder e=toexecute.get(i);
	    if(false && e.t_after<prevt)
	    {
		throw new RuntimeException("The data is not sorted in queue:"+
					   " currenttime="+tradertime);
	    }
	    prevt=e.t_after;
	    String txt_exe;
	    if(tradertime>=prevt)
	    {
		txt_exe="execute at any price";
		if(!FLAG_FORCE_EXECUTION_AT_ANY_PRICE)
		{
		    txt_exe=null;
		    if(e.ticker==ticker)
		    { // if we scan trades make sure the buffer is of correct asset
			final Trade [] tb=trades.getBuffer();
			final int tsize=trades.getSize();

			// same ticker, scan trades
			if(e.order.getSharesSign()>0)
			{ // buy order
			    final double p=e.order.price+(FLAG_PRICE_EQUAL_ISOK?1:-1)*OrdersFlow.EPS;
			    for(int j=tsize;j-->0;)
			    {
				final Trade t=tb[j];
				if(t.time<=prevt)
				{
				    break;
				}
				if(t.price<=p)
				{
				    txt_exe=Trade.getTradeDescription(t);
				    break;
				}
			    }
			}
			else
			{ // sell order
			    final double p=e.order.price-(FLAG_PRICE_EQUAL_ISOK?1:-1)*OrdersFlow.EPS;
			    for(int j=tsize;j-->0;)
			    {
				final Trade t=tb[j];
				if(t.time<=prevt)
				{
				    break;
				}
				if(t.price>=p)
				{
				    txt_exe=Trade.getTradeDescription(t);
				    break;
				}
			    }
			}
		    }
		}
	    }
	    else
	    {
		txt_exe=null;
		System.err.println("DELAY: tradertime="+tradertime+" e.t_after="+e.t_after);
	    }

	    if(txt_exe!=null)
	    {
		e.execute(txt_exe,tradertime);
	    }
	    else
	    {
		e.errIfProcessed();
		if(i>n_left)
		{
		    // swap elements
		    toexecute.set(i,toexecute.get(n_left));
		    toexecute.set(n_left,e).errIfNotProcessed();
		}
		else if(i<n_left)
		{
		    throw new IllegalStateException();
		}
		n_left++;
	    }
	}
	if(false && this.min_delay<=0 && n_left>0)
	{
	    throw new RuntimeException("Some events are missed.");
	}
	for(int i=n;i-->n_left;)
	{
	    toexecute.remove(i).errIfNotProcessed();
	}
    }



    @Override
    public void hookForDebugging(final Object data,final long tradertime)
    {

	final Object [] d=(Object [])data;
	final String ticker=(String)d[0];
	final Distribution<Trade> trades=(Distribution<Trade>)d[1];
	final double p_buy=((Double)d[2]).doubleValue(),p_sell=((Double)d[3]).doubleValue();
	processExecute(tradertime,ticker,trades);
    }

    private long id=100;
    long getNextID()
    {
	return id++;
    }


    public long sendNewOrder_ToBroker(final String ticker,final BOrder o)
    {
	//final long t0=o.original_time;
	final OrderDataStored dor=hookForInternalUsagegetOdrerDatra(o.getID());
	final long t0=dor==null||dor.thisorder==null?o.getTime():dor.thisorder.original_time;
	final double d=r.nextDouble();
	if(EMULATE_FAILURES && d<=PROBABLITY_EXECUTIONFAILED1)
	{
	    System.err.println("BROKER: emulate EXECUTIONFAILED1 ");
	    return -1;
	}
	final ExecuteOrder eo=new ExecuteOrder(ticker,
					       t0,
					       t0+min_delay,
					       o,
					       EMULATE_FAILURES && d<=PROBABLITY_EXECUTIONFAILED2);
	toexecute.add(eo);
	return eo.orderId;
    }

    public long cancelAndReplaceOrder_ToBroker(final long ordertocancel,final BOrder o)
    {
	cancelOrder_ToBroker(ordertocancel);
	return sendNewOrder_ToBroker(null,o);
    }

    public long cancelOrder_ToBroker(final long ordertocancel)
    {
	final double d=r.nextDouble();
	if(EMULATE_FAILURES &&  d<=PROBABLILITY_CANCELMISSED)
	{
	    System.err.println("BROKER: emulate CANCELMISSED for "+ordertocancel);
	    return ordertocancel; // as we missed cancel, report everything is OK, but do nothing.
	}
	if(EMULATE_FAILURES &&  d<=PROBABLILITY_CANCELFAILED)
	{
	    System.err.println("BROKER: emulate CANCELFAILED for "+ordertocancel);
	    return -1; // as cancel failed
	}
	// actual cancel
	final OrderDataStored dor=hookForInternalUsagegetOdrerDatra(ordertocancel);
	final long t0=dor==null||dor.thisorder==null?0:dor.thisorder.original_time;
	final CancelOrder co=new CancelOrder(t0,t0+min_delay,ordertocancel);
	tocancel.add(co);
	return ordertocancel;
    }


}
