package com.polytechnik.broker;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */

import com.polytechnik.realtimebook.BookItem;

/** An interface to a broker to add/remove trades.
 *  @version $Id: GBroker.java,v 1.75 2012/01/23 15:05:58 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class GBroker implements BrokerInterface,ToBrokerInterface,FromBrokerInterface {



    static class OrderDataStored {
	final AOrder thisorder;
	final TraderContainer s;
	OrderDataStored(final AOrder thisorder,final TraderContainer s)
	{
	    this.thisorder=thisorder;
	    this.s=s;
	}
    }


    /** A map of order ID -> Order & Positions.
     */
    private final com.polytechnik.utils.LongHashMap<OrderDataStored> globalsent=
	new com.polytechnik.utils.LongHashMap<OrderDataStored>(203,0.3f);


    /** A list of already processed executions.
     */
    private final java.util.HashMap<String,AOrder> processes_executions=
	new java.util.HashMap<String,AOrder>();

    public void hookForDebugging(final Object data,final long currenttime){}

    public long addNewOrder(final AOrder o,final TraderContainer s)
    {
	//System.err.println("addNewOrder: THREAD="+Thread.currentThread());

	final long id;
	synchronized(this.globalsent)
	{
	    id=sendNewOrder_ToBroker(s.ticker,o);
	    if(id<0)
	    {
		// no race because the order is just created
		// no reference point to it
		o.setAttrib(AOrder.SET_REJECTED,AOrder.STATUS_CONFIRMED);
		logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		       "Rejected orderid="+id+" "+o.getDescription());
		return -1;
	    }
	    o.setID(id);
	    final OrderDataStored d=this.globalsent.put(id,new OrderDataStored(o,s));
	    if(d!=null)
	    {
		this.globalsent.put(id,d);
		throw new IllegalStateException("The order "+o.getDescription()+
						" is already sent for processing."+
						"\n"+s.positions.dumpData());
	    }

	synchronized(s.positions)
	{
	    // put the order immediately to a book.
	    final AOrder oe=s.positions.book.put(id,o);
	    if(oe!=null)
	    {
		s.positions.book.put(id,oe);
		throw new IllegalStateException("The order="+o.getDescription()+
						" is already in a book."+
						"\n"+s.positions.dumpData());
	    }
	    o.setAttrib(AOrder.SET_ADDED,AOrder.STATUS_MESSAGEPOSTED);
	    logEvent(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		     "ORDER:ADDED "+o.getDescription());
	}

	} // END synchronized(this.globalsent)

	return id;
    }

    public long cancelReplaceOrder(final long ordertocancel,final AOrder o,final TraderContainer s)
    {
	final long id;
	synchronized(this.globalsent)
	{
	    id=cancelAndReplaceOrder_ToBroker(ordertocancel,o);
	    if(id<0)
	    {
		// no race because the order is just created
		// no reference point to it
		o.setAttrib(AOrder.SET_REJECTED,AOrder.STATUS_CONFIRMED);
		logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		       "Rejected orderid="+id+" for ordertocancel="+
		       ordertocancel+" "+o.getDescription());
		return -1;
	    }
	    o.setID(id);
	    final OrderDataStored dor=this.globalsent.get(ordertocancel);
	    if(dor==null||dor.s!=s)
	    {
		throw new IllegalStateException("The order="+ordertocancel+
						" does not exist, or invalid s="+s+" dor.s="+
						(dor!=null?dor.s:null)+
						"; cannot cancel."+"\n"+s.positions.dumpData());
	    }
	    final OrderDataStored d=this.globalsent.put(id,new OrderDataStored(o,s));
	    if(d!=null)
	    {
		this.globalsent.put(id,d);
		throw new IllegalStateException("The order "+o.getDescription()+
						" is already sent for processing."+
						"\n"+s.positions.dumpData());
	    }

	synchronized(s.positions)
	{
	    // cancel old order
	    final AOrder oen=s.positions.book.get(ordertocancel);
	    if(oen==null)
	    {
		throw new IllegalStateException("The order="+ordertocancel+
						" does not exist, cannot cancel. new order id="+
						id+"\n"+s.positions.dumpData());
	    }
	    if(oen.isCancelled())
	    {
		logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		       "An attempt to cancel an inactive order="+ordertocancel+"; "+s.positions.dumpData());
	    }
	    oen.setAttrib(AOrder.SET_CANCELLED,AOrder.STATUS_MESSAGEPOSTED);

	    // add new order
	    // put the order immediately to a book.
	    final AOrder oe=s.positions.book.put(id,o);
	    if(oe!=null)
	    {
		s.positions.book.put(id,oe);
		throw new IllegalStateException("The order "+o.getDescription()+
						" is already in a book."+
						"\n"+s.positions.dumpData());
	    }
	    o.setAttrib(AOrder.SET_ADDED,AOrder.STATUS_MESSAGEPOSTED);
	    logEvent(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		     "ORDER:ADDED (C/R) "+o.getDescription());
	}
	} // END synchronized(this.globalsent)


	return id;
    }

    OrderDataStored hookForInternalUsagegetOdrerDatra(final long ordertocancel)
    {
	return this.globalsent.get(ordertocancel);
    }

    public long cancelOrder(final long ordertocancel,final TraderContainer s)
    {
	final long cancelid;
	synchronized(this.globalsent)
	{
	    cancelid=cancelOrder_ToBroker(ordertocancel);
	    if(cancelid<0)
	    {
		logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		       "Invalid cancelid="+cancelid+
		       " for ordertocancel="+ordertocancel);
		return -1;
	    }
	    final OrderDataStored dor=this.globalsent.get(ordertocancel);
	    if(dor==null||dor.s!=s)
	    {
		throw new IllegalStateException("The order="+ordertocancel+
						" does not exist, or invalid s="+s+" dor.s="+
						(dor!=null?dor.s:null)+
						"; cannot cancel."+"\n"+s.positions.dumpData());
	    }

	synchronized(s.positions)
	{
	    // cancel the order immediately from a book.
	    final AOrder oen=s.positions.book.get(ordertocancel);
	    if(oen==null)
	    {
		throw new IllegalStateException("The order="+ordertocancel+" does not exist, cannot cancel."+
						"\n"+s.positions.dumpData());
	    }
	    if(oen.isCancelled())
	    {
		logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		       "An attempt to cancel an inactive order="+ordertocancel+"; "+s.positions.dumpData());
	    }
	    // cancel old order
	    oen.setAttrib(AOrder.SET_CANCELLED,AOrder.STATUS_MESSAGEPOSTED);
	    logEvent(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		     "ORDER:CANCEL "+oen.getDescription());
	}
	} // END synchronized(this.globalsent)


	return cancelid;
    }



    public int cancelPerTickerOrders(final long [] orderstocancel,final TraderContainer s)
    {
	int n_err=0;
	int nc=0;
	final long [] oc=new long [orderstocancel.length];
	synchronized(this.globalsent)
	{
	    for(int i=0;i<oc.length;i++)
	    {
		final long cancelid=cancelOrder_ToBroker(orderstocancel[i]);
		if(cancelid<0)
		{
		    n_err++;
		    logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
			   "Rejected cancelid="+cancelid+
			   " for ordertocancel="+orderstocancel[i]+
			   " ticker="+s.ticker);
		    continue;
		}
		else
		{
		    oc[nc++]=orderstocancel[i];
		}
		final OrderDataStored dor=this.globalsent.get(orderstocancel[i]);
		if(dor==null||dor.s!=s)
		{
		    throw new IllegalStateException("The order="+orderstocancel[i]+
						    " does not exist, or invalid s="+s+" dor.s="+
						    (dor!=null?dor.s:null)+
						    "; cannot cancel."+"\n"+s.positions.dumpData());
		}
	    }

	synchronized(s.positions)
	{
	    for(int i=0;i<nc;i++)
	    {
		// cancel the order immediately from a book.
		final AOrder oen=s.positions.book.get(oc[i]);
		if(oen==null)
		{
		    throw new IllegalStateException("The order="+oc[i]+
						    " does not exist, cannot cancel."+
						    "\n"+s.positions.dumpData());
		}
		if(oen.isCancelled())
		{
		    n_err++;
		    logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
			   "An attempt to cancel an inactive order="+oc[i]+"; ticker="+s.ticker);
		}
		// cancel old order
		oen.setAttrib(AOrder.SET_CANCELLED,AOrder.STATUS_MESSAGEPOSTED);
		logEvent(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
			 "ORDER:CANCEL "+oen.getDescription());
	    }
	}

	if(n_err>0)
	{
	    logErr(BookItem.getTimeDiag(s.getTraderCurrentTime())+":"+
		   "POSITIONS="+s.positions.dumpData());
	}
	} // END synchronized(this.globalsent)

	return nc;
    }

    /** An order executed confirmation callback from broker.
     */
    public void orderExecuted_FromBroker(final ATrade t,final int shares_left,final long timestamp)
    {
	final OrderDataStored d;
	synchronized(this.globalsent)
	{
	    d=this.globalsent.get(t.order_id);
	    if(d==null)
	    {
		throw new IllegalStateException("The order="+t.id+
						" is not found in the list of submitted."+
						" time="+BookItem.getTimeDiag(timestamp));
	    }
	    final AOrder oexecuted=this.processes_executions.get(t.id);
	    if(oexecuted==null)
	    {
		this.processes_executions.put(t.id,d.thisorder);
	    }
	    else
	    {
		logErr(BookItem.getTimeDiag(timestamp)+":"+
		       "Re-play of already executed "+t.id+" for "+d.s.ticker);
		if(oexecuted!=d.thisorder)
		{
		    throw new IllegalStateException("object mismatch. oexecuted="+oexecuted+
						    " saved order="+d.thisorder+
						    "\n"+d.s.positions.dumpData());
		}
		// skip book modification, otherwise we may count trades multiple times
		return;
	    }

	synchronized(d.s.positions)
	{
	    final AOrder o=d.thisorder;
	    if(o!=d.s.positions.book.get(t.order_id)||
	       t.order_id!=o.getID())
	    {
		throw new RuntimeException("object mismatch. book object="+d.s.positions.book.get(t.order_id)+
					   " obj="+o+
					   " order "+o.getDescription()+
					   " trade "+t.getDescription()+
					   "\n"+d.s.positions.dumpData());
	    }

	    if(o.isCancelled())
	    {
		logErr(BookItem.getTimeDiag(timestamp)+":"+
		       "Executing a cancelled/rejected order="+t.order_id+
		       " trade "+t.getDescription()+
		       " order "+o.getDescription()+
		       " Probable reason: MISSED TRADE caught (probably) during re-play."+
		       "\n"+d.s.positions.dumpData());
	    }
	    if(t.shares<=0)
	    {
		throw new RuntimeException("Invalid  shares="+t.shares+
					   " trade "+t.getDescription()+
					   " order="+o.getDescription()+
					   "\n"+d.s.positions.dumpData());
	    }
	    final int new_shares=o.getShares()-t.shares;
	    if(new_shares<0)
	    {
		logErr(BookItem.getTimeDiag(timestamp)+":"+
		       "Invalid new_shares="+new_shares+
		       " trade "+t.getDescription()+
		       " order "+o.getDescription()+
		       "\n"+d.s.positions.dumpData());
	    }
	    o.setShares(new_shares);
	    t.setOrder(o);
	    o.setTime(timestamp);
	    d.s.positions.trades.add(t);
	    logEvent(BookItem.getTimeDiag(timestamp)+":"+
		     "TRADE:EXECUTED"+t.getDescription());
	    // shares_left<0 means unknown
	    if(shares_left>=0 && new_shares!=shares_left)
	    {
		throw new BrokerFailureException("Invalid shares left mismatch: new_shares="+new_shares+
						 " trade "+t.getDescription()+
						 " order "+o.getDescription()+
						 "\n"+d.s.positions.dumpData());
	    }
	}

	} // END synchronized(this.globalsent)

	d.s.notifyTrader();
    }


    /** An order accepted callback.
     */
    public void orderAdded_FromBroker(final BOrder bo,final long timestamp)
    {

	final OrderDataStored d;
	synchronized(this.globalsent)
	{
	    final long boid=bo.getID();
	    d=this.globalsent.get(boid);
	    if(d==null)
	    {
		throw new IllegalStateException("The order="+boid+
						" is not found in the list of submitted."+
						" time="+timestamp);
	    }

	synchronized(d.s.positions)
	{
	    final AOrder o=d.thisorder;
	    if(boid!=o.getID()||
	       o!=d.s.positions.book.get(boid))
	    {
		throw new RuntimeException("id or object mismatch. order id="+o.getID()+
					   " boid="+boid+
					   " book order object="+d.s.positions.book.get(boid)+
					   " order obj="+o+
					   "\n"+d.s.positions.dumpData());
	    }
	    if(/* o.getShares()!=bo.getShares()|| */
	       o.side!=bo.side||
	       o.getID()!=boid)
	    {
		throw new IllegalStateException("The order="+boid+" for "+d.s.ticker+
						" does not match saves order="+o.getID()+
						" shares="+bo.getShares()+":"+o.getShares()+
						" original shares="+o.original_shares+
						" side="+bo.side+":"+o.side+
						"\n"+d.s.positions.dumpData());
	    }
	    if(o.isCancelled())
	    {
		logErr(BookItem.getTimeDiag(timestamp)+":"+
		       "Adding a cancelled/rejected order="+boid+
		       " for "+d.s.ticker+
		       " time="+bo.getTime()+
		       "\n"+d.s.positions.dumpData());
	    }
	    o.setAttrib(AOrder.SET_ADDED,AOrder.STATUS_CONFIRMED);
	    o.setTime(timestamp);
	    logEvent(BookItem.getTimeDiag(timestamp)+":"+
		     "ORDER:ADDED:CONFIRMED "+o.getDescription());
	}

	} // END synchronized(this.globalsent)

	d.s.notifyTrader();
    }



    /** An order cancelled callback.
     */
    public void orderCancelled_FromBroker(final long ordertocancel,final String text,final long timestamp)
    {
	final OrderDataStored d;
	synchronized(this.globalsent)
	{
	    d=this.globalsent.get(ordertocancel);
	    if(d==null)
	    {
		throw new IllegalStateException("The order="+ordertocancel+
						" is not found in the list of submitted. time="+timestamp);
	    }

	synchronized(d.s.positions)
	{
	    final AOrder o=d.thisorder;
	    if(ordertocancel!=o.getID()||
	       o!=d.s.positions.book.get(ordertocancel))
	    {
		throw new RuntimeException("id or object mismatch. order id="+o.getID()+
					   " ordertocancel="+ordertocancel+
					   " book order object="+d.s.positions.book.get(ordertocancel)+
					   " order obj="+o+
					   "\n"+d.s.positions.dumpData());
	    }

	    o.setAttrib(AOrder.SET_CANCELLED,AOrder.STATUS_CONFIRMED);
	    o.setTime(timestamp);
	    logEvent(BookItem.getTimeDiag(timestamp)+":"+
		     "ORDER:CANCELLED:CONFIRMED "+o.getDescription()+"; reason="+text);
	}
	} // END synchronized(this.globalsent)

	d.s.notifyTrader();
    }


    /** A reject callback from broker.
     */
    public void orderRejected_FromBroker(final long id,final String text,final long timestamp)
    {
	final OrderDataStored d;
	synchronized(this.globalsent)
	{
	    d=this.globalsent.get(id);
	    if(d==null)
	    {
		throw new IllegalStateException("The order="+id+
						" is not found in the list of submitted. time="+timestamp);
	    }

	synchronized(d.s.positions)
	{
	    final AOrder o=d.thisorder;
	    if(id!=o.getID()||
	       o!=d.s.positions.book.get(id))
	    {
		throw new RuntimeException("id or object mismatch. order "+o.getDescription()+
					   " ordertocancel="+id+
					   " book order object="+d.s.positions.book.get(id)+
					   " order obj="+o+
					   "\n"+d.s.positions.dumpData());
	    }
	    o.setAttrib(AOrder.SET_REJECTED,AOrder.STATUS_CONFIRMED);
	    o.setTime(timestamp);
	    logEvent(BookItem.getTimeDiag(timestamp)+":"+
		     "ORDER:REJECTED:CONFIRMED "+o.getDescription()+(text!=null?";reason="+text:"; reason unknown"));
	}
	} // END synchronized(this.globalsent)

	d.s.notifyTrader();
    }


    public void logErr(final String str)
    {
	// for now
	System.err.println("GBroker:ERROR:"+getClass().getName()+":"+str);
    }


    public void logEvent(final String str)
    {
	// for now
	System.err.println("GBroker:EVENT:"+getClass().getName()+":"+str);
    }



}
