package com.polytechnik.broker;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.BookItem;
import com.polytechnik.realtimebook.TraderInterface;

/** A class that performs actual trading.
 *  @version $Id: TradingAlgorithm.java,v 1.148 2014/10/23 22:10:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class TradingAlgorithm<ET extends TradingAlgorithm.TradingEvent> {

/** Events we receive.
 */
public abstract static class TradingEvent implements AOrderCancellationCheck {
    public final long time;
    public final double price;
    final int side;
    public final boolean flag_close_postion;
    public TradingEvent(final int side,
			final long time,
			final double price,
			final boolean flag_close_postion)
    {
	this.side=side;
	this.time=time;
	this.price=price;
	this.flag_close_postion=flag_close_postion;
	if(side!=BOrder.SIDE_BUY&&side!=BOrder.SIDE_SELL_SHORT)
	{
	    throw new IllegalArgumentException();
	}
    }
    public abstract String getDescription();
}


    public abstract ET getEmergencyClose(final ET event_ontherun,final Positions.PositionsInfo pinfo);
    public abstract ET getOpportunityEvents(final ET event_ontherun,final Positions.PositionsInfo pinfo);
    public abstract ET getExitEvents(final ET event_ontherun,final Positions.PositionsInfo pinfo);

    protected ET getEventOntherunDiag()
    {
	return this.eventontherun;
    }

    protected static final boolean PRINT_DEBUG=true;

    private final TraderContainer sd;
    public TradingAlgorithm(final TraderContainer sd)
    {
	this.sd=sd;
    }

    // used for marking orders.
    private int strategy_id=-1;
    private int getStrategyID()
    {
	return this.strategy_id;
    }
    private static int strategies=7001;
    private void nextStrategy()
    {
	this.strategy_id=strategies++;
    }



    // a rudimentary final state automate
    protected static final int
	SGOAL_WATCHOPPORTUNITY=807,
	SGOAL_LIQUIDATE=808,
	SGOAL_HOLD=809;
    private int goal=SGOAL_LIQUIDATE;
    private ET eventontherun=null;


    private AOrder convertEventToOrder(final Positions.PositionsInfo pinfo,final ET event,final String comments)
    {
	if(event==null)
	{
	    return null;
	}
	final String s=comments!=null?" "+comments:"";
	if(event.flag_close_postion)
	{ // exit
	    final int position_shares_signed=pinfo.trade_info.current_shares;
	    if(position_shares_signed==0)
	    {
		// dummy order, 0 size, just signal to cancel all pending orders
		return new AOrder(
		    pinfo.time,
		    event,
		    event.price,
		    0,
		    pinfo.getEffShares()>0?BOrder.SIDE_SELL:BOrder.SIDE_BUY_COVER, // does not matter, size=0
		    new OrderAttributes(this.sd.ticker,event,getStrategyID(),"VIRTUAL CLOSE:"+pinfo.getMarketInfo()+s));
	    }
	    final int shares,side;
	    if(position_shares_signed>0)
	    { // long position, sell
		shares=position_shares_signed;
		side=BOrder.SIDE_SELL;
	    }
	    else
	    { // short position, buy to cover.
		shares=-position_shares_signed;
		side=BOrder.SIDE_BUY_COVER;
	    }
	    // close position
	    return new AOrder(pinfo.time,
			      event,
			      event.price,
			      shares,
			      side,
			      new OrderAttributes(this.sd.ticker,event,getStrategyID(),pinfo.getMarketInfo()+s));
	}

	// opportunity
	final int shares,side;
	if(event.side==BOrder.SIDE_BUY)
	{
	    // buy
	    shares=this.sd.moneymanager.getAvailableLong(pinfo.time,
							 pinfo.p_buy,
							 pinfo.p_sell,
							 (int)pinfo.avail_buy,
							 (int)pinfo.avail_sell);
	    side=BOrder.SIDE_BUY;
	}
	else if(event.side==BOrder.SIDE_SELL_SHORT)
	{
	    shares=this.sd.moneymanager.getAvailableShort(pinfo.time,
							  pinfo.p_buy,
							  pinfo.p_sell,
							  (int)pinfo.avail_buy,
							  (int)pinfo.avail_sell);
	    side=BOrder.SIDE_SELL_SHORT;
	}
	else
	{
	    throw new IllegalArgumentException();
	}

	return new AOrder(
	    pinfo.time,
	    event,
	    event.price,
	    shares,
	    side,
	    new OrderAttributes(this.sd.ticker,event,getStrategyID(),pinfo.getMarketInfo()+s));
    }

    /** Post a trade if event occured.
     *	@return next time to check opportunity.
     */
    public long tradeIfOpportunityIsGood(final BrokerInterface broker,
					 final long currenttime,
					 final double p_buy,
					 final double p_sell,
					 final long avail_buy,
					 final long avail_sell)
    {
	testing_reportPostedOrders(null); // clear memory
	//System.err.println(" goal="+goal);

	final Positions.PositionsInfo pinfo;
	final boolean empty_book;
	synchronized(this.sd.positions)
	{
	    pinfo=this.sd.positions.getInfo(currenttime,p_buy,p_sell,avail_buy,avail_sell);
	    empty_book=(pinfo.order_info.orders_active==null && pinfo.trade_info.current_shares==0);
	    if(empty_book && goal==SGOAL_LIQUIDATE)
	    {
		goal=SGOAL_WATCHOPPORTUNITY;
		this.eventontherun=null;
		nextStrategy();
		this.sd.positions.initiateNewStrategy(pinfo.time);
	    }
	}
	final int shares_eff=pinfo.getEffShares();

	// report current orders & trades info to MoneyManager
	this.sd.moneymanager.notifyPositionsInfo(pinfo);

	String comments=null;
	ET event=null;
	// default next check when we have a non-empty position
	long tau=this.sd.moneymanager.getTimeNextCheck_OpenPostition(pinfo.time);
	final boolean time_OK=this.sd.moneymanager.isTimeOK(pinfo.time);

	if(goal==SGOAL_WATCHOPPORTUNITY)
	{
	    if(getEventOntherunDiag()!=null){throw new RuntimeException("null!=eventontherun="+getEventOntherunDiag().getDescription());}
	    if(!empty_book)
	    {
		// watch opportunity got some "come late" positions, close immediately
		goal=SGOAL_LIQUIDATE;
		event=getEmergencyClose(getEventOntherunDiag(),pinfo);
		comments="Come late position close immediately.";
	    }
	    else
	    {
		event=getOpportunityEvents(getEventOntherunDiag(),pinfo);
		if(event!=null && time_OK)
		{
		    comments="Opportunity found";
		    // if we found an opportunity - switch state to hold a position
		    goal=SGOAL_HOLD;
		    if(PRINT_DEBUG)
		    {
			System.err.println("OPPORTUNITY Watchopportunity found new order="+
					   (event!=null?event.getDescription():"null")+
					   " time="+BookItem.getTimeDiag(pinfo.time)+
					   " effshares_existing="+pinfo.getEffShares()+
					   " Market="+pinfo.getMarketInfo());
		    }
		}
		else
		{
		    if(false && PRINT_DEBUG)
		    {
			System.err.println("OPPORTUNITY Watchopportunity SKIPPED="+
					   (event!=null?event.getDescription():"null")+
					   " time="+BookItem.getTimeDiag(pinfo.time)+
					   " effshares_existing="+pinfo.getEffShares()+
					   " orders_active="+(pinfo.order_info.orders_active!=null?java.util.Arrays.toString(pinfo.order_info.orders_active):"N/A")+
					   " Market="+pinfo.getMarketInfo());
		    }
		    event=null;
		    tau=this.sd.moneymanager.getTimeNextCheck_WatchOpportunity(pinfo.time);
		}
	    }
	}
	else if(goal==SGOAL_HOLD)
	{
	    if(getEventOntherunDiag()==null){throw new RuntimeException("null==eventontherun");}
	    if(false)
	    {
		System.err.println("HOLD: time="+BookItem.getTimeDiag(pinfo.time)+
				   " Market="+pinfo.getMarketInfo()+
				   " getEventOntherunDiag()="+getEventOntherunDiag().getDescription());
	    }

	    if(empty_book)
	    { // speedup exit
		if(PRINT_DEBUG)
		{
		    System.err.println("EMPTY BOOK CLOSE SPEEDUP SWITCH. order="+
				       (event!=null?event.getDescription():"null")+
				       " ticker="+this.sd.ticker+
				       " time="+BookItem.getTimeDiag(pinfo.time)+
				       " effshares_existing="+shares_eff+
				       " Market="+pinfo.getMarketInfo());
		}

		goal=SGOAL_LIQUIDATE;
		this.sd.moneymanager.notifyPositionsInfo(null);
		// speedup switch to watch for new opportunitues
		return tradeIfOpportunityIsGood(broker,
						currenttime,
						p_buy,
						p_sell,
						avail_buy,
						avail_sell);
	    }

	    if((shares_eff>0 && getEventOntherunDiag().side==BOrder.SIDE_SELL_SHORT)||
	       (shares_eff<0 && getEventOntherunDiag().side==BOrder.SIDE_BUY)||
	       shares_eff==0)
	    {
		if(PRINT_DEBUG)
		{
		    System.err.println("Wrong SIGN: shares_eff="+shares_eff+
				       " pinfo.trade_info.current_shares="+pinfo.trade_info.current_shares+
				       " pinfo.order_info.shares_buys="+pinfo.order_info.shares_buys+
				       " pinfo.order_info.shares_sells="+pinfo.order_info.shares_sells+
				       " N active orders="+(pinfo.order_info.orders_active!=null?String.valueOf(pinfo.order_info.orders_active.length):"N/A")+
				       " time="+BookItem.getTimeDiag(pinfo.time)+
				       " effshares_existing="+pinfo.getEffShares()+
				       " getEventOntherunDiag()="+getEventOntherunDiag().getDescription()+
				       " Market="+pinfo.getMarketInfo());
		}
		event=getEmergencyClose(getEventOntherunDiag(),pinfo);
		comments="Wrong SIGN. shares_eff="+shares_eff+" side="+getEventOntherunDiag().side+" Force position close, previous close failed or order is cancelled.";
	    }
	    else if(!time_OK)
	    {
		// close everything at the end of day
		event=getEmergencyClose(getEventOntherunDiag(),pinfo);
		comments="End of day or time drift force close";
	    }
	    else
	    {
		// we are holding positions
		// check whether to continue to hold.
		event=getExitEvents(getEventOntherunDiag(),pinfo);
		if(PRINT_DEBUG)
		{
		    System.err.println("EHOLD,event="+(event==null?"null":event.getDescription()));
		}
		if(event!=null) comments="Regular exit from opportunity";
	    }

	    if(event!=null)
	    { // regular exit or emergency exit
		goal=SGOAL_LIQUIDATE;
	    }
	}
	else if(goal==SGOAL_LIQUIDATE)
	{
	    if(getEventOntherunDiag()==null){throw new RuntimeException("null==eventontherun");}
	    // when liquidate
	    // if valid close order present - do nothing
	    // if no valid order exist - and new.
	    if(pinfo.trade_info.current_shares==0)
	    {
		event=getEmergencyClose(getEventOntherunDiag(),pinfo);
		comments="Liquidation 0 size resubmit.";
	    }
	    else
	    {
		final int norders;
		if(pinfo.trade_info.current_shares>0)
		{
		    norders=pinfo.order_info.n_sells;
		}
		else if(pinfo.trade_info.current_shares<0)
		{
		    norders=pinfo.order_info.n_buys;
		}
		else
		{
		    throw new IllegalArgumentException();
		}
		if(pinfo.order_info.orders_active==null ||
		   pinfo.order_info.orders_to_die!=null ||
		   norders<=0)
		{ // liquidate resubmit
		    event=getEmergencyClose(getEventOntherunDiag(),pinfo);
		    comments="Liquidation resubmit, some orders were not executed.";
		    if(PRINT_DEBUG)
		    {
			System.err.println("EFORCE,event="+(event==null?"null":event.getDescription()));
		    }
		}
		else
		{ // check exit re-submit
		    event=getExitEvents(getEventOntherunDiag(),pinfo);
		    if(event!=null) comments="Exit re-submit";
		    if(PRINT_DEBUG)
		    {
			System.err.println("ELQD,event="+(event==null?"null":event.getDescription()));
		    }
		}
	    }
	}
	else
	{
	    throw new IllegalStateException();
	}


	// cancel all or some of orders
	if(pinfo.order_info.orders_active!=null && (event!=null && event.flag_close_postion))
	{ // close re-submit, probably at different price
	    broker.cancelPerTickerOrders(pinfo.order_info.orders_active,this.sd);
	}
	else if(pinfo.order_info.orders_to_die!=null)
	{
	    broker.cancelPerTickerOrders(pinfo.order_info.orders_to_die,this.sd);
	}

	final AOrder order=convertEventToOrder(pinfo,event,comments);
	if(event!=null){this.eventontherun=event;}

	// add new order if any
	if(order!=null && order.getShares()>0)
	{
	    // when closing position 0 size order indicate
	    // to cancel all previous order, skip such order from posting to broker.
	    // actually post to broker
	    broker.addNewOrder(order,this.sd);
	    if(PRINT_DEBUG)
	    {
		System.err.println("ADD ORDER order="+
				   (order!=null?order.getDescription():"null")+
				   " time="+BookItem.getTimeDiag(pinfo.time)+
				   " effshares_existing="+pinfo.getEffShares()+
				   " Market="+pinfo.getMarketInfo());
	    }
	}

	if(PRINT_DEBUG)
	{
	    if(order!=null)
	    {
		System.err.println("before testing_reportPostedOrders: order="+order.getDescription()+
				   " Market="+pinfo.getMarketInfo());
	    }

	}
	// testing hook
	testing_reportPostedOrders(order);



	if(PRINT_DEBUG)
	{ // dump positions after trading is done
	    if(currenttime>=ttodump)
	    {
		System.err.println("REPORT:DUMP STATUS:"+
				   " ticker="+this.sd.ticker+
				   " time="+BookItem.getTimeDiag(pinfo.time)+
				   " money_spent="+pinfo.trade_info.money_p_and_l+
				   " unrel P&L="+pinfo.trade_info.unrealized_p_and_l+
				   " shares="+pinfo.trade_info.shares+
				   " effshares_existing="+pinfo.getEffShares()+
				   " timeOK="+time_OK+
				   " Market="+pinfo.getMarketInfo());

		testing_saveOrdersAndTrades("Temporarily diagnostics for "+this.sd.ticker+
					    " dump at "+BookItem.getTimeDiag(pinfo.time)+" timeOK="+time_OK,
					    pinfo);
		// dump diag every few minutes
		ttodump=currenttime+MoneyManager.FREQUENCY_REPORTING;
	    }
	}

	// clear money monager
	this.sd.moneymanager.notifyPositionsInfo(null);

	return tau;
    }
    private long ttodump=0; // variable to control reports & diagnostincs reports printing frequency.


    /** A hook for debugging. */
    protected void testing_reportPostedOrders(final AOrder order)
    {
    }

    protected void testing_saveOrdersAndTrades(final String txt,final Positions.PositionsInfo pinfo)
    {
    }

}
