package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** A simple container for common basis operations.
 * Includes the basis itself {@link #getABasis} and variation-by-parts
 * helpers for specific integration weight and basis used.
 * @version $Id: CommonBasisOperations.java,v 1.11 2022/09/05 02:14:05 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public interface CommonBasisOperations {

    /** Return the moments {@literal <Qk>}, {@code k=0..n-1},
     * for the measure and basis used.
     */
    double [] getdtMoments(final int n);

    /** Return time--variation {@literal |psi>}. Typically a linear function 
     * on {@code psi} and the derivative of {@code psi} multiplied
     * by a weight-dependent polynomial.
     */
    double [] getEDPsi(final double [] psi);

    /** Integrate a polynomial as \(R(x)\omega(t)=\int_{-\infty}^{t} P(x(t))\omega(t)dt\).
     * The result \(R(x)\) is weight-dependent and does not exist for some bases,
     * the class {@link CommonBasisOperations} is intendent to be used 
     * only for the weights and bases with analytically available integration.
     */
    double [] getPolynomialWIntegratedDt(final double [] polynomial);

    /** The basis.
     */
    OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();

    /** Obtain {@literal \(\left<FQ_k\right>\)} moments 
     * from {@literal \(\left<\frac{dF}{dt}Q_k\right>\)} moments
     * using integration by parts and {@link #getPolynomialWIntegratedDt}.
     * The default implementation uses {@link #getTransformFMomentsFromDF} helper method.
     * @see com.polytechnik.trading.DataInterfaceFDFMoments#setFMoments
     */
    default double [] getFMomentsFromDF(final double [] dFmoments){
	return CommonBasisOperations.getTransformFMomentsFromDF(this::getPolynomialWIntegratedDt,dFmoments);
    }

    /** Obtain {@literal \(\left<\frac{dF}{dt}Q_k\right>\)} moments 
     * from {@literal \(\left<FQ_k\right>\)} moments.
     * @see #getEDPsi
     * @see com.polytechnik.trading.DataInterfaceFDFMoments#setDFMoments
     */
    double [] getDFMomentsFromF(final double [] Fmoments);

    WIntegrator getEmptyWIntegrator(final int ntm,final double atau,final int nobservables);

    default double [] getPsi2WIntegratedDt(final double [] psi)
    {
	return getPolynomialWIntegratedDt(getABasis().B.mult2Pol(psi,psi));
    }

/** A helper method to obtain {@literal \(\left<FQ_k\right>\)} moments 
 * from {@literal \(\left<\frac{dF}{dt}Q_k\right>\)} moments.
 * Can be used for other than {@link CommonBasisOperations} data sources.
 */
    public static double [] getTransformFMomentsFromDF(java.util.function.Function<double [],double []> getPolynomialWIntegratedDt,final double [] dFmoments){
	final double [] FMoments=new double [dFmoments.length];
	for(int k=FMoments.length;--k>=0;) {
	    final double [] a=new double [k+1];
	    a[k]=-1; // "-" sign integration by parts
	    FMoments[k]=TMatr.mulVVn1(getPolynomialWIntegratedDt.apply(a),dFmoments);
	}
	return FMoments;	
    }

}
