package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.Polynomials;

/** A class to calculate commonly used moments.
 * It has an internal state {@link WIntegrator} and methods for
 * updating the internal state with new {@code t,P,dv} coming.
 * @version $Id: CommonlyUsedMoments.java,v 1.91 2022/09/14 14:39:30 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class CommonlyUsedMoments implements CommonBasisOperations {

    /** Recurrent integrator.
     */
    private final WIntegrator wi;

    /** Price offset to increase numerical stability.
     */
    private double p_offset;

    /** For dp/dt calculation.*/
    private double P_lastsaved;

    /** For d lambda^{IH}/dt calculation using secondary sampling.*/
    private double IH_lastsaved,P_for_IH_lastsaved;

    /** Time of last trade added, may be different from 
     * last basis time reference {@link WIntegrator#currenttime_previous}.
     */
    private long t_lastsaved;
    
    /** Initial state.
     */
    public CommonlyUsedMoments(final int ntm,final double atau){
	super();
	this.wi=getEmptyWIntegrator(ntm,atau,18);
	this.p_offset=0;
	// As wi.sum is incorrect (we do not track it's offset changes) -- set to NaN to avoid using the result.
	java.util.Arrays.fill(wi.sum,Double.NaN);
	wi.sum[0/* dv */]=wi.sum[8 /* dp */]=wi.sum[10/* da */]=wi.sum[15 /* dIH */]=0; // these do not use any offset

	// for dp/dt, dIH/dt, and dpIH/dt
	this.P_lastsaved=Double.NaN;
	this.t_lastsaved=Long.MIN_VALUE;
	this.IH_lastsaved=0;
	this.P_for_IH_lastsaved=Double.NaN;
    }


    /** A trivial update with a single triple (t,P,dv) coming.
     * The method shifts the basis to correspond to {@code currenttime};
     * new value of {@link #getPOffset} can be also set to increase numerical stability.
     * Internally the method adjusts the basis using {@link #updateBasisToNewTimeReference},
     * then calls {@link #addObservationNoBasisShift} to add an observation.
     */
    public void updateWithSingleObservation(long currenttime,long t_last, double P_last, /*long*/double dv_last){

	// Update basis offsets.
	// disabled: if dV=0 do not change price offset.
	this.updateBasisToNewTimeReference(currenttime,(dv_last>0||true)?P_last:getPOffset());
	// add single observation
	this.addObservationNoBasisShift(currenttime,t_last,P_last,dv_last);
    }

    public void updateBasisToNewTimeReference(final long currenttime,final double new_p_offset){

	// 1: Value shift offsets to (t-currenttime) and (p-new_p_offset)
	final double dtnow=(currenttime-wi.currenttime_previous)/wi.atau,
	    dpoffset=(new_p_offset-getPOffset());
	final double []
	    i=wi.integration_moments[0],
	    pi=wi.integration_moments[1],
	    p2i=wi.integration_moments[2],
	    p3i=wi.integration_moments[3],
	    Ti=wi.integration_moments[4],
	    Tpi=wi.integration_moments[5],
	    Tp2i=wi.integration_moments[6],
	    Tp3i=wi.integration_moments[7],
	    dp=wi.integration_moments[8],
	    Tdp=wi.integration_moments[9],
	    da=wi.integration_moments[10],
	    pda=wi.integration_moments[11],
	    Tda=wi.integration_moments[12],
	    VdP=wi.integration_moments[13],
	    AdP=wi.integration_moments[14],
	    // three last are used with secondary sampling
	    dIH=wi.integration_moments[15],
	    pdIH=wi.integration_moments[16],
	    dPIH=wi.integration_moments[17];
	
	// new price offset
	Polynomials.shiftPOffsetInPlace(wi.ntm,i,pi,p2i,p3i,dpoffset);
	Polynomials.shiftPOffsetInPlace(wi.ntm,Ti,Tpi,Tp2i,Tp3i,dpoffset);
	for(int k=0;k<wi.ntm;k++){
	    // shift to new time offset
	    Ti[k]  -= dtnow*i[k];
	    Tpi[k] -= dtnow*pi[k];
	    Tp2i[k]-= dtnow*p2i[k];
	    Tp3i[k]-= dtnow*p3i[k];
	    Tdp[k] -= dtnow*dp[k];
	    Tda[k] -= dtnow*da[k];
	    // here is price offset, just a few array to update
	    pda[k] -= dpoffset*da[k];
	    pdIH[k] -= dpoffset*dIH[k];
	    dPIH[k] -= dpoffset*dIH[k];
	}
	this.p_offset=new_p_offset;
	// Note! wi.sum is incorrect because of always changing offsets!!!
	// we do not use it thus do not apply formulas similar to that of above
	// to obtain proper wi.sum values.

	// 2: Basis functions shift argument offset
	wi.updateSavedMomentsToNewTimeReference(currenttime);
    }

    private long counter_addObservationNoBasisShift=0,counter_addIHObservationSecondarySampling=0;
    public void addObservationNoBasisShift(final long currenttime,final long t,final double P,final double dv){
	counter_addObservationNoBasisShift++;

	final double ttnow=(t-currenttime)/wi.atau,p=(P-getPOffset()),pdv=p*dv,p2dv=p*pdv,p3dv=p*p2dv,
	    dp,pprev
	    ;
	if(this.t_lastsaved<=Long.MIN_VALUE){
	    dp=0;
	    pprev=p;
	} else {
	    dp=P-this.P_lastsaved;
	    pprev=(this.P_lastsaved-getPOffset());
	}
	final double da=Math.abs(dp), // poor man dv
	    pda=p*da, //,p2da=p*pda,p3da=p*p2da
	    vdp= -dv*dp, // VdP/dt
	    adp= -da*dp // AdP/dt
	    ;

	// adjust volume-offset here
	final double []
	     dP=wi.integration_moments[8],
	    VdP=wi.integration_moments[13],
	    AdP=wi.integration_moments[14];
	    
	for(int k=0;k<wi.ntm;k++){
	    VdP[k]-=dv*dP[k];
	    AdP[k]-=da*dP[k];
	}

	// arguments, the order:
	// dv:[0], pdv:[1], p2dv:[2], p3dv:[3], Tdv:[4], Tpdv:[5], Tp2dv:[6], Tp3dv:[7], dp:[8], T*dp:[9], da:[10], p*da:[11], T*da:[12], VdP:[13], AdP:[14], dIH[15], pdIH[16], dPIH[17] (dIH moments will be set later with addIHObservationSecondarySampling)
	if(this.t_lastsaved>t) throw new RuntimeException("time-unordered input will screw dp/dt calculation t="+t+" t_lastsaved="+t_lastsaved);
	this.P_lastsaved=P;
	this.t_lastsaved=t;
	wi.addSingleObservation(currenttime,t,new double []{dv,pdv,p2dv,p3dv,ttnow*dv,ttnow*pdv,ttnow*p2dv,ttnow*p3dv,dp,ttnow*dp,da,pda,ttnow*da,vdp,adp,0,0,0});	
    }

    /** An experiment to calculate
     * \( \frac{d}{dt}I^H \),  \( \frac{d}{dt} P I^H \), and \( P\frac{d}{dt}I^H \)
     * moments by using secondary sampling; the \( I^H \) is calculated from directly sampled moments
     * and then used as it were a regular observable.
     * The call of {@link addIHObservationSecondarySampling} should immediately 
     * follow {@link #updateWithSingleObservation}.
     * We use incremental calculation of moments.
     * @see com.polytechnik.trading.CommonlyUsedMoments_TestFromSample#fromSeqRecurrentlyTestSecondarySampling
     */
    public void addIHObservationSecondarySampling(final double IH){
	if(counter_addObservationNoBasisShift!=counter_addIHObservationSecondarySampling+1
	   || this.t_lastsaved<=Long.MIN_VALUE)
	    throw new IllegalStateException("An invalid order of calls: counter_addObservationNoBasisShift="+counter_addObservationNoBasisShift+" counter_addIHObservationSecondarySampling="+counter_addIHObservationSecondarySampling+" t_lastsaved="+this.t_lastsaved);
	counter_addIHObservationSecondarySampling++;

	// for now: take internally saved values.
	final long currenttime=wi.currenttime_previous,t=this.t_lastsaved;
	final double P=this.P_lastsaved,Pprev=this.P_for_IH_lastsaved;

	// check constraint
	if(t!=this.t_lastsaved||currenttime!=wi.currenttime_previous)
	    throw new IllegalStateException("Time mismatch: currenttime="+currenttime+" wi.currenttime_previous="+wi.currenttime_previous+" t="+t+" this.t_lastsaved="+this.t_lastsaved);

	double p=(P-getPOffset()),pprev=(Pprev-getPOffset()),
	    dIH=IH-this.IH_lastsaved,
	    dPIH=!Double.isFinite(pprev)?p*dIH:(p*IH-pprev*this.IH_lastsaved);
	if(Double.isFinite(IH)){
	    this.IH_lastsaved=IH;
	    this.P_for_IH_lastsaved=P;
	}
	if(!Double.isFinite(dIH)){
	    dIH=0;
	    dPIH=0;
	}
	wi.addSingleObservationToSomeObservables(currenttime,t,new double []{dIH,p*dIH,dPIH},15);
    }
    
    /** Price Offset. Selected close to last price to simplify calculations.
     * Average price would probably provide a slightly better numerical stability.
     * All price moments are calculated relatively this offset,
     * this greatly increases numerical stability.
     * To obtain "actual" price from the values stored in moments data --
     * add the value returned by this method.
     * @see com.polytechnik.trading.DataInterfaceDirect#getPOffset
     */
    public double getPOffset(){return this.p_offset;}

    /** Last price as per {@link #addObservationNoBasisShift} call.
     * No price offset used. To use it with saved price moments -- use 
     * the value {@link #getPLast_withoutPOffset}-{@link #getPOffset}.
     */
    public double getPLast_withoutPOffset(){return this.P_lastsaved;}

    // If output is modifed -- integrator state will be incorrect.
    // clone() the data if the fields to be changed to public.

    /** Volume moments {@literal <I*Qk>}. Corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_I} and price power=0.
     * @see #getdAMoments
     */
    public double [] getdVMoments(){return wi.integration_moments[0/* dv */].clone();}

    public double getTotalVolume(){return wi.sum[0/* dv */];}
    public double getTotalA(){return wi.sum[10/* da */];}

    /** {@literal <p*I*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_I} and price power=1.
     */
    public double [] getPdVMoments(){return wi.integration_moments[1/* (P-p_offset)*dv */].clone();}

    /** {@literal <p^2*I*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_I} and price power=2.
     */
    public double [] getP2dVMoments(){return wi.integration_moments[2/* (P-p_offset)^2*dv */].clone();}

    /** {@literal <p^3*I*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_I} and price power=3.
     */
    public double [] getP3dVMoments(){return wi.integration_moments[3/* (P-p_offset)^3*dv */].clone();}

    /** {@literal <T*I*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_TI} and price power=0.
     */
    public double [] getTdVMoments(){return wi.integration_moments[4/* (t-tnow)*dv */].clone();}

    /** {@literal <p*T*I*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_TI} and price power=1.
     */
    public double [] getTPdVMoments(){return wi.integration_moments[5/* (t-tnow)*(P-p_offset)*dv */].clone();}

    /** {@literal <p^2*T*I*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_TI} and price power=2.
     */
    public double [] getTP2dVMoments(){return wi.integration_moments[6/* (t-tnow)*(P-p_offset)^2*dv */].clone();}

    /** {@literal <p^3*T*I*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_TI} and price power=3.
     */
    public double [] getTP3dVMoments(){return wi.integration_moments[7/* (t-tnow)*(P-p_offset)^3*dv */].clone();}

    /** Price change moments {@literal <dp/dt*Qk>}. Corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_DP} and price power=1.
     */
    public double [] getdpMoments(){return wi.integration_moments[8/* dp */].clone();}

    /** {@literal <T*dp/dt*Qk>}, corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_TDP} and price power=1.
     */
    public double [] getTdpMoments(){return wi.integration_moments[9/* T*dp */].clone();}

    /** A poor man volume, {@literal <abs(dp)/dt*Qk>}. Instead of a number of shares traded
     * the absolute value of every tick price 
     * change is used as integration measure.
     * Can be used instead of volume with mixed results, see
     * <a href="https://arxiv.org/abs/1510.05510">Fig. 6</a>.
     * Corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_A} and price power=0.
     * @see #getdVMoments
     */
    public double [] getdAMoments(){return wi.integration_moments[10/* da */].clone();}

    /** The moments {@literal <p*abs(dp)/dt*Qk>}. Instead of a number of shares traded
     * the absolute value of every tick price 
     * change is used as integration measure.    
     * Corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_A} and price power=1.
     * @see #getPdVMoments
     */
    public double [] getPdAMoments(){return wi.integration_moments[11/* (P-p_offset)*dA */].clone();}


    /** {@literal <A*dP/dt *Qk>}.
     */
    public double [] getAdPMoments(){return wi.integration_moments[14/* A*dP/dt */].clone();}

    
    /** {@literal <T*abs(dp)/dt*Qk>}.
     * @see #getdAMoments
     */
    public double [] getTdAMoments(){return wi.integration_moments[12/* T*dA */].clone();}

    
    /** Obtained using integration by parts from {@link #getdtMoments}.
     */
    public double [] getTdtMoments(){return getFMomentsFromDF(getdtMoments());}

    /** Obtained using integration by parts from {@link #getdVMoments}.
     */
    public double [] getVdtMoments(){return getFMomentsFromDF(getdVMoments());}

    /** The moments of a constant {@literal <Qk>}, 
     * analytically obtained exact value.
     * Corresponds to {@link com.polytechnik.trading.DataInterfaceDirect#setMT}.
     */
    public double [] getdtMoments(){return getdtMoments(wi.ntm);}

    
    /** {@literal <V*dP/dt *Qk>}.
     * Corresponds to {@link com.polytechnik.trading.DataInterfaceDirectTradesHistorical#M_VdPdt} and price power=1.
     * @see com.polytechnik.trading.CommonlyUsedMoments_TestFromSample#fromSeqRecurrentlyTestSecondarySampling
     */
    public double [] getVdPMoments(){return wi.integration_moments[13/* V*dP/dt */].clone();}

    /** An experimental method to obtain \( \frac{d}{dt}I^H \) moments using secondary sampling.
     * Requires {@link #addIHObservationSecondarySampling} to be properly called on every tick.
     * @see #addIHObservationSecondarySampling
     */
    public double [] getdIHdtMoments(){
	if(counter_addObservationNoBasisShift!=counter_addIHObservationSecondarySampling)
	    throw new IllegalStateException("An invalid order of calls: counter_addObservationNoBasisShift="+counter_addObservationNoBasisShift+" counter_addIHObservationSecondarySampling="+counter_addIHObservationSecondarySampling);
	return wi.integration_moments[15/* dIH/dt */].clone();
    }

    /** An experimental method to obtain  \( P\frac{d}{dt}I^H \) moments using secondary sampling.
     * Requires {@link #addIHObservationSecondarySampling} to be properly called on every tick.
     * @see #addIHObservationSecondarySampling
     */
    public double [] getPdIHdtMoments(){
	if(counter_addObservationNoBasisShift!=counter_addIHObservationSecondarySampling)
	    throw new IllegalStateException("An invalid order of calls: counter_addObservationNoBasisShift="+counter_addObservationNoBasisShift+" counter_addIHObservationSecondarySampling="+counter_addIHObservationSecondarySampling);
	return wi.integration_moments[16/* (P-p_offset)*dIH/dt */].clone();
    }

    /** An experimental method to obtain \( \frac{d}{dt}PI^H \) moments using secondary sampling.
     * Requires {@link #addIHObservationSecondarySampling} to be properly called on every tick.
     * For secondary sampling where the \( I^H \) can be of various meaning
     * it is better to use \( \frac{d}{dt}PI^H \) rather than
     * \( I^H\frac{d}{dt}P \) as in {@link #getVdPMoments}
     * for the reason of unit test simplification.
     * The \( I^H\frac{d}{dt}P \) can be obtained as {@link #getdPIHdtMoments}-{@link getPdIHdtMoments}.
     * @see #addIHObservationSecondarySampling
     * @see #getVdPMoments
     */
    public double [] getdPIHdtMoments(){
	if(counter_addObservationNoBasisShift!=counter_addIHObservationSecondarySampling)
	    throw new IllegalStateException("An invalid order of calls: counter_addObservationNoBasisShift="+counter_addObservationNoBasisShift+" counter_addIHObservationSecondarySampling="+counter_addIHObservationSecondarySampling);
	return wi.integration_moments[17/* d(P-p_offset)*IH/dt */].clone();
    }

    /** For unit tests to imitate initial 
     * input to match an observable's derivative.
     * In application the initial value is zero and this method is not required.
     */
    public void setInitIHForSecondarySampling(final double IH_lastsaved){
	if(counter_addObservationNoBasisShift!=0||
	   counter_addIHObservationSecondarySampling!=0)
	    throw new IllegalStateException("The call is allowed only for empty integrator");
	this.IH_lastsaved=IH_lastsaved;
    }
    
} // END OF class CommonlyUsedMoments
