package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.VolMatrix;

/**
 * Directional indicator in the state: between a spike in {@code I=dV/dt} and "now".
 * @see com.polytechnik.trading.IntervalSincePsiM
 * @version $Id: FreeMoneyForAll.java,v 1.172 2022/09/29 21:00:05 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public class FreeMoneyForAll {


    /** Regular moving average.
     */
    public final double pv_average;

    /** Standard deviation of {@link #pv_average}.
     */
    public final double stdev_pv_average;

    /** Price in the state of global max {@code I} .
     */
    public final double pv_IH,pt_IH;

    /** Time to "now" (in tau units) in the state of global max {@code I} .
     */
    public final double T_IH,Tv_IH;

  
    private static final boolean FORCE_SELFTEST_BY_DEFALT=false;


    public FreeMoneyForAll(final int n,final CommonlyUsedMoments sst,final boolean flagUseSecondarySamplingFordIHdt){
	this(n,sst,flagUseSecondarySamplingFordIHdt,(FORCE_SELFTEST_BY_DEFALT? (sst.getTotalVolume()>100000?1e-6:0) : 0 ));
    }

    public final PFuture pFV,pFA;
    
/**
* Constructor.
* @param n Basis dimension.
* @param sst Internal state with moments.
* @param flagUseSecondarySamplingFordIHdt Whether to use secondary sampling.
* @param eps If positive -- run internal selftests with this eps,
* typical value is about {@code 1e-5}. 
* May fail and throw {@link RuntimeException} if 
* internal results do not match within this {@code eps}.
*/
public FreeMoneyForAll(final int n,
		       final CommonlyUsedMoments sst,
		       final boolean flagUseSecondarySamplingFordIHdt,
		       final double eps)
{

    final double []
	imom=sst.getdVMoments(),
	tmom=sst.getdtMoments(),
	pimom=sst.getPdVMoments(),
	VdPmom=sst.getVdPMoments(),
	dpmom=sst.getdpMoments();
    final double p_offset=sst.getPOffset(),
	p_last=sst.getPLast_withoutPOffset()-p_offset;
 
    
    final var M=sst.getABasis();
    final double []    
	QQt=M.B.getQQMatr(n,tmom),
	QQi=M.B.getQQMatr(n,imom);
    final var dmi=new IandDM(n,QQt,QQi,M,sst::getPolynomialWIntegratedDt,eps);
    final int ind_IH=n-1;// maximal eigenvalue

    // We need to know I.sH for secondary sampling
    if(flagUseSecondarySamplingFordIHdt){
	// secondary sampling for dIH/dt
	sst.addIHObservationSecondarySampling(dmi.I.sH);
	// Secondary sampling moments
	//    sst.getdIHdtMoments()
	//    sst.getPdIHdtMoments()
	//    sst.getdPIHdtMoments()
	// are now available.
	// These moments are available only when flagUseSecondarySamplingFordIHdt=true.
	// They are properly set by sst.addIHObservationSecondarySampling call above
	// that must follow each sst.updateWithSingleObservation call.
    }
  

    final double [] QQdp=M.B.getQQMatr(n,dpmom);
    this.pFV=new PFuture(dmi,
			 p_offset,
			 p_last,
			 // input moment matrices for eigenstate problem
			 QQt,
			 QQi,
			 // input moment matrices for other calculations
			 M.B.getQQMatr(n,sst.getPdVMoments()),
			 M.B.getQQMatr(n,sst.getTdVMoments()),
			 M.B.getQQMatr(n,sst.getVdtMoments()),
			 QQdp,
			 M.B.getQQMatr(n,sst.getVdPMoments()),
			 sst.getTotalVolume(),
			 M,sst::getPolynomialWIntegratedDt,sst::getEDPsi,eps);

    // ==== Surrogate Volume
    final double [] amom=sst.getdAMoments(),
	QQa=M.B.getQQMatr(n,amom),
	QQA=M.B.getQQMatr(n,sst.getFMomentsFromDF(amom));
    final var dma=new IandDM(n,QQt,QQa,M,sst::getPolynomialWIntegratedDt,eps);
    this.pFA=new PFuture(dma,
			 p_offset,
			 p_last,
			 // input moment matrices for eigenstate problem
			 QQt,
			 QQa,
			 // input moment matrices for other calculations
			 M.B.getQQMatr(n,sst.getPdAMoments()),
			 M.B.getQQMatr(n,sst.getTdAMoments()),
			 QQA,
			 QQdp,
			 M.B.getQQMatr(n,sst.getAdPMoments()),
			 sst.getTotalA(),
			 M,sst::getPolynomialWIntegratedDt,sst::getEDPsi,eps);


    // ========================================================
    // Other values 
    final double []
	// always available
	QQpi=M.B.getQQMatr(n,pimom),
	QQp2i=M.B.getQQMatr(n,sst.getP2dVMoments()),
	QQpt=M.B.getQQMatr(n,TMatr.sumaAbB(p_last,tmom,1,sst.getFMomentsFromDF(dpmom))),
	// The sign is negative as we use negative (t-tnow) Radau multiplier
	// in moments definition.
	QQT=M.B.getQQMatr(n,sst.getTdtMoments()),
	QQV=M.B.getQQMatr(n,sst.getVdtMoments()),
	QQTi=M.B.getQQMatr(n,sst.getTdVMoments());

    this.pv_average=p_offset+QQpi[0]/QQi[0];
    // were it not for p_offset high order moments would be unstable
    this.stdev_pv_average=Math.sqrt(QQp2i[0]/QQi[0]-Math.pow(QQpi[0]/QQi[0],2));

    this.pv_IH=p_offset+dmi.I.getFPsi(ind_IH,QQpi)/dmi.I.ev[ind_IH];
    this.pt_IH=p_offset+dmi.I.getFPsi(ind_IH,QQpt);
    this.T_IH= -dmi.I.getFPsi(ind_IH,QQT);
    this.Tv_IH= -dmi.I.getFPsi(ind_IH,QQTi)/dmi.I.ev[ind_IH];   
}

}
