package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.PolynomialsRatio;
import com.polytechnik.utils.PolynomialRootsFinder;
/**
 * The state of {@literal
\( \max I= \frac{\left<\psi|QQi|\psi\right>}{\left<\psi|QQt|\psi\right>} \)
}
among localized states
{@literal \( \left|\psi\right> = \frac{1}{\mathrm{norm}}G^{-1}Q(x) \) }.
* The problem is reduced to maximizing a ratio of two polynomials
* of the degree {@code 2n-2}. The derivative roots of this ratio
* correspond to the roots of a polynomial of the degree {@code 4n-5},
* a solution is then selected among all available roots of the derivative.
* This polynomial of the degree {@code 4n-5} has leading coefficient 0,
* thus actual order is {@code 4n-6}, the number of coefficients in polynomial is {@code 1+4n-6}.
* @see EVXData
* @version $Id: ILocalizedStates.java,v 1.13 2022/08/08 20:10:45 mal Exp $
* @author Vladislav Malyshkin mal@gromco.com
*/
public class ILocalizedStates {

    public final SelectMinMaxFValuesOnDerivativeRoots res;

    public ILocalizedStates(
	final int n,

	// input moment matrices
	final double [] QQt,
	final double [] QQi,
	final double [] InvG,

	// basis operations
	final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,

	// for unit test
	final double eps){

	final double []
	    PolNom=M.B.getKK(n,TMatr.mulMMM(n,InvG,QQi,InvG)),
	    PolDenom=M.B.getKK(n,TMatr.mulMMM(n,InvG,QQt,InvG));
	double [] coefs=TMatr.sumAcB(
		M.B.mult2Pol(M.B.getDifferentiated(PolNom),PolDenom),
		-1,
		M.B.mult2Pol(M.B.getDifferentiated(PolDenom),PolNom));
	if(eps>0){
	    System.err.println("ILocalizedStates:n="+n+" coefs="+java.util.Arrays.toString(coefs)+" coefs.length="+coefs.length);
	    UnitTests.checkDoubleRel(coefs[coefs.length-1],"coefs[coefs.length-1]",0,"0",eps);
	}
	if(false){
	    // check the highest order coefficient is not too small
	    final double MINHIGHORDERCOEF=1e-5;
	    for(int i=coefs.length;--i>=2;){
		//System.err.println("i="+i+" coefs[i]="+coefs[i]);
		if(!(Math.abs(coefs[i])>MINHIGHORDERCOEF)){
		    //System.err.println("remove "+coefs[i]+" coefs="+java.util.Arrays.toString(coefs));
		    coefs=java.util.Arrays.copyOf(coefs,i);
		} else break;
	    }
	} else {
	    // just remove highest order coefficient that is 0
	    coefs=java.util.Arrays.copyOf(coefs,coefs.length-1);
	}

	this.res=new SelectMinMaxFValuesOnDerivativeRoots(
	    n,
	    coefs,
	    InvG,
	    psi->TMatr.getAMatrA(psi,QQi,n)/TMatr.getAMatrA(psi,QQt,n),
	    M,
	    eps);

	if(eps>0){
	    final var I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	    final var polsRatio=new PolynomialsRatio(PolDenom,PolNom,M.B);
	    int n_real_roots=0;
	    for(int k=0;k<res.r.nroots;k++){
		if(Math.abs(res.r.im_root[k])<=0){
		    n_real_roots++;
		    final double x=res.r.re_root[k],
			scaleF=Math.max(I.sH,Math.abs(polsRatio.getF(x)));
		    // may give 1e-3 error
		    UnitTests.checkDoubleRel(polsRatio.getDF(x)/scaleF,"DF",0,"0",eps);
		}
	    }
	    // OK only if G=QQt
	    if(true) UnitTests.checkDoubleRel(I.s0,"I.s0",polsRatio.getF(I.x0),"I(x0)",eps);
	    // always OK
	    UnitTests.checkDoubleRel(polsRatio.getF(res.x_Fmin),"I(res.x_Fmin)",res.Fmin,"res.Fmin",eps);
	    UnitTests.checkDoubleRel(polsRatio.getDF(res.x_Fmin)/Math.max(I.sH,Math.abs(polsRatio.getF(res.x_Fmin))),"DF(res.x_Fmin)",0,"0",eps);
	    if(polsRatio.getD2F(res.x_Fmin)<0) throw new RuntimeException("res.x_Fmin="+res.x_Fmin+" D2="+polsRatio.getD2F(res.x_Fmin));
	    UnitTests.checkDoubleRel(polsRatio.getF(res.x_Fmax),"I(res.x_Fmax)",res.Fmax,"res.Fmax",eps);
	    UnitTests.checkDoubleRel(polsRatio.getDF(res.x_Fmax)/Math.max(I.sH,Math.abs(polsRatio.getF(res.x_Fmax))),"DI(res.x_Fmax)",0,"0",eps);
	    if(polsRatio.getD2F(res.x_Fmax)>0) throw new RuntimeException("res.x_Fmax="+res.x_Fmax+" D2="+polsRatio.getD2F(res.x_Fmax));

	    System.err.println("ILocalizedStates: I.sH="+I.sH+" Imax="+this.res.Fmax+" I.sL="+I.sL+" Imin="+this.res.Fmin+" n="+n+" coefs.length="+coefs.length+" 1+(4n-6)="+(1+4*n-6)+" roots="+res.r.nroots+" real roots="+n_real_roots);

	}

    }

public static class SelectMinMaxFValuesOnDerivativeRoots {
    public final double Fmin,x_Fmin,Fmax,x_Fmax;
    public final double [] psi_Fmin,psi_Fmax;
    final PolynomialRootsFinder.PRoots r;
    SelectMinMaxFValuesOnDerivativeRoots(
	final int n,
	final double [] coefs,
	final double [] InvG,
	final java.util.function.ToDoubleFunction<double []>getF,
	final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	// for unit test
	final double eps){

	this.r=M.getXNodes(coefs);

	// we need to find min/max. treat complex roots as they were real
	double [] psiFmin=null,psiFmax=null;
	double Fmin=0,Fmax=0,xFmin=Double.NaN,xFmax=Double.NaN;
	for(int k=0;k<r.nroots;k++){
	    final double [] qk=M.B.getX0Moments(n,r.re_root[k]),
		psik=TMatr.getMatr_x_A(n,InvG,qk);
	    final double fk=getF.applyAsDouble(psik);
	    TMatr.scaleVectorInPlace(psik,1/Math.sqrt(TMatr.mulVV(qk,psik)));
	    if(psiFmin==null||fk<=Fmin){
		psiFmin=psik;
		Fmin=fk;
		xFmin=r.re_root[k];
	    }
	    if(psiFmax==null||fk>=Fmax){
		psiFmax=psik;
		Fmax=fk;
		xFmax=r.re_root[k];
	    }
	}
	this.Fmin=Fmin;
	this.x_Fmin=xFmin;
	this.psi_Fmin=psiFmin==null?new double [n]:psiFmin;
	this.Fmax=Fmax;
	this.x_Fmax=xFmax;
	this.psi_Fmax=psiFmax==null?new double [n]:psiFmax;	    
    }
	    
} // END OF class SelectMinMaxFValuesOnDerivativeRoots 
}
