package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;

/** Obtain a density matrix for the state {@code psi} of max I:
 * "since psi till now".
 * @version $Id: IandDM.java,v 1.10 2022/09/30 16:19:34 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class IandDM {

    /** Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}
     * solution in original basis \(Q_k\).
     */
    public final EVXData I;

    /** The eigenstate selected. The one corresponding
     * to maximal eigenvalue.
     */
    public final double []  psiM;

    public final double []
	pol_QQDensityMatrix,
	QQDensityMatrix,
	pol_QQDensityMatrix2,
	QQDensityMatrix2;

    /** Gram matrix inverse. A helper.
     */
    public final double [] IQQt;
    
    public IandDM(final int n,
	   // input moment matrices
	   final double [] QQt,
	   final double [] QQi,
	   // basis
	   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	   final java.util.function.Function<double [],double []> getPolynomialWIntegratedDt,
	   // for unit tests
	   final double eps){
	
	this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int ind_IH=n-1;// maximal eigenvalue
	this.psiM=I.getPsi(ind_IH);

	final double [] UNIT=TMatr.getUnitMatrix(n);
	// Gram matrix inverse can be obtained without solving a linear system.
	IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT);
	   
	// A mixed state with |QQDensityMatrix| density matrix obtained as "an interval since |psiM> till now".

	// As the calculations require multiplication operator -- perfrom them in Q-basis,
	// where multiplication is readily available.
	//
	// Integrate psiM^2 omega(x) dt: time since psiM spike
	// dmu=pol_QQDensityMatrix(x(t))*omega(t)dt
	pol_QQDensityMatrix=getPolynomialWIntegratedDt.apply(M.B.mult2Pol(psiM,psiM));
	
	// Construct the measure corresponding to pol_QQDensityMatrix polynomial
	// using Theorem 3 of https://arxiv.org/abs/1807.06007 : find the measure generating given polynomial,
	// Find measure moments, then create density matrix corresponding to this measure.
	QQDensityMatrix=M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,pol_QQDensityMatrix,IQQt));


	// try double integration
	pol_QQDensityMatrix2=getPolynomialWIntegratedDt.apply(pol_QQDensityMatrix);
	QQDensityMatrix2=M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,pol_QQDensityMatrix2,IQQt));
	
	if(eps>0){
	    // Check IQQt is correct
	    UnitTests.checkDoubleArrayRel1(TMatr.mulMM(n,QQt,IQQt),"QQt * IQQt",UNIT,"UNIT",eps);

	    // Check normalizing. Analytic integration expression (assuming omega(x0)=1).
	    UnitTests.checkDoubleRel(M.B.getD0(pol_QQDensityMatrix,I.x0),"\\int psiM^2(x) omega(x(t)) dt",1,"1",eps);
	    UnitTests.checkDoubleRel(M.B.getD0(pol_QQDensityMatrix2,I.x0),"pol2",TMatr.getSpurMM(n,IQQt,QQDensityMatrix),"spur(IQQt*QQDensityMatrix)",eps);
	    
	    // Check the polynomial
	    UnitTests.checkDoubleArrayRel1(pol_QQDensityMatrix,"pol_QQDensityMatrix",M.B.getKK(n,TMatr.mulMMM(n,IQQt,QQDensityMatrix,IQQt)),"Q(x) G^{-1}*DM*G^{-1} Q(x)",eps);
	    UnitTests.checkDoubleArrayRel1(pol_QQDensityMatrix2,"pol_QQDensityMatrix2",M.B.getKK(n,TMatr.mulMMM(n,IQQt,QQDensityMatrix2,IQQt)),"Q(x) G^{-1}*DM2*G^{-1} Q(x)",eps);

	    // Check orthogonality
	    EVXData.test_orthogonality("IandDM.I.ev",n,I.ev,QQi,QQt,eps);

	}
    }
}
