package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.PolynomialsRatio;
import com.polytechnik.utils.Sum2DerivativesCalculateable;

/**
 * The state of {@literal
\( \max I -V/T= \frac{\left<\psi|QQi|\psi\right>}{\left<\psi|QQt|\psi\right>}
-
\frac{\left<\psi|QQV|\psi\right>}{\left<\psi|QQT|\psi\right>}
 \)
}
among localized states
{@literal \( \left|\psi\right> = \frac{1}{\mathrm{norm}}G^{-1}Q(x) \) }.
* The problem is reduced to maximizing a ratio of two polynomials
* of the degree {@code 4n-4}. The derivative roots of this ratio
* correspond to the roots of a polynomial of the degree {@code 8n-9},
* a solution is then selected among all available roots of the derivative.
* This polynomial of the degree {@code 8n-9} has leading coefficient 0,
* thus actual order is {@code 8n-10}, the number of coefficients in polynomial is {@code 1+8n-10}.
* @see ILocalizedStates
 * @version $Id: ImVTLocalizedStates.java,v 1.24 2022/08/08 20:10:45 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public class ImVTLocalizedStates {

    public final ILocalizedStates.SelectMinMaxFValuesOnDerivativeRoots res;

    public ImVTLocalizedStates(
	final int n,

	// input moment matrices
	final double [] QQt,
	final double [] QQi,
	final double [] QQT,
	final double [] QQV,

	final double [] InvG,

	// basis operations
	final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,

	// for unit test
	final double eps){

	final double []
	    IPolNom=M.B.getKK(n,TMatr.mulMMM(n,InvG,QQi,InvG)),
	    IPolDenom=M.B.getKK(n,TMatr.mulMMM(n,InvG,QQt,InvG)),
	    VTPolNom=M.B.getKK(n,TMatr.mulMMM(n,InvG,QQV,InvG)),
	    VTPolDenom=M.B.getKK(n,TMatr.mulMMM(n,InvG,QQT,InvG)),
	    PolDenom=M.B.mult2Pol(IPolDenom,VTPolDenom),
	    PolNom=TMatr.sumAcB(
		M.B.mult2Pol(IPolNom,VTPolDenom),
		-1,
		M.B.mult2Pol(VTPolNom,IPolDenom));
	double [] coefs=TMatr.sumAcB(
	    M.B.mult2Pol(M.B.getDifferentiated(PolNom),PolDenom),
	    -1,
	    M.B.mult2Pol(M.B.getDifferentiated(PolDenom),PolNom));
	if(eps>0){
	    System.err.println("ImVTLocalizedStates:n="+n+" coefs="+java.util.Arrays.toString(coefs)+" coefs.length="+coefs.length);
	    UnitTests.checkDoubleRel(coefs[coefs.length-1],"coefs[coefs.length-1]",0,"0",eps);
	}

	// just remove highest order coefficient that is 0
	coefs=java.util.Arrays.copyOf(coefs,coefs.length-1);


	this.res=new ILocalizedStates.SelectMinMaxFValuesOnDerivativeRoots(
	    n,
	    coefs,
	    InvG,
	    psi->(TMatr.getAMatrA(psi,QQi,n)/TMatr.getAMatrA(psi,QQt,n)
		  -TMatr.getAMatrA(psi,QQV,n)/TMatr.getAMatrA(psi,QQT,n)
		  ),
	    M,
	    eps);

	if(eps>0){
	    final double epsDF=Math.max(eps,1e-1);
	    final var I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	    final var polsRatio=new Sum2DerivativesCalculateable(
		1,new PolynomialsRatio(IPolDenom,IPolNom,M.B),
		-1,new PolynomialsRatio(VTPolDenom,VTPolNom,M.B));
	    int n_real_roots=0;
	    for(int k=0;k<res.r.nroots;k++){
		if(Math.abs(res.r.im_root[k])<=0){
		    n_real_roots++;
		    final double x=res.r.re_root[k],
			scaleF=Math.max(I.sH,Math.abs(polsRatio.getF(x)));
		    // may give 1e-3 error
		    //UnitTests.checkDoubleRel(polsRatio.getDF(x)/scaleF,"DF",0,"0",epsDF);
		}
	    }
	    UnitTests.checkDoubleRel(polsRatio.getF(res.x_Fmin),"ImVT(res.x_Fmin)",res.Fmin,"res.Fmin",eps);
	    //UnitTests.checkDoubleRel(polsRatio.getDF(res.x_Fmin)/Math.max(I.sH,Math.abs(polsRatio.getF(res.x_Fmin))),"DImVT(res.x_Fmin)",0,"0",epsDF);
	    //if(polsRatio.getD2F(res.x_Fmin)<0) throw new RuntimeException("res.x_Fmin="+res.x_Fmin+" D2="+polsRatio.getD2F(res.x_Fmin));
	    UnitTests.checkDoubleRel(polsRatio.getF(res.x_Fmax),"ImVT(res.x_Fmax)",res.Fmax,"res.Fmax",eps);
	    //UnitTests.checkDoubleRel(polsRatio.getDF(res.x_Fmax)/Math.max(I.sH,Math.abs(polsRatio.getF(res.x_Fmax))),"DImVT(res.x_Fmax)",0,"0",epsDF);
	    //if(polsRatio.getD2F(res.x_Fmax)>0) throw new RuntimeException("res.x_Fmax="+res.x_Fmax+" D2="+polsRatio.getD2F(res.x_Fmax));

	    System.err.println("ImVTLocalizedStates: ImVTmax="+this.res.Fmax+" ImVTmin="+this.res.Fmin+" n="+n+" coefs.length="+coefs.length+" 1+(8n-10)="+(1+8*n-10)+" roots="+res.r.nroots+" real roots="+n_real_roots);

	}
    }
}
