package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.VolMatrix;

/**
 * A helper class to calculate commonly used matrices from volume-weighted sampled moments
 * using integration by parts.
 * @version $Id: MatricesFromPI.java,v 1.54 2022/09/25 08:54:49 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public class MatricesFromPI {

    /** A helper function to obtain \(|p|\) operator from \(|pI|\) as
     * \(|I^{-1/2}|pI|I^{-1/2}|\), crude and simple approach.
     * The main advantage of the approach is that 
     * the eigenvalues 
     * {@literal \(\left|p|\psi\right>=\lambda\left|\psi\right>\)}
     * of obtained matrix are the same as of the 
     * problem {@literal \(\left|pI|\psi\right>=\lambda\left|I|\psi\right>\)}.
     * The method is not a good one for derivatives. 
     * The {@link #getbQQIpowdpdtFromQQpi_withLastP} typically 
     * gives a much better result for derivatives.
     * @param bQQpi The matrix {@literal \( \left<\psi^{[j]}|pI|\psi^{[k]}\right> \)}
     * where {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}.
     * The matrix {@code QQpi}={@literal \( \left<Q_j|pI|Q_k\right>\)} in original basis \(Q_k\)
     * can be converted to {@literal \( \left|\psi^{[i]}\right> \)} -basis {@code bQQpi }using 
     * {@link EVXData#convertMatrFromQtoB_Basis 
         {@code EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev)}
       }.
     * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}
     * solution.
     * @return {@code bQQp} matrix.
     */
    public static double [] getbQQpFrombQQpi(final double [] bQQpi,final EVXData I){
	// operator |I^{-1/2}|pI|I^{-1/2}| (the flagDivide=true)
	return getbQQpIpowFrombQQp(bQQpi,I,true);
    }

    /** A helper function to obtain \(|pI|\) operator from \(|p|\) as
     * \(|I^{1/2}|p|I^{1/2}|\).
     * @param bQQp The matrix {@literal \( \left<\psi^{[j]}|p|\psi^{[k]}\right> \)}
     * where {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}.
     * The matrix {@code QQp}={@literal \( \left<Q_j|p|Q_k\right>\)} in original basis \(Q_k\)
     * can be converted to {@literal \( \left|\psi^{[i]}\right> \)} -basis {@code bQQp} using 
     * {@link EVXData#convertMatrFromQtoB_Basis 
         {@code EVXData.convertMatrFromQtoB_Basis(n,QQp,I.ev)}
       }.
     * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}
     * solution.
     * @return {@code bQQpi} matrix.
     */
    public static double [] getbQQpiFrombQQp(final double [] bQQp,final EVXData I){
	// operator |I^{1/2}|pI|I^{1/2}| (the flagDivide=false)
	return getbQQpIpowFrombQQp(bQQp,I,false);
    }
    
    /** Transform {@code bQQp} matrix either to \(|I^{-1/2}|p|I^{-1/2}|\) 
     * if {@code flagDivide=true} or to \(|I^{1/2}|p|I^{1/2}|\)
     * if {@code flagDivide=false}.
     * @param bQQp The matrix in I-basis.
     * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}
     * solution.
     * @param flagDivide If {@code true} construct {@code bQQp[j,k]/sqrt(I.ev[j]*I.ev[k])},
     * if {@code false} construct {@code bQQp[j,k]*sqrt(I.ev[j]*I.ev[k])}.
     */
    private static double [] getbQQpIpowFrombQQp(final double [] bQQp,final EVXData I,final boolean flagDivide){
	final int n=I.neq;
	final double [] bQQpIpow=bQQp.clone(),sqrtI=new double [n];
	for(int k=0;k<n;k++) {
	    final double diagIk=I.ev[k],s=sqrtI[k]=Math.sqrt(diagIk);
	    for(int l=0,ok=k*n;l<=k;l++){
		final double fact=(l==k?diagIk:s*sqrtI[l]),m;
		if(flagDivide){
		    m=bQQpIpow[ok+l]/fact;
		} else {
		    m=bQQpIpow[ok+l]*fact;
		}
		bQQpIpow[ok+l]=bQQpIpow[l*n+k]=m;
	    }
	}
	return bQQpIpow;
    }


/** Obtain regular {@literal \(\left<\psi^{[j]}\Big|\frac{dp}{dt}\Big|\psi^{[k]}\right> \)}
 * from {@literal \( \left<Q_j|p|Q_k\right>\)} using integration by parts.
 * The method can also be used for the calculation of {@literal \(\left<\psi^{[j]}\Big|\frac{dpI}{dt}\Big|\psi^{[k]}\right> \)}:
 * just put {@code QQpi} instead of {@code QQp} and {@code Ifuture*p_last} instead of {@code p_last}.
 * @param QQp The matrix {@literal \( \left<Q_j|p|Q_k\right>\)} in original basis.
 * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)} solution.
 * @param getEDPsi Time-variation of {@literal \( \left|\psi\right> \)},
 * corresponds to {@link CommonBasisOperations#getEDPsi} that is a one of
 * {@link WIntegratorLegendreShifted#getEDPsi},
 * {@link WIntegratorLaguerre#getEDPsi},
 * or {@link WIntegratorMonomials#getEDPsi}
 * depending on basis used.
 * @param QQt The matrix {@literal \( \left<Q_j|Q_k\right>\)} in original basis.
 * @param p_last Edge (at \(x_0\)) value of \( p \) to calculate \( dp/dt \), denoted \( P^{last} \).
 @return The matrix {@literal \(
P^{last}\psi^{[j]}(x_0)\psi^{[k]}(x_0)\omega(x_0)-
\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|p\Big|\psi^{[k]}\right>
-\left<\psi^{[j]}\Big|p\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}.
@see VolMatrix#getDiffMatrix
*/
public static double [] getbQQdpdtFromQQp_withLastP(final double [] QQp,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi,final double [] QQt,final double p_last){
    if(true){
	// a better method to calculate
	return getbQQdpdtIpowFromQQpi(TMatr.sumaAbB(p_last,QQt,-1,QQp),I,getEDPsi,(psiIndex,DpsiIndex)->1,false);
    } else {
	// same thing
	final int n=I.neq;
	final double [] bQQdpdt=getbQQdpdtIpowFromQQpi(QQp,I,getEDPsi,(psiIndex,DpsiIndex)->1,false);
	for(int j=0;j<n;j++){
	    for(int k=0;k<n;k++){
		final int o=j*n+k;
		// assume omega(x0)=1
		bQQdpdt[o]=p_last*I.psik_at_x0[j]*I.psik_at_x0[k]-bQQdpdt[o];
	    }
	}
	return bQQdpdt;
    }
}

/** Obtain {@literal \(\left<\psi^{[j]}\Big|\frac{dp}{dt}\Big|\psi^{[k]}\right> \)}
 * from {@literal \( \left<Q_j|pI|Q_k\right>\)},
 * similar to {@link VolMatrix#getDiffMatrix} 
 * but the original moments are calculated with \( I\) weigth.
 * The approximation is
 * {@literal
\begin{align}
\left<\psi^{[j]}\Big|\frac{dp}{dt}\Big|\psi^{[k]}\right>&\approx
p(x_0)\psi^{[j]}(x_0)\psi^{[k]}(x_0)\omega(x_0)-
\frac{1}{\lambda^{[k]}}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
-\frac{1}{\lambda^{[j]}}\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\end{align}
}
* @param QQpi The matrix {@literal \( \left<Q_j|pI|Q_k\right>\)} in original basis.
* @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)} solution.
* @param getEDPsi Time-variation of {@literal \( \left|\psi\right> \)}
* with weight; corresponds to {@link CommonBasisOperations#getEDPsi}
* and {@link VolMatrix#getDPsi}.

@return The matrix {@literal \(
\frac{1}{\lambda^{[k]}}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
+\frac{1}{\lambda^{[j]}}\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}. Out of integral term is not included and the sign is the same as in 
{@link VolMatrix#getDiffMatrix}.
@see VolMatrix#getDiffMatrix
    */
private static double [] getbQQdpdtFromQQpi(final double [] QQpi,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi){
    return getbQQdpdtIpowFromQQpi(QQpi,I,getEDPsi,(psiIndex,DpsiIndex)->I.ev[psiIndex],true);
}

/** Obtain {@literal \(\left<\psi^{[j]}\Big|I\frac{dp}{dt}\Big|\psi^{[k]}\right> \)}
 * from {@literal \( \left<Q_j|pI|Q_k\right>\)}.
 @return The matrix {@literal \(
\frac{\sqrt{\lambda^{[j]}}}{\sqrt{\lambda^{[k]}}}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
+\frac{\sqrt{\lambda^{[k]}}}{\sqrt{\lambda^{[j]}}}\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}. Out of integral term is not included.
*/
private static double [] getbQQIdpdtFromQQpi(final double [] QQpi,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi){
    return getbQQdpdtIpowFromQQpi(QQpi,I,getEDPsi,(psiIndex,DpsiIndex)->Math.sqrt(I.ev[DpsiIndex]/I.ev[psiIndex]),false);
}

/** An attempt to obtain {@literal \(\left<\psi^{[j]}\Big|I^2\frac{dp}{dt}\Big|\psi^{[k]}\right> \)}
 * from {@literal \( \left<Q_j|pI|Q_k\right>\)}.
 @return The matrix {@literal \(
\lambda^{[j]}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
+\lambda^{[k]}\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}. Out of integral term is not included.
*/
private static double [] getbQQI2dpdtFromQQpi(final double [] QQpi,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi){
    return getbQQdpdtIpowFromQQpi(QQpi,I,getEDPsi,(psiIndex,DpsiIndex)->I.ev[DpsiIndex],false);
}


/** An attempt to obtain {@literal \(\left<\psi^{[j]}\Big|\frac{1}{I}\frac{dp}{dt} \Big|\psi^{[k]}\right> \)}
 * from {@literal \( \left<Q_j|pI|Q_k\right>\)}.
 @return The matrix {@literal \(
\frac{1}{{\lambda^{[k]}}^{3/2}\sqrt{\lambda^{[j]}}}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
+\frac{1}{{\lambda^{[j]}}^{3/2}\sqrt{\lambda^{[k]}}}\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}. Out of integral term is not included.
*/
private static double [] getbQQdpdtDivIFromQQpi(final double [] QQpi,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi){
    return getbQQdpdtIpowFromQQpi(QQpi,I,getEDPsi,(psiIndex,DpsiIndex)->I.ev[psiIndex]*Math.sqrt(I.ev[DpsiIndex]*I.ev[psiIndex]),true);
}

/** An attempt to obtain {@literal \(\left<\psi^{[j]}\Big|\frac{1}{I^2}\frac{dp}{dt}\Big|\psi^{[k]}\right> \)}
 * from {@literal \( \left<Q_j|pI|Q_k\right>\)}.
 @return The matrix {@literal \(
\frac{1}{{\lambda^{[k]}}^{2}\lambda^{[j]}}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
+\frac{1}{{\lambda^{[j]}}^{2}\lambda^{[k]}}\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}. Out of integral term is not included.
*/
private static double [] getbQQdpdtDivI2FromQQpi(final double [] QQpi,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi){
    return getbQQdpdtIpowFromQQpi(QQpi,I,getEDPsi,(psiIndex,DpsiIndex)->Math.pow(I.ev[psiIndex],2)*I.ev[DpsiIndex],true);
}
    
@FunctionalInterface private interface GetPsiDpsiFactor {
    double get(int psiIndex,int DpsiIndex);
}

private static double [] getbQQdpdtIpowFromQQpi(final double [] QQpi,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi,final GetPsiDpsiFactor factor,final boolean flagDivide){
	final int n=I.neq;
	if(QQpi.length!=n*n) throw new IllegalArgumentException();
	final double [][] psi=new double [n][],Dpsi=new double [n][];
	final double [] bQQdpdt=new double [n*n];
	for(int j=0;j<n;j++){
	    psi[j]=I.getPsi(j);
	    Dpsi[j]=getEDPsi.apply(psi[j]);
	}
	for(int j=0;j<n;j++){
	    for(int k=0;k<n;k++){
		final double factor_psikDpsij=factor.get(k,j),factor_psijDpsik=factor.get(j,k);
		if(flagDivide){
		    bQQdpdt[j*n+k]=TMatr.getAMatrB(Dpsi[j],psi[k],QQpi,n)/factor_psikDpsij+TMatr.getAMatrB(psi[j],Dpsi[k],QQpi,n)/factor_psijDpsik;
		} else {
		    bQQdpdt[j*n+k]=TMatr.getAMatrB(Dpsi[j],psi[k],QQpi,n)*factor_psikDpsij+TMatr.getAMatrB(psi[j],Dpsi[k],QQpi,n)*factor_psijDpsik;
		}
	    }
	}
	return bQQdpdt;
    }


    /** A helper method to obtain actual
     * {@literal \( \left<\psi^{[j]}\Big|I^{\mathrm{Ipow}}\frac{dp}{dt}\Big|\psi^{[j]}\right> \)},
     * including out of integral term,
     * using integration by parts.
     * The method adjusts last value to 0 to compensate out of integral term, then calls 
     * {@link #getbQQdpdtIpowFromQQpi}.
     * @param Ipow the power of \( I \) in the result.
     * @param QQpi The matrix {@literal \( \left<Q_j|pI|Q_k\right>\)} in original basis.
     * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)} solution.
     * @param getEDPsi Time-variation of {@literal \( \left|\psi\right> \)},
     * corresponds to {@link CommonBasisOperations#getEDPsi} that is a one of
     * {@link WIntegratorLegendreShifted#getEDPsi},
     * {@link WIntegratorLaguerre#getEDPsi},
     * or {@link WIntegratorMonomials#getEDPsi}
     * depending on basis used.
     * @param QQi The matrix {@literal \( \left<Q_j|I|Q_k\right>\)} in original basis.
     * @param p_last The last value of p.
     * @return The matrix {@literal \(
\frac{\left(\lambda^{[j]}\lambda^{[k]}\right)^{\mathrm{Ipow}/2}}{\lambda^{[k]}}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|(\mathrm{p\_last}-p)I\Big|\psi^{[k]}\right>
+\frac{\left(\lambda^{[j]}\lambda^{[k]}\right)^{\mathrm{Ipow}/2}}{\lambda^{[j]}}\left<\psi^{[j]}\Big|(\mathrm{p\_last}-p)I\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}.
    */
    public static double [] getbQQIpowdpdtFromQQpi_withLastP(
	final double Ipow,
	final double [] QQpi,
	final EVXData I,
	final java.util.function.Function<double [],double []> getEDPsi,
	final double [] QQi,
	final double p_last){
	// add last price to compensate out of integral term
	final double [] mQQpi_PLastI=TMatr.sumaAbB(-1,QQpi,p_last,QQi);
	// optimization
	if(Ipow==0) return getbQQdpdtFromQQpi(mQQpi_PLastI,I,getEDPsi);
	else if(Ipow==1) return getbQQIdpdtFromQQpi(mQQpi_PLastI,I,getEDPsi);
	else if(Ipow==2) return getbQQI2dpdtFromQQpi(mQQpi_PLastI,I,getEDPsi);
	else if(Ipow== -1) return getbQQdpdtDivIFromQQpi(mQQpi_PLastI,I,getEDPsi);
	else if(Ipow== -2) return getbQQdpdtDivI2FromQQpi(mQQpi_PLastI,I,getEDPsi);
	// general case
	return getbQQdpdtIpowFromQQpi(mQQpi_PLastI,I,getEDPsi,(psiIndex,DpsiIndex)->Math.pow(I.ev[psiIndex]*I.ev[DpsiIndex],Ipow/2)/I.ev[psiIndex],false);		    
    }

    /** A helper method to obtain something like
     * {@literal \( \left<\psi^{[j]}\Big|p\frac{dI}{dt}\Big|\psi^{[k]}\right> \)},
     * exact meaning is not clear: when \(p=const\) we get an exact \( dI/dt \),
     * and when \( I=const \) we get all zero.
     * Typical  usage: call with {@code QQpi=QQt} to obtain 
     * {@literal \(\left<\psi^{[j]}\Big|\frac{dI}{dt}\Big|\psi^{[k]}\right>\)}.
     * @param QQp The matrix {@literal \( \left<Q_j|p|Q_k\right>\)} in original basis.
     * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)} solution.
     * @param getEDPsi Time-variation of {@literal \( \left|\psi\right> \)}.
     * @param Ifuture Future value of \( I \) to calculate \( dI/dt \), denoted \(I^F\), 
     * typically the maximal eigenvalue of {@code I}.
     * @return The matrix {@literal \(
\left<\psi^{[j]}\Big|p\frac{dI}{dt}\Big|\psi^{[k]}\right>=
\left(I^F-\lambda^{[k]}\right)
\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|p\Big|\psi^{[k]}\right>
+
\left(I^F-\lambda^{[j]}\right)
\left<\psi^{[j]}\Big|p\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}.
    */
    public static double [] getbQQpdIdtFromQQp(
	final double [] QQp,
	final EVXData I,
	final java.util.function.Function<double [],double []> getEDPsi,
	final double Ifuture){
	
	return getbQQpdIdtFromQQp(QQp,I,getEDPsi,Ifuture,false);
    }


    /** A helper method to obtain something like
     * {@literal \( \left<\psi^{[j]}\Big|p\frac{dI}{dt}\Big|\psi^{[k]}\right> \)},
     * exact meaning is not clear: when \(p=const\) we get an exact \( dI/dt \),
     * and when \( I=const \) we get all zero.
     * Typical  usage: call with {@code QQpi=QQi} to obtain 
     * {@literal \(\left<\psi^{[j]}\Big|\frac{dI}{dt}\Big|\psi^{[k]}\right>\)}.
     * @param QQpi The matrix {@literal \( \left<Q_j|pI|Q_k\right>\)} in original basis.
     * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)} solution.
     * @param getEDPsi Time-variation of {@literal \( \left|\psi\right> \)}.
     * @param Ifuture Future value of \( I \) to calculate \( dI/dt \), denoted \(I^F\), 
     * typically the maximal eigenvalue of {@code I}.
     * @return The matrix {@literal \(
\left<\psi^{[j]}\Big|p\frac{dI}{dt}\Big|\psi^{[k]}\right>=
\left(\frac{I^F}{\lambda^{[k]}}-1\right)
\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
+
\left(\frac{I^F}{\lambda^{[j]}}-1\right)
\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}.
    */
    public static double [] getbQQpdIdtFromQQpi(
	final double [] QQpi,
	final EVXData I,
	final java.util.function.Function<double [],double []> getEDPsi,
	final double Ifuture){
	
	return getbQQpdIdtFromQQp(QQpi,I,getEDPsi,Ifuture,true);
    }

    private static double [] getbQQpdIdtFromQQp(
	final double [] QQp,
	final EVXData I,
	final java.util.function.Function<double [],double []> getEDPsi,
	final double Ifuture,
	final boolean flagDivI){

	final int n=I.neq;
	if(QQp.length!=n*n) throw new IllegalArgumentException();
	final double [][] psi=new double [n][],Dpsi=new double [n][];
	final double [] bQQdIdt=new double [n*n];
	for(int j=0;j<n;j++){
	    psi[j]=I.getPsi(j);
	    Dpsi[j]=getEDPsi.apply(psi[j]);
	}
	for(int j=0;j<n;j++){
	    for(int k=0;k<n;k++){
		final double factor_psikDpsij,factor_psijDpsik;
		if(flagDivI){
		    factor_psikDpsij=Ifuture/I.ev[k]-1;
		    factor_psijDpsik=Ifuture/I.ev[j]-1;
		} else {
		    factor_psikDpsij=Ifuture-I.ev[k];
		    factor_psijDpsik=Ifuture-I.ev[j];
		}
		bQQdIdt[j*n+k]=TMatr.getAMatrB(Dpsi[j],psi[k],QQp,n)*factor_psikDpsij+TMatr.getAMatrB(psi[j],Dpsi[k],QQp,n)*factor_psijDpsik;
	    }
	}
	return bQQdIdt;
    }


/** An attempt to obtain {@literal \(\left<\psi^{[j]}\Big|\frac{d}{dt}\frac{p}{I}\Big|\psi^{[k]}\right> \)}
 * from {@literal \( \left<Q_j|pI|Q_k\right>\)} using integration by parts.
 * @param QQpi The matrix {@literal \( \left<Q_j|pI|Q_k\right>\)} in original basis.
 * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)} solution.
 * @param getEDPsi Time-variation of {@literal \( \left|\psi\right> \)},
 * corresponds to {@link CommonBasisOperations#getEDPsi} that is a one of
 * {@link WIntegratorLegendreShifted#getEDPsi},
 * {@link WIntegratorLaguerre#getEDPsi},
 * or {@link WIntegratorMonomials#getEDPsi}
 * depending on basis used.
 * @param p_last Edge (at \(x_0\)) value of \(p\), denoted \( P^{last} \).
 * @param Ifuture Future value of \( I \) to calculate \( dI/dt \), denoted \(I^F\).
 @return The matrix {@literal \(
\frac{P^{last}}{I^F}\psi^{[j]}(x_0)\psi^{[k]}(x_0)\omega(x_0)-
\frac{1}{{\lambda^{[k]}}^2}\left<\mathrm{getEDPsi}(\psi^{[j]})\Big|pI\Big|\psi^{[k]}\right>
-\frac{1}{{\lambda^{[j]}}^2}\left<\psi^{[j]}\Big|pI\Big|\mathrm{getEDPsi}(\psi^{[k]})\right>
\)}.
@see VolMatrix#getDiffMatrix
*/
    public static double [] getbQQDtpDivIFromQQpi_withLastP(final double [] QQpi,final EVXData I,final java.util.function.Function<double [],double []> getEDPsi,final double p_last,final double Ifuture){
	
	final int n=I.neq;
	final double [] bQQdpdivI=getbQQdpdtIpowFromQQpi(QQpi,I,getEDPsi,(psiIndex,DpsiIndex)->Math.pow(I.ev[psiIndex],2),true);
	for(int j=0;j<n;j++){
	    for(int k=0;k<n;k++){
		final int o=j*n+k;
		// assume omega(x0)=1
		bQQdpdivI[o]=(p_last/Ifuture)*I.psik_at_x0[j]*I.psik_at_x0[k]-bQQdpdivI[o];
	    }
	}
	return bQQdpdivI;
    }
}
