package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVSolver;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;

/** A class to split dI/dt like operator on positive and negative subspaces.
 * Simplified version.
 * @version $Id: SplitdIdt.java,v 1.13 2022/09/26 15:22:53 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class SplitdIdt {

    public static class SplitData {
	public final double f_spurM,f_posdI_spurM,f_negdI_spurM;
	SplitData(final double f_spurM,
		  final double f_posdI_spurM,
		  final double f_negdI_spurM){
	    this.f_spurM=f_spurM;
	    this.f_posdI_spurM=f_posdI_spurM;
	    this.f_negdI_spurM=f_negdI_spurM;
	}

	/** For unit test.
	 */
	public static void match(final SplitData a,final SplitData b,final double eps){
	    UnitTests.checkDoubleRel1(a.f_spurM,"a.f_spurM",b.f_spurM,"b.f_spurM",eps);
	    UnitTests.checkDoubleRel1(a.f_posdI_spurM,"a.f_posdI_spurM",b.f_posdI_spurM,"b.f_posdI_spurM",eps);
	    UnitTests.checkDoubleRel1(a.f_negdI_spurM,"a.f_negdI_spurM",b.f_negdI_spurM,"b.f_negdI_spurM",eps);
	}
    }

    private final int n;
    private final double [] bQQ_selPosdIdt,bQQ_selNegdIdt,
	ev_bQQdIdt,bbQQ_selPosdIdt,bbQQ_selNegdIdt;
    private final double eps;
    
    public SplitdIdt(final int n,
		     final double [] bQQdIdt,
		     final EVSolver EV,
		     // for unit tests
		     final double eps){
	this.n=n;
	this.eps=eps;

	ev_bQQdIdt=EV.getEV(n,bQQdIdt,true);
	bbQQ_selPosdIdt=new double [n*n];
	bbQQ_selNegdIdt=new double [n*n];
	// Set projection operators in |dI/dt|psi>=lambda|psi> basis.
	for(int k=0;k<n;k++){
	    final int diagk=k*n+k;
	    if(ev_bQQdIdt[k]>0){
		bbQQ_selPosdIdt[diagk]=1;
	    } else if(ev_bQQdIdt[k]<=0){
		bbQQ_selNegdIdt[diagk]=1;
	    }
	}
	final double [] UNIT=TMatr.getUnitMatrix(n);
	// convert back to b-basis
	this.bQQ_selPosdIdt=EVXData.convertMatrFromBtoQ_Basis(n,bbQQ_selPosdIdt,ev_bQQdIdt,UNIT);
	this.bQQ_selNegdIdt=EVXData.convertMatrFromBtoQ_Basis(n,bbQQ_selNegdIdt,ev_bQQdIdt,UNIT);

	if(eps>0){	    
	    System.err.println("SplitdIdt: test OK EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_bQQdIdt,n))+" n="+n+" eps="+eps);
	}			   
    }


    public SplitData getSplitted(final double [] bQQf,final double [] bQQDensityMatrix){
	final var s=new SplitData(
	    TMatr.getSpurMM(n,bQQf,bQQDensityMatrix),
	    TMatr.getSpurMMM(n,bQQf,bQQ_selPosdIdt,bQQDensityMatrix),
	    TMatr.getSpurMMM(n,bQQf,bQQ_selNegdIdt,bQQDensityMatrix)
				  );
	if(eps>0){
	    final double T_M=TMatr.getSpurM(n,bQQDensityMatrix);
	    UnitTests.checkDoubleRel1(s.f_spurM,"s.f_spurM",s.f_posdI_spurM+s.f_negdI_spurM,"s.f_posdI_spurM+s.f_negdI_spurM",eps);
	    final double []
		bbQQDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,bQQDensityMatrix,ev_bQQdIdt),
		bbQQf=EVXData.convertMatrFromQtoB_Basis(n,bQQf,ev_bQQdIdt);

	    UnitTests.checkDoubleRel1(TMatr.getSpurMM(n,bbQQf,bbQQDensityMatrix),"spur(n,bbQQf*bbQQDensityMatrix)",s.f_spurM,"s.f_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurM(n,bbQQDensityMatrix),"spur(bbQQDensityMatrix)",T_M,"T_M",eps);
	    
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQf,bbQQ_selPosdIdt,bbQQDensityMatrix),"spur(bbQQf*bbQQ_selPosdIdt*bbQQDensityMatrix)",s.f_posdI_spurM,"s.f_posdI_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQf,bbQQ_selNegdIdt,bbQQDensityMatrix),"spur(bbQQf*bbQQ_selNegdIdt*bbQQDensityMatrix)",s.f_negdI_spurM,"s.f_negdI_spurM",eps);

	    System.err.println("SplitdIdt:getSplitted: basis invariance OK n="+n+" eps="+eps);
	}
	return s;
    }
   
}
