package com.polytechnik.freemoney;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
/** A simple accumulator with integration.
 * Subclasses contain recurrent code for fast calculation of the moments.
 * Basis recurrence relations used in 
 * {@link WIntegratorMonomials} correspond to {@link com.polytechnik.trading.QVMData},
 * {@link WIntegratorLaguerre} correspond to {@link com.polytechnik.trading.QVMDataL}, and
 * {@link WIntegratorLegendreShifted} correspond to {@link com.polytechnik.trading.QVMDataP}.
 *  @version $Id: WIntegrator.java,v 1.32 2022/09/03 19:02:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class WIntegrator {

    /** The time the moments are calculated on.
     */
    public long currenttime_previous=Long.MIN_VALUE;
    /** The number of moments {@literal <fQk>} to calculate, k=0 &hellip; ntm-1.
     */
    public final int ntm;
    /** Timeseries number.
     */
    public final int nobservables;
    /** Time scale of integration measure.
     */
    public final double atau;
    /** Integration moments. The integration_moments[i][k]
     * is the Fdt[i]--th moments {@literal <F[i]Qk>}.
     */
    public final double [][] integration_moments;
    /** The sum[i] is the sum of all posted Fdt[i].
     */
    public final double [] sum;
    /**
     * @param ntm The number of moments to caclucate.
     * @param atau Time scale.
     * @param nobservables The number of observables.
     */
    public WIntegrator(final int ntm,final double atau,final int nobservables){
	this.ntm=ntm;
	this.nobservables=nobservables;
	this.atau=atau;
	this.integration_moments=new double [nobservables][ntm];
	this.sum=new double [nobservables];
    }

    /** Used internally to calculate basis function argument.
     */
    public abstract double getXFromTime(final long currenttime,final long t);

    /** The method shifts basis functions offset as the time progresses.
     */
    public abstract void updateSavedMomentsToNewTimeReference(final long currenttime);

    /** Add a single observation at time t to all stored series.
     * No basis time-shift is performed, the {@link updateSavedMomentsToNewTimeReference}
     * should be called separately.
     * @param currenttime Current time.
     * @param t Event time.
     * @param Fdt All observables, array dimension must be equal to the {@link #nobservables}.
     */
    public void addSingleObservation(final long currenttime,long t,final double [] Fdt){
	if(Fdt.length!=nobservables) throw new IllegalArgumentException();
	addSingleObservationToSomeObservables(currenttime,t,Fdt,0);
    }

    /** Add a single observation at time t to a range of stored series.
     * No basis time-shift is performed, the {@link updateSavedMomentsToNewTimeReference}
     * should be called separately.
     * @param currenttime Current time.
     * @param t Event time.
     * @param Fdt The observables, to be added starting with staringSerieIndex (inclusive).
     * @param staringSerieIndex The serie index, the data will be added to {@link #integration_moments}[staringSerieIndex]..{@link #integration_moments}[staringSerieIndex+Fdt.length-1].
     */
    public abstract void addSingleObservationToSomeObservables(final long currenttime,long t,final double [] Fdt,final int staringSerieIndex);
}
