package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;

/** A class to store Eigenproblem-adjusted solution.
 * @version $Id: AdjustedStateToUnitaryWithEigenproblem.java,v 1.58 2024/04/19 21:40:13 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class AdjustedStateToUnitaryWithEigenproblem implements UAdjustment {
    final ApplyFunToGramMatrix gram;

    /** A criteria of all Gram eigenvalues equals to 1: \( \sum 1/\lambda^{[i]}_G \)
     * If all are equal to one -- the values is {@code nC} (the minimal), otherwise it is greater.
     */
    final double sum1Lam;

    /** A matrix such that u_adj  = T u_orig.
     * Would it be better to make it {@code lazy val} style?
     */
    final double [] GramSqM05;
    AdjustedStateToUnitaryWithEigenproblem(final ApplyFunToGramMatrix gram){
	this.gram=gram;
	this.GramSqM05=gram.applyFun((ind,lam)->1/Math.sqrt(lam));
	double sum=0;
	for(int i=0;i<gram.nC;i++){sum+= 1/gram.ev_m[i];}
	this.sum1Lam=sum;
    }
    @Override public double [] getUOrig(){return gram.u_orig;}
    @Override public double [] getUAdj(){return TMatr.mul_MATrans_x_MB(gram.nC,GramSqM05,gram.nC,gram.u_orig,gram.nX);}

    /** For unit test. */
    @Override public AdjustedStateToUnitaryWithSVD getSVDForTesting(){
	return new AdjustedStateToUnitaryWithSVD(
	    KGOSolutionVectorXVectorF.EV.getSVD(gram.nC,gram.nX,getUOrig()),
	    getUOrig(),
	    KGOSolutionVectorXVectorF.EV.getSVD(gram.nC,gram.nX,getUAdj()),
	    getUAdj());
    }


     // ========== !!! BELOW ARE INTERNAL FUNCTIONS AND UNIT TESTS !!! ==============

     /**
     * To remove along with {@link KGOIterationalMultipleTransforms}
     * and {@link KGOIterationalGEVAdj_Q_denominator}.
     */
    static double [] convertSKtoNewBasis(final int nC,final int nX,final double [] SK, final double [] to_new_basis){
	final int n=nC*nX;
	if(SK.length!=n*n || to_new_basis.length!=nC*nC) throw new IllegalArgumentException();
	final double [] SKnew=new double [n*n];
	for(int s=0;s<nC;s++){
	for(int sp=0;sp<nC;sp++){
	for(int k=0;k<nX;k++){
	for(int kp=0;kp<nX;kp++){
	    double sum=0;
	    for(int j=0;j<nC;j++){
	    for(int jp=0;jp<nC;jp++){
		sum+=to_new_basis[s*nC+j]*SK[(j*nX+k)*n+(jp*nX+kp)]*to_new_basis[sp*nC+jp];
	    }}
	    SKnew[(s*nX+k)*n+(sp*nX+kp)]=sum;
	}}}}
	return SKnew;
    }

    /** Create SK matrix such that
     * SK[j,k,jp,kp]=Gram[j,jp] delta_{k kp}.
     */
    static double [] createSKFromGram(final int nC,final int nX,final double [] Gram){
	final int n=nC*nX;
	if(Gram.length!=nC*nC) throw new IllegalArgumentException();
	final double [] SKGram=new double [n*n];
	for(int j=0;j<nC;j++){
	for(int jp=0;jp<nC;jp++){
	    final double fact=Gram[j*nC+jp];
	    for(int k=0;k<nX;k++){
		SKGram[(j*nX+k)*n+(jp*nX+k)]=fact;
	    }
	}}
	return SKGram;
    }

    /**
     * For unit test.
     */
    static double [] getSKasMatrixInUState(final int nC,final int nX,final double [] SK, final double [] u){
	final int n=nC*nX;
	if(SK.length!=n*n || u.length!=n) throw new IllegalArgumentException();
	final double [] SKinUstate=new double [nC*nC];
	for(int s=0;s<nC;s++){
	for(int sp=0;sp<nC;sp++){
	    double sum=0;
	    for(int k=0;k<nX;k++){
	    for(int kp=0;kp<nX;kp++){
		sum+=u[s*nX+k]*SK[(s*nX+k)*n+(sp*nX+kp)]*u[sp*nX+kp];
	    }}
	    SKinUstate[s*nC+sp]=sum;
	}}
	return SKinUstate;
    }

}
