package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.DataReadObservationVectorXVectorF;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;

/** A demo to recover a mapping from \( \mathbf{x} \to \mahtbf{f} \).
 * This is a demo. No quality regularization!
 * If any Gram matrix is degenerated --
 * the code fails since no quality regularization is used!
 * @version $Id: DemoRecoverMapping.java,v 1.8 2024/05/12 10:39:02 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DemoRecoverMapping {

    /** To turn on a selftest. */
    static final double eps=1e-7;


static class Res {
    final double [] u;
    final int nF,nX;
    Res(final double [] u,
	final int nF,
	final int nX){
	this.u=u;
	this.nF=nF;
	this.nX=nX;
    }
}

static Res getUFromConfig(final KGOConfig c) throws java.io.IOException {
    if(c.validateOptions().size()>0) throw new IllegalArgumentException();

    if(c.max_multiindex_x>=0 || c.max_multiindex_f>=0) throw new IllegalArgumentException("multiindex is not supported");
    // ignore c.approximation and c.SKtype
    //if(c.approximation!=KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalSubspaceLinearConstraints) throw new IllegalArgumentException("approximation "+c.approximation+" is not supported");


    // read sample
    final var sampV=DataReadObservationVectorXVectorF.getObservations(
	c.data_file_to_build_model_from,
	c.columns_number,
	c.getXIndexesSequence(),
	c.getFIndexesSequence(),
	c.indcol_weight,
	c.indcol_label);

    // a very crude way to calculate. no proper regularization, nothing.
    final double []
	Gx=sampV.getFun1Fun2Averaged(o->o.x,o->o.x,o->1),
	Gf=sampV.getFun1Fun2Averaged(o->o.f,o->o.f,o->1);
    // crude regularization
    final int nX=sampV.xHeader.size(),nF=sampV.fHeader.size();
    // a regularization is non-unique. Note! The result does not depend on regularization used.
    final double [] Rx,RxInv,Rf,RfInv;
    if(true){
	// trivial way
	final ApplyFunToMatrix aGx=new ApplyFunToMatrix(nX,Gx),aGf=new ApplyFunToMatrix(nF,Gf);
	Rx=aGx.applyFun((i,lam)->Math.sqrt(1/lam));
	RxInv=aGx.applyFun((i,lam)->Math.sqrt(lam));
	Rf=aGf.applyFun((i,lam)->Math.sqrt(1/lam));
	RfInv=aGf.applyFun((i,lam)->Math.sqrt(lam));
    } else {
	// Gram Schmidt
	final var rx=new GramSchmidtTest(nX,Gx,eps);
	Rx=rx.basis;
	RxInv=rx.basisInv;
	final var rf=new GramSchmidtTest(nF,Gf,eps);
	Rf=rf.basis;
	RfInv=rf.basisInv;
    }
    if(eps>0){
	UnitTests.checkDoubleArrayRel1(TMatr.mulMM(nX,Rx,RxInv),"Rx*RxInv",TMatr.getUnitMatrix(nX),"I",eps);
	UnitTests.checkDoubleArrayRel1(TMatr.mulMM(nF,Rf,RfInv),"Rf*RfInv",TMatr.getUnitMatrix(nF),"I",eps);
	UnitTests.checkDoubleArrayRel1(
	    sampV.getFun1Fun2Averaged(o->TMatr.getMatr_x_A(nX,Rx,o.x),o->TMatr.getMatr_x_A(nX,Rx,o.x),o->1),
	    "Rx Rx",TMatr.getUnitMatrix(nX),"I",eps);
	UnitTests.checkDoubleArrayRel1(
	    sampV.getFun1Fun2Averaged(o->TMatr.getMatr_x_A(nF,Rf,o.f),o->TMatr.getMatr_x_A(nF,Rf,o.f),o->1),
	    "Rf Rf",TMatr.getUnitMatrix(nF),"I",eps);
    }

    // slow but simple calculation
    final double [] Sfxfx=sampV.getFun1Fun2Averaged(
	o->TMatr.dyadicMatrix(TMatr.getMatr_x_A(nF,Rf,o.f),TMatr.getMatr_x_A(nX,Rx,o.x)),
	o->TMatr.dyadicMatrix(TMatr.getMatr_x_A(nF,Rf,o.f),TMatr.getMatr_x_A(nX,Rx,o.x)),
	o->1);
    // find u in regularized basis
    final var kgo=new KGOIterationalSubspaceLinearConstraints(nF,nX,Sfxfx,0*eps);
    final double [] uRfRx=kgo.aep.getUAdj();
    // convert to original basis
    final double [] u=
	TMatr.mul_MA_x_MB(nF,
			  TMatr.mul_MA_x_MB(nF,RfInv,nF,uRfRx,nX),
			  nX,
			  Rx,
			  nX);
    if(eps>0){
	UnitTests.checkDoubleArrayRel1(
	    Gf,"Gf",
	    EVXData.convertMatrFromQtoB_Basis(nX,Gx,u,nF,0),"u Gx uT",eps);
    }

    //System.err.println("u="+java.util.Arrays.toString(u));
    return new Res(u,nF,nX);
}

public static void main(String [] args) throws java.io.IOException {
    final var c=new KGOConfig(args);
    System.err.println("description\n"+c.getDescription());
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }
    final var res=getUFromConfig(c);
    // print u
    System.out.println("Recovered u:");
    for(int i=0;i<res.nF;i++){
	final var v=EVXData.getPsiFromBasis(i,res.u,res.nX,0);
	System.out.println(java.util.Arrays.toString(v));
    }
}


    /** For unit tests. A very crude. Do not use in production! */
    static class GramSchmidtTest {
	final double [] basis;
	final int n;
	final double [] basisInv;

	GramSchmidtTest(int n,double [] Gram,final double eps){
	    this.n=n;
	    if(Gram.length!=n*n) throw new IllegalArgumentException();
	    final double [][] basis=new double [n][n];
	    for(int j=0;j<n;j++) basis[j][j]=1;

	    for(int j=0;j<n;j++)
	    {
		// BEGIN find pivot
		int jSel=j;double n2Sel=0;
		for(int jn=j;jn<n;jn++){
		    final double l2=TMatr.getAMatrA(basis[jn],Gram,n);
		    if(l2>=n2Sel){
			n2Sel=l2;
			jSel=jn;
		    }
		}
		// END find pivot
		// swap j and jSel
		final double [] bprev=basis[jSel];
		basis[jSel]=basis[j];
		basis[j]=bprev;
		TMatr.scaleVectorInPlace(bprev,1/Math.sqrt(n2Sel));
		// remove bprev from basis
		for(int i=j+1;i<n;i++){
		    TMatr.sumP1(basis[i],bprev,-TMatr.getAMatrB(basis[i],bprev,Gram,n));
		}
	    }
	    this.basis=new double [n*n];
	    for(int i=0;i<n;i++){
		System.arraycopy(basis[i],0,this.basis,n*i,n);
	    }
	    this.basisInv=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,this.basis);

	    if(eps>0){
		for(int i=0;i<n;i++){
		    for(int j=0;j<n;j++){
			UnitTests.checkDoubleRel1(TMatr.getAMatrB(basis[i],basis[j],Gram,n),"<i|G|j>",
						  i==j?1:0,i==j?"1":"0",eps);
		    }
		}
		System.err.println("GramSchmidtTest: orthogonality OK n="+n+" eps="+eps);
	    }
	}
    }// END of class GramSchmidtTest

}
