package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.DataReadObservationVectorXF;
import com.polytechnik.utils.DataReadObservationVectorXVectorF;
import com.polytechnik.utils.ObservationVectorXVectorF;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.kgo.DemoRecoverMapping.GramSchmidtTest;
import static com.polytechnik.utils.RNConfig.getVal;

/** A demo to recover unitary matrix \(\mathcal{U}\) from a sequence:
 * $$ X^{(l+1)}=\mathcal{U}X^{(l)} $$
 * This is a demo. No quality regularization! No constant injected to the basis.
 * If Gram matrix \( G_{ij}=\sum\limits_{l} X_i^{(l)}X_j^{(l)} \) is degenerated --
 * the code fails since no quality regularization is used!
 * @version $Id: DemoRecoverUnitaryMatrixFromSeq.java,v 1.9 2024/05/15 15:01:47 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DemoRecoverUnitaryMatrixFromSeq {

    /** To turn on a selftest. */
    static final double eps=1e-7;




/** Command line options for {@link DemoRecoverUnitaryMatrixFromSeq}.
 * @version $Id: DemoRecoverUnitaryMatrixFromSeq.java,v 1.9 2024/05/15 15:01:47 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
static class RUConfig {
    public final int indcol_x_start,indcol_x_end;
    public final int indcol_label,columns_number;
    public final int flag_print_verbosity;

    /** In/Out file related info.
     */
    public final String output_files_prefix,data_file_to_build_unitarymodel_from;

    protected final java.util.List<String> opts_remained_unprocessed=new java.util.ArrayList<String>();
    public RUConfig(final String [] args)
    {
	int indcol_x_start=-1,indcol_x_end=-1,
	    indcol_label=-1,columns_number=-1;
	String data_file_to_build_unitarymodel_from=null,
	    output_files_prefix="";

	int flag_print_verbosity=2;
	final java.util.ArrayList<String> data_file_evaluation=new java.util.ArrayList<>();
	for(String a:args)
	{
	    String v=null;
	    if(null!=(v=getVal(a,"--data_file_to_build_unitarymodel_from="))) data_file_to_build_unitarymodel_from=v;
	    else if(null!=(v=getVal(a,"--data_file_evaluation="))) data_file_evaluation.add(v);
	    else if(null!=(v=getVal(a,"--output_files_prefix="))) output_files_prefix=v;
	    else if(null!=(v=getVal(a,"--flag_print_verbosity="))) flag_print_verbosity=Integer.valueOf(v);
	    else if(null!=(v=getVal(a,"--data_cols="))){
		final String [] k=com.polytechnik.utils.ReadWriteUtils.extractnames(v,":");
		if(k.length!=3) throw new RuntimeException("(numcols:xstart,xend:label) colums index="+v+" is invalid. The format is like 15:0,3:0");
		columns_number=Integer.valueOf(k[0]);
		final String [] kx=com.polytechnik.utils.ReadWriteUtils.extractnames(k[1],",");
		indcol_label=Integer.valueOf(k[2]);
		if(kx.length!=2) throw new RuntimeException("(numcols:xstart,xend:label) colums index="+v+" is invalid. The format is like 15:0,3:0");
		indcol_x_start=Integer.valueOf(kx[0]);
		indcol_x_end=Integer.valueOf(kx[1]);
	    }
	    else{
		opts_remained_unprocessed.add(a);
	    }
	}
	this.columns_number=columns_number;
	this.indcol_x_start=indcol_x_start;
	this.indcol_x_end=indcol_x_end;
	this.indcol_label=indcol_label;

	this.data_file_to_build_unitarymodel_from=data_file_to_build_unitarymodel_from;
	this.output_files_prefix=output_files_prefix;
	this.flag_print_verbosity=flag_print_verbosity;
    }
    public java.util.List<String> validateOptions()
    {
	final java.util.List<String> errs=new java.util.ArrayList<String>();
	if(data_file_to_build_unitarymodel_from==null) errs.add("no --data_file_to_build_unitarymodel_from= specified");
	if(columns_number<=0||
	   indcol_x_start<0||indcol_x_end<0||
	   indcol_x_start>indcol_x_end||indcol_x_end>=columns_number||
	   indcol_label >=columns_number ) errs.add("no --data_cols=numcols:xstart,xend:label columns numbers specified for the file --data_file_to_build_unitarymodel_from=");
	if(opts_remained_unprocessed.size()>0)
	    errs.add("There are unknown options found:"+opts_remained_unprocessed.toString());
	return errs;
    }
    public int [] getXIndexesSequence(){
	final int [] inX=new int [indcol_x_end-indcol_x_start+1];
	for(int k=indcol_x_start,ind=0;k<=indcol_x_end;k++){
	    inX[ind++]=k;
	}
	return inX;
    }
    public String getDescription()
    {
	return "#BEGIN{KGOConfig}\n"+
	    "data_file_to_build_unitarymodel_from="+data_file_to_build_unitarymodel_from+" # input file name\n"+
	    "output_files_prefix="+output_files_prefix+"\n"+
	    "data_cols="+columns_number+":"+indcol_x_start+","+indcol_x_end+":"+indcol_label+" # (numcols:xstart,xend:label) the file="+data_file_to_build_unitarymodel_from+" columns index (base 0)\n"+
	    "flag_print_verbosity="+flag_print_verbosity+" # (2 standard output (default), 0 skip X in output)\n"+
	    "#END{KGOConfig}\n";
    }
} // END OF class RUConfig

public static double [] getUFromArgsGramMatrixChannel(final String [] args) throws java.io.IOException {
    return getUFromConfigGramMatrixChannel(new RUConfig(args));
}
public static double [] getUFromArgsUnitMatrixChannel(final String [] args) throws java.io.IOException {
    return getUFromConfigUnitMatrixChannel(new RUConfig(args));
}

static double [] getUFromConfigGramMatrixChannel(final RUConfig c) throws java.io.IOException {
    if(c.validateOptions().size()>0) throw new IllegalArgumentException();

    // read sample
    final var samp=DataReadObservationVectorXF.getObservations(
	c.data_file_to_build_unitarymodel_from,
	c.columns_number,
	c.getXIndexesSequence(),
	-1, // indcol_f
	-1, // indcol_weight
	c.indcol_label);

    // BEGIN Create X and F data from common sample
    final var fHeader=new java.util.ArrayList<String>();
    for(var hX:samp.xHeader){
	fHeader.add("F"+hX);
    }
    final var dSam=new java.util.ArrayList<ObservationVectorXVectorF>();
    for(int l=0,M=samp.data.size();l<M-1;l++){
	final var sThis=samp.data.get(l);
	final var sNext=samp.data.get(l+1);
	dSam.add(new ObservationVectorXVectorF(sThis.label,sThis.x,sNext.x,sThis.weight));
    }
    final var sampV=new DataReadObservationVectorXVectorF(dSam,samp.xHeader,fHeader);
    // END Create X and F data from common sample


    // a very crude way to calculate. no proper regularization, nothing.
    final double []
	Gx=sampV.getFun1Fun2Averaged(o->o.x,o->o.x,o->1),
	Gf=sampV.getFun1Fun2Averaged(o->o.f,o->o.f,o->1);
    // crude regularization
    final int nX=sampV.xHeader.size(),nF=sampV.fHeader.size();

    // regularization is non-unique. Note! The result does not depend on regularization used.
    final double [] Rx,RxInv,Rf,RfInv;
    if(true){
	// trivial way
	final ApplyFunToMatrix aGx=new ApplyFunToMatrix(nX,Gx),aGf=new ApplyFunToMatrix(nF,Gf);
	Rx=aGx.applyFun((i,lam)->Math.sqrt(1/lam));
	RxInv=aGx.applyFun((i,lam)->Math.sqrt(lam));
	Rf=aGf.applyFun((i,lam)->Math.sqrt(1/lam));
	RfInv=aGf.applyFun((i,lam)->Math.sqrt(lam));
    } else {
	// Gram Schmidt
	final var rx=new GramSchmidtTest(nX,Gx,eps);
	Rx=rx.basis;
	RxInv=rx.basisInv;
	final var rf=new GramSchmidtTest(nF,Gf,eps);
	Rf=rf.basis;
	RfInv=rf.basisInv;
    }
    if(eps>0){
	UnitTests.checkDoubleArrayRel1(TMatr.mulMM(nX,Rx,RxInv),"Rx*RxInv",TMatr.getUnitMatrix(nX),"I",eps);
	UnitTests.checkDoubleArrayRel1(TMatr.mulMM(nF,Rf,RfInv),"Rf*RfInv",TMatr.getUnitMatrix(nF),"I",eps);
	UnitTests.checkDoubleArrayRel1(
	    sampV.getFun1Fun2Averaged(o->TMatr.getMatr_x_A(nX,Rx,o.x),o->TMatr.getMatr_x_A(nX,Rx,o.x),o->1),
	    "Rx Rx",TMatr.getUnitMatrix(nX),"I",eps);
	UnitTests.checkDoubleArrayRel1(
	    sampV.getFun1Fun2Averaged(o->TMatr.getMatr_x_A(nF,Rf,o.f),o->TMatr.getMatr_x_A(nF,Rf,o.f),o->1),
	    "Rf Rf",TMatr.getUnitMatrix(nF),"I",eps);
    }

    // slow but simple calculation
    final double [] Sfxfx=sampV.getFun1Fun2Averaged(
	o->TMatr.dyadicMatrix(TMatr.getMatr_x_A(nF,Rf,o.f),TMatr.getMatr_x_A(nX,Rx,o.x)),
	o->TMatr.dyadicMatrix(TMatr.getMatr_x_A(nF,Rf,o.f),TMatr.getMatr_x_A(nX,Rx,o.x)),
	o->1);
    // find u in regularized basis
    final var kgo=new KGOIterationalSubspaceLinearConstraints(nF,nX,Sfxfx,0*eps);
    final double [] uRfRx=kgo.aep.getUAdj();
    // convert to original basis
    if(nX!=nF) throw new IllegalArgumentException();
    final int n=nX;
    final double [] u=TMatr.mulMMM(n,RfInv,uRfRx,Rx);

    //System.err.println("u="+java.util.Arrays.toString(u));
    return u;
}


static double [] getUFromConfigUnitMatrixChannel(final RUConfig c) throws java.io.IOException {
    if(c.validateOptions().size()>0) throw new IllegalArgumentException();

    // read sample
    final var samp=DataReadObservationVectorXF.getObservations(
	c.data_file_to_build_unitarymodel_from,
	c.columns_number,
	c.getXIndexesSequence(),
	-1, // indcol_f
	-1, // indcol_weight
	c.indcol_label);

    // BEGIN Create X and F data from common sample
    final var fHeader=new java.util.ArrayList<String>();
    for(var hX:samp.xHeader){
	fHeader.add("F"+hX);
    }
    final var dSam=new java.util.ArrayList<ObservationVectorXVectorF>();
    for(int l=0,M=samp.data.size();l<M-1;l++){
	final var sThis=samp.data.get(l);
	final var sNext=samp.data.get(l+1);
	dSam.add(new ObservationVectorXVectorF(sThis.label,sThis.x,sNext.x,sThis.weight));
    }
    final var sampV=new DataReadObservationVectorXVectorF(dSam,samp.xHeader,fHeader);
    // END Create X and F data from common sample

    // no regularization, traditional unitary learning
    final int nX=sampV.xHeader.size(),nF=sampV.fHeader.size();

    // slow but simple calculation
    final double [] Sfxfx=sampV.getFun1Fun2Averaged(
	o->TMatr.dyadicMatrix(o.f,o.x),
	o->TMatr.dyadicMatrix(o.f,o.x),
	o->1);
    final var kgo=new KGOIterationalSubspaceLinearConstraints(nF,nX,Sfxfx,0*eps);
    if(nX!=nF) throw new IllegalArgumentException();
    final int n=nX;
    final double [] u=kgo.aep.getUAdj();

    //System.err.println("u="+java.util.Arrays.toString(u));
    return u;
}

static void printU(final String txt,final double [] u){
    System.out.println(txt);
    final int n=(int)Math.sqrt(u.length);
    if(n*n!=u.length) throw new IllegalArgumentException();
    for(int i=0;i<n;i++){
	final var v=EVXData.getPsiFromBasis(i,u,n,0);
	System.out.println(java.util.Arrays.toString(v));
    }
}

public static void main(String [] args) throws java.io.IOException {
    final var c=new RUConfig(args);
    System.err.println("description\n"+c.getDescription());
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }
    final double [] uGram=getUFromConfigGramMatrixChannel(c);
    printU("Recovered Gram u:",uGram);

    final double [] uUnit=getUFromConfigUnitMatrixChannel(c);
    printU("Recovered Unit u:",uUnit);

}


}
