package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.DataReadObservationVectorXF;
import com.polytechnik.utils.DataReadObservationVectorXVectorF;
import com.polytechnik.utils.ObservationVectorXVectorF;
import com.polytechnik.utils.ObservationVectorXF;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.Chebyshev;
import com.polytechnik.utils.Legendre;
import com.polytechnik.utils.LegendreShifted;
import com.polytechnik.utils.BasisPolynomials;
import com.polytechnik.utils.Linsystems;
import com.polytechnik.utils.PolynomialRootsConfederateMatrix;
import com.polytechnik.kgo.DemoRecoverMapping.GramSchmidtTest;
import static com.polytechnik.utils.RNConfig.getVal;

/** A demo of scalar f(x) interpolation.
 * @version $Id: DemoScalarInterpolation.java,v 1.16 2024/05/12 17:13:14 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DemoScalarInterpolation {

    /** To turn on a selftest. */
    static final double eps=1e-7;

   static final BasisPolynomials
       CBX=new Legendre(),
       CBF=new LegendreShifted();



/** Command line options for {@link DemoScalarInterpolation}.
 * @version $Id: DemoScalarInterpolation.java,v 1.16 2024/05/12 17:13:14 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
static class RConfig {
    public final int indcol_x,indcol_f,indcol_w;;
    public final int indcol_label,columns_number;
    public final int max_multiindex_x,max_multiindex_f;
    public final int flag_print_verbosity;

    /** In/Out file related info.
     */
    public final String output_files_prefix,data_file_to_read_from;

    protected final java.util.List<String> opts_remained_unprocessed=new java.util.ArrayList<String>();
    public RConfig(final String [] args)
    {
	int indcol_x=-1,indcol_f=-1,indcol_w=-1,
	    indcol_label=-1,columns_number=-1;
	int max_multiindex_x=-1,max_multiindex_f=-1;
	String data_file_to_read_from=null,
	    output_files_prefix="";

	int flag_print_verbosity=2;
	final java.util.ArrayList<String> data_file_evaluation=new java.util.ArrayList<>();
	for(String a:args)
	{
	    String v=null;
	    if(null!=(v=getVal(a,"--data_file_to_read_from="))) data_file_to_read_from=v;
	    else if(null!=(v=getVal(a,"--output_files_prefix="))) output_files_prefix=v;
	    else if(null!=(v=getVal(a,"--flag_print_verbosity="))) flag_print_verbosity=Integer.valueOf(v);
	    else if(null!=(v=getVal(a,"--data_cols="))){
		final String [] k=com.polytechnik.utils.ReadWriteUtils.extractnames(v,":");
		if(k.length!=5) throw new RuntimeException("(numcols:xcol:fcol:wcol:label) colums index="+v+" is invalid. The format is like 15:3:4:-1:0");
		columns_number=Integer.valueOf(k[0]);
		indcol_x=Integer.valueOf(k[1]);
		indcol_f=Integer.valueOf(k[2]);
		indcol_w=Integer.valueOf(k[3]);
		indcol_label=Integer.valueOf(k[4]);
	    }
	    else if(null!=(v=getVal(a,"--max_multiindex_x="))) max_multiindex_x=Integer.valueOf(v);
	    else if(null!=(v=getVal(a,"--max_multiindex_f="))) max_multiindex_f=Integer.valueOf(v);
	    else{
		opts_remained_unprocessed.add(a);
	    }
	}
	this.columns_number=columns_number;
	this.indcol_x=indcol_x;
	this.indcol_f=indcol_f;
	this.indcol_w=indcol_w;
	this.indcol_label=indcol_label;
	this.max_multiindex_x=max_multiindex_x;
	this.max_multiindex_f=max_multiindex_f;

	this.data_file_to_read_from=data_file_to_read_from;
	this.output_files_prefix=output_files_prefix;
	this.flag_print_verbosity=flag_print_verbosity;
    }
    public java.util.List<String> validateOptions()
    {
	final java.util.List<String> errs=new java.util.ArrayList<String>();
	if(data_file_to_read_from==null) errs.add("no --data_file_to_read_from= specified");
	if(columns_number<=0 ||
	   indcol_x<0||indcol_x>=columns_number ||
	   indcol_f<0||indcol_f>=columns_number ||
	   indcol_label<0 || indcol_label >=columns_number ||
	   indcol_w >= columns_number
	   ) errs.add("no --data_cols=numcols:xcol:fcol:wcol:label columns numbers specified for the file --data_file_to_read_from=");
	if(max_multiindex_x<=0) errs.add("no --max_multiindex_x=");
	if(max_multiindex_f<=0) errs.add("no --max_multiindex_f=");
	if(opts_remained_unprocessed.size()>0)
	    errs.add("There are unknown options found:"+opts_remained_unprocessed.toString());
	return errs;
    }
    public String getDescription()
    {
	return "#BEGIN{KGOConfig}\n"+
	    "data_file_to_read_from="+data_file_to_read_from+" # input file name\n"+
	    "output_files_prefix="+output_files_prefix+"\n"+
	    "data_cols="+columns_number+":"+indcol_x+":"+indcol_f+":"+indcol_w+":"+indcol_label+" # (numcols:xcol:fcol:wcol:label) the file="+data_file_to_read_from+" columns index (base 0)\n"+
	    "max_multiindex_x="+max_multiindex_x+"\n"+
	    "max_multiindex_f="+max_multiindex_f+"\n"+
	    "flag_print_verbosity="+flag_print_verbosity+" # (2 standard output (default), 0 skip X in output)\n"+
	    "#END{KGOConfig}\n";
    }
} // END OF class RConfig



static class FPRes {
    final double f,P;
    FPRes(final double f,final double P){
	this.f=f;
	this.P=P;
    }
}
static class Res {
    final double [] GfInv,GxInv,Gf;
    final double [] u;
    final int nF,nX;
    final DataReadObservationVectorXVectorF sampV;
    /** For diag. */
    final double [] fGx,fx;
    Res(final double [] GfInv,
	final double [] GxInv,
	final double [] Gf,
	final double [] u,
	final int nF,
	final int nX,
	final double [] fGx,
	final double [] fx,
	final DataReadObservationVectorXVectorF sampV){
	this.GfInv=GfInv;
	this.GxInv=GxInv;
	this.u=u;
	this.nF=nF;
	this.nX=nX;
	// for tests
	this.Gf=Gf;
	this.fGx=fGx;
	this.fx=fx;
	this.sampV=sampV;
    }

    double evaluatef_LeastSquares(double x){
	final double [] X=CBX.getX0Moments(nX,x);
	return TMatr.getAMatrB(X,fx,GxInv,nX);
    }

    double evaluatef_RN(double x){
	final double [] X=CBX.getX0Moments(nX,x);
	return TMatr.getAMatrA(X,TMatr.mulMMM(nX,GxInv,fGx,GxInv),nX)/
	    TMatr.getAMatrA(X,GxInv,nX);
    }

    double getP(double f,double x){
	final double [] X=CBX.getX0Moments(nX,x);
	final double oKx=TMatr.getAMatrA(X,GxInv,nX);
	final double [] a=TMatr.getVectorScaled(
	    TMatr.mul_MA_x_MB(
	    nF,
	    GfInv,
	    nF,
	    TMatr.mul_MA_x_MB(nF,u,nX,X,1),
	    1),
	    1/Math.sqrt(oKx));
	final double [] F=CBF.getX0Moments(nF,f);

	return Math.pow(TMatr.mulVV(F,a),2)/TMatr.getAMatrA(F,GfInv,nF);
    }

    FPRes getFPmax(double x){
	final double [] X=CBX.getX0Moments(nX,x);
	final double oKx=TMatr.getAMatrA(X,GxInv,nX);
	final double [] a=TMatr.getVectorScaled(
	    TMatr.mul_MA_x_MB(
	    nF,
	    GfInv,
	    nF,
	    TMatr.mul_MA_x_MB(nF,u,nX,X,1),
	    1),
	    1/Math.sqrt(oKx));
	final double [] nom=CBF.mult2Pol(a,a),
	    denom=CBF.getKK(nF,GfInv);
	if(eps>0){
	    for(double ft:new double[]{0.02,0.1,0.3}){
		final double Pv=getP(ft,x),
		    Pp=CBF.getD0(nom,ft)/CBF.getD0(denom,ft);
		UnitTests.checkDoubleRel1(Pv,"Pv",Pp,"Pp",eps);
		System.err.println("Probability OK");
	    }
	}
	final double []
	    diffP=TMatr.sumAcB(
		CBF.mult2Pol(CBF.getDifferentiated(nom),denom),
		-1,
		CBF.mult2Pol(CBF.getDifferentiated(denom),nom)),
	    remove0=java.util.Arrays.copyOf(diffP,diffP.length-1);
	System.err.println("diffP="+java.util.Arrays.toString(diffP)+" remove0="+java.util.Arrays.toString(remove0));
	final var rf=new PolynomialRootsConfederateMatrix(CBF);
	final var roots=rf.findRoots(remove0);
	double fPmax=Double.NaN;
	double pMax=0;
	int indSel=-1;
	// ckeck all of them
	for(int nr=0;nr<roots.re_root.length;nr++){
	    final double f=roots.re_root[nr];
	    final double p=CBF.getD0(nom,f)/CBF.getD0(denom,f);
	    if(Double.isNaN(fPmax)||p>=pMax){
		pMax=p;
		fPmax=f;
		indSel=nr;
	    }
	}
	if(eps>0 && indSel>=0){
	    //if(roots.im_root[indSel]!=0) throw new IllegalStateException();
	}
	return new FPRes(fPmax,pMax);
    }
    FPRes getFPmaxLin(double x){
	final double [] X=CBX.getX0Moments(nX,x);
	final double oKx=TMatr.getAMatrA(X,GxInv,nX);
	final double [] a=TMatr.getVectorScaled(
	    TMatr.mul_MA_x_MB(
	    nF,
	    GfInv,
	    nF,
	    TMatr.mul_MA_x_MB(nF,u,nX,X,1),
	    1),
	    1/Math.sqrt(oKx));
	final double P=TMatr.getAMatrA(a,Gf,nF);
	final double [] Fp=TMatr.getMatr_x_A(nF,Gf,a),
	    mP1=CBF.convertFunctionTo_Monomials_from_Basis(new double []{0,1});
	final double
	    bP0=CBF.calculate(0,0),
	    rpf=(Fp[1]/Fp[0])*bP0;
	final double f=(rpf-mP1[0])/mP1[1];
	return new FPRes(f,P);
    }

}// END OF class Res

static Res getResFromConfig(final RConfig c) throws java.io.IOException {
    if(c.validateOptions().size()>0) throw new IllegalArgumentException();

    // read sample
    final var samp=DataReadObservationVectorXF.getObservations(
	c.data_file_to_read_from,
	c.columns_number,
	new int []{c.indcol_x},
	c.indcol_f, // indcol_f
	c.indcol_w, // indcol_w
	c.indcol_label);

    // Create vector to vector sample.
    // A very crude method, no argument scale used to bring to [-1:1] range.
    // BEGIN Create X and F vector data from common sample
    final java.util.ArrayList<String>
	xHeader=new java.util.ArrayList<>(),
	fHeader=new java.util.ArrayList<>();
    for(int i=0;i<c.max_multiindex_x;i++){xHeader.add("XT"+i);}
    for(int i=0;i<c.max_multiindex_f;i++){fHeader.add("FT"+i);}
    final var dSam=new java.util.ArrayList<ObservationVectorXVectorF>();
    for(var o:samp.data){
	final double [] X=CBX.getX0Moments(xHeader.size(),o.x[0]),
	    F=CBF.getX0Moments(fHeader.size(),o.f);
	dSam.add(new ObservationVectorXVectorF_WithExtra(o.label,X,F,o.weight,o));
    }
    final var sampV=new DataReadObservationVectorXVectorF(dSam,xHeader,fHeader);
    // END Create X and F data from common sample


    // a very crude way to calculate. no proper regularization, just a test.
    final double []
	Gx=sampV.getFun1Fun2Averaged(o->o.x,o->o.x,o->o.weight),
	Gf=sampV.getFun1Fun2Averaged(o->o.f,o->o.f,o->o.weight);
    // a crude way to regularize
    final int nX=sampV.xHeader.size(),nF=sampV.fHeader.size();
    final double [] Rx,RxInv,Rf,RfInv;
    if(true){
	// trivial way
	final ApplyFunToMatrix aGx=new ApplyFunToMatrix(nX,Gx),aGf=new ApplyFunToMatrix(nF,Gf);
	Rx=aGx.applyFun((i,lam)->Math.sqrt(1/lam));
	RxInv=aGx.applyFun((i,lam)->Math.sqrt(lam));
	Rf=aGf.applyFun((i,lam)->Math.sqrt(1/lam));
	RfInv=aGf.applyFun((i,lam)->Math.sqrt(lam));
    } else {
	// Gram Schmidt
	final var rx=new GramSchmidtTest(nX,Gx,eps);
	Rx=rx.basis;
	RxInv=rx.basisInv;
	final var rf=new GramSchmidtTest(nF,Gf,eps);
	Rf=rf.basis;
	RfInv=rf.basisInv;
    }
    if(eps>0){
	UnitTests.checkDoubleArrayRel1(TMatr.mulMM(nX,Rx,RxInv),"Rx*RxInv",TMatr.getUnitMatrix(nX),"I",eps);
	UnitTests.checkDoubleArrayRel1(TMatr.mulMM(nF,Rf,RfInv),"Rf*RfInv",TMatr.getUnitMatrix(nF),"I",eps);
	UnitTests.checkDoubleArrayRel1(
	    sampV.getFun1Fun2Averaged(o->TMatr.getMatr_x_A(nX,Rx,o.x),o->TMatr.getMatr_x_A(nX,Rx,o.x),o->o.weight),
	    "Rx Rx",TMatr.getUnitMatrix(nX),"I",eps);
	UnitTests.checkDoubleArrayRel1(
	    sampV.getFun1Fun2Averaged(o->TMatr.getMatr_x_A(nF,Rf,o.f),o->TMatr.getMatr_x_A(nF,Rf,o.f),o->o.weight),
	    "Rf Rf",TMatr.getUnitMatrix(nF),"I",eps);
    }


    // slow but simple calculation
    final double [] SfxfxKfKx=sampV.getFun1Fun2Averaged(
	o->TMatr.dyadicMatrix(TMatr.getMatr_x_A(nF,Rf,o.f),TMatr.getMatr_x_A(nX,Rx,o.x)),
	o->TMatr.dyadicMatrix(TMatr.getMatr_x_A(nF,Rf,o.f),TMatr.getMatr_x_A(nX,Rx,o.x)),
	o->{
	    final double [] fT=TMatr.getMatr_x_A(nF,Rf,o.f),xT=TMatr.getMatr_x_A(nX,Rx,o.x);
	    // Christoffel function in regualized basis is simple
	    return o.weight/(TMatr.mulVV(fT,fT)*TMatr.mulVV(xT,xT));
	});
    // find u in regularized basis
    final var kgo=new KGOIterationalSubspaceLinearConstraints(nF,nX,SfxfxKfKx,0*eps);

    final double [] uRfRx=kgo.aep.getUAdj();
    // convert to original basis
    final double [] u=
	TMatr.mul_MA_x_MB(nF,
			  TMatr.mul_MA_x_MB(nF,RfInv,nF,uRfRx,nX),
			  nX,
			  Rx,
			  nX);
    if(eps>0){
	UnitTests.checkDoubleArrayRel1(
	    Gf,"Gf",
	    EVXData.convertMatrFromQtoB_Basis(nX,Gx,u,nF,0),"u Gx uT",eps);
    }


    // a klugde to get standard approximations
    final double []
	fGx=sampV.getFun1Fun2Averaged(o->o.x,o->o.x,o->{
		if(o instanceof ObservationVectorXVectorF_WithExtra op){
		    return op.weight*op.orig.f;
		} else throw new IllegalArgumentException();}),
	fx=sampV.getFun1Fun2Averaged(o->o.x,o->new double []{1},o->{
		if(o instanceof ObservationVectorXVectorF_WithExtra op){
		    return op.weight*op.orig.f;
		} else throw new IllegalArgumentException();});

    return new Res(Linsystems.getInvertedMatrix(nF,nF,Gf),
		   Linsystems.getInvertedMatrix(nX,nX,Gx),
		   Gf,
		   u,
		   nF,
		   nX,
		   fGx,fx,sampV);
}

public static void main(String [] args) throws java.io.IOException {
    final var c=new RConfig(args);
    System.err.println("description\n"+c.getDescription());
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }

    final var res=getResFromConfig(c);


    // print u
    System.out.println("Recovered u:");
    for(int i=0;i<res.nF;i++){
	final var v=EVXData.getPsiFromBasis(i,res.u,res.nX,0);
	System.out.println(java.util.Arrays.toString(v));
    }

    printEvaluation(c,res.sampV,res,c.output_files_prefix+"scal.csv");
}


static void printEvaluation(
    final RConfig c,
    final DataReadObservationVectorXVectorF sampV,
    final Res res,
    final String outFname) throws java.io.IOException{

    final String s=DataReadObservationVectorXF.SEPARATOR;
    System.err.println("Writing to "+outFname+" evaluation_data model is built from "+c.data_file_to_read_from);
    final java.io.PrintWriter out=new java.io.PrintWriter(outFname);

    out.println("|label:0"+s+"x:1"+s+"f:2"+s+"fLS:3"+s+"fRN:4"+s+"fPmax:5"+s+"Pmax:6"+s+"LfPmax:7"+s+"LPmax:8");
    for(var o:sampV.data){
	if(o instanceof ObservationVectorXVectorF_WithExtra op){
	    final double x=op.orig.x[0];
	    final var fpm=res.getFPmax(x);
	    final var fpl=res.getFPmaxLin(x);
	    out.println(String.valueOf(op.label)+s+x+s+op.orig.f+s+res.evaluatef_LeastSquares(x)+
			s+res.evaluatef_RN(x)+
			s+fpm.f+
			s+fpm.P+
			s+fpl.f+
			s+fpl.P
			);
	} else throw new IllegalArgumentException();
    }
    out.close();
}


private static class ObservationVectorXVectorF_WithExtra extends ObservationVectorXVectorF {
	final ObservationVectorXF orig;
	ObservationVectorXVectorF_WithExtra(
	     final String label,
	     final double [] x,
	     final double [] f,
	     final double weight,
	     final ObservationVectorXF orig){
	     super(label,x,f,weight);
	     this.orig=orig;
	 }
}// END OF class ObservationVectorXVectorF_WithExtra



}
