package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2021
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.XXMatrs;
import com.polytechnik.utils.VectorXVectorFPointAccess;
import com.polytechnik.utils.VectorXVectorFPointEvaluatable;
import com.polytechnik.utils.Linsystems;

/** A simple test-solution to a vector X to vector F problem.
 * The X- and F- spaces are assumed to be regularized (XX and FF are not degenerated).
 * @version $Id: DirectProjectionSolutionVectorXVectorF.java,v 1.24 2024/04/02 01:29:09 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DirectProjectionSolutionVectorXVectorF implements  VectorXVectorFPointEvaluatable {
    final int nX,nF;
    final double [] FF/* nF nF */,XX/* nX nX */,XF /* nX nF */;
    final double [] Gfm1,Gxm1;

    /** Rank matrices Eq. (86).
     */
    final XXMatrs rankMatrices;

    /** Some temporary matrices.
     */
    final double [] Gxm1Xf,Gxm1XfGfm1,Gxm1XfGfm1XfGxm1,XfGfm1;

    /**
     * @param nX X-space dimension.
     * @param nF F-space dimension.
     * @param FF The matrix  {@literal <FF>} nF x nF dimension.
     * @param XX The matrix  {@literal <XX>} nX x nX dimension.
     * @param XF The matrix  {@literal <XF>} nX x nF dimension.
     * The X- and F- spaces are assumed to be regularized (non-degenerated),
     * but not necessary orthogonalized.
     */
    public DirectProjectionSolutionVectorXVectorF(final int nX,
						  final int nF,
						  final double [] FF,
						  final double [] XX,
						  final double [] XF){
	this.nX=nX;
	this.nF=nF;
	this.FF=FF;if(FF.length!=nF*nF) throw new IllegalArgumentException();
	this.XX=XX;if(XX.length!=nX*nX) throw new IllegalArgumentException();
	this.XF=XF;if(XF.length!=nX*nF) throw new IllegalArgumentException();
	this.Gxm1=Linsystems.getInvertedMatrix(nX,nX,XX);
	this.Gfm1=Linsystems.getInvertedMatrix(nF,nF,FF);

	this.rankMatrices=getFFandFFproj();

	// temp matrices
	Gxm1Xf=getGxm1Xf();
	Gxm1XfGfm1=getGxm1XfGfm1();
	Gxm1XfGfm1XfGxm1=getGxm1XfGfm1XfGxm1();
	XfGfm1=getXfGfm1();
    }

    private XXMatrs getFFandFFproj(){
	final double [] tmp=TMatr.mul_MA_x_MB(nX,Gxm1,nX,XF,nF),
	    XfGxm1Xf=TMatr.mul_MATrans_x_MB(nF,XF,nX,tmp,nF);
	return new XXMatrs(nF,nF,XfGxm1Xf,FF);
    }

    /** Covered rank, corresponds to the sum in Eq. (87).
     */
    double getProjectedTotalRankWithoutSpectrum(){
	return TMatr.getSpurMM(nF,rankMatrices.XfX,Gfm1);
    }

    private double [] getXfGfm1(){
	return TMatr.mul_MA_x_MB(nX,XF,nF,Gfm1,nF);
    }

    private double [] getGxm1Xf(){
	return TMatr.mul_MA_x_MB(nX,Gxm1,nX,XF,nF);
    }

    private double [] getGxm1XfGfm1(){
	final double [] Gxm1XfGfm1=TMatr.mul_MA_x_MB(nX,getGxm1Xf(),nF,Gfm1,nF);
	if(true){
	    final double [] tmpx=TMatr.mul_MATrans_x_MB(nF,XF,nX,Gxm1XfGfm1,nF);
	    System.err.println("Spur(Gxm1XfGfm1 Xf)="+TMatr.getSpurM(nF,tmpx));
	}
	return Gxm1XfGfm1;
    }

    private double [] getGxm1XfGfm1XfGxm1(){
	final double [] Gxm1XfGfm1XfGxm1=TMatr.mul_MA_x_MBTrans(nX,getGxm1XfGfm1(),nF,getGxm1Xf(),nX);
	if(true){
	    System.err.println("Spur(Gxm1XfGfm1XfGxm1 XX)="+TMatr.getSpurMM(nX,Gxm1XfGfm1XfGxm1,XX));
	}
	return Gxm1XfGfm1XfGxm1;
     }

    public VectorXVectorFPointAccess evaluateAt(final double [] X){
	if(X.length!=nX) throw new IllegalArgumentException();
	final double [] LeastSquaresFvalues=TMatr.mul_MA_x_MB(1,X,nX,Gxm1Xf,nF);
	final double [] bGfm1=TMatr.getMatr_x_A(nF,Gfm1,LeastSquaresFvalues);
	final double _ChristoffelOatX=TMatr.getAMatrA(X,Gxm1XfGfm1XfGxm1,nX);


	return new VectorXVectorFPointAccess() {
	    /** The Prob(g|y); if the Error=0 then it is=1, corresponds to Eq. (99).
	     * For Direct Projection model getFProbabilityAtX(getFLeastSquaresX())=1.
	     */
	    public double getFProbabilityAtX(final double [] F){
		if(F.length!=nF) throw new IllegalArgumentException();
		return Math.pow(TMatr.mulVV(bGfm1,F),2)/(getChristoffelOatF(F)*getChristoffelOatX());
	    }
	    public double getChristoffelOatX(){return _ChristoffelOatX;}
	    public double getChristoffelOatF(final double [] F){return TMatr.getAMatrA(F,Gfm1,nF);}
	    /** Equals to 1 for Direct Projection model.
	     */
	    public double getMaximalProbability(){return TMatr.getAMatrA(LeastSquaresFvalues,Gfm1,nF)/getChristoffelOatX();}
	    public double [] getPsiFNonNormalized(){return bGfm1.clone();}
	    public double [] getFLeastSquaresX(){return LeastSquaresFvalues.clone();}
	    public double [] getFofMaximalProbability(){return getFLeastSquaresX();}
	};
    }

}
