package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.DataReadObservationVectorXF;
import com.polytechnik.utils.DataReadObservationVectorXVectorF;
import com.polytechnik.utils.RegularizedVectorXVectorF;

/** A simple Knowledge Generalizing Operator model. A "driver" class.
  * @version $Id: KGO.java,v 1.18 2024/04/08 22:11:10 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
  */
public class KGO {

public static void printKGOEvaluation(
    final String [] comand_line_arguments,
    final String kgo_data_file_to_build_model_from,
    final KGOSolutionVectorXVectorF kgo,
    final int flag_print_verbosity,
    final DataReadObservationVectorXVectorF ro,
    final String evaluation_data_file,
    final java.util.function.Function<String,String> getoutputfilename) throws java.io.IOException{

    final String s=DataReadObservationVectorXF.SEPARATOR;
    final String fout=getoutputfilename.apply(new java.io.File(evaluation_data_file).getName());
    System.err.println("Writing to "+fout+" evaluation_data "+evaluation_data_file+" model is built from "+kgo_data_file_to_build_model_from);
    final java.io.PrintWriter out=new java.io.PrintWriter(fout);
    out.println("|@KGO_command_line_arguments:"+s+java.util.Arrays.toString(comand_line_arguments));
    out.println("|@version:"+s+"$Id: KGO.java,v 1.18 2024/04/08 22:11:10 mal Exp $");
    out.println("|");
    out.println("|MODEL");
    out.println("|kgo_data_file_to_build_model_from"+s+kgo_data_file_to_build_model_from);
    out.println("|totalweight"+s+kgo.reg.X.totalweight);
    out.println("|noriginal x"+s+kgo.reg.X.noriginal);
    out.println("|nX"+s+kgo.reg.X.nX);
    out.println("|noriginal f"+s+kgo.reg.F.noriginal);
    out.println("|nF"+s+kgo.reg.F.nX);


    if(flag_print_verbosity>=3){

    }
    out.println("|");
    out.println("|EVALUATION");
    out.println("|kgo_data_file_to_build_model_from"+s+kgo_data_file_to_build_model_from);
    out.println("|evaluation_data_file"+s+evaluation_data_file);
    out.println("|sample stat:");
    out.println("|totalweight"+s+kgo.reg.X.totalweight);
    out.println("|points"+s+ro.data.size());
    out.println("|");
    out.print("|");
    out.print("Observation Label"+s);
    if(flag_print_verbosity>=1){
	for(String h:ro.xHeader){
	    out.print(h);out.print(s);
	}
    }
    if(flag_print_verbosity>=2){
	for(String h:ro.fHeader){
	    out.print(h);out.print(s);
	}
    }
    out.print("weight"+s+"ChristoffelX"+s+"ChristoffelXadj"+s+"ChristoffelF"+s+"Max P"+s+"P(observed f)"+s);
    for(String h:ro.fHeader){
	out.print("MP_"+h);out.print(s);
    }
    out.println();

    // print points
    for(var o:ro.data){
	if(o.x.length!=kgo.reg.X.noriginal) throw new RuntimeException("X Header/Data length mismatch");
	if(o.f.length!=kgo.reg.F.noriginal) throw new RuntimeException("F Header/Data length mismatch");
	out.print((o.label!=null?o.label:"??")+s);
	if(flag_print_verbosity>=1){
	    for(double x:o.x){
		out.print(x);out.print(s);
	    }
	}
	if(flag_print_verbosity>=2){
	    for(double f:o.f){
		out.print(f);out.print(s);
	    }
	}
	final var e=kgo.evaluateAt(o.x);
	out.print(""+o.weight+
		  s+(1/e.getChristoffelOatX())+
		  s+(1/e.getChristoffelOatXDimensionAdjusted())+
		  s+(1/e.getChristoffelOatF(o.f))+
		  s+e.getMaximalProbability()+
		  s+e.getFProbabilityAtX(o.f)
		  );

	final var fmaxP=e.getFofMaximalProbability();
	for(double fr:fmaxP){
	    out.print(s+fr);
	}
	out.println();
    }
    out.close();
}


/** A driver for Knowledge Generalizing Operator model.
 */
public static void main(final String [] args) throws java.io.IOException
{
    final var c=new KGOConfig(args);
    System.err.println("description\n"+c.getDescription());
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }

    // Read the data from c.data_file_to_build_model_from and test data from c.data_file_evaluation
    DataReadObservationVectorXVectorF d=null,
	d_test[]=new DataReadObservationVectorXVectorF[c.data_file_evaluation.length];
    for(int k=-1;k<c.data_file_evaluation.length;k++){
	final var fname=k<0?c.data_file_to_build_model_from:c.data_file_evaluation[k];
	final var v=DataReadObservationVectorXVectorF.getObservations(
	    fname,
	    c.columns_number,
	    c.getXIndexesSequence(),
	    c.getFIndexesSequence(),
	    c.indcol_weight,
	    c.indcol_label);
	if(k<0) d=v;
	else {
	    d_test[k]=v;
	    if(!d.xHeader.equals(d_test[k].xHeader))
		throw new RuntimeException("The files data_file_to_build_model_from="+c.data_file_to_build_model_from+" and data_file_evaluation="+c.data_file_evaluation[k]+" have different X-headers:"+d.xHeader+" vs. "+d_test[k].xHeader);
	    if(!d.fHeader.equals(d_test[k].fHeader))
		throw new RuntimeException("The files data_file_to_build_model_from="+c.data_file_to_build_model_from+" and data_file_evaluation="+c.data_file_evaluation[k]+" have different F-headers:"+d.fHeader+" vs. "+d_test[k].fHeader);
	}
    }


    if(c.max_multiindex_x>0||c.max_multiindex_f>0){
	// Multiindex products are required.
	final RegularizedVectorXVectorF dr=d.getDataRegularized_OLDSCALARCODE(
	    KGOSolutionVectorXVectorF.STD_REGULARIZATION_METHOD,
	    KGOSolutionVectorXVectorF.STD_REGULARIZATION_METHOD
									      );
	for(int k=-1;k<c.data_file_evaluation.length;k++){
	    final var v=dr.getAttribsProductsMultiIndexed(
		c.max_multiindex_x,
		c.max_multiindex_f,
		k<0?d:d_test[k]);
	    if(k<0) d=v;
	    else {
		d_test[k]=v;
		if(!d.xHeader.equals(d_test[k].xHeader))
		    throw new RuntimeException("The files data_file_to_build_model_from="+c.data_file_to_build_model_from+" and data_file_evaluation="+c.data_file_evaluation[k]+" have different X-headers:"+d.xHeader+" vs. "+d_test[k].xHeader);
		if(!d.fHeader.equals(d_test[k].fHeader))
		    throw new RuntimeException("The files data_file_to_build_model_from="+c.data_file_to_build_model_from+" and data_file_evaluation="+c.data_file_evaluation[k]+" have different F-headers:"+d.fHeader+" vs. "+d_test[k].fHeader);
	    }
	}
    }

    // force a selftest
    //TestKGO.EPS_selftest=1e-5;//eps;

    final KGOSolutionVectorXVectorF kgo=
	KGOSolutionVectorXVectorF.createFromSample(d,c.approximation,c.SKtype);

    // Evaluate training data
    printKGOEvaluation(args,
		       c.data_file_to_build_model_from,
		       kgo,
		       c.flag_print_verbosity,
		       d,
		       c.data_file_to_build_model_from,
		       (fn)->c.output_files_prefix+fn+".KGO.csv");

    for(int i=0;i<d_test.length;i++){
	var ro=d_test[i];
	if(!ro.xHeader.equals(d.xHeader)) throw new RuntimeException("X header mesmatch");
	if(!ro.fHeader.equals(d.fHeader)) throw new RuntimeException("F header mesmatch");

	printKGOEvaluation(args,
			   c.data_file_to_build_model_from,
			   kgo,
			   c.flag_print_verbosity,
			   ro,
			   c.data_file_evaluation[i],
			   (fn)->c.output_files_prefix+fn+".KGO.csv");
    }
}


}
