package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.EVSolver;

/**
 * A container to store a vector corresponding to a KGO operator.
 * It has several methods to "adjust" stored arbitrary solution
 * to become partially unitary KGO operator.
 * @version $Id: KGOEVSelection.java,v 1.63 2024/04/08 22:11:10 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOEVSelection {
    /** Index of original eigenvector, \(ind \in [0\dots nC*nX-1] \).
     */
    public final int ind;
    /** Original eigenvector, an operator is extracted from it row by row.
     */
    public final double [] eVector;
    /** The eigenvalue of this state.
     */
    public final double eLambda;
    /** The dimensions, {@code eVector.length=nX*nC}.
     */
    public final int nC,nX;
    KGOEVSelection(final int ind,
		   final double [] eVector,
		   final double eLambda,
		   final int nC,
		   final int nX){
	this.ind=ind;
	this.eVector=eVector;
	this.eLambda=eLambda;
	this.nC=nC;
	this.nX=nX;
    }


    /** Adjust the {@link #eVector} to be partially unitary.
     * The result should be the same (within sign) as the one from {@link #getSVDAdjustedTo1()}.
     * This method is faster and returns the data
     * with additional constraint-related info that is useful in iteration process.
     * @see #getSVDAdjustedTo1()
     */
    public AdjustedStateToUnitaryWithEigenproblem getEVAdjustedTo1(){
	// to have the norm equals to Gram matrix spur
	final double [] u_orig=TMatr.getVectorScaled(eVector,Math.sqrt(nC));

	return new AdjustedStateToUnitaryWithEigenproblem(new ApplyFunToGramMatrix(nC,nX,u_orig));
    }

    /** Adjust all singular values to 1.
     * The result should be the same as the one from {@link #getEVAdjustedTo1()}.
     * This method is typically used for only for unit tests.
     * @see #getEVAdjustedTo1()
     */
    public AdjustedStateToUnitaryWithSVD getSVDAdjustedTo1(){
	return getSVDAdjustedPartially(1.0);
    }


    static java.util.List<KGOEVSelection> selectTopEV(final int ntop,final double [] ev,final int nC,final int nX){
	final int n=nC*nX;
	// for now -- try max ntop eigenvalues
	final var a=new java.util.ArrayList<KGOEVSelection>();
	for(int k=n,ne=(n-ntop);--k>=ne;){
	    if(k>=0) a.add(
		new KGOEVSelection(k,EVXData.getPsiFromBasis(k,ev,n),ev[k],nC,nX)
			   );
	}
	return a;
    }


    // ========== !!! BELOW ARE UNIT TESTS AND OLD VERSIONS !!! ==============

    /** Adjust singular values partially.
     * @param adjustmentLevel A value between [0:1],
     * 0 -- do not adjust, 1 -- adjust all to 1, for other
     * values they are set to Sigma*(1-adjustmentLevel)+1*(adjustmentLevel),
     * then normalized to nC=sum Sigma^2.
     * @return Adjusted SVD.
     */
    private AdjustedStateToUnitaryWithSVD getSVDAdjustedPartially(final double adjustmentLevel){
	final double eps=TestKGO.EPS_selftest;

	// to have the norm equals to Gram matrix spur
	final double [] u_orig=TMatr.getVectorScaled(eVector,Math.sqrt(nC));
	final var svd_orig=EV.getSVD(nC,nX,u_orig);

	// adjust all SVD according to adjustmentLevel given
	final double [] S_adj=new double [svd_orig.S.length];
	for(int j=0;j<S_adj.length;j++){
	    S_adj[j]=svd_orig.S[j]*(1-adjustmentLevel)+1*(adjustmentLevel);
	}
	// scale to full dimension
	TMatr.scaleVectorInPlace(S_adj,Math.sqrt(nC/TMatr.mulVV(S_adj,S_adj)));

	final var svd_adj=new EVSolver.SVDResult(
	    svd_orig.m,
	    svd_orig.n,
	    svd_orig.U,
	    S_adj,
	    svd_orig.VT,
	    svd_orig.flagFailure);
	final double [] u_adj=AdjustedStateToUnitaryWithSVD.getMatrixConstructedWithAdjustedSingularValues(svd_orig,svd_adj.S);

	if(eps>0){
	    UnitTests.checkDoubleRel1(TMatr.mulVV(svd_adj.S,svd_adj.S),"sum S^2",nC,"nC",eps);
	    UnitTests.checkDoubleArrayRel1(u_orig,"u_orig",AdjustedStateToUnitaryWithSVD.getMatrixConstructedWithAdjustedSingularValues(svd_orig,svd_orig.S),"u_orig from SVD",eps);
	    UnitTests.checkDoubleRel1(TMatr.mulVV(u_orig,u_orig),"<u_orig|u_orig>",nC,"nC",eps);
	    UnitTests.checkDoubleRel1(TMatr.mulVV(u_adj,u_adj),"<u_adj|u_adj>",nC,"nC",eps);
	}

	return new AdjustedStateToUnitaryWithSVD(svd_orig,u_orig,svd_adj,u_adj);
    }



    /** Adjust all singular values to +-1, find the best combination of signs.
     * For not very bad initial state the best result corresponds to
     * all signs 1 (the same is with all -1), as this is a minimal adjustment
     * of all positive singular values returned by a SVD solver.
     * There are \( 2^{nC} \) check of +-1 combinations what may be really slow.
     */
    AdjustedStateToUnitaryWithSVD getSVDAdjustedToPlusMinus1(final double [] SK){
	final int n=nC*nX;
	if(SK.length!=n*n) throw new IllegalArgumentException();
	final double eps=TestKGO.EPS_selftest;

	// to have the norm equals to Gram matrix spur
	final double [] u_orig=TMatr.getVectorScaled(eVector,Math.sqrt(nC));
	final var svd_orig=EV.getSVD(nC,nX,u_orig);
	final double [] SKinS=AdjustedStateToUnitaryWithSVD.convertFunctionalFromOriginalToSigmaBasis(svd_orig,SK);

	if(eps>0){
	    final double Fun=TMatr.getAMatrA(u_orig,SK,n)/nC,
		FunS=TMatr.getAMatrA(svd_orig.S,SKinS,nC)/nC;
	    // check SVD expansion is OK
	    UnitTests.checkDoubleRel1(Fun,"Fun",FunS,"FunS",eps);
	    System.err.println("SKinS Match OK");
	    final double [] evSKinS=EV.getEV(nC,SKinS,false);
	    // if this is global max
	    //UnitTests.checkDoubleRel1(evSKinS[nC-1],"evSKinS[nC-1]",FunS,"FunS",eps);
	    System.err.println(" SK eigenvalues in S:"+java.util.Arrays.toString(evSKinS)+" sum="+TMatr.sumElements(evSKinS)+" FunS="+FunS+" Sigmas="+java.util.Arrays.toString(svd_orig.S));
	}

	// adjust all SVD to +=1, find the best combination of signs
	class ScanRecursively {
	    final double [] S_adj=new double [svd_orig.S.length];
	    final double [] s_signs=new double [svd_orig.S.length];
	    {
		java.util.Arrays.fill(s_signs,1);
		System.arraycopy(s_signs,0,S_adj,0,svd_orig.S.length);
	    }
	    double Fmax=TMatr.getAMatrA(S_adj,SKinS,nC)/nC;

	    void next(int ind){
		if(ind>=nC){
		    final double fun=TMatr.getAMatrA(s_signs,SKinS,nC)/nC;
		    System.err.println(java.util.Arrays.toString(s_signs)+" FUNs="+fun+" nC="+nC);
		    if(fun>=Fmax){
			Fmax=fun;
			System.arraycopy(s_signs,0,S_adj,0,svd_orig.S.length);
		    }
		    return;
		}
		s_signs[ind]=1;
		next(ind+1);
		s_signs[ind]=-1;
		next(ind+1);
	    }
	}
	final var sr=new ScanRecursively();
	// Start scanning with 1 to have first sign always 1
	// as the result is invariant with respect
	// to sign change of all singular values. Twice fewer checks.
	sr.next(1);

	final var svd_adj=new EVSolver.SVDResult(
	    svd_orig.m,
	    svd_orig.n,
	    svd_orig.U,
	    sr.S_adj,
	    svd_orig.VT,
	    svd_orig.flagFailure);

	// reconstruct u_adj from adjusted singular values
	final double [] u_adj=AdjustedStateToUnitaryWithSVD.getMatrixConstructedWithAdjustedSingularValues(svd_orig,svd_adj.S);

	if(eps>0){
	    UnitTests.checkDoubleArrayRel1(u_orig,"u_orig",AdjustedStateToUnitaryWithSVD.getMatrixConstructedWithAdjustedSingularValues(svd_orig,svd_orig.S),"u_orig from SVD",eps);
	    final double [] s1=new double[nC];java.util.Arrays.fill(s1,1);
	    final double Finit=TMatr.getAMatrA(s1,SKinS,nC)/nC;
	    UnitTests.checkDoubleRel1(TMatr.mulVV(u_orig,u_orig),"<u_orig|u_orig>",nC,"nC",eps);
	    UnitTests.checkDoubleRel1(TMatr.mulVV(u_adj,u_adj),"<u_adj|u_adj>",nC,"nC",eps);
	    System.err.println("Selected signs="+java.util.Arrays.toString(sr.S_adj)+" F="+sr.Fmax+" Finit="+Finit+" of signs="+java.util.Arrays.toString(s1));
	}

	return new AdjustedStateToUnitaryWithSVD(svd_orig,u_orig,svd_adj,u_adj);
    }


    /** Do not adjust singular values.
     */
    public AdjustedStateToUnitaryWithSVD getSVDNonAdjusted(){
	return getSVDAdjustedPartially(0.0);
    }

    private static final EVSolver EV=KGOSolutionVectorXVectorF.EV;
}
