package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;

/**
 * A Knowledge Generalized Operator iteration optimization,
 * a different version of iteration.
 * Simple version of constraints:
 * we use a special \( Q_{ij} \)
 * in the denominator \( \sum\limits_{i,j=0}^{nC-1}\sum\limits_{k=0}^{nX-1}Q_{ij}u_{ik}u_{jk} \).
 * The \( Q_{ij} \) is calculated from \( \lambda_{ij} \)
 * of previous  iterations,
 * in the simplest case it is selected
 * as  \( Q=const \lambda_{ij} \)
 * corresponding to the best global maximum candidate.
 * This version can combine together
 * {@link KGOIterationalLagrangeMultipliersInDenominatorU}
 * that uses \( u_{ik}Q_{ij}u_{jk} \) in denominator and
 * {@link KGOIterationalSimpleOptimizationU}
 * that adds \( u_{ik}\lambda_{ij}u_{jk} \) term.
 * This class can combine them together.
 * @version $Id: KGOIterationalGEVAdj_Q_denominator.java,v 1.47 2024/05/04 13:23:57 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalGEVAdj_Q_denominator {
    private static final int N_iterations=40;
    private static final boolean FORCE_DIAG_DUMP=true;


    public final KGOEVSelection evSelected;
    public final UAdjustment aep;
    public KGOIterationalGEVAdj_Q_denominator(final int nC,
					      final int nX,
					      final double [] SK,
					      final double eps)
{
    final int n=nC*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    //if(true) throw new RuntimeException();


    final double [] SKwithLambda=new double [n*n];
    final java.util.function.Function<double [],double []> adjSKWithLambda=curLambda->{
	// construct SK adjusted with lambda
	System.arraycopy(SK,0,SKwithLambda,0,SK.length);
	for(int j=0;j<nC;j++){
	for(int jp=0;jp<nC;jp++){
	    final double l=curLambda[j*nC+jp];
	    for(int k=0;k<nX;k++){
		SKwithLambda[(j*nX+k)*n+(jp*nX+k)]-=l;
	    }}
	}
	return SKwithLambda;
    };

    class LocalMax {
	final double Fadj,Forig;
	final KGOEVSelection e;
	final AdjustedStateToUnitaryWithEigenproblem aep;
	final int itNum;
	LocalMax(final KGOEVSelection e,int itNum){
	    this.e=e;
	    final double [] u_orig=TMatr.getVectorScaled(e.eVector,Math.sqrt(nC)/Math.sqrt(TMatr.mulVV(e.eVector,e.eVector)));
	    this.aep=new AdjustedStateToUnitaryWithEigenproblem(new ApplyFunToGramMatrix(nC,nX,u_orig));
	    this.Fadj=TMatr.getAMatrA(aep.getUAdj(),SK,n);
	    this.Forig=TMatr.getAMatrA(aep.getUOrig(),SK,n);
	    this.itNum=itNum;

	}
	boolean isValid(){return Double.isFinite(Fadj);}
	boolean isArgLocalBetter(LocalMax s){
	    // for now
	    return s.Fadj>=this.Fadj;
	}
	boolean isArgGlobalBetter(LocalMax s){
	    return s.Fadj>=this.Fadj;
	}
    }

    double [] lambda=new double [nC*nC],Q=TMatr.getUnitMatrix(nC);
    LocalMax globMax=null,prevLocMax=null;
    for(int it=0;it<N_iterations;it++){
	// solve regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getGEV_Generalized(
	    n,
	    adjSKWithLambda.apply(lambda),
	    AdjustedStateToUnitaryWithEigenproblem.createSKFromGram(nC,nX,Q),true);
	System.err.println("EV evSKLambda="+java.util.Arrays.toString(java.util.Arrays.copyOf(evSKLambda,n)));

	// BEGIN select the eigenvector among several
	LocalMax locMax=null;
	boolean flagGlobalMax=false;
	final var evAdj=new java.util.ArrayList<LocalMax>();
	int nNegLambda=0;
	// select eigenvalue
	for(var e:KGOEVSelection.selectTopEV(n, // n -- try all ; try top 3?
					     evSKLambda,nC,nX)){
	    if(e.eLambda<=0) nNegLambda++;
	    // Try all eigenstates, including all with negative eigenvalue
	    if(it<=0 ||  true || locMax==null){
		final var s=new LocalMax(e,it);
		evAdj.add(s);
		if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(s)){
		    // found global max (across iterations)
		    globMax=locMax=s;
		    flagGlobalMax=true;
		    //Q=getLambda.apply(s.aep.getUAdj());
		}
		else if(!flagGlobalMax && (
			    locMax==null || !locMax.isValid() || locMax.isArgLocalBetter(s)
					   )){
		    // some local max
		    locMax=s;
		}

		if(eps>0){
		    // check projection max
		    final double uOrigUadj=TMatr.mulVV(s.aep.getUAdj(),s.aep.getUOrig());
		    if(true){
			double sumSQRT=0;
			for(int k=0;k<nC;k++) sumSQRT+=Math.sqrt(s.aep.gram.ev_m[k]);
			UnitTests.checkDoubleRel1(sumSQRT,"sumSQRT",uOrigUadj,"uOrigUadj",eps);
		    }
		    UnitTests.checkDoubleRel(TMatr.getSpurM(nC,s.aep.gram.m),"spur G",nC,"nC",eps);
		    TestKGO.testOrthogonality("Adj Orthogonality",nC,s.aep.getUAdj(),nX,eps);
		    System.err.println("it="+it+":Try e.ind="+e.ind+" ("+n+") e.eLambda="+e.eLambda+" Forig="+s.Forig+" Fadj="+s.Fadj+" sum1Lam="+s.aep.sum1Lam+" <uOrig|Uadj>="+uOrigUadj+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(s.aep.gram.ev_m,nC)));
		}
	    }
	}
	// END select the eigenvector among several
	// A different way to select eigenvalues
	// largers Fadj on top
	java.util.Collections.sort(evAdj,(a,b)->Double.compare(b.Fadj,a.Fadj));

	if(eps>0 && prevLocMax!=null){
	    // Diag convergence.
	    System.err.println("it="+it+":CheckPrev <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+
			       " <prev_orig|orig>="+TMatr.mulVV(prevLocMax.aep.getUOrig(),locMax.aep.getUOrig())+
			       " <prev_adj|adj>="+TMatr.mulVV(prevLocMax.aep.getUAdj(),locMax.aep.getUAdj())+
			       " Forig="+TMatr.getAMatrA(locMax.aep.getUOrig(),SK,n)+
			       " prev_Forig="+TMatr.getAMatrA(prevLocMax.aep.getUOrig(),SK,n)+
			       " Fadj="+TMatr.getAMatrA(locMax.aep.getUAdj(),SK,n)+
			       " prev_Fadj="+TMatr.getAMatrA(prevLocMax.aep.getUAdj(),SK,n));
	}
	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+":Selected ind="+locMax.e.ind+" ("+n+") eLambda="+locMax.e.eLambda+"; Forig="+locMax.Forig+" Fadj="+locMax.Fadj+" sum1Lam="+locMax.aep.sum1Lam+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	// BEGIN calculate new values of Lagrange multipliers

	if(true){
	    final double [][] u=new double [1][],
		vBasis=new double [Math.min(evAdj.size(),nC*(nC+1)/2+0)][];
	    for(int p=0;p<vBasis.length;p++){
		if(p<u.length) u[p]=evAdj.get(p).aep.getUAdj();
		/*if(p<u.length) vBasis[p]=evAdj.get(p).aep.getUOrig();
		  else vBasis[p]=evAdj.get(p).aep.getUOrig();*/
		vBasis[p]=evAdj.get(p).aep.getUOrig();
	    }
	    lambda=LagrangeMultipliersPartialSubspace.getLambdaForSubspace(nC,nX,SK,u[0],vBasis);
	    if(eps>0){
		System.err.println("KGOIterationalGEVAdj_Q_denominator:lambda ev="+java.util.Arrays.toString(KGOSolutionVectorXVectorF.EV.getEV(nC,lambda,false))+"");

	    }
	    // check positively definite
	    final double [] ev_lam=KGOSolutionVectorXVectorF.EV.getEV(nC,lambda,false);
	    for(int k=0;k<ev_lam.length;k++){
		if(!(ev_lam[k]>0)){
		    lambda=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,locMax.aep.getUAdj());
		    break;
		}
	    }
	} else {
	    lambda=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,locMax.aep.getUAdj());
	    if(eps>0){
		// test lambda
		// the value of FadjLambda is zero as lambda spur is equal to Fadj
		final double FadjLambda=TMatr.getAMatrA(locMax.aep.getUAdj(),adjSKWithLambda.apply(lambda),n);
		UnitTests.checkDoubleRel1(FadjLambda,"FadjLambda",0,"0",eps);
		UnitTests.checkDoubleRel1(locMax.Fadj,"Fadj",TMatr.getSpurM(nC,lambda),"spur(lambda)",eps);
	    }
	    if(eps>0){
		final var evlam=new ApplyFunToMatrix(nC,lambda);
		final double [] mtransf=java.util.Arrays.copyOfRange(evlam.ev_m,nC,evlam.ev_m.length);
		final double [] v=TMatr.mul_MA_x_MB(nC,mtransf,nC,locMax.aep.getUAdj(),nX);
		final double [] SKv=AdjustedStateToUnitaryWithEigenproblem.convertSKtoNewBasis(nC,nX,SK,mtransf/*TMatr.getMatrTrans(nC,mtransf)*/);
		TestKGO.testOrthogonality("Adj Orthogonality",nC,v,nX,eps);
		UnitTests.checkDoubleRel(locMax.Fadj,"Fadj",TMatr.getAMatrA(v,SKv,n),"<v|SKv|v>",eps);
		final double [] lambdaV=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SKv,v);
		// test the matrix is diag in lambda basis
		UnitTests.checkDoubleArrayRel(java.util.Arrays.copyOf(evlam.ev_m,nC),"evLam",TMatr.getMatrixDiagElements(nC,lambdaV),"diag(lambdaV)",eps);
		UnitTests.checkDoubleArrayRel1(TMatr.sumAcB(lambdaV,-1,TMatr.getDiagMatrix(TMatr.getMatrixDiagElements(nC,lambdaV))),"offdiag lambdaV",new double [nC*nC],"0",eps);
		//System.err.println("KGOIterationalGEVAdj_Q_denominator: lambdaV="+java.util.Arrays.toString(lambdaV)+" ev="+java.util.Arrays.toString(java.util.Arrays.copyOf(evlam.ev_m,nC)));
		//final double [] vt=TMatr.mul_MATrans_x_MB(nC,mtransf,nC,locMax.aep.getUAdj(),nX);
		//System.err.println("KGOIterationalGEVAdj_Q_denominator: Fadj="+TMatr.getAMatrA(v,SKv,n)+" Fv="+TMatr.getAMatrA(v,SK,n)+" vt="+TMatr.getAMatrA(vt,SK,n));
	    }
	}
	// END calculate new values of Lagrange multipliers

	if(eps>0){
	    System.err.println("KGOIterationalGEVAdj_Q_denominator: it="+it+" Fadj="+locMax.Fadj+" Forig="+locMax.Forig+" ind="+locMax.e.ind+" ("+n+") <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	prevLocMax=locMax;
	if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(locMax)) {
	    globMax=locMax;
	    //Q=lambda;
	}
	Q=lambda;
	//Q=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,locMax.aep.getUAdj());
	//Q=com.polytechnik.utils.Linsystems.getInvertedMatrix(nC,nC,lambda);
	//Q=new ApplyFunToMatrix(nC,lambda).applyFun((ind,lam)->Math.pow(lam,0));
	if(eps>0){
//	    final var evinvGram=new ApplyFunToMatrix(nC,com.polytechnik.utils.Linsystems.getInvertedMatrix(nC,nC,locMax.aep.gram.m));
//	    final double [] mtransf=java.util.Arrays.copyOfRange(evinvGram.ev_m,nC,evinvGram.ev_m.length);
//	    final double [] v=
	}
    }
    evSelected=globMax.e;
    aep=globMax.aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalGEVAdj_Q_denominator: RESULT:Selected ind="+evSelected.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" at iteration="+globMax.itNum+" N_iterations="+N_iterations);
    }
}

}
