package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;

/**
 * A Knowledge Generalized Operator iteration optimization,
 * a try with Lagrange multipliers in the denominator;
 * calculations directly in u_orig basis.
 * Since the variation of Rayleigh quotients
 * $$
\mathcal{F}=\frac{
\sum\limits_{j,j^{\prime}=0}^{nC-1}\sum\limits_{k,k^{\prime}=0}^{nX-1}
u_{jk}
S_{jk;j^{\prime}k^{\prime}}
u_{j^{\prime}k^{\prime}}}
{
\sum\limits_{j,j^{\prime}=0}^{nC-1}\sum\limits_{k=0}^{nX-1}
u_{jk}Q_{jj^{\prime}}u_{j^{\prime}k}
}
\xrightarrow[{u}]{\quad }\max
$$
gives
$$
0=\sum\limits_{j^{\prime}=0}^{nC-1}\sum\limits_{k^{\prime}=0}^{nX-1}
S_{sq;j^{\prime}k^{\prime}}u_{j^{\prime}k^{\prime}}
- \mathcal{F} \sum\limits_{j^{\prime}=0}^{nC-1} Q_{sj^{\prime}} u_{j^{\prime}q}
$$
one can use Lagrange multipliers \(\lambda_{ij}\)
in denomination in place of \(Q_{ij}\)
instead of considering regular
$$
\mathcal{L}=\sum\limits_{j,j^{\prime}=0}^{nC-1}\sum\limits_{k,k^{\prime}=0}^{nX-1}
u_{jk}
S_{jk;j^{\prime}k^{\prime}}
u_{j^{\prime}k^{\prime}}
+   \sum\limits_{j,j^{\prime}=0}^{nC-1}
   \lambda_{jj^{\prime}}\left[\delta_{jj^{\prime}}-\sum\limits_{k^{\prime}=0}^{nX-1}u_{jk^{\prime}} u_{j^{\prime}k^{\prime}} \right]
$$
 * The result is almost the same as with {@link KGOIterationalSimpleOptimizationU}
 * wich uses regular Lagrangian approach.
 * @see KGOIterationalSimpleOptimizationU
 * @see KGOIterationalGEVAdj_Q_denominator
 * @version $Id: KGOIterationalLagrangeMultipliersInDenominatorU.java,v 1.50 2024/04/25 22:17:36 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalLagrangeMultipliersInDenominatorU {
    private static final int N_iterations=40;
    private static final boolean FORCE_DIAG_DUMP=true;

    public final KGOEVSelection evSelected;
    public final UAdjustment aep;
    public KGOIterationalLagrangeMultipliersInDenominatorU(
	final int nC,final int nX,
	final double [] SK,
	final double eps)
{
    final int n=nC*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    //if(true) throw new RuntimeException();

    final java.util.function.Function<double [],double []> getMatrixInDenominator=denominatorLambda->{
	return AdjustedStateToUnitaryWithEigenproblem.createSKFromGram(nC,nX,denominatorLambda);
    };

    final java.util.function.Function<double [],double []> getLambdaInDenominator=u_at->{
	// need only Hermitian parts
	final double [] denomScalProduct=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,u_at);
	return TMatr.scaleVectorInPlace(denomScalProduct,nC/TMatr.getSpurM(nC,denomScalProduct));
    };

    class LocalMax {
	final double Fadj,Forig;
	final KGOEVSelection e;
	final AdjustedStateToUnitaryWithEigenproblem aep;
	final int itNum;
	LocalMax(final KGOEVSelection e,int itNum){
	    this.e=e;
	    final double [] u_orig=TMatr.getVectorScaled(e.eVector,Math.sqrt(nC)/Math.sqrt(TMatr.mulVV(e.eVector,e.eVector)));
	    this.aep=new AdjustedStateToUnitaryWithEigenproblem(new ApplyFunToGramMatrix(nC,nX,u_orig));
	    this.Fadj=TMatr.getAMatrA(aep.getUAdj(),SK,n);
	    this.Forig=TMatr.getAMatrA(aep.getUOrig(),SK,n);
	    this.itNum=itNum;

	}
	boolean isValid(){return Double.isFinite(Fadj);}
	boolean isArgLocalBetter(LocalMax s){
	    // for now
	    return s.Fadj>=this.Fadj;
	}
	boolean isArgGlobalBetter(LocalMax s){
	    return s.Fadj>=this.Fadj;
	}
    }

    double [] denominator=TMatr.getUnitMatrix(nC);
    LocalMax globMax=null,prevLocMax=null;
    for(int it=0;it<N_iterations;it++){
	// solve generalized regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getGEV_Generalized(n,SK,getMatrixInDenominator.apply(denominator),true);
	//System.err.println("EV evSKLambda="+java.util.Arrays.toString(java.util.Arrays.copyOf(evSKLambda,n)));

	// BEGIN select the eigenvector among several
	LocalMax locMax=null;
	boolean flagGlobalMax=false;
	// select eigenvalue
	for(var e:KGOEVSelection.selectTopEV(n, /* n -- try all ; try top 3*/evSKLambda,nC,nX)){
	    if(it<=0 || true || locMax==null){
		final var s=new LocalMax(e,it);
		if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(s)){
		    // found global max (across iterations)
		    globMax=locMax=s;
		    flagGlobalMax=true;
		}
		else if(!flagGlobalMax && (
			    locMax==null || !locMax.isValid() || locMax.isArgLocalBetter(s)
					   )){
		    // some local max
		    locMax=s;
		}

		if(eps>0){
		    /*
		      final var rq=new RayleighQuotient.RQVariationInMatrixForm(s.aep.getUAdj(),n,SK,getMatrixInDenominator.apply(getLambdaInDenominator.apply(s.aep.getUAdj())),n);
		      final double [] iD2=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,rq.d2);
		      final double varD= -0.5*TMatr.getAMatrA(rq.d1,iD2,n);
		      System.err.println("it="+it+" e.ind="+e.ind+" varD="+varD+" sumd1^2="+TMatr.mulVV(rq.d1,rq.d1));
		    */
		    //UnitTests.checkDoubleRel1(TMatr.mulVV(s.aep.getUOrig(),s.aep.getUOrig()),"UOrig GramDiag",nC,"nC",eps);
		    //UnitTests.checkDoubleRel1(TMatr.getAMatrA(s.aep.getUOrig(),getMatrixInDenominator.apply(denominator),n),"UOrig Gram",nC,"nC",eps);
		    UnitTests.checkDoubleRel(TMatr.getSpurM(nC,s.aep.gram.m),"spur G",nC,"nC",eps);
		    TestKGO.testOrthogonality("Adj EV ",nC,s.aep.getUAdj(),nX,eps);
		    System.err.println("it="+it+" e.ind="+e.ind+" e.eLambda="+e.eLambda+" Forig="+s.Forig+" Fadj="+s.Fadj+" <u_orig|u_adj>="+TMatr.mulVV(s.aep.getUAdj(),s.aep.getUOrig())+" GEV="+java.util.Arrays.toString(java.util.Arrays.copyOf(s.aep.gram.ev_m,nC)));
		}
	    }
	}
	// END select the eigenvector among several

	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+":Selected ind="+locMax.e.ind+" ("+n+") eLambda="+locMax.e.eLambda+"; Forig="+locMax.Forig+" Fadj="+locMax.Fadj+" sum1Lam="+locMax.aep.sum1Lam+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	// BEGIN calculate new values of Lagrange multipliers
	denominator=getLambdaInDenominator.apply(locMax.aep.getUAdj());
	// END calculate new values of Lagrange multipliers

	if(eps>0){
	    System.err.println("KGOIterationalLagrangeMultipliersInDenominatorU: it="+it+" Fadj="+locMax.Fadj+" Forig="+locMax.Forig+" ind="+locMax.e.ind+" ("+n+") <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	prevLocMax=locMax;
	if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(locMax)) {
	    globMax=locMax;
	}
    }
    evSelected=globMax.e;
    aep=globMax.aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalLagrangeMultipliersInDenominatorU: RESULT:Selected ind="+evSelected.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" at iteration="+globMax.itNum+" N_iterations="+N_iterations);
    }
}

}
