package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;

/**
 * A Knowledge Generalized Operator iteration optimization,
 * partial space Lagrange multipliers calculation.
 * @see LagrangeMultipliersPartialSubspace#getLambdaForSubspace
 * @version $Id: KGOIterationalLagrangeMultipliersPartialSubspace.java,v 1.65 2024/05/04 13:23:57 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalLagrangeMultipliersPartialSubspace {
    private static final int N_iterations=40;
    private static final boolean FORCE_DIAG_DUMP=true;

    public final KGOEVSelection evSelected;
    public final UAdjustment aep;
    public KGOIterationalLagrangeMultipliersPartialSubspace(
	final int nC,final int nX,
	final double [] SK,
	final double eps)
{
    final int n=nC*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    //if(true) throw new RuntimeException();

    final double [] SKwithLambda=new double [n*n];
    final java.util.function.Function<double [],double []> adjSKWithLambda=curLambda->{
	// construct SK adjusted with lambda
	System.arraycopy(SK,0,SKwithLambda,0,SK.length);
	for(int j=0;j<nC;j++){
	for(int jp=0;jp<nC;jp++){
	    final double l=curLambda[j*nC+jp];
	    for(int k=0;k<nX;k++){
		SKwithLambda[(j*nX+k)*n+(jp*nX+k)]-=l;
	    }}
	}
	return SKwithLambda;
    };

    class LocalMax {
	final double Fadj,Forig;
	final KGOEVSelection e;
	final AdjustedStateToUnitaryWithEigenproblem aep;
	final int itNum;
	LocalMax(final KGOEVSelection e,int itNum){
	    this.e=e;
	    this.aep=e.getEVAdjustedTo1();
	    this.Fadj=TMatr.getAMatrA(aep.getUAdj(),SK,n);
	    this.Forig=TMatr.getAMatrA(aep.getUOrig(),SK,n);
	    this.itNum=itNum;

	}
	boolean isValid(){return Double.isFinite(Fadj);}
	boolean isArgLocalBetter(LocalMax s){
	    // for now
	    return s.Fadj>=this.Fadj;
	    //return s.aep.sum1Lam<=this.aep.sum1Lam;
	}
	boolean isArgGlobalBetter(LocalMax s){
	    return s.Fadj>=this.Fadj;
	}
    } // END OF class LocalMax

    class CalculateLambdaBasisSubset {
	final LocalMax loc;
	final double [] lambda;
	CalculateLambdaBasisSubset(
	    final LocalMax loc,
	    final double [] evSKLambda,
	    final double [] lambdaOld,
	    final int nToSelect){

	    this.loc=loc;
	    final double [] u=loc.aep.getUAdj();
	    final double muU=(loc.Fadj-TMatr.getSpurM(nC,lambdaOld))/nC;
	    // calculate b=var L/var u
	    final double [] b=TMatr.mul_MA_x_MB(n,SK,n,u,1);
	    TMatr.sumP1(b,TMatr.mul_MA_x_MB(nC,lambdaOld,nC,u,nX),-1);
	    TMatr.sumP1(b,u,-muU);

	    if(eps>0){
		final double [] SKLm=adjSKWithLambda.apply(TMatr.sumAcB(lambdaOld,muU,TMatr.getUnitMatrix(nC)));
		UnitTests.checkDoubleRel1(TMatr.getAMatrA(u,SKLm,n),"<u|SKLm|u>",0,"0",eps);
		UnitTests.checkDoubleArrayRel1(TMatr.mul_MA_x_MB(n,SKLm,n,u,1),"|SKLm|u>",b,"b",eps);
		System.err.println("labmdaOld matrix eigenvalues="+java.util.Arrays.toString(KGOSolutionVectorXVectorF.EV.getEV(nC,lambdaOld,false))+" (degeneracy diag)");
	    }

	    class BProjV {
		final double [] v;
		final double projV;
		final int indV;
		BProjV(final int indV){
		    this.indV=indV;
		    this.v=EVXData.getPsiFromBasis(indV,evSKLambda,n);
		    this.projV=TMatr.mulVV(b,v);
		}
	    } // END OF class BProjV

	    final var arr=new BProjV [n];
	    for(int p=0;p<n;p++){arr[p]=new BProjV(p);}
	    // Largest on top
	    // take abs for sort, less propably to underflow that squared
	    //java.util.Arrays.sort(arr,(ax,bx)->Double.compare(Math.abs(bx.projV),Math.abs(ax.projV)));
	    java.util.Arrays.sort(arr,(ax,bx)->Double.compare(evSKLambda[bx.indV],evSKLambda[ax.indV]));

	    if(eps>0){
		double sumProj2=0,sumFadjWithLambda=0;
		double [] sumB=new double [n];
		for(int p=0;p<n;p++){
		    final double uproj2=Math.pow(TMatr.mulVV(u,arr[p].v),2);
		    sumFadjWithLambda += evSKLambda[arr[p].indV]*uproj2;
		    sumProj2 += uproj2;
		    TMatr.sumP1(sumB,arr[p].v,arr[p].projV);
		    System.err.println("p="+p+" indV="+arr[p].indV+" uproj2="+uproj2+" projV="+arr[p].projV);
		}
		UnitTests.checkDoubleRel(sumProj2,"sumProj2",TMatr.mulVV(u,u),"<u|u>",eps);
		UnitTests.checkDoubleRel(sumFadjWithLambda,"sumFadjWithLambda",nC*muU,"nC*muU",eps);
		UnitTests.checkDoubleArrayRel1(sumB,"sumB",b,"b",eps);
	    }
	    final double [][] vBasis=new double [nToSelect+1][];
	    for(int p=0,o=0;p<vBasis.length;p++){
		if(p<=0) vBasis[p]=u;
		else vBasis[p]=arr[o++].v;
	    }
	    this.lambda=LagrangeMultipliersPartialSubspace.getLambdaForSubspace(nC,nX,SK,u,vBasis);
	}
    } // END OF class CalculateLambdaBasisSubset


    double [] lambda=new double [nC*nC];
    LocalMax globMax=null,prevLocMax=null;
    for(int it=0;it<N_iterations;it++){
	// solve regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getEV(n,adjSKWithLambda.apply(lambda),true);
	//System.err.println("EV evSKLambda="+java.util.Arrays.toString(java.util.Arrays.copyOf(evSKLambda,n)));

	// BEGIN select the eigenvector among several
	LocalMax locMax=null;
	boolean flagGlobalMax=false;
	final var evAdj=new java.util.ArrayList<LocalMax>();
	int nNegLambda=0;
	// select eigenvalue
	for(var e:KGOEVSelection.selectTopEV(n, // n -- try all ; try top 3?
					     evSKLambda,nC,nX)){
	    if(e.eLambda<=0) nNegLambda++;
	    // How many negaive lambda to consider? Try 4.
	    if(it<=0 || true || nNegLambda<=4 || locMax==null){
		final var s=new LocalMax(e,it);
		evAdj.add(s);
		if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(s)){
		    // found global max (across iterations)
		    globMax=locMax=s;
		    flagGlobalMax=true;
		}
		else if(!flagGlobalMax && (
			    locMax==null || !locMax.isValid() || locMax.isArgLocalBetter(s)
					   )){
		    // some local max
		    locMax=s;
		}

		if(eps>0){
		    // check projection max
		    final double uOrigUadj=TMatr.mulVV(s.aep.getUAdj(),s.aep.getUOrig());
		    if(true){
			double sumSQRT=0;
			for(int k=0;k<nC;k++) sumSQRT+=Math.sqrt(s.aep.gram.ev_m[k]);
			UnitTests.checkDoubleRel1(sumSQRT,"sumSQRT",uOrigUadj,"uOrigUadj",eps);
		    }
		    TestKGO.testOrthogonality("Adj Orthogonality",nC,s.aep.getUAdj(),nX,eps);
		    System.err.println("it="+it+":Try e.ind="+e.ind+" ("+n+") e.eLambda="+e.eLambda+" Forig="+s.Forig+" Fadj="+s.Fadj+" sum1Lam="+s.aep.sum1Lam+" <uOrig|Uadj>="+uOrigUadj+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(s.aep.gram.ev_m,nC)));
		}
	    }
	}
	// END select the eigenvector among several
	// A different way to select eigenvalues
	// largers Fadj on top
	java.util.Collections.sort(evAdj,(a,b)->Double.compare(b.Fadj,a.Fadj));

	if(eps>0 && prevLocMax!=null){
	    // Diag convergence.
	    System.err.println("it="+it+":CheckPrev <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+
			       " <prev_orig|orig>="+TMatr.mulVV(prevLocMax.aep.getUOrig(),locMax.aep.getUOrig())+
			       " <prev_adj|adj>="+TMatr.mulVV(prevLocMax.aep.getUAdj(),locMax.aep.getUAdj())+
			       " Forig="+TMatr.getAMatrA(locMax.aep.getUOrig(),SK,n)+
			       " prev_Forig="+TMatr.getAMatrA(prevLocMax.aep.getUOrig(),SK,n)+
			       " Fadj="+TMatr.getAMatrA(locMax.aep.getUAdj(),SK,n)+
			       " prev_Fadj="+TMatr.getAMatrA(prevLocMax.aep.getUAdj(),SK,n));
	}
	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+":Selected ind="+locMax.e.ind+" ("+n+") eLambda="+locMax.e.eLambda+" Forig="+locMax.Forig+" Fadj="+locMax.Fadj+" sum1Lam="+locMax.aep.sum1Lam+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	// BEGIN calculate new values of Lagrange multipliers
	if(false){
	    lambda=new CalculateLambdaBasisSubset(locMax,evSKLambda,lambda,nC*(nC+1)/2+7).lambda;
	}
	else if(true){
	    final double [] u=evAdj.get(0).aep.getUAdj();
	    final double [][] vBasis=new double [Math.min(evAdj.size(),nC*(nC+1)/2+3)][];
	    for(int p=0;p<vBasis.length;p++){
		//if(p<1) vBasis[p]=u;
		/*if(p<u.length) vBasis[p]=evAdj.get(p).aep.getUOrig();
		  else vBasis[p]=evAdj.get(p).aep.getUOrig();*/
		/*if(p<=0)vBasis[p]=evAdj.get(p).aep.getUOrig();
		  else vBasis[p]=evAdj.get(p).aep.getUOrig();*/
		vBasis[p]=evAdj.get(p).aep.getUOrig();
	    }
	    final double [] uOrig=evAdj.get(0).aep.getUOrig(),uAdj=evAdj.get(0).aep.getUAdj();
	    System.err.println("<uOrig|uAdj>="+TMatr.mulVV(uOrig,uAdj)+" <uOrig|uOrig>="+TMatr.mulVV(uOrig,uOrig)+"  <uAdj|uAdj>="+TMatr.mulVV(uAdj,uAdj));
	    if(eps>0){
		// to print diag
		new CalculateLambdaBasisSubset(evAdj.get(0),evSKLambda,lambda,n/*nC*(nC+1)/2+6*/);
	    }
	    lambda=LagrangeMultipliersPartialSubspace.getLambdaForSubspace(nC,nX,SK,u,vBasis);
	} else if(true){
	    final double [][] u=new double [1][],
		vBasis=new double [Math.min(evAdj.size(),nC*(nC+1)/2+3)][];
	    for(int p=0;p<vBasis.length;p++){
		if(p<u.length) u[p]=evAdj.get(p).aep.getUAdj();
		/*if(p<u.length) vBasis[p]=evAdj.get(p).aep.getUOrig();
		  else vBasis[p]=evAdj.get(p).aep.getUOrig();*/
		vBasis[p]=evAdj.get(p).aep.getUOrig();
	    }
	    lambda=LagrangeMultipliersPartialSubspace.getLambdaForSubspace(nC,nX,SK,u[0],vBasis);
	} else if(false){
	    final double [] uAdj=evAdj.get(0).aep.getUAdj(),u=new double [n];
	    final double [][] vBasis=new double [Math.min(evAdj.size(),nC*(nC+1)/2+3)][];
	    for(int p=0;p<vBasis.length;p++){
		final var l=evAdj.get(p);
		final double [] v=l.e.eVector;
		TMatr.sumP1(u,v,TMatr.mulVV(v,uAdj));
		vBasis[p]=l.aep.getUOrig();
	    }
	    for(int p=0;p<evAdj.size();p++){
		final var evp=evAdj.get(p);
		System.err.println("ind="+evAdj.get(0).e.ind+" p="+p+" evp.ind="+evp.e.ind+" <u|uAdj>="+TMatr.mulVV(evp.e.eVector,uAdj));
	    }
	    System.err.println("<u|uAdj>="+TMatr.mulVV(u,uAdj)+" <u|u>="+TMatr.mulVV(u,u)+"  <uAdj|uAdj>="+TMatr.mulVV(uAdj,uAdj));

	    lambda=LagrangeMultipliersPartialSubspace.getLambdaForSubspace(nC,nX,SK,u,vBasis);
	} else if(false){
	    // a different approximation for lambda. Diagonal elements are the same.
	    // works worse than the standard one
	    lambda=AdjustedStateToUnitaryWithEigenproblem.getSKasMatrixInUState(nC,nX,SK,locMax.aep.getUAdj());
	} else {
	    // old version
	    lambda=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,locMax.aep.getUAdj());
	    if(eps>0){
		// test lambda
		// the value of FadjLambda is zero as lambda spur is equal to Fadj
		final double FadjLambda=TMatr.getAMatrA(locMax.aep.getUAdj(),adjSKWithLambda.apply(lambda),n);
		UnitTests.checkDoubleRel1(FadjLambda,"FadjLambda",0,"0",eps);
		UnitTests.checkDoubleRel1(locMax.Fadj,"Fadj",TMatr.getSpurM(nC,lambda),"spur(lambda)",eps);
		// test different method
		UnitTests.checkDoubleArrayRel1(lambda,"lambda",new CalculateLambdaBasisSubset(locMax,evSKLambda,lambda,n).lambda,"lambdaTest",eps);
	    }
	}
	// END calculate new values of Lagrange multipliers

	if(eps>0){
	    System.err.println("KGOIterationalLagrangeMultipliersPartialSubspace: it="+it+" Fadj="+locMax.Fadj+" Forig="+locMax.Forig+" ind="+locMax.e.ind+" ("+n+") <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	prevLocMax=locMax;
	if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(locMax)) {
	    globMax=locMax;
	}
    } // end iterations

    evSelected=globMax.e;
    aep=globMax.aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalLagrangeMultipliersPartialSubspace: RESULT:Selected ind="+evSelected.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" at iteration="+globMax.itNum+" N_iterations="+N_iterations);
    }
}

}
