package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.Linsystems;

/**
 * A Knowledge Generalized Operator iteration optimization,
 * a different version of lambda.
 * Simple version of constraints:
 * \( \sum\limits_{k=0}^{n-1}u_{ik}u_{jk}=\delta_{ij} \), for \(i,j=0\dots m-1\).
 * A try without SVD.
 * @version $Id: KGOIterationalLambda2.java,v 1.74 2024/05/01 09:26:34 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalLambda2 {
    private static final int N_iterations=40;
    private static final boolean FORCE_DIAG_DUMP=true;

    /** Lambda calculation method.
     */
    private static enum  LambdaCalculationMethod {
	MUL_ITSELF,
	MUL_U_ADJ,
	MUL_B,
	BB
	;
	static final LambdaCalculationMethod METHOD=MUL_ITSELF;//MUL_U_ADJ;//MUL_ITSELF; // MUL_B
    }

    public final KGOEVSelection evSelected;
    public final UAdjustment aep;
    public KGOIterationalLambda2(final int nC,final int nX,
				 final double [] SK,
				 final double eps)
{
    final int n=nC*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    //if(true) throw new RuntimeException();

    final double [] SKwithLambda=new double [n*n];
    final java.util.function.Function<double [],double []> adjSKWithLambda=curLambda->{
	// construct SK adjusted with lambda
	System.arraycopy(SK,0,SKwithLambda,0,SK.length);
	for(int j=0;j<nC;j++){
	for(int jp=0;jp<nC;jp++){
	    final double l=curLambda[j*nC+jp];
	    for(int k=0;k<nX;k++){
		SKwithLambda[(j*nX+k)*n+(jp*nX+k)]-=l;
	    }}
	}
	return SKwithLambda;
    };

    final java.util.function.Function<double [],double []> getLambda=(u_at)->{
	final double [] b=TMatr.mul_MATrans_x_MB(n,SK,n,u_at,1),
	bb=TMatr.mul_MA_x_MBTrans(nC,b,nX,b,nC),
	lambda=switch(LambdaCalculationMethod.METHOD){
	  case MUL_ITSELF -> {
	    // multiplication by itself (from lambda^2)
	    final double [] ev_bb=KGOSolutionVectorXVectorF.EV.getEV(nC,bb,true),
		diag_sqrt_EV=new double [nC*nC];
	    // all eigenvalues must be positive
	    for(int j=0;j<nC;j++){diag_sqrt_EV[j*nC+j]=Math.sqrt(ev_bb[j]);}
	    // obtained matrix is always Hermitian
	    final var lambdaNew=EVXData.convertMatrFromBtoQ_Basis(
		nC,
		diag_sqrt_EV,
		ev_bb,
		TMatr.getUnitMatrix(nC));

	    if(eps>0){
		// THIS IS OLD VERSION
		final double [] OLD_lambda=TMatr.selectHermitianPartInPlace(nC,TMatr.mul_MA_x_MBTrans(nC,u_at,nX,b,nC));
		//System.err.println("ev_SQRT="+java.util.Arrays.toString(TMatr.getMatrixDiagElements(nC,diag_sqrt_EV))+" SUM OLD="+TMatr.getSpurM(nC,OLD_lambda)+" SUM_NEW="+TMatr.getSpurM(nC,diag_sqrt_EV));
		UnitTests.checkDoubleArrayRel1(TMatr.selectAntiHermitianPartInPlace(nC,lambdaNew.clone()),"AH lambda",new double [nC*nC],"0",eps);
		UnitTests.checkDoubleArrayRel1(TMatr.mul_MA_x_MB(nC,lambdaNew,nC,lambdaNew,nC),"lambda^2",bb,"bb",eps);
		UnitTests.checkDoubleRel1(TMatr.getSpurM(nC,OLD_lambda),"spur OLD_lambda",TMatr.mulVV(u_at,b),"Fadj",eps);
	    }
	    yield lambdaNew;
	  }
	case MUL_U_ADJ -> {
	    // multiplication by u_adj
	    yield TMatr.selectHermitianPartInPlace(nC,TMatr.mul_MA_x_MBTrans(nC,u_at,nX,b,nC));
	}
	case MUL_B -> {
	    // multiplication by b_jk. works well
	    final double [] bu=TMatr.mul_MA_x_MBTrans(nC,b,nX,u_at,nC),
		invbu=Linsystems.getInvertedMatrix(nC,nC,bu);
	    yield TMatr.selectHermitianPartInPlace(nC,TMatr.mul_MA_x_MB(nC,invbu,nC,bb,nC));
	}
	case BB -> {
	    // bb
	    yield TMatr.getVectorScaled(bb,nC/TMatr.mulVV(u_at,b));
	}
	default -> throw new IllegalArgumentException("METHOD="+LambdaCalculationMethod.METHOD);
      };
      return lambda;
    };

    double [] lambda=new double [nC*nC];
//    if(false){
//	final var kgo=new KGOIterationalLinearConstraints(nC,nX,SK,eps);
//	lambda=getLambda.apply(kgo.aep.getUAdj());
//    }
    KGOEVSelection result_evsel=null;
    UAdjustment result_aep=null,prev_aep=null;
    double itmaxSK=0;
    for(int it=0;it<N_iterations;it++){
	// solve regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getEV(n,adjSKWithLambda.apply(lambda),true);
	//System.err.println("EV evSKLambda="+java.util.Arrays.toString(java.util.Arrays.copyOf(evSKLambda,n)));

	// BEGIN select the eigenvector among several
	double maxSKadj=0;
	KGOEVSelection evsel=null;
	UAdjustment aepsel=null;

	final var saveaepsel=new java.util.ArrayList<UAdjustment>();
	int nNegLambda=0;
	// select eigenvalue
	for(var e:KGOEVSelection.selectTopEV(n, /* n -- try all ; try top 3*/evSKLambda,nC,nX)){
	    if(e.eLambda<=0) nNegLambda++;
	    if(it<=0 || /* true || */ nNegLambda<=4 || evsel==null){
		final var s=e.getEVAdjustedTo1();
		final double curSKadj=TMatr.getAMatrA(s.getUAdj(),SK,n);
		if(evsel==null|| !Double.isFinite(maxSKadj) || curSKadj>=maxSKadj){
		    evsel=e;
		    aepsel=s;
		    maxSKadj=curSKadj;
		}

		if(eps>0){
		    saveaepsel.add(s);
		    final double curSKorig=TMatr.getAMatrA(s.getUOrig(),SK,n);
		    TestKGO.testOrthogonality("Adj SVD ",nC,s.getUAdj(),nX,eps);
		    System.err.println("it="+it+" e.ind="+e.ind+" e.eLambda="+e.eLambda+" Forig="+curSKorig+" Fadj="+curSKadj);
		}
	    }
	}
	// END select the eigenvector among several

	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+" Selected ind="+evsel.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aepsel.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aepsel.getUAdj(),SK,n));
	}
	if(eps>0 && prev_aep!=null){
	    // Diag convergence.
	    System.err.print("it="+it+" <u_orig|u_adj>="+TMatr.mulVV(aepsel.getUAdj(),aepsel.getUOrig()));
	    System.err.print(" <prev_orig|orig>="+TMatr.mulVV(prev_aep.getUOrig(),aepsel.getUOrig()));
	    System.err.print(" <prev_adj|adj>="+TMatr.mulVV(prev_aep.getUAdj(),aepsel.getUAdj()));
	    System.err.println();
	    for(int i=0;i<saveaepsel.size();i++){
		for(int j=i;i<saveaepsel.size();i++){
		    System.err.println("<o"+i+"|o"+j+">="+TMatr.mulVV(saveaepsel.get(i).getUOrig(),saveaepsel.get(j).getUOrig())+" <a"+i+"|a"+j+">="+TMatr.mulVV(saveaepsel.get(i).getUAdj(),saveaepsel.get(j).getUAdj()));
		}
	    }
	}

	// BEGIN calculate new values of Lagrange multipliers
	lambda=getLambda.apply(aepsel.getUAdj());
	// END calculate new values of Lagrange multipliers

	if(eps>0){
	    final double
		Fadj=TMatr.getAMatrA(aepsel.getUAdj(),SK,n),
		Forig=TMatr.getAMatrA(aepsel.getUOrig(),SK,n);
	    // test lambda
	    final double [] testSK=adjSKWithLambda.apply(lambda);
	    // the value of FadjLambda is zero as lambda spur is equal to Fadj
	    final double FadjLambda=TMatr.getAMatrA(aepsel.getUAdj(),testSK,n);
	    //UnitTests.checkDoubleRel1(FadjLambda,"FadjLambda",0,"0",eps);
	    //UnitTests.checkDoubleRel1(Fadj,"Fadj",TMatr.getSpurM(nC,lambda),"spur(lambda)",eps);

	    System.err.println("KGOIterationalLambda2: it="+it+" Fadj="+Fadj+" Forig="+Forig+" ind="+evsel.ind+" ("+n+") <u_orig|u_adj>="+TMatr.mulVV(aepsel.getUAdj(),aepsel.getUOrig()));
	}

	prev_aep=aepsel;
	// adjusted value is equal to lambda spur
	final double Fadj=TMatr.getAMatrA(aepsel.getUAdj(),SK,n);
	if(result_evsel==null || !Double.isFinite(itmaxSK) || Fadj>=itmaxSK) {
	    result_evsel=evsel;
	    result_aep=aepsel;
	    itmaxSK=Fadj;
	}
    }
    evSelected=result_evsel;
    aep=result_aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalLambda2: RESULT: Selected ind="+evSelected.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" N_iterations="+N_iterations+" LAMBDA SELECTION="+LambdaCalculationMethod.METHOD);
    }
}



}
