package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;

/**
 * A class to find KGO operator trying linear constraints.
 * @version $Id: KGOIterationalLinearConstraints.java,v 1.83 2024/04/26 17:41:21 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalLinearConstraints {
    private static final int N_iterations=80;
    private static final boolean FORCE_DIAG_DUMP=true;

    private static final double MIN_eVectorLastElement=1e-9;

    public final KGOEVSelection evSelected;
    public final UAdjustment aep;

    KGOIterationalLinearConstraints(final int nC,final int nX,
				    final double [] SK,
				    final double eps)
{
    final int n=nC*nX,na=n+1;
    if(SK.length!=n*n) throw new IllegalArgumentException();

    final double [] SKwithLambdaLin=new double [na*na];
    final java.util.function.Function<LambdaLin,double []> adjSKWithLambdaLin=curLambdaLin->{
	// construct SK adjusted with lambdaLin (add one more degree of freedom)
	for(int i=0,iSK=na-1;i<na;i++){
	    if(i<iSK){
		System.arraycopy(SK,i*n,SKwithLambdaLin,i*na,n);
		SKwithLambdaLin[i*na+na-1]= -curLambdaLin.b[i];
	    } else {
		for(int k=0;k<n;k++) SKwithLambdaLin[i*na+k]= -curLambdaLin.b[k];
		SKwithLambdaLin[i*na+na-1]=curLambdaLin.c;
	    }
	}
	return SKwithLambdaLin;
    };

    LambdaLin lambdaLin=new LambdaLin(n);
    KGOEVSelectionSeparatelyLast result_evsel=null;
    UAdjustment result_aep=null;
    double itmaxSK=0;
    for(int it=0;it<N_iterations;it++){
	// solve regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getEV(na,adjSKWithLambdaLin.apply(lambdaLin),true);

	// BEGIN select the eigenvector among several
	double maxSKadj=0;
	KGOEVSelectionSeparatelyLast evsel=null;
	AdjustedStateToUnitaryWithEigenproblem aepsel=null;
	// select eigenvalue
	for(var e:KGOEVSelectionSeparatelyLast.selectTopEVOneMoreDegree(na, /* na -- try all ; otherwise try top 3*/evSKLambda,nC,nX)){
	    if(it<=0 || true || evsel==null){
		if(eps>0){
		    UnitTests.checkDoubleRel(Math.pow(e.eVectorLastElement,2)+Math.pow(e.originalUNorm,2),"|original|",1,"1",eps);
		    System.err.println("e.ind="+e.ind+" e.eLambda="+e.eLambda+" e.eVectorLastElement="+e.eVectorLastElement+" e.originalUNorm="+e.originalUNorm);
		    if(e.originalUNorm>0){
			final double _lam=TMatr.getAMatrA(e.eVector,SK,n)*Math.pow(e.originalUNorm,2)+
			    -2*e.originalUNorm*e.eVectorLastElement*TMatr.mulVV(e.eVector,lambdaLin.b)+
			    +Math.pow(e.eVectorLastElement,2)*lambdaLin.c;
			UnitTests.checkDoubleRel1(_lam,"_lam",e.eLambda,"e.eLambda",eps);
		    }
		}
		if(e.originalUNorm<=0) continue;
		final var s=e.getEVAdjustedTo1();
		final double curSKadj=TMatr.getAMatrA(s.getUAdj(),SK,n);
		if(evsel==null || !Double.isFinite(maxSKadj) || curSKadj>=maxSKadj){
		    evsel=e;
		    aepsel=s;
		    maxSKadj=curSKadj;
		}
		if(eps>0){
		    final double curSKorig=TMatr.getAMatrA(s.getUOrig(),SK,n);

		    TestKGO.testOrthogonality("Adj SVD ",nC,s.getUAdj(),nX,eps);
		    System.err.println("it="+it+" e.ind="+e.ind+" e.eLambda="+e.eLambda+" Forig="+curSKorig+" Fadj="+curSKadj+" e.eVectorLastElement="+e.eVectorLastElement);

		    final double scaleUadj,lastEl;
		    if(!(Math.abs(e.eVectorLastElement)>MIN_eVectorLastElement)){
			scaleUadj=Math.sqrt((nC-1.0)/nC)/Math.sqrt(nC);
			lastEl=1/Math.sqrt(nC);
		    } else {
			scaleUadj=e.originalUNorm/Math.sqrt(nC);
			lastEl=e.eVectorLastElement;
		    }
		    // Check lambda--adjusted test.
		    final var ll=LambdaLin.fromAdjustedState(
			s.getUAdj(),
			SK,
			scaleUadj,
			lastEl,
			eps);
		    final double []
			u1=getUaddOneDegree(s.getUAdj(),scaleUadj,lastEl),
			testSK=adjSKWithLambdaLin.apply(ll);
		    UnitTests.checkDoubleRel1(TMatr.mulVV(u1,u1),"<u1|u1>",1,"1",eps);
		    UnitTests.checkDoubleRel1(TMatr.getAMatrA(u1,testSK,na),"<u1|testSK|u1>",0,"0",eps);
		    UnitTests.checkDoubleArrayRel1(TMatr.mul_MA_x_MB(na,testSK,na,u1,1),"testSK u1",new double [na],"0",eps);
		}
	    }
	}
	// END select the eigenvector among several

	// BEGIN calculate new values of Lagrange multipliers
	final double scaleUadj,lastEl;
	if(true||!(Math.abs(evsel.eVectorLastElement)>MIN_eVectorLastElement)){
	    final double fact =0.4;
	    scaleUadj=Math.sqrt((nC-fact)/nC)/Math.sqrt(nC);
	    lastEl=Math.sqrt(fact)/Math.sqrt(nC);
	} else {
	    scaleUadj=evsel.originalUNorm/Math.sqrt(nC);
	    lastEl=evsel.eVectorLastElement;
	}
	lambdaLin= LambdaLin.fromAdjustedState(aepsel.getUAdj(),
					       SK,
					       scaleUadj,
					       lastEl,
					       eps);
	// END calculate new values of Lagrange multipliers

	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+" Selected ind="+evsel.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aepsel.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aepsel.getUAdj(),SK,n)+" evsel.eVectorLastElement="+evsel.eVectorLastElement);
	}

	if(eps>0){
	    // test basis invariance
	    final double
		Fadj=TMatr.getAMatrA(aepsel.getUAdj(),SK,n),
		_Fadj=TMatr.mulVV(aepsel.getUAdj(),lambdaLin.b)*lastEl/scaleUadj,
		__Fadj=lambdaLin.c*Math.pow(lastEl/scaleUadj,2),
		Forig=TMatr.getAMatrA(aepsel.getUOrig(),SK,n);
	    // test lambda
	    final double [] u1=getUaddOneDegree(aepsel.getUAdj(),scaleUadj,lastEl),
		testSK=adjSKWithLambdaLin.apply(lambdaLin);
	    final double FadjLambda=TMatr.getAMatrA(u1,testSK,na);
	    UnitTests.checkDoubleRel1(FadjLambda,"FadjLambda",0,"0",eps);
	    UnitTests.checkDoubleRel1(Fadj,"Fadj",_Fadj,"_Fadj",eps);
	    UnitTests.checkDoubleRel1(Fadj,"Fadj",__Fadj,"__Fadj",eps);
	    UnitTests.checkDoubleArrayRel1(TMatr.mul_MA_x_MB(na,testSK,na,u1,1),"testSK u1",new double [na],"0",eps);
	}
	final double Fadj=TMatr.getAMatrA(aepsel.getUAdj(),SK,n);
	if(result_evsel==null || !Double.isFinite(itmaxSK) || Fadj>=itmaxSK) {
	    result_evsel=evsel;
	    result_aep=aepsel;
	    itmaxSK=Fadj;
	}
    }
    evSelected=result_evsel;
    aep=result_aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalLinearConstraints: RESULT: Selected ind="+evSelected.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" N_iterations="+N_iterations);
    }
}

private static double [] getUaddOneDegree(final double [] u,
					  final double scaleToOrig,
					  final double eVectorLastElement){
    final double [] u1=java.util.Arrays.copyOf(u,u.length+1);
    TMatr.scaleVectorInPlace(u1,scaleToOrig);
    u1[u.length]=eVectorLastElement;
    return u1;
}
private static class KGOEVSelectionSeparatelyLast extends KGOEVSelection {
    final double eVectorLastElement,originalUNorm;
    final double [] eVector_original;
    KGOEVSelectionSeparatelyLast(final int ind,
				 final double [] eVector_original,
				 final double eLambda,
				 final int nC,
				 final int nX){
	super(ind,new double [nC*nX],eLambda,nC,nX);
	this.eVector_original=eVector_original;
	final int n=nC*nX,na=n+1;
	if(eVector_original.length!=na) throw new IllegalArgumentException();
	System.arraycopy(eVector_original,0,eVector,0,n);
	this.eVectorLastElement=eVector_original[n];
	this.originalUNorm=Math.sqrt(TMatr.mulVV(eVector,eVector));
	TMatr.scaleVectorInPlace(eVector,1/originalUNorm);
    }

    static java.util.List<KGOEVSelectionSeparatelyLast> selectTopEVOneMoreDegree(final int ntop,final double [] ev,final int nC,final int nX){
	final int na=nC*nX+1;
	if(ev.length!=na+na*na) throw new IllegalArgumentException();
	// for now -- try max ntop eigenvalues
	final var a=new java.util.ArrayList<KGOEVSelectionSeparatelyLast>();
	for(int k=na,ne=(na-ntop);--k>=ne;){
	    if(k>=0){
		a.add(
		    new KGOEVSelectionSeparatelyLast(k,EVXData.getPsiFromBasis(k,ev,na),ev[k],nC,nX)
		      );
	    }
	}
	return a;
    }
} // END OF class KGOEVSelectionSeparatelyLast


    private static class LambdaLin {
	final double [] b;
	final double c,toOriginalEVectorScaleFactor,eVectorLastElement;
	LambdaLin(final double [] b,
		  final double c,
		  final double toOriginalEVectorScaleFactor,
		  final double eVectorLastElement){
	    this.b=b;
	    this.c=c;
	    this.toOriginalEVectorScaleFactor=toOriginalEVectorScaleFactor;
	    this.eVectorLastElement=eVectorLastElement;
	}
	LambdaLin(final int n){
	    this(new double [n],0,1,1);
	}
	static LambdaLin fromAdjustedState(final double [] u_adj,
					   final double [] SK,
					   final double toOriginalEVectorScaleFactor,
					   final double eVectorLastElement,
					   final double eps){
	    final int n=u_adj.length;
	    if(SK.length!=n*n) throw new IllegalArgumentException();
	    if(eps>0){
		UnitTests.checkDoubleRel(Math.pow(eVectorLastElement,2)+Math.pow(toOriginalEVectorScaleFactor,2)*TMatr.mulVV(u_adj,u_adj),"|original|",1,"1",eps);
	    }
	    final double scaleTo=toOriginalEVectorScaleFactor/eVectorLastElement;
	    final double [] b=TMatr.scaleVectorInPlace(TMatr.mul_MATrans_x_MB(n,SK,n,u_adj,1),scaleTo);
	    final double c=TMatr.mulVV(b,u_adj)*scaleTo;
	    return new LambdaLin(b,c,toOriginalEVectorScaleFactor,eVectorLastElement);
	}
    } // END OF class LambdaLin


}
