package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.EVXData;

/**
 * A Knowledge Generalized Operator iteration optimization,
 * directly in u_orig basis.
 * The algorithm uses an adjustment from simplified to full version of constraints:
 * \( \sum\limits_{k=0}^{nX-1}u_{ik}u_{jk}=\delta_{ij} \), for \(i,j=0\dots nC-1\).
 * The Lagrange multipliers are calculated with
 * {@link LagrangeMultipliersPartialSubspace#calculateRegularLambda}. With this
 * \( \lambda_{ij} \) the algorithm does not converge well, but a good
 * enough solution can be found.
 * A try without SVD.
 * @see KGOIterationalLagrangeMultipliersPartialSubspace
 * @see KGOIterationalLagrangeMultipliersInDenominatorU
 * @version $Id: KGOIterationalSimpleOptimizationU.java,v 1.84 2024/04/26 17:41:21 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalSimpleOptimizationU {
    private static final int N_iterations=40;
    private static final boolean FORCE_DIAG_DUMP=true;

    public final KGOEVSelection evSelected;
    public final UAdjustment aep;
    public KGOIterationalSimpleOptimizationU(final int nC,final int nX,
					     final double [] SK,
					     final double eps)
{
    final int n=nC*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    //if(true) throw new RuntimeException();

    final double [] SKwithLambda=new double [n*n];
    final java.util.function.Function<double [],double []> adjSKWithLambda=curLambda->{
	// construct SK adjusted with lambda
	System.arraycopy(SK,0,SKwithLambda,0,SK.length);
	for(int j=0;j<nC;j++){
	for(int jp=0;jp<nC;jp++){
	    final double l=curLambda[j*nC+jp];
	    for(int k=0;k<nX;k++){
		SKwithLambda[(j*nX+k)*n+(jp*nX+k)]-=l;
	    }}
	}
	return SKwithLambda;
    };

    final java.util.function.Function<double [],double []>
	getLambda=u_at->LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,u_at);

    class LocalMax {
	final double Fadj,Forig;
	final KGOEVSelection e;
	final AdjustedStateToUnitaryWithEigenproblem aep;
	final int itNum;
	LocalMax(final KGOEVSelection e,int itNum){
	    this.e=e;
	    this.aep=e.getEVAdjustedTo1();
	    this.Fadj=TMatr.getAMatrA(aep.getUAdj(),SK,n);
	    this.Forig=TMatr.getAMatrA(aep.getUOrig(),SK,n);
	    this.itNum=itNum;

	}
	boolean isValid(){return Double.isFinite(Fadj);}
	boolean isArgLocalBetter(LocalMax s){
	    // for now
	    return s.Fadj>=this.Fadj;
	}
	boolean isArgGlobalBetter(LocalMax s){
	    return s.Fadj>=this.Fadj;
	}
    }

    double [] lambda=new double [nC*nC];
    LocalMax globMax=null,prevLocMax=null;
    for(int it=0;it<N_iterations;it++){
	// solve regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getEV(n,adjSKWithLambda.apply(lambda),true);
	//System.err.println("EV evSKLambda="+java.util.Arrays.toString(java.util.Arrays.copyOf(evSKLambda,n)));

	// BEGIN select the eigenvector among several
	LocalMax locMax=null;
	boolean flagGlobalMax=false;
	int nNegLambda=0;
	// select eigenvalue
	for(var e:KGOEVSelection.selectTopEV(n, /* n -- try all ; try top 3*/evSKLambda,nC,nX)){
	    if(e.eLambda<=0) nNegLambda++;
	    // How many negaive lambda to consider? Try 4.
	    if(it<=0 || nNegLambda<=4 || locMax==null){
		final var s=new LocalMax(e,it);
		if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(s)){
		    // found global max (across iterations)
		    globMax=locMax=s;
		    flagGlobalMax=true;
		}
		else if(!flagGlobalMax && (
			    locMax==null || !locMax.isValid() || locMax.isArgLocalBetter(s)
					   )){
		    // some local max
		    locMax=s;
		}

		// Calculate first and second derivatives
		// at the point s.aep.getUAdj()
		if(eps>0){
		    // Both first and second variations are 0.
		    final double [] lamE=getLambda.apply(s.aep.getUAdj());
		    final var rq=new RayleighQuotient.RQVariationInMatrixForm(s.aep.getUAdj(),n,adjSKWithLambda.apply(lamE),TMatr.getUnitMatrix(n),n);
		    final double [] iD2=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,rq.d2),
			del=TMatr.getVectorScaled(TMatr.getMatr_x_A(n,iD2,rq.d1),-1);
		    //System.err.println("uadj="+java.util.Arrays.toString(s.aep.getUAdj()));
		    //System.err.println("del="+java.util.Arrays.toString(del));

		    // Note! iD2 is not positively definite!
		    final double varD= -0.5*TMatr.getAMatrA(rq.d1,iD2,n);
		    System.err.println("DTEST:it="+it+" e.ind="+e.ind+" Newton varD="+varD+" ==0!!!");
		    UnitTests.checkDoubleRel1(varD,"varD",0,"0",eps);
		    UnitTests.checkDoubleRel1(rq.d0,"<uAdj|SK-lam|uAdj>",0,"0",eps);
		    UnitTests.checkDoubleRel1(TMatr.getAMatrA(s.aep.getUAdj(),SK,n),"<uAdj|SK|uAdj>",TMatr.getSpurM(nC,lamE),"Spur LamE",eps);
		    UnitTests.checkDoubleArrayRel1(s.aep.getUAdj(),"s.aep.getUAdj()",del,"del",eps);

		    // This is the reason of a problem with algorithm convergence.
		    System.err.println("DTEST:it="+it+" e.ind="+e.ind+" Newton varD="+varD+" sumd1^2="+TMatr.mulVV(rq.d1,rq.d1)+" SPUR D2="+TMatr.getSpurM(n,rq.d2));
		    // other diag
		    /*
		    final double [] ev_iDs=KGOSolutionVectorXVectorF.EV.getEV(n,TMatr.selectHermitianPartInPlace(n,iD2.clone()),false);
		    System.err.println("DTEST:it="+it+" ev_iDs="+java.util.Arrays.toString(ev_iDs));
		    System.err.println("DTEST:it="+it+" e.ind="+e.ind+" D2EV="+java.util.Arrays.toString(KGOSolutionVectorXVectorF.EV.getEV(n,rq.d2,false)));
		    System.err.println("DTEST:it="+it+" e.ind="+e.ind+" rq.d1="+java.util.Arrays.toString(rq.d1)+" sumd1^2="+TMatr.mulVV(rq.d1,rq.d1));
		    System.err.println("DTEST:it="+it+" e.ind="+e.ind+" rq.d2="+java.util.Arrays.toString(rq.d2)+" sumd2^2="+TMatr.mulVV(rq.d2,rq.d2));
		    */
		}

		if(eps>0){
		    final double uOrigUadj=TMatr.mulVV(s.aep.getUAdj(),s.aep.getUOrig());
		    if(true){
			// check projection max
			double sumSQRT=0;
			for(int k=0;k<nC;k++) sumSQRT+=Math.sqrt(s.aep.gram.ev_m[k]);
			UnitTests.checkDoubleRel1(sumSQRT,"sumSQRT",uOrigUadj,"uOrigUadj",eps);
		    }
		    TestKGO.testOrthogonality("Adj EV ",nC,s.aep.getUAdj(),nX,eps);

		    System.err.println("it="+it+":Try e.ind="+e.ind+" ("+n+") e.eLambda="+e.eLambda+" Forig="+s.Forig+" Fadj="+s.Fadj+" sum1Lam="+s.aep.sum1Lam+" <uOrig|Uadj>="+uOrigUadj+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(s.aep.gram.ev_m,nC)));
		}
		if(eps>0){
		    // Check lambda--adjusted test.
		    // This test is not really useful.
		    final double []
			SKlaminU=AdjustedStateToUnitaryWithEigenproblem.getSKasMatrixInUState(
			    nC,
			    nX,
			    adjSKWithLambda.apply(getLambda.apply(s.aep.getUAdj())),
			    s.aep.getUAdj())
			//,ev_SKlaminU=KGOSolutionVectorXVectorF.EV.getEV(nC,SKlaminU,true)
			;
		    UnitTests.checkDoubleRel1(TMatr.sumElements(SKlaminU),"sum SKlaminU",0,"0",eps);
		    for(int i=0;i<nC;i++){
			double srow=0,scol=0;
			for(int j=0;j<nC;j++){
			    srow+=SKlaminU[j*nC+i];
			    scol+=SKlaminU[i*nC+j];
			}
			UnitTests.checkDoubleRel1(srow,"srow",0,"0",eps);
			UnitTests.checkDoubleRel1(scol,"scol",0,"0",eps);
		    }
		    //System.err.println("SKlaminU="+java.util.Arrays.toString(SKlaminU));
		    // one of the eigenvalues is always 0
		    //System.err.println("it="+it+" e.ind="+e.ind+" ev SKlaminU="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_SKlaminU,nC)));
		}
		// Check the number of postitive/negative eigenvalues
		// of the second derivative. Also check the spur.
		//final double [] ev_d2=KGOSolutionVectorXVectorF.EV.getEV(n,rq.d2,false);
		//int nPlus=0,nMinus=0;
		//for(int k=0;k<n;k++) {if(ev_d2[k]<=0) nMinus++; else nPlus++;}
		//System.err.println("it="+it+" e.ind="+e.ind+" spur d2="+TMatr.getSpurM(n,rq.d2)+" nPlus="+nPlus+" nMinus="+nMinus);
	    }
	}
	// END select the eigenvector among several

	if(eps>0 && prevLocMax!=null){
	    // Diag convergence.
	    System.err.println("it="+it+":CheckPrev <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+
			       " <prev_orig|orig>="+TMatr.mulVV(prevLocMax.aep.getUOrig(),locMax.aep.getUOrig())+
			       " <prev_adj|adj>="+TMatr.mulVV(prevLocMax.aep.getUAdj(),locMax.aep.getUAdj())+
			       " Forig="+TMatr.getAMatrA(locMax.aep.getUOrig(),SK,n)+
			       " prev_Forig="+TMatr.getAMatrA(prevLocMax.aep.getUOrig(),SK,n)+
			       " Fadj="+TMatr.getAMatrA(locMax.aep.getUAdj(),SK,n)+
			       " prev_Fadj="+TMatr.getAMatrA(prevLocMax.aep.getUAdj(),SK,n));
	}
	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+":Selected ind="+locMax.e.ind+" ("+n+") eLambda="+locMax.e.eLambda+" Forig="+locMax.Forig+" Fadj="+locMax.Fadj+" sum1Lam="+locMax.aep.sum1Lam+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	// BEGIN calculate new values of Lagrange multipliers
	lambda=getLambda.apply(locMax.aep.getUAdj());
	// END calculate new values of Lagrange multipliers

	if(eps>0){
	    // test lambda
	    // the value of FadjLambda is zero as lambda spur is equal to Fadj
	    final double FadjLambda=TMatr.getAMatrA(locMax.aep.getUAdj(),adjSKWithLambda.apply(lambda),n);
	    UnitTests.checkDoubleRel1(FadjLambda,"FadjLambda",0,"0",eps);
	    UnitTests.checkDoubleRel1(locMax.Fadj,"locMax.Fadj",TMatr.getSpurM(nC,lambda),"spur(lambda)",eps);

	    System.err.println("KGOIterationalSimpleOptimizationU: it="+it+" Fadj="+locMax.Fadj+" Forig="+locMax.Forig+" ind="+locMax.e.ind+" ("+n+") <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig()));
	}
	if(eps>0){
	    // test different method of lambda calculation
	    final double [][] vBasis=new double [n][];
	    for(int p=0;p<n;p++) vBasis[p]=EVXData.getPsiFromBasis(p,evSKLambda,n);
	    final double [] lambdaTest=LagrangeMultipliersPartialSubspace.getLambdaForSubspace(
		nC,
		nX,
		SK,
		locMax.aep.getUAdj(),
		vBasis);
	    UnitTests.checkDoubleArrayRel1(lambda,"lambda",lambdaTest,"lambdaTest",eps);
	    System.err.println("KGOIterationalSimpleOptimizationU: it="+it+" lambda test OK");
	}

	prevLocMax=locMax;
	if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(locMax)) {
	    globMax=locMax;
	}
    } // end iterations

    evSelected=globMax.e;
    aep=globMax.aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalGEVAdj_Q: RESULT:Selected ind="+evSelected.ind+" ("+n+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" at iteration="+globMax.itNum+" N_iterations="+N_iterations);
    }
}



}
