package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.EliminateLinearConstraints_HomegrownLUFactorization;

/**
 * A Knowledge Generalized Operator iteration optimization,
 * a try with extra linear constraints incorporated with
 * {@link com.polytechnik.utils.EliminateLinearConstraints_HomegrownLUFactorization Gaussian elimination}
 * to incorporate
 * {@link LinearConstraints#getOrthogonalOffdiag0DiagEq linear constraints}
 * into generalized eigenvalue problem.
 * This seems to be the best version.
 * @see KGOIterationalSimpleOptimizationU
 * @see com.polytechnik.utils.EliminateLinearConstraints_HomegrownLUFactorization
 * @see LinearConstraints#getOrthogonalOffdiag0DiagEq
 * @version $Id: KGOIterationalSubspaceLinearConstraints.java,v 1.12 2024/05/12 22:45:04 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalSubspaceLinearConstraints {
    private static final int N_iterations=40;
    private static final boolean FORCE_DIAG_DUMP=true;

    public final KGOEVSelection evSelected;
    public final UAdjustment aep;
    public KGOIterationalSubspaceLinearConstraints(
	final int nC,final int nX,
	final double [] SK,
	final double eps)
{
    final int n=nC*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    //if(true) throw new RuntimeException();

    final double [] SKwithLambda=new double [n*n];
    final java.util.function.Function<double [],double []> adjSKWithLambda=curLambda->{
	// construct SK adjusted with lambda
	System.arraycopy(SK,0,SKwithLambda,0,SK.length);
	for(int j=0;j<nC;j++){
	for(int jp=0;jp<nC;jp++){
	    final double l=curLambda[j*nC+jp];
	    for(int k=0;k<nX;k++){
		SKwithLambda[(j*nX+k)*n+(jp*nX+k)]-=l;
	    }}
	}
	return SKwithLambda;
    };

    class LocalMax {
	final double Fadj,Forig;
	final KGOEVSelection e;
	final AdjustedStateToUnitaryWithEigenproblem aep;
	final int itNum;
	LocalMax(final KGOEVSelection e,int itNum){
	    this.e=e;
	    final double [] u_orig=TMatr.getVectorScaled(e.eVector,Math.sqrt(nC/TMatr.mulVV(e.eVector,e.eVector)));
	    this.aep=new AdjustedStateToUnitaryWithEigenproblem(new ApplyFunToGramMatrix(nC,nX,u_orig));
	    this.Fadj=TMatr.getAMatrA(aep.getUAdj(),SK,n);
	    this.Forig=TMatr.getAMatrA(aep.getUOrig(),SK,n);
	    this.itNum=itNum;

	}
	boolean isValid(){return Double.isFinite(Fadj);}
	boolean isArgLocalBetter(LocalMax s){
	    // for now
	    return s.Fadj>=this.Fadj;
	}
	boolean isArgGlobalBetter(LocalMax s){
	    return s.Fadj>=this.Fadj;
	}
    }

    // A convenience basis to work with; orthogonal basis also increases numerical stability.
    final double [] ev_SK=KGOSolutionVectorXVectorF.EV.getEV(n,SK,true),
	ev_SK_basis=java.util.Arrays.copyOfRange(ev_SK,n,ev_SK.length);
    class CreateSubspaceOrtU {

	/** Current iteration. */
	final double [] u;

	/** A basis with linear constraints eliminated,
	 * a matrix of newDim*n size.
	 */
	final double [] fromSubspaceToOrig;

	/** Problem dimension after elimination of linear constraints.
	 */
	final int newDim;

	/**
	 * @param u Solution iteration.
	 */
	CreateSubspaceOrtU(final double [] u){
	    this.u=u;

	    // BEGIN create linear constraints from UV vectors
	    // a vector to check projections to
	    final double [] xProj=u;

	    final double [][]
		linearConstrains=LinearConstraints.getOrthogonalOffdiag0DiagEq(nC,nX,xProj,ev_SK_basis);
	    // END create linear constraints from ProjV vectors

	    // Gaussian elimination object from linear constraints.
	    final var t=new EliminateLinearConstraints_HomegrownLUFactorization(n,linearConstrains,condMaxToFail_LinearSystem);

	    // BEGIN create a basis for eigenvalue problem.
	    this.newDim=t.dimXnew;
	    this.fromSubspaceToOrig=TMatr.mul_MATrans_x_MB(t.dimXnew,t.fromXnewToXorigAsMatrix(),n,ev_SK_basis,n);
	    // END create a basis for eigenvalue problem.

	    // BEGIN unit test
	    if(eps>0){
		// test basis transform OK
		final double [] mSKs=EVXData.convertMatrFromQtoB_Basis(n,SK,fromSubspaceToOrig,newDim,0),
		    mQs=TMatr.mul_MA_x_MBTrans(newDim,fromSubspaceToOrig,n,fromSubspaceToOrig,newDim),
		    mSKv=EVXData.convertMatrFromQtoB_Basis(n,SK,ev_SK_basis,n,0);

		for(int p=0;p<newDim;p++){
		    final double [] vbpX=EVXData.getPsiFromBasis(p,fromSubspaceToOrig,n,0);
		    // test that basis function vbpX satisfy required linear constraints
		    LinearConstraints.testConstraintsOK_getOrthogonalOffdiag0DiagEq(nC,nX,xProj,vbpX,eps);

		    // check basis
		    final double [] xnewpZ=new double [newDim];xnewpZ[p]=1;
		    final double []
			xnewpXfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewpZ,1),
			xnewpV=t.fromXnewToXorig(xnewpZ),
			xnewpX=EVXData.convertPsiFromBtoQ_Basis(n,xnewpV,ev_SK);
		    UnitTests.checkDoubleArrayRel1(xnewpX,"xnewpX",vbpX,"vbpX",eps);
		    UnitTests.checkDoubleArrayRel1(xnewpX,"xnewpX",xnewpXfm,"xnewpXfm",eps);
		    // test Gram matrix is the same
		    for(int i=0;i<newDim;i++){
			final double [] xnewiZ=new double [newDim];xnewiZ[i]=1;
			final double []
			    xnewiXfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewiZ,1),
			    xnewiV=t.fromXnewToXorig(xnewiZ),
			    xnewiX=EVXData.convertPsiFromBtoQ_Basis(n,xnewiV,ev_SK);
			UnitTests.checkDoubleArrayRel1(xnewiX,"xnewiX",xnewiXfm,"xnewiXfm",eps);
			UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpX,xnewiX),"<xnewpX|xnewiX>",TMatr.mulVV(xnewpV,xnewiV),"<xnewpV|xnewiV>",eps);
			UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpX,xnewiX),"<xnewpX|xnewiX>",TMatr.getAMatrB(xnewpZ,xnewiZ,mQs,newDim),"<xnewpZ|mQs|xnewiZ>",eps);
			UnitTests.checkDoubleRel1(TMatr.getAMatrB(xnewpX,xnewiX,SK,n),"<xnewpX|SK|xnewiX>",TMatr.getAMatrB(xnewpV,xnewiV,mSKv,n),"<xnewpV|mSKv|xnewiV>",eps);

			// test matrix conversion formulas
			UnitTests.checkDoubleRel1(TMatr.getAMatrB(xnewpXfm,xnewiXfm,SK,n),"<xnewpXfm|SK|xnewiXfmx>",TMatr.getAMatrB(xnewpZ,xnewiZ,mSKs,newDim),"<xnewpZ|mSKs|xnewiZ>",eps);
			UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpXfm,xnewiXfm),"<xnewpXfm|xnewiXfm>",TMatr.getAMatrB(xnewpZ,xnewiZ,mQs,newDim),"<xnewpZ|mQs|xnewiZ>",eps);
		    }
		}
		System.err.println("CreateSubspaceOrtU: nEliminated="+t.nEliminated+" newDim="+newDim+" n="+n);
	    }
	    // END unit test
	}
    } // END OF class CreateSubspaceOrtU


    int nA=n;
    double [] T=ev_SK_basis,lambda=new double [nC*nC];
    LocalMax globMax=null,prevLocMax=null;
    for(int it=0;it<N_iterations;it++){
	// solve generalized regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getGEV_Generalized(
		nA,
		/*SKs=*/EVXData.convertMatrFromQtoB_Basis(n,adjSKWithLambda.apply(lambda),T,nA,0),
		/*Qs=*/TMatr.mul_MA_x_MBTrans(nA,T,n,T,nA),
		true);

	// BEGIN select the eigenvector among several
	LocalMax locMax=null;
	boolean flagGlobalMax=false;
	int nNegLambda=0,nFromTopLambda=0;
	// select eigenvalue, start with the maximal
	for(int p=nA;--p>=0;){
	    final double eLambda=evSKLambda[p];
	    nFromTopLambda++;if(eLambda<=0) nNegLambda++;
	    // Eigenvalue selection affects whether the global maximum will be found.
	    // A good heuristics may be to always take the specific eigenvalue
	    // (highest, second highest, etc.) and then compare the results.
	    // Currently we test only for maximal eigenvalue, add heuristics & tries later.
	    if(it<=0 || nFromTopLambda<=1 || locMax==null){
		final var e=new KGOEVSelection(
		    p,
		    TMatr.mul_MATrans_x_MB(n,T,nA,EVXData.getPsiFromBasis(p,evSKLambda,nA),1),
		    eLambda,
		    nC,nX);
		final var s=new LocalMax(e,it);
		if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(s)){
		    // found global max (across iterations)
		    globMax=locMax=s;
		    flagGlobalMax=true;
		}
		else if(!flagGlobalMax && (
			    locMax==null || !locMax.isValid() || locMax.isArgLocalBetter(s)
					   )){
		    // some local max
		    locMax=s;
		}

		if(eps>0){
		    UnitTests.checkDoubleRel(TMatr.getSpurM(nC,s.aep.gram.m),"spur G",nC,"nC",eps);
		    TestKGO.testOrthogonality("Adj EV ",nC,s.aep.getUAdj(),nX,eps);
		    System.err.println("it="+it+" e.ind="+e.ind+" ("+nA+") e.eLambda="+e.eLambda+" Forig="+s.Forig+" Fadj="+s.Fadj+" <u_orig|u_adj>="+TMatr.mulVV(s.aep.getUAdj(),s.aep.getUOrig())+" GEV="+java.util.Arrays.toString(java.util.Arrays.copyOf(s.aep.gram.ev_m,nC)));
		}
	    }
	}
	// END select the eigenvector among several

	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+":Selected ind="+locMax.e.ind+" ("+nA+") eLambda="+locMax.e.eLambda+"; Forig="+locMax.Forig+" Fadj="+locMax.Fadj+" sum1Lam="+locMax.aep.sum1Lam+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	// BEGIN calculate new values of Lagrange multipliers
	final var l=new CreateSubspaceOrtU(locMax.aep.getUAdj());
	lambda=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,l.u);
	nA=l.newDim;
	T=l.fromSubspaceToOrig;
	// END calculate new values of Lagrange multipliers

	if(eps>0){
	    System.err.println("KGOIterationalSubspaceLinearConstraints: it="+it+" Fadj="+locMax.Fadj+" Forig="+locMax.Forig+" ind="+locMax.e.ind+" ("+nA+") <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	prevLocMax=locMax;
	if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(locMax)) {
	    globMax=locMax;
	}
    }
    evSelected=globMax.e;
    aep=globMax.aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalSubspaceLinearConstraints: RESULT:Selected ind="+evSelected.ind+" ("+nA+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" at iteration="+globMax.itNum+" N_iterations="+N_iterations);
    }
}


    /** Minimal norm for Gaussian elimination.
     */
    private static final double condMaxToFail_LinearSystem=1e-10;



}
