package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.EliminateLinearConstraints_HomegrownLUFactorization;

/**
 * A Knowledge Generalized Operator iteration optimization,
 * a try with extra linear constraints incorporated with
 * {@link com.polytechnik.utils.EliminateLinearConstraints_HomegrownLUFactorization Gaussian elimination}.
 * This class has a number of implemented algorithms for linear constraints,
 * the one that should be used -- the {@link LinConstraint#UV_OFFDIAG0DIAGEQ}.
 * The {@link LinConstraint#UV_ALL} provides identical results
 * and the other type of constraints do not converge at all. This class is a test and should not
 * be used, proper equivalent implementation is {@link KGOIterationalSubspaceLinearConstraints}.
 * @see KGOIterationalSubspaceLinearConstraints
 * @see com.polytechnik.utils.EliminateLinearConstraints_HomegrownLUFactorization
 * @see LinearConstraints#getOrthogonalOffdiag0DiagEq
 * @version $Id: KGOIterationalTrySubspaceOrthogonal.java,v 1.51 2024/05/12 22:45:04 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOIterationalTrySubspaceOrthogonal {
    private static final int N_iterations=40;
    private static final boolean FORCE_DIAG_DUMP=true;

    public final KGOEVSelection evSelected;
    public final UAdjustment aep;
    public KGOIterationalTrySubspaceOrthogonal(
	final int nC,final int nX,
	final double [] SK,
	final double eps)
{
    final int n=nC*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    //if(true) throw new RuntimeException();


    final double [] SKwithLambda=new double [n*n];
    final java.util.function.Function<double [],double []> adjSKWithLambda=curLambda->{
	// construct SK adjusted with lambda
	System.arraycopy(SK,0,SKwithLambda,0,SK.length);
	for(int j=0;j<nC;j++){
	for(int jp=0;jp<nC;jp++){
	    final double l=curLambda[j*nC+jp];
	    for(int k=0;k<nX;k++){
		SKwithLambda[(j*nX+k)*n+(jp*nX+k)]-=l;
	    }}
	}
	return SKwithLambda;
    };

    class LocalMax {
	final double Fadj,Forig;
	final KGOEVSelection e;
	final AdjustedStateToUnitaryWithEigenproblem aep;
	final int itNum;
	LocalMax(final KGOEVSelection e,int itNum){
	    this.e=e;
	    final double [] u_orig=TMatr.getVectorScaled(e.eVector,Math.sqrt(nC/TMatr.mulVV(e.eVector,e.eVector)));
	    this.aep=new AdjustedStateToUnitaryWithEigenproblem(new ApplyFunToGramMatrix(nC,nX,u_orig));
	    this.Fadj=TMatr.getAMatrA(aep.getUAdj(),SK,n);
	    this.Forig=TMatr.getAMatrA(aep.getUOrig(),SK,n);
	    this.itNum=itNum;

	}
	boolean isValid(){return Double.isFinite(Fadj);}
	boolean isArgLocalBetter(LocalMax s){
	    // for now
	    return s.Fadj>=this.Fadj;
	}
	boolean isArgGlobalBetter(LocalMax s){
	    return s.Fadj>=this.Fadj;
	}
    }


    // A convenience basis to work with; orthogonal basis also increases numerical stability.
    final double [] ev_SK=KGOSolutionVectorXVectorF.EV.getEV(n,SK,true),
	ev_SK_basis=java.util.Arrays.copyOfRange(ev_SK,n,ev_SK.length);
    class CreateSubspaceOrtU {

	/** A basis with linear constraints eliminated,
	 * a matrix of newDim*n size.
	 */
	final double [] fromSubspaceToOrig;

	/** Problem dimension after elimination of linear constraints
	 * and adding {@code u} to the basis.
	 */
	final int newDim;

	/** Lagrange multipliers. */
	final double [] lambda;

	/**
	 * @param u Solution iteration.
	 * @param linConstraintType The type of linear constraints to apply.
	 */
	CreateSubspaceOrtU(final double [] u,final LinConstraint linConstraintType){

	    // new Lagrange multipliers
	    this.lambda=LagrangeMultipliersPartialSubspace.calculateRegularLambda(nC,nX,SK,u);

	    // BEGIN create linear constraints from UV vectors
	    // a vector to check projections to
	    final double [] xProj=switch(linConstraintType) {
	    case BV_ORT,BV_ALL -> {
		// b-projection set to zero
		final double [] bx=TMatr.scaleVectorInPlace(TMatr.sumAcB(TMatr.mul_MA_x_MB(n,SK,n,u,1),-1,TMatr.mul_MA_x_MB(nC,lambda,nC,u,nX)),1/TMatr.getSpurM(nC,lambda));
		if(eps>0){
		    UnitTests.checkDoubleRel(TMatr.mulVV(bx,u),"<bx|u>",0,"0",eps);
		    UnitTests.checkDoubleArrayRel1(TMatr.selectHermitianPartInPlace(nC,TMatr.mul_MA_x_MBTrans(nC,bx,nX,u,nC)),"Herm(bx|u)",new double [nC*nC],"0",eps);
		}
		yield bx;
	    }
	    // u-projection set to zero; inject u itself later
	    case UV_ORT,UV_ORT_NOINJECT,UV_ALL,UV_OFFDIAG0DIAGEQ -> u;
	    };

	    final double [][] linearConstrains=switch(linConstraintType) {
		case UV_ORT,UV_ORT_NOINJECT,BV_ORT -> LinearConstraints.getVectorOrthogonal(nC,nX,xProj,ev_SK_basis);
		case UV_OFFDIAG0DIAGEQ -> LinearConstraints.getOrthogonalOffdiag0DiagEq(nC,nX,xProj,ev_SK_basis);
		case UV_ALL,BV_ALL -> LinearConstraints.getOrthogonalAllProj0(nC,nX,xProj,ev_SK_basis);
	    };
	    // END create linear constraints from ProjV vectors

	    // Gaussian elimination object from linear constraints.
	    final var t=new EliminateLinearConstraints_HomegrownLUFactorization(n,linearConstrains,condMaxToFail_LinearSystem);

	    // BEGIN create a basis for eigenvalue problem.
	    switch(linConstraintType) {
	    case BV_ORT,BV_ALL,UV_ORT_NOINJECT,UV_OFFDIAG0DIAGEQ:
		// no any u-inject, just copy basis from t
		this.newDim=t.dimXnew;
		this.fromSubspaceToOrig=TMatr.mul_MATrans_x_MB(t.dimXnew,t.fromXnewToXorigAsMatrix(),n,ev_SK_basis,n);
		break;
	    case UV_ORT,UV_ALL:
		// injection is not a good idea, a better option is to use the constraints that do not require injection.
		// u and the basis from t, inject u to basis
		this.newDim=t.dimXnew+1;
		this.fromSubspaceToOrig=new double[newDim*n];
		// u as first vector
		System.arraycopy(u,0,this.fromSubspaceToOrig,0,n);
		// the rest as V
		final double [] matrV=TMatr.mul_MATrans_x_MB(t.dimXnew,t.fromXnewToXorigAsMatrix(),n,ev_SK_basis,n);
		System.arraycopy(matrV,0,this.fromSubspaceToOrig,n,matrV.length);
		break;
	    default: throw new IllegalArgumentException();
	    }
	    // END create a basis for eigenvalue problem.

	    // BEGIN unit test
	    if(eps>0){
		// check a matrix for: sum diag0, all diag 0, all 0.
		final java.util.function.Consumer<double[]>testPVp=
		    (pVp)->{
		    switch(linConstraintType) {
		    case UV_ORT,UV_ORT_NOINJECT,BV_ORT:
		    UnitTests.checkDoubleRel1(TMatr.getSpurM(nC,pVp),"spur(PVp)",0,"0",eps);
		    break;
		    case UV_ALL,BV_ALL:
		    UnitTests.checkDoubleArrayRel1(pVp,"Herm(PVp)",new double [nC*nC],"0",eps);
		    break;
		    case UV_OFFDIAG0DIAGEQ:
		    for(int j=0;j<nC;j++){
			for(int i=0;i<nC;i++){
			    if(i!=j) UnitTests.checkDoubleRel1(pVp[i*nC+j],"pVp[i*nC+j]",0,"0",eps);
			}
			if(j>0){
			    UnitTests.checkDoubleRel1(pVp[j*nC+j],"pVp[j*nC+j]",pVp[(j-1)*nC+(j-1)],"pVp[(j-1)*nC+(j-1)]",eps);
			}
		    }
		    break;
		    default: throw new IllegalArgumentException();
		    }
		};

		// test basis transform OK
		final double [] mSKs=EVXData.convertMatrFromQtoB_Basis(n,SK,fromSubspaceToOrig,newDim,0),
		    mQs=TMatr.mul_MA_x_MBTrans(newDim,fromSubspaceToOrig,n,fromSubspaceToOrig,newDim),
		    mSKv=EVXData.convertMatrFromQtoB_Basis(n,SK,ev_SK_basis,n,0);
		switch(linConstraintType){
		case UV_ORT,UV_ALL:
		    // test with injection
		    final double [] uX=EVXData.getPsiFromBasis(0,fromSubspaceToOrig,n,0),
			uV=EVXData.convertPsiFromQtoB_Basis(n,uX,ev_SK,TMatr.getUnitMatrix(n));
		    UnitTests.checkDoubleArrayRel1(uX,"uX",u,"u",eps);
		    for(int p=0;p<newDim;p++){
			final double [] vbpX=EVXData.getPsiFromBasis(p,fromSubspaceToOrig,n,0),
			    PVp=TMatr.selectHermitianPartInPlace(nC,TMatr.mul_MA_x_MBTrans(nC,uX,nX,vbpX,nC));
			if(p<=0) UnitTests.checkDoubleArrayRel1(PVp,"Herm(PVp)",TMatr.getUnitMatrix(nC),"I",eps);
			else testPVp.accept(PVp);

			final double [] xnewpZ=new double [newDim];xnewpZ[p]=1;
			final double [] xnewpXfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewpZ,1);
			double [] xnewpV=null,xnewpX=null;
			if(p>0){
			    // off by 1 to test t.fromXnewToXorig
			    xnewpV=t.fromXnewToXorig(java.util.Arrays.copyOfRange(xnewpZ,1,xnewpZ.length));
			    xnewpX=EVXData.convertPsiFromBtoQ_Basis(n,xnewpV,ev_SK);
			    UnitTests.checkDoubleArrayRel1(xnewpX,"xnewpX",vbpX,"vbpX",eps);
			    UnitTests.checkDoubleArrayRel1(xnewpX,"xnewpX",xnewpXfm,"xnewpXfm",eps);
			    UnitTests.checkDoubleRel1(TMatr.mulVV(uV,xnewpV),"<uV|xnewpV>",0,"0",eps);
			    UnitTests.checkDoubleRel1(TMatr.mulVV(u,xnewpX),"<u|xnewpX>",0,"0",eps);
			}
			// test Gram matrix is the same
			for(int i=0;i<newDim;i++){
			    final double [] xnewiZ=new double [newDim];xnewiZ[i]=1;
			    final double [] xnewiXfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewiZ,1);
			    if(i>0 && p>0){
				final double [] // off by 1 to test t.fromXnewToXorig
				    xnewiV=t.fromXnewToXorig(java.util.Arrays.copyOfRange(xnewiZ,1,xnewiZ.length)),
				    xnewiX=EVXData.convertPsiFromBtoQ_Basis(n,xnewiV,ev_SK);
				UnitTests.checkDoubleArrayRel1(xnewiX,"xnewiX",xnewiXfm,"xnewiXfm",eps);
				UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpX,xnewiX),"<xnewpX|xnewiX>",TMatr.mulVV(xnewpV,xnewiV),"<xnewpV|xnewiV>",eps);
				UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpX,xnewiX),"<xnewpX|xnewiX>",TMatr.getAMatrB(xnewpZ,xnewiZ,mQs,newDim),"<xnewpZ|mQs|xnewiZ>",eps);
				UnitTests.checkDoubleRel1(TMatr.getAMatrB(xnewpX,xnewiX,SK,n),"<xnewpX|SK|xnewiX>",TMatr.getAMatrB(xnewpV,xnewiV,mSKv,n),"<xnewpV|mSKv|xnewiV>",eps);
			    }
			    // test matrix conversion formulas
			    UnitTests.checkDoubleRel1(TMatr.getAMatrB(xnewpXfm,xnewiXfm,SK,n),"<xnewpXfm|SK|xnewiXfmx>",TMatr.getAMatrB(xnewpZ,xnewiZ,mSKs,newDim),"<xnewpZ|mSKs|xnewiZ>",eps);
			    UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpXfm,xnewiXfm),"<xnewpXfm|xnewiXfm>",TMatr.getAMatrB(xnewpZ,xnewiZ,mQs,newDim),"<xnewpZ|mQs|xnewiZ>",eps);
			}
		    }
		    break;
		case BV_ORT,BV_ALL,UV_ORT_NOINJECT,UV_OFFDIAG0DIAGEQ:
		    // test without injection
		    final double [] bX=xProj,
			bV=EVXData.convertPsiFromQtoB_Basis(n,bX,ev_SK,TMatr.getUnitMatrix(n));
		    for(int p=0;p<newDim;p++){
			final double [] vbpX=EVXData.getPsiFromBasis(p,fromSubspaceToOrig,n,0),
			    PVp=TMatr.selectHermitianPartInPlace(nC,TMatr.mul_MA_x_MBTrans(nC,bX,nX,vbpX,nC));
			testPVp.accept(PVp);

			// check basis
			final double [] xnewpZ=new double [newDim];xnewpZ[p]=1;
			final double []
			    xnewpXfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewpZ,1),
			    xnewpV=t.fromXnewToXorig(xnewpZ),
			    xnewpX=EVXData.convertPsiFromBtoQ_Basis(n,xnewpV,ev_SK);
			UnitTests.checkDoubleArrayRel1(xnewpX,"xnewpX",vbpX,"vbpX",eps);
			UnitTests.checkDoubleArrayRel1(xnewpX,"xnewpX",xnewpXfm,"xnewpXfm",eps);
			// fails for UV_OFFDIAG0DIAGEQ
			if(linConstraintType!=LinConstraint.UV_OFFDIAG0DIAGEQ){
			    UnitTests.checkDoubleRel1(TMatr.mulVV(bV,xnewpV),"<uV|xnewpV>",0,"0",eps);
			    UnitTests.checkDoubleRel1(TMatr.mulVV(bX,xnewpX),"<u|xnewpX>",0,"0",eps);
			}
			// test Gram matrix is the same
			for(int i=0;i<newDim;i++){
			    final double [] xnewiZ=new double [newDim];xnewiZ[i]=1;
			    final double []
				xnewiXfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewiZ,1),
				xnewiV=t.fromXnewToXorig(xnewiZ),
				xnewiX=EVXData.convertPsiFromBtoQ_Basis(n,xnewiV,ev_SK);
			    UnitTests.checkDoubleArrayRel1(xnewiX,"xnewiX",xnewiXfm,"xnewiXfm",eps);
			    UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpX,xnewiX),"<xnewpX|xnewiX>",TMatr.mulVV(xnewpV,xnewiV),"<xnewpV|xnewiV>",eps);
			    UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpX,xnewiX),"<xnewpX|xnewiX>",TMatr.getAMatrB(xnewpZ,xnewiZ,mQs,newDim),"<xnewpZ|mQs|xnewiZ>",eps);
			    UnitTests.checkDoubleRel1(TMatr.getAMatrB(xnewpX,xnewiX,SK,n),"<xnewpX|SK|xnewiX>",TMatr.getAMatrB(xnewpV,xnewiV,mSKv,n),"<xnewpV|mSKv|xnewiV>",eps);

			    // test matrix conversion formulas
			    UnitTests.checkDoubleRel1(TMatr.getAMatrB(xnewpXfm,xnewiXfm,SK,n),"<xnewpXfm|SK|xnewiXfmx>",TMatr.getAMatrB(xnewpZ,xnewiZ,mSKs,newDim),"<xnewpZ|mSKs|xnewiZ>",eps);
			    UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpXfm,xnewiXfm),"<xnewpXfm|xnewiXfm>",TMatr.getAMatrB(xnewpZ,xnewiZ,mQs,newDim),"<xnewpZ|mQs|xnewiZ>",eps);
			}
		     }
		    break;
		default: throw new IllegalArgumentException();
		}
		System.err.println("CreateSubspaceOrtU: nEliminated="+t.nEliminated+" newDim="+newDim+" n="+n);
	    }
	    // END unit test
	}
    } // END OF class CreateSubspaceOrtU


    int nA=n;
    double [] T=ev_SK_basis,lambda=new double [nC*nC];
    LocalMax globMax=null,prevLocMax=null;
    for(int it=0;it<N_iterations;it++){
	// solve generalized regular eigenproblem
	final double [] evSKLambda=KGOSolutionVectorXVectorF.EV.getGEV_Generalized(
		nA,
		/*SKs=*/EVXData.convertMatrFromQtoB_Basis(n,adjSKWithLambda.apply(lambda),T,nA,0),
		/*Qs=*/TMatr.mul_MA_x_MBTrans(nA,T,n,T,nA),
		true);

	// BEGIN select the eigenvector among several
	LocalMax locMax=null;
	boolean flagGlobalMax=false;
	int nNegLambda=0,nFromTopLambda=0;
	// select eigenvalue, start with the maximal
	for(int p=nA;--p>=0;){
	    final double eLambda=evSKLambda[p];
	    nFromTopLambda++;if(eLambda<=0) nNegLambda++;
	    if(it<=0 || nFromTopLambda <=1 /*true || nNegLambda<=4 */ || locMax==null){
		final var e=new KGOEVSelection(
		    p,
		    TMatr.mul_MATrans_x_MB(n,T,nA,EVXData.getPsiFromBasis(p,evSKLambda,nA),1),
		    eLambda,
		    nC,nX);
		final var s=new LocalMax(e,it);
		if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(s)){
		    // found global max (across iterations)
		    globMax=locMax=s;
		    flagGlobalMax=true;
		}
		else if(!flagGlobalMax && (
			    locMax==null || !locMax.isValid() || locMax.isArgLocalBetter(s)
					   )){
		    // some local max
		    locMax=s;
		}

		if(eps>0){
		    UnitTests.checkDoubleRel(TMatr.getSpurM(nC,s.aep.gram.m),"spur G",nC,"nC",eps);
		    TestKGO.testOrthogonality("Adj EV ",nC,s.aep.getUAdj(),nX,eps);
		    System.err.println("it="+it+" e.ind="+e.ind+" ("+nA+") e.eLambda="+e.eLambda+" Forig="+s.Forig+" Fadj="+s.Fadj+" <u_orig|u_adj>="+TMatr.mulVV(s.aep.getUAdj(),s.aep.getUOrig())+" GEV="+java.util.Arrays.toString(java.util.Arrays.copyOf(s.aep.gram.ev_m,nC)));
		}
	    }
	}
	// END select the eigenvector among several

	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+":Selected ind="+locMax.e.ind+" ("+nA+") eLambda="+locMax.e.eLambda+"; Forig="+locMax.Forig+" Fadj="+locMax.Fadj+" sum1Lam="+locMax.aep.sum1Lam+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	final var l=new CreateSubspaceOrtU(locMax.aep.getUAdj(),LinConstraint.UV_OFFDIAG0DIAGEQ /*UV_ALL UV_ALL_NOINJECT UV_ORT BV_ALL  BV_ORT UV_ALL*/);
	// BEGIN calculate new values of Lagrange multipliers
	lambda=l.lambda;
	nA=l.newDim;
	T=l.fromSubspaceToOrig;
	// END calculate new values of Lagrange multipliers

	if(eps>0){
	    System.err.println("KGOIterationalTrySubspaceOrthogonal: it="+it+" Fadj="+locMax.Fadj+" Forig="+locMax.Forig+" ind="+locMax.e.ind+" ("+nA+") <u_orig|u_adj>="+TMatr.mulVV(locMax.aep.getUAdj(),locMax.aep.getUOrig())+" Gram EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(locMax.aep.gram.ev_m,nC)));
	}

	prevLocMax=locMax;
	if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(locMax)) {
	    globMax=locMax;
	}
    }
    evSelected=globMax.e;
    aep=globMax.aep;
    if(FORCE_DIAG_DUMP||eps>0||KGOOptimizationResult.FORCE_DIAG_RESULT){
	System.err.println("KGOIterationalTrySubspaceOrthogonal: RESULT:Selected ind="+evSelected.ind+" ("+nA+")  Forig="+TMatr.getAMatrA(aep.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(aep.getUAdj(),SK,n)+" at iteration="+globMax.itNum+" N_iterations="+N_iterations);
    }
}


    /** Minimal norm for Gaussian elimination.
     */
    private static final double condMaxToFail_LinearSystem=1e-10;

    /** Possible linear constraints to apply. */
    private static enum LinConstraint {
	/** New basis orthogonal to u, single constraint. Actually does nothing
	 * as we construct the basis fron u and linear constraints basis.
	 */
	UV_ORT,
	UV_ORT_NOINJECT,
	/** Set all Herm(UV) matrix to zero, {@code nC*(nC+1)/2} constraints.
	   @see LinearConstraints#getOrthogonalAllProj0
	 */
	UV_ALL,
	/**
	   @see LinearConstraints#getOrthogonalOffdiag0DiagEq
	 */
	UV_OFFDIAG0DIAGEQ,

	/** Set all projections to vector \(
b_{sq}=\sum\limits_{j^{\prime}=0}^{nC-1}\sum\limits_{k^{\prime}=0}^{nX-1}
S_{sq;j^{\prime}k^{\prime}}u_{j^{\prime}k^{\prime}}
-  \sum\limits_{j^{\prime}=0}^{nC-1} \lambda_{sj^{\prime}} u_{j^{\prime}q}
\) to zero:
\(0=\sum\limits_{i=0}^{nC-1}\sum\limits_{q=0}^{nX-1}b_{iq}v_{iq}\), one linear constraint.
	*/
	BV_ORT,
	/** Set to zero all diagonal \(0=\mathrm{Herm}\left(\sum\limits_{q=0}^{nX-1}b_{iq}v_{jq}\right)\),
	 * {@code nC*(nC+1)/2} linear constraints.
	 */
	BV_ALL
	;
    }


}
