package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;

/**
 * A Knowledge Generalized Operator optimization result.
 * @version $Id: KGOOptimizationResult.java,v 1.88 2024/05/07 19:52:34 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOOptimizationResult {

/** The list of possible approximations.
 */
public static enum Approximation {
    MAXEV_SVDADJ,
    MAXEV_EVADJ,
    MAXEV_NOADJ,
    LISTEV_NOADJ,
    LISTEV_SVDADJ,
    LISTEV_EVADJ,
    /** Adjust least squares to unitary with Gram matrix eigenstates.
     */
    LSQ_EVADJ,
    /** Adjust least squares to unitary with SVD.
     */
    LSQ_SVDADJ,
    LSQ_SVDADJ_CHECKSIGNS,
    /** Non-adjusted, regular least squares.
     */
    LSQ_NOADJ,

    /** Lagrange multipliers iteration with \(Q_{ij}=\delta_{ij}\).
     */
    ITERATIONS_KGOIterationalSimpleOptimizationU,
    /** Lagrange multipliers and variate states orthogonal U.
     */
    ITERATIONS_KGOIterationalTrySubspaceOrthogonal,
    /** The only iteration that works. */
    ITERATIONS_KGOIterationalSubspaceLinearConstraints,
    /** The same as {@link ITERATIONS_KGOIterationalSimpleOptimizationU} but uses different basis,
     * the result shuld be the same.
     */
    ITERATIONS_KGOIterationalMultipleTransforms,

    /** Calculate Lagrange multipliers on subspace. */
    ITERATIONS_KGOIterationalLagrangeMultipliersPartialSubspace,

    /** Lagrange multipliers with custom denominator.
     */
    ITERATIONS_KGOIterationalGEVAdj_Q_denominator,
    /** Lagrange multipliers in denominator. Very similar (but not identical)
     * to {@link ITERATIONS_KGOIterationalSimpleOptimizationU}.
     */
    ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU,


    ITERATIONS_KGOIterationalLambda2,

    ITERATIONS_KGOIterationalLinearConstraints,
    ITERATIONS_KGOIterationalLinearConstraintsE,
    ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom,



    /** Currently broken.
     */
    ITERATIONS_KGOIterationalContractionMapping
    ;
    static final Approximation DEFAULT=ITERATIONS_KGOIterationalLinearConstraintsE;

    private static final java.util.List<Approximation> BROKENLIST=
	    java.util.List.of(
		// slow convergence
		ITERATIONS_KGOIterationalLinearConstraints,
		ITERATIONS_KGOIterationalLinearConstraintsE,
		ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom,
		// poor convergence
		ITERATIONS_KGOIterationalSimpleOptimizationU,
		// the same this but less stable
		ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU,
		// the same as ITERATIONS_KGOIterationalSimpleOptimizationU,
		ITERATIONS_KGOIterationalMultipleTransforms,
		// work in progress
		ITERATIONS_KGOIterationalLagrangeMultipliersPartialSubspace,
		// work in progress
		ITERATIONS_KGOIterationalTrySubspaceOrthogonal,
		ITERATIONS_KGOIterationalSubspaceLinearConstraints,
		// work in progress, currently broken
		ITERATIONS_KGOIterationalGEVAdj_Q_denominator,
		// not actually a good idea
		ITERATIONS_KGOIterationalLambda2,
		// this is completely broken
		ITERATIONS_KGOIterationalContractionMapping
					     );

    public static java.util.List<Approximation> valuesSkipBroken(){
	return java.util.Arrays.stream(values()).filter(e -> !BROKENLIST.contains(e)).toList();
    }
}

static final boolean FORCE_DIAG_RESULT=true;


/** Initial state before solution adjustement.
 */
final KGOEVSelection evSelected;

/** Actual solution.
 * Has both: intitial and adjusted solutions
 * satisfying the full set of constraints.
 */
final UAdjustment uadj;

/** Approximation used, all possible values are listed in {@link Approximation}.
 * @see Approximation
 */
final Approximation approximation;

KGOOptimizationResult(
    final int nF,final int nX,
    final double [] SK,
    final KGOEVSelection evSelected,
    final UAdjustment uadj,
    final Approximation approximation)
{
    final int n=nF*nX;
    if(SK.length!=n*n) throw new IllegalArgumentException();
    if(nF>nX) throw new IllegalArgumentException("nF="+nF+">nX="+nX+" is disabled.");
    this.evSelected=evSelected;
    this.uadj=uadj;
    this.approximation=approximation;

    if(FORCE_DIAG_RESULT){
	System.err.println("KGOOptimizationResult: RESULT: Selected ind="+evSelected.ind+" ("+n+")  Forig="+TMatr.getAMatrA(uadj.getUOrig(),SK,n)+" Fadj="+TMatr.getAMatrA(uadj.getUAdj(),SK,n)+" approximation="+approximation);
    }
}


    /** Select from eigenvalues list.
     * @see Approximation#LISTEV_NOADJ
     */
    public static KGOOptimizationResult fromListEV(
	final int nF,final int nX,
	final double [] SK,
	final java.util.function.ToDoubleFunction<double []> selectEV,
	final Approximation approximation,
	final double eps){
	final int n=nF*nX;
	if(SK.length!=n*n) throw new IllegalArgumentException();
	if(nF>nX) throw new IllegalArgumentException("nF="+nF+">nX="+nX+" is disabled.");
	if(approximation!=Approximation.LISTEV_NOADJ) throw new IllegalArgumentException("approximation="+approximation);

	// solve regular eigenproblem
	final double [] evSK=KGOSolutionVectorXVectorF.EV.getEV(n,SK,true);
	double maxSK=0;
	KGOEVSelection evSelected=null;
	UAdjustment aep=null;
	for(var e:KGOEVSelection.selectTopEV(n/* try them all */,evSK,nF,nX)){
	    final var ucur=e.getSVDNonAdjusted();
	    final double curSK=selectEV.applyAsDouble(ucur.getUOrig());
	    if(evSelected==null|| !Double.isFinite(maxSK) || curSK>=maxSK){
		evSelected=e;
		aep=ucur;
		maxSK=curSK;
	    }
	    if(eps>0){
		System.err.println("fromListEV "+e.ind+" curSK="+curSK);
	    }
	}
	return new KGOOptimizationResult(nF,nX,SK,evSelected,aep,approximation);
    }

    /** Common one shot approximations for SVD and eigenvalue adjustments.
     * @see Approximation#MAXEV_EVADJ
     * @see Approximation#LISTEV_SVDADJ
     * @see Approximation#LISTEV_EVADJ
     * @see Approximation#MAXEV_SVDADJ
     * @see Approximation#MAXEV_NOADJ
     */
    public static KGOOptimizationResult fromEVProblemOneShotSVDAdj(
	final int nF,final int nX,
	final double [] SK,
	final Approximation approximation,
	final double eps){

	final int n=nF*nX;
	if(SK.length!=n*n) throw new IllegalArgumentException();
	if(nF>nX) throw new IllegalArgumentException("nF="+nF+">nX="+nX+" is disabled.");

	// solve regular eigenproblem
	final double [] evSK=KGOSolutionVectorXVectorF.EV.getEV(n,SK,true);

	if(approximation==Approximation.MAXEV_EVADJ) {
	    final var evSelected=KGOEVSelection.selectTopEV(1,evSK,nF,nX).get(0);
	    final var aep=evSelected.getEVAdjustedTo1();
	    return new KGOOptimizationResult(nF,nX,SK,evSelected,aep,approximation);
	}  else if(approximation==Approximation.LISTEV_SVDADJ||
		   approximation==Approximation.LISTEV_EVADJ){
	    double maxSK=0;
	    KGOEVSelection evsel=null;
	    UAdjustment uadjsel=null;
	    // select eigenvalue
	    for(var e:KGOEVSelection.selectTopEV(n/* try them all */,evSK,nF,nX)){
		final UAdjustment us=switch(approximation){
		case LISTEV_SVDADJ -> {// SVD type of adjustment
		    final var s=e.getSVDAdjustedTo1();
		    if(eps>0){
			final double origcurSK=TMatr.getAMatrA(s.getUOrig(),SK,n),
			    curSK=TMatr.getAMatrA(s.getUAdj(),SK,n);
			UnitTests.checkDoubleRel(e.eLambda*nF,"e.eLambda*nF",origcurSK,"origcurSK",eps);
			UnitTests.checkDoubleRel(TMatr.mulVV(s.svd_orig.S,s.svd_orig.S),"sum (S^2)",nF,"nF",eps);
			System.err.println("EV "+e.ind+" orig F="+origcurSK+" adj F="+curSK+" SVD orig="+java.util.Arrays.toString(s.svd_orig.S));
		    }
		    yield s;
		}
		default ->{ // EV type of adjustment
		    final var s=e.getEVAdjustedTo1();
		    if(eps>0){
			final double origcurSK=TMatr.getAMatrA(s.getUOrig(),SK,n),
			    curSK=TMatr.getAMatrA(s.getUAdj(),SK,n),
			    gramEV[]=java.util.Arrays.copyOf(s.gram.ev_m,nF);
			UnitTests.checkDoubleRel(e.eLambda*nF,"e.eLambda*nF",origcurSK,"origcurSK",eps);
			UnitTests.checkDoubleRel(TMatr.sumElements(gramEV),"sum (gramEV)",nF,"nF",eps);
			System.err.println("EV "+e.ind+" orig F="+origcurSK+" adj F="+curSK+" Gram EV orig="+java.util.Arrays.toString(gramEV));
		    }
		    yield s;
		}
		};

		final double curSK=TMatr.getAMatrA(us.getUAdj(),SK,n);
		if(evsel==null || !Double.isFinite(maxSK) || curSK>=maxSK){
		    evsel=e;
		    uadjsel=us;
		    maxSK=curSK;
		}
	    }
	    return new KGOOptimizationResult(nF,nX,SK,evsel,uadjsel,approximation);
	}

	final KGOEVSelection evSelected;
	final AdjustedStateToUnitaryWithSVD svd;
	if(approximation==Approximation.MAXEV_SVDADJ){
	    evSelected=KGOEVSelection.selectTopEV(1,evSK,nF,nX).get(0);
	    svd=evSelected.getSVDAdjustedTo1();
	} else if(approximation==Approximation.MAXEV_NOADJ){
	    evSelected=KGOEVSelection.selectTopEV(1,evSK,nF,nX).get(0);
	    svd=evSelected.getSVDNonAdjusted();
	} else {
	    throw new IllegalArgumentException("Not Implemented "+approximation);
	}

	if(eps>0){
	    final double Forig=TMatr.getAMatrA(svd.getUOrig(),SK,n),
		Fadj=TMatr.getAMatrA(svd.getUAdj(),SK,n);
	    final double [] topNfEV=java.util.Arrays.copyOfRange(evSK,n-nF,n);
	    UnitTests.checkDoubleRel(evSelected.eLambda,"evSelected.eLambda",Forig/nF,"Forig/nF",eps);
	    UnitTests.checkDoubleRel(TMatr.mulVV(svd.svd_orig.S,svd.svd_orig.S),"sum (S^2)",nF,"origcurSK",eps);
	    System.err.println("KGOApproximateOptimization: approximation="+approximation+" evSelected.ind="+evSelected.ind+" (of "+n+") Fadj/nF="+(Fadj/nF)+" Forig/nF="+(Forig/nF)+" Flambda="+evSelected.eLambda+" topEV("+nF+" of "+n+")="+java.util.Arrays.toString(topNfEV)+" SVD orig="+java.util.Arrays.toString(svd.svd_orig.S)+" approximation="+approximation+" eps="+eps);
	}
	return new KGOOptimizationResult(nF,nX,SK,evSelected,svd,approximation);
    }

    /** Least sqaures with an adjustements and without.
     * @see Approximation#LSQ_SVDADJ
     * @see Approximation#LSQ_EVADJ
     * @see Approximation#LSQ_NOADJ
     * @see Approximation#LSQ_SVDADJ_CHECKSIGNS
     */
    public static KGOOptimizationResult fromLeastSquaresOneShot(
	final int nF,final int nX,
	final double [] SK,
	final double [] FX,
	final Approximation approximation,
	final double eps){

	final int n=nF*nX;
	if(SK.length!=n*n||FX.length!=n) throw new IllegalArgumentException();
	if(nF>nX) throw new IllegalArgumentException("nF="+nF+">nX="+nX+" is disabled.");

	final var v=TMatr.getVectorScaled(FX,1/Math.sqrt(TMatr.mulVV(FX,FX)));
	final KGOEVSelection evSelected=new KGOEVSelection(-1,v,TMatr.getAMatrA(v,SK,n),nF,nX);
	final AdjustedStateToUnitaryWithSVD svd;

	if(approximation==Approximation.LSQ_SVDADJ){
	    svd=evSelected.getSVDAdjustedTo1();
	} else if(approximation==Approximation.LSQ_EVADJ){
	    final var aep=evSelected.getEVAdjustedTo1();
	    return new KGOOptimizationResult(nF,nX,SK,evSelected,aep,approximation);
	} else if(approximation==Approximation.LSQ_NOADJ){
	    svd=evSelected.getSVDNonAdjusted();
	} else if(approximation==Approximation.LSQ_SVDADJ_CHECKSIGNS){
	    svd=evSelected.getSVDAdjustedToPlusMinus1(SK);
	} else {
	    throw new IllegalArgumentException("Not Implemented "+approximation);
	}


	if(eps>0){
	    UnitTests.checkDoubleRel(evSelected.eLambda,"evSelected.eLambda",TMatr.getAMatrA(svd.getUOrig(),SK,n)/nF,"origcurSK",eps);
	    UnitTests.checkDoubleRel(TMatr.mulVV(svd.svd_orig.S,svd.svd_orig.S),"sum (S^2)",nF,"origcurSK",eps);
	    System.err.println("KGOApproximateOptimization: approximation="+approximation+" evSelected.ind="+evSelected.ind+" (of "+n+") original F="+evSelected.eLambda+" SVD orig="+java.util.Arrays.toString(svd.svd_orig.S)+" eps="+eps);
	}
	return new KGOOptimizationResult(nF,nX,SK,evSelected,svd,approximation);
    }

    /** Iteration methods.
     * @see Approximation#ITERATIONS_KGOIterationalLagrangeMultipliersPartialSubspace
     * @see Approximation#ITERATIONS_KGOIterationalGEVAdj_Q_denominator
     * @see Approximation#ITERATIONS_KGOIterationalSimpleOptimizationU
     * @see Approximation#ITERATIONS_KGOIterationalMultipleTransforms
     * @see Approximation#ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU
     * @see Approximation#ITERATIONS_KGOIterationalTrySubspaceOrthogonal
     * @see Approximation#ITERATIONS_KGOIterationalSubspaceLinearConstraints

     * @see Approximation#ITERATIONS_KGOIterationalLambda2
     * @see Approximation#ITERATIONS_KGOIterationalLinearConstraints
     * @see Approximation#ITERATIONS_KGOIterationalLinearConstraintsE
     * @see Approximation#ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom
     * @see Approximation#ITERATIONS_KGOIterationalContractionMapping
     */
    public static KGOOptimizationResult fromIterationsLagrangeMultipliers(
	final int nF,final int nX,
	final double [] SK,
	final Approximation approximation,
	final double eps){

	return switch(approximation){
	case ITERATIONS_KGOIterationalLagrangeMultipliersPartialSubspace ->{
	    final var kgo=new KGOIterationalLagrangeMultipliersPartialSubspace(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalGEVAdj_Q_denominator -> {
	    final var kgo=new KGOIterationalGEVAdj_Q_denominator(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalSimpleOptimizationU -> {
	    final var kgo=new KGOIterationalSimpleOptimizationU(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalMultipleTransforms -> {
	    final var kgo=new KGOIterationalMultipleTransforms(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU -> {
	    final var kgo=new KGOIterationalLagrangeMultipliersInDenominatorU(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalTrySubspaceOrthogonal -> {
	    final var kgo=new KGOIterationalTrySubspaceOrthogonal(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalSubspaceLinearConstraints-> {
	    final var kgo=new KGOIterationalSubspaceLinearConstraints(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
       /* Those not workig well are below. */
	case ITERATIONS_KGOIterationalLambda2 -> {
	    final var kgo=new KGOIterationalLambda2(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalLinearConstraints -> {
	    final var kgo=new KGOIterationalLinearConstraints(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalLinearConstraintsE -> {
	    final var kgo=new KGOIterationalLinearConstraintsE(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom -> {
	    final var kgo=new KGOIterationalLinearConstraintsExtraDegreesOfFreedom(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	case ITERATIONS_KGOIterationalContractionMapping -> {
	    // Currently completely broken.
	    final var kgo=new KGOIterationalContractionMapping(nF,nX,SK,eps);
	    yield new KGOOptimizationResult(nF,nX,SK,kgo.evSelected,kgo.aep,approximation);
	}
	default -> throw new IllegalArgumentException("Not Implemented "+approximation);
	};
    }

}
