package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Linsystems;
import com.polytechnik.utils.ObservationVectorXVectorF;
import com.polytechnik.utils.EVSolver;
import com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis;
import com.polytechnik.utils.VectorXVectorFPointEvaluatable;
import com.polytechnik.utils.VectorXVectorFPointAccess;
import com.polytechnik.utils.RegularizedVectorXVectorF;
import com.polytechnik.utils.RegularizedVectorX;
import com.polytechnik.utils.RegularizationMethod;
import com.polytechnik.utils.DataReadObservationVectorXVectorF;

/** Knowledge Generalizing Operator solution to a vector X to vector F problem.
 * The X- and F- spaces are assumed to be regularized: XX and FF are not degenerated and diagonal.
 * @version $Id: KGOSolutionVectorXVectorF.java,v 1.136 2024/05/07 08:18:57 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOSolutionVectorXVectorF implements  VectorXVectorFPointEvaluatable {
    /** Standard EV solver. */
    static final EVSolver EV=new OrthogonalPolynomialsMonomialsBasis().EV;

    final RegularizedVectorXVectorF reg;
    public final int nX,nF;
    private final double [] fF, xX;
    private final double [] S_FXFX /*  nF*NX x nF*nX */;

    final KGOOptimizationResult.Approximation approximation;

    final SKFromSampleCalculationMethod SKtype;

    /** Contributing subspace.
     */
    private final double [] CX;

    /** Old "projective" version.
     */
    private final double [] FX /* nF x nX */;

    /** Maximal possible coverage among all projective-type transforms. */
    public final double FTOT;

    /** A coverage corresponding to "Direct Projection" type of expansion:
     * least squares value is calculated first, then a f-localized function
     * is built at this value. The value may be greater than {@link #FTOT}.
     */
    public final double FLS;

    /** A coverage corresponding to found mapping.
     */
    public final double FCOVU;

    private final VectorXVectorFPointEvaluatableCommon evaluator;
    /** For unit tests. */
    private final AdjustedStateToUnitaryWithSVD test_svd;


    /**
     * @param ro Read Observations (sample).
     * @param regularization_methodX regularizatuion method of {@code x}.
     * @param regularization_methodF regularizatuion method of {@code f}.
     * @param eps Run a selftest with this {@code eps} if positive.
     */
    KGOSolutionVectorXVectorF(
	final DataReadObservationVectorXVectorF ro,
	final RegularizationMethod regularization_methodX,
	final RegularizationMethod regularization_methodF,
	final KGOOptimizationResult.Approximation approximation,
	final SKFromSampleCalculationMethod SKtype,
	final double eps){

	final var t0=System.currentTimeMillis();System.err.println("before regularization");
	this.reg=ro.getDataRegularized_OLDSCALARCODE(regularization_methodX,regularization_methodF);
	final var t1=System.currentTimeMillis();System.err.println("after regularization dt="+(t1-t0));
	this.nX=reg.X.nX;
	this.nF=reg.F.nX;
	if(nF>nX) throw new RuntimeException("Currently nF>nX is not supported: nF="+nF+" nX="+nX);

	// These can be calculated without secondary sampling.
	// But as we use secondaty sampling anyway -- simplify the code.
	// Old projective version. Used in unit tests only
	FX=reg.getFXmoments(ro);
	// needed for F->forig transform
	fF=ro.getFun1Fun2Averaged(o->o.f,o->reg.F.getRegularizedBasisXFromOriginalX(o.f),o->1);
	if(fF.length!=reg.F.noriginal*nF) throw new IllegalArgumentException();
	// needed  for X->xorig transform
	xX=ro.getFun1Fun2Averaged(o->o.x,o->reg.X.getRegularizedBasisXFromOriginalX(o.x),o->1);
	if(xX.length!=reg.X.noriginal*nX) throw new IllegalArgumentException();

	final var t2=System.currentTimeMillis();System.err.println("regular moments done dt="+(t2-t1));

	CX=TestKGO.orthogonalizeU(nF,nX,FX);
	if(eps>0){
	    final int nC=CX.length/nX;
	    TestKGO.testOrthogonality("CX test",nC,CX,nX,eps);
	}

	// Christoffel function product moments can be calculated
	// only using secondary sampling.
	final double [] m_FXFX=switch(SKtype) {
	// two Christoffel functions product
	// with adjusted number of degrees of freedom
	case CHRISTOFFEL_DEGREESOFFREEDOMADJUSTED ->
	        reg.getChristoffelFunctionsProductMomentsInFXBasis_nCdegreesOfFreedom(ro,CX);
	// two Christoffel functions product
	case CHRISTOFFEL ->
	        reg.getChristoffelFunctionsProductMomentsInFXBasis(ro);
	// a single (F) Christoffel function
	case FXFX_F_CHRISTOFFEL -> false?
	        // regular
	        reg.getDyadicProductMomentsInFXBasis(ro,o->{
		final double [] rF=reg.F.getRegularizedBasisXFromOriginalX(o.f);
		return 1/TMatr.mulVV(rF,rF);
		    })
	        :
	        // speed optimized
	        getFXFX_F_Christoffel_SPEED_Optimization(ro)
	        ;
	// No Christoffel function, plain FXFX
	case FXFX -> reg.getDyadicProductMomentsInFXBasis(ro,o->1);
	default -> throw new IllegalArgumentException("SKtype="+SKtype);
	};

	if(m_FXFX.length!=nX*nF*nX*nF) throw new IllegalArgumentException();

	// Create the tensor for U|S|U
	// unit Gram matrix makes tensor the same
	S_FXFX=m_FXFX;
	final var t3=System.currentTimeMillis();System.err.println("Christoffel function moments done dt="+(t3-t2));

	this.SKtype=SKtype;
	// Algorithm groups (there specific algrithm selection is inside group processor)
	final KGOOptimizationResult kgoOpt=switch(approximation) {
	// Least squares one-step adjustment
	case LSQ_SVDADJ,
	     LSQ_EVADJ,
	     LSQ_NOADJ,
	     LSQ_SVDADJ_CHECKSIGNS
	     ->
	       KGOOptimizationResult.fromLeastSquaresOneShot(nF,nX,S_FXFX,FX,approximation,eps);
	// check all eigenvalues with regular coverage
	case LISTEV_NOADJ ->
	   KGOOptimizationResult.fromListEV(
	       nF,nX,S_FXFX,ucur->getFtotalForTransform(ro,ucur),approximation,eps);
	// Iterational
	case ITERATIONS_KGOIterationalSimpleOptimizationU, // more or less OK
	     ITERATIONS_KGOIterationalMultipleTransforms, // same thing but not well tested
	     ITERATIONS_KGOIterationalTrySubspaceOrthogonal, // work in progress
	     ITERATIONS_KGOIterationalSubspaceLinearConstraints, // work in progress
	     ITERATIONS_KGOIterationalLagrangeMultipliersPartialSubspace, // work in progress
	     ITERATIONS_KGOIterationalGEVAdj_Q_denominator, // work in progress
	     ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU, // the concept works but gives no improvement
	     ITERATIONS_KGOIterationalLambda2,
	     ITERATIONS_KGOIterationalLinearConstraints, // work in progress
	     ITERATIONS_KGOIterationalLinearConstraintsE, // work in progress
	     ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom,
	     ITERATIONS_KGOIterationalContractionMapping // requires more work, currently broken
	     ->
	       KGOOptimizationResult.fromIterationsLagrangeMultipliers(nF,nX,S_FXFX,approximation,eps);
	// All others (eigenstate bases with one step adjustment).
	// Will fail inside if not supported.
	default ->KGOOptimizationResult.fromEVProblemOneShotSVDAdj(nF,nX,S_FXFX,approximation,eps);
	};
	this.approximation=kgoOpt.approximation;
	final var t4=System.currentTimeMillis();System.err.println("Calculations done dt="+(t4-t3));


	// The best coverage among all projections.
	this.FTOT=TMatr.getSpurM(nX,getXF_KFF_FX(reg.getFChristoffelFunctionFMoments(ro)));
	// The coverage of states localized at least squares value,
	// this is the coverage of the "Direct Projection" model.
	this.FLS=getFtotalForTransform(ro,FX);
	this.FCOVU=getFtotalForTransform(ro,kgoOpt.uadj.getUAdj());

	// possibly set it to null if eps=0 as no test run is expected
	this.test_svd=kgoOpt.uadj.getSVDForTesting();
	this.evaluator=new VectorXVectorFPointEvaluatableCommon(kgoOpt.uadj,reg.X,reg.F,CX,this::forigFromF,FX,SKtype);

	if(eps>0){
	    selfTest(ro,eps);
	}
    }

    double [] forigFromF(final double [] F){
	if(F.length!=nF) throw new IllegalArgumentException();
	return TMatr.mul_MA_x_MBTrans(reg.F.noriginal,fF,F.length,F,1);
    }

    double getFtotalForTransform(final DataReadObservationVectorXVectorF ro,
				 final double [] u_jk){
	if(u_jk.length!=nF*nX) throw new IllegalArgumentException();
	return ro.data.stream().mapToDouble(o -> {
		final double[] Xl=reg.X.getRegularizedBasisXFromOriginalX(o.x),
		    Fl=reg.F.getRegularizedBasisXFromOriginalX(o.f),
		    FLSl=TMatr.mul_MA_x_MB(nF,u_jk,nX,Xl,1);
		return o.weight*Math.pow(TMatr.mulVV(Fl,FLSl),2)/
		    (TMatr.mulVV(Fl,Fl)*TMatr.mulVV(FLSl,FLSl));
	    }).sum();
    }

    public VectorXVectorFPointEvaluatableCommon.VectorXVectorFPointAccessWithDimensionAdjusted evaluateAt(final double [] xorig){return evaluator.evaluateAt(xorig);}

    private double [] getXF_KFF_FX(final double [] KFF){
	return TMatr.mul_MATrans_x_MB(nX,FX,nF,TMatr.mul_MA_x_MB(nF,KFF,nF,FX,nX),nX);
    }


    /** Optimized version for speed, without re-calculation and extra function calls.
     */
    private double [] getFXFX_F_Christoffel_SPEED_Optimization(final DataReadObservationVectorXVectorF ro){
	final int n=nF*nX;
	final double [] scalprods=new double [n*n];
	for(ObservationVectorXVectorF o:ro.data)
	{
	    final double []
		F=reg.F.getRegularizedBasisXFromOriginalX(o.f),
		dyadic=TMatr.dyadicMatrix(F,reg.X.getRegularizedBasisXFromOriginalX(o.x));
	    final double cfactor=o.weight/TMatr.mulVV(F,F);
	    for(int j=0,of=0;j<dyadic.length;j++){
		final double jfact=dyadic[j]*cfactor;
		for(int k=0;k<dyadic.length;k++,of++){
		    scalprods[of]+=dyadic[k]*jfact;
		}
	    }
	}
	return scalprods;
    }



    /** Contributing subspace old projective approach.
     */
    private void testProjectiveContributingBasisAndFTOTCoverage(final DataReadObservationVectorXVectorF ro,final double eps){

	// match FTOT
	final double _FTOT=ro.data.stream().mapToDouble(o -> {
		final double[] X=reg.X.getRegularizedBasisXFromOriginalX(o.x),
		    F=reg.F.getRegularizedBasisXFromOriginalX(o.f),
		    FprojX=TMatr.mul_MATrans_x_MB(nX,FX,nF,F,1);
		return o.weight*TMatr.mulVV(FprojX,FprojX)/TMatr.mulVV(F,F);
	    }).sum();
	UnitTests.checkDoubleRel(FTOT,"FTOT",_FTOT,"_FTOT",eps);
	// test contributing subspace is the maximal
	final int nC=CX.length/nX;
	// orthogonality
	TestKGO.testOrthogonality("CX ort",nC,CX,nX,eps);

	final double [] CF=TMatr.mul_MA_x_MBTrans(nC,CX,nX,FX,nF);
	final double __FTOT=ro.data.stream().mapToDouble(o -> {
		final double[] F=reg.F.getRegularizedBasisXFromOriginalX(o.f),
		    FprojC=TMatr.mul_MA_x_MB(nC,CF,nF,F,1);
		return o.weight*TMatr.mulVV(FprojC,FprojC)/TMatr.mulVV(F,F);
	    }).sum();
	UnitTests.checkDoubleRel(FTOT,"FTOT",__FTOT,"__FTOT",eps);

	// Test eigenvalue-based contributing subspace
	final double [] CX_ev=TestKGO.getContributingSubspaceEV(getXF_KFF_FX(reg.getFChristoffelFunctionFMoments(ro)),nF,nX,eps);
	final int nC_ev=CX_ev.length/nX;
	TestKGO.testOrthogonality("CX_ev ort",nC_ev,CX_ev,nX,eps);
	// test the contributing subspace is the same
	for(int j=0;j<nC;j++){
	    double s=0;
	    for(int jp=0;jp<nC_ev;jp++){
		s+=Math.pow(TMatr.mulVV(nX,CX,j*nX,CX_ev,jp*nX),2);
	    }
	    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
	}

	// test FTOT coverage
	final double [] CF_ev=TMatr.mul_MA_x_MBTrans(nC_ev,CX_ev,nX,FX,nF);
	final double ___FTOT=ro.data.stream().mapToDouble(o -> {
		final double[] F=reg.F.getRegularizedBasisXFromOriginalX(o.f),
		    FprojC=TMatr.mul_MA_x_MB(nC_ev,CF_ev,nF,F,1);
		return o.weight*TMatr.mulVV(FprojC,FprojC)/TMatr.mulVV(F,F);
	    }).sum();
	UnitTests.checkDoubleRel(FTOT,"FTOT",___FTOT,"___FTOT",eps);

	final double _FLS=ro.data.stream().mapToDouble(
	    o -> o.weight*getDirectProjectionProbability(evaluateAt(o.x),o.f)
						       ).sum();
	UnitTests.checkDoubleRel(FLS,"FLS",_FLS,"_FLS",eps);

	// Test adjusted Christoffel function

	final double [] iFFa=Linsystems.getInvertedMatrix(nF,nF,TMatr.mul_MA_x_MBTrans(nF,FX,nX,FX,nF)),
	    XF_iFFa_FX=TMatr.mul_MATrans_x_MB(nX,FX,nF,TMatr.mul_MA_x_MB(nF,iFFa,nF,FX,nX),nX);
	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
	    "ChristoffelCX",
	    ro,o->{
		final var r=reg.X.getRegularizedBasisXFromOriginalX(o.x);
		return 1/TMatr.getAMatrA(r,XF_iFFa_FX,r.length);
	    },
	    ro,o->{
		final var r=reg.X.getRegularizedBasisXFromOriginalX(o.x);
		final var C=TMatr.mul_MA_x_MB(this.CX.length/nX,this.CX,nX,r,1);
		return 1/TMatr.mulVV(C,C);
	    },
	    eps);
    }

    private void matchMatrices(final DataReadObservationVectorXVectorF ro,final double eps){
	// Make sure the matrices are diagonal
	final double [] XX=ro.getFun1Fun2Averaged(o->reg.X.getRegularizedBasisXFromOriginalX(o.x),
						  o->reg.X.getRegularizedBasisXFromOriginalX(o.x),
						  o->1);
	UnitTests.checkDoubleArrayRel1(XX,"XX",TMatr.getUnitMatrix(reg.X.nX),"1",eps);
	final double [] FF=ro.getFun1Fun2Averaged(o->reg.F.getRegularizedBasisXFromOriginalX(o.f),
						  o->reg.F.getRegularizedBasisXFromOriginalX(o.f),
						  o->1);
	UnitTests.checkDoubleArrayRel1(FF,"FF",TMatr.getUnitMatrix(reg.F.nX),"1",eps);
	final double [] KCKF_FXFX=ro.getFun1Fun2Averaged(
	    o->TMatr.dyadicMatrix(reg.F.getRegularizedBasisXFromOriginalX(o.f),reg.X.getRegularizedBasisXFromOriginalX(o.x)),
	    o->TMatr.dyadicMatrix(reg.F.getRegularizedBasisXFromOriginalX(o.f),reg.X.getRegularizedBasisXFromOriginalX(o.x)),
	    o->{final var e=evaluateAt(o.x);
		return 1/(
		    switch(SKtype){
		    case CHRISTOFFEL_DEGREESOFFREEDOMADJUSTED ->
			e.getChristoffelOatXDimensionAdjusted()*e.getChristoffelOatF(o.f);
		    case CHRISTOFFEL ->
			e.getChristoffelOatX()*e.getChristoffelOatF(o.f);
		    case FXFX_F_CHRISTOFFEL -> e.getChristoffelOatF(o.f);
		    case FXFX -> 1;
		    default -> throw new IllegalArgumentException();
		    });}
							 );
	UnitTests.checkDoubleArrayRel1(KCKF_FXFX,"KCKF_FXFX",S_FXFX,"S_FXFX",eps);
	System.err.println("KGOSolutionVectorXVectorF:matchMatrices OK nX="+nX+" nF="+nF+" eps="+eps);
    }

    private void matchSampleValues(final DataReadObservationVectorXVectorF ro,final double eps){

	final double [] fX=ro.getFun1Fun2Averaged(o->o.f,o->reg.X.getRegularizedBasisXFromOriginalX(o.x),o->1);

	// x-expansion in X should be exact
	TestObservationSequenceFromSampleVectorXVectorF.matchArrayFunctionOnTwoSamples(
	    null/*"LeastSquares x in X"*/,
	    ro,o->leastSquaresO(xX,reg.X,o.x),
	    ro,o->o.x,eps);
	// f-expansion in F should be exact
	TestObservationSequenceFromSampleVectorXVectorF.matchArrayFunctionOnTwoSamples(
	    null/*"LeastSquares f in F"*/,
	    ro,o->leastSquaresO(fF,reg.F,o.f),
	    ro,o->o.f,eps);

	// Test sample values to match:
	// Least squares
	TestObservationSequenceFromSampleVectorXVectorF.matchArrayFunctionOnTwoSamples(
	    null/*"LeastSquares f in X"*/,
	    ro,o->leastSquaresO(fX,reg.X,o.x),
	    ro,o->evaluateAt(o.x).getFLeastSquaresX(),eps);
	// Christoffel X
	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
			    null,//"ChristoffelX",
			    ro,o->christoffel(reg.X,o.x),
			    ro,o->1/evaluateAt(o.x).getChristoffelOatX(),
			    eps);
	// Christoffel X dimesion adjusted
	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
			    null,//"ChristoffelX",
			    ro,o->{
				final var C=TMatr.mul_MA_x_MB(CX.length/nX,CX,nX,reg.X.getRegularizedBasisXFromOriginalX(o.x),1);
				// for unit Gram matrix
				return 1/TMatr.mulVV(C,C);
			    },
			    ro,o->1/evaluateAt(o.x).getChristoffelOatXDimensionAdjusted(),
			    eps);
	// Christoffel F
	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
			    null,//"ChristoffelF",
			    ro,o->christoffel(reg.F,o.f),
			    ro,o->1/evaluateAt(o.x).getChristoffelOatF(o.f),
			    eps);

	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
			    null,//"Probability at o.x",
			    ro,o->getFProbabilityAtX_CalculateInSVDBasis(o.x).applyAsDouble(o.f),
			    ro,o->evaluateAt(o.x).getFProbabilityAtX(o.f),
			    eps);
	TestObservationSequenceFromSampleVectorXVectorF.matchArrayFunctionOnTwoSamples(
	    null/*"f of maximal probability"*/,
	    ro,o->getFofMaximalProbability_CalculateInSVDBasis(o.x),
	    ro,o->evaluateAt(o.x).getFofMaximalProbability(),eps);


	// For this model the maximal probability is 1
	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
			    null,//"P(FmaxP)",
			    ro,o->{
				final var e=evaluateAt(o.x);
				return e.getFProbabilityAtX(e.getFofMaximalProbability());
			    },
			    ro,o->evaluateAt(o.x).getMaximalProbability(), // always 1 for adjusted number of the degrees of freedom
			    eps);

	if(approximation==KGOOptimizationResult.Approximation.LSQ_NOADJ){
	    TestObservationSequenceFromSampleVectorXVectorF.matchArrayFunctionOnTwoSamples(
		null/*"LeastSquares as transform"*/,
		ro,o->evaluateAt(o.x).getFofMaximalProbability(),
		ro,o->evaluateAt(o.x).getFLeastSquaresX(),eps);
	}

	System.err.println("KGOSolutionVectorXVectorF:matchSampleValues OK nX="+nX+" nF="+nF+" eps="+eps);
    }

    private void selfTest(final DataReadObservationVectorXVectorF ro,final double eps){

	matchMatrices(ro,eps);
	matchSampleValues(ro,eps);

	// Old version 1: contributing subspace
	testProjectiveContributingBasisAndFTOTCoverage(ro,eps);

	// Old version 2: "Direct Projection" least squares-like version
	TestKGO.matchDirectProjection(this,ro,eps);

	System.err.println("KGOSolutionVectorXVectorF:test OK nX="+nX+" nF="+nF+" weight="+ro.datastat.totalweight+" approximation="+approximation+
			   " SVD_adj="+java.util.Arrays.toString(test_svd.svd_adj.S)+
			   " FTOT="+FTOT+" FLS="+FLS+" FCOVU="+FCOVU+" eps="+eps);
    }

    /** Returns DirectProjection (old least squares version) for unit test.
     */
    DirectProjectionSolutionVectorXVectorF getDirectProjection_FXbasis(){
	return new DirectProjectionSolutionVectorXVectorF(
	    nX,nF,TMatr.getUnitMatrix(nF),TMatr.getUnitMatrix(nX),TMatr.getMatrTrans(nF,nX,FX)
							  );
    }

    double getDirectProjectionProbability(
	final VectorXVectorFPointAccess e,
	final double [] forig){
	final double [] F=reg.F.getRegularizedBasisXFromOriginalX(forig),
	    FLS=reg.F.getRegularizedBasisXFromOriginalX(e.getFLeastSquaresX());
	// for unit Gram matrix
	return Math.pow(TMatr.mulVV(F,FLS),2)/
	    (TMatr.mulVV(F,F)*TMatr.mulVV(FLS,FLS));
    }

    /** For unit test to match {@link VectorXVectorFPointAccess#getFProbabilityAtX}.
     */
    final java.util.function.ToDoubleFunction<double[]>
	getFProbabilityAtX_CalculateInSVDBasis(final double [] xorig){
	  final double [] X=reg.X.getRegularizedBasisXFromOriginalX(xorig),
	      tXS=TMatr.getVAB(
		  test_svd.svd_adj.S,
		  java.util.Arrays.copyOf(TMatr.mul_MA_x_MB(nX,test_svd.svd_adj.VT,X.length,X,1),nF));
	  return forig->{
	      final double [] F=reg.F.getRegularizedBasisXFromOriginalX(forig),
		  tF=TMatr.mul_MATrans_x_MB(nF,test_svd.svd_adj.U,F.length,F,1);
	      return Math.pow(TMatr.mulVV(tF,tXS),2)/
		  (TMatr.mulVV(tF,tF)
		   *
		   switch(SKtype){
		   case CHRISTOFFEL_DEGREESOFFREEDOMADJUSTED,FXFX_F_CHRISTOFFEL,FXFX -> TMatr.mulVV(tXS,tXS);
		   case CHRISTOFFEL -> TMatr.mulVV(X,X);
		   default -> throw new IllegalArgumentException();
		   });
	  };
    }
    /** For unit test to match {@link VectorXVectorFPointAccess#getFProbabilityAtX}.
     */
    double [] getFofMaximalProbability_CalculateInSVDBasis(final double [] xorig){
	final double [] X=reg.X.getRegularizedBasisXFromOriginalX(xorig),
	    tX=TMatr.mul_MA_x_MB(nX,test_svd.svd_adj.VT,X.length,X,1),
	    tF=TMatr.getVAB(test_svd.svd_adj.S,java.util.Arrays.copyOf(tX,nF)),
	    F=TMatr.mul_MA_x_MB(nF,test_svd.svd_adj.U,tF.length,tF,1),
	    FconstA=TMatr.getVectorScaled(F,reg.F.normalizedconstbasisfunction/F[reg.F.indexConst]);
	//System.err.println("f="+java.util.Arrays.toString(forigFromF(FconstA)));
	return forigFromF(FconstA);
    }

    static double [] leastSquaresO(
	final double [] FX,
	final RegularizedVectorX reg,
	final double [] xorig){

	final var X=reg.getRegularizedBasisXFromOriginalX(xorig);
	final int nF=FX.length/X.length;
	if(nF*X.length!=FX.length) throw new RuntimeException();
	// for unit Gram matrix
	return TMatr.mul_MA_x_MBTrans(nF,FX,X.length,X,1);
    }
    static double christoffel(
	final RegularizedVectorX reg,
	final double [] xorig){
	final var r=reg.getRegularizedBasisXFromOriginalX(xorig);
	// for unit Gram matrix
	return 1/TMatr.mulVV(r,r);
    }


    public static KGOSolutionVectorXVectorF createFromSample(
	final DataReadObservationVectorXVectorF ro,
	final KGOOptimizationResult.Approximation approximation,
	final SKFromSampleCalculationMethod SKtype){
	return new KGOSolutionVectorXVectorF(
	    ro,
	    STD_REGULARIZATION_METHOD,
	    STD_REGULARIZATION_METHOD,
	    approximation,
	    SKtype,
	    0);
    }

    static final RegularizationMethod STD_REGULARIZATION_METHOD=RegularizationMethod.LIN;


    /** Trivial test run once. */
    public static void main(String [] args){
	final var r=new java.util.Random(134);

	//KGOOptimizationResult.Approximation.LISTEV_NOADJ,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalContractionMapping,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalMultipleTransforms,
	var approximation=KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalSimpleOptimizationU;
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLagrangeMultipliersPartialSubspace,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLambda2,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLinearConstraints,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalGEVAdj_Q_denominator,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLinearConstraintsE,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom,
	//KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU,
	//KGOOptimizationResult.Approximation.LISTEV_SVDADJ,
	//KGOOptimizationResult.Approximation.LSQ_SVDADJ_CHECKSIGNS,

	if(args.length==1){
	    approximation=KGOOptimizationResult.Approximation.valueOf(args[0]);
	}

	final double eps=1e-5;
	// force selftests globally
	//TestKGO.EPS_selftest=eps;
	final TestKGO t=new TestKGO(r,13000+r.nextInt(1000),eps);
	t.runOne(19,4,approximation);
	// disable forced selftest
	//TestKGO.EPS_selftest=0;
    }

}
