package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.RegularizedVectorXVectorF;
import com.polytechnik.utils.RegularizationMethod;
import com.polytechnik.utils.DataReadObservationVectorXVectorF;
import com.polytechnik.utils.VectorXVectorFPointEvaluatable;

/** Knowledge Generalizing Operator solution to a vector X to vector F problem.
 * The X- and F- spaces are assumed to be regularized: XX and FF are not degenerated and diagonal.
 * @version $Id: KGOSubspaceSolutionVectorXVectorF.java,v 1.16 2024/05/01 09:41:35 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class KGOSubspaceSolutionVectorXVectorF implements  VectorXVectorFPointEvaluatable {

    /** The list of implemented subspaces.
     */
    public static enum  XSubspace {
	DIRECT_PROJECTION,
	Kf_COVERAGE
	;
	public static final XSubspace DEFAULT=DIRECT_PROJECTION;

	private static final java.util.List<XSubspace> BROKENLIST=
	java.util.List.of(
			  );
	public static java.util.List<XSubspace> valuesSkipBroken(){
	    return java.util.Arrays.stream(values()).filter(e -> !BROKENLIST.contains(e)).toList();
	}

    }


    final RegularizedVectorXVectorF reg;
    public final int nX,nF;
    private final double [] fF, xX;

    /** Chrisfoffel function moments.
     */
    private final double [] S_FCFC /*  nF*nC x nF*nC */;

    final KGOOptimizationResult.Approximation approximation;
    final XSubspace xSubspace;

    /** Contributing subspace (projective).
     */
    private final double [] CX_projective;

    /** Old "projective" version.
     */
    private final double [] FX /* nF x nX */;

    /** K(f) Christoffel function moments.
     */
    private final double [] F_Kf_F;

    /** Maximal possible coverage among all projective-type transforms. */
    public final double FTOT;

    /** A coverage corresponding to "Direct Projection" type of expansion:
     * least squares value is calculated first, then a f-localized function
     * is built at this value. The value may be greater than {@link #FTOT}.
     */
    public final double FLS;

    /** Direct projection via {@link #CX} subspace.
     */
    public final double FCCXLS;

    /** Selected contributing subspace.
     */
    private final double [] CX;

    /** A coverage corresponding to found mapping.
     */
    public final double FCOVU;


    //private final VectorXVectorFPointEvaluatableCommon evaluator;


    /**
     * @param ro Read Observations (sample).
     * @param regularization_methodX regularizatuion method of {@code x}.
     * @param regularization_methodF regularizatuion method of {@code f}.
     * @param eps Run a selftest with this {@code eps} if positive.
     */
    KGOSubspaceSolutionVectorXVectorF(
	final DataReadObservationVectorXVectorF ro,
	final RegularizationMethod regularization_methodX,
	final RegularizationMethod regularization_methodF,
	final KGOOptimizationResult.Approximation approximation,
	final XSubspace xSubspace,
	final double eps){
	final var t0=System.currentTimeMillis();System.err.println("before regularization approximation="+approximation+" xSubspace="+xSubspace);
	this.reg=ro.getDataRegularized_OLDSCALARCODE(regularization_methodX,regularization_methodF);
	final var t1=System.currentTimeMillis();System.err.println("after regularization dt="+(t1-t0));
	this.nX=reg.X.nX;
	this.nF=reg.F.nX;
	if(nF>nX) throw new RuntimeException("Currently nF>nX is not supported: nF="+nF+" nX="+nX);

	// These can be calculated without secondary sampling.
	// But as we use secondaty sampling anyway -- simplify the code.
	// Old projective version. Used in unit tests only
	FX=reg.getFXmoments(ro);
	// needed for F->forig transform
	fF=ro.getFun1Fun2Averaged(o->o.f,o->reg.F.getRegularizedBasisXFromOriginalX(o.f),o->1);
	if(fF.length!=reg.F.noriginal*nF) throw new IllegalArgumentException();
	// needed  for X->xorig transform
	xX=ro.getFun1Fun2Averaged(o->o.x,o->reg.X.getRegularizedBasisXFromOriginalX(o.x),o->1);
	if(xX.length!=reg.X.noriginal*nX) throw new IllegalArgumentException();

	final var t2=System.currentTimeMillis();System.err.println("regular moments done dt="+(t2-t1));

	CX_projective=TestKGO.orthogonalizeU(nF,nX,FX);
	if(eps>0){
	    final int nC=CX_projective.length/nX;
	    TestKGO.testOrthogonality("CX test",nC,CX_projective,nX,eps);
	}
	// The best coverage among all projections.
	F_Kf_F=reg.getFChristoffelFunctionFMoments(ro);
	this.FTOT=TMatr.getSpurM(nX,getXF_KFF_FX(F_Kf_F));
	// The coverage of states localized at least squares value,
	// this is the coverage of the "Direct Projection" model.
	this.FLS=getFtotalForTransform(ro,FX);

	// select the subspace
	CX=switch(xSubspace){
	 case DIRECT_PROJECTION ->CX_projective;
	 case Kf_COVERAGE -> {
	    final double [] X_Kf_X=reg.getFChristoffelFunctionXMoments(ro);
	    final var contribMC=new ContributingSubspaceMaxCoverage(nF,nX,X_Kf_X);
	    if(eps>0){
		TestKGO.testOrthogonality("contribMC.CX test",contribMC.CX.length/nX,contribMC.CX,nX,eps);
	    	System.err.println("KGOSolutionVectorXVectorF:test OK nX="+nX+" nF="+nF+" approximation="+approximation+" xSubspace="+xSubspace+" weight="+ro.datastat.totalweight+" FTOT="+FTOT+" FLS(direct projection)="+FLS+" contribMC.mSum="+contribMC.mSum+" FsumAll="+contribMC.totalSum+" contribEV="+java.util.Arrays.toString(java.util.Arrays.copyOf(contribMC.ev_X_Kf_X,nX))+" eps="+eps);
	    }
	    yield contribMC.CX;
	 }
	 default -> throw new IllegalArgumentException("xSubspace="+xSubspace);
	};
	final int nC=CX.length/nX;
	FCCXLS=getFtotalForTransform(ro,TMatr.mul_MA_x_MB(nF,TMatr.mul_MA_x_MBTrans(nF,FX,nX,CX,nC),nC,CX,nX));

	if(eps>0){
	    TestKGO.testOrthogonality("CX test",nC,CX,nX,eps);
	}
	this.approximation=approximation;
	this.xSubspace=xSubspace;

	/** Calculate f-Chrisfoffel function moments using secondary sampling.
	 */
	final double [] m_FCFC=ro.getFun1Fun2Averaged(
	    o->TMatr.dyadicMatrix(reg.F.getRegularizedBasisXFromOriginalX(o.f),TMatr.mul_MA_x_MB(nC,CX,nX,reg.X.getRegularizedBasisXFromOriginalX(o.x),1)),
	    o->TMatr.dyadicMatrix(reg.F.getRegularizedBasisXFromOriginalX(o.f),TMatr.mul_MA_x_MB(nC,CX,nX,reg.X.getRegularizedBasisXFromOriginalX(o.x),1)),
	    o->{final double [] rF=reg.F.getRegularizedBasisXFromOriginalX(o.f),
		    rCX=TMatr.mul_MA_x_MB(nC,CX,nX,reg.X.getRegularizedBasisXFromOriginalX(o.x),1);
		return 1/(TMatr.mulVV(rF,rF)*TMatr.mulVV(rCX,rCX));
	    });
	S_FCFC=m_FCFC;
	final var t3=System.currentTimeMillis();System.err.println("Christoffel function moments done dt="+(t3-t2));


	// Algorithm groups (there specific algrithm selection is inside group processor)
	final KGOOptimizationResult kgoOpt=switch(approximation) {
	// Least squares one-step adjustment
	case LSQ_SVDADJ,
	     LSQ_EVADJ,
	     LSQ_NOADJ,
	     LSQ_SVDADJ_CHECKSIGNS
	     ->
	     KGOOptimizationResult.fromLeastSquaresOneShot(nF,nC,S_FCFC,TMatr.mul_MA_x_MBTrans(nF,FX,nX,CX,nC),approximation,eps);
	// Iterational
	case ITERATIONS_KGOIterationalSimpleOptimizationU, // more or less OK
	     ITERATIONS_KGOIterationalLambda2,
	     ITERATIONS_KGOIterationalLinearConstraints, // work in progress
	     ITERATIONS_KGOIterationalLinearConstraintsE, // work in progress
	     ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom,
	     ITERATIONS_KGOIterationalMultipleTransforms, // same thing but not well tested
	     ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU, // the concept works but gives no improvement
	     ITERATIONS_KGOIterationalContractionMapping // requires more work, currently broken
	     ->
	       KGOOptimizationResult.fromIterationsLagrangeMultipliers(nF,nC,S_FCFC,approximation,eps);
	// All others (eigenstate bases with one step adjustment).
	// Will fail inside if not supported.
	default ->KGOOptimizationResult.fromEVProblemOneShotSVDAdj(nF,nC,S_FCFC,approximation,eps);
	};
	final var t4=System.currentTimeMillis();System.err.println("Calculations done dt="+(t4-t3));

	FCOVU=getFtotalForTransform(ro,TMatr.mul_MA_x_MB(nF,kgoOpt.uadj.getUAdj(),nC,CX,nX));

	if(eps>0){
	    System.err.println("<uadj|S|uadj>="+TMatr.getAMatrA(kgoOpt.uadj.getUAdj(),S_FCFC,nF*nC)+" FCOVU="+FCOVU);
	    selfTest(ro,eps);
	}
    }

    double getFtotalForTransform(final DataReadObservationVectorXVectorF ro,
				 final double [] u_jk){
	if(u_jk.length!=nF*nX) throw new IllegalArgumentException();
	return ro.data.stream().mapToDouble(o -> {
		final double[] Xl=reg.X.getRegularizedBasisXFromOriginalX(o.x),
		    Fl=reg.F.getRegularizedBasisXFromOriginalX(o.f),
		    FLSl=TMatr.mul_MA_x_MB(nF,u_jk,nX,Xl,1);
		return o.weight*Math.pow(TMatr.mulVV(Fl,FLSl),2)/
		    (TMatr.mulVV(Fl,Fl)*TMatr.mulVV(FLSl,FLSl));
	    }).sum();
    }

    double [] forigFromF(final double [] F){
	if(F.length!=nF) throw new IllegalArgumentException();
	return TMatr.mul_MA_x_MBTrans(reg.F.noriginal,fF,F.length,F,1);
    }


    public VectorXVectorFPointEvaluatableCommon.VectorXVectorFPointAccessWithDimensionAdjusted evaluateAt(final double [] xorig){
	//return evaluator.evaluateAt(xorig);
	return null;// FIXME
    }

    private double [] getXF_KFF_FX(final double [] KFF){
	return TMatr.mul_MATrans_x_MB(nX,FX,nF,TMatr.mul_MA_x_MB(nF,KFF,nF,FX,nX),nX);
    }


    public static KGOSubspaceSolutionVectorXVectorF createFromSample(
	final DataReadObservationVectorXVectorF ro,
	final KGOOptimizationResult.Approximation approximation,
	final XSubspace xSubspace){
	return new KGOSubspaceSolutionVectorXVectorF(
	    ro,
	    STD_REGULARIZATION_METHOD,
	    STD_REGULARIZATION_METHOD,
	    approximation,
	    xSubspace,
	    0);
    }

    private static final RegularizationMethod STD_REGULARIZATION_METHOD=RegularizationMethod.LIN;

    private void selfTest(final DataReadObservationVectorXVectorF ro,final double eps){
	System.err.println("KGOSolutionVectorXVectorF:test OK nX="+nX+" nF="+nF+
			   " approximation="+approximation+" xSubspace="+xSubspace+
			   " weight="+ro.datastat.totalweight+
			   " FTOT="+FTOT+" FLS(direct projection)="+FLS+" FCCXLS="+FCCXLS+" FCOVU="+FCOVU+" eps="+eps);
    }


    /** Trivial test run once. */
    public static void main(String [] args){
	final var r=new java.util.Random(134);
	final double eps=1e-5;
	// force a selftest
	TestKGO.EPS_selftest=eps;
	final TestKGO t=new TestKGO(r,13000+r.nextInt(1000),eps);
	t.runOneKGOSubspace(19,4);
	// disable selftest
	TestKGO.EPS_selftest=0;
    }

}
