package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Linsystems;

/**
 * A class to handle Lagrange multipliers in partial subspace.
 * @version $Id: LagrangeMultipliersPartialSubspace.java,v 1.33 2024/05/07 05:27:28 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
class LagrangeMultipliersPartialSubspace {


    /** Calculate regular lambda as
     * \( \lambda_{ij}=\mathrm{Herm} \sum\limits_{k=k^{\prime}=0}^{nX-1}\sum\limits_{j^{\prime}=0}^{nC-1} u_{ik}S_{jk;j^{\prime}k^{\prime}} u_{j^{\prime}k^{\prime}} \).
     * The \( u_{jk} \) is assumed to be orthogonal:
     * \( \delta_{ij}=\sum\limits_{k=0}^{nX-1} u_{ik}u_{jk} \).
     * @param nC First dimension of  \( u_{jk} \).
     * @param nX Second dimension of  \( u_{jk} \).
     * @param SK Data tensor.
     * @param u The state in which to calculate Lagrange multipliers. The state must be orthogonal  \( \delta_{ij}=\sum\limits_{k=0}^{nX-1} u_{ik}u_{jk} \).
     * @return Lagrange multipliers.
     */
    public static double [] calculateRegularLambda(final int nC,final int nX,final double [] SK, final double [] u){
	final int n=nC*nX;
	if(SK.length!=n*n || u.length!=n) throw new IllegalArgumentException();
	// need only Hermitian part
	return TMatr.selectHermitianPartInPlace(
	    nC,
	    TMatr.mul_MA_x_MBTrans(
		nC,u,nX,TMatr.mul_MA_x_MB(n,SK,n,u,1),nC)
						);
    }


    /** Given UV matrix of nC*nC obtain corresponding coefficients (as vector)
     * by Hermitian lambda.
     * @return Coefficients by Lambda, as a vector of nC*(nC+1)/2 length.
     */
    private static double [] getVectorizedCoefsByLambda(final int nC,final double [] UV){
	if(UV.length!=nC*nC) throw new IllegalArgumentException();
	final double [] vecLambdaCoefs=new double [nC*(nC+1)/2];
	for(int o=0,j=0;j<nC;j++){
	    for(int i=0;i<=j;i++,o++){
		if(i==j) vecLambdaCoefs[o]=UV[j*nC+i];
		else vecLambdaCoefs[o]=UV[j*nC+i]+UV[i*nC+j];
	    }
	}
	return vecLambdaCoefs;
    }


    /** Unvectorize lambda.
     */
    private static double [] getLambdaAsMatrix(final int nC,final double [] vecLambda){
	if(vecLambda.length != nC*(nC+1)/2) throw new IllegalArgumentException();
	final double [] lambda=new double [nC*nC];
	for(int o=0,j=0;j<nC;j++){
	    for(int i=0;i<=j;i++,o++){
		if(i==j) lambda[j*nC+i]=vecLambda[o];
		else lambda[j*nC+i]=lambda[i*nC+j]=vecLambda[o]; // 0.5 is already included!
	    }
	}
	return lambda;
    }

    /** Similar to {@link #calculateRegularLambda} but calculates Lagrange multipliers
     * using only weights for projections to a given subspace.
     * Originally, the \( \lambda_{ij} \) is obtained from the minimization problem (method {@link #calculateRegularLambda}
$$
\sum\limits_{i=0}^{nC-1}\sum\limits_{q=0}^{nX-1}
\left|
\sum\limits_{j^{\prime}=0}^{nC-1}\sum\limits_{k^{\prime}=0}^{nX-1}
S_{iq;j^{\prime}k^{\prime}} u_{j^{\prime}k^{\prime}}
-
\sum\limits_{j=0}^{nC-1} \frac{\lambda_{ij}+\lambda_{ji}}{2} u_{jq}
\right|^2
 \xrightarrow[\lambda_{ij}]{\quad }\min
$$
If we have some chosen basis \( v^{[p]}_{jk} \) then we may consider
$$
\sum\limits_{p=0}^{nP-1}
\left|
\sum\limits_{i=0}^{nC-1}\sum\limits_{q=0}^{nX-1}
v^{[p]}_{iq}\left[
\sum\limits_{j^{\prime}=0}^{nC-1}\sum\limits_{k^{\prime}=0}^{nX-1}
S_{iq;j^{\prime}k^{\prime}} u_{j^{\prime}k^{\prime}}
-
\sum\limits_{j=0}^{nC-1} \frac{\lambda_{ij}+\lambda_{ji}}{2} u_{jq}
\right]
\right|^2
 \xrightarrow[\lambda_{ij}]{\quad }\min
$$
If \( v^{[p]}_{jk} \) is a full basis \( p=0\dots nC*nX-1\),
then two methods of \( \lambda_{ij} \) calculation are identical. The reason why subspace
may be beneficial to use -- the  \( \lambda_{ij} \), Hermitian matrix, has \(nC(nC+1)/2\) independent
parameters. Lagrangian variation has \(nC*nX\) equations.
One may choose a subspace, the variation  projections on which must be set to zero in the first place. This technique may be used to improve algorithm convergence.
Work in progress. Linear system may be degenerated. The concept itself
may not be a good one. A better approach seems to add
additional linear constraints on \(u_{jk}\), see
{@link KGOIterationalSubspaceLinearConstraints} and
{@link LinearConstraints#getOrthogonalOffdiag0DiagEq}.
     * @param nC First dimension of  \( u_{jk} \).
     * @param nX Second dimension of  \( u_{jk} \).
     * @param SK Data tensor.
     * @param u The state in which to calculate Lagrange multipliers.
     * @param vBasis A subspace using which to calculate Lagrange multipliers, keep
     * {@code vBasis.length >= nC*(nC+1)/2}
     * to avoid linear system degeneracy.
     * @return Lagrange multipliers \( nC \times nC \) matrix,
     * if {@code vBasis} is a full basis the result is identical to
     * {@link #calculateRegularLambda}.
     */
    static double [] getLambdaForSubspace(
	final int nC,
	final int nX,
	final double [] SK,
	final double [] u,
	final double [][] vBasis){

	if(false){
	    final var ls=new LinearSystemCoefficients(nC,nX,SK,u,vBasis);
	    return getLambdaSolvingLinearSystem(nC,ls.matrLinSystem,ls.rPart);
	} else {
	    // same thing
	    return getLambdaForSubspaceAverageSeveralU(nC,nX,SK,new double [][]{u},vBasis);
	}
    }

    /** A try to build Lagrange multipliers not for specific u, but averaged for several states
     * from {@code uToAverage}, obtain averaged Lagrange multiplers over several solution candidates.
     */
    private static double [] getLambdaForSubspaceAverageSeveralU(
	final int nC,
	final int nX,
	final double [] SK,
	final double [][] uToAverage,
	final double [][] vBasis){

	if(uToAverage.length<1) throw new IllegalArgumentException();
	final int lamLen=nC*(nC+1)/2;
	final double [] rPart=new double[lamLen],matrLinSystem=new double[lamLen*lamLen];

	for(int pu=0;pu<uToAverage.length;pu++){
	    final double [] u=uToAverage[pu];
	    final var ls=new LinearSystemCoefficients(nC,nX,SK,u,vBasis);
	    TMatr.sumP1(rPart,ls.rPart,1);
	    TMatr.sumP1(matrLinSystem,ls.matrLinSystem,1);
	}
	return getLambdaSolvingLinearSystem(nC,matrLinSystem,rPart);
    }

    /** Solve linear system and convert obtained lambda-vector to Hermitian matrix.
     */
    private static double [] getLambdaSolvingLinearSystem(
	final int nC,
	final double [] matrLinSystem,
	final double [] rPart){

	final int lamLen=nC*(nC+1)/2;
	if(rPart.length!=lamLen || matrLinSystem.length!=lamLen*lamLen) throw new IllegalArgumentException();
	final double [] matr=matrLinSystem.clone(),rp=rPart.clone();

	if(true){
	    // diag linear system degeneracy
	    System.err.println("LagrangeMultipliersPartialSubspace:linear system matrix eigenvalues="+java.util.Arrays.toString(KGOSolutionVectorXVectorF.EV.getEV(lamLen,matr,false))+" (degeneracy diag)");
	}

	// solve linear system
	final int [] ipvt=new int [lamLen];
	final int pinfo=Linsystems.dgefa_m1(matr,lamLen,lamLen,ipvt);
	if(pinfo!=0){
	    java.util.Arrays.fill(rp,Double.NaN);
	} else {
	    Linsystems.dgesl_m1(matr,lamLen,lamLen,ipvt,rp,1);
	}
	// convert lambda vector to Hermitian matrix
	return getLambdaAsMatrix(nC,rp);
    }

/** Store linear system coefficeints.
 */
private static class LinearSystemCoefficients {
    final double [] rPart,matrLinSystem;
    LinearSystemCoefficients(final int nC,
			     final int nX,
			     final double [] SK,
			     final double [] u,
			     final double [][] vBasis){
	final int lamLen=nC*(nC+1)/2,n=nC*nX;
	if(u.length!=n || SK.length!=n*n) throw new IllegalArgumentException();
	// will be degenerated
	//if(vBasis.length<lamLen) throw new IllegalArgumentException();
	this.rPart=new double[lamLen];
	this.matrLinSystem=new double[lamLen*lamLen];

	final double [] b=TMatr.mul_MA_x_MB(n,SK,n,u,1);
	for(int p=0;p<vBasis.length;p++){
	    final double [] v=vBasis[p];
	    if(v.length!=n) throw new IllegalArgumentException();
	    final double vNorm2=1/*TMatr.mulVV(v,v)*/,bv=TMatr.mulVV(b,v);
	    final double [] lamCoefs=getVectorizedCoefsByLambda(nC,TMatr.mul_MA_x_MBTrans(nC,u,nX,v,nC));
	    if(lamCoefs.length!=lamLen) throw new IllegalArgumentException();
	    for(int j=0;j<lamLen;j++){
		rPart[j] += bv*lamCoefs[j]/vNorm2;
		for(int i=0;i<lamLen;i++){
		    matrLinSystem[j*lamLen+i] += lamCoefs[j]*lamCoefs[i]/vNorm2;
		}
	    }
	}
    }
} // END OF class LinearSystemCoefficients

}
