package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;

/**
 * A class to calculated various linear constraints.
 * @version $Id: LinearConstraints.java,v 1.13 2024/05/06 08:54:46 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class LinearConstraints {

    /** Zero scalar product  \( 0=\sum\limits_{j=0}^{nC-1}\sum\limits_{k=0}^{nX-1} u_{jk}v_{jk} \),
     * a single conditions.
     */
    public static double [][] getVectorOrthogonal(final int nC,final int nX,final double [] u,final double [] fullBasis){
	final int n=nX*nC;
	if(u.length!=n || fullBasis.length!=n*n) throw new IllegalArgumentException();

	final double [][] linearConstrains=new double [1][n];
	 for(int p=0;p<n;p++){
	     final double [] v=EVXData.getPsiFromBasis(p,fullBasis,n,0),
		 ProjV=TMatr.mul_MA_x_MBTrans(nC,u,nX,v,nC);
	     linearConstrains[0][p] = 2*TMatr.getSpurM(nC,ProjV);
	 }
	 return linearConstrains;
    }

    /** All \(0=\mathrm{Herm}\left(\sum\limits_{k=0}^{nX-1}u_{jk}v_{ik}\right) \),
     * {@code nC*(nC+1)/2} conditions. Used for unit tests.
     */
    static double [][] getOrthogonalAllProj0(final int nC,final int nX,final double [] u,final double [] fullBasis){
	final int n=nX*nC;
	if(u.length!=n || fullBasis.length!=n*n) throw new IllegalArgumentException();

	final double [][] linearConstrains=new double [nC*(nC+1)/2][n];
	 for(int p=0;p<n;p++){
	     final double [] v=EVXData.getPsiFromBasis(p,fullBasis,n,0),
		 ProjV=TMatr.mul_MA_x_MBTrans(nC,u,nX,v,nC);
	     for(int o=0,j=0;j<nC;j++){
		 for(int i=0;i<=j;i++,o++){
		     linearConstrains[o][p] = ProjV[j*nC+i]+ProjV[i*nC+j];
		 }
	     }
	 }
	 return linearConstrains;
    }

    /** Zero offdiagonal and all equal diagonal elements
of the matrix \( \mathrm{Herm}\left(\sum\limits_{k=0}^{nX-1}u_{jk}v_{ik}\right) \),
     * there are {@code (nC-1)*(nC+2)/2} conditions.
     */
    public static double [][] getOrthogonalOffdiag0DiagEq(final int nC,final int nX,final double [] u,final double [] fullBasis){
	final int n=nX*nC;
	if(u.length!=n || fullBasis.length!=n*n) throw new IllegalArgumentException();

	final double [][] linearConstrains=new double [(nC-1)*(nC+2)/2][n];
	for(int p=0;p<n;p++){
	    final double [] v=EVXData.getPsiFromBasis(p,fullBasis,n,0),
		ProjV=TMatr.mul_MA_x_MBTrans(nC,u,nX,v,nC);
	    //final double sumDiagA=2*TMatr.getSpurM(nC,ProjV)/nC;
	    for(int o=0,j=0;j<nC;j++){
		// zero offdiagonal elements
		for(int i=0;i<j;i++){
		    linearConstrains[o++][p] = ProjV[j*nC+i]+ProjV[i*nC+j];
		}
		if(j>0){
		    // equal diagonal elements
		    linearConstrains[o++][p] = ProjV[j*nC+j]-ProjV[(j-1)*nC+(j-1)];
		    //linearConstrains[o++][p] = sumDiagA-2*ProjV[(j-1)*nC+(j-1)];
		}
	    }
	}
	return linearConstrains;
    }


    /**
     * Check a matrix for: all offdiag 0, all diag are the same.
     */
    static void testConstraintsOK_getOrthogonalOffdiag0DiagEq(final int nC,final int nX,final double [] u,final double [] v,final double eps){
	final double [] PVp=TMatr.selectHermitianPartInPlace(nC,TMatr.mul_MA_x_MBTrans(nC,u,nX,v,nC));
	for(int j=0;j<nC;j++){
	    // offdiagonal elements are 0
	    for(int i=0;i<nC;i++){
		if(i!=j) UnitTests.checkDoubleRel1(PVp[i*nC+j],"PVp[i*nC+j]",0,"0",eps);
	    }
	    // diagonal elements are the same
	    if(j>0){
		UnitTests.checkDoubleRel1(PVp[j*nC+j],"PVp[j*nC+j]",PVp[(j-1)*nC+(j-1)],"PVp[(j-1)*nC+(j-1)]",eps);
	    }
	}
    }

}
