package com.polytechnik.kgo;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.RegularizationMethod;
import com.polytechnik.utils.VectorXVectorFPointAccess;
import com.polytechnik.utils.DataReadObservationVectorXVectorF;
import com.polytechnik.utils.RegularizedVectorXVectorF;


/** A class for {@link KGOSolutionVectorXVectorF} unit test.
  * @version $Id: TestKGO.java,v 1.72 2024/05/07 12:55:49 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
 */
public class TestKGO {

    final java.util.Random r;
    final int size;
    final double eps;
    TestKGO(final java.util.Random r,final int size,final double eps){
	this.r=r;
	this.size=size;
	this.eps=eps;
    }

    static double [] getContributingSubspaceEV(final double [] XF_KFF_FX,final int nF,final int nX,final double eps){
	final double [] ev=KGOSolutionVectorXVectorF.EV.getEV(nX,XF_KFF_FX,true);
	if(eps>0){
	    final double [] eValues=java.util.Arrays.copyOf(ev,nX);
	    System.err.println("getContributingSubspaceEV: eigenvalues="+java.util.Arrays.toString(eValues)+" sum(EV)="+TMatr.sumElements(eValues));
	    final int n0=nX-Math.min(nF,nX);
	    UnitTests.checkDoubleArrayRel1(java.util.Arrays.copyOfRange(eValues,0,n0),"first "+n0+" ev",new double[n0],"0",eps);
	}
	// copy top nF eigenvectors
	final int nC=Math.min(nF,nX);
	final double [] CX=new double [nC*nX];
	for(int j=0,indEv=nX-1;j<nC;j++,indEv--){
	    final double [] e=EVXData.getPsiFromBasis(indEv,ev,nX);
	    System.arraycopy(e,0,CX,j*nX,nX);
	}
	return CX;
    }

    static double [] orthogonalizeU(final int nF,final int nX,final double [] FX){
	if(FX.length!=nF*nX) throw new RuntimeException();
	final double [][] basis=new double [nF][nX];
	for(int j=0;j<nF;j++){
	    System.arraycopy(FX,j*nX,basis[j],0,nX);
	}
	// A homegrown orthogonalization
	final double NORM_MIN=1e-12;
	int nFound=0;
	for(int j=0;j<nF;j++){
	    // find the state with max norm
	    double norm2=0;
	    int indMaxNorm=j;
	    for(int jj=j;jj<nF;jj++){
		final double s=TMatr.mulVV(basis[jj],basis[jj]);
		if(s>norm2||jj==j){
		    norm2=s;
		    indMaxNorm=jj;
		}
	    }
	    final double sqnorm2=Math.sqrt(norm2);
	    //System.err.println("getContributingSubspaceLIN: j="+j+" nF="+nF+" nX="+nX+" sqnorm2="+sqnorm2+" indMaxNorm="+indMaxNorm);
	    if(sqnorm2>NORM_MIN){
		final var v=basis[indMaxNorm];
		basis[indMaxNorm]=basis[j];
		basis[j]=TMatr.scaleVectorInPlace(v,1/sqnorm2);
		// remove found vector from remained subspace
		for(int jj=j+1;jj<nF;jj++){
		    TMatr.sumP1(basis[jj],v,-TMatr.mulVV(v,basis[jj]));
		}
		nFound++;
	    } else {break;}
	}
	final double [] CX=new double [nFound*nX];
	for(int j=0;j<nFound;j++){
	    System.arraycopy(basis[j],0,CX,j*nX,nX);
	}
	return CX;
    }

    static void testOrthogonality(final String txt,final int nC,final double [] basis,final int nX,final double eps){
	if(basis.length!=nC*nX) throw new IllegalArgumentException();
	for(int j=0;j<nC;j++){
	    final double [] psij=EVXData.getPsiFromBasis(j,basis,nX,0);
	    for(int k=0;k<nC;k++){
		final double [] psik=EVXData.getPsiFromBasis(k,basis,nX,0);
		final double scal=TMatr.mulVV(psij,psik);
		UnitTests.checkDoubleRel1(scal,"<psiC"+j+"|psiC"+k+">",j==k?1:0,j==k?"1":"0",eps);
	    }
	}
	if(txt!=null) System.err.println(txt+"KGOSolutionVectorXVectorF:basis orthogonality OK nC="+nC+" nX="+nX+" eps="+eps);
    }


    static java.util.List<RegularizedVectorXVectorF> getAllRegularizations(
	final DataReadObservationVectorXVectorF ro){
	return java.util.List.of(
	    ro.getDataRegularized_OLDSCALARCODE(RegularizationMethod.LIN,RegularizationMethod.LIN),
	    ro.getDataRegularized_OLDSCALARCODE(RegularizationMethod.LIN,RegularizationMethod.EV),
	    ro.getDataRegularized_OLDSCALARCODE(RegularizationMethod.EV,RegularizationMethod.LIN),
	    ro.getDataRegularized_OLDSCALARCODE(RegularizationMethod.EV,RegularizationMethod.EV));
    }

    void testBases(final int nx,final int nf){
	final var ro=TestObservationSequenceFromSampleVectorXVectorF.getDegeneratedXandF_VectorXVectorF(nx,nf,r,size);
	final boolean fStartsTheSame=true;
	TestObservationSequenceFromSampleVectorXVectorF.testBasisConversion(
	    java.util.List.of(
		new TestObservationSequenceFromSampleVectorXVectorF.RegularizationWithSample(getAllRegularizations(ro.ndeg),ro.ndeg),
		new TestObservationSequenceFromSampleVectorXVectorF.RegularizationWithSample(getAllRegularizations(ro.deg),ro.deg)
			      ),fStartsTheSame,eps);
	System.err.println("testBases(): OK nx="+nx+" nf="+nf+" eps="+eps);
    }


    void runOne(final int nx,final int nf,final KGOOptimizationResult.Approximation approximation){
	final var SKtype=SKFromSampleCalculationMethod
	    //.DEFAULT
	    .CHRISTOFFEL_DEGREESOFFREEDOMADJUSTED
	    ;
	final var ro=TestObservationSequenceFromSampleVectorXVectorF.getDegeneratedXandF_VectorXVectorF(nx,nf,r,size);
	final KGOSolutionVectorXVectorF
	    kgoL=new KGOSolutionVectorXVectorF(
		ro.ndeg,
		RegularizationMethod.LIN,
		RegularizationMethod.LIN,
		approximation,
		SKtype,
		eps);
     }

    void runOneKGOSubspace(final int nx,final int nf){
	final var ro=TestObservationSequenceFromSampleVectorXVectorF.getDegeneratedXandF_VectorXVectorF(nx,nf,r,size);
	for(var xSubspace:KGOSubspaceSolutionVectorXVectorF.XSubspace.valuesSkipBroken()){
	    final KGOSubspaceSolutionVectorXVectorF
		kgoL=new KGOSubspaceSolutionVectorXVectorF(
		    ro.ndeg,
		    RegularizationMethod.LIN,
		    RegularizationMethod.LIN,
		    //KGOOptimizationResult.Approximation.MAXEV_NOADJ,
		    //KGOOptimizationResult.Approximation.MAXEV_EVADJ,
		    KGOOptimizationResult.Approximation.LSQ_EVADJ,
		    //KGOOptimizationResult.Approximation.LSQ_NOADJ,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalContractionMapping,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalMultipleTransforms,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalSimpleOptimizationU,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLambda2,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLinearConstraints,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLinearConstraintsE,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLinearConstraintsExtraDegreesOfFreedom,
		    //KGOOptimizationResult.Approximation.ITERATIONS_KGOIterationalLagrangeMultipliersInDenominatorU,
		    //KGOOptimizationResult.Approximation.LISTEV_SVDADJ,
		    //KGOOptimizationResult.Approximation.LSQ_SVDADJ_CHECKSIGNS,
		    xSubspace,
		    eps);
	}
    }


    void testKGOSimple(final int nx,final int nf){
	final var ro=TestObservationSequenceFromSampleVectorXVectorF.getDegeneratedXandF_VectorXVectorF(nx,nf,r,size);
	for(var SKtype:SKFromSampleCalculationMethod.valuesSkipBroken()){
	for(var approximation : KGOOptimizationResult.Approximation.valuesSkipBroken()){
	final KGOSolutionVectorXVectorF
	    kgoL=new KGOSolutionVectorXVectorF(ro.ndeg,RegularizationMethod.LIN,RegularizationMethod.LIN,approximation,SKtype,eps),
	    dkgoL=new KGOSolutionVectorXVectorF(ro.deg,RegularizationMethod.LIN,RegularizationMethod.LIN,approximation,SKtype,eps),
	    kgoE=new KGOSolutionVectorXVectorF(ro.ndeg,RegularizationMethod.EV,RegularizationMethod.EV,approximation,SKtype,eps),
	    dkgoE=new KGOSolutionVectorXVectorF(ro.deg,RegularizationMethod.EV,RegularizationMethod.EV,approximation,SKtype,eps);

	// Direct projection trivial test
	matchDirectProjection(dkgoL,ro.deg,eps);

	// match with different regularization to check the invariance
	matchTwoKGO(kgoL,ro.ndeg,dkgoL,ro.deg,true);
	matchTwoKGO(kgoE,ro.ndeg,dkgoE,ro.deg,true);
	matchTwoKGO(kgoL,ro.ndeg,dkgoE,ro.deg,true);
	matchTwoKGO(kgoE,ro.ndeg,dkgoL,ro.deg,true);
	matchTwoKGO(kgoE,ro.ndeg,kgoE,ro.ndeg,true);
	matchTwoKGO(dkgoL,ro.deg,dkgoE,ro.deg,true);

	}
	}
	System.err.println("testKGOSimple(): OK nx="+nx+" nf="+nf+" eps="+eps);
    }

    void matchTwoKGO(final KGOSolutionVectorXVectorF  kgo,final DataReadObservationVectorXVectorF ro,
		     final KGOSolutionVectorXVectorF  dkgo,final DataReadObservationVectorXVectorF dro,
		     final boolean fStartsTheSame){

	UnitTests.checkDoubleRel1(kgo.nX,"kgo.nX",dkgo.nX,"dkgo.nX",eps);
	UnitTests.checkDoubleRel1(kgo.FTOT,"kgo.FTOT",dkgo.FTOT,"dkgo.FTOT",eps);
	UnitTests.checkDoubleRel1(kgo.FLS,"kgo.FLS",dkgo.FLS,"dkgo.FLS",eps);

	// evaluate every point
	matchKGOValues(kgo,ro,dkgo,dro,fStartsTheSame);

	System.err.println("matchTwoKGO(): two KGOSolutionVectorXVectorF matched OK  eps="+eps);
    }
    void matchKGOValues(final KGOSolutionVectorXVectorF  kgo,final DataReadObservationVectorXVectorF ro,
			final KGOSolutionVectorXVectorF  dkgo,final DataReadObservationVectorXVectorF dro,
			final boolean fStartsTheSame){
	if(ro.data.size()!=dro.data.size()) throw new RuntimeException();
	UnitTests.checkDoubleRel(ro.datastat.totalweight,"ro.datastat.totalweight",dro.datastat.totalweight,"dro.datastat.totalweight",eps);

	final int nFToCheck=Math.min(kgo.reg.F.noriginal,dkgo.reg.F.noriginal);
	final var idro=dro.data.iterator();
	// evaluate every point
	for(var o:ro.data) {
	    final var od=idro.next();
	    final VectorXVectorFPointAccess ekgo=kgo.evaluateAt(o.x),edkgo=dkgo.evaluateAt(od.x);
	    // match Christoffel functions
	    UnitTests.checkDoubleRel1(ekgo.getChristoffelOatX(),"ekgo.getChristoffelOatX()",edkgo.getChristoffelOatX(),"edkgo.getChristoffelOatX()",eps);
	    UnitTests.checkDoubleRel1(ekgo.getChristoffelOatF(o.f),"ekgo.getChristoffelOatF(o.f)",edkgo.getChristoffelOatF(od.f),"edkgo.getChristoffelOatF(od.f)",eps);
	    // probability
	    UnitTests.checkDoubleRel1(ekgo.getFProbabilityAtX(o.f),"ekgo.getFProbabilityAtX(o.f) approximation="+kgo.approximation,edkgo.getFProbabilityAtX(od.f),"edkgo.getFProbabilityAtX(od.f) approximation="+dkgo.approximation,eps);
	    UnitTests.checkDoubleRel1(ekgo.getMaximalProbability(),"ekgo.getMaximalProbability()",edkgo.getMaximalProbability(),"edkgo.getMaximalProbability()",eps);
	    // f-values
	    if(fStartsTheSame){
		UnitTests.checkDoubleArrayRel1(
		    java.util.Arrays.copyOf(ekgo.getFLeastSquaresX(),nFToCheck),"ekgo.getFLeastSquaresX()",
		    java.util.Arrays.copyOf(edkgo.getFLeastSquaresX(),nFToCheck),"edkgo.getFLeastSquaresX()",
		    eps);
		UnitTests.checkDoubleArrayRel1(
		    java.util.Arrays.copyOf(ekgo.getFofMaximalProbability(),nFToCheck),"ekgo.getFofMaximalProbability()",
		    java.util.Arrays.copyOf(edkgo.getFofMaximalProbability(),nFToCheck),"edkgo.getFofMaximalProbability()",
		    eps);
	    }
	    // add more tests here
	}
	System.err.println("matchKGOValues(): two KGOSolutionVectorXVectorF matched OK eps="+eps);
    }


    static void matchDirectProjection(final KGOSolutionVectorXVectorF  kgo,final DataReadObservationVectorXVectorF ro,final double eps){
	final var dp=kgo.getDirectProjection_FXbasis();
	final java.util.function.Function<double[],VectorXVectorFPointAccess> dpevaluateAtxorig=
	    xorig->dp.evaluateAt(kgo.reg.X.getRegularizedBasisXFromOriginalX(xorig));

	TestObservationSequenceFromSampleVectorXVectorF.matchArrayFunctionOnTwoSamples(
	    "Match Direct Projection LS",
	    ro,o->kgo.evaluateAt(o.x).getFLeastSquaresX(),
	    ro,o->kgo.forigFromF(dpevaluateAtxorig.apply(o.x).getFLeastSquaresX()),
	    eps);
	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
	    "Match Direct Projection Probability",
	    ro,o->kgo.getDirectProjectionProbability(kgo.evaluateAt(o.x),o.f),
	    ro,o->dpevaluateAtxorig.apply(o.x).getFProbabilityAtX(kgo.reg.F.getRegularizedBasisXFromOriginalX(o.f)),
	    eps);
	TestObservationSequenceFromSampleVectorXVectorF.matchFunctionOnTwoSamples(
	    "Match Direct Projection Max Probability",
	    ro,o->1,
	    ro,o->{
		final var e=kgo.evaluateAt(o.x);
		final double [] fLS=e.getFLeastSquaresX(),
		    FLS=kgo.reg.F.getRegularizedBasisXFromOriginalX(fLS);
		return dpevaluateAtxorig.apply(o.x).getFProbabilityAtX(FLS);
	    },
	    eps);
	System.err.println("matchDirectProjection(): matched OK eps="+eps);
    }


    static void runAllTests(final java.util.Random r,final double eps){
	System.err.println("TestKGO:starting tests; ver=$Id: TestKGO.java,v 1.72 2024/05/07 12:55:49 mal Exp $");
	final long t0=System.currentTimeMillis();

	for(int stest=0;stest<10;stest++){
	    final TestKGO t=new TestKGO(r,2300+r.nextInt(1000),eps);

	    for(int nx: java.util.List.of(2,9,16)){
		for(int nf: java.util.List.of(1,2,3,16,9)){
		    if(nf>nx) continue;
		    // test basis
		    t.testBases(nx,nf);
		    t.testKGOSimple(nx,nf);
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TestKGO: tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: TestKGO.java,v 1.72 2024/05/07 12:55:49 mal Exp $");
    }


    public static void main(String [] args){

	final double eps=1e-5;
	// force a selftest
	TestKGO.EPS_selftest=eps;
	// run the selftest
	runAllTests(new java.util.Random(129L),eps);
	// disable selftest
	TestKGO.EPS_selftest=0;
    }

    /** No "final", a kludgy way to make a test from other classes.
     * An alternative can be constructor's argument, a factory, or an injection.
     */
    static double EPS_selftest=0;//1e-6;

}
