package com.polytechnik.lapack;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TestEigenvalues;

/** Eigenvectors problem. Call Lapack converted to java. Possible conversion problems.
 *
 *  @version $Id: Eigenvalues.java,v 1.35 2022/10/20 18:31:14 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public class Eigenvalues {


public static double [] getEV3_dstebz(final int n,final double [] diag,final int diag_off,final double [] offdiag,final int offdiag_off,final boolean flag_ev)
{
    if(n<=0) return new double []{};

    final double [] work,res;
    final int [] iwork=new int [n*5];
    final char sort;
    if(flag_ev)
    {
	res=new double [n+n*n];
	work=new double [n*5];
	sort='B';
    }
    else
    {
	res=new double [n];
	work=new double [n*4];
	sort='E';
    }

    final intW m=new intW(0),info=new intW(0),nsplit=new intW(0);
    Dstebz.dstebz(
	'A',sort,n,
	Double.NaN,Double.NaN, // VL, VU
	-1,-1, // IL,IU
	-1, // ABSTOL, use default
	diag,diag_off,offdiag,offdiag_off, // diag & offdiag. No need to clone arrays, the data is preserved.
	m, // number of eigenvalues calculated
	nsplit,
	res,0, // result, first m.val elements
	iwork,0+0,
	iwork,0+n,
	work,0,
	iwork,0+n+n,
	info);
    //System.err.println("getEV:n="+n+" work[0]="+work[0]+" iwork[0]="+iwork[0]+" m.val="+m.val+" res="+java.util.Arrays.toString(res));
    //System.err.println("nsplit="+nsplit.val);

    if(m.val!=n||info.val!=0) {java.util.Arrays.fill(res,Double.NaN);return res;}

    if(!flag_ev){return res;}

    Dstein.dstein(n,
		  diag,diag_off,offdiag,offdiag_off, // diag & offdiag. No need to clone arrays, the data is preserved.
		  m.val,
		  res,0,
		  iwork,0+0,
		  iwork,0+n,
		  res,n,n,
		  work,0,
		  iwork,0+n+n,
		  iwork,0+n+n+n,
		  info);
    if(info.val!=0) {java.util.Arrays.fill(res,Double.NaN);return res;}

    Util.sortArrayWithLoad(n,res,0+0,n,res,n,0+n,-1,null,-1,-1);
    
    return res;
}


public static double [] getEV3_dstemr(final int n,final double [] diag,final int diag_off,final double [] offdiag,final int offdiag_off,final boolean flag_ev)
{
    if(n<=0) return new double []{};
    final double [] work,res;
    final int [] iwork,isplit;
    final char jobz;
    // work/iwork sizes is 18/10, or 12/8 if no eigenvalues required.
    if(flag_ev)
    {
	work=new double [n*18];
	res=new double [n+n*n];
	iwork=new int [n*10];
	isplit=new int [n*2];
	jobz='V';
    }
    else
    {
	work=new double [n*12];
	res=new double [n];
	iwork=new int [n*8];
	isplit=null;
	jobz='N';
    }
    final double [] d=new double [2*n];
    System.arraycopy(diag,diag_off,d,0,n);
    System.arraycopy(offdiag,offdiag_off,d,n,n-1);

    final booleanW tryrac=new booleanW(true);
    final intW m=new intW(0),info=new intW(0);
    Dstemr.dstemr(
	jobz,'A',n,
	d,0,d,n, // diag & offdiag.
	Double.NaN,Double.NaN, // VL, VU
	-1,-1, // IL,IU
	m,
	res,0,
        res,n,n,n,
	isplit,0,
	tryrac,
	work,0,work.length,
	iwork,0,iwork.length,
	info);

    if(m.val!=n||info.val!=0) java.util.Arrays.fill(res,Double.NaN);
    return res;
}


/** First n element -eigenvalues, if eigenvectors are required, then n eigenvectors follow.
 */
public static double [] getEV_dsyevx(final int n,final double [] matr,final int lda_matr,final boolean flag_ev)
{
    if(n<=0) return new double []{};
    final double [] mt=new double [lda_matr*n];
    System.arraycopy(matr,0,mt,0,mt.length);
    final double [] work=new double [n*8];
    final int [] iwork=new int [n*5],ifail;
    final double [] ev;
    final char job;
    if(flag_ev)
    {
	ifail=new int [n];
	ev=new double [n*(n+1)];
	job='V';
    }
    else
    {
	ifail=null;
	ev=new double [n];
	job='N';
    }
    final intW m=new intW(0),info=new intW(0);
    Dsyevx.dsyevx(
	job,'U',
	n,mt,0,lda_matr,
	0,
	m,
	ev,0,
	ev,n,n,
	work,0,work.length,
	iwork,0,
	ifail,0,
	info);

    if(m.val!=n||info.val!=0) java.util.Arrays.fill(ev,Double.NaN);
    return ev;
}


/** First n element -eigenvalues, if eigenvectors are required, then n eigenvectors follow.
 */
public static double [] getEV_dsyevr(final int n,final double [] matr,final int lda_matr,final boolean flag_ev)
{
    if(n<=0) return new double []{};
    final double [] mt=new double [lda_matr*n];
    System.arraycopy(matr,0,mt,0,mt.length);
    final double [] work=new double [n*26];
    final int [] iwork=new int [n*10],isupp;
    final double [] ev;
    final char job;
    if(flag_ev)
    {
	isupp=new int [2*n];
	ev=new double [n*(n+1)];
	job='V';
    }
    else
    {
	isupp=null;
	ev=new double [n];
	job='N';
    }
    final intW m=new intW(0),info=new intW(0);
    Dsyevr.dsyevr(
	job,'A','U',
	n,mt,0,lda_matr,
	0,0,-1,-1,
	0, // ABSTOL
	m,
	ev,0,
	ev,n,n,
	isupp,0,
	work,0,work.length,
	iwork,0,iwork.length,
	info);

    if(m.val!=n||info.val!=0) java.util.Arrays.fill(ev,Double.NaN);
    return ev;
}

/** Generalized eigenvalues.
 *  First n element -eigenvalues, if eigenvectors are required, then n eigenvectors follow.
 */
public static double [] getGEV_dsygvx(final int n,final double [] A,final int lda_A,final double [] B,final int lda_B,final boolean flag_ev)
{
    if(n<=0) return new double []{};
    final int o=lda_A*n,om=o+lda_B*n,nwork=n*8;
    final double [] mt=new double [om+nwork];
    System.arraycopy(A,0,mt,0,o);
    System.arraycopy(B,0,mt,o,lda_B*n);
    final int [] iwork=new int [n*5],ifail;
    final double [] ev;
    final char job;
    if(flag_ev)
    {
	ifail=new int [n];
	ev=new double [n*(n+1)];
	job='V';
    }
    else
    {
	ifail=null;
	ev=new double [n];
	job='N';
    }
    final intW m=new intW(0),info=new intW(0);
    Dsygvx.dsygvx(1,
	job,'U',
	n,mt,0,lda_A,mt,o,lda_B,
	0,
	m,
	ev,0,
	ev,n,n,
	mt,om,nwork,
	iwork,0,
	ifail,0,
	info);

    if(m.val!=n||info.val!=0) java.util.Arrays.fill(ev,Double.NaN);
    return ev;
}


/** Generalized eigenvalues.
 *  @return First n element -eigenvalues, if eigenvectors are required, then n eigenvectors follow.
 */
public static double [] getGEV_dsygvx_dsyevr(final int n,final double [] A,final int lda_A,final double [] B,final int lda_B,final boolean flag_ev)
{
    if(n<=0) return new double []{};
    final int o=lda_A*n,om=o+lda_B*n,nwork=n*26;
    final double [] mt=new double [om+nwork];
    System.arraycopy(A,0,mt,0,o);
    System.arraycopy(B,0,mt,o,lda_B*n);
    final int [] iwork=new int [n*10];
    final double [] ev;
    final char job;
    if(flag_ev)
    {
	ev=new double [n*(n+1)];
	job='V';
    }
    else
    {
	ev=new double [n];
	job='N';
    }
    final intW m=new intW(0),info=new intW(0);
    Dsygvx_dsyevr.dsygvr(1,
	job,'U',
	n,mt,0,lda_A,mt,o,lda_B,
	0, // ABSTOL
	m,
	ev,0,
	ev,n,n,
	mt,om,nwork,
	iwork,0,
	info);

    if(m.val!=n||info.val!=0) java.util.Arrays.fill(ev,Double.NaN);
    return ev;
}



/** Calls Dgees.dgees to calculate eigenvalues of non-symmetric matrix.
 * @param n Dimension.
 * @param matr Matrix, on exit matr is clobbered.
 * @param lda_matr Leading dimension of matr.
 * @param wr[n] Real part of eigenvalues.
 * @param wi[n] Imaginary part of eigenvalues.
 * @return error code, equal to 0 if success.
 */
public static int getNonSymEV_dgees(final int n,final double [] matr,final int lda_matr,final double [] wr,final double [] wi)
{
    if(n==0) return 0;

    final double [] work=new double [3*n];
    final intW info=new intW(0);
    Dgees.dgees('N',
		n,
		matr,0,lda_matr,
		wr,0,
		wi,0,
		null,0,
		n,
		work,0,work.length,
		info);
    return info.val;
}


public static void main(String [] args)
{
    final double eps=1e-13;
    final java.util.Random r=new java.util.Random(1167);
    final int n_tests=1000;
    for(int n=1;n<50;n++)
    {
	TestEigenvalues.test_getEV(Eigenvalues::getEV_dsyevx,n_tests,n,r,eps);
	TestEigenvalues.test_getEV(Eigenvalues::getEV_dsyevr,n_tests,n,r,eps);
	TestEigenvalues.test_getEV3(Eigenvalues::getEV3_dstebz,n_tests,n,r,eps);
	TestEigenvalues.test_getEV3(Eigenvalues::getEV3_dstemr,n_tests,n,r,eps);
	TestEigenvalues.test_getGEV(Eigenvalues::getGEV_dsygvx,n_tests,n,r,eps);
	TestEigenvalues.test_getGEV(Eigenvalues::getGEV_dsygvx_dsyevr,n_tests,n,r,eps);
    }
}



}
