package com.polytechnik.outils;
/*
 * (C) Vladislav Malyshkin 2010
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.*;

/** A crude and simple Newton style root finder algorithm.
 *  Used for simple(non-degenerated) real roots finding only.
 *  Can work in an orthogonal polynomial basis.
 *  @version $Id: PolynomialSimpleRootsFinder.java,v 1.19 2020/02/06 10:29:41 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

public abstract class PolynomialSimpleRootsFinder implements PolynomialRootsFinder  {
    final BasisFunctionsCalculatable B;
    final int maxiterperstep,miniter;
    final double x0,eps1;
    double [] coefs;
    public PolynomialSimpleRootsFinder(final BasisFunctionsCalculatable B,
				       final int maxiterperstep,
				       final int miniter,
				       final double x0,
				       final double eps1)
    {
	this.B=B;
	this.maxiterperstep=maxiterperstep;
	this.miniter=miniter;
	this.x0=x0;
	this.eps1=eps1;
	this.coefs=null;

    }
    public PolynomialSimpleRootsFinder(final BasisFunctionsCalculatable B)
    {
	this(B,100,17,1.1,1e-3);
    }
/** Finds only real roots. im=0.
 */
public PRoots findRoots(final double [] polynomial_coefs){
    final int degree=polynomial_coefs.length-1;
    final double [] re=new double [degree],im=new double [degree],op=polynomial_coefs.clone();
    final int nroots=_findRoots(degree,op,re,im);
    return new PRoots(nroots,op,re,im);
}
    int _findRoots(final int n,
		   final double [] polynomial_coefs,
		   final double [] re_root,
		   final double [] im_root)
    {
	coefs=new double [n+1];
	System.arraycopy(polynomial_coefs,0,coefs,0,n+1);

	int r=0;
	double xinit=x0;
	for(int nn=n;nn>=1;nn--)
	{
	    double x=xinit;
	  rf:{
	     double fx=f(x);
	     int s;
	     if(fx<0) s=-1;
	     else if(fx>0) s=1;
	     else break rf;

	     boolean ftest=true;
	     double xL=Double.NaN,xR=Double.NaN;
	     int sL=0,sR=0;
	     boolean bounded=false;
	     for(int k=0,ke=maxiterperstep+miniter+2,kx=0;k<ke;k++)
	     {
		 double dfx=df(x);
		 if(!(Math.abs(dfx)>0)){dfx=-fx;}
		 final double dx1=fx/dfx;
		 double x1=x-dx1;
		 if(nn<=1){x=x1; break rf;}
		 double fx1=f(x1);
		 int s1;
		 if(fx1>0) s1=1;
		 else if(fx1<0) s1=-1;
		 else {x=x1; /*System.err.println("Break FX1 fx1="+fx1+" x="+x1);*/ break rf;}
		 if(!bounded)
		 {
		     final double xx;
		     final int ss;
		     if(s1*s<0)
		     {
			 xx=x1;ss=s1;
		     }
		     else
		     {
			final double x1a=x1-dx1;
			final double fx1a=f(x1a);
			final int s1a;
			if(fx1a>0) s1a=1;
			else if(fx1a<0) s1a=-1;
			else {x=x1a; break rf;}
			if(s1a*s<0){ss=s1a;xx=x1a;}
			else{ss=s1;xx=x1;}
		     }
		     if(ss*s<0)
		     {
			 if(x<xx){xL=x;xR=xx;sL=s;sR=ss;}
			 else{xL=xx;xR=x;sL=ss;sR=s;}
			 bounded=true;
		     }
		     //System.err.println("Not bounded nn="+nn+" k="+k+" ftest="+ftest+" kx="+kx+" miniter="+miniter);
		 }
		 else
		 {
		     final double xm=0.5*(xL+xR),fm=f(xm);
		     final int sm;
		     if(fm>0) sm=1;
		     else if(fm<0) sm=-1;
		     else {x=xm; break rf;}
		     // adjust x1 if bissection is better
		     //System.err.println("Bounded nn="+nn+" fx1("+x1+")="+fx1+" fm("+xm+")="+fm);
		     //System.err.println("Bounded nn="+nn+" k="+k+" ftest="+ftest+" kx="+kx+" miniter="+miniter);
		     if(x1<xL || x1>xR || Math.abs(fx1)>Math.abs(fm))
		     {
			 x1=xm;
			 fx1=fm;
			 s1=sm;
			 //System.err.println("Bissection is better fx1("+x1+")="+fx1+" fm("+xm+")="+fm);
		     }
		     if(s1==sL){xL=x1;}else{xR=x1;}
		 }
		 final double xx=x;
		 x=x1;
		 fx=fx1;
		 s=s1;

		 if(Math.abs(xx-x)<=eps1)
		 {
		     ftest=false;
		 }
		 else kx=0;
		 if(!ftest)
		 {
		     //System.err.println(" EE: kx="+kx+" miniter="+miniter);
		     if(++kx>=miniter){break rf;}
		 }
	     }
	     //System.err.println("EXIT R");
	     // root not found, skip iterations
	     return r;
	  } // end rf
	  //System.err.println("nn="+nn+" x="+x);
	  re_root[r]=x;
	  im_root[r]=0;
	  coefs=doSdiv1AndScale(coefs,x);
	  //xinit=x;
	  r++;
	}
	return r;
    }

    public double f(final double x)
    {
	return B.getD0(coefs,x);
    }
    public double df(final double x)
    {
	return B.getD1(coefs,x);
    }
    public double [] doSdiv1AndScale(final double [] p,final double x0)
    {
	final SDivRes sdr=B.sdiv1(p,x0);
	final int n=p.length-1;
	final double [] res=sdr.q;
	final double s=res[n-1];
	for(int k=n;--k>=0;)
	{
	    res[k]/=s;
	}
	return res;
    }

    static void test_RealRootsFinder(final int nmax,
				     final PolynomialSimpleRootsFinder thisfinder,
				     final double eps)
    {
	final long t0=System.currentTimeMillis();
	final PolynomialSimpleRootsFinder rf=thisfinder;
	final PolynomialRootsJenkinsTraubORIG jrf=new PolynomialRootsJenkinsTraubORIG();
	for(int n=1;n<=nmax;n++)
	{
	    final double [] pol=new double [n+1];
	    pol[n]=1;
	    final PolynomialRootsFinder.PRoots u=rf.findRoots(pol);
	    final double [] re=u.re_root,im=u.im_root;
	    SortArrays.sortArrayWithLoad(n,re,null);
	    final PolynomialRootsFinder.PRoots j=jrf.findRoots(rf.B.convertFunctionTo_Monomials_from_Basis(pol));
	    final double [] jre=j.re_root,jim=j.im_root;
	    SortArrays.sortArrayWithLoad(n,jre,null);
	    if(u.nroots!=j.nroots) throw new RuntimeException("u.nroots="+u.nroots+" j.nroots="+j.nroots+" n="+n);
	    UnitTests.checkDoubleArrayRel(re,"re",jre,"jre",eps);
	    System.err.println("test_RealRootsFinder: n="+n+" matched OK:");
	}
	final long t1=System.currentTimeMillis();
	System.err.println("PolynomialSimpleRootsFinder:test_RealRootsFinder: OK eps="+eps+
			   " in "+(t1-t0)+" milliseconds.");
    }


}
