package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Pseudocorrelation (with {@literal |o>} component removed) in Hilbert space.
 *  This is a nom-optimized version, used for unit test.
 *  @version $Id: ACorr.java,v 1.75 2022/10/30 21:29:53 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class ACorr {
public final double x_a_a,x_a_b,x_b_b,osp;
public ACorr(final double x_a_b,final double x_a_a,final double x_b_b)
{
    this.x_a_b=x_a_b;
    this.x_a_a=x_a_a;
    this.x_b_b=x_b_b;
    this.osp=x_a_b/Math.sqrt(x_a_a*x_b_b);
}


/** Standard definition of average.
 */
public static ACorr getACorr_STD(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final double [] QQt=M.B.getQQMatr(n,t);
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n,n,ipvt);

    final double [] ptmp=new double [n];
    solveLinsystem(n,ptmp,QQt,b,ipvt,pinfo);
    final double x_a_b=TMatr.mulVVn(n,a,ptmp),x_b_b=TMatr.mulVVn(n,b,ptmp);
    solveLinsystem(n,ptmp,QQt,a,ipvt,pinfo);

    return new ACorr(x_a_b,TMatr.mulVVn(n,a,ptmp),x_b_b);
}

/** Remove {@literal |o>} from a &amp; b.
 */
public static ACorr getACorr_F(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0)
{
    final double [] QQt=M.B.getQQMatr(n,t),o0=M.B.getX0Moments(n,x0),o=new double [n];
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n,n,ipvt);
    solveLinsystem(n,o,QQt,o0,ipvt,pinfo);

    final double o_o=TMatr.mulVVn(n,o,o0),a_o=TMatr.mulVVn(n,a,o),b_o=TMatr.mulVVn(n,b,o);
    final double [] aa=new double [n],bb=new double [n];
    for(int k=0;k<n;k++)
    {
	aa[k]=a[k]-o0[k]*a_o/(o_o);
	bb[k]=b[k]-o0[k]*b_o/(o_o);
    }

    final double [] ptmp=o0;
    solveLinsystem(n,ptmp,QQt,bb,ipvt,pinfo);
    final double x_a_b=TMatr.mulVVn(n,aa,ptmp),x_b_b=TMatr.mulVVn(n,bb,ptmp);
    solveLinsystem(n,ptmp,QQt,aa,ipvt,pinfo);

    return new ACorr(x_a_b,TMatr.mulVVn(n,aa,ptmp),x_b_b);
}



public static ACorr getACorrAverage_Psi(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0)
{
    final double [] psi=new double [n];
    psi[0]=1;
    return getACorr_Psi(psi,a,b,t,M,x0);
}

/** Remove {@literal |o>} from psi.
 */
public static ACorr getACorr_Psi(final double [] psi,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0)
{
    final int n=psi.length;
    final double [] QQt=M.B.getQQMatr(n,t),o=M.B.getX0Moments(n,x0),QQMt=QQt.clone();
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQMt,n,n,ipvt);
    solveLinsystem(n,o,QQMt,o,ipvt,pinfo);
    final double [] psix=TMatr.sumAcB(psi,-M.B.getD0(psi,x0)/M.B.getD0(o,x0),o);
    TMatr.scaleVectorInPlace(psix,1/Math.sqrt(TMatr.getAMatrB(n,psix,psix,QQt,n)));

    final double [] pa=TMatr.getMatr_x_A(n,M.B.getQQMatr(n,a),psix),
	pb=TMatr.getMatr_x_A(n,M.B.getQQMatr(n,b),psix),ptmp=QQt;
    solveLinsystem(n,ptmp,QQMt,pb,ipvt,pinfo);
    final double x_a_b=TMatr.mulVVn(n,pa,ptmp),x_b_b=TMatr.mulVVn(n,pb,ptmp);
    solveLinsystem(n,ptmp,QQMt,pa,ipvt,pinfo);

    return new ACorr(x_a_b,TMatr.mulVVn(n,pa,ptmp),x_b_b);
}


/** Pseudocorrelation without modification psi.
 * x_a_b={@literal <psi|QQa|(QQt)^-1|QQb|psi>}
 */
public static ACorr getACorr_NotModifiedPsi(final double [] psi,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int n=psi.length;
    final double [] QQMt=M.B.getQQMatr(n,t);
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQMt,n,n,ipvt);

    final double [] pa=TMatr.getMatr_x_A(n,M.B.getQQMatr(n,a),psi),
	pb=TMatr.getMatr_x_A(n,M.B.getQQMatr(n,b),psi),ptmp=new double [n];
    solveLinsystem(n,ptmp,QQMt,pb,ipvt,pinfo);
    final double x_a_b=TMatr.mulVVn(n,pa,ptmp),x_b_b=TMatr.mulVVn(n,pb,ptmp);
    solveLinsystem(n,ptmp,QQMt,pa,ipvt,pinfo);

    return new ACorr(x_a_b,TMatr.mulVVn(n,pa,ptmp),x_b_b);
}
/** Pseudocorrelation version which is defined as:
 * x_a_b={@literal <psi|QQa|(QQt)^-1|QQb|psi>-<psi|QQa|psi><psi|QQb|psi>}.
 * The psi should be normalized as {@literal <psi|QQt|psi>=1}.
 */
public static ACorr getACorr_OrtPsiFromQQ(final double [] psi,final double [] QQa,final double [] QQb,final double [] QQt)
{
    final int n=psi.length;
    final double [] QQMt=QQt.clone();
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQMt,n,n,ipvt);

    final double [] pa=new double [n],pb=new double [n],ptmp=new double [n];
    TMatr.multiply_MA_x_MB(pa,1,psi,n,QQa,n);
    TMatr.multiply_MA_x_MB(pb,n,QQb,n,psi,1);
    final double Aa=TMatr.mulVVn(n,pa,psi),Ab=TMatr.mulVVn(n,pb,psi);

    solveLinsystem(n,ptmp,QQMt,pb,ipvt,pinfo);
    final double x_a_b=TMatr.mulVVn(n,pa,ptmp)-Aa*Ab,x_b_b=TMatr.mulVVn(n,pb,ptmp)-Ab*Ab;
    solveLinsystem(n,ptmp,QQMt,pa,ipvt,pinfo);
    final double x_a_a=TMatr.mulVVn(n,pa,ptmp)-Aa*Aa;

    return new ACorr(x_a_b,x_a_a,x_b_b);
}


private static double [] getXmX0P(final double [] P,OrthogonalPolynomialsBasisFunctionsCalculatable<?extends BasisFunctionsCalculatable> M,final double x0)
{
    final double [] m=M.getxP(P);
    TMatr.sumP1(m,P,-x0);
    return m;
}
private static void solveLinsystem(final int n,final double [] res,final double [] M,final double [] x,final int [] ipvt,final int pinfo)
{
    if(res!=x) System.arraycopy(x,0,res,0,n);
    if(pinfo!=0) java.util.Arrays.fill(res,0,n,Double.NaN);
    else com.polytechnik.utils.Linsystems.dgesl_m1(M,n,n,ipvt,res,1);
}
private static void solveMm1A(final int n,final double [] res,final double [] M,final double [] A,final int [] ipvt,final int pinfo,final double [] tmp)
{
    if(pinfo!=0) {java.util.Arrays.fill(res,0,n*n,Double.NaN);return;}
    for(int i=0;i<n;i++)
    {
	System.arraycopy(A,i*n,tmp,0,n);
	com.polytechnik.utils.Linsystems.dgesl_m1(M,n,n,ipvt,tmp,0);
	System.arraycopy(tmp,0,res,i*n,n);
    }
}

public static ACorr getACorrAverage_Psi_and_Basis(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0)
{
    final double [] psi=new double [n];
    psi[0]=1;
    return getACorr_Psi_and_Basis(psi,a,b,t,M,x0);
}

/** Remove {@literal |o>} from psi and basis.
 */
public static ACorr getACorr_Psi_and_Basis(final double [] psi,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0)
{
    final int n0=psi.length;
    if(n0<=1)
    {
	return new ACorr(0,0,0);
    }
    final double [] QQt=M.B.getQQMatr(n0,t),o=M.B.getX0Moments(n0,x0);
    final int [] ipvt=new int [n0];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n0,n0,ipvt);
    solveLinsystem(n0,o,QQt,o,ipvt,pinfo);
    final double [] psix=TMatr.sumAcB(psi,-M.B.getD0(psi,x0)/M.B.getD0(o,x0),o);

    final int nd=n0-1;
    final double [] one=new double []{1};one[0]/=M.B.getD0(one,0);
    final double [] mul2=getXmX0P(getXmX0P(one,M,x0),M,x0);
    final double [] QQMt=M.B.getQQMatr(nd,M.B.getPQkMomentsFromQkMoments(mul2,t));

    final double [] psiX=M.B.sdiv1(psix,x0).q;
    TMatr.scaleVectorInPlace(psiX,1/Math.sqrt(TMatr.getAMatrA(nd,psiX,QQMt,nd)));

    if(false) {
     final double [] QQMa=M.B.getQQMatr(nd,M.B.getPQkMomentsFromQkMoments(mul2,a)),
	 QQMb=M.B.getQQMatr(nd,M.B.getPQkMomentsFromQkMoments(mul2,b)),
	 IQQMt=com.polytechnik.utils.Linsystems.getInvertedMatrix(nd,nd,QQMt);
     return new ACorr(TMatr.getAMatrA(psiX,TMatr.mulMMM(nd,QQMa,IQQMt,QQMb),nd),
		      TMatr.getAMatrA(psiX,TMatr.mulMMM(nd,QQMa,IQQMt,QQMa),nd),
		      TMatr.getAMatrA(psiX,TMatr.mulMMM(nd,QQMb,IQQMt,QQMb),nd));
    }
    // optimization
    final double [] pa=TMatr.getMatr_x_A(nd,M.B.getQQMatr(nd,M.B.getPQkMomentsFromQkMoments(mul2,a)),psiX),
	pb=TMatr.getMatr_x_A(nd,M.B.getQQMatr(nd,M.B.getPQkMomentsFromQkMoments(mul2,b)),psiX),ptmp=QQt;
    final int pminfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQMt,nd,nd,ipvt);
    solveLinsystem(nd,ptmp,QQMt,pb,ipvt,pminfo);
    final double x_a_b=TMatr.mulVVn(nd,pa,ptmp),x_b_b=TMatr.mulVVn(nd,pb,ptmp);
    solveLinsystem(nd,ptmp,QQMt,pa,ipvt,pminfo);

    return new ACorr(x_a_b,TMatr.mulVVn(nd,pa,ptmp),x_b_b);

}


/** Spur like pseudoaverage.
 */
public static ACorr getACorr_SPUR(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final double [] QQa=M.B.getQQMatr(n,a),QQb=M.B.getQQMatr(n,b),QQt=M.B.getQQMatr(n,t);
    if(false)
    {
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt);
	final double x_a_b=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][] {IQQt,QQa,IQQt,QQb})),
	    x_a_a=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][] {IQQt,QQa,IQQt,QQa})),
	    x_b_b=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][] {IQQt,QQb,IQQt,QQb}));
	return new ACorr(x_a_b,x_a_a,x_b_b);
    }
    // optimization
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n,n,ipvt);
    final double [] tmp=new double [n];
    solveMm1A(n,QQa,QQt,QQa,ipvt,pinfo,tmp);
    solveMm1A(n,QQb,QQt,QQb,ipvt,pinfo,tmp);
    return new ACorr(TMatr.getSpurMM(n,QQa,QQb),
		     TMatr.getSpurMM(n,QQa,QQa),
		     TMatr.getSpurMM(n,QQb,QQb));
}


/** Commutator.
 */
static double [][] getCommutatorVarPsiX0(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0)
{
    final double [] QQa=M.B.getQQMatr(n,a),QQb=M.B.getQQMatr(n,b),QQt=M.B.getQQMatr(n,t),o0=M.B.getX0Moments(n,x0);

    if(false)
    {
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt);
	final double [] psi0=TMatr.getMatr_x_A(n,IQQt,o0);
	TMatr.scaleVectorInPlace(psi0,1/Math.sqrt(TMatr.mulVVn(n,psi0,o0)));
	final double [] varpsi0=TMatr.getMatr_x_A(
	    n,
	    TMatr.mulMM(
		n,IQQt,
		TMatr.sumAcB(
		    TMatr.mulMMM(n,QQa,IQQt,QQb),
		    -1,
		    TMatr.mulMMM(n,QQb,IQQt,QQa))),
	    psi0);
	return new double [][]{psi0,varpsi0};
    }

    // optimization
    final double [] psi0=new double [n];
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n,n,ipvt);
    solveLinsystem(n,psi0,QQt,o0,ipvt,pinfo);
    TMatr.scaleVectorInPlace(psi0,1/Math.sqrt(TMatr.mulVVn(n,psi0,o0)));
    double [] tmp=TMatr.getMatr_x_A(n,QQb,psi0);
    solveLinsystem(n,tmp,QQt,tmp,ipvt,pinfo);
    final double [] tmp2=TMatr.getMatr_x_A(n,QQa,tmp);
    tmp=TMatr.getMatr_x_A(n,QQa,psi0);
    solveLinsystem(n,tmp,QQt,tmp,ipvt,pinfo);
    final double [] tmp3=TMatr.getMatr_x_A(n,QQb,tmp);

    TMatr.sumP1(tmp2,tmp3,-1);
    solveLinsystem(n,tmp2,QQt,tmp2,ipvt,pinfo);
    return new double [][]{psi0,tmp2};
}



/** Commutator with derivative.
 */
public static ACorr getCommutatorVarDPsiX0(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0)
{
    return getCommutatorVarDPsiX0(n,a,b,t,M,x0,false);
}



/** Commutator averages infinitesimal shift operator {@literal <shift(psi)|C|psi>}.
 *  Depending on flag_sigma2_onD different normalising is used.
 */
public static ACorr getCommutatorVarDPsiX0(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0,final boolean flag_sigma2_onD)
{
    final double [] QQa=M.B.getQQMatr(n,a),QQb=M.B.getQQMatr(n,b),QQt=M.B.getQQMatr(n,t),o0=M.B.getX0Moments(n,x0);

    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n,n,ipvt);
    final double [] psi0=new double [n];
    solveLinsystem(n,psi0,QQt,o0,ipvt,pinfo);
    TMatr.scaleVectorInPlace(psi0,1/Math.sqrt(TMatr.mulVVn(n,psi0,o0)));

    double [] Dpsi0=new double [n],tmp=new double [n],tmp2=o0;
    final double fd=M.B.getD1(psi0,x0)/M.B.getD0(psi0,x0);
    for(int k=0;k<n;k++){tmp[k]=1;tmp2[k]=M.B.getD1(tmp,x0)-fd*M.B.getD0(tmp,x0);tmp[k]=0;}
    solveLinsystem(n,Dpsi0,QQt,tmp2,ipvt,pinfo);
    TMatr.scaleVectorInPlace(Dpsi0,1/Math.sqrt(TMatr.mulVVn(n,Dpsi0,tmp2)));
    //System.err.println("Dpsi0(x0)="+M.B.getD0(Dpsi0,x0)+" must be 0");
    final double [] psi,Dpsi;
    if(flag_sigma2_onD){psi=Dpsi0;Dpsi=psi0;}
    else{psi=psi0;Dpsi=Dpsi0;}

    tmp=TMatr.getMatr_x_A(n,QQb,psi);
    solveLinsystem(n,tmp,QQt,tmp,ipvt,pinfo);
    tmp2=TMatr.getMatr_x_A(n,QQa,tmp);

    tmp=TMatr.getMatr_x_A(n,QQa,psi);
    solveLinsystem(n,tmp,QQt,tmp,ipvt,pinfo);
    final double [] tmp3=TMatr.getMatr_x_A(n,QQb,tmp);
    TMatr.sumP1(tmp2,tmp3,-1);

    solveLinsystem(n,tmp,QQt,tmp2,ipvt,pinfo);
    final double x_a_b=TMatr.mulVVn(n,tmp2,Dpsi),x_b_b=TMatr.mulVVn(n,tmp,tmp2),x_a_a=1;

    return new ACorr(flag_sigma2_onD?-x_a_b:x_a_b,x_a_a,x_b_b);
}


/** Two point commutator. Similar to getCommutatorVarDPsiX0, but uses
 *  finite shift operator instead of infinitesimal one.
 */
public static ACorr getCommutatorPsiX0PsiX1(final int n,final double [] a,final double [] b,final double [] t,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double x0,final double x1)
{
    final double [] QQa=M.B.getQQMatr(n,a),QQb=M.B.getQQMatr(n,b),QQt=M.B.getQQMatr(n,t),o0=M.B.getX0Moments(n,x0),o1=M.B.getX0Moments(n,x1);

    final double [] psi0=new double [n];
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n,n,ipvt);
    solveLinsystem(n,psi0,QQt,o0,ipvt,pinfo);
    final double o_o_0=TMatr.mulVVn(n,psi0,o0);

    double [] tmp=TMatr.getMatr_x_A(n,QQb,psi0);
    solveLinsystem(n,tmp,QQt,tmp,ipvt,pinfo);
    double [] tmp2=TMatr.getMatr_x_A(n,QQa,tmp);

    tmp=TMatr.getMatr_x_A(n,QQa,psi0);
    solveLinsystem(n,tmp,QQt,tmp,ipvt,pinfo);
    final double [] tmp3=TMatr.getMatr_x_A(n,QQb,tmp);

    TMatr.sumP1(tmp2,tmp3,-1);
    solveLinsystem(n,tmp,QQt,tmp2,ipvt,pinfo);
    final double x_a_b=TMatr.mulVVn(n,tmp,o1),x_b_b=TMatr.mulVVn(n,tmp,tmp2);
    solveLinsystem(n,tmp,QQt,o1,ipvt,pinfo);
    final double o_o_1=TMatr.mulVVn(n,o1,tmp);

    return new ACorr(x_a_b/Math.sqrt(o_o_0*o_o_1),1,x_b_b/o_o_0);
}

public static ACorr getACorr_ABW(final int n,final double [] a,final double [] b,final double [] w)
{
    double x_a_a=0,x_a_b=0,x_b_b=0;
    for(int k=0;k<n;k++)
    {
	final double wg=w!=null?w[k]:1;
	x_a_a+=a[k]*a[k]*wg;
	x_a_b+=a[k]*b[k]*wg;
	x_b_b+=b[k]*b[k]*wg;
    }
    return new ACorr(x_a_b,x_a_a,x_b_b);
}

public static void matchACorr(final ACorr a,final ACorr b,final double eps)
{
    UnitTests.checkDoubleRel(a.x_a_b,"a.x_a_b",b.x_a_b,"b.x_a_b",eps);
    UnitTests.checkDoubleRel(a.x_a_a,"a.x_a_a",b.x_a_a,"b.x_a_a",eps);
    UnitTests.checkDoubleRel(a.x_b_b,"a.x_b_b",b.x_b_b,"b.x_b_b",eps);
    System.err.println("ACorr:matchACorr test OK, eps="+eps);
}

} // END OF class ACorr
