package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.EVSolver;

/** Split a basis to +/-, then solve I-> max in each of them (conditional extremum).
 * @version $Id: BasisSplit.java,v 1.6 2018/12/22 18:38:31 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class BasisSplit {

public final ImaxInSelectedBasis sel_plus,sel_minus;

public BasisSplit(final int n,final double [] ev,final java.util.function.DoublePredicate ev_selection_condition,
		  final double [] QQi,final EVSolver EV,final double eps)
{
    if(ev.length!=n+n*n || QQi.length!=n*n) throw new IllegalArgumentException();
    final double [] basis_plus=new double [n*n],basis_minus=new double [n*n];
    int n_basis_plus=0,n_basis_minus=0;
    for(int k=0;k<n;k++) {
	if(ev_selection_condition.test(ev[k])){
	    System.arraycopy(ev,n+k*n,basis_plus,n_basis_plus*n,n);
	    n_basis_plus++;
	} else {
	    System.arraycopy(ev,n+k*n,basis_minus,n_basis_minus*n,n);
	    n_basis_minus++;
	}
    }
    this.sel_plus=new ImaxInSelectedBasis(n,java.util.Arrays.copyOf(basis_plus,n_basis_plus*n),n_basis_plus,QQi,EV,eps);
    this.sel_minus=new ImaxInSelectedBasis(n,java.util.Arrays.copyOf(basis_minus,n_basis_minus*n),n_basis_minus,QQi,EV,eps);
}

/** A class to solve I -> max problem in a subset of the original basis.
 */
public static class ImaxInSelectedBasis {
    public final double [] psiH,psiL,selected_basis;
    public final double I_H,I_L;
    public final int n_elements_in_basis;
    public ImaxInSelectedBasis(final int n,final double [] selected_basis,final int n_elements_in_basis,
			final double [] QQi,final EVSolver EV,final double eps){
	this.n_elements_in_basis=n_elements_in_basis;
	this.selected_basis=selected_basis;
	if(n_elements_in_basis==0){
	    this.psiH=this.psiL=new double [n];java.util.Arrays.fill(psiH,Double.NaN);
	    this.I_H=this.I_L=Double.NaN;
	} else {
	    final int ixH=n_elements_in_basis-1;
	    final double [] bbQQi=EVXData.convertMatrFromQtoB_Basis(n,QQi,selected_basis,n_elements_in_basis,0),
		ev_bbQQi=EV.getEV(n_elements_in_basis,bbQQi,true),
		bbpsiH=EVXData.getPsiFromBasis(ixH,ev_bbQQi,n_elements_in_basis),
		bbpsiL=EVXData.getPsiFromBasis(0,ev_bbQQi,n_elements_in_basis);
	    this.psiH=EVXData.convertPsiFromBtoQ_Basis(n_elements_in_basis,bbpsiH,n,selected_basis,0);
	    this.psiL=EVXData.convertPsiFromBtoQ_Basis(n_elements_in_basis,bbpsiL,n,selected_basis,0);
	    this.I_H=ev_bbQQi[ixH];
	    this.I_L=ev_bbQQi[0];
	    if(eps>0){
		for(int j=0;j<n_elements_in_basis;j++) for(int k=0;k<n_elements_in_basis;k++){
		  UnitTests.checkDoubleRel(TMatr.getAMatrB(n,selected_basis,j*n,selected_basis,k*n,QQi,n),"<psij|I|psik>",bbQQi[j*n_elements_in_basis+k],"<psij|I|psik>",eps);
		}
	    }
	}
	if(eps>0){
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psiH,QQi,n),"<psiH|I|psiH>",I_H,"I_H",eps);
	    System.err.println("ImaxInSelectedBasis: test OK, n="+n+" n_elements_in_basis="+n_elements_in_basis+" I_H="+I_H);
	}
    }
} // END OF class ImaxInSelectedBasis


}
