package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.SortArrays;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.freemoney.MatricesFromPI;

/**
 * Try to predict the direction using basis transfrom to {@literal |I|psi>=lambda|psi>} basis.
 * We consider the state since a spike in I.
 * The spike is a pure state, integration by parts (since spike till now) gives a mixed state.
 * Obtained density matrix is then tried with several attributes.
 * The most promising one seems top be pI with diagonal elements set to zero (in I-basis).
 * @see InIStateIHDensity
 * @see DM_DI
 * @see DM_DIf
 * @version $Id: BasisTransformPiTiPrediction.java,v 1.65 2022/08/16 10:14:24 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public class BasisTransformPiTiPrediction {
    final EVXData I;
    /** Index of eigenstate selected as a spike.
     */
    public final int ind_M;

    public final double p_M,T_M;

    /** An attribute calcuilated in "since spike till now" mixed state.
     */
    public final DirectionalInfoWithQuadrature dirPi_M,dirTi_M,dirP_M,dirT_M,
	dirDPfromPi_M,dirDPfromPiwithdiag_M,
	dirDP_M,dirDPwithdiag_M;
    public double dp_spurDM_M,dpI_spurDM_M,p_dpI_spur0DM_M,Ddp_dIF_spur,Dp_dIF_spur;

    BasisTransformPiTiPrediction(final int n,
				 final double p_offset,
				 final double p_last,

				 // four input moment matrices
				 final double [] QQt,
				 final double [] QQi,
				 final double [] QQpi,
				 final double [] QQTi,

				 // just a try
				 final double [] QQdp,
				 final double [] QQpt,

				 // basis operations
				 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				 final java.util.function.Function<double [],double []> getPolynomialWIntegratedDt,
				 final java.util.function.Function<double [],double []> getEDPsi,

				 // for unit test
				 final double [] tmom,
				 final double eps){
	final VolMatrix VOLATILITY_EDPsi=getEDPsi::apply;

	this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final double [] bQQTi=EVXData.convertMatrFromQtoB_Basis(n,QQTi,I.ev);
	ind_M=selectIStateIndex(bQQTi,I,eps);

	this.p_M=p_offset+I.getFPsi(ind_M,QQpi)/I.ev[ind_M];
	this.T_M=I.getFPsi(ind_M,QQTi)/I.ev[ind_M];

	final double [] bpsi0=EVXData.convertPsiFromQtoB_Basis(n,I.psi0,I.ev,QQt),
	    psiM=I.getPsi(ind_M),
	    bpsiM=EVXData.convertPsiFromQtoB_Basis(n,psiM,I.ev,QQt),
	    bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	    bQQp=MatricesFromPI.getbQQpFrombQQpi(bQQpi,I),
	    bQQT=MatricesFromPI.getbQQpFrombQQpi(bQQTi,I),
	    bQQdp_fromPi=getbQQdpFrombQQpi(bQQpi,I,QQt,p_last,VOLATILITY_EDPsi),
	    bQQdp=EVXData.convertMatrFromQtoB_Basis(n,QQdp,I.ev),
	    bQQi=TMatr.getDiagMatrix(I.getEValues()),
	    QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	    bQQ0iF=TMatr.scaleVectorInPlace(EVXData.convertMatrFromQtoB_Basis(n,QQ0,I.ev),I.ev[ind_M]),
	    bQQDiffI=EVXData.convertMatrFromQtoB_Basis(n,VOLATILITY_EDPsi.getDiffMatrix(n,QQi),I.ev),
	    bQQDiffPI=EVXData.convertMatrFromQtoB_Basis(n,VOLATILITY_EDPsi.getDiffMatrix(n,QQpi),I.ev),
	    bQQdidtF=TMatr.sumAcB(bQQ0iF,-1,bQQDiffI);

	//System.err.println("bQQdp_fromPi="+java.util.Arrays.toString(bQQdp_fromPi));
	final double []
	    bQQpi_offdiag=bQQpi.clone(),
	    bQQTi_offdiag=bQQTi.clone(),
	    bQQp_offdiag=bQQp.clone(),
	    bQQT_offdiag=bQQT.clone(),
	    bQQdp_fromPi_offdiag=bQQdp_fromPi.clone(),
	    bQQdp_offdiag=bQQdp.clone();

	// remove diagonal elements
	for(int k=0;k<n;k++) {
	    bQQpi_offdiag[k+n*k]=0;
	    bQQTi_offdiag[k+n*k]=0;
	    bQQp_offdiag[k+n*k]=0;
	    bQQT_offdiag[k+n*k]=0;
	    bQQdp_fromPi_offdiag[k+n*k]=0;
	    bQQdp_offdiag[k+n*k]=0;
	}

	final double []
	    UNIT=TMatr.getUnitMatrix(n),
	    // Gram matrix inverse can be obtained without solving a linear system.
	    IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT),
	    // Integrate psiM^2 omega(x) dt: time since psiM spike
	    psiM2Integrated=getPolynomialWIntegratedDt.apply(M.B.mult2Pol(psiM,psiM)),
	    // Construct the measure corresponding to psiM2Integrated polynomial
	    // using Theorem 3 of https://arxiv.org/abs/1807.06007 : find the measure generating given polynomial,
	    // and create density matrix corresponding to this measure.
	    QQDensityMatrix=M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,psiM2Integrated,IQQt)),
	    // Convert it to b-basis to avoid extra IQQt in spur to simplify calculations.
	    bQQDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,QQDensityMatrix,I.ev);

	// various attributes just to check
	dirPi_M=new DirectionalInfoWithQuadrature(n,bQQpi_offdiag,bQQDensityMatrix,true,M.EV,eps);
	dirTi_M=new DirectionalInfoWithQuadrature(n,bQQTi_offdiag,bQQDensityMatrix,true,M.EV,eps);
	dirP_M=new DirectionalInfoWithQuadrature(n,bQQp_offdiag,bQQDensityMatrix,true,M.EV,eps);
	dirT_M=new DirectionalInfoWithQuadrature(n,bQQT_offdiag,bQQDensityMatrix,true,M.EV,eps);

	dirDPfromPi_M=new DirectionalInfoWithQuadrature(n,bQQdp_fromPi_offdiag,bQQDensityMatrix,true,M.EV,eps);
	dirDPfromPiwithdiag_M=new DirectionalInfoWithQuadrature(n,bQQdp_fromPi,bQQDensityMatrix,true,M.EV,eps);
	dirDP_M=new DirectionalInfoWithQuadrature(n,bQQdp_offdiag,bQQDensityMatrix,true,M.EV,eps);
	dirDPwithdiag_M=new DirectionalInfoWithQuadrature(n,bQQdp,bQQDensityMatrix,true,M.EV,eps);

	dp_spurDM_M=TMatr.getSpurMM(n,bQQdp,bQQDensityMatrix);
	dpI_spurDM_M=TMatr.getSpurMMM(n,bQQi,bQQdp,bQQDensityMatrix)/
	    TMatr.getSpurMM(n,bQQi,bQQDensityMatrix)*TMatr.getSpurM(n,bQQDensityMatrix);
	// equals to p_M
	p_dpI_spur0DM_M=p_offset+TMatr.getSpurMM(n,bQQDiffPI,bQQDensityMatrix)/TMatr.getSpurMM(n,bQQ0iF,bQQDensityMatrix);
	Ddp_dIF_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQdp,bQQdidtF,bQQDensityMatrix}));
	Dp_dIF_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQp,bQQdidtF,bQQDensityMatrix}));

	if(eps>0){
	    // match dI/dt
	    UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdidtF,bQQDensityMatrix),"Spur(bQQdidtF*bQQDensityMatrix)",0,"0",eps);
	    final double [] bQQdpidtF=TMatr.sumaAbB((p_dpI_spur0DM_M-p_offset),bQQ0iF,-1,bQQDiffPI);
	    UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdpidtF,bQQDensityMatrix),"Spur(bQQdpidtF*bQQDensityMatrix)",0,"0",eps);
	}

	if(eps>0){
	    final double []
		_IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
		QQpi_offdiag=EVXData.convertMatrFromBtoQ_Basis(n,bQQpi_offdiag,I.ev,QQt),
		QQTi_offdiag=EVXData.convertMatrFromBtoQ_Basis(n,bQQTi_offdiag,I.ev,QQt);

	    // Inverse Gram matrix check
	    UnitTests.checkDoubleArrayRel1(IQQt,"IQQt",_IQQt,"_IQQt",eps);
	    UnitTests.checkDoubleArrayRel1(TMatr.mulMM(n,QQt,IQQt),"QQt * IQQt",UNIT,"UNIT",eps);

	    // Spur invariance (match calculations in original basis).
	    UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,QQpi_offdiag),"spur IQQt*QQpi_offdiag",0,"0",eps);
	    UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,QQTi_offdiag),"spur IQQt*QQTi_offdiag",0,"0",eps);
	    UnitTests.checkDoubleRel(TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQDensityMatrix,IQQt,QQpi_offdiag})),"spur IQQt QQDensityMatrix IQQt QQpi_offdiag",dirPi_M.dir,"dirPi_M.dir",eps);
	    UnitTests.checkDoubleRel(TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQDensityMatrix,IQQt,QQTi_offdiag})),"spur IQQt QQDensityMatrix IQQt QQTi_offdiag",dirTi_M.dir,"dirTi_M,dir",eps);

	    final double
		psiM2Integrated_average=TMatr.mulVV(psiM2Integrated,tmom),
		spur_QQDensityMatrix_IQQt=TMatr.getSpurMM(n,IQQt,QQDensityMatrix);
	    System.err.println("<psiM2Integrated>="+psiM2Integrated_average+" spur="+spur_QQDensityMatrix_IQQt);
	    UnitTests.checkDoubleRel(psiM2Integrated_average,"psiM2Integrated_average",spur_QQDensityMatrix_IQQt,"spur_QQDensityMatrix_IQQt",eps);


	    // Match density mattrix state as a difference betweeen "last" and M states.
	    final double []
		QQdidt0=TMatr.sumaAbB(I.s0,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQi)),
		bQQdidt0=EVXData.convertMatrFromQtoB_Basis(n,QQdidt0,I.ev);
	    UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdidt0,bQQDensityMatrix),"Spur(bQQdidt0*bQQDensityMatrix)",I.s0-I.ev[ind_M],"I.s0-I.ev[ind_M]",eps);
	    // spur with dp/dt is just price difference
	    final double pt_M=TMatr.getAMatrA(psiM,QQpt,n),
		_dp_spurDM=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdp,IQQt,QQDensityMatrix}));
	    UnitTests.checkDoubleRel(_dp_spurDM,"_dp_spurDM",p_last-pt_M,"p_last-pt_M",eps);
	    UnitTests.checkDoubleRel(dp_spurDM_M,"dp_spurDM_M",p_last-pt_M,"p_last-pt_M",eps);
	    System.err.println("BasisTransformPiTiPrediction: test OK n="+n+" eps="+eps);
	}
    }

    /** This method is incorrect and should not be used,
     * use {@link MatricesFromPI#getbQQIpowdpdtFromQQpi_withLastP} as proper alternative.
     * @see MatricesFromPI#getbQQIpowdpdtFromQQpi_withLastP
     */
    private static double [] getbQQdpFrombQQpi(final double [] bQQpi,final EVXData I,final double [] QQt,final double p_last,final VolMatrix VOLATILITY_EDPsi){
	final int n=I.neq;
	final double [] bQQp=MatricesFromPI.getbQQpFrombQQpi(TMatr.sumAcB(bQQpi,-p_last,TMatr.getDiagMatrix(I.getEValues())),I);
	final double []
	    QQp=EVXData.convertMatrFromBtoQ_Basis(n,bQQp,I.ev,QQt),
	    QQdp=VOLATILITY_EDPsi.getDiffMatrix(n,QQp),
	    bQQdp=EVXData.convertMatrFromQtoB_Basis(n,QQdp,I.ev);
	return TMatr.scaleVectorInPlace(bQQdp,-1);
    }


    /** A helper function to select the closest to "now"
     * local maximum in {@link #I}.
     * @param bQQTi The matrix {@literal \( \left<\psi^{[j]}|(t_{now}-t)I|\psi^{[k]}\right> \)}
     * where {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}.
     * The matrix {@literal \( \left<Q_j|(t_{now}-t)I|Q_k\right>\)} in original basis \(Q_k\)
     * can be converted to {@literal \( \left|\psi^{[i]}\right> \)} -basis using
     * {@link EVXData#convertMatrFromQtoB_Basis
         {@code EVXData.convertMatrFromQtoB_Basis(n,QQTi,I.ev)}
       }.
     * @param I Eigenproblem {@literal \( \left|I|\psi^{[i]}\right>=\lambda^{[i]}\left|\psi^{[i]}\right> \)}
     * solution in original basis \(Q_k\).
     * @param eps The value of {@code eps} for selftest,
     * set to {@code 0} to disable.
     * @return The index of selected eigenstate.
     */
private static int getFirstLocalIMaxIndex(final double [] bQQTi,final EVXData I, final double eps){
	final int n=I.neq;
	// find first local maximum in I
	final double [] T_dist=TMatr.getMatrixDiagElements(n,bQQTi);
	for(int k=0;k<n;k++) T_dist[k]/=I.ev[k];
	// Sort them according to T_dist distance.
	// An alternative of K2Distance typically gives very similar result,
	// but sometimes violates triangle inequality, see Distance.java:test_Self().
	final int [] ind=SortArrays.sortArrayIndexes(T_dist);
	int ev_Index= -1;
	for(int k : ind){
	    if(ev_Index<0 || I.ev[k]>=I.ev[ev_Index]){
		ev_Index=k;
	    } else {
		// decreasing I
		break;
	    }
	}
	if(eps>0){
	    System.err.println("BasisTransformPiTiPrediction.getFirstLocalIMaxIndex:selected ind_M="+ev_Index+" distances="+java.util.Arrays.toString(T_dist)+" I.ev="+java.util.Arrays.toString(I.getEValues())+" distance sorted as="+java.util.Arrays.toString(ind));
	}
	if(ev_Index<0) ev_Index=n-1;
	return ev_Index;
    }


private static int getFirstIgtI0(final double [] bQQTi,final EVXData I, final double eps){
	final int n=I.neq;
	// find first local maximum in I
	final double [] T_dist=TMatr.getMatrixDiagElements(n,bQQTi);
	for(int k=0;k<n;k++) T_dist[k]/=I.ev[k];
	// Sort them according to T_dist distance.
	// An alternative of K2Distance typically gives very similar result,
	// but sometimes violates triangle inequality, see Distance.java:test_Self().
	final int [] ind=SortArrays.sortArrayIndexes(T_dist);
	int ev_Index= -1;
	for(int k : ind){
	    if(I.ev[k]>I.s0){
		ev_Index=k;
		break;
	    }
	}
	if(ev_Index<0) ev_Index=n-1;
	if(eps>0){
	    System.err.println("BasisTransformPiTiPrediction.getFirstLocalIMaxIndex:selected ind_M="+ev_Index+" distances="+java.util.Arrays.toString(T_dist)+" I.ev="+java.util.Arrays.toString(I.getEValues())+" distance sorted as="+java.util.Arrays.toString(ind));
	}
	return ev_Index;
    }

private static int selectIStateIndex(final double [] bQQTi,final EVXData I, final double eps){
    	if(false) return getFirstIgtI0(bQQTi,I,eps);
	if(false) return getFirstLocalIMaxIndex(bQQTi,I,eps);// the state of local max I state most close to "now"
	// the state of maximal I, global I maximum.
	return I.neq-1;
    }



static void match(final BasisTransformPiTiPrediction a,final BasisTransformPiTiPrediction b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    // in this version output prices (public fields) are without an offset
    UnitTests.checkDoubleRel(a.p_M,"a.p_M",b.p_M,"b.p_M",eps);
    UnitTests.checkDoubleRel(a.T_M,"a.T_M",b.T_M,"b.T_M",eps);
    UnitTests.checkDoubleRel(a.dp_spurDM_M,"a.dp_spurDM_M",b.dp_spurDM_M,"b.dp_spurDM_M",eps);
    UnitTests.checkDoubleRel(a.dpI_spurDM_M,"a.dpI_spurDM_M",b.dpI_spurDM_M,"b.dpI_spurDM_M",eps);


    DirectionalInfoWithQuadrature.match(a.dirPi_M,b.dirPi_M,eps);
    DirectionalInfoWithQuadrature.match(a.dirTi_M,b.dirTi_M,eps);
    DirectionalInfoWithQuadrature.match(a.dirP_M,b.dirP_M,eps);
    DirectionalInfoWithQuadrature.match(a.dirT_M,b.dirT_M,eps);
    DirectionalInfoWithQuadrature.match(a.dirDPfromPi_M,b.dirDPfromPi_M,eps);
    DirectionalInfoWithQuadrature.match(a.dirDPfromPiwithdiag_M,b.dirDPfromPiwithdiag_M,eps);
    DirectionalInfoWithQuadrature.match(a.dirDP_M,b.dirDP_M,eps);
    DirectionalInfoWithQuadrature.match(a.dirDPwithdiag_M,b.dirDPwithdiag_M,eps);

    System.err.println("BasisTransformPiTiPrediction test OK n="+a.I.neq+" eps="+eps);
}


}
