package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.realtimebook.*;

/** Rurik measure (continued).
 *  Like BQVM but now for multi-stocks
 *  @version $Id: CQVM.java,v 1.33 2022/11/20 13:21:11 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class CQVM<DTYPE> {

final DTYPE [] odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;

/** Rurik evolution. */
final double saL,waL,saH,waH,sa0,K00t;

CQVM(final int neq,final DTYPE [] odata)
{
    this.odata=odata;
    this.access=getAccess();

    final int nmoms=2*neq-1,ns=odata.length;

    final double [][] moms=new double [2*ns][nmoms];
    for(int k=0,o=0;k<ns;k++)
    {
	// moments for k-th instrument
	access.setM(moms[o++],0,access.M_I,odata[k],nmoms);
	access.setM(moms[o++],0,access.M_T,odata[k],nmoms);
    }
    final double [][] QQm=access.getQQ_Matrs(neq,1,moms);

    final int isel=neq-1;
    final double [] momrts=new double [nmoms],momrvs=new double [nmoms],momx=new double [nmoms];
    for(int k=0,o=0;k<ns;k++)
    { // over all instruments
	final double [] QQv=QQm[o++],QQt_sam=QQm[o++];
	final double [] evI=solveGeneralizedEV(neq,QQv,QQt_sam);
	final double IE=evI[isel],F=getF(odata[k]);
	access.setM(momrts,0,access.M_RTS,odata[k],nmoms);
	access.setM(momrvs,0,access.M_RVS,odata[k],nmoms);
	for(int l=nmoms;--l>=0;){momx[l]+=F*(momrts[l]*IE-momrvs[l]);}
    }
    final double x0=getX0();
    final double [] momt=momrts; // re-use temporary array
    // t-moments exact values
    access.setMT(momt,nmoms);
    final double [] QQf[]=access.getQQ_Matrs(neq,1,new double [][]{momx,momt}),
	QQx=QQf[0],QQt=QQf[1],r=getOKDiagT(neq,QQt);
    K00t=access.fx(r,x0);
    sa0=TMatr.getAMatrA(neq,r,QQx,neq)/K00t;
    final double sqrt_o_o=Math.sqrt(K00t);

    final double [] evS=getSymEVDiagT(neq,QQx,QQt); // use exact QQt
    saL=evS[0];
    saH=evS[neq-1];
    System.arraycopy(evS,neq+(0)*neq,r,0,neq);
    waL=access.fx(r,x0)/sqrt_o_o;
    System.arraycopy(evS,neq+(neq-1)*neq,r,0,neq);
    waH=access.fx(r,x0)/sqrt_o_o;
}


double [] getOKDiagT(final int n,final double [] QQt)
{
    final double x0=getX0();
    final double [] r=new double [n],o=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=1;o[k]=access.fx(r,x0);r[k]=0;
	o[k]/=QQt[k*n+k]; // assume QQt is diagonal
    }
    return o;
}
double [] getSymEVDiagT(final int n,final double [] QQv,final double [] QQt)
{
    return EVXData.getSymEVDiagT(n,QQv,QQt,DeprecatedHelperEVXData.EV);
}
double [] solveGeneralizedEV(final int n,final double [] A,final double [] B)
{
    return DeprecatedHelperEVXData.solveGeneralizedEV(n,A,B);
}
/** Return 1 in a model of 100% correlation. */
double getF(final DTYPE odata){return 1;}



abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract double getX0();

static abstract class CQVMTest<DTYPE> {

int neq;
long currenttime,tau;
Trade [][] tba;
int [] tba_size;

abstract long getTau(final Trade [] tb);
abstract CQVM<DTYPE> getCQVM1();
abstract CQVM<DTYPE> getCQVM();
abstract BQVM<?> getBQVM(final int ni);

final String name;
final java.util.Random r;
final double eps;
CQVMTest(final String name,
	 final java.util.Random r,final double eps)
{
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Common()
{

    System.err.println(name+":CQVMTest:test_Common(): test OK neq="+neq+" neq_i="+tba.length+" eps="+eps);
}

void setData(final int tsize,final int neq,final int neq_i)
{
    this.neq=neq;
    this.tba=new Trade [neq_i][];
    this.tba_size=new int [neq_i];
    long currenttime=-1;
    for(int i=0;i<neq_i;i++)
    {
	final Trade [] tb=tba[i]=CalculateKendall.TradeEMoments.getRandomTrades(tsize+r.nextInt(tsize+1),r);
	if(i<=0)
	{
	    this.tau=getTau(tb);
	}
	tba_size[i]=tb.length;
	if(currenttime<tb[tb.length-1].time)
	{
	    currenttime=tb[tb.length-1].time;
	}
    }
    this.currenttime=currenttime;
}

void test_BQVM1()
{
    final CQVM<DTYPE> q=getCQVM1();
    final BQVM<?> s=getBQVM(0);

    UnitTests.checkDoubleRel(q.K00t,"q.K00t",s.K00t,"s.K00t",eps);
    UnitTests.checkDoubleRel(q.sa0,"q.sa0",s.DS.s0,"s.DS.s0",eps);
    UnitTests.checkDoubleRel(Math.abs(q.waL),"q.waL",Math.abs(s.DS.wL),"s.DS.wL",eps);
    UnitTests.checkDoubleRel(q.saL,"q.saL",s.DS.sL,"s.DS.sL",eps);
    UnitTests.checkDoubleRel(Math.abs(q.waH),"q.waH",Math.abs(s.DS.wH),"s.DS.wH",eps);
    UnitTests.checkDoubleRel(q.saH,"q.saH",s.DS.sH,"s.DS.sH",eps);

    System.err.println(name+":CQVMTest:test_BQVM1(): test OK neq="+neq+" eps="+eps);
}
void test_BQVM()
{
    final CQVM<DTYPE> q=getCQVM();
    final double [] QQx=new double [neq*neq];
    BQVM<?> s=null;
    double sa0=0;
    for(int k=0;k<tba.length;k++)
    {
	s=getBQVM(k);
	UnitTests.checkDoubleRel(q.K00t,"q.K00t",s.K00t,"s.K00t",eps);
	final double F=q.getF(q.odata[k]);
	for(int l=QQx.length;--l>=0;) {QQx[l]+=F*(s.ITsam.sH*s.sQQMrts[l]-s.sQQMrvs[l]);}
	sa0+=F*s.DS.s0;
    }
    final double [] evS=DeprecatedHelperEVXData.solveGeneralizedEV(neq,QQx,s.sQQt),r=new double [neq];
    final double x0=s.getX0(),sqrt_o_o=Math.sqrt(s.K00t),saL=evS[0],saH=evS[neq-1];
    System.arraycopy(evS,neq+(0)*neq,r,0,neq);
    final double waL=s.access.fx(r,x0)/sqrt_o_o;
    System.arraycopy(evS,neq+(neq-1)*neq,r,0,neq);
    final double waH=s.access.fx(r,x0)/sqrt_o_o;

    UnitTests.checkDoubleRel(q.sa0,"q.sa0",sa0,"sa0",eps);
    UnitTests.checkDoubleRel(q.saL,"q.saL",saL,"saL",eps);
    UnitTests.checkDoubleRel(Math.abs(q.waL),"q.waL",Math.abs(waL),"waL",eps);
    UnitTests.checkDoubleRel(q.saH,"q.saH",saH,"saH",eps);
    UnitTests.checkDoubleRel(Math.abs(q.waH),"q.waH",Math.abs(waH),"waH",eps);

    System.err.println(name+":CQVMTest:test_BQVM(): test SUM OK neq="+neq+" neq_i="+tba.length+" eps="+eps);
}


void test_Calculations(final int tsize,final int neq,final int neq_i)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":CQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq,neq_i);
    test_Common();
    test_BQVM1();
    test_BQVM();

    System.err.println(name+":CQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max,final int neq_i_max)
{
    System.err.println(name+":CQVMTest: starting tests; ver=$Id: CQVM.java,v 1.33 2022/11/20 13:21:11 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    for(int neq_i=1;neq_i<=neq_i_max;neq_i++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(950+r.nextInt(850),neq,neq_i);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":CQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: CQVM.java,v 1.33 2022/11/20 13:21:11 mal Exp $");
}
} // END OF class CQVMTest

}// END OF class CQVM
