package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;



/** Calculate orders volume for slices in time space.
 *  Similar to CalculateTradeVolume, but we do not merge them yet,
 *  both classes are still evolving.
 *  @version $Id: CalculateOrdersVolumeFixedThreshold.java,v 1.14 2020/02/06 00:09:22 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class CalculateOrdersVolumeFixedThreshold {


    private final Distribution<?extends BookItem> orders;
    CalculateOrdersVolumeFixedThreshold(final Distribution<?extends BookItem> orders)
    {
	this.orders=orders;
    }


	long []
	// volumes
	    v_buy_tau=null,
	    v_sell_tau=null;


    double [] book_median_buy_tau=null,book_median_sell_tau=null;



    void setData(final int nt)
    {
	if(nt<0)
	{
	    this.v_buy_tau=this.v_sell_tau=null;
	    this.book_median_buy_tau=this.book_median_sell_tau=null;
	}
	else
	{
	    this.v_buy_tau=new long [nt];
	    this.v_sell_tau=new long [nt];
	    this.book_median_buy_tau=new double [nt];
	    this.book_median_sell_tau=new double [nt];
	}
    }


    void calculateOrdersVolume(final long [] delays,
			       final long currenttime,
			       final double frac,
			       final double [] levels_buy,
			       final double [] levels_sell)
    {
	if(frac <0.0 || frac>1.0)
	{
	    throw new IllegalArgumentException("frac="+frac);
	}

	final int nt=delays.length;
	setData(nt);
	if(nt!=levels_buy.length || nt!=levels_sell.length)
	{
	    throw new IllegalArgumentException();
	}
	final BookItem [] tb=this.orders.getBuffer();
	final int tsize=this.orders.getSize();
	if(tsize>tb.length)
	{
	    throw new RuntimeException("This should never happen");
	}


	final boolean PRINT_DATA_DEBUG=false;
	//currenttime==34344233000000L;
	//currenttime==32894696000000L;
	//currenttime==34580422000000L;
	//currenttime>=9*3600*TraderInterface.TIME_SCALE && tsize>0;


	// order array by time.
	// the buffer is considered poisoned after this.
	com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tb,0,tsize-1,CompareItems.cmp_by_time);


	if(PRINT_DATA_DEBUG)
	{
	    System.err.println("=="+getClass().getName()+"============BOOK VOLUME t=\t"+currenttime+"\tNPRDERS=\t"+tsize+
			       "\ttsize=\t"+tsize);
	}
	if(PRINT_DATA_DEBUG)
	{
	    System.err.println("num\t"+
			       "time\t"+
			       "shares\t"+
			       "buy_sell_indicator\t"+
			       "price\t");
	}

	final int [] tau_interval_range_ptr=new int [nt];
	for(int i=nt;i-- >0;)
	{
	    // flag not set
	    tau_interval_range_ptr[i]=-1;
	    this.book_median_buy_tau[i]=this.book_median_sell_tau[i]=Double.NaN;
	}
	// time interval ranges
	long tbefore=currenttime-delays[0];
	long prevt=-1L;
        gloop:for(int i=tsize,cur_pos=0;i-- >0;)
	{
	    final BookItem t=tb[i];
	    final long tt=t.getTime();
	    if(prevt>=0 && tt>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+
					   prevt+" and "+tt);
	    }
	    prevt=tt;
	    if(PRINT_DATA_DEBUG)
	    {
		System.err.println(String.valueOf(i)+"\t"+
				   tt+"\t"+
				   t.shares+"\t"+
				   t.buy_sell_indicator+"\t"+
				   t.price);
	    }
	    for(;!(tt>=tbefore);)
	    {
		if(++cur_pos>=nt)
		{
		    break gloop;
		}
		tbefore=currenttime-delays[cur_pos];
	    }
	    tau_interval_range_ptr[cur_pos]=i;
	}

	// calculate the volume for median calculation
	final long [] ba=this.v_buy_tau,sa=this.v_sell_tau;
	for(int cur_pos=0;cur_pos<nt;cur_pos++)
	{
	    final int ptr_start=tau_interval_range_ptr[cur_pos];
	    if(ptr_start<0)
	    {
		// no points for this interval
		// previous interval is the same as this one
		if(cur_pos>0)
		{
		    ba[cur_pos]=ba[cur_pos-1];
		    sa[cur_pos]=sa[cur_pos-1];
		}
		continue;
	    }
	    final double price_buy_cutoff=levels_buy[cur_pos],price_sell_cutoff=levels_sell[cur_pos];
	    for(int i=tsize-1;i>=ptr_start;i--)
	    {
		final BookItem t=tb[i];
		final int s=t.shares;
		if(t.buy_sell_indicator==BookItem.BOOK_BUY)
		{
		    if(!(t.price>=price_buy_cutoff))
		    {
			continue;
		    }
		    ba[cur_pos]+=s;
		}
		else if(t.buy_sell_indicator==BookItem.BOOK_SELL)
		{
		    if(!(t.price<=price_sell_cutoff))
		    {
			// typically should never happen
			continue;
		    }
		    sa[cur_pos]+=s;
		}
	    }
	}


	    // calculate median in price space for slices in time space
	    for(int cur_pos=0;cur_pos<nt;cur_pos++)
	    {
		if(cur_pos>0 &&((ba[cur_pos]<ba[cur_pos-1])||(sa[cur_pos]<sa[cur_pos-1])))
		{
		    throw new IllegalStateException();
		}
		final int ptr_start=tau_interval_range_ptr[cur_pos];
		if(ptr_start<0)
		{
		    // no points for this interval
		    // previous interval is the same as this one
		    continue;
		}
		final long tnotafter=currenttime-delays[cur_pos];
		// order array by type and price.
		// the buffer is considered poisoned after this.
		com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tb,ptr_start,tsize-1,Distribution.cmp_orders_by_price_and_type);

		// median for buy
		final double price_buy_cutoff=levels_buy[cur_pos];
		long vol=0,prevvol=0,lev=(long)(frac*ba[cur_pos]);
		double prev_price=Double.NaN;
		for(int i=ptr_start;i<tsize;i++)
		{
		    final BookItem t=tb[i];
		    if(t.buy_sell_indicator!=BookItem.BOOK_BUY)
		    {
			break;
		    }
		    if(prev_price<t.price || t.getTime()<tnotafter) // will be false for NaN
		    {
			throw new RuntimeException("Ordering is broken");
		    }
		    else if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
		    {
			if(i>ptr_start)
			{
			    final int i1=i,i2=i-1;
			    if(tb[i2].price<tb[i1].price)
			    {
				throw new RuntimeException(
				    "Ordering is broken"+
				    " i1="+i1+" p1="+tb[i1].price+
				    " i2="+i2+" p2="+tb[i2].price);
			    }
			    else
			    {
				if(tb[i2].price==tb[i1].price)
				{
				    if(tb[i2].time>tb[i1].time)
				    {
					throw new RuntimeException(
					    "Ordering is broken"+
					    " i1="+i1+" t1="+tb[i1].time+
					    " i2="+i2+" t2="+tb[i2].time);
				    }
				    else if(tb[i2].time==tb[i1].time && (tb[i2].orderreferencenumber>tb[i1].orderreferencenumber))
				    {
					throw new RuntimeException(
					    "Ordering is broken"+
					    " i1="+i1+" orderreferencenumber1="+tb[i1].orderreferencenumber+
					    " i2="+i2+" orderreferencenumber2="+tb[i2].orderreferencenumber);
				    }
				}
			    }
			}
		    }
		    prev_price=t.price;
		    if(!(t.price>=price_buy_cutoff))
		    {
			// typically should never happen
			continue;
		    }
		    vol+=t.shares;
		    if(vol<0)
		    {
			throw new RuntimeException("Integer overflow currenttime="+currenttime);
		    }
		    if(vol>lev && lev>=prevvol)
		    {
			this.book_median_buy_tau[cur_pos]=t.price;
			// we can add break here.
			// do not do this to check constrain later
		    }
		    prevvol=vol;
		}
		if(vol!=ba[cur_pos])
		{
		    throw new RuntimeException("Broken invariant. This should never happen. currenttime="+currenttime);
		}

		// median for sell
		final double price_sell_cutoff=levels_sell[cur_pos];
		vol=prevvol=0;
		lev=(long)(frac*sa[cur_pos]);
		prev_price=Double.NaN;
		for(int i=tsize-1;i>=ptr_start;i--)
		{
		    final BookItem t=tb[i];
		    if(t.buy_sell_indicator!=BookItem.BOOK_SELL)
		    {
			break;
		    }
		    if(prev_price>t.price || t.getTime()<tnotafter) // will be false for NaN
		    {
			throw new RuntimeException("Ordering is broken");
		    }
		    else if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
		    {
			if(i<tsize-1)
			{
			    final int i1=i,i2=i+1;
			    if(tb[i2].price>tb[i1].price)
			    {
				throw new RuntimeException(
				    "Ordering is broken"+
				    " i1="+i1+" p1="+tb[i1].price+
				    " i2="+i2+" p2="+tb[i2].price);
			    }
			    else
			    {
				if(tb[i2].price==tb[i1].price)
				{
				    if(tb[i2].time>tb[i1].time)
				    {
					throw new RuntimeException(
					    "Ordering is broken"+
					    " i1="+i1+" t1="+tb[i1].time+
					    " i2="+i2+" t2="+tb[i2].time);
				    }
				    else if(tb[i2].time==tb[i1].time && (tb[i2].orderreferencenumber>tb[i1].orderreferencenumber))
				    {
					throw new RuntimeException(
					    "Ordering is broken"+
					    " i1="+i1+" orderreferencenumber1="+tb[i1].orderreferencenumber+
					    " i2="+i2+" orderreferencenumber2="+tb[i2].orderreferencenumber);
				    }
				}
			    }
			}
		    }
		    prev_price=t.price;
		    if(!(t.price<=price_sell_cutoff))
		    {
			// typically should never happen
			continue;
		    }
		    vol+=t.shares;
		    if(vol<0)
		    {
			throw new RuntimeException("Integer overflow currenttime="+currenttime);
		    }
		    if(vol>lev && lev>=prevvol)
		    {
			this.book_median_sell_tau[cur_pos]=t.price;
			// we can add break here.
			// do not do this to check constrain later
		    }
		    prevvol=vol;
		}
		if(vol!=sa[cur_pos])
		{
		    throw new RuntimeException("Broken invariant. This should never happen. currenttime="+currenttime);
		}
	    }

            // check previous.
	    for(int i=0;i<nt;i++)
	    {
		final int i1=i-1;
		final long vb=this.v_buy_tau[i]-(i1>=0?this.v_buy_tau[i1]:0),
		    vs=this.v_sell_tau[i]-(i1>=0?this.v_sell_tau[i1]:0);
		if(tau_interval_range_ptr[i]<0 && (
		       !Double.isNaN(this.book_median_buy_tau[i])||
		       !Double.isNaN(this.book_median_sell_tau[i])||
		       vb>0||vs>0))
		{
		    throw new RuntimeException("Broken constrains");
		}
		if(i1>=0)
		{
		    // we did not calculate intervals if they have no points.
		    // use previously calculated intervals
		    if(Double.isNaN(this.book_median_buy_tau[i]))
		    {
			this.book_median_buy_tau[i]=this.book_median_buy_tau[i1];
		    }
		    if(Double.isNaN(this.book_median_sell_tau[i]))
		    {
			this.book_median_sell_tau[i]=this.book_median_sell_tau[i1];
		    }
		}

		if(Double.isNaN(this.book_median_buy_tau[i])!=(this.v_buy_tau[i]==0))
		{
		    throw new RuntimeException("buy_vol."+delays[i]+"="+this.v_buy_tau[i]+
					       " mismatch with 0; currenttime="+currenttime);
		}
		if(Double.isNaN(this.book_median_sell_tau[i])!=(this.v_sell_tau[i]==0))
		{
		    throw new RuntimeException("sell_vol."+delays[i]+"="+this.v_sell_tau[i]+
					       " mismatch with 0; currenttime="+currenttime);
		}
	    }



	    if(PRINT_DATA_DEBUG)
	    {
		System.err.println("tau\t"+
				   "vol_buy\t"+
				   "vol_sell\t"+
				   "vol_total\t"+
				   "\t"+
				   "median_buy\t"+
				   "median_sell\t"+
				   "\t"+
				   "lev_buy\t"+
				   "lev_sell");
		for(int i=0;i<nt;i++)
		{
		  System.err.println(String.valueOf(delays[i])+"\t"+
				     this.v_buy_tau[i]+"\t"+
				     this.v_sell_tau[i]+"\t"+
				     (this.v_buy_tau[i]+this.v_sell_tau[i])+"\t"+
				     "\t"+
				     this.book_median_buy_tau[i]+"\t"+
				     this.book_median_sell_tau[i]+"\t"+
				     "\t"+
				     levels_buy[i]+"\t"+
				     levels_sell[i]);
		}
	    }
    }

}
