package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.freemoney.CommonlyUsedMoments;
import com.polytechnik.realtimebook.*;

/** For unit tests of {@link CommonlyUsedMoments}.
 * @version $Id: CommonlyUsedMoments_TestFromSample.java,v 1.32 2022/09/14 21:24:37 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class CommonlyUsedMoments_TestFromSample  {

    @FunctionalInterface public interface InitStorage<RES> {
	RES apply(int n, double atau);
    }

    public static <T extends CommonlyUsedMoments> T fromSeqDirectly(
	final InitStorage<T> init,
	final int n,
	final long currenttime,
	final long tau,
	final Trade [] tb){

	final T m=init.apply(n,tau);

	double p_offset=Double.NaN;
	for(int k=tb.length;--k>=tb.length-6 && k>=0;) p_offset=tb[k].price+0.11; // some random offset
	m.updateBasisToNewTimeReference(currenttime,p_offset);

	for(var t:tb){
	    m.addObservationNoBasisShift(currenttime,t.time,t.price,t.shares);
	}
	return m;
    }


    public static <T extends CommonlyUsedMoments> T fromSeqRecurrently(
	final InitStorage<T> init,
	final int n,
	final long currenttime,
	final long tau,
	final Trade [] tb){

	final T m=init.apply(n,tau);
	for(var t:tb){
	    m.updateWithSingleObservation(t.time,t.time,t.price,t.shares);
	}

	final double p_offset=m.getPOffset()-0.061; // to test price offset invariance
	m.updateBasisToNewTimeReference(currenttime,p_offset);
	return m;
    }


    public static <T extends CommonlyUsedMoments> T fromSeqRecurrentlyTestSecondarySampling(
	final InitStorage<T> init,
	final int n,
	final long currenttime,
	final long tau,
	final Trade [] tb,
	final double eps){

	long vtotal=0;
	double atotal=0;
	Trade tprev=null;
	for(var t:tb){
	    vtotal+=t.shares;
	    if(tprev!=null) atotal+=Math.abs(t.price-tprev.price);
	    tprev=t;
	}
	final T m=init.apply(n,tau),ma=init.apply(n,tau),
	    mtestdPIH=init.apply(n,tau),
	    mtestVdP=init.apply(n,tau);
	// Set initial IH to match other moments
	m.setInitIHForSecondarySampling(0);
	ma.setInitIHForSecondarySampling(0);
	mtestdPIH.setInitIHForSecondarySampling(0);
	mtestVdP.setInitIHForSecondarySampling(-vtotal);

	long V=0;
	double a=0;
	tprev=null;
	for(var t:tb){
	    //System.err.println("!!! k="+/*k+*/" V="+(V+t.shares)+" t.shares="+t.shares+" p="+t.price+" vtotal="+vtotal);
	    final double ashares=tprev==null?0:Math.abs(t.price-tprev.price);
	    tprev=t;
	    m.updateWithSingleObservation(t.time,t.time,t.price,t.shares);
	    ma.updateWithSingleObservation(t.time,t.time,t.price,ashares);
	    mtestdPIH.updateWithSingleObservation(t.time,t.time,t.price,t.shares);
	    mtestVdP.updateWithSingleObservation(t.time,t.time,t.price,t.shares);
	    V+=t.shares;
	    a+=ashares;

	    // imitate secondary sampling to match I, pI, and VdP/dt moments
	    m.addIHObservationSecondarySampling(V);
	    ma.addIHObservationSecondarySampling(a);
	    mtestdPIH.addIHObservationSecondarySampling(V*t.price);
	    mtestVdP.addIHObservationSecondarySampling(V-vtotal);
	}

	final double p_offset=m.getPOffset()-0.061; // to test price offset invariance
	m.updateBasisToNewTimeReference(currenttime,p_offset);
	ma.updateBasisToNewTimeReference(currenttime,p_offset);
	mtestdPIH.updateBasisToNewTimeReference(currenttime,p_offset);
	mtestVdP.updateBasisToNewTimeReference(currenttime,p_offset);

	if(eps>0){
	    UnitTests.checkDoubleRel1(m.getTotalVolume(),"m.getTotalVolume()",vtotal,"vtotal",eps);
	    UnitTests.checkDoubleArrayRel1(m.getdVMoments(),"m.getdVMoments()",m.getdIHdtMoments(),"m.getdIHdtMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(m.getPdVMoments(),"m.getPdVMoments()",m.getPdIHdtMoments(),"m.getPdIHdtMoments()",eps);
	    final double [] VdpFromdPIH=TMatr.sumAcBdE(m.getdPIHdtMoments(),-1,m.getPdIHdtMoments(),-vtotal,m.getdpMoments());
	    UnitTests.checkDoubleArrayRel1(m.getVdPMoments(),"m.getVdPMoments()",VdpFromdPIH,"VdpFromdPIH",eps);

	    final double [] mcheckdPIH=TMatr.sumAcB(mtestdPIH.getdIHdtMoments(),-m.getPOffset(),m.getdIHdtMoments());
	    UnitTests.checkDoubleArrayRel1(mcheckdPIH,"mcheckdPIH",m.getdPIHdtMoments(),"m.getdPIHdtMoments()",eps);

	    // test VdP/dt
	    UnitTests.checkDoubleArrayRel1(mtestVdP.getdIHdtMoments(),"mtestVdP.getdIHdtMoments()",m.getdVMoments(),"m.getdVMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(mtestVdP.getPdIHdtMoments(),"mtestVdP.getPdIHdtMoments()",m.getPdVMoments(),"m.getPdVMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(TMatr.sumAcB(mtestVdP.getdPIHdtMoments(),-1,mtestVdP.getPdIHdtMoments()),"mtestVdP.getdPIHdtMoments()-mtestVdP.getPdIHdtMoments()",m.getVdPMoments(),"m.getVdPMoments()",eps);
	    System.err.println("CommonlyUsedMoments_TestFromSample: secondary sampling matched OK n="+n+" eps="+eps);
	}
	// check a-moments
	if(eps>0){
	    UnitTests.checkDoubleRel1(m.getTotalA(),"m.getTotalA()",atotal,"atotal",eps);
	    UnitTests.checkDoubleArrayRel1(ma.getdVMoments(),"ma.getdVMoments()",m.getdAMoments(),"m.getdAMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(ma.getPdVMoments(),"ma.getPdVMoments()",m.getPdAMoments(),"m.getPdAMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(ma.getVdPMoments(),"ma.getVdPMoments()",m.getAdPMoments(),"m.getAdPMoments()",eps);

	    UnitTests.checkDoubleArrayRel1(ma.getdVMoments(),"ma.getdVMoments()",ma.getdIHdtMoments(),"ma.getdIHdtMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(ma.getPdVMoments(),"ma.getPdVMoments()",ma.getPdIHdtMoments(),"ma.getPdIHdtMoments()",eps);
	    final double [] AdpFromdPIH=TMatr.sumAcBdE(ma.getdPIHdtMoments(),-1,ma.getPdIHdtMoments(),-atotal,ma.getdpMoments());
	    UnitTests.checkDoubleArrayRel1(ma.getVdPMoments(),"ma.getVdPMoments()",AdpFromdPIH,"AdpFromdPIH",eps);
	    System.err.println("CommonlyUsedMoments_TestFromSample: da moments matched OK n="+n+" eps="+eps);
	}

	return m;
    }

    static void test_CalculateMoments_matchSMomentsData(
	final SMomentsData mm,
	final java.util.List<? extends CommonlyUsedMoments> toTest,
	final double eps){
	for(var mD:toTest){
	    UnitTests.checkDoubleArrayRel(mm.imom,"mm.imom",mD.getdVMoments(),"mD.getdVMoments()",eps);
	    UnitTests.checkDoubleArrayRel(mm.tmom,"mm.tmom",mD.getdtMoments(),"mD.getdtMoments()",eps);
	    UnitTests.checkDoubleArrayRel(mm.Tmom,"mm.Tmom",mD.getTdtMoments(),"mD.getTdtMoments()",eps);
	    UnitTests.checkDoubleArrayRel(mm.Vmom,"mm.Vmom",mD.getVdtMoments(),"mD.getVdtMoments()",eps);
	    UnitTests.checkDoubleArrayRel(mm.Timom,"mm.Timom",mD.getTdVMoments(),"mD.getTdVMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(mm.dpmom,"mm.dpmom",mD.getdpMoments(),"mD.getdpMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(mm.Tdpmom,"mm.Tdpmom",mD.getTdpMoments(),"mD.getTdpMoments()",eps);
	    UnitTests.checkDoubleArrayRel1(mm.amom,"mm.amom",mD.getdAMoments(),"mD.getdAMoments()",eps);
	    // VdP/dt
	    UnitTests.checkDoubleArrayRel1(mm.VdPdtmom,"mm.VdPdtmom",mD.getVdPMoments(),"mD.getVdPMoments()",eps);

	    // p*I
	    final double dpoffset=mm.p_offset-mD.getPOffset();
	    final double []
		i=mD.getdVMoments(),Pi=mD.getPdVMoments(),P2i=mD.getP2dVMoments(),P3i=mD.getP3dVMoments(),
		V=mD.getFMomentsFromDF(i),pV=mD.getFMomentsFromDF(Pi),p2V=mD.getFMomentsFromDF(P2i),p3V=mD.getFMomentsFromDF(P3i),
		Ti=mD.getTdVMoments(),TPi=mD.getTPdVMoments(),TP2i=mD.getTP2dVMoments(),TP3i=mD.getTP3dVMoments(),
		a=mD.getdAMoments(),pa=mD.getPdAMoments();

	    com.polytechnik.utils.Polynomials.shiftPOffsetInPlace(i.length,i,Pi,P2i,P3i,dpoffset);
	    com.polytechnik.utils.Polynomials.shiftPOffsetInPlace(V.length,V,pV,p2V,p3V,dpoffset);
	    com.polytechnik.utils.Polynomials.shiftPOffsetInPlace(Ti.length,Ti,TPi,TP2i,TP3i,dpoffset);
	    TMatr.sumP1(pa,a,-dpoffset);

	    UnitTests.checkDoubleArrayRel(mm.imom,"mm.imom",i,"mD:i",eps);
	    UnitTests.checkDoubleArrayRel(mm.pimom,"mm.pimom",Pi,"mD:Pi",eps);
	    UnitTests.checkDoubleArrayRel(mm.p2imom,"mm.p2imom",P2i,"mD:P2i",eps);
	    UnitTests.checkDoubleArrayRel(mm.p3imom,"mm.p3imom",P3i,"mD:P3i",eps);

	    UnitTests.checkDoubleArrayRel(mm.amom,"mm.amom",a,"mD:a",eps);
	    UnitTests.checkDoubleArrayRel(mm.pamom,"mm.pamom",pa,"mD:pa",eps);

	    UnitTests.checkDoubleArrayRel(mm.Vmom,"mm.Vmom",V,"mD:V",eps);
	    UnitTests.checkDoubleArrayRel(mm.pVmom,"mm.pVmom",pV,"mD:pV",eps);
	    UnitTests.checkDoubleArrayRel(mm.p2Vmom,"mm.p2Vmom",p2V,"mD:p2V",eps);
	    UnitTests.checkDoubleArrayRel(mm.p3Vmom,"mm.p3Vmom",p3V,"mD:p3V",eps);

	    UnitTests.checkDoubleArrayRel(mm.Timom,"mm.Timom",Ti,"mD:Ti",eps);
	    UnitTests.checkDoubleArrayRel(mm.Tpimom,"mm.Tpimom",TPi,"mD:TPi",eps);
	    UnitTests.checkDoubleArrayRel(mm.Tp2imom,"mm.Tp2imom",TP2i,"mD:TP2i",eps);
	    UnitTests.checkDoubleArrayRel(mm.Tp3imom,"mm.Tp3imom",TP3i,"mD:TP3i",eps);
	}

	System.err.println("test_CalculateMoments_matchSMomentsData(): test OK eps="+eps);
    }

}
