package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2009
 * This file is under GPL version 3.
 *
 */

import java.io.StreamTokenizer;
import com.polytechnik.utils.ReadWriteUtils;
import com.polytechnik.realtimebook.TraderInterface;

/** Configuration reader.
 *  @version $Id: ConfigurationReader.java,v 1.33 2022/07/28 21:03:02 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class ConfigurationReader {


    static class CommonParameters {
	final String INPUT_DIR,OUTPUT_DIR;
	final int NEQ_T;
	final java.util.HashMap<String,ModelParameters.ModelParametersData> data=
	    new java.util.HashMap<String,ModelParameters.ModelParametersData>();
	CommonParameters(final String INPUT_DIR,
			 final String OUTPUT_DIR,
			 final int NEQ_T)
	{
	    this.INPUT_DIR=INPUT_DIR;
	    this.OUTPUT_DIR=OUTPUT_DIR;
	    this.NEQ_T=NEQ_T;
	}
    }


    static CommonParameters getCommonParams(final StreamTokenizer tk)
	throws java.io.IOException
    {
	final String [] to_read=new String [20];

	String in=null,out=null;
	int neq_t=-1;
	for(;tk.ttype!=StreamTokenizer.TT_EOF;)
	{
	    final int n=ReadWriteUtils.readOneLine(tk,to_read);
	    if(n<=0)
	    {
		continue;
	    }
	    final String fchar=to_read[0];
	    if(fchar.startsWith("#"))
	    {
		continue;
	    }
	    if(fchar.equals("!INSTRUMENT!"))
	    {
		break;
	    }
	    if(n<2)
	    {
		throw new RuntimeException("No value for "+fchar);
	    }
	    final String val=to_read[1];
	    if(fchar.equals("NEQ_T"))
	    {
		neq_t=Integer.parseInt(val);
	    }
	    else if(fchar.equals("INPUT_DIR"))
	    {
		in=val;
	    }
	    else if(fchar.equals("OUTPUT_DIR"))
	    {
		final String alt_out=System.getProperty("tr.OUTPUT_DIR");
		if(alt_out!=null)
		{
		    out=alt_out;
		}
		else
		{
		    out=val;
		}
	    }
	    else
	    {
		throw new RuntimeException("Unknown value "+fchar);
	    }
	}
	if(in!=null && out!=null && neq_t>0)
	{
	    return new CommonParameters(in,out,neq_t);
	}
	throw new RuntimeException("Some values are not set in config: INPUT_DIR="+in+" OUTPUT_DIR="+out+" NEQ_t="+neq_t);
    }

    static final int N_COLUMNS=3;

    static void checkHeader(final int n,final String [] d)
    {
	if(n!=N_COLUMNS)
	{
	    throw new RuntimeException(
		"Invalid header "+
		java.util.Arrays.toString(
		    java.util.Arrays.copyOf(d,n)));
	}
	if(!d[0].equals("TICKER")||
	   !d[1].equals("SHARES")||
	   !d[2].equals("LEV=THR"))
	{
	    throw new RuntimeException(
		"Invalid header "+
		java.util.Arrays.toString(
		    java.util.Arrays.copyOf(d,n)));
	}
    }

    static void readOneTicker(final CommonParameters c,
			      final int n,
			      final String [] d)
    {
	if(n!=N_COLUMNS)
	{
	    throw new RuntimeException(
		"Line "+
		java.util.Arrays.toString(
		    java.util.Arrays.copyOf(d,n)));
	}
	final String ticker=d[0],shares=d[1],levs=d[2];
	final ModelParameters.ModelParametersData dold=c.data.get(ticker);
	if(dold!=null)
	{
	    throw new RuntimeException("Duplicated ticker"+ticker);
	}
	final int s=Integer.parseInt(shares);
	final String [] p=com.polytechnik.utils.ReadWriteUtils.extractnames(levs,";");
	if(p.length<=0)
	{
	    throw new RuntimeException("invalid levs="+levs);
	}

	double thr_enter=Double.NaN,thr_exit=Double.NaN,
	    thrw_enter=Double.NaN,thrw_exit=Double.NaN;
	long tau=-1;
	for(String sp:p)
	{
	    final String [] pl=com.polytechnik.utils.ReadWriteUtils.extractnames(sp,"=\t\r\n ");
	    if(pl.length!=2)
	    {
		throw new RuntimeException("invalid levs="+levs);
	    }
	    final String nm=pl[0],v=pl[1];
	    //System.err.println("nm="+nm+" v="+v);
	    if(nm.equals("thr_enter"))
	    {
		thr_enter=Double.parseDouble(v);
	    }
	    else if(nm.equals("thr_exit"))
	    {
		thr_exit=Double.parseDouble(v);
	    }
	    else if(nm.equals("thrw_enter"))
	    {
		thrw_enter=Double.parseDouble(v);
	    }
	    else if(nm.equals("thrw_exit"))
	    {
		thrw_exit=Double.parseDouble(v);
	    }
	    else if(nm.equals("tau"))
	    {
		tau=TraderInterface.TIME_SCALE*Long.parseLong(v);
	    }
	    else
	    {
		throw new RuntimeException("invalid levs: "+levs+" nm="+nm);
	    }
	}
	if(Double.isNaN(thr_enter)
	   || Double.isNaN(thr_exit)
	   || Double.isNaN(thrw_enter)
	   || Double.isNaN(thrw_exit)
	   || tau<=0)
	{
	    throw new RuntimeException(
		"invalid input="+levs+
		" some of levels are not set:"+
		" thr_enter="+thr_enter+
		" thr_exit="+thr_exit+
		" thrw_enter="+thrw_enter+
		" thrw_exit="+thrw_exit+
		" tau="+tau);
	}


	c.data.put(ticker,new ModelParameters.ModelParametersData(c.INPUT_DIR,
								  c.OUTPUT_DIR,
								  true,
								  s,
								  s,
								  ticker,
								  thr_enter,
								  thr_exit,
								  thrw_enter,
								  thrw_exit,
								  tau,
								  c.NEQ_T));
    }

    static void readStockInfo(final CommonParameters c,
			      final StreamTokenizer tk)
	throws java.io.IOException
    {
	boolean header_not_checked=true;
	final String [] to_read=new String [20];

	for(;tk.ttype!=StreamTokenizer.TT_EOF;)
	{
	    final int n=ReadWriteUtils.readOneLine(tk,to_read);
	    if(n<=0)
	    {
		continue;
	    }
	    final String fchar=to_read[0];
	    if(fchar.startsWith("#"))
	    {
		continue;
	    }
	    if(header_not_checked)
	    {
		checkHeader(n,to_read);
		header_not_checked=false;
		continue;
	    }
	    readOneTicker(c,n,to_read);
	}
    }





    static CommonParameters readConfig(final String configfile)
	throws java.io.IOException
    {
	final java.io.BufferedReader br=
	    new java.io.BufferedReader(new java.io.FileReader(configfile));
	final StreamTokenizer tk=ReadWriteUtils.getTokenizer(",",br);
	tk.quoteChar('\"');
	final CommonParameters c=getCommonParams(tk);
	readStockInfo(c,tk);
	br.close();
	return c;
    }
}
