package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** d^2/dt^2 P is back. Consider the |P| operator modified in a way
 * it has the same eigenvalues as the |I| operator, |P|=sum {@literal |psik>P_k<psik|}. Then consider
 * the second variation (the first one is zero) in the state of max I, the |psiH>.
 * @version $Id: DIselD2P.java,v 1.43 2022/08/08 20:10:45 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselD2P {
    public final EVXData I;
    public final double p_offset,
    // The https://arxiv.org/abs/1510.05510 answers.
	p_IH,var2p_IH,
    // new answer
	dirpI_plus,dirpI_minus,dirpV_plus,dirpV_minus
	;
    /** For debugging and unit tests.
     */
    public final double [] QQp_diagI;

    /** Constructor.
     * @param eps Run a selftest with this eps of postive.
     */
public DIselD2P(final int n,
		final double p_offset,
		final double [] QQt,final double [] QQi,
		final double [] QQpm,final double [] QQm,
		final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		final VolMatrix VOLATILITY_EDPsi,
		final double eps)
{
    this.p_offset=p_offset;
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] bQQpm=EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev),
	bQQm=EVXData.convertMatrFromQtoB_Basis(n,QQm,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),UNIT=TMatr.getUnitMatrix(n),
	bQQp_diagI=new double [n*n];
    // Pure Dynamic Impact Approximation: P to be diagonal in the basis of diagonal I.
    for(int k=0;k<n;k++) bQQp_diagI[k+n*k]=bQQpm[k+n*k]/bQQm[k+n*k];
    this.p_IH=bQQp_diagI[ixH+n*ixH];
    this.QQp_diagI=EVXData.convertMatrFromBtoQ_Basis(n,bQQp_diagI,I.ev,QQt);

    final double [] d2I=new double [n],d2p=new double [n],te=new double [n],ONE=M.B.getONE();
    for(int k=0;k<n;k++){
	final double [] psik=I.getPsi(k),bDpsik=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(psik),I.ev,QQt);
	bDpsik[k]=0; // remove psik projection, now bpsik and bDpsik are orthogonal 2D eigenvalues solution
	final double norm2bDpsik=TMatr.mulVV(bDpsik,bDpsik);
	// changes in p and I, the scale is norm2bDpsik.
	d2I[k]=TMatr.getAMatrA(bDpsik,bQQi,n)/norm2bDpsik-bQQi[k+n*k]; // bQQi is diagonal
	d2p[k]=TMatr.getAMatrA(bDpsik,bQQp_diagI,n)/norm2bDpsik-bQQp_diagI[k+n*k]; // bQQp_diagI is diagonal
	// Lebesgue quadrature weight
	te[k]=Math.pow(TMatr.getAMatrB(psik,ONE,QQt,n),2);
	if(eps>0){
	    final double [] bpsik=new double [n],_bDpsik=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(psik),I.ev,QQt);
	    bpsik[k]=1;
	    final double [] rqPk=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(bpsik,_bDpsik,null,bQQp_diagI,UNIT,n),
		rqIk=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(bpsik,_bDpsik,null,bQQi,UNIT,n);
	    UnitTests.checkDoubleRel(rqIk[0],"rqIk[0]",I.ev[k],"I.ev[k]",eps);
	    UnitTests.checkDoubleRel(rqIk[1],"rqIk[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(rqIk[2],"rqIk[2]",2*d2I[k]*norm2bDpsik,"2*d2I[k]*norm2bDpsik",eps);
	    UnitTests.checkDoubleRel(rqPk[0],"rqPk[0]",bQQp_diagI[k*n+k],"bQQp_diagI[k*n+k]",eps);
	    UnitTests.checkDoubleRel(rqPk[1],"rqPk[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(rqPk[2],"rqPk[2]",2*d2p[k]*norm2bDpsik,"2*d2p[k]*norm2bDpsik",eps);
	    UnitTests.checkDoubleRel(2*TMatr.getAMatrB(bpsik,_bDpsik,UNIT,n),"2*<psi|D(psi)>",Math.pow(I.psik_at_x0[k],2),"psik^2(x0)",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(bpsik,bDpsik,UNIT,n),"<psi|Dpsi_ort>",0,"0",eps);
	    UnitTests.checkDoubleArrayRel1(bpsik,"bpsik",EVXData.convertPsiFromQtoB_Basis(n,psik,I.ev,QQt),"bpsik",eps);
	}
    }
    if(eps>0){
	UnitTests.checkDoubleRel(TMatr.sumElements(te),"sum(te)",TMatr.getAMatrA(ONE,QQt,n),"<1>",eps);
    }
    double dirpI_plus=0,dirpI_minus=0,dirpV_plus=0,dirpV_minus=0;
    for(int k=0;k<n;k++){
	if(d2I[k]<0){
	    // negative second derivative is a max of I
	    dirpV_plus+=d2I[k]*d2p[k]*te[k];
	    dirpI_plus+=d2I[k]*d2p[k];
	} else {
	    // positive second derivative is a min of I
	    dirpV_minus+=d2I[k]*d2p[k]*te[k];
	    dirpI_minus+=d2I[k]*d2p[k];
	}
    }
    this.dirpI_plus=dirpI_plus;
    this.dirpV_plus=dirpV_plus;
    this.dirpI_minus=dirpI_minus;
    this.dirpV_minus=dirpV_minus;
    this.var2p_IH=d2p[ixH];


    // For debugging and unit tests
    if(true)
    {
	final double [] bDpsi=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(I.psiH),I.ev,QQt),bpsi=new double [n];bpsi[ixH]=1;
	final double [] rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(bpsi,bDpsi,null,bQQp_diagI,UNIT,n);
	final double [] bDpsi_plus=new double [n],bDpsi_minus=new double [n];
	double w_plus=0,w_minus=0,w_0=0,lw_plus=0,lw_minus=0;
	for(int k=0;k<n;k++){
	    final double DpsiH_psik=bDpsi[k],w=Math.pow(DpsiH_psik,2),
		lam_P=bQQp_diagI[k+n*k]-bQQp_diagI[ixH+n*ixH];
	    if(k==ixH){w_0+=w;}
	    else if(lam_P>0) {w_plus+=w;lw_plus+=w*lam_P;bDpsi_plus[k]+=DpsiH_psik;}
	    else {w_minus+=w;lw_minus+=w*lam_P;bDpsi_minus[k]+=DpsiH_psik;}
	}
	if(eps>0){
	    final double _p_IH=TMatr.getAMatrA(I.psiH,QQpm,n)/TMatr.getAMatrA(I.psiH,QQm,n);
	    UnitTests.checkDoubleRel(p_IH,"p_IH",_p_IH,"_p_IH",eps);
	    final double [] Dpsi=VOLATILITY_EDPsi.getDPsi(I.psiH),basispt=I.ev.clone(),
		rrq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(I.psiH,Dpsi,null,QQp_diagI,QQt,n);
	    for(int k=0;k<n;k++) basispt[k]=I.getFPsi(k,QQpm)/I.getFPsi(k,QQm);
	    EVXData.test_orthogonality("QQp_diagI",n,basispt,QQp_diagI,QQt,eps);
	    UnitTests.checkDoubleRel(rq[0],"rq[0]",p_IH,"p_IH",eps);
	    UnitTests.checkDoubleRel(rq[1],"rq[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(rq[2],"rq[2]",2*(lw_plus+lw_minus),"2*(lw_plus+lw_minus)",eps);
	    UnitTests.checkDoubleRel(w_0+w_plus+w_minus,"w_0+w_plus+w_minus",TMatr.getAMatrA(Dpsi,QQt,n),"<Dpsi|Dpsi>",eps);
	    UnitTests.checkDoubleRel(rrq[0],"rrq[0]",p_IH,"p_IH",eps);
	    UnitTests.checkDoubleRel(rrq[1],"rrq[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(rrq[2],"rrq[2]",rq[2],"rq[2]",eps);

	    UnitTests.checkDoubleArrayRel1(
		EVXData.convertPsiFromBtoQ_Basis(n,DeprecatedHelperTMatr.sumAB(n,bDpsi_plus,bDpsi_minus),I.ev),"bDpsi_plus+bDpsi_minus",
		TMatr.sumAcB(Dpsi,-TMatr.getAMatrB(Dpsi,I.psiH,QQt,n),I.psiH),"|Dpsi_ort>",eps);
	}
    }



    // Different type of variation, divide psi_{ILow} by x-xk, like Mu2DynMicroImpact.java
    // Find the roots of |psi>.
    // For the |psi>, corresponding to min/max all the roots are simple, real, distinct.
    final double [] psi=I.psiL,x=M.getXNodes(psi).re_root,wmu=new double [x.length],dpmu=new double [x.length],t=new double [x.length],bQQpi_diagI=TMatr.mulMM(n,bQQp_diagI,bQQi);
    final double lam=I.sL;
    // Build Gauss quadrature for the measure dmu=dv-lam*dt.
    for(int k=0;k<x.length;k++)
    {
	final double xk=x[k];
	final double [] l=M.B.sdiv1(psi,xk).q;
	final double nrmlk=M.B.getD0(l,xk);
	TMatr.scaleVectorInPlace(l,1/nrmlk);
	final double [] bl=EVXData.convertPsiFromQtoB_Basis(n,java.util.Arrays.copyOf(l,n),I.ev,QQt);
	final double m0i=TMatr.getAMatrA(bl,bQQi,n),m0t=TMatr.getAMatrA(bl,UNIT,n);
	t[k]=m0t;
	wmu[k]=m0i-lam*m0t;
	dpmu[k]=TMatr.getAMatrA(bl,bQQpi_diagI,n)-lam*TMatr.getAMatrA(bl,bQQp_diagI,n);
    }
    final double sumdpmu=TMatr.sumElements(dpmu);
    // Check the quadrature is built OK.
    if(eps>0)
    {
	final double [] o=new double []{1},ox=M.B.getaxbP(o,1,0);
	final double [] QQmu=TMatr.sumAcB(QQi,-lam,QQt),mu=wmu;
	UnitTests.checkDoubleRel(TMatr.sumElements(mu),"sum(mu)",TMatr.getAMatrA(o,QQmu,n),"QQmu[0]",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(x,mu),"sum(x*mu)",TMatr.getAMatrB(o,ox,QQmu,n),"XQQmu[0]",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(x,mu),"sum(x^2*mu)",TMatr.getAMatrA(ox,QQmu,n),"X^2QQmu[0]",eps);
	System.err.println("DIselD2P: invariant internal test OK n="+n+" lam="+lam+" eps="+eps);
    }


    if(eps>0){
	System.err.println("DIselD2P: test OK n="+n+" eps="+eps);
    }
}
//final double w_plus,w_minus,w_0,lw_plus,lw_minus,sumdpmu,QQd_plus[],QQd_minus[];


}
