package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** d^2/dt^2 P is back. Consider the |P| operator modified in a way
 * it has the same eigenvalues as the |I| operator, |P|=sum {@literal |psik>P_k<psik|}. Then consider
 * the second variation (the first one is zero) in the state of max I, the |psiH>.
 * Similar to DIselD2PstateIH.java
 * @version $Id: DIselD2PstateIH.java,v 1.12 2022/08/08 20:10:46 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselD2PstateIH {
    public final EVXData I;
    public final double p_offset,
    // The https://arxiv.org/abs/1510.05510 answers.
	p_IH,
    // new answer
	var2p
	;

    /** For debugging and unit tests.
     */
    public final double [] QQp_diagI;

public DIselD2PstateIH(final int n,
		       final double p_offset,
		       final double [] QQt,final double [] QQi,
		       final double [] QQpm,final double [] QQm,
		       final boolean flag_pure_dynamic_impact,
		       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		       final VolMatrix VOLATILITY_EDPsi,
		final double eps)
{
    this.p_offset=p_offset;
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double []
	bDpsi=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(I.psiH),I.ev,QQt),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),UNIT=TMatr.getUnitMatrix(n),
	bQQp_diagI,bpsi=new double [n];
    bpsi[ixH]=1;
    if(flag_pure_dynamic_impact) {
	bQQp_diagI=new double [n*n];
	final double [] bQQpm=EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev),
	    bQQm=EVXData.convertMatrFromQtoB_Basis(n,QQm,I.ev);
	for(int k=0;k<n;k++) bQQp_diagI[k+n*k]=bQQpm[k+n*k]/bQQm[k+n*k];
    } else {
	if(QQm==QQt){
	    bQQp_diagI=EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev);
	} else if(QQm==QQi) {
	    final double [] invsqrtI=new double [n*n];
	    for(int k=0;k<n;k++) invsqrtI[k+n*k]=1/Math.sqrt(I.ev[k]);
	    bQQp_diagI=TMatr.mulMMM(n,invsqrtI,EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev),invsqrtI);
	} else throw new IllegalArgumentException("for flag_pure_dynamic_impact="+flag_pure_dynamic_impact+" only QQ=QQi or QQm=QQt are supported");
	for(int k=0;k<n;k++) if(k!=ixH) bQQp_diagI[k+ixH*n]=bQQp_diagI[ixH+k*n]=0;
    }

    this.p_IH=TMatr.getAMatrA(bpsi,bQQp_diagI,n);
    final double [] rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(bpsi,bDpsi,null,bQQp_diagI,UNIT,n);
    this.var2p=rq[2];

    // for debugging
    this.QQp_diagI=EVXData.convertMatrFromBtoQ_Basis(n,bQQp_diagI,I.ev,QQt);

    if(eps>0){
	final double _p_IH=TMatr.getAMatrA(I.psiH,QQpm,n)/TMatr.getAMatrA(I.psiH,QQm,n);
	UnitTests.checkDoubleRel(p_IH,"p_IH",_p_IH,"_p_IH",eps);
	final double [] Dpsi=VOLATILITY_EDPsi.getDPsi(I.psiH),
	    rrq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(I.psiH,Dpsi,null,QQp_diagI,QQt,n);
	if(flag_pure_dynamic_impact){
	    final double [] basispt=I.ev.clone();
	    for(int k=0;k<n;k++) basispt[k]=I.getFPsi(k,QQpm)/I.getFPsi(k,QQm);
	    EVXData.test_orthogonality("QQp_diagI",n,basispt,QQp_diagI,QQt,eps);
	} else {
	    for(int k=0;k<n;k++){
		final double pk=k==ixH?p_IH:0;
		UnitTests.checkDoubleRel(TMatr.getAMatrB(n,I.psiH,0,I.ev,n+k*n,QQp_diagI,n),"<psiH|P|psik>",pk,"pk",eps);
	    }
	}
	UnitTests.checkDoubleRel(rq[0],"rq[0]",p_IH,"p_IH",eps);
	UnitTests.checkDoubleRel(rq[1],"rq[1]",0,"0",eps);
	//UnitTests.checkDoubleRel(rq[2],"rq[2]",2*(lw_plus+lw_minus),"2*(lw_plus+lw_minus)",eps);
	//UnitTests.checkDoubleRel(w_0+w_plus+w_minus,"w_0+w_plus+w_minus",TMatr.getAMatrA(Dpsi,QQt,n),"<Dpsi|Dpsi>",eps);
	UnitTests.checkDoubleRel(rrq[0],"rrq[0]",p_IH,"p_IH",eps);
	UnitTests.checkDoubleRel(rrq[1],"rrq[1]",0,"0",eps);
	UnitTests.checkDoubleRel(rrq[2],"rrq[2]",rq[2],"rq[2]",eps);

    }


    if(eps>0){
	System.err.println("DIselD2PstateIH: test OK n="+n+" eps="+eps);
    }
}


}
