package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.LebesgueQuadratureWithEVData;

/** Selected dI/dt states.
 * The main idea of DIselDM is to "split" the {@literal |psiH>} state (max I) to
 * the two states with {positive,negative} dI/dt,
 * note that {@literal <psiH|dI/dt|psiH>}=0, for this reason such two states do always exist.
 * The 2n-2 degree polynomial Weight(x)=2*psiH(x)*(D(psiH(x))- {@literal <D(psiH)|psiH>}psiH(x))
 * corresponds to dI/dt in psiH state with the I^f=I^{[IH]} boundary condition,
 * the relations {@literal <Weight>=0} ; {@literal <I*Weight>}=0 are held because the psiH is an eigenvector of I.
 * Then use https://arxiv.org/abs/1807.06007 appendix A theory
 * to build an operator isomorphic to Weight(x)=sum_i lambda^{[i]} (psi^{[i]}(x))^2.
 * Find the operator positive/negative eigenvalues and corresponding eigenvectors
 * to build the density matrix.
 * @version $Id: DIselDM.java,v 1.41 2022/08/07 15:44:01 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselDM {
public final EVXData I;
public final double p_offset,
// The https://arxiv.org/abs/1510.05510 answers.
    p_IH,dpdtvari_IH
    ;

/** Other answers, correspond to various density matrix tries.
 */
public final DMstateAnswer sIH,sDM_diplus,sDM_diminus,sDIH_diplus,sDIH_diminus;

DIselDM(final int n,final double p_offset,final double p_last,
	final double [] QQt,final double [] QQi,
	final double [] QQpi,
	final double [] QQdpdt,
	final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	final VolMatrix VOLATILITY_EDPsi,
	final double eps)
{
    this.p_offset=p_offset;
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];

    final double [] psiH=I.psiH,DpsiH=VOLATILITY_EDPsi.getDPsi(psiH);
    if(eps>0){
	// check infinitesimal operator
	UnitTests.checkDoubleRel(2*TMatr.getAMatrB(psiH,DpsiH,QQt,n),"2*<psiH|DpsiH>",Math.pow(I.psik_at_x0[ixH],2),"psiH^2(x0)",eps);
    }
    // Remove |psiH> component from |DpsiH>
    TMatr.sumP1(DpsiH,psiH,-(true?0.5*Math.pow(I.psik_at_x0[ixH],2):TMatr.getAMatrB(psiH,DpsiH,QQt,n)));
    // The old result: https://arxiv.org/abs/1709.06759 , formulae (31): Delta P in |psiH> state.
    this.dpdtvari_IH=-2*I.getFPsiPsi(ixH,DpsiH,QQpi)/I.ev[ixH];

    // The new answers
    final double [] Weight=TMatr.scaleVectorInPlace(M.B.mult2Pol(psiH,DpsiH),2);
    if(Weight.length!=2*n-1) throw new IllegalArgumentException();
    final double [] UNIT=TMatr.getUnitMatrix(n),
	// Gram matrix inverse can be obtained without solving a linear system.
	IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT);
	//IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt);
    this.DensityMatrix=M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,Weight,IQQt));
    final double []
	// Convert QQpi, QQdpdt, and etc. to I.ev basis to avoid IQQt=G^{-1} required for a non-orthogonal basis.
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	bQQdpdt=EVXData.convertMatrFromQtoB_Basis(n,QQdpdt,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),
	bDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,DensityMatrix,I.ev);
    /* Split the |DensityMatrix| state on two, corresponding to positive and negative dI/dt.
       In contrast with DIselIH.java, here, we do this manipulating the density matrix directly.
    */
    // The calculations in the basis of I=dV/dt operator.
    final double [] ev_bDensityMatrix=M.EV.getEV(n,bDensityMatrix,true),
	bbDM_diplus=new double [n*n],bbDM_diminus=new double [n*n];
    for(int k=0;k<n;k++) {
	final double ev_dm=ev_bDensityMatrix[k];
	if(ev_dm>0){
	    bbDM_diplus[k+n*k]=ev_dm;
	} else {
	    bbDM_diminus[k+n*k]=ev_dm;
	}
    }
    final double [] bDM_diplus=EVXData.convertMatrFromBtoQ_Basis(n,bbDM_diplus,ev_bDensityMatrix,UNIT),
	bDM_diminus=EVXData.convertMatrFromBtoQ_Basis(n,bbDM_diminus,ev_bDensityMatrix,UNIT);
    // OK, now the bDM_{diplus,diminus} are obtained in the basis of I=dV/dt eigenvectors.

    final double [] bpsiH=new double [n];bpsiH[ixH]=1;
    // Old https://arxiv.org/abs/1510.05510 answer, with the density matrix |psiH><psiH|
    sIH=DMstateAnswer.mixedState(n,TMatr.getQQaa(bpsiH),-1,bQQi,bQQpi,bQQdpdt,bpsiH);
    // Two new answers with the density matrix.
    sDM_diplus=DMstateAnswer.mixedState(n,bDM_diplus,-1,bQQi,bQQpi,bQQdpdt,bpsiH);
    sDM_diminus=DMstateAnswer.mixedState(n,bDM_diminus,-1,bQQi,bQQpi,bQQdpdt,bpsiH);
    sDIH_diplus=DMstateAnswer.pureState(n,EVXData.getPsiFromBasis(n-1,ev_bDensityMatrix,n),UNIT,bQQi,bQQpi,bQQdpdt,bpsiH);
    sDIH_diminus=DMstateAnswer.pureState(n,EVXData.getPsiFromBasis(0,ev_bDensityMatrix,n),UNIT,bQQi,bQQpi,bQQdpdt,bpsiH);




    // For illustration purposes obtain the same density matrices in the original Q-basis.
    DM_diplus=EVXData.convertMatrFromBtoQ_Basis(n,bDM_diplus,I.ev,QQt);
    DM_diminus=EVXData.convertMatrFromBtoQ_Basis(n,bDM_diminus,I.ev,QQt);
    DIH_diplus=EVXData.convertMatrFromBtoQ_Basis(n,TMatr.getQQaa(EVXData.getPsiFromBasis(n-1,ev_bDensityMatrix,n)),I.ev,QQt);
    DIH_diminus=EVXData.convertMatrFromBtoQ_Basis(n,TMatr.getQQaa(EVXData.getPsiFromBasis(0,ev_bDensityMatrix,n)),I.ev,QQt);
    // Some selftests folow
    if(eps>0){
	UnitTests.checkDoubleArrayRel1(TMatr.sumAcB(DM_diplus,1,DM_diminus),"DM_diplus+DM_diminus",DensityMatrix,"DensityMatrix",eps);
    }
    if(eps>0){
	// check the eigenvector condition
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bDensityMatrix,bQQi),"spur(DM*bQQi)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bDensityMatrix,UNIT),"spur(DM)",0,"0",eps);
	UnitTests.checkDoubleRel(2*TMatr.getAMatrB(psiH,DpsiH,QQi,n),"<psiH|I|DpsiH>",0,"0",eps);
	UnitTests.checkDoubleRel(2*TMatr.getAMatrB(psiH,DpsiH,QQt,n),"<psiH|DpsiH>",0,"0",eps);
	// Check wavefunction <-> density matrix match
	for(double [] QQ:new double [][]{QQpi,QQdpdt}){
	    final double h_psi=2*TMatr.getAMatrB(psiH,DpsiH,QQ,n);
	    final double [] bQQ=EVXData.convertMatrFromQtoB_Basis(n,QQ,I.ev);
	    final double h_dm=TMatr.getSpurMM(n,TMatr.sumAcB(bDM_diplus,1,bDM_diminus),bQQ);
	    UnitTests.checkDoubleRel(h_psi,"h_psi",h_dm,"h_dm",eps);
	}
    }
    if(eps>0){
	// Match the Lebesgue quadrature.
	final LebesgueQuadratureWithEVData DM=new LebesgueQuadratureWithEVData(n,DensityMatrix,QQt,M);
	final double [] _Weight=new double [2*n-1],dp_in_DM=DM.getLambdaMatrixDiagonalsInPsiStates(QQdpdt);
	double dp_diplus=0,dp_diminus=0;
	for(int k=0;k<n;k++){
	    final double [] psi=EVXData.getPsiFromBasis(k,DM.ev,n);
	    TMatr.sumP1(_Weight,M.B.mult2Pol(psi,psi),DM.f[k]);
	    if(DM.f[k]>0) dp_diplus+=dp_in_DM[k]; else dp_diminus+=dp_in_DM[k];
	}
	UnitTests.checkDoubleArrayRel1(_Weight,"_Weight",Weight,"Weight",eps);
	UnitTests.checkDoubleRel(DM.sumLambdaMatrixDiagonalsInPsiStates(QQi),"spur(DM*QQi)",0,"0",eps);
	UnitTests.checkDoubleRel(DM.sumLambdaMatrixDiagonalsInPsiStates(QQt),"spur(DM*QQt)",0,"0",eps);
	UnitTests.checkDoubleRel(dp_diplus,"dp_diplus",TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpdt,IQQt,DM_diplus})),"spur",eps);
	UnitTests.checkDoubleRel(dp_diminus,"dp_diminus",TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpdt,IQQt,DM_diminus})),"spur",eps);
	UnitTests.checkDoubleRel(dp_diplus,"dp_diplus",sDM_diplus.dpdt,"sDM_diplus.dpdt",eps);
	UnitTests.checkDoubleRel(dp_diminus,"dp_diminus",sDM_diminus.dpdt,"sDM_diminus.dpdt",eps);

	// Match the old result: https://arxiv.org/abs/1709.06759 , formulae (31): Delta P=-rq[1]/I.sH
	final double [] rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(psiH,VOLATILITY_EDPsi.getDPsi(psiH),null,QQpi,QQt,n);
	UnitTests.checkDoubleRel(-rq[1],"-rq[1]",dpdtvari_IH*I.sH,"dpdtvari_IH*I.sH",eps);
	UnitTests.checkDoubleRel(rq[1],"rq[1]",TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQpi,IQQt,DM_diplus}))+TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQpi,IQQt,DM_diminus})),"spur(bQQpi*bDM_diplus)+spur(bQQpi*bDM_diminus)",eps);
	UnitTests.checkDoubleRel(rq[1],"rq[1]",TMatr.getSpurMM(n,bDM_diplus,bQQpi)+TMatr.getSpurMM(n,bDM_diminus,bQQpi),"spur(bDM_diplus*bQQpi)+spur(bDM_diminus*bQQpi)",eps);
	// The I.sH * Delta P = spur(bDM_diplus*bQQpi)+spur(bDM_diminus*bQQpi) matched OK.

	System.err.println("DIselDM: spur <psiH|dI/dt|psiH> matches n="+n+" I="+java.util.Arrays.toString(I.getEValues())+" ev_bDensityMatrix="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_bDensityMatrix,n))+" DM.f="+java.util.Arrays.toString(DM.f)+" eps="+eps);
    }
    if(eps>0){
	// test slightly different version
	// final DIselDMs diselDMs=new DIselDMs(n,p_offset,p_last,QQt,QQi,QQpt,QQt_norm,QQpi,QQi_norm,QQdpdt,M,VOLATILITY_EDPsi,eps);
	// test the diselDMs match.
    }



}
    final double [] DensityMatrix,DM_diplus,DM_diminus,DIH_diplus,DIH_diminus;

}
