package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Selected basis, the states of positive/negative QQDEVseparation eigenvalues,
 * then solve I-> max using selected basis.
 * @version $Id: DIselIH.java,v 1.100 2022/08/05 11:25:31 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselIH {
public final EVXData I;
public final double p_offset,
// The https://arxiv.org/abs/1510.05510 answers.
    p_IH,dpdt_IH,dEVseparation_IH,
// New answers
    wpsiH_projection_simple_plus,
    wpsiH_projection_simple_minus
    ;

/** The answers, corresponding to rho^+ and rho^-, the wpsiH_projection
 * field is the projection of interest.
 */
public final DMstateAnswer sIH,sIH_plus,sIH_minus;

public static final int METHOD_FULL_EV_QQDEV=711,METHOD_2EV_QQDEV=712,METHOD_2EV_QQ0=713;

/**
 * @param n The dimension.
 * @param calculation_method one of METHOD_FULL_EV_QQDEV or METHOD_2EV_QQDEV.
 * @param QQDEVseparation eigenvalues separation matrix, the dS/dt.
 * The null value is special, if QQDEVseparation==null, then the -(p-p^{[IH]})I operator is used.
 * @param QQt {@literal <Q_j|Q_k>}.
 * @param QQi {@literal <Q_j|I|Q_k>}.
 * @param QQpi {@literal <Q_j|pI|Q_k>}.
 * @param QQdpdt {@literal <Q_j|dp/dt|Q_k>}.
 * @param M basis operations.
 * @param eps Run a selftest with this eps if &gt;0.
 */
public DIselIH(final int n,
	       final int calculation_method,
	       final double [] QQDEVseparation,
	       final double p_offset,final double p_last,
	       final double [] QQt,final double [] QQi,final double [] QQpi,
	       final double [] QQdpdt,
	       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	       final double eps)
{
    this.p_offset=p_offset;
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];
    this.dpdt_IH=I.getFPsi(ixH,QQdpdt);
    this.QQDEVseparation=QQDEVseparation!=null?QQDEVseparation:TMatr.sumaAbB(p_IH,QQi,-1,QQpi);
    this.dEVseparation_IH=I.getFPsi(ixH,this.QQDEVseparation);

    // The new answers
    final double []
	// Convert QQDEVseparation, QQdpdt, and etc. to I.ev basis to avoid IQQt=G^{-1} required for non-orthogonal basis.
	bQQDEVseparation=EVXData.convertMatrFromQtoB_Basis(n,this.QQDEVseparation,I.ev),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	bQQdpdt=EVXData.convertMatrFromQtoB_Basis(n,QQdpdt,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues());

    final double [] UNIT=TMatr.getUnitMatrix(n),bpsi_plus,bpsi_minus,bpsiH=new double [n];bpsiH[ixH]=1;
    if(calculation_method==METHOD_FULL_EV_QQDEV){
	// The calculations in the basis of I=dV/dt operator.
	double wpsiH_projection_simple_plus=0,wpsiH_projection_simple_minus=0;
	final double [] ev_bQQDEVseparation=M.EV.getEV(n,bQQDEVseparation,true);
	// Simplistic try (|psiH> projection).
	for(int k=0;k<n;k++){
	    final double w=Math.pow(DeprecatedHelperTMatr.mulVVo(n,ev_bQQDEVseparation,n+k*n,bpsiH,0),2);
	    if(ev_bQQDEVseparation[k]>dEVseparation_IH){wpsiH_projection_simple_plus+=w;}
	    else{wpsiH_projection_simple_minus+=w;}
	}
	this.wpsiH_projection_simple_plus=wpsiH_projection_simple_plus;
	this.wpsiH_projection_simple_minus=wpsiH_projection_simple_minus;
	// More generic approach
	final BasisSplit di_pm=new BasisSplit(n,ev_bQQDEVseparation,(ev) -> ev>dEVseparation_IH,bQQi,M.EV,eps);
	bpsi_plus=di_pm.sel_plus.psiH;
	bpsi_minus=di_pm.sel_minus.psiH;
	if(eps>0){
	    System.err.println("DIselIH: ev_bQQDEVseparation="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_bQQDEVseparation,n))+" eps="+eps);
	}
    } else if(calculation_method==METHOD_2EV_QQDEV||calculation_method==METHOD_2EV_QQ0){
	// dim=2 eigenvalues problem, the basis is |a>=|bpsiH> and |b>=|bQQDEVseparation|bpsiH> or |bpsi0>
	final double [] a=bpsiH,b=calculation_method==METHOD_2EV_QQ0?
	    TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o))
	    :
	    TMatr.getMatr_x_A(n,bQQDEVseparation,bpsiH);
	final int d=2;
	final double
	    a_a=1, // <bpsiH|bpsiH>
	    a_b=TMatr.mulVV(a,b),
	    b_b=TMatr.mulVV(b,b)
	;
	final double [] Sep2=new double []{TMatr.getAMatrA(a,bQQDEVseparation,n),
					   TMatr.getAMatrB(a,b,bQQDEVseparation,n),
					   TMatr.getAMatrB(b,a,bQQDEVseparation,n),
					   TMatr.getAMatrA(b,bQQDEVseparation,n)},
	    Mt=new double []{a_a,a_b,a_b,b_b},
	    ev_Sep2=M.EV.getGEV_Generalized(d,Sep2,Mt,true);
	bpsi_minus=TMatr.sumaAbB(ev_Sep2[d+0*d+0],a,ev_Sep2[d+0*d+1],b);
	bpsi_plus=TMatr.sumaAbB(ev_Sep2[d+1*d+0],a,ev_Sep2[d+1*d+1],b);
	this.wpsiH_projection_simple_plus=Math.pow(TMatr.mulVV(bpsiH,bpsi_plus),2);
	this.wpsiH_projection_simple_minus=Math.pow(TMatr.mulVV(bpsiH,bpsi_minus),2);
	if(eps>0){
	    // match the skewness
	    UnitTests.checkDoubleRel(wpsiH_projection_simple_minus-wpsiH_projection_simple_plus,"wpsiH_projection_simple_minus-wpsiH_projection_simple_plus",(2*dEVseparation_IH-ev_Sep2[0]-ev_Sep2[d-1])/(ev_Sep2[0]-ev_Sep2[d-1]),"Gamma skewness",eps);
	    System.err.println("DIselIH: d="+d+" ev_Sep2="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_Sep2,d))+" n="+n+" eps="+eps);
	}
    }
    else throw new IllegalArgumentException("calculation_method="+calculation_method);


    // Old https://arxiv.org/abs/1510.05510 answer, with the density matrix |psiH><psiH|
    sIH=DMstateAnswer.pureState(n,bpsiH,UNIT,bQQi,bQQpi,bQQdpdt,bpsiH);
    // Conditional I-> max solutions, corresponding to positive and negative eigenvalues of bQQDEVseparation.
    sIH_plus=DMstateAnswer.pureState(n,bpsi_plus,UNIT,bQQi,bQQpi,bQQdpdt,bpsiH);
    sIH_minus=DMstateAnswer.pureState(n,bpsi_minus,UNIT,bQQi,bQQpi,bQQdpdt,bpsiH);

    // For illustration purposes obtain the density matrices in the original Q-basis.
    DMIH_plus=EVXData.convertMatrFromBtoQ_Basis(n,TMatr.getQQaa(bpsi_plus),I.ev,QQt);
    DMIH_minus=EVXData.convertMatrFromBtoQ_Basis(n,TMatr.getQQaa(bpsi_minus),I.ev,QQt);


    if(eps>0){
	UnitTests.checkDoubleRel(wpsiH_projection_simple_plus+wpsiH_projection_simple_minus,"wpsiH_projection_simple_plus+wpsiH_projection_simple_minus",1,"1",eps);
	if(QQDEVseparation==null){
	    UnitTests.checkDoubleRel(dEVseparation_IH,"dEVseparation_IH",0,"0",eps);
	}
	System.err.println("DIselIH: OK n="+n+" I="+java.util.Arrays.toString(I.getEValues())+" QQDEVseparation[0]="+this.QQDEVseparation[0]+" dEVseparation_IH="+this.dEVseparation_IH+" calculation_method="+calculation_method+" eps="+eps);
    }


}

final double [] QQDEVseparation,DMIH_plus,DMIH_minus;

public static void matchDIselIH(final DIselIH a,final DIselIH b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.dpdt_IH,"a.dpdt_IH",b.dpdt_IH,"b.dpdt_IH",eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.wpsiH_projection_simple_plus,"a.wpsiH_projection_simple_plus",b.wpsiH_projection_simple_plus,"b.wpsiH_projection_simple_plus",eps);
    UnitTests.checkDoubleRel(a.wpsiH_projection_simple_minus,"a.wpsiH_projection_simple_minus",b.wpsiH_projection_simple_minus,"b.wpsiH_projection_simple_minus",eps);
    DMstateAnswer.matchDMstateAnswer(a.sIH,a.p_offset,b.sIH,b.p_offset,eps);
    DMstateAnswer.matchDMstateAnswer(a.sIH_plus,a.p_offset,b.sIH_plus,b.p_offset,eps);
    DMstateAnswer.matchDMstateAnswer(a.sIH_minus,a.p_offset,b.sIH_minus,b.p_offset,eps);
    System.err.println("DIselIH.matchDIselIH test OK n="+a.I.neq+" eps="+eps);
}

}
