package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.ProbabilityCorrelation;

/** Simplistic version of selected basis {@literal |psiH>} and {@literal |dS/dt|psiH>}.
 * @version $Id: DIselIH2PositiondI.java,v 1.20 2022/08/05 11:25:31 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselIH2PositiondI {
public final EVXData I;
public final double p_offset,
// The https://arxiv.org/abs/1510.05510 answers.
    p_IH
    ;
/** New answer
 */
public final ProbabilityCorrelation corr_Position_dI;

/**
 * @param n The dimension.
 * @param QQt {@literal <Q_j|Q_k>}.
 * @param QQi {@literal <Q_j|I|Q_k>}.
 * @param QQpi {@literal <Q_j|pI|Q_k>}.
 * @param M basis operations.
 * @param eps Run a selftest with this eps if &gt;0.
 */
public DIselIH2PositiondI(final int n,
		   final double p_offset,
		   final double [] QQt,final double [] QQi,final double [] QQpi,
		   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		   final VolMatrix VOLATILITY_EDPsi,
		   final double eps)
{
    this.p_offset=p_offset;
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	QQdidt=TMatr.sumaAbB(I.sH,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQi)),
	bQQdidt=EVXData.convertMatrFromQtoB_Basis(n,QQdidt,I.ev),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),UNIT=TMatr.getUnitMatrix(n),
	bQQposition=TMatr.sumaAbB(p_IH,bQQi,-1,bQQpi);
    final double [] bpsiH=new double [n];bpsiH[ixH]=1;
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQi,n),"<psiH|I|psiH>",I.sH,"I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQposition,n),"<psiH|bQQposition|psiH>",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQdidt,n),"<psiH|bQQdidt|psiH>",0,"0",eps);
    }
    final double [] a=bpsiH,b=TMatr.getMatr_x_A(n,bQQdidt,bpsiH),
	Mt=new double [] {
	    TMatr.getAMatrB(a,a,UNIT,n), //=1
	    TMatr.getAMatrB(a,b,UNIT,n), //=0
	    TMatr.getAMatrB(b,a,UNIT,n), //=0
	    TMatr.getAMatrB(b,b,UNIT,n)
	},
	Mposition=new double []{
	    TMatr.getAMatrB(a,a,bQQposition,n), //=0
	    TMatr.getAMatrB(a,b,bQQposition,n),
	    TMatr.getAMatrB(b,a,bQQposition,n),
	    TMatr.getAMatrB(b,b,bQQposition,n)
	},
        MdI=new double []{
	    TMatr.getAMatrB(a,a,bQQdidt,n), //=0
	    TMatr.getAMatrB(a,b,bQQdidt,n),
	    TMatr.getAMatrB(b,a,bQQdidt,n),
	    TMatr.getAMatrB(b,b,bQQdidt,n)
	};
    /** Probability correlation relation.
     */
    this.corr_Position_dI=new ProbabilityCorrelation(Mt,Mposition,MdI);
    if(eps>0)
    {
	System.err.println("DIselIH2PositiondI: Mt="+java.util.Arrays.toString(Mt)+
			   " Mposition="+java.util.Arrays.toString(Mposition)+
			   " MdI="+java.util.Arrays.toString(MdI)
			   );
	final int d=2;
	// Calculate directly:
	final double [] ev_position=M.EV.getGEV_Generalized(d,Mposition,Mt,true),
	    ev_dI=M.EV.getGEV_Generalized(d,MdI,Mt,true),
	    psi_position_L=EVXData.getPsiFromBasis(0,ev_position,d),
	    psi_position_H=EVXData.getPsiFromBasis(d-1,ev_position,d),
	    psi_dI_L=EVXData.getPsiFromBasis(0,ev_dI,d),
	    psi_dI_H=EVXData.getPsiFromBasis(d-1,ev_dI,d);
	final double w_LL=Math.pow(TMatr.getAMatrB(psi_position_L,psi_dI_L,Mt,d),2),
	    w_LH=Math.pow(TMatr.getAMatrB(psi_position_L,psi_dI_H,Mt,d),2),
	    w_HL=Math.pow(TMatr.getAMatrB(psi_position_H,psi_dI_L,Mt,d),2),
	    w_HH=Math.pow(TMatr.getAMatrB(psi_position_H,psi_dI_H,Mt,d),2);
	UnitTests.checkDoubleRel(w_LL+w_LH+w_HL+w_HH,"w_LL+w_LH+w_HL+w_HH",corr_Position_dI.W,"corr_Position_dI.W",eps);
	UnitTests.checkDoubleRel(w_LL-w_LH-w_HL+w_HH,"w_LL-w_LH-w_HL+w_HH",corr_Position_dI.corr*d,"corr_Position_dI.corr*d",eps);
	System.err.println("DIselIH2PositiondI: ev_position="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_position,d))+" ev_dI="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_dI,d)));
    }


    if(eps>0){
	System.err.println("DIselIH2PositiondI: OK n="+n+" corr_Position_dI.corr="+corr_Position_dI.corr+" I="+java.util.Arrays.toString(I.getEValues())+" Mposition="+java.util.Arrays.toString(Mposition)+" MdI="+java.util.Arrays.toString(MdI)+" eps="+eps);
    }


}


public static void matchDIselIH2PositiondI(final DIselIH2PositiondI a,final DIselIH2PositiondI b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    ProbabilityCorrelation.matchProbabilityCorrelation(a.corr_Position_dI,b.corr_Position_dI,eps);

    System.err.println("DIselIH2PositiondI.matchDIselIH2PositiondI test OK n="+a.I.neq+" eps="+eps);
}

}
