package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Simplistic version of selected basis {@literal |psiH>} and {@literal |dS/dt|psiH>}.
 * In contrast with the DIselIH.java, here we find the min/max of price, not (p-p_IH)*I.
 * @version $Id: DIselIH2pi.java,v 1.20 2022/07/01 17:28:19 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselIH2pi {
public final EVXData I;
public final double p_offset,
// The https://arxiv.org/abs/1510.05510 answers.
    p_IH,pf_IH,dEVseparation_IH,
// New answers
    wpsiHP_projection_simple_plus,
    wpsiHP_projection_simple_minus,
    GammaP,
    wpsiHPI_projection_simple_plus,
    wpsiHPI_projection_simple_minus,
    GammaPI
    ;

/**
 * @param n The dimension.
 * @param QQt {@literal <Q_j|Q_k>}.
 * @param QQi {@literal <Q_j|I|Q_k>}.
 * @param QQpi {@literal <Q_j|pI|Q_k>}.
 * @param M basis operations.
 * @param eps Run a selftest with this eps if &gt;0.
 */
public DIselIH2pi(final int n,
		  final boolean flag_Add_Impact_From_The_future,
		  final boolean flag_Use_Psi0_As_Petrubation,
		  final double p_offset,final double p_last,
		  final double [] QQt,final double [] QQi,final double [] QQpi,
		  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		  final double eps)
{
    this.p_offset=p_offset;
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];
    final double I_0=I.s0,I_0_feq=flag_Add_Impact_From_The_future?I.sH:I_0,dI=I_0_feq-I_0,
	dV=dI/I.o_o,Pf=p_last/* does have the offset. */;
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	QQi_feq=TMatr.sumAcB(QQi,dV,QQ0),
	QQpi_feq=TMatr.sumAcB(QQpi,Pf*dV,QQ0),
	bQQi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQi_feq,I.ev),
	bQQpi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQpi_feq,I.ev),
	// Diagonalize "future" |I|:
	bbasis=M.EV.getEV(n,bQQi_feq,true),
	bpsiHf=EVXData.getPsiFromBasis(ixH,bbasis,n),
	bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)) /*EVXData.convertPsiFromQtoB_Basis(n,I.psi0,QQt,I.ev) */
	;
    final double If_IH=bbasis[ixH];
    this.pf_IH=TMatr.getAMatrA(bpsiHf,bQQpi_feq,n)/If_IH;
    final double [] bQQDEVseparation=TMatr.sumaAbB(pf_IH,bQQi_feq,-1,bQQpi_feq);
    this.dEVseparation_IH=TMatr.getAMatrA(bpsiHf,bQQDEVseparation,n);
    if(eps>0) {
	final double [] psiHf=EVXData.convertPsiFromBtoQ_Basis(n,bpsiHf,I.ev);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi_feq,n),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psiHf,QQi_feq,n),"<psiHf|If|psiHf>",If_IH,"If_IH",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psiHf,QQi,n)+dI*Math.pow(TMatr.getAMatrB(psiHf,I.psi0,QQt,n),2),"<psiHf|If|psiHf>",If_IH,"If_IH",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psiHf,QQt,n),"<psiHf|psiHf>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psiHf,QQpi_feq,n),"<psiHf|QQpi_feq|psiHf>",pf_IH*If_IH,"pf_IH*If_IH",eps);
    }


    // The new answers

    final double [] UNIT=TMatr.getUnitMatrix(n);

    // dim=2 eigenvalues problem, the basis is |a>=|bpsiHf> and |b>=|bQQDEVseparation|bpsiHf> or bpsi0:
    final double [] a=bpsiHf,b=flag_Use_Psi0_As_Petrubation?
	bpsi0
	:
	TMatr.getMatr_x_A(n,bQQDEVseparation,bpsiHf);
    final int d=2;
    final double [] Sep2pi=new double []{
	TMatr.getAMatrB(a,a,bQQDEVseparation,n),
	TMatr.getAMatrB(a,b,bQQDEVseparation,n),
	TMatr.getAMatrB(b,a,bQQDEVseparation,n),
	TMatr.getAMatrB(b,b,bQQDEVseparation,n)},
    Sep2i=new double []{
	TMatr.getAMatrB(a,a,bQQi_feq,n),
	TMatr.getAMatrB(a,b,bQQi_feq,n),
	TMatr.getAMatrB(b,a,bQQi_feq,n),
	TMatr.getAMatrB(b,b,bQQi_feq,n)},
    ev_Sep2=M.EV.getGEV_Generalized(d,Sep2pi,Sep2i,true),
    bpsiP_minus=TMatr.sumaAbB(ev_Sep2[d+0*d+0],a,ev_Sep2[d+0*d+1],b),
    bpsiP_plus=TMatr.sumaAbB(ev_Sep2[d+1*d+0],a,ev_Sep2[d+1*d+1],b);
    this.wpsiHP_projection_simple_plus=Math.pow(TMatr.getAMatrB(bpsiHf,bpsiP_plus,bQQi_feq,n),2)/If_IH;
    this.wpsiHP_projection_simple_minus=Math.pow(TMatr.getAMatrB(bpsiHf,bpsiP_minus,bQQi_feq,n),2)/If_IH;
    this.GammaP=(2*0-ev_Sep2[0]-ev_Sep2[d-1])/(ev_Sep2[0]-ev_Sep2[d-1]);
    if(eps>0){
	UnitTests.checkDoubleRel(wpsiHP_projection_simple_plus+wpsiHP_projection_simple_minus,"wpsiHP_projection_simple_plus+wpsiHP_projection_simple_minus",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiHf,bQQDEVseparation,n),"<bpsiHf|bQQDEVseparation|bpsiHf>",0,"0",eps);
	UnitTests.checkDoubleRel(wpsiHP_projection_simple_plus-wpsiHP_projection_simple_minus,"wpsiHP_projection_simple_plus-wpsiHP_projection_simple_minus",-GammaP,"-GammaP",eps);

    }
    final double [] Mt=new double []{
	    TMatr.getAMatrB(a,a,UNIT,n),
	    TMatr.getAMatrB(a,b,UNIT,n),
	    TMatr.getAMatrB(b,a,UNIT,n),
	    TMatr.getAMatrB(b,b,UNIT,n)},
    ev_Sep2t=M.EV.getGEV_Generalized(d,Sep2pi,Mt,true),
    bpsiPI_minus=TMatr.sumaAbB(ev_Sep2t[d+0*d+0],a,ev_Sep2t[d+0*d+1],b),
    bpsiPI_plus=TMatr.sumaAbB(ev_Sep2t[d+1*d+0],a,ev_Sep2t[d+1*d+1],b);
    wpsiHPI_projection_simple_plus=Math.pow(TMatr.getAMatrB(bpsiHf,bpsiPI_plus,UNIT,n),2);
    wpsiHPI_projection_simple_minus=Math.pow(TMatr.getAMatrB(bpsiHf,bpsiPI_minus,UNIT,n),2);
    GammaPI=(2*0-ev_Sep2t[0]-ev_Sep2t[d-1])/(ev_Sep2t[0]-ev_Sep2t[d-1]);
    if(eps>0){
	UnitTests.checkDoubleRel(wpsiHPI_projection_simple_plus+wpsiHPI_projection_simple_minus,"wpsiHPI_projection_simple_plus+wpsiHPI_projection_simple_minus",1,"1",eps);
	UnitTests.checkDoubleRel(wpsiHPI_projection_simple_plus-wpsiHPI_projection_simple_minus,"wpsiHPI_projection_simple_plus-wpsiHPI_projection_simple_minus",-GammaPI,"-GammaPI",eps);
    }
    if(eps>0){
	final DIselIH ih=new DIselIH(n,flag_Use_Psi0_As_Petrubation?DIselIH.METHOD_2EV_QQ0:DIselIH.METHOD_2EV_QQDEV,null,p_offset,p_last,QQt,QQi_feq,QQpi_feq,QQt,M,eps);
	UnitTests.checkDoubleRel(wpsiHPI_projection_simple_plus,"wpsiHPI_projection_simple_plus",ih.wpsiH_projection_simple_plus,"ih.wpsiH_projection_simple_plus",eps);
	UnitTests.checkDoubleRel(wpsiHPI_projection_simple_minus,"wpsiHPI_projection_simple_minus",ih.wpsiH_projection_simple_minus,"ih.wpsiH_projection_simple_minus",eps);
    }
    if(eps>0){
	System.err.println("DIselIH2pi: OK n="+n+" I="+java.util.Arrays.toString(I.getEValues())+" If_IH="+If_IH+" flag_Add_Impact_From_The_future="+flag_Add_Impact_From_The_future+" flag_Use_Psi0_As_Petrubation="+flag_Use_Psi0_As_Petrubation+" GammaP="+GammaP+" GammaPI="+GammaPI+" dEVseparation_IH="+this.dEVseparation_IH+" eps="+eps);
    }


}


public static void matchDIselIH2pi(final DIselIH2pi a,final DIselIH2pi b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.pf_IH,"a.pf_IH",b.pf_IH+p_offset_shift,"b.pf_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.wpsiHP_projection_simple_plus,"a.wpsiHP_projection_simple_plus",b.wpsiHP_projection_simple_plus,"b.wpsiHP_projection_simple_plus",eps);
    UnitTests.checkDoubleRel(a.wpsiHP_projection_simple_minus,"a.wpsiHP_projection_simple_minus",b.wpsiHP_projection_simple_minus,"b.wpsiHP_projection_simple_minus",eps);
    UnitTests.checkDoubleRel(a.GammaP,"a.GammaP",b.GammaP,"b.GammaP",eps);

    UnitTests.checkDoubleRel(a.wpsiHPI_projection_simple_plus,"a.wpsiHPI_projection_simple_plus",b.wpsiHPI_projection_simple_plus,"b.wpsiHPI_projection_simple_plus",eps);
    UnitTests.checkDoubleRel(a.wpsiHPI_projection_simple_minus,"a.wpsiHPI_projection_simple_minus",b.wpsiHPI_projection_simple_minus,"b.wpsiHPI_projection_simple_minus",eps);
    UnitTests.checkDoubleRel(a.GammaPI,"a.GammaPI",b.GammaPI,"b.GammaPI",eps);

    System.err.println("DIselIH2pi.matchDIselIH2pi test OK n="+a.I.neq+" eps="+eps);
}

}
