package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.LebesgueQuadratureWithEVData;
import com.polytechnik.utils.LebesgueQuadratureJointDistribution;

/** General +- projections of dI/dt and dS/dt into a density matrix state.
 * This arroach corresponds to Lebesgue Quadratures density matrix correlation
 * from the https://arxiv.org/abs/1807.08197 .
 * @version $Id: DIselIHPositiondI.java,v 1.62 2022/08/07 15:44:02 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselIHPositiondI {
public final EVXData I;
public final double p_offset,
// The https://arxiv.org/abs/1510.05510 answers.
    p_IH
    ;
/** New answer
 */
public final double dm_spur,position_eq,w_pp,w_pm,w_mp,w_mm,
    dS_in_dI_plus,dS_in_dI_minus,dI_in_dS_plus,dI_in_dS_minus
    ;
public final double [] lambda_dS,lambda_dI,dm_corr;
public DIselIHPositiondI(
    final EVXData I,
    final double p_offset,
    final double p_IH,
    final double dm_spur,final double position_eq,
    final double w_pp,final double w_pm,final double w_mp,final double w_mm,
    final double [] lambda_dS,final double [] lambda_dI,final double [] dm_corr,
    final double dS_in_dI_plus,final double dS_in_dI_minus,final double dI_in_dS_plus,final double dI_in_dS_minus){
    this.I=I;this.p_offset=p_offset;this.p_IH=p_IH;
    this.dm_spur=dm_spur;this.position_eq=position_eq;this.w_pp=w_pp;this.w_pm=w_pm;this.w_mp=w_mp;this.w_mm=w_mm;
    this.lambda_dS=lambda_dS;this.lambda_dI=lambda_dI;this.dm_corr=dm_corr;
    this.dS_in_dI_plus=dS_in_dI_plus;this.dS_in_dI_minus=dS_in_dI_minus;
    this.dI_in_dS_plus=dI_in_dS_plus;this.dI_in_dS_minus=dI_in_dS_minus;
}


/**
 * Use the density matrix {@literal |psiH><psiH|} and dS/dt = -(p-p_IH)*I
 * @param n The dimension.
 * @param p_offset Price offset.
 * @param QQt {@literal <Q_j|Q_k>}.
 * @param QQi {@literal <Q_j|I|Q_k>}.
 * @param QQpm {@literal <Q_j|pm|Q_k>} price operator relatively some measure, e.g. m=I.
 * @param QQm {@literal <Q_j|m|Q_k>}.
 * @param M basis operations.
 * @param VOLATILITY_EDPsi Need it for infinitesimal timeshift to calculate dI/dt.
 * @param eps Run a selftest with this eps if >0.
 */
public static DIselIHPositiondI getLocalInPsiH(
    final int n,
    final double p_offset,final double p_last,
    final double [] QQt,final double [] QQi,final double [] QQpm,final double [] QQm,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
    final VolMatrix VOLATILITY_EDPsi,
    final boolean use_dynamicimpact,
    final double eps)
{
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double []
	bQQdidt=EVXData.convertMatrFromQtoB_Basis(n,new QQdidtMatrix(QQdidtMatrix.BOUNDARY_IeqIH,I,QQi,M,VOLATILITY_EDPsi).QQdidt,I.ev),UNIT=TMatr.getUnitMatrix(n),
	bQQpm=EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev),
	bQQm=EVXData.convertMatrFromQtoB_Basis(n,QQm,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),
	bQQp_diagI=new double [n*n];
    for(int k=0;k<n;k++) bQQp_diagI[k+n*k]=bQQpm[k+n*k]/bQQm[k+n*k];
    final double p_IH=bQQp_diagI[ixH+n*ixH];
    final double [] QQp_diagI=EVXData.convertMatrFromBtoQ_Basis(n,bQQp_diagI,I.ev,QQt),
	QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	bQQposition=use_dynamicimpact?
	EVXData.convertMatrFromQtoB_Basis(n,TMatr.sumaAbB(p_IH,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQp_diagI)),I.ev)
	   :TMatr.sumaAbB(p_IH,bQQm,-1,bQQpm),
	bpsiH=new double [n];bpsiH[ixH]=1;
    final double position_eq=0,dm_spur=TMatr.mulVV(bpsiH,bpsiH);

    if(eps>0)
    {
	final double mh=TMatr.getAMatrA(bpsiH,bQQm,n);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQi,n),"<psiH|I|psiH>",I.sH,"I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQpm,n),"<psiH|pm|psiH>",p_IH*mh,"p_IH*I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQposition,n),"<psiH|bQQposition|psiH>",position_eq,"position_eq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQdidt,n),"<psiH|bQQdidt|psiH>",0,"0",eps);
    }

    final double [] ev_bQQposition=M.EV.getEV(n,bQQposition,true),
	ev_bQQdidt=M.EV.getEV(n,bQQdidt,true),dm_corr=new double [n*n];
    double dI_in_dS_plus=0,dI_in_dS_minus=0,dS_in_dI_plus=0,dS_in_dI_minus=0,w_pp=0,w_pm=0,w_mp=0,w_mm=0;
    for(int j=0;j<n;j++)
    {
	final double w_position=DeprecatedHelperTMatr.mulVVo(n,bpsiH,0,ev_bQQposition,n+n*j);
	for(int k=0;k<n;k++)
	{
	    final double w_position_dI=DeprecatedHelperTMatr.mulVVo(n,ev_bQQposition,n+n*j,ev_bQQdidt,n+n*k),
		w_dI=DeprecatedHelperTMatr.mulVVo(n,bpsiH,0,ev_bQQdidt,n+n*k),
		w=w_position*w_dI*w_position_dI;
	    dm_corr[j*n+k]=w;
	    if(ev_bQQposition[j]>(position_eq/dm_spur)){
		dI_in_dS_plus+=ev_bQQdidt[k]*w;
		if(ev_bQQdidt[k]>0){w_pp+=w;}else{w_pm+=w;}
	    } else {
		dI_in_dS_minus+=ev_bQQdidt[k]*w;
		if(ev_bQQdidt[k]>0){w_mp+=w;}else{w_mm+=w;}
	    }
	    if(ev_bQQdidt[k]>0){
		dS_in_dI_plus+=ev_bQQposition[j]*w;
	    } else {
		dS_in_dI_minus+=ev_bQQposition[j]*w;
	    }
	}
    }

    if(eps>0){
	UnitTests.checkDoubleRel(dm_spur,"dm_spur",1,"1",eps);
	UnitTests.checkDoubleRel(w_pp+w_pm+w_mp+w_mm,"sum(w_pp+w_pm+w_mm+w_mm)",1,"1",eps);
	UnitTests.checkDoubleRel(dI_in_dS_plus+dI_in_dS_minus,"dI_in_dS_plus+dI_in_dS_minus",0,"0",eps);
	UnitTests.checkDoubleRel(dS_in_dI_plus+dS_in_dI_minus,"dS_in_dI_plus+dS_in_dI_minus",position_eq,"position_eq",eps);
	// Match the Lebesgue Quadratures density matrix correlation from the https://arxiv.org/abs/1807.08197
	final LebesgueQuadratureJointDistribution jd=
	    new LebesgueQuadratureJointDistribution(
		new LebesgueQuadratureWithEVData(
		    n,EVXData.convertMatrFromBtoQ_Basis(n,bQQposition,I.ev,QQt),QQt,M),
		new LebesgueQuadratureWithEVData(
		    n,EVXData.convertMatrFromBtoQ_Basis(n,bQQdidt,I.ev,QQt),QQt,M));
	final double [] DMpsiH=EVXData.convertMatrFromBtoQ_Basis(n,TMatr.getQQaa(bpsiH),I.ev,QQt),
	    _dm_corr=jd.getDensityMatrixCorrelation(DMpsiH),C=new double [n];java.util.Arrays.fill(C,1);
	UnitTests.checkDoubleArrayRel1(dm_corr,"dm_corr",_dm_corr,"_dm_corr",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psiH,jd.ff.QQf,n),"position_eq",position_eq,"position_eq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(n,jd.ff.ev,C,_dm_corr,n),"spur dS*rho",position_eq,"position_eq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psiH,jd.gg.QQf,n),"0",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(n,C,jd.gg.ev,_dm_corr,n),"spur dI*rho",0,"0",eps);

	// match SepSkewnes
	final SepSkewness ss=new SepSkewness(I.psiH,jd.ff.QQf,QQt,M.EV,eps*0);
	UnitTests.checkDoubleRel(w_pp+w_pm,"w_pp+w_pm",ss.w_plus,"ss.w_plus",eps);
	UnitTests.checkDoubleRel(w_mp+w_mm,"w_mp+w_mm",ss.w_minus,"ss.w_minus",eps);
    }

    if(eps>0){
	System.err.println("DIselIHPositiondI: dm_corr="+java.util.Arrays.toString(dm_corr)+" I="+java.util.Arrays.toString(I.getEValues()));
	System.err.println("DIselIHPositiondI: OK n="+n+" eps="+eps);
    }

    return new DIselIHPositiondI(I,p_offset,p_IH,
				 dm_spur,position_eq,w_pp,w_pm,w_mp,w_mm,
				 java.util.Arrays.copyOf(ev_bQQposition,n),java.util.Arrays.copyOf(ev_bQQdidt,n),dm_corr,
				 dS_in_dI_plus,dS_in_dI_minus,dI_in_dS_plus,dI_in_dS_minus);
}



/**
 * Use the density matrix "after max spike in I" and dS/dt = (F(x)p')'
 * Consider the I=dV/dt and the state psiH(x) corresponding to max lambda_{I}.
 * Then obtain F(x)=\int_{-inf}^{x)psiH^2(x)omega(x)dx where F(tnow)=1 follow from psiH normalizing,
 * this F(x) can be density matrix expanded as F(x)=rho(x,x).
 * @param n The dimension.
 * @param p_offset Price offset.
 * @param QQt {@literal <Q_j|Q_k>}.
 * @param QQi {@literal <Q_j|I|Q_k>}.
 * @param QQpi {@literal <Q_j|pI|Q_k>}.
 * @param QQdpdt {@literal <Q_j|dp/dt|Q_k>}.
 * @param M basis operations.
 * @param VOLATILITY_EDPsi Need it for infinitesimal timeshift to calculate dI/dt.
 * @param getPsi2WIntegratedDt an integrator of psi^2(x)omega(x) to obtain "after the psi" state.
 * @param eps Run a selftest with this eps if &gt;0.
 */
public static DIselIHPositiondI getAfterPsiH(
    final int n,
    final double p_offset,
    final double [] QQt,final double [] QQi,final double [] QQpi,final double [] QQdpdt,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
    final VolMatrix VOLATILITY_EDPsi,
    final java.util.function.Function<double [],double []> getPsi2WIntegratedDt,
    final boolean use_p_last_as_equilibrium,
    final boolean use_average_twice,
    final double eps)
{
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];
    final double [] psiWeight=getPsi2WIntegratedDt.apply(I.psiH);
    if(psiWeight.length!=2*n-1) throw new IllegalArgumentException();
    final double [] UNIT=TMatr.getUnitMatrix(n),
	// Gram matrix inverse can be obtained without solving a linear system.
	IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT),
	//IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	bDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,psiWeight,IQQt)),I.ev),
	bQQdpdt=EVXData.convertMatrFromQtoB_Basis(n,QQdpdt,I.ev),
	bQQdidt=EVXData.convertMatrFromQtoB_Basis(n,new QQdidtMatrix(QQdidtMatrix.BOUNDARY_IeqIH,I,QQi,M,VOLATILITY_EDPsi).QQdidt,I.ev),
	bQQposition=use_p_last_as_equilibrium?
	bQQdpdt
	:
	TMatr.sumAcB(bQQdpdt,-TMatr.getSpurMM(n,bQQdpdt,bDensityMatrix),TMatr.getQQaa(I.psik_at_x0))
	;
    final double position_eq=use_p_last_as_equilibrium?TMatr.getSpurMM(n,bQQposition,bDensityMatrix):0,
	dm_spur1=TMatr.getSpurM(n,bDensityMatrix),
	dm_weight=use_average_twice?TMatr.getSpurMM(n,bDensityMatrix,bDensityMatrix):dm_spur1;
    if(eps>0)
    {
	System.err.println("use_p_last_as_equilibrium="+use_p_last_as_equilibrium+" position_eq="+position_eq);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQposition,bDensityMatrix),"spur |bQQposition|bDensityMatrix|",position_eq,"position_eq",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdidt,bDensityMatrix),"spur |bQQdidt|bDensityMatrix|",0,"0",eps);
    }

    final double [] ev_bQQposition=M.EV.getEV(n,bQQposition,true),
	ev_bQQdidt=M.EV.getEV(n,bQQdidt,true),dm_corr=new double [n*n];
    double dI_in_dS_plus=0,dI_in_dS_minus=0,dS_in_dI_plus=0,dS_in_dI_minus=0,w_pp=0,w_pm=0,w_mp=0,w_mm=0;
    for(int j=0;j<n;j++)
    {
	for(int k=0;k<n;k++)
	{
	    final double w_position_dI=TMatr.getAMatrB(n,ev_bQQposition,n+n*j,ev_bQQdidt,n+n*k,UNIT,n),
		dm_position_dI=TMatr.getAMatrB(n,ev_bQQposition,n+n*j,ev_bQQdidt,n+n*k,bDensityMatrix,n),
		w=use_average_twice?Math.pow(dm_position_dI,2):w_position_dI*dm_position_dI;
	    dm_corr[j*n+k]=w;
	    if(ev_bQQposition[j]>(position_eq/dm_spur1)){
		dI_in_dS_plus+=ev_bQQdidt[k]*w*(use_average_twice?ev_bQQposition[j]:1);
		if(ev_bQQdidt[k]>0){w_pp+=w;}else{w_pm+=w;}
	    } else {
		dI_in_dS_minus+=ev_bQQdidt[k]*w*(use_average_twice?ev_bQQposition[j]:1);
		if(ev_bQQdidt[k]>0){w_mp+=w;}else{w_mm+=w;}
	    }
	    if(ev_bQQdidt[k]>0){
		dS_in_dI_plus+=ev_bQQposition[j]*w*(use_average_twice?ev_bQQdidt[k]:1);
	    } else {
		dS_in_dI_minus+=ev_bQQposition[j]*w*(use_average_twice?ev_bQQdidt[k]:1);;
	    }
	}
    }

    if(eps>0){
	UnitTests.checkDoubleRel(TMatr.sumElements(dm_corr),"sum(dm_corr)",dm_weight,"dm_weight",eps);
	UnitTests.checkDoubleRel(w_pp+w_pm+w_mp+w_mm,"sum(w_pp+w_pm+w_mm+w_mm)",dm_weight,"dm_weight",eps);
	if(!use_average_twice){
	    UnitTests.checkDoubleRel(dI_in_dS_plus+dI_in_dS_minus,"dI_in_dS_plus+dI_in_dS_minus",0,"0",eps);
	    UnitTests.checkDoubleRel(dS_in_dI_plus+dS_in_dI_minus,"dS_in_dI_plus+dS_in_dI_minus",position_eq,"position_eq",eps);
	}

	// Match the Lebesgue Quadratures density matrix correlation from the https://arxiv.org/abs/1807.08197
	final LebesgueQuadratureJointDistribution jd=
	    new LebesgueQuadratureJointDistribution(
		new LebesgueQuadratureWithEVData(
		    n,EVXData.convertMatrFromBtoQ_Basis(n,bQQposition,I.ev,QQt),QQt,M),
		new LebesgueQuadratureWithEVData(
		    n,EVXData.convertMatrFromBtoQ_Basis(n,bQQdidt,I.ev,QQt),QQt,M));
	final double [] DM=EVXData.convertMatrFromBtoQ_Basis(n,bDensityMatrix,I.ev,QQt),
	    _dm_corr=use_average_twice?jd.get_M_M_Correlation(DM,DM):jd.getDensityMatrixCorrelation(DM),
	    C=new double [n];java.util.Arrays.fill(C,1);
	UnitTests.checkDoubleArrayRel1(dm_corr,"dm_corr",_dm_corr,"_dm_corr",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{jd.ff.QQf,IQQt,DM,IQQt})),"spur dS*rho",position_eq,"position_eq",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{jd.gg.QQf,IQQt,DM,IQQt})),"spur dI*rho",0,"0",eps);
	if(!use_average_twice){
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(n,jd.ff.ev,C,_dm_corr,n),"spur dS*rho",position_eq,"position_eq",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(n,C,jd.gg.ev,_dm_corr,n),"spur dI*rho",0,"0",eps);
	}
	if(use_average_twice){
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(n,jd.ff.ev,jd.gg.ev,_dm_corr,n),"spur dS*rho*dI*rho",dS_in_dI_plus+dS_in_dI_minus,"dS_in_dI_plus+dS_in_dI_minus",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(n,jd.ff.ev,jd.gg.ev,_dm_corr,n),"spur dS*rho*dI*rho",dI_in_dS_plus+dI_in_dS_minus,"dI_in_dS_plus+dI_in_dS_minus",eps);
	}
    }

    if(eps>0){
	System.err.println("DIselIHPositiondI: dm_spur1="+dm_spur1+" dm_weight="+dm_weight+" dm_corr="+java.util.Arrays.toString(dm_corr)+" I="+java.util.Arrays.toString(I.getEValues()));
	System.err.println("DIselIHPositiondI: OK n="+n+" eps="+eps);
    }

    return new DIselIHPositiondI(I,p_offset,p_IH,
				 dm_weight,position_eq,w_pp,w_pm,w_mp,w_mm,
				 java.util.Arrays.copyOf(ev_bQQposition,n),java.util.Arrays.copyOf(ev_bQQdidt,n),dm_corr,
				 dS_in_dI_plus,dS_in_dI_minus,dI_in_dS_plus,dI_in_dS_minus);
}

/** Dynamic impact answer.
 */
public static DIselIHPositiondI getAfterPsiH(
    final int n,
    final double p_offset,final double p_last,
    final double [] QQt,final double [] QQi,final double [] QQpm,final double [] QQm,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
    final VolMatrix VOLATILITY_EDPsi,
    final java.util.function.Function<double [],double []> getPsi2WIntegratedDt,
    final boolean use_p_last_as_equilibrium,
    final double eps)
{
    final int ixH=n-1; // maximal eigenvalue index
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double []
	bQQi=TMatr.getDiagMatrix(I.getEValues()),UNIT=TMatr.getUnitMatrix(n),
	bQQpm=EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev),
	bQQm=EVXData.convertMatrFromQtoB_Basis(n,QQm,I.ev),
	bQQp_diagI=new double [n*n];
    for(int k=0;k<n;k++) bQQp_diagI[k+n*k]=bQQpm[k+n*k]/bQQm[k+n*k];
    final double I_0_feq=I.sH,p_IH=bQQp_diagI[ixH+n*ixH],p_eq=use_p_last_as_equilibrium?p_last:p_IH;
    final double [] QQp_diagI=EVXData.convertMatrFromBtoQ_Basis(n,bQQp_diagI,I.ev,QQt),
	QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	bQQdpdt=EVXData.convertMatrFromQtoB_Basis(n,TMatr.sumaAbB(p_eq,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQp_diagI)),I.ev),
	bQQdidt=EVXData.convertMatrFromQtoB_Basis(n,TMatr.sumaAbB(I_0_feq,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQi)),I.ev);

    final double [] psiWeight=getPsi2WIntegratedDt.apply(I.psiH);
    if(psiWeight.length!=2*n-1) throw new IllegalArgumentException();
    final double []
	// Gram matrix inverse can be obtained without solving a linear system.
	IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT),
	//IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	bDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,psiWeight,IQQt)),I.ev);
    if(eps>0){
	System.err.println("psiWeight(x0)="+M.B.getD0(psiWeight,I.x0));
	final double [] QQdpdt=EVXData.convertMatrFromBtoQ_Basis(n,bQQdpdt,I.ev,QQt),
	    DensityMatrix=M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,psiWeight,IQQt)),
	    bDiffDensityMatix=EVXData.convertMatrFromQtoB_Basis(n,VOLATILITY_EDPsi.getDiffMatrix(n,DensityMatrix),I.ev);
	final EVXData DM=EVXData.getEVXDataFromQQ(n,DensityMatrix,QQt,M);
	double s=0,si=0;
	final double [] psi2=new double [2*(n-1)+1];
	for(int k=0;k<n;k++){
	    final double [] psi=DM.getPsi(k),Dpsi=VOLATILITY_EDPsi.getDPsi(psi);
	    s+=DM.ev[k]*2*TMatr.getAMatrB(psi,Dpsi,QQp_diagI,n);
	    si+=DM.ev[k]*2*TMatr.getAMatrB(psi,Dpsi,QQi,n);
	    TMatr.sumP1(psi2,M.B.mult2Pol(psi,Dpsi),2*DM.ev[k]);
	}
	System.err.println("s="+s+" p_IH="+p_IH+" si="+si+" I.sH="+I.sH);
	System.err.println("psi2a="+java.util.Arrays.toString(psi2));
	System.err.println("psi2o="+java.util.Arrays.toString(M.B.mult2Pol(I.psiH,I.psiH)));
	for(int k=0;k<n;k++)
	{
	    final double pk=I.getFPsi(k,QQp_diagI),psik20=Math.pow(I.psik_at_x0[k],2);
	    System.err.println("k="+k+" <psik|QQdpdt|psik>="+I.getFPsi(k,QQdpdt)+" (p_eq-lamk)*psi^x(x0)="+((p_eq-pk)*psik20));
	}
	System.err.println("spur="+TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,DensityMatrix,IQQt,QQdpdt}))+" bDiffDensityMatix="+java.util.Arrays.toString(bDiffDensityMatix)+" spur="+TMatr.getSpurM(n,bDiffDensityMatix));
	System.err.println(" bQQdpdt="+java.util.Arrays.toString(bQQdpdt)+" <>="+TMatr.getAMatrA(I.psiH,QQdpdt,n));
	System.err.println("use_p_last_as_equilibrium="+use_p_last_as_equilibrium);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psiH,QQp_diagI,n),"p_IH",p_IH,"p_IH",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdidt,bDensityMatrix),"spur |bQQdidt|bDensityMatrix|",I_0_feq-I.sH,"I_0_feq-I.sH",eps);
	// THIS NOT WORKING!!!
	if(false) UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdpdt,bDensityMatrix),"spur |bQQdpdt|bDensityMatrix|",p_eq-p_IH,"p_eq-p_IH",eps);
    }
    return null;

}
/** Dyn version, det dp/dt elements to 0.
 */
public static DIselIHPositiondI getLocalInPsiHdyn(
    final int n,
    final double p_offset,
    final double [] QQt,final double [] QQi,final double [] QQpi,final double [] QQdpdt,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
    final VolMatrix VOLATILITY_EDPsi,
    final java.util.function.Function<double [],double []> getPsi2WIntegratedDt,
    final double eps)
{
     // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];
    final double [] psiWeight=getPsi2WIntegratedDt.apply(I.psiH);
    if(psiWeight.length!=2*n-1) throw new IllegalArgumentException();
    final double [] UNIT=TMatr.getUnitMatrix(n),
	// Gram matrix inverse can be obtained without solving a linear system.
	IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT),
	//IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	bDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,psiWeight,IQQt)),I.ev),
	bQQdpdt=EVXData.convertMatrFromQtoB_Basis(n,QQdpdt,I.ev),
	bQQdidt=EVXData.convertMatrFromQtoB_Basis(n,new QQdidtMatrix(QQdidtMatrix.BOUNDARY_IeqIH,I,QQi,M,VOLATILITY_EDPsi).QQdidt,I.ev);
    // Dynamic impact approximation
    for(int k=0;k<n;k++) bQQdpdt[k+n*k]=0;
    final double [] bQQposition=bQQdpdt;
    final double position_eq=TMatr.getSpurMM(n,bQQposition,bDensityMatrix),
	dm_spur1=TMatr.getSpurM(n,bDensityMatrix),dm_weight=dm_spur1;

    if(eps>0){
	System.err.println("psiWeight(x0)="+M.B.getD0(psiWeight,I.x0));
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdidt,bDensityMatrix),"spur |bQQdidt|bDensityMatrix|",I.sH-I.sH,"I.sH-I.sH",eps);
	//UnitTests.checkDoubleRel(TMatr.getSpurMM(n,bQQdpdt,bDensityMatrix),"spur |bQQdpdt|bDensityMatrix|",p_eq-p_IH,"p_eq-p_IH",eps);
    }

    final boolean use_average_twice=false;
    final double [] ev_bQQposition=M.EV.getEV(n,bQQposition,true),
	ev_bQQdidt=M.EV.getEV(n,bQQdidt,true),dm_corr=new double [n*n];
    double dI_in_dS_plus=0,dI_in_dS_minus=0,dS_in_dI_plus=0,dS_in_dI_minus=0,w_pp=0,w_pm=0,w_mp=0,w_mm=0;
    for(int j=0;j<n;j++)
    {
	for(int k=0;k<n;k++)
	{
	    final double w_position_dI=TMatr.getAMatrB(n,ev_bQQposition,n+n*j,ev_bQQdidt,n+n*k,UNIT,n),
		dm_position_dI=TMatr.getAMatrB(n,ev_bQQposition,n+n*j,ev_bQQdidt,n+n*k,bDensityMatrix,n),
		w=use_average_twice?Math.pow(dm_position_dI,2):w_position_dI*dm_position_dI;
	    dm_corr[j*n+k]=w;
	    if(ev_bQQposition[j]>(position_eq/dm_spur1)){
		dI_in_dS_plus+=ev_bQQdidt[k]*w*(use_average_twice?ev_bQQposition[j]:1);
		if(ev_bQQdidt[k]>0){w_pp+=w;}else{w_pm+=w;}
	    } else {
		dI_in_dS_minus+=ev_bQQdidt[k]*w*(use_average_twice?ev_bQQposition[j]:1);
		if(ev_bQQdidt[k]>0){w_mp+=w;}else{w_mm+=w;}
	    }
	    if(ev_bQQdidt[k]>0){
		dS_in_dI_plus+=ev_bQQposition[j]*w*(use_average_twice?ev_bQQdidt[k]:1);
	    } else {
		dS_in_dI_minus+=ev_bQQposition[j]*w*(use_average_twice?ev_bQQdidt[k]:1);;
	    }
	}
    }

    return new DIselIHPositiondI(I,p_offset,p_IH,
				 dm_weight,position_eq,w_pp,w_pm,w_mp,w_mm,
				 java.util.Arrays.copyOf(ev_bQQposition,n),java.util.Arrays.copyOf(ev_bQQdidt,n),dm_corr,
				 dS_in_dI_plus,dS_in_dI_minus,dI_in_dS_plus,dI_in_dS_minus);
}


public static void matchDIselIHPositiondI(final DIselIHPositiondI a,final DIselIHPositiondI b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.dm_spur,"a.dm_spur",b.dm_spur,"b.dm_spur",eps);

    UnitTests.checkDoubleRel(a.dI_in_dS_plus,"a.dI_in_dS_plus",b.dI_in_dS_plus,"b.dI_in_dS_plus",eps);
    UnitTests.checkDoubleRel(a.dI_in_dS_minus,"a.dI_in_dS_minus",b.dI_in_dS_minus,"b.dI_in_dS_minus",eps);
    UnitTests.checkDoubleRel(a.dS_in_dI_plus,"a.dS_in_dI_plus",b.dS_in_dI_plus,"b.dS_in_dI_plus",eps);
    UnitTests.checkDoubleRel(a.dS_in_dI_minus,"a.dS_in_dI_minus",b.dS_in_dI_minus,"b.dS_in_dI_minus",eps);
    UnitTests.checkDoubleRel(a.w_pp,"a.w_pp",b.w_pp,"b.w_pp",eps);
    UnitTests.checkDoubleRel(a.w_pm,"a.w_pm",b.w_pm,"b.w_pm",eps);
    UnitTests.checkDoubleRel(a.w_mp,"a.w_mp",b.w_mp,"b.w_mp",eps);
    UnitTests.checkDoubleRel(a.w_mm,"a.w_mm",b.w_mm,"b.w_mm",eps);

    System.err.println("DIselIHPositiondI.matchDIselIHPositiondI test OK n="+a.I.neq+" eps="+eps);
}

}
