package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.VolMatrix;

/** Selected basis, the states of positive/negative QQDEVseparation eigenvalues,
 * then solve I-&gt; max using selected basis.
 * @version $Id: DIselIHskewness.java,v 1.16 2022/08/05 11:25:32 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselIHskewness {
public final EVXData I;
public final double p_offset,
// The https://arxiv.org/abs/1510.05510 answers.
    p_IH
    ;

/** The new answers.
 */
public final Skewness skIH,skIH_plus,skIH_minus,sk_dim;
public final double sk_dI_plus,sk_dI_minus,pI_plus,pI_minus,var2p_dim;

/**
 * @param n The dimension.
 * @param QQt {@literal <Q_j|Q_k>}.
 * @param QQi {@literal <Q_j|I|Q_k>}.
 * @param QQpi {@literal <Q_j|pI|Q_k>}.
 * @param M basis operations.
 * @param eps Run a selftest with this eps if &gt;0.
 */
public DIselIHskewness(final int n,
		       final double p_offset,
		       final double [] QQt,final double [] QQi,final double [] QQpi,final double [] QQp2i,final double [] QQp3i,
		       final double [] QQpt,
		       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		       final VolMatrix VOLATILITY_EDPsi,
		       final double eps)
{
    this.p_offset=p_offset;
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];

    final double []
	bQQdidt=EVXData.convertMatrFromQtoB_Basis(n,new QQdidtMatrix(QQdidtMatrix.BOUNDARY_IeqIH,I,QQi,M,VOLATILITY_EDPsi).QQdidt,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,QQp2i,I.ev),
	bQQp3i=EVXData.convertMatrFromQtoB_Basis(n,QQp3i,I.ev),
	bQQp_diagI=new double [n*n],var2I=new double [n],var2p=new double [n],normdd=new double [n],
	ev_bQQdidt=M.EV.getEV(n,bQQdidt,true);
    for(int k=0;k<n;k++){bQQp_diagI[k+n*k]=bQQpi[k+n*k]/I.ev[k];}
    for(int k=0;k<n;k++){
	final double [] bDpsi=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(I.getPsi(k)),I.ev,QQt);
	bDpsi[k]=0; // remove psik projection, now bpsi and bDpsi are orthogonal 2D eigenvalues solution
	final double dd=normdd[k]=TMatr.mulVV(bDpsi,bDpsi);
	var2I[k]=TMatr.mulV2V(bDpsi,I.getEValues())/dd-I.ev[k];
	var2p[k]=TMatr.getAMatrA(bDpsi,bQQp_diagI,n)/dd-bQQp_diagI[k+n*k];
    }
    final double [] bpsiH=new double [n];bpsiH[ixH]=1;
    final double dIsplit=0;
    if(eps>0){
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQi,n),"<psiH|I|psiH>",I.sH,"I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQpi,n),"<psiH|pI|psiH>",p_IH*I.sH,"p_IH*I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiH,bQQdidt,n),"<psiH|dI/dt|psiH>",dIsplit,"dIsplit",eps);
    }
    final BasisSplit di_pm=new BasisSplit(n,ev_bQQdidt,(ev) -> ev>dIsplit,bQQi,M.EV,eps);
    this.skIH=Skewness.getSkewnessFromMonomialMoments(
	TMatr.getAMatrA(bpsiH,bQQi,n),
	TMatr.getAMatrA(bpsiH,bQQpi,n),
	TMatr.getAMatrA(bpsiH,bQQp2i,n),
	TMatr.getAMatrA(bpsiH,bQQp3i,n));
    this.skIH_plus=Skewness.getSkewnessFromMonomialMoments(
	TMatr.getAMatrA(di_pm.sel_plus.psiH,bQQi,n),
	TMatr.getAMatrA(di_pm.sel_plus.psiH,bQQpi,n),
	TMatr.getAMatrA(di_pm.sel_plus.psiH,bQQp2i,n),
	TMatr.getAMatrA(di_pm.sel_plus.psiH,bQQp3i,n));
    this.skIH_minus=Skewness.getSkewnessFromMonomialMoments(
	TMatr.getAMatrA(di_pm.sel_minus.psiH,bQQi,n),
	TMatr.getAMatrA(di_pm.sel_minus.psiH,bQQpi,n),
	TMatr.getAMatrA(di_pm.sel_minus.psiH,bQQp2i,n),
	TMatr.getAMatrA(di_pm.sel_minus.psiH,bQQp3i,n));
    double wpsiH_projection_simple_plus=0,sk_dI_plus=0,wpsiH_projection_simple_minus=0,sk_dI_minus=0;
    final double [] bQQpt=EVXData.convertMatrFromQtoB_Basis(n,QQpt,I.ev);
    double pI_plus=0,pI_minus=0;
    for(int k=0;k<n;k++)
    {
	final double w=Math.pow(DeprecatedHelperTMatr.mulVVo(n,ev_bQQdidt,n+k*n,bpsiH,0),2),
	    m0=TMatr.getAMatrA(n,ev_bQQdidt,n+k*n,bQQi,n),
	    m1=TMatr.getAMatrA(n,ev_bQQdidt,n+k*n,bQQpi,n),
	    m2=TMatr.getAMatrA(n,ev_bQQdidt,n+k*n,bQQp2i,n),
	    m3=TMatr.getAMatrA(n,ev_bQQdidt,n+k*n,bQQp3i,n);
	final Skewness sk=Skewness.getSkewnessFromMonomialMoments(m0,m1,m2,m3);
	if(ev_bQQdidt[k]>dIsplit){wpsiH_projection_simple_plus+=w;sk_dI_plus+=w*sk.sk;}
	else{wpsiH_projection_simple_minus+=w;sk_dI_minus+=w*sk.sk;}
	final double pt=TMatr.getAMatrA(n,ev_bQQdidt,n+k*n,bQQpt,n),
	    dpi=(m1-m0*pt);
	if(ev_bQQdidt[k]>dIsplit){pI_plus+=w*dpi;}
	else{pI_minus+=w*dpi;}

    }
    this.sk_dI_plus=sk_dI_plus/wpsiH_projection_simple_plus;
    this.sk_dI_minus=sk_dI_minus/wpsiH_projection_simple_minus;
    this.pI_plus=pI_plus;
    this.pI_minus=pI_minus;
    if(eps>0)
    {
	final DIselIH diselih=new DIselIH(
	    n,DIselIH.METHOD_FULL_EV_QQDEV,
	    EVXData.convertMatrFromBtoQ_Basis(n,bQQdidt,I.ev,QQt),
	    p_offset,p_offset,QQt,QQi,QQpi,QQt,M,eps);
	UnitTests.checkDoubleRel(diselih.sIH_plus.wi,"diselih.sIH_plus.wi",skIH_plus.w1+skIH_plus.w2,"skIH_plus.w1+skIH_plus.w2",eps);
	UnitTests.checkDoubleRel(diselih.sIH_plus.pi,"diselih.sIH_plus.pi",skIH_plus.xa,"skIH_plus.xa",eps);
	UnitTests.checkDoubleRel(diselih.sIH_minus.wi,"diselih.sIH_minus.wi",skIH_minus.w1+skIH_minus.w2,"skIH_minus.w1+skIH_minus.w2",eps);
	UnitTests.checkDoubleRel(diselih.sIH_minus.pi,"diselih.sIH_minus.pi",skIH_minus.xa,"skIH_minus.xa",eps);
	UnitTests.checkDoubleRel(diselih.wpsiH_projection_simple_plus,"diselih.wpsiH_projection_simple_plus",wpsiH_projection_simple_plus,"wpsiH_projection_simple_plus",eps);
	UnitTests.checkDoubleRel(diselih.wpsiH_projection_simple_minus,"diselih.wpsiH_projection_simple_minus",wpsiH_projection_simple_minus,"wpsiH_projection_simple_minus",eps);
    }
    final int inddi=com.polytechnik.utils.SortArrays.sortArrayIndexes(var2I)[0];
    this.sk_dim=Skewness.getSkewnessFromMonomialMoments(
	bQQi[inddi+n*inddi],
	bQQpi[inddi+n*inddi],
	bQQp2i[inddi+n*inddi],
	bQQp3i[inddi+n*inddi]);
    this.var2p_dim=var2p[inddi];

    if(eps>0){
	System.err.println("DIselIHskewness: OK n="+n+" I="+java.util.Arrays.toString(I.getEValues())+" ev_bQQdidt="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_bQQdidt,n))+" eps="+eps);
    }
}

public static void matchDIselIHskewness(final DIselIHskewness a,final DIselIHskewness b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.var2p_dim,"a.var2p_dim",b.var2p_dim,"b.var2p_dim",eps);
    Skewness.matchSkewness(a.skIH.getSameSkewnessXShifted(a.p_offset),b.skIH.getSameSkewnessXShifted(b.p_offset),eps);
    Skewness.matchSkewness(a.sk_dim.getSameSkewnessXShifted(a.p_offset),b.sk_dim.getSameSkewnessXShifted(b.p_offset),eps);
    Skewness.matchSkewness(a.skIH_plus.getSameSkewnessXShifted(a.p_offset),b.skIH_plus.getSameSkewnessXShifted(b.p_offset),eps);
    Skewness.matchSkewness(a.skIH_minus.getSameSkewnessXShifted(a.p_offset),b.skIH_minus.getSameSkewnessXShifted(b.p_offset),eps);
    System.err.println("DIselIHskewness.matchDIselIHskewness test OK n="+a.I.neq+" eps="+eps);
}

}
