package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.ProbabilityCorrelation;

/** Try (P-p_last)^2/I -> min state.
 * @version $Id: DIselP2divI.java,v 1.14 2022/08/08 20:10:46 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselP2divI {
    public final EVXData I;
    public final double p_offset,
    // The https://arxiv.org/abs/1510.05510 answers.
	p_IH,
    // other answers
	p_ppl2divI_min,var0_pi,var1_pi,pi,pV,V
	;
    final ProbabilityCorrelation pc;

public DIselP2divI(final int n,
		   final double p_offset,final double p_last,
		   final double [] QQt,final double [] QQi,
		   final double [] QQpi,final double [] QQp2i,
		   final double [] QQV,final double [] QQpV,
		   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		   final VolMatrix VOLATILITY_EDPsi,
		   final double eps)
{
    this.p_offset=p_offset;
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);

    final double [] bQQi=TMatr.getDiagMatrix(I.getEValues()),UNIT=TMatr.getUnitMatrix(n),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),bQQppli=bQQpi.clone(),
	bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,QQp2i,I.ev),bQQppl2i=bQQp2i.clone();
    this.p_IH=bQQpi[ixH+n*ixH]/I.ev[ixH];
    final double pl_m_poffset=p_last;// p_offset is already included in p_last;
    com.polytechnik.utils.Polynomials.shiftPOffsetInPlace(n*n,bQQi,bQQppli,bQQppl2i,null,pl_m_poffset);
    final double [] invbQQi=new double [n*n],invSqrtbQQi=new double [n*n];
    for(int k=0;k<n;k++){
	invbQQi[k+n*k]=1/I.ev[k];
	invSqrtbQQi[k+n*k]=1/Math.sqrt(I.ev[k]);
    }
    final double [] bQQppldivI=TMatr.mulMMM(n,invbQQi,bQQppli,invbQQi),
	bQQppl2divI=TMatr.mulMMM(n,invbQQi,bQQppl2i,invbQQi);
    final double [] ev_bQQppl2divI=M.EV.getEV(n,bQQppl2divI,true),
	bpsi=EVXData.getPsiFromBasis(0,ev_bQQppl2divI,n);
    final double [] psi=EVXData.convertPsiFromBtoQ_Basis(n,bpsi,I.ev),
	Dpsi=VOLATILITY_EDPsi!=null?VOLATILITY_EDPsi.getDPsi(psi):new double [n],
	bDpsi=EVXData.convertPsiFromQtoB_Basis(n,Dpsi,I.ev,QQt);
    final double I_ppl2divI_min=TMatr.getAMatrA(bpsi,bQQi,n);
    p_ppl2divI_min=TMatr.getAMatrA(bpsi,bQQpi,n)/I_ppl2divI_min;
    final double [] rqpplI=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(bpsi,bDpsi,null,bQQppli,UNIT,n);
    pi=TMatr.getAMatrA(bpsi,bQQpi,n)/TMatr.getAMatrA(bpsi,bQQi,n);
    pV=TMatr.getAMatrA(psi,QQpV,n)/TMatr.getAMatrA(psi,QQV,n);
    V=TMatr.getAMatrA(psi,QQV,n);
    var0_pi=rqpplI[0];
    var1_pi=rqpplI[1];

    if(eps>0){
	final double [] rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(psi,Dpsi,null,EVXData.convertMatrFromBtoQ_Basis(n,bQQppli,I.ev,QQt),QQt,n);
	UnitTests.checkDoubleRel(rqpplI[0],"rqpplI[0]",rq[0],"rq[0]",eps);
	UnitTests.checkDoubleRel(rqpplI[1],"rqpplI[1]",rq[1],"rq[1]",eps);
	UnitTests.checkDoubleRel(rqpplI[2],"rqpplI[2]",rq[2],"rq[2]",eps);
    }

    final double [] bQQp=TMatr.mulMMM(n,invSqrtbQQi,bQQpi,invSqrtbQQi),
	bDpsiort=TMatr.sumAcB(bDpsi,-TMatr.mulVV(bDpsi,bpsi),bpsi);
    pc=new ProbabilityCorrelation(
	    new double [] {
		TMatr.getAMatrB(bpsi,bpsi,UNIT,n),
		TMatr.getAMatrB(bpsi,bDpsi,UNIT,n),
		TMatr.getAMatrB(bDpsi,bpsi,UNIT,n),
		TMatr.getAMatrB(bDpsi,bDpsi,UNIT,n)
	    },new double []{
		TMatr.getAMatrB(bpsi,bpsi,bQQp,n),
		TMatr.getAMatrB(bpsi,bDpsi,bQQp,n),
		TMatr.getAMatrB(bDpsi,bpsi,bQQp,n),
		TMatr.getAMatrB(bDpsi,bDpsi,bQQp,n)
	    },
	    new double []{
		TMatr.getAMatrB(bpsi,bpsi,bQQi,n),
		TMatr.getAMatrB(bpsi,bDpsi,bQQi,n),
		TMatr.getAMatrB(bDpsi,bpsi,bQQi,n),
		TMatr.getAMatrB(bDpsi,bDpsi,bQQi,n)
	    });
    if(eps>0){
	UnitTests.checkDoubleRel(TMatr.getAMatrB(bDpsiort,bpsi,UNIT,n),"<bpsi|bDpsiort>",0,"0",eps);
	final double [] QQp=EVXData.convertMatrFromBtoQ_Basis(n,bQQp,I.ev,QQt);
	final ProbabilityCorrelation pco=new ProbabilityCorrelation(
	    new double [] {
		TMatr.getAMatrB(psi,psi,QQt,n),
		TMatr.getAMatrB(psi,Dpsi,QQt,n),
		TMatr.getAMatrB(Dpsi,psi,QQt,n),
		TMatr.getAMatrB(Dpsi,Dpsi,QQt,n)
	    },new double []{
		TMatr.getAMatrB(psi,psi,QQp,n),
		TMatr.getAMatrB(psi,Dpsi,QQp,n),
		TMatr.getAMatrB(Dpsi,psi,QQp,n),
		TMatr.getAMatrB(Dpsi,Dpsi,QQp,n)
	    },
	    new double []{
		TMatr.getAMatrB(psi,psi,QQi,n),
		TMatr.getAMatrB(psi,Dpsi,QQi,n),
		TMatr.getAMatrB(Dpsi,psi,QQi,n),
		TMatr.getAMatrB(Dpsi,Dpsi,QQi,n)
	    });
	ProbabilityCorrelation.matchProbabilityCorrelation(pc,pco,eps);
    }

    if(eps>0){
	System.err.println("DIselP2divI: test OK n="+n+" eps="+eps);
    }
}

}
