package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.ProbabilityCorrelation;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** P and I probability correlations in the state of max di.
 * @version $Id: DIselPIProbabilityCorrelation.java,v 1.9 2022/08/08 20:10:46 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselPIProbabilityCorrelation {
    public final EVXData I;
    public final double p_offset,
    // The https://arxiv.org/abs/1510.05510 answers.
	p_IH,
    // new answer
	p_dim
	;
public final ProbabilityCorrelation pc;


public DIselPIProbabilityCorrelation(final int n,
		       final double p_offset,
		       final double [] QQt,final double [] QQi,
		       final double [] QQpm,final double [] QQm,
		       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		       final VolMatrix VOLATILITY_EDPsi,
		       final double eps)
{
    this.p_offset=p_offset;
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] var2I=new double [n],normdd=new double [n];
    for(int k=0;k<n;k++){
	final double [] bDpsi=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(I.getPsi(k)),I.ev,QQt);
	bDpsi[k]=0; // remove psik projection, now bpsi and bDpsi are orthogonal 2D eigenvalues solution
	final double dd=normdd[k]=TMatr.mulVV(bDpsi,bDpsi);
	var2I[k]=TMatr.mulV2V(bDpsi,I.getEValues())/dd-I.ev[k];
    }
    final double [] bQQp;
    if(QQm==QQt){
	bQQp=EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev);
    } else if(QQm==QQi) {
	final double [] invsqrtI=new double [n*n];
	for(int k=0;k<n;k++) invsqrtI[k+n*k]=1/Math.sqrt(I.ev[k]);
	bQQp=TMatr.mulMMM(n,invsqrtI,EVXData.convertMatrFromQtoB_Basis(n,QQpm,I.ev),invsqrtI);
    } else throw new IllegalArgumentException("only QQm=QQi or QQm=QQt are supported");
    this.p_IH=bQQp[ixH+n*ixH];

    final int inddim=com.polytechnik.utils.SortArrays.sortArrayIndexes(var2I)[0];
    this.p_dim=bQQp[inddim+n*inddim];
    final double [] a=new double [n],b=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(I.getPsi(inddim)),I.ev,QQt);
    a[inddim]=1;
    b[inddim]=0;
    final double [] bQQi=TMatr.getDiagMatrix(I.getEValues()),UNIT=TMatr.getUnitMatrix(n);
    this.pc=new ProbabilityCorrelation(
	new double []{
	    TMatr.getAMatrB(a,a,UNIT,n),
	    TMatr.getAMatrB(a,b,UNIT,n),
	    TMatr.getAMatrB(b,a,UNIT,n),
	    TMatr.getAMatrB(b,b,UNIT,n)
	},
	new double []{
	    TMatr.getAMatrB(a,a,bQQp,n),
	    TMatr.getAMatrB(a,b,bQQp,n),
	    TMatr.getAMatrB(b,a,bQQp,n),
	    TMatr.getAMatrB(b,b,bQQp,n)
	},
	new double []{
	    TMatr.getAMatrB(a,a,bQQi,n),
	    TMatr.getAMatrB(a,b,bQQi,n),
	    TMatr.getAMatrB(b,a,bQQi,n),
	    TMatr.getAMatrB(b,b,bQQi,n)
	});
    if(eps>0){
	UnitTests.checkDoubleRel(p_IH,"p_IH",TMatr.getAMatrA(I.psiH,QQpm,n)/TMatr.getAMatrA(I.psiH,QQm,n),"p_IH",eps);
	UnitTests.checkDoubleRel(pc.iL-pc.iH,"pc.iL-pc.iH",var2I[inddim],"var2I[inddim]",eps);
    }

    if(eps>0){
	System.err.println("DIselPIProbabilityCorrelation: test OK n="+n+" eps="+eps);
    }
}
}
