package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Simplistic version of selected basis {@literal |psiH>} and {@literal |dS/dt|psiH>}.
 * In contrast with the {@link DIselIH}, here we find the min/max of price, not (p-p_IH)*I.
 * @version $Id: DIselPsi0.java,v 1.9 2022/07/01 17:28:19 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DIselPsi0 {
public final EVXData I;
public final double p_offset,
// The https://arxiv.org/abs/1510.05510 answers.
    p_IH,p_0f,dEVseparation_0,
// New answers
    wpsi0f_projection_simple_plus,
    wpsi0f_projection_simple_minus,
    GammaPI
    ;

/**
 * @param n The dimension.
 * @param QQt {@literal <Q_j|Q_k>}.
 * @param QQi {@literal <Q_j|I|Q_k>}.
 * @param QQpi {@literal <Q_j|pI|Q_k>}.
 * @param M basis operations.
 * @param eps Run a selftest with this eps if &gt;0.
 */
public DIselPsi0(final int n,
		  final boolean flag_Add_Impact_From_The_future,
		  final double p_offset,final double p_last,
		  final double [] QQt,final double [] QQi,final double [] QQpi,
		  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		  final double eps)
{
    this.p_offset=p_offset;
    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1; // maximal eigenvalue index
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];
    final double I_0=I.s0,I_0_feq=flag_Add_Impact_From_The_future?I.sH:I_0,dI=I_0_feq-I_0,
	dV=dI/I.o_o,Pf=p_last/* does have the offset. */;
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	QQi_feq=TMatr.sumAcB(QQi,dV,QQ0),
	QQpi_feq=TMatr.sumAcB(QQpi,Pf*dV,QQ0),
	bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)) /*EVXData.convertPsiFromQtoB_Basis(n,I.psi0,QQt,I.ev) */,
	bQQi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQi_feq,I.ev),
	bQQpi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQpi_feq,I.ev),
	bPsi=bpsi0;
    this.p_0f=TMatr.getAMatrA(bPsi,bQQpi_feq,n)/TMatr.getAMatrA(bPsi,bQQi_feq,n);
    double wpsi0f_projection_simple_plus=0,wpsi0f_projection_simple_minus=0;
    final double [] bQQDEVseparation=TMatr.sumaAbB(p_0f,bQQi_feq,-1,bQQpi_feq),
	// Diagonalize dS:
	ev_bQQDEVseparation=M.EV.getEV(n,bQQDEVseparation,true);
    this.dEVseparation_0=TMatr.getAMatrA(bPsi,bQQDEVseparation,n);
    // Simplistic try (|psi0> projection).
    for(int k=0;k<n;k++){
	final double w=Math.pow(DeprecatedHelperTMatr.mulVVo(n,ev_bQQDEVseparation,n+k*n,bPsi,0),2);
	if(ev_bQQDEVseparation[k]>0){wpsi0f_projection_simple_plus+=w;}
	else{wpsi0f_projection_simple_minus+=w;}
    }
    this.wpsi0f_projection_simple_plus=wpsi0f_projection_simple_plus;
    this.wpsi0f_projection_simple_minus=wpsi0f_projection_simple_minus;
    this.GammaPI=(this.wpsi0f_projection_simple_plus-this.wpsi0f_projection_simple_minus);

    if(eps>0) {
	UnitTests.checkDoubleRel(this.wpsi0f_projection_simple_plus+this.wpsi0f_projection_simple_minus,"this.wpsi0f_projection_simple_plus+this.wpsi0f_projection_simple_minus",1,"1",eps);
	final double [] psi0f=I.psi0;
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi_feq,n),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bPsi,bQQDEVseparation,n),"<bPsi|QQDEVseparation|bPsi>",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi0f,QQt,n),"<psi0f|psi0f>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi0f,QQpi_feq,n),"<psi0f|QQpi_feq|psi0f>",p_0f*I_0_feq,"p_0f*I_0_feq",eps);
    }



    if(eps>0){
	System.err.println("DIselPsi0: OK n="+n+" I="+java.util.Arrays.toString(I.getEValues())+" flag_Add_Impact_From_The_future="+flag_Add_Impact_From_The_future+" GammaPI="+GammaPI+" dEVseparation_0="+this.dEVseparation_0+" eps="+eps);
    }


}


public static void matchDIselPsi0(final DIselPsi0 a,final DIselPsi0 b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.p_0f,"a.p_0f",b.p_0f+p_offset_shift,"b.p_0f+p_offset_shift",eps);

    UnitTests.checkDoubleRel(a.wpsi0f_projection_simple_plus,"a.wpsi0f_projection_simple_plus",b.wpsi0f_projection_simple_plus,"b.wpsi0f_projection_simple_plus",eps);
    UnitTests.checkDoubleRel(a.wpsi0f_projection_simple_minus,"a.wpsi0f_projection_simple_minus",b.wpsi0f_projection_simple_minus,"b.wpsi0f_projection_simple_minus",eps);
    UnitTests.checkDoubleRel(a.GammaPI,"a.GammaPI",b.GammaPI,"b.GammaPI",eps);

    System.err.println("DIselPsi0.matchDIselPsi0 test OK n="+a.I.neq+" eps="+eps);
}

}
