package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.LebesgueQuadratureWithEVData;

/** Density matrix for the state: "after max spike in I".
 * Consider the I=dV/dt and the state psiH(x) corresponding to max lambda_{I}.
 * Then obtain F(x)=\int_{-inf}^{x)psiH^2(x)omega(x)dx where F(tnow)=1 follow from psiH normalizing,
 * this F(x) can be density matrix expanded as F(x)=rho(x,x);
 * after the rho(x,y) is obtained - consider Spur(dp/dt|rho|dI/dt|rho)
 * and similar type of answers.
 * The Spur(dp/dt|rho) is euqual to p_last-pt_H, the https://arxiv.org/abs/1510.05510 result.
 *  @version $Id: DM_DI.java,v 1.34 2022/11/20 12:41:32 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class DM_DI {
public final EVXData I,DM;
    public final double p_offset,
    // The https://arxiv.org/abs/1510.05510 answers.
    p_IH,pt_IH,dpdt_IH,V_IH,T_IH,
    // The old "pure state"--like answers.
    dp_spurDMIH, // =p_last-pt_IH
    di_spurDMIH, // =I.s0-I.sH
	dp_wI_spurDMIH,dp_winvI_spurDMIH,
	dpdi_spurDMIH,dpI_spurDMIH,
    // The new "density matrix" answers,
    // for a pure state they are equal to the corresponding products of dp_spurDMIH and di_spurDMIH.
    dpdi_spur2DMIH,dpdp_spur2DMIH,didi_spur2DMIH,
    // separated I_low->I_high & I_high->I_low
    dpdi_spur2DMIH_LOWtoHIGH,dpdi_spur2DMIH_HIGHtoLOW,
    dpdi_spurDMIH_LOWtoHIGH,dpdi_spurDMIH_HIGHtoLOW,
    // with I-state, just a try
    dpI_spur2DMIH,dpdi_LOWtoHIGH,dpdi_HIGHtoLOW
    ;
    DM_DI(final int n,final double p_offset,final double p_last,
	  final double [] QQt,final double [] QQpt,
	  final double [] QQi,final double [] QQpi,
	  final double [] QQdpdt,
	  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	  final VolMatrix VOLATILITY_EDPsi,
	  final java.util.function.Function<double [],double []> getPsi2WIntegratedDt,
	  // not null only for unit test
	  final RMomentsData moms,
	  final double eps)
    {
	this.p_offset=p_offset;
	    // The https://arxiv.org/abs/1510.05510 answers:
    final int ixH=n-1;
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_IH=I.getFPsi(ixH,QQpi)/I.ev[ixH];
    this.pt_IH=I.getFPsi(ixH,QQpt);
    this.dpdt_IH=I.getFPsi(ixH,QQdpdt);

    // New answers
    final double I_0=I.s0,I_0_feq=I.sH,pI_0=TMatr.getAMatrA(I.psi0,QQpi,n);
    final double [] psiH=I.psiH,psiWeight=getPsi2WIntegratedDt.apply(psiH);
    if(psiWeight.length!=2*n-1) throw new IllegalArgumentException();
    final double [] UNIT=TMatr.getUnitMatrix(n),
	// Gram matrix inverse can be obtained without solving a linear system.
	IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT),
	//IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	DensityMatrix=M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,psiWeight,IQQt)),
	QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	QQdidt0=TMatr.sumaAbB(I_0,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQi)),
	QQdpidt0=TMatr.sumaAbB(pI_0,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQpi)),
	// convert QQdidt0, QQdpdt, and DensityMatrix to I.ev basis to avoid IQQt=G^{-1} required for non-orthogonal basis.
	bQQdidt0=EVXData.convertMatrFromQtoB_Basis(n,QQdidt0,I.ev),
	bQQdpidt0=EVXData.convertMatrFromQtoB_Basis(n,QQdpidt0,I.ev),
	bQQpt=EVXData.convertMatrFromQtoB_Basis(n,QQpt,I.ev),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),
	bQQdpdt=EVXData.convertMatrFromQtoB_Basis(n,QQdpdt,I.ev),
	bDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,DensityMatrix,I.ev),
	bQQinvI=new double [n*n];
    for(int k=0;k<n;k++) bQQinvI[k+n*k]=1/I.ev[k];
    // Spur calculation.

    this.dpdi_spur2DMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQdidt0,bDensityMatrix,bQQdpdt,bDensityMatrix}));
    // normalizing, a crude approach
    this.dpdp_spur2DMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQdpdt,bDensityMatrix,bQQdpdt,bDensityMatrix}));
    this.didi_spur2DMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQdidt0,bDensityMatrix,bQQdidt0,bDensityMatrix}));
    // pure state--like answers
    this.dp_spurDMIH=TMatr.getSpurMM(n,bQQdpdt,bDensityMatrix);
    this.dpI_spurDMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQi,bQQdpdt,bDensityMatrix}));
    this.dp_wI_spurDMIH=dpI_spurDMIH
	*TMatr.getSpurM(n,bDensityMatrix)/TMatr.getSpurMM(n,bQQi,bDensityMatrix);
    this.dp_winvI_spurDMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQinvI,bQQdpdt,bDensityMatrix}))
	*TMatr.getSpurM(n,bDensityMatrix)/TMatr.getSpurMM(n,bQQinvI,bDensityMatrix);

    this.di_spurDMIH=TMatr.getSpurMM(n,bQQdidt0,bDensityMatrix);
    this.dpdi_spurDMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQdidt0,bDensityMatrix,bQQdpdt}));
    // To match old calculations (time and volume of psiH state).
    this.V_IH=TMatr.getSpurMM(n,bDensityMatrix,bQQi);
    this.T_IH=TMatr.getSpurM(n,bDensityMatrix);

    // just a try
    this.dpI_spur2DMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQi,bDensityMatrix,bQQdpdt,bDensityMatrix}));
    if(eps>0)
    {
	// obtain the same spur answer without basis conversion
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,QQdidt0,IQQt),"spur(didt0)",0,"0",eps);
	final double _dpdi_spur2DMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix,IQQt,QQdpdt,IQQt,DensityMatrix})),
	    _dpI_spur2DMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQi,IQQt,DensityMatrix,IQQt,QQdpdt,IQQt,DensityMatrix})),
	    _dpI_spurDMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQi,IQQt,QQdpdt,IQQt,DensityMatrix})),
	    _di_spurDMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix})),
	    _dp_spurDMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpdt,IQQt,DensityMatrix})),
	    _T_IH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQt,IQQt,DensityMatrix})),
	    _V_IH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQi,IQQt,DensityMatrix}));

	UnitTests.checkDoubleRel(_dpdi_spur2DMIH,"_dpdi_spur2DMIH",dpdi_spur2DMIH,"dpdi_spur2DMIH",eps);
	UnitTests.checkDoubleRel(_dpI_spur2DMIH,"_dpI_spur2DMIH",dpI_spur2DMIH,"dpI_spur2DMIH",eps);
	UnitTests.checkDoubleRel(_dpI_spurDMIH,"_dpI_spurDMIH",dpI_spurDMIH,"dpI_spurDMIH",eps);
	UnitTests.checkDoubleRel(_di_spurDMIH,"_di_spurDMIH",di_spurDMIH,"di_spurDMIH",eps);
	UnitTests.checkDoubleRel(_dp_spurDMIH,"_dp_spurDMIH",dp_spurDMIH,"dp_spurDMIH",eps);
	UnitTests.checkDoubleRel(_di_spurDMIH,"_di_spurDMIH",I_0-I_0_feq,"I_0-I_0_feq",eps);
	UnitTests.checkDoubleRel(_dp_spurDMIH,"_dp_spurDMIH",p_last-pt_IH,"p_last-pt_IH",eps);
	// Match direct averaging, without density matrix used.
	//final RMomentsData moms=getRMoments(2*n-1,odata);
	UnitTests.checkDoubleRel(_dp_spurDMIH,"_dp_spurDMIH",TMatr.mulVVn1(psiWeight,moms.dpmom),"<dpmom>",eps);
	UnitTests.checkDoubleRel(_V_IH,"_V_IH",TMatr.mulVVn1(psiWeight,moms.imom),"<imom>",eps);
	UnitTests.checkDoubleRel(_T_IH,"_T_IH",TMatr.mulVVn1(psiWeight,moms.tmom),"<imom>",eps);
	UnitTests.checkDoubleRel(V_IH,"V_IH",TMatr.mulVVn1(psiWeight,moms.imom),"<imom>",eps);
	UnitTests.checkDoubleRel(T_IH,"T_IH",TMatr.mulVVn1(psiWeight,moms.tmom),"<tmom>",eps);
	final double [] bpsiH=new double[n];bpsiH[ixH]=1;
	final double psiHDMpsiH=TMatr.getAMatrA(psiH,DensityMatrix,n),
	    bpsiHDMpsiH=TMatr.getAMatrA(bpsiH,bDensityMatrix,n);
	System.err.println("psiHDMpsiH="+psiHDMpsiH+" bpsiHDMpsiH="+bpsiHDMpsiH);
	System.err.println("DM_DI: spur matches n="+n+" eps="+eps);
    }
    // need for unit test only
    this.DM=EVXData.getEVXDataFromQQ(n,DensityMatrix,QQt,M);

    if(eps>0)
    {
	// To match https://arxiv.org/abs/1807.06007 , Appendix A:
	final LebesgueQuadratureWithEVData lDM=M.getDensityMatrixProducingGivenPolynomial(n,psiWeight,QQt);
	UnitTests.checkDoubleArrayRel(lDM.f,"lDM.f",DM.getEValues(),"DM.ev",eps);
	UnitTests.checkDoubleArrayRel1(lDM.QQf,"lDM.QQf",DensityMatrix,"DensityMatrix",eps);
	EVXData.test_orthogonality("lDM",n,lDM.ev,lDM.QQf,lDM.QQ,eps);
    }

    // one more try, separate I_low->I_high & I_high->I_low
    final double [] ev_bQQdidt0=M.EV.getEV(n,bQQdidt0,true),
	bDensityMatrix_bQQdpdt_bDensityMatrix=TMatr.mulMatrices(n,new double [][]{bDensityMatrix,bQQdpdt,bDensityMatrix}),
	bQQdpdt_bDensityMatrix=TMatr.mulMatrices(n,new double [][]{bQQdpdt,bDensityMatrix}),
	bbQQdidt0_LOWtoHIGH=new double [n*n],bbQQdidt0_HIGHtoLOW=new double [n*n];
    for(int k=0;k<n;k++){
	final double lambda_didt0=ev_bQQdidt0[k];
	// in ev_bQQdidt0 basis the dI/dt is diagonal, split it on two matrices: LOWtoHIGH & HIGHtoLOW
	if(lambda_didt0>0){
	    bbQQdidt0_LOWtoHIGH[k+n*k]=lambda_didt0;
	}
	else {
	    bbQQdidt0_HIGHtoLOW[k+n*k]=lambda_didt0;
	}
    }
    // convert the matrices from ev_bQQdidt0 back to I.ev basis
    final double [] bQQdidt0_LOWtoHIGH=EVXData.convertMatrFromBtoQ_Basis(n,bbQQdidt0_LOWtoHIGH,ev_bQQdidt0,UNIT),
	bQQdidt0_HIGHtoLOW=EVXData.convertMatrFromBtoQ_Basis(n,bbQQdidt0_HIGHtoLOW,ev_bQQdidt0,UNIT);
    this.dpdi_spur2DMIH_LOWtoHIGH=TMatr.getSpurMM(n,bQQdidt0_LOWtoHIGH,bDensityMatrix_bQQdpdt_bDensityMatrix);
    this.dpdi_spur2DMIH_HIGHtoLOW=TMatr.getSpurMM(n,bQQdidt0_HIGHtoLOW,bDensityMatrix_bQQdpdt_bDensityMatrix);
    this.dpdi_spurDMIH_LOWtoHIGH=TMatr.getSpurMM(n,bQQdidt0_LOWtoHIGH,bQQdpdt_bDensityMatrix);
    this.dpdi_spurDMIH_HIGHtoLOW=TMatr.getSpurMM(n,bQQdidt0_HIGHtoLOW,bQQdpdt_bDensityMatrix);
    if(eps>0)
    {
	// FAILS
	//UnitTests.checkDoubleArrayRel1(TMatr.sumAB(n*n,bQQdpdt_LOWtoHIGH,bQQdpdt_HIGHtoLOW),"bQQdpdt_LOWtoHIGH+bQQdpdt_HIGHtoLOW",bQQdpdt,"bQQdpdt",eps);
    }

    double dpdi_LOWtoHIGH=0,dpdi_HIGHtoLOW=0;
    for(int k=0;k<n;k++){
	final double lambda_didt0=ev_bQQdidt0[k],
	    dp=TMatr.getAMatrA(n,ev_bQQdidt0,n+k*n,bQQdpdt,n),
	    w=TMatr.getAMatrA(n,ev_bQQdidt0,n+k*n,bDensityMatrix,n);
	if(lambda_didt0>0){dpdi_LOWtoHIGH+=lambda_didt0*dp*w*w;}
	else {dpdi_HIGHtoLOW+=lambda_didt0*dp*w*w;}
    }
    this.dpdi_LOWtoHIGH=dpdi_LOWtoHIGH;
    this.dpdi_HIGHtoLOW=dpdi_HIGHtoLOW;
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.sumElements(java.util.Arrays.copyOf(ev_bQQdidt0,n)),"sum(ev_bQQdidt0)",0,"0",eps);
	UnitTests.checkDoubleRel(dpdi_spur2DMIH_LOWtoHIGH+dpdi_spur2DMIH_HIGHtoLOW,"dpdi_spur2DMIH_LOWtoHIGH+dpdi_spur2DMIH_HIGHtoLOW",dpdi_spur2DMIH,"dpdi_spur2DMIH",eps);
	UnitTests.checkDoubleRel(dpdi_spurDMIH_LOWtoHIGH+dpdi_spurDMIH_HIGHtoLOW,"dpdi_spurDMIH_LOWtoHIGH+dpdi_spurDMIH_HIGHtoLOW",dpdi_spurDMIH,"dpdi_spurDMIH",eps);
	//UnitTests.checkDoubleRel(dp_spurDMIH_LOWtoHIGH+dp_spurDMIH_HIGHtoLOW,"dp_spurDMIH_LOWtoHIGH+dp_spurDMIH_HIGHtoLOW",dp_spurDMIH,"dp_spurDMIH",eps); // FAILS
	// calculate same thing without basis conversion
	double _dpdi_spur2DMIH_LOWtoHIGH=0,_dpdi_spur2DMIH_HIGHtoLOW=0;
	for(int k=0;k<n;k++){
	    final double lambda_didt0=ev_bQQdidt0[k],
	    b=TMatr.getAMatrA(n,ev_bQQdidt0,n+k*n,bDensityMatrix_bQQdpdt_bDensityMatrix,n);
	    if(lambda_didt0>0){
		_dpdi_spur2DMIH_LOWtoHIGH+=b*lambda_didt0;
	    } else {
		_dpdi_spur2DMIH_HIGHtoLOW+=b*lambda_didt0;
	    }
	}
	UnitTests.checkDoubleRel(dpdi_spur2DMIH_LOWtoHIGH,"dpdi_spur2DMIH_LOWtoHIGH",_dpdi_spur2DMIH_LOWtoHIGH,"_dpdi_spur2DMIH_LOWtoHIGH",eps);
	UnitTests.checkDoubleRel(dpdi_spur2DMIH_HIGHtoLOW,"dpdi_spur2DMIH_HIGHtoLOW",_dpdi_spur2DMIH_HIGHtoLOW,"_dpdi_spur2DMIH_HIGHtoLOW",eps);
    }

    final double []
	pI_in_didt0=TMatr.getMatrixDiagElementsInBasis(n,ev_bQQdidt0,bQQpi),
	pt_in_didt0=TMatr.getMatrixDiagElementsInBasis(n,ev_bQQdidt0,bQQpt),
	dpidt0_in_didt0=TMatr.getMatrixDiagElementsInBasis(n,ev_bQQdidt0,bQQdpidt0),
	I_in_didt0=TMatr.getMatrixDiagElementsInBasis(n,ev_bQQdidt0,bQQi),
	rho_in_didt0=TMatr.getMatrixDiagElementsInBasis(n,ev_bQQdidt0,bDensityMatrix);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.mulVVn1(rho_in_didt0,ev_bQQdidt0),"sum<lambda <psi|rho|psi>",I_0-I_0_feq,"I_0-I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(rho_in_didt0),"sum(rho_in_didt0)",T_IH,"T_IH",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(java.util.Arrays.copyOf(ev_bQQdidt0,n)),"sum(ev_bQQdidt0)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQdidt0),"spur(bQQdidt0)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQdpidt0),"spur(bQQdpidt0)",0,"0",eps);
	UnitTests.checkDoubleRel(bQQpt[ixH+n*ixH],"bQQpt[ixH+n*ixH]",pt_IH,"pt_IH",eps);
	final EVXData DIdt0=EVXData.getEVXDataFromQQ(n,QQdidt0,QQt,M);
	for(int k=0;k<n;k++){
	    final double [] psi=EVXData.convertPsiFromBtoQ_Basis(n,EVXData.getPsiFromBasis(k,ev_bQQdidt0,n),I.ev);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,QQpt,n),"pt",pt_in_didt0[k],"pt_in_didt0[k]",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,QQdpidt0,n),"dpi",dpidt0_in_didt0[k],"dpi_in_didt0[k]",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,DensityMatrix,n),"rho",rho_in_didt0[k],"rho_in_didt0[k]",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,QQdidt0,n),"didt0",ev_bQQdidt0[k],"ev_bQQdidt0[k]",eps);
	    UnitTests.checkDoubleRel(ev_bQQdidt0[k],"ev_bQQdidt0[k]",DIdt0.ev[k],"DIdt0",eps);
	    UnitTests.checkDoubleArrayRel1(EVXData.getMSign(psi),"psi",EVXData.getMSign(DIdt0.getPsi(k)),"DIdt0.getPsi(k)",eps);
	}
	System.err.println("DM_DI: spur dpidt0 matches n="+n+" eps="+eps);
    }

    double dpi_LOWtoHIGH_rho=0,dpi_HIGHtoLOW_rho=0,dpi_LOWtoHIGH=0,dpi_HIGHtoLOW=0;
    for(int k=0;k<n;k++){
	final double lambda_didt0=ev_bQQdidt0[k],w=rho_in_didt0[k],
	    c=dpidt0_in_didt0[k]-lambda_didt0*pt_in_didt0[k];
	if(lambda_didt0>0){
	    dpi_LOWtoHIGH_rho+=c*w;
	    dpi_LOWtoHIGH+=c;
	} else {
	    dpi_HIGHtoLOW_rho+=c*w;
	    dpi_HIGHtoLOW+=c;
	}
    }
    this.dpi_LOWtoHIGH_rho=dpi_LOWtoHIGH_rho;
    this.dpi_HIGHtoLOW_rho=dpi_HIGHtoLOW_rho;
    this.dpi_LOWtoHIGH=dpi_LOWtoHIGH;
    this.dpi_HIGHtoLOW=dpi_HIGHtoLOW;
    double pit_fplus=0,pit_fminus=0,pitDM_fplus=0,pitDM_fminus=0;
    for(int k=0;k<n;k++){
	final double lambda_didt0=ev_bQQdidt0[k],w=rho_in_didt0[k],
	    c=pI_in_didt0[k]-I_in_didt0[k]*pt_in_didt0[k];
	if(lambda_didt0>0){
	    pit_fplus+=c;
	    pitDM_fplus+=c*w;
	}
	else {
	    pit_fminus+=c;
	    pitDM_fminus+=c*w;
	}
    }
    this.pit_fplus=pit_fplus;
    this.pit_fminus=pit_fplus;
    this.pitDM_fplus=pit_fplus;
    this.pitDM_fminus=pit_fplus;
    this.DensityMatrix_psiHintegrated=DensityMatrix;
}
    final double dpi_LOWtoHIGH_rho,dpi_HIGHtoLOW_rho,dpi_LOWtoHIGH,dpi_HIGHtoLOW;
    final double pit_fplus,pit_fminus,pitDM_fplus,pitDM_fminus;
    final double [] DensityMatrix_psiHintegrated;

static void matchDM_DI(final DM_DI a,final DM_DI b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    EVXData.matchEVXData(a.DM,b.DM,false,eps);
    UnitTests.checkDoubleRel(a.dpdt_IH,"a.dpdt_IH",b.dpdt_IH,"b.dpdt_IH",eps);
    UnitTests.checkDoubleRel(a.T_IH,"a.T_IH",b.T_IH,"b.T_IH",eps);
    UnitTests.checkDoubleRel(a.V_IH,"a.V_IH",b.V_IH,"b.T_IH",eps);
    UnitTests.checkDoubleRel(a.dpdi_spur2DMIH,"a.dpdi_spur2DMIH",b.dpdi_spur2DMIH,"b.dpdi_spur2DMIH",eps);
    UnitTests.checkDoubleRel(a.dpdi_spur2DMIH_LOWtoHIGH,"a.dpdi_spur2DMIH_LOWtoHIGH",b.dpdi_spur2DMIH_LOWtoHIGH,"b.dpdi_spur2DMIH_LOWtoHIGH",eps);
    UnitTests.checkDoubleRel(a.dpdi_spur2DMIH_HIGHtoLOW,"a.dpdi_spur2DMIH_HIGHtoLOW",b.dpdi_spur2DMIH_HIGHtoLOW,"b.dpdi_spur2DMIH_HIGHtoLOW",eps);
    UnitTests.checkDoubleRel(a.dpI_spur2DMIH,"a.dpI_spur2DMIH",b.dpI_spur2DMIH,"b.dpI_spur2DMIH",eps);

    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.pt_IH,"a.pt_IH",b.pt_IH+p_offset_shift,"b.pt_IH+p_offset_shift",eps);
    //UnitTests.checkDoubleRel(a.p_average,"a.p_average",b.p_average+p_offset_shift,"b.p_average+p_offset_shift",eps);
    //UnitTests.checkDoubleRel(a.pt_average,"a.pt_average",b.pt_average+p_offset_shift,"b.pt_average+p_offset_shift",eps);

    System.err.println("DM_DI.matchDM_DI test OK n="+a.I.neq+" eps="+eps);
}
}
