package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.*;
/** A container to store directional information.
 * Given attribute matrix and density matrix the direction {@link #dir} is the Spur of their product.
 * Normalizing the {@link #dir} requires either to diagonalize attribute matrix
 * as for {@link #dirNorm} or doing spur
 * of matrix products as {@link #dirNormSK} using
 * a quadrature {@link #sk}.
 * @version $Id: DirectionalInfoWithQuadrature.java,v 1.23 2022/08/08 20:10:46 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class DirectionalInfoWithQuadrature {

    /** Directional info: {@code Spur(bQQf*bQQDensityMatrix)}.
     * Can be expressed as {@link sk}.x1*{@link sk}.w1+{@link sk}.x2*{@link sk}.w2.
     */
    public final double dir;

    /** A norm for {@link #dir} as {@code Spur(bQQfabs*bQQDensityMatrix) }
     * where {@code bQQfabs} is {@code bQQf} with all eigenvalues set to their absolute values.
     * This norm is slow to calculate as it requires solving an eigenvalue problem,
     * one can use {@link #dirNormSK} as a (slightly worse) fast alternative.
     */
    public final double dirNorm;

    /** The {@code Spur(bQQf*bQQf*bQQDensityMatrix)}.
     */
    public final double mom2;

    /** The {@code Spur(bQQf*bQQf*bQQf*bQQDensityMatrix)}.
     */
    public final double mom3;

    /** Attribute matrix eigenvalues.
     * For use in unit tests and etc.
     */
    public final double [] attributeEValues;

    /** The Spur of input matrix.
     */
    public final double densityMatrixSpur,attributeSpur;


    /** The {@link Skewness} calculated on the moments:
     * {@link #densityMatrixSpur}, {@link #dir}, {@link #mom2}, {@link mom3}.
     */
    public final Skewness sk;


    /** A norm for {@link #dir} as sqrt(Spur(bQQf*bQQf*bQQDensityMatrix)*Spur(bQQDensityMatrix)),
     * specifically {@code dirNormW}=sqrt({@link #mom2}*{@link #densityMatrixSpur}).
     */
    public final double dirNormW;


    /** A norm for {@link #dir}. Calculated as
     * abs({@link sk}.x1*{@link sk}.w1)+abs({@link sk}.x2*{@link sk}.w2),
     * note that {@link #dir}={@link sk}.x1*{@link sk}.w1+{@link sk}.x2*{@link sk}.w2.
     */
    public final double dirNormSK;


    /** Whether the {@link #dirNorm} and {@link attributeEValues}
     * are available.
     */
    public final boolean flagUseEstimateWithEValues;

    /** Obtain directional information from attribute matrix and density matrix.
     * @param n Basis dimenstion.
     * @param bQQf Attribute matrix in an orthonormal basis.
     * @param bQQDensityMatrix State density matrix in the same basis.
     * @param flagUseEstimateWithEValues If false -- the eigenvalue problem is not solved hence
     * {@link #dirNorm} and {@link attributeEValues} are not availbale, use {@link #dirNormSK} instead.
     * @param EV Eigenproblem solver.
     * @param eps Perform a unit test with this eps if positive.
     */
    DirectionalInfoWithQuadrature(final int n,
				  final double [] bQQf,
				  final double [] bQQDensityMatrix,
				  final boolean flagUseEstimateWithEValues,
				  final EVSolver EV,
				  final double eps){

	// Regular spur.
	this.attributeSpur=TMatr.getSpurM(n,bQQf);
	// Moments with bQQDensityMatrix
	this.densityMatrixSpur=TMatr.getSpurM(n,bQQDensityMatrix);
	// Directional information
	this.dir=TMatr.getSpurMM(n,bQQf,bQQDensityMatrix);
	// High order moments
	this.mom2=TMatr.getSpurMMM(n,bQQf,bQQf,bQQDensityMatrix);
	this.mom3=TMatr.getSpurMM(n,TMatr.mulMMM(n,bQQf,bQQf,bQQf),bQQDensityMatrix);
	this.sk=Skewness.getSkewnessFromMonomialMoments(densityMatrixSpur,dir,mom2,mom3);
	this.dirNormW=Math.sqrt(mom2*densityMatrixSpur);
	this.dirNormSK=Math.abs(sk.x1*sk.w1)+Math.abs(sk.x2*sk.w2);

	this.flagUseEstimateWithEValues=flagUseEstimateWithEValues;

	if(flagUseEstimateWithEValues){
	    // Normalize directional info
	    final double [] ev_bQQf=EV.getEV(n,bQQf,true);
	    this.attributeEValues=java.util.Arrays.copyOf(ev_bQQf,n);
	    // convert density matrix to this new basis
	    final double [] bbQQDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,bQQDensityMatrix,ev_bQQf);
	    double sNormDir=0,sDir=0;
	    for(int k=0;k<n;k++){
		sDir+=attributeEValues[k]*bbQQDensityMatrix[k+n*k];
		// For the norm treat all eigenvalues of bQQf as positive.
		sNormDir+=Math.abs(attributeEValues[k])*bbQQDensityMatrix[k+n*k];
	    }
	    this.dirNorm=sNormDir;

	    // If no eigenvalues -- these tests do not make any sense.
	    if(eps>0){
		// basis invariance check
		UnitTests.checkDouble(attributeSpur,"attributeSpur",TMatr.sumElements(attributeEValues),"sum(attributeEValues)",eps);
		UnitTests.checkDoubleRel(densityMatrixSpur,"densityMatrixSpur",TMatr.getSpurM(n,bbQQDensityMatrix),"Spur bbQQDensityMatrix",eps);
		UnitTests.checkDoubleRel(dir,"dir",TMatr.getSpurMM(n,TMatr.getDiagMatrix(attributeEValues),bbQQDensityMatrix),"Spur diagF*bbQQDensityMatrix",eps);
		UnitTests.checkDoubleRel(dir,"dir",sDir,"sDir",eps);
		UnitTests.checkDoubleRel(mom2,"mom2",TMatr.mulV2V(attributeEValues,TMatr.getMatrixDiagElements(n,bbQQDensityMatrix)),"Spur diagF*diagF*bbQQDensityMatrix",eps);
		UnitTests.checkDoubleRel(mom3,"mom3",TMatr.mulVVV(attributeEValues,TMatr.getVAB(attributeEValues,attributeEValues),TMatr.getMatrixDiagElements(n,bbQQDensityMatrix)),"Spur diagF*diagF*diagF*bbQQDensityMatrix",eps);
	    }
	} else {
	    // No eigenproblem to use
	    this.dirNorm=Double.NaN;
	    this.attributeEValues=new double [n];
	    java.util.Arrays.fill(this.attributeEValues,Double.NaN);
	}
	if(eps>0){
	    // Quadrature check
	    UnitTests.checkDoubleRel(densityMatrixSpur,"densityMatrixSpur",sk.w1+sk.w2,"sk.w1+sk.w2",eps);
	    UnitTests.checkDoubleRel(dir,"dir",sk.x1*sk.w1+sk.x2*sk.w2,"sk.x1*sk.w1+sk.x2*sk.w2",eps);
	    UnitTests.checkDoubleRel(mom2,"mom2",Math.pow(sk.x1,2)*sk.w1+Math.pow(sk.x2,2)*sk.w2,"sk.x1^2*sk.w1+sk.x2^2*sk.w2",eps);
	    UnitTests.checkDoubleRel(mom3,"mom3",Math.pow(sk.x1,3)*sk.w1+Math.pow(sk.x2,3)*sk.w2,"sk.x1^3*sk.w1+sk.x2^3*sk.w2",eps);

	    System.err.println("DirectionalInfo: test OK n="+n+" dir="+dir+" dirNorm="+dirNorm+" dirNormSK="+dirNormSK+" dirNormW="+dirNormW+" densityMatrixSpur="+densityMatrixSpur+" attributeSpur="+attributeSpur+" attributeEValues="+java.util.Arrays.toString(attributeEValues)+" eps="+eps);
	}
    }

     /** For unit test.
     */
    public static void match(final DirectionalInfoWithQuadrature a,final DirectionalInfoWithQuadrature b,final double eps){
	match(a,b,true,eps);
    }

    /** For unit test.
     */
    public static void match(final DirectionalInfoWithQuadrature a,final DirectionalInfoWithQuadrature b,final boolean flagUseEstimateWithEValues,final double eps){

	UnitTests.checkDouble(a.attributeSpur,"a.attributeSpur",b.attributeSpur,"b.attributeSpur",eps);

	// moments
	UnitTests.checkDoubleRel(a.densityMatrixSpur,"a.densityMatrixSpur",b.densityMatrixSpur,"b.densityMatrixSpur",eps);
	UnitTests.checkDoubleRel(a.dir,"a.dir",b.dir,"b.dir",eps);
	UnitTests.checkDoubleRel(a.mom2,"a.mom2",b.mom2,"b.mom2",eps);
	UnitTests.checkDoubleRel(a.mom3,"a.mom3",b.mom3,"b.mom3",eps);

	// skewness from these moments
	Skewness.matchSkewness(a.sk,b.sk,eps);
	UnitTests.checkDoubleRel(a.dirNormSK,"a.dirNormSK",b.dirNormSK,"b.dirNormSK",eps);
	UnitTests.checkDoubleRel(a.dirNormW,"a.dirNormW",b.dirNormW,"b.dirNormW",eps);

	if(flagUseEstimateWithEValues){
	    UnitTests.checkDoubleRel(a.dirNorm,"a.dirNorm",b.dirNorm,"b.dirNorm",eps);
	    UnitTests.checkDoubleArrayRel1(a.attributeEValues,"a.attributeEValues",b.attributeEValues,"b.attributeEValues",eps);
	}
	System.err.println("DirectionalInfoWithQuadrature.match: test OK n="+a.attributeEValues.length+" eps="+eps);
    }
}
