package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;

/** More John Pierpont Morgan and finite variations.
 *  @version $Id: Distance.java,v 1.68 2022/11/19 11:08:57 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 *  Note: According to unit test results on random data
 *  transitivity of the comparison based on distmatr may not hold (very seldom).
 *  This is a problem to discuss.
 */
public abstract class Distance {

    final int [] t_order;
    final double [] evv_distancesorted;
    Distance(final int n,final double [] basis,final int basisLD,final int basisOff)
    {
	final double [] distmatr=getDistMatr(n,basis,basisLD,basisOff);

	// Sort eigenvectors according to distmatr distance.
	this.t_order=com.polytechnik.utils.SortArrays.sortArrayIndexes(
	    n,new com.polytechnik.utils.IntPTRCompare(){
		public int compare(final int p1,final int p2)
		{
		    if(p1==p2) return 0;
		    // inverse order to have "latest" at index=0
		    final int s=distmatr[n*p1+p2]<0?1:-1;
		    return s;
		}
	    });
	// Save the result in proper order.
	// This is equivalent to applying permutation operator
	// corresponding to distmatr generated distance.
	this.evv_distancesorted=new double [n*basisLD];
	for(int k=0;k<n;k++)
	{
	    System.arraycopy(basis,basisOff+t_order[k]*basisLD,this.evv_distancesorted,k*basisLD,basisLD);
	}
    }
abstract double getK2Distance(final double [] a,final double [] b);
double [] getDistMatr(final int n,final double [] basis,final int basisLD,final int basisOff)
{
    final double [] distmatr;
    // Obtain distance matrix: anti-Hermitian part of getGTDistanceMatr generated matrix.
    TMatr.scaleVectorInPlace(TMatr.selectAntiHermitianPartInPlace(n,distmatr=getGTDistanceMatr(n,basis,basisLD,basisOff)),2);
    return distmatr;
}

double [] getGTDistanceMatr(final int n,final double [] basis,final int basisLD,final int basisOFF)
{
    final double [] k2matr=new double [n*n];
    final double [] psik=new double [basisLD],psil=new double [basisLD];

    for(int k=0;k<n;k++)
    {
	System.arraycopy(basis,basisOFF+k*basisLD,psik,0,basisLD);
	for(int l=0;l<n;l++)
	{
	    System.arraycopy(basis,basisOFF+l*basisLD,psil,0,basisLD);
	    k2matr[k*n+l]=getK2Distance(psik,psil);
	}
    }
    return k2matr;
}


double [] getOrderedEMfromQQ(final double [] QQ,final int ldQQ)
{
    final int n=t_order.length,basisLD=n<=0?0:evv_distancesorted.length/n;
    final double [] m=new double [n];
    for(int k=0;k<n;k++)
    {
	m[k]=TMatr.getAMatrB(basisLD,evv_distancesorted,k*basisLD,basisLD,evv_distancesorted,k*basisLD,QQ,ldQQ);
    }
    return m;
}

void test_Self(final double eps)
{
    final int n=t_order.length,basisLD=n<=0?0:evv_distancesorted.length/n;
    final double [] gtmatr=getGTDistanceMatr(n,evv_distancesorted,basisLD,0),
	ltmatr=TMatr.getMatrTrans(n,gtmatr),
	distmatr=new double [gtmatr.length],distmatr2,hermatr2;
    TMatr.scaleVectorInPlace(TMatr.selectAntiHermitianPartInPlace(n,distmatr2=gtmatr.clone()),2);
    TMatr.selectHermitianPartInPlace(n,hermatr2=gtmatr.clone());
    TMatr.scaleVectorInPlace(hermatr2,2);
    for(int k=0;k<gtmatr.length;k++)
    {
	// make sure hermitian part of gtmatr is equal exactly to 1
	UnitTests.checkDoubleRel(gtmatr[k]+ltmatr[k],"gtmatr[k]+ltmatr[k]",1,"1",eps);
	UnitTests.checkDoubleRel(hermatr2[k],"hermatr2[k]",1,"1",eps);
	distmatr[k]=gtmatr[k]-ltmatr[k];
    }
    UnitTests.checkDoubleArrayRel(distmatr2,"distmatr2",distmatr,"distmatr",0*eps); // to match exactly
    System.err.println("Sorted distance matrix="+java.util.Arrays.toString(distmatr));
    System.err.println("t_order="+java.util.Arrays.toString(t_order));
    // check the distmatr matrix is anti-symmetric
    for(int k=0;k<n;k++)
    {
	UnitTests.checkDoubleRel(distmatr[k*n+k],"distmatr[k*n+k]",0,"0",eps);
	for(int l=0;l<k;l++)
	{
	    UnitTests.checkDoubleRel(distmatr[k*n+l],"distmatr[k*n+l]",-distmatr[l*n+k],"-distmatr[l*n+k]",eps);
	}
    }

/* Example: triangle inequality violation:
   failed distance matrix (unsorted) from unit test:
   [0.0,               -0.9050667173692115,  -0.8881454834785325,-0.9130118149233837,
    0.9050667173692115, 0.0,                 -0.0822836350907803, 0.020144552654933312,
    0.8881454834785325, 0.0822836350907803,   0.0,               -0.10607117905107627,
    0.9130118149233837,-0.020144552654933312, 0.10607117905107627,0.0]
    what lead to:
    a0<a1 a0<a2 a0<a3
    a1<a2 a1>a3
    a2<a3
    this transforms to a3<a1 a1<a2 and a2<a3 , and give a contradiction: a2>a3 and a2<a3.
    This means this distance matrix cannot be sorted (no permutation operator exist).
*/
    // Note: transitivity condition may not hold
    final double [] psik=new double [basisLD],psil=new double [basisLD];
    for(int k=0;k<n;k++)
    {
	System.arraycopy(evv_distancesorted,k*basisLD,psik,0,basisLD);
	double wprev=-1;
	for(int l=0;l<k;l++)
	{
	    System.arraycopy(evv_distancesorted,l*basisLD,psil,0,basisLD);
	    final double wgt=getK2Distance(psik,psil),wlt=getK2Distance(psil,psik);
	    UnitTests.checkDoubleRel(distmatr[k*n+l],"distmatr[k*n+l]",wgt-wlt,"wgt-wlt",eps);
	    final double wcur=distmatr[k*n+l];
	    if(!(wprev<wcur))
	    {
		// relatively seldom event
		System.err.println("HARD ORDERING DOES NOTH HOLD: (k,l)=("+k+","+l+") wprev="+wprev+" wcur="+wcur);
	    }
	    wprev=wcur;
	    if(!(wgt<wlt))
	    {
		//throw new RuntimeException(
		System.err.println("TRANSITIVITY DOES NOT HOLD!!!\n"+
				   "wgt(k="+k+",l="+l+")="+wgt+"\twgt(k="+k+",l="+l+")="+wlt+"\tt_order="+java.util.Arrays.toString(t_order)+" gtmatr[k*n+l]="+gtmatr[k*n+l]+" ltmatr[k*n+l]="+ltmatr[k*n+l]);
	    }
	}
    }

    System.err.println("Distance:test_Self:$Revision: 1.68 $ test OK n="+n+" eps="+eps);
}
} // END OF class Distance
