package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.PolynomialRootsFinder;
import com.polytechnik.utils.Quadrature;
import com.polytechnik.realtimebook.*;

/** Extension of Rurik: Raiser &amp; Vidka with volatility.
 *  @version $Id: DjQVM.java,v 1.103 2022/08/08 20:10:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class DjQVM<DTYPE> {

final DTYPE odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final EVXData I,P2Ri,P2Rt,P2V,VRt,R2R,RI,tsI;

final double
    P2Ri_dp_i,P2Ri_vs,P2Ri_rvs,P2Ri_norm,P2Rt_dp,P2Rt_vs,P2Rt_rvs,P2Rt_norm,P2Ri_i_dp_i,P2Rt_dp_i;
final double x_P2Ri_i_dp_i,x_P2Ri_dp_i_dp,x_P2Ri_i_i_i;
final double xIL,xIH,xFIL,xFILn,xFIH,xFIHn,xRFIL,xRFIH;

final double n1,n2,n3,VRt_dp_i,VRt_vs,VRt_rvs,
    xL_P2Rt,
    xH_P2Rt,
	xL_P2V,
	xH_P2V,
	xL_VRt,
    xH_VRt,
    dpL_P2V,
    dpH_P2V,lh_P2V;
final double fi0,fdp0,fIdp0,fvs0,frvs0,fix,fdpx,fIdpx,fvsx,frvsx,
    fRix,fRdpx,fRIdpx,fRvsx,fRrvsx,fRi0,fRdp0,fRIdp0,fRvs0,fRrvs0;

final double fP1_0,fP1_x,fP1_xn;
final double fP1_xFIL,fP1_xFIH;
final double [] Xx,Wx,Wdp,Wvs,Wms;
final double fXx,fWdp,fWvs,fWx;
final double [] qXx,qWx,qWdp,qWvs;
final double qfXx,qfWdp,qfWvs,qfWx;
final double [] rXx,rWx,rWdp,rWvs;
final double rfXx,rfWdp,rfWvs,rfWx;

DjQVM(final int neq,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();
    this.M=getABasis();
    final int nmoms=2*neq-1+1+1;
    final double [] imom=new double [nmoms],tmom=new double [nmoms],
	dpmom=new double [nmoms],dp2mom=new double [nmoms],
	vsmom=new double [nmoms],tsmom=new double [nmoms],rvsmom=new double [nmoms];
    access.setMT(tmom,nmoms);
    access.setM(imom,0,access.M_I,odata,nmoms);
    access.setM(dpmom,1,access.M_DP,odata,nmoms);
    access.setM(dp2mom,2,access.M_DP,odata,nmoms);
    access.setM(vsmom,0,access.M_VS,odata,nmoms);
    access.setM(tsmom,0,access.M_TS,odata,nmoms);
    access.setM(rvsmom,0,access.M_RVS,odata,nmoms);
    final double [] P1mom=new double [nmoms],P2mom=new double [nmoms];
    access.setFMoments(P1mom,dpmom,nmoms);
    access.setFMoments(P2mom,dp2mom,nmoms);
    final double [] Vmom=new double [nmoms];
    access.setFMoments(Vmom,imom,nmoms); for(int k=0;k<nmoms;k++) Vmom[k]=-Vmom[k];


    final double [] Rtmom=M.getRadauMomentsFromMoments(tmom),
	Rimom=M.getRadauMomentsFromMoments(imom),
	Rdpmom=M.getRadauMomentsFromMoments(dpmom),
	Rvsmom=M.getRadauMomentsFromMoments(vsmom),
	Rrvsmom=M.getRadauMomentsFromMoments(rvsmom);

    final double x0=getX0(),p_offset=getPOffset();
    this.I=DeprecatedHelperEVXData.getEVXData(neq,imom,tmom,access,x0);
    this.RI=DeprecatedHelperEVXData.getEVXData(neq,Rimom,Rtmom,access,x0);
    this.P2Ri=DeprecatedHelperEVXData.getEVXData(neq,P2mom,Rimom,access,x0);
    this.P2Rt=DeprecatedHelperEVXData.getEVXData(neq,P2mom,Rtmom,access,x0);
    this.P2V=DeprecatedHelperEVXData.getEVXData(neq,P2mom,Vmom,access,x0);
    this.VRt=DeprecatedHelperEVXData.getEVXData(neq,Vmom,Rtmom,access,x0);
    this.tsI=DeprecatedHelperEVXData.getEVXData(neq,tsmom,imom,access,x0);

    {
	final double [] R2tmom=M.getRadauMomentsFromMoments(Rtmom),
	    RVmom=M.getRadauMomentsFromMoments(Vmom);
	this.R2R=DeprecatedHelperEVXData.getEVXData(neq,R2tmom,Rtmom,access,x0);

	final double [] R2QQt=M.B.getQQMatr(neq,R2tmom),
	    RQQt=M.B.getQQMatr(neq,Rtmom),
	    QQV=M.B.getQQMatr(neq,Vmom),
	    RQQV=M.B.getQQMatr(neq,RVmom),
	    QQP1=M.B.getQQMatr(neq,P1mom),
	    QQP2=M.B.getQQMatr(neq,P2mom);
	xL_P2Rt=TMatr.getAMatrA(neq,P2Rt.psiL,R2QQt,neq);
	xH_P2Rt=TMatr.getAMatrA(neq,P2Rt.psiH,R2QQt,neq);
	n1=TMatr.getAMatrA(neq,P2Rt.psiH,RQQt,neq);
	xL_P2V=TMatr.getAMatrA(neq,P2V.psiL,RQQV,neq);
	xH_P2V=TMatr.getAMatrA(neq,P2V.psiH,RQQV,neq);
	n2=TMatr.getAMatrA(neq,P2V.psiH,QQV,neq);
	dpL_P2V=TMatr.getAMatrA(neq,P2V.psiL,QQP1,neq);
	dpH_P2V=TMatr.getAMatrA(neq,P2V.psiH,QQP1,neq);
	lh_P2V=TMatr.getAMatrA(neq,P2V.psiH,QQP2,neq);

	xL_VRt=TMatr.getAMatrA(neq,VRt.psiL,R2QQt,neq);
	xH_VRt=TMatr.getAMatrA(neq,VRt.psiH,R2QQt,neq);
	n3=TMatr.getAMatrA(neq,VRt.psiH,RQQt,neq);
    }

    final double [] RQQt=M.B.getQQMatr(neq,Rtmom),
	RQQi=M.B.getQQMatr(neq,Rimom),
	RQQdp=M.B.getQQMatr(neq,Rdpmom),
	RQQvs=M.B.getQQMatr(neq,Rvsmom),
	RQQrvs=M.B.getQQMatr(neq,Rrvsmom),
	IRQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,RQQt),
	RQQdp_IRQQt_RQQi=TMatr.mulMMM(neq,RQQdp,IRQQt,RQQi),
	RQQi_IRQQt_RQQdp_IRQQt_RQQi=TMatr.mulMMM(neq,RQQi,IRQQt,RQQdp_IRQQt_RQQi);
    P2Ri_i_dp_i=TMatr.getAMatrA(neq,P2Ri.psiL,RQQi_IRQQt_RQQdp_IRQQt_RQQi,neq);
    P2Ri_dp_i=TMatr.getAMatrA(neq,P2Ri.psiL,RQQdp_IRQQt_RQQi,neq);
    P2Ri_vs=TMatr.getAMatrA(neq,P2Ri.psiL,RQQvs,neq);
    P2Ri_rvs=TMatr.getAMatrA(neq,P2Ri.psiL,RQQrvs,neq);
    P2Ri_norm=TMatr.getAMatrA(neq,P2Ri.psiL,RQQi,neq);

    x_P2Ri_i_dp_i=TMatr.getAMatrA(neq,P2Ri.psiL,TMatr.mulMatrices(neq,new double [][]{RQQi,IRQQt,RQQdp,IRQQt,RQQi}),neq);
    x_P2Ri_dp_i_dp=TMatr.getAMatrA(neq,P2Ri.psiL,TMatr.mulMatrices(neq,new double [][]{RQQdp,IRQQt,RQQi,IRQQt,RQQdp}),neq);
    x_P2Ri_i_i_i=TMatr.getAMatrA(neq,P2Ri.psiL,TMatr.mulMatrices(neq,new double [][]{RQQi,IRQQt,RQQi,IRQQt,RQQi}),neq);

    P2Rt_dp=TMatr.getAMatrA(neq,P2Rt.psiL,RQQdp,neq);
    P2Rt_vs=TMatr.getAMatrA(neq,P2Rt.psiL,RQQvs,neq);
    P2Rt_rvs=TMatr.getAMatrA(neq,P2Rt.psiL,RQQrvs,neq);
    P2Rt_norm=TMatr.getAMatrA(neq,P2Rt.psiL,RQQt,neq);
    P2Rt_dp_i=TMatr.getAMatrA(neq,P2Rt.psiL,RQQdp_IRQQt_RQQi,neq);

    VRt_dp_i=TMatr.getAMatrA(neq,VRt.psiH,RQQdp_IRQQt_RQQi,neq);
    VRt_vs=TMatr.getAMatrA(neq,VRt.psiH,RQQvs,neq);
    VRt_rvs=TMatr.getAMatrA(neq,VRt.psiH,RQQrvs,neq);

    final double [] psi=new double [neq],xI=new double [neq];
    for(int k=0;k<neq;k++)
    {
	System.arraycopy(I.ev,neq+k*neq,psi,0,neq);
	xI[k]=TMatr.getAMatrA(neq,psi,RQQt,neq);
    }

    this.xIL=getXF(I.psiL);
    this.xIH=getXF(I.psiH);
    double [] ff=getXFI(I.psiL);
    this.xFIL=ff[0];
    this.xFILn=ff[1];
    ff=getXFI(I.psiH);
    this.xFIH=ff[0];
    this.xFIHn=ff[1];

    simom=imom;
    stmom=tmom;
    sdpmom=dpmom;
    svsmom=vsmom;


    if(true)
    {
	final double [] QQt=M.B.getQQMatr(neq,tmom),
	    QQi=M.B.getQQMatr(neq,imom),
	    QQdp=M.B.getQQMatr(neq,dpmom),
	    QQvs=M.B.getQQMatr(neq,vsmom),
	    QQrvs=M.B.getQQMatr(neq,rvsmom),
	    QQP1=M.B.getQQMatr(neq,P1mom),
	    IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQt);
	if(true)
	{
	    final double [] xmom=new double [nmoms],msmom=new double [nmoms];
	    for(int k=0;k<nmoms;k++) {
		xmom[k]=imom[k]-I.sL*tmom[k];
		msmom[k]=vsmom[k]-I.sL*tsmom[k];
	    }
	    sxmom=xmom;
	    final double [] QQx=M.B.getQQMatr(neq,xmom),QQms=M.B.getQQMatr(neq,msmom);
	    final double [][] r=getRoots(I.psiL);
	    final double [] xL=r!=null?r[0]:null;
	    final int indX=getIndX0(xL,x0,isEdgeAtRight());
	    final double [][] Wres=getIntegralEstimation(xL,I.psiL,new double [][]{QQdp,QQvs,QQms},QQx,neq,M);
	    Xx=xL;
	    Wdp=Wres[0];
	    Wvs=Wres[1];
	    Wms=Wres[2];
	    Wx=Wres[3];

	    final double [][] qWres=getIntegralEstimationQ1(xL,I.psiL,new double [][]{dpmom,vsmom},xmom);
	    qXx=xL;
	    qWdp=qWres[0];
	    qWvs=qWres[1];
	    qWx=qWres[2];

	    if(indX>=0)
	    {
		fXx=Xx[indX];
		fWdp=Wdp[indX];
		fWvs=Wvs[indX];
		fWx=Wx[indX];
		qfXx=qXx[indX];
		qfWdp=qWdp[indX];
		qfWvs=qWvs[indX];
		qfWx=qWx[indX];
	    }
	    else
	    {
		fXx=fWdp=fWvs=fWx=qfXx=qfWdp=qfWvs=qfWx=Double.NaN;
	    }

	    final double [][] Wrad=getRadauIntegrals(neq,xmom,new double [][]{dpmom,vsmom},false);
	    rWdp=Wrad[0];
	    rWvs=Wrad[1];
	    rXx=Wrad[2];
	    rWx=Wrad[3];
	    if(neq>1)
	    {
		rfXx=rXx[neq-2];
		rfWdp=rWdp[neq-2];
		rfWvs=rWvs[neq-2];
		rfWx=rWx[neq-2];
	    }
	    else {rfXx=rfWdp=rfWvs=rfWx=Double.NaN;}
	}

	final double [] KQQi=TMatr.mulMatrices(neq,new double [][]{IQQt,QQi,IQQt}),
	    KQQdp=TMatr.mulMatrices(neq,new double [][]{IQQt,QQdp,IQQt}),
	    KQQIdp=TMatr.mulMatrices(neq,new double [][]{IQQt,QQdp,IQQt,QQi,IQQt}),
	    KQQvs=TMatr.mulMatrices(neq,new double [][]{IQQt,QQvs,IQQt}),
	    KQQrvs=TMatr.mulMatrices(neq,new double [][]{IQQt,QQrvs,IQQt}),
	    KQQP1=TMatr.mulMatrices(neq,new double [][]{IQQt,QQP1,IQQt});

	final double [] kkt=M.B.getKK(neq,IQQt),
	    kki=M.B.getKK(neq,KQQi),kkdp=M.B.getKK(neq,KQQdp),
	    kkIdp=M.B.getKK(neq,KQQIdp),kkvs=M.B.getKK(neq,KQQvs),kkrvs=M.B.getKK(neq,KQQrvs),
	    kkP1=M.B.getKK(neq,KQQP1);
	fi0=M.B.getD0(kki,x0)/M.B.getD0(kkt,x0);
	fdp0=M.B.getD0(kkdp,x0)/M.B.getD0(kkt,x0);
	fIdp0=M.B.getD0(kkIdp,x0)/M.B.getD0(kkt,x0);
	fvs0=M.B.getD0(kkvs,x0)/M.B.getD0(kkt,x0);
	frvs0=M.B.getD0(kkrvs,x0)/M.B.getD0(kkt,x0);
	fP1_0=p_offset+M.B.getD0(kkP1,x0)/M.B.getD0(kkt,x0);

	{
	final double x=this.xFIL;
	fix=M.B.getD0(kki,x)/M.B.getD0(kkt,x);
	fdpx=M.B.getD0(kkdp,x)/M.B.getD0(kkt,x);
	fIdpx=M.B.getD0(kkIdp,x)/M.B.getD0(kkt,x);
	fvsx=M.B.getD0(kkvs,x)/M.B.getD0(kkt,x);
	frvsx=M.B.getD0(kkrvs,x)/M.B.getD0(kkt,x);
	fP1_x=p_offset+M.B.getD0(kkP1,x)/M.B.getD0(kkt,x);
	fP1_xFIL=p_offset+M.B.getD0(kkP1,this.xFIL)/M.B.getD0(kkt,this.xFIL);
	fP1_xFIH=p_offset+M.B.getD0(kkP1,this.xFIH)/M.B.getD0(kkt,this.xFIH);
	}

	fP1_xn=p_offset+M.B.getD0(kkP1,this.xFILn)/M.B.getD0(kkt,this.xFILn);
    }

    if(true)
    {
	this.xRFIL=getXFI(RI.psiL)[0];
	this.xRFIH=getXFI(RI.psiH)[0];

	final double [] KRQQi=TMatr.mulMatrices(neq,new double [][]{IRQQt,RQQi,IRQQt}),
	    KRQQdp=TMatr.mulMatrices(neq,new double [][]{IRQQt,RQQdp,IRQQt}),
	    KRQQIdp=TMatr.mulMatrices(neq,new double [][]{IRQQt,RQQdp,IRQQt,RQQi,IRQQt}),
	    KRQQvs=TMatr.mulMatrices(neq,new double [][]{IRQQt,RQQvs,IRQQt}),
	    KRQQrvs=TMatr.mulMatrices(neq,new double [][]{IRQQt,RQQrvs,IRQQt});
	final double [] kkt=M.B.getKK(neq,IRQQt),
	    kki=M.B.getKK(neq,KRQQi),kkdp=M.B.getKK(neq,KRQQdp),
	    kkIdp=M.B.getKK(neq,KRQQIdp),kkvs=M.B.getKK(neq,KRQQvs),kkrvs=M.B.getKK(neq,KRQQrvs);
	final double x=this.xRFIL;
	fRix=M.B.getD0(kki,x)/M.B.getD0(kkt,x);
	fRdpx=M.B.getD0(kkdp,x)/M.B.getD0(kkt,x);
	fRIdpx=M.B.getD0(kkIdp,x)/M.B.getD0(kkt,x);
	fRvsx=M.B.getD0(kkvs,x)/M.B.getD0(kkt,x);
	fRrvsx=M.B.getD0(kkrvs,x)/M.B.getD0(kkt,x);

	fRi0=M.B.getD0(kki,x0)/M.B.getD0(kkt,x0);
	fRdp0=M.B.getD0(kkdp,x0)/M.B.getD0(kkt,x0);
	fRIdp0=M.B.getD0(kkIdp,x0)/M.B.getD0(kkt,x0);
	fRvs0=M.B.getD0(kkvs,x0)/M.B.getD0(kkt,x0);
	fRrvs0=M.B.getD0(kkrvs,x0)/M.B.getD0(kkt,x0);
    }

}

final double [] simom,stmom,sxmom,sdpmom,svsmom;


static double [][] getIntegralEstimation(final double [] x,final double [] psi,final double [][] QQ,final double [] QQmu,final int nmQQ,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int n=psi.length-1;
    final int nx=(x==null?0:x.length);
    final double [][] w=new double [QQ.length+1][nx];
    for(int k=0;k<nx;k++)
    {
	final double xk=x[k];
	final double [] p=M.B.sdiv1(psi,xk).q;
	final double f0=M.B.getD0(p,xk),wn=TMatr.getAMatrA(n,p,QQmu,nmQQ),w0=wn/(f0*f0);
	w[QQ.length][k]=w0;
	for(int j=0;j<QQ.length;j++)
	{
	    w[j][k]=TMatr.getAMatrA(n,p,QQ[j],nmQQ)/wn;
	}
    }
    return w;
}
double [][] getIntegralEstimationQ1(final double [] x,final double [] psi,final double [][] moms,final double [] xmom)
{
    final int n=(x==null?0:x.length);

    final double [][] w=new double [moms.length+1][n];
    for(int k=0;k<n;k++)
    {
	final double xk=x[k];
	final double [] p=M.B.sdiv1(psi,xk).q;
	final double f0=M.B.getD0(p,xk),wn=TMatr.mulVVn1(p,xmom),w0=Math.abs(wn/f0); // abs for rounding errors
	w[moms.length][k]=w0;
	for(int j=0;j<moms.length;j++)
	{
	    w[j][k]=(TMatr.mulVVn1(p,moms[j])/f0)/Math.sqrt(w0);
	}
    }
    return w;
}


double [][] getRadauIntegrals(final int nd,final double [] xmom,final double [][] moms,final boolean flag_psi2)
{
    final int n=nd-1; // number of roots
    final double x0=getX0();
    final double [] QQx=M.B.getQQMatr(n,xmom);
    final double [] ok=access.getOK(n,QQx,x0);
    final double [] xr=M.getXNodes(ok).re_root;
    //System.err.println(" n="+n+" ok.length="+ok.length+" xr="+java.util.Arrays.toString(xr));
    final double [] x=new double [n];
    System.arraycopy(xr,0,x,0,n-1);
    x[n-1]=x0; // set radau node
    final double [][] w=new double [moms.length+2][n];
    for(int k=0;k<n;k++)
    {
	final double xk=x[k];
	final double [] psi;
	if(k>=(n-1)) {psi=ok;}
	else{psi=M.getRadauMultiplied(M.B.sdiv1(ok,xk).q);}
	w[moms.length][k]=x[k];
	final double f0=M.B.getD0(psi,xk);
	if(flag_psi2)
	{
	    final double wn=TMatr.getAMatrA(psi.length,psi,M.B.getQQMatr(psi.length,xmom),psi.length);
	    w[moms.length+1][k]=wn/(f0*f0);
	    for(int j=0;j<moms.length;j++)
	    {
		w[j][k]=TMatr.getAMatrA(psi.length,psi,M.B.getQQMatr(psi.length,moms[j]),psi.length)/wn;
	    }
	}
	else
	{
	    w[moms.length+1][k]=TMatr.mulVVn1(psi,xmom)/f0;
	    final double sqw=Math.sqrt(w[moms.length+1][k]);
	    for(int j=0;j<moms.length;j++)
	    {
		w[j][k]=(TMatr.mulVVn1(psi,moms[j])/f0)/sqw;
	    }
	}
    }
    return w;
}


abstract boolean isEdgeAtRight();
void sortArray(double [] re,double [] im)
{
    if(isEdgeAtRight()) {com.polytechnik.utils.SortArrays.sortArrayWithLoadDesc(re.length,re,im);}
    else {com.polytechnik.utils.SortArrays.sortArrayWithLoad(re.length,re,im);}
}

static int getIndX0(final double [] x,final double x0,final boolean flag_right)
{
    final double ECUT=0e-4;
    if(x!=null)
    {
	for(int k=0;k<x.length;k++)
	{
	    final double xk=x[k];
	    if(flag_right)
	    {
		if(xk<x0+ECUT) return k;
	    }
	    else
	    {
		if(xk>x0-ECUT) return k;
	    }
	}
    }
    return -1;
}
static int getIndX0_NonSortedX(final double [] x,final double x0,final boolean flag_right)
{
    if(x==null) return -1;
    final double ECUT=0e-4;
    int ind=-1;
    double dmin=Double.POSITIVE_INFINITY;
    for(int k=0;k<x.length;k++)
    {
	final double xk=x[k],d;
	if(flag_right){d=x0-xk;}
	else{d=xk-x0;}
	if(!(d>ECUT)) continue;
	if(d<=dmin){ind=k;dmin=d;}
    }
    return ind;
}
static int getIndMax(final double [] f)
{
    int ind=-1;
    if(f==null) return ind;
    double fmax=Double.NEGATIVE_INFINITY;
    for(int k=0;k<f.length;k++)
    {
	if(f[k]>=fmax)
	{
	    fmax=f[k];
	    ind=k;
	}
    }
    return ind;
}

static int getIndXClose(final double [] x,final double xx,final double x0,final boolean flag_right)
{
    final double ECUT=1e-4;
    if(x!=null)
    {
	int indx=-1;
	double diff=Double.POSITIVE_INFINITY;
	for(int k=0;k<x.length;k++)
	{
	    final double xk=x[k];
	    if(flag_right)
	    {
		if(!(xk<=x0+ECUT)) continue;
	    }
	    else
	    {
		if(!(xk>=x0-ECUT)) continue;
	    }

	    final double d=Math.abs(xk-xx);
	    if(d<=diff)
	    {
		diff=d;
		indx=k;
	    }
	}
	return indx;
    }
    return -1;
}

double [] getXFI(final double [] psi)
{
    final double ECUT=1e-4;
    final double [][] r=getRoots(psi);
    if(r!=null)
    {
    final double [] x=r[0];
    for(int k=0;k<x.length;k++)
    {
	final double xk=x[k];
	if(isEdgeAtRight())
	{
	    if(xk<=getX0()+ECUT) return new double []{xk,k<x.length-1?x[k+1]:Double.NaN};
	}
	else
	{
	    if(xk>=getX0()-ECUT) return new double []{xk,k<x.length-1?x[k+1]:Double.NaN};
	}
    }
    }
    return new double []{Double.NaN,Double.NaN};
}

double getXF(final double [] psi)
{
    final double [][] r=getRoots(psi);
    if(r==null) return Double.NaN;
    return r[0][0];
}
double [][] getRoots(final double [] psi)
{
    final int n=psi.length-1;
    if(n<=0) return null;
    final PolynomialRootsFinder.PRoots u=M.getPolynomialRootsFinderInBasis().findRoots(psi);
    final double [] re_root=u.re_root,im_root=u.im_root;
    if(u.nroots!=n) return null;
    sortArray(re_root,im_root);
    return new double [][]{re_root,im_root};
}


abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double getX0();
abstract double getPOffset();



static abstract class DjQVMTest<DTYPE> extends QTest {

abstract DjQVM<DTYPE> getDjQVM();
abstract BjQVM<DTYPE> getBjQVM();
abstract BiQVM<DTYPE> getBiQVM();
abstract ZQVM<DTYPE> getZQVM();
abstract double [][] getMonomialMoments(final DTYPE o);
DjQVMTest(final String name,final java.util.Random r,final double eps)
{
    super(name,r,eps);
}

void test_R2R(final double [] ev)
{
}


abstract double [] getDpol(final double [] kkt,final double [] kki);
void print_RadonNykodym(final int n,final double [] tmom,final double [] imom,final DjQVM<DTYPE> q)
{
    final double [] QQt=q.M.B.getQQMatr(n,tmom),QQi=q.M.B.getQQMatr(n,imom);
    final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	IQQi=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQi);
    final double [] kkt=q.M.B.getKK(n,IQQt),kki=q.M.B.getKK(n,IQQi);
    final double [] p=getDpol(kkt,kki);

    System.err.println("p="+java.util.Arrays.toString(p));
    final PolynomialRootsFinder.PRoots u=q.M.getPolynomialRootsFinderInBasis().findRoots(p);
    final double [] re=u.re_root,im=u.im_root;
    if(u.nroots!=(p.length-1)){
	System.err.println("skipping");
	return;
	//throw new RuntimeException("nL="+nL+" p.length-1="+(p.length-1));
    }
    q.sortArray(re,im);
    int nr=0;
    for(int k=0;k<re.length;k++)
    {
	if(Math.abs(im[k])<eps)
	{
	    nr++;
	    System.err.println("ROOT_"+k+":"+re[k]+","+im[k]);
	    //UnitTests.checkDoubleRel(q.M.B.getD0(p,re[k]),"p(x)",0,"0",eps);
	}
    }
    System.err.println("nr="+nr+" n="+n);
}


void test_quadratures(final String txt,final double [] psi,final double lambda,final int sign,final double [] Xx,final double [] Wx,final DjQVM<DTYPE> q,final double [][] mom_test,final double [][] int_mom_test)
{
    final int n=psi.length-1;
    final double [][] rl=q.getRoots(psi);
    if(rl!=null)
    {
	System.err.println(txt+":im="+java.util.Arrays.toString(rl[1]));
	System.err.println(txt+":re="+java.util.Arrays.toString(rl[0]));
    }
    final double [] xmom=new double [q.simom.length];
    for(int k=0;k<xmom.length;k++)
    {
	xmom[k]=(q.simom[k]-lambda*q.stmom[k])*sign;
    }
    final double psiA=TMatr.mulVVn1(psi,xmom);
    System.err.println("AVER: psiA="+psiA);
    if(true)
    {
	final double [] QQx=q.M.B.getQQMatr(psi.length,xmom);
	final double [] tpsi=new double [psi.length];
	System.err.println("psi="+java.util.Arrays.toString(psi));
	for(int k=0;k<psi.length;k++)
	{
	    tpsi[k]=1;
	    final double aver=TMatr.getAMatrB(psi.length,psi,tpsi,QQx,psi.length);
	    System.err.println("<"+k+"|psi>="+aver);
	    tpsi[k]=0;
	}
	if(Xx!=null)
	{
	for(int k=0;k<Xx.length;k++)
	{
	    final double [] p=q.M.B.sdiv1(psi,Xx[k]).q;
	    final double pp=TMatr.getAMatrB(p.length,p,p,QQx,psi.length);
	    System.err.println("x["+k+"]="+Xx[k]+" <sdx|sdx>="+pp);
	}
	}
    }

    final Quadrature qa=q.M.getGaussQuadratureFromMoments(n,xmom);
    final double [] XK=qa.x,W=qa.w,reR=rl[0];
    q.sortArray(XK,W);

    System.err.println(txt+":XK="+java.util.Arrays.toString(XK));
    System.err.println(txt+": W="+java.util.Arrays.toString(W));
    UnitTests.checkDoubleArrayRel(XK,"XK",reR,"reR",eps);
    System.err.println("lambda="+lambda+" sumW="+TMatr.sumElements(W));

    final double [] wsq=q.getIntegralEstimationQ1(XK,psi,new double [][]{xmom},xmom)[0];
    for(int k=0;k<wsq.length;k++) {wsq[k]=Math.pow(wsq[k],2);}
    UnitTests.checkDoubleArrayRel1(W,"W",wsq,"wsq",eps);

    if(Wx!=null)
    {
	UnitTests.checkDoubleArrayRel(Wx,"Wx",W,"W",eps);
    }
    if(Xx!=null)
    {
	UnitTests.checkDoubleArrayRel(Xx,"Xx",XK,"XK",eps);
    }
    if(mom_test!=null)
    {
	final double [] QQ=q.M.B.getQQMatr(n,xmom),QQx=q.M.B.getQQMatr(n,q.M.B.getXQkMomentsFromQkMoments(xmom));
	final double [] ev=q.M.EV.getGEV_Generalized(n,QQx,QQ,true);
	final double [] tpsi=new double [n],f=new double [n];
	for(int j=0;j<mom_test.length;j++)
	{
	    final double [] QQf=q.M.B.getQQMatr(n,mom_test[j]);
	    for(int k=0;k<n;k++)
	    {
		final int ind=q.isEdgeAtRight()?n-1-k:k;
		System.arraycopy(ev,n+ind*n,tpsi,0,n);
		f[k]=TMatr.getAMatrA(n,tpsi,QQf,n);
	    }
	    UnitTests.checkDoubleArrayRel(f,"f",int_mom_test[j],"int_mom_test["+j+"]",eps);
	}
	System.err.println("calculated averages match");
    }


    System.err.println(name+":DjQVMTest:test_quadratures():"+txt+": test OK n="+n+" eps="+eps);
}

void test_radauquadratures(final String txt,final int n,final double [] xmom,final double [] Xx,final double [] Wx,final DjQVM<DTYPE> q,final double [][] mom_test,final double [][] int_mom_test)
{
    final Quadrature rquad=q.M.getRadauQuadratureFromMoments(n-2,xmom);
    // n-1 nodes.
    System.err.println("xmom="+java.util.Arrays.toString(xmom));
    final double [] x=rquad.x,w=rquad.w;
    System.err.println("x="+java.util.Arrays.toString(x)+"\nw="+java.util.Arrays.toString(w));
    System.err.println("Xx="+java.util.Arrays.toString(Xx)+"\nWx="+java.util.Arrays.toString(Wx));
    if(Double.isNaN(x[0]))
    {
	System.err.println(name+":DjQVMTest:test_radauquadratures():"+txt+": x=NaN, skipping test, n="+n+" eps="+eps);
	return;
    }
    UnitTests.checkDoubleArrayRel(x,"x",Xx,"Xx",eps);
    UnitTests.checkDoubleArrayRel(w,"w",Wx,"Wx",eps);

    final double [] wsq=q.getRadauIntegrals(n,xmom,new double [][]{xmom},false)[0];
    for(int k=0;k<(n-1);k++) {wsq[k]=Math.pow(wsq[k],2);}
    UnitTests.checkDoubleArrayRel(w,"w",wsq,"wsq",eps);

    final double [] QQx=q.M.B.getQQMatr(n-1,xmom),
	IQQx=com.polytechnik.utils.Linsystems.getInvertedMatrix(n-1,n-1,QQx);
    final double [] f=new double [n-1];
    for(int j=0;j<mom_test.length;j++)
    {
	final double [] v=TMatr.getMatr_x_A(n-1,IQQx,java.util.Arrays.copyOf(mom_test[j],n-1));
	for(int k=0;k<x.length;k++) {f[k]=q.M.B.getD0(v,x[k])*Math.sqrt(w[k]);}
	UnitTests.checkDoubleArrayRel(f,"f",int_mom_test[j],"int_mom_test["+j+"]",eps);
    }

    System.err.println(name+":DjQVMTest:test_radauquadratures():"+txt+": test OK n="+n+" eps="+eps);
}



void test_Common()
{
    final DjQVM<DTYPE> q=getDjQVM();

    UnitTests.checkDoubleRel(q.P2Ri_norm,"q.P2Ri_norm",1,"1",eps);
    UnitTests.checkDoubleRel(q.P2Rt_norm,"q.P2Rt_norm",1,"1",eps);

    UnitTests.checkDoubleRel(q.n1,"q.n1",1,"1",eps);
    UnitTests.checkDoubleRel(q.n2,"q.n2",1,"1",eps);
    UnitTests.checkDoubleRel(q.lh_P2V,"q.lh_P2V",q.P2V.sH,"q.P2V.sH",eps);
    UnitTests.checkDoubleRel(q.n3,"q.n3",1,"1",eps);

    UnitTests.checkDoubleRel(q.x_P2Ri_i_dp_i,"q.x_P2Ri_i_dp_i",q.P2Ri_i_dp_i,"q.P2Ri_i_dp_i",eps);
    test_R2R(q.R2R.ev);


    System.err.println("I.s0="+q.I.s0);
    for(int j=0;j<2;j++)
    {
	final double [] psi=j<=0?q.I.psiL:q.I.psiH;
	final double lam=j<=0?q.I.sL:q.I.sH;
	final double [][] r=q.getRoots(psi);
	if(r!=null)
	{
	    final double [] zero=new double [r[0].length];
	    System.err.println("F_"+j+":lam="+lam+" re="+java.util.Arrays.toString(r[0])+" im="+java.util.Arrays.toString(r[1]));
	    UnitTests.checkDoubleArrayRel(r[1],"im",zero,"zero",eps);
	}
    }


    UnitTests.checkDoubleRel(q.RI.s0,"q.RI.s0",q.fRi0,"q.fRi0",eps);
    UnitTests.checkDoubleRel(q.fXx,"q.fXx",q.xFIL,"q.xFIL",eps);


    final double [] w1=new double [q.I.psiL.length-1];java.util.Arrays.fill(w1,1);
    test_quadratures("I.psiL",q.I.psiL,q.I.sL,1,q.Xx,q.Wx,q,new double [][]{q.sdpmom,q.svsmom,q.sxmom,q.M.B.getXQkMomentsFromQkMoments(q.sxmom)},new double [][]{q.Wdp,q.Wvs,w1,q.Xx});
    test_quadratures("I.psiH",q.I.psiH,q.I.sH,-1,null,null,q,null,null);


    System.err.println("radau:q.rfXx="+q.rfXx);
    final double [] wsq=new double [q.rWx.length];
    for(int k=0;k<wsq.length;k++) {wsq[k]=Math.sqrt(q.rWx[k]);}
    // may be negatively defined
    test_radauquadratures("RadL",neq,q.sxmom,q.rXx,q.rWx,q,new double [][]{q.sdpmom,q.svsmom,q.sxmom},new double [][]{q.rWdp,q.rWvs,wsq});


    if(q.Xx.length!=q.rXx.length) throw new RuntimeException("d mismatch");
    if(q.Xx.length!=(neq-1)) throw new RuntimeException("d mismatch");


    UnitTests.checkDoubleArrayRel1(q.Wx,"q.Wx",q.qWx,"q.qWx",eps);

    print_RadonNykodym(neq,q.stmom,q.simom,q);

    System.err.println(name+":DjQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}



void test_BjQVM()
{
    final DjQVM<DTYPE> q=getDjQVM();
    final BjQVM<DTYPE> s=getBjQVM();

    EVXData.matchEVXData(q.I,s.I,eps);
    EVXData.matchEVXData(q.RI,s.RI,eps);
    EVXData.matchEVXData(q.P2Ri,s.P2Ri,eps);
    EVXData.matchEVXData(q.P2Rt,s.P2Rt,eps);
    EVXData.matchEVXData(q.P2V,s.P2V,eps);

    System.err.println(name+":DjQVMTest:test_BjQVM(): test OK neq="+neq+" eps="+eps);
}

void test_BiQVM()
{
    final DjQVM<DTYPE> q=getDjQVM();
    final BiQVM<DTYPE> s=getBiQVM();

    EVXData.matchEVXData(q.I,s.I,eps);
    EVXData.matchEVXData(q.tsI,s.tsI,eps);
    UnitTests.checkDoubleRel(q.fP1_0,"q.fP1_0",s.Pt0,"s.Pt0",eps);

    System.err.println(name+":DjQVMTest:test_BiQVM(): test OK neq="+neq+" eps="+eps);
}



void test_ZQVM()
{
    final DjQVM<DTYPE> q=getDjQVM();
    final ZQVM<DTYPE> s=getZQVM();


    UnitTests.checkDoubleRel(q.fi0,"q.fi0",s.nA0,"s.nA0",eps);
    UnitTests.checkDoubleRel(q.fvs0,"q.fvs0",s.ntsvs0,"s.ntsvs0",eps);
    UnitTests.checkDoubleRel(q.frvs0,"q.frvs0",s.ntrvs0,"s.ntrvs0",eps);
    UnitTests.checkDoubleRel(q.fIdp0,"q.fIdp0",s.nIdp0,"s.nIdp0",eps);

    System.err.println(name+":DjQVMTest:test_ZQVM(): test OK neq="+neq+" eps="+eps);
}



void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":DjQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_BjQVM();
    test_BiQVM();
    test_ZQVM();

    System.err.println(name+":DjQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":DjQVMTest: starting tests; ver=$Id: DjQVM.java,v 1.103 2022/08/08 20:10:46 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(950+r.nextInt(850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":DjQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: DjQVM.java,v 1.103 2022/08/08 20:10:46 mal Exp $");
}

} // END OF class DjQVMTest


}
