package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;

/** Book in time/price space with Radau measure.
 *  @version $Id: DoQVM.java,v 1.12 2020/02/06 00:09:22 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class DoQVM<DTYPE> {



final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;
private final DTYPE odata_buy,odata_sell;
final EVXData buy,sell;
final double p_buy,p_sell;
DoQVM(final int neq,final DTYPE odata_buy,final DTYPE odata_sell)
{
    this.odata_buy=odata_buy;
    this.odata_sell=odata_sell;
    this.access=getAccess();
    this.p_buy=getPEdge(this.odata_buy);
    this.p_sell=getPEdge(this.odata_sell);
    final double poffset_buy=getPOffset(this.odata_buy),poffset_sell=getPOffset(this.odata_sell);
    final int nmoms=2*neq-1;
    final double [] pmoms_buy=new double [nmoms],pmoms_sell=new double [nmoms];
    final double [] tmom=new double [nmoms];
    final double dpebuy=this.p_buy-poffset_buy,dpesell=this.p_sell-poffset_sell;
    access.setM(pmoms_buy,1,access.M_I,this.odata_buy,nmoms);
    access.setM(tmom,0,access.M_I,this.odata_buy,nmoms);
    for(int k=nmoms;--k>=0;) {pmoms_buy[k]=dpebuy*tmom[k]-pmoms_buy[k];}
    access.setM(pmoms_sell,1,access.M_I,this.odata_sell,nmoms);
    access.setM(tmom,0,access.M_I,this.odata_sell,nmoms);
    for(int k=nmoms;--k>=0;) {pmoms_sell[k]=pmoms_sell[k]-dpesell*tmom[k];}
    access.setMT(tmom,nmoms);
    final double [][] QQ=access.getQQ_Matrs(neq,1,new double [][]{tmom,pmoms_buy,pmoms_sell});
    this.buy=DeprecatedHelperEVXData.getEVXDataFromQQfDiagQQa(neq,QQ[1],QQ[0],access,getX0());
    this.sell=DeprecatedHelperEVXData.getEVXDataFromQQfDiagQQa(neq,QQ[2],QQ[0],access,getX0());

    // for unit tests
    u_tmom=tmom;u_pmoms_buy=pmoms_buy;u_pmoms_sell=pmoms_sell;
}

final double [] u_tmom,u_pmoms_buy,u_pmoms_sell;

abstract double getPOffset(final DTYPE odata);
abstract double getPEdge(final DTYPE odata);
abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract double getX0();



static abstract class DoQVMTest<DTYPE> {

int neq;
long currenttime,tau;
Trade [] tbuy,tsell;
BookItem [] tbuy_byp,tsell_byp;

abstract DoQVM<DTYPE> getDoQVM();
abstract BQVM<?> getBQVM();
abstract long getTau();

final String name;
final java.util.Random r;
final double eps;
DoQVMTest(final String name,
	 final java.util.Random r,final double eps)
{
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Common()
{
    final DoQVM<DTYPE> q=getDoQVM();

    for(int k=tbuy.length;--k>=0;)
    {
	if(tbuy[k].price>q.p_buy) throw new RuntimeException("tbuy["+k+"].price="+tbuy[k].price+" q.p_buy="+q.p_buy);
    }
    for(int k=tsell.length;--k>=0;)
    {
	if(tsell[k].price<q.p_sell) throw new RuntimeException("tsell["+k+"].price="+tsell[k].price+" q.p_sell="+q.p_sell);
    }

    System.err.println(name+":DoQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}
void setData(final int buysize,final int sellsize,final int neq)
{
    this.neq=neq;
    this.tbuy=CalculateKendall.TradeEMoments.getRandomTrades(buysize,r);
    tbuy_byp=tbuy.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tbuy_byp,0,tbuy_byp.length-1,Distribution.cmp_by_price_lowest_down);
    this.tsell=CalculateKendall.TradeEMoments.getRandomTrades(sellsize,r);
    tsell_byp=tsell.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tsell_byp,0,tsell_byp.length-1,Distribution.cmp_by_price_lowest_up);
    this.currenttime=Math.max(tbuy[tbuy.length-1].time,tsell[tsell.length-1].time);
    this.tau=getTau();
    //System.err.println("tau="+tau);
}

void test_BQVM()
{
    final DoQVM<DTYPE> q=getDoQVM();
    final BQVM<?> b=getBQVM();
    EVXData.matchEVXData(q.buy,DeprecatedHelperEVXData.getEVXData(neq,q.u_pmoms_buy,q.u_tmom,q.access,q.getX0()),eps);
    EVXData.matchEVXData(q.sell,DeprecatedHelperEVXData.getEVXData(neq,q.u_pmoms_sell,q.u_tmom,q.access,q.getX0()),eps);

    EVXData.matchEVXData(q.buy,b.getEVXData(neq,q.u_pmoms_buy,q.u_tmom),eps);
    EVXData.matchEVXData(q.sell,b.getEVXData(neq,q.u_pmoms_sell,q.u_tmom),eps);

    System.err.println(name+":DoQVMTest:test_BQVM(): test OK neq="+neq+" eps="+eps);
}

void test_Calculations(final int tbuysize,final int tsellsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":DoQVMTest:test_Calculations(): test with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);

    setData(tbuysize,tsellsize,neq);
    test_Common();
    test_BQVM();

    System.err.println(name+":DoQVMTest:test_Calculations(): test OK with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":DoQVMTest: starting tests; ver=$Id: DoQVM.java,v 1.12 2020/02/06 00:09:22 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(850),1950+r.nextInt(850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":DoQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: DoQVM.java,v 1.12 2020/02/06 00:09:22 mal Exp $");
}
} // END OF class DoQVMTest



} // END of class DoQVM
