package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;

/** Like DynVarFiniteC.get_PnLCV_DynVarFiniteC
 * but uses only infinitesimal variation (var_pn_IX=0,var2_pn_IX).
 * Note that for psi^2(x0)=0, e.g. psi^2=r*phi^2
 * we have: var2_pn_IX=<C D^2 psi^2> - <Cpsi^2> <D^2 psi^2>.
 * for psi(x0)=0 <D^2 psi^2>=0, but for Radau measue psi^2=r*phi^2
 * <D^2 psi^2>=-phi^2(x0)!=0.
 *  @version $Id: DynCPnL.java,v 1.22 2022/07/28 12:27:15 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynCPnL {
    final double I_0,I_IL,I_IH,p_IX,I_IX,var_pI_IX,var_pn_IX,var2_pn_IX,test_var2_pn_IX;
    DynCPnL(final double I_0,final double I_IL,final double I_IH,final double p_IX,final double I_IX,final double var_pI_IX,final double var_pn_IX,final double var2_pn_IX,final double test_var2_pn_IX)
    {
	this.I_0=I_0;
	this.I_IL=I_IL;
	this.I_IH=I_IH;
	this.p_IX=p_IX;
	this.I_IX=I_IX;
	this.var_pI_IX=var_pI_IX;
	//System.err.println("var_pI_IX="+var_pI_IX);
	this.var_pn_IX=var_pn_IX;
	this.var2_pn_IX=var2_pn_IX;
	this.test_var2_pn_IX=test_var2_pn_IX;
    }

static DynCPnL get0_DynCPnL(final int n,
			    final double [] QQV,final double [] QQt,final double [] QQpV,
			    final VolMatrix VOLATILITY_EDPsi,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");
    final double [] QQi=VOLATILITY_EDPsi.getDiffMatrix(n,QQV),QQpi=VOLATILITY_EDPsi.getDiffMatrix(n,QQpV);

    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M);
    final int i0=0,iL=1,iX=n-1;
    final double [] psi=EVXData.getPsiFromBasis(iX,basis,n),Dpsi=VOLATILITY_EDPsi.getDPsi(psi),D2psi=VOLATILITY_EDPsi.getDPsi(Dpsi);
    final double I_0=basis[i0],I_IL=basis[iL],I_IH=basis[n-1],
	pI_IX=TMatr.getAMatrA(psi,QQpi,n),I_IX=basis[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpV,-p_IX,QQV),
	rq=RayleighQuotient.getRQVariationValues(psi,Dpsi,D2psi,QQpnl,QQt,n),
	QQpI=TMatr.sumAcB(QQpi,-p_IX,QQi);
    return new DynCPnL(I_0,I_IL,I_IH,p_IX,I_IX,TMatr.getAMatrB(psi,Dpsi,QQpI,n),
		       rq[1],rq[2],
		       // alternative rq[2] calculation
		       2*TMatr.getAMatrB(psi,Dpsi,QQpI,n));
}

/** Radau measure is back.
 */
static DynCPnL getR_DynCPnL(final int n,
			    final double [] QQV,final double [] QQt,final double [] QQpV,
			    final VolMatrix VOLATILITY_EDPsi,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");
    final double [] QQi=VOLATILITY_EDPsi.getDiffMatrix(n,QQV),QQpi=VOLATILITY_EDPsi.getDiffMatrix(n,QQpV);
    final int nd=n-1,iX=nd-1;
    final double [] basis=M.EV.getGEV_Generalized(nd,VolMatrix.getRadauMatrix(n,QQi,M),VolMatrix.getRadauMatrix(n,QQt,M),true),
	psi=EVXData.getPsiFromBasis(iX,basis,nd),
	Dpsi=VOLATILITY_EDPsi.getDPsi(psi),
	D2psi=VOLATILITY_EDPsi.getDPsi(Dpsi),
	Rpsi=M.getRadauMultiplied(psi),
	DRpsi=VOLATILITY_EDPsi.getDPsi(Rpsi),
	D2Rpsi=VOLATILITY_EDPsi.getDPsi(DRpsi);
    final double pI_IX=TMatr.getAMatrB(psi,Rpsi,QQpi,n),I_IX=basis[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpV,-p_IX,QQV),
	rq=RayleighQuotient.getRQVariationValues(psi,Dpsi,D2psi,Rpsi,DRpsi,D2Rpsi,QQpnl,QQt,n),
	QQpnlO=TMatr.sumAcB(QQpnl,-TMatr.getAMatrB(Rpsi,psi,QQpnl,n),QQt),
	QQpI=TMatr.sumAcB(QQpi,-p_IX,QQi);
    if(false)
    {
	System.err.println("DynCPnL: phi(x0^2)="+Math.pow(M.B.getD0(psi,M.getRadauRoot()),2)+
			   "=-<D^2 r*phi^2>="+(-1*(TMatr.getAMatrB(Rpsi,D2psi,QQt,n)+2*TMatr.getAMatrB(DRpsi,Dpsi,QQt,n)+TMatr.getAMatrB(D2Rpsi,psi,QQt,n))));
    }
    return new DynCPnL(Double.NaN,basis[0],basis[nd-1],p_IX,I_IX,
		       TMatr.getAMatrB(Rpsi,Dpsi,QQpI,n)+TMatr.getAMatrB(psi,DRpsi,QQpI,n),
		       rq[1],rq[2],
		       // alternative rq[2] calculation
		       (TMatr.getAMatrB(Rpsi,D2psi,QQpnlO,n)+2*TMatr.getAMatrB(DRpsi,Dpsi,QQpnlO,n)+TMatr.getAMatrB(D2Rpsi,psi,QQpnlO,n))
		       );
}

/** Try without psi^2(x0)=0 boundary condition.
 * may lead to first variation being !=0, what is incorrect.
 */
static DynCPnL get_DynCPnL(final int n,
			    final double [] QQV,final double [] QQt,final double [] QQpV,
			    final VolMatrix VOLATILITY_EDPsi,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");
    final double [] QQi=VOLATILITY_EDPsi.getDiffMatrix(n,QQV),QQpi=VOLATILITY_EDPsi.getDiffMatrix(n,QQpV);

    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int iX=n-1;
    final double [] psi=I.psiH,Dpsi=VOLATILITY_EDPsi.getDPsi(psi),D2psi=VOLATILITY_EDPsi.getDPsi(Dpsi);
    final double pI_IX=I.getFPsi(iX,QQpi),I_IX=I.ev[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpV,-p_IX,QQV),
	rq=RayleighQuotient.getRQVariationValues(psi,Dpsi,D2psi,QQpnl,QQt,n),
	QQpnlO=TMatr.sumAcB(QQpnl,-TMatr.getAMatrA(psi,QQpnl,n),QQt),
	QQpI=TMatr.sumAcB(QQpi,-p_IX,QQi);
    if(false)
    {
	System.err.println("get_DynCPnL: rq[1]="+rq[1]+" rq[2]="+rq[2]);
    }
    return new DynCPnL(I.s0,I.sL,I.sH,p_IX,I_IX,
		       TMatr.getAMatrB(psi,Dpsi,QQpI,n),
		       rq[1],rq[2],
		       // alternative rq[2] calculation
		       2*(TMatr.getAMatrB(psi,D2psi,QQpnlO,n)+TMatr.getAMatrB(Dpsi,Dpsi,QQpnlO,n)));
}

static void matchDynVarFiniteC(final DynCPnL a,final DynVarFiniteC b,final double eps)
{
    UnitTests.checkDoubleRel(a.I_0,"a.I_0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.I_IL,"a.I_IL",b.I_IL,"b.I_IL",eps);
    UnitTests.checkDoubleRel(a.I_IH,"a.I_IH",b.I_IH,"b.I_IH",eps);
    UnitTests.checkDoubleRel(a.I_IX,"a.I_IX",b.I_IX,"b.I_IX",eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IX,"b.p_IX",eps);
    UnitTests.checkDoubleRel(a.var2_pn_IX,"a.var2_pn_IX",b.var2_pn_IX,"b.var2_pn_IX",eps);
    UnitTests.checkDoubleRel(a.var_pn_IX,"a.var_pn_IX",0,"0",eps);
    UnitTests.checkDoubleRel(a.var2_pn_IX,"a.var2_pn_IX",a.test_var2_pn_IX,"a.test_var2_pn_IX",eps);

    System.err.println("DynCPnL:matchDynVarFiniteC test OK eps="+eps);
}

static void matchDynRPnL(final DynCPnL a,final DynRPnL b,final double eps)
{
    UnitTests.checkDoubleRel(a.I_IL,"a.I_IL",b.I_L,"b.I_L",eps);
    UnitTests.checkDoubleRel(a.I_IH,"a.I_IH",b.I_H,"b.I_H",eps);
    UnitTests.checkDoubleRel(a.I_IX,"a.I_IX",b.I_H,"b.I_H",eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.var_pI_IX,"a.var_pI_IX",b.var_p_IH,"b.var_p_IH",eps);
    UnitTests.checkDoubleRel(a.var_pn_IX,"a.var_pn_IX",0,"0",eps);

    UnitTests.checkDouble(a.var2_pn_IX,"a.var2_pn_IX",a.test_var2_pn_IX,"a.test_var2_pn_IX",eps);
    System.err.println("a.var2_pn_IX="+a.var2_pn_IX+" a.test_var2_pn_IX="+a.test_var2_pn_IX);
    System.err.println("a.var_pn_IX="+a.var_pn_IX+" 2*b.var_p_IH="+(2*b.var_p_IH));

    System.err.println("DynCPnL:matchDynVarFiniteC test OK eps="+eps);
}

static void matchDynVar2C(final DynCPnL a,final DynVar2C b,final double eps)
{
    UnitTests.checkDoubleRel(a.I_0,"a.I_0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.I_IL,"a.I_IL",b.I_IL,"b.I_IL",eps);
    UnitTests.checkDoubleRel(a.I_IH,"a.I_IH",b.I_IH,"b.I_IH",eps);
    UnitTests.checkDoubleRel(a.I_IX,"a.I_IX",b.I_IX,"b.I_IX",eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IX,"b.p_IX",eps);
    UnitTests.checkDoubleRel(a.var_pI_IX,"a.var_pI_IX",0.5*b.var_pI_IX,"0.5*b.var_pI_IX",eps);
    //UnitTests.checkDoubleRel(a.var_pn_IX,"a.var_pn_IX",0,"0",eps);

    //UnitTests.checkDouble(a.var2_pn_IX,"a.var2_pn_IX",a.test_var2_pn_IX,"a.test_var2_pn_IX",eps);
    System.err.println("a.var2_pn_IX="+a.var2_pn_IX+" a.test_var2_pn_IX="+a.test_var2_pn_IX);
    System.err.println("a.var_pn_IX="+a.var_pn_IX+" 2*b.var_pI_IX="+(2*b.var_pI_IX));

    System.err.println("DynCPnL:matchDynVar2C test OK eps="+eps);
}

}

