package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/** Consider dI/dt. Some kind similar to DynVar2I.java.
 * After Dirk Laurie <dirk@aims.ac.za> notes about
 * differential equation on w(x) on Nov, 14, 2015.
 *  @version $Id: DynDI.java,v 1.20 2022/07/01 17:28:19 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynDI {
    final EVXData I;
    final double p_IX,dI_DIL,P_DIL,dI_DIH,P_DIH;
DynDI(final EVXData I,final double p_IX,
      final double dI_DIL,final double P_DIL,
      final double dI_DIH,final double P_DIH)
{
    this.I=I;
    this.p_IX=p_IX;
    this.dI_DIL=dI_DIL;
    this.P_DIL=P_DIL;
    this.dI_DIH=dI_DIH;
    this.P_DIH=P_DIH;
}


static DynDI get_DynDI(final int n,
		       final double [] QQi,final double [] QQt,
		       final double [] QQpi,
		       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		       final VolMatrix VOLATILITY_EDPsi,
		       final double eps)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int iX=n-1;
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double I_IX=I.ev[iX],pI_IX=I.getFPsi(iX,QQpi),p_IX=pI_IX/I_IX,If=I_IX;

    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	QQDi=VOLATILITY_EDPsi.getDiffMatrix(n,QQi);
    TMatr.sumP1(QQDi,QQ0,-If);
    final double [] bQQDi=EVXData.convertMatrFromQtoB_Basis(n,QQDi,I.ev);
    final double [] evDi=M.EV.getEV(n,bQQDi,true);
    if(eps>0)
    {
	for(int k=0;k<n;k++)
	{
	    // diagonal elements are known analytically
	    final double iik=(I.ev[k]-I.ev[iX])*Math.pow(M.B.getD0(I.getPsi(k),I.x0),2);
	    System.err.println("DynDI:bQQDi["+k+","+k+"]="+bQQDi[k*n+k]+" iik="+(iik)+" evDi["+k+"]="+evDi[k]);
	    UnitTests.checkDoubleRel(bQQDi[k*n+k],"bQQDI[k,k]",iik,"(I.ev[k]-I.ev[iX])*psi_k(x0)^2",eps);
	}
    }

    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev);
    final int iDIL=0,iDIH=n-1;
    final double I_DIL=TMatr.mulV2V(EVXData.getPsiFromBasis(iDIL,evDi,n),I.getEValues()),
	P_DIL=TMatr.getAMatrA(n,evDi,n+(iDIL)*n,bQQpi,n)/I_DIL,
	I_DIH=TMatr.mulV2V(EVXData.getPsiFromBasis(iDIH,evDi,n),I.getEValues()),
	P_DIH=TMatr.getAMatrA(n,evDi,n+(iDIH)*n,bQQpi,n)/I_DIH;
    return new DynDI(I,p_IX,evDi[iDIL],P_DIL,evDi[iDIH],P_DIH);
}


public static void matchDynVar2I(final DynDI a,final DynVar2I b,final double eps)
{
    UnitTests.checkDoubleRel(a.I.s0,"a.I.s0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.I.sL,"a.I.sL",b.I_IL,"b.I_IL",eps);
    UnitTests.checkDoubleRel(a.I.sH,"a.I.sH",b.I_IH,"b.I_IH",eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IX,"b.p_IX",eps);
    UnitTests.checkDoubleRel(a.dI_DIL,"a.dI_DIL",b.Di_DIL,"b.Di_DIL",eps);
    UnitTests.checkDoubleRel(a.P_DIL,"a.P_DIL",b.P_DIL,"b.P_DIL",eps);
    UnitTests.checkDoubleRel(a.dI_DIH,"a.dI_DIH",b.Di_DIH,"b.Di_DIH",eps);
    UnitTests.checkDoubleRel(a.P_DIH,"a.P_DIH",b.P_DIH,"b.P_DIH",eps);

    System.err.println("DynDI:matchDynVar2I test OK  eps="+eps);
}

}
