package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *
 *  @version $Id: DynEVdI0.java,v 1.19 2022/07/28 12:27:15 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class DynEVdI0 {



    public final EVXData I;
    public final double p_offset,p_last__with_p_offset_included,p_IH,I_0_feq,dI,Pf,PfdI,PfdI0,var_pI_IX,var_pI_spur;
    DynEVdI0(final EVXData I,final double p_offset,final double p_last__with_p_offset_included,
	     final double p_IH,final double I_0_feq,
	     final double Pf,final double PfdI,final double PfdI0,final double var_pI_IX,final double var_pI_spur,
	     final double eps)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_last__with_p_offset_included=p_last__with_p_offset_included;
	this.p_IH=p_IH;
	this.I_0_feq=I_0_feq;
	final int n=I.neq;
	this.dI=I_0_feq-I.s0;
	this.Pf=Pf;
	this.PfdI=PfdI;
	this.PfdI0=PfdI0;
	this.var_pI_IX=var_pI_IX;
	this.var_pI_spur=var_pI_spur;
    }


/**
 * Solve {@literal <psi|I|psi>/<psi|psi>}=lambda extremum problem, find eigenstates &amp; eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-{@literal <psi0|I|psi0>} (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + {@literal |psi0>dI<psi0|}
 * Find the eigenstates of this new I^{f} matrix,
 * in dynamic impact approximation its eigenstates should be also the P^{f} operator eigenstates.
 *
 * Then consider sum({@literal <psik|pI|psik>}) (dynamic impact approximation)
 * and variate it ith respect to infinitesimal psi shift
 * to find Pf.
 */
public static DynEVdI0 get_DynEVdI0(final MomentsData mm,
				    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				    final VolMatrix VOLATILITY_EDPsi,
				    final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final int ixH=n-1; // the state with maximal I
    final double I_0=I.s0,I_0_feq=I.ev[ixH],dI=I_0_feq-I_0,adI=dI/I.o_o;

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // QQi_feq = QQi + |psi0>dI<psi0|
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	QQi_feq=TMatr.sumAcB(mm.QQi,adI,QQ0);
    final EVXData If=EVXData.getEVXDataFromQQ(n,QQi_feq,mm.QQt,M);
    double spdi=0,sdi=0,sw=0,sdidt=0,spdidt=0,swsdidt=0,sdidt0=0,spdidt0=0,swsdidt0=0;
    for(int k=0;k<n;k++)
    {
	final double [] psi=If.getPsi(k),Dpsi=VOLATILITY_EDPsi.getDPsi(psi);
	final double [] rqpi=RayleighQuotient.getRQVariationValues(psi,Dpsi,null,mm.QQpi,mm.QQt,n),
	    rqi0=TMatr.scaleVectorInPlace(RayleighQuotient.getRQVariationValues(psi,Dpsi,null,QQ0,mm.QQt,n),adI);
	spdi+=rqpi[1];
	sdi+=rqi0[1];
	sw+=rqi0[0];
	final double fdidt=I_0_feq*TMatr.getAMatrA(psi,QQ0,n)-2*adI*TMatr.getAMatrB(psi,Dpsi,QQ0,n);
	sdidt+=fdidt;
	spdidt+=2*TMatr.getAMatrB(psi,Dpsi,mm.QQpi,n);
	swsdidt+=fdidt-2*TMatr.getAMatrB(psi,Dpsi,mm.QQi,n);
	final double wk=Math.pow(TMatr.getAMatrB(If.psi0,psi,mm.QQt,n),2);
	sdidt0+=(I_0_feq-wk*dI)*wk;
	spdidt0+=rqpi[0]*wk;
	swsdidt0+=(I_0_feq-wk*dI-TMatr.getAMatrA(psi,mm.QQi,n))*wk;
	if(eps>0)
	{
	    final double [] rqi=RayleighQuotient.getRQVariationValues(psi,Dpsi,null,mm.QQi,mm.QQt,n);
	    UnitTests.checkDoubleRel(rqi[0]+rqi0[0],"(If)",If.ev[k],"If.ev[k]",eps);
	    UnitTests.checkDoubleRel(rqi[1]+rqi0[1],"var(If)",0,"0",eps);
	    UnitTests.checkDoubleRel(rqi0[0]/adI,"rqi0[0]/adI",Math.pow(M.B.getD0(psi,I.x0),2),"psi^2(x0)",eps);
	}
    }
    final double Pf=-spdi/(sdi),PfdI=spdidt/sdidt,PfdI0=spdidt0/sdidt0;
    if(eps>0)
    {
	UnitTests.checkDoubleRel(sw,"sw",dI,"dI",eps);
	UnitTests.checkDoubleRel(swsdidt,"swsdidt",0,"0",eps);
	UnitTests.checkDoubleRel(swsdidt0,"swsdidt0",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi_feq,n),"<psi0|QQi_feq|psi0>",I_0_feq,"I_0_feq",eps);
    }

    double var_pI_spur=0,var_pI_IX=0;
    for(int k=0;k<n;k++)
    {
	final double [] psi=I.getPsi(k),Dpsi=VOLATILITY_EDPsi.getDPsi(psi);
	final double [] rqpi=RayleighQuotient.getRQVariationValues(psi,Dpsi,null,mm.QQpi,mm.QQt,n);
	var_pI_spur+=rqpi[1];
	if(k==ixH) var_pI_IX=rqpi[1];
	if(eps>0)
	{
	    final double [] rqi=RayleighQuotient.getRQVariationValues(psi,Dpsi,null,mm.QQi,mm.QQt,n);
	    UnitTests.checkDoubleRel(rqi[0],"(I)",I.ev[k],"I.ev[k]",eps);
	    UnitTests.checkDoubleRel(rqi[1],"var(I)",0,"0",eps);
	}
    }
    if(eps>0)
    {
	final DynVar2C c=DynVar2C.get_DynVar2C(n,mm.QQi,mm.QQt,mm.QQpi,
					       M,VOLATILITY_EDPsi,DynVar2C.SELECTION_IH);
	UnitTests.checkDoubleRel(var_pI_IX,"var_pI_IX",c.var_pI_IX,"c.var_pI_IX",eps);
	UnitTests.checkDoubleRel(var_pI_spur,"var_pI_spur",c.var_pI_spur,"c.var_pI_spur",eps);
    }
    if(eps>0)
    {
	System.err.println("DynEVdI0.get_DynEVdI0 selftest OK n="+n+" sdi/adI="+(sdi/adI)+" eps="+eps);
    }

    return new DynEVdI0(I,mm.p_offset,mm.p_last,
			TMatr.getAMatrA(I.psiH,mm.QQpi,n)/I.sH,
			I_0_feq,Pf,PfdI,PfdI0,var_pI_IX,var_pI_spur,eps);
}


static void matchDynEVdI(final DynEVdI0 a,final DynEVdI b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);
    UnitTests.checkDoubleRel(a.Pf,"a.Pf",b.Pf+p_offset_shift,"b.Pf+p_offset_shift",eps);
    System.err.println("DynEVdI0.matchDynEVdI test OK n="+a.I.neq+" eps="+eps);
}
}
