package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;


/** <I\psi^2>=\lambda<\psi^2> is back.
 *  @version $Id: DynIVT.java,v 1.14 2022/07/13 09:29:50 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynIVT {
    final double I_0,p_IL,I_IL,p_IH,I_IH,pa_IL,Ia_IL,pa_IH,Ia_IH,peq_IL,peq_IH;
    final double [] basis;
    final boolean flag_0;
    final double V_IH,V_IL,pV_IH,pV_IL,T_IH,T_IL,p_offset;

    DynIVT(final int n,
	   final double [] QQi,final double [] QQt,final double [] QQpi,
	   final double [] QQV,final double [] QQT,final double [] QQpV,
	   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	   final boolean flag_0,
	   final double p_offset)
    {
	this.p_offset=p_offset;
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");
	this.flag_0=flag_0;
	final int iL,iH;
	if(flag_0)
	{
	    iL=1;iH=n-1;
	    this.basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M);
	    I_0=basis[0];
	}
	else
	{
	    iL=0;iH=n-1;
	    final EVXData ev=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	    this.basis=ev.ev;
	    I_0=ev.s0;
	}

	{
	I_IL=basis[iL];
	final double pI=TMatr.getAMatrA(n,basis,n+iL*n,QQpi,n),
	    V=TMatr.getAMatrA(n,basis,n+iL*n,QQV,n),pV=TMatr.getAMatrA(n,basis,n+iL*n,QQpV,n),
	    T=TMatr.getAMatrA(n,basis,n+iL*n,QQT,n);
	p_IL=pI/I_IL;
	Ia_IL=V/T;
	pa_IL=pV/V;
	peq_IL=(pI-pV/T)/(I_IL-Ia_IL);
	this.V_IL=V;
	this.T_IL=T;
	this.pV_IL=pV;
	}

	{
	I_IH=basis[iH];
	final double pI=TMatr.getAMatrA(n,basis,n+iH*n,QQpi,n),
	    V=TMatr.getAMatrA(n,basis,n+iH*n,QQV,n),pV=TMatr.getAMatrA(n,basis,n+iH*n,QQpV,n),
	    T=TMatr.getAMatrA(n,basis,n+iH*n,QQT,n);
	p_IH=pI/I_IH;
	Ia_IH=V/T;
	pa_IH=pV/V;
	peq_IH=(pI-pV/T)/(I_IH-Ia_IH);
	this.V_IH=V;
	this.T_IH=T;
	this.pV_IH=pV;
	}
    }

    void test_Integration(final int n,final double [] imom,final double [] tmom,final double [] pimom,
			  final java.util.function.Function<double [],double []> intg,
			  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			  final double eps)
    {
	final int iL,iH;
	if(flag_0){iL=1;iH=n-1;}
	else{iL=0;iH=n-1;}
	{
	final double [] psi=EVXData.getPsiFromBasis(iL,basis,n),ipsi2=intg.apply(psi);
	final double T=TMatr.mulVVn1(ipsi2,tmom),V=TMatr.mulVVn1(ipsi2,imom),pV=TMatr.mulVVn1(ipsi2,pimom);
	UnitTests.checkDoubleRel(Ia_IL,"Ia_IL",V/T,"V/T",eps);
	UnitTests.checkDoubleRel(pa_IL,"pa_IL",pV/V,"pV/V",eps);
	UnitTests.checkDoubleRel(M.B.getD0(ipsi2,M.getRadauRoot()),"ipsi2(x0)",1,"1",eps);
	TMatr.scaleVectorInPlace(ipsi2,1/T);
	final double Iv=TMatr.mulVVn1(ipsi2,imom);
	if(Iv<I_IL)
	{
	    if(!flag_0) throw new RuntimeException("Iv="+Iv+" I_IL="+I_IL);
	    else System.err.println("DynIVT:test_Integration:Iv="+Iv+" I_IL="+I_IL+" flag_0="+flag_0);
	}
	}
	{
	final double [] psi=EVXData.getPsiFromBasis(iH,basis,n),ipsi2=intg.apply(psi);
	final double T=TMatr.mulVVn1(ipsi2,tmom),V=TMatr.mulVVn1(ipsi2,imom),pV=TMatr.mulVVn1(ipsi2,pimom);
	UnitTests.checkDoubleRel(Ia_IH,"Ia_IH",V/T,"V/T",eps);
	UnitTests.checkDoubleRel(pa_IH,"pa_IH",pV/V,"pV/V",eps);
	UnitTests.checkDoubleRel(M.B.getD0(ipsi2,M.getRadauRoot()),"ipsi2(x0)",1,"1",eps);
	TMatr.scaleVectorInPlace(ipsi2,1/T);
	final double Iv=TMatr.mulVVn1(ipsi2,imom);
	if(Iv>I_IH)
	{
	    if(!flag_0) throw new RuntimeException("Iv="+Iv+" I_IH="+I_IH);
	    else System.err.println("DynIVT:test_Integration:Iv="+Iv+" I_IH="+I_IH+" flag_0="+flag_0);
	}
	}

	System.err.println("DynIVT:test_Integration: test OK, n="+n+" eps="+eps);
    }

public static void matchPnLDiagProj(final PnLDiagProj a,final DynIVT b,final double eps)
{
    final int n=a.I.neq;
    //EVXData.matchEVXData(a.I,b.I,false,eps);
    final double dp_offset=b.p_offset-a.p_offset;
    UnitTests.checkDoubleArrayRel(a.I.getEValues(),"a.I.getEValues()",java.util.Arrays.copyOf(b.basis,n),"b.basis",eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+dp_offset,"b.p_IH+dp_offset",eps);
    UnitTests.checkDoubleRel(a.p[0],"a.p[0]",b.p_IL+dp_offset,"b.p_IL+dp_offset",eps);
    // For historical reason DynIVT (called from YpnlQVM) has opposite sign of V and p_offset=p_last.
    UnitTests.checkDoubleRel(a.V[0],"a.V[0]",-b.V_IL,"-b.V_IL",eps);
    UnitTests.checkDoubleRel(a.pV[0],"a.pV[0]",-(b.pV_IL+dp_offset*b.V_IL),"-b.pV_IL+offset",eps);
    UnitTests.checkDoubleRel(a.V[n-1],"a.V[n-1]",-b.V_IH,"-b.V_IH",eps);
    UnitTests.checkDoubleRel(a.pV[n-1],"a.pV[n-1]",-(b.pV_IH+dp_offset*b.V_IH),"-b.pV_IH+offset",eps);

    System.err.println("DynIVT.matchPnLDiagProj: test OK, n="+n+" dp_offset="+dp_offset+" eps="+eps);
}
}
