package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/**
 * Try Radau Version.
 *  @version $Id: DynRVarSP0.java,v 1.15 2022/07/14 23:48:04 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynRVarSP0 {
/** Set to a small number about 1e-5 to force constrains check with given precision.
 */
private static final double EPS_FLAG_CHECKCONSTRAINS=1e-5 * 0 ;

final double [] I,DI,o,pI,DpI;
final double v0,p0,pe0,pdI;
/** A class to store liquidity deficit data and capital flow.
 */
DynRVarSP0(final double [] I,final double [] DI,final double [] o,final double [] pI,final double [] DpI)
{
    this.I=I;
    this.DI=DI;
    this.o=o;
    this.pI=pI;
    this.DpI=DpI;
    final double o_o=TMatr.sumElements(o),di_o=TMatr.sumElements(DI),dpi_o=TMatr.sumElements(DpI);
    this.v0=-di_o/o_o;
    this.p0=dpi_o/di_o;
    this.pe0=TMatr.mulVV(pI,o)/TMatr.mulVV(I,o);
    double sd=0;
    for(int k=0;k<DI.length;k++)
    {
	sd+=(pI[k]/I[k])*DI[k];
    }
    this.pdI=sd/di_o;
}


/**
 */
static DynRVarSP0 geRt_DynRVarSP0(final int n,
				  final double [] QQi,final double [] QQt,final double [] QQpi,
				  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int nd=n-1;
    final double [] RQQi=VolMatrix.getRadauMatrix(n,QQi,M),
	RQQt=VolMatrix.getRadauMatrix(n,QQt,M),
	RQQDi=VOLATILITY_EDPsi.getDRadauMatrix(n,QQi,M),
	RQQpi=VolMatrix.getRadauMatrix(n,QQpi,M),
	RQQDpi=VOLATILITY_EDPsi.getDRadauMatrix(n,QQpi,M);
    final double x0=M.getRadauRoot();
    final double [] evR=M.EV.getGEV_Generalized(nd,RQQi,RQQt,true);
    final double [] I=new double [nd],DI=new double [nd],o=new double [nd],pI=new double [nd],DpI=new double [nd];
    for(int k=0;k<nd;k++)
    {
	final double [] psi=EVXData.getPsiFromBasis(k,evR,nd);
	I[k]=evR[k]; //<psi|rI|psi>/<psi|r|psi>
	DI[k]=TMatr.getAMatrA(psi,RQQDi,nd); // <r*psi^2* dI/dt>
	o[k]=-Math.pow(M.B.getD0(psi,x0),2); // (r*psi^2)'_{x0}
	pI[k]=TMatr.getAMatrA(psi,RQQpi,nd); // <psi|rpI|psi>
	DpI[k]=TMatr.getAMatrA(psi,RQQDpi,nd);
	if(EPS_FLAG_CHECKCONSTRAINS>0)
	{
	    final double eps=EPS_FLAG_CHECKCONSTRAINS;
	    final double [] Rpsi=M.getRadauMultiplied(psi),
		Dpsi=VOLATILITY_EDPsi.getDPsi(psi),DRpsi=VOLATILITY_EDPsi.getDPsi(Rpsi),
		RQQD0=VOLATILITY_EDPsi.getDRadauMatrix(n,TMatr.getQQaa(M.B.getX0Moments(n,x0)),M),
		RQQD2t=VolMatrix.getRadauMatrix(n,VOLATILITY_EDPsi.getDiff2Matrix(n,QQt),M);
	    final double Dpsi2X0=(M.B.getD0(M.B.mult2Pol(Rpsi,Dpsi),x0)+
				  M.B.getD0(M.B.mult2Pol(psi,DRpsi),x0)),
		_Dpsi2X0=TMatr.getAMatrA(nd,evR,nd+k*nd,RQQD2t,nd),
		o0=TMatr.getAMatrA(nd,evR,nd+k*nd,RQQD0,nd);


	    System.err.println("DynRVarSP0.get_DynRVarSP0:k="+k+" Dpsi2X0="+Dpsi2X0+" _Dpsi2X0="+_Dpsi2X0+" o0="+o0+" o[k]="+o[k]+" DI[k]="+DI[k]+" o[k]="+o[k]+" I[k]="+I[k]);
	    UnitTests.checkDoubleRel(Dpsi2X0,"Dpsi2X0",o[k],"o[k]",eps);
	    UnitTests.checkDoubleRel(_Dpsi2X0,"_Dpsi2X0",o[k],"o[k]",eps);
	    UnitTests.checkDoubleRel(o0,"o0",o[k],"o[k]",eps);
	    UnitTests.checkDoubleRel(I[k],"I[k]",TMatr.getAMatrB(psi,Rpsi,QQi,n),"<psi|rI|psi>",eps);
	    UnitTests.checkDoubleRel(DI[k],"DI[k]",TMatr.getAMatrB(psi,DRpsi,QQi,n)+TMatr.getAMatrB(Rpsi,Dpsi,QQi,n),"<psi|r dI/dt|psi>",eps);
	    UnitTests.checkDoubleRel(pI[k],"pI[k]",TMatr.getAMatrB(psi,Rpsi,QQpi,n),"<psi|rpI|psi>",eps);
	    UnitTests.checkDoubleRel(DpI[k],"DpI[k]",TMatr.getAMatrB(psi,DRpsi,QQpi,n)+TMatr.getAMatrB(Rpsi,Dpsi,QQpi,n),"<psi|r dpI/dt|psi>",eps);
	    //if(DI[k]>0)throw new RuntimeException("DI[k]="+DI[k]); // not true
	}
    }
    return new DynRVarSP0(I,DI,o,pI,DpI);
}


}

