package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.VolMatrix;

/** Skwewness at psiH.
 *  @version $Id: DynRki.java,v 1.16 2020/09/27 20:33:11 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynRki {
    final double [] eI;
    final double I_0,I_L,I_H,p_IH,ps_IH,vola_IH,pm_IH;
    final Skewness sk;
    DynRki(final double [] eI,
	   final double I_0,final double I_L,final double I_H,
	   final double p_IH,final double ps_IH,final double vola_IH,final double pm_IH,
	   final Skewness sk)
    {
	this.eI=eI;
	this.I_0=I_0;
	this.I_L=I_L;
	this.I_H=I_H;
	this.p_IH=p_IH;
	this.ps_IH=ps_IH;
	this.vola_IH=vola_IH;
	this.pm_IH=pm_IH;
	this.sk=sk;
    }

static DynRki get0_DynRki(final int n,final double [] QQi,final double [] QQt,final double [] QQf,
			  final double [] QQpf,final double [] QQp2f,final double [] QQp3f,
			  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int iL=1,iH=n-1,i0=0;
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M),
	psi=EVXData.getPsiFromBasis(iH,basis,n);

    final double m0=QQf==QQi?basis[iH]:TMatr.getAMatrA(psi,QQf,n),
	m1=TMatr.getAMatrA(psi,QQpf,n),
	m2=TMatr.getAMatrA(psi,QQp2f,n),
	m3=TMatr.getAMatrA(psi,QQp3f,n);
    final Skewness sk=Skewness.getSkewnessFromMonomialMoments(m0,m1,m2,m3);
    final double I_0=basis[i0],I_L=basis[iL],I_H=basis[iH],
	p_IH=m1/m0,
	vola_IH=m2+p_IH*(-2*m1+p_IH*m0),
	ps_IH=(m3+p_IH*(-2*m2+p_IH*m1))/vola_IH,
	pm_IH=0.5*(m3-m2*p_IH)/vola_IH;
    return new DynRki(basis,I_0,I_L,I_H,p_IH,ps_IH,vola_IH,pm_IH,sk);
}

static DynRki getR_DynRki(final int n,final double [] QQi,final double [] QQt,final double [] QQf,
			  final double [] QQpf,final double [] QQp2f,final double [] QQp3f,
			  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int nd=n-1,iL=0,iH=nd-1;
    final double [] basis=M.EV.getGEV_Generalized(nd,VolMatrix.getRadauMatrix(n,QQi,M),
						  VolMatrix.getRadauMatrix(n,QQt,M),true),
	psi=EVXData.getPsiFromBasis(iH,basis,nd),Rpsi=M.getRadauMultiplied(psi);

    final double m0=QQf==QQi?basis[iH]:TMatr.getAMatrB(psi,Rpsi,QQf,n),
	m1=TMatr.getAMatrB(psi,Rpsi,QQpf,n),
	m2=TMatr.getAMatrB(psi,Rpsi,QQp2f,n),
	m3=TMatr.getAMatrB(psi,Rpsi,QQp3f,n);
    final Skewness sk=Skewness.getSkewnessFromMonomialMoments(m0,m1,m2,m3);
    final double I_0=Double.NaN,I_L=basis[iL],I_H=basis[iH],
	p_IH=m1/m0,
	vola_IH=m2+p_IH*(-2*m1+p_IH*m0),
	ps_IH=(m3+p_IH*(-2*m2+p_IH*m1))/vola_IH,
	pm_IH=0.5*(m3-m2*p_IH)/vola_IH;
    return new DynRki(basis,I_0,I_L,I_H,p_IH,ps_IH,vola_IH,pm_IH,sk);
}

static DynRki get_DynRki(final int n,final double [] QQi,final double [] QQt,final double [] QQf,
			  final double [] QQpf,final double [] QQp2f,final double [] QQp3f,
			  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final EVXData ev=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] psi=ev.psiH;

    final double m0=QQf==QQi?ev.sH:TMatr.getAMatrA(psi,QQf,n),
	m1=TMatr.getAMatrA(psi,QQpf,n),
	m2=TMatr.getAMatrA(psi,QQp2f,n),
	m3=TMatr.getAMatrA(psi,QQp3f,n);
    final Skewness sk=Skewness.getSkewnessFromMonomialMoments(m0,m1,m2,m3);
    final double I_0=ev.s0,I_L=ev.sL,I_H=ev.sH,
	p_IH=m1/m0,
	vola_IH=m2+p_IH*(-2*m1+p_IH*m0),
	ps_IH=(m3+p_IH*(-2*m2+p_IH*m1))/vola_IH,
	pm_IH=0.5*(m3-m2*p_IH)/vola_IH;
    //System.err.println("m0="+m0+"\tm1="+m1+"\tm2="+m3+"\tm3="+m3+"\tp_IH="+p_IH+"\tps_IH="+ps_IH+"\tpm_IH="+pm_IH+"\tsk.xa="+sk.xa+"\tsk.xm="+sk.xm+" -0.5b="+(-0.5*(m2*m1-m3*m0)/(m2*m0-m1*m1)));
    return new DynRki(ev.ev,I_0,I_L,I_H,p_IH,ps_IH,vola_IH,pm_IH,sk);
}

static void matchDynRki(final DynRki a,final DynRki b,final double eps)
{
    UnitTests.checkDoubleRel(a.I_L,"a.I_L",b.I_L,"b.I_L",eps);
    UnitTests.checkDoubleRel(a.I_H,"a.I_H",b.I_H,"b.I_H",eps);
    UnitTests.checkDoubleRel(a.I_0,"a.I_0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.vola_IH,"a.vola_IH",b.vola_IH,"b.vola_IH",eps);
    UnitTests.checkDoubleRel(a.ps_IH,"a.ps_IH",b.ps_IH,"b.ps_IH",eps);
    UnitTests.checkDoubleRel(a.pm_IH,"a.pm_IH",b.pm_IH,"b.pm_IH",eps);
    UnitTests.checkDoubleRel(a.sk.xa,"a.sk.xa",b.sk.xa,"b.sk.xa",eps);
    UnitTests.checkDoubleRel(a.sk.xm,"a.sk.xm",b.sk.xm,"b.sk.xm",eps);
    UnitTests.checkDoubleRel(a.sk.sk,"a.sk.sk",b.sk.sk,"b.sk.sk",eps);
    for(DynRki x:new DynRki[]{a,b})
    {
	UnitTests.checkDoubleRel(x.sk.xa,"x.sk.xa",x.p_IH,"x.p_IH",eps);
	UnitTests.checkDoubleRel(x.sk.xm,"x.sk.xm",x.pm_IH,"x.pm_IH",eps);
	UnitTests.checkDoubleRel(2*x.pm_IH-x.p_IH,"2*x.pm_IH-x.p_IH",x.ps_IH,"x.ps_IH",eps);
    }
    System.err.println("DynRki:matchDynRki test OK eps="+eps);
}

}
