package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/**
 *  Try <psi|I|Dpsi> & <psi|I|psi> eigenfunctions.
 *  @version $Id: DynSDPnL.java,v 1.24 2022/08/05 11:25:32 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynSDPnL {

    final EVXData I;
    final double p_IX,pd_IX,I_IX,DI_IX,avar_p_IX,
	p_DIL,pd_DIL,I_DIL,DI_DIL,avar_p_DIL,
	p_DIH,pd_DIH,I_DIH,DI_DIH,avar_p_DIH,
	dist2_IX_DIL,dist2_IX_DIH;

    DynSDPnL(final EVXData I,
	     final double p_IX,final double pd_IX,final double I_IX,final double DI_IX,final double avar_p_IX,
	     final double p_DIL,final double pd_DIL,final double I_DIL,final double DI_DIL,final double avar_p_DIL,
	     final double p_DIH,final double pd_DIH,final double I_DIH,final double DI_DIH,final double avar_p_DIH,
	     final double dist2_IX_DIL,final double dist2_IX_DIH)
    {
	this.I=I;
	this.p_IX=p_IX;
	this.pd_IX=pd_IX;
	this.I_IX=I_IX;
	this.DI_IX=DI_IX;
	this.avar_p_IX=avar_p_IX;
	this.p_DIL=p_DIL;
	this.pd_DIL=pd_DIL;
	this.I_DIL=I_DIL;
	this.DI_DIL=DI_DIL;
	this.avar_p_DIL=avar_p_DIL;
	this.p_DIH=p_DIH;
	this.pd_DIH=pd_DIH;
	this.I_DIH=I_DIH;
	this.DI_DIH=DI_DIH;
	this.avar_p_DIH=avar_p_DIH;
	this.dist2_IX_DIL=dist2_IX_DIL;
	this.dist2_IX_DIH=dist2_IX_DIH;
    }

/** The concept here is to variate
 * var <psi|(p-p*)I|psi>/<psi|psi> with p*=<psi|pI|psi>/<psi|I|psi>.
 * With this definition of p* the psi is now not required to be
 * an eigenfuction of I, so we can try different eigentfuctions.
 * Within a positive constant the variation is a difference between
 * <psi|pI|Dpsi>/<psi|I|Dpsi> and p*.
 * The value of 2<psi|I|Dpsi> at psi=psiH (eigenfunction of I)
 * for Dpsi=-d/dt is I_H*psiH(x0)^2. If we use min/max eigenfuction
 * of <psiA |I|DpsiB>+<DpsiA|I|psiB> the value can be only greater than that.
 */
static DynSDPnL get_DynSDPnL(
    final int n,
    final double [] QQi,final double [] QQt, final double [] QQpi,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi,final K2Distance distance2)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int iX=n-1;
    final double [] bQQDi=EVXData.convertMatrFromQtoB_Basis(n,VOLATILITY_EDPsi.getDiffMatrix(n,QQi),I.ev),
	bQQDpi=EVXData.convertMatrFromQtoB_Basis(n,VOLATILITY_EDPsi.getDiffMatrix(n,QQpi),I.ev),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues());

    final double I_IX=I.ev[iX],pI_IX=bQQpi[iX*n+iX],p_IX=pI_IX/I_IX,
	DI_IX=bQQDi[iX*n+iX],pdI_IX=bQQDpi[iX*n+iX],pd_IX=pdI_IX/DI_IX,
	avar_p_IX=pdI_IX-p_IX*DI_IX;

    final double [] id=M.EV.getEV(n,bQQDi,true);
    final int iDIL=0,iDIH=n-1;
    final double I_DIL=TMatr.getAMatrA(n,id,n+iDIL*n,bQQi,n),
	pI_DIL=TMatr.getAMatrA(n,id,n+iDIL*n,bQQpi,n),p_DIL=pI_DIL/I_DIL,
	DI_DIL=id[iDIL],pdI_DIL=TMatr.getAMatrA(n,id,n+iDIL*n,bQQDpi,n),pd_DIL=pdI_DIL/DI_DIL,
	avar_p_DIL=pdI_DIL-p_DIL*DI_DIL,
	I_DIH=TMatr.getAMatrA(n,id,n+iDIH*n,bQQi,n),
	pI_DIH=TMatr.getAMatrA(n,id,n+iDIH*n,bQQpi,n),p_DIH=pI_DIH/I_DIH,
	DI_DIH=id[iDIH],pdI_DIH=TMatr.getAMatrA(n,id,n+iDIH*n,bQQDpi,n),pd_DIH=pdI_DIH/DI_DIH,
	avar_p_DIH=pdI_DIH-p_DIH*DI_DIH;

    final double dist2_IX_DIL=distance2.getDistance2(I.psiH,EVXData.convertPsiFromBtoQ_Basis(n,EVXData.getPsiFromBasis(iDIL,id,n),I.ev)),
	dist2_IX_DIH=distance2.getDistance2(I.psiH,EVXData.convertPsiFromBtoQ_Basis(n,EVXData.getPsiFromBasis(iDIH,id,n),I.ev));

    return new DynSDPnL(I,
			p_IX,pd_IX,I_IX,DI_IX,avar_p_IX,
			p_DIL,pd_DIL,I_DIL,DI_DIL,avar_p_DIL,
			p_DIH,pd_DIH,I_DIH,DI_DIH,avar_p_DIH,
			dist2_IX_DIL,dist2_IX_DIH);
}


static DynSDPnL test_get_DynSDPnL(
    final int n,
    final double [] QQi,final double [] QQt, final double [] QQpi,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi,final K2Distance distance2)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int iX=n-1;
    final double [] QQDi=VOLATILITY_EDPsi.getDiffMatrix(n,QQi),QQDpi=VOLATILITY_EDPsi.getDiffMatrix(n,QQpi);
    final double I_IX=I.ev[iX],pI_IX=I.getFPsi(iX,QQpi),p_IX=pI_IX/I_IX,
	DI_IX=I.getFPsi(iX,QQDi),pdI_IX=I.getFPsi(iX,QQDpi),pd_IX=pdI_IX/DI_IX,
	avar_p_IX=pdI_IX-p_IX*DI_IX;

    final EVXData ID=EVXData.getEVXDataFromQQ(n,QQDi,QQt,M);
    final int iDIL=0,iDIH=n-1;
    final double I_DIL=ID.getFPsi(iDIL,QQi),pI_DIL=ID.getFPsi(iDIL,QQpi),p_DIL=pI_DIL/I_DIL,
	DI_DIL=ID.sL,pdI_DIL=ID.getFPsi(iDIL,QQDpi),pd_DIL=pdI_DIL/DI_DIL,
	avar_p_DIL=pdI_DIL-p_DIL*DI_DIL,
	I_DIH=ID.getFPsi(iDIH,QQi),pI_DIH=ID.getFPsi(iDIH,QQpi),p_DIH=pI_DIH/I_DIH,
	DI_DIH=ID.sH,pdI_DIH=ID.getFPsi(iDIH,QQDpi),pd_DIH=pdI_DIH/DI_DIH,
	avar_p_DIH=pdI_DIH-p_DIH*DI_DIH;

    final double dist2_IX_DIL=distance2.getDistance2(I.psiH,ID.psiL),
	dist2_IX_DIH=distance2.getDistance2(I.psiH,ID.psiH);
    if(true)
    {
	final double eps=1e-5;
	UnitTests.checkDoubleRel(DI_IX,"DI_IX",I_IX*Math.pow(M.B.getD0(I.psiH,I.x0),2),"I_IX*psi_IX(x0)^2",eps);
	UnitTests.checkDoubleRel(DI_DIL,"DI_DIL",2*TMatr.getAMatrB(ID.psiL,VOLATILITY_EDPsi.getDPsi(ID.psiL),QQi,n),"2<psi_DIL|I|Dpsi_DIL>",eps);
	UnitTests.checkDoubleRel(DI_DIH,"DI_DIH",2*TMatr.getAMatrB(ID.psiH,VOLATILITY_EDPsi.getDPsi(ID.psiH),QQi,n),"2<psi_DIH|I|Dpsi_DIH>",eps);
	System.err.println("DynSDPnL DI_IX="+DI_IX+" DI_DIH="+DI_DIH+" DI_DIL="+DI_DIL);
	if(DI_IX>DI_DIH) throw new RuntimeException("DI_IX"+DI_IX+" DI_DIH="+DI_DIH);
	System.err.println("DynSDPnL:test OK n="+n+" eps="+eps);
    }

    return new DynSDPnL(I,
			p_IX,pd_IX,I_IX,DI_IX,avar_p_IX,
			p_DIL,pd_DIL,I_DIL,DI_DIL,avar_p_DIL,
			p_DIH,pd_DIH,I_DIH,DI_DIH,avar_p_DIH,
			dist2_IX_DIL,dist2_IX_DIH);
}

public static void matchDynSDPnL(final DynSDPnL a,final DynSDPnL b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,false,eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IX,"b.p_IX",eps);
    UnitTests.checkDoubleRel(a.pd_IX,"a.pd_IX",b.pd_IX,"b.pd_IX",eps);
    UnitTests.checkDoubleRel(a.I_IX,"a.I_IX",b.I_IX,"b.I_IX",eps);
    UnitTests.checkDoubleRel(a.DI_IX,"a.DI_IX",b.DI_IX,"b.DI_IX",eps);
    UnitTests.checkDoubleRel(a.avar_p_IX,"a.avar_p_IX",b.avar_p_IX,"b.avar_p_IX",eps);

    UnitTests.checkDoubleRel(a.p_DIL,"a.p_DIL",b.p_DIL,"b.p_DIL",eps);
    UnitTests.checkDoubleRel(a.pd_DIL,"a.pd_DIL",b.pd_DIL,"b.pd_DIL",eps);
    UnitTests.checkDoubleRel(a.I_DIL,"a.I_DIL",b.I_DIL,"b.I_DIL",eps);
    UnitTests.checkDoubleRel(a.DI_DIL,"a.DI_DIL",b.DI_DIL,"b.DI_DIL",eps);
    UnitTests.checkDoubleRel(a.avar_p_DIL,"a.avar_p_DIL",b.avar_p_DIL,"b.avar_p_DIL",eps);

    UnitTests.checkDoubleRel(a.p_DIH,"a.p_DIH",b.p_DIH,"b.p_DIH",eps);
    UnitTests.checkDoubleRel(a.pd_DIH,"a.pd_DIH",b.pd_DIH,"b.pd_DIH",eps);
    UnitTests.checkDoubleRel(a.I_DIH,"a.I_DIH",b.I_DIH,"b.I_DIH",eps);
    UnitTests.checkDoubleRel(a.DI_DIH,"a.DI_DIH",b.DI_DIH,"b.DI_DIH",eps);
    UnitTests.checkDoubleRel(a.avar_p_DIH,"a.avar_p_DIH",b.avar_p_DIH,"b.avar_p_DIH",eps);

    UnitTests.checkDouble(a.dist2_IX_DIL,"a.dist2_IX_DIL",b.dist2_IX_DIL,"b.dist2_IX_DIL",eps);
    UnitTests.checkDouble(a.dist2_IX_DIH,"a.dist2_IX_DIH",b.dist2_IX_DIH,"b.dist2_IX_DIH",eps);

    System.err.println("DynSDPnL:matchDynSDPnL test OK neq="+a.I.neq+" eps="+eps);
}
public static void testVar1(final DynSDPnL a,
    final int n,final double [] QQi,final double [] QQt,
    final double [] QQpi,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi,final double eps)
{
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M),
	ID=EVXData.getEVXDataFromQQ(n,VOLATILITY_EDPsi.getDiffMatrix(n,QQi),QQt,M);
    {
    final DynVar1 v=new DynVar1(I.psiH,QQi,QQt,QQpi,VOLATILITY_EDPsi);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",v.p,"v.p",eps);
    UnitTests.checkDoubleRel(a.pd_IX,"a.pd_IX",v.pd,"v.pd",eps);
    UnitTests.checkDoubleRel(a.I_IX,"a.I_IX",v.I,"v.I",eps);
    UnitTests.checkDoubleRel(a.DI_IX,"a.DI_IX",v.DI,"v.DI",eps);
    UnitTests.checkDoubleRel(a.avar_p_IX,"a.avar_p_IX",v.avar_p,"v.avar_p",eps);
    UnitTests.checkDoubleRel(a.avar_p_IX,"a.avar_p_IX",2*v.a.x_a_b,"2*v.a.x_a_b",eps);
    }
    {
    final DynVar1 v=new DynVar1(ID.psiL,QQi,QQt,QQpi,VOLATILITY_EDPsi);
    UnitTests.checkDoubleRel(a.p_DIL,"a.p_DIL",v.p,"v.p",eps);
    UnitTests.checkDoubleRel(a.pd_DIL,"a.pd_DIL",v.pd,"v.pd",eps);
    UnitTests.checkDoubleRel(a.I_DIL,"a.I_DIL",v.I,"v.I",eps);
    UnitTests.checkDoubleRel(a.DI_DIL,"a.DI_DIL",v.DI,"v.DI",eps);
    UnitTests.checkDoubleRel(a.avar_p_DIL,"a.avar_p_DIL",v.avar_p,"v.avar_p",eps);
    UnitTests.checkDoubleRel(a.avar_p_DIL,"a.avar_p_DIL",2*v.a.x_a_b,"2*v.a.x_a_b",eps);
    }
    {
    final DynVar1 v=new DynVar1(ID.psiH,QQi,QQt,QQpi,VOLATILITY_EDPsi);
    UnitTests.checkDoubleRel(a.p_DIH,"a.p_DIH",v.p,"v.p",eps);
    UnitTests.checkDoubleRel(a.pd_DIH,"a.pd_DIH",v.pd,"v.pd",eps);
    UnitTests.checkDoubleRel(a.I_DIH,"a.I_DIH",v.I,"v.I",eps);
    UnitTests.checkDoubleRel(a.DI_DIH,"a.DI_DIH",v.DI,"v.DI",eps);
    UnitTests.checkDoubleRel(a.avar_p_DIH,"a.avar_p_DIH",v.avar_p,"v.avar_p",eps);
    UnitTests.checkDoubleRel(a.avar_p_DIH,"a.avar_p_DIH",2*v.a.x_a_b,"2*v.a.x_a_b",eps);
    }
    System.err.println("DynSDPnL:testVar1 test OK neq="+a.I.neq+" eps="+eps);
}

}
