package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;

class DynVH2PnL {

    final double p_IH,var_p_IH;
    final double var2tt_pnlL,var2tt_pnlH,var2tt0_pnlL,var2tt0_pnlH,
	var2pt_pnlL,var2pt_pnlH,
	pnlL,pnlH,pnl0;
    final EVXData I;

    DynVH2PnL(final EVXData I,
	      final double p_IH,final double var_p_IH,
	      final double var2tt_pnlL,final double var2tt_pnlH,
	      final double var2tt0_pnlL,final double var2tt0_pnlH,
	      final double var2pt_pnlL,final double var2pt_pnlH,
	      final double pnlL,final double pnlH,final double pnl0
	      )
    {
	this.I=I;
	this.p_IH=p_IH;
	this.var_p_IH=var_p_IH;
	this.var2tt_pnlL=var2tt_pnlL;
	this.var2tt_pnlH=var2tt_pnlH;
	this.var2tt0_pnlL=var2tt0_pnlL;
	this.var2tt0_pnlH=var2tt0_pnlH;
	this.var2pt_pnlL=var2pt_pnlL;
	this.var2pt_pnlH=var2pt_pnlH;
	this.pnlL=pnlL;
	this.pnlH=pnlH;
	this.pnl0=pnl0;
    }

/** A remiscent of DynVHPnL.java.
 * Consider I|psi>=\lambda |psi>, corresponding to {max \lambda ; psiH}.
 * Calculate p_IH=<psiH|pI|psiH>/<psiH|I|psiH>,
 * then consider |psi> states, corresponding to min/max:
 * P&L=<psi|(p-p_IH) I |psi>/<psi|psi>
 * and consider second P&L variations of two forms:
 * 1) D(psi) ; D(p_IH)
 * 2) D(psi) ; D(psi)
 * First P&L variation over D(psi) is zero.
 */
static DynVH2PnL get_DynVH2PnL(final int n,
			       final double [] QQi,final double [] QQt,
			       final double [] QQpi,
			       final VolMatrix VOLATILITY_EDPsi,
			       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			       final double eps)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");

    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int iX=n-1; // maximal I
    final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
    final double pI_IH=I.getFPsi(iX,QQpi),p_IH=pI_IH/I.ev[iX],
	var_p_IH=2*(I.getFPsiPsi(iX,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iX,Dpsi_IH,QQt));

    final double [] QQpnl=TMatr.sumAcB(QQpi,-p_IH,QQi);
    final EVXData Ipnl=EVXData.getEVXDataFromQQ(n,QQpnl,QQt,M);
    final double [] psipnlL=Ipnl.psiL,DpsipnlL=VOLATILITY_EDPsi.getDPsi(psipnlL),
	psipnlH=Ipnl.psiH,DpsipnlH=VOLATILITY_EDPsi.getDPsi(psipnlH);
    final double []
	rqpnlL=RayleighQuotient.getRQVariationValues(psipnlL,DpsipnlL,null,QQpnl,QQt,n),
	rqpnlH=RayleighQuotient.getRQVariationValues(psipnlH,DpsipnlH,null,QQpnl,QQt,n),
	rq0pnlL=RayleighQuotient.getRQVariationValues(psipnlL,Ipnl.psi0,null,QQpnl,QQt,n),
	rq0pnlH=RayleighQuotient.getRQVariationValues(psipnlH,Ipnl.psi0,null,QQpnl,QQt,n),
	rqIpnlL=RayleighQuotient.getRQVariationValues(psipnlL,DpsipnlL,null,QQi,QQt,n),
	rqIpnlH=RayleighQuotient.getRQVariationValues(psipnlH,DpsipnlH,null,QQi,QQt,n);
    final double pnlL=Ipnl.sL,var2tt_pnlL=rqpnlL[2],var2tt0_pnlL=rq0pnlL[2],
	pnlH=Ipnl.sH,var2tt_pnlH=rqpnlH[2],var2tt0_pnlH=rq0pnlH[2],
	pnl0=TMatr.getAMatrA(Ipnl.psi0,QQpnl,n),
	var2pt_pnlL=rqIpnlL[1],var2pt_pnlH=rqIpnlH[1];
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_IH,QQpnl,n),"<psiH|PnL|psiH>",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_IH,QQt,n),"<psiH|psiH>",1,"1",eps);
	UnitTests.checkDoubleRel(rqpnlL[0],"rqpnlL[0]",pnlL,"pnlL",eps);
	UnitTests.checkDoubleRel(rqpnlH[0],"rqpnlH[0]",pnlH,"pnlH",eps);
	UnitTests.checkDoubleRel(rqpnlL[1],"rqpnlL[1]",0,"0",eps);
	UnitTests.checkDoubleRel(rqpnlH[1],"rqpnlH[1]",0,"0",eps);
	UnitTests.checkDoubleRel(rqpnlL[2],"rqpnlL[2]",2*(TMatr.getAMatrA(DpsipnlL,QQpnl,n)-Ipnl.sL*TMatr.getAMatrA(DpsipnlL,QQt,n)),"var2 PnL L",eps);
	UnitTests.checkDoubleRel(rqpnlH[2],"rqpnlH[2]",2*(TMatr.getAMatrA(DpsipnlH,QQpnl,n)-Ipnl.sH*TMatr.getAMatrA(DpsipnlH,QQt,n)),"var2 PnL H",eps);

	UnitTests.checkDoubleRel(pnl0,"pnl0",Ipnl.s0,"Ipnl.s0",eps);
	UnitTests.checkDoubleRel(rq0pnlL[0],"rq0pnlL[0]",pnlL,"pnlL",eps);
	UnitTests.checkDoubleRel(rq0pnlH[0],"rq0pnlH[0]",pnlH,"pnlH",eps);
	UnitTests.checkDoubleRel(rq0pnlL[1],"rq0pnlL[1]",0,"0",eps);
	UnitTests.checkDoubleRel(rq0pnlH[1],"rq0pnlH[1]",0,"0",eps);
	UnitTests.checkDoubleRel(rq0pnlL[2],"rq0pnlL[2]",2*(TMatr.getAMatrA(Ipnl.psi0,QQpnl,n)-Ipnl.sL*TMatr.getAMatrA(Ipnl.psi0,QQt,n)),"var2_0 PnL L",eps);
	UnitTests.checkDoubleRel(rq0pnlH[2],"rq0pnlH[2]",2*(TMatr.getAMatrA(Ipnl.psi0,QQpnl,n)-Ipnl.sH*TMatr.getAMatrA(Ipnl.psi0,QQt,n)),"var2_0 PnL H",eps);

	System.err.println("DynVH2PnL: test OK, n="+n+" eps="+eps);

    }


    return new DynVH2PnL(I,p_IH,var_p_IH,
			 var2tt_pnlL,var2tt_pnlH,
			 var2tt0_pnlL,var2tt0_pnlH,
			 var2pt_pnlL,var2pt_pnlH,
			 pnlL,pnlH,pnl0);
}

/*
public static void matchDynHPnL(final DynVH2PnL a,final DynHPnL b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,false,eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.var_p_IH,"a.var_p_IH",b.var_p_IH,"b.var_p_IH",eps);
    UnitTests.checkDoubleRel(a.ib_IH,"a.ib_IH",b.ib_IH,"b.ib_IH",eps);

    UnitTests.checkDoubleRel(a.var_p_IH,"a.var_p_IH",a.wk[a.I.neq-1],"a.wk[a.I.neq-1]",eps);
    UnitTests.checkDoubleRel(b.var_p_IH,"b.var_p_IH",b.wk[b.I.neq-1],"b.wk[b.I.neq-1]",eps);
    UnitTests.checkDoubleRel(a.sw_p_IH,"a.sw_p_IH",b.sw_p_IH,"b.sw_p_IH",eps);
    UnitTests.checkDoubleRel(a.pb_IH,"a.pb_IH",b.pb_IH,"b.pb_IH",eps);
    UnitTests.checkDoubleRel(a.ib_IH,"a.ib_IH",b.ib_IH,"b.ib_IH",eps);
    UnitTests.checkDoubleRel(a.pd_IH,"a.pd_IH",b.pd_IH,"b.pd_IH",eps);
    UnitTests.checkDoubleRel(a.id_IH,"a.id_IH",b.id_IH,"b.id_IH",eps);

    System.err.println("DynVH2PnL:matchDynVHPnL test OK neq="+a.I.neq+" eps="+eps);
}
*/

public static void matchDynVarPnLSpur(final DynVH2PnL a,final DynVarPnLSpur b,final double eps)
{
    final int n=a.I.neq;
    EVXData.matchEVXData(a.I,b.I,false,false,eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.var_p_IH,"a.var_p_IH",b.var_p_IH,"b.var_p_IH",eps);
    UnitTests.checkDoubleRel(a.var2tt_pnlL,"a.var2tt_pnlL",b.var2PnL[0],"b.var2PnL[0]",eps);
    UnitTests.checkDoubleRel(a.var2tt_pnlH,"a.var2tt_pnlH",b.var2PnL[n-1],"b.var2PnL[n-1]",eps);
    UnitTests.checkDoubleRel(a.pnlL,"a.pnlL",b.Ipnl.sL,"b.Ipnl.sL",eps);
    UnitTests.checkDoubleRel(a.pnlH,"a.pnlH",b.Ipnl.sH,"b.Ipnl.sH",eps);

    System.err.println("DynVH2PnL:matchDynVarPnLSpur test OK n="+n+" eps="+eps);
}

}
