package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Used for unit tests var1.
 * Takes J=<psi|(p-p*)I|psi>/<psi|psi>, where p*=<psi|pI|psi>/<psi|I|psi>
 * and variate over infinitesimal shift of psi, var psi=D(psi),
 * what gives avar_p=2(<psi|pI|Dpsi>-p*<psi|I|Dpsi>). (There are no <psi|Dpsi>
 * term because of J(psi)=0). Given the psi is
 * choosen such as <psi|I|Dpsi> positive(e.g. I eigenvector)
 * this gives avar_p=2<psi|I|Dpsi>*(pd-p*), where pd=<psi|pI|Dpsi>/<psi|I|Dpsi>.
 * As a |psi> some eigenvectors can be used
 * (either I eigenvectors or (D^T I +I D) eigenvectors).
 */
class DynVar1 {
    final double p,pd,I,DI,avar_p,var_I,var_pI;
    final ACorr a;

    DynVar1(final double [] psi,
	    final double [] QQi,final double [] QQt,final double [] QQpi,
	    final VolMatrix VOLATILITY_EDPsi)
    {
	final int n=psi.length;
	if(n<2||QQt.length!=(n*n)) throw new IllegalArgumentException("n="+n+"<2");

	final double pI=TMatr.getAMatrA(psi,QQpi,n);
	this.I=TMatr.getAMatrA(psi,QQi,n);
	this.p=pI/I;
	final double [] Dpsi=VOLATILITY_EDPsi.getDPsi(psi);
	this.DI=2*TMatr.getAMatrB(psi,Dpsi,QQi,n);
	final double DpI=2*TMatr.getAMatrB(psi,Dpsi,QQpi,n),dt=2*TMatr.getAMatrB(psi,Dpsi,QQt,n);
	this.pd=DpI/this.DI;
	this.var_I=DI-I*dt; // dt=psi(x0)^2
	this.var_pI=DpI-pI*dt;
	// note: avar_p=2*a.x_a_b
	this.avar_p=(DpI-p*DI);
	this.a=ACorr.getACorr_OrtPsiFromQQ(psi,TMatr.sumAcB(QQpi,-p,QQi),TMatr.mulMM(n,QQt,VOLATILITY_EDPsi.getDMatrix(n)),QQt);
    }

    /** Radau version.
     * Note: "a" here is incorrect.
     */
    DynVar1(final double [] psi,
	    final double [] QQi,final double [] QQt,final double [] QQpi,
	    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	    final VolMatrix VOLATILITY_EDPsi)
    {
	final int nd=psi.length,n=nd+1;
	if(n<2||QQt.length!=(n*n)) throw new IllegalArgumentException("n="+n+"<2");
	final double [] Rpsi=M.getRadauMultiplied(psi);

	final double pI=TMatr.getAMatrB(psi,Rpsi,QQpi,n);
	this.I=TMatr.getAMatrB(psi,Rpsi,QQi,n);
	this.p=pI/I;
	final double [] Dpsi=VOLATILITY_EDPsi.getDPsi(psi),DRpsi=VOLATILITY_EDPsi.getDPsi(Rpsi);
	this.DI=TMatr.getAMatrB(Rpsi,Dpsi,QQi,n)+TMatr.getAMatrB(psi,DRpsi,QQi,n);
	final double DpI=TMatr.getAMatrB(Rpsi,Dpsi,QQpi,n)+TMatr.getAMatrB(psi,DRpsi,QQpi,n),
	    dt=(TMatr.getAMatrB(Rpsi,Dpsi,QQt,n)+TMatr.getAMatrB(psi,DRpsi,QQt,n));
	this.pd=DpI/this.DI;
	this.var_I=DI-I*dt; // dt is 0 for Radau
	this.var_pI=DpI-pI*dt;
	this.avar_p=(DpI-p*DI);
	// note the "a" here is incorrect because D operation changes the basis.
	this.a=ACorr.getACorr_OrtPsiFromQQ(psi,VolMatrix.getRadauMatrix(n,TMatr.sumAcB(QQpi,-p,QQi),M),
					   VOLATILITY_EDPsi.getDRadauMatrix(n,QQt,M), // =0 !!!!
					   VolMatrix.getRadauMatrix(n,QQt,M));
    }
}
