package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: DynVar2DIDynImpact.java,v 1.13 2022/07/28 11:15:38 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVar2DIDynImpact {
    final EVXData I;
    final double p_IL,p_IH,pa_sum;
    final double var2_ppIH,peq_I0,peq_oI0;
    final ACorr a,aa;
    DynVar2DIDynImpact(final EVXData I,
		       final double [] dp2,final double [] dI2,final double [] p,final double [] psi2x0,
		       final double [] dpdI,final double [] dpdp,final double [] dIdI,
		       final double [] dpI2,final double peq_I0,final double peq_oI0)
{
    this.I=I;
    this.p_IL=p[0];
    this.p_IH=p[p.length-1];
    this.pa_sum=TMatr.mulVVn1(p,I.ev)/TMatr.sumElements(I.getEValues());
    this.a=new ACorr(TMatr.mulVV(dp2,dI2),TMatr.mulVV(dp2,dp2),TMatr.mulVV(dI2,dI2));
    this.aa=new ACorr(TMatr.sumElements(dpdI),TMatr.sumElements(dpdp),TMatr.sumElements(dIdI));
    var2_ppIH=dpI2[dpI2.length-1];
    this.peq_I0=peq_I0;
    this.peq_oI0=peq_oI0;

    if(PRINT_DIAG)
    {
	System.err.println("DynVar2DIDynImpact: I0="+I.s0+" = spur(I,psi^2(x0))/spur(psi^2(x0))="+(TMatr.mulVVn1(psi2x0,I.ev)/TMatr.sumElements(psi2x0)));
    }
}

    static final boolean PRINT_DIAG=false;

/**
 * For psi, I eigenvector: I|psi>=lambda |psi>
 * variate <psi|I|psi>/<psi|psi> and <psi|pI|psi>/<psi|psi>
 * on infinitesimal psi time-shift variation |Dpsi>=|D|psi>,
 * then consider dynamic impact through density matrix state.
 */
static DynVar2DIDynImpact get_DynVar2DIDynImpact(final int n,
					     final double [] QQi,final double [] QQt,final double [] QQpi,
					     final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] dp2=new double [n],dI2=new double [n],dpI2=new double [n],_dpI2=new double [n],p=new double [n],pI=new double [n],psi2x0=new double [n],
	dpdI=new double [n],dpdp=new double [n],dIdI=new double [n];
    for(int k=0;k<I.neq;k++)
    {
	pI[k]=I.getFPsi(k,QQpi);
	p[k]=pI[k]/I.ev[k];
    }
    double si0=0,spi0=0,spio0=0;
    for(int k=0;k<I.neq;k++)
    {
	final double di0=2*(I.s0-I.ev[k]),dpi0=2*(TMatr.getAMatrA(I.psi0,QQpi,n)-pI[k]);
	if(di0<0)
	{
	    si0+=di0;
	    spi0+=dpi0;
	}
	else
	{
	    si0-=di0;
	    spi0-=dpi0;
	}
	final double [] psik=I.getPsi(k),Dpsik=VOLATILITY_EDPsi.getDPsi(psik);
	final double psik2x0=2*TMatr.getAMatrB(psik,Dpsik,QQt,n);
	psi2x0[k]=psik2x0;

	double sdI=0,sdp=0,sdpI=0,sdpdI=0,sdpdp=0,sdIdI=0,sddpI=0,sdpI0=0,sdI0=0;
	for(int l=0;l<I.neq;l++)
	{
	    if(k==l) continue;

	    final double w=2*Math.pow(I.getFPsiPsi(l,Dpsik,QQt),2),
		jdp=p[l]-p[k],jdI=I.ev[l]-I.ev[k],jdpI=pI[l]-pI[k],
		w0=2*Math.pow(I.getFPsiPsi(l,I.psi0,QQt),2);
	    sdI0+=w0*jdI;
	    sdpI0+=w0*jdpI;
	    sdI+=w*jdI;
	    sdp+=w*jdp;
	    sdpI+=w*jdpI;
	    sddpI+=w*jdp*I.ev[l];
	    sdpdI+=w*jdp*jdI;
	    sdpdp+=w*jdp*jdp;
	    sdIdI+=w*jdI*jdI;
	}
	dp2[k]=sdp;
	dI2[k]=sdI;
	dpI2[k]=sdpI-p[k]*sdI;
	_dpI2[k]=sddpI;
	dpdI[k]=sdpdI;
	dpdp[k]=sdpdp;
	dIdI[k]=sdIdI;

	if(di0<0)
	{
	    spio0+=sdpI0;
	}
	else
	{
	    spio0-=sdpI0;
	}

	if(PRINT_DIAG)
	{
	    System.err.println("DynVar2DIDynImpact: sdI0="+sdI0+" di0="+di0);
	    System.err.println("DynVar2DIDynImpact: sdpI0="+sdpI0+" dpi0="+dpi0);
	}
    }
    final double peq_I0=spi0/si0,peq_oI0=spio0/si0;
    if(PRINT_DIAG)
    {
	System.err.println("dI2="+java.util.Arrays.toString(dI2));
	System.err.println("dpI2="+java.util.Arrays.toString(dpI2));
	System.err.println("_dpI2="+java.util.Arrays.toString(_dpI2));
    }
    return new DynVar2DIDynImpact(I,dp2,dI2,p,psi2x0,dpdI,dpdp,dIdI,dpI2,peq_I0,peq_oI0);
}


static void matchDynVar2DynImpact(final DynVar2DIDynImpact a,final DynVar2DynImpact b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.p_IL,"a.p_IL",b.p_IL,"b.p_IL",eps);
    UnitTests.checkDoubleRel(a.pa_sum,"a.pa_sum",b.pa_sum,"b.pa_sum",eps);
    UnitTests.checkDoubleRel(a.a.x_b_b,"a.a.x_b_b",b.a.x_b_b,"b.a.x_b_b",eps);

    System.err.println("DynVar2DIDynImpact:matchDynVar2DynImpact test OK n="+a.I.neq+" eps="+eps);
}


}

