package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: DynVar2DynImpactDI.java,v 1.39 2022/07/28 22:30:49 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVar2DynImpactDI {
    final double I_0,I_IL,I_IH,p_IH,var_pI_IH;
    final double dp_plus,dp_minus,dpI_plus,dpI_minus,dpdI_plus,dpdI_minus,
	dp_dIL,dp_dIH,dI_dIL,dI_dIH;
    final ACorr a_dpdI,a_dpdII;
    final double [] dI,dp,i,p,psi2x0;
    DynVar2DynImpactDI(final double I_0,final double I_IL,final double I_IH,final double p_IH,final double var_pI_IH,
		       final double [] dI,
		       final double [] dp,final double [] i,final double [] p,final double [] psi2x0)
{
    this.I_0=I_0;
    this.I_IL=I_IL;
    this.I_IH=I_IH;
    this.p_IH=p_IH;
    this.var_pI_IH=var_pI_IH;
    this.dI=dI;
    this.dp=dp;
    this.i=i;
    this.p=p;
    this.psi2x0=psi2x0;
    final int n=dp.length;
    dp_dIL=dp[0];
    dp_dIH=dp[n-1];
    dI_dIL=dI[0];
    dI_dIH=dI[n-1];

    final double [] dII=new double [n];
    double dp_plus=0,dp_minus=0,dpI_plus=0,dpI_minus=0,dpdI_plus=0,dpdI_minus=0;
    for(int k=0;k<n;k++)
    {
	final double sgn=Math.signum(dI[k]);
	if(sgn>0)
	{
	    dp_plus+=dp[k];
	    dpI_plus+=dp[k]*i[k];
	    dpdI_plus+=dp[k]*dI[k];
	}
	else
	{
	    dp_minus+=dp[k];
	    dpI_minus+=dp[k]*i[k];
	    dpdI_minus+=dp[k]*dI[k];
	}
	dII[k]=dI[k]/i[k];
    }
    this.dp_plus=dp_plus;
    this.dp_minus=dp_minus;
    this.dpI_plus=dpI_plus;
    this.dpI_minus=dpI_minus;
    this.dpdI_plus=dpdI_plus;
    this.dpdI_minus=dpdI_minus;
    this.a_dpdI=ACorr.getACorr_ABW(n,dp,dI,null);
    this.a_dpdII=ACorr.getACorr_ABW(n,dp,dII,null);
}


/**
 * On psi state, dI/dt eigenvector: dI/dt|psi>=lambda |psi>,
 * calculate dp/dt (here I=dV/dt).
 * Experimental evidences, presented in my
 * "Market Dynamics. On Supply and Demand Concepts, Feb 14, 2016, arXiv:1602.04423"
 * paper show that price sensitivity to execution flow (dynamic impact: dp <-> dI)
 * is much greater than price sensitivity to volume (regular impact: dp <-> dV).
 * From this observation immediately follows that |psi> states of dI/dt operator
 * should correspond to the states of large dp/dt.
 * Calculation of <Q_i|dI/dt|Q_j> matrix can be done by parts
 * (possible only for some measures, Dirk Laurie personal communications):
 * <Q_i|dI/dt|Q_j> = I^{*} Q_i(x0) Q_j(x0) - (<Q_i|I|D(Q_j)> + <D(Q_i)|I|Q_j>)
 * but edge value I^{*} is still unknown and cannot be determined from dynamic equation.
 * I previously tried I_0, I_{IH} and other values without much success.
 * Let us now try to use I^{*}=0 (Aug, 13 2016, an idea I got in Pier 62, Hudson River Park)
 * only for the reason it simplifies calculations.
 * The low I_0 states (enter condition) have I^{*}=0 as a good approximation,
 * and for high I_0 states (exit condition) specific I^{*} value is often irrelevant.
 * Then use matrix averages (introduced in arXiv:1510.05510 Appendix E,
 * which are related to quantum mechanics density matrix arXiv:1512.03219)
 * for dynamic impact estimation (as correlation of dp/dt and dI/dt on above selected |psi> states).
 * Matrix average approach remove one "differentiation" e.g. matrix average of dI/dt behave like I.
 * A question arise whether all |psi> states or only the states with dI/dt>0 should be used
 * for dynamic impact calculation? Numerical experiments seems to show, that
 * only growing execution flow dI/dt>0 states are relevant to dynamic impact calculation
 * and dI/dt<0 states seems to be "relaxational" ones and possess no dynamical information.
 */
static DynVar2DynImpactDI get_DynVar2DynImpactDI(final int n,
						 final double [] QQi,final double [] QQt,final double [] QQpi,
						 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int i0=0,iL=1,iH=n-1;
    // Use a basis with psi(x0)=0 boundary conditions. Result is basis-invariant.
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M),
	psi_IH=EVXData.getPsiFromBasis(iH,basis,n),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
    final double p_IH=TMatr.getAMatrA(psi_IH,QQpi,n)/basis[iH],
	var_pI_IH=2*(TMatr.getAMatrB(psi_IH,Dpsi_IH,QQpi,n)-p_IH*TMatr.getAMatrB(psi_IH,Dpsi_IH,QQi,n));

    // Assume I^{*}=0
    final double [] bQQDi=EVXData.convertMatrFromQtoB_Basis(n,TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDiffMatrix(n,QQi),-1),basis,n,n+0),
	bQQi=EVXData.convertMatrFromQtoB_Basis(n,QQi,basis,n,n+0),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,basis,n,n+0);
    final double [] DIev=M.EV.getEV(n,bQQDi,true),
	dp=new double [n],i=new double [n],p=new double [n],psi2x0=new double [n];
    for(int k=0;k<n;k++)
    {
	final double [] bpsik=EVXData.getPsiFromBasis(k,DIev,n),
	    bDpsik=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,bpsik,basis)),basis,QQt);
	final double pIk=TMatr.getAMatrA(bpsik,bQQpi,n),Ik=TMatr.getAMatrA(bpsik,bQQi,n);
	i[k]=Ik;
	p[k]=pIk/Ik;
	dp[k]=(-2*TMatr.getAMatrB(bpsik,bDpsik,bQQpi,n)-p[k]*DIev[k])/i[k];
	psi2x0[k]=2*TMatr.mulVV(bpsik,bDpsik);
    }
    return new DynVar2DynImpactDI(basis[i0],basis[iL],basis[iH],p_IH,var_pI_IH,
				  java.util.Arrays.copyOf(DIev,n),dp,i,p,psi2x0);
}

/** Same thing, but calculate in original basis (used for unit test).
 */
static DynVar2DynImpactDI test_get_DynVar2DynImpactDI(final int n,
						 final double [] QQi,final double [] QQt,final double [] QQpi,
						 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi,
						 final double eps)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    // Assume I^{*}=0.
    final double [] QQDi=TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDiffMatrix(n,QQi),-1);
    final EVXData DI=EVXData.getEVXDataFromQQ(n,QQDi,QQt,M);
    final double [] dp=new double [n],i=new double [n],p=new double [n],psi2x0=new double [n];
    for(int k=0;k<n;k++)
    {
	final double [] psik=DI.getPsi(k),Dpsik=VOLATILITY_EDPsi.getDPsi(psik);
	final double pIk=TMatr.getAMatrA(psik,QQpi,n),Ik=TMatr.getAMatrA(psik,QQi,n);
	i[k]=Ik;
	p[k]=pIk/Ik;
	dp[k]=(-2*TMatr.getAMatrB(psik,Dpsik,QQpi,n)-p[k]*DI.ev[k])/i[k];
	psi2x0[k]=2*TMatr.getAMatrB(psik,Dpsik,QQt,n);
	if(eps>0)
	{
	    final double [] rq=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQpi,QQi,n);
	    UnitTests.checkDoubleRel(p[k],"p[k]",rq[0],"rq[0]",eps);
	    UnitTests.checkDoubleRel(dp[k],"dp[k]",-rq[1],"-rq[1]",eps);
	    UnitTests.checkDoubleRel(psi2x0[k],"psi2x0[k]",Math.pow(M.B.getD0(psik,DI.x0),2),"psik(x0)^2",eps);
	    UnitTests.checkDoubleRel(DI.ev[k],"DI.ev[k]",-2*TMatr.getAMatrB(psik,Dpsik,QQi,n),"-2<psik|I|Dpsik>",eps);
	    System.err.println("DynVar2DynImpactDI: test OK n="+n+" k="+k+" eps="+eps);
	}
    }

    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M),
	psi_IH=EVXData.getPsiFromBasis(n-1,basis,n),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
    final double I_IH=basis[n-1],p_IH=TMatr.getAMatrA(psi_IH,QQpi,n)/I_IH,
	var_pI_IH=2*(TMatr.getAMatrB(psi_IH,Dpsi_IH,QQpi,n)-p_IH*TMatr.getAMatrB(psi_IH,Dpsi_IH,QQi,n));
    if(eps>0)
    {
	final double [] rq=RayleighQuotient.getRQVariationValues(psi_IH,Dpsi_IH,null,QQpi,QQi,n);
	UnitTests.checkDoubleRel(p_IH,"p_IH",rq[0],"rq[0]",eps);
	UnitTests.checkDoubleRel(var_pI_IH,"var_pI_IH",I_IH*rq[1],"I_IH*rq[1]",eps);
	final double [] ee=DI.getEValues();
	System.err.println("DynVar2DynImpactDI: SUM(i)="+TMatr.sumElements(i)+" i="+java.util.Arrays.toString(i));
	System.err.println("DynVar2DynImpactDI: SUM(dp)="+TMatr.sumElements(dp)+" dp="+java.util.Arrays.toString(dp));
	System.err.println("DynVar2DynImpactDI: SUM(DI)="+TMatr.sumElements(ee)+" DI="+java.util.Arrays.toString(ee));
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	for(int k=0;k<n;k++)
	{
	    final double [] psik=I.getPsi(k);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psik,QQDi,n),"<psik|DI|psik>",-I.ev[k]*Math.pow(M.B.getD0(psik,I.x0),2),"-I.ev[k]*psik(x0)^2",eps);
	}
	System.err.println("DynVar2DynImpactDI: all test OK n="+n+" eps="+eps);
    }

    return new DynVar2DynImpactDI(basis[0],basis[1],I_IH,p_IH,var_pI_IH,
				  DI.getEValues(),dp,i,p,psi2x0);
}

static void match(final DynVar2DynImpactDI a,final DynVar2DynImpactDI b, final double eps)
{
    final int n=a.i.length;
    UnitTests.checkDoubleRel(a.I_0,"a.I_0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.I_IL,"a.I_IL",b.I_IL,"b.I_IL",eps);
    UnitTests.checkDoubleRel(a.I_IH,"a.I_IH",b.I_IH,"b.I_IH",eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);

    UnitTests.checkDoubleArrayRel(a.dI,"a.dI",b.dI,"b.dI",eps);
    UnitTests.checkDoubleArrayRel(a.dp,"a.dp",b.dp,"b.dp",eps);
    UnitTests.checkDoubleArrayRel(a.i,"a.i",b.i,"b.i",eps);
    UnitTests.checkDoubleArrayRel(a.p,"a.p",b.p,"b.p",eps);
    UnitTests.checkDoubleArrayRel(a.psi2x0,"a.psi2x0",b.psi2x0,"b.psi2x0",eps);

    ACorr.matchACorr(a.a_dpdI,b.a_dpdI,eps);
    ACorr.matchACorr(a.a_dpdII,b.a_dpdII,eps);

    System.err.println("DynVar2DynImpactDI.match: test OK n="+n+" eps="+eps);
}

}

