package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: DynVar2Proj.java,v 1.22 2022/07/28 12:27:16 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVar2Proj {
    final EVXData I;
    final double p_IL,p_IH,p_EQ;
    DynVar2Proj(final EVXData I,final double p_IL,final double p_IH,final double p_EQ)
{
    this.I=I;
    this.p_IL=p_IL;
    this.p_IH=p_IH;
    this.p_EQ=p_EQ;
}

    static boolean PRINT_DIAG=false;

/**
 * For psi, I eigenvector: I|psi>=lambda |psi>
 * variate <psi|I|psi>/<psi|psi>
 * then condider infinitesimal psi variation |Dpsi>=|D|psi>,
 * projected to orthogonal D1 direction: |qpsi>=|pI|psi> - |psi> <psi|pI|psi> ,
 * |dpsi>=|Dpsi> - |qpsi><Dpsi|qpsi>/<qpsi|qpsi>.
 * For such |dpsi> variation we have
 * (according to my assumption in "Market Dynamics. On Supply and Demand Concepts" paper,
 * that |psi> states of I extremums should also be the states of price extremum)
 * 0=D1=2(<psi|pI|dpsi> - <psi|pI|psi><psi|dpsi>)
 * and D2 variation can be considered
 * D2=2(<dpsi|pI|dpsi>-<psi|pI|psi><dpsi|dpsi>),
 * that estimate the equilibrium price:
 * P_eq=(<dpsi|pI|dpsi>-<psi|pI|psi><dpsi|dpsi>)/(<dpsi|I|dpsi>-<psi|I|psi><dpsi|dpsi>).
 */
static DynVar2Proj get_DynVar2Proj(final int n,
				   final double [] QQi,final double [] QQt,final double [] QQpi,
				   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi,
				   final double eps_selftest)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int iH=n-1;
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double p_IL=TMatr.getAMatrA(I.psiL,QQpi,n)/I.sL,
	p_IH=TMatr.getAMatrA(I.psiH,QQpi,n)/I.sH;
    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),bqpsik=new double [n];
    double si=0,spi=0;
    for(int k=0;k<I.neq;k++)
    {
	final double [] bDpsik=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(I.getPsi(k)),I.ev,QQt);
	final double pI=bQQpi[k+n*k];
	System.arraycopy(bQQpi,k*n,bqpsik,0,n);
	bqpsik[k]-=pI;
	final double [] bqDpsik=TMatr.sumAcB(bDpsik,-TMatr.mulVV(bDpsik,bqpsik)/TMatr.mulVV(bqpsik,bqpsik),bqpsik);
	final double bdd=TMatr.mulVV(bqDpsik,bqDpsik);
	final double v2i=2*(TMatr.mulV2V(bqDpsik,I.getEValues())-I.ev[k]*bdd),
	    v2pi=2*(TMatr.getAMatrA(bqDpsik,bQQpi,n)-pI*bdd),
	    var_pI=2*TMatr.mulVV(bqpsik,bDpsik);
	if(v2i<0)
	{
	    si+=v2i;
	    spi+=v2pi;
	}
	else
	{
	    si-=v2i;
	    spi-=v2pi;
	}

	if(PRINT_DIAG||eps_selftest>0)
	{
	    final double eps=eps_selftest>0?eps_selftest:1e-5;
	    UnitTests.checkDoubleRel(TMatr.mulVV(bqpsik,bqDpsik),"<bqpsik|bqDpsik>",0,"0",eps);

	    final double [] psik=I.getPsi(k),Dpsik=VOLATILITY_EDPsi.getDPsi(psik),
		IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
		qpsik=TMatr.getMatr_x_A(n,IQQt,
					TMatr.sumAcB(TMatr.getMatr_x_A(n,QQpi,psik),
						     -I.getFPsi(k,QQpi),
						     TMatr.getMatr_x_A(n,QQt,psik))),
		qdpsik=TMatr.sumAcB(Dpsik,-TMatr.getAMatrB(Dpsik,qpsik,QQt,n)/TMatr.getAMatrB(qpsik,qpsik,QQt,n),qpsik);

	    UnitTests.checkDoubleRel(TMatr.getAMatrB(qpsik,qdpsik,QQt,n),"<qpsik|qdpsik>",0,"0",eps);
	    UnitTests.checkDoubleRel(2*TMatr.getAMatrB(qpsik,Dpsik,QQt,n),"2<qpsik|Dpsik>",var_pI,"var_pI",eps);

	    final double [] rqI=RayleighQuotient.getRQVariationValues(psik,qdpsik,null,QQi,QQt,n),
		rqpI=RayleighQuotient.getRQVariationValues(psik,qdpsik,null,QQpi,QQt,n);

	    //System.err.println("DynVar2Proj: rqI[1]="+rqI[1]+" rqI[2]="+rqI[2]+" v2i="+v2i);
	    //System.err.println("DynVar2Proj: rqpI[1]="+rqpI[1]+" rqpI[2]="+rqpI[2]+" v2pi="+v2pi);

	    UnitTests.checkDoubleRel(rqI[0],"rqI[0]",I.ev[k],"I.ev[k]",eps);
	    UnitTests.checkDoubleRel(rqI[1],"rqI[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(rqI[2],"rqI[2]",v2i,"v2i",eps);
	    UnitTests.checkDoubleRel(rqpI[0],"rqpI[0]",pI,"pI",eps);
	    UnitTests.checkDoubleRel(rqpI[1],"rqpI[1]",0,"0",eps);
	    if(n>2) UnitTests.checkDoubleRel(rqpI[2],"rqpI[2]",v2pi,"v2pi",eps);

	    final double [] rI=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQi,QQt,n),
		rpI=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQpi,QQt,n);
	    UnitTests.checkDoubleRel(rI[0],"rI[0]",I.ev[k],"I.ev[k]",eps);
	    UnitTests.checkDoubleRel(rI[1],"rI[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(rpI[0],"rpI[0]",pI,"pI",eps);
	    UnitTests.checkDoubleRel(rpI[1],"rpI[1]",var_pI,"var_pI",eps);

	    System.err.println("DynVar2Proj: test OK k="+k+" n="+n+" eps="+eps);
	}
    }
    final double p_EQ=spi/si;

    return new DynVar2Proj(I,p_IL,p_IH,p_EQ);
}

/*
static void matchDynVar2C(final DynVar2Proj a,final DynVar2C b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IX,"b.p_IX",eps);
    UnitTests.checkDoubleRel(a.varPi,"a.varPi",b.var_pI_sum,"b.var_pI_sum",eps);
    UnitTests.checkDoubleRel(a.varAPi,"a.varAPi",b.var_pI_sumA,"b.var_pI_sumA",eps);
    UnitTests.checkDoubleRel(a.varHPi,"a.varHPi",b.var_pI_IX,"b.var_pI_IX",eps);
    System.err.println("DynVarPiPt:matchDynVar2C test OK n="+a.I.neq+" eps="+eps);
}
*/


}

