package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.VolMatrix;

/**
 * Radau measure (x0-x)dmu is back.
 * Similar to DynVar2DISkewness.java, but uses Radau measure
 * in a style of DynRPnL.java from May-Aug 2015.
 *  @version $Id: DynVar2RDISkewness.java,v 1.52 2022/07/28 22:30:49 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVar2RDISkewness {
    final EVXData I;
    final double p_offset,p_last__with_p_offset_included,p_IH,I_0_feq,sD;
    final Skewness RDIHf_S,RDILf_S,RDIf_S[];

    DynVar2RDISkewness(final EVXData I,final double p_offset,final double p_last__with_p_offset_included,
		       final double p_IH,final double I_0_feq,
		       final Skewness [] RDIf_S,final double [] rates,
		       final double eps)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_last__with_p_offset_included=p_last__with_p_offset_included;
	this.p_IH=p_IH;
	this.I_0_feq=I_0_feq;
	this.RDIf_S=RDIf_S;
	this.RDIHf_S=RDIf_S[RDIf_S.length-1];
	this.RDILf_S=RDIf_S[0];
	double sr=0,sskr=0;
	for(int k=0;k<RDIf_S.length;k++)
	{
	    sr+=rates[k];
	    sskr+=rates[k]*RDIf_S[k].sk;
	}
	this.sD=sskr/sr;
    }

static DynVar2RDISkewness get_DynVar2RDISkewness(final MomentsData mm,
						 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
						 final VolMatrix VOLATILITY_EDPsi,
						 final boolean flag_dII,
						 final double eps)
{
    return get_DynVar2RDISkewness(
	mm.n,mm.p_offset,mm.p_last,
	mm.QQt,mm.QQi,mm.QQi,mm.QQpi,mm.QQp2i,mm.QQp3i,M,VOLATILITY_EDPsi,flag_dII,eps);
}

/**
 * Find |psi> state, dI/dt eigenvector: dI^{f}/dt|psi>=lambda |psi>,
 * Calculation of <Q_i|dI^{f}/dt|Q_j> matrix can be done by parts:
 * <Q_i|dI^{f}/dt|Q_j> = I^{*} Q_i(x0) Q_j(x0) - (<Q_i|I^{f}|D(Q_j)> + <D(Q_i)|I^{f}|Q_j>).
 * We take psi^2(x)=(x0-x)(\sum a_k Q_k(x))^2. This give the condition psi^2(x0)=0,
 * what does not require I^{*} knowledge, but dI enters the matrix.
 * To calculate price moments pi_m=<p^m I^{f} psi^2(x)> the knowledge of P^{f}
 * is not required, because of psi^2(x0)=0.
 * Then we calculate the skewness of p.
 * Compare this with DynVar2DynImpactDI.java where dp/dt is calculated,
 * what is incorrect. The correct approach is to look
 * at distribution skewness, not at dp/dt as price does not determine market dynamics.
 *
 * No optimization in this version. Just a try.
 * The code is similar to DynVar2DISkewness.java, but now with Radau measure.
 */
static DynVar2RDISkewness get_DynVar2RDISkewness(
    final int n,
    final double p_offset,final double p_last,
    final double [] QQt,final double [] QQi,
    final double [] QQp0i,final double [] QQp1i,final double [] QQp2i,final double [] QQp3i,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
    final VolMatrix VOLATILITY_EDPsi,
    final boolean flag_dII,
    final double eps)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0,dV=dI/I.o_o,
	p_last__with_p_offset_included=p_last;
    final int nd=n-1;
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),QQ0nd=DeprecatedHelperEVXData.getOneVectorRemovedFromMatrix(n,QQ0,n-1),
	QQi_feq=TMatr.sumAcB(QQi,dV,QQ0),
	rQQi=VolMatrix.getRadauMatrix(n,QQi,M),
	rQQp0i=VolMatrix.getRadauMatrix(n,QQp0i,M),
	rQQp1i=VolMatrix.getRadauMatrix(n,QQp1i,M),
	rQQp2i=VolMatrix.getRadauMatrix(n,QQp2i,M),
	rQQp3i=VolMatrix.getRadauMatrix(n,QQp3i,M),
	rQQDi_=TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDRadauMatrix(n,QQi,M),-1),
	rQQDif=TMatr.sumAcB(rQQDi_,dV,QQ0nd),
	rQQt=VolMatrix.getRadauMatrix(n,QQt,M),
	rG=flag_dII?rQQi:rQQt,
	rDbasis=M.EV.getGEV_Generalized(nd,rQQDif,rG,true);

    final Skewness RDIf_S[]=new Skewness [nd];
    for(int k=0;k<nd;k++)
    {
	RDIf_S[k]=Skewness.getSkewnessFromMonomialMoments(
	    TMatr.getAMatrA(nd,rDbasis,nd+k*nd,rQQp0i,nd),
	    TMatr.getAMatrA(nd,rDbasis,nd+k*nd,rQQp1i,nd),
	    TMatr.getAMatrA(nd,rDbasis,nd+k*nd,rQQp2i,nd),
	    TMatr.getAMatrA(nd,rDbasis,nd+k*nd,rQQp3i,nd));
    }
    final double [] rates=java.util.Arrays.copyOf(rDbasis,nd);
    if(!flag_dII) for(int k=0;k<nd;k++) rates[k]/=(RDIf_S[k].w1+RDIf_S[k].w2);

    if(eps>0)
    {
	// Check orthogonality.
	EVXData.test_orthogonality("DynVar2RDISkewness.get_DynVar2RDISkewness:rDbasis",nd,rDbasis,rQQDif,rG,eps);
	// Check values on |psi0>
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi_feq,n),"<pis0|QQI_feq|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi,n),"<pis0|QQI|psi0>",I.s0,"I.s0",eps);
	// Check eigenvalues.
	for(int k=0;k<nd;k++)
	{
	    final double [] psi=EVXData.getPsiFromBasis(k,rDbasis,nd),Rpsi=M.getRadauMultiplied(psi);
	    UnitTests.checkDoubleRel(-(TMatr.getAMatrB(psi,VOLATILITY_EDPsi.getDPsi(Rpsi),QQi_feq,n)+TMatr.getAMatrB(Rpsi,VOLATILITY_EDPsi.getDPsi(psi),QQi_feq,n)),"-<I D(rpsi^2(x))>",rDbasis[k],"rDbasis[k]",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(psi,VOLATILITY_EDPsi.getDPsi(Rpsi),QQt,n)+TMatr.getAMatrB(Rpsi,VOLATILITY_EDPsi.getDPsi(psi),QQt,n),"<D(rpsi^2(x)>",0,"0",eps);
	    UnitTests.checkDoubleRel(2*TMatr.getAMatrB(psi,VOLATILITY_EDPsi.getDPsi(psi),QQt,n),"<D(psi^2(x)>",Math.pow(M.B.getD0(psi,I.x0),2),"psi^2(x0)",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,QQ0,n),"psi^2(x0)",Math.pow(M.B.getD0(psi,I.x0),2),"psi^2(x0)",eps);
	    // Quadrature weights sum have to match.
	    UnitTests.checkDoubleRel(RDIf_S[k].w1+RDIf_S[k].w2,"RDIf_S[k].w1+RDIf_S[k].w2",TMatr.getAMatrB(psi,Rpsi,QQp0i,n),"<psi|QQp0i|psi>",eps);
	    UnitTests.checkDoubleRel(RDIf_S[k].xa*(RDIf_S[k].w1+RDIf_S[k].w2),"RDIf_S[k].xa*(RDIf_S[k].w1+RDIf_S[k].w2)",TMatr.getAMatrB(psi,Rpsi,QQp1i,n),"<psi|QQp1i|psi>",eps);
	}

	// dI does not change the moments calculated directly from sample.
	UnitTests.checkDoubleArrayRel(VolMatrix.getRadauMatrix(n,QQi_feq,M),"rQQi",rQQi,"rQQi",eps);
	// Two ways to calculate rQQDif matrix.
	UnitTests.checkDoubleArrayRel(TMatr.scaleVectorInPlace(VolMatrix.getRadauMatrix(n,VOLATILITY_EDPsi.getDiffMatrix(n,QQi_feq),M),-1),"rQQDif",rQQDif,"rQQDif",eps);
	// Check rQQDif=dV*QQ0+rQQDi to match exactly the matrix calculated from QQi_feq
	UnitTests.checkDoubleArrayRel(TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDRadauMatrix(n,QQi_feq,M),-1),"rQQDif",rQQDif,"rQQDif",eps);


	System.err.println("DynVar2RDISkewness: test OK n="+n+" dI="+dI+" rates="+java.util.Arrays.toString(rates)+" sum(rates)="+TMatr.sumElements(rates)+" flag_dII="+flag_dII+" eps="+eps);
    }
    return new DynVar2RDISkewness(
	I,p_offset,p_last__with_p_offset_included,
	TMatr.getAMatrA(I.psiH,QQp1i,n)/TMatr.getAMatrA(I.psiH,QQp0i,n),I_0_feq,
	RDIf_S,rates,eps);
}

}

