package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/**
 *  Try correlation at psiH.
 *  @version $Id: DynVarCD.java,v 1.33 2022/07/28 22:30:49 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVarCD {
    final EVXData I;
    final ACorr a;
    final double p_IX,X_IX,I_IX;
DynVarCD(final EVXData I,final ACorr a,final double p_IX,final double X_IX,final double I_IX)
{
    this.I=I;
    this.a=a;
    this.p_IX=p_IX;
    this.X_IX=X_IX;
    this.I_IX=I_IX;
}

private static final boolean FLAG_RUN_SELFTEST=false;

/**
 *
 */
static DynVarCD get_DynVarCD(final int n,
			     final double [] QQi,final double [] QQt,
			     final double [] QQpi,final double [] QQp2i,
			     final double [] QQXi,final double [] QQX2i,
			     final double [] QQpXi,
			     final double p_offset,
			     final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int iX=n-1;
    final double I_IX=I.ev[iX],pI_IX=I.getFPsi(iX,QQpi),p_IX=pI_IX/I_IX,X_IX=I.getFPsi(iX,QQXi)/I_IX;
    final double pscale=p_IX+p_offset,
	x_a_b=I.getFPsi(iX,QQpXi)-pI_IX*X_IX,
	x_a_a=(I.getFPsi(iX,QQp2i)-pI_IX*p_IX)/pscale,
	x_b_b=(I.getFPsi(iX,QQX2i)-I_IX*X_IX*X_IX)*pscale;

    return new DynVarCD(I,new ACorr(x_a_b,x_a_a,x_b_b),p_IX,X_IX,I_IX);
}


/** Calculate (<psiH|QQpi|psik>,<psik|QQXi|psiH>) pseudocorrelation,
 * Here a.osp is the same as DynVHPnL.getme_DynVHPnL.sw_p_IH.
 * My theory tells that if <psik|QQV|psiH> is used instead of <psik|QQXi|psiH>
 * the result should be pretty much the same (need to test).
 */
static DynVarCD getXp_DynVarCD(final int n,
			       final double [] QQi,final double [] QQt,
			       final double [] QQpi,final double [] QQXi,
			       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");

    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	bQQXi=EVXData.convertMatrFromQtoB_Basis(n,QQXi,I.ev);

    final int ind_toremove=n-1;
    final double [] bPpsi=new double [n],bXpsi=new double [n];
    for(int k=0;k<n;k++) {
	bPpsi[k]=bQQpi[n*ind_toremove+k];
	bXpsi[k]=bQQXi[n*k+ind_toremove];
    }
    final double I_IX=I.ev[ind_toremove],p_IX=bPpsi[ind_toremove]/I_IX,X_IX=bXpsi[ind_toremove]/I_IX;
    bXpsi[ind_toremove]=bPpsi[ind_toremove]=0;
    final double x_a_b=TMatr.mulVV(bXpsi,bPpsi),x_a_a=TMatr.mulVV(bXpsi,bXpsi),x_b_b=TMatr.mulVV(bPpsi,bPpsi);
    if(FLAG_RUN_SELFTEST)
    {
	final double eps=1e-5;
	final double [] eeP=M.EV.getEV(n,TMatr.getQQaa(bPpsi),true);
	UnitTests.checkDoubleRel(eeP[n-1],"eeP[n-1]",x_b_b,"x_b_b",eps);
	final double [] eeX=M.EV.getEV(n,TMatr.getQQaa(bXpsi),true);
	UnitTests.checkDoubleRel(eeX[n-1],"eeX[n-1]",x_a_a,"x_a_a",eps);
	System.err.println("eeP & eeX match OK");
	final double [] eeXP=M.EV.getEV(n,TMatr.getQQab(n,bPpsi,bXpsi),true);
	System.err.println("x_a_b="+x_a_b+" ; eeXP="+java.util.Arrays.toString(java.util.Arrays.copyOf(eeXP,n)));
    }
    return new DynVarCD(I,new ACorr(x_a_b,x_a_a,x_b_b),p_IX,X_IX,I_IX);
}

/** Calculate (<psiH|QQpi|psik>,<psik|QQi|D(psiH)>) pseudocorrelation.
 * Note: old version (<psiH|QQpi|psik>,<psik|QQt|D(psiH)>)
 * corresponds exactly to DynVHPnL.var_p_IH and DynVar2C.var_pI_IX.
 */
static DynVarCD getDp_DynVarCD(final int n,
			       final double [] QQi,final double [] QQt,
			       final double [] QQpi,
			       final boolean flag_ID,
			       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");
    final int ind_toremove=n-1;
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] bPpsi=new double [n],
	bDpsi=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(I.getPsi(ind_toremove)),I.ev,flag_ID?QQi:QQt),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev);
    for(int k=0,nr=n*ind_toremove;k<n;k++) {
	bPpsi[k]=bQQpi[nr+k];
    }
    final double I_IX=I.ev[ind_toremove],p_IX=bPpsi[ind_toremove]/I_IX,X_IX=bDpsi[ind_toremove]/(flag_ID?I_IX:1);
    if(FLAG_RUN_SELFTEST)
    {
	final double eps=1e-5;
	UnitTests.checkDoubleRel(X_IX,"X_IX",Math.pow(M.B.getD0(I.psiH,M.getRadauRoot()),2)/2,"psiH(x0)^2/2",eps);
	final double [] Dpsi=VOLATILITY_EDPsi.getDPsi(I.getPsi(ind_toremove)),
	    btestD=new double [n],btest2D=new double [n];
	for(int k=0;k<n;k++)
	{
	    btestD[k]=I.getFPsiPsi(k,Dpsi,flag_ID?QQi:QQt);
	    btest2D[k]=(flag_ID?I.ev[k]:1)*I.getFPsiPsi(k,Dpsi,QQt);
	}
	UnitTests.checkDoubleArrayRel(bDpsi,"bDpsi",btestD,"btestD",eps);
	UnitTests.checkDoubleArrayRel(bDpsi,"bDpsi",btest2D,"btest2D",eps);
	System.err.println("getDp_DynVarCD: match OK");
    }
    bDpsi[ind_toremove]=bPpsi[ind_toremove]=0;
    final double x_a_b=TMatr.mulVV(bDpsi,bPpsi),x_a_a=TMatr.mulVV(bDpsi,bDpsi),x_b_b=TMatr.mulVV(bPpsi,bPpsi);
    return new DynVarCD(I,new ACorr(x_a_b,x_a_a,x_b_b),p_IX,X_IX,I_IX);
}

/** Try (<psiH|QQC|psik>,<psik|QQV|psiH>) pseudocorrelation,
 *  here QQC=QQpV-p_IX*QQV.
 */
static DynVarCD getCV_DynVarCD(final int n,
			       final double [] QQi,final double [] QQt,final double [] QQpi,
			       final double [] QQV,final double [] QQpV,
			       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<2) throw new IllegalArgumentException("n="+n+"<2");

    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int ind_toremove=n-1;
    final double I_IX=I.ev[ind_toremove],p_IX=I.getFPsi(ind_toremove,QQpi)/I_IX;
    final double [] bQQc=EVXData.convertMatrFromQtoB_Basis(n,TMatr.sumAcB(QQpV,-p_IX,QQV),I.ev),
	bQQV=EVXData.convertMatrFromQtoB_Basis(n,QQV,I.ev);

    final double [] bCpsi=new double [n],bVpsi=new double [n];
    for(int k=0;k<n;k++) {
	bCpsi[k]=bQQc[n*ind_toremove+k];
	bVpsi[k]=bQQV[n*k+ind_toremove];
    }
    final double X_IX=bVpsi[ind_toremove];
    if(FLAG_RUN_SELFTEST)
    {
	System.err.println("bVpsi="+java.util.Arrays.toString(bVpsi));
	System.err.println("bCpsi="+java.util.Arrays.toString(bCpsi));
	final double eps=1e-5;
	final double [] btestC=new double [n],btestV=new double [n];
	for(int k=0;k<n;k++)
	{
	    btestV[k]=I.getFPsiPsi(k,ind_toremove,QQV);
	    btestC[k]=I.getFPsiPsi(k,ind_toremove,QQpV)-p_IX*btestV[k];
	}
	UnitTests.checkDoubleArrayRel(bVpsi,"bVpsi",btestV,"btestV",eps);
	UnitTests.checkDoubleArrayRel(bCpsi,"bCpsi",btestC,"btestV",eps);
	System.err.println("getCV_DynVarCD: match OK");
    }
    bVpsi[ind_toremove]=bCpsi[ind_toremove]=0;
    final double x_a_b=TMatr.mulVV(bVpsi,bCpsi),x_a_a=TMatr.mulVV(bVpsi,bVpsi),x_b_b=TMatr.mulVV(bCpsi,bCpsi);

    return new DynVarCD(I,new ACorr(x_a_b,x_a_a,x_b_b),p_IX,X_IX,I_IX);
}



public static void matchDynVHPnL(final DynVarCD a,final DynVHPnL b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,false,eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.a.x_a_b,"a.a.x_a_b",b.var_p_IH,"b.var_p_IH",eps);
    UnitTests.checkDoubleRel(a.a.osp,"a.a.osp",b.sw_p_IH,"b.sw_p_IH",eps);

    System.err.println("DynVarCD:matchDynVHPnL test OK neq="+a.I.neq+" eps="+eps);
}

}

