package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;

/**
 *  @version $Id: DynVarFiniteC.java,v 1.34 2022/07/28 12:27:16 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVarFiniteC {
    final double I_0,I_IL,I_IH,p_IX,I_IX,var_pn_IX,var2_pn_IX,pd_cL,Id_cL,pd_cH,Id_cH,distance2_cL_cH,distance2_IH_cH,distance2_IH_cL;
    DynVarFiniteC(final double I_0,final double I_IL,final double I_IH,final double p_IX,final double I_IX,final double var_pn_IX,final double var2_pn_IX,
		  final double pd_cL,final double Id_cL,final double pd_cH,final double Id_cH,final double distance2_cL_cH,final double distance2_IH_cH,final double distance2_IH_cL)
{
    this.I_0=I_0;
    this.I_IL=I_IL;
    this.I_IH=I_IH;
    this.p_IX=p_IX;
    this.I_IX=I_IX;
    this.var_pn_IX=var_pn_IX;
    this.var2_pn_IX=var2_pn_IX;
    this.pd_cL=pd_cL;
    this.Id_cL=Id_cL;
    this.pd_cH=pd_cH;
    this.Id_cH=Id_cH;
    this.distance2_cL_cH=distance2_cL_cH;
    this.distance2_IH_cH=distance2_IH_cH;
    this.distance2_IH_cL=distance2_IH_cL;
}


/**
 * For psiH: I|psiH>=lambda_max |psiH>, I_IH=lambda_max,
 * then calculate p_IH=<psiH|pI|psiH>/I_IH.
 * Use P&L=<psi|(p-p_IH)V|psi>/<psi|psi> what lead to psi_cL and psi_cH states.
 * (Note the infinitesimal variation of psi and V=I give
 * well known answer, see YkQVM & DynVHPnL.var_p_IH
 * with the answer: <pisH|pI|DpsiH> - <pisH|DpsiH><pisH|pI|psiH> (here DpsiH=d/dt psiH).)
 * But now we take finite variation, instead of infinitesimal one.
 * Also note that DynVarC results Id_{cL,cH}=I_IH-<psi_{cL,cH}|I|psi{cL,cH}>.
 *
 */
static DynVarFiniteC get_PnLX_DynVarFiniteC(final int n,
					    final double [] QQi,final double [] QQt,final double [] QQpi,
					    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
					    final VolMatrix VOLATILITY_EDPsi,
					    final K2Distance distance2)

{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int iX=n-1;
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH),D2psi_IH=VOLATILITY_EDPsi.getDPsi(Dpsi_IH);

    final double pI_IX=I.getFPsi(iX,QQpi),I_IX=I.ev[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpi,-p_IX,QQi),
	rq=RayleighQuotient.getRQVariationValues(psi_IH,Dpsi_IH,D2psi_IH,QQpnl,QQt,n),
	bQQpnl=EVXData.convertMatrFromQtoB_Basis(n,QQpnl,I.ev,n,n+0);

    // P&L min/max
    final double [] pbasis=M.EV.getEV(n,bQQpnl,true);
    final double pd_cL=pbasis[0],Id_cL=I_IX-TMatr.mulV2V(EVXData.getPsiFromBasis(0,pbasis,n),I.getEValues()),
	pd_cH=pbasis[n-1],Id_cH=I_IX-TMatr.mulV2V(EVXData.getPsiFromBasis(n-1,pbasis,n),I.getEValues());
    final double [] psi_cL=EVXData.convertPsiFromBtoQ_Basis(n,EVXData.getPsiFromBasis(0,pbasis,n),I.ev),
	psi_cH=EVXData.convertPsiFromBtoQ_Basis(n,EVXData.getPsiFromBasis(n-1,pbasis,n),I.ev);

    return new DynVarFiniteC(I.s0,I.sL,I.sH,p_IX,I_IX,rq[1],rq[2],pd_cL,Id_cL,pd_cH,Id_cH,
			     distance2.getDistance2(psi_cL,psi_cH),
			     distance2.getDistance2(psi_IH,psi_cH),
			     distance2.getDistance2(psi_IH,psi_cL));
}

static DynVarFiniteC test_get_PnLX_DynVarFiniteC(final int n,
						 final double [] QQi,final double [] QQt,final double [] QQpi,
						 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
						 final VolMatrix VOLATILITY_EDPsi,
						 final K2Distance distance2)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int iX=n-1;
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH),D2psi_IH=VOLATILITY_EDPsi.getDPsi(Dpsi_IH);
    final double pI_IX=I.getFPsi(iX,QQpi),I_IX=I.ev[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpi,-p_IX,QQi),
	rq=RayleighQuotient.getRQVariationValues(psi_IH,Dpsi_IH,D2psi_IH,QQpnl,QQt,n);
    final EVXData ev=EVXData.getEVXDataFromQQ(n,QQpnl,QQt,M);
    final double [] psi_cL=ev.getPsi(0),psi_cH=ev.getPsi(n-1);
    return new DynVarFiniteC(I.s0,I.sL,I.sH,p_IX,I_IX,rq[1],rq[2],
			     ev.sL,I_IX-TMatr.getAMatrA(psi_cL,QQi,n),
			     ev.sH,I_IX-TMatr.getAMatrA(psi_cH,QQi,n),
			     distance2.getDistance2(psi_cL,psi_cH),
			     distance2.getDistance2(psi_IH,psi_cH),
			     distance2.getDistance2(psi_IH,psi_cL));

}

static DynVarFiniteC getR_PnLX_DynVarFiniteC(final int n,
					     final double [] QQi,final double [] QQt,final double [] QQpi,
					     final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
					     final VolMatrix VOLATILITY_EDPsi,
					     final K2Distance distance2)

{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int nd=n-1,iX=nd-1;
    final double [] basis=M.EV.getGEV_Generalized(nd,VolMatrix.getRadauMatrix(n,QQi,M),VolMatrix.getRadauMatrix(n,QQt,M),true),
	psi=EVXData.getPsiFromBasis(iX,basis,nd),
	Dpsi=VOLATILITY_EDPsi.getDPsi(psi),
	D2psi=VOLATILITY_EDPsi.getDPsi(Dpsi),
	Rpsi=M.getRadauMultiplied(psi),
	DRpsi=VOLATILITY_EDPsi.getDPsi(Rpsi),
	D2Rpsi=VOLATILITY_EDPsi.getDPsi(DRpsi);
    final double pI_IX=TMatr.getAMatrB(psi,Rpsi,QQpi,n),I_IX=basis[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpi,-p_IX,QQi),
	RQQpnl=VolMatrix.getRadauMatrix(n,QQpnl,M),
	rq=RayleighQuotient.getRQVariationValues(psi,Dpsi,D2psi,Rpsi,DRpsi,D2Rpsi,QQpnl,QQt,n),
	bRQQpnl=EVXData.convertMatrFromQtoB_Basis(nd,RQQpnl,basis,nd,nd+0);

    // P&L min/max
    final double [] pbasis=M.EV.getEV(nd,bRQQpnl,true),
	psi_cL=EVXData.convertPsiFromBtoQ_Basis(nd,EVXData.getPsiFromBasis(0,pbasis,nd),basis),
	Rpsi_cL=M.getRadauMultiplied(psi_cL),
	psi_cH=EVXData.convertPsiFromBtoQ_Basis(nd,EVXData.getPsiFromBasis(nd-1,pbasis,nd),basis),
	Rpsi_cH=M.getRadauMultiplied(psi_cH);
    final double pd_cL=pbasis[0],Id_cL=I_IX-TMatr.getAMatrB(psi_cL,Rpsi_cL,QQi,n),
	pd_cH=pbasis[nd-1],Id_cH=I_IX-TMatr.getAMatrB(psi_cH,Rpsi_cH,QQi,n);

    return new DynVarFiniteC(Double.NaN,basis[0],basis[nd-1],p_IX,I_IX,rq[1],rq[2],pd_cL,Id_cL,pd_cH,Id_cH,
			     distance2.getK2(psi_cL,Rpsi_cL,psi_cH,Rpsi_cH)-distance2.getK2(psi_cH,Rpsi_cH,psi_cL,Rpsi_cL),
			     distance2.getK2(psi,Rpsi,psi_cH,Rpsi_cH)-distance2.getK2(psi_cH,Rpsi_cH,psi,Rpsi),
			     distance2.getK2(psi,Rpsi,psi_cL,Rpsi_cL)-distance2.getK2(psi_cL,Rpsi_cL,psi,Rpsi));
}
static DynVarFiniteC test_getR_PnLX_DynVarFiniteC(final int n,
					     final double [] QQi,final double [] QQt,final double [] QQpi,
					     final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
					     final VolMatrix VOLATILITY_EDPsi,
					     final K2Distance distance2)

{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int nd=n-1,iX=nd-1;
    final double [] RQQt=VolMatrix.getRadauMatrix(n,QQt,M),
	basis=M.EV.getGEV_Generalized(nd,VolMatrix.getRadauMatrix(n,QQi,M),RQQt,true),
	psi_IH=EVXData.getPsiFromBasis(iX,basis,nd),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH),D2psi_IH=VOLATILITY_EDPsi.getDPsi(Dpsi_IH),
	Rpsi_IH=M.getRadauMultiplied(psi_IH),DRpsi_IH=VOLATILITY_EDPsi.getDPsi(Rpsi_IH),D2Rpsi_IH=VOLATILITY_EDPsi.getDPsi(DRpsi_IH);

    final double pI_IX=TMatr.getAMatrB(psi_IH,Rpsi_IH,QQpi,n),I_IX=basis[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpi,-p_IX,QQi),
	rq=RayleighQuotient.getRQVariationValues(psi_IH,Dpsi_IH,D2psi_IH,Rpsi_IH,DRpsi_IH,D2Rpsi_IH,QQpnl,QQt,n),
	pbasis=M.EV.getGEV_Generalized(nd,VolMatrix.getRadauMatrix(n,QQpnl,M),RQQt,true),
	psi_cL=EVXData.getPsiFromBasis(0,pbasis,nd),
	Rpsi_cL=M.getRadauMultiplied(psi_cL),
	psi_cH=EVXData.getPsiFromBasis(nd-1,pbasis,nd),
	Rpsi_cH=M.getRadauMultiplied(psi_cH);
    return new DynVarFiniteC(Double.NaN,basis[0],basis[nd-1],p_IX,I_IX,rq[1],rq[2],
			     pbasis[0],I_IX-TMatr.getAMatrB(psi_cL,Rpsi_cL,QQi,n),
			     pbasis[nd-1],I_IX-TMatr.getAMatrB(psi_cH,Rpsi_cH,QQi,n),
			     distance2.getK2(psi_cL,Rpsi_cL,psi_cH,Rpsi_cH)-distance2.getK2(psi_cH,Rpsi_cH,psi_cL,Rpsi_cL),
			     distance2.getK2(psi_IH,Rpsi_IH,psi_cH,Rpsi_cH)-distance2.getK2(psi_cH,Rpsi_cH,psi_IH,Rpsi_IH),
			     distance2.getK2(psi_IH,Rpsi_IH,psi_cL,Rpsi_cL)-distance2.getK2(psi_cL,Rpsi_cL,psi_IH,Rpsi_IH));
}


/** Similar to get_PnLX_DynVarFiniteC, but uses psi_IH(x0)=0
 * boundary condition.
 * Note that in a very important case pV & V the P&L matrix first variation
 * is exactly =0 if psiH(x0)=0.
 * I | psiH>= lambda |psiH>, I_IH=lambda_max, p_IH=<psi_IH|pI|psi_IH>/I_IH.
 * then C=\int (p-p_IH)Idt, and
 * var <psi_IH|C|psi_IH>/<psi_IH|psi_IH>=2*<psi_IH|C|Dpsi_IH>-<psi_IH|C|psi_IH>*2*<psi_IH|Dpsi_IH>.
 * The term <psi_IH|Dpsi_IH>=0 if psi_IH(x0)=0 and Dpsi_IH=d/dt psi_IH.
 * and the term <psi_IH|C|Dpsi_IH>=0 by p_IH definition.
 *
 */
static DynVarFiniteC get_PnLCV_DynVarFiniteC(final int n,
					     final double [] QQi,final double [] QQt,final double [] QQpi,
					     final double [] QQV,final double [] QQpV,
					     final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
					     final VolMatrix VOLATILITY_EDPsi,
					     final K2Distance distance2)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int i0=0,iX=n-1;
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M),
	psi_IH=EVXData.getPsiFromBasis(iX,basis,n);
    final double pI_IX=TMatr.getAMatrA(psi_IH,QQpi,n),I_IX=basis[iX],p_IX=pI_IX/I_IX;
    final double [] QQpnl=TMatr.sumAcB(QQpV,-p_IX,QQV),
	Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH),
	rq=RayleighQuotient.getRQVariationValues(psi_IH,Dpsi_IH,VOLATILITY_EDPsi.getDPsi(Dpsi_IH),QQpnl,QQt,n);
    TMatr.sumP1(QQpnl,QQt,-TMatr.getAMatrA(psi_IH,QQpnl,n));
    final double [] bQQpnl=EVXData.convertMatrFromQtoB_Basis(n,QQpnl,basis),
	bQQV=EVXData.convertMatrFromQtoB_Basis(n,QQV,basis);

    final double [] pbasis=M.EV.getEV(n,bQQpnl,true);
    final double V_IX=bQQV[n*iX+iX],pd_cL=pbasis[0],Id_cL=V_IX-TMatr.getAMatrA(n,pbasis,n+0*n,bQQV,n),
	pd_cH=pbasis[n-1],Id_cH=V_IX-TMatr.getAMatrA(n,pbasis,n+(n-1)*n,bQQV,n);
    final double [] psi_cL=EVXData.convertPsiFromBtoQ_Basis(n,EVXData.getPsiFromBasis(0,pbasis,n),basis),
	psi_cH=EVXData.convertPsiFromBtoQ_Basis(n,EVXData.getPsiFromBasis(n-1,pbasis,n),basis);

     return new DynVarFiniteC(basis[0],basis[1],basis[n-1],p_IX,I_IX,rq[1],rq[2],
			      pd_cL,Id_cL,pd_cH,Id_cH,
			      distance2.getDistance2(psi_cL,psi_cH),
			      distance2.getDistance2(psi_IH,psi_cH),
			      distance2.getDistance2(psi_IH,psi_cL));
}

/** Unit test for get_PnLCV_DynVarFiniteC.
 */
static DynVarFiniteC test_get_PnLCV_DynVarFiniteC(final int n,
						  final double [] QQi,final double [] QQt,final double [] QQpi,
						  final double [] QQV,final double [] QQpV,
						  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
						  final VolMatrix VOLATILITY_EDPsi,
						  final K2Distance distance2)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final int i0=0,iX=n-1;
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M),
	psi_IH=EVXData.getPsiFromBasis(iX,basis,n),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH),
	D2psi_IH=VOLATILITY_EDPsi.getDPsi(Dpsi_IH);
    final double pI_IX=TMatr.getAMatrA(psi_IH,QQpi,n),I_IX=basis[iX],
	p_IX=pI_IX/I_IX,V_IX=TMatr.getAMatrA(psi_IH,QQV,n);
    final double [] QQpnl=TMatr.sumAcB(QQpV,-p_IX,QQV),
	rq=RayleighQuotient.getRQVariationValues(psi_IH,Dpsi_IH,D2psi_IH,QQpnl,QQt,n);
    TMatr.sumP1(QQpnl,QQt,-TMatr.getAMatrA(psi_IH,QQpnl,n));
    final EVXData ev=EVXData.getEVXDataFromQQ(n,QQpnl,QQt,M);
    final double [] psi_cL=ev.getPsi(0),psi_cH=ev.getPsi(n-1);
    final double pd_cL=ev.sL,Id_cL=V_IX-ev.getFPsi(0,QQV),
	pd_cH=ev.sH,Id_cH=V_IX-ev.getFPsi(n-1,QQV);
    if(false)
    {
	// is psi_IH P&L eigenfunction?
	System.err.println("QQpnl | psi_IH>="+java.util.Arrays.toString(TMatr.getMatr_x_A(n,QQpnl,psi_IH)));
	System.err.println("QQt | psi_IH>="+java.util.Arrays.toString(TMatr.getMatr_x_A(n,QQt,psi_IH)));
    }

    return new DynVarFiniteC(basis[0],basis[1],basis[n-1],p_IX,I_IX,rq[1],rq[2],pd_cL,Id_cL,pd_cH,Id_cH,
			     distance2.getDistance2(psi_cL,psi_cH),
			     distance2.getDistance2(psi_IH,psi_cH),
			     distance2.getDistance2(psi_IH,psi_cL));
}


static void matchDynVarFiniteC(final DynVarFiniteC a,final DynVarFiniteC b,final double eps)
{
    UnitTests.checkDoubleRel(a.I_0,"a.I_0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.I_IL,"a.I_IL",b.I_IL,"b.I_IL",eps);
    UnitTests.checkDoubleRel(a.I_IH,"a.I_IH",b.I_IH,"b.I_IH",eps);
    UnitTests.checkDoubleRel(a.I_IX,"a.I_IX",b.I_IX,"b.I_IX",eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IX,"b.p_IX",eps);
    UnitTests.checkDouble(a.var_pn_IX/a.I_IH,"a.var_pn_IX",b.var_pn_IX/b.I_IH,"b.var_pn_IX",eps);
    UnitTests.checkDoubleRel(a.var2_pn_IX,"a.var2_pn_IX",b.var2_pn_IX,"b.var2_pn_IX",eps);
    UnitTests.checkDoubleRel(a.Id_cL,"a.Id_cL",b.Id_cL,"b.Id_cL",eps);
    UnitTests.checkDoubleRel(a.pd_cL,"a.pd_cL",b.pd_cL,"b.pd_cL",eps);
    UnitTests.checkDoubleRel(a.Id_cH,"a.Id_cH",b.Id_cH,"b.Id_cH",eps);
    UnitTests.checkDoubleRel(a.pd_cH,"a.pd_cH",b.pd_cH,"b.pd_cH",eps);
    UnitTests.checkDouble(a.distance2_cL_cH,"a.distance2_cL_cH",b.distance2_cL_cH,"b.distance2_cL_cH",eps);
    UnitTests.checkDouble(a.distance2_IH_cH,"a.distance2_IH_cH",b.distance2_IH_cH,"b.distance2_IH_cH",eps);
    UnitTests.checkDouble(a.distance2_IH_cL,"a.distance2_IH_cL",b.distance2_IH_cL,"b.distance2_IH_cL",eps);
    System.err.println("DynVarFiniteC:matchDynVarFiniteC test OK eps="+eps);
}

static void matchDynVarC(final DynVarFiniteC a,final DynVarC b,final double eps)
{
    UnitTests.checkDoubleRel(a.I_0,"a.I_0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.I_IL,"a.I_IL",b.I_IL,"b.I_IL",eps);
    UnitTests.checkDoubleRel(a.I_IH,"a.I_IH",b.I_IH,"b.I_IH",eps);
    UnitTests.checkDoubleRel(a.I_IX,"a.I_IX",b.I_IX,"b.I_IX",eps);
    UnitTests.checkDoubleRel(a.p_IX,"a.p_IX",b.p_IX,"b.p_IX",eps);
    UnitTests.checkDoubleRel(a.Id_cL,"a.Id_cL",b.Id_cL,"b.Id_cL",eps);
    UnitTests.checkDoubleRel(a.pd_cL,"a.pd_cL",b.p_cL,"b.pd_cL",eps);
    UnitTests.checkDoubleRel(a.Id_cH,"a.Id_cH",b.Id_cH,"b.Id_cH",eps);
    UnitTests.checkDoubleRel(a.pd_cH,"a.pd_cH",b.p_cH,"b.pd_cH",eps);
    System.err.println("DynVarFiniteC:matchDynVarC test OK eps="+eps);
}

}

