package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.SortArrays;

/**
 * A combination of P&L trading strategy with pseudoaverages (see DynVH2PnL.java for min/max version).
 * @version $Id: DynVarPnLSpur.java,v 1.18 2022/07/28 12:27:16 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
class DynVarPnLSpur {
    final EVXData I,Ipnl;
    final double p_IH,var_p_IH,sumvar2PnL,sumPnL,sumv2PnLProjIH,sumPnLProjIH;
    final double [] var2PnL;
    DynVarPnLSpur(final EVXData I,final EVXData Ipnl,
		  final double p_IH,final double var_p_IH,
		  final double [] var2PnL,final double [] w_ih_pnl
		      )
    {
	this.I=I;
	this.Ipnl=Ipnl;
	this.var2PnL=var2PnL;
	this.p_IH=p_IH;
	this.var_p_IH=var_p_IH;
	this.sumvar2PnL=TMatr.sumElements(var2PnL);
	this.sumPnL=TMatr.sumElements(Ipnl.getEValues());
	this.sumv2PnLProjIH=TMatr.mulVV(w_ih_pnl,var2PnL);
	this.sumPnLProjIH=TMatr.mulVVn1(w_ih_pnl,Ipnl.ev);
    }

/** Consider |I|psi>=\lambda |psi> states,
 * find the one giving max \lambda and corresponding psiH.
 * Find p_{IH}=<psiH|pI|psiH>/<psiH|pI|psiH> and then
 * consider P&L matrix =|(p-p_{IH})I| , corresponding to "exit at psiH" trading strategy,
 * find its eigenstates (\lambda_{P&L}, psi_{P&L}).
 * Variate <psi_{P&L}|P&L|psi_{P&L}>/<psi_{P&L}|psi_{P&L}>,
 * first variation is zero (because it is an eigenstate),
 * second variation of P&L matrix is non-zero, and try to use
 * infinitesimal (by time) second variation as direction indicator.
 * See DynVH2PnL.java for an earlier version of this idea (there only min/max eigenstates are used).
 *
 */
static DynVarPnLSpur get_DynVarPnLSpur(final int n,
					 final double [] QQi,final double [] QQt,final double [] QQpi,
					 final VolMatrix VOLATILITY_EDPsi,
					 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
					 final double eps)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");
	final int i0=0,iL=1,iH=n-1;
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int iX=n-1;
	final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
	final double pI_IH=I.getFPsi(iX,QQpi),p_IH=pI_IH/I.ev[iX],
	    var_p_IH=2*(I.getFPsiPsi(iX,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iX,Dpsi_IH,QQt));
	final double [] QQpnl=TMatr.sumAcB(QQpi,-p_IH,QQi);
	final EVXData Ipnl=EVXData.getEVXDataFromQQ(n,QQpnl,QQt,M);

	final double [] var2PnL=new double [n],w_ih_pnl=new double [n],I_pnl=new double [n];
	for(int k=0;k<n;k++)
	{
	    final double [] psi=Ipnl.getPsi(k),Dpsi=VOLATILITY_EDPsi.getDPsi(psi);
	    final double [] rq=RayleighQuotient.getRQVariationValues(psi,Dpsi,null,QQpnl,QQt,n);

	    var2PnL[k]=rq[2];
	    w_ih_pnl[k]=Math.pow(TMatr.getAMatrB(psi_IH,psi,QQt,n),2);
	    I_pnl[k]=TMatr.getAMatrA(psi,QQi,n);
	    if(eps>0)
	    {
		UnitTests.checkDoubleRel(rq[0],"rq[0]",Ipnl.ev[k],"Ipnl.ev[k]",eps);
		UnitTests.checkDoubleRel(rq[1],"rq[1]",0,"0",eps);
		UnitTests.checkDoubleRel(rq[2],"rq[2]",2*(TMatr.getAMatrA(Dpsi,QQpnl,n)-Ipnl.ev[k]*TMatr.getAMatrA(Dpsi,QQt,n)),"2*(<Dpsi|P&L|Dpsi>-P&L*<Dpsi|Dpsi>",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,QQt,n),"<psi|psi>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_IH,QQpnl,n),"<psi_IH|P&L|psi_IH>",0,"0",eps);
	    }
	}
	if(eps>0)
	{
	    UnitTests.checkDoubleRel(TMatr.sumElements(w_ih_pnl),"sum <psi_IH|psi>^2",1,"1",eps);
	    System.err.println("DynVarPnLSpur:  test OK n="+n+" eps="+eps);
	}
	return new DynVarPnLSpur(I,Ipnl,p_IH,var_p_IH,var2PnL,w_ih_pnl);
    }
}
