package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: DynVarSP0.java,v 1.21 2020/09/27 20:33:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVarSP0 {
/** Set to a small number about 1e-5 to force constrains check with given precision.
 */
private static final double EPS_FLAG_CHECKCONSTRAINS=1e-5 * 0 ;

final EVXData I;
final double [] o,w,d,pI;
final double dirC,dirP,I0,dirPC,Pi0;
/** A class to store liquidity deficit data and capital flow direction dirC.
 * Low liqudity: I.wL^2>0.5,
 * High liquidity: I.wH^2>0.5
 */
DynVarSP0(final EVXData I,final double [] o,final double [] w,final double [] d,final double [] pI)
{
    this.I=I;
    this.o=o;
    this.w=w;
    this.d=d;
    this.pI=pI;
    final double o_o=TMatr.sumElements(o),w_o=TMatr.sumElements(w),d_o=TMatr.sumElements(d);
    // Volume flow.
    this.I0=w_o/o_o;
    // Change in capital to total volume flow.
    this.dirPC=d_o/w_o; // Worst !!!
    // Change in capital.
    this.dirC=d_o/o_o;
    double dirP=0,pe0=0;
    for(int k=0;k<I.neq;k++){
	dirP+=d[k]/I.ev[k];
	pe0+=pI[k]*o[k];
    }
    // Change in price (all d[k] are divided by I.ev[k]).
    this.dirP=dirP/o_o;
    this.Pi0=pe0/w_o;

    // Have to match, debugging & diag code.
    if(EPS_FLAG_CHECKCONSTRAINS>0)
    {
	final double eps=EPS_FLAG_CHECKCONSTRAINS;
	final double wL2=(o[0]/o_o),wH2=(o[o.length-1]/o_o);
	System.err.println("DynVarSP0: n="+I.neq+" I.s0="+I.s0+"=I0="+I0+" I.o_o="+I.o_o+"=o_o="+o_o+
			   " wL2="+wL2+"=I.wL^2="+Math.pow(I.wL,2)+" wH2="+wH2+"=I.wH^2="+Math.pow(I.wH,2));
	UnitTests.checkDoubleRel(I.s0,"I.s0",I0,"I0",eps);
	UnitTests.checkDoubleRel(Math.pow(I.wL,2),"I.wL^2",wL2,"wL2",eps);
	UnitTests.checkDoubleRel(Math.pow(I.wH,2),"I.wH^2",wH2,"wH2",eps);
    }
}


/** "Evening Train" version from Aug 28,2015. Instead of just one |psiH>
 * try to sum over all |psik>, the eigenfunctions of I|psik>=I_k|psik>
 * dirC=(2 sum_k <psik|pI|D(psik)>-<psik|pI|psik><psik|D(psik)>)/o_o
 * Note that 2<psik|D(psik)>=psik(x0^2) and the dirC has actually a form
 * of price differences multipled by weight wk=I_k*psik(x0)^2:
 * (sum_k (<psik|pI|D(psik)>/<psik|I|D(psik)>-<psik|pI|psik>/<psik|I|psik>)*wk)/o_o.
 * Note the <psik|I|D(psik)>=0.5*I*psik(x0)^2 and <psik|I|psik>=I_k ,
 * I0= (sum_k I_k*psik(x0)^2)/o_o and o_o=sum_k psik(x0)^2.
 * As one can see the dirC is similar to I0, but uses price variation
 * weighted with wk, and I0 is a ratio of two reproducing kernels.
 */
static DynVarSP0 get_DynVarSP0(final int n,
			       final double [] QQi,final double [] QQt,final double [] QQpi,
			       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi)
{
    //if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] w=new double [n],d=new double [n],o=new double [n],pI=new double [n];
    for(int k=0;k<n;k++)
    {
	final double [] psi=I.getPsi(k),Dpsi=VOLATILITY_EDPsi.getDPsi(psi);
	final double Ik=I.ev[k],psik2x0=2*I.getFPsiPsi(k,Dpsi,QQt);
	pI[k]=I.getFPsi(k,QQpi);
	o[k]=psik2x0;
	w[k]=Ik*psik2x0;
	d[k]=2*I.getFPsiPsi(k,Dpsi,QQpi)-pI[k]*psik2x0;
	if(EPS_FLAG_CHECKCONSTRAINS>0)
	{
	    final double eps=EPS_FLAG_CHECKCONSTRAINS;
	    System.err.println("DynVarSP0.get_DynVarSP0:k="+k+" psik2x0="+psik2x0+" psik(x0)^2="+Math.pow(M.B.getD0(psi,I.x0),2));
	    UnitTests.checkDoubleRel(psik2x0,"psik2x0",Math.pow(M.B.getD0(psi,I.x0),2),"psik(x0)^2",eps);
	}
    }
    return new DynVarSP0(I,o,w,d,pI);
}

/** Like get_DynVarSP0, but uses p_last.
 */
static DynVarSP0 get_pl_DynVarSP0(final int n,
				  final double [] QQi,final double [] QQt,final double [] QQpi,
				  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double p_last)
{
    //if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] w=new double [n],d=new double [n],o=new double [n],pI=new double [n];
    for(int k=0;k<n;k++)
    {
	final double [] psi=I.getPsi(k);
	final double Ik=I.ev[k],psik2x0=Math.pow(M.B.getD0(psi,I.x0),2);
	pI[k]=I.getFPsi(k,QQpi);
	o[k]=psik2x0;
	w[k]=Ik*psik2x0;
	d[k]=(p_last*Ik-pI[k])*psik2x0;
    }
    return new DynVarSP0(I,o,w,d,pI);
}

}

