package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 * Variate <psi|V|psi>/<psi|psi>.
 *  @version $Id: DynVarV.java,v 1.14 2020/09/27 20:33:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVarV {
    final EVXData V;
    final double pi_sum,pv_sum;
    final double [] I,dpI;
    DynVarV(final EVXData V,final double pi_sum,final double pv_sum,final double [] I,final double [] dpI)
{
    this.V=V;
    this.pi_sum=pi_sum;
    this.pv_sum=pv_sum;
    this.I=I;
    this.dpI=dpI;
}
    static boolean PRINT_DIAG=false;

/**
 * Note: QQV & QQpV must be set to have V(x0)=0 & pV(x0)=0.
 */
static DynVarV get_DynVarV(final int n,
			   final double [] QQV,final double [] QQt,final double [] QQpV,
			   final double [] test_QQi,final double [] test_QQpi,final double eps_selftest,
			   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final VolMatrix VOLATILITY_EDPsi)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData V=EVXData.getEVXDataFromQQ(n,QQV,QQt,M);
    final double [] QQi=VOLATILITY_EDPsi.getDiffMatrix(n,QQV),QQpi=VOLATILITY_EDPsi.getDiffMatrix(n,QQpV);
    final double [] dpI=new double [n],I=new double [n];
    double si=0,spi=0,sV=0,spV=0;
    for(int k=0;k<n;k++)
    {
	final double [] psik=V.getPsi(k),Dpsik=VOLATILITY_EDPsi.getDPsi(psik);
	final double ik=TMatr.getAMatrA(psik,QQi,n),pik=TMatr.getAMatrA(psik,QQpi,n),
	    pVk=TMatr.getAMatrA(psik,QQpV,n),psik2_x0=2*TMatr.getAMatrB(psik,Dpsik,QQt,n);
	final double var2V=2*(TMatr.getAMatrA(Dpsik,QQV,n)-V.ev[k]*TMatr.getAMatrA(Dpsik,QQt,n)),
	    pVkx0=pVk*psik2_x0,var1pV=(pik-pVkx0);
	dpI[k]=var1pV;
	I[k]=ik;
	si+=ik;
	spi+=pik;
	// Simplified formulae, here V(x0)=0 is required to have ik=V.ev[k]*psik2_x0,
	// otherwise V.ev[k] should be adjusted on V offset.
	sV+=ik; /* +=V.ev[k]*psik2_x0 */
	spV+=pVkx0;
	if(PRINT_DIAG||eps_selftest>0)
	{
	    final double eps=eps_selftest>0?eps_selftest:1e-5;
	    final double Vk=TMatr.getAMatrA(psik,QQV,n);
	    UnitTests.checkDoubleRel(Vk,"Vk",V.ev[k],"V.ev[k]",eps);
	    UnitTests.checkDoubleRel(ik-Vk*psik2_x0,"ik-Vk*psik2_x0",0,"0",eps);
	    final double [] rqV=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQV,QQt,n);
	    UnitTests.checkDoubleRel(Vk,"Vk",rqV[0],"rqV[0]",eps);
	    UnitTests.checkDoubleRel(0,"0",rqV[1],"rqV[1]",eps);
	    UnitTests.checkDoubleRel(var2V,"var2V",rqV[2],"rqV[2]",eps);
	    final double [] rqpV=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQpV,QQt,n);
	    UnitTests.checkDoubleRel(var1pV,"var1pV",rqpV[1],"rqpV[1]",eps);

	    final double _psik2_x0=Math.pow(M.B.getD0(psik,V.x0),2);
	    UnitTests.checkDoubleRel(psik2_x0,"psik2_x0",_psik2_x0,"_psik2_x0",eps);
	    UnitTests.checkDoubleRel(ik,"ik",V.ev[k]*psik2_x0,"V.ev[k]*psik2_x0",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psik,QQt,n),"<psik|psik>",1,"1",eps);

	    System.err.println("DynVarV: test OK k="+k+" n="+n+" eps="+eps);
	}
    }
    final double pi_sum=spi/si,pv_sum=spV/sV;

    if(PRINT_DIAG||eps_selftest>0)
    {
	final double eps=eps_selftest>0?eps_selftest:1e-5;
	UnitTests.checkDoubleArrayRel(QQi,"QQi",test_QQi,"test_QQi",eps);
	UnitTests.checkDoubleArrayRel(QQpi,"QQpi",test_QQpi,"test_QQpi",eps);
	UnitTests.checkDoubleRel(pi_sum-pv_sum,"pi_sum-pv_sum",TMatr.sumElements(dpI)/TMatr.sumElements(I),"sum(dpI)/sum(I)",eps);

	final DynVar2C d=DynVar2C.get_DynVar2C(n,QQi,QQt,QQpi,M,VOLATILITY_EDPsi,DynVar2C.SELECTION_IH);
	// Spur should be invariant to basis choice.
	UnitTests.checkDoubleRel(pi_sum,"pi_sum",d.pa_sum,"d.pa_sum",eps);
	System.err.println("DynVarV: d.I.sH="+d.I.sH+" I="+java.util.Arrays.toString(I));

	System.err.println("DynVarV: test OK DynVar2C OK ; n="+n+" eps="+eps);
    }
    return new DynVarV(V,pi_sum,pv_sum,I,dpI);
}


}

