package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;

/**
 *  @version $Id: DynVarY.java,v 1.26 2022/11/16 08:59:02 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynVarY {
    private static final boolean FLAG_PRINT_DEBUG=false;
    final double I_0,I_IL,I_IH,p_IH,
	p_vI_min,p_vI_max,p_vI_A,p_vI_AA;
    DynVarY(final double I_0,final double I_IL,final double I_IH,final double p_IH,
	    final double p_vI_min,final double p_vI_max,final double p_vI_A,final double p_vI_AA)
    {
	this.I_0=I_0;
	this.I_IL=I_IL;
	this.I_IH=I_IH;
	this.p_IH=p_IH;
	this.p_vI_min=p_vI_min;
	this.p_vI_max=p_vI_max;
	this.p_vI_A=p_vI_A;
	this.p_vI_AA=p_vI_AA;
    }


/** Future I_0 assumed to be equal I_IH with psi(x0)=0 boundary condition.
 * Under this assumption future QQi is obtained as
 * QQi_future=QQi_init+ QQ0*((I_IH-I_0)/\sum psi(x0)^2)
 * and eignenvectors obtained.
 */
static DynVarY get_DynVarY(final int n,
			   final double [] QQi,final double [] QQt,final double [] QQpi,
			   final VolMatrix VOLATILITY_EDPsi,
			   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			   final double eps)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");
	final int i0=0,iL=1,iH=n-1;
	final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M);
	final double I_IH=basis[iH],pI_IH=TMatr.getAMatrA(n,basis,n+iH*n,QQpi,n),p_IH=pI_IH/I_IH;
	final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,M.getRadauRoot()));
	final double I_0=basis[i0],I_IL=basis[iL],dI0=I_IH-I_0,ifact=dI0/TMatr.getAMatrA(n,basis,n+i0*n,QQ0,n);
	final double [] QQifuture=TMatr.sumAcB(QQi,ifact,QQ0),
	    // For now calculate evF in original basis. Optimize later for the use of above calculated basis.
	    evF=M.EV.getGEV_Generalized(n,QQifuture,QQt,true);

	final double [] varIk=new double [n],varpIk=new double [n],varDk=new double [n];
	int k_vI_min=-1,k_vI_max=-1;
	double sp_vI=0,s_vI=0;
	for(int k=0;k<n;k++)
	{
	    final double [] psik=EVXData.getPsiFromBasis(k,evF,n),Dpsik=VOLATILITY_EDPsi.getDPsi(psik);
	    final double psi2kx0=2*TMatr.getAMatrB(psik,Dpsik,QQt,n);
	    varIk[k]=2*TMatr.getAMatrB(psik,Dpsik,QQi,n)-TMatr.getAMatrA(psik,QQi,n)*psi2kx0;
	    varpIk[k]=2*TMatr.getAMatrB(psik,Dpsik,QQpi,n)-TMatr.getAMatrA(psik,QQpi,n)*psi2kx0;
	    varDk[k]=2*TMatr.getAMatrB(psik,Dpsik,QQ0,n)-Math.pow(psi2kx0,2);
	    if(k<=0 || varIk[k]<=varIk[k_vI_min]) k_vI_min=k;
	    if(k<=0 || varIk[k]>=varIk[k_vI_max]) k_vI_max=k;
	    if(varIk[k]>0){s_vI+=varIk[k];sp_vI+=varpIk[k];}else{s_vI-=varIk[k];sp_vI-=varpIk[k];}
	}
	final double p_vI_min=varpIk[k_vI_min]/varIk[k_vI_min],
	    p_vI_max=varpIk[k_vI_max]/varIk[k_vI_max],p_vI_A=sp_vI/s_vI,
	    p_vI_AA=TMatr.sumElements(varpIk)/TMatr.sumElements(varIk);
	if(FLAG_PRINT_DEBUG||eps>0)
	{
	    final double [] psiH=EVXData.getPsiFromBasis(iH,basis,n),
		psi0=EVXData.getPsiFromBasis(i0,basis,n);
	    System.err.println("DynVarY:<psi0|QQi|psi0>="+TMatr.getAMatrA(psi0,QQi,n));
	    System.err.println("DynVarY:<psiH|QQi|psiH>="+TMatr.getAMatrA(psiH,QQi,n));
	    System.err.println("DynVarY:<psi0|QQifuture|psi0>="+TMatr.getAMatrA(psi0,QQifuture,n));
	    System.err.println("DynVarY:<psiH|QQifuture|psiH>="+TMatr.getAMatrA(psiH,QQifuture,n));
	    UnitTests.checkDoubleRel(I_IH,"I_IH",TMatr.getAMatrA(psi0,QQifuture,n),"<psi0|QQifuture|psi0>",eps);
	    UnitTests.checkDoubleRel(I_IH,"I_IH",TMatr.getAMatrA(psiH,QQifuture,n),"<psiH|QQifuture|psiH>",eps);
	    final double x0=M.getRadauRoot();
	    final double [] psi2x0=new double [n],vIk=new double [n],vpIk=new double [n];
	    for(int k=0;k<n;k++)
	    {
		final double [] psik=EVXData.getPsiFromBasis(k,evF,n),
		    Dpsik=VOLATILITY_EDPsi.getDPsi(psik);
		final double
		    vIF=2*(TMatr.getAMatrB(psik,Dpsik,QQifuture,n)-
			   TMatr.getAMatrA(psik,QQifuture,n)*TMatr.getAMatrB(psik,Dpsik,QQt,n)),
		    vI=2*(TMatr.getAMatrB(psik,Dpsik,QQi,n)-
				   TMatr.getAMatrA(psik,QQi,n)*TMatr.getAMatrB(psik,Dpsik,QQt,n)),
		    vpI=2*(TMatr.getAMatrB(psik,Dpsik,QQpi,n)-
				   TMatr.getAMatrA(psik,QQpi,n)*TMatr.getAMatrB(psik,Dpsik,QQt,n)),
		    psikx0=M.B.getD0(psik,x0),Dpsikx0=M.B.getD0(Dpsik,x0),
		    vI0=ifact*(2*psikx0*Dpsikx0-Math.pow(psikx0,4)),
		    _vI0=ifact*(2*TMatr.getAMatrB(psik,Dpsik,QQ0,n)-Math.pow(TMatr.getAMatrA(psik,QQ0,n),2));
		psi2x0[k]=Math.pow(psikx0,2);
		vIk[k]=vI;
		vpIk[k]=vpI;
		UnitTests.checkDoubleRel(vIF,"vIF",0,"0",eps);
		UnitTests.checkDoubleRel1(vI,"vI",-vI0,"-vI0",eps);
		UnitTests.checkDoubleRel1(vI,"vI",-_vI0,"-_vI0",eps);
		UnitTests.checkDoubleRel1(vI,"vI",-_vI0,"-_vI0",eps);
		UnitTests.checkDoubleRel1(vI0,"vI0",ifact*varDk[k],"ifact*varDk[k]",eps);
		UnitTests.checkDoubleRel1(vpI,"vpI",varpIk[k],"varpIk[k]",eps);
		System.err.println("k="+k+" evF[k]="+evF[k]+" vI="+vI+" vI0="+vI0);
	    }
	    System.err.println("SUM(varDk)="+TMatr.sumElements(varDk)+" SUM(psi^2(x0))="+TMatr.sumElements(psi2x0));
	    UnitTests.checkDoubleRel(p_vI_AA,"p_vI_AA",TMatr.sumElements(vpIk)/(-ifact*TMatr.sumElements(varDk)),"sum(vpIk)/sum(vIk)",eps);
	    UnitTests.checkDoubleArrayRel(vIk,"vIk",varIk,"varIk",eps);
	    UnitTests.checkDoubleArrayRel(vpIk,"vpIk",varpIk,"varpIk",eps);
	}



	return new DynVarY(I_0,I_IL,I_IH,p_IH,p_vI_min,p_vI_max,p_vI_A,p_vI_AA);
    }


}

