package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;

/**
 *  @version $Id: DynYp1.java,v 1.15 2020/09/27 20:33:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynYp1 {
    private static final boolean FLAG_PRINT_DEBUG=true;
    final EVXData I;
    final double p_IH,p_IJp,r_IJp,I_IJp,var_p_IH,var_pt_IJp;
    DynYp1(final EVXData I,
	   final double p_IH,final double p_IJp,final double r_IJp,final double I_IJp,final double var_p_IH,final double var_pt_IJp)
    {
	this.I=I;
	this.p_IH=p_IH;
	this.p_IJp=p_IJp;
	this.r_IJp=r_IJp;
	this.I_IJp=I_IJp;
	this.var_p_IH=var_p_IH;
	this.var_pt_IJp=var_pt_IJp;
    }

private static final int N_iterations=10;

static DynYp1 get_DynYp1(final int n,
			 final double [] QQi,final double [] QQt,final double [] QQpi,
			 final double [] QQr,final double [] QQpr,final double [] QQp2r,
			 final VolMatrix VOLATILITY_EDPsi,
			 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");

	// First stage: in time space.
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int iH=n-1;
	final double [] psi_IH=I.getPsi(iH),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
	final double pI_IH=I.getFPsi(iH,QQpi),p_IH=pI_IH/I.ev[iH],
	    var_p_IH=2*(I.getFPsiPsi(iH,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iH,Dpsi_IH,QQt));

	// standart transform ev M ev^T
	final double [] bQQr=EVXData.convertMatrFromQtoB_Basis(n,QQr,I.ev),
	    bQQpr=EVXData.convertMatrFromQtoB_Basis(n,QQpr,I.ev),
	    bQQp2r=EVXData.convertMatrFromQtoB_Basis(n,QQp2r,I.ev),
	    bQQi=new double [n*n],bQQt=new double [n*n],bQQvv=new double [n*n];
	for(int k=0;k<n;k++){bQQi[k+n*k]=I.ev[k];bQQt[k+n*k]=1;}

	// Second stage: iterate in price space.
	final int iJ=0;
	double p_IJp=p_IH;
	double [] pev=null;
	for(int it=0;it<N_iterations;it++)
	{
	    for(int k=0;k<bQQvv.length;k++) bQQvv[k]=p_IJp*(p_IJp*bQQr[k]-2*bQQpr[k])+bQQp2r[k];
	    pev=EVXData.getSymEVDiagT(n,bQQvv,bQQi,M.EV);
	    p_IJp=TMatr.getAMatrA(n,pev,n+iJ*n,bQQpr,n)/TMatr.getAMatrA(n,pev,n+iJ*n,bQQr,n);
	}

	final double [] psi_IJp=EVXData.getPsiFromBasis(iJ,pev,n),
	    Dpsi_IJp=EVXData.convertPsiFromQtoB_Basis(
		n,VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,psi_IJp,I.ev)),I.ev,QQt);
	for(int k=0;k<bQQvv.length;k++) bQQvv[k]=p_IJp*bQQr[k]-bQQpr[k];
	final double r_IJp=pev[iJ],I_IJp=TMatr.getAMatrA(psi_IJp,bQQi,n)/TMatr.getAMatrA(psi_IJp,bQQt,n),
	    var_pt_IJp=I_IJp*2*(TMatr.getAMatrB(psi_IJp,Dpsi_IJp,bQQvv,n)-
			  TMatr.getAMatrA(psi_IJp,bQQvv,n)*TMatr.getAMatrB(psi_IJp,Dpsi_IJp,bQQi,n));
	if(FLAG_PRINT_DEBUG)
	{
	    System.err.println("DynYp1:Dpsi="+java.util.Arrays.toString(VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,psi_IJp,I.ev))));
	    System.err.println("DynYp1: <psi|QQvv|Dpsi>="+TMatr.getAMatrB(psi_IJp,Dpsi_IJp,bQQvv,n));
	    System.err.println("DynYp1: <psi_IJp|I|Dpsi_IJp>="+TMatr.getAMatrB(psi_IJp,Dpsi_IJp,bQQi,n));
	    System.err.println("DynYp1: <psi_IJp|var|Dpsi_IJp>="+TMatr.getAMatrB(psi_IJp,Dpsi_IJp,bQQvv,n));
	    System.err.println("DynYp1: <psi_IJp|var|psi_IJp>="+TMatr.getAMatrA(psi_IJp,bQQvv,n));
	}
	return new DynYp1(I,p_IH,p_IJp,r_IJp,I_IJp,var_p_IH,var_pt_IJp);
    }



static void match_DynYp(final DynYp1 a,DynYp b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.var_p_IH,"a.var_p_IH",b.var_p_IH,"b.var_p_IH",eps);
    UnitTests.checkDoubleRel(a.p_IJp,"a.p_IJp",b.p_IJp,"b.p_IJp",eps);
    UnitTests.checkDoubleRel(a.r_IJp,"a.r_IJp",b.r_IJp,"b.r_IJp",eps);
    UnitTests.checkDoubleRel(a.I_IJp,"a.I_IJp",b.I_IJp,"b.I_IJp",eps);
    UnitTests.checkDoubleRel(a.var_pt_IJp,"a.var_pt_IJp",b.var_pt_IJp,"b.var_pt_IJp",eps);
    System.err.println("DynYp1:match_DynYp test OK n="+a.I.neq+" eps="+eps);
}

}

