package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;

/**
 *  @version $Id: DynYpNoIter.java,v 1.22 2020/09/27 20:33:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynYpNoIter {
    private static final boolean FLAG_PRINT_DEBUG=false;
    final EVXData I;
    final double p_IH,var_p_IH,var_pp_IpH,varp_pp_IpH;
    DynYpNoIter(final EVXData I,
		final double p_IH,final double var_p_IH,
		final double var_pp_IpH,final double varp_pp_IpH)
    {
	this.I=I;
	this.p_IH=p_IH;
	this.var_p_IH=var_p_IH;
	this.var_pp_IpH=var_pp_IpH;
	this.varp_pp_IpH=varp_pp_IpH;
    }

/**
 * Consider I|psi>=\lambda |psi>, corresponding to {max \lambda ; psiH}.
 * Calculate p_IH=<psiH|pI|psiH>/<psiH|I|psiH>,
 * then consider |psi> states, corresponding to max (not min!) volatility:
 * <psi|(p-p_IH)^2 I |psi>/<psi|psi>.
 * Then variate it twice: over psi (var(psi)=Dpsi),
 * and then over p_IH. First variation=0, the second one is non-0.
 * Hope it will be better than old answer (originating back to 2013):
 * var (p_IH) =2*(<psiH|pI|DpsiH>-<psiH|pI|psiH><psiH|DpsiH>)
 * used in  DynVHPnL.var_p_IH, DynVar2C.var_pI_IX and similar.
 * See also YkQVM.java for this old type of variations
 * and the DynYp.java where
 * min_{pp} max_{psi} <psi|(p-pp)^2 I |psi>/<psi|psi>
 * is considered (but without much success).
 */
static DynYpNoIter get_DynYpNoIter(final int n,
				   final double [] QQi,final double [] QQt,
				   final double [] QQpi,final double [] QQp2i,
				   final VolMatrix VOLATILITY_EDPsi,
				   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				   final double eps)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");

	// First stage: in time space.
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int iX=n-1; // maximal I
	final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
	final double pI_IH=I.getFPsi(iX,QQpi),p_IH=pI_IH/I.ev[iX],
	    var_p_IH=2*(I.getFPsiPsi(iX,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iX,Dpsi_IH,QQt));

	double [] bQQpx2I=new double [n*n],bQQpx1I=new double [n*n];
	final double p_volatility=p_IH;
	for(int k=0;k<bQQpx2I.length;k++) {
	    bQQpx2I[k]=p_volatility*(p_volatility*QQi[k]-2*QQpi[k])+QQp2i[k];
	    bQQpx1I[k]=(QQpi[k]-p_volatility*QQi[k]); // opposite sign for convenience
	}
	final double [] psi_IpH,Dpsi_IpH,ev_Ip,bQQt;
	if(false)
	{ // Non--optimized version (in original basis).
	    bQQt=QQt;
	    ev_Ip=EVXData.getEVXDataFromQQ(n,bQQpx2I,bQQt,M).ev; // optimize later
	    psi_IpH=EVXData.getPsiFromBasis(n-1,ev_Ip,n);
	    Dpsi_IpH=VOLATILITY_EDPsi.getDPsi(psi_IpH);
	}
	else
	{ // Optimized version (in I.ev basis).
	    bQQpx2I=EVXData.convertMatrFromQtoB_Basis(n,bQQpx2I,I.ev);
	    bQQpx1I=EVXData.convertMatrFromQtoB_Basis(n,bQQpx1I,I.ev);
	    ev_Ip=M.EV.getEV(n,bQQpx2I,true);
	    psi_IpH=EVXData.getPsiFromBasis(n-1,ev_Ip,n);
	    Dpsi_IpH=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,psi_IpH,I.ev)),I.ev,QQt);
	    bQQt=TMatr.getUnitMatrix(n);
	}
	final double varp_pp_IpH=TMatr.getAMatrA(psi_IpH,bQQpx1I,n),
	    var_pp_IpH=2*(TMatr.getAMatrB(psi_IpH,Dpsi_IpH,bQQpx1I,n)-
			  varp_pp_IpH*TMatr.getAMatrB(psi_IpH,Dpsi_IpH,bQQt,n));
	if(eps>0 || FLAG_PRINT_DEBUG)
	{
	    final double [] rqIp=RayleighQuotient.getRQVariationValues(psi_IpH,Dpsi_IpH,null,bQQpx2I,bQQt,n);
	    UnitTests.checkDoubleRel(rqIp[0],"rqIp[0]",ev_Ip[(n-1)],"Ip_sH",eps);
	    UnitTests.checkDoubleRel(rqIp[1],"rqIp[1]",0,"0",eps);
	    final double [] rq1Ip=RayleighQuotient.getRQVariationValues(psi_IpH,Dpsi_IpH,null,bQQpx1I,bQQt,n);
	    UnitTests.checkDoubleRel(rq1Ip[0],"rq1Ip[0]",varp_pp_IpH,"varp_pp_IpH",eps);
	    UnitTests.checkDoubleRel(rq1Ip[1],"rq1Ip[1]",var_pp_IpH,"var_pp_IpH",eps);

	    final double [] rqI=RayleighQuotient.getRQVariationValues(psi_IH,Dpsi_IH,null,QQpi,QQt,n);
	    UnitTests.checkDoubleRel(rqI[0],"rqI[0]",pI_IH,"pI_IH",eps);
	    UnitTests.checkDoubleRel(rqI[1],"rqI[1]",var_p_IH,"var_p_IH",eps);

	    System.err.println("DynYpNoIter: test OK, n="+n+" eps="+eps);

	}
	return new DynYpNoIter(I,p_IH,var_p_IH,var_pp_IpH,varp_pp_IpH);
    }


static void match_DynYp(final DynYpNoIter a,final DynYp b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.var_p_IH,"a.var_p_IH",b.var_p_IH,"b.var_p_IH",eps);
    UnitTests.checkDoubleRel(a.var_pp_IpH,"a.var_pp_IpH",b.var_pp_IpH,"b.var_pp_IpH",eps);
    UnitTests.checkDoubleRel(a.varp_pp_IpH,"a.varp_pp_IpH",b.varp_pp_IpH,"b.varp_pp_IpH",eps);
    System.err.println("DynYpNoIter:match_DynYp test OK n="+a.I.neq+" eps="+eps);
}


}

