package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.outils.FunctionCalculateable;
import com.polytechnik.outils.GoldenMin;

/**
 *  @version $Id: DynYpm.java,v 1.39 2022/08/05 11:25:32 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class DynYpm {
    private static final boolean FLAG_PRINT_DEBUG=true;
    final EVXData I;
    final double p_IH,var_p_IH,sigma_p,pe_VOL,I_VOL,p_VOL,var_pe_VOL;
    DynYpm(final EVXData I,
	   final double p_IH,final double var_p_IH,final double sigma_p,
	   final double pe_VOL,final double I_VOL,final double p_VOL,final double var_pe_VOL)
    {
	this.I=I;
	this.p_IH=p_IH;
	this.var_p_IH=var_p_IH;
	this.sigma_p=sigma_p;
	this.p_VOL=p_VOL;
	this.I_VOL=I_VOL;
	this.pe_VOL=pe_VOL;
	this.var_pe_VOL=var_pe_VOL;
    }

private static final int N_iterations=20,N_lintests=10;

static DynYpm test_get_DynYpm(final int n,
			 final double [] QQi,final double [] QQt,final double [] QQpi,final double [] QQp2i,
			      final boolean flag_J_H,
			 final VolMatrix VOLATILITY_EDPsi,
			 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");

	// First stage: in time space.
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int iX=flag_J_H?n-1:0; // maximal I
	final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
	final double pI_IH=I.getFPsi(iX,QQpi),p_IH=pI_IH/I.ev[iX],
	    var_p_IH=2*(I.getFPsiPsi(iX,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iX,Dpsi_IH,QQt));
	final double [] QQp2iS=new double [n*n];
	for(int k=0;k<QQp2iS.length;k++) QQp2iS[k]=p_IH*(p_IH*QQi[k]-2*QQpi[k])+QQp2i[k];
	double sp2=0;
	for(int k=0;k<n;k++){
	    sp2+=I.getFPsi(k,QQp2iS)/I.ev[k];
	}
	final double sigma_p=Math.sqrt(sp2/n);
	final int iJ=flag_J_H?n-1:0;

	class FCalc implements FunctionCalculateable {
	    double [] psi=null;
	    public double getF(final double p){
		for(int k=0;k<QQp2iS.length;k++) QQp2iS[k]=p*(p*QQi[k]-2*QQpi[k])+QQp2i[k];
		final double [] e=M.EV.getGEV_Generalized(n,QQp2iS,QQt,true);
		psi=EVXData.getPsiFromBasis(iJ,e,n);
		// calculate I on eigenvector of maximal(not minimal!) volatility
		return (flag_J_H?1:-1)*TMatr.getAMatrA(psi,QQi,n);
	    }
	}
	final FCalc fc=new FCalc();

	if(FLAG_PRINT_DEBUG)
	{
	    double Imin=Double.POSITIVE_INFINITY,p_Imin=Double.NaN;
	    for(double p=p_IH-2*sigma_p;!(p>p_IH+2*sigma_p);p+=0.02)
	    {
		final double i=fc.getF(p);
		//System.err.println("p="+p+" fc(p)="+i+" I_IH="+I.sH);
		if(i<Imin) {Imin=i;p_Imin=p;}
	    }
	    System.err.println("DynYpm:Imin="+Imin+" IH="+I.sH+" p_Imin="+p_Imin+" p_IH="+p_IH+" flag_J_H="+flag_J_H);
	    final double pr=GoldenMin.findMin(fc,p_IH-2*sigma_p,p_IH+2*sigma_p,N_iterations);
	    System.err.println("DynYpm:I(pr)="+((flag_J_H?1:-1)*fc.getF(pr))+" IH="+I.sH+" IL="+I.sL+" pr="+pr+" p_IH="+p_IH);
	}
	final double pe_VOL=GoldenMin.findMin(fc,p_IH-2*sigma_p,p_IH+2*sigma_p,N_iterations),
	    I_VOL=(flag_J_H?1:-1)*fc.getF(pe_VOL);
	if(FLAG_PRINT_DEBUG)
	{
	    final double [] rq=RayleighQuotient.getRQVariationValues(fc.psi,VOLATILITY_EDPsi.getDPsi(fc.psi),null,QQp2iS,QQt,n);
	    System.err.println("DynYpm: P2:rq[0]="+rq[0]+" rq[1]="+rq[1]+" must be =0");
	}
	final double p_VOL=TMatr.getAMatrA(fc.psi,QQpi,n)/TMatr.getAMatrA(fc.psi,QQi,n);
	for(int k=0;k<QQp2iS.length;k++) QQp2iS[k]=QQpi[k]-pe_VOL*QQi[k];
	final double [] rq=RayleighQuotient.getRQVariationValues(fc.psi,VOLATILITY_EDPsi.getDPsi(fc.psi),null,QQp2iS,QQt,n);
	final double var_pe_VOL=rq[1];
	if(FLAG_PRINT_DEBUG)
	{
	    System.err.println("DynYpm: pe_VOL="+pe_VOL+" p_VOL="+p_VOL+" p_IH="+p_IH+" rq[0]="+rq[0]+" sigma_p="+sigma_p);
	}

	return new DynYpm(I,p_IH,var_p_IH,sigma_p,pe_VOL,I_VOL,p_VOL,var_pe_VOL);
    }

static DynYpm get_DynYpm(final int n,
			 final double [] QQi,final double [] QQt,final double [] QQpi,final double [] QQp2i,
			 final boolean flag_J_H,
			 final VolMatrix VOLATILITY_EDPsi,
			 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");

	// First stage: in time space.
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int iX=flag_J_H?n-1:0; // maximal I
	final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
	final double pI_IH=I.getFPsi(iX,QQpi),p_IH=pI_IH/I.ev[iX],
	    var_p_IH=2*(I.getFPsiPsi(iX,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iX,Dpsi_IH,QQt));
	// standart transform ev M ev^T
	final double [] bQQi=TMatr.getDiagMatrix(I.getEValues()),
	    bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	    bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,QQp2i,I.ev),
	    bQQs=new double [n*n];

	final int iJ=flag_J_H?n-1:0;
	class FCalc implements FunctionCalculateable {
	    void setMatr(final double p){
		for(int k=0;k<bQQs.length;k++) bQQs[k]=p*(p*bQQi[k]-2*bQQpi[k])+bQQp2i[k];
	    }
	    double getPVolatilitySpur(final double p)
	    {
		setMatr(p);
		double s=0;
		for(int k=0;k<n;k++) s+=bQQs[k+n*k]/I.ev[k];
		return Math.sqrt(s/n);
	    }
	    double [] e=null;
	    public double getF(final double p){
		setMatr(p);
		e=M.EV.getEV(n,bQQs,true);
		// calculate I on eigenvector of maximal(not minimal!) volatility
		return (flag_J_H?1:-1)*TMatr.mulV2V(EVXData.getPsiFromBasis(iJ,e,n),I.getEValues());
	    }
	}
	final FCalc fc=new FCalc();
	final double sigma_p=fc.getPVolatilitySpur(p_IH),
	    pe_VOL=GoldenMin.findMin(fc,p_IH-2*sigma_p,p_IH+2*sigma_p,N_iterations),I_VOL=(flag_J_H?1:-1)*fc.getF(pe_VOL);
	final double [] psi=EVXData.getPsiFromBasis(iJ,fc.e,n),
	    Dpsi=EVXData.convertPsiFromQtoB_Basis(
		n,VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,psi,I.ev)),I.ev,QQt);
	TMatr.sumP1(Dpsi,psi,-TMatr.mulVV(psi,Dpsi));
	final double p_VOL=TMatr.getAMatrA(psi,bQQpi,n)/I_VOL,
	    var_pe_VOL=2*(TMatr.getAMatrB(psi,Dpsi,bQQpi,n)-pe_VOL*TMatr.getAMatrB(psi,Dpsi,bQQi,n));

	return new DynYpm(I,p_IH,var_p_IH,sigma_p,pe_VOL,I_VOL,p_VOL,var_pe_VOL);
    }

static DynYpm test_get_DynYpi(final int n,
			      final double [] QQi,final double [] QQt,final double [] QQpi,final double [] QQp2i,
			      final VolMatrix VOLATILITY_EDPsi,
			      final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");

	// First stage: in time space.
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int iX=n-1; // maximal I
	final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH),QQp2iS=new double [n*n];
	final double pI_IH=I.getFPsi(iX,QQpi),p_IH=pI_IH/I.ev[iX],
	    var_p_IH=2*(I.getFPsiPsi(iX,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iX,Dpsi_IH,QQt));

	final int iJ=0; // minimal volatility
	class FCalc implements FunctionCalculateable {
	    void setMatr(final double p)
	    {
		for(int k=0;k<QQp2iS.length;k++) QQp2iS[k]=p*(p*QQi[k]-2*QQpi[k])+QQp2i[k];
	    }
	    double getPVolatilitySpur(final double p)
	    {
		setMatr(p);
		double s=0;
		for(int k=0;k<n;k++){
		    s+=I.getFPsi(k,QQp2iS)/I.ev[k];
		}
		return Math.sqrt(s/n);
	    }
	    double [] psi=null;
	    EVXData J=null;
	    public double getI(final double p){
		setMatr(p);
		J=EVXData.getEVXDataFromQQ(n,QQp2iS,QQi,M);
		psi=J.getPsi(iJ);
		// calculate I on eigenvector of minimal volatility
		return 1/TMatr.getAMatrA(psi,QQt,n);
	    }
	    public double getF(final double p){return -getI(p);/* "-" sign to find max I */}
	}
	final FCalc fc=new FCalc();
	final double sigma_p=fc.getPVolatilitySpur(p_IH);
	int ind_Imin=0;double itstep=2*sigma_p/N_lintests,Imax=Double.NEGATIVE_INFINITY;
	for(int it=-N_lintests;it<=N_lintests;it++)
	{
	    final double p=p_IH+it*itstep,i=fc.getI(p);
	    if(i>=Imax){Imax=i;ind_Imin=it;}
	}

	final double pe_VOL=GoldenMin.findMin(fc,p_IH+(ind_Imin-1)*itstep,p_IH+(ind_Imin+1)*itstep,N_iterations),I_VOL=fc.getI(pe_VOL);
	final double p_VOL=TMatr.getAMatrA(fc.psi,QQpi,n)/TMatr.getAMatrA(fc.psi,QQi,n);
	final double [] QQpp=TMatr.sumAcB(QQpi,pe_VOL,QQi);
	final double [] rq=RayleighQuotient.getRQVariationValues(fc.psi,VOLATILITY_EDPsi.getDPsi(fc.psi),null,QQpp,QQi,n);
	final double var_pe_VOL=rq[1];

	if(FLAG_PRINT_DEBUG)
	{
	    System.err.println("DynYpi: I.sL="+I.sL+" I.sH="+I.sH+" I_VOL="+I_VOL+" pe_VOL="+pe_VOL+" p_VOL="+p_VOL+" pert="+DeprecatedHelperEVXData.getPertr1Term(fc.J,iJ,QQt,QQp2iS));
	    final double [] rqt=RayleighQuotient.getRQVariationValues(fc.psi,VOLATILITY_EDPsi.getDPsi(fc.psi),null,QQp2iS,QQi,n);
	    final double [] dpsi=VOLATILITY_EDPsi.getDPsi(fc.psi);
	    TMatr.sumP1(dpsi,fc.psi,-TMatr.getAMatrB(fc.psi,dpsi,QQi,n));
	    System.err.println("DynYpi: P2:rqt[0]="+rqt[0]+" rqt[1]="+rqt[1]+" must be =0"+" varp1="+(2*TMatr.getAMatrB(fc.psi,dpsi,QQpp,n))+" var_pe_VOL="+var_pe_VOL);
	}

	return new DynYpm(I,p_IH,var_p_IH,sigma_p,pe_VOL,I_VOL,p_VOL,var_pe_VOL);
    }

static DynYpm get_DynYpi(final int n,
			 final double [] QQi,final double [] QQt,final double [] QQpi,final double [] QQp2i,
			 final VolMatrix VOLATILITY_EDPsi,
			 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	if(n<2) throw new IllegalArgumentException("n="+n+"<2");

	// First stage: in time space.
	final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
	final int iX=n-1; // maximal I
	final double [] psi_IH=I.getPsi(iX),Dpsi_IH=VOLATILITY_EDPsi.getDPsi(psi_IH);
	final double pI_IH=I.getFPsi(iX,QQpi),p_IH=pI_IH/I.ev[iX],
	    var_p_IH=2*(I.getFPsiPsi(iX,Dpsi_IH,QQpi)-pI_IH*I.getFPsiPsi(iX,Dpsi_IH,QQt));
	// standart transform ev M ev^T
	final double [] bQQi=TMatr.getDiagMatrix(I.getEValues()),
	    bQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,I.ev),
	    bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,QQp2i,I.ev),
	    bQQs=new double [n*n];

	final int iJ=0;
	class FCalc implements FunctionCalculateable {
	    void setMatr(final double p){
		for(int k=0;k<bQQs.length;k++) bQQs[k]=p*(p*bQQi[k]-2*bQQpi[k])+bQQp2i[k];
	    }
	    double getPVolatilitySpur(final double p)
	    {
		setMatr(p);
		double s=0;
		for(int k=0;k<n;k++) s+=bQQs[k+n*k]/I.ev[k];
		return Math.sqrt(s/n);
	    }
	    double [] e=null;
	    public double getI(final double p){
		setMatr(p);
		e=EVXData.getSymEVDiagT(n,bQQs,bQQi,M.EV);
		// calculate I on eigenvector of minimal volatility
		return 1/DeprecatedHelperTMatr.mulVVo(n,e,n+iJ*n,e,n+iJ*n);
	    }
	    public double getF(final double p){return -getI(p);/* "-" sign to find max I */}
	}
	final FCalc fc=new FCalc();
	final double sigma_p=fc.getPVolatilitySpur(p_IH);
	int ind_Imin=0;double itstep=2*sigma_p/N_lintests,Imax=Double.NEGATIVE_INFINITY;
	for(int it=-N_lintests;it<=N_lintests;it++)
	{
	    final double p=p_IH+it*itstep,i=fc.getI(p);
	    if(i>=Imax){Imax=i;ind_Imin=it;}
	    //System.err.println("DynYpi: i="+i+" p="+p+" p_IH="+p_IH+" Imax="+Imax+" I.sH="+I.sH);
	}
	//System.err.println("DynYpi:ind_Imin="+ind_Imin);
	final double pe_VOL=GoldenMin.findMin(fc,p_IH+(ind_Imin-1)*itstep,p_IH+(ind_Imin+1)*itstep,N_iterations),I_VOL=fc.getI(pe_VOL);
	final double [] psi=EVXData.getPsiFromBasis(iJ,fc.e,n),
	    Dpsi=EVXData.convertPsiFromQtoB_Basis(
		n,VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,psi,I.ev)),I.ev,QQt);
	for(int k=0;k<bQQs.length;k++) bQQs[k]=bQQpi[k]-pe_VOL*bQQi[k];
	final double p_VOL=TMatr.getAMatrA(psi,bQQpi,n)/TMatr.getAMatrA(psi,bQQi,n),
	    var_pe_VOL=2*(TMatr.getAMatrB(psi,Dpsi,bQQs,n)-
			  TMatr.getAMatrA(psi,bQQs,n)*TMatr.getAMatrB(psi,Dpsi,bQQi,n));
	//System.err.println("DynYpi:<psi|QQs|psi>="+TMatr.getAMatrA(psi,bQQs,n));

	return new DynYpm(I,p_IH,var_p_IH,sigma_p,pe_VOL,I_VOL,p_VOL,var_pe_VOL);
    }


static void match_DynYpm(final DynYpm a,DynYpm b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,eps);
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH,"b.p_IH",eps);
    UnitTests.checkDoubleRel(a.var_p_IH,"a.var_p_IH",b.var_p_IH,"b.var_p_IH",eps);
    UnitTests.checkDoubleRel(a.sigma_p,"a.sigma_p",b.sigma_p,"b.sigma_p",eps);
    UnitTests.checkDoubleRel(a.pe_VOL,"a.pe_VOL",b.pe_VOL,"b.pe_VOL",eps);
    UnitTests.checkDoubleRel(a.I_VOL,"a.I_VOL",b.I_VOL,"b.I_VOL",eps);
    UnitTests.checkDoubleRel(a.p_VOL,"a.p_VOL",b.p_VOL,"b.p_VOL",eps);
    UnitTests.checkDoubleRel(a.var_pe_VOL,"a.var_pe_VOL",b.var_pe_VOL,"b.var_pe_VOL",eps);

    System.err.println("DynYpm:match_DynYpm test OK n="+a.I.neq+" eps="+eps);
}

}

