package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Time evolution as commutator acting on |psi0>.
 *  @version $Id: EComm.java,v 1.90 2022/11/19 11:08:57 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class EComm {
final EVXData I;
final double [] psi0,varpsi0,Dpsi0;
final double Dsens,Dsens_rho,Dsens_nrm,IDsens,xDsens,xDsens_rho,xxDsens,xxDsens_rho,xxDsens_Rrho,spDsens,spDsens_rho;

EComm(final int n,final double [] imom,final double [] tmom,final double [] fmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final double [] QQt=M.B.getQQMatr(n,tmom),QQi=M.B.getQQMatr(n,imom);
    this.I=isDiagQQt()?EVXData.getEVXDataFromQQfDiagQQa(n,QQi,QQt,M):
	EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double x0=this.I.x0;

    // convert I.psi0 to new basis
    this.psi0=getPsi0(this.I);
    this.Dpsi0=getDpsi(this.psi0);
    final double [] tmppsi=new double [n],psi0B=EVXData.convertPsiFromQtoB_Basis(n,this.psi0,this.I.ev,QQt),
	Dpsi0B=EVXData.convertPsiFromQtoB_Basis(n,this.Dpsi0,this.I.ev,QQt);
    // OK, now psi0B is the psi0 and Dpsi0B is Dpsi0 in the basis of I.ev eigenvectors.

    // Remove |psi0B> from |Dpsi0B> for the consistency of normalizing.
    // Not really important because <varpsi0B|psi0B>=0.
    TMatr.sumP1(Dpsi0B,psi0B,-TMatr.mulVV(Dpsi0B,psi0B));

    // QQi & QQt are diagonal in new basis.
    java.util.Arrays.fill(QQi,0);
    final double [] xDpsi0B=new double [n],xxDpsi0B=new double [n];
    for(int k=0;k<n;k++){
	QQi[k*n+k]=this.I.ev[k]; // QQi in new basis is diagonal.
	System.arraycopy(this.I.ev,n+k*n,tmppsi,0,n);
	xDpsi0B[k]=M.B.getD0(getDpsi(tmppsi),x0);
	xxDpsi0B[k]=M.B.getD1(tmppsi,x0);
    }
    TMatr.sumP1(xDpsi0B,psi0B,-TMatr.mulVV(xDpsi0B,psi0B));
    TMatr.sumP1(xxDpsi0B,psi0B,-TMatr.mulVV(xxDpsi0B,psi0B));

    // Convert the remaining matrices to new basis (only QQf left).
    final double [] QQf=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,fmom),this.I.ev);

    final double [] QQc=TMatr.mulMM(n,QQi,QQf);
    // Anti-Hermitian part (QQi QQf - QQf QQi) is responsible for time evolution.
    TMatr.scaleVectorInPlace(TMatr.selectAntiHermitianPartInPlace(n,QQc),2);

    // Time evolution of psi0: |varpsi0B> = QQc|psi0B>.
    final double [] varpsi0B=TMatr.getMatr_x_A(n,QQc,psi0B);
    // Correlation to infinitesimal time shift operator.
    this.Dsens=TMatr.mulVV(varpsi0B,Dpsi0B);
    final double b0b0=TMatr.mulVV(varpsi0B,varpsi0B);
    this.Dsens_rho=this.Dsens/Math.sqrt(b0b0*TMatr.mulVV(Dpsi0B,Dpsi0B));
    this.Dsens_nrm=this.Dsens/Math.sqrt(sumAB2(n,varpsi0B,Dpsi0B));

    this.IDsens=TMatr.getAMatrB(varpsi0B,psi0B,QQi,n);
    this.xDsens=TMatr.mulVV(varpsi0B,xDpsi0B);
    this.xDsens_rho=this.xDsens/Math.sqrt(b0b0*TMatr.mulVV(xDpsi0B,xDpsi0B));
    this.xxDsens=TMatr.mulVV(varpsi0B,xxDpsi0B);
    this.xxDsens_rho=this.xxDsens/Math.sqrt(b0b0*TMatr.mulVV(xxDpsi0B,xxDpsi0B));
    final double [] varxxDpsi0B=TMatr.getMatr_x_A(n,QQc,xxDpsi0B);
    this.xxDsens_Rrho=this.xxDsens/Math.sqrt(TMatr.mulVV(varxxDpsi0B,varxxDpsi0B)*TMatr.mulVV(psi0B,psi0B));

    // Convert |varpsi0B> to original Q-basis.
    this.varpsi0=EVXData.convertPsiFromBtoQ_Basis(n,varpsi0B,this.I.ev);

    double sp=0,sigma2_d=0,sigma2_c=0;
    for(int k=0;k<n;k++)
    {
	final double [] psik=this.I.getPsi(k),
	    psikB=EVXData.convertPsiFromQtoB_Basis(n,psik,this.I.ev,QQt),
	    DpsikB=EVXData.convertPsiFromQtoB_Basis(n,getDpsi(psik),this.I.ev,QQt);
	TMatr.sumP1(DpsikB,psikB,-TMatr.mulVV(DpsikB,psikB));
	double [] tmp=TMatr.getMatr_x_A(n,QQc,psikB);
	sp+=TMatr.mulVV(tmp,DpsikB);
	sigma2_d+=TMatr.mulVV(DpsikB,DpsikB);
	sigma2_c+=TMatr.mulVV(tmp,tmp);
    }
    this.spDsens=sp;
    this.spDsens_rho=sp/Math.sqrt(sigma2_d*sigma2_c);

    // for debugging
    if(false)
    {
	final double eps=1e-5;
	final double [] ppvarpsi0=new double [n],pppsi0=new double [n],ppDpsi0=new double [n];
	for(int k=0;k<n;k++)
	{
	    final double [] psik=this.I.getPsi(k);
	    TMatr.sumP1(ppvarpsi0,psik,varpsi0B[k]);
	    TMatr.sumP1(pppsi0,psik,psi0B[k]);
	    TMatr.sumP1(ppDpsi0,psik,Dpsi0B[k]);
	}
	System.err.println("ppvarpsi0="+java.util.Arrays.toString(ppvarpsi0));
	System.err.println("this.varpsi0="+java.util.Arrays.toString(this.varpsi0));
	System.err.println("pppsi0="+java.util.Arrays.toString(pppsi0));
	System.err.println("psi0="+java.util.Arrays.toString(psi0));
	UnitTests.checkDoubleArrayRel(ppvarpsi0,"ppvarpsi0",this.varpsi0,"this.varpsi0",eps);
	UnitTests.checkDoubleArrayRel(pppsi0,"pppsi0",psi0,"psi0",eps);
	// does not match because we removed |psi0B> from |Dpsi0B>
	//UnitTests.checkDoubleArrayRel(ppDpsi0,"ppDpsi0",Dpsi0,"Dpsi0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(varpsi0B,psi0B),"<psi0B|varpsi0B>",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(Dpsi0B,psi0B,QQc,n),"<Dpsi0B|QQc|psi0B>",Dsens,"Dsens",eps);

	System.err.println("ab2="+sumAB2(n,varpsi0B,Dpsi0B));
	System.err.println("ab2sum0="+sumAB2(n,varpsi0B,TMatr.sumAcB(Dpsi0B,n*n*1.023,psi0B)));
	if(psi0!=I.psi0)UnitTests.checkDoubleRel(sumAB2(n,varpsi0B,Dpsi0B),"ab2",sumAB2(n,varpsi0B,TMatr.sumAcB(Dpsi0B,n*n*1.023,psi0B)),"ab2sum0",eps);

	System.err.println("EComm:constructor OK n="+n+" eps="+eps);
    }

    this.simom=imom;
    this.stmom=tmom;
    this.sfmom=fmom;
    this.sM=M;
}

double [] getPsi0(final EVXData j){return j.psi0;}

/** For future optimization.
 */
boolean isDiagQQt(){return false;}


static double sumAB2(final int n,final double [] a,final double [] b)
{
    double sd2=0;
    for(int k=0;k<n;k++) sd2+=Math.pow(a[k]*b[k],2);
    return sd2;
}

abstract double [] getDpsi(final double [] psi);

final double [] simom,stmom,sfmom;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> sM;

void match_EiComm(final EiComm e,final double eps)
{
    final int n=this.I.neq;
    EVXData.matchEVXData(this.I,e.I,false,true,eps);
    UnitTests.checkDoubleRel(e.Dsens_Rrho,"e.Dsens_Rrho",this.xxDsens_Rrho,"this.xxDsens_Rrho",eps);
    UnitTests.checkDoubleRel(e.Dsens_rho,"e.Dsens_rho",this.xxDsens_rho,"this.xxDsens_rho",eps);
    System.err.println("EComm:match_EiComm: test OK n="+n+" eps="+eps);

}

void test_Self(final double eps)
{
    final int n=this.I.neq;
    final double [] QQt=sM.B.getQQMatr(n,stmom),QQi=sM.B.getQQMatr(n,simom),QQf=sM.B.getQQMatr(n,sfmom),IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt);
    final double [] QQC=TMatr.sumAcB(
	TMatr.mulMatrices(n,new double [][] {QQi,IQQt,QQf}),
	-1,
	TMatr.mulMatrices(n,new double [][] {QQf,IQQt,QQi})),
	QQc=TMatr.mulMM(n,IQQt,QQC);

    final double x0=this.I.x0;
    final double [] varpsi=TMatr.getMatr_x_A(n,QQc,psi0);
    System.err.println("this.varpsi0(x0)="+sM.B.getD0(this.varpsi0,x0));
    System.err.println("varpsi(x0)="+sM.B.getD0(varpsi,x0));
    UnitTests.checkDoubleArrayRel1(varpsi,"varpsi",this.varpsi0,"this.varpsi0",eps);
    UnitTests.checkDoubleRel(TMatr.getAMatrB(psi0,varpsi0,QQt,n),"<psi0|QQt|varpsi0>",0,"0",eps);
    UnitTests.checkDoubleRel(TMatr.getAMatrB(Dpsi0,varpsi,QQt,n),"<Dpsi0|QQt|varpsi0>",Dsens,"Dsens",eps);
    double sp=0,sigma2_d=0,sigma2_c=0;
    for(int k=0;k<n;k++)
    {
	final double [] psik=this.I.getPsi(k),Dpsik=getDpsi(psik);
	sp+=TMatr.getAMatrB(Dpsik,psik,QQC,n);
	sigma2_c+=TMatr.getAMatrA(psik,TMatr.mulMM(n,TMatr.getMatrTrans(n,QQC),QQc),n);
	TMatr.sumP1(Dpsik,psik,-TMatr.getAMatrB(Dpsik,psik,QQt,n));
	sigma2_d+=TMatr.getAMatrA(Dpsik,QQt,n);
    }
    UnitTests.checkDoubleRel(sp,"sp",spDsens,"spDsens",eps);
    UnitTests.checkDoubleRel(sp/Math.sqrt(sigma2_d*sigma2_c),"sp_rho",spDsens_rho,"spDsens_rho",eps);

    if(this.psi0==this.I.psi0)
    {
	UnitTests.checkDoubleRel(sM.B.getD0(varpsi,x0)/this.I.s0,"varpsi0(x0)",0,"0",eps);
	UnitTests.checkDoubleRel(sM.B.getD0(this.varpsi0,x0),"this.varpsi0(x0)",0,"0",eps);
	// Integration by parts and boundary condition varpsi0(x0)=0.
	UnitTests.checkDoubleRel(TMatr.getAMatrB(getDpsi(psi0),varpsi0,QQt,n),"<D(psi0)|varpsi0>",-TMatr.getAMatrB(getDpsi(varpsi0),psi0,QQt,n),"-<D(varpsi0)|psi0>",eps);

	System.err.println("IDsens="+IDsens+" xDsens="+xDsens+" xxDsens="+xxDsens);

	System.err.println("o_o="+this.I.o_o+" Dsens="+Dsens+" xDsens="+xDsens+" xDsens/Dsens="+(xDsens/Dsens));
	UnitTests.checkDoubleRel(this.Dsens,"this.Dsens",-this.xDsens/Math.sqrt(this.I.o_o),"-this.xDsens/sqrt(this.I.o_o)",eps);

	/*
	final double [] QQJ=TMatr.mulMM(n,QQc,IQQt);
	final double [] kk=saccess.getKK(n,new double [][]{QQJ})[0];
	UnitTests.checkDoubleRel(saccess.fx(kk,x0)/this.I.s0,"kk(x0)",0,"0",eps);
	UnitTests.checkDoubleRel(saccess.Dfx(kk,x0)/this.I.o_o,"Dkk(x0)",0,"0",eps);
	System.err.println("QQJ="+java.util.Arrays.toString(QQJ));
	for(int k=0;k<n;k++) for(int l=0;l<=k;l++) QQJ[k*n+l]=0;
	UnitTests.checkDoubleRel(saccess.Dfx(saccess.getKK(n,new double [][]{QQJ})[0],x0)/this.I.o_o,"varDkk(x0)",this.xDsens,"this.xDsens",eps);
	*/



	System.err.println("EComm: extra checks for the case psi0==I.psi0 are OK");
    }

    System.err.println("EComm:test_Self: test OK n="+n+" eps="+eps);
}

} // END OF class EComm
