package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVSolver;

/** Convert to spectrum.
 *  @version $Id: EVX2Data.java,v 1.18 2022/11/20 12:41:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class EVX2Data {
    final double sIL,sIH,sfL,sfH,w_IL_fL,w_IL_fH,w_IH_fL,w_IH_fH;
    final double [] ev;
    final int neq;
    final ACorr a;

    EVX2Data(final int neq,
	     final double sIL,final double sIH,final double sfL,final double sfH,
	     final double w_IL_fL,final double w_IL_fH,final double w_IH_fL,final double w_IH_fH,
	     final double [] ev,
	     final double x_If,final double x_II,final double x_ff)
    {
	this.neq=neq;
	this.sIL=sIL;
	this.sIH=sIH;
	this.sfL=sfL;
	this.sfH=sfH;
	this.w_IL_fL=w_IL_fL;
	this.w_IL_fH=w_IL_fH;
	this.w_IH_fL=w_IH_fL;
	this.w_IH_fH=w_IH_fH;
	this.ev=ev;
	this.a=new ACorr(x_If,x_II,x_ff);
    }

public static void matchEVX2Data(final EVX2Data a,final EVX2Data b,final boolean flag_dimcheck,final boolean flag_psicheck,final double eps)
{
    UnitTests.checkDoubleRel(a.sIL,"a.sIL",b.sIL,"b.sIL",eps);
    UnitTests.checkDoubleRel(a.sIH,"a.sIH",b.sIH,"b.sIH",eps);
    UnitTests.checkDoubleRel(a.sfL,"a.sfL",b.sfL,"b.sfL",eps);
    UnitTests.checkDoubleRel(a.sfH,"a.sfH",b.sfH,"b.sfH",eps);

    UnitTests.checkDoubleRel(Math.abs(a.w_IL_fL),"a.w_IL_fL",Math.abs(b.w_IL_fL),"b.w_IL_fL",eps);
    UnitTests.checkDoubleRel(Math.abs(a.w_IL_fH),"a.w_IL_fH",Math.abs(b.w_IL_fH),"b.w_IL_fH",eps);
    UnitTests.checkDoubleRel(Math.abs(a.w_IH_fL),"a.w_IH_fL",Math.abs(b.w_IH_fL),"b.w_IH_fL",eps);
    UnitTests.checkDoubleRel(Math.abs(a.w_IH_fH),"a.w_IH_fH",Math.abs(b.w_IH_fH),"b.w_IH_fH",eps);
    ACorr.matchACorr(a.a,b.a,eps);

    if(flag_dimcheck)
    {
	UnitTests.checkDoubleRel(a.neq,"a.neq",b.neq,"b.neq",0);
	UnitTests.checkDoubleArrayRel(java.util.Arrays.copyOf(a.ev,a.neq),"a.ev",java.util.Arrays.copyOf(b.ev,b.neq),"b.ev",eps);
    }

    if(flag_psicheck)
    {
	UnitTests.checkDoubleRel(a.neq,"a.neq",b.neq,"b.neq",0);
	for(int k=0;k<a.neq;k++)
	{
	    final double [] psiA=EVXData.getPsiFromBasis(k,a.ev,a.neq),
		psiB=EVXData.getPsiFromBasis(k,b.ev,b.neq);
	    UnitTests.checkDoubleArrayRel(EVXData.getMSign(psiA),"psiA",EVXData.getMSign(psiB),"psiB",eps);
	}
    }
    System.err.println("EVX2Data:matchEVX2Data test OK neq="+a.neq+" eps="+eps);
}



public static EVX2Data getEVX2DataPsi0FromQQ(final int n,final double [] QQi,final double [] QQt,final double [] QQf,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int nd=n-1,nd1=nd-1;
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M),
	bQQf=EVXData.convertMatrFromQtoB_Basis(n,QQf,basis,nd,n+n+0);
    double x_If=0,x_II=0;
    for(int k=0;k<nd;k++)
    {
	x_If+=basis[k+1]*bQQf[k*nd+k];
	x_II+=Math.pow(basis[k+1],2);
    }
    final double [] evf=M.EV.getEV(nd,bQQf,true);
    return new EVX2Data(n,basis[1],basis[n-1],evf[0],evf[nd-1],
			evf[nd+(0*nd+0)],evf[nd+(nd1*nd+0)],evf[nd+(0*nd+nd1)],evf[nd+(nd1*nd+nd1)],
			basis,x_If,x_II,TMatr.getSpurMM(nd,bQQf,bQQf));
}

public static EVX2Data getEVX2DataFromQQ(final int n,final double [] QQi,final double [] QQt,final double [] QQf,final EVSolver EV)
{
    final int n1=n-1;
    final double [] basis=EV.getGEV_Generalized(n,QQi,QQt,true),
	bQQf=EVXData.convertMatrFromQtoB_Basis(n,QQf,basis);
    double x_If=0,x_II=0;
    for(int k=0;k<n;k++)
    {
	x_If+=basis[k]*bQQf[k*n+k];
	x_II+=Math.pow(basis[k],2);
    }
    final double [] evf=EV.getEV(n,bQQf,true);
    return new EVX2Data(n,basis[0],basis[n-1],evf[0],evf[n-1],
			evf[n+(0*n+0)],evf[n+(n1*n+0)],evf[n+(0*n+n1)],evf[n+(n1*n+n1)],
			basis,x_If,x_II,TMatr.getSpurMM(n,bQQf,bQQf));
}


} // END OF class EVX2Data
