package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Spectrum interpolation.
 *  @version $Id: EVXSpectrum.java,v 1.14 2022/07/28 11:26:35 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class EVXSpectrum {
    final double sL,sH;
    final int neq;
    final double [] ev;
    final double [] wpsi;
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;

    EVXSpectrum(final int neq,
		final double sL,final double sH,
		final double [] ev, final double [] wpsi,
		final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	this.neq=neq;
	this.sL=sL;
	this.sH=sH;
	this.ev=ev;
	this.wpsi=wpsi;
	this.M=M;
    }
public double [] getPsi(final int ind_lev)
{
    if(ind_lev<0||ind_lev>=neq)
	throw new RuntimeException("ind_lev="+ind_lev+" neq="+neq);
    final double [] psi=new double [neq];
    System.arraycopy(this.ev,neq+ind_lev*neq,psi,0,neq);
    return psi;
}
public double getE(final int ind_lev)
{
    if(ind_lev<0||ind_lev>=neq)
	throw new RuntimeException("ind_lev="+ind_lev+" neq="+neq);
    return this.ev[ind_lev];
}

private double getKSpectrInterpolation(final int k,final double lamx)
{
  final double ek=getE(k);
  double s=1;
  for(int l=0;l<neq;l++)
  {
      if(k!=l)
      {
	  final double el=getE(l);
	  s*=(lamx-el)/(ek-el);
      }
  }
  return s;
}

public double [] getPsiLamX(final double lamx)
{
    final int n=neq;
    final double [] psi=new double [n];
    for(int k=0;k<n;k++)
    {
	DeprecatedHelperTMatr.sumP1(psi,this.ev,n+k*n,n,getKSpectrInterpolation(k,lamx)/wpsi[k]);
    }
    //System.err.println("psi="+java.util.Arrays.toString(psi)+" wpsi="+java.util.Arrays.toString(wpsi));
    return psi;
}
public double [] getPsi2LamX(final double lamx)
{
    final int n=neq;
    final double [] psi=new double [n],psi2=new double[2*n-1];
    for(int k=0;k<n;k++)
    {
	System.arraycopy(this.ev,n+k*n,psi,0,n);
	final double [] ppsi=M.B.mult2Pol(psi,psi);
	//System.err.println("getKSpectrInterpolation("+k+","+lamx+")="+getKSpectrInterpolation(k,lamx)+" E(k)="+getE(k));
	DeprecatedHelperTMatr.sumP1(psi2,ppsi,0,ppsi.length,getKSpectrInterpolation(k,lamx));
    }
    return psi2;
}

public static EVXSpectrum getEVXSpectrumFromQQ(final int n,final double [] QQf,final double [] QQa,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final double [] ev=M.EV.getGEV_Generalized(n,QQf,QQa,true);
    final double [] one=new double []{1};
    one[0]/=M.B.getD0(one,0);
    final double [] wpsi=new double [n];
    for(int k=0;k<n;k++)
    {
	wpsi[k]=TMatr.getAMatrB(one.length,one,0,n,ev,n+k*n,QQa,n);
    }
    //System.err.println("Cwpsi="+java.util.Arrays.toString(wpsi)+" ev="+java.util.Arrays.toString(ev));
    return  new EVXSpectrum(n,ev[0],ev[n-1],ev,wpsi,M);
}


} // END OF class EVXSpectrum
