package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Time evolution as perturbation to |psi0>.
 *  John Pierpont Morgan is dead.
 *  @version $Id: EaComm.java,v 1.13 2022/06/27 10:50:02 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class EaComm {

final double I_0,I_L,I_H,Istar,vstar,f0,f1,f2;

EaComm(final int n,final double [] imom,final double [] tmom,final double [] fimom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");
    final double [] basis=getBasis(n,imom,tmom,M);

    final int i0=0,iL=1,iH=n-1;
    this.I_0=basis[i0];
    this.I_L=basis[iL];
    this.I_H=basis[iH];
    final double psi0x0;
    {
	final double x0=M.getRadauRoot();
	psi0x0=M.B.getD0(EVXData.getPsiFromBasis(i0,basis,n),x0);
    }

    final double [] QQi=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,imom),basis),
	QQfi=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,fimom),basis);

    this.f0=QQfi[i0+n*i0]/QQi[i0+n*i0];
    final double [] Fi=new double [n],Ii=new double [n];
    for(int k=0;k<n;k++){
	if(k==i0){Fi[k]=Ii[k]=0;}
	else {
	    Fi[k]=QQfi[k+n*i0];
	    Ii[k]=QQi[k+n*i0];
	}
    }
    this.f1=TMatr.mulVV(Fi,Ii)/TMatr.mulVV(Ii,Ii);
    this.f2=TMatr.getAMatrA(Ii,QQfi,n)/TMatr.getAMatrA(Ii,QQi,n);

    final double b=2*TMatr.mulVV(Ii,Ii);
    final double Ieq=Math.sqrt(b);
    this.Istar=Ieq-this.I_0;
    this.vstar=this.Istar/Math.pow(psi0x0,2);

    this.simom=imom;
    this.stmom=tmom;
    this.sfimom=fimom;
    this.M=M;
    this.sn=n;
}
final double [] simom,stmom,sfimom;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final int sn;


double [] getBasis(final int n,final double [] imom,final double [] tmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    return EdComm.getIRadauBasisPsi0(n,imom,tmom,M);
}


void test_Self(final double eps)
{
    final int n=sn;
    final double [] basis=getBasis(n,simom,stmom,M);
    final int i0=0;
    final double [] psi0=EVXData.getPsiFromBasis(i0,basis,n);

    final double [] QQt=M.B.getQQMatr(n,stmom),QQi=M.B.getQQMatr(n,simom),QQfi=M.B.getQQMatr(n,sfimom);

    final double f0=TMatr.getAMatrA(psi0,QQfi,n)/TMatr.getAMatrA(psi0,QQi,n);
    double sfi1=0,si1=0,sfi2=0,si2=0;
    for(int k=0;k<n;k++)
    {
	if(k==i0) continue;
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	sfi1+=TMatr.getAMatrB(psi0,psik,QQfi,n)*TMatr.getAMatrB(psi0,psik,QQi,n);
	si1+=TMatr.getAMatrB(psi0,psik,QQi,n)*TMatr.getAMatrB(psi0,psik,QQi,n);
	for(int m=0;m<n;m++)
	{
	    if(m==i0) continue;
	    final double [] psim=EVXData.getPsiFromBasis(m,basis,n);
	    sfi2+=TMatr.getAMatrB(psi0,psik,QQi,n)*TMatr.getAMatrB(psik,psim,QQfi,n)*TMatr.getAMatrB(psi0,psim,QQi,n);
	    si2+=TMatr.getAMatrB(psi0,psik,QQi,n)*TMatr.getAMatrB(psik,psim,QQi,n)*TMatr.getAMatrB(psi0,psim,QQi,n);
	}
    }
    final double f1=sfi1/si1,f2=sfi2/si2;

    UnitTests.checkDoubleRel(f0,"f0",this.f0,"this.f0",eps);
    UnitTests.checkDoubleRel(f1,"f1",this.f1,"this.f1",eps);
    UnitTests.checkDoubleRel(f2,"f2",this.f2,"this.f2",eps);

    final double b=2*si1,Istar=Math.sqrt(b)-TMatr.getAMatrA(psi0,QQi,n);
    UnitTests.checkDoubleRel(Istar,"Istar",this.Istar,"this.Istar",eps);
    System.err.println("Istar="+Istar+" b="+b);
    //if(Istar>0) throw new RuntimeException();

    System.err.println("EaComm:test_Self: test OK n="+n+" eps="+eps);
}




} // END OF class EaComm
