package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Time evolution as commutator acting on |psi>.
 *  Uses infinitesimal shift operators.
 *  @version $Id: EcComm.java,v 1.44 2022/11/19 11:08:57 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class EcComm {

public final double I_0,I_L,I_H,DsensL,DsensL_rho,DsensH,DsensH_rho,Dsens0,Dsens0_rho;

EcComm(final int n,final double [] imom,final double [] tmom,final double [] fmom,final double [] gmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");
    final double [] basis=getBasis(n,imom,tmom,M);

    final int i0=0,iL=1,iH=n-1;
    this.I_0=basis[i0];
    this.I_L=basis[iL];
    this.I_H=basis[iH];

    // QQt is diagonal in new basis.
    final double [] QQt=M.B.getQQMatr(n,tmom),
	QQg=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,gmom),basis),
	QQf=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,fmom),basis);

    final double [] psiL=new double [n],psiH=new double [n],psi0=new double [n];
    System.arraycopy(basis,n+iL*n,psiL,0,n);
    System.arraycopy(basis,n+iH*n,psiH,0,n);
    System.arraycopy(basis,n+i0*n,psi0,0,n);
    final double [] psiLB=new double [n],psiHB=new double [n],psi0B=new double [n];
    psiLB[iL]=1;psiHB[iH]=1;psi0B[i0]=1;
    final double [] DpsiLB=EVXData.convertPsiFromQtoB_Basis(n,getDPsi(psiL),basis,QQt),
	DpsiHB=EVXData.convertPsiFromQtoB_Basis(n,getDPsi(psiH),basis,QQt),
	Dpsi0B=EVXData.convertPsiFromQtoB_Basis(n,getDPsi(psi0),basis,QQt);
    TMatr.sumP1(DpsiLB,psiLB,-TMatr.mulVV(DpsiLB,psiLB));
    TMatr.sumP1(DpsiHB,psiHB,-TMatr.mulVV(DpsiHB,psiHB));
    TMatr.sumP1(Dpsi0B,psi0B,-TMatr.mulVV(Dpsi0B,psi0B));

    final double [] QQc=TMatr.mulMM(n,QQg,QQf);
    // Anti-Hermitian part (QQg QQf - QQf QQg) is responsible for time evolution.
    TMatr.scaleVectorInPlace(TMatr.selectAntiHermitianPartInPlace(n,QQc),2);

    final double [] varpsiLB=TMatr.getMatr_x_A(n,QQc,psiLB);
    this.DsensL=TMatr.mulVV(DpsiLB,varpsiLB);
    this.DsensL_rho=this.DsensL/Math.sqrt(TMatr.mulVV(DpsiLB,DpsiLB)*TMatr.mulVV(varpsiLB,varpsiLB));

    final double [] varpsiHB=TMatr.getMatr_x_A(n,QQc,psiHB);
    this.DsensH=TMatr.mulVV(DpsiHB,varpsiHB);
    this.DsensH_rho=this.DsensH/Math.sqrt(TMatr.mulVV(DpsiHB,DpsiHB)*TMatr.mulVV(varpsiHB,varpsiHB));

    final double [] varpsi0B=TMatr.getMatr_x_A(n,QQc,psi0B);
    this.Dsens0=TMatr.mulVV(Dpsi0B,varpsi0B);
    this.Dsens0_rho=this.Dsens0/Math.sqrt(TMatr.mulVV(Dpsi0B,Dpsi0B)*TMatr.mulVV(varpsi0B,varpsi0B));


    this.simom=imom;
    this.stmom=tmom;
    this.sfmom=fmom;
    this.sgmom=gmom;
    this.M=M;
    this.sn=n;
}
final double [] simom,stmom,sfmom,sgmom;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final int sn;

abstract double [] getDPsi(final double [] psi);

double [] getBasis(final int n,final double [] imom,final double [] tmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    return EdComm.getIRadauBasisPsi0(n,imom,tmom,M);
}

void test_Basis(final double eps)
{
    final int n=sn;
    final double [] basis=getBasis(n,simom,stmom,M),QQt=M.B.getQQMatr(n,stmom);
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	for(int l=0;l<n;l++)
	{
	    final double [] psil=EVXData.getPsiFromBasis(l,basis,n);
	    if(k!=l) UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQt,n),"<psik|QQt|psil>",0,"0",eps);
	    if(k!=l) UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,getDPsi(psil),QQt,n),"<psik|QQt|dpsil>",-TMatr.getAMatrB(getDPsi(psik),psil,QQt,n),"-<dpsik|QQt|psil>",eps);
	}
    }
    System.err.println("EcComm:test_Basis: test OK n="+n+" eps="+eps);
}

void test_Self(final double eps)
{
    test_Basis(eps);
    test_EComm(eps);

    final int n=sn;
    final double [] QQt=M.B.getQQMatr(n,stmom),QQg=M.B.getQQMatr(n,sgmom),QQf=M.B.getQQMatr(n,sfmom),IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt);
    final double [] QQC=TMatr.sumAcB(
	TMatr.mulMMM(n,QQg,IQQt,QQf),
	-1,
	TMatr.mulMMM(n,QQf,IQQt,QQg)),
	QQc=TMatr.mulMM(n,IQQt,QQC);

    final double [] basis=getBasis(n,simom,stmom,M),psi=new double [n];
    for(int in:new int []{0,1,n-1})
    {
	System.arraycopy(basis,n+in*n,psi,0,n);
	final double [] varpsi=TMatr.getMatr_x_A(n,QQc,psi);
	final double d=in==0?Dsens0:(in==1?DsensL:DsensH);

	UnitTests.checkDoubleRel(TMatr.getAMatrB(psi,varpsi,QQt,n)/I_L,"<psi|QQt|varpsi>",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(getDPsi(psi),varpsi,QQt,n),"<Dpsi|QQt|varpsi>",d,"d",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(getDPsi(psi),varpsi,QQt,n),"<D(psi)|varpsi>",-TMatr.getAMatrB(psi,getDPsi(varpsi),QQt,n),"-<psi|D(varpsi)>",eps);
    }

    System.err.println("EcComm:test_Self: test OK n="+n+" eps="+eps);
}

void test_EComm(final double eps)
{
    final int n=sn;
    final double [] basis=getBasis(n,simom,stmom,M),psi=new double [n];
    for(int in:new int []{0,1,n-1})
    {
	System.arraycopy(basis,n+in*n,psi,0,n);
	final EComm s=new EComm(sn,sgmom,stmom,sfmom,M){
	    double [] getDpsi(final double [] apsi){return EcComm.this.getDPsi(apsi);}
	    double [] getPsi0(final EVXData j){return psi;}
	};
	final double d=in==0?Dsens0:(in==1?DsensL:DsensH);
	final double d_rho=in==0?Dsens0_rho:(in==1?DsensL_rho:DsensH_rho);
	UnitTests.checkDoubleRel(s.Dsens,"s.Dsens",d,"d",eps);
	UnitTests.checkDoubleRel(s.Dsens_rho,"s.Dsens_rho",d_rho,"d_rho",eps);
    }
    System.err.println("EcComm:test_EComm: test OK n="+sn+" eps="+eps);
}

} // END OF class EcComm
