package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.SortArrays;
import com.polytechnik.utils.IntPTRCompare;
import com.polytechnik.utils.Quadrature;

/** Time evolution as commutator acting on |psi0>.
 *  Uses finite variation in contrast with EComm which uses infinitesimal shits.
 *  @version $Id: EdComm.java,v 1.109 2022/11/20 12:41:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class EdComm {
final double [] aPdir,axPdir,Pdir,Pdir_rho,Px,Pdir_Rrho,di,df,di_Rrho,df_Rrho;
public final double x0,dir;

EdComm(final int n,final double [] imom,final double [] tmom,final double [] fmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final boolean isEdgeAtRight)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");
    final double [] basis=getBasis(n,tmom,M,isEdgeAtRight);
    // QQt is diagonal in new basis.
    final double [] QQi=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,imom),basis),
	QQf=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,fmom),basis);

    final double [] QQc=TMatr.mulMM(n,QQi,QQf);
    // Anti-Hermitian part (QQi QQf - QQf QQi) is responsible for time evolution.
    TMatr.scaleVectorInPlace(TMatr.selectAntiHermitianPartInPlace(n,QQc),2);

    this.aPdir=new double [n];
    this.axPdir=new double [n];
    this.Pdir=new double [n];
    this.Pdir_rho=new double [n];
    this.Pdir_Rrho=new double [n];
    this.Px=new double [n];
    this.di=new double [n];
    this.di_Rrho=new double [n];
    this.df=new double [n];
    this.df_Rrho=new double [n];

    final int ind_edge=0;
    final double nrz=DeprecatedHelperTMatr.mulVVo(n,QQc,ind_edge*n,QQc,ind_edge*n);
    for(int k=0;k<n;k++)
    {
	this.Px[k]=basis[k];
	this.Pdir[k]=QQc[k*n+ind_edge];
	double sa=0;for(int l=0;l<n;l++) sa+=QQc[k*n+l]*(l>=k?-1:1);
	this.aPdir[k]=sa;
	sa=0;for(int l=0;l<n;l++) if(l!=k) sa+=QQc[k*n+l]/(basis[k]-basis[l]);
	this.axPdir[k]=sa*(isEdgeAtRight?1:-1);
	this.Pdir_rho[k]=this.Pdir[k]/Math.sqrt(nrz);
	final double nrzR=DeprecatedHelperTMatr.mulVVo(n,QQc,k*n,QQc,k*n);
	this.Pdir_Rrho[k]=this.Pdir[k]/Math.sqrt(nrzR);
	this.di[k]=QQi[k*n+ind_edge];
	this.di_Rrho[k]=this.di[k]/Math.sqrt(DeprecatedHelperTMatr.mulVVo(n,QQi,k*n,QQi,k*n)-Math.pow(QQi[k*n+k],2));
	this.df[k]=QQf[k*n+ind_edge];
	this.df_Rrho[k]=this.df[k]/Math.sqrt(DeprecatedHelperTMatr.mulVVo(n,QQf,k*n,QQf,k*n)-Math.pow(QQf[k*n+k],2));
    }
    this.dir=n>1?this.Pdir_Rrho[1]:Double.NaN;
    this.x0=this.Px[0];


    this.simom=imom;
    this.stmom=tmom;
    this.sfmom=fmom;
    this.sM=M;
    this.sisEdgeAtRight=isEdgeAtRight;
}
final double [] simom,stmom,sfmom;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> sM;
final boolean sisEdgeAtRight;

double [] getBasis(final int n,final double [] tmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final boolean isEdgeAtRight)
{
    return getRadauBasis(n,tmom,M,isEdgeAtRight);
}


/** Utility function to get regular Radau basis.
 */
static double [] getRadauBasis(final int n,final double [] tmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final boolean isEdgeAtRight)
{
    final double x0=M.getRadauRoot();

    final int nR=n-1;
    final double [] Rtmom=M.getRadauMomentsFromMoments(tmom),XRtmom=M.B.getXQkMomentsFromQkMoments(Rtmom);
    final double [] Rev=M.EV.getGEV_Generalized(nR,M.B.getQQMatr(nR,XRtmom),M.B.getQQMatr(nR,Rtmom),true);
    final int [] ind=SortArrays.sortArrayIndexes(nR,isEdgeAtRight?IntPTRCompare.sortDescending(Rev):IntPTRCompare.sortAscending(Rev));
    final double [] basis=new double [n+n*n],psiR=new double [nR];
    for(int k=-1;k<ind.length;k++)
    {
	final double x;
	final double [] psi;
	if(k<0)
	{
	    final int j0=ind[0];
	    System.arraycopy(Rev,nR+j0*nR,psiR,0,nR);
	    psi=M.getxP(psiR);
	    DeprecatedHelperTMatr.sumP1(psi,psiR,nR,-Rev[j0]);
	    x=x0;
	}
	else
	{
	    final int j0=ind[k];
	    System.arraycopy(Rev,nR+j0*nR,psiR,0,nR);
	    psi=M.getxP(psiR);
	    DeprecatedHelperTMatr.sumP1(psi,psiR,nR,-x0);
	    x=Rev[j0];
	}
	final int k1=k+1;
	basis[k1]=x;
	TMatr.scaleVectorInPlace(psi,1/M.B.getD0(psi,x));
	TMatr.scaleVectorInPlace(psi,1/Math.sqrt(TMatr.mulVVn1(psi,tmom)));
	System.arraycopy(psi,0,basis,n+k1*n,n);
    }
    return basis;
}

void test_Radau(final double eps)
{
    final int n=Px.length;
    final double [] basis=getRadauBasis(n,stmom,sM,sisEdgeAtRight);
    final Quadrature Q=sM.getRadauQuadratureFromMoments(n-1,stmom);
    final double [] x=Q.x,w=Q.w;
    if(sisEdgeAtRight) SortArrays.sortArrayWithLoadDesc(x.length,x,w);
    else SortArrays.sortArrayWithLoad(x.length,x,w);
    UnitTests.checkDoubleArrayRel(x,"x",Px,"Px",eps);
    final double [] psi=new double [n],pp=new double [n],QQt=sM.B.getQQMatr(n,stmom);
    for(int k=0;k<n;k++)
    {
	System.arraycopy(basis,n+k*n,psi,0,n);
	final double wk=1/Math.pow(sM.B.getD0(psi,x[k]),2);
	UnitTests.checkDoubleRel(wk,"wk",w[k],"w[k]",eps);
	for(int l=0;l<n;l++)
	{
	    System.arraycopy(basis,n+l*n,pp,0,n);
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(psi,pp,QQt,n),"<psik|psil>",k==l?1:0,k==l?"1":"0",eps);
	    if(k!=l) UnitTests.checkDoubleRel(sM.B.getD0(pp,Px[k]),"psil(xk)",0,"0",eps);
	}
	if(k>0 && !(sisEdgeAtRight?Px[k-1]>Px[k]:Px[k-1]<Px[k]))
	{
	    throw new RuntimeException("not sorted, basis="+java.util.Arrays.toString(basis));
	}
	if(sM.B.getD0(psi,Px[k])<=0)
	{
	    throw new RuntimeException("k="+k+" psi("+x[k]+")="+sM.B.getD0(psi,x[k])+"<0");
	}
    }
    System.err.println("EdComm:test_Radau: x="+java.util.Arrays.toString(Px));

    System.err.println("EdComm:test_Radau: test OK n="+n+" eps="+eps);
}

/** Utility function to get I*Radau basis.
 *  First n elements are I_k, then psi0,psiL,...psiH follows.
 */
static double [] getIRadauBasisPsi0(final int n,final double [] imom,final double [] tmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final double x0=M.getRadauRoot();

    final int nR=n-1;
    double [] Rtmom=M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(tmom)),Rimom=M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(imom));
    final double [] Rev=M.EV.getGEV_Generalized(nR,M.B.getQQMatr(nR,Rimom),M.B.getQQMatr(nR,Rtmom),true);
    final double [] basis=new double [n+n*n],psiR=new double [nR];
    final double [] QQt=M.B.getQQMatr(n,tmom);
    for(int k=nR;--k>=0;)
    {
	final int j0=k;
	System.arraycopy(Rev,nR+j0*nR,psiR,0,nR);
	final double [] psi=M.getxP(psiR);
	DeprecatedHelperTMatr.sumP1(psi,psiR,nR,-x0);
	final int k1=k+1;
	basis[k1]=Rev[j0];
	final double w=TMatr.mulVVn1(psi,tmom);
	if(w<0) TMatr.scaleVectorInPlace(psi,-1);
	System.arraycopy(psi,0,basis,n+k1*n,n);
	if(false)
	{
	    final double [] rpsi=M.getRadauMultiplied(psi);
	    System.err.println("basis["+k1+"]="+basis[k1]+" s1R/w1R="+(TMatr.mulVVn1(psiR,Rimom)/TMatr.mulVVn1(psiR,Rtmom))+
			       " match "+(TMatr.mulVVn1(rpsi,imom)/TMatr.mulVVn1(rpsi,tmom))+
			       " and not necessary match to s1/w1="+(TMatr.mulVVn1(psi,imom)/TMatr.mulVVn1(psi,tmom)));
	}
    }
    final double [] psi0=M.B.getX0Moments(n,x0);
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQt,n,n,ipvt);
    if(pinfo!=0){java.util.Arrays.fill(psi0,Double.NaN);}
    else {com.polytechnik.utils.Linsystems.dgesl_m1(QQt,n,n,ipvt,psi0,1);}
    final double o_o=M.B.getD0(psi0,x0),w=TMatr.mulVVn1(psi0,tmom);
    TMatr.scaleVectorInPlace(psi0,(w<0?-1:1)/Math.sqrt(o_o));
    System.arraycopy(psi0,0,basis,n+0,n);
    final double [] psi02=M.B.mult2Pol(psi0,psi0);
    //System.err.println("n="+n+" imom.length="+imom.length+" psi02.length="+psi02.length+" psi0.length="+psi0.length);
    basis[0]=TMatr.mulVVn1(psi02,imom);
    return basis;
}

static double [] getIRadauBasisPsi0IOrthogonalFromQQ(final int n,final double [] QQi,final double [] QQt,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M);
    final double [] psi0=EdComm.getPsi0QQtOrthogonalBasisAtX0(n,QQi,n-1,basis,n+n,true,M);
    final double I0=1/TMatr.getAMatrA(psi0,QQt,n);
    TMatr.scaleVectorInPlace(psi0,Math.sqrt(I0));
    basis[0]=I0;
    System.arraycopy(psi0,0,basis,n,n);
    return basis;
}

static double [] getIRadauBasisPsi0FromQQ(final int n,final double [] QQi,final double [] QQt,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final double x0=M.getRadauRoot();

    final int nR=n-1;
    final double [] psiR=new double [nR],basis=new double [n+n*n],trans=basis;
    for(int k=0;k<nR;k++)
    {
	psiR[k]=1;
	final double [] psi=M.getxP(psiR);
	DeprecatedHelperTMatr.sumP1(psi,psiR,nR,-x0);
	System.arraycopy(psi,0,trans,k*n,n);
	psiR[k]=0;
    }
    final double [] RQQi=EVXData.convertMatrFromQtoB_Basis(n,QQi,trans,nR,0),
	RQQt=EVXData.convertMatrFromQtoB_Basis(n,QQt,trans,nR,0);

    final double [] Rev=M.EV.getGEV_Generalized(nR,RQQi,RQQt,true),bQ=new double [n*nR];
    DeprecatedHelperTMatr.multiply_MA_x_MB(bQ,nR,Rev,nR,nR+0,nR,trans,n,n,0);
    System.arraycopy(bQ,0,basis,n+1*n,n*nR); // Copy eigenvectors, all except 0-th are now set.
    System.arraycopy(Rev,0,basis,1,nR); // Copy eignevalues, all except 0-th are now set.

    // Calculate |psi0> and <psi0|I|psi0>.
    final double [] psi0;
    if(true)
    {
	psi0=getPsi0QQtOrthogonalBasisAtX0(n,QQt,nR,basis,n+n,false,M);
    }
    else
    {
	psi0=M.B.getX0Moments(n,x0);
	final double [] tmpQQt=new double [n*n];final int [] ipvt=new int [n];
	System.arraycopy(QQt,0,tmpQQt,0,n*n);
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(tmpQQt,n,n,ipvt);
	if(pinfo!=0){java.util.Arrays.fill(psi0,Double.NaN);}
	else {com.polytechnik.utils.Linsystems.dgesl_m1(tmpQQt,n,n,ipvt,psi0,1);}
	final double o_o=M.B.getD0(psi0,x0);
	TMatr.scaleVectorInPlace(psi0,1/Math.sqrt(o_o));
    }
    System.arraycopy(psi0,0,basis,n+0,n);
    basis[0]=TMatr.getAMatrA(psi0,QQi,n);
    return basis;
}

static double [] getPsi0QQtOrthogonalBasisAtX0(final int n,final double [] QQt,final int nInBasis,final double [] basis,final int basis_offset,final boolean flag_normalize_basis,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
  final double x0=M.getRadauRoot();
  final double [] psi0=M.B.getX0Moments(n,x0);
  for(int k=0;k<n;k++){psi0[k]/=Math.sqrt(QQt[k+n*k]);}
  for(int k=0;k<nInBasis;k++){DeprecatedHelperTMatr.sumP1(psi0,basis,basis_offset+k*n,n,-TMatr.getAMatrB(n,psi0,0,basis,basis_offset+k*n,QQt,n)/(flag_normalize_basis?TMatr.getAMatrA(n,basis,basis_offset+k*n,QQt,n):1));}
  TMatr.scaleVectorInPlace(psi0,1/Math.sqrt(TMatr.getAMatrA(psi0,QQt,n)));
  if(M.B.getD0(psi0,x0)<0) TMatr.scaleVectorInPlace(psi0,-1);
  return psi0;
}


void test_IRadauQQ(final double eps)
{
    final int n=Px.length,nd=n-1;
    final double [] basis=getIRadauBasisPsi0(n,simom,stmom,sM),
	QQi=sM.B.getQQMatr(n,simom),QQt=sM.B.getQQMatr(n,stmom),
	basisQQ=getIRadauBasisPsi0FromQQ(n,QQi,QQt,sM);
    DeprecatedHelperEVXData.test_orthogonalitySKIP0I("basis",n,basis,QQi,QQt,eps);
    DeprecatedHelperEVXData.test_orthogonalitySKIP0I("basisQQ",n,basisQQ,QQi,QQt,eps);
    for(int k=0;k<n;k++)
    {
	UnitTests.checkDoubleRel(basis[k],"basis[k]",basisQQ[k],"basisQQ[k]",eps);
	final double [] psi=EVXData.getPsiFromBasis(k,basis,n),psiQQ=EVXData.getPsiFromBasis(k,basisQQ,n);
	UnitTests.checkDoubleArrayRel1(EVXData.getMSign(psi),"psi",EVXData.getMSign(psiQQ),"psiQQ",eps);
    }

    final double x0=sM.getRadauRoot();
    final double [][] RRQQ=BQVM.getQQ0RemovedA(n,new double [][]{QQt,QQi},sM,x0);
    final double [] ev=sM.EV.getGEV_Generalized(nd,RRQQ[1],RRQQ[0],true);
    EVXData.test_orthogonality("getQQ0RemovedA:ev",nd,ev,RRQQ[1],RRQQ[0],eps);
    for(int k=0;k<nd;k++)
    {
	UnitTests.checkDoubleRel(basisQQ[k+1],"basisQQ[k+1]",ev[k],"ev[k]",eps);
    }

    System.err.println("EdComm:test_IRadauQQ: test OK n="+n+" eps="+eps);
}

void test_IRadau(final double eps)
{
    final int n=Px.length;
    final double x0=sM.getRadauRoot();
    final double [] basis=getIRadauBasisPsi0(n,simom,stmom,sM);
    System.err.println("test_IRadau spectrum="+java.util.Arrays.toString(java.util.Arrays.copyOf(basis,n)));
    final double [] psi=new double [n],pp=new double [n],QQt=sM.B.getQQMatr(n,stmom),QQi=sM.B.getQQMatr(n,simom);
    System.err.println("newx0moms="+java.util.Arrays.toString(sM.B.getX0Moments(simom.length,x0)));
//    System.err.println("newRmomi="+java.util.Arrays.toString(TMatr.sumAcB(simom,Math.abs(basis[n-1])*1e6*Math.abs(stmom[0]),sM.B.getX0Moments(simom.length,x0))));
//    System.err.println("newRQQi="+java.util.Arrays.toString(sM.B.getQQMatr(n,TMatr.sumAcB(simom,Math.abs(basis[n-1])*1e6*Math.abs(stmom[0]),sM.B.getX0Moments(simom.length,x0)))));
//    System.err.println("newRQQt="+java.util.Arrays.toString(QQt));
    final EVXData a=EVXData.getEVXDataFromQQ(n,sM.B.getQQMatr(n,TMatr.sumAcB(simom,Math.abs(basis[n-1])*1e6*Math.abs(stmom[0]),sM.B.getX0Moments(simom.length,x0))),QQt,sM);
    System.err.println("test_IRadau R spectrum="+java.util.Arrays.toString(a.getEValues()));
    for(int k=0;k<n;k++)
    {
	System.arraycopy(basis,n+k*n,psi,0,n);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(psi,psi,QQi,n),"<psi|QQi|psi>",basis[k],"basis[k]",eps);
	for(int l=0;l<n;l++)
	{
	    System.arraycopy(basis,n+l*n,pp,0,n);
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(psi,pp,QQt,n),"<psik|psil>",k==l?1:0,k==l?"1":"0",eps);
	}

	final double w1=TMatr.mulVVn1(psi,stmom);
	if(w1<=0)
	{
	    throw new RuntimeException("k="+k+" w1="+w1+"<0");
	}
	if(k!=0) // Match a: the limit case when contrib to I at x0 is large.
	{
	    UnitTests.checkDoubleRel(a.ev[k-1],"a.ev["+(k-1)+"]",basis[k],"basis[k]",eps);
	    final double [] psia=a.getPsi(k-1);
	    TMatr.scaleVectorInPlace(psia,TMatr.mulVVn1(psia,stmom)<0?-1:1);
//	    System.err.println("<psia|QQt|psia>="+TMatr.getAMatrA(psia,QQt,n));
//	    System.err.println("<psi|QQt|psi>="+TMatr.getAMatrA(psi,QQt,n));
	    UnitTests.checkDoubleArrayRel1(psi,"psi",psia,"psia",eps);
	}
	else
	{
	    final double [] psia=a.getPsi(n-1);
	    TMatr.scaleVectorInPlace(psia,TMatr.mulVVn1(psia,stmom)<0?-1:1);
//	    System.err.println("<psia|QQt|psia>="+TMatr.getAMatrA(psia,QQt,n));
//	    System.err.println("<psi|QQt|psi>="+TMatr.getAMatrA(psi,QQt,n));
	    UnitTests.checkDoubleArrayRel1(psi,"psi",psia,"psia",eps);
	    UnitTests.checkDoubleRel(basis[0],"basis[0]",TMatr.getAMatrA(psia,QQi,n),"<psia|QQi|psia>",eps);

	}
    }
    System.err.println("EdComm:test_IRadau: test OK n="+n+" eps="+eps);
}

void test_IRadau_MkQVM_X0Psi_Psix0(final double[][] MkQVM_X0Psi_Psix0,final double [] MkQVM_X0Psi_psi0,final double eps)
{
    final int n=Px.length;
    if(n-1!=MkQVM_X0Psi_Psix0.length) throw new IllegalArgumentException("n="+n+" MkQVM_X0Psi_Psix0.length="+MkQVM_X0Psi_Psix0.length);
    final double [] basis=getIRadauBasisPsi0(n,simom,stmom,sM);
    final double [] psi=new double [n];
    for(int k=0;k<MkQVM_X0Psi_Psix0.length;k++)
    {
	System.arraycopy(basis,n+(k+1)*n,psi,0,n);
	UnitTests.checkDoubleArrayRel(EVXData.getMSign(psi),"psi",EVXData.getMSign(MkQVM_X0Psi_Psix0[k]),"MkQVM_X0Psi_Psix0[k]",eps);
    }
    System.arraycopy(basis,n+0,psi,0,n);
    UnitTests.checkDoubleArrayRel(EVXData.getMSign(psi),"psi",EVXData.getMSign(MkQVM_X0Psi_psi0),"MkQVM_X0Psi_psi0",eps);
    System.err.println("EdComm:IRadau_MkQVM_X0Psi_Psix0: test OK n="+n+" eps="+eps);
}

void test_ACorr(final double eps)
{
    final int n=Px.length;
    for(int k=1;k<n;k++)
    {
	final ACorr a=ACorr.getCommutatorPsiX0PsiX1(n,simom,sfmom,stmom,sM,Px[0],Px[k]);
	UnitTests.checkDoubleRel(a.x_a_b,"a.x_a_b",Pdir[k],"Pdir[k]",eps);
	System.err.println("test_ACorr:x="+Px[k]+" a.x_a_b="+a.x_a_b+" Pdir[k]="+Pdir[k]);
	UnitTests.checkDoubleRel(a.osp,"a.osp",Pdir_rho[k],"Pdir_rho[k]",eps);
	final ACorr b=ACorr.getCommutatorPsiX0PsiX1(n,simom,sfmom,stmom,sM,Px[k],Px[0]);
	UnitTests.checkDoubleRel(-b.x_a_b,"-b.x_a_b",Pdir[k],"Pdir[k]",eps);
	UnitTests.checkDoubleRel(-b.osp,"-b.osp",Pdir_Rrho[k],"Pdir_Rrho[k]",eps);
    }
    final int ind_x=1;
    double s2=0,sd=0;
    for(int k=0;k<n;k++)
    {
	final ACorr a=ACorr.getCommutatorPsiX0PsiX1(n,simom,sfmom,stmom,sM,Px[k],Px[ind_x]);
	if(k==0) sd=a.x_a_b;
	s2+=Math.pow(a.x_a_b,2);
    }
    UnitTests.checkDoubleRel(this.dir,"this.dir",sd/Math.sqrt(s2),"sd/sqrt(s2)",eps);

    System.err.println("EdComm:test_ACorr: test OK n="+n+" eps="+eps);
}

void test_I(final double eps)
{
    final int n=Px.length;
    final double [] basis=getBasis(n,stmom,sM,sisEdgeAtRight);
    final double [] QQi=sM.B.getQQMatr(n,simom),QQf=sM.B.getQQMatr(n,sfmom),psi=new double [n];
    final int ind_edge=0;
    for(int k=0;k<n;k++)
    {
	UnitTests.checkDoubleRel(TMatr.getAMatrB(n,basis,n+k*n,basis,n+ind_edge*n,QQi,n),"<psi0|i|psik>",di[k],"di[k]",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(n,basis,n+k*n,basis,n+ind_edge*n,QQf,n),"<psi0|f|psik>",df[k],"df[k]",eps);
    }

    System.err.println("EdComm:test_I: test OK n="+n+" eps="+eps);
}

void test_Self(final double eps)
{
    final int n=Px.length;
    test_Radau(eps);
    test_IRadau(eps);
    test_IRadauQQ(eps);
    test_ACorr(eps);
    test_I(eps);
    System.err.println("EdComm:test_Common: test OK n="+n+" eps="+eps);
}


} // END OF class EdComm
