package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Time evolution as perturbation to |psi0>.
 *  @version $Id: EfiComm.java,v 1.33 2022/06/27 13:32:45 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class EfiComm {

public final double I_0,I_L,I_H,f0,fL,fH,df0,dfL,dfH,fi0,fiL,fiH;
final double dC0,dC0_rho,dCf0,dCf0_rho,fv,fvn;

final TSTwoLevels L0,H0,DL,DH;

EfiComm(final int n,final double [] imom,final double [] tmom,final double [] fmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");
    final double [] basis=getBasis(n,imom,tmom,M);

    final int i0=0,iL=1,iH=n-1;
    this.I_0=basis[i0];
    this.I_L=basis[iL];
    this.I_H=basis[iH];

    // QQt is diagonal in new basis.
    final double [] QQt=M.B.getQQMatr(n,tmom),
	QQi=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,imom),basis),
	QQf=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,fmom),basis);
    final double [] D=DeprecatedHelperEVXData.convertOperatorFromQtoB_Basis(n,getDMatr(n),QQt,basis,n);
    final double [] DD=D.clone();
    //System.err.println("D="+java.util.Arrays.toString(D));
     // Anti-Hermitian part is responsible for time evolution.
    for(int k=0;k<n;k++) {
	QQi[k*n+k]=D[k*n+k]=0; // can do because only psi0(x0)!=0
	for(int l=0;l<k;l++) QQi[k*n+l]=-QQi[l*n+k]; // first order perturbation
    }

    final double [] psiLB=new double [n],psiHB=new double [n],psi0B=new double [n];
    psiLB[iL]=1;psiHB[iH]=1;psi0B[i0]=1;
    final double [] Dpsi0B=TMatr.getMatr_x_A(n,D,psi0B),DpsiLB=TMatr.getMatr_x_A(n,D,psiLB),DpsiHB=TMatr.getMatr_x_A(n,D,psiHB);

    /*
    final double [] psi=new double [n];
    System.arraycopy(basis,n+i0*n,psi,0,n);
    final double [] Dpsi0B=EVXData.convertPsiFromQtoB_Basis(n,getDPsi(psi),QQt,basis);
    System.arraycopy(basis,n+iL*n,psi,0,n);
    final double [] DpsiLB=EVXData.convertPsiFromQtoB_Basis(n,getDPsi(psi),QQt,basis);
    System.arraycopy(basis,n+iH*n,psi,0,n);
    final double [] DpsiHB=EVXData.convertPsiFromQtoB_Basis(n,getDPsi(psi),QQt,basis);
    */

    double s0=0,sL=0,sH=0;
    for(int k=0;k<n;k++)
    {
	if(k!=i0){s0+=QQf[i0*n+k]*Dpsi0B[k];}
	if(k!=iL){sL+=QQf[iL*n+k]*DpsiLB[k];}
	if(k!=iH){sH+=QQf[iH*n+k]*DpsiHB[k];}
    }
    this.f0=QQf[i0*n+i0];
    this.fL=QQf[iL*n+iL];
    this.fH=QQf[iH*n+iH];
    this.df0=-2*TMatr.getAMatrB(psi0B,TMatr.getMatr_x_A(n,DD,psi0B),QQf,n);
    this.dfL=-2*TMatr.getAMatrB(psiLB,TMatr.getMatr_x_A(n,DD,psiLB),QQf,n);
    this.dfH=-2*TMatr.getAMatrB(psiHB,TMatr.getMatr_x_A(n,DD,psiHB),QQf,n);
    this.fi0=s0;
    this.fiL=sL;
    this.fiH=sH;

    final double [] QQc=TMatr.mulMatrices(n,new double [][] {QQi,D}),QQfc=TMatr.mulMatrices(n,new double [][] {QQi,QQf,D}),QQfD=TMatr.mulMatrices(n,new double [][] {QQf,D});
    this.dC0=QQc[i0*n+i0];
    this.dCf0=QQfc[i0*n+i0];
    double si=0,sd=0,sdf=0;
    for(int k=0;k<n;k++)
    {
	if(k!=i0){si+=Math.pow(QQi[k*n+i0],2);sd+=Math.pow(D[k*n+i0],2);sdf+=Math.pow(QQfD[k*n+i0],2);}
    }
    this.dC0_rho=this.dC0/Math.sqrt(si*sd);
    this.dCf0_rho=this.dCf0/Math.sqrt(si*sdf);
    final double [] varpsi0B=TMatr.getMatr_x_A(n,QQi,psi0B);
    this.fv=TMatr.getAMatrA(varpsi0B,QQf,n);
    this.fvn=this.fv/TMatr.mulVV(varpsi0B,varpsi0B);


    final double [] psi0=EVXData.getPsiFromBasis(i0,basis,n),
	psiL=EVXData.getPsiFromBasis(iL,basis,n),
	psiH=EVXData.getPsiFromBasis(iH,basis,n);

    L0=TSTwoLevels.test_getTSTwoLevels(imom,tmom,fmom,M,psiL,psi0);
    H0=TSTwoLevels.test_getTSTwoLevels(imom,tmom,fmom,M,psiH,psi0);
    DL=TSTwoLevels.test_getTSTwoLevels(imom,tmom,fmom,M,getDPsi(psiL),psiL);
    DH=TSTwoLevels.test_getTSTwoLevels(imom,tmom,fmom,M,getDPsi(psiH),psiH);

    this.simom=imom;
    this.stmom=tmom;
    this.sfmom=fmom;
    this.M=M;
    this.sn=n;
}
final double [] simom,stmom,sfmom;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final int sn;

abstract double [] getDPsi(final double [] psi);

/** Calculate d/dt matrix elements.
 */
double [] getDMatr(final int n)
{
    final double [] D=new double [n*n],p=new double [n];

    for(int k=0;k<n;k++)
    {
	p[k]=1;
	final double [] d=getDPsi(p);
	if(d.length!=n) throw new RuntimeException();
	p[k]=0;
	for(int j=d.length;--j>=0;)
	{
	    D[j*n+k]=d[j];
	}
    }
    return D;
}


double [] getBasis(final int n,final double [] imom,final double [] tmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    return EdComm.getIRadauBasisPsi0(n,imom,tmom,M);
}

/** Test dCf0 without getDPsi().
 */
void test_dC_rho(final double eps)
{
    final int n=sn;
    final double x0=M.getRadauRoot();
    final double [] basis=getBasis(n,simom,stmom,M);
    final double [] QQi=M.B.getQQMatr(n,simom),QQt=M.B.getQQMatr(n,stmom),QQf=M.B.getQQMatr(n,sfmom);
    final int i0=0,iL=1,iH=n-1;
    final double [] psi0=EVXData.getPsiFromBasis(i0,basis,n),Dpsi0=new double [n];
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	TMatr.sumP1(Dpsi0,psik,M.B.getD1(psik,x0));
    }
    final double sp=M.B.getD1(psi0,x0)>0?-1:1;
    TMatr.sumP1(Dpsi0,psi0,-TMatr.getAMatrB(psi0,Dpsi0,QQt,n));

    double sfi=0,sf2=0,si2=0,sid=0,sd2=0;
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	final double i=TMatr.getAMatrB(psi0,psik,QQi,n),f=TMatr.getAMatrB(psik,Dpsi0,QQf,n),d=TMatr.getAMatrB(psik,Dpsi0,QQt,n);
	if(k!=i0){
	    sfi+=i*f;
	    sid+=i*d;
	    sf2+=f*f;
	    si2+=i*i;
	    sd2+=d*d;
	}
    }
    final double dcf_rho=sp*sfi/Math.sqrt(si2*sf2);
    UnitTests.checkDoubleRel(dcf_rho,"dcf_rho",dCf0_rho,"dCf0_rho",eps);
    final double dc_rho=sp*sid/Math.sqrt(si2*sd2);
    UnitTests.checkDoubleRel(dc_rho,"dc_rho",dC0_rho,"dC0_rho",eps);
    System.err.println("dc_rho="+dc_rho+" dcf_rho="+dcf_rho);

    System.err.println("EfiComm:test_dC_rho: test OK n="+n+" eps="+eps);
}

void test_dC(final double eps)
{
    final int n=sn;
    final double [] basis=getBasis(n,simom,stmom,M);
    final double [] QQi=M.B.getQQMatr(n,simom),QQt=M.B.getQQMatr(n,stmom),QQf=M.B.getQQMatr(n,sfmom);
    final int i0=0,iL=1,iH=n-1;
    final double [] psi0=EVXData.getPsiFromBasis(i0,basis,n),Dpsi0=getDPsi(psi0),varpsi0=new double [n];
    TMatr.sumP1(Dpsi0,psi0,-TMatr.getAMatrB(psi0,Dpsi0,QQt,n));
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	if(k!=i0){
	    TMatr.sumP1(varpsi0,psik,TMatr.getAMatrB(psi0,psik,QQi,n));
	}
    }
    final double dc=TMatr.getAMatrB(Dpsi0,varpsi0,QQt,n),dcf=TMatr.getAMatrB(Dpsi0,varpsi0,QQf,n);
    UnitTests.checkDoubleRel(dc,"dc",dC0,"dC0",eps);
    UnitTests.checkDoubleRel(dcf,"dcf",dCf0,"dCf0",eps);
    UnitTests.checkDoubleRel(fv,"fv",TMatr.getAMatrA(varpsi0,QQf,n),"<varpsi0|QQf|varpsi0>",eps);
    //System.err.println("TMatr.getAMatrA(varpsi0,QQt,n)="+TMatr.getAMatrA(varpsi0,QQt,n));
    UnitTests.checkDoubleRel(fvn,"fvn",TMatr.getAMatrA(varpsi0,QQf,n)/TMatr.getAMatrA(varpsi0,QQt,n),"<varpsi0|QQf|varpsi0>/<varpsi0|QQt|varpsi0>",eps);
    System.err.println("EfiComm:test_dC: test OK n="+n+" eps="+eps);
}

void test_Self(final double eps)
{
    test_dC(eps);
    test_dC_rho(eps);
    final int n=sn;
    final double [] basis=getBasis(n,simom,stmom,M);
    final double [] QQi=M.B.getQQMatr(n,simom),QQt=M.B.getQQMatr(n,stmom),QQf=M.B.getQQMatr(n,sfmom),D=getDMatr(n);
    final int i0=0,iL=1,iH=n-1;
    final double [] psi0=EVXData.getPsiFromBasis(i0,basis,n),psiL=EVXData.getPsiFromBasis(iL,basis,n),
	psiH=EVXData.getPsiFromBasis(iH,basis,n);
    double s0=0,sL=0,sH=0;
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	if(k!=i0) s0+=TMatr.getAMatrB(psi0,psik,QQf,n)*TMatr.getAMatrB(psi0,getDPsi(psik),QQt,n);
	if(k!=iL) sL+=TMatr.getAMatrB(psiL,psik,QQf,n)*TMatr.getAMatrB(psiL,getDPsi(psik),QQt,n);
	if(k!=iH) sH+=TMatr.getAMatrB(psiH,psik,QQf,n)*TMatr.getAMatrB(psiH,getDPsi(psik),QQt,n);
	UnitTests.checkDoubleArrayRel(getDPsi(psik),"getDPsi(psik)",TMatr.getMatr_x_A(n,D,psik),"D psik",eps);
	for(int l=0;l<n;l++)
	{
	    final double [] psil=EVXData.getPsiFromBasis(l,basis,n);
	    if(k!=l)
	    {
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,getDPsi(psil),QQt,n),"<psik|Dpsil>",-TMatr.getAMatrB(psil,getDPsi(psik),QQt,n),"-<psil|Dpsik>",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQt,n),"<psik|psil>",0,"0",eps);
		if(k!=i0 && l!=i0) UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQi,n),"<psik|psil>",0,"0",eps);
		//System.err.println("<psi"+k+"|QQi|psi"+l+">="+TMatr.getAMatrB(psik,psil,QQi,n));
	    }
	}
	UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psik,QQt,n),"<psik|psik>",1,"1",eps);
    }
    UnitTests.checkDoubleRel(fi0,"fi0",-s0,"-s0",eps);
    UnitTests.checkDoubleRel(fiL,"fiL",-sL,"-sL",eps);
    UnitTests.checkDoubleRel(fiH,"fiH",-sH,"-sH",eps);

    UnitTests.checkDoubleRel(TMatr.getAMatrB(psiL,getDPsi(psiL),QQt,n),"<psiL|DpsiL>",0,"0",eps);
    UnitTests.checkDoubleRel(TMatr.getAMatrB(psiH,getDPsi(psiH),QQt,n),"<psiH|DpsiH>",0,"0",eps);

    L0.match_Values(new EjComm(n,simom,stmom,sfmom,M).tL0,eps);
    H0.match_Values(new EjComm(n,simom,stmom,sfmom,M).tH0,eps);
    //DL.test_Self(eps);
    //DH.test_Self(eps);

    System.err.println("EfiComm:test_Self: test OK n="+n+" eps="+eps);
}

} // END OF class EfiComm
