package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** John Pierpont Morgan is back acting on |dp>.
 *  @version $Id: EjpComm.java,v 1.27 2022/08/07 15:44:02 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class EjpComm {

final double I_0,I_L,I_H;
final EVXData F;
final double I_FL,I_FH;
final double F_0;
final double w_IH_FL,w_IH_FH,w_IL_FL,w_IL_FH,fi,ft;
final ACorr acorr,bcorr,ccorr;

EjpComm(final int n,final double [] imom,final double [] tmom,final double [] fmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");
    final double [] basis=getBasis(n,imom,tmom,M);

    final int i0=0,iL=1,iH=n-1;
    this.I_0=basis[i0];
    this.I_L=basis[iL];
    this.I_H=basis[iH];

    final int nd=n-1;
    final double [] basis_noi0=DeprecatedHelperEVXData.getBasisOneVectorRemoved(n,basis,n,i0);
    final double [] QQf=M.B.getQQMatr(n,fmom),jQQf=EVXData.convertMatrFromQtoB_Basis(n,QQf,basis_noi0,nd,0),
	jQQt,jQQi;
    final boolean flag_diag_basis=true;
    if(!flag_diag_basis) {
	jQQt=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,tmom),basis_noi0,nd,0);
	jQQi=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,imom),basis_noi0,nd,0);
	this.F=EVXData.getEVXDataFromQQ(nd,jQQf,jQQt,M);
    }
    else { // optimization: jQQt & jQQi are diagonal in new basis
	jQQt=new double [nd*nd];
	jQQi=new double [nd*nd];
	for(int k=0;k<nd;k++){jQQt[k*nd+k]=1;jQQi[k*nd+k]=basis[k+1];}
	this.F=EVXData.getEVXDataFromQQfDiagQQa(nd,jQQf,jQQt,M);
    }

    this.I_FL=TMatr.getAMatrA(F.psiL,jQQi,nd);
    this.I_FH=TMatr.getAMatrA(F.psiH,jQQi,nd);
    this.F_0=TMatr.getAMatrA(EVXData.getPsiFromBasis(i0,basis,n),QQf,n);
    this.w_IH_FL=F.psiL[F.neq-1];
    this.w_IH_FH=F.psiH[F.neq-1];
    this.w_IL_FL=F.psiL[0];
    this.w_IL_FH=F.psiH[0];
    this.acorr=new ACorr(TMatr.getSpurMM(nd,jQQf,jQQi),
			 TMatr.getSpurMM(nd,jQQf,jQQf),
			 TMatr.getSpurMM(nd,jQQi,jQQi));
    this.ccorr=new ACorr(TMatr.getSpurMM(nd,jQQf,jQQt),
			 TMatr.getSpurMM(nd,jQQf,jQQf),
			 TMatr.getSpurMM(nd,jQQt,jQQt));
    final double [] invjQQi=new double [nd*nd];
    if(flag_diag_basis) {
	for(int k=0;k<nd;k++){invjQQi[k*nd+k]=1/jQQi[k*nd+k];}
    }
    this.bcorr=new ACorr(TMatr.getSpurMM(nd,jQQf,invjQQi),
			 TMatr.getSpurMM(nd,jQQf,jQQf),
			 TMatr.getSpurMM(nd,invjQQi,invjQQi));
    double sft=0,sfi=0,st=0,si=0;
    for(int k=0;k<nd;k++){
	sft+=jQQf[k*nd+k]*jQQt[k*nd+k];
	sfi+=jQQf[k*nd+k]*jQQi[k*nd+k];
	st+=jQQt[k*nd+k];
	si+=jQQi[k*nd+k];
    }
    this.ft=sft/st;
    this.fi=sfi/si;

    this.simom=imom;
    this.stmom=tmom;
    this.sfmom=fmom;
    this.M=M;
}
final double [] simom,stmom,sfmom;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;


double [] getBasis(final int n,final double [] imom,final double [] tmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    return EdComm.getIRadauBasisPsi0(n,imom,tmom,M);
}




void test_Self(final double eps,final DeprecatedHelperDataDirectAccessWithBasis<?> access)
{
    final int n=F.neq+1;
    final EVXData bq=BQVM.getEVXData0Removed(n,sfmom,stmom,access,F.x0);
    EVXData.matchEVXData(F,bq,false,false,eps);

    final double [] QQt=M.B.getQQMatr(n,stmom),QQi=M.B.getQQMatr(n,simom);
    UnitTests.checkDoubleRel(TMatr.getAMatrA(bq.psiL,QQi,n),"<psiL|I|psiL>",I_FL,"I_FL",eps);
    UnitTests.checkDoubleRel(TMatr.getAMatrA(bq.psiL,QQt,n),"<psiL|1|psiL>",1,"1",eps);
    UnitTests.checkDoubleRel(TMatr.getAMatrA(bq.psiH,QQi,n),"<psiH|I|psiH>",I_FH,"I_FH",eps);
    UnitTests.checkDoubleRel(TMatr.getAMatrA(bq.psiH,QQt,n),"<psiH|1|psiH>",1,"1",eps);

    final EVXData iq=BQVM.getEVXData0Removed(n,simom,stmom,access,F.x0);
    UnitTests.checkDoubleRel(Math.abs(TMatr.getAMatrB(iq.psiL,bq.psiL,QQt,n)),"<psiIL|psiFL>",Math.abs(w_IL_FL),"w_IL_FL",eps);
    UnitTests.checkDoubleRel(Math.abs(TMatr.getAMatrB(iq.psiL,bq.psiH,QQt,n)),"<psiIL|psiFH>",Math.abs(w_IL_FH),"w_IL_FH",eps);
    UnitTests.checkDoubleRel(Math.abs(TMatr.getAMatrB(iq.psiH,bq.psiL,QQt,n)),"<psiIH|psiFL>",Math.abs(w_IH_FL),"w_IH_FL",eps);
    UnitTests.checkDoubleRel(Math.abs(TMatr.getAMatrB(iq.psiH,bq.psiH,QQt,n)),"<psiIH|psiFH>",Math.abs(w_IH_FH),"w_IH_FH",eps);

    final double [] RRfmom=M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(sfmom)),
	RRimom=M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(simom)),
	RRtmom=M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(stmom));

    final ACorr a=ACorr.getACorr_SPUR(F.neq,RRfmom,RRimom,RRtmom,M);
    ACorr.matchACorr(a,acorr,eps);
    final ACorr i=ACorr.getACorr_SPUR(F.neq,RRimom,RRtmom,RRtmom,M);
    UnitTests.checkDoubleRel(a.x_a_b/i.x_a_b,"a.x_a_b/i.x_a_b",fi,"fi",eps);

    final ACorr at=ACorr.getACorr_SPUR(F.neq,RRfmom,RRtmom,RRtmom,M),
	t=ACorr.getACorr_SPUR(F.neq,RRtmom,RRtmom,RRtmom,M);
    UnitTests.checkDoubleRel(at.x_a_b/t.x_a_b,"at.x_a_b/t.x_a_b",ft,"ft",eps);


    System.err.println("EjpComm:test_Self: test OK n="+n+" eps="+eps);
}




} // END OF class EjpComm
