package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */

import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.ReadWriteUtils;

/** A very crude backtesting paper-trading example.
 * !!! DO NOT USE FOR ACTUAL TRADING !!!
 * This model trades a basket of assets
 * and price marked at last price. Using last price as a mark is actually not a good idea,
 * but this program is just a demo code, so price marks are read from ITCH file.
 * The code is synchroneous for simplicity and reproducibility of the results.
 * For actual trading async code is required (broken at this point).
 * Using synchroneous code in actual trading would immediately lead to the disaster.
 *
 * Enter/exit conditions
 * are implemented according to UMdQVM.java algorithm originating from
 * Jan, 13, 2015 version,
 * see the http://arxiv.org/abs/1510.05510 , section
 * "Directional Liquidity Deficit with Total Market Capitalization".
 * More advanced versions with actual models for broker and money manager at this point
 * have the code BROKEN (compiles, but not working) and need to be fixed.
 * I will fix Broker, MoneyManager and convert Trader to use the signal generated by the UMdQVM.java,
 * total capital liquidity deficit trading algorithm (the only working one),
 * as soon as I manage to get back to trading.
 * (Minimal infrastructure: equity HFT--like account at Lime (acquired in 2011 by WEDBUSH, Inc.)
 * and Nasdaq ITCH feed (colocation in Carteret, New Jersey)).
 *
 * Regarding derivatives. There is an option trading
 * algorithm, that might work (see http://arxiv.org/abs/1510.05510 "Volatility Trading" section),
 * but I never have an access to options HFT trading account (Ray was really scared to open a such
 * after my (and Eugeny Yakushev) initial trading losses even on equities).
 * Actual option trading was never tested. Nobody knows. All the juice might go to
 * the commissions paid to build a straddle. But don not be afraid. The UMdQVM.java
 * approach never short volatility, so I promise not to piss away more capital than you have on the account.
 *
 * According to the http://arxiv.org/abs/1510.05510 prophecy, section "Speculations",
 * this liquidity deficit with hedging equity trading strategy is expected to work better and better
 * as more and more market participants start using it.
 * But there are will be complete market destabilization in the end.
 *
 * @version $Id: ExampleTestMultiStock.java,v 1.54 2016/03/17 00:35:58 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class ExampleTestMultiStock {

private static final boolean PRINT_DEBUG=true;

/** The parameters for simplistic trading algorithm.
 */
static class Config {
    final String [] assets;
    final int neq;
    final long tau;
    final double shares;
    final String strategyname,OUTPUT_DIR;
    Config(final String [] assets,
	   final int neq,
	   final long tau,
	   final double shares,
	   final String strategyname,
	   final String OUTPUT_DIR)
    {
	this.assets=assets;
	this.neq=neq;
	this.tau=tau;
	this.shares=shares;
	this.strategyname=strategyname;
	this.OUTPUT_DIR=OUTPUT_DIR;
    }
    static final long MILLISECOND_SCALE=TraderInterface.TIME_SCALE/1000;
    static Config read(final String fname) throws java.io.IOException
    {
	final java.util.ResourceBundle p=ReadWriteUtils.getProperties(fname);
	final String OUTPUT_DIR=ReadWriteUtils.getparameter("OUTPUT_DIR",p,true);
	final String strategyname=ReadWriteUtils.getparameter("strategyname",p,true);
	final String [] assets=ReadWriteUtils.extractnames(ReadWriteUtils.getparameter("assets",p,true),":\t\r\n ");
	final int neq=Integer.valueOf(ReadWriteUtils.getparameter("neq",p,true));
	final long tau=Long.valueOf(ReadWriteUtils.getparameter("tau",p,true))*MILLISECOND_SCALE;
	final double shares=Double.valueOf(ReadWriteUtils.getparameter("shares",p,true));
	return new Config(assets,neq,tau,shares,strategyname,OUTPUT_DIR);
    }
    int isTickerActive(final String t)
    {
	for(int k=0;k<assets.length;k++)
	{
	    if(t.equals(assets[k])) return k;
	}
	return -1;
    }
    String getDescription()
    {
	final StringBuilder out=new StringBuilder();
	out.append("assets="+java.util.Arrays.toString(assets)+"\n");
	out.append("neq="+neq+"\n");
	out.append("tau="+(tau/MILLISECOND_SCALE)+" milliseconds\n");
	out.append("shares="+shares+"\n");
	out.append("strategyname="+strategyname+"\n");
	out.append("OUTPUT_DIR="+OUTPUT_DIR);
	return out.toString();
    }
}// END OF class Config

static class ExampleTradingStrategy {
    final Config config;
    final ExampleStockStatusData [] assets;
    final Trade [][] assets_buffer;
    final int [] assets_buffer_size;
    ExampleTradingStrategy(final Config config){
	this.config=config;
	this.assets=new ExampleStockStatusData [config.assets.length];
	this.assets_buffer=new Trade [config.assets.length][];
	this.assets_buffer_size=new int [config.assets.length];
    }
    void setAsset(final ExampleStockStatusData a)
    {
	final int ind=config.isTickerActive(a.ot.ticker);
	//System.err.println("a.ot.ticker="+a.ot.ticker+" ind="+ind+" config="+config.getDescription());
	if(ind<0) throw new IllegalArgumentException("a.ot.ticker not found in "+config.getDescription());
	if(this.assets[ind]!=null)
	    throw new IllegalArgumentException("asset a.ot.ticker is already set for "+this.assets[ind]);
	this.assets[ind]=a;
    }
    void validateStrategyParameters()
    {
	for(int k=0;k<assets.length;k++)
	{
	    if(assets[k]==null)
		throw new RuntimeException("no asset "+config.assets[k]+" is set for "+config.getDescription());
	}
    }
    long currenttime=-1;
    void copyData(final long currenttime)
    {
	if(this.currenttime>=0) throw new RuntimeException("this.currenttime="+this.currenttime+" race is somewhere");
	this.currenttime=currenttime;
	for(int k=0;k<assets.length;k++)
	{
	    assets[k].copyData();
	    assets_buffer[k]=assets[k].savedtrades.getBuffer();
	    assets_buffer_size[k]=assets[k].savedtrades.getSize();
	}
    }
    void calculate()
    {
	final UMdQVMLegendreShifted c=
	    UMdQVMLegendreShifted.calculate(config.neq,
					    currenttime,
					    config.tau,
					    assets_buffer,
					    assets_buffer_size,
					    this.umdqvmp);
	this.umdqvmp=c;
	updateStrategy(currenttime);
    }
    void clearData()
    {
	if(this.currenttime<0) throw new RuntimeException("this.currenttime="+this.currenttime+" race is somewhere");
	this.currenttime=-1;
	for(int k=0;k<assets.length;k++)
	{
	    assets[k].clearData();
	    assets_buffer[k]=null;
	    assets_buffer_size[k]=0;
	}
    }
    private static final ExampleTradingStrategy [] SNONE=new ExampleTradingStrategy [0];
    static ExampleTradingStrategy [] getStrategiesByTicker(final String ticker,final ExampleTradingStrategy [] strategies)
    {
	int n=0;
	final ExampleTradingStrategy [] s=new ExampleTradingStrategy [strategies.length];
	for(int k=0;k<strategies.length;k++)
	{
	    if(strategies[k].config.isTickerActive(ticker)>=0) s[n++]=strategies[k];
	}
	if(n<=0) return SNONE;
	if(n>=s.length) return s;
	final ExampleTradingStrategy [] s2=new ExampleTradingStrategy [n];
	System.arraycopy(s,0,s2,0,n);
	return s2;
    }
    UMdQVMLegendreShifted umdqvmp=null;
    UMdQVM.State enter_state=null;
    boolean isTimeTradeOK(final long currenttime)
    {
	return ExampleMultiStockTrader.T_START<= currenttime && currenttime <= ExampleMultiStockTrader.T_END;
    }
    void updateStrategy(final long currenttime)
    {
	if(enter_state==null)
	{
	    if(isTimeTradeOK(currenttime)  && umdqvmp.lam_0 <= umdqvmp.lam_L)
	    {
		enter_state=umdqvmp.st;
		// FIXME open trade
	    }
	}
	else if(umdqvmp.lam_0 >= umdqvmp.lam_H || currenttime>=ExampleMultiStockTrader.T_GLOBAL_END)
	{
	    // close "enter_state" position with exit "umdqvmp.st"
	    final double pnl=enter_state.getPnL(umdqvmp.st);
	    System.err.println("pnl="+pnl); // FIXME
	    enter_state=null;
	}
    }
    java.io.PrintWriter outstate=null;
    void dumpState()
    {
	if(outstate==null)
	{
	    final String fname=config.OUTPUT_DIR+config.strategyname+" with "+java.util.Arrays.toString(config.assets)+".csv";
	    System.err.println("Opening startegy state file="+fname);
	    try {
		outstate=new java.io.PrintWriter(fname);
	    } catch(final java.io.IOException e) {
		throw new RuntimeException(e);
	    }
	    int cn=0;
	    outstate.print("|time:"+(++cn)+"\tlam_0:"+(++cn)+"\tlam_L:"+(++cn)+"\tlam_H:"+(++cn)+"\tsi_H:"+(++cn)+"\tsi_0:"+(++cn)+"\t");
	    for(int k=0;k<config.assets.length;k++)
	    {
		outstate.print(config.assets[k]+":p_last:"+(++cn)+"\t"+
			       config.assets[k]+":i_cH:"+(++cn)+"\t"+
			       config.assets[k]+":p_cH:"+(++cn)+"\t"+
			       config.assets[k]+":dcap_cH:"+(++cn)+"\t");
	    }
	    outstate.println();
	}
	outstate.print(String.valueOf(currenttime)+"\t"+umdqvmp.lam_0+"\t"+umdqvmp.lam_L+
		       "\t"+umdqvmp.lam_H+"\t"+umdqvmp.si_H+"\t"+umdqvmp.si_0+"\t");
	 for(int k=0;k<config.assets.length;k++)
	 {
	     outstate.print(""+umdqvmp.st.pH_last[k]+"\t"+
			    umdqvmp.st.i[k]+"\t"+
			    (umdqvmp.st.cap[k]/umdqvmp.st.i[k])+"\t"+
			    umdqvmp.st.dcap[k]+"\t"
			    );
	 }
	 outstate.println();
    }
    void closeDump()
    {
	if(outstate!=null){
	    outstate.println("|close state DONE");
	    outstate.close();
	    outstate=null;
	}
    }

    /** Call trading algorithm.
     */
    void tryTrading(final int eventid,final String origin_ticker)
    {
	// too many events in output
	if(true && !(eventid==TraderInterface.UPD_EXECUTIONORDER_TRADE)) return;
	//FIXME
	System.err.println("t="+currenttime+" origin_ticker="+origin_ticker+
			   " eventid="+eventid+
			   " assets="+java.util.Arrays.toString(config.assets)+
			   " n assets="+assets.length);
	calculate();
	dumpState();
	// FIXME DO TRADING
    }
} // END OF class ExampleTradingStrategy

static class ExampleStockStatusData {
    final OrdersAndTrades ot;
    ExampleStockStatusData(final OrdersAndTrades ot)
    {
	this.ot=ot;
    }
    final Distribution<Trade> savedtrades=new Distribution<Trade>();
    BookItem best_buy=null,best_sell=null;
    Trade p_last=null;
    void copyData()
    {
	best_buy=ot.orderbook.getBestOrder(BookItem.BOOK_BUY);
	best_sell=ot.orderbook.getBestOrder(BookItem.BOOK_SELL);
	p_last=ot.tradebook.getLastAddedElement();
	ot.tradebook.copyTrades(savedtrades);
    }
    void clearData()
    {
	savedtrades.clearAll();
	best_buy=best_sell=null;
	p_last=null;
    }
}

/** Implement TraderInterface interface to be able to read ITCH data
 * and call event processing on update event.
 */
static class ExampleMultiStockTrader implements TraderInterface {
    static final long T_START=(9*3600+30*60)*TraderInterface.TIME_SCALE, // 9:30
	T_END=(15*3600+55*60)*TraderInterface.TIME_SCALE, // 15:55
	T_GLOBAL_END=T_END+20*TraderInterface.TIME_SCALE; // 15:55 + 20 seconds

    final ExampleTradingStrategy [] strategies;
    ExampleMultiStockTrader(final Config [] configs)
    {
	this.strategies=new ExampleTradingStrategy [configs.length];
	for(int k=0;k<configs.length;k++) strategies[k]=new ExampleTradingStrategy(configs[k]);
    }

    public void allowTrading(final long currenttime,final boolean on_off){}
    public void allowTrading(final OrdersAndTrades ot,final long currenttime,final int value,final String text){}
    public void setMarketMaker(final OrdersAndTrades ot,
			       final long currenttime,
			       final String mpid,
			       final boolean flag_primary,
			       final boolean flag_ok){}
    private boolean creation_finished=false;
    ExampleTradingStrategy [] getStrategies(final OrdersAndTrades ot,final long currenttime)
    {
	ExampleTradingStrategy [] s=ot.getData(); // Store data in OrdersAndTrades for simplicity.
	if(s==null)
	{
	    if(!this.creation_finished)
	    {
		s=ExampleTradingStrategy.getStrategiesByTicker(ot.ticker,this.strategies);
		ot.setActive(s.length>0); // do not store trades for unused stocks
		if(true || s.length>0)
		{
		    ot.setData(s);
		    for(ExampleTradingStrategy e:s) e.setAsset(new ExampleStockStatusData(ot));
		    this.storeMultiStockStatusDatabyticker.put(ot.ticker,s);
		}
	    }
	    else
	    {
		throw new RuntimeException("Late creation of ticker="+ot.ticker+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	}
	else
	{
	    if(s.length>0 && !ot.getActive())
	    {
		throw new RuntimeException("inactive "+ot.ticker+" has "+s.length+" active strategies"+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	    if(s.length<=0 && ot.getActive())
	    {
		throw new RuntimeException("active "+ot.ticker+" has "+s.length+" no active strategies"+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	}
	return s;
    }

    public void setInstrument(final OrdersAndTrades ot,
			      final long currenttime,
			      final boolean flag_statusOK,
			      final int round_lot_shares,
			      final boolean round_lot_only,
			      final String market)
    {
	//System.err.println("setInstrument:ticker="+ot.ticker+" ot="+ot);
	getStrategies(ot,currenttime);
    }

    public void update(final OrdersAndTrades ot,
		       final long currenttime,
		       final int eventid)
    {
	if(!this.creation_finished)
	{ // first update
	    this.creation_finished=true;
	    for(ExampleTradingStrategy [] s:this.storeMultiStockStatusDatabyticker.values())
	    {
		for(ExampleTradingStrategy e:s) e.validateStrategyParameters();
	    }
	    if(PRINT_DEBUG)
	    { // Dump all active stocks for debugging
		System.err.println("=========== ALL STOCKS SETTINGS ========");
		for (java.util.Map.Entry<String, ExampleTradingStrategy []> entry : this.storeMultiStockStatusDatabyticker.entrySet())
		{
		    System.err.println( entry.getKey()+" is set ; used in "+entry.getValue().length+" strategies");
		}
	    }
	}
	final ExampleTradingStrategy [] strategies=getStrategies(ot,currenttime);
	if( // here we trade only asset--related events. Good enough, but may change later.
	    strategies.length>0
	    )
	{
	    long taum=-1;
	    for(ExampleTradingStrategy s: strategies)
	    {
		s.copyData(currenttime);
		s.tryTrading(eventid,ot.ticker);
		s.clearData();
		if(s.config.tau>=taum) taum=s.config.tau;
	    }
	    ot.clearBefore(currenttime-3*taum);
	}
    }

    void closeDiag()
    {
	for(ExampleTradingStrategy [] s:this.storeMultiStockStatusDatabyticker.values())
	{
	    for(ExampleTradingStrategy e:s) e.closeDump();
	}
    }

    private final java.util.Map<String,ExampleTradingStrategy []> storeMultiStockStatusDatabyticker=
	new java.util.HashMap<String,ExampleTradingStrategy []>();
} // END OF class ExampleMultiStockTrader


private final com.polytechnik.itch.ITCHParser parser;
private final ExampleMultiStockTrader trader;
ExampleTestMultiStock(final Config [] configs)
{
    this.trader=new ExampleMultiStockTrader(configs);
    this.parser=new com.polytechnik.itch.ITCHParser(this.trader);
}

void runParser(final String file) throws java.io.IOException
{
    new com.polytechnik.itch.FileFeeder(parser,file).processStream();
    this.trader.closeDiag();
}

public static void main(String [] args)
    throws Exception
{

    if(args.length<=1)
    {
	System.err.println("java -server -Xmx6500000K 'config_file1 config_file2 ' itch_file_1 itch_file_2 ... ");
	System.exit(1);
    }

    final long t0=System.currentTimeMillis();
    System.err.println("===Version $Id: ExampleTestMultiStock.java,v 1.54 2016/03/17 00:35:58 mal Exp $");
    System.err.println("Trading Signals Algorithm: "+UMdQVM.VERSION);
    System.err.println("Started="+new java.util.Date(t0));

    final String [] config_files=ReadWriteUtils.extractnames(args[0],":\t\r\n ");
    System.err.println("Read config files "+java.util.Arrays.toString(config_files));
    final Config [] c=new Config [config_files.length];
    for(int k=0;k<c.length;k++)
    {
	c[k]=Config.read(config_files[k]);
	System.err.println("------CONFIG #"+k+":\n"+c[k].getDescription());
	System.err.println("------END CONFIG");
    }
    System.err.println("===END Version");

    int n=0;
    // trade on each file
    for(int i=1;i<args.length;i++)
    {
	final String f=args[i];
	System.err.println("Processing "+f+" by $Id: ExampleTestMultiStock.java,v 1.54 2016/03/17 00:35:58 mal Exp $");

	final ExampleTestMultiStock tr=new ExampleTestMultiStock(c);
	tr.runParser(f);

	System.err.println("ExampleTestMultiStock: DONE processing "+f);
	System.err.flush();

	n++;
    }

    final long t1=System.currentTimeMillis();
    System.err.println("Finished="+new java.util.Date(t1)+" Started="+new java.util.Date(t0));
    System.err.println("Processed "+n+" files in "+
		       ((t1-t0)/1000L)+
		       " seconds. by $Id: ExampleTestMultiStock.java,v 1.54 2016/03/17 00:35:58 mal Exp $");

}
}
