package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.outils.BissectionRootFinder;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Quadrature;

/** Book in price space &amp; Haar measure.
 *  @version $Id: HaarQVM.java,v 1.22 2022/06/17 20:37:51 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class HaarQVM {


final double pbuy,psell,k0buy_pbuy,k0buy_psell,k0sell_psell,k0sell_pbuy,peq,
    k0buy_peq,k0sell_peq;
private final double [] Kbuy,Ksell;
private final double poffset_buy,poffset_sell;

final DeprecatedHelperDataDirectAccessWithBasis<?> access;
final BookItem [] odata_buy,odata_sell;final int nbuy,nsell;
HaarQVM(final int neq,final BookItem [] odata_buy,final int nbuy,final BookItem [] odata_sell,final int nsell)
{
    this.access=getAccess();
    this.odata_buy=odata_buy;this.nbuy=nbuy;
    this.odata_sell=odata_sell;this.nsell=nsell;

    final int nmoms=2*neq-1;
    this.pbuy=this.poffset_buy=nbuy>0?odata_buy[0].price:Double.NaN;
    this.psell=this.poffset_sell=nsell>0?odata_sell[0].price:Double.NaN;

    final double [] pmoms_buy=getHaarMoments(nmoms,poffset_buy,odata_buy,nbuy);
    final double [] pmoms_sell=getHaarMoments(nmoms,poffset_sell,odata_sell,nsell);

    final double [][] QQ=access.getQQ_Matrs(neq,1,new double [][]{pmoms_buy,pmoms_sell});
    final double [] iQQbuy=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQ[0]),
	iQQsell=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQ[1]);
    final double [][] kkk=access.getKK(neq,new double [][] {iQQbuy,iQQsell});
    this.Kbuy=kkk[0];
    this.Ksell=kkk[1];

    this.k0buy_pbuy=access.fx(this.Kbuy,pbuy-poffset_buy);
    this.k0buy_psell=access.fx(this.Kbuy,psell-poffset_buy);
    this.k0sell_psell=access.fx(this.Ksell,psell-poffset_sell);
    this.k0sell_pbuy=access.fx(this.Ksell,pbuy-poffset_sell);
    this.peq=getPEQ();
    this.k0buy_peq=access.fx(this.Kbuy,peq-poffset_buy);
    this.k0sell_peq=access.fx(this.Ksell,peq-poffset_sell);
}
DeprecatedHelperQVMDataBLpDirectAccessWithBasis getAccess(){return new DeprecatedHelperQVMDataBLpDirectAccessWithBasis();}

static double Ip(final double p){return 1;}

static double [] getHaarMoments(final int nmoms,final double p_offset,final BookItem [] odata,final int nitems)
{
    final double [] pmom=new double [nmoms];
    double p=nitems>0?odata[0].price:Double.NaN,ttau=0;
    long vk=0;
    for(int k=0;k<nitems;k++)
    {
	final BookItem b=odata[k];
	vk+=b.shares;
	if((k+1>=nitems)||!(p==odata[k+1].price))
	{ // next level
	    ttau+=vk/Ip(p);
	    final double mu=vk/ttau,dp=p-p_offset;
	    //System.err.println("MU:p="+p+" dp="+dp+" mu="+mu+" vk="+vk+" ttau="+ttau);
	    double f=1;
	    for(int s=0;s<nmoms;s++)
	    {
		pmom[s]+=f*mu;
		f*=dp;
	    }
	    if(!(k+1>=nitems))
	    {
		vk=0;
		p=odata[k+1].price;
	    }
	}
    }
    //System.err.println("pmom="+java.util.Arrays.toString(pmom));
    return pmom;
}



double getPEQ()
{
    //System.err.println("pbuy="+pbuy+" psell="+psell);
    if(!(psell>pbuy)) return Double.NaN;
    final double S=(psell-pbuy)/5;
    final int NS=250;
    double pxbuy=pbuy;
    for(int k=0;k<NS;k++)
    {
	//System.err.println("pxbuy: k="+k+" f="+(access.fx(this.Kbuy,pxbuy-poffset_buy)-access.fx(this.Ksell,pxbuy-poffset_sell)));
	if(access.fx(this.Kbuy,pxbuy-poffset_buy)<access.fx(this.Ksell,pxbuy-poffset_sell)){break;}
	else if(k>=NS-1) return Double.NaN;
	pxbuy-=S;
    }
    double pxsell=psell;
    for(int k=0;k<NS;k++)
    {
	if(access.fx(this.Kbuy,pxsell-poffset_buy)>access.fx(this.Ksell,pxsell-poffset_sell)){break;}
	else if(k>=NS-1) return Double.NaN;
	pxsell+=S;
    }
    return new BissectionRootFinder(){
	public double getF(final double x)
	{
	    return access.fx(Kbuy,x-poffset_buy)-access.fx(Ksell,x-poffset_sell);
	}
    }.doBissection(pxbuy,pxsell);
}


public static HaarQVM calculate(final int neq,
				      final long currenttime,
				      final BookItem [] buy,
				      final int buysize,
				      final BookItem [] sell,
				      final int sellsize)
{
    return new HaarQVM(neq,buy,buysize,sell,sellsize);
}


static class HaarQVMTest {

int neq;
long currenttime;
Trade [] tbuy,tsell;
BookItem [] tbuy_byp,tsell_byp;

HaarQVM getHaarQVM(){return calculate(neq,currenttime,tbuy_byp,tbuy_byp.length,tsell_byp,tsell_byp.length);}

final String name;
final java.util.Random r;
final double eps;
HaarQVMTest(final String name,
	    final java.util.Random r,final double eps)
{
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Common()
{
    final HaarQVM q=getHaarQVM();

    for(int k=tbuy.length;--k>=0;)
    {
	if(tbuy[k].price>q.pbuy) throw new RuntimeException("tbuy["+k+"].price="+tbuy[k].price+" q.pbuy="+q.pbuy);
    }
    for(int k=tsell.length;--k>=0;)
    {
	if(tsell[k].price<q.psell) throw new RuntimeException("tsell["+k+"].price="+tsell[k].price+" q.psell="+q.psell);
    }

    System.err.println(name+":HaarQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}
void setData(final int buysize,final int sellsize,final int neq)
{
    this.neq=neq;
    this.tbuy=CalculateKendall.TradeEMoments.getRandomTradesQuantumPrice(buysize,r,-1,-2.2,9.98);
    tbuy_byp=tbuy.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tbuy_byp,0,tbuy_byp.length-1,Distribution.cmp_by_price_lowest_down);
    this.tsell=CalculateKendall.TradeEMoments.getRandomTradesQuantumPrice(sellsize,r,-1,2.2,10.1);
    tsell_byp=tsell.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tsell_byp,0,tsell_byp.length-1,Distribution.cmp_by_price_lowest_up);
    this.currenttime=Math.max(tbuy[tbuy.length-1].time,tsell[tsell.length-1].time);
}

double [] getHaarMomentsTest(final int nmoms,final double poffset,final BookItem [] odata,final int nitems)
{
    final double [] pmom=new double [nmoms];
    double p=Double.NaN;
    long v=0,vk=0;
    for(int k=0;k<=nitems;k++)
    {
	if(k>=nitems||!(p==odata[k].price))
	{
	    if(k>0)
	    {
		final double dp=p-poffset,mu=vk/(double)v;
		//System.err.println("TEST:p="+p+" dp="+dp+" mu="+mu+" vk="+vk+" v="+v);
		double f=1;
		for(int s=0;s<nmoms;s++)
		{
		    pmom[s]+=f*mu;
		    f*=dp;
		}
	    }
	    if(!(k>=nitems)) {p=odata[k].price;vk=0;}
	}
	if(!(k>=nitems)){vk+=odata[k].shares;v+=odata[k].shares;}
    }
    return pmom;
}

OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getOrtPolBasis(final double p0,final boolean flag_buy)
{
    return new com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis()
    {
	@Override
	public double [] getRadauMultiplied(final double [] nodes0)
	{
	    final double [] res=getxP(nodes0);
	    if(flag_buy){for(int k=0;k<=nodes0.length;k++) res[k]=(k>=nodes0.length?0:p0*nodes0[k])-res[k];}
	    else{for(int k=0;k<=nodes0.length;k++) res[k]=res[k]-(k>=nodes0.length?0:p0*nodes0[k]);}
	    return res;
	}
    };
}
void test_Pol(final BookItem [] tb,final boolean flag_buy,final double p0,final double k00,final double [] kk,final double kk_p_offset)
{
    final double p_offset=kk_p_offset+(1-2*r.nextDouble());
    final double [] moments_pv=getHaarMomentsTest(2*neq+1,p_offset,tb,tb.length);
    //System.err.println("moments_pv="+java.util.Arrays.toString(moments_pv));
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getOrtPolBasis(p0-p_offset,flag_buy);
    final double [][] p=M.getOrthogonalPolynomialsFirstKind(neq+1,moments_pv);
    final Quadrature xx=M.getRadauQuadratureFromPolynomials(neq-1,p);
    final double [] x=xx.x,w=xx.w;
    if(w.length!=neq) throw new RuntimeException(" w.length="+w.length+" neq="+neq);
    final double w0=w[w.length-1],x0=x[x.length-1]+p_offset;
    final double [] K=new double [2*neq-1];
    //System.err.println("K.length="+K.length+" kk.length="+kk.length+" neq="+neq);
    for(int k=0;k<neq;k++)
    {
	DeprecatedHelperTMatr.sumP2(K,p[k]);
    }

    UnitTests.checkDoubleRel(w0,"w0",1/k00,"1/k00",eps);
    UnitTests.checkDoubleRel(x0,"x0",p0,"p0",eps);
    UnitTests.checkDoubleArrayRel(Polynomials.getPolynomialShifted(kk,-(kk_p_offset-p_offset)),"kk",K,"K",eps);

    System.err.println(name+":HaarQVMTest:test_Pol(): test OK neq="+neq+" eps="+eps);
}

void test_OrtPol()
{
    final HaarQVM q=getHaarQVM();
    test_Pol(tbuy_byp,true,q.pbuy,q.k0buy_pbuy,q.Kbuy,q.poffset_buy);
    test_Pol(tsell_byp,false,q.psell,q.k0sell_psell,q.Ksell,q.poffset_sell);

    System.err.println("pbuy="+q.pbuy+" psell="+q.psell+" peq="+q.peq+
		       "\nk0buy_peq="+q.k0buy_peq+" k0sell_peq="+q.k0sell_peq);

    if(q.peq>q.pbuy) test_Pol(tbuy_byp,true,q.peq,q.k0buy_peq,q.Kbuy,q.poffset_buy);
    else System.err.println("Skipping buy test:peq="+q.peq+" pbuy="+q.pbuy);
    if(q.peq<q.psell) test_Pol(tsell_byp,false,q.peq,q.k0sell_peq,q.Ksell,q.poffset_sell);
    else System.err.println("Skipping sell test:peq="+q.peq+" psell="+q.psell);

    System.err.println(name+":HaarQVMTest:test_OrtPol(): test OK neq="+neq+" eps="+eps);
}

void test_Calculations(final int tbuysize,final int tsellsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":HaarQVMTest:test_Calculations(): test with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);

    setData(tbuysize,tsellsize,neq);
    test_Common();
    test_OrtPol();

    System.err.println(name+":HaarQVMTest:test_Calculations(): test OK with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":HaarQVMTest: starting tests; ver=$Id: HaarQVM.java,v 1.22 2022/06/17 20:37:51 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(850+r.nextInt(100),950+r.nextInt(100),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":HaarQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: HaarQVM.java,v 1.22 2022/06/17 20:37:51 mal Exp $");
}
} // END OF class HaarQVMTest


public static void main(String [] args)
{
    new HaarQVMTest("HaarQVMTest",new java.util.Random(6871L),1e-4).runTests(5);
} // END main

} // END of class HaarQVM
