package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.ProbabilityCorrelation;
import com.polytechnik.utils.Polynomials;


/** Try to add an assymetry (obtained as \int_{-inf}^{x)psi^2(x)omega(x)dx.
 * Consider the I=dV/dt and the state psiH(x) of max I and corresponding p_IH. Then obtain
 * F(x)=\int_{-inf}^{x)psiH2(x)omega(x)dx where F(tnow)=1 follow from psiH normalizing.
 * Then consider g=\int g*(1-F(x))*psiH2(x)*omega(x)*dx and \int g*F(x)*psiH2(x)*omega(x)*dx
 * where g can be price, I or etc.
 * @version $Id: IHState.java,v 1.40 2022/07/12 07:32:33 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public abstract class IHState<DTYPE> {

public static class State {
    public final long seq;
    State(final State prev)
    {
	this.seq=(prev==null?0:1+prev.seq);
    }
}
public final State sst;
public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getABasis();
public final EVXData I,Ibi,Ibt,A,IA;
public final double p_offset,p_last,p_IH,pt_IH,dp_IH,p_IHb,i_IHb,dp_IHb,p_IHf,i_IHf,dp_IHf,dp_IAH;
public final ProbabilityCorrelation pib,pif;
final DTYPE odata,odata2;

final double dpab,dpaf,dpa,dpIbi,p_IHbi,p_ILbt,pab_IH;


/** Not worth a function or costructor argument.
 */
private static double EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;

public IHState(final int n,final DTYPE odata,final DTYPE odata2,final State rst)
{

    final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING; // turn on selftests if required
    this.odata=odata;
    this.odata2=odata2;
    final RMomentsData mom=getRMomentsInitial(2*n-1,odata),mm=mom.asQQMatrices(n,M);
    p_offset=mom.p_offset;
    p_last=mom.p_last;
    //System.err.println("mm.tmom="+java.util.Arrays.toString(mm.tmom)+" mm.imom="+java.util.Arrays.toString(mm.imom));
    final int ixH=n-1;
    this.I=EVXData.getEVXDataFromQQ(n,mm.imom,mm.tmom,M);
    this.p_IH=I.getFPsi(ixH,mm.pimom)/I.ev[ixH];
    this.pt_IH=I.getFPsi(ixH,mm.ptmom);
    this.dp_IH=I.getFPsi(ixH,mm.dpmom);
    final int n2=n;
    final RMomentsData mmpsiQQb=getPsiMoments(2*n2-1,mom,I.psiH,true).asQQMatrices(n2,M),
	mmpsiQQf=getPsiMoments(2*n2-1,mom,I.psiH,false).asQQMatrices(n2,M);
    i_IHb=TMatr.getAMatrA(I.psiH,mmpsiQQb.imom,mmpsiQQb.n);
    p_IHb=TMatr.getAMatrA(I.psiH,mmpsiQQb.pimom,mmpsiQQb.n)/i_IHb;
    dp_IHb=TMatr.getAMatrA(I.psiH,mmpsiQQb.dpmom,mmpsiQQb.n)/TMatr.getAMatrA(I.psiH,mmpsiQQb.tmom,mmpsiQQb.n);
    i_IHf=TMatr.getAMatrA(I.psiH,mmpsiQQf.imom,mmpsiQQf.n);
    p_IHf=TMatr.getAMatrA(I.psiH,mmpsiQQf.pimom,mmpsiQQf.n)/i_IHf;
    dp_IHf=TMatr.getAMatrA(I.psiH,mmpsiQQf.dpmom,mmpsiQQf.n)/TMatr.getAMatrA(I.psiH,mmpsiQQf.tmom,mmpsiQQf.n);
    //System.err.println("tmoms B="+java.util.Arrays.toString(getPsiMoments(2*n2-1,mom,I.psiH,true).tmom));
    //System.err.println("tmoms F="+java.util.Arrays.toString(getPsiMoments(2*n2-1,mom,I.psiH,false).tmom));
    pib=getPIprobabilitycorrelation(I.psiH,mmpsiQQb);
    pif=getPIprobabilitycorrelation(I.psiH,mmpsiQQf);

    // old (and not very successful) experiments
    this.IA=EVXData.getEVXDataFromQQ(n,mm.imom,mm.amom,M);
    dp_IAH=TMatr.getAMatrA(IA.psiH,mm.dpmom,n);
    this.A=EVXData.getEVXDataFromQQ(n,mm.amom,mm.tmom,M);
    this.pab_IH=mmpsiQQb.pimom[0]/mmpsiQQb.imom[0];
    dpab=mmpsiQQb.dpmom[0];
    dpaf=mmpsiQQf.dpmom[0];
    dpa=mm.dpmom[0];
    final int ixHb=n2-1;
    if(mmpsiQQb.n!=mm.n) throw new RuntimeException();
    Ibi=EVXData.getEVXDataFromQQ(n2,mmpsiQQb.imom,mm.tmom,M);
    Ibt=EVXData.getEVXDataFromQQ(n2,mm.imom,mmpsiQQb.tmom,M);
    dpIbi=Ibi.getFPsi(ixHb,mmpsiQQb.dpmom);
    p_IHbi=Ibi.getFPsi(ixHb,mmpsiQQb.pimom)/Ibi.ev[ixHb];
    p_ILbt=Ibt.getFPsi(0,mm.pimom)/Ibt.ev[0];
    this.sst=new State(rst);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(i_IHb+i_IHf,"i_IHb+i_IHf",I.sH,"I.sH",eps);
	UnitTests.checkDoubleRel((p_IHb*i_IHb+p_IHf*i_IHf)/I.sH,"p_IH",p_IH,"p_IH",eps);
	UnitTests.checkDoubleRel(dp_IHb+dp_IHf,"dp_IHb+dp_IHf",2*dp_IH,"2*dp_IH",eps);
	runSelfTest(n,n2,mom,eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psiH,mmpsiQQb.tmom,n2),"1/2",0.5,"0.5",eps);
	System.err.println("IHState: mmpsiQQb.tmom[0]="+mmpsiQQb.tmom[0]+" mmpsiQQf.tmom[0]="+mmpsiQQf.tmom[0]);
    }
}
RMomentsData getRMomentsInitial(final int nmoms,final DTYPE odata)
{
    return RMomentsData.get_MomentsData(nmoms,odata,getAccess());
}
abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract RMomentsData getPsiMoments(final int nmoms2,final RMomentsData mom,final double [] psi,final boolean flag_gt);

static ProbabilityCorrelation getPIprobabilitycorrelation(final double [] psi,final RMomentsData mm)
{
	if(!mm.flagQQ) throw new IllegalArgumentException();
	final int n=mm.n;
//	System.err.println("p_offset="+p_offset+" i="+TMatr.getAMatrA(psi,imom,n)+" pi="+TMatr.getAMatrA(psi,pimom,n)+" p2i="+TMatr.getAMatrA(psi,p2imom,n));
	return new ProbabilityCorrelation(
	    TMatr.getAMatrA(psi,mm.tmom,n),
	    TMatr.getAMatrA(psi,mm.ptmom,n),
	    TMatr.getAMatrA(psi,mm.p2tmom,n),
	    TMatr.getAMatrA(psi,mm.ptmom,n),
	    TMatr.getAMatrA(psi,mm.p2tmom,n),
	    TMatr.getAMatrA(psi,mm.p3tmom,n),
	    TMatr.getAMatrA(psi,mm.imom,n),
	    TMatr.getAMatrA(psi,mm.pimom,n),
	    TMatr.getAMatrA(psi,mm.p2imom,n),true);
}

void runSelfTest(final int n,final int n2,final RMomentsData mom,final double eps)
{
    UnitTests.checkDoubleRel(dpa,"dpa",dpab+dpaf,"dpab+dpaf",eps);
    final double [] psic=new double [n];psic[0]=1/Math.sqrt(mom.tmom[0]);
    final RMomentsData mmCf=getPsiMoments(2*n2-1,mom,psic,false);
    if(odata2!=null)
    { // basis exp(t)*Q(t)
	final RMomentsData mom2=getRMomentsInitial(2*n2-1,odata2);
	UnitTests.checkDoubleRel(mmCf.dpmom[0],"mmCf.dpmom[0]",mom2.dpmom[0],"mom2.dpmom[0]",eps);
	UnitTests.checkDoubleRel(mmCf.imom[0],"mmCf.imom[0]",mom2.imom[0],"mom2.imom[0]",eps);
	final DataInterfaceToMoments<DTYPE> a=getAccess();
	final double dp_offset=a.getPOffset(odata2)-a.getPOffset(odata);
	// shift mom2 price offset to the same one as mom
	Polynomials.shiftPOffsetInPlace(mom2.n,mom2.tmom,mom2.ptmom,mom2.p2tmom,mom2.p3tmom,-dp_offset);
	Polynomials.shiftPOffsetInPlace(mom2.n,mom2.imom,mom2.pimom,mom2.p2imom,mom2.p3imom,-dp_offset);
	Polynomials.shiftPOffsetInPlace(mom2.n,mom2.amom,mom2.pamom,mom2.p2amom,mom2.p3amom,-dp_offset);
	// different normalizing dt/tau
	TMatr.scaleVectorInPlace(mom2.imom,2);TMatr.scaleVectorInPlace(mom2.pimom,2);TMatr.scaleVectorInPlace(mom2.p2imom,2);TMatr.scaleVectorInPlace(mom2.p3imom,2);
	TMatr.scaleVectorInPlace(mom2.amom,2);TMatr.scaleVectorInPlace(mom2.pamom,2);TMatr.scaleVectorInPlace(mom2.p2amom,2);TMatr.scaleVectorInPlace(mom2.p3amom,2);
	TMatr.scaleVectorInPlace(mom2.dpmom,2);
	for(int k=0;k<1;k++){
	    for(double [][] rr:new double [][][]{
		    {mmCf.dpmom,mom2.dpmom},
		    {mmCf.imom,mom2.imom},
		    {mmCf.pimom,mom2.pimom},
		    {mmCf.p2imom,mom2.p2imom},
		    {mmCf.p3imom,mom2.p3imom},
		    {mmCf.amom,mom2.amom},
		    {mmCf.pamom,mom2.pamom},
		    {mmCf.p2amom,mom2.p2amom},
		    {mmCf.p3amom,mom2.p3amom},
		    {mmCf.tmom,mom2.tmom},
		    {mmCf.ptmom,mom2.ptmom},
		    {mmCf.p2tmom,mom2.p2tmom},
		    {mmCf.p3tmom,mom2.p3tmom},
		    {mmCf.tmom,mom2.tmom}})
	    {
		UnitTests.checkDoubleRel(rr[0][k],"rr[0][1]",0.5*rr[1][k],"0.5*rr[1][k]",eps);
	    }
	}
    }
    else
    { // basis exp(t)*Q(exp(t))
	for(int k=0;k<mmCf.n;k++){
	    for(double [][] rr:new double [][][]{
		    {mmCf.dpmom,mom.dpmom},
		    {mmCf.imom,mom.imom},
		    {mmCf.pimom,mom.pimom},
		    {mmCf.p2imom,mom.p2imom},
		    {mmCf.p3imom,mom.p3imom},
		    {mmCf.amom,mom.amom},
		    {mmCf.pamom,mom.pamom},
		    {mmCf.p2amom,mom.p2amom},
		    {mmCf.p3amom,mom.p3amom},
		    {mmCf.tmom,mom.tmom},
		    {mmCf.ptmom,mom.ptmom},
		    {mmCf.p2tmom,mom.p2tmom},
		    {mmCf.p3tmom,mom.p3tmom},
		    {mmCf.tmom,mom.tmom}})
	    {
		UnitTests.checkDoubleRel(rr[0][k],"rr[0][1]",M.B.getXQkMomentsFromQkMoments(rr[1])[k],"R(rr[1])",eps);
	    }
	}
    }
    System.err.println("IHState: selftest OK n="+n+" eps="+eps);
}

/** A class for unit test.
*/
static abstract class IHStateTest<DTYPE> extends QTest {

abstract IHState<DTYPE> getIHState();
abstract IkQVM<DTYPE> getIkQVM();
abstract IAImpact<DTYPE> getIAImpact();

IHStateTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}


void test_Extra(){}

void test_IkQVM()
{
    final int n=neq;
    final IHState<DTYPE> q=getIHState();
    final IkQVM<DTYPE> s=getIkQVM();

    System.err.println(name+":IHStateTest.test_IkQVM(): test OK n="+n+" eps="+eps);
}
void test_IAImpact()
{
    final int n=neq;
    final IHState<DTYPE> q=getIHState();
    final IAImpact<DTYPE> s=getIAImpact();
    EVXData.matchEVXData(q.I,s.I,false,eps);
    EVXData.matchEVXData(q.IA,s.IA,false,eps);
    UnitTests.checkDoubleRel(q.dp_IAH,"q.dp_IAH",s.dp_IAH,"s.dp_IAH",eps);

    System.err.println(name+":IHStateTest.test_IAImpact(): test OK n="+n+" eps="+eps);
}

void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":IHStateTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Extra();
    test_IkQVM();
    test_IAImpact();

    System.err.println(name+":IHStateTest.test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_min,final int neq_max)
{
    System.err.println(name+":IHStateTest: starting tests; ver=$Id: IHState.java,v 1.40 2022/07/12 07:32:33 mal Exp $");
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=eps;
    final long t0=System.currentTimeMillis();
    for(int neq=neq_min;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;
    final long t1=System.currentTimeMillis();
    System.err.println(name+":IHStateTest.main: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: IHState.java,v 1.40 2022/07/12 07:32:33 mal Exp $");
}
} // END OF class IHStateTest


static void matchIHState(final IHState a,final IHState b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    EVXData.matchEVXData(a.A,b.A,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.pt_IH,"a.pt_IH",b.pt_IH+p_offset_shift,"b.pt_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.p_IHb,"a.p_IHb",b.p_IHb+p_offset_shift,"b.p_IHb+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.p_IHf,"a.p_IHf",b.p_IHf+p_offset_shift,"b.p_IHf+p_offset_shift",eps);
    ProbabilityCorrelation.matchProbabilityCorrelationWithOffset(a.pib,a.p_offset,b.pib,b.p_offset,eps);
    ProbabilityCorrelation.matchProbabilityCorrelationWithOffset(a.pif,a.p_offset,b.pif,b.p_offset,eps);

    UnitTests.checkDoubleRel(a.dpab,"a.dpab",b.dpab,"b.dpab",eps);
    UnitTests.checkDoubleRel(a.dpaf,"a.dpaf",b.dpaf,"b.dpaf",eps);
    UnitTests.checkDoubleRel(a.pab_IH,"a.pab_IH",b.pab_IH+p_offset_shift,"b.pab_IH+p_offset_shift",eps);
    System.err.println("IHState.matchIHState test OK n="+a.I.neq+" eps="+eps);
}

}
