package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;

/** Impact QVM calculations.
 *  @version $Id: IQVM.java,v 1.120 2019/01/06 23:33:47 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class IQVM<DTYPE> {

public final double io,jo,q_eq,sigma2_eq,x_q_eq,x_sigma2_eq,i_i,x_i_i,qf,qfo;

private final int mtype;
final DTYPE odata;
final DeprecatedHelperDataInterface<DTYPE> access;
IQVM(final int neq_t,final DTYPE odata,final int mtype)
{
    this.odata=odata;
    this.access=getAccess(neq_t);
    this.mtype=mtype;

    final double [] i=new double [neq_t],qi=new double [neq_t],o=new double [neq_t];
    setMom(qi,o,mtype);
    for(int k=neq_t;--k>=0;){i[k]=qi[k]+o[k];qi[k]-=o[k];}
    access.setM(o,0,access.M_O,odata);
    io=access.get_f_o(i);
    jo=access.get_f_o(qi);
    i_i=access.get_f_g(i,i);
    final double i_qi=access.get_f_g(i,qi);
    q_eq=i_qi/i_i;
    sigma2_eq=1-q_eq*q_eq;
    TMatr.sumP1(i,o,-io/access.get_o_o());
    x_i_i=access.get_f_g(i,i);
    final double x_i_qi=access.get_f_g(i,qi);
    x_q_eq=x_i_qi/x_i_i;
    x_sigma2_eq=1-x_q_eq*x_q_eq;
    // Radon-Nikodym style measure derivative.
    qf=(i_qi-x_i_qi)/(i_i-x_i_i);
    qfo=jo/io;
}

abstract DeprecatedHelperDataInterface<DTYPE> getAccess(final int neq_t);


static final int MT_BS=97,MT_PRICE=99;
static final int default_MT=MT_BS;
private void setMom(final double [] u,final double [] d,final int mtype)
{
    if(mtype==MT_BS)
    {
	access.setM(u,0,access.M_ISELL,odata);
	access.setM(d,0,access.M_IBUY,odata);
    }
    else if(mtype==MT_PRICE)
    {
	access.setM(u,1,access.M_DPPLUS,odata);
	access.setM(d,1,access.M_DPMINUS,odata);
	for(int k=access.get_NT();--k>=0;) d[k]=-d[k];
    }
    else throw new RuntimeException("mtype="+mtype);
}



static abstract class IQVMTest<DTYPE> {

abstract IQVM<DTYPE> getIQVMWithRandomData(final int neq_t,final java.util.Random r,final int tsize);
abstract EQVM<DTYPE> getEQVM(final int neq_t,final DTYPE odata);
final String name;
final double eps,epssum;
IQVMTest(final String name,final double eps,final double epssum)
{
    this.name=name;
    this.eps=eps;
    this.epssum=epssum;
}


void matchIQVM(final int tsize,final java.util.Random r,final int neq_t)
{
    System.err.println(name+":IQVMTest:matchIQVM(): neq_t="+neq_t);
    final IQVM<DTYPE> q=getIQVMWithRandomData(neq_t,r,tsize);

    final double [] i=new double [neq_t],qi=new double [neq_t],
	q2i=new double [neq_t],o=new double [neq_t];
    q.setMom(q2i,o,q.mtype);
    for(int k=0;k<neq_t;k++){i[k]=q2i[k]+o[k];qi[k]=q2i[k]-o[k];q2i[k]=i[k];}
    final DeprecatedHelperDataInterface<DTYPE> access=q.access;
    access.setM(o,0,access.M_O,q.odata);

    final double i_i=access.get_f_g(i,i),
	q_eq=access.get_f_g(i,qi)/i_i,
	sigma2_eq=access.get_f_g(i,q2i)/i_i-q_eq*q_eq,
	o_o=access.get_o_o(),io=access.get_f_o(i);
    UnitTests.checkDoubleRel(q.i_i,"q.i_i",i_i,"i_i",eps);
    UnitTests.checkDoubleRel(q.q_eq,"q.q_eq",q_eq,"q_eq",eps);
    UnitTests.checkDoubleRel(q.sigma2_eq,"q.sigma2_eq",sigma2_eq,"sigma2_eq",eps);
    UnitTests.checkDoubleRel(q.io,"q.io",io,"io",eps);

    TMatr.sumP1(i,o,-io/o_o);
    TMatr.sumP1(qi,o,-access.get_f_o(qi)/o_o);
    TMatr.sumP1(q2i,o,-access.get_f_o(q2i)/o_o);
    UnitTests.checkDoubleRel(access.get_f_o(i),"<i|o>",0,"0",eps);
    UnitTests.checkDoubleRel(access.get_f_o(qi),"<qi|o>",0,"0",eps);
    UnitTests.checkDoubleRel(access.get_f_o(q2i),"<q2i|o>",0,"0",eps);

    final double x_i_i=access.get_f_g(i,i),
	x_q_eq=access.get_f_g(i,qi)/x_i_i,
	x_sigma2_eq=access.get_f_g(i,q2i)/x_i_i-x_q_eq*x_q_eq;
    UnitTests.checkDoubleRel(q.x_i_i,"q.x_i_i",x_i_i,"x_i_i",eps);
    UnitTests.checkDoubleRel(q.x_q_eq,"q.x_q_eq",x_q_eq,"x_q_eq",eps);
    UnitTests.checkDoubleRel(q.x_sigma2_eq,"q.x_sigma2_eq",x_sigma2_eq,"x_sigma2_eq",eps);

    // check Radon-Nikodym measure match
    UnitTests.checkDoubleRel(q.qf,"q.qf",q.jo/q.io,"q.jo/q.io",eps);
    UnitTests.checkDoubleRel(q.qf,"q.qf",q.qfo,"q.qfo",eps);

    System.err.println(name+":IQVMTest:matchIQVM(): ALL TESTS OK, neq_t="+neq_t+" q.mtype="+q.mtype+" eps="+eps);
}

void matchEQVM(final int tsize,final java.util.Random r,final int neq_t)
{
    System.err.println(name+":IQVMTest:matchEQVM(): neq_t="+neq_t);
    final IQVM<DTYPE> q=getIQVMWithRandomData(neq_t,r,tsize);
    final EQVM<DTYPE> s=getEQVM(neq_t,q.odata);


    UnitTests.checkDoubleRel(q.io,"q.io",s.io,"s.io",eps);
    UnitTests.checkDoubleRel(q.i_i,"q.i_i",s.scaleo_io*s.scaleo_io/q.access.get_o_o(),"s.scaleo_io^2/o_o",eps);
    UnitTests.checkDoubleRel(q.x_i_i,"q.x_i_i",s.scale_io0*s.scale_io0/q.access.get_o_o(),"s.scale_io0^2/o_o",eps);

    if(q.mtype==MT_BS)
    {
	UnitTests.checkDoubleRel(q.q_eq,"q.q_eq",s.j2_bs/q.i_i,"s.j2_bs/q.i_i",eps);
	UnitTests.checkDoubleRel(q.x_q_eq,"q.x_q_eq",s.j2x_bs/q.x_i_i,"s.j2x_bs/q.x_i_i",eps);
	UnitTests.checkDoubleRel(q.jo,"q.jo",s.j2_bs_o,"s.j2_bs_o",eps);
    }

    System.err.println(name+":IQVMTest:matchEQVM(): TEST OK, neq_t="+neq_t);
}


void checkExternal(final int tsize,final java.util.Random r,final int neq_t){}
void checkCalculations(final int tsize,final java.util.Random r,final int neq_t)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":IQVMTest:checkCalculations(): test with N points="+tsize+
		       " neq_t="+neq_t+" eps="+eps+" epssum="+epssum);
    matchEQVM(tsize,r,neq_t);
    matchIQVM(tsize,r,neq_t);
    checkExternal(tsize,r,neq_t);
    System.err.println(name+":IQVMTest:checkCalculations(): test OK with N points="+tsize+
		       " neq_t="+neq_t+" eps="+eps+" epssum="+epssum);
}
void runTests()
{
    System.err.println(name+":IQVMTest: starting tests; ver=$Id: IQVM.java,v 1.120 2019/01/06 23:33:47 mal Exp $");
    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-79091);
    for(int neq_t=3;neq_t<=12;neq_t++)
    {
	for(int i=0;i<118;i++)
	{
	    checkCalculations(2210+r.nextInt(2210),r,neq_t);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":IQVMTest: tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: IQVM.java,v 1.120 2019/01/06 23:33:47 mal Exp $");
}
} // END OF class IQVMTest

}// END OF class IQVM
