package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2017
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.ProbabilityCorrelation;

/**
 * A simple try of I(p) impact from the future asymmetry.
 * First variation answer.
 * @version $Id: ImpactFuture2Asymmetry.java,v 1.36 2022/11/17 10:50:23 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class ImpactFuture2Asymmetry {

/** For testing, better use ProbabilityCorrelation.
 */
final MuF mL,mH;
final double di,normbeta,lamconst;
final ProbabilityCorrelation pc;
final double var_f;
public ImpactFuture2Asymmetry(final double p0t,final double p1t,final double p2t,
			      final double p0i,final double p1i,final double p2i,
			      final double d0i,final double Pf,
			      final double eps)
{
    this.di=d0i;
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=
	new OrthogonalPolynomialsMonomialsBasisX(1,-Pf);
    final VolMatrix varpsi=new VolMatrix(){
	    public double [] getDPsi(final double [] phi){
		// In price space. Regular differentiation.
		return M.B.getDifferentiated(phi);
	    }
    };
    final int d=2;
    final double [] Mi=new double []{p0i,p1i,p1i,p2i}, // <I p^j p^k>
	Mpf=M.B.getQQMatr(d,M.B.getX0Moments(2*d-1,Pf)),
	dMi=TMatr.getVectorScaled(Mpf,di),Mif=TMatr.sumAcB(Mi,1,dMi),
	Mt=new double []{p0t,p1t,p1t,p2t} // <p^j p^k>
    ;

    final EVXData mI=EVXData.getEVXDataFromQQ(d,Mif,Mt,M);
    final double [] dphi0=varpsi.getDPsi(mI.psi0),
	rqf=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(mI.psi0,dphi0,null,Mif,Mt,d);
    this.var_f=rqf[1];
    if(eps>0)
    {
	final double [] rq0=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(mI.psi0,dphi0,null,dMi,Mt,d);
	UnitTests.checkDoubleRel(rq0[1],"rq0[1]",0,"0",eps);
	final double [] rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(mI.psi0,dphi0,null,Mi,Mt,d);
	UnitTests.checkDoubleRel(rq[1],"rq[1]",rqf[1],"rqf[1]",eps);
    }
    final double [] ev=M.EV.getGEV_Generalized(d,Mif,Mt,true); // eigenvalues problem
    this.mL=new MuF(EVXData.getPsiFromBasis(0,ev,d), // eigenvector low I
		    ev[0],Mi,dMi,Mpf,Mt,M,varpsi,Pf,eps);
    this.mH=new MuF(EVXData.getPsiFromBasis(d-1,ev,d), // eigenvector high I
		    ev[d-1],Mi,dMi,Mpf,Mt,M,varpsi,Pf,eps);
    // Scale as {mL,mH}.beta*(di/normbeta)
    normbeta=Math.pow(mL.psi_at_Pf,2)+Math.pow(mH.psi_at_Pf,2);
    lamconst=(p0i+d0i)/p0t;
    final double [] MDb=TMatr.sumaAbB(1/normbeta,varpsi.getDiffMatrix(d,Mpf),-0.5,varpsi.getDiffMatrix(d,Mt));
    this.pc=new ProbabilityCorrelation(Mt,MDb,Mif,M,true);
    if(eps>0)
    {
	// Check beta difference to match exactly.
	final double bd=TMatr.getAMatrA(mL.psi,MDb,d)-TMatr.getAMatrA(mH.psi,MDb,d);
	UnitTests.checkDoubleRel(bd,"bd",(mL.beta-mH.beta)/normbeta,"(mL.beta-mH.beta)/normbeta",eps);
	UnitTests.checkDoubleRel(pc.corr*(pc.pL-pc.pH),"*(pc.pL-pc.pH)",bd,"bd",eps);
	// check the matrix
	UnitTests.checkDoubleArrayRel1(dMi,"dMi",new double []{di,Pf*di,Pf*di,Math.pow(Pf,2)*di},"dMi",eps);
	UnitTests.checkDoubleRel(Mif[0]/Mt[0],"Mif[0]/Mt[0]",lamconst,"lamconst",eps);
	System.err.println("mL.beta="+mL.beta+" mH.beta="+mH.beta);
	final double [] D=varpsi.getDiffMatrix(d,Mpf),Dt=varpsi.getDiffMatrix(d,Mt);
	final ImpactFuture2Asymmetry i=
	    new ImpactFuture2Asymmetry(p0t,p1t,p2t,p0i,p1i,p2i,0,Pf,0);
	for(int j=0;j<2;j++)
	{
	    final double [] e;final MuF _mL,_mH;
	    if(j<=0) {e=ev;_mL=mL;_mH=mH;}
	    else{e=M.EV.getGEV_Generalized(d,Mi,Mt,true);_mL=i.mL;_mH=i.mH;}
	    final double [] mD=EVXData.convertMatrFromQtoB_Basis(d,D,e),
		mDt=EVXData.convertMatrFromQtoB_Basis(d,Dt,e),
		mMpf=EVXData.convertMatrFromQtoB_Basis(d,Mpf,e);
	    UnitTests.checkDoubleRel(mD[0]-mMpf[0]*mDt[0],"beta0",_mL.beta,"_mL.beta",eps);
	    UnitTests.checkDoubleRel(mD[d*d-1]-mMpf[d*d-1]*mDt[d*d-1],"beta1",_mH.beta,"_mH.beta",eps);
	    final double _Pfe=(_mH.p_psiroot*Math.pow(_mL.psi_at_Pf,2)+_mL.p_psiroot*Math.pow(_mH.psi_at_Pf,2))/
		(Math.pow(_mL.psi_at_Pf,2)+Math.pow(_mH.psi_at_Pf,2)),
		_Pfie=(_mH.p_psiroot*_mL.lam*Math.pow(_mL.psi_at_Pf,2)+_mL.p_psiroot*_mH.lam*Math.pow(_mH.psi_at_Pf,2))/(_mL.lam*Math.pow(_mL.psi_at_Pf,2)+_mH.lam*Math.pow(_mH.psi_at_Pf,2));
	    //UnitTests.checkDoubleRel(Pf,"Pf",_Pfie,"_Pfie j="+j,eps);
	}
	final ImpactFuture2Asymmetry i0= // Try |psi0><psi0| state:
	    new ImpactFuture2Asymmetry(p0t,p1t,p2t,1,Pf,Pf*Pf,d0i,Pf,0);
	UnitTests.checkDoubleRel(i0.mL.beta,"i0.mL.beta",0,"0",eps);
	UnitTests.checkDoubleRel(i0.mH.beta,"i0.mH.beta",0,"0",eps);
	System.err.println("ImpactFuture2Asymmetry:sum.beta="+(mL.beta+mH.beta)+" i.sum.beta="+(i.mL.beta+i.mH.beta)+" i0.mL.beta="+i0.mL.beta+" i0.mH.beta="+i0.mH.beta);
    }

}

/** For unit test.
 */
static class MuF {
    final double beta,lam,p_psiroot,psi_at_Pf;
    final double [] psi;
    MuF(final double [] psi,
	final double lam,
	final double [] QQi,final double [] dQQi,final double [] QQpf,final double [] QQt,
	final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	final VolMatrix varpsi,
	final double Pf,
	final double eps)
    {
	final int d=2;
	this.psi=psi;
	if(psi.length!=d || QQi.length!=d*d || QQt.length!=d*d)
	    throw new IllegalArgumentException();
	this.lam=lam;
	this.psi_at_Pf=M.B.getD0(psi,Pf);
	final double [] Dpsi=varpsi.getDPsi(psi);
	this.p_psiroot=-M.B.getD0(psi,0)/M.B.getD0(Dpsi,0);
	this.beta=2*(TMatr.getAMatrB(psi,Dpsi,QQpf,d)-
		     TMatr.getAMatrA(psi,QQpf,d)*TMatr.getAMatrB(psi,Dpsi,QQt,d));

	if(eps>0)
	{
	    // check |psi> is the eigenvalue:
	    final double [] QQif=TMatr.sumAcB(QQi,1,dQQi);
	    final double varIf=2*(TMatr.getAMatrB(psi,Dpsi,QQif,d)-lam*TMatr.getAMatrB(psi,Dpsi,QQt,d)),
		varI=2*(TMatr.getAMatrB(psi,Dpsi,QQi,d)-TMatr.getAMatrA(psi,QQi,d)*TMatr.getAMatrB(psi,Dpsi,QQt,d));
	    UnitTests.checkDoubleRel1(varIf,"varIf",0,"0",eps);
	    UnitTests.checkDoubleRel1(varI,"varI",-beta*dQQi[0],"-beta*dQQi[0]",eps);
	    UnitTests.checkDoubleRel1(Math.pow(psi_at_Pf,2),"psi_at_Pf^2",TMatr.getAMatrA(psi,QQpf,d),"TMatr.getAMatrA(psi,QQpf,d)",eps);
	    System.err.println("ImpactFuture2Asymmetry.MuF: <psi|Dpsi>="+TMatr.getAMatrB(psi,Dpsi,QQt,d)+" D<psi|psi0>^2="+TMatr.getAMatrB(psi,Dpsi,QQpf,d)+" <psi|psi0>^2="+TMatr.getAMatrA(psi,QQpf,d));
	    System.err.println("ImpactFuture2Asymmetry.MuF: test OK, lam="+lam+" eps="+eps);
	}
    }
} // END OF public static class MuF
} // END OF public class ImpactFuture2Asymmetry
