package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.ProbabilityCorrelation;
import com.polytechnik.realtimebook.*;

/**
 *  @version $Id: ImpactQVMMuse.java,v 1.24 2022/07/15 19:57:10 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class ImpactQVMMuse<DTYPE> {

public static class State {
    public final long seq;
    State(final State prev)
    {
	this.seq=(prev==null?0:1+prev.seq);
    }
}
public final State sst;

public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
public final PnLSimpleSkewness pnlss;
public final Skewness SK_P_IH;
public final PnLdIDSk pnldidsk;
public final PnLFutureSk pnlfutureSk;
public final EVXData I,P;
public final double p_offset;
final DTYPE odata;

ImpactQVMMuse(final int n,final DTYPE odata,final State rst)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");

    this.odata=odata;
    this.M=getABasis();
    final MomentsData mm=getMomentsData(n);
    this.p_offset=mm.p_offset;
    final int ixH=n-1; // max I, the eigenvalues are sorted.

    // Moving average (familiar solutions), correspond to n=2.
    this.pnlss=new PnLSimpleSkewness(getMomentsData(2),M);

    // Our "old" solution p_IH: price corresponding the state of max I of past sample.
    this.I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    // P eigenstates (out of curiousity)
    this.P=EVXData.getEVXDataFromQQ(n,mm.QQpi,mm.QQi,M);

    if(false)
    {
	// if you need dp/dt in the state of max I it is
	final double dp_IH=I.getFPsi(ixH,mm.QQdp);
	// you may also llok at ImpactIStatesExample for various price impacts.
    }

    // The SK_P_IH.xa is equal to p_IH.
    this.SK_P_IH=Skewness.getSkewnessFromMonomialMoments(
	I.ev[ixH],
	I.getFPsi(ixH,mm.QQpi),
	I.getFPsi(ixH,mm.QQp2i),
	I.getFPsi(ixH,mm.QQp3i));

    // Naive "dynamic impact" answer.
    // The pnldidsk.Pf_from_pt_true_pi is "trivial Pf" and pnldidsk.deg_from_pt_true_pi is "trivial degeneracy degree".
    // Can be also used for unit tests to match the pnlss.SK_P_average(=pnldidsk.SK_movaver__nodI) and SK_P_IH, see test_ImpactQVM() below.
    this.pnldidsk=PnLdIDSk.get_PnLdIDSk(mm,M,0);

    // balance of "future skewness"
    this.pnlfutureSk=PnLFutureSk.get_PnLFutureSk(mm,M,0);

    this.sst=new State(rst);
}
protected MomentsData getMomentsData(final int n){return MomentsData.get_MomentsData(n,M,odata,getAccess());}

abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();


static abstract class ImpactQVMMuseTest<DTYPE> extends QTest {

abstract ImpactQVMMuse<DTYPE> getImpactQVMMuse();
abstract ImpactQVM<DTYPE> getImpactQVM();

ImpactQVMMuseTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}


void test_Extra(){}


void test_ImpactQVM()
{
    final int n=neq;
    final ImpactQVMMuse<DTYPE> q=getImpactQVMMuse();
    PnLFutureSk.get_PnLFutureSk(q.getMomentsData(n),q.M,eps);
    Skewness.matchSkewness(q.pnlss.SK_P_average,q.pnldidsk.SK_movaver__nodI,eps);
    EVXData.matchEVXData(q.I,q.pnldidsk.I,false,eps);

    final ImpactQVM<DTYPE> s=getImpactQVM();
    if(s==null)
    {
	System.err.println(name+":ImpactQVMMuseTest.test_ImpactQVM(): skipping test OK n="+n+" as getImpactQVM()=null eps="+eps);
	return;
    }

    EVXData.matchEVXData(q.I,s.pnldiskewness_nodI.I,false,eps);
    PnLSimpleSkewness.matchPnLSimpleSkewness(q.pnlss,s.pnlss,eps);
    Skewness.matchSkewness(q.pnlss.SK_P_average.getSameSkewnessXShifted(q.pnlss.p_offset),s.pnldidsk.SK_movaver__nodI.getSameSkewnessXShifted(s.pnldidsk.p_offset),eps);
    Skewness.matchSkewness(q.SK_P_IH.getSameSkewnessXShifted(q.p_offset),s.pnldiskewness_nodI.IHf_S.getSameSkewnessXShifted(s.pnldiskewness_nodI.p_offset),eps);

    PnLdISkewness.matchPnLdIDSk(s.pnldiskewness,PnLdIDSk.get_PnLdIDSk(q.getMomentsData(n),q.M,eps),eps);
    PnLdISkewness.matchPnLdIDSk(s.pnldiskewness,q.pnldidsk,eps);

    System.err.println(name+":ImpactQVMMuseTest.test_ImpactQVM(): test OK n="+n+" eps="+eps);
}



void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":ImpactQVMMuseTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Extra();
    test_ImpactQVM();

    System.err.println(name+":ImpactQVMMuseTest.test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_min,final int neq_max)
{
    System.err.println(name+":ImpactQVMMuseTest: starting tests; ver=$Id: ImpactQVMMuse.java,v 1.24 2022/07/15 19:57:10 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=neq_min;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":ImpactQVMMuseTest.main: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: ImpactQVMMuse.java,v 1.24 2022/07/15 19:57:10 mal Exp $");
}
} // END OF class ImpactQVMMuseTest




}// END OF class ImpactQVMMuse
