package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/**
 * dI/dt=d^2V/dt^2 operator eigenstates:
 * |dI/dt|psi>=lambda|psi> with Radau measure.
 *  @version $Id: InDIState.java,v 1.21 2022/07/14 23:48:04 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class InDIState {
public final EVXData I;
public final double dIL,dIH;
public final Mu2 mu2dIL,mu2dIH;

public final double [] psik_I_psik,psik_pI_psik,psik_p2I_psik,psik_t_psik,psik_pt_psik,psik_p2t_psik;
final double [] rbasis;
final Mu2 [] mu2;
/** Consider the I(p) in the state of dI/dt.
 * @param mm Input matrices.
 * @param M Basis operator.
 * @param VOLATILITY_EDPsi Volatility operator.
 * @param flag_R If true use Radau measure (x-x0)*dmu, if false use the (x-x0)^2*dmu measure.
 * @param flag_Add_Impact_From_The_future Whether to add an impact from the future (has an effect only when flag_R=true).
 * @param eps Run selftest with this eps, if positive.
 */
public InDIState(final IMomentsData mm,
		 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		 final VolMatrix VOLATILITY_EDPsi,
		 final boolean flag_R,
		 final boolean flag_Add_Impact_From_The_future,
		 final double eps)
{
    final int n=mm.n,nd=n-1;
    if(n<2 || (!flag_R && flag_Add_Impact_From_The_future))
	throw new IllegalArgumentException();
    this.I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=flag_Add_Impact_From_The_future?I.sH:I_0,dI=I_0_feq-I_0,dV=dI/I.o_o;
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),QQi_feq=TMatr.sumAcB(mm.QQi,dV,QQ0);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi_feq,n),"<psi|I^{f}|psi>",I_0_feq,"I_0_feq",eps);
    }
    final double [] RQQdi,RQQt;
    if(flag_R)
    {
	RQQdi=TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDRadauMatrix(n,QQi_feq,M),-1); // -1 to have regular sign for dI/dt
	RQQt=VolMatrix.getRadauMatrix(n,mm.QQt,M);
    }
    else
    {
	RQQdi=TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDRadau2Matrix(n,QQi_feq,M),-1);
	RQQt=VolMatrix.getRadau2Matrix(n,mm.QQt,M);
    }
    rbasis=M.EV.getGEV_Generalized(nd,RQQdi,RQQt,true);
    dIL=rbasis[0];
    dIH=rbasis[nd-1];
    psik_I_psik=new double [nd];
    psik_pI_psik=new double [nd];
    psik_p2I_psik=new double [nd];
    psik_t_psik=new double [nd];
    psik_pt_psik=new double [nd];
    psik_p2t_psik=new double [nd];
    mu2=new Mu2 [nd];
    for(int k=0;k<nd;k++)
    {
	final double [] psi=EVXData.getPsiFromBasis(k,rbasis,nd),Rpsi=M.getRadauMultiplied(psi),_psi=flag_R?psi:Rpsi;
	psik_I_psik[k]=TMatr.getAMatrB(_psi,Rpsi,mm.QQi,n);
	psik_pI_psik[k]=TMatr.getAMatrB(_psi,Rpsi,mm.QQpi,n);
	psik_p2I_psik[k]=TMatr.getAMatrB(_psi,Rpsi,mm.QQp2i,n);
	psik_t_psik[k]=TMatr.getAMatrB(_psi,Rpsi,mm.QQt_sample,n);
	psik_pt_psik[k]=TMatr.getAMatrB(_psi,Rpsi,mm.QQpt_sample,n);
	psik_p2t_psik[k]=TMatr.getAMatrB(_psi,Rpsi,mm.QQp2t_sample,n);
	final double nf=1/psik_t_psik[k]; // sample and exact give slightly different normalizing
	mu2[k]=new Mu2(psik_t_psik[k]*nf,psik_pt_psik[k]*nf,psik_p2t_psik[k]*nf,
		       psik_I_psik[k],psik_pI_psik[k],psik_p2I_psik[k],eps);
	if(eps>0)
	{
	    UnitTests.checkDoubleRel(TMatr.getAMatrB(_psi,Rpsi,mm.QQt,n),"<psi|RQQt|psi>",1,"1",eps);
	    UnitTests.checkDoubleRel(-(TMatr.getAMatrB(Rpsi,VOLATILITY_EDPsi.getDPsi(_psi),QQi_feq,n)+
				       TMatr.getAMatrB(_psi,VOLATILITY_EDPsi.getDPsi(Rpsi),QQi_feq,n)),"<psi|RdI/dt|psi>",rbasis[k],"rbasis[k]",eps);
	}
    }
    mu2dIL=mu2[0];
    mu2dIH=mu2[nd-1];
    if(eps>0)
    {
      if(flag_Add_Impact_From_The_future)
      {
	  // this one has an impact from the future.
	  final DynVar2RDISkewness ds=DynVar2RDISkewness.get_DynVar2RDISkewness(
	      n,mm.p_offset,mm.p_last,mm.QQt,mm.QQi,
	      mm.QQi,mm.QQpi,mm.QQp2i,mm.QQp3i,M,VOLATILITY_EDPsi,false,eps),
	  dst=DynVar2RDISkewness.get_DynVar2RDISkewness(
	      n,mm.p_offset,mm.p_last,mm.QQt,mm.QQi,
	      mm.QQt_sample,mm.QQpt_sample,mm.QQp2t_sample,mm.QQp3t_sample,M,VOLATILITY_EDPsi,false,eps);
	  for(int k=0;k<nd;k++)
	  {
	      final com.polytechnik.utils.Skewness sk=ds.RDIf_S[k],st=dst.RDIf_S[k];
	      UnitTests.checkDoubleRel(psik_I_psik[k],"psik_I_psik[k]",sk.w1+sk.w2,"sk.w1+sk.w2",eps);
	      UnitTests.checkDoubleRel(psik_pI_psik[k],"psik_pI_psik[k]",sk.xa*(sk.w1+sk.w2),"sk.xa*(sk.w1+sk.w2)",eps);
	      UnitTests.checkDoubleRel(psik_p2I_psik[k],"psik_p2I_psik[k]",Math.pow(sk.x1,2)*sk.w1+Math.pow(sk.x2,2)*sk.w2,"Math.pow(sk.x1,2)*sk.w1+Math.pow(sk.x2,2)*sk.w2",eps);
	      UnitTests.checkDoubleRel(psik_t_psik[k],"psik_t_psik[k]",st.w1+st.w2,"st.w1+st.w2",eps);
	      UnitTests.checkDoubleRel(psik_pt_psik[k],"psik_pt_psik[k]",st.xa*(st.w1+st.w2),"st.xa*(st.w1+st.w2)",eps);
	      UnitTests.checkDoubleRel(psik_p2t_psik[k],"psik_p2t_psik[k]",Math.pow(st.x1,2)*st.w1+Math.pow(st.x2,2)*st.w2,"Math.pow(st.x1,2)*st.w1+Math.pow(st.x2,2)*st.w2",eps);
	  }
      }
      else for(double [][] mma:new double [][][]{
	      {mm.QQi,psik_I_psik},{mm.QQpi,psik_pI_psik},{mm.QQp2i,psik_p2I_psik},
	      {mm.QQt_sample,psik_t_psik},{mm.QQpt_sample,psik_pt_psik},{mm.QQp2t_sample,psik_p2t_psik}
	  })
      {
	final double [] QQ=mma[0],f=mma[1],tomatch[];
	final DREVXData dr;
	if(flag_R)
	{
	    // this one to match
	    final DynRPnL d=DynRPnL.get_DynRPnL(n,mm.QQi,mm.QQt,mm.QQpi,QQ,VOLATILITY_EDPsi,M);
	    dr=DREVXData.get_DREVXData(n,mm.QQi,mm.QQt,mm.QQpi,QQ,mm.QQt,VOLATILITY_EDPsi,M);
	    tomatch=new double [][]{d.evDIRadau,dr.ev.ev};
	} else {
	    dr=DREVXData.get0_DREVXData(n,mm.QQi,mm.QQt,mm.QQpi,QQ,mm.QQt,VOLATILITY_EDPsi,M);
	    tomatch=new double [][]{dr.ev.ev};
	}
	UnitTests.checkDoubleRel(f[0],"f[0]",dr.f_H,"dr.f_H",eps);
	UnitTests.checkDoubleRel(f[nd-1],"f[nd-1]",dr.f_L,"dr.f_L",eps);
	for(double [] xm:tomatch)for(int j=0;j<nd;j++)
	{
	    // the DynRPnL actually calculate -dI/dt
	    UnitTests.checkDoubleRel(rbasis[j],"rbasis[j]",-xm[nd-j-1],"-xm[nd-j-1]",eps);
	}
      }
      System.err.println("InDIState: test OK n="+n+" flag_R="+flag_R+" flag_Add_Impact_From_The_future="+flag_Add_Impact_From_The_future+" eps="+eps);
    }
}

public InDIState(final IMomentsData mm,
		 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		 final VolMatrix VOLATILITY_EDPsi)
{
    this(mm,M,VOLATILITY_EDPsi,true,true,0);
}
}
