package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/**
 * dI/dt=d^2V/dt^2 operator eigenstates:
 * |dI/dt|psi>=lambda|psi>.
 *  @version $Id: InDIStateI0.java,v 1.35 2022/11/20 12:41:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class InDIStateI0 {
public final EVXData I;
/** Consider the I(p) in the state of dI/dt.
 * @param mm Input matrices.
 * @param M Basis operator.
 * @param VOLATILITY_EDPsi Volatility operator.
 * @param flag_Add_Impact_From_The_future Whether to add an impact from the future (has an effect only when flag_R=true).
 * @param eps Run selftest with this eps, if positive.
 */
public InDIStateI0(final IMomentsData mm,
		   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		   final VolMatrix VOLATILITY_EDPsi,
		   final boolean flag_Add_Impact_From_The_future,
		   final double eps)
{
    final int n=mm.n;
    this.I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=flag_Add_Impact_From_The_future?I.sH:I_0,dI=I_0_feq-I_0,dV=dI/I.o_o;
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),QQi_feq=TMatr.sumAcB(mm.QQi,dV,QQ0),
	bpsiC=EVXData.convertPsiFromQtoB_Basis(n,java.util.Arrays.copyOf(new double []{1/Math.sqrt(mm.QQt[0])},n),I.ev,mm.QQt),
	bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	bQQdp=EVXData.convertMatrFromQtoB_Basis(n,mm.QQdpdt,I.ev),
	bQQdi=EVXData.convertMatrFromQtoB_Basis(n,TMatr.sumaAbB(I_0_feq,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQi_feq)),I.ev);

    final double [] bbasis=M.EV.getEV(n,bQQdi,true);
    double sp=0,sm=0,sap=0,sam=0,sa=0;
    for(int k=0;k<n;k++)
    {
	final double ac2=Math.pow(DeprecatedHelperTMatr.mulVVo(n,bpsiC,0,bbasis,n+k*n),2),e=TMatr.getAMatrA(n,bbasis,n+k*n,bQQdp,n);
	sa+=ac2;
	if(bbasis[k]>0) {sp+=e;sap+=ac2*e;} else {sm+=e;sam+=ac2*e;}
    }
    this.dpP=sp;this.dpaP=sap;this.dpM=sm;this.dpaM=sam;
    this.dV_shares=dV;

    if(eps>0)
    { // test a model matrix with |psi0> being an eigenvector
	final double [] ipsi0=TMatr.getMatr_x_A(n,mm.QQi,I.psi0),tpsi0=TMatr.getMatr_x_A(n,mm.QQt,I.psi0),
	    _QQii0=TMatr.sumaAbB(1,mm.QQi,-1/I.s0,TMatr.getQQaa(ipsi0));
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,_QQii0,n),"<psi0|_QQii0|psi0>",0,"0",eps);
	final double [] _QQiev0=TMatr.sumaAbB(1,_QQii0,0.4551*I.s0,TMatr.getQQaa(tpsi0)), // some matrix with |psi0> as eigenvector
	    IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt);
	final double _i0=TMatr.getAMatrA(I.psi0,_QQiev0,n),
	    _psi0[]=TMatr.getMatr_x_A(n,IQQt,TMatr.scaleVectorInPlace(TMatr.getMatr_x_A(n,_QQiev0,I.psi0),1/_i0)),
	    _QQdi[]=TMatr.sumaAbB(_i0,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,_QQiev0)),
	    _QQdi0[]=TMatr.sumaAbB(TMatr.getAMatrA(I.psi0,QQ0,n),QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQ0));
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,_QQdi),"spur(IQQt*_QQdi)",0,"0",eps);
	UnitTests.checkDoubleArrayRel1(EVXData.getMSign(I.psi0),"I.psi0",EVXData.getMSign(_psi0),"_psi0",eps);
	final EVXData _dI=EVXData.getEVXDataFromQQ(n,_QQdi,mm.QQt,M),_dI0=EVXData.getEVXDataFromQQ(n,_QQdi0,mm.QQt,M);
	UnitTests.checkDoubleRel(TMatr.sumElements(_dI.getEValues()),"spur(bQQdi)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(_dI0.getEValues()),"spur(bQQdi0)",0,"0",eps);
	System.err.println("InDIStateI0:model dI with psi0 dI  spectrum="+java.util.Arrays.toString(_dI.getEValues())+" <psi_diL|psi0>="+_dI.wL+" <psi_diH|psi0>="+_dI.wH+" di0="+_dI.s0);
	System.err.println("InDIStateI0:model dI with psi0 dI0 spectrum="+java.util.Arrays.toString(_dI0.getEValues())+" <psi_di0L|psi0>="+_dI0.wL+" <psi_di0H|psi0>="+_dI0.wH+" di0="+_dI0.s0);

	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQdi),"spur(bQQdi)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),"spur(bQQdi)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi_feq,n),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	// Check spur invariance.
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQdp),"Spur(bQQdp)",sm+sp,"sm+sp",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQdpdt),"Spur(QQdpdt)",sm+sp,"sm+sp",eps);

	UnitTests.checkDoubleRel(sa,"sa",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiC,bQQdp,n),"<bpsiC|dp|psiC>",mm.QQdpdt[0]/mm.QQt[0],">dp/dt>",eps);
	// FIXME, not working
	//UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQdi,n),"<psi0|QQdi|psi0>",0,"0",eps);

	System.err.println("InDIStateI0: dI spectrum="+java.util.Arrays.toString(java.util.Arrays.copyOf(bbasis,n)));
	System.err.println("InDIStateI0: test OK n="+n+" flag_Add_Impact_From_The_future="+flag_Add_Impact_From_The_future+" eps="+eps);
    }
}
    final double dpP,dpaP,dpM,dpaM,dV_shares;
}
