package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.LebesgueQuadratureWithEVData;
import com.polytechnik.utils.LebesgueQuadratureJointDistribution;

/**
 * Consider I=dV/dt and (p-p_{IH})I operators eigenstates.
 *  @version $Id: InIStateIHDensity.java,v 1.81 2022/11/19 11:08:57 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class InIStateIHDensity {
    public final double p_offset,p_IH,dpdt_IH,T_IH,V_IH;
    public final EVXData I,DM,DI,DI0;

public InIStateIHDensity(final int n,final double p_offset,
			 final double [] QQt,final double [] QQi,final double [] QQpi,final double [] QQdpdt,
			 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			 final VolMatrix VOLATILITY_EDPsi,
			 final java.util.function.Function<double [],double []> getPsi2WIntegratedDt,
			 final double eps)
{
    if(n<=1) throw new IllegalArgumentException();
    this.p_offset=p_offset;
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH;
    final double [] psiH=I.psiH,QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0));
    p_IH=TMatr.getAMatrA(psiH,QQpi,n)/I.sH;
    dpdt_IH=TMatr.getAMatrA(psiH,QQdpdt,n);
    final double [] psiWeight=getPsi2WIntegratedDt.apply(psiH);
    if(psiWeight.length!=2*n-1) throw new IllegalArgumentException();
    final double [] UNIT=TMatr.getUnitMatrix(n),
	// Gram matrix inverse can be obtained without solving a linear system.
	IQQt=EVXData.convertMatrFromBtoQ_Basis(n,UNIT,I.ev,UNIT),
	//IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	DensityMatrix=M.B.getQQMatr(n,M.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,psiWeight,IQQt));
    // Density matrix eigenstates, note that psiWeight= sum DM[k]* (DMpsik(x)^2)
    DM=EVXData.getEVXDataFromQQ(n,DensityMatrix,QQt,M);
    T_IH=TMatr.sumElements(DM.getEValues());
    V_IH=TMatr.sumElements(DM.getLambdaFPsi(QQi));

    // dp and dI in DM states.
    final double [] QQdidt0=TMatr.sumaAbB(I_0,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQi));
    dpdi_spur2DMIH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix,IQQt,QQdpdt,IQQt,DensityMatrix}));
    if(true)
    { // same thing in different basis, because it is orthogonal, no need to invert QQt
	final double [] bQQdidt0=EVXData.convertMatrFromQtoB_Basis(n,QQdidt0,I.ev),
	    bQQdpdt=EVXData.convertMatrFromQtoB_Basis(n,QQdpdt,I.ev),
	    bDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,DensityMatrix,I.ev);
	// same thing as dpdi_spur2DMIH
	dpdi_spur2DMIHa=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{bQQdidt0,bDensityMatrix,bQQdpdt,bDensityMatrix}));
	if(false)
	{
	    System.err.println("monomials:psiWeigth="+java.util.Arrays.toString(M.B.convertFunctionTo_Monomials_from_Basis(psiWeight)));
	    System.err.println("bDensityMatrix="+java.util.Arrays.toString(bDensityMatrix));
	    System.err.println("bQQdidt0="+java.util.Arrays.toString(bQQdidt0));
	    System.err.println("bQQdpdt="+java.util.Arrays.toString(bQQdpdt));
	    System.err.println("QQt="+java.util.Arrays.toString(QQt));
	    System.err.println("I.ev="+java.util.Arrays.toString(I.getEValues()));
	}
    }

    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQdidt0),DM.ev),"spur(Di0 rho)",I.s0-I.sH,"I.s0-I.sH",eps);
	final double spur_DM_dI0=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix}));
	UnitTests.checkDoubleRel(spur_DM_dI0,"spur(Di0 rho)",I.s0-I.sH,"I.s0-I.sH",eps);
	UnitTests.checkDoubleRel(dpdi_spur2DMIH,"dpdi_spur2DMIH",dpdi_spur2DMIHa,"dpdi_spur2DMIHa",eps);
	final double [] pp=M.B.getKK(n,TMatr.mulMMM(n,IQQt,DensityMatrix,IQQt));
	System.err.println("pp="+java.util.Arrays.toString(pp)+" psiWeight="+java.util.Arrays.toString(psiWeight));
	UnitTests.checkDoubleArrayRel(pp,"pp",psiWeight,"psiWeight",eps);
    }

    dpdi0dm=TMatr.mulVVV(DM.getFPsi(QQdidt0),DM.getFPsi(QQdpdt),DM.getEValues());
    dpdi0dmdm=TMatr.mulVVV(DM.getFPsi(QQdidt0),DM.getFPsi(QQdpdt),TMatr.squareArrayElements(DM.getEValues()));

    // Same thing from the Lebesgue quadrature.
    final LebesgueQuadratureWithEVData lDM=M.getDensityMatrixProducingGivenPolynomial(n,psiWeight,QQt);
    final double [] dI0_lDM=lDM.getMatrixDiagonalsInPsiStates(QQdidt0),dp_lDM=lDM.getMatrixDiagonalsInPsiStates(QQdpdt);
    final double _dpdi0dmdm=TMatr.mulVVV(dI0_lDM,dp_lDM,TMatr.squareArrayElements(lDM.f));
    if(eps>0){
	UnitTests.checkDoubleRel(TMatr.mulVV(dI0_lDM,lDM.f),"spurt(lDm*dI0)",I.s0-I.sH,"I.s0-I.sH",eps);
	UnitTests.checkDoubleRel(dpdi0dmdm,"dpdi0dmdm",_dpdi0dmdm,"_dpdi0dmdm",eps);
    }

    final double [] QQdif=TMatr.sumaAbB(I_0_feq,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQi));
    DI=EVXData.getEVXDataFromQQ(n,QQdif,QQt,M);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psiH,QQdif,n),"<psiH|dI/dt|psiH>",0,"0",eps);
    }
    final double I_IHdm=(I_0_feq*TMatr.mulVV(DI.getFPsi(DensityMatrix),DI.getFPsi(QQ0)));
    p_IHdm=TMatr.mulVV(DI.getFPsi(DensityMatrix),DI.getFPsi(VOLATILITY_EDPsi.getDiffMatrix(n,QQpi)))
	/I_IHdm;
    final double [] _QQdi=VOLATILITY_EDPsi.getDiffMatrix(n,QQi);
    if(eps>0)
    {
	final double [] I_in_DMstate=DM.getFPsi(QQi),DM_in_Istate=I.getFPsi(DensityMatrix);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DI.getFPsi(DensityMatrix),DI.ev),"spur(DM*DI)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQdif),DM.ev),"spur(DM*DI)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQi),DM.ev),"I in DMstate",TMatr.mulVVn1(I.getFPsi(DensityMatrix),I.ev),"DM in Istate",eps);
	final double [] _psiw=new double[psiWeight.length];
	final EVXData _DI=EVXData.getEVXDataFromQQ(n,_QQdi,QQt,M);
	double si=0;
	for(int k=0;k<n;k++)
	{
	    final double [] psi=DM.getPsi(k);
	    TMatr.sumP1(_psiw,M.B.mult2Pol(psi,psi),DM.ev[k]);
	    si+=_DI.ev[k]*_DI.getFPsi(k,DensityMatrix);
	}
	// psi matches
	UnitTests.checkDoubleArrayRel1(_psiw,"_psiw",psiWeight,"psiWeight",eps);
	UnitTests.checkDoubleRel(si,"si",I.sH,"I.sH",eps);
	System.err.println("sum(DM)="+TMatr.sumElements(DM.getEValues())+" sum(DM_in_Istate)="+TMatr.sumElements(DM_in_Istate)+" DM_in_Istate="+java.util.Arrays.toString(DM_in_Istate)+"\t"+"I_in_DMstate="+java.util.Arrays.toString(I_in_DMstate));
	UnitTests.checkDoubleRel(TMatr.sumElements(DM.getEValues()),"sum(DM.ev)",TMatr.sumElements(DM_in_Istate),"TMatr.sumElements(DM_in_Istate)",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(I.getEValues()),"sum(I.ev)",TMatr.sumElements(I_in_DMstate),"TMatr.sumElements(I_in_DMstate)",eps);
	System.err.println("InIStateIHDensity: I.s0="+I.s0+" DM.ev="+java.util.Arrays.toString(DM.getEValues())+" test OK n="+n+" eps="+eps);

	final double dI_inDIstate=TMatr.mulVV(DI.getFPsi(DensityMatrix),DI.getFPsi(QQdif));
	UnitTests.checkDoubleRel(dI_inDIstate,"dI_inDIstate",0,"0",eps);
	final double ifeq=I_0_feq*TMatr.mulVV(DI.getFPsi(DensityMatrix),DI.getFPsi(QQ0)),
	    _ifeq=TMatr.mulVV(DI.getFPsi(DensityMatrix),DI.getFPsi(VOLATILITY_EDPsi.getDiffMatrix(n,QQi)));
	UnitTests.checkDoubleRel(ifeq,"ifeq",_ifeq,"_ifeq",eps);
	UnitTests.checkDoubleRel(I_IHdm,"I_IHdm",/*I.sH*/ifeq,"ifeq",eps);
    }

    final EVXData DPDT=EVXData.getEVXDataFromQQ(n,QQdpdt,QQt,M);
    dpdt_inDPstate=TMatr.mulVV(DPDT.getFPsi(DensityMatrix),DPDT.getFPsi(QQdpdt));
    dpdt_inIstate=TMatr.mulVV(I.getFPsi(DensityMatrix),I.getFPsi(QQdpdt));
    dpdtI_inIstate=TMatr.mulVVV(I.getFPsi(DensityMatrix),I.getFPsi(QQdpdt),I.getEValues());
    dpdt_inDIstate=TMatr.mulVV(DI.getFPsi(DensityMatrix),DI.getFPsi(QQdpdt));
    dpdt_inDMstate=TMatr.mulVV(DM.getFPsi(DensityMatrix),DM.getFPsi(QQdpdt));
    final double [] _QQdpi=VOLATILITY_EDPsi.getDiffMatrix(n,QQpi),
	QQdpif=TMatr.sumaAbB(I_0_feq*p_IH,QQ0,-1,VOLATILITY_EDPsi.getDiffMatrix(n,QQpi));

    if(eps>0)
    {
	UnitTests.checkDoubleRel(dpdt_inDPstate,"dpdt_inDPstate",dpdt_inDMstate,"dpdt_inDMstate",eps);
    }
    dpdt_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpdt,IQQt,DensityMatrix}));
    final double [] QQdpdtf=TMatr.sumAcB(QQdpdt,-dpdt_spur,QQ0);
    if(eps>0)
    {
	final double dpdtf_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpdtf,IQQt,DensityMatrix}));
	UnitTests.checkDoubleRel(dpdtf_spur,"dpdtf_spur",0,"0",eps);
    }

    dpdtI_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQi,IQQt,QQdpdt,IQQt,DensityMatrix}));
    dpdtdI_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdif,IQQt,QQdpdt,IQQt,DensityMatrix}));
    dpdtdIdm_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,DensityMatrix,IQQt,QQdif,IQQt,QQdpdt,IQQt,DensityMatrix}));
    DpDI_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdif,IQQt,DensityMatrix,IQQt,QQdpdt,IQQt,DensityMatrix}));
    DpfDI_spur=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdif,IQQt,DensityMatrix,IQQt,QQdpdtf,IQQt,DensityMatrix}));
    // match old result to spur answer:
    final double I_IH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,_QQdi,IQQt,DensityMatrix})),
	pI_IH=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,_QQdpi,IQQt,DensityMatrix})),
	I0_IH=I_0_feq*TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQ0,IQQt,DensityMatrix}));
    if(eps>0)
    {
	UnitTests.checkDoubleRel(I_IH,"I_IH",I.sH,"I.sH",eps);
	UnitTests.checkDoubleRel(I0_IH,"I0_IH",I.sH,"I.sH",eps);
	UnitTests.checkDoubleRel(p_IH,"p_IH",pI_IH/I.sH,"pI_IH/I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQdif),DM.ev),"spur(Dif rho)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQdpdtf),DM.ev),"spur(Dpf rho)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQdpif),DM.ev),"spur(Dpif rho)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQ0),DM.ev),"spur(QQ0 rho)",1,"1",eps);
    }
    final double [] QQdpDM=TMatr.scaleVectorInPlace(TMatr.mulMatrices(n,new double [][]{DensityMatrix,IQQt,QQdpdtf}),0.5),
	QQdIDM=TMatr.scaleVectorInPlace(TMatr.mulMatrices(n,new double [][]{DensityMatrix,IQQt,QQdif}),0.5);
    TMatr.scaleVectorInPlace(TMatr.selectHermitianPartInPlace(n,QQdpDM),2);
    TMatr.scaleVectorInPlace(TMatr.selectHermitianPartInPlace(n,QQdIDM),2);
    DpfDI_spur_HERM=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpDM,IQQt,QQdIDM}));
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,QQdpDM,IQQt),"spur(Ddp rho)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,QQdIDM,IQQt),"spur(DdI rho)",0,"0",eps);
    }

    DI0=EVXData.getEVXDataFromQQ(n,QQdidt0,QQt,M);
    final double [] DI0_psiaverage=DI0.getPsiAveraged(QQt,M),DI0_dpdt=DI0.getFPsi(QQdpdt);
    double dp_LOWtoHIGH=0,dp_HIGHtoLOW=0;
    for(int k=0;k<n;k++)
    {
	if(DI0.ev[k]>0) dp_LOWtoHIGH+=DI0_dpdt[k]*Math.pow(DI0_psiaverage[k],2);
	else dp_HIGHtoLOW+=DI0_dpdt[k]*Math.pow(DI0_psiaverage[k],2);
    }
    this.dp_LOWtoHIGH=dp_LOWtoHIGH;this.dp_HIGHtoLOW=dp_HIGHtoLOW;
    final double [] tau_in__Di0=DI0.getFPsi(DensityMatrix);
    final double dp_dI__DI0=TMatr.mulVVV(tau_in__Di0,DI0.getEValues(),TMatr.getVAB(tau_in__Di0,DI0.getFPsi(QQdpdt)));

    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.sumElements(DI0.getEValues()),"sum(DI0.ev)",0,"0",eps);
	UnitTests.checkDoubleRel(-TMatr.mulVVn1(DI0.getFPsi(DensityMatrix),DI0.ev),"-spur(Di0 rho)",I.sH-I.s0,"I.sH-I.s0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(DM.getFPsi(QQdidt0),DM.ev),"spur(Di0 rho)",I.s0-I.sH,"I.s0-I.sH",eps);
    }


    DpDI0_corr=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,QQdpdt,IQQt,DensityMatrix}))
	*TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,DensityMatrix}))
	-
	TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix}))
	*TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpdt,IQQt,DensityMatrix}));
    DpDI0_spur2=TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix,IQQt,QQdpdt,IQQt,DensityMatrix}))
	-
	TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix}))
	*TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdpdt,IQQt,DensityMatrix}));

    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,TMatr.mulMatrices(n,new double [][]{IQQt,QQdidt0,IQQt,DensityMatrix})),"spur(Di0 rho)",I.s0-I.sH,"I.s0-I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(DM.getLambdaFPsi(QQdidt0)),"DMspur(Di0 rho)",I.s0-I.sH,"I.s0-I.sH",eps);
	final double _DpDI0_corr=
	    TMatr.sumElements(DM.getLambdaFPsi(TMatr.mulMatrices(n,new double [][]{QQdidt0,IQQt,QQdpdt})))
	    *TMatr.sumElements(DM.getLambdaFPsi(QQt))
	    -
	    TMatr.sumElements(DM.getLambdaFPsi(QQdidt0))
	    *TMatr.sumElements(DM.getLambdaFPsi(QQdpdt));
	UnitTests.checkDoubleRel(DpDI0_corr,"DpDI0_corr",_DpDI0_corr,"_DpDI0_corr",eps);
    }
    final LebesgueQuadratureJointDistribution jdist=
	new LebesgueQuadratureJointDistribution(
	    new LebesgueQuadratureWithEVData(n,QQdidt0,QQt,M),
	    new LebesgueQuadratureWithEVData(n,QQdpdt,QQt,M));
    final double [] corr_dm=jdist.getDensityMatrixCorrelation(DensityMatrix),
	corr_dm_dm=jdist.get_M_M_Correlation(DensityMatrix,DensityMatrix);
    double sfg=TMatr.sumElements(jdist.getFGWeightedWithCorrelation(corr_dm)),
	ssfg=TMatr.sumElements(jdist.getFGWeightedWithCorrelation(corr_dm_dm)),
	sfg2=TMatr.sumElements(TMatr.getVAB(jdist.getFGWeightedWithCorrelation(corr_dm),corr_dm));
    if(eps>0)
    {
	final LebesgueQuadratureWithEVData PDM=new LebesgueQuadratureWithEVData(n,DensityMatrix,QQt,M);
	UnitTests.checkDoubleRel(TMatr.sumElements(PDM.f),"sum(PDM.f)",TMatr.sumElements(corr_dm),"TMatr.sumElements(corr_dm)",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(PDM.f,PDM.f),"sum(PDM.f*PDM.f)",TMatr.sumElements(corr_dm_dm),"TMatr.sumElements(corr_dm_dm)",eps);
	double ss=0,s=0;
	for(int i=0;i<n;i++) for(int j=0;j<n;j++){
		s+=jdist.ff.ev[i]*TMatr.getAMatrB(n,jdist.ff.ev,n+i*n,jdist.gg.ev,n+j*n,DensityMatrix,n)*TMatr.getAMatrB(n,jdist.ff.ev,n+i*n,jdist.gg.ev,n+j*n,jdist.ff.QQ,n)*jdist.gg.ev[j];
		ss+=jdist.ff.ev[i]*Math.pow(TMatr.getAMatrB(n,jdist.ff.ev,n+i*n,jdist.gg.ev,n+j*n,DensityMatrix,n),2)*jdist.gg.ev[j];
	}
	UnitTests.checkDoubleRel(sfg,"sfg",s,"s",eps);
	UnitTests.checkDoubleRel(ssfg,"ssfg",ss,"ss",eps);
    }
    DpDI0_jointS=sfg;
    DpDI0_jointSS=ssfg;
    DpDI0_jointS2=sfg2;
    final double [] taus_in_DI0=jdist.ff.getMatrixDiagonalsInPsiStates(DensityMatrix),
	dI_in_DI0=jdist.ff.f,
	dp_in_DI0=jdist.ff.getMatrixDiagonalsInPsiStates(QQdpdt);
    dpdt_dI_DI0=TMatr.mulVVV(taus_in_DI0,dI_in_DI0,dp_in_DI0);
    dp_dI_DI0=TMatr.mulVVV(taus_in_DI0,dI_in_DI0,TMatr.getVAB(dp_in_DI0,taus_in_DI0));
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.mulVV(taus_in_DI0,dI_in_DI0),"sum(taus_in_DI0*dI_in_DI0)",I.s0-I.sH,"I.s0-I.sH",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(dp_in_DI0),"sum(tdp_in_DI0)",TMatr.sumElements(jdist.gg.f),"TMatr.sumElements(jdist.gg.f)",eps);
	UnitTests.checkDoubleRel(dp_dI_DI0,"dp_dI_DI0",dp_dI__DI0,"dp_dI__DI0",eps);
    }

}
    final double p_IHdm,dpdt_inDPstate,dpdt_inIstate,dpdtI_inIstate,dpdt_inDIstate,dpdt_inDMstate,dpdt_spur,dpdtI_spur,dpdtdI_spur,dpdtdIdm_spur,DpDI_spur,DpfDI_spur,DpfDI_spur_HERM,DpDI0_corr,DpDI0_spur2,dpdi_spur2DMIH,dpdi_spur2DMIHa,DpDI0_jointS2,DpDI0_jointS,DpDI0_jointSS,dpdi0dm,dpdi0dmdm,dpdt_dI_DI0,dp_dI_DI0,dp_LOWtoHIGH,dp_HIGHtoLOW;


static void matchDM_DI(final InIStateIHDensity a,final DM_DI b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    EVXData.matchEVXData(a.DM,b.DM,false,eps);
    UnitTests.checkDoubleRel(a.dpdt_IH,"a.dpdt_IH",b.dpdt_IH,"b.dpdt_IH",eps);
    UnitTests.checkDoubleRel(a.T_IH,"a.T_IH",b.T_IH,"b.T_IH",eps);
    UnitTests.checkDoubleRel(a.V_IH,"a.V_IH",b.V_IH,"b.T_IH",eps);
    UnitTests.checkDoubleRel(a.dpdi_spur2DMIH,"a.dpdi_spur2DMIH",b.dpdi_spur2DMIH,"b.dpdi_spur2DMIH",eps);
    UnitTests.checkDoubleRel(a.dpdi_spur2DMIHa,"a.dpdi_spur2DMIHa",b.dpdi_spur2DMIH,"b.dpdi_spur2DMIH",eps);

    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    //UnitTests.checkDoubleRel(a.pt_IH,"a.pt_IH",b.pt_IH+p_offset_shift,"bt.p_IH+p_offset_shift",eps);

    System.err.println("InIStateIHDensity.matchDM_DI test OK n="+a.I.neq+" eps="+eps);
}

static void matchInIStateIHDensity(final InIStateIHDensity a,final InIStateIHDensity b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    EVXData.matchEVXData(a.DM,b.DM,false,eps);
    UnitTests.checkDoubleRel(a.dpdt_IH,"a.dpdt_IH",b.dpdt_IH,"b.dpdt_IH",eps);
    UnitTests.checkDoubleRel(a.T_IH,"a.T_IH",b.T_IH,"b.T_IH",eps);
    UnitTests.checkDoubleRel(a.V_IH,"a.V_IH",b.V_IH,"b.T_IH",eps);
    UnitTests.checkDoubleRel(a.dpdi_spur2DMIH,"a.dpdi_spur2DMIH",b.dpdi_spur2DMIH,"b.dpdi_spur2DMIH",eps);
    UnitTests.checkDoubleRel(a.dpdi_spur2DMIHa,"a.dpdi_spur2DMIHa",b.dpdi_spur2DMIHa,"b.dpdi_spur2DMIHa",eps);

    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    //UnitTests.checkDoubleRel(a.pt_IH,"a.pt_IH",b.pt_IH+p_offset_shift,"bt.p_IH+p_offset_shift",eps);

    System.err.println("InIStateIHDensity.matchInIStateIHDensity test OK n="+a.I.neq+" eps="+eps);
}

}
