package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;

/**
 * Consider I=dV/dt and (p-p_{IH})I operators eigenstates.
 *  @version $Id: InIStatePI.java,v 1.39 2022/07/01 17:28:20 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class InIStatePI {
public final EVXData I;
public final Skewness PnLsk,PnLskL,PnLskP,PnLskLP,PnLskf;
public final double p_offset,p_IH,p1_IH,pd_IH,dirHmu,dirHsk2,dirHlsk2,dir,pdir,dirPnL,dirPnLsk,dirsk2,dirHsk2_l,dp_a,dp_Ia;

public InIStatePI(final IMomentsData mm,
		  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		  final double eps)
{
    final int n=mm.n;
    if(n<=1) throw new IllegalArgumentException();
    this.p_offset=mm.p_offset;
    this.I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double [] psiH=I.psiH,_pIpsiaverage=I.getPsiAveraged(mm.QQpi,M),_Ipsiaverage=I.getPsiAveraged(mm.QQi,M),_psiaverage=I.getPsiAveraged(mm.QQt,M),tauis=TMatr.squareArrayElements(_psiaverage);
    p_IH=TMatr.getAMatrA(psiH,mm.QQpi,n)/I.sH;
    p1_IH=_pIpsiaverage[n-1]/_Ipsiaverage[n-1];
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.sumElements(tauis),"sum tau",mm.QQt[0],"mm.QQt[0]",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(tauis,I.ev),"sum I.ev tau",mm.QQi[0],"mm.QQi[0]",eps);
    }
    double si=0,spi=0,sw=0,pkf[]=new double [n+1],ikf[]=new double [n+1],dpk[]=new double [n];;
    for(int k=0,iH=n-1;k<n;k++)
    {
	final double w0=Math.pow(I.getFPsiPsi(k,I.psi0,mm.QQt),2),pi=I.getFPsi(k,mm.QQpi),i=I.ev[k];
	dpk[k]=I.getFPsi(k,mm.QQdpdt);
	ikf[k]=i*w0;
	pkf[k]=pi/i;
	final double w=k==iH?(1-w0):-w0;
	si+=i*w;
	spi+=pi*w;
	sw+=w;
	if(eps>0)
	{
	    UnitTests.checkDoubleRel(_Ipsiaverage[k],"_Ipsiaverage[k]",I.ev[k]*_psiaverage[k],"I.ev[k]*_psiaverage[k]",eps);
	}
    }
    pd_IH=spi/si;
    pkf[n]=pd_IH;
    ikf[n]=I.sH-I.s0;
    dp_a=TMatr.mulVV(tauis,dpk);
    dp_Ia=TMatr.mulVVV(tauis,dpk,I.getEValues());
    PnLskf=Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(ikf),
						   TMatr.mulVV(pkf,ikf),
						   TMatr.mulV2V(pkf,ikf),
						   TMatr.mulVVV(pkf,TMatr.squareArrayElements(pkf),ikf));
    if(eps>0)
    {
	UnitTests.checkDoubleRel(si,"si",I.sH-I.s0,"I.sH-I.s0",eps);
	UnitTests.checkDoubleRel(sw,"sw",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(ikf),"sum(ikf)",I.sH,"I.sH",eps);
	UnitTests.checkDoubleRel(PnLskf.xa,"PnLskf.xa",p_IH,"p_IH",eps);
    }
    final double [] QQPnL=TMatr.sumAcB(mm.QQpi,-p_IH,mm.QQi);
    final EVXData PNL=EVXData.getEVXDataFromQQ(n,QQPnL,mm.QQt,M);
    final double [] psiaverage=PNL.getPsiAveraged(mm.QQt,M);
    double spnl=0,smupnl=0;
    for(int k=0;k<n;k++)
    {
	final double mc=PNL.ev[k]*Math.pow(PNL.getFPsiPsi(k,I.psiH,mm.QQt),2);
	spnl+=mc;
	smupnl+=mc*Math.pow(psiaverage[k],2);
    }
    dirHmu=smupnl;
    if(eps>0)
    {
	UnitTests.checkDoubleRel(spnl,"spnl",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(psiaverage,psiaverage),"sum <psik>^2",mm.QQt[0],"mm.QQt[0]",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(psiaverage,PNL.getEValues()),"sum pnlk*<psik>^2",QQPnL[0],"QQPnL[0]",eps);
    }

    final double Pf=mm.p_last/* does have the offset. */,dIp=(I.sH-I.s0)*(Pf-p_IH);
    dirHsk2=(2*0-PNL.sL-PNL.sH)/(PNL.sL-PNL.sH);
    dirHlsk2=(2*dIp-PNL.sL-PNL.sH)/(PNL.sL-PNL.sH);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psiH,QQPnL,n),"<psiH|PnL|psiH>",0,"0",eps);
    }
    int ind_L=0,ind_H=0;
    for(int k=0;k<n;k++)
    {
	if(k<n-1 && PNL.ev[k]<=0 && PNL.ev[k+1]>=0){
	    ind_L=k;
	    ind_H=k+1;
	}
    }
    dirPnL=Math.pow(TMatr.getAMatrB(psiH,PNL.getPsi(ind_H),mm.QQt,n),2)-
	Math.pow(TMatr.getAMatrB(psiH,PNL.getPsi(ind_L),mm.QQt,n),2);
    dirPnLsk=(PNL.ev[ind_H]+PNL.ev[ind_L])/(PNL.ev[ind_H]-PNL.ev[ind_L]);

    final double [] wk=new double [n],pnlk=new double [n],iwk=new double [n],pnldpk=new double [n];
    double sdir=0,psdir=0;
    for(int k=0;k<n;k++)
    {
	final double i=PNL.getFPsi(k,mm.QQi);
	wk[k]=Math.pow(TMatr.getAMatrB(psiH,PNL.getPsi(k),mm.QQt,n),2); // <pisH|psik>^2
	pnlk[k]=PNL.ev[k]; // <psik|PnL|psik>
	psdir+=(pnlk[k]/PNL.getFPsi(k,mm.QQi))*wk[k];
	if(pnlk[k]<=0){
	    sdir-=wk[k];
	} else if(pnlk[k]>=0){
	    sdir+=wk[k];
	}
	iwk[k]=i*wk[k];
	pnldpk[k]=pnlk[k]/i;
    }
    PnLsk=Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(wk),
						  TMatr.mulVV(pnlk,wk),
						  TMatr.mulV2V(pnlk,wk),
						  TMatr.mulVVV(pnlk,TMatr.squareArrayElements(pnlk),wk));
    PnLskP=Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(iwk),
						  TMatr.mulVV(pnldpk,iwk),
						  TMatr.mulV2V(pnldpk,iwk),
						  TMatr.mulVVV(pnldpk,TMatr.squareArrayElements(pnldpk),iwk));
    this.dir=sdir;
    this.pdir=psdir;
    if(eps>0)
    {
	UnitTests.checkDoubleRel(PnLsk.xa,"PnLsk.xa",0,"0",eps);
	UnitTests.checkDoubleRel(PnLskP.xa,"PnLskP.xa",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(wk,pnlk),"sum(wk,pnlk)",0,"0",eps);
	System.err.println("InIStatePI: I.s0="+I.s0+" dirPnL="+dirPnL+" dirPnLsk="+dirPnLsk+" dir="+dir+" PnLsk.xa="+PnLsk.xa);
    }
    final double [] QQPnL_l=TMatr.sumAcB(mm.QQpi,-Pf,mm.QQi);
    final EVXData PNL_l=EVXData.getEVXDataFromQQ(n,QQPnL_l,mm.QQt,M);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psiH,QQPnL_l,n),"<psiH|PnL_l|psiH>",I.sH*(p_IH-Pf),"I.sH*(p_IH-Pf)",eps);
    }
    for(int k=0;k<n;k++)
    {
	final double i=PNL_l.getFPsi(k,mm.QQi);
	wk[k]=Math.pow(TMatr.getAMatrB(psiH,PNL_l.getPsi(k),mm.QQt,n),2); // <pisH|psik>^2
	pnlk[k]=PNL_l.ev[k]; // <psik|PnL|psik>
	iwk[k]=i*wk[k];
	pnldpk[k]=pnlk[k]/i;
    }
    PnLskL=Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(wk),
						   TMatr.mulVV(pnlk,wk),
						   TMatr.mulV2V(pnlk,wk),
						   TMatr.mulVVV(pnlk,TMatr.squareArrayElements(pnlk),wk));
    PnLskLP=Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(iwk),
						    TMatr.mulVV(pnldpk,iwk),
						    TMatr.mulV2V(pnldpk,iwk),
						    TMatr.mulVVV(pnldpk,TMatr.squareArrayElements(pnldpk),iwk));
    dirsk2=(2*0-PNL_l.sL-PNL_l.sH)/(PNL_l.sL-PNL_l.sH);
    dirHsk2_l=(2*TMatr.getAMatrA(I.psiH,QQPnL_l,n)-PNL_l.sL-PNL_l.sH)/(PNL_l.sL-PNL_l.sH);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(PnLskL.xa,"PnLskL.xa",I.sH*(p_IH-Pf),"I.sH*(p_IH-Pf)",eps);
	//UnitTests.checkDoubleRel(PnLskLP.xa,"PnLskLP.xa",(p_IH-Pf),"(p_IH-Pf)",eps);
	System.err.println("InIStatePI: I.s0="+I.s0+" PnLskL.xa="+PnLskL.xa);
    }
}


static void matchInIStatePIo(final InIStatePI a,final InIStatePIo b,final double eps)
{

    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.PnLskL.sk,"a.PnLskL.sk",b.PnLskL.sk,"b.PnLskL.sk",eps);
    Skewness.matchSkewness(a.PnLskL,b.PnLskL,eps); // may fail on xa rounding errors

    System.err.println("InIStatePI.matchInIStatePIo test OK n="+a.I.neq+" p_offset_shift="+p_offset_shift+" eps="+eps);
}
}
