package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;

/**
 * dI/dt=d^2V/dt^2 operator eigenstates:
 * |dI/dt|psi>=lambda|psi>.
 *  @version $Id: InIStatePIf.java,v 1.14 2022/08/07 15:44:02 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class InIStatePIf {
public final EVXData I;
final Skewness PnLsk;
final double dirp;
/** Consider the projections to |psi0>.
 * @param mm Input matrices.
 * @param M Basis operator.
 * @param flag_Add_Impact_From_The_future Whether to add an impact from the future (has an effect only when flag_R=true).
 * @param eps Run selftest with this eps, if positive.
 */
public InIStatePIf(final IMomentsData mm,
		   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		   final boolean flag_Add_Impact_From_The_future,
		   final double eps)
{
    final int n=mm.n;
    this.I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=flag_Add_Impact_From_The_future?I.sH:I_0,dI=I_0_feq-I_0,
	dV=dI/I.o_o,Pf=mm.p_last/* does have the offset. */;
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),QQi_feq=TMatr.sumAcB(mm.QQi,dV,QQ0),
	bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	QQpi_feq=TMatr.sumAcB(mm.QQpi,Pf*dV,QQ0),
	bQQi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQi_feq,I.ev),
	bQQpi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQpi_feq,I.ev);
    final double I0=TMatr.getAMatrA(bpsi0,bQQi_feq,n),pI0=TMatr.getAMatrA(bpsi0,bQQpi_feq,n);
    final double [] bQQpnl=TMatr.sumAcB(bQQpi_feq,-pI0/I0,bQQi_feq);
    if(eps>0)
    {
	UnitTests.checkDoubleRel(I0,"I0",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQpnl,n),"<psi0|QQpnl|psi0>",0,"0",eps);
    }
    final double [] bbasispnl=M.EV.getEV(n,bQQpnl,true),wk=new double [n],pnlk=new double [n];
    double sumi=0,sumpi=0,wsumi=0,wsumpi=0,wsump=0;
    for(int k=0;k<n;k++)
    {
	wk[k]=Math.pow(DeprecatedHelperTMatr.mulVVo(n,bpsi0,0,bbasispnl,n+k*n),2);
	pnlk[k]=bbasispnl[k];
	final double i=TMatr.getAMatrA(n,bbasispnl,n+k*n,bQQi_feq,n);
	sumi+=i;
	sumpi+=pnlk[k];
	wsumi+=wk[k]*i;
	wsumpi+=wk[k]*pnlk[k];
	wsump+=wk[k]*(pnlk[k]/i);
    }
    dirp=wsump;
    PnLsk=Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(wk),
						  TMatr.mulVV(pnlk,wk),
						  TMatr.mulV2V(pnlk,wk),
						  TMatr.mulVVV(pnlk,TMatr.squareArrayElements(pnlk),wk));
    if(eps>0)
    {
	UnitTests.checkDoubleRel(sumi,"sumi",TMatr.getSpurM(n,bQQi_feq),"spur(bQQi_feq)",eps);
	UnitTests.checkDoubleRel(sumpi,"sumpi",TMatr.getSpurM(n,bQQpnl),"spur(bQQpnl)",eps);
	UnitTests.checkDoubleRel(wsumpi,"wsumpi",0,"0",eps);
	UnitTests.checkDoubleRel(PnLsk.xa,"PnLsk.xa",0,"0",eps);
	//UnitTests.checkDoubleRel(wsumi,"wsumi",I_0_feq,"I_0_feq",eps);
	System.err.println("InIStatePIf:test OK n="+n+" flag_Add_Impact_From_The_future="+flag_Add_Impact_From_The_future+" eps="+eps);
    }
}
}
