package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;

/**
 * dI/dt=d^2V/dt^2 operator eigenstates:
 * |dI/dt|psi>=lambda|psi>.
 *  @version $Id: InIStatePIf2.java,v 1.15 2022/08/07 15:44:02 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class InIStatePIf2 {
public final EVXData I;
final Skewness PnLsk;
final double dirp,diri;
/** Consider the projections to |psi0>.
 * @param mm Input matrices.
 * @param M Basis operator.
 * @param eps Run selftest with this eps, if positive.
 */
public InIStatePIf2(final IMomentsData mm,
		    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		    final double eps)
{
    final int n=mm.n;
    this.I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0,
	dV=dI/I.o_o,Pf=mm.p_last/* does have the offset. */;
    final double [] QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),QQi_feq=TMatr.sumAcB(mm.QQi,dV,QQ0),
	bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	QQpi_feq=TMatr.sumAcB(mm.QQpi,Pf*dV,QQ0),
	bQQi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQi_feq,I.ev),
	bQQpi_feq=EVXData.convertMatrFromQtoB_Basis(n,QQpi_feq,I.ev);
    final double [] bQQpnl=TMatr.sumAcB(bQQpi_feq,-Pf,bQQi_feq),
	bbasispnl=M.EV.getEV(n,bQQpnl,true),
	wk=new double [n],pnlk=java.util.Arrays.copyOf(bbasispnl,n);
    double pnl0=TMatr.getAMatrA(bpsi0,bQQpnl,n),sumi=0,wsumi=0,sumpnl=0,wsumpnl=0,wsump=0;
    for(int k=0;k<n;k++)
    {
	final double i=TMatr.getAMatrA(n,bbasispnl,n+k*n,bQQi_feq,n),pnl=pnlk[k],w=Math.pow(DeprecatedHelperTMatr.mulVVo(n,bpsi0,0,bbasispnl,n+k*n),2);
	wk[k]=w;
	sumi+=i;
	sumpnl+=pnl;
	wsumi+=w*i;
	wsumpnl+=w*pnl;
	wsump+=w*(pnl/i);
    }
    dirp=wsump-pnl0/I_0_feq;
    diri=wsumi-I_0_feq;
    PnLsk=Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(wk),
						  TMatr.mulVV(pnlk,wk),
						  TMatr.mulV2V(pnlk,wk),
						  TMatr.mulVVV(pnlk,TMatr.squareArrayElements(pnlk),wk));
    if(eps>0)
    {
	// spur invariance check
	UnitTests.checkDoubleRel(sumi,"sumi",TMatr.getSpurM(n,bQQi_feq),"spur(bQQi_feq)",eps);
	UnitTests.checkDoubleRel(sumpnl,"sumpnl",TMatr.getSpurM(n,bQQpnl),"spur(bQQpnl)",eps);
	// known matrix elements check
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(PnLsk.xa,"PnLsk.xa",pnl0,"pnl0",eps);
	UnitTests.checkDoubleRel(wsumpnl,"wsumpnl",pnl0,"pnl0",eps);
	UnitTests.checkDoubleRel(wsumpnl,"wsumpnl",TMatr.mulVVn1(wk,bbasispnl),"sum(pnl_k*w_k)",eps);
	System.err.println("InIStatePIf2:test OK n="+n+" diri="+diri+" eps="+eps);
    }
}
}
