package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;

/**
 * Consider I=dV/dt and V operator eigenstates. Optimized version.
 *  @version $Id: InIStateVT.java,v 1.8 2022/06/27 14:47:45 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class InIStateVT {
public final EVXData I;
public final double p_offset,p_IH,p_Vmax,p_Tmax,d_pIH,d_pItau;

public InIStateVT(final int n,final double p_offset,
		  final double [] QQt,final double [] QQi,final double [] QQpi,
		  final VolMatrix VOLATILITY_EDPsi,
		  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		  final double eps)
{
    //if(n<=1) throw new IllegalArgumentException();
    this.p_offset=p_offset;
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int iI=n-1; // the state of max I
    this.p_IH=I.getFPsi(iI,QQpi)/I.ev[iI];
    final double [] ONE=M.B.getONE();
    final double [] wtau=new double [n];
    for(int k=0;k<n;k++)
    {
	wtau[k]=Math.pow(TMatr.getAMatrB(ONE.length,ONE,0,n,I.ev,n+k*n,QQt,n),2);
    }

    int ind_Vmax=0,ind_Tmax=0;
    for(int k=0;k<n;k++)
    {
	if(wtau[k]*I.ev[k]>=wtau[ind_Vmax]*I.ev[ind_Vmax]) ind_Vmax=k;
	if(wtau[k]>=wtau[ind_Tmax]) ind_Tmax=k;
    }
    p_Vmax=I.getFPsi(ind_Vmax,QQpi)/I.ev[ind_Vmax];
    p_Tmax=I.getFPsi(ind_Tmax,QQpi)/I.ev[ind_Tmax];

    double [] d=new double[n];
    for(int k=0;k<n;k++)
    {
	final double [] psi=I.getPsi(k);
	final double [] rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(psi,VOLATILITY_EDPsi.getDPsi(psi),null,QQpi,QQt,n);
	d[k]=rq[1];
	if(eps>0)
	{
	    final double [] Dpsi=VOLATILITY_EDPsi.getDPsi(psi);
	    final double d1=2*(TMatr.getAMatrB(psi,Dpsi,QQpi,n)-
			       TMatr.getAMatrA(psi,QQpi,n)*TMatr.getAMatrB(psi,Dpsi,QQt,n));
	    UnitTests.checkDoubleRel(d1,"d1",rq[1],"rq[1]",eps);
	}
    }
    d_pIH=d[iI];
    d_pItau=TMatr.mulVV(d,wtau);

    if(eps>0)
    {
	UnitTests.checkDoubleRel(TMatr.mulVVn1(wtau,I.ev),"sum(wtai*I)",QQi[0]/Math.pow(ONE[0],2),"V",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(wtau),"sum(wtai)",QQt[0]/Math.pow(ONE[0],2),"T",eps);
	System.err.println("InIStateVT: test OK n="+n+" sum(wtau)="+TMatr.sumElements(wtau)+" eps="+eps);
    }
}

}
