package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.PolynomialRootsFinder;
import com.polytechnik.utils.Skewness;


/**
 * Find {@literal <psi_y (p-p_{last})*I*psi_y>}=0 states. Found states are localized at x=y.
 * @version $Id: IstatesLocalized.java,v 1.44 2022/08/08 20:10:47 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class IstatesLocalized {

    /** A class to store one p=p_last match.
     */
    public static class PriceMatchStorage {
	public final double x,wr0,i;
	public final double [] psi_M;
	public final Skewness sk;
	public PriceMatchStorage(final double x,
				 final double wr0,
				 final double i,
				 final double ip,
				 final double ip2,
				 final double ip3,
				 final double [] psi_M){
	    this.x=x;this.wr0=wr0;this.i=i;this.psi_M=psi_M;
	    this.sk=Skewness.getSkewnessFromMonomialMoments(i,ip,ip2,ip3);
	}
    } // END OF class PriceMatchStorage

    public final double p_offset,p_last,p_IH,sk_Iweighted;

    /** The p=p_last matches,
     * time ordered, earliest in time to be at 0-th element index.
     * The x is not necessary on the measure support.
     */
    public final PriceMatchStorage [] priceMatches;
    /** The skewness in weighted moments. */
    public Skewness sk_Mweighted;

    /** The |I|psi>=lambda |psi> solution.
     */
    public final EVXData I;
public IstatesLocalized(final int n,
			final double p_offset,final double p_last,
			final double [] QQt,final double [] QQi,final double [] QQpi,final double [] QQp2i,final double [] QQp3i,
			final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			final double eps)
{
    this.p_offset=p_offset;
    this.p_last=p_last;
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final int ixH=n-1; // maximal eigenvalue index
    this.p_IH=I.getFPsi(ixH,QQpi)/I.getFPsi(ixH,QQi);

    final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	QQpnl=TMatr.sumAcB(QQpi,-p_last,QQi),pol=M.B.getKK(n,TMatr.mulMMM(n,IQQt,QQpnl,IQQt));
    if(eps>0){
	UnitTests.checkDoubleRel(I.getFPsi(ixH,QQpnl),"I.getFPsi(ixH,QQpnl)",I.sH*(p_IH-p_last),"I.sH*(p_IH-p_last)",eps);
	UnitTests.checkDoubleRel(M.B.getD0(pol,I.x0)/I.o_o,"pI0-pI_last",TMatr.getAMatrA(I.psi0,QQpnl,n),"<psi0|PnL|psi0>",eps);
    }
    final PolynomialRootsFinder.PRoots u=M.getPolynomialRootsFinderInBasis().findRoots(pol);
    final java.util.ArrayList<PriceMatchStorage> pm=new java.util.ArrayList<PriceMatchStorage>();
    for(int r=0;r<u.nroots;r++){
	if(Math.abs(u.im_root[r])<=EPS_IMROOT){
	    // real root, save localized at x=u.im_root[r] the state:
	    final double [] psi=TMatr.getMatr_x_A(n,IQQt,M.B.getX0Moments(n,u.re_root[r]));
	    TMatr.scaleVectorInPlace(psi,1/Math.sqrt(TMatr.getAMatrA(psi,QQt,n)));
	    pm.add(new PriceMatchStorage(u.re_root[r],Math.pow(TMatr.getAMatrB(psi,I.psi0,QQt,n),2),
					 TMatr.getAMatrA(psi,QQi,n),
					 TMatr.getAMatrA(psi,QQpi,n),
					 TMatr.getAMatrA(psi,QQp2i,n),
					 TMatr.getAMatrA(psi,QQp3i,n),
					 psi));
	    if(eps>0){
		final PriceMatchStorage xr=pm.get(pm.size()-1);
		UnitTests.checkDoubleRel(xr.sk.xa,"xr.sk.xa",p_last,"p_last",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(xr.psi_M,QQt,n),"<psi_M|psi_M>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(xr.psi_M,QQpnl,n),"<psi_M|QQpnl|psi_M>",0,"0",eps);
	    }
	}
	if(eps>0){
	    System.err.println("IstatesLocalized: ROOT=("+u.re_root[r]+","+u.im_root[r]+") pol(root)="+M.B.getD0(pol,u.re_root[r]));
	}
    }
    final double [] ONE=M.B.getONE();
    final double [] RONE=M.getRadauMultiplied(ONE);
    final boolean flag_dir_right=M.B.getD1(RONE,0)<0;
    // earliest in time to be at 0-th element index
    pm.sort(new java.util.Comparator<PriceMatchStorage>(){
	    public int compare(PriceMatchStorage o1, PriceMatchStorage o2){
		return Double.compare(o1.x,o2.x)*(flag_dir_right?1:-1);
	    }
	});
    this.priceMatches=pm.toArray(PriceMatchStorage[]::new);
    // averging the skewness answers
    double sskw=0,sw=0;
    for(int k=0;k<priceMatches.length;k++){
	sw+=priceMatches[k].i;
	sskw+=priceMatches[k].i*priceMatches[k].sk.sk;
    }
    sk_Iweighted=sskw/sw;

    // averaging moments in the pm states combined (corresponds to some density matrix).
    double spm0=0,spm1=0,spm2=0,spm3=0;
    for(int k=0;k<priceMatches.length;k++){
	final double [] psi=priceMatches[k].psi_M;
	spm0+=TMatr.getAMatrA(psi,QQi,n);
	spm1+=TMatr.getAMatrA(psi,QQpi,n);
	spm2+=TMatr.getAMatrA(psi,QQp2i,n);
	spm3+=TMatr.getAMatrA(psi,QQp3i,n);
    }
    sk_Mweighted=Skewness.getSkewnessFromMonomialMoments(spm0,spm1,spm2,spm3);


    if(eps>0){
	final double [] roots=new double [priceMatches.length],wr0=new double [priceMatches.length];
	for(int k=0;k<roots.length;k++) {
	    roots[k]=priceMatches[k].x;
	    wr0[k]=priceMatches[k].wr0;
	}
	System.err.println("IstatesLocalized: test OK I.sH="+I.sH+" x roots="+java.util.Arrays.toString(roots)+" flag_dir_right="+flag_dir_right+" wr0="+java.util.Arrays.toString(wr0)+" n="+n+" eps="+eps);
    }
}

public PriceMatchStorage getPriceMatchMaxI(){
    int indimax=-1;
    for(int r=0;r<priceMatches.length;r++){
	final double i=priceMatches[r].i;
	if(i>=I.s0 && (indimax<0 || i>=priceMatches[indimax].i)){
	    indimax=r;
	}
    }
    return (indimax<0)?null:priceMatches[indimax];
}

static final double EPS_IMROOT=1e-6;
}
