package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Quadrature;
import com.polytechnik.utils.RecurrenceAB;
import com.polytechnik.realtimebook.*;

/** Impact QVM calculations with the basis of second kind.
 *  @version $Id: LQVM.java,v 1.114 2022/08/08 20:10:47 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class LQVM<DTYPE> {

/** Weights and nodes for two basises.
 *  1-dv/dt = (wttnodes-witnodes)/wttnodes
 *  1-dt/dv = (wiinodes-wtinodes)/wiinodes
 *  dp/dt   = dptnodes/wttnodes
 *  dp/dv   = dpinodes/wiinodes
 */
final double [] xinodes,wiinodes,wtinodes,dpinodes,dsinodes,xtnodes,wttnodes,witnodes,dptnodes,dstnodes,dapinodes,dtinodes
    ;
final double [] dptnodes2,vsinodes2,dpinodes2;

/** Standard deviations.
 *  dp2t    = <dp/dt|dp/dt>_t
 *  dp2i    = <dp/dv|dp/dv>_i
 *  dw2t    = <1-dv/dt|1-dv/dt>_t
 *  dw2i    = <1-dt/dv|1-dt/dv>_i
 *  dpd2t   = <dp/dt-dp/dv|dp/dt-dp/dv>_t
 *  dpd2i   = <dp/dt-dp/dv|dp/dt-dp/dv>_i
 *  ds2t    = <dvs/dt|dvs/dt>_t
 *  ds2i    = <dvs/dv|dvs/dv>_i
 */
final double dp2t,dp2i,dw2t,dw2i,dpd2t,dpd2i,ds2t,ds2i,dt2i;
final double sigma2_dpi0,sigma2_dpt0,sigma2_dsi0,sigma2_dst0,dpi0,dpt0,dsi0,dst0,dti0,sigma2_dti0;
final double dp00,v00;

final DTYPE odata;
final DataInterfaceToMoments<DTYPE> access;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> basis;
LQVM(final int neq,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();
    this.basis=getBasis();

    final double [] mom=new double [2*neq+1];
    access.setM(mom,0,access.M_I,odata,mom.length);
    final double [][] p=basis.getOrthogonalPolynomialsFirstKind(neq+1,mom);
    final double [] tmom=new double [mom.length];
    access.setMT(tmom,tmom.length);

    final double [][] nw=getINodes(p);
    sortNodes(xinodes=nw[0],wiinodes=nw[1]);
    final double [] inodes=nw[2];
    wtinodes=new double [wiinodes.length];

    final double [][] nt=getTNodes(neq);
    sortNodes(xtnodes=nt[0],wttnodes=nt[1]);
    final double [] tnodes=nt[2];
    witnodes=new double [wttnodes.length];
    if(tnodes.length!=inodes.length) throw new IllegalArgumentException();

    final double [] mom1=new double [mom.length],moma=new double [mom.length],
	moms=new double [mom.length],momts=new double [mom.length],
	momtmp=new double [mom.length];
    // set dp
    access.setM(moma,1,access.M_DPPLUS,odata,mom.length);
    access.setM(mom1,1,access.M_DPMINUS,odata,mom.length);
    for(int k=mom.length;--k>=0;) moma[k]=moma[k]-mom1[k];
    dp00=moma[0];
    access.setM(mom1,1,access.M_DP,odata,mom1.length);
    access.setM(momts,0,access.M_TS,odata,mom1.length);
    access.setM(moms,0,access.M_ISELL,odata,moms.length);
    access.setM(momtmp,0,access.M_IBUY,odata,moms.length);
    v00=moms[0]+momtmp[0];for(int k=moms.length;--k>=0;) moms[k]=(moms[k]-momtmp[k])/v00;
    dpinodes=new double [xinodes.length];
    dsinodes=new double [xinodes.length];
    dtinodes=new double [xinodes.length];
    dapinodes=new double [xinodes.length];
    dptnodes=new double [xtnodes.length];
    dstnodes=new double [xtnodes.length];
    dptnodes2=new double [xtnodes.length];
    vsinodes2=new double [xtnodes.length];
    dpinodes2=new double [xtnodes.length];
    final double [] QQdp=basis.B.getQQMatr(xtnodes.length,mom1),
	QQvs=basis.B.getQQMatr(xtnodes.length,moms),
	QQv=basis.B.getQQMatr(xtnodes.length,mom),
	QQidp=TMatr.mulMMM(xtnodes.length,QQdp,com.polytechnik.utils.Linsystems.getInvertedMatrix(xtnodes.length,xtnodes.length,basis.B.getQQMatr(xtnodes.length,tmom)),QQv);

    if(xtnodes.length!=neq) throw new RuntimeException("xtnodes.length="+xtnodes.length+" neq="+neq);

    double sdp2t=0,sdp2i=0,sdw2t=0,sdw2i=0,sds2t=0,sds2i=0,sdt2i=0;
    final double [] stinodes=new double [xtnodes.length],sitnodes=new double [xinodes.length]; // to calculate dp2d
    for(int k=0;k<xinodes.length;k++)
    {
	double [] tmp=basis.B.sdiv1(tnodes,xtnodes[k]).q;
	if(tmp.length!=tnodes.length-1) throw new RuntimeException();
	final double ft=(basis.B.getD0(tmp,xtnodes[k]));
	dptnodes[k]=TMatr.mulVVn1(tmp,mom1)/ft;
	dstnodes[k]=TMatr.mulVVn1(tmp,moms)/ft;
	dptnodes2[k]=TMatr.getAMatrA(tmp,QQdp,xtnodes.length)/ft/ft/wttnodes[k];

	// like dv/dt
	witnodes[k]=TMatr.mulVVn1(tmp,mom)/ft/mom[0];
	// like (dp/dt)^2
	sdp2t+=Math.pow(dptnodes[k],2)/wttnodes[k];
	sds2t+=Math.pow(dstnodes[k],2)/wttnodes[k];
	// like (1-dv/dt)^2
	sdw2t+=Math.pow(wttnodes[k]-witnodes[k],2)/wttnodes[k];
	final double fft=(dptnodes[k]/wttnodes[k])/ft;
	for(int l=0;l<xtnodes.length;l++){
	    stinodes[l]+=fft*(basis.B.getD0(tmp,xinodes[l]));
	}

	tmp=basis.B.sdiv1(inodes,xinodes[k]).q;
	if(tmp.length!=inodes.length-1) throw new RuntimeException();
	final double fi=(basis.B.getD0(tmp,xinodes[k]));
	dpinodes[k]=TMatr.mulVVn1(tmp,mom1)/fi;
	dsinodes[k]=TMatr.mulVVn1(tmp,moms)/fi;
	dapinodes[k]=TMatr.mulVVn1(tmp,moma)/fi;
	dtinodes[k]=TMatr.mulVVn1(tmp,momts)/fi;

	vsinodes2[k]=TMatr.getAMatrA(tmp,QQvs,xinodes.length)/fi/fi/wiinodes[k];
	dpinodes2[k]=TMatr.getAMatrA(tmp,QQidp,xinodes.length)/fi/fi/(v00*wiinodes[k]);

	// like dt/dv
	wtinodes[k]=tmp[0]/fi;
	// like (dp/dv)^2
	sdp2i+=Math.pow(dpinodes[k],2)/wiinodes[k];
	sds2i+=Math.pow(dsinodes[k],2)/wiinodes[k];
	sdt2i+=Math.pow(dtinodes[k],2)/wiinodes[k];
	// like (1-dt/dv)^2
	sdw2i+=Math.pow(wiinodes[k]-wtinodes[k],2)/wiinodes[k];
	final double ffi=(dpinodes[k]/wiinodes[k])/fi;
	for(int l=0;l<xinodes.length;l++){
	    sitnodes[l]+=ffi*(basis.B.getD0(tmp,xtnodes[l]));
	}
    }
    this.dp2t=sdp2t;
    this.dp2i=sdp2i;
    this.dw2t=sdw2t;
    this.dw2i=sdw2i;
    this.ds2t=sds2t;
    this.ds2i=sds2i;
    this.dt2i=sdt2i;
    double sdpd2t=0,sdpd2i=0;
    for(int k=0;k<xinodes.length;k++)
    {
	sdpd2t+=dptnodes[k]*(dptnodes[k]/wttnodes[k]-2*sitnodes[k])+Math.pow(sitnodes[k],2)*wttnodes[k];
	sdpd2i+=dpinodes[k]*(dpinodes[k]/wiinodes[k]-2*stinodes[k])+Math.pow(stinodes[k],2)*wiinodes[k];
    }
    this.dpd2t=sdpd2t;
    this.dpd2i=sdpd2i;

    dpi0=dpinodes[0]/Math.sqrt(wiinodes[0]);
    sigma2_dpi0=dp2i-dpi0*dpi0;
    dpt0=dptnodes[0]/Math.sqrt(wttnodes[0]);
    sigma2_dpt0=dp2t-dpt0*dpt0;
    dsi0=dsinodes[0]/Math.sqrt(wiinodes[0]);
    sigma2_dsi0=ds2i-dsi0*dsi0;
    dst0=dstnodes[0]/Math.sqrt(wttnodes[0]);
    sigma2_dst0=ds2t-dst0*dst0;
    dti0=dtinodes[0]/Math.sqrt(wiinodes[0]);
    sigma2_dti0=dt2i-dti0*dti0;
}

abstract double getPOffset();
abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getBasis();
abstract void sortNodes(final double [] x,final double [] w);
abstract double [][] getINodes(final double [][] p);
static double [][] scaleWInd1(final double [][] nd)
{
    final double [] w=nd[1];
    final double s=TMatr.sumElements(w);
    for(int k=0;k<w.length;k++) {w[k]/=s;}
    return nd;
}
abstract double [][] getTNodes(final int n);
/** @returns true if x1 is after x2.
 */
abstract boolean isAfter(final double x1,final double x2);

static abstract class LQVMTest<DTYPE> {

final com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis M;
int neq;
long currenttime,tau;
Trade [] tb;
abstract void setData(final int tsize,final int neq);

abstract LQVM<DTYPE> getLQVM();
abstract void test_SQVM();
abstract void test_OLD_QVM();
abstract void test_Moments();
double [][] getNI(final int n,final double [] mom)
{
    final Quadrature.GaussQuadratureWithNodesPolynomialData q=
	M.getGaussQuadratureFromPolynomials(n,M.getOrthogonalPolynomialsFirstKind(n+1,mom));
    return scaleWInd1(new double [][]{q.x,q.w,q.nodespolynomial});
}



final String name;
final java.util.Random r;
final double eps;
LQVMTest(final String name,
	 final com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis M,
	 final java.util.Random r,final double eps)
{
    this.M=M;
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Common()
{
    final LQVM<DTYPE> q=getLQVM();
    //UnitTests.checkDoubleRel(q.dp2i,"q.dp2i",q.dp2t,"q.dp2t",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(q.dpinodes),"q.dpinodes",TMatr.sumElements(q.dptnodes),"q.dptnodes",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(q.wiinodes),"q.wiinodes",1,"1",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(q.witnodes),"q.witnodes",1,"1",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(q.wttnodes),"q.wttnodes",1,"1",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(q.wtinodes),"q.wtinodes",1,"1",eps);

    System.err.println(name+":LQVMTest:test_Commons(): test OK neq="+neq+" eps="+eps+" class="+q.getClass().getName());
}

void test_dp2i(final boolean flag_radau)
{
    final LQVM<DTYPE> q=getLQVM();
    final double [] mom_i=new double [2*neq+1],mom1=new double [mom_i.length];
    q.access.setM(mom1,1,q.access.M_DP,q.odata,mom1.length);
    q.access.setM(mom_i,0,q.access.M_I,q.odata,mom_i.length);
    final double [] mom_t=new double [mom_i.length];mom_t[0]=10;
    final double [][] pi,pt;
    if(flag_radau)
    {
	final RecurrenceAB i_ab=q.basis.getAB(q.basis.getOrthogonalPolynomialsFirstKind(neq+1,mom_i));
	final RecurrenceAB abi=q.basis.getABRadauModified(neq+1,i_ab);
	pi=q.basis.getOrthogonalPolynomialsFirstKindFromAB(abi);
	final RecurrenceAB t_ab=q.basis.getAB(q.basis.getOrthogonalPolynomialsFirstKind(neq+1,mom_t));
	final RecurrenceAB abt=q.basis.getABRadauModified(neq+1,t_ab);
	pt=q.basis.getOrthogonalPolynomialsFirstKindFromAB(abt);

    }
    else
    {
	pi=q.basis.getOrthogonalPolynomialsFirstKind(neq,mom_i);
	pt=q.basis.getOrthogonalPolynomialsFirstKind(neq,mom_t);
    }
    // normalize to 1
    for(int k=pt.length;--k>=0;)
    for(int l=pt[k].length;--l>=0;)
    {
	pt[k][l]/=pt[0][0];
	pi[k][l]/=pi[0][0];
    }

    test_D("dp2t dw2t dpd2t",q.dp2t,q.dw2t,q.dpd2t,pt,pi,mom1,mom_t,mom_i);
    test_D("dp2i dw2i dpd2i",q.dp2i,q.dw2i,q.dpd2i,pi,pt,mom1,mom_i,mom_t);

    System.err.println(name+":LQVMTest:test_dp2i(): test OK neq="+neq+" eps="+eps+" class="+q.getClass().getName());
}

private void test_D(final String txt,
		    final double test_dp2,final double test_dw2,final double test_dpd2,
		    final double [][] pt,final double [][] pv,
		    final double [] mom_p,final double [] mom_t,final double [] mom_v)
{
    final double [] dp_pt=new double [neq],dv=new double [neq],dp_pv=new double [neq];
    for(int k=0;k<neq;k++)
    {
	dp_pt[k]=TMatr.mulVVn(k+1,pt[k],mom_p);
	dv[k]=TMatr.mulVVn(k+1,pt[k],mom_v)/mom_v[0];
	dp_pv[k]=TMatr.mulVVn(k+1,pv[k],mom_p);
    }

    UnitTests.checkDoubleRel(test_dp2,"test_dp2",TMatr.mulVV(dp_pt,dp_pt),"<dp|dp>_t",eps);
    UnitTests.checkDoubleRel(test_dw2,"test_dw2",TMatr.mulVV(dv,dv)-1,"<1-dv/dt|1-dv/dt>_t",eps);

    double dpd2=0;
    for(int k=neq;--k>=0;)
    {
	dpd2+=dp_pt[k]*dp_pt[k];
	final double [] tmp=TMatr.expandOverLowerDiagBasis(pv[k],pt);
	for(int l=k;l>=0;l--) dpd2+=-2*dp_pt[l]*dp_pv[k]*tmp[l];
	for(int m=neq;--m>=0;)
	{
	    final double [] tmp2=TMatr.expandOverLowerDiagBasis(pv[m],pt);
	    for(int l=Math.min(m,k);l>=0;l--) dpd2+=dp_pv[k]*tmp[l]*tmp2[l]*dp_pv[m];
	}
    }
    UnitTests.checkDoubleRel(test_dpd2,"test_dpd2",dpd2,"<dp/dt-dp/dv|dp/dt-dp/dv>_t",eps);

    System.err.println(name+":LQVMTest:test_D():"+txt+": test <dp|dp> <1-dv/dt|1-dv/dt> <dp/dt-dp/dv|dp/dt-dp/dv> OK, neq="+neq+" eps="+eps);
}


void test_Monomials(final double [] mom_i,final double [] mom_t,final double [] mom_dp,final double [] mom_s)
{
    final LQVM<DTYPE> q=getLQVM();
    final double [][] ni=getNI(neq,mom_i);
    final double [] xi,wii,wti;
    q.sortNodes(xi=ni[0],wii=ni[1]);
    wti=M.getWeightsByIntegration(ni[2],xi,TMatr.getVectorScaled(mom_t,1/mom_t[0]));
    final double [] dpinodes=M.getWeightsByIntegration(ni[2],xi,mom_dp);
    UnitTests.checkDoubleArrayRel(q.xinodes,"q.xinodes",xi,"xi",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(wii),"sum(wii)",1,"1",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(wti),"sum(wti)",1,"1",eps);
    UnitTests.checkDoubleArrayRel(q.wiinodes,"q.wiinodes",wii,"wii",eps);
    UnitTests.checkDoubleArrayRel1(q.wtinodes,"q.wtinodes",wti,"wti",eps);
    UnitTests.checkDoubleArrayRel1(q.dpinodes,"q.dpinodes",dpinodes,"dpinodes",eps);

    final double [][] nt=getNI(neq,mom_t);
    final double [] xt,wtt,wit;
    q.sortNodes(xt=nt[0],wtt=nt[1]);
    wit=M.getWeightsByIntegration(nt[2],xt,TMatr.getVectorScaled(mom_i,1/mom_i[0]));
    final double [] dptnodes=M.getWeightsByIntegration(nt[2],xt,mom_dp);
    UnitTests.checkDoubleArrayRel(q.xtnodes,"q.xtnodes",xt,"xt",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(wtt),"sum(wtt)",1,"1",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(wit),"sum(wit)",1,"1",eps);
    UnitTests.checkDoubleArrayRel(q.wttnodes,"q.wttnodes",wtt,"wtt",eps);
    UnitTests.checkDoubleArrayRel1(q.witnodes,"q.witnodes",wit,"wit",eps);
    UnitTests.checkDoubleArrayRel1(q.dptnodes,"q.dptnodes",dptnodes,"dptnodes",eps);

    // check direct calculations, bypassing orthogonal polynomials basis
    final double [] dpdt=DeprecatedHelperEVXData.getLinsysSol(neq,TMatr.getVectorScaled(mom_t,1/mom_t[0]),0,mom_dp,0),
	dpdv=DeprecatedHelperEVXData.getLinsysSol(neq,TMatr.getVectorScaled(mom_i,1/mom_i[0]),0,mom_dp,0),
	dsdv=DeprecatedHelperEVXData.getLinsysSol(neq,TMatr.getVectorScaled(mom_i,1/mom_i[0]),0,mom_s,0),
	dvdt=DeprecatedHelperEVXData.getLinsysSol(neq,TMatr.getVectorScaled(mom_t,1/mom_t[0]),0,TMatr.getVectorScaled(mom_i,1/mom_i[0]),0),
	dtdv=DeprecatedHelperEVXData.getLinsysSol(neq,TMatr.getVectorScaled(mom_i,1/mom_i[0]),0,TMatr.getVectorScaled(mom_t,1/mom_t[0]),0);
    for(int k=0;k<neq;k++)
    {
	UnitTests.checkDoubleRel(Polynomials.getD0(dpdv,xi[k]),"dpdv(xi[k])",q.dpinodes[k]/q.wiinodes[k],"q.dpinodes[k]/q.wiinodes[k]",eps);
	UnitTests.checkDoubleRel(Polynomials.getD0(dsdv,xi[k]),"dsdv(xi[k])",q.dsinodes[k]/q.wiinodes[k],"q.dsinodes[k]/q.wiinodes[k]",eps);
	UnitTests.checkDoubleRel(Polynomials.getD0(dtdv,xi[k]),"dtdv(xi[k])",q.wtinodes[k]/q.wiinodes[k],"q.wtinodes[k]/q.wiinodes[k]",eps);
	UnitTests.checkDoubleRel(Polynomials.getD0(dpdt,xt[k]),"dpdt(xt[k])",q.dptnodes[k]/q.wttnodes[k],"q.dptnodes[k]/q.wttnodes[k]",eps);
	UnitTests.checkDoubleRel(Polynomials.getD0(dvdt,xt[k]),"dvdt(xt[k])",q.witnodes[k]/q.wttnodes[k],"q.witnodes[k]/q.wttnodes[k]",eps);
    }

  if(neq>1)
  {
    final double [] diff=new double [neq];TMatr.sumP1(diff,dpdv,1);TMatr.sumP1(diff,dpdt,-1);
    UnitTests.checkDoubleRel(TMatr.mulVVn1(Polynomials.mult2Pol(dpdv,dpdv),mom_i)/mom_i[0],"<dp/dv|dp/dv>_i",q.dp2i,"q.dp2i",eps);
    UnitTests.checkDoubleRel(TMatr.mulVVn1(Polynomials.mult2Pol(dsdv,dsdv),mom_i)/mom_i[0],"<ds/dv|ds/dv>_i",q.ds2i,"q.ds2i",eps);
    UnitTests.checkDoubleRel(TMatr.mulVVn1(Polynomials.mult2Pol(dpdt,dpdt),mom_t)/mom_t[0],"<dp/dt|dp/dt>_t",q.dp2t,"q.dp2t",eps);
    UnitTests.checkDoubleRel(TMatr.mulVVn1(Polynomials.mult2Pol(diff,diff),mom_i)/mom_i[0],"<dp/dv-dp/dt|dp/dv-dp/dt>_i",q.dpd2i,"q.dpd2i",eps);
    UnitTests.checkDoubleRel(TMatr.mulVVn1(Polynomials.mult2Pol(diff,diff),mom_t)/mom_t[0],"<dp/dv-dp/dt|dp/dv-dp/dt>_t",q.dpd2t,"q.dpd2t",eps);
    java.util.Arrays.fill(diff,0);TMatr.sumP1(diff,dtdv,1);diff[0]-=1;
    UnitTests.checkDoubleRel(TMatr.mulVVn1(Polynomials.mult2Pol(diff,diff),mom_i)/mom_i[0],"<1-dt/dv|1-dt/dv>_i",q.dw2i,"q.dw2i",eps);
    java.util.Arrays.fill(diff,0);TMatr.sumP1(diff,dvdt,1);diff[0]-=1;
    UnitTests.checkDoubleRel(TMatr.mulVVn1(Polynomials.mult2Pol(diff,diff),mom_t)/mom_t[0],"<1-dv/dt|1-dv/dt>_t",q.dw2t,"q.dw2t",eps);
  }

    System.err.println(name+":LQVMTest:test_Monomials(): test OK neq="+neq+" eps="+eps+" class="+q.getClass().getName());
}

void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":LQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_OLD_QVM();
    test_SQVM();
    test_Moments();

    System.err.println(name+":LQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":LQVMTest: starting tests; ver=$Id: LQVM.java,v 1.114 2022/08/08 20:10:47 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<118;i++)
	{
	    test_Calculations(7250+r.nextInt(7550),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":LQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: LQVM.java,v 1.114 2022/08/08 20:10:47 mal Exp $");
}
} // END OF class LQVMTest



}// END OF class LQVM
