package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.ReadWriteUtils;
import com.polytechnik.realtimebook.*;

/** Test whether the John Pierpont Morgan was right.
 *  @version $Id: LwQVM.java,v 1.32 2022/06/27 07:19:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class LwQVM<DTYPE> {

final DTYPE odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;

EVXData I;
PEVar tip_IL;
ProbVar itpi_IL;
final FVar itPm_sam;

LwQVM(final int n,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();

    {
    final int nmoms=2*n-1;
    final double [] imom_sam=new double [nmoms],pimom_sam=new double [nmoms],
	tmom_sam=new double [nmoms],ptmom_sam=new double [nmoms],
	pertmom=getX0moms(nmoms);
    access.setM(imom_sam,0,access.M_I,odata,nmoms);
    access.setM(pimom_sam,1,access.M_I,odata,nmoms);
    access.setM(tmom_sam,0,access.M_T,odata,nmoms);
    access.setM(ptmom_sam,1,access.M_T,odata,nmoms);
    this.itPm_sam=new FVar(n,imom_sam,tmom_sam,pertmom,pimom_sam,ptmom_sam);
    }

    lateInit(n);

}
void lateInit(final int n)
{
    final int nmoms=2*n-1;
    final double [][] QQ=access.getQQ_Matrs(n,1,getMoments(nmoms));
    final double [] QQt=QQ[0],QQi=QQ[1],QQpt=QQ[2],QQpi=QQ[3];
    this.I=DeprecatedHelperEVXData.getEVXDataFromQQ(n,QQi,QQt,access,getX0());
    this.tip_IL=new PEVar(this.I.psiL,QQt,QQi,QQpt,0,1);
    this.itpi_IL=new ProbVar(this.I,0,QQt,QQi,QQpi);
}

double getD1G(final int ind_lev,final EVXData J,final double [] QQ)
{
    final double [] psi=J.getPsi(ind_lev),pert=DeprecatedHelperEVXData.getPsiPertr(J,ind_lev,QQ);
    return getK2(pert,psi,psi,psi)-getK2(psi,psi,pert,psi);
}

double [] getX0moms(final int nmoms)
{
    return DeprecatedHelperEVXData.getX0Moms(nmoms,access,getX0());
}


class PEVar {
final EVXData J;
final double [] f,varf,senspsi,sensft;
PEVar(final double [] psiB,final double [] QQt,final double [] QQi,final double [] QQft,final int indt,final int indi)
{
    final int n=psiB.length,nmoms=2*n-1;
    final double [][] SQQ=access.getQQ_Matrs(n,1,getSMoments(nmoms,psiB,psiB));
    final double [] SQQt=SQQ[indt],SQQi=SQQ[indi];
    this.J=DeprecatedHelperEVXData.getEVXDataFromQQ(n,SQQt,QQt,access,getX0());
    this.f=new double [n];
    this.varf=new double [n];
    this.sensft=new double [n];
    this.senspsi=new double [n];
    final double [] QQpert=new double [n*n];
    for(int k=0;k<n;k++)
    {
	final double ek=this.J.ev[k];
	this.f[k]=TMatr.getAMatrB(n,this.J.ev,n+k*n,this.J.ev,n+k*n,QQft,n);
	this.sensft[k]=getD1G(k,this.J,QQft);
	double s=0;
	for(int l=0;l<n;l++)
	{
	    if(l==k) continue;
	    final double el=this.J.ev[l];
	    s+=((TMatr.getAMatrB(n,this.J.ev,n+k*n,this.J.ev,n+l*n,SQQi,n)-ek*TMatr.getAMatrB(n,this.J.ev,n+k*n,this.J.ev,n+l*n,QQi,n))/(ek-el))*
		TMatr.getAMatrB(n,this.J.ev,n+k*n,this.J.ev,n+l*n,QQft,n);
	}
	this.varf[k]=2*s;
	for(int l=QQpert.length;--l>=0;){QQpert[l]=SQQi[l]-ek*QQi[l];}
	this.senspsi[k]=getD1G(k,this.J,QQpert);
    }
}

void match_LkQVM_PEData(final LkQVM<?>.PEData s,final double eps)
{
   UnitTests.checkDoubleArrayRel(this.f,"this.f",s.f,"s.f",eps);
   UnitTests.checkDoubleArrayRel(this.varf,"this.varf",s.varf,"s.varf",eps);
   UnitTests.checkDoubleArrayRel(this.senspsi,"this.senspsi",s.senspsi,"s.senspsi",eps);
   UnitTests.checkDoubleArrayRel(this.sensft,"this.sensft",s.sensft,"s.sensft",eps);
   EVXData.matchEVXData(this.J,s.J,false,false,eps);
   System.err.println("LwQVM:PEVar:match_LkQVM_PEData: test OK neq="+J.neq+" eps="+eps);
}
} // END OF class PEVar



class ProbVar {
final EVXData J,I;
final double [] probpert,probsens,fi,senspsi;
ProbVar(final EVXData I,final int ind_lev,final double [] QQt,final double [] QQi,final double [] QQipert)
{
    this.I=I;
    final int n=I.neq,nmoms=2*n-1;
    final double [] psi=I.getPsi(ind_lev),psipert=DeprecatedHelperEVXData.getPsiPertr(I,ind_lev,QQipert);
    final double [][] QQ=access.getQQ_Matrs(n,1,new double [][]{getSTMoments(nmoms,psi,psi),getSTMoments(nmoms,psi,psipert)});
    final double [] QQj=QQ[0],QQjpert=TMatr.scaleVectorInPlace(QQ[1],2);
    this.J=DeprecatedHelperEVXData.getEVXDataFromQQ(n,QQj,QQt,access,getX0());
    this.probpert=new double [n];
    this.probsens=new double [n];
    this.fi=new double [n];
    this.senspsi=new double [n];
    for(int k=0;k<n;k++)
    {
	this.probpert[k]=this.J.getFPsi(k,QQjpert);
	this.probsens[k]=getD1G(k,this.J,QQjpert);
	this.fi[k]=this.J.getFPsi(k,QQipert)/this.J.getFPsi(k,QQi);
	this.senspsi[k]=getD1G(k,I,QQipert);
    }
}
void match_LkQVM_PEData(final LkQVM<?>.PEData s,final double sp_offset,final double eps)
{
   final int n=this.J.neq;
   EVXData.matchEVXData(this.J,s.J,false,false,eps);

   final double dof=getPOffset()-sp_offset;
   for(int k=0;k<n;k++)
   {
       UnitTests.checkDoubleRel(this.fi[k],"this.fi[k]",s.fi[k]-dof,"s.fi[k]-dof",eps);
   }

   System.err.println("LwQVM:ProbVar:match_LkQVM_PEData: test OK n="+n+" eps="+eps);
}
void match_MkQVM_FPertPsi(final MkQVM<?>.FPertPsi s,final double eps)
{
   final int n=this.J.neq;
   EVXData.matchEVXData(this.I,s.J,false,false,eps);
   UnitTests.checkDoubleArrayRel(this.senspsi,"this.senspsi",s.senspsi,"s.senspsi",eps);

   System.err.println("LwQVM:ProbVar:match_MkQVM_FPertPsi: test OK n="+n+" eps="+eps);
}
void match_LkQVM_ProbVarData(final LkQVM<?>.ProbVarData s,final double sp_offset,final double eps)
{
   final int n=this.J.neq;
   EVXData.matchEVXData(this.I,s.I,false,false,eps);
   EVXData.matchEVXData(this.J,s.J,false,false,eps);

   final double dof=getPOffset()-sp_offset;
   for(int k=0;k<n;k++)
   {
       UnitTests.checkDoubleRel(this.fi[k],"this.fi[k]",s.fi[k]-dof,"s.fi[k]-dof",eps);
   }
   UnitTests.checkDoubleArrayRel(this.probpert,"this.probpert",s.probpert,"s.probpert",eps);
    UnitTests.checkDoubleRel(this.senspsi[s.ind_lev],"this.senspsi[s.ind_lev]",s.d1G,"s.d1G",eps);

   System.err.println("LwQVM:ProbVar:match_LkQVM_ProbVarData: test OK n="+n+" eps="+eps);
}
} // END OF class ProbVar




class FVar {
final EVXData I;
final double [] varf,varFn;
FVar(final int n,final double [] imom,final double [] tmom,final double [] pertmom,final double [] fimom,final double [] ftmom)
{
    final double [][] QQ=access.getQQ_Matrs(n,1,new double [][]{imom,tmom,pertmom,fimom,ftmom});
    final double [] QQi=QQ[0],QQt=QQ[1],QQpert=QQ[2],QQfi=QQ[3],QQft=QQ[4];
    this.I=DeprecatedHelperEVXData.getEVXDataFromQQ(n,QQi,QQt,access,getX0());
    this.varf=new double [n];
    this.varFn=new double [n];
    for(int k=0;k<n;k++)
    {
	final double ek=this.I.ev[k];
	final double [] psi=this.I.getPsi(k),psipert=DeprecatedHelperEVXData.getPsiPertr(this.I,k,QQpert);
	this.varf[k]=TMatr.getAMatrB(psi,psipert,QQfi,n)-ek*TMatr.getAMatrB(psi,psipert,QQft,n);
	this.varf[k]*=2;
	this.varFn[k]=this.varf[k]/TMatr.getAMatrB(psi,psi,QQpert,n);
    }
}

void match_MkQVM_FPertPsi(final MkQVM<?>.FPertPsi s,final double eps)
{
   final int n=this.I.neq;
   EVXData.matchEVXData(this.I,s.J,false,false,eps);
   UnitTests.checkDoubleArrayRel(this.varf,"this.varf",s.varf,"s.varf",eps);
   UnitTests.checkDoubleArrayRel(this.varFn,"this.varFn",s.varFn,"s.varFn",eps);

   System.err.println("LwQVM:FVar:match_MkQVM_FPertPsi: test OK n="+n+" eps="+eps);
}

} // END OF class FVar




double [][] getMoments(final int nmoms)
{
    final double [] Ptmom=new double [nmoms],dpmom=new double [nmoms],tmom=new double [nmoms],imom=new double [nmoms],pimom=new double [nmoms];

    access.setM(dpmom,1,access.M_DP,odata,nmoms);
    access.setFMoments(Ptmom,dpmom,nmoms);
    access.setM(imom,0,access.M_I,odata,nmoms);
    access.setM(pimom,1,access.M_I,odata,nmoms);
    access.setMT(tmom,nmoms);

    return new double [][]{tmom,imom,Ptmom,pimom};
}

abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract double getX0();
abstract double getPLast();
abstract double getPOffset();
abstract double [][] getSMoments(final int nmoms,final double [] psiS1,final double [] psiS2);
abstract double [] getSTMoments(final int nmoms,final double [] psiS1,final double [] psiS2);
abstract double getK2(final double [] a,final double [] aa,final double [] b,final double [] bb);

static abstract class LwQVMTest<DTYPE> extends QTest {

LwQVMTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}

abstract LwQVM<DTYPE> getLwQVM();
abstract LkQVM<?> getLkQVM();
abstract MkQVM<?> getMkQVM();

void test_Common()
{
    final int n=neq,nmoms=2*n-1,ntests=1000;
    final LwQVM<DTYPE> q=getLwQVM();
    for(int t=0;t<ntests;t++)
    {
	final double [] psik=new double [n-r.nextInt(n)],
	    psil=new double [n-r.nextInt(n)];
	for(int k=0;k<psik.length;k++) psik[k]=r.nextDouble();
	for(int k=0;k<psil.length;k++) psil[k]=r.nextDouble();
	UnitTests.checkDoubleArrayRel(q.getSMoments(nmoms,psik,psil)[0],"q.getSMoments(nmoms,psik,psil)[0]",q.getSTMoments(nmoms,psik,psil),"q.getSTMoments(nmoms,psik,psil)",eps);
    }

    EVXData.matchEVXData(q.tip_IL.J,q.itpi_IL.J,true,true,eps);

    System.err.println(name+":LwQVMTest:test_Common(): test OK neq="+neq+" ntests="+ntests+" eps="+eps);
}
void test_LkQVM()
{
    final int n=neq;
    final LwQVM<DTYPE> q=getLwQVM();
    final LkQVM<?> s=getLkQVM();
    q.tip_IL.match_LkQVM_PEData(s.pedt_IL,eps);
    q.itpi_IL.match_LkQVM_PEData(s.pedt_IL,s.getPOffset(),eps);
    q.itpi_IL.match_LkQVM_ProbVarData(s.itpi_IL,s.getPOffset(),eps);

    System.err.println(name+":LwQVMTest:test_LkQVM(): test OK n="+n+" eps="+eps);
}
void test_MkQVM()
{
    final int n=neq;
    final LwQVM<DTYPE> q=getLwQVM();
    final MkQVM<?> s=getMkQVM();
    EVXData.matchEVXData(q.I,s.I,false,false,eps);
    q.itpi_IL.match_MkQVM_FPertPsi(s.itPvF,eps);
    q.itPm_sam.match_MkQVM_FPertPsi(s.itPm_sam,eps);

    System.err.println(name+":LwQVMTest:test_MkQVM(): test OK n="+n+" eps="+eps);
}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":LwQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_LkQVM();
    test_MkQVM();

    System.err.println(name+":LwQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":LwQVMTest: starting tests; ver=$Id: LwQVM.java,v 1.32 2022/06/27 07:19:13 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":LwQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: LwQVM.java,v 1.32 2022/06/27 07:19:13 mal Exp $");
}
} // END OF class LwQVMTest




}// END OF class LwQVM
