package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;

/** A class to store parameters for model prediction.
 *  In a future - read per stock parameters based on
 *  statistical analysis. For now - the same for all.
 *  @version $Id: ModelParameters.java,v 1.298 2019/02/27 20:35:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class ModelParameters {

    static final boolean PRINT_DATA_DEBUG=true;

    static final long DISTANCE_TIME_SCALE=TraderInterface.TIME_SCALE/1000L;
    static final long T_SUMM_SCALE=TraderInterface.TIME_SCALE/100000L; // 0.01 ms

    /* Use 0.75% as a single global threshold for valid orders.
     */
    private static final double [] price_validorders_thresholds_ratio_global=new double []{1.0,1.0001,1.00025,1.0005,1.001,1.002,1.003,1.005,1.01};
    /* All possible delays.
     */
    private static final long [] tau_embedding_global=new long []{
	2*TraderInterface.TIME_SCALE,
	4*TraderInterface.TIME_SCALE,
	8*TraderInterface.TIME_SCALE,16*TraderInterface.TIME_SCALE,32*TraderInterface.TIME_SCALE,64*TraderInterface.TIME_SCALE,128*TraderInterface.TIME_SCALE,256*TraderInterface.TIME_SCALE ,512*TraderInterface.TIME_SCALE,1024*TraderInterface.TIME_SCALE
	//128*TraderInterface.TIME_SCALE
        //,2048*TraderInterface.TIME_SCALE
    };

    private static final long [] tau_embedding_global_consequent=new long []{
//	2*TraderInterface.TIME_SCALE,
//	16*TraderInterface.TIME_SCALE,
//	32*TraderInterface.TIME_SCALE,
//	64*TraderInterface.TIME_SCALE,
	128*TraderInterface.TIME_SCALE,
	256*TraderInterface.TIME_SCALE
    };


    private static final int NEQ_T_DEFAULT=12;
    static final double TRY_PCUT=0.5;


    public static class ModelParametersData {
	final boolean FLAG_NO_BOOK=false;


	final double MINIMAL_SPREAD=0.01;  // 1 cent is typical for nasdaq


	// discretization
	final double MEDIAN_LEVEL=1.00001,MINMAX_LEVEL=1.00001,DTAU_LEVEL=1.00001;
	/** Equal to zero _DIV means use fixed level.
	 */
	final int FRAC_MEDIAN_DIV=3,FRAC_MINMAX_DIV=3,FRAC_DTAU_DIV=3;


	// training, 1 sec for now
	final long tau_training_quant=TraderInterface.TIME_SCALE /* /10 */ /* 16 */ /* 0L */;

	final long [] tau_embedding=tau_embedding_global;
	final long [] tau_embedding_consequent=tau_embedding_global_consequent; // for now

	final long tau_max_back_antecedent=tau_embedding[tau_embedding.length-1];

	// what data to keep
	final long tau_readdata_clearBefore_delay=
	    //(3600+3600/2)*TraderInterface.TIME_SCALE+ // 1.5 hour
	    2*tau_max_back_antecedent+ // window size
	    3600*TraderInterface.TIME_SCALE/1; // 5minutes

	// a threshold for book valid trade
	final double good_price_book_range_filter_antecedent=1.3; // test 1% FIXME // 10% - almost everything.
	// the number of different prices.
	final int good_price_n_points_max=10;


	final int VOLUME_FACTOR=1;


	final int N_TLEV=12;

	final int N_TRADES_MIN=20;

	private boolean checked=false;


	final int MAX_LONG,MAX_SHORT;

	final boolean flag_active;
	final String read_knowledgedir,write_knowledgedir;
	final double [] scales=price_validorders_thresholds_ratio_global;


	final double thr_enter,thr_exit,thrw_enter,thrw_exit;
	final long tau;
	final int neq_t;

	ModelParametersData(final String read_knowledgedir,
			    final String write_knowledgedir,
			    final boolean flag_active,
			    final int MAX_LONG,
			    final int MAX_SHORT,
			    final String ticker,

			    final double thr_enter,
			    final double thr_exit,
			    final double thrw_enter,
			    final double thrw_exit,
			    final long tau,
			    final int neq_t)
	{
	    this.read_knowledgedir=read_knowledgedir;
	    this.write_knowledgedir=write_knowledgedir;
	    this.flag_active=flag_active;
	    this.MAX_LONG=MAX_LONG;
	    this.MAX_SHORT=MAX_SHORT;


	    this.thr_enter=thr_enter;
	    this.thr_exit=thr_exit;
	    this.thrw_enter=thrw_enter;
	    this.thrw_exit=thrw_exit;
	    this.tau=tau;
	    this.neq_t=neq_t;

	    if(PRINT_DATA_DEBUG && this.flag_active)
	    {
		System.err.println("=================== INITIALIZE "+ticker+" active settings");
		System.err.println("\tread_knowledgedir="+read_knowledgedir);
		System.err.println("\twrite_knowledgedir="+write_knowledgedir);
		System.err.println("\tflag_active="+flag_active);
		System.err.println("\tFLAG_NO_BOOK="+this.FLAG_NO_BOOK);
		System.err.println("\tMAX_LONG="+MAX_LONG);
		System.err.println("\tMAX_SHORT="+MAX_SHORT);

		System.err.println("\tthr_enter="+this.thr_enter);
		System.err.println("\tthr_exit="+this.thr_exit);
		System.err.println("\tthrw_enter="+this.thrw_enter);
		System.err.println("\tthrw_exit="+this.thrw_exit);
		System.err.println("\ttau="+this.tau);
		System.err.println("\tneq_t="+this.neq_t);
		if(this.tau_embedding==null)
		{
		    System.err.println("\ttau_embedding=null");
		}
		else
		{
		    System.err.print("\ttau_embedding=");
		    for(long t:this.tau_embedding)
		    {
			System.err.print(String.valueOf(t)+";");
		    }
		    System.err.println();
		}

	    }
	}


    double [] getBuyLevels(final double scale)
    {
	final double [] levels=new double [this.scales.length];
	for(int i=levels.length;i-->0;)
	{
	    levels[i]=scale/this.scales[i];
	}
	return levels;
    }


    double [] getSellLevels(final double scale)
    {
	final double [] levels=new double [this.scales.length];
	for(int i=levels.length;i-->0;)
	{
	    levels[i]=scale*this.scales[i];
	}
	return levels;
    }

	double [] getDLevels(final double scale)
    {
	final double [] levels=new double [this.scales.length];
	for(int i=levels.length;i-->0;)
	{
	    levels[i]=scale*(this.scales[i]-1);
	}
	return levels;
    }


	String [] getLevelsIDs()
	{
	    final String [] id=new String [this.scales.length];
	    for(int i=id.length;i-->0;)
	    {
		id[i]=String.valueOf(this.scales[i]);
	    }
	    return id;
	}


	double [][] getDeltaOKAndFailure(final double p_buy,
					 final double p_sell)
	{
	    final double F_RATIO=3;
	    final double av=0.5*(p_buy+p_sell);
	    final double [] l_ok=getSellLevels(1.0);
	    final int n=l_ok.length;
	    final double [] d_ok=new double [n],d_fail=new double [n];
	    for(int i=0;i<n;i++)
	    {
		d_ok[i]=(l_ok[i]-1)*av;
		d_fail[i]=d_ok[i]/F_RATIO;
	    }
	    return new double [][]{d_ok,d_fail};
	}
    }



    private final java.util.HashMap<String,ModelParametersData> specialparams;
    final String out_dir;
    public ModelParameters(final String read_knowledgedir,
			   final String write_knowledgedir,
			   final String [] only_allowed_stocks)
	throws java.io.IOException
    {
	this.specialparams=new java.util.HashMap<String,ModelParametersData>(4001,0.3f);
	this.out_dir=write_knowledgedir;
	for(String s : only_allowed_stocks)
	{
	    final ModelParametersData m=
		new ModelParametersData(
		    read_knowledgedir,
		    write_knowledgedir,
		    true,
		    100,
		    100,
		    s,
		    Double.NaN,
		    Double.NaN,
		    Double.NaN,
		    Double.NaN,
		    -1,
		    NEQ_T_DEFAULT);
	    this.specialparams.put(s,m);
	}
    }

    static double getDouble(final String key,
			    final java.util.ResourceBundle data)
	throws java.io.IOException
    {
	final String d=data.getString(key);
	if(d.length()<=2||d.charAt(0)!='"'||d.charAt(d.length()-1)!='"')
	{
	    throw new IllegalArgumentException(d);
	}
	return Double.parseDouble(d.substring(1,d.length()-1));
    }


    public ModelParametersData getParameters(final String ticker)
    {
	final ModelParametersData m=this.specialparams.get(ticker);
	if(m!=null)
	{
	    if(!m.checked)
	    {
		m.checked=true;
		// to avoid keys duplication
		this.specialparams.put(ticker,m);
	    }
	    return m;
	}

	// skip all except selected
	return new ModelParametersData(null,null,false,100,100,ticker,
				       Double.NaN,
				       Double.NaN,
				       Double.NaN,
				       Double.NaN,
				       -1,-1); // default parameters.
    }


    public ModelParameters(final String configfile)
	throws java.io.IOException
    {
	final ConfigurationReader.CommonParameters c=
	    ConfigurationReader.readConfig(configfile);
	this.specialparams=c.data;
	this.out_dir=c.OUTPUT_DIR;
    }

}
