package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.ProbabilityCorrelation;

/**
 * A simple try of I(p) asymmetry. The expansion I(p+d) on |psi> state give a polynomial on d, that has:
 * 0-th order equal to lambda,
 * 1-st order is zero as we varying on an eigenstate
 * 2-nd order has constant sign because of min/max eigenstate used and
 * 3-rd order carry directional information!
 * It is rather difficult to obtain Rayleigh quotient for third order variation in general case,
 * but if: d=2, psi=psi(p), and \delta psi=d/dp psi(p),
 * the answer becomes a very simple (because then \delta \delta psi=0);
 * (I still do not have an answer in general case d>2).
 * There is a relation of the dinamic impact asymmetry beta
 * to probability correlation, corr=(beta^{[IL]}-beta^{[IH]})/(lambda^[1]_Dp-lambda^[0]_Dp).
 * See ProbabilityCorrelation code for corr correlation calculion without beta.
 * @version $Id: Mu2P2Asymmetry.java,v 1.56 2022/07/28 08:04:53 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class Mu2P2Asymmetry {

/** For testing, better use ProbabilityCorrelation pc.
 */
final MuRQ mL,mH;
final double betaS,IbetaS;
public final ProbabilityCorrelation pc,pci,pcA,pcDpA,pcDA;
Mu2P2Asymmetry(final double p0t,final double p1t,final double p2t,
	       final double p0i,final double p1i,final double p2i,
	       final double eps)
{
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=true? // regular monomials basis
	com.polytechnik.utils.DiffSkewness.MB:
	new com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis();
    // In price space. Regular differentiation.
    final VolMatrix varpsi=M.B::getDifferentiated;
    final int d=2;
    final double [] Mi=new double []{p0i,p1i,p1i,p2i}, // <I p^j p^k>
	Mt=new double []{p0t,p1t,p1t,p2t} // <p^j p^k>
    ;
    // Probability correlation of I and d/dp operators.
	this.pc=new ProbabilityCorrelation(Mt,varpsi.getDiffMatrix(d,Mt),Mi,M,true);
	this.pci=new ProbabilityCorrelation(Mt,varpsi.getDiffMatrix(d,Mi),Mi,M,true);
    final double [] Ma=TMatr.sumAcB(Mi,-Mi[0]/Mt[0],Mt);
    this.pcA=new ProbabilityCorrelation(Mt,Ma,Mi,M,true);
    this.pcDpA=new ProbabilityCorrelation(Mt,varpsi.getDiffMatrix(d,Mt),Ma,M,true);
    this.pcDA=new ProbabilityCorrelation(Mt,varpsi.getDiffMatrix(d,Ma),Mi,M,true);

    // Second approach. Should give the same result.
    final double [] ev=M.EV.getGEV_Generalized(d,Mi,Mt,true); // eigenvalues problem
    this.mL=new MuRQ(EVXData.getPsiFromBasis(0,ev,d), // eigenvector low I
		     ev[0],Mi,Mt,M,varpsi,eps);
    this.mH=new MuRQ(EVXData.getPsiFromBasis(d-1,ev,d), // eigenvector high I
		     ev[d-1],Mi,Mt,M,varpsi,eps);
    betaS=mL.beta+mH.beta;
    IbetaS=mL.lam*mL.beta+mH.lam*mH.beta;
    if(eps>0)
    {
	UnitTests.checkDoubleRel(betaS,"mbetaS",0,"0",eps);
    }

    // Probability correlation is equal to scaled beta^{[IL]}.
    final double corr_frombeta=(mL.beta-mH.beta)/(pc.pH-pc.pL);
    // Unit tests.
    if(eps>0)
    {
	// special case for d=2, the beta to differ only in sign.
	UnitTests.checkDoubleRel(mL.beta,"mL.beta",-mH.beta,"-mH.beta",eps);
	// generalized skewness of I to match
	final com.polytechnik.utils.Skewness sk=
	    com.polytechnik.utils.Skewness.getGSkewnessFromMonomialMoments(p0t,p1t,p2t,p0i,p1i,p2i);
	UnitTests.checkDoubleRel(sk.sk,"sk.sk",(2*Mi[0]/Mt[0]-ev[0]-ev[1])/(ev[0]-ev[1]),"skewness",eps);
	UnitTests.checkDoubleRel(sk.sk,"sk.sk",-(mL.wmu+mH.wmu)/(mL.wmu-mH.wmu),"skewness",eps);
	UnitTests.checkDoubleRel(sk.sk,"sk.sk",mL.wC-mH.wC,"mL.wC-mH.wC",eps);
	// average I to match
	UnitTests.checkDoubleRel(mL.lamconst,"mL.lamconst",mH.lamconst,"mH.lamconst",eps);
	UnitTests.checkDoubleRel(mL.wC+mH.wC,"mL.wC+mH.wC",1,"1",eps);
	System.err.println("Mu2P2Asymmetry: mL.beta="+mL.beta+" mH.beta="+mH.beta);
	System.err.println("Mu2P2Asymmetry: mL.D="+mL.D+" mH.D="+mH.D);

	// probability correlation to match
	UnitTests.checkDoubleRel(corr_frombeta,"corr_frombeta",pc.corr,"pc.corr",eps);
	// |d/dp| operator: average and skewness are zero
	//UnitTests.checkDoubleRel(pc.skewnessLike_p,"pc.skewnessLike_p",0,"0",eps);
	//UnitTests.checkDoubleRel(pc.p_aver,"pc.p_aver",0,"0",eps);
	// the skewness of I to match
	//UnitTests.checkDoubleRel(pc.skewnessLike_i,"pc.skewnessLike_i",sk.sk,"sk.sk",eps);

	final com.polytechnik.utils.Skewness skIdp=
	    com.polytechnik.utils.Skewness.getGSkewness(Mt,new double []{0,p0i,p0i,2*p1i},M);
	//UnitTests.checkDoubleRel(pci.skewnessLike_p,"pci.skewnessLike_p",skIdp.sk,"skIdp.sk",eps);
	//UnitTests.checkDoubleRel(pc.skewnessLike_p,"pc.skewnessLike_p",0,"0",eps);

	final Mu2 m2=new Mu2(p0t,p1t,p2t,p0i,p1i,p2i,eps);
	System.err.println("Mu2P2Asymmetry: m2.pc.corr="+m2.pc.corr+" pc.corr="+pc.corr);
	System.err.println("Mt="+java.util.Arrays.toString(Mt));
	System.err.println("Dp="+java.util.Arrays.toString(varpsi.getDiffMatrix(d,Mt)));
	System.err.println("Mi="+java.util.Arrays.toString(Mi));
	System.err.println("IDp="+java.util.Arrays.toString(varpsi.getDiffMatrix(d,Mi)));
	//System.err.println("pc.skewnessLike_p="+pc.skewnessLike_p+" pc.skewnessLike_i="+pc.skewnessLike_i);
	System.err.println("Mu2P2Asymmetry: test OK, eps="+eps);
    }
}

/** For unit test.
 */
static class MuRQ {
    final double beta,D,wmu,lam,lamconst,p_psiroot,pt,pi,wC;
    final double [] varAt0;
    MuRQ(final double [] psi,
	 final double lam,
	 final double [] QQi,final double [] QQt,
	 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	 final VolMatrix varpsi,
	 final double eps)
    {
	final int d=2;
	if(psi.length!=d || QQi.length!=d*d || QQt.length!=d*d)
	    throw new IllegalArgumentException();
	this.lam=lam;
	this.lamconst=QQi[0]/QQt[0];
	this.wmu=QQi[0]-lam*QQt[0]; // =(lamconst-lam)*QQt[0]
	final double [] o=new double []{1},ox=M.B.getaxbP(o,1,0);
	this.pt=TMatr.getAMatrB(ox,o,QQt,d)/TMatr.getAMatrA(o,QQt,d);
	this.pi=TMatr.getAMatrB(ox,o,QQi,d)/TMatr.getAMatrA(o,QQi,d);
	this.wC=Math.pow(TMatr.getAMatrB(psi,o,QQt,d),2)/TMatr.getAMatrA(o,QQt,d);
	final double [] Dpsi=varpsi.getDPsi(psi);
	this.p_psiroot=-M.B.getD0(psi,0)/M.B.getD0(Dpsi,0);
	final double [] nomc=new double []{
	    TMatr.getAMatrA(psi,QQi,d),
	    2*TMatr.getAMatrB(psi,Dpsi,QQi,d),
	    TMatr.getAMatrA(Dpsi,QQi,d)},
	  dnomc=new double []{
	    TMatr.getAMatrA(psi,QQt,d),
	    2*TMatr.getAMatrB(psi,Dpsi,QQt,d),
	    TMatr.getAMatrA(Dpsi,QQt,d)};
	if(eps>0)
	{
	    // check |psi> is the eigenvalue:
	    UnitTests.checkDoubleRel(nomc[0]/dnomc[0],"nomc[0]/dnomc[0]",lam,"lam",eps);
	    UnitTests.checkDoubleRel(nomc[1]/lam,"nomc[1]/lam",dnomc[1],"dnomc[1]",eps);
	    //System.err.println("nomc="+java.util.Arrays.toString(nomc)+" dnomc="+java.util.Arrays.toString(dnomc));
	    // check Gauss quadrature for dv-lam*dt
	    final double [] QQmu=TMatr.sumAcB(QQi,-lam,QQt);
	    UnitTests.checkDoubleRel(p_psiroot,"p_psiroot",QQmu[0*d+1]/wmu,"<p>",eps);
	    UnitTests.checkDoubleRel(Math.pow(p_psiroot,2),"p_psiroot^2",QQmu[1*d+1]/wmu,"<p^2>",eps);
	}
	TMatr.scaleVectorInPlace(nomc,1/nomc[0]);
	TMatr.scaleVectorInPlace(dnomc,1/dnomc[0]);
	this.D=nomc[2]-dnomc[2];
	this.beta=-dnomc[1]; // =-2*<psi|d/dP|psi> ; the dimension is: [1/P]
	this.varAt0=new double []{1,0,D,beta*D};

	if(eps>0)
	{
	    UnitTests.checkDoubleRel(nomc[1],"nomc[1]",dnomc[1],"dnomc[1]",eps);
	    UnitTests.checkDoubleRel((lam/Math.pow(Dpsi[0],2))*varAt0[2],"(lam/Dpsi^2)*varAt0[2]",wmu,"wmu",eps);
	    final double [] rq=TMatr.scaleVectorInPlace(com.polytechnik.utils.RayleighQuotient.getRQVariationValues(psi,Dpsi,null,QQi,QQt,d),1/lam);
	    UnitTests.checkDoubleRel(rq[0],"rq[0]",1,"1",eps);
	    UnitTests.checkDoubleRel(rq[1],"rq[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(0.5*rq[2],"0.5*rq[2]",varAt0[2],"varAt0[2]",eps);
	    final com.polytechnik.utils.PolynomialsRatio pr=
		new com.polytechnik.utils.PolynomialsRatio(dnomc,nomc,M.B);
	    UnitTests.checkDoubleRel(pr.getF(0),"pr.getF(0)",1,"1",eps);
	    UnitTests.checkDoubleRel(pr.getDF(0),"pr.getDF(0)",0,"0",eps);
	    UnitTests.checkDoubleRel(0.5*pr.getD2F(0),"0.5*pr.getD2F(0)",varAt0[2],"varAt0[2]",eps);
	    final com.polytechnik.utils.PolynomialsRatio Dpr=
		new com.polytechnik.utils.PolynomialsRatio(
		    M.B.mult2Pol(dnomc,dnomc),
		    TMatr.sumAcB(M.B.mult2Pol(varpsi.getDPsi(nomc),dnomc),-1,M.B.mult2Pol(varpsi.getDPsi(dnomc),nomc)),M.B);
	    UnitTests.checkDoubleRel(Dpr.getF(0),"Dpr.getF(0)",0,"0",eps);
	    UnitTests.checkDoubleRel(0.5*Dpr.getDF(0),"0.5*Dpr.getDF(0)",varAt0[2],"varAt0[2]",eps);
	    UnitTests.checkDoubleRel((1.0/6)*Dpr.getD2F(0),"(1/6)*Dpr.getD2F(0)",varAt0[3],"varAt0[3]",eps);

	    System.err.println("Mu2P2Asymmetry.MuRQ: test OK, varAt0="+java.util.Arrays.toString(varAt0)+" lam="+lam+" eps="+eps);
	}
    }
    static void matchMuRQ(final MuRQ q,final double p_offset_q,final MuRQ s,final double p_offset_s,final double eps)
    {
	UnitTests.checkDoubleRel(q.wmu,"q.wmu",s.wmu,"s.wmu",eps);
	UnitTests.checkDoubleRel(q.beta,"q.beta",s.beta,"s.beta",eps);
	UnitTests.checkDoubleRel(q.D,"q.beta",s.D,"s.beta",eps);
	UnitTests.checkDoubleRel(q.lam,"q.lam",s.lam,"s.lam",eps);
	UnitTests.checkDoubleRel(q.lamconst,"q.lamconst",s.lamconst,"s.lamconst",eps);
	UnitTests.checkDoubleArrayRel(q.varAt0,"q.varAt0",s.varAt0,"s.varAt0",eps);

	final double p_offset_shift=p_offset_s-p_offset_q;
	UnitTests.checkDoubleRel(q.p_psiroot,"q.p_psiroot",s.p_psiroot+p_offset_shift,"s.p_psiroot+p_offset_shift",eps);
	UnitTests.checkDoubleRel(q.pt+1,"q.pt+1",s.pt+1+p_offset_shift,"s.pt+1+p_offset_shift",eps);
	UnitTests.checkDoubleRel(q.pi,"q.pi",s.pi+p_offset_shift,"s.pi+p_offset_shift",eps);
	System.err.println("Mu2P2Asymmetry.MuRQ.matchMuRQ: test OK, eps="+eps);
    }
} // END OF public static class MuRQ
public static void matchMu2P2Asymmetry(final Mu2P2Asymmetry q,final double p_offset_q,final Mu2P2Asymmetry s,final double p_offset_s,final double eps)
{
    MuRQ.matchMuRQ(q.mL,p_offset_q,s.mL,p_offset_s,eps);
    MuRQ.matchMuRQ(q.mH,p_offset_q,s.mH,p_offset_s,eps);
    UnitTests.checkDoubleRel(q.mL.beta,"q.mL.beta",-q.mH.beta,"-q.mH.beta",eps);
    UnitTests.checkDoubleRel(s.mL.beta,"s.mL.beta",-s.mH.beta,"-s.mH.beta",eps);
    ProbabilityCorrelation.matchProbabilityCorrelation(q.pc,s.pc,eps);
    System.err.println("Mu2P2Asymmetry.matchMu2P2Asymmetry: test OK, eps="+eps);
}
} // END OF public class Mu2P2Asymmetry
