package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/**
 * Try dmu=dv-lam*dt measure, calculate beta for every root, then weight obtained betas.
 * @version $Id: Mu2P2Set.java,v 1.14 2022/06/27 07:19:13 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class Mu2P2Set {
    public final EVXData I;
    public final Mu2P2Asymmetry [] dia;
    public final double [] dia_wmu;
    public final double corr_weighted;

Mu2P2Set(final EVXData I,final Mu2P2Asymmetry [] dia,final double [] dia_wmu,final double eps)
{
    this.I=I;
    this.dia=dia;
    this.dia_wmu=dia_wmu;
    double sum_w_corr=0,sum_w=0;
    for(int k=0;k<dia.length;k++) {
	sum_w+=dia_wmu[k];
	sum_w_corr+=dia[k].pc.corr*dia_wmu[k]; // betaSum=mL.beta=-mH.beta
	if(eps>0)
	{
	    UnitTests.checkDoubleRel(dia[k].mL.beta,"dia[k].mL.beta",-dia[k].mH.beta,"-dia[k].mH.beta",eps);
	}
    }
    this.corr_weighted=sum_w_corr/sum_w;
}


public static Mu2P2Set getMu2P2Set_MU(final IMomentsData mm,
				      final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				      final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final Mu mu=new Mu(I.psiL,I.sL,mm.QQt,mm.QQi,
		       mm.QQt_sample,mm.QQpt_sample,mm.QQp2t_sample,
		       mm.QQi,mm.QQpi,mm.QQp2i,M,eps);
    return new Mu2P2Set(I,mu.dia,mu.wmu,eps);
}

public static class Mu {
    public final double lam;
    public final double [] x,wmu;
    public final Mu2P2Asymmetry [] dia;
Mu(final double [] psi,final double lam,final double [] QQt,final double [] QQi,
   final double [] QQp0t,final double [] QQp1t,final double [] QQp2t,
   final double [] QQp0i,final double [] QQp1i,final double [] QQp2i,
   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double eps)
{
    final int n=psi.length,d=2;
    this.lam=lam;
    this.x=M.getXNodes(psi).re_root;
    this.wmu=new double [x.length];
    this.dia=new Mu2P2Asymmetry [x.length];
    // The measure dmu=dv-lam*dt.
    for(int k=0;k<x.length;k++)
    {
	final double xk=x[k];
	final double [] l=M.B.sdiv1(psi,xk).q;
	final double nrmlk=M.B.getD0(l,xk);
	TMatr.scaleVectorInPlace(l,1/nrmlk);
	final double m0i=TMatr.getAMatrA(l,QQi,n),m0t=TMatr.getAMatrA(l,QQt,n);
	this.wmu[k]=m0i-lam*m0t;
	this.dia[k]=new Mu2P2Asymmetry(
	    TMatr.getAMatrA(l,QQp0t,n),TMatr.getAMatrA(l,QQp1t,n),TMatr.getAMatrA(l,QQp2t,n),
	    TMatr.getAMatrA(l,QQp0i,n),TMatr.getAMatrA(l,QQp1i,n),TMatr.getAMatrA(l,QQp2i,n),eps);
    }
    if(eps>0)
    {
	// Test Gauss quadratures for dmu=dv-lam*dt
	final double [] QQmu=TMatr.sumAcB(QQi,-lam,QQt),o=new double []{1},ox=M.B.getaxbP(o,1,0);
	UnitTests.checkDoubleRel(TMatr.sumElements(wmu),"sum(wmu)",TMatr.getAMatrA(o,QQmu,n),"QQmu[0]",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(x,wmu),"sum(x*wmu)",TMatr.getAMatrB(o,ox,QQmu,n),"x QQmu[0]",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(x,wmu),"sum(x^2*wmu)",TMatr.getAMatrA(ox,QQmu,n),"x^2 QQmu[0]",eps);
	System.err.println("Mu2P2Set: test OK, n="+n+" eps="+eps);
    }
}
} // END OF public static class Mu

}
