package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import java.util.*;
import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.ReadWriteUtils;

/** A class to store parameters for model prediction.
 *  In a future - read per stock parameters based on
 *  statistical analysis. For now - the same for all.
 *  @version $Id: MultiStockModelParameters.java,v 1.38 2022/06/07 10:52:26 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class MultiStockModelParameters {

    static final long MILLISECOND_SCALE=TraderInterface.TIME_SCALE/1000;

    static class MultiStockModelParametersData {

	final int MAX_LONG=100,MAX_SHORT=100; // FIXME

	final int nd;
	final long tau;
	final double thr_enter,thr_exit;
	final String ticker;
	final String [] indexstocks;
	final boolean flag_active;
	final boolean flag_towatch;

	final long tau_QUANT;
	final long tau_readdata_clearBefore_delay;
	final String OUTPUT_DIR;

	MultiStockModelParametersData(final String ticker,
				      final int nd,
				      final long tau,
				      final double thr_enter,
				      final double thr_exit,
				      final String [] indexstocks,
				      final long tau_QUANT,
				      final boolean flag_active,
				      final boolean flag_towatch,
				      final String OUTPUT_DIR)
	{
	    this.ticker=ticker;
	    this.nd=nd;
	    this.tau=tau;
	    this.thr_enter=thr_enter;
	    this.thr_exit=thr_exit;
	    this.indexstocks=indexstocks;
	    this.tau_QUANT=tau_QUANT;
	    this.flag_active=flag_active;
	    this.flag_towatch=flag_towatch;
	    this.OUTPUT_DIR=OUTPUT_DIR;
	    // 100*tau_QUANT+200 sec. Think later.
	    this.tau_readdata_clearBefore_delay=100*this.tau_QUANT+200*TraderInterface.TIME_SCALE;
	    if((this.tau<=0) != (!this.flag_active))
	    {
		throw new RuntimeException("invalid tau="+this.tau+" for flag_active="+this.flag_active);
	    }
	}
	String dumpToSting()
	{
	    return "MultiStockModelParametersData:ticker="+ticker+
		"\n\t nd="+nd+" thr_enter="+thr_enter+" thr_exit="+thr_exit+
		" flag_active="+flag_active+" flag_towatch="+flag_towatch+" OUTPUT_DIR="+OUTPUT_DIR+
		"\n\ttau="+(tau<=0?tau:(tau/MILLISECOND_SCALE))+" tau_QUANT="+(tau_QUANT<=0?tau_QUANT:(tau_QUANT/MILLISECOND_SCALE))+
		"\n\tind="+java.util.Arrays.toString(indexstocks);
	}
    }

    static String [] removeSelfMatches(final String [] ind,final String ticker)
    {
	final java.util.ArrayList<String> sa=new java.util.ArrayList<String>();
	for(String in:ind) // remove duplicates
	{
	    if(!in.equals(ticker)) sa.add(in);
	}
	return sa.toArray(String[]::new);
    }

    MultiStockModelParametersData getParameters(final String ticker)
    {
	final MultiStockModelParametersData m=this.active.get(ticker);
	if(m!=null)
	{
	    return m;
	}
	return new MultiStockModelParametersData(ticker,-1,-1,Double.NaN,Double.NaN,null,-1,false,this.tickers_to_watch.contains(ticker),null);
    }


    private java.util.Set<String> getTickersToWatch()
    {
	final java.util.LinkedHashSet<String> sa=new java.util.LinkedHashSet<String>();
	for(MultiStockModelParametersData m:this.active.values())
	{
	    sa.add(m.ticker);
	    for(String x:m.indexstocks)
	    {
		sa.add(x);
	    }
	}
	return sa;
    }


    final int DEFAULT_ND;
    final double DEFAULT_THR_ENTER,DEFAULT_THR_EXIT;
    final long DEFAULT_TAU,T_QUANT;
    final String OUTPUT_DIR;
    private final java.util.Map<String,MultiStockModelParametersData> active;
    private final java.util.Set<String> tickers_to_watch;
    MultiStockModelParameters(final int DEFAULT_ND,
			      final double DEFAULT_THR_ENTER,
			      final double DEFAULT_THR_EXIT,
			      final long DEFAULT_TAU,
			      final long T_QUANT,
			      final String OUTPUT_DIR,
			      final java.util.Map<String,MultiStockModelParametersData> active)
    {
	this.DEFAULT_ND=DEFAULT_ND;
	this.DEFAULT_THR_ENTER=DEFAULT_THR_ENTER;
	this.DEFAULT_THR_EXIT=DEFAULT_THR_EXIT;
	this.DEFAULT_TAU=DEFAULT_TAU;
	this.T_QUANT=T_QUANT;
	this.OUTPUT_DIR=OUTPUT_DIR;
	this.active=active;
	this.tickers_to_watch=getTickersToWatch();

    }


    static MultiStockModelParameters read(final String fname) throws java.io.IOException
    {
	final ResourceBundle p=ReadWriteUtils.getProperties(fname);
	//System.err.println("keys"+p.keySet());
	final String alt_out=System.getProperty("tr.OUTPUT_DIR");
	final String OUTPUT_DIR=alt_out!=null?alt_out:ReadWriteUtils.getparameter("_OUTPUT_DIR",p,true);
	final int DEFAULT_ND=Integer.valueOf(ReadWriteUtils.getparameter("_DEFAULT_ND",p,true));
	final double DEFAULT_THR_ENTER=Double.valueOf(ReadWriteUtils.getparameter("_DEFAULT_THR_ENTER",p,true));
	final double DEFAULT_THR_EXIT=Double.valueOf(ReadWriteUtils.getparameter("_DEFAULT_THR_EXIT",p,true));
	final long T_QUANT=Long.valueOf(ReadWriteUtils.getparameter("_T_QUANT",p,true))*MILLISECOND_SCALE;
	final long DEFAULT_TAU=Long.valueOf(ReadWriteUtils.getparameter("_DEFAULT_TAU",p,true))*MILLISECOND_SCALE;

	final String globalindstr=ReadWriteUtils.getparameter("_GLOBALINDS",p,true);
	final java.util.Map<String,String> gind=new java.util.HashMap<String,String>();
	if(globalindstr!=null)
	{
	    final String [] globalinds=ReadWriteUtils.extractnames(globalindstr,"\t ");
	    for(int i=0;i<globalinds.length;i++)
	    {
		final String ind_name=globalinds[i],ind=ReadWriteUtils.getparameter("_"+ind_name,p,true);
		if(gind.get(ind_name)!=null)
		{
		    throw new RuntimeException("duplicate global_index="+ind_name+" list="+globalindstr);
		}
		gind.put(ind_name,ind);
	    }
	}

	final String tickersstr=ReadWriteUtils.getparameter("_TICKERSTOTRADE",p,true);
	final String [] tickers=ReadWriteUtils.extractnames(tickersstr,"\t ");
	final String [] tickerprops=ReadWriteUtils.readParameters(".",tickers,p);

	java.util.Map<String,MultiStockModelParametersData> a=
	    new java.util.HashMap<String,MultiStockModelParametersData>();
	final java.util.Map<String,String> m=new java.util.HashMap<String,String>();
	for(int i=0;i<tickers.length;i++)
	{
	    final String s=tickerprops[i],ticker=tickers[i];
	    m.clear();
	    final String [] td=ReadWriteUtils.extractnames(s,";");
	    for(String ss: td)
	    {
		final String [] pl=ReadWriteUtils.extractnames(ss,"=\t\r\n ");
		if(pl.length!=2)
		{
		    throw new RuntimeException("For ticker="+ticker+" invalid string="+ss+" found, opts="+s);
		}
		if(m.get(pl[0])!=null)
		{
		    throw new RuntimeException("For ticker="+ticker+" duplicated papameter="+pl[0]+" found, opts="+s);
		}
		m.put(pl[0],pl[1]);
	    }
	    final String nd_str=m.get("nd");
	    final int nd=nd_str==null?DEFAULT_ND:Integer.valueOf(nd_str);

	    final String thr_enter_str=m.get("thr_enter");
	    final double thr_enter=thr_enter_str==null?DEFAULT_THR_ENTER:Double.valueOf(thr_enter_str);

	    final String thr_exit_str=m.get("thr_exit");
	    final double thr_exit=thr_exit_str==null?DEFAULT_THR_EXIT:Double.valueOf(thr_exit_str);

	    final long tau;
	    final String tau_str=m.get("tau");
	    if(tau_str==null){tau=DEFAULT_TAU;}
	    else{tau=Long.valueOf(tau_str)*MILLISECOND_SCALE;}

	    String ind_str=m.get("ind");
	    if(ind_str==null)
	    {
		throw new RuntimeException("For ticker="+ticker+" no ind= found, opts="+s);
	    }
	    final String ind_fromglobal=gind.get(ind_str);
	    if(ind_fromglobal!=null) ind_str=ind_fromglobal;
	    final String [] ind=ReadWriteUtils.extractnames(ind_str,",\t\r\n ");

	    if(a.get(ticker)!=null)
	    {
		throw new RuntimeException("For ticker="+ticker+" duplicated ticker="+ticker+" found, opts="+s+" check _TICKERSTOTRADE");
	    }
	    a.put(ticker,new MultiStockModelParametersData(ticker,nd,tau,thr_enter,thr_exit,removeSelfMatches(ind,ticker),T_QUANT,true,true,OUTPUT_DIR));
	}
	return new MultiStockModelParameters(DEFAULT_ND,DEFAULT_THR_ENTER,DEFAULT_THR_EXIT,DEFAULT_TAU,T_QUANT,OUTPUT_DIR,a);
    }

    public String dumpToSting()
    {
	final StringBuilder buf=new StringBuilder();
	final java.util.ArrayList<MultiStockModelParametersData> ts=
	    new java.util.ArrayList<MultiStockModelParametersData>(this.active.values());
	Collections.sort(ts,new java.util.Comparator<MultiStockModelParametersData>(){
		public int compare(final MultiStockModelParametersData o1,
				   final MultiStockModelParametersData o2)
		                  {return o1.ticker.compareTo(o2.ticker);}});
	for(MultiStockModelParametersData mm:ts)
	{
	    buf.append(mm.dumpToSting());
	    buf.append("\n");
	}
	return buf.toString();
    }


    public static void main(final String [] args) throws java.io.IOException
    {
	if(args.length!=1)
	{
	    System.err.println("MultiStockModelParameters.main config_file");
	    System.exit(1);
	}
	final String config_file=args[0];
	System.err.println("Reading config_file="+config_file);
	final MultiStockModelParameters ma=read(config_file);

	// dump diag to stderr
	System.err.println(ma.dumpToSting());
    }

}
