package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.broker.*;
import com.polytechnik.outils.*;

/** A simple multi-stock trader.
 *  @version $Id: MultiStockTrader.java,v 1.37 2022/05/29 19:17:55 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class MultiStockTrader implements TraderInterface {
    /** The number of pre-set threads. */
    private static final int N_WORKERS=10;


    private static final boolean PRINT_DEBUG=true;
    /** Whether not to wait for explicit 'status OK' message.
     */
    static final boolean DEFAULT_ALLOW_TO_TRADE=false;

    private final BrokerInterface BROKER;

    private final WorkersPool WORKERSPOOL;
    private final MultiStockModelParameters mparameters;

    final int dump_trading_diagnostics;
    final boolean flag_realtime;
    final String description;

    private final FindOpportunityTaskMultiplexer findopportunity;
    public MultiStockTrader(final int dump_trading_diagnostics,
			    final MultiStockModelParameters mparameters,
			    final BrokerInterface BROKER,
			    final boolean flag_singlethread,
			    final boolean flag_realtime,
			    final String description)
    {
	// single threaded for training
	this.WORKERSPOOL=new WorkersPool(flag_singlethread?0:N_WORKERS,false);
	this.dump_trading_diagnostics=dump_trading_diagnostics;
	this.mparameters=mparameters;
	this.BROKER=BROKER;
	this.findopportunity=new FindOpportunityTaskMultiplexer();
	this.flag_realtime=flag_realtime;
	this.description=description;
    }

    private final java.util.Map<String,MultiStockStatusData> storeMultiStockStatusDatabyticker=
	new java.util.HashMap<String,MultiStockStatusData>();
    private boolean creation_finished=false;
    private final MultiStockStatusData getStatusData(final OrdersAndTrades ot,final long currenttime)
    {
	MultiStockStatusData h=ot.getData(); // Store data in OrdersAndTrades for simplicity.
	if(h==null)
	{
	    final MultiStockModelParameters.MultiStockModelParametersData mp=
		this.mparameters.getParameters(ot.ticker);
	    h=new MultiStockStatusData(ot.ticker,this.flag_realtime,
				       new MoneyManager(mp.MAX_LONG,mp.MAX_SHORT),
				       this,mp,ot);
	    ot.setData(h);
	    ot.setActive(h.mparametersdata.flag_towatch); // do not store trades for unused stocks
	    if(!this.creation_finished)
	    {
		this.storeMultiStockStatusDatabyticker.put(ot.ticker,h);
	    }
	    else
	    {
		throw new RuntimeException("Late creation of ticker="+ot.ticker+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	    if(h.mparametersdata.flag_active)
	    {
		BROKER.readInitialPosition(h); // initial position from broker
	    }
	}
	if(h.ticker!=ot.ticker || h.ot!=ot)
	{
	    throw new RuntimeException("Ticker or object mismatch:"+
				       " ticker1="+ot.ticker+
				       " and ticker2="+h.ticker+
				       " ot1="+h.ot+" ot2="+ot);
	}
	if(h.DIAG!=null && this.creation_finished)
	{
	    h.DIAG.setTicker();
	}
	h.setTraderCurrentTime(currenttime);
	return h;
    }

    void closeAllOutputs()
    {
	for(MultiStockStatusData h:this.storeMultiStockStatusDatabyticker.values())
	{
	    if(h.mparametersdata.flag_active && this.creation_finished)
	    {
		h.DIAG.closeTicker();
	    }
	}
    }


    boolean allow_trading=DEFAULT_ALLOW_TO_TRADE;
    public void allowTrading(final long currenttime,final boolean on_off)
    {
	this.allow_trading=on_off;
    }


    public void allowTrading(final OrdersAndTrades ot,final long currenttime,final int value,final String text)
    {
	final MultiStockStatusData sd=getStatusData(ot,currenttime);
	System.err.println("MultiStockTrader:allowTrading: for ticker="+sd.ticker+" value="+value+" msg="+text);
	if(value==TraderInterface.SEC_TRADING_OK) sd.setAllowTrading(true);
	else if(value==TraderInterface.SEC_TRADING_DISABLED) sd.setAllowTrading(false);
	// shorts can be handled separately
    }


    public void setMarketMaker(final OrdersAndTrades ot,
			       final long currenttime,
			       final String mpid,
			       final boolean flag_primary,
			       final boolean flag_ok)
    {
	// not implemented yet
    }


    public void setInstrument(final OrdersAndTrades ot,
			      final long currenttime,
			      final boolean flag_statusOK,
			      final int round_lot_shares,
			      final boolean round_lot_only,
			      final String market)
    {
	//System.err.println("TICKER="+ot.ticker);
	getStatusData(ot,currenttime).setStatus(flag_statusOK,round_lot_shares,round_lot_only,market);
    }


    /** Note: works as a multiplexor, do not save anything inside!!!
     */
    private class FindOpportunityTaskMultiplexer implements WorkerTask {
	private MultiStockStatusData sd=null;
	private long id=-1;
	private final Opportunity [] opp;

	FindOpportunityTaskMultiplexer()
	{
	    this.opp=new Opportunity [N_WORKERS];
	    for(int i=this.opp.length;i-- >0;)
	    {
		this.opp[i]=new Opportunity();
	    }
	}
	void setParameters(final MultiStockStatusData sd,final long id)
	{
	    if(this.sd!=null)
	    {
		throw new RuntimeException("Something is terrible wrong with threading for ticker="+sd.ticker+
					   " currently processing ticker="+this.sd.ticker);
	    }
	    this.sd=sd;
	    this.id=id;
	}

	public void initializeTask(final int tasknum)
	{
	    // called synchroneously, no race can occur.
	    final MultiStockStatusData sd=this.sd;
	    final long currenttime=sd.getTraderCurrentTime();
	    if(sd.processstarted || this.id<0)
	    {
		throw new RuntimeException("Something is terrible wrong with threading for ticker="+
					   sd.ticker+" currenttime="+BookItem.getTimeDiag(currenttime)+
					   " time_processstarted="+BookItem.getTimeDiag(sd.time_processstarted)+
					   " id="+this.id+
					   " sd.processstarted="+sd.processstarted);
	    }
	    sd.processstarted=true;
	    sd.time_processstarted=currenttime;

	    // copy all data to trade opportunity checker
	    this.opp[tasknum].setData(sd,this.id);

	    // clear internal state, because the multiplexer is re-used.
	    this.sd=null;
	    this.id=-1;
	}
	public void cleanupTask(final int tasknum){}

	public void performTask(final int tasknum)
	{
	    final Opportunity checker=this.opp[tasknum];
	    final long magicid=checker.getMagicID();
	    final MultiStockStatusData sd=checker.sd;
	    final long currenttime=sd.getTraderCurrentTime();

	    long tau=-1L;
	    boolean sameidok=false;

	    try {
		if(!sd.processstarted)
		{
		    throw new RuntimeException("Something is terrible wrong for ticker="+sd.ticker+
					       " currenttime="+BookItem.getTimeDiag(currenttime)+" processstarted="+sd.processstarted);
		}

		// prepare internal structures
		checker.prepareData();

		final Opportunity.CalculationResults cd=checker.calculateEverything();
		//System.err.println("checker.sd.umsqvml="+checker.sd.umsqvml);
		//System.err.println("checker.sd.umsqvmp="+checker.sd.umsqvmp);

		BROKER.hookForDebugging(new Object []{sd.ticker,
						      checker.savedtrades,
						      Double.valueOf(cd.p_buy),
						      Double.valueOf(cd.p_sell)},currenttime);

		// actual fine-grained calculations in predictor
		sd.PREDICTOR.updateInternalState(cd);

		// trade if opportunity is good.
		tau=sd.PREDICTOR.tradeIfOpportunityIsGood(BROKER,
							  cd.time,
							  cd.p_buy,
							  cd.p_sell,
							  cd.vol_p_buy,
							  cd.vol_p_sell);

		sd.DIAG.dumpTickData(cd);
	    } finally {
		// Clear opportunity checker data after we are done.
		sameidok=(magicid==checker.clear());
		// Clear predictor data.
		sd.PREDICTOR.clear();
		// Set status, no race can occur if just one processor is started for one ticker.
		sd.time_processstarted=0;
		if(tau>=0)
		{
		    sd.setNextCheck(tau);
		}
		sd.processstarted=false;
	    }
	    if(tau<0L)
	    {
		throw new RuntimeException("Invalid tau="+tau+" for ticker="+sd.ticker+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	    if(!sameidok)
	    { // no exception & id mismatch
		throw new RuntimeException("Something is terrible wrong with threading for ticker="+sd.ticker+" time="+currenttime+" time_processstarted="+sd.time_processstarted+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	}
    } // END OF CLASS FindOpportunityTask


    private long id=0;
    private long getID()
    {
	return ++this.id;
    }

    public void update(final OrdersAndTrades ot,
		       final long currenttime,
		       final int eventid)
    {
	final MultiStockStatusData sd=getStatusData(ot,currenttime);

	if(!this.creation_finished)
	{ // init all depenent tickers during first "update"
	    MultiStockStatusData.setAllStockReferences(this.storeMultiStockStatusDatabyticker);
	    this.creation_finished=true;
	    if(PRINT_DEBUG)
	    { // Dump all active stocks for debuffing
		System.err.println("=========== ALL STOCKS SETTINGS ========");
		for(MultiStockStatusData s:this.storeMultiStockStatusDatabyticker.values())
		{
		    System.err.println(s.mparametersdata.dumpToSting());
		}
	    }
	}

	final MultiStockStatusData [] m=sd.references;
	if(m!=null)
	{
	    for(MultiStockStatusData mm:m)
	    {
		if(!mm.mparametersdata.flag_active)
		{
		    throw new RuntimeException("The ticker="+mm.ticker+" is not active, but received an event on ticker="+sd.ticker+
					       " currenttime="+BookItem.getTimeDiag(currenttime));
		}
		updateDependent(mm.ot,currenttime,eventid,ot.ticker);
	    }
	}
	else
	{
	    if(sd.mparametersdata.flag_active)
	    {
		throw new RuntimeException("For active ticker="+sd.ticker+" no dependent ticker(s) found"+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	}
    }


    private void updateDependent(final OrdersAndTrades ot,
				 final long currenttime,
				 final int eventid,
				 final String ticker_called_from)
    {
	final MultiStockStatusData sd=getStatusData(ot,currenttime);
	/*
	if(sd.mparametersdata.flag_active)
	{
	    System.err.println("eventid="+eventid+
			       " ticker="+ot.ticker+
			       " currenttime="+BookItem.getTimeDiag(currenttime)+
			       " tradebook.getSize()="+ot.tradebook.getSize()+
			       " orderbook.getSize()="+ot.orderbook.getSize()+
			       " ticker_called_from="+ticker_called_from);
	}
	*/
	if(sd.ot!=ot)
	{
	    throw new RuntimeException("Something is terrible wrong with OrdersAndTrades: "+
				       " ot1="+ot+" ot2="+sd.ot+
				       " ticker="+sd.ticker+" currenttime="+BookItem.getTimeDiag(currenttime));
	}

	if(sd.processstarted)
	{ // to avoid two processors of the same stock
	    if(PRINT_DEBUG && currenttime>sd.mparametersdata.tau_QUANT+sd.time_processstarted)
	    {
		   System.err.println(
		       "MultiStockTrader: work skipped because a one for ticker="+
		       ot.ticker+" is already running started at t="+
		       BookItem.getTimeDiag(sd.time_processstarted)+
		       " now="+BookItem.getTimeDiag(currenttime)+
		       " diff="+(currenttime-sd.time_processstarted)+
		       " tau_QUANT="+sd.mparametersdata.tau_QUANT+
		       " ticker_called_from="+ticker_called_from);
	    }
	    return; // Pending excecution, skip this work to avoid races.
	}



	long tclear=-1;
	if(!sd.mparametersdata.flag_towatch)
	{
	    if(ot.tradebook.getSize()>0)
	    {
		throw new RuntimeException("There should be no trades here "+
					   " ticker="+ot.ticker+
					   " ot.tradebook.getSize()="+ot.tradebook.getSize()+
					   " ticker_called_from="+ticker_called_from+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
	    }
	}
	else if(sd.shouldCheck(currenttime))
	{
	    // perform trade opportunity check
	    // set data available to workers
	    this.findopportunity.setParameters(sd,getID());
	    // perform work, run  asynchronously in a different thread.
	    final boolean work_started=this.WORKERSPOOL.performWork(this.findopportunity);
	    if(PRINT_DEBUG)
	    {
		if(!work_started)
		{
		   System.err.println("MultiStockTrader: work skipped becauseus too "+
				      "many tasks are already started; ticker="+
				      ot.ticker+" time="+currenttime+
				      " ticker_called_from="+ticker_called_from);
		}
	    }

	    if(sd.mparametersdata.tau_readdata_clearBefore_delay>=0)
	    {
		tclear=currenttime-sd.mparametersdata.tau_readdata_clearBefore_delay;
	    }
	    if(!work_started)
	    {
		tclear=-1; // do not clear old data if some work is skipped
	    }
	}

	// clear old trades
	if(tclear>=0L)
	{
	    /*
	    System.err.println("CLEAR ticker="+sd.ticker+
			       " lastprocessed="+sd.pastcalculations.training_last_data_processed+
			       " tlast="+(sd.pastcalculations.training_last_data_processed!=null?
					  sd.pastcalculations.training_last_data_processed.getTime():-1)+
			       " dt="+((currenttime-tclear)/1e9));
	    */
	    ot.clearBefore(tclear);
	}

    }

}
