package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.PolynomialRootsFinder;
import com.polytechnik.realtimebook.*;

/** Impact QVM calculations with kernel {@literal <Q(x)|Q(x')>} polynomials.
 *  Here we have {@literal <a|b>}=a_i ({@literal <QQ>}_mu)_{ij} b_j
 *  @version $Id: OQVM.java,v 1.58 2022/07/28 12:27:17 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class OQVM<DTYPE> {

final DTYPE odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;

final double [] kk,kksell,kkbuy,ok,c,cp,cp2,cdp,cds,kkeq;
final double wa,xeq,peq,p2eq,dLp,dpL,dpLE,dpLE_sigma2,dpLE0_sigma2,xeqdVdT,xoutdVdT,dIxout,dIx0,dIxend,dpxeq,w00,
    dLs,dsL,dsLE,dsLE_sigma2,dsLE0_sigma2,dsxeq;
final double Ik,dIkL,dIt,dIv,xeqsell,xeqbuy;

final double dbsIk,dbsIkL,dbsIksL,dbsIkbL,dpmIk,dpmIkL,dpmIksL,dpmIkbL;
final double dIa,dbsa,dpma,kka,kkasell,kkabuy,kka0,kkasell0,kkabuy0;

OQVM(final int neq,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();

    final int nmoms=2*neq-1;
    final double [] mom=new double [nmoms],pmom=new double [nmoms],p2mom=new double [nmoms],dp=new double [nmoms],ds=new double [nmoms];
    access.setM(mom,0,access.M_I,odata,nmoms);
    access.setM(pmom,1,access.M_I,odata,nmoms);
    access.setM(p2mom,2,access.M_I,odata,nmoms);
    access.setM(ds,0,access.M_ISELL,odata,nmoms);
    access.setM(dp,0,access.M_IBUY,odata,nmoms);
    for(int k=nmoms;--k>=0;) ds[k]=ds[k]-dp[k];
    access.setM(dp,1,access.M_DP,odata,nmoms);
    w00=mom[0];

    final double [][] kkm=getKK(neq,mom,new double [][]{mom,pmom,p2mom,dp,ds});
    kk=kkm[0];
    ok=kkm[1];
    c=kkm[2];
    cp=kkm[3];
    cp2=kkm[4];
    cdp=kkm[5];
    cds=kkm[6];
    kkeq=getKKEQ(neq);
    xeq=getXEq(kk);
    wa=access.fx(c,xeq);
    final double dp_eq=access.fx(cp,xeq);
    //System.err.println("CC:xeq="+xeq+" cp="+java.util.Arrays.toString(cp));
    //System.err.println("dp="+java.util.Arrays.toString(dp));
    peq=dp_eq+getPOffset();
    p2eq=access.fx(cp2,xeq)-dp_eq*dp_eq;

    {
    final double [] dpx=getDPX(xeq,cdp,ok);
    dLp=dpx[0];
    dpL=dpx[1];
    dpLE=dpL/Math.sqrt(dLp);
    dpLE_sigma2=dpx[2];
    dpLE0_sigma2=dpx[3];
    dpxeq=dpx[4];
    }

    {
    final double [] dsx=getDPX(xeq,cds,ok);
    dLs=dsx[0];
    dsL=dsx[1];
    dsLE=dsL/Math.sqrt(dLs);
    dsLE_sigma2=dsx[2];
    dsLE0_sigma2=dsx[3];
    dsxeq=dsx[4];
    }

    final double [] xx=getXdVdT(kk,kkeq);
    xeqdVdT=xx[0];
    xoutdVdT=xx[1];
    dIxout=access.fx(kkeq,xoutdVdT)/access.fx(kk,xoutdVdT);
    dIx0=access.fx(kkeq,getX0())/access.fx(kk,getX0());
    dIxend=kkeq[kkeq.length-1]/kk[kk.length-1];

    final double [] di=getDs(kk,kkeq);
    Ik=w00*di[0];
    dIkL=di[1];
    dIv=di[2];
    dIt=di[3];
    dIa=w00*di[4];

    {
    final double [] momisell=new double [nmoms],momibuy=new double [nmoms];
    access.setM(momisell,0,access.M_ISELL,odata,nmoms);
    access.setM(momibuy,0,access.M_IBUY,odata,nmoms);
    final double [] kks=kksell=getKK(neq,momisell,null)[0],kkb=kkbuy=getKK(neq,momibuy,null)[0];
    xeqsell=getXEq(kks);
    xeqbuy=getXEq(kkb);
    final double [] dbs=getDs(kks,kkb);
    dbsIk=dbs[0];
    dbsIkL=dbs[1];
    dbsIksL=dbs[2];
    dbsIkbL=dbs[3];
    dbsa=dbs[4];
    }
    kka=kk[kk.length-1]/kkeq[kkeq.length-1];
    kkasell=kksell[kk.length-1]/kkeq[kkeq.length-1];
    kkabuy=kkbuy[kk.length-1]/kkeq[kkeq.length-1];
    kka0=1/access.fx(kk,getX0());
    kkasell0=1/access.fx(kksell,getX0());
    kkabuy0=1/access.fx(kkbuy,getX0());

    {
    final double [] momplus=new double [nmoms],momminus=new double [nmoms];
    access.setM(momplus,1,access.M_DPPLUS,odata,nmoms);
    access.setM(momminus,1,access.M_DPMINUS,odata,nmoms);
    for(int k=nmoms;--k>=0;) momminus[k]=-momminus[k];
    final double [] kks=getKK(neq,momplus,null)[0],kkb=getKK(neq,momminus,null)[0];
    final double [] dbs=getDs(kks,kkb);
    dpmIk=dbs[0];
    dpmIkL=dbs[1];
    dpmIksL=dbs[2];
    dpmIkbL=dbs[3];
    dpma=dbs[4];
    }

}

double [][] getKK(final int n,final double [] moms,final double [][] extra_moms)
{
    final double [] QQ=getQQ_Matr(n,moms);
    final double [] kk=new double [2*n-1],ok=new double [n];
    final double [][] res=new double [2+(extra_moms!=null?extra_moms.length:0)][];
    res[0]=kk;res[1]=ok;
    for(int l=res.length;--l>=2;) res[l]=new double [n];
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQ,n,n,ipvt);
    if(pinfo!=0) {
	for(double [] l:res) java.util.Arrays.fill(l,Double.NaN);
	return res;
    }

    final double x0=getX0();
    final double [] tmp=new double [2*n-1],r=new double [n];
    for(int i=0;i<n;i++)
    {
	r[i]=1;
	ok[i]=access.fx(r,x0);
	com.polytechnik.utils.Linsystems.dgesl_m1(QQ,n,n,ipvt,r,0);
	for(int j=0;j<n;j++)
	{
	    final int na=access.setQQCoefs(i,j,tmp);
	    DeprecatedHelperTMatr.sumP1(kk,tmp,na,r[j]);
	    r[j]=0;
	}
    }

    for(int l=res.length;--l>=1;) {
	if(l>=2) System.arraycopy(extra_moms[l-2],0,res[l],0,n);
	com.polytechnik.utils.Linsystems.dgesl_m1(QQ,n,n,ipvt,res[l],1);
    }
    return res;
}
double [] getDs(final double [] kv,final double [] kt)
{
    if(kv.length!=kt.length) throw new RuntimeException("kv.length="+kv.length+" kt.length="+kt.length);
    final double x0=getX0();
    final double Kt=access.fx(kt,x0),Kv=access.fx(kv,x0),
	dKt=access.Dfx(kt,x0),dKv=access.Dfx(kv,x0);
    return new double []{
	Kt/Kv,
	dKt/Kt-dKv/Kv,
	dKv/Kv,
	dKt/Kt,
	kt[kt.length-1]/kv[kv.length-1] // x->infinity asymptotic
    };
}

abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract double getXEq(final double [] kk);
abstract double [] getKKEQ(final int n);
abstract double [] getDPX(final double xeq,final double [] dpv,final double [] ok);
abstract double [] getXdVdT(final double [] kv,final double [] kt);
abstract double getX0();
abstract double getPOffset();

/** Calculates <Q|Q> matrix.
 */
double [] getQQ_Matr(final int n,final double [] mom)
{
    return access.getQQ_Matrs(n,1,new double [][]{mom})[0];
}


static abstract class OQVMTest<DTYPE> extends QTest {

abstract OQVM<DTYPE> getOQVM();
abstract NQVM<DTYPE> getNQVM();
abstract double [][] getMonomialMoments(DTYPE o);

OQVMTest(final String name,
	 final java.util.Random r,final double eps)
{
    super(name,r,eps);
}
void test_Common()
{
    final OQVM<DTYPE> q=getOQVM();
    if(neq>1)
    {
	UnitTests.checkDoubleRel(q.wa,"q.wa",1,"1",eps);
	final double [] v=convertToMonomials(q.kk),t=convertToMonomials(q.kkeq);
	System.err.println("v="+java.util.Arrays.toString(v)+" t="+java.util.Arrays.toString(t)+" q,w00="+q.w00+" q.xeqdVdT="+q.xeqdVdT);
	final double d=q.w00*(Polynomials.getD1(v,q.xeqdVdT)*Polynomials.getD0(t,q.xeqdVdT)
		       -Polynomials.getD1(t,q.xeqdVdT)*Polynomials.getD0(v,q.xeqdVdT))/
	               Math.pow(Polynomials.getD0(t,q.xeqdVdT),1);
	UnitTests.checkDoubleRel(d,"d",0,"0",eps);
    }
    System.err.println(name+":OQVMTest:test_Commons(): test OK neq="+neq+" eps="+eps);
}

static double [] getKK(final int n,final double [] moms)
{
    final double [][] vv=NQVM.NQVMTest.getMM_Matr(n,moms,1);
    final double [][] mvv=com.polytechnik.utils.Linsystems.getInvertedMatrix(vv);
    final double [] kk=new double [2*n-1];
    for(int k=0;k<n;k++)
    {
	for(int j=0;j<n;j++)
	{
	    kk[k+j]+=mvv[k][j];
	}
    }
    return kk;
}
static double [] getKK_fromQQ(final int n,final double [] QQ)
{
    final double [] kk=new double [2*n-1];
    for(int k=0;k<n;k++)
    {
	for(int j=0;j<n;j++)
	{
	    kk[k+j]+=QQ[k*n+j];
	}
    }
    return kk;
}

static double [] getKM(final int n,final double [] moms,final double [] m)
{
    return getM1A(n,NQVM.NQVMTest.getMM_Matr(n,moms,1),m);
}
static double [] getM1A(final int n,final double [][] matr,final double [] m)
{
    final double [] r=new double [n];
    System.arraycopy(m,0,r,0,n);
    final double [][] vv=new double [n][n];
    for(int k=0;k<n;k++) System.arraycopy(matr,0,vv,0,n);
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa(vv,n,ipvt);
    if(pinfo!=0) java.util.Arrays.fill(r,Double.NaN);
    else com.polytechnik.utils.Linsystems.dgesl(vv,n,ipvt,r,1);
    return r;
}

abstract boolean isDescending();
void test_Monomials_BasisExpansion()
{
    final OQVM<DTYPE> q=getOQVM();
    final double [][] amoms=getMonomialMoments(q.odata);
    final double [] mom_v=amoms[0],mom_pv=amoms[1],mom_p2v=amoms[2],mom_t=amoms[3],mom_dp=amoms[4];

    final double [] kk=getKK(neq,mom_v);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.kk),"conv(q.kk)",kk,"kk",eps);


    final double [] c=getKM(neq,mom_v,mom_v);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.c),"conv(q.c)",c,"c",eps);
    final double [] ca=new double [neq]; ca[0]=1;
    UnitTests.checkDoubleArrayRel(c,"c",ca,"ca",eps);
    System.err.println("q.c="+java.util.Arrays.toString(q.c)+" q.cp="+java.util.Arrays.toString(q.cp)+" q.cp2="+java.util.Arrays.toString(q.cp2));

    final double [] cp=getKM(neq,mom_v,mom_pv);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.cp),"conv(q.cp)",cp,"cp",eps);
    final double [] cp2=getKM(neq,mom_v,mom_p2v);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.cp2),"conv(q.cp2)",cp2,"cp2",eps);
    final double [] cdp=getKM(neq,mom_v,mom_dp);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.cdp),"conv(q.cdp)",cdp,"cdp",eps);
    final double [] ok=NQVM.NQVMTest.getOK_frommoments(neq,mom_v,q.getX0());
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.ok),"conv(q.ok)",ok,"ok",eps);

    final double dpLE_sigma2=DeprecatedHelperTMatr.getA_M_A(neq,cdp,NQVM.NQVMTest.getMM_Matr(neq,mom_t,1));
    UnitTests.checkDoubleRel(q.dpLE_sigma2,"q.dpLE_sigma2",dpLE_sigma2,"dpLE_sigma2",eps);


    final double [] kkt=getKK(neq,mom_t);
    UnitTests.checkDoubleRel(q.dIa,"q.dIa",(kkt[kkt.length-1]/mom_t[0])/(kk[kk.length-1]/mom_v[0]),"kkt/kkv",eps);

    if(neq>1)
    {
	final double ks=(isDescending()?1:-1)*Polynomials.getD1(kk,q.getX0());
	UnitTests.checkDoubleRel(Math.signum(ks),"signum(d/dt kk)",1,"1",eps);

	final double xeq=getXEQ(kk);
	UnitTests.checkDoubleRel(q.xeq,"q.xeq",xeq,"xeq",eps);

	UnitTests.checkDoubleRel(q.peq,"q.peq",q.getPOffset()+Polynomials.getD0(cp,xeq),"cp(xeq)",eps);
	UnitTests.checkDoubleRel(Polynomials.getD1(kk,xeq),"d/dx kk(xeq)",0,"0",eps);
	final double d2=Polynomials.getD2(kk,xeq);
	if(!(d2>0))
	{
	    throw new RuntimeException("d2="+d2);
	}
	System.err.println("xeq="+xeq+" q.xeq="+q.xeq+" d2="+d2);
    }

    System.err.println(name+":OQVMTest:test_Monomials_BasisExpansion(): test OK neq="+neq+" eps="+eps+" class="+q.getClass().getName());
}

double getXEQ(final double [] kk)
{
    final double [] dk=Polynomials.getPolynomialDifferentiated(kk);
    final PolynomialRootsFinder.PRoots u=new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG().findRoots(dk);
    final double [] x=u.re_root,y=u.im_root;

    double xeq;
    if(isDescending())
    {
	xeq=Double.NEGATIVE_INFINITY;
	for(int k=0;k<u.nroots;k++) if(x[k]>=xeq) xeq=x[k];
    }
    else
    {
	xeq=Double.POSITIVE_INFINITY;
	for(int k=0;k<u.nroots;k++) if(x[k]<=xeq) xeq=x[k];
    }
    System.err.println(" Re x="+java.util.Arrays.toString(x)+" Im x="+java.util.Arrays.toString(y));
    return xeq;
}

abstract double [] getdLdpLsigma2(final double [] dp,final double [] ok,final double [] mom_t,final double xeq);
void test_Monomials_DP()
{
    if(neq<=1) return;

    final OQVM<DTYPE> q=getOQVM();
    final double [][] amoms=getMonomialMoments(q.odata);
    final double [] mom_v=amoms[0],mom_t=amoms[3],mom_dp=amoms[4];

    final double [] cdp=getKM(neq,mom_v,mom_dp),kk=getKK(neq,mom_v);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.cdp),"conv(q.cdp)",cdp,"cdp",eps);
    final double [] ok=NQVM.NQVMTest.getOK_frommoments(neq,mom_v,q.getX0());
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.ok),"conv(q.ok)",ok,"ok",eps);
    final double xeq=getXEQ(kk);
    UnitTests.checkDoubleRel(q.xeq,"q.xeq",xeq,"xeq",eps);
    final double [] dLdpLsigma2=getdLdpLsigma2(cdp,ok,mom_t,xeq);
    final double dL=dLdpLsigma2[0],dpL=dLdpLsigma2[1],sigma2=dLdpLsigma2[2],sigma2_0=dLdpLsigma2[3];

    UnitTests.checkDoubleRel(q.dLs,"q.dLs",dL,"dLs",eps);
    UnitTests.checkDoubleRel(q.dLp,"q.dLp",dL,"dLp",eps);
    UnitTests.checkDoubleRel(q.dpL,"q.dpL",dpL,"dpL",eps);
    UnitTests.checkDoubleRel(q.dpLE_sigma2,"q.dpLE_sigma2",sigma2,"sigma2",eps);
    UnitTests.checkDoubleRel(q.dpLE0_sigma2,"q.dpLE0_sigma2",sigma2_0,"sigma2_0",eps);

    System.err.println(name+":OQVMTest:test_Monomials_DP(): test OK neq="+neq+" eps="+eps);
}

void test_NQVM()
{
    final OQVM<DTYPE> q=getOQVM();
    final NQVM<DTYPE> s=getNQVM();
    final double w0=1/q.access.fx(q.kk,q.getX0());
    final double dp0=q.access.fx(q.cdp,q.getX0());
    System.err.println("cdp="+java.util.Arrays.toString(q.cdp)+" w00="+s.w00);
    System.err.println("cdp*w00="+java.util.Arrays.toString(TMatr.getVectorScaled(q.cdp,s.w00)));

    UnitTests.checkDoubleRel(w0,"w0",s.w0,"s.w0",eps);
    UnitTests.checkDoubleRel(q.kka0,"q.kka0",s.w0,"s.w0",eps);
    UnitTests.checkDoubleRel(1/q.access.fx(q.ok,q.getX0()),"1/ok(x0)",s.w0,"s.w0",eps);
    UnitTests.checkDoubleRel(dp0,"dp0",s.dpi0r/s.w00/Math.sqrt(s.wii0r),"s.dpi0r/s.w00/Math.sqrt(s.wii0r)",eps);

     UnitTests.checkDoubleRel(q.dpLE0_sigma2,"q.dpLE0_sigma2",s.sigma2_dpexcess0r/s.w00/s.w00,"s.sigma2_dpexcess0r/s.w00/s.w00",eps);

    System.err.println(name+":OQVMTest:test_NQVM(): test OK neq="+neq+" eps="+eps);
}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":OQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_Monomials_BasisExpansion();
    test_NQVM();
    test_Monomials_DP();

    System.err.println(name+":OQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":OQVMTest: starting tests; ver=$Id: OQVM.java,v 1.58 2022/07/28 12:27:17 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<118;i++)
	{
	    test_Calculations(1250+r.nextInt(7550),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":OQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: OQVM.java,v 1.58 2022/07/28 12:27:17 mal Exp $");
}
} // END OF class OQVMTest

}// END OF class OQVM
