package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/** Basis with one variable not in t-basis.
 *  As a basis we use {Q_0, {(p-pl)Q_k}} basis. This is equivalent to
 *  {Q_0/(p-pl) ,{Q_k}} basis with (p-pl)^2 weight; k=0..n-2.
 *  @version $Id: PEVXData.java,v 1.27 2022/11/20 12:41:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PEVXData {


static double [] getQQMatrInBasisFromMoments(final int n,final double [] m0,final double [] m1,final double [] m2,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int nd=n-1;
    final double [] q0=new double []{1},QQ2=M.B.getQQMatr(nd,m2),QQ=new double [n*n];
    final double vq0=M.B.getD0(q0,0);
    QQ[0+n*0]=vq0*m0[0];
    for(int k=nd;--k>=0;)
    {
	QQ[k+1+n*0]=QQ[0+(k+1)*n]=vq0*m1[k];
	System.arraycopy(QQ2,k*nd,QQ,1+(k+1)*n,nd);
    }
    return QQ;
}
/** Same as getQQMatrInBasisFromMoments. The dimension of input matrices is n-1.
 */
static double [] getQQMatrInBasisFromQQ(final int n,final double [] QQ0,final double [] QQ1,final double [] QQ2)
{
    final int nd=n-1,ndnd=nd*nd;
    if(ndnd!=QQ0.length||ndnd!=QQ1.length||ndnd!=QQ2.length)
	throw new RuntimeException("n="+n+" QQ0.length="+QQ0.length+" QQ1.length="+QQ1.length+" QQ2.length="+QQ2.length);
    final double [] QQ=new double [n*n];
    QQ[0+n*0]=QQ0[0+nd*0];
    for(int k=nd;--k>=0;)
    {
	QQ[k+1+n*0]=QQ[0+(k+1)*n]=QQ1[k+nd*0];
	System.arraycopy(QQ2,k*nd,QQ,1+(k+1)*n,nd);
    }
    return QQ;
}


static double [] getX0MomsInBasis(final int n,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    // At the edge p=pl and all (p-pl)Q_k components are equal to 0. The Q_0 componet is a constant
    final double [] fx0=new double [n],q0=new double []{1};
    fx0[0]=M.B.getD0(q0,0);
    return fx0;
}


public static EVXData getPEVXDataFromQQ(final int n,final double [] QQf,final double [] QQa,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    return _getPEVXDataFromQQ(n,M.EV.getGEV_Generalized(n,QQf,QQa,true),getX0MomsInBasis(n,M));
}

private static EVXData _getPEVXDataFromQQ(final int n,final double [] evS,final double [] fx0)
{

    final double [] psiL=new double [n],psiH=new double [n],psi0=new double [n],psik2_at_x0=new double [n];
    double s_s_o_o=0,ss0=0,s_wL=0,s_wH=0;
    for(int k=0;k<n;k++)
    {
	final double fkx0=DeprecatedHelperTMatr.mulVVo(n,fx0,0,evS,n+(k)*n),f2kx0=fkx0*fkx0;
	DeprecatedHelperTMatr.sumP1(psi0,evS,n+(k)*n,n,fkx0);
	if(k<=0) s_wL=fkx0;
	if(k>=n-1) s_wH=fkx0;
	s_s_o_o+=f2kx0;
	ss0+=f2kx0*evS[k];
    }
    final double sL=evS[0],sH=evS[n-1];
    System.arraycopy(evS,n+(0)*n,psiL,0,n);
    System.arraycopy(evS,n+(n-1)*n,psiH,0,n);
    final double sqrt_s_s_o_o=Math.sqrt(s_s_o_o);
    final double wL=s_wL/sqrt_s_s_o_o,wH=s_wH/sqrt_s_s_o_o,s0=ss0/s_s_o_o;
    for(int k=0;k<n;k++) psi0[k]/=sqrt_s_s_o_o;

    return new EVXData(n,wL,sL,psiL,wH,sH,psiH,psi0,evS,psik2_at_x0,s0,s_s_o_o,Double.NaN);
}


/** Calculate basis |o> and and remaining.
 *  The matrices should be generated by getQQMatrInBasisFromMoments or getQQMatrInBasisFromQQ,
 *  what give matrices with all k!=0 elements having no |o> component.
 */
static double [] getIRadauBasisPsi0FromQQInBasis(final int n,final double [] QQi,final double [] QQt,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int nR=n-1;
    final double [] RQQi=new double [nR*nR],RQQt=new double [nR*nR];
    for(int k=0;k<nR;k++)
    {
	System.arraycopy(QQi,1+(k+1)*n,RQQi,k*nR,nR);
	System.arraycopy(QQt,1+(k+1)*n,RQQt,k*nR,nR);
    }
    final double [] Rev=M.EV.getGEV_Generalized(nR,RQQi,RQQt,true);
    final double [] basis=new double [n+n*n];
    for(int k=0;k<nR;k++)
    {
	System.arraycopy(Rev,nR+k*nR,basis,n+1+(k+1)*n,nR);
    }
    System.arraycopy(Rev,0,basis,1,nR); // Copy eignevalues, all except 0-th are now set.

    // Calculate |psi0> and <psi0|I|psi0>.
    final double [] psi0=getX0MomsInBasis(n,M);
    for(int k=0;k<n;k++){psi0[k]/=Math.sqrt(QQt[k+n*k]);}
    for(int k=0;k<nR;k++){DeprecatedHelperTMatr.sumP1(psi0,basis,n+(k+1)*n,n,-TMatr.getAMatrB(n,psi0,0,basis,n+(k+1)*n,QQt,n));}
    TMatr.scaleVectorInPlace(psi0,1/Math.sqrt(TMatr.getAMatrA(psi0,QQt,n)));
    System.arraycopy(psi0,0,basis,n+0,n);
    basis[0]=TMatr.getAMatrA(psi0,QQi,n);
    return basis;
}


static void test_E0_and_Norm(final int n,final double [] mt0,final double [] mt1,final double [] mt2,
		    final double [] mi0,final double [] mi1,final double [] mi2,
		    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double eps)
{
    final double [] QQt=getQQMatrInBasisFromMoments(n,mt0,mt1,mt2,M),QQi=getQQMatrInBasisFromMoments(n,mi0,mi1,mi2,M);
    if(true)
    {
	final int nd=n-1;
	final double [] QQta=getQQMatrInBasisFromQQ(n,M.B.getQQMatr(nd,mt0),M.B.getQQMatr(nd,mt1),M.B.getQQMatr(nd,mt2));
	UnitTests.checkDoubleArrayRel(QQt,"QQt",QQta,"QQta",eps);
	final double [] QQia=getQQMatrInBasisFromQQ(n,M.B.getQQMatr(nd,mi0),M.B.getQQMatr(nd,mi1),M.B.getQQMatr(nd,mi2));
	UnitTests.checkDoubleArrayRel(QQi,"QQi",QQia,"QQia",eps);
    }
    final EVXData q=getPEVXDataFromQQ(n,QQi,QQt,M);
    final double [] e0basis=getIRadauBasisPsi0FromQQInBasis(n,QQi,QQt,M);
    UnitTests.checkDoubleRel(q.s0,"q.s0",e0basis[0],"e0basis[0]",eps);
    UnitTests.checkDoubleArrayRel(EVXData.getMSign(q.psi0),"q.psi0",EVXData.getMSign(EVXData.getPsiFromBasis(0,e0basis,n)),"psi0 from e0basis",eps);
    if(e0basis[1]<q.sL) throw new RuntimeException("e0basis[1]="+e0basis[1]+"<q.sL="+q.sL);
    if(e0basis[n-1]>q.sH) throw new RuntimeException("e0basis[n-1]="+e0basis[n-1]+">q.sH="+q.sH);


    DeprecatedHelperEVXData.test_orthogonalitySKIP0I("e0basis",n,e0basis,QQi,QQt,eps);
    EVXData.test_orthogonality("q.ev",n,q.ev,QQi,QQt,eps);

    if(true){
	final int nd=n-1;
	final double [] RQQt=M.B.getQQMatr(nd,mt2),RQQi=M.B.getQQMatr(nd,mi2);
	final double [] Rbasis=M.EV.getGEV_Generalized(nd,RQQi,RQQt,true);
	for(int k=0;k<nd;k++)
	{
	    UnitTests.checkDoubleRel(Rbasis[k],"Rbasis[k]",e0basis[k+1],"e0basis[k+1]",eps);
	}
    }
    System.err.println("PEVXData:test_E0_and_Norm test OK n="+n+" eps="+eps);
}


} // END OF class PEVXData
