package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.realtimebook.*;

/** Test whether the John Pierpont Morgan was right.
 *  @version $Id: PkQVM.java,v 1.25 2022/07/28 22:30:50 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class PkQVM<DTYPE> {

final DTYPE odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;

final EVXData I,P2t;
final double x,Vs,V,sens;
final PkxVol PVt,PVt_sam,PVi;

PkQVM(final int n,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();
    this.M=getABasis();

    final double pl_m_poffset=getPLast()-getPOffset();
    final int nmoms=2*n-1;
    final double [] P1tmom,P2tmom;
    {
    final double [] dp0mom=new double [nmoms],dp1mom=new double [nmoms],dp2mom=new double [nmoms];
    access.setM(dp0mom,0,access.M_DP,odata,nmoms);
    access.setM(dp1mom,1,access.M_DP,odata,nmoms);
    access.setM(dp2mom,2,access.M_DP,odata,nmoms);
    Polynomials.shiftPOffsetInPlace(nmoms,dp0mom,dp1mom,dp2mom,null,pl_m_poffset);
    P1tmom=getFmom(dp1mom);
    P2tmom=getFmom(dp2mom);
    }

    final double [] imom=new double [nmoms],tmom=new double [nmoms+1];
    access.setM(imom,0,access.M_RVS,odata,nmoms);
    final double [] Vsmom=getFmom(imom);
    access.setM(imom,0,access.M_I,odata,nmoms);
    final double [] Vmom=getFmom(imom);
    access.setMT(tmom,tmom.length);

    final double [][] QQ=access.getQQ_Matrs(n,1,new double [][]{tmom,imom,Vsmom,P1tmom,P2tmom,Vmom,M.B.getXQkMomentsFromQkMoments(tmom)});
    final double [] QQt=QQ[0],QQi=QQ[1],QQVs=QQ[2],QQP1t=QQ[3],QQP2t=QQ[4],QQV=QQ[5],QQX=QQ[6];

    this.I=DeprecatedHelperEVXData.getEVXDataFromQQ(n,QQi,QQt,access,getX0());
    this.P2t=DeprecatedHelperEVXData.getEVXDataFromQQ(n,QQP2t,QQt,access,getX0());
    final double [] psi=this.P2t.psiL,pert=DeprecatedHelperEVXData.getPsiPertr(this.P2t,0,QQP1t);
    this.x=TMatr.getAMatrA(psi,QQX,n);
    this.Vs=TMatr.getAMatrA(psi,QQVs,n);
    this.V=TMatr.getAMatrA(psi,QQV,n);
    this.sens=getK2(pert,psi,psi,psi)-getK2(psi,psi,pert,psi);
    this.PVt=new PkxVol(n,tmom,P1tmom,P2tmom,M);
    {
	final double [] p0mom=new double [nmoms+1],p1mom=new double [nmoms],p2mom=new double [nmoms];
	access.setM(p0mom,0,access.M_I,odata,p0mom.length);
	access.setM(p1mom,1,access.M_I,odata,nmoms);
	access.setM(p2mom,2,access.M_I,odata,nmoms);
	Polynomials.shiftPOffsetInPlace(nmoms,p0mom,p1mom,p2mom,null,pl_m_poffset);
	this.PVi=new PkxVol(n,p0mom,p1mom,p2mom,M);
	access.setM(p0mom,0,access.M_T,odata,p0mom.length);
	access.setM(p1mom,1,access.M_T,odata,nmoms);
	access.setM(p2mom,2,access.M_T,odata,nmoms);
	Polynomials.shiftPOffsetInPlace(nmoms,p0mom,p1mom,p2mom,null,pl_m_poffset);
//	System.err.println("P1tmom="+java.util.Arrays.toString(P1tmom));
//	System.err.println("p1mom_sam="+java.util.Arrays.toString(p1mom));
//	System.err.println("P2tmom="+java.util.Arrays.toString(P2tmom));
//	System.err.println("p2mom_sam="+java.util.Arrays.toString(p2mom));
	this.PVt_sam=new PkxVol(n,p0mom,p1mom,p2mom,M);
    }
}

double [] getFmom(final double [] moms)
{
    final double [] fmoms=new double [moms.length];
    access.setFMoments(fmoms,moms,fmoms.length);
    //TMatr.scaleVectorInPlace(fmoms,-1);
    return fmoms;
}

abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double getX0();
abstract double getPOffset();
abstract double getPLast();
abstract double getK2(final double [] a,final double [] aa,final double [] b,final double [] bb);

static abstract class PkQVMTest<DTYPE> extends QTest {

abstract PkQVM<DTYPE> getPkQVM();

abstract DkQVM<?> getDkQVM();

PkQVMTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}

void test_PP()
{
    final PkQVM<DTYPE> q=getPkQVM();
    final int n=q.I.neq,nmoms=2*n-1;
    final double [] tmom=new double [nmoms+1],imom=new double [nmoms],amom=new double [nmoms],dpmom=new double [nmoms];
    q.access.setMT(tmom,tmom.length);
    q.access.setM(imom,0,q.access.M_I,q.odata,nmoms);
    q.access.setM(amom,0,q.access.M_A,q.odata,nmoms);
    q.access.setM(dpmom,1,q.access.M_DP,q.odata,nmoms);

    for(double [] r:new double [][]{tmom,amom})
    {
	for(double [] l:new double [][]{q.M.B.getXQkMomentsFromQkMoments(tmom),imom,dpmom})
	{
	    final EVXData E=EVXData.getEVXDataFromQQ(n,q.M.B.getQQMatr(n,l),q.M.B.getQQMatr(n,r),q.M);
	    for(int k=0;k<n;k++)
	    {
		final double [] psi=E.getPsi(k),psi2=q.M.B.mult2Pol(psi,psi);
		final double s1=TMatr.mulVVn1(psi,l),w1=TMatr.mulVVn1(psi,r),
		    s2=TMatr.mulVVn1(psi2,l),w2=TMatr.mulVVn1(psi2,r);
		System.err.println("ev["+k+"]="+E.ev[k]+"\ts1/w1="+(s1/w1)+"\ts2/w2="+(s2/w2)+" w1="+w1+" w2="+w2);
		UnitTests.checkDoubleRel(s2/w2,"s2/w2",E.ev[k],"E.ev[k]",eps);
		UnitTests.checkDoubleRel(s1/w1,"s1/w1",E.ev[k],"E.ev[k]",eps);
	    }
	}
    }

    System.err.println(name+":PkQVMTest:test_PP(): test OK n="+n+" eps="+eps);
}

void test_P2t()
{
    final PkQVM<DTYPE> q=getPkQVM();
    final int n=q.I.neq,nmoms=2*n-1;
    final double pl_m_poffset=q.getPLast()-q.getPOffset();
    final double [] p0mom=new double [nmoms],p1mom=new double [nmoms],p2mom=new double [nmoms];
    q.access.setM(p0mom,0,q.access.M_DP,q.odata,nmoms);
    q.access.setM(p1mom,1,q.access.M_DP,q.odata,nmoms);
    q.access.setM(p2mom,2,q.access.M_DP,q.odata,nmoms);
    Polynomials.shiftPOffsetInPlace(nmoms,p0mom,p1mom,p2mom,null,pl_m_poffset);
    final double [] P1mom=q.getFmom(p1mom),P2mom=q.getFmom(p2mom),tmom=new double [nmoms+1];
    q.access.setMT(tmom,tmom.length);
    final PkxVol s=new PkxVol(n,tmom,P1mom,P2mom,q.M);
    EVXData.matchEVXData(q.P2t,s.P2,false,true,eps);
    UnitTests.checkDoubleRel(q.x,"q.x",s.x,"s.x",eps);
    System.err.println(name+":PkQVMTest:test_P2t(): test OK n="+n+" eps="+eps);
}

void test_Common()
{
    final PkQVM<DTYPE> q=getPkQVM();
    final int n=q.I.neq;
    System.err.println("VOLtp="+java.util.Arrays.toString(q.P2t.getEValues()));
    System.err.println("VOLt="+java.util.Arrays.toString(q.PVt.P2.getEValues()));
    System.err.println("VOLt_sam="+java.util.Arrays.toString(q.PVt_sam.P2.getEValues()));
    System.err.println("VOLi_sam="+java.util.Arrays.toString(q.PVi.P2.getEValues()));

    EVXData.matchEVXData(q.P2t,q.PVt.P2,true,true,eps);

    System.err.println(name+":PkQVMTest:test_Common(): test OK neq="+n+" eps="+eps);
}

double getSignToMatchDkQVMBAsis(){return 1;}

void test_DkQVM()
{
  final PkQVM<DTYPE> q=getPkQVM();
  final DkQVM<?> s=getDkQVM();

  EVXData.matchEVXData(q.I,s.IT,false,true,eps);
  EVXData.matchEVXData(q.P2t,s._P2tt_XX.P2x,false,true,eps);

  UnitTests.checkDoubleRel(q.x,"q.x",getSignToMatchDkQVMBAsis()*(s._P2tt_XX.XP[0]),"s._P2tt_XX.XP[0]",eps);

  System.err.println(name+":PkQVMTest:test_DkQVM(): test OK neq="+neq+" eps="+eps);
}




void test_Extra(){}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":PkQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_DkQVM();
    test_Extra();
    test_P2t();
    test_PP();

    System.err.println(name+":PkQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":PkQVMTest: starting tests; ver=$Id: PkQVM.java,v 1.25 2022/07/28 22:30:50 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":PkQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: PkQVM.java,v 1.25 2022/07/28 22:30:50 mal Exp $");
}
} // END OF class PkQVMTest




}// END OF class PkQVM
