package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVSolver;
import com.polytechnik.outils.DerivativesAtCalculateable;
import com.polytechnik.outils.SimpleRootsFinder;

/** P&L calculation (like PnLICalc, nut no I(p)).
 *  Solving the equition P&L_pnlH(p)+P&L_pnlL(p)=0.
 *  @version $Id: PnLCalc.java,v 1.4 2020/02/06 09:53:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class PnLCalc implements DerivativesAtCalculateable {

    static final int MAXITER=100;
    static final double EPSP=1e-5;

    final int n;
    final double [] bQQi,bQQpi;
    final EVSolver EV;
    final double [] tmp;
    PnLCalc(final int n,final double [] bQQi,final double [] bQQpi,final EVSolver EV)
    {
	this.n=n;
	this.tmp=new double [n*n];
	this.bQQi=bQQi;
	this.bQQpi=bQQpi;
	this.EV=EV;
    }
    double p,I_L,I_H,pnl_L,pnl_H;
    public void calculateAt(final double p)
    {
	this.p=p;
	for(int k=n*n;--k>=0;) tmp[k]=bQQpi[k]-p*bQQi[k];
	final double [] ev=EV.getEV(n,tmp,true);
	final int kL=0,kH=n-1;
	this.I_L=TMatr.getAMatrA(n,ev,n+kL*n,bQQi,n);
	this.I_H=TMatr.getAMatrA(n,ev,n+kH*n,bQQi,n);
	this.pnl_L=ev[kL];
	this.pnl_H=ev[kH];
    }
    public double getF(){return -(pnl_L+pnl_H);}
    public double getDF(){return (I_L+I_H);}
    String getAsText()
    {
	return "PnLCalc:p="+p+" I_L="+I_L+" I_H="+I_H+" getF()="+getF()+" pnl_L="+pnl_L+" pnl_H="+pnl_H+" getDF()="+getDF()+" n="+n;
    }

    double iterateNewton(final double p0)
    {
	return SimpleRootsFinder.findRoots(this,p0,MAXITER,EPSP);
    }

    /** Note! Clober the result.
     */
    void testDPrint(final double px,final double e,final double eps)
    {
	calculateAt(px+e);
	final double apnl_L=-pnl_L,apnl_H=-pnl_H,a_F=getF();
	calculateAt(px-e);
	final double bpnl_L=-pnl_L,bpnl_H=-pnl_H,b_F=getF();
	calculateAt(px);
	final double I_L_diff=(apnl_L-bpnl_L)/(2*e),I_H_diff=(apnl_H-bpnl_H)/(2*e),DF_diff=(a_F-b_F)/(2*e);
	System.err.println("testDPrint:px="+px+" I_L="+I_L+" I_H="+I_H+" getF()="+getF()+" n="+n);
	System.err.println("testDPrint:I_L="+I_L+" -(pnl_L(px+e)-pnl_L(px-e))/2e="+I_L_diff);
	System.err.println("testDPrint:I_H="+I_H+" -(pnl_H(px+e)-pnl_H(px-e))/2e="+I_H_diff);
	System.err.println("testDPrint:px="+px+" getDF="+getDF()+" DF from F="+DF_diff+" getF()="+getF()+" n="+n);
	if(getDF()<0)
	{
	    throw new RuntimeException("getDF()="+getDF()+" must be positive");
	}

	if(eps>=0)
	{ // match values
	    UnitTests.checkDoubleRel(I_L,"I_L",I_L_diff,"I_L_diff",eps);
	    UnitTests.checkDoubleRel(I_H,"I_H",I_H_diff,"I_H_diff",eps);
	    UnitTests.checkDoubleRel(getDF(),"getDF()",DF_diff,"DF_diff",eps);
	}
	iterateNewton(px);
	System.err.println("testDPrint:newton:"+getAsText());
    }
}
