package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;

/**
 * Calculate future skewness of I. Then find equilibrium price,
 * requiring future pI skewness to be equal future I skewness.
 *  @version $Id: PnLFutureSk.java,v 1.22 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLFutureSk {

public final EVXData I;
public final double p_offset,p_IH,P0,I_0_feq,dI,GammaPast,GammaFuture,degeneracy,Pf_SameGammaFuture;
PnLFutureSk(final EVXData I,final double p_offset,
	    final double p_IH,final double pI_0,final double I_0_feq,
	    final double [] psiFk_psi0,final double [] psiFk_I_psiFk,final double [] psiFk_pI_psiFk,
	    final double [] bbasis,
	    final double eps)
{
    this.I=I;
    this.p_offset=p_offset;
    this.p_IH=p_IH;
    this.P0=pI_0/I.s0;
    this.I_0_feq=I_0_feq;
    this.dI=I_0_feq-I.s0;
    final int n=I.neq,iL=0,iH=n-1;
    this.GammaPast=(2*I.s0-I.sL-I.sH)/(I.sL-I.sH);
    this.GammaFuture=(2*I_0_feq-bbasis[iL]-bbasis[iH])/(bbasis[iL]-bbasis[iH]);
    final double [] w0k=TMatr.squareArrayElements(psiFk_psi0);
    final double nomPf=2-w0k[iL]-w0k[iH],nomC=2*pI_0-psiFk_pI_psiFk[iL]-psiFk_pI_psiFk[iH],
	denomPf=w0k[iL]-w0k[iH],denomC=psiFk_pI_psiFk[iL]-psiFk_pI_psiFk[iH];
    // solve the equation: (nomC+dI*Pf*nomPf)/(denomC+dI*Pf*denomPf)=GammaFuture
    this.degeneracy=(nomPf-GammaFuture*denomPf);
    final double PfdI=(GammaFuture*denomC-nomC)/degeneracy;
    this.Pf_SameGammaFuture=PfdI/dI;

    if(eps>0)
    {
	if(n==2)
	{
	    UnitTests.checkDoubleRel(GammaPast,"GammaPast",Math.pow(I.wL,2)-Math.pow(I.wH,2),"Weights diff",eps);
	    UnitTests.checkDoubleRel(GammaFuture,"GammaFuture",w0k[iL]-w0k[iH],"Weights diff",eps);
	}
	// check I_0_feq
	UnitTests.checkDoubleRel(TMatr.mulVVn1(w0k,bbasis),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(w0k),"<psi0|psi0>",1,"1",eps);
	for(int k=0;k<n;k++)
	{
	    UnitTests.checkDoubleRel(psiFk_I_psiFk[k]+dI*Math.pow(psiFk_psi0[k],2),"<psiFk|If|PsiFk>",bbasis[k],"bbasis[k]",eps);
	}
	System.err.println("PnLFutureSk.PnLFutureSk selftest2 OK degeneracy="+degeneracy+" dI="+dI+" Pf_SameGammaFuture="+Pf_SameGammaFuture+" n="+n+" eps="+eps);
    }
}


/**
 * Solve <psi|I|psi>/<psi|psi>=lambda extremum problem, find eigenstates & eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-<psi0|I|psi0> (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + |psi0>dI<psi0|
 * Find the eigenstates of this new I^{f} matrix,
 * in dynamic impact approximation its eigenstates should be also the P^{f} operator eigenstates.
 *
 * For the moments of |psik> state calculate the price, manipulating with the skewness
 * for IL & IH only.
 */
public static PnLFutureSk get_PnLFutureSk(final MomentsData mm,
					  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
					  final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final int ixH=n-1; // the state with maximal I
    final double I_0=I.s0,I_0_feq=I.ev[ixH],dI=I_0_feq-I_0;

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev);
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];

    // OK, now obtained: bQQi_feq: |I| oprator with an impact from the future, and bQQpi: |pI| operator without an impact from the future.

    // Diagonalize "future" |I|:
    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true),
	psiFk_psi0=new double [n],psiFk_I_psiFk=new double[n],psiFk_pI_psiFk=new double[n];
    for(int k=0;k<n;k++)
    {
	psiFk_psi0[k]=DeprecatedHelperTMatr.mulVVo(n,bbasis,n+k*n,bpsi0,0);
	psiFk_I_psiFk[k]=TMatr.mulV2V(EVXData.getPsiFromBasis(k,bbasis,n),I.getEValues());
	psiFk_pI_psiFk[k]=TMatr.getAMatrA(n,bbasis,n+k*n,bQQpi,n);
    }
    final double pI0=TMatr.getAMatrA(bpsi0,bQQpi,n);

    if(eps>0)
    {
	EVXData.test_orthogonality("PnLFutureSk.get_PnLFutureSk:I.ev",n,I.ev,mm.QQi,mm.QQt,eps); // old basis
	EVXData.test_orthogonality("PnLFutureSk.get_PnLFutureSk:bbasis",n,bbasis,bQQi_feq,TMatr.getUnitMatrix(n),eps); // new basis
	// check some matrix elements
	UnitTests.checkDoubleRel(TMatr.mulVV(psiFk_psi0,psiFk_psi0),"<psi0|psi0>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(psiFk_psi0,java.util.Arrays.copyOf(bbasis,n)),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(bpsi0,bpsi0),"<psi0|psi0>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(bpsi0,I.getEValues()),"<psi0|I|psi0>",I.s0,"I.s0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|I^f|psi0>",I_0_feq,"I_0_feq",eps);
	// check spur invariance
	UnitTests.checkDoubleRel(TMatr.sumElements(I.getEValues()),"sum(I.ev)",TMatr.sumElements(psiFk_I_psiFk),"sum(psiFk_I_psiFk)",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(I.getEValues())+dI,"sum(I.ev)+dI",TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),"sum(I^{f})",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQpi),"Spur(bQQpi)",TMatr.sumElements(psiFk_pI_psiFk),"sum(psiFk_pI_psiFk)",eps);
	System.err.println("PnLFutureSk.get_PnLFutureSk selftest OK n="+n+" eps="+eps);
    }

    final PnLFutureSk s=new PnLFutureSk(I,mm.p_offset,
					bQQpi[ixH+n*ixH]/I.ev[ixH],pI0,I_0_feq,
					psiFk_psi0,psiFk_I_psiFk,psiFk_pI_psiFk,bbasis,
					eps);
    if(eps>0)
    {
	if(n==2)
	{
	    final Skewness g=Skewness.getG0Skewness(mm.QQt,mm.QQi,M);
	    UnitTests.checkDoubleRel(g.sk,"skewness(I)",s.GammaPast,"s.GammaPast",eps);
	    UnitTests.checkDoubleRel(g.xa,"average(I)",s.I.s0,"s.I.s0",eps);
	    UnitTests.checkDoubleRel(g.x1,"IL",s.I.sL,"s.I.sL",eps);
	    UnitTests.checkDoubleRel(g.x2,"IH",s.I.sH,"s.I.sH",eps);
	}
	final double dIPf=dI*s.Pf_SameGammaFuture;
	final double [] bQQ0=TMatr.getQQaa(bpsi0),bQQpi_feq=TMatr.sumAcB(bQQpi,dIPf,bQQ0),
	     w0k=TMatr.squareArrayElements(psiFk_psi0);
	final double piLF=TMatr.getAMatrA(n,bbasis,n+0*n,bQQpi_feq,n),
	    piHF=TMatr.getAMatrA(n,bbasis,n+(ixH)*n,bQQpi_feq,n),
	    pi0F=TMatr.getAMatrA(bpsi0,bQQpi_feq,n),
	    GammaF=(2*pi0F-piLF-piHF)/(piLF-piHF),
	    GammaiF=(2*I_0_feq-bbasis[0]-bbasis[ixH])/(bbasis[0]-bbasis[ixH]);
	UnitTests.checkDoubleRel(piLF,"piLF",psiFk_pI_psiFk[0]+w0k[0]*dIPf,"piLF",eps);
	UnitTests.checkDoubleRel(piHF,"piHF",psiFk_pI_psiFk[ixH]+w0k[ixH]*dIPf,"piHF",eps);
	UnitTests.checkDoubleRel(pi0F,"pi0F",pI0+dIPf,"pi0F",eps);
	UnitTests.checkDoubleRel(GammaF,"GammaF",s.GammaFuture,"s.GammaFuture",eps);
	UnitTests.checkDoubleRel(GammaF,"GammaF",GammaiF,"GammaiF",eps);
	System.err.println("PnLFutureSk.get_PnLFutureSk selftest GammaF OK n="+n+" eps="+eps);
    }
    return s;
}


}
