package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;

/**
 *  @version $Id: PnLI0.java,v 1.64 2022/07/01 21:23:39 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLI0 {
    public final EVXData I;
    public final double p_offset,p_IH,I_0,I_L,I_H,I_0_eq,I_H_eq,P_eq,P_phi0,P1_eq,rf,PeqqdI,Peqq;
    PnLI0(final EVXData I,final double p_offset,final double p_IH,
	  final double I_0, final double I_L, final double I_H,
	  final double I_0_eq,final double I_H_eq,
	  final double P_eq,final double P_phi0,final double P1_eq,final double rf,
	  final double PeqqdI)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.I_0=I_0;
	this.I_L=I_L;
	this.I_H=I_H;
	this.I_0_eq=I_0_eq;
	this.I_H_eq=I_H_eq;
	this.P_eq=P_eq;
	this.P_phi0=P_phi0;
	this.P1_eq=P1_eq;
	this.rf=rf;
	this.PeqqdI=PeqqdI;
	this.Peqq=PeqqdI/(I_0_eq-I_0);
    }

/** Same as get_PnLI0_NO_0_atBoundary
 * but "future" equilibrium I_0_eq is calculated with boundary conditions.
 */
static PnLI0 get_PnLI0(final MomentsData mm,
		       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		       final double eps)
{
    return get_PnLI0(mm,M,true,eps);
}

private static PnLI0 get_PnLI0(final MomentsData mm,
		       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		       final boolean use_threshold_with_boundary_condition,
		       final double eps)
{
    final int n=mm.n,iH=n-1,i0=0;
    if(n<2||i0!=0) throw new RuntimeException("n="+n+"<2");
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,mm.QQi,mm.QQt,M),
	bQQi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQi,basis),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,basis);
    final double [] bphi=new double [n],bpsi0=new double [n];
    System.arraycopy(bQQi,i0*n,bphi,0,n);
    bphi[i0]=0;
    TMatr.scaleVectorInPlace(bphi,1/Math.sqrt(TMatr.mulVV(bphi,bphi)));
    bpsi0[i0]=1;
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0_eq;
    if(use_threshold_with_boundary_condition)
    {
	I_0_eq=basis[iH];
    }
    else
    {
	// For get_PnLI0_NO_0_atBoundary unit test we need
	// a threshold calculated without psi(x0)=0 boundary condition.
	// With this threshold the P_eq & I_H_eq match
	// exactly the ones calculated by the get_PnLI0_NO_0_atBoundary.
	I_0_eq=I.sH;
    }
    final double p_IH=bQQpi[iH+n*iH]/basis[iH],
	psi0_I_phi=TMatr.getAMatrB(bpsi0,bphi,bQQi,n),
	phi_I_phi=TMatr.mulV2V(bphi,java.util.Arrays.copyOf(basis,n)), // the i0-th element is 0
	phi_pI_phi=TMatr.getAMatrA(bphi,bQQpi,n),
	psi0_pI_phi=TMatr.getAMatrB(bpsi0,bphi,bQQpi,n),
	psi0_pI_psi0=TMatr.getAMatrA(bpsi0,bQQpi,n),
	beta_eq=-psi0_I_phi/(phi_I_phi-I_0_eq),
	dI_H_eq=beta_eq*(2*psi0_I_phi+beta_eq*(phi_I_phi-I_0_eq)),
	I_H_eq=I_0_eq+dI_H_eq,
	P1_eq=DeprecatedHelperTMatr.mulVVo(n-1,bQQpi,i0*n+1,bQQi,i0*n+1)/DeprecatedHelperTMatr.mulVVo(n-1,bQQi,i0*n+1,bQQi,i0*n+1),
	vvP_eq=(phi_pI_phi+psi0_pI_phi/beta_eq)/I_0_eq,
	P_eq=(psi0_pI_phi/beta_eq+phi_pI_phi)/I_0_eq,
	PeqqdI=psi0_pI_phi/beta_eq+phi_pI_phi-psi0_pI_psi0;
    if(eps>0)
    {
	DeprecatedHelperEVXData.test_orthogonalitySKIP0I("PnLI0.get_PnLI0",n,basis,mm.QQi,mm.QQt,eps);
	final double x0=M.getRadauRoot();
	final double [] i0_I_k=new double [n],i0_pI_k=new double [n];
	for(int k=i0+1;k<n;k++){
	    i0_I_k[k]=TMatr.getAMatrB(n,basis,n+i0*n,basis,n+k*n,mm.QQi,n);
	    i0_pI_k[k]=TMatr.getAMatrB(n,basis,n+i0*n,basis,n+k*n,mm.QQpi,n);
	    final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	    UnitTests.checkDoubleRel(M.B.getD0(psik,x0),"psik(x0)",0,"0",eps);
	}
	final double _p_IH=TMatr.getAMatrA(n,basis,n+iH*n,mm.QQpi,n)/TMatr.getAMatrA(n,basis,n+iH*n,mm.QQi,n),
	    i0_eq=I_0_eq,
	    jjp=TMatr.mulVV(i0_I_k,i0_pI_k),
	    jj=TMatr.mulVV(i0_I_k,i0_I_k),
	    jjI=TMatr.mulV2V(i0_I_k,java.util.Arrays.copyOf(basis,n)),
	    _jjI=TMatr.getAMatrA(i0_I_k,bQQi,n),
	    _jjpI=TMatr.getAMatrA(i0_I_k,bQQpi,n),
	    abeta_eq=-jj/(jjI-i0_eq*jj),
	    ih_eq=i0_eq+jj*abeta_eq,
	    p1_eq=jjp/jj,
	    vvp_eq=(jjp/abeta_eq+_jjpI)/(i0_eq*jj),
	    const_eq=(jj/abeta_eq+_jjI)/(i0_eq*jj);
	System.err.println("PnLI0.get_PnLI0 i0_eq="+i0_eq+" ih_eq="+ih_eq+"  beta_eq="+beta_eq+" must be >0 i0_I_k="+java.util.Arrays.toString(i0_I_k));
	UnitTests.checkDoubleRel(p_IH,"p_IH",_p_IH,"_p_IH",eps);

	UnitTests.checkDoubleRel(I_0_eq,"I_0_eq",i0_eq,"i0_eq",eps);
	UnitTests.checkDoubleRel(1,"1",const_eq,"const_eq",eps);
	UnitTests.checkDoubleRel(jjI,"jjI",_jjI,"_jjI",eps);
	if(beta_eq<0) throw new RuntimeException("beta_eq="+beta_eq+" <0");
	UnitTests.checkDoubleRel(I_H_eq,"I_H_eq",ih_eq,"ih_eq",eps);
	UnitTests.checkDoubleRel(P1_eq,"P1_eq",p1_eq,"p1_eq",eps);
	UnitTests.checkDoubleRel(vvP_eq,"vvP_eq",vvp_eq,"vvp_eq",eps);
	UnitTests.checkDoubleRel(vvP_eq,"vvP_eq",P_eq,"P_eq",eps);
	UnitTests.checkDoubleRel(PeqqdI,"PeqqdI",P_eq*I_0_eq-psi0_pI_psi0,"P_eq*I_0_eq-psi0_pI_psi0",eps);
    }

    return new PnLI0(I,mm.p_offset,p_IH,basis[i0],basis[1],basis[iH],
		     I_0_eq,I_H_eq,P_eq,phi_pI_phi/phi_I_phi,psi0_pI_phi/psi0_I_phi,(psi0_I_phi/Math.sqrt(/*phi_phi*/1))/I_0_eq,
		     PeqqdI);
}

/** Consider <psi|I|psi>/<psi|psi>=lambda extremums, the solutions of I|psi>=lambda|psi>.
 * Then consider psi0=\sum psi_k(x)*psi_k(x0)/sqrt(\sum psi_k(x0)*psi_k(x0))
 * this function is othogonal to all psi with psi(x0)=0 boundary condition.
 * Future assumption: equilibrium <psi0|I|psi0> will be equal to max lambda=I_H.
 * Consider |phi'>=\sum_k <psi0|I|psik> |psi_k>, defining the "direction" of max I change from "now",
 * with psi0 subtracted |phi>=|phi'>-<phi'|psi0>|psi0> (here phi(x0)=0)
 * and maximize <psi0+beta*phi|I|psi0+beta*phi>/<psi0+beta*phi|psi0+beta*phi>.
 * over "beta" to obtain I_H_eq, which is always greater than I_H (this is correct as
 * long equilibrium <psi0|I|psi0> is assumed to be greater equal than I_H).
 * Then consider price and I on the state |psi0+beta_eq*phi>.
 * The equilibrium price is chosen as following:
 * beta_eq is chosen as the one, providing max I=I_H_eq (the maximization above).
 * Then consider pI and find virtual "last" price P_eq.
 * See "Market Dynamics. On Supply and Demand Concepts, Feb 14, 2016, arXiv:1602.04423",
 * where I guess (based on experimental observations) that "pI" and "I"
 * have an extremum on the same |psi> state (in dynamic impact approximation).
 */
public static PnLI0 get_PnLI0_NO_0_atBoundary(final MomentsData mm,
		       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		       final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev);
    final double p_IH=bQQpi[(n-1)+n*(n-1)]/I.sH;

    final double [] bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(TMatr.mulVV(I.psik_at_x0,I.psik_at_x0))),bphi=TMatr.getVAB(I.getEValues(),bpsi0);
    TMatr.sumP1(bphi,bpsi0,-TMatr.mulVV(bphi,bpsi0));

    final double I_0_eq=I.sH,
	psi0_phi=TMatr.mulVV(bpsi0,bphi),psi0_I_phi=TMatr.mulVVV(bpsi0,I.getEValues(),bphi),
	phi_phi=TMatr.mulVV(bphi,bphi),phi_I_phi=TMatr.mulVVV(bphi,I.getEValues(),bphi),
	psi0_pI_phi=TMatr.getAMatrB(bpsi0,bphi,bQQpi,n),
	phi_pI_phi=TMatr.getAMatrA(bphi,bQQpi,n),
	psi0_pI_psi0=TMatr.getAMatrA(bpsi0,bQQpi,n),

	beta_eq=-psi0_I_phi/(phi_I_phi-I_0_eq*phi_phi),
	I_H_eq=I_0_eq+beta_eq*(2*psi0_I_phi+beta_eq*(phi_I_phi-I_0_eq*phi_phi)),
	vvP_eq=(phi_pI_phi+psi0_pI_phi/beta_eq)/(I_0_eq*phi_phi),
	P_eq=(psi0_pI_phi/beta_eq+phi_pI_phi)/(I_0_eq*phi_phi),
	PeqqdI=((psi0_pI_phi/beta_eq)+phi_pI_phi)/phi_phi-psi0_pI_psi0;

    if(eps>0)
    {
	final double [] phi=EVXData.convertPsiFromBtoQ_Basis(n,bphi,I.ev),_bQQi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQi,I.ev);
	System.err.println("Ip="+(beta_eq* (2*psi0_I_phi+beta_eq*phi_I_phi))+" I_H_eq="+I_H_eq+" dI="+(I_H_eq-I_0_eq));
	UnitTests.checkDoubleRel(M.B.getD0(phi,I.x0),"phi(x0)",0,"0",eps);
	UnitTests.checkDoubleRel(psi0_phi,"<bpsi0|bphi>",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(bpsi0,bpsi0),"<bpsi0|bpsi0>",1,"1",eps);
	UnitTests.checkDoubleRel((TMatr.getAMatrA(bphi,_bQQi,n)+TMatr.getAMatrB(bpsi0,bphi,_bQQi,n)/beta_eq)/(I_0_eq*phi_phi),"P=const",1,"1",eps);
	UnitTests.checkDoubleRel(I_H_eq,"I_H_eq",I_0_eq+beta_eq*psi0_I_phi,"I_0_eq+beta_eq*<psi0|I|phi>",eps);
	UnitTests.checkDoubleRel(phi_phi,"<bphi|bphi>",TMatr.getAMatrA(phi,mm.QQt,n),"<phi|phi>",eps);
	UnitTests.checkDoubleRel(psi0_I_phi,"<bpsi0|I|bphi>",TMatr.getAMatrB(I.psi0,phi,mm.QQi,n),"<psi0|I|phi>",eps);
	UnitTests.checkDoubleRel(phi_I_phi,"<bphi|I|bphi>",TMatr.getAMatrA(phi,mm.QQi,n),"<phi|I|phi>",eps);
	UnitTests.checkDoubleRel(psi0_pI_psi0,"<bpsi0|pI|bpsi0>",TMatr.getAMatrA(I.psi0,mm.QQpi,n),"<psi0|pI|psi0>",eps);
	UnitTests.checkDoubleRel(psi0_pI_phi,"<bpsi0|pI|bphi>",TMatr.getAMatrB(I.psi0,phi,mm.QQpi,n),"<psi0|pI|phi>",eps);
	UnitTests.checkDoubleRel(phi_pI_phi,"<bphi|pI|bphi>",TMatr.getAMatrA(phi,mm.QQpi,n),"<phi|pI|phi>",eps);
	// make sure that <phi|phi>=<phi|I|psi0>
	UnitTests.checkDoubleRel(phi_phi,"<bphi|bphi>",psi0_I_phi,"<bpsi0|I|bphi>",eps);
	UnitTests.checkDoubleRel(beta_eq,"beta_eq",1/(I_0_eq-phi_I_phi/phi_phi),"1/(I_0_eq-<phi|I|phi>/<phi|phi>",eps);
	UnitTests.checkDoubleRel(vvP_eq,"vvP_eq",P_eq,"P_eq",eps);
	UnitTests.checkDoubleRel(PeqqdI,"PeqqdI",P_eq*I_0_eq-psi0_pI_psi0,"P_eq*I_0_eq-psi0_pI_psi0",eps);
    }
    if(eps>0)
    {
	final PnLI0 p=get_PnLI0(mm,M,false,eps);
	UnitTests.checkDoubleRel(I.s0,"I.s0",p.I_0,"p.I_0",eps);
	UnitTests.checkDoubleRel(I_0_eq,"I_0_eq",p.I_0_eq,"p.I_0_eq",eps);
	UnitTests.checkDoubleRel(I_H_eq,"I_H_eq",p.I_H_eq,"p.I_H_eq",eps);
	UnitTests.checkDoubleRel(P_eq,"P_eq",p.P_eq,"p.P_eq",eps);
	System.err.println("PnLI0.get_PnLI0_NO_0_atBoundary mathch get_PnLI0 OK n="+n+" eps="+eps);
    }
    if(eps>0)
    { // P_eq calculation require only I_0_eq to be known
      // from eigenvalues problem. When I_0_eq is known
      // the result can be obtained just with QQt inverted.
      // UNIT TEST: P_eq calculations in original basis without eigenstates expansion.
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt),
	    phi=TMatr.getMatr_x_A(n,IQQt,TMatr.getMatr_x_A(n,mm.QQi,I.psi0));
	TMatr.sumP1(phi,I.psi0,-TMatr.getAMatrB(phi,I.psi0,mm.QQt,n));
	UnitTests.checkDoubleRel(M.B.getD0(phi,I.x0),"phi(x0)",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrB(phi,I.psi0,mm.QQt,n),"<phi|psi0>",0,"0",eps);
	final double _psi0_I_phi=TMatr.getAMatrB(I.psi0,phi,mm.QQi,n),
	    _phi_I_phi=TMatr.getAMatrA(phi,mm.QQi,n),
	    _phi_phi=TMatr.getAMatrA(phi,mm.QQt,n),
	    _psi0_pI_phi=TMatr.getAMatrB(I.psi0,phi,mm.QQpi,n),
	    _phi_pI_phi=TMatr.getAMatrA(phi,mm.QQpi,n),
	    _beta_eq=-_psi0_I_phi/(_phi_I_phi-I_0_eq*_phi_phi),
	    _ih_eq=I_0_eq+_beta_eq*_psi0_I_phi,
	    _vvP_eq=(_phi_pI_phi+_psi0_pI_phi/_beta_eq)/(I_0_eq*_phi_phi),
	    _P_eq=(_phi_pI_phi+_psi0_pI_phi/_beta_eq)/(_phi_I_phi+_psi0_I_phi/_beta_eq);
	UnitTests.checkDoubleRel(I_H_eq,"I_H_eq",_ih_eq,"_ih_eq",eps);
	UnitTests.checkDoubleRel(vvP_eq,"vvP_eq",_vvP_eq,"_vvP_eq",eps);
	UnitTests.checkDoubleRel(P_eq,"P_eq",_P_eq,"_P_eq",eps);
	//if(_beta_eq<0) throw new RuntimeException("_beta_eq="+_beta_eq+" <0");
	System.err.println("PnLI0.get_PnLI0_NO_0_atBoundary mathc without eigenproblem OK n="+n+" eps="+eps);
    }
    if(eps>0)
    {
	EVXData.test_orthogonality("PnLI0.get_PnLI0_NO_0_atBoundary",n,I.ev,mm.QQi,mm.QQt,eps);
	System.err.println("PnLI0.get_PnLI0_NO_0_atBoundary selftest OK n="+n+" eps="+eps);
    }

    return new PnLI0(I,mm.p_offset,p_IH,I.s0,I.sL,I.sH,
		     I_0_eq,I_H_eq,
		     P_eq,phi_pI_phi/phi_I_phi,psi0_pI_phi/psi0_I_phi,
		     (psi0_I_phi/Math.sqrt(phi_phi))/I_0_eq,PeqqdI);
}

public static void matchPnLI0(final PnLI0 a,final PnLI0 b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0,"a.I_0",b.I_0,"b.I_0",eps);
    UnitTests.checkDoubleRel(a.I_L,"a.I_L",b.I_L,"b.I_L",eps);
    UnitTests.checkDoubleRel(a.I_H,"a.I_H",b.I_H,"b.I_H",eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);

    UnitTests.checkDoubleRel(a.I_0_eq,"a.I_0_eq",b.I_0_eq,"b.I_0_eq",eps);
    UnitTests.checkDoubleRel(a.I_H_eq,"a.I_H_eq",b.I_H_eq,"b.I_H_eq",eps);
    UnitTests.checkDoubleRel(a.rf,"a.rf",b.rf,"b.rf",eps);
    UnitTests.checkDoubleRel(a.P_eq,"a.P_eq",b.P_eq+p_offset_shift,"b.P_eq+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.Peqq,"a.Peqq",b.Peqq+p_offset_shift,"b.Peqq+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.P_phi0,"a.P_phi0",b.P_phi0+p_offset_shift,"b.P_phi0+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.P1_eq,"a.P1_eq",b.P1_eq+p_offset_shift,"b.P1_eq+p_offset_shift",eps);
    System.err.println("PnLI0.matchPnLI0 test OK n="+a.I.neq+" eps="+eps);
}

public static void matchPnLI0e(final PnLI0 a,final PnLI0e b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.I_0_eq,"a.I_0_eq",b.I_0_feq,"b.I_0_feq",eps);
    UnitTests.checkDoubleRel(a.Peqq,"a.Peqq",b.Peqq+p_offset_shift,"b.Peqq+p_offset_shift",eps);

    System.err.println("PnLI0.matchPnLI0e test OK n="+a.I.neq+" eps="+eps);
}
}
