package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLI0B.java,v 1.22 2022/07/28 12:27:18 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLI0B {

    public final EVXData I;
    public final double p_offset,p_IH,pb_IHb,pb_IH,Pf_IH,P0f;
    PnLI0B(final EVXData I,final double p_offset,final double p_IH,final double pb_IHb,final double pb_IH,
	   final double Pf_IH,final double P0f
	   )
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.pb_IHb=pb_IHb;
	this.pb_IH=pb_IH;
	this.Pf_IH=Pf_IH;
	this.P0f=P0f;
    }

/**
 * Test version of "retarted potential" version.
 * Instead of considering "advanced potential"
 * I^{f}= I + |psi0>dI<psi0|  with dI=lambda^{IH}-<psi0|I|psi0>
 * consider dI=-<psi0|I|psi0>
 * Then look at p_IH.
 * Do not expect much, this is just a try.
 */
public static PnLI0B get_PnLI0B(final MomentsData mm,
				  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				  final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=-I_0;
    final int ixH=n-1;

    final double [] bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev),bpsiH=new double [n];
    bpsiH[ixH]=1;
    final double pI_0=TMatr.getAMatrA(bpsi0,bQQpi,n);

    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o));
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];
    final double [] bQQpi_feq=TMatr.sumAcB(bQQpi,-pI_0,TMatr.getQQaa(bpsi0)),
	bbasis=M.EV.getEV(n,bQQi_feq,true);
    final double pb_IHb=TMatr.getAMatrA(n,bbasis,n+ixH*n,bQQpi_feq,n)/bbasis[ixH],
	pb_IH=TMatr.getAMatrA(bpsiH,bQQpi_feq,n)/TMatr.getAMatrA(bpsiH,bQQi_feq,n);

    double sn=0,sd=0,wh=TMatr.mulVV(bpsiH,bpsi0),sb0=0,sI0=0,spI0=0,s4=0;
    for(int k=0;k<n;k++)
    {
	sb0+=bbasis[k]*Math.pow(DeprecatedHelperTMatr.mulVVo(n,bbasis,n+k*n,bpsi0,0),2);
	final double wk=bpsi0[k];
	s4+=Math.pow(wk,4);
	sI0+=bQQi_feq[k+n*k]*Math.pow(wk,2);
	spI0+=bQQpi_feq[k+n*k]*Math.pow(wk,2);
	if(k==ixH) continue;
	sd+=Math.pow(wh*wk,2);
	sn+=bQQpi_feq[ixH+k*n]*wh*wk;
    }
    //System.err.println("sb0="+sb0+" <psi0|bQQi_feq|psi0>="+TMatr.getAMatrA(bpsi0,bQQi_feq,n));
    final double Pf_IH=-sn/(sd*I.s0),I0_f=sI0/(1-s4),pI0_f=spI0/(1-s4);
    if(eps>0)
    {
	EVXData.test_orthogonality("PnLI0B.get_PnLI0B:I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQpi_feq,n),"<psi0|pIb|psi0",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|Ib|psi0",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(bpsi0,I.getEValues()),"<psi0|I|psi0",I.s0,"I.s0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(n,bbasis,n+ixH*n,bQQi_feq,n),"<psiH|Ib|psiH>",bbasis[ixH],"bbasis[ixH]",eps);
	UnitTests.checkDoubleArrayRel(bQQpi,"bQQpi",TMatr.sumAcB(bQQpi_feq,pI_0,TMatr.getQQaa(bpsi0)),"bQQpi_feq+ |psi0>pI_0<psi0|",eps);
	UnitTests.checkDoubleRel(sd,"sd",wh*wh*(1-wh*wh),"wh^2*(1-wh^2)",eps);
	UnitTests.checkDoubleRel(sb0,"sb0",0,"0",eps);
	UnitTests.checkDoubleRel(I0_f,"I0_f",I.s0,"I.s0",eps);
	System.err.println("PnLI0B.get_PnLI0B selftest OK n="+n+" eps="+eps);
    }

    return new PnLI0B(I,mm.p_offset,bQQpi[ixH+n*ixH]/I.ev[ixH],pb_IHb,pb_IH,Pf_IH,pI0_f/I0_f);
}


}
