package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLI0IH.java,v 1.23 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLI0IH {

    public final EVXData I;
    public final double p_offset,p_IH,I_0_feq,PfdI,Pf,Pfi;
    PnLI0IH(final EVXData I,final double p_offset,final double p_IH,final double I_0_feq,final double PfdI,
	    final double [] ev_I_feq,final double [] ev_pI_feq)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.I_0_feq=I_0_feq;
	final double dI=I_0_feq-I.s0;
	this.PfdI=PfdI;
	this.Pf=PfdI/dI;
	this.Pfi=(ev_pI_feq[1]-ev_pI_feq[0])/(ev_I_feq[1]-ev_I_feq[0]);
    }

/**
 * Solve <psi|I|psi>/<psi|psi>=lambda extremum problem, find eigenfunction & eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-<psi0|I|psi0> (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + |psi0>dI<psi0|
 * solve I^{f}->max in 2D basis: (psi^{IH}, psi0)
 * Then consider pI^{f} (future pI): pI^{f}= pI + |psi0>P^{f}dI<psi0|
 * to choose P^{f} in a way I^{f} and pI^{f} have the same eigenvectors
 * in 2D subspace (psi^{IH}, psi0).
 */
public static PnLI0IH get_PnLI0IH(final MomentsData mm,
				  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				  final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final int ixH=n-1,indB=1; // the state with maximal I

    final double [] bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	bpsiH=new double [n],bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev);
    bpsiH[ixH]=1;
    final double [] bphi=TMatr.sumAcB(bpsiH,-TMatr.mulVV(bpsiH,bpsi0),bpsi0);

    final double psi0_I_phi=TMatr.mulVVV(bpsi0,I.getEValues(),bphi),
	phi_I_phi=TMatr.mulV2V(bphi,I.getEValues()),
	phi_phi=TMatr.mulVV(bphi,bphi),
	psi0_pI_psi0=TMatr.getAMatrA(bpsi0,bQQpi,n),
	psi0_pI_phi=TMatr.getAMatrB(bpsi0,bphi,bQQpi,n),
	phi_pI_phi=TMatr.getAMatrA(bphi,bQQpi,n);
    final double [] evf=M.EV.getGEV_Generalized(2,new double []{I_0_feq,psi0_I_phi,psi0_I_phi,phi_I_phi},
						new double []{1,0,0,phi_phi},true);
    final double beta=evf[2+2*indB+1]/evf[2+2*indB+0],nrm=1+Math.pow(beta,2)*phi_phi,
	PfdI=nrm*(psi0_pI_phi/beta+phi_pI_phi)/phi_phi
	 - beta*(2*psi0_pI_phi+beta*phi_pI_phi) - psi0_pI_psi0
	;
    final double [] mpI_feq=new double []{psi0_pI_psi0+PfdI,psi0_pI_phi,psi0_pI_phi,phi_pI_phi},
	ev_pI_feq=new double []{TMatr.getAMatrA(2,evf,2+2*0,mpI_feq,2),TMatr.getAMatrA(2,evf,2+2*1,mpI_feq,2)};

    if(eps>0)
    {
	EVXData.test_orthogonality("PnLI0IH.get_PnLI0IH:I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	// Check spur invariance
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQi),"Spur(IQQt,QQi)",TMatr.sumElements(I.getEValues()),"sum(I.ev)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQpi),"Spur(IQQt,QQpi)",TMatr.getSpurM(n,bQQpi),"Spur(bQQpi)",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(bpsi0,bphi),"<psi0|phi>",0,"0",eps);

	// Create bQQpi_feq (operator of "future" pI), with the |bpsi0>PfdI<bpsi0| contribution from the future:
	// bQQpi_feq = bQQpi + |bpsi0>PfdI<bpsi0|
	final double [] bQQpi_feq=TMatr.sumAcB(bQQpi,PfdI,TMatr.getQQaa(bpsi0)),
	    psiB0=EVXData.getPsiFromBasis(0,evf,2),
	    psiB1=EVXData.getPsiFromBasis(1,evf,2),
	    evpf=M.EV.getGEV_Generalized(2,new double []{
		TMatr.getAMatrA(bpsi0,bQQpi_feq,n),
		TMatr.getAMatrB(bpsi0,bphi,bQQpi_feq,n),
		TMatr.getAMatrB(bpsi0,bphi,bQQpi_feq,n),
		TMatr.getAMatrA(bphi,bQQpi_feq,n)},
		new double []{1,0,0,phi_phi},true);
	System.err.println("PnLI0IH.get_PnLI0IH: I_0_feq="+I_0_feq+" evf[indB]="+evf[indB]+" PfdI="+PfdI+" beta="+beta);
	System.err.println("PnLI0IH.get_PnLI0IH: evf="+java.util.Arrays.toString(evf));
	System.err.println("PnLI0IH.get_PnLI0IH: evpf="+java.util.Arrays.toString(evpf));
	if(evf[1]<I_0_feq) throw new RuntimeException(" evf[1]<I_0_feq="+I_0_feq);
	if(evf[0]>I_0_feq) throw new RuntimeException(" evf[0]>I_0_feq="+I_0_feq);
	final double [] mevpf=new double [2];
	for(int indpb=0;indpb<2;indpb++)
	{
	    final double [] p=EVXData.getPsiFromBasis(indpb,evpf,2);
	    final boolean not_match_B0=UnitTests.checkDoubleRelNoFail(p[1]/p[0],psiB0[1]/psiB0[0],eps),
		not_match_B1=UnitTests.checkDoubleRelNoFail(p[1]/p[0],psiB1[1]/psiB1[0],eps);
	    // The eigenfunctions have to match
	    if(not_match_B0 && not_match_B1)
	    {
		throw new RuntimeException("Psi does not match");
	    }
	    if(not_match_B0) mevpf[1]=evpf[indpb];
	    if(not_match_B1) mevpf[0]=evpf[indpb];
	}
	UnitTests.checkDoubleArrayRel(mevpf,"mevpf",ev_pI_feq,"ev_pI_feq",eps);
	System.err.println("PnLI0IH.get_PnLI0IH selftest OK n="+n+" eps="+eps);
    }

    return new PnLI0IH(I,mm.p_offset,bQQpi[ixH+n*ixH]/I.ev[ixH],I_0_feq,PfdI,evf,ev_pI_feq);
}

public static PnLI0IH test_get_PnLI0IH(final MomentsData mm,
				       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				       final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final int ixH=n-1,indB=1;// the state with maximal I

    final double psi0_psiH=TMatr.getAMatrB(I.psi0,I.psiH,mm.QQt,n),
	psi0_If_psiH=TMatr.getAMatrB(I.psi0,I.psiH,mm.QQi,n)+dI*psi0_psiH,
	psiH_If_psiH=I.sH+dI*Math.pow(psi0_psiH,2),
	psi0_pI_psi0=TMatr.getAMatrA(I.psi0,mm.QQpi,n),
	psi0_pI_psiH=TMatr.getAMatrB(I.psi0,I.psiH,mm.QQpi,n),
	psiH_pI_psiH=TMatr.getAMatrA(I.psiH,mm.QQpi,n);

    final double [] md=new double []{1,psi0_psiH,psi0_psiH,1},
	mIf=new double []{I_0_feq,psi0_If_psiH,psi0_If_psiH,psiH_If_psiH},
        evf=M.EV.getGEV_Generalized(2,mIf,md,true),
	    mpi=new double []{psi0_pI_psi0,psi0_pI_psiH,psi0_pI_psiH,psiH_pI_psiH},
	    m0=new double []{1,psi0_psiH,psi0_psiH,Math.pow(psi0_psiH,2)};

    // Find Pf, such that
    // bQQpi + |bpsi0>PfdI<bpsi0| and bQQi + |bpsi0>dI<bpsi0|
    // have the same eigenvector projections.
    final double a_0=evf[2+2*indB+0],a_H=evf[2+2*indB+1],
	r_0=md[0]*a_0+md[1]*a_H,r_H=md[2]*a_0+md[3]*a_H,
	PfdI=-((mpi[0]*a_0+mpi[1]*a_H)/r_0-(mpi[2]*a_0+mpi[3]*a_H)/r_H)/
	((m0[0]*a_0+m0[1]*a_H)/r_0-(m0[2]*a_0+m0[3]*a_H)/r_H);
    final double [] mpI_feq=TMatr.sumAcB(mpi,PfdI,m0),
	ev_pI_feq=new double []{TMatr.getAMatrA(2,evf,2+2*0,mpI_feq,2),TMatr.getAMatrA(2,evf,2+2*1,mpI_feq,2)};
    if(eps>0)
    {
	EVXData.test_orthogonality("PnLI0IH.test_get_PnLI0IH:I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	final double []
	    psiB0=EVXData.getPsiFromBasis(0,evf,2),
	    psiB1=EVXData.getPsiFromBasis(1,evf,2),
	    evpf=M.EV.getGEV_Generalized(2,mpI_feq,md,true);
	final double beta=a_H/(a_0+psi0_psiH*a_H);

	System.err.println("PnLI0IH.test_get_PnLI0IH: I_0_feq="+I_0_feq+" evf[indB]="+evf[indB]+" PfdI="+PfdI+" beta="+beta);
	System.err.println("PnLI0IH.test_get_PnLI0IH: evf="+java.util.Arrays.toString(evf));
	System.err.println("PnLI0IH.test_get_PnLI0IH: evpf="+java.util.Arrays.toString(evpf));
	if(evf[1]<I_0_feq) throw new RuntimeException(" evf[1]<I_0_feq="+I_0_feq);
	if(evf[0]>I_0_feq) throw new RuntimeException(" evf[0]>I_0_feq="+I_0_feq);
	final double [] mevpf=new double [2];
	for(int indpb=0;indpb<2;indpb++)
	{
	    final double [] p=EVXData.getPsiFromBasis(indpb,evpf,2);
	    final boolean not_match_B0=UnitTests.checkDoubleRelNoFail(p[1]/p[0],psiB0[1]/psiB0[0],eps),
		not_match_B1=UnitTests.checkDoubleRelNoFail(p[1]/p[0],psiB1[1]/psiB1[0],eps);
	    // The eigenfunctions have to match
	    if(not_match_B0 && not_match_B1)
	    {
		throw new RuntimeException("Psi does not match");
	    }
	    if(not_match_B0) mevpf[1]=evpf[indpb];
	    if(not_match_B1) mevpf[0]=evpf[indpb];
	}
	UnitTests.checkDoubleArrayRel(mevpf,"mevpf",ev_pI_feq,"ev_pI_feq",eps);
	System.err.println("PnLI0IH.test_get_PnLI0IH selftest OK n="+n+" eps="+eps);
    }

    return new PnLI0IH(I,mm.p_offset,psiH_pI_psiH/I.sH,I_0_feq,PfdI,evf,ev_pI_feq);
}

public static void matchPnLI0IH(final PnLI0IH a,final PnLI0IH b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);
    final double p_offset_shift=b.p_offset-a.p_offset,dI=b.I_0_feq-b.I.s0;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.PfdI,"a.PfdI",b.PfdI+p_offset_shift*dI,"b.PfdI+p_offset_shift*dI",eps);
    UnitTests.checkDoubleRel(a.Pfi,"a.Pfi",b.Pfi+p_offset_shift,"b.Pfi+p_offset_shift",eps);
    System.err.println("PnLI0IH.matchPnLI0IH test OK n="+a.I.neq+" eps="+eps);
}
}
