package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLI0e.java,v 1.31 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLI0e {

    public final EVXData I;
    public final double p_offset,p_IH,I_0_feq,PeqqdI,Peqq,PfdI,Pf;
    PnLI0e(final EVXData I,final double p_offset,final double p_IH,final double I_0_feq,
	   final double PeqqdI,final double PfdI)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.I_0_feq=I_0_feq;
	final double dI=I_0_feq-I.s0;
	this.PeqqdI=PeqqdI;
	this.Peqq=PeqqdI/dI;
	this.PfdI=PfdI;
	this.Pf=PfdI/dI;
    }

/**
 * Solve <psi|I|psi>/<psi|psi>=lambda extremum problem, find eigenfunction & eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-<psi0|I|psi0> (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + |psi0>dI<psi0|
 * Consider |phi> such that |b>=|I^{f}|psi0>, |phi>=|b>-<psi0|b>|psi0>
 * <psi0+beta*phi|I^f|psi0+beta*phi>/<psi0+beta*phi|psi0+beta*phi> -> max
 * find beta (similar to PnLI0, but exact, not perturbation).
 * Then find Pf requiring future pI operator pI^{f} = pI+|psi0>Pf*dI<psi0|
 * to have the same eigenfunction (projected to alpha|psi0>+beta|phi> subspace).
 */
public static PnLI0e get_PnLI0e(final MomentsData mm,
				final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				final boolean flag_LOW_I,
				final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final int ixH=n-1,indB=flag_LOW_I?0:1; // the state with maximal I

    final double [] bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	bphi=TMatr.getVAB(I.getEValues(),bpsi0),bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev);
    TMatr.sumP1(bphi,bpsi0,-TMatr.mulVV(bphi,bpsi0));

    final double psi0_I_phi=TMatr.mulVVV(bpsi0,I.getEValues(),bphi),
	phi_I_phi=TMatr.mulV2V(bphi,I.getEValues()),
	phi_phi=TMatr.mulVV(bphi,bphi),
	psi0_pI_psi0=TMatr.getAMatrA(bpsi0,bQQpi,n),
	psi0_pI_phi=TMatr.getAMatrB(bpsi0,bphi,bQQpi,n),
	phi_pI_phi=TMatr.getAMatrA(bphi,bQQpi,n);
    final double [] evf=M.EV.getGEV_Generalized(2,new double []{I_0_feq,psi0_I_phi,psi0_I_phi,phi_I_phi},
						new double []{1,0,0,phi_phi},true);
    final double beta=evf[2+2*indB+1]/evf[2+2*indB+0],nrm=1+Math.pow(beta,2)*phi_phi,
	PfdI=nrm*(psi0_pI_phi/beta+phi_pI_phi)/phi_phi
	 - beta*(2*psi0_pI_phi+beta*phi_pI_phi) - psi0_pI_psi0,
	beta0=-psi0_I_phi/(phi_I_phi-I_0_feq*phi_phi),
	PeqqdI=((psi0_pI_phi/beta0)+phi_pI_phi)/phi_phi-psi0_pI_psi0;

    if(eps>0)
    {
	EVXData.test_orthogonality("PnLI0e.get_PnLI0e:I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	// Check spur invariance
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQi),"Spur(IQQt,QQi)",TMatr.sumElements(I.getEValues()),"sum(I.ev)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQpi),"Spur(IQQt,QQpi)",TMatr.getSpurM(n,bQQpi),"Spur(bQQpi)",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(bpsi0,bphi),"<psi0|phi>",0,"0",eps);
	UnitTests.checkDoubleRel(phi_phi,"<phi|phi>",psi0_I_phi,"<psi0|I|phi>",eps);

	// Create bQQpi_feq (operator of "future" pI), with the |bpsi0>PfdI<bpsi0| contribution from the future:
	// bQQpi_feq = bQQpi + |bpsi0>PfdI<bpsi0|
	final double [] bQQpi_feq=TMatr.sumAcB(bQQpi,PfdI,TMatr.getQQaa(bpsi0)),
	    psiB=EVXData.getPsiFromBasis(indB,evf,2),
	    evpf=M.EV.getGEV_Generalized(2,new double []{
		TMatr.getAMatrA(bpsi0,bQQpi_feq,n),
		TMatr.getAMatrB(bpsi0,bphi,bQQpi_feq,n),
		TMatr.getAMatrB(bpsi0,bphi,bQQpi_feq,n),
		TMatr.getAMatrA(bphi,bQQpi_feq,n)},
		new double []{1,0,0,phi_phi},true),
	    psipB0=EVXData.getPsiFromBasis(0,evpf,2),
	    psipB1=EVXData.getPsiFromBasis(1,evpf,2);
	System.err.println("PnLI0e.get_PnLI0e: I_0_feq="+I_0_feq+" evf[indB]="+evf[indB]+" PfdI="+PfdI+" PeqqdI="+PeqqdI+" beta="+beta+" flag_LOW_I="+flag_LOW_I);
	System.err.println("PnLI0e.get_PnLI0e: evf="+java.util.Arrays.toString(evf));
	System.err.println("PnLI0e.get_PnLI0e: evpf="+java.util.Arrays.toString(evpf));
	if(evf[1]<I_0_feq) throw new RuntimeException(" evf[1]<I_0_feq="+I_0_feq);
	if(evf[0]>I_0_feq) throw new RuntimeException(" evf[0]>I_0_feq="+I_0_feq);
	// one eigenfunction have to match
	if(UnitTests.checkDoubleRelNoFail(psiB[1]/psiB[0],psipB0[1]/psipB0[0],eps)
	   &&
	   UnitTests.checkDoubleRelNoFail(psiB[1]/psiB[0],psipB1[1]/psipB1[0],eps))
	{
	    throw new RuntimeException("Psi does not match");
	}

	System.err.println("PnLI0e.get_PnLI0e selftest OK n="+n+" eps="+eps);
    }

    return new PnLI0e(I,mm.p_offset,bQQpi[ixH+n*ixH]/I.ev[ixH],I_0_feq,PeqqdI,PfdI);
}

public static PnLI0e test_get_PnLI0e(final MomentsData mm,
				final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				final boolean flag_LOW_I,
				final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final int ixH=n-1,indB=flag_LOW_I?0:1; // the state with maximal I

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),
	bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev);
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];

    final double
	psi0_psi0=1,
	psi0_If_psi0=TMatr.getAMatrA(bpsi0,bQQi_feq,n),
	psi0_If_If_psi0=TMatr.getAMatrA(bpsi0,TMatr.mulMatrices(n,new double [][]{bQQi_feq,bQQi_feq}),n),
	psi0_If_If_If_psi0=TMatr.getAMatrA(bpsi0,TMatr.mulMatrices(n,new double [][]{bQQi_feq,bQQi_feq,bQQi_feq}),n),
	psi0_pI_psi0=TMatr.getAMatrA(bpsi0,bQQpi,n),
	psi0_If_pI_psi0=TMatr.getAMatrA(bpsi0,TMatr.mulMatrices(n,new double [][]{bQQi_feq,bQQpi}),n),
	psi0_If_pI_If_psi0=TMatr.getAMatrA(bpsi0,TMatr.mulMatrices(n,new double [][]{bQQi_feq,bQQpi,bQQi_feq}),n);
    final double [] md=new double []{psi0_psi0,psi0_If_psi0,psi0_If_psi0,psi0_If_If_psi0},
	evf=M.EV.getGEV_Generalized(2,new double []{psi0_If_psi0,psi0_If_If_psi0,psi0_If_If_psi0,psi0_If_If_If_psi0},md,true),
	mpi=new double []{psi0_pI_psi0,psi0_If_pI_psi0,psi0_If_pI_psi0,psi0_If_pI_If_psi0},
	m0=new double []{psi0_psi0,psi0_If_psi0,psi0_If_psi0,Math.pow(psi0_If_psi0,2)};

    // Find Pf, such that
    // bQQpi + |bpsi0>PfdI<bpsi0| and bQQi + |bpsi0>dI<bpsi0|
    // have the same eigenvector projection.
    final double a_0=evf[2+2*indB+0],a_i0=evf[2+2*indB+1],
	r_0=md[0]*a_0+md[1]*a_i0,r_i0=md[2]*a_0+md[3]*a_i0,
	PfdI=-((mpi[0]*a_0+mpi[1]*a_i0)/r_0-(mpi[2]*a_0+mpi[3]*a_i0)/r_i0)/
	((m0[0]*a_0+m0[1]*a_i0)/r_0-(m0[2]*a_0+m0[3]*a_i0)/r_i0);
    final double PeqqdI;
    {
	final double [] bphi=TMatr.getVAB(bpsi0,I.getEValues());
	TMatr.sumP1(bphi,bpsi0,-TMatr.mulVV(bphi,bpsi0));
	final double psi0_I_phi=TMatr.mulVVV(bpsi0,I.getEValues(),bphi),
	    phi_I_phi=TMatr.mulV2V(bphi,I.getEValues()),
	    phi_phi=TMatr.mulVV(bphi,bphi),
	    psi0_pI_phi=TMatr.getAMatrB(bpsi0,bphi,bQQpi,n),
	    phi_pI_phi=TMatr.getAMatrA(bphi,bQQpi,n),
	    beta0=-psi0_I_phi/(phi_I_phi-I_0_feq*phi_phi);
	PeqqdI=((psi0_pI_phi/beta0)+phi_pI_phi)/phi_phi-psi0_pI_psi0;
    }

    if(eps>0)
    {
	EVXData.test_orthogonality("PnLI0e.test_get_PnLI0e:I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	final double [] mpi_feq=TMatr.sumAcB(mpi,PfdI,m0),
	    psiB=EVXData.getPsiFromBasis(indB,evf,2),
	    evpf=M.EV.getGEV_Generalized(2,mpi_feq,md,true),
	    psipB0=EVXData.getPsiFromBasis(0,evpf,2),
	    psipB1=EVXData.getPsiFromBasis(1,evpf,2);

	// Create bQQpi_feq (operator of "future" pI), with the |bpsi0>PfdI<bpsi0| contribution from the future:
	// bQQpi_feq = bQQpi + |bpsi0>PfdI<bpsi0|
	final double [] bQQpi_feq=TMatr.sumAcB(bQQpi,PfdI,TMatr.getQQaa(bpsi0));

	System.err.println("PnLI0e.test_get_PnLI0e: I_0_feq="+I_0_feq+" evf[indB]="+evf[indB]+" PfdI="+PfdI+" PeqqdI="+PeqqdI+" beta="+(psiB[1]/(psiB[0]+I_0_feq*psiB[1]))+" flag_LOW_I="+flag_LOW_I);
	System.err.println("PnLI0e.test_get_PnLI0e: evf="+java.util.Arrays.toString(evf));
	System.err.println("PnLI0e.test_get_PnLI0e: evpf="+java.util.Arrays.toString(evpf));
	if(evf[1]<I_0_feq) throw new RuntimeException(" evf[1]<I_0_feq="+I_0_feq);
	if(evf[0]>I_0_feq) throw new RuntimeException(" evf[0]>I_0_feq="+I_0_feq);
	// one eigenfunction have to match
	if(UnitTests.checkDoubleRelNoFail(psiB[1]/psiB[0],psipB0[1]/psipB0[0],eps)
	   &&
	   UnitTests.checkDoubleRelNoFail(psiB[1]/psiB[0],psipB1[1]/psipB1[0],eps))
	{
	    throw new RuntimeException("Psi does not match");
	}

	System.err.println("PnLI0e.test_get_PnLI0e selftest OK n="+n+" eps="+eps);
    }

    return new PnLI0e(I,mm.p_offset,bQQpi[ixH+n*ixH]/I.ev[ixH],I_0_feq,PeqqdI,PfdI);
}

public static void matchPnLI0e(final PnLI0e a,final PnLI0e b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);
    final double p_offset_shift=b.p_offset-a.p_offset,dI=b.I_0_feq-b.I.s0;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.PeqqdI,"a.PeqqdI",b.PeqqdI+p_offset_shift*dI,"b.PeqqdI+p_offset_shift*dI",eps);
    UnitTests.checkDoubleRel(a.PfdI,"a.PfdI",b.PfdI+p_offset_shift*dI,"b.PfdI+p_offset_shift*dI",eps);
    System.err.println("PnLI0e.matchPnLI0e test OK n="+a.I.neq+" eps="+eps);
}
}
