package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVSolver;
import com.polytechnik.outils.DerivativesAtCalculateable;
import com.polytechnik.outils.SimpleRootsFinder;

/** P&L and I(p) calculation. Solving the equition P&L_pnlH(p)-P&L_pnlL(p)->min,
 *  what is equivalent to I_pnlH(p)-I_pnlL(p)=0.
 *  @version $Id: PnLICalc.java,v 1.5 2020/02/06 09:53:13 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class PnLICalc implements DerivativesAtCalculateable {

    static final int MAXITER=100;
    static final double EPSP=1e-5;

    final int n;
    final double [] bQQi,bQQpi;
    final EVSolver EV;
    final double [] tmp;
    PnLICalc(final int n,final double [] bQQi,final double [] bQQpi,final EVSolver EV)
    {
	this.n=n;
	this.tmp=new double [n*n];
	this.bQQi=bQQi;
	this.bQQpi=bQQpi;
	this.EV=EV;
    }
    double p,I_L,I_H,dIdp_L,dIdp_H;
    public void calculateAt(final double p)
    {
	this.p=p;
	for(int k=n*n;--k>=0;) tmp[k]=bQQpi[k]-p*bQQi[k];
	final double [] ev=EV.getEV(n,tmp,true);
	final int kL=0,kH=n-1;
	double dIdpL=0,dIdpH=0;
	for(int k=0;k<n;k++)
	{
	    if(k!=kL)
	    {
		dIdpL-=2*Math.pow(TMatr.getAMatrB(n,ev,n+kL*n,ev,n+k*n,bQQi,n),2)/(ev[kL]-ev[k]);
	    }
	    if(k!=kH)
	    {
		dIdpH-=2*Math.pow(TMatr.getAMatrB(n,ev,n+kH*n,ev,n+k*n,bQQi,n),2)/(ev[kH]-ev[k]);
	    }
	}
	this.I_L=TMatr.getAMatrA(n,ev,n+kL*n,bQQi,n);
	this.I_H=TMatr.getAMatrA(n,ev,n+kH*n,bQQi,n);
	this.dIdp_L=dIdpL; // always positive
	this.dIdp_H=dIdpH; // always negative
    }
    public double getF(){return I_L-I_H;}
    public double getDF(){
	return dIdp_L-dIdp_H; // always positive
    }
    String getAsText()
    {
	return "PnLICalc:p="+p+" I_L="+I_L+" I_H="+I_H+" getF()="+getF()+" dIdp_L="+dIdp_L+" dIdp_H="+dIdp_H+" getDF()="+getDF()+" n="+n;
    }

    double iterateNewton(final double p0)
    {
	return SimpleRootsFinder.findRoots(this,p0,MAXITER,EPSP);
    }

    /** Note! Clober the result.
     */
    void testDPrint(final double px,final double e,final double eps)
    {
	calculateAt(px+e);
	final double aI_L=I_L,aI_H=I_H,adIdp_L=dIdp_L,adIdp_H=dIdp_H,a_F=getF();
	calculateAt(px-e);
	final double bI_L=I_L,bI_H=I_H,bdIdp_L=dIdp_L,bdIdp_H=dIdp_H,b_F=getF();
	calculateAt(px);
	final double dIdp_L_diff=(aI_L-bI_L)/(2*e),dIdp_H_diff=(aI_H-bI_H)/(2*e),DF_diff=(a_F-b_F)/(2*e);
	System.err.println("testDPrint:px="+px+" I_L="+I_L+" I_H="+I_H+" getF()="+getF()+" n="+n);
	System.err.println("testDPrint:dIdp_L="+dIdp_L+" (I_L(px+e)-I_L(px-e))/2e="+dIdp_L_diff);
	System.err.println("testDPrint:dIdp_H="+dIdp_H+" (I_H(px+e)-I_H(px-e))/2e="+dIdp_H_diff);
	System.err.println("testDPrint:px="+px+" getDF="+getDF()+" DF from F="+DF_diff+" getF()="+getF()+" n="+n);
	if(getDF()<0)
	{
	    throw new RuntimeException("getDF()="+getDF()+" must be positive");
	}
	if(dIdp_L<0)
	{
	     throw new RuntimeException("dIdp_L="+dIdp_L+" must be positive");
	}
	if(dIdp_H>0)
	{
	     throw new RuntimeException("dIdp_H()="+dIdp_H+" must be negative");
	}
	if(eps>=0)
	{ // match values
	    UnitTests.checkDoubleRel(dIdp_L,"dIdp_L",dIdp_L_diff,"dIdp_L_diff",eps);
	    UnitTests.checkDoubleRel(dIdp_H,"dIdp_H",dIdp_H_diff,"dIdp_H_diff",eps);
	    UnitTests.checkDoubleRel(getDF(),"getDF()",DF_diff,"DF_diff",eps);
	}
	iterateNewton(px);
	System.err.println("testDPrint:newton:"+getAsText());
    }
}
