package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.freemoney.WIntegrator;

/**
 * A factory for {@link PnLILowIHighTPVFactory.PnLILowIHighTPV}.
 * @version $Id: PnLILowIHighTPVFactory.java,v 1.31 2022/07/28 07:29:39 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public abstract class PnLILowIHighTPVFactory {


public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getABasis();

final int n;
final double [] QQt;
final double atau;
/**
* A Factory.
* @param n Basis dimension.
* @param odata Regular moments internal state.
* @param rst Previous tick internal state.
*/
public PnLILowIHighTPVFactory(final int n,final double atau)
{
    this.n=n;
    this.atau=atau;
    this.QQt=M.B.getQQMatr(n,getDtMoments(2*n-1));
}

/** Next state, it should be called on every tick or even more frequently.
 * For "dummy" calls (to identify low I events) use last price as P and dV=0.
 */
public PnLILowIHighTPV getNext(final long currenttime,final double P,final long dV,PnLILowIHighStateSaverWithWI state){
    return new PnLILowIHighTPV(currenttime,P,dV,state!=null?state:new PnLILowIHighStateSaverWithWI(getEmptyWIntegrator(5)));
}

/**
 * Calculated state.
 * Note that all moments are calculated using  {@link  WIntegrator} without much dependencies.
 * @version $Id: PnLILowIHighTPVFactory.java,v 1.31 2022/07/28 07:29:39 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public class PnLILowIHighTPV {

/** The |I|psi>=lambda |psi> solution.
 */
public final EVXData I;

public final double
// Helper data and moving averages.
    p_last,pi_average,
// The https://arxiv.org/abs/1510.05510 results (the state of max I).
    p_IH
    ;
    /** The state.
     */
public final PnLILowIHighStateSaverWithWI sst;

public final EVXData dIHpnl;
/** The <a href="https://arxiv.org/abs/1903.11530">scalp price</a>
 * style results.
 */
public final double DP_since_dIH_H,AbsDP_since_dIH_H,
    P_at_dIH_H,DP_at_dIH_H,AbsDP_at_dIH_H,
    dP,dIH
    ;

PnLILowIHighTPV(final long currenttime,final double P,final long dV,final PnLILowIHighStateSaverWithWI rst){

    // An integrator rst.wi will be re-used as sst.wi, the rst will be discarded.
    // This is an optimization, update integrator "in-place".
    rst.wi.updateSavedMomentsToNewTimeReference(currenttime);
    rst.addSingleObservation_PdV(currenttime,P,dV);

    final double [] QQi=M.B.getQQMatr(n,rst.getDvMoments()),QQpi=M.B.getQQMatr(n,rst.getPDvMoments());
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.p_last=P;
    this.pi_average=QQpi[0]/QQi[0];
    final double [] psiH=I.getPsi(n-1/* max I state */),psi0=I.psi0;
    this.p_IH=TMatr.getAMatrA(psiH,QQpi,n)/TMatr.getAMatrA(psiH,QQi,n);

    this.sst=new PnLILowIHighStateSaverWithWI(rst,currenttime,P,dV,I);
    /*final double*/ dIH=rst.I==null?sst.I.sH:sst.I.sH-rst.I.sH;dP=sst.P;//-rst.P;
    if(dIH>0 && Double.isFinite(dP)){
	sst.addSingleObservation_PnLs(/*currenttime=*/currenttime,dP*dIH,dP*dIH,dIH);
    }

    final double [] QQdIHpnl=M.B.getQQMatr(n,sst.getDIHpnlMoments());
    this.dIHpnl=EVXData.getEVXDataFromQQ(n,QQdIHpnl,QQt,M);
    // A new version with dIHpnl.psiH
    final double [] psiHdIH2Integrated=getPsi2WIntegratedDt(dIHpnl.psiH);
    this.DP_since_dIH_H=TMatr.mulVV(psiHdIH2Integrated,sst.getPcDIHpnlMoments());
    //TMatr.mulVV(psiHdIH2Integrated,sst.getDIHpnlMoments());
    this.AbsDP_since_dIH_H=TMatr.mulVV(psiHdIH2Integrated,sst.getAbsPcDIHpnlMoments());
    //TMatr.mulVV(psiHdIH2Integrated,sst.getDIHpnlMoments());

    // d/dt at |psiH>
    final double [] QQPcdIHpnl=M.B.getQQMatr(n,sst.getPcDIHpnlMoments()),
	QQAbsPcdIHpnl=M.B.getQQMatr(n,sst.getAbsPcDIHpnlMoments()),
	dIHpnl_DpsiH=getEDPsi(dIHpnl.psiH);
    this.P_at_dIH_H=TMatr.getAMatrA(dIHpnl.psiH,QQPcdIHpnl,n)/dIHpnl.sH;
    this.DP_at_dIH_H=2*(TMatr.getAMatrB(dIHpnl.psiH,dIHpnl_DpsiH,QQPcdIHpnl,n)/dIHpnl.sH-P_at_dIH_H*TMatr.getAMatrB(dIHpnl.psiH,dIHpnl_DpsiH,QQt,n));
    this.AbsDP_at_dIH_H=TMatr.getAMatrA(dIHpnl.psiH,QQAbsPcdIHpnl,n)/dIHpnl.sH;

}

} // END OF class PnLILowIHighTPV



abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
/** Square a wavefunction, then integrate by parts.
 */
abstract double [] getPsi2WIntegratedDt(final double [] psi);
abstract WIntegrator getEmptyWIntegrator(int dim);
abstract double [] getDtMoments(int nmoms);
abstract double [] getEDPsi(final double [] psi);


}
