package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.freemoney.*;

/**
 * Try PnL In psiH state again.
 * @version $Id: PnLInPsiHstate.java,v 1.224 2024/04/15 06:15:33 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public abstract class PnLInPsiHstate<DTYPE> {



/** "Regular", not scalp--moments. The object has
 * an internal state to recurrently speed up the calculations.
 */
final DTYPE odata;

public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getABasis();


    /** Not worth a function or constructor argument.
     */
    private static double EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;

    /** Actual prices (without an offset).
     */
    public final double P_last,pv_average,pt_average,Tv_average;

    /** Used internally. */
    private final double p_offset,p_last;

    public final EVXData I,VT,TimVT;
    public final double pi_IH;

    public final TwoTopI twoI;

    public final BasisTransformPiTiPrediction PiTi;
    public final IntervalSincePsiM SM,SM_localized,SMImVT_localized,SMImVT_Istate,SMVT,SMTimVT_L,SMTimVT_H;
    public final PsiMSelection PSIM;
    public final DaDv dadv,dadt;
    public final PFuture pFV;

    private final com.polytechnik.utils.VolMatrix voltlty=this::getEDPsi;

    /** Misc tests. */
    public final MiscTests MT;
    /** More tests. */
    private final ATestForBasisUnitaryTransforms BP;
/**
* Constructor.
* @param n Basis dimension.
* @param odata Regular moments internal state.
*/
PnLInPsiHstate(final int n,final DTYPE odata)
{

    this.odata=odata;
    final SMomentsData mm=getSMoments(2*n-1).asQQMatrices(n,M);
    // After asQQMatrices mm has the matrices (not moments) stored.
    final double []
	QQt=mm.tmom,
	QQi=mm.imom,
	QQpi=mm.pimom,
	QQp2i=mm.p2imom,
	QQp3i=mm.p3imom,
	QQdp=mm.dpmom,
	QQp1t=mm.p1tmom,QQp2t=mm.p2tmom,
	QQp0t=mm.p0tmom,
	QQa=mm.amom,QQpa=mm.pamom,
	QQTi=TMatr.getVectorScaled(mm.Timom,-1),
	QQTpi=TMatr.getVectorScaled(mm.Tpimom,-1),
	QQTp2i=TMatr.getVectorScaled(mm.Tp2imom,-1),
	QQT=TMatr.getVectorScaled(mm.Tmom,-1),
	QQV=TMatr.getVectorScaled(mm.Vmom,-1),
	QQTdp=TMatr.getVectorScaled(mm.Tdpmom,-1),
	QQVdP=TMatr.getVectorScaled(mm.VdPdtmom,-1),
	IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt);

    // turn on selftests if required
    final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING;
    //final double eps=mm.totalvolume>10000?1e-5:0;

    p_offset=mm.p_offset;
    p_last=mm.p_last;
    this.P_last=p_offset+p_last; // actual price
    this.pv_average=p_offset+QQpi[0]/QQi[0];
    this.pt_average=p_offset+QQp1t[0]/QQp0t[0];
    this.Tv_average=QQTi[0]/QQi[0];
    //System.err.println("p_last="+p_last);


    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    this.VT=EVXData.getEVXDataFromQQ(n,QQV,QQT,M);
    this.TimVT=EVXData.getEVXDataFromQQ(n,TMatr.sumAcB(QQTi,-1,QQV),QQT,M);
    this.pi_IH=p_offset+TMatr.getAMatrA(I.psiH,QQpi,n)/I.sH;


    twoI=new TwoTopI(I,p_offset,p_last,QQTi,QQpi,QQV,QQdp);
    final var dm=new IandDM(n,QQt,QQi,M,this::getPolynomialWIntegratedDt,eps);

    this.pFV=new PFuture(dm,
			 p_offset,
			 p_last,
			 // input moment matrices for eigenstate problem
			 QQt,
			 QQi,
			 // input moment matrices for other calculations
			 QQpi,
			 TMatr.getVectorScaled(QQTi,-1),
			 TMatr.getVectorScaled(QQV,-1),
			 QQdp,
			 TMatr.getVectorScaled(QQVdP,-1),
			 mm.totalvolume,
			 M,this::getPolynomialWIntegratedDt,this::getEDPsi,eps);


    // try to check dV/da & da/dt
    this.dadv=new DaDv(n,
		       p_offset,
		       p_last,
		       QQt,
		       QQi,
		       QQTi,

		       QQi,
		       QQpi,
		       QQa,
		       QQpa,
		       M,eps);
    this.dadt=new DaDv(n,
		       p_offset,
		       p_last,
		       QQt,
		       QQi,
		       QQTi,

		       QQp0t,
		       QQp1t,
		       QQa,
		       QQpa,
		       M,eps);

    // actual calculations
    this.PiTi=new BasisTransformPiTiPrediction(n,p_offset,p_last,QQt,QQi,QQpi,QQTi,QQdp,QQp1t,M,this::getPolynomialWIntegratedDt,this::getEDPsi,eps>0?getSMoments(2*n-1).tmom:null,eps);

    java.util.function.Function<double [],IntervalSincePsiM> getSincePsiM=
	psiM->new IntervalSincePsiM(n,p_offset,p_last,TMatr.getVectorScaled(psiM,1/Math.sqrt(TMatr.getAMatrA(psiM,QQt,n))),QQt,QQi,QQpi,QQTi,QQTpi,QQdp,QQp1t,QQT,QQV,QQTdp,M,this::getPolynomialWIntegratedDt,this::getEDPsi,eps>0?getSMoments(2*n-1).tmom:null,eps>0?getSMoments(2*n-1).imom:null,eps);


    this.SM=getSincePsiM.apply(I.psiH);
    this.SM_localized=getSincePsiM.apply(new ILocalizedStates(n,QQt,QQi,IQQt,M,eps).res.psi_Fmax);
    this.SMImVT_localized=getSincePsiM.apply(new ImVTLocalizedStates(n,QQt,QQi,QQT,QQV,IQQt,M,eps).res.psi_Fmax);
    this.SMImVT_Istate=getSincePsiM.apply(PsiMSelection.getPsi_MaxImVdivT(I,EVXData.convertMatrFromQtoB_Basis(n,QQV,I.ev),EVXData.convertMatrFromQtoB_Basis(n,QQT,I.ev),eps));
    this.SMVT=getSincePsiM.apply(VT.psiH);
    this.SMTimVT_H=getSincePsiM.apply(TimVT.psiH);
    this.SMTimVT_L=getSincePsiM.apply(TimVT.psiL);
    this.PSIM=new PsiMSelection(n,p_offset,p_last,QQt,QQi,QQpi,QQTi,QQTpi,QQdp,QQp1t,QQT,QQV,QQTdp,M,this::getPolynomialWIntegratedDt,this::getEDPsi,eps);


    if(false &&
       EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING==0 &&
       /*mPnL.flag_solution_exists && */
       mm.currenttime>(long)(t0dumpinhours*3600e9)
       ){
	if(PSI_WAS_DUMPED){
	    System.err.println("PnLInPsiHstate: WAVEFUNCTIONS WERE ALREADY DUMPED, EXITING!!!");
	    throw new RuntimeException("Exit from execution in the middle");
	} else {
	    System.err.println("PnLInPsiHstate: DUMPING WAVEFUNCTIONS at t="+com.polytechnik.realtimebook.BookItem.getTimeHHMMSS(mm.currenttime)+" P_last="+P_last);
	    PnLInPsiHstateTest.dumpWavefunctionOnce(
		this,
		QQt,
		QQi,
		QQpi,
		java.util.List.of(pFV.QQIdp_toDump,pFV.QQdIdt_toDump,pFV.QQDP_toDump,QQdp),
		mm.currenttime);
	    PSI_WAS_DUMPED=true;
	}
    }

    this.MT=new MiscTests(mm);
    this.BP=new ATestForBasisUnitaryTransforms(n,p_offset,QQt,QQi,QQpi,M,this::getEDPsi,eps);

}
    private static final double t0dumpinhours=9.98;//9.97;//11.2;//10.0;//10.11;//11.0;//9.97;
    private static boolean PSI_WAS_DUMPED=false;

/** Obtain regular moments.
 */
abstract SMomentsData getSMoments(final int nmoms);

/** Time-basis of the moments from {@link #getSMoments}.
 */
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();

/** Time--variate {@literal |psi>}, used for comparison with old results.
 */
abstract double [] getEDPsi(final double [] psi);
abstract double [] getPolynomialWIntegratedDt(final double [] polynomial);


/** Convert postion change dS=(p-p_M)*I*psi^2(t)omega(t)dt
 * to position S and obtain it's moments.
 */
abstract double [] getPositionMoments(final int n,final double p_eq,final double [] psiM);

/** A distance between two wavefunctions. Typically joint distibution is better than simple comparison {@literal <psi|x|psi>}.
 */
abstract K2Distance getK2Distance();

/** A class with {@literal |D(Ti)|psi>=lamnda|i|psi>} solution.
 */
class OneStateSaved {
    final double [] psiM,DpsiM;
    final double lambda,i,T,pa_i,pa_var;
    final int ev_index;
    final EVXData J;
    OneStateSaved(final int n,
		  final int ev_index,
		  final double [] QQt,

		  final double [] QQi,
		  final double [] QQpi,

		  final double [] QQTi,
		  final double [] QQTpi){
	this.ev_index=ev_index;
	final double [] QQDTi=voltlty.getDiffMatrix(n,QQTi),QQDTpi=voltlty.getDiffMatrix(n,QQTpi);
	this.J=EVXData.getEVXDataFromQQ(n,QQDTi,QQi,M);
	this.psiM=J.getPsi(ev_index);
	this.DpsiM=getEDPsi(psiM);
	this.lambda=J.ev[ev_index];
	final double ni=TMatr.getAMatrA(psiM,QQi,n), // always =1
	    pni=TMatr.getAMatrA(psiM,QQpi,n);
	this.i=ni/TMatr.getAMatrA(psiM,QQt,n);
	this.T=TMatr.getAMatrA(psiM,QQTi,n)/ni;
	this.pa_i=p_offset+pni/ni;
	this.pa_var=p_offset+(TMatr.getAMatrA(psiM,QQDTpi,n)-pni)/((lambda-1)*ni);

	final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING;
	if(eps>0){
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psiM,QQDTi,n),"<psiM|QQDTi|psiM>",lambda*ni,"lambda*ni",eps);
	    final double dtpi=TMatr.getAMatrA(psiM,QQDTpi,n),
		var0P=dtpi-lambda*TMatr.getAMatrA(psiM,QQpi,n);
	    UnitTests.checkDoubleRel(var0P,"var0P",(pa_var-pa_i)*(lambda-1)*ni,"(pa_var-pa_i)*(lambda-1)*ni",eps);
	    UnitTests.checkDoubleRel(dtpi,"dtpi",2*TMatr.getAMatrB(psiM,DpsiM,QQTpi,n),"2*<psiM|QQTpi|DpsiM>",eps);

	    UnitTests.checkDoubleRel(TMatr.getAMatrB(psiM,DpsiM,QQDTi,n)-lambda*TMatr.getAMatrB(psiM,DpsiM,QQi,n),"var1",0,"0",eps);
	    final var rq=RayleighQuotient.getRQVariationValues(psiM,DpsiM,null,QQDTi,QQi,n);
	    UnitTests.checkDoubleRel(rq[0],"rq[0]",lambda,"lambda",eps);
	    UnitTests.checkDoubleRel(rq[1],"rq[1]",0,"0",eps);

	    System.err.println("OneStateSaved: test OK n="+n+" lambda="+lambda+" T="+T+" i="+i+" ev="+java.util.Arrays.toString(J.getEValues())+" eps="+eps);
	}
    }
}

/** A class with {@literal |D(Ti)+c*i|psi>=lambda|psi>} solution.
 */
class OneCTIstate {
    final EVXData J;
    final double [] psiM,DpsiM;
    final double cFactor,lambda,i,T,pa_i,pa_var;
    final int ev_index;
    OneCTIstate(final int n,
		final int ev_index,
		final double cFactor,
		final double [] QQt,

		final double [] QQi,
		final double [] QQpi,

		final double [] QQTi,
		final double [] QQTpi){
	this.ev_index=ev_index;
	this.cFactor=cFactor;

	final double []
	    QQDTi=voltlty.getDiffMatrix(n,QQTi),
	    QQDTpi=voltlty.getDiffMatrix(n,QQTpi),
	    QQcTi=TMatr.sumAcB(QQDTi,cFactor,QQi),
	    QQcpTi=TMatr.sumAcB(QQDTpi,cFactor,QQpi);
	this.J=EVXData.getEVXDataFromQQ(n,QQcTi,QQt,M);
	this.psiM=J.getPsi(ev_index);
	this.DpsiM=getEDPsi(psiM);
	this.lambda=J.ev[ev_index];
	this.i=J.getFPsi(ev_index,QQi);
	this.T=J.getFPsi(ev_index,QQTi)/i;
	this.pa_i=p_offset+J.getFPsi(ev_index,QQpi)/i;
	this.pa_var=Double.NaN;

	final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING;
	if(eps>0){

	    final var rq=RayleighQuotient.getRQVariationValues(psiM,DpsiM,null,QQcTi,QQt,n);
	    UnitTests.checkDoubleRel(rq[0],"rq[0]",lambda,"lambda",eps);
	    UnitTests.checkDoubleRel(rq[1],"rq[1]",0,"0",eps);

	    System.err.println("OneCTIstate: test OK n="+n+" lambda="+lambda+" T="+T+" i="+i+" ev="+java.util.Arrays.toString(J.getEValues())+" eps="+eps);
	}
    }
}


class OneIState {
    final double [] psiM,DpsiM,QQS_M;
    final double i,p_M,S_M,var_S_M;
    final int ev_index;
    OneIState(final double [] _psiM,
	      final int ev_index,
	      final double [] QQt,
	      final double [] QQi,
	      final double [] QQpi){
	this.ev_index=ev_index;
	final int n=_psiM.length;
	this.psiM=TMatr.getVectorScaled(_psiM,1/Math.sqrt(TMatr.getAMatrA(_psiM,QQt,n)));
	this.DpsiM=getEDPsi(this.psiM);
	this.i=TMatr.getAMatrA(this.psiM,QQi,n);
	this.p_M=TMatr.getAMatrA(this.psiM,QQpi,n)/i;
	this.QQS_M=M.B.getQQMatr(n,getPositionMoments(2*n-1,p_M,this.psiM));
	this.S_M=TMatr.getAMatrA(this.psiM,QQS_M,n);
	this.var_S_M=2*(TMatr.getAMatrB(this.psiM,this.DpsiM,QQS_M,n)-S_M*TMatr.getAMatrB(this.psiM,this.DpsiM,QQt,n));

    }
}

    /** Misc test. */
public class MiscTests {
    final OneStateSaved DTII_L,DTII_H;
    final OneCTIstate TDI_L,TDI_H;;
    final double OLD_p0I,OLD_p0t,OLD_pi0,OLD_svar,OLD_svar_pt,OLD_svar_feq,OLD_svar_pt_feq;


    final EVXData VT,TiV,PiPt,If;

    final java.util.ArrayList<OneIState> statesM=new java.util.ArrayList<OneIState>();

    //public final com.polytechnik.utils.IstatesConditional mPnL;
    public final double i_TiVH,Pi_TiVH,Pv_TiVH,P_TiVH,Ti_TiVH,varTi_TiVH,Pvar_TiVH,
	pV_VTH,pI_VTH,TI_VTH,
	dpdtM,rqPi_0,rqPi_1,rqTi_0,rqTi_1,rqPTi_0,rqPTi_1;
    MiscTests(final SMomentsData mm){
    final int n=I.neq;
    final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING;

    // After asQQMatrices mm has the matrices (not moments) stored.
    final double [] QQt=mm.tmom,QQi=mm.imom,QQpi=mm.pimom,QQdp=mm.dpmom,QQp1t=mm.p1tmom,QQp0t=mm.p0tmom,
	QQTi=TMatr.getVectorScaled(mm.Timom,-1),
	QQTpi=TMatr.getVectorScaled(mm.Tpimom,-1),
	QQV=TMatr.getVectorScaled(mm.Vmom,-1),
	QQT=TMatr.getVectorScaled(mm.Tmom,-1),
	QQpV=TMatr.getVectorScaled(mm.pVmom,-1);

    double sPi=0,sPt=0,si=0,st=0,svar=0,svar_pt=0;
    for(int k=0;k<n;k++){
	final double [] psik=I.getPsi(k),Dpsik=getEDPsi(psik);
	final double wk0=Math.pow(TMatr.getAMatrB(psik,I.psi0,QQt,n),2),
	    pt=TMatr.getAMatrA(psik,QQp1t,n)/TMatr.getAMatrA(psik,QQp0t,n),
	    pi=TMatr.getAMatrA(psik,QQpi,n)/I.ev[k];
	si+=wk0*I.ev[k];
	sPi+=wk0*TMatr.getAMatrA(psik,QQpi,n);
	sPt+=wk0*pt;
	st+=wk0;

	final var rqI=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQi,QQt,n);
	if(/* I.ev[k]>I.s0 */rqI[2]<0){
	    svar_pt+=(pi-pt)*I.ev[k]*wk0;

	    final var rq=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQpi,QQt,n);
	    svar+=rq[1];
	}
	if(eps>0){
	    UnitTests.checkDoubleRel(rqI[1],"rqI[1]",0,"0",eps);
	}
    }
    if(eps>0){
	UnitTests.checkDoubleRel(st,"st",1,"1",eps);
	UnitTests.checkDoubleRel(si,"si",I.s0,"I.s0",eps);
    }
    OLD_p0I=p_offset+sPi/si;
    OLD_p0t=p_offset+sPt/st;
    OLD_pi0=p_offset+TMatr.getAMatrA(I.psi0,QQpi,n)/I.s0;
    OLD_svar=svar;
    OLD_svar_pt=svar_pt;

    final double [] psiM=I.getPsi(n-1),DpsiM=getEDPsi(psiM);
    final double T_Ma=TMatr.getAMatrA(psiM,QQTi,n)/TMatr.getAMatrA(psiM,QQi,n),
	p_Ma=TMatr.getAMatrA(psiM,QQpi,n)/TMatr.getAMatrA(psiM,QQi,n);
    final double [] QQTima=TMatr.sumAcB(QQTi,-T_Ma,QQi),
	QQpima=TMatr.sumAcB(QQpi,-p_Ma,QQi),QQTPima=new double [n*n];
    for(int k=0;k<QQTPima.length;k++){
	QQTPima[k]=QQTpi[k]-p_Ma*QQTi[k]-T_Ma*QQpi[k]+T_Ma*p_Ma*QQi[k];
    }
    dpdtM=TMatr.getAMatrA(psiM,QQdp,n);
    final double []
	rqPi=RayleighQuotient.getRQVariationValues(psiM,DpsiM,null,QQpima,QQi,n),
	rqTi=RayleighQuotient.getRQVariationValues(psiM,DpsiM,null,QQTima,QQi,n),
	rqPTi=RayleighQuotient.getRQVariationValues(psiM,DpsiM,null,QQTPima,QQt,n);
    rqPi_0=rqPi[0];
    rqPi_1=rqPi[1];
    rqTi_0=rqTi[0];
    rqTi_1=rqTi[1];
    rqPTi_0=rqPTi[0];
    rqPTi_1=rqPTi[1];

    statesM.add(new OneIState(I.psiH,n-1,QQt,QQi,QQpi));
    statesM.add(new OneIState(I.psiL,0,QQt,QQi,QQpi));

    if(eps>0){
	EVXData.test_orthogonality("PnLInPsiHstate.I.ev",n,I.ev,QQi,QQt,eps);
    }



    this.TiV=EVXData.getEVXDataFromQQ(n,QQTi,QQV,M);
    int indPsiSel=n-1;

/*    double imax=0;
    for(int k=0;k<n;k++){
	final double [] psi=TiV.getPsi(k);
	final double i=TMatr.getAMatrA(psi,QQi,n)/TMatr.getAMatrA(psi,QQt,n);
	if(i>=imax){
	    imax=i;
	    indPsiSel=k;
	}
    }
    System.err.println(" I.sH="+I.sH+" imax="+imax+" indPsiSel="+indPsiSel);
*/
    final double  [] psiTiV=TiV.getPsi(indPsiSel);
    TMatr.scaleVectorInPlace(psiTiV,1/Math.sqrt(TMatr.getAMatrA(psiTiV,QQt,n)));
    final double [] DpsiTiV=getEDPsi(psiTiV);
    i_TiVH=TMatr.getAMatrA(psiTiV,QQi,n);
    Pi_TiVH=p_offset+TMatr.getAMatrA(psiTiV,QQpi,n)/i_TiVH;

    Ti_TiVH=TMatr.getAMatrA(psiTiV,QQTi,n);
    P_TiVH=p_offset+TMatr.getAMatrA(psiTiV,QQTpi,n)/Ti_TiVH;
    Pv_TiVH=p_offset+TMatr.getAMatrA(psiTiV,QQpV,n)/TMatr.getAMatrA(psiTiV,QQV,n);

    varTi_TiVH=TMatr.getAMatrB(psiTiV,DpsiTiV,QQTi,n);
    Pvar_TiVH=p_offset+TMatr.getAMatrB(psiTiV,DpsiTiV,QQTpi,n)/varTi_TiVH;

    statesM.add(new OneIState(psiTiV,-2,QQt,QQi,QQpi));

    this.VT=EVXData.getEVXDataFromQQ(n,QQV,QQT,M);
    this.pV_VTH=p_offset+TMatr.getAMatrA(VT.psiH,QQpV,n)/TMatr.getAMatrA(VT.psiH,QQV,n);
    this.pI_VTH=p_offset+TMatr.getAMatrA(VT.psiH,QQpi,n)/TMatr.getAMatrA(VT.psiH,QQi,n);
    this.TI_VTH=TMatr.getAMatrA(VT.psiH,QQTi,n)/TMatr.getAMatrA(VT.psiH,QQi,n);

    if(eps>0){
	final double vt=TMatr.getAMatrA(VT.psiH,QQV,n)/TMatr.getAMatrA(VT.psiH,QQT,n),
	    i=TMatr.getAMatrA(VT.psiH,QQi,n)/TMatr.getAMatrA(VT.psiH,QQt,n);
	UnitTests.checkDoubleRel(vt,"vt",i,"i",eps);
    }


    if(eps>0){
	EVXData.test_orthogonality("PnLInPsiHstate.TiV.ev",n,TiV.ev,QQTi,QQV,eps);
	// make sure common relations hold:
	final double varV=TMatr.getAMatrB(psiTiV,DpsiTiV,QQV,n),valV=TMatr.getAMatrA(psiTiV,QQV,n);
	UnitTests.checkDoubleRel(2*varV,"2*varV",i_TiVH,"i_TiVH",eps);
	UnitTests.checkDoubleRel(2*varTi_TiVH,"2*varTi_TiVH",TiV.sH*i_TiVH,"TiV.sH*i_TiVH",eps);
	UnitTests.checkDoubleRel(Ti_TiVH,"Ti_TiVH",valV*varTi_TiVH/varV,"valV*varTi_TiVH/varV",eps);
	final var rq=RayleighQuotient.getRQVariationValues(psiTiV,DpsiTiV,null,QQTi,QQV,n);
	UnitTests.checkDoubleRel(rq[0],"rq[0]",TiV.sH,"TiV.sH",eps);
	UnitTests.checkDoubleRel(rq[1],"rq[1]",0,"0",eps);
	final var rqP=RayleighQuotient.getRQVariationValues(psiTiV,DpsiTiV,null,QQTpi,QQV,n);
	UnitTests.checkDoubleRel(rqP[0],"rqP[0]",(P_TiVH-p_offset)*Ti_TiVH/valV,"(P_TiVH-p_offset)*Ti_TiVH/valV",eps);
	UnitTests.checkDoubleRel(0.5*rqP[1],"0.5*rqP[1]",(Pvar_TiVH-P_TiVH)*varTi_TiVH/valV,"(Pvar_TiVH-P_TiVH)*varTi_TiVH/valV",eps);

    }

    DTII_L=new OneStateSaved(n,0, QQt, QQi,QQpi, QQTi,QQTpi);
    statesM.add(new OneIState(DTII_L.psiM,-100,QQt,QQi,QQpi));
    DTII_H=new OneStateSaved(n,n-1, QQt, QQi,QQpi, QQTi,QQTpi);
    statesM.add(new OneIState(DTII_H.psiM,-100,QQt,QQi,QQpi));
    TDI_L=new OneCTIstate(n,0,-1.0,  QQt, QQi,QQpi, QQTi,QQTpi);
    TDI_H=new OneCTIstate(n,n-1,-1.0,  QQt, QQi,QQpi, QQTi,QQTpi);
    statesM.add(new OneIState(TDI_L.psiM,-100,QQt,QQi,QQpi));
    statesM.add(new OneIState(TDI_H.psiM,-100,QQt,QQi,QQpi));

    if(eps>0){
	final double [] QQDTi=voltlty.getDiffMatrix(n,QQTi),QQDTpi=voltlty.getDiffMatrix(n,QQTpi);
	// check the matrix
	for(int j=0;j<n;j++){
	    final var psij=I.getPsi(j);
	    for(int k=0;k<n;k++){
		final var psik=I.getPsi(k);
		for(var s:new double [][][]{{QQTi,QQDTi},{QQTpi,QQDTpi}}){
		    final double [] QQm=s[0],QQDm=s[1];
		    final double ejk=TMatr.getAMatrB(psij,getEDPsi(psik),QQm,n)+TMatr.getAMatrB(psik,getEDPsi(psij),QQm,n);
		    UnitTests.checkDoubleRel(TMatr.getAMatrB(psij,psik,QQDm,n),"<psij|QQDm|psik>",ejk,"ejk",eps);
		    UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psij,QQDm,n),"<psik|QQDm|psij>",ejk,"ejk",eps);
		}
	    }
	}
	// known relations
	UnitTests.checkDoubleArrayRel(voltlty.getDiffMatrix(n,QQT),"D(QQT)",QQt,"QQt",eps);
	UnitTests.checkDoubleArrayRel(voltlty.getDiffMatrix(n,QQV),"D(QQV)",QQi,"QQi",eps);
	System.err.println("D(Ti)=lambda I.ev="+java.util.Arrays.toString(DTII_H.J.getEValues()));
    }


    final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	QQPImPt=TMatr.sumAcB(QQpi,-1,TMatr.selectHermitianPartInPlace(n,TMatr.mulMatrices(n,new double [][]{QQp1t,IQQt,QQi})));
    PiPt=EVXData.getEVXDataFromQQ(n,QQPImPt,QQt,M);
    statesM.add(new OneIState(PiPt.psiL,-100,QQt,QQi,QQpi));
    statesM.add(new OneIState(PiPt.psiH,-100,QQt,QQi,QQpi));

    if(eps>0){
	UnitTests.checkDoubleArrayRel(QQt,"QQt",QQp0t,"QQp0t",eps);
	for(int k=0;k<n;k++){
	    final double [] psik=I.getPsi(k);
	    final double Ipipt=TMatr.getAMatrA(psik,QQPImPt,n),
		pt=TMatr.getAMatrA(psik,QQp1t,n)/TMatr.getAMatrA(psik,QQp0t,n),
		pi=TMatr.getAMatrA(psik,QQpi,n)/TMatr.getAMatrA(psik,QQi,n);
	    UnitTests.checkDoubleRel(Ipipt,"Ipipt",(pi-pt)*TMatr.getAMatrA(psik,QQi,n),"(pi-pt)*I",eps);
	}
    }

    {
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final double []
	QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	QQi_feq=TMatr.sumAcB(QQi,dI/I.o_o,QQ0),
	QQpi_feq=TMatr.sumAcB(QQpi,p_last*dI/I.o_o,QQ0);
    If=EVXData.getEVXDataFromQQ(n,QQi_feq,QQt,M);
    double svar_feq=0,svar_pt_feq=0;
    for(int k=0;k<n;k++){
	final double [] psik=If.getPsi(k),Dpsik=getEDPsi(psik);
	final double wk0=Math.pow(TMatr.getAMatrB(psik,If.psi0,QQt,n),2),
	    pt=TMatr.getAMatrA(psik,QQp1t,n)/TMatr.getAMatrA(psik,QQp0t,n),
	    pi=TMatr.getAMatrA(psik,QQpi_feq,n)/If.ev[k];

	final var rqI=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQi_feq,QQt,n);
	if(/* If.ev[k]>If.s0 */rqI[2]<0){
	    svar_pt_feq+=(pi-pt)*If.ev[k]*wk0;

	    final var rq=RayleighQuotient.getRQVariationValues(psik,Dpsik,null,QQpi_feq,QQt,n);
	    svar_feq+=rq[1];
	}
	if(eps>0){
	    UnitTests.checkDoubleRel(rqI[1],"rqI[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(If.psi0,QQi_feq,n),"<psi0|QQi_feq|psi0>",I_0_feq,"I_0_feq",eps);
	}
    }
    OLD_svar_feq=svar_feq;
    OLD_svar_pt_feq=svar_pt_feq;
    }

    }
}

/** A class for unit test.
*/
static abstract class PnLInPsiHstateTest<DTYPE> extends QTest {

abstract PnLInPsiHstate<DTYPE> getPnLInPsiHstate();

PnLInPsiHstateTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}
void test_Regular()
{
    final int n=neq;
    final PnLInPsiHstate<DTYPE> q=getPnLInPsiHstate();

    System.err.println(name+":PnLInPsiHstateTest.test_Regular(): test OK n="+n+" eps="+eps);
}
void test_Extra(){}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":PnLInPsiHstate:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Regular();
    test_Extra();
    test_IvsVT_CalculateMoments();
    test_IvsVT_CalculateMoments_Extra();
    test_FreeMoneyForAll();

    System.err.println(name+":PnLInPsiHstate.test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_min,final int neq_max)
{
    System.err.println(name+":PnLInPsiHstate: starting tests; ver=$Id: PnLInPsiHstate.java,v 1.224 2024/04/15 06:15:33 mal Exp $");
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=eps;
    final long t0=System.currentTimeMillis();
    for(int neq=neq_min;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;
    final long t1=System.currentTimeMillis();
    System.err.println(name+":PnLInPsiHstate.main: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: PnLInPsiHstate.java,v 1.224 2024/04/15 06:15:33 mal Exp $");
}

private static String getFnameToDump(final String out_fname_prefix,final int n,long tau,final long t){

    final double t0dumpinhours=t/(3600.0*com.polytechnik.realtimebook.TraderInterface.TIME_SCALE);
    final String out_fname_key=out_fname_prefix+"_n_"+n+"_tdump_"+t0dumpinhours+"_tau_"+tau;
    return out_fname_key+"_t_"+t+".dump.csv";
}

/** At given tick dump several variables RN-interpolated at X as {@literal <psiX|f|psiX>}
 * along with {@literal |psi_M>} at X.
 */
static void dumpWavefunctionOnce(final PnLInPsiHstate<?> ref,
				 final double [] QQt,
				 final double [] QQi,
				 final double [] QQpi,
				 java.util.List<double []> mQQ,
				 final long currenttime
			  ){

    // a kludgy way to get t->x, do it in debug instead of adding an abstract method in main code
    final WIntegrator wi;
    if(ref instanceof PnLInPsiHstateLegendreShifted r) wi=new WIntegratorLegendreShifted(0,r.odata.tau,0);
    else if(ref instanceof PnLInPsiHstateLaguerre r) wi=new WIntegratorLaguerre(0,r.odata.d.tau,0);
    else if(ref instanceof PnLInPsiHstateMonomials r) wi=new WIntegratorMonomials(0,r.odata.d.tau,0);
    else throw new RuntimeException("unknown class "+ref);

    final int n=ref.I.neq;
    final String outFilename=getFnameToDump("_"+ref.getClass().getSimpleName(),n,(long)wi.atau,currenttime);
    System.err.println("PnLInPsiHstate: dumping wavefunctions to "+outFilename);

    java.io.PrintWriter out=null;
    try {
	out=new java.io.PrintWriter(new java.io.FileWriter(outFilename));
    } catch (Throwable e) {throw new RuntimeException(e);}
    out.print("|1:t\t2:x\t3:I\t4:piRN\t5:<psi|psiH>^2\t6:<psi|psiL>^2");
    int v=0,k=7;
    for(var s:mQQ){
	out.print("\t"+(k++)+":F_"+v);
	v++;
    }
    out.println();

    final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt);
    for(long t=currenttime;
	t>=(long)(8.1*3600.0*com.polytechnik.realtimebook.TraderInterface.TIME_SCALE);
	// dump once a second
	t-= 1*com.polytechnik.realtimebook.TraderInterface.TIME_SCALE){
	final double x=wi.getXFromTime(currenttime,t);
	final double [] psiX=TMatr.getMatr_x_A(n,IQQt,ref.M.B.getX0Moments(n,x));
	TMatr.scaleVectorInPlace(psiX,1/Math.sqrt(TMatr.getAMatrA(psiX,QQt,n)));
	final double i=TMatr.getAMatrA(psiX,QQi,n),
	    piRN=ref.p_offset+TMatr.getAMatrA(psiX,QQpi,n)/i,
	    projL=Math.pow(TMatr.getAMatrB(psiX,ref.I.psiL,QQt,n),2),
	    projH=Math.pow(TMatr.getAMatrB(psiX,ref.I.psiH,QQt,n),2);
	out.print(""+t+"\t"+x+"\t"+i+"\t"+piRN+"\t"+projH+"\t"+projL);
	for(var s:mQQ){
	    final double F=TMatr.getAMatrA(psiX,s,n);
	    out.print("\t"+F);
	}
	out.println();
    }
    out.close();
}


    abstract java.util.List<? extends CommonlyUsedMoments> getMomentsToMatch();

    void test_IvsVT_CalculateMoments_Extra(){}
    void test_IvsVT_CalculateMoments(){

	final int n=neq;
	final PnLInPsiHstate<DTYPE> q=getPnLInPsiHstate();
	final var toMatch=getMomentsToMatch();
	final SMomentsData mm=q.getSMoments(toMatch.get(0).getdVMoments().length);
	// match the moments
	CommonlyUsedMoments_TestFromSample.test_CalculateMoments_matchSMomentsData(
	    mm,toMatch,eps);
	System.err.println(name+":test_IvsVT_CalculateMoments(): test OK with N points="+tb.length+" neq="+neq+" eps="+eps);
    }

void test_FreeMoneyForAll(){
    final int n=neq;
    final boolean flagUseSecondarySamplingFordIHdt=false;
    final PnLInPsiHstate<DTYPE> q=getPnLInPsiHstate();
    for(var mD:getMomentsToMatch()){
	IntervalSincePsiM.matchFreeMoneyForAll(q.SM,new FreeMoneyForAll(n,mD,flagUseSecondarySamplingFordIHdt,eps),eps);
	PFuture.match(q.pFV,new FreeMoneyForAll(n,mD,flagUseSecondarySamplingFordIHdt,eps).pFV,eps);
    }
    System.err.println(name+":test_FreeMoneyForAll(): test OK with N points="+tb.length+" neq="+neq+" eps="+eps);
}

} // END OF class PnLInPsiHstateTest

static void matchPnLInPsiHstate(final PnLInPsiHstate<?> a,final PnLInPsiHstate<?> b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    EVXData.matchEVXData(a.MT.TiV,b.MT.TiV,false,eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_last,"a.p_last",b.p_last+p_offset_shift,"b.p_last+p_offset_shift",eps);
    // in this version output prices (public fields) are without an offset
    UnitTests.checkDoubleRel(a.pi_IH,"a.pi_IH",b.pi_IH,"b.pi_IH",eps);
    UnitTests.checkDoubleRel(a.pv_average,"a.pv_average",b.pv_average,"b.pv_average",eps);
    UnitTests.checkDoubleRel(a.pt_average,"a.pt_average",b.pt_average,"b.pt_average",eps);

    if(a.MT.statesM.size()!=b.MT.statesM.size()) throw new RuntimeException("a.MT.statesM.size()="+a.MT.statesM.size()+" b.MT.statesM.size()="+b.MT.statesM.size());
    for(int k=0;k<a.MT.statesM.size();k++){
	final var sa=a.MT.statesM.get(k);
	final var sb=b.MT.statesM.get(k);
	UnitTests.checkDoubleRel(sa.p_M,"sa.p_M",sb.p_M+p_offset_shift,"sb.p_M+p_offset_shift",eps);
	UnitTests.checkDoubleRel(sa.i,"sa.i",sb.i,"sb.i",eps);
	UnitTests.checkDoubleRel(sa.S_M,"sa.S_M",sb.S_M,"sb.S_M",eps);
	UnitTests.checkDoubleRel(sa.var_S_M,"sa.var_S_M",sb.var_S_M,"sb.var_S_M",eps);
    }

    BasisTransformPiTiPrediction.match(a.PiTi,b.PiTi,eps);
    IntervalSincePsiM.match(a.SM,b.SM,eps);
    IntervalSincePsiM.match(a.SMVT,b.SMVT,eps);
    IntervalSincePsiM.match(a.SMTimVT_H,b.SMTimVT_H,eps);
    IntervalSincePsiM.match(a.SMTimVT_L,b.SMTimVT_L,eps);

    System.err.println("PnLInPsiHstate.matchPnLInPsiHstate test OK p_offset_shift="+p_offset_shift+" n="+a.I.neq+" nM="+a.MT.statesM.size()+" eps="+eps);
}


}
