package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.ProbabilityCorrelation;

/**
 * @version $Id: PnLSimpleSkewness.java,v 1.21 2022/07/15 19:57:10 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLSimpleSkewness {
    public final EVXData I,P;
    public final double p_offset,p_IH,p_IL,p_B,P0;
    public final Skewness SK_P_average,gSK_I_average,gSK_P_average;
    public final ProbabilityCorrelation PCORR_average,PCORR_dpI;
    PnLSimpleSkewness(final MomentsData mm,
		      final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	final int n=mm.n;
	if(n!=2) throw new RuntimeException("n="+n);
	this.p_offset=mm.p_offset;
	// Moving average (familiar solution).
	// SK_P_average.xa=mm.QQpi[0]/mm.QQi[0] is equal to exponential moving average fix tau time scale.
	SK_P_average=Skewness.getSkewnessFromMonomialMoments(
	    mm.QQi[0]/mm.QQt[0],
	    mm.QQpi[0]/mm.QQt[0],
	    mm.QQp2i[0]/mm.QQt[0],
	    mm.QQp3i[0]/mm.QQt[0]);
	I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
	p_IH=TMatr.getAMatrA(I.psiH,mm.QQpi,n)/I.sH;
	p_IL=TMatr.getAMatrA(I.psiL,mm.QQpi,n)/I.sL;
	final double [] psiB=M.getRadauMultiplied(new double [] {1});
	p_B=TMatr.getAMatrA(psiB,mm.QQpi,n)/TMatr.getAMatrA(psiB,mm.QQi,n);
	P0=TMatr.getAMatrA(I.psi0,mm.QQpi,n)/I.s0;
	// Generalized skewness of P
	gSK_P_average=Skewness.getGSkewness(mm.QQi,mm.QQpi,M);
	P=EVXData.getEVXDataFromQQ(n,mm.QQpi,mm.QQi,M);
	// Generalized skewness of I
	gSK_I_average=Skewness.getGSkewness(mm.QQt,mm.QQi,M);

	// Some calculations out of curiosity:
	// Probability correlation
	// note t- and i- moments are swapped for P to have ths same basis as 1/I
	PCORR_average=new ProbabilityCorrelation(mm.QQi,mm.QQpi,mm.QQt,M,true);
	// This one dp- dependence is no good, just for the reference.
	// The PCORR_dpI.skewnessLike_i is approximately equal to -PCORR_average.skewnessLike_i
	// (as the moments are not swapped) and slightly better than -PCORR_average.skewnessLike_i.
	PCORR_dpI=new ProbabilityCorrelation(mm.QQt,mm.QQdp,mm.QQi,M,true);

    }
static void matchPnLSimpleSkewness(final PnLSimpleSkewness a,final PnLSimpleSkewness b,final double eps)
{
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.p_IL,"a.p_IL",b.p_IL+p_offset_shift,"b.p_IL+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.p_B,"a.p_B",b.p_B+p_offset_shift,"b.p_B+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.P0,"a.P0",b.P0+p_offset_shift,"b.P0+p_offset_shift",eps);
    EVXData.matchEVXData(a.I,b.I,false,eps);
    Skewness.matchSkewness(a.SK_P_average.getSameSkewnessXShifted(a.p_offset),b.SK_P_average.getSameSkewnessXShifted(b.p_offset),eps);
    Skewness.matchSkewness(a.gSK_I_average,b.gSK_I_average,eps);
    Skewness.matchSkewness(a.gSK_P_average.getSameSkewnessXShifted(a.p_offset),b.gSK_P_average.getSameSkewnessXShifted(b.p_offset),eps);

    ProbabilityCorrelation.matchProbabilityCorrelationWithOffset(a.PCORR_average,a.p_offset,b.PCORR_average,b.p_offset,eps);
    ProbabilityCorrelation.matchProbabilityCorrelation(a.PCORR_dpI,b.PCORR_dpI,eps);

    System.err.println("PnLSimpleSkewness.matchPnLSimpleSkewness test OK n="+a.I.neq+" eps="+eps);
}
}

