package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.RayleighQuotient;

/**
 * @version $Id: PnLVarVolatilitypi.java,v 1.16 2022/07/28 11:15:39 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLVarVolatilitypi {
    public final EVXData I;
    final double p_offset;
    final double [] pI,varpI,p,Volatility,varVolatility;
    final double p_IH,varVol_IH,p_A,varVol_A;
    PnLVarVolatilitypi(final EVXData I,final double p_offset,
		       final double [] pI,final double [] varpI,
		       final double [] Volatility,final double [] varVolatility)
{
    this.I=I;
    final int n=I.neq;
    this.p_offset=p_offset;
    this.pI=pI;
    this.varpI=varpI;
    this.p=new double [n];
    for(int k=0;k<n;k++) p[k]=pI[k]/I.ev[k];
    this.Volatility=Volatility;
    this.varVolatility=varVolatility;


    final int iH=n-1;
    p_IH=p[iH];
    varVol_IH=varVolatility[iH];
    p_A=TMatr.sumElements(pI)/TMatr.sumElements(I.getEValues());
    varVol_A=TMatr.sumElements(varVolatility);

}

/**
 * Find (eigenvector,eigenvalue) pairs (lambda,|psi>)
 * of generalized eigenvalue problem I|psi>=lambda |psi>.
 * (Typically the calculations are performed in price basis)
 * Find the state psi_IH, corresponding to max lambda
 * and the price p_IH, corresponding to it.
 * Then consider stdev--type volatility <(p-p_IH)^2 *I*psi_IH(p)^2>/<psi_IH(p)^2>
 * and variate it over the p.
 */
public static PnLVarVolatilitypi get_PnLVarVolatilitypi(final MomentsData mm,
							final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
							final VolMatrix VOLATILITY_EDPsi,
							final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);

    final double [] QQpIvol=new double [n*n];
    final double [] pI=new double[n],varpI=new double [n],Volatility=new double[n],varVolatility=new double[n];
    for(int k=0;k<n;k++)
    {
	pI[k]=I.getFPsi(k,mm.QQpi);
	final double pk=pI[k]/I.ev[k];
	for(int l=0;l<QQpIvol.length;l++) QQpIvol[l]=mm.QQp2i[l]+pk*(-2*mm.QQpi[l]+pk*mm.QQi[l]);
	Volatility[k]=I.getFPsi(k,QQpIvol);
	final double [] Psi=I.getPsi(k),varPsi=VOLATILITY_EDPsi.getDPsi(Psi);
	final double psivarpsiQQt=TMatr.getAMatrB(Psi,varPsi,mm.QQt,n);
	varVolatility[k]=2*(TMatr.getAMatrB(Psi,varPsi,QQpIvol,n)-Volatility[k]*psivarpsiQQt);
	varpI[k]=2*(TMatr.getAMatrB(Psi,varPsi,mm.QQpi,n)-pI[k]*psivarpsiQQt);
	if(eps>0)
	{
	    final double [] rq=RayleighQuotient.getRQVariationValues(Psi,varPsi,null,QQpIvol,mm.QQt,n);
	    UnitTests.checkDoubleRel(rq[0],"rq[0]",Volatility[k],"Volatility[k]",eps);
	    UnitTests.checkDoubleRel(rq[1],"rq[1]",varVolatility[k],"varVolatility[k]",eps);
	    final double [] rqpI=RayleighQuotient.getRQVariationValues(Psi,varPsi,null,mm.QQpi,mm.QQt,n);
	    UnitTests.checkDoubleRel(rqpI[0],"rqpI[0]",pI[k],"pI[k]",eps);
	    UnitTests.checkDoubleRel(rqpI[1],"rqpI[1]",varpI[k],"varpI[k]",eps);
	    if(false)
	    {
		System.err.println("Psi="+java.util.Arrays.toString(Psi));
		System.err.println("varPsi="+java.util.Arrays.toString(varPsi));
		// rqpI[1]=0 not holds even for price basis
		UnitTests.checkDoubleRel(rqpI[1],"rqpI[1]",0,"0",eps);
	    }
	    final DynVolPnL d=new DynVolPnL(n,k,mm.QQt,mm.QQi,mm.QQpi,mm.QQp2i,M);
	    UnitTests.checkDoubleRel(d.volP2EB,"d.volP2EB",Volatility[k],"Volatility[k]",eps);
	    UnitTests.checkDoubleRel(d.pEB,"d.pEB",pk,"pk",eps);
	}
    }
    if(eps>0)
    {
	// Check eigenvalues problem is solved OK.
	EVXData.test_orthogonality("PnLVarVolatilitypi.I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	System.err.println("PnLVarVolatilitypi.get_PnLVarVolatilitypi: all internal tests OK, all invariance tests OK n="+n+" eps="+eps);
    }

    return new PnLVarVolatilitypi(I,mm.p_offset,pI,varpI,Volatility,varVolatility);
}


}

