package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.DiffSkewness;

/**
 * Skewness on dI/dt states is back.
 *  @version $Id: PnLdI2Skewness.java,v 1.30 2022/08/05 11:25:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLdI2Skewness {


public final EVXData I;
public final double p_offset,p_IH,I_0_feq;

PnLdI2Skewness(final EVXData I,final double p_offset,final double p_last__with_p_offset_included,
	       final double p_IH,final double I_0_feq,
	       final DiffSkewness [] SK_rates,final double [] rates,
	       final double eps)
{
    this.I=I;
    this.p_IH=p_IH;
    this.p_offset=p_offset;
    final int n=I.neq;
    this.I_0_feq=I_0_feq;
}



/**
 * Solve {@literal <psi|d/dt I^f|psi>/<psi|psi>}  {@literal ->} max
 * extremum problem, (assume I^{*}=max I)
 * Then consider skewness (calculated from quadratures built) on:
 *  {@literal <psi|p^m I|psi>} moments (price--space regular skewness).
 */
public static PnLdI2Skewness get_PnLdI2Skewness(final MomentsData mm,
						final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
						final VolMatrix VOLATILITY_EDPsi,
						final double eps)
{
    final int n=mm.n,ixH=n-1;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0,dV=dI/I.o_o,
	p_last__with_p_offset_included=mm.p_last;

    final double [] bQQDi=EVXData.convertMatrFromQtoB_Basis(n,TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDiffMatrix(n,mm.QQi),-1),I.ev),
	bQQ0=TMatr.getQQaa(I.psik_at_x0),
	bQQD0=EVXData.convertMatrFromQtoB_Basis(n,TMatr.scaleVectorInPlace(VOLATILITY_EDPsi.getDiffMatrix(n,TMatr.getQQaa(M.B.getX0Moments(n,I.x0))),-1),I.ev),
	bQQi=TMatr.getDiagMatrix(I.getEValues()),
	bQQi_feq=TMatr.sumAcB(bQQi,dV,bQQ0),
	bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev),
	bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,mm.QQp2i,I.ev),
	bQQp3i=EVXData.convertMatrFromQtoB_Basis(n,mm.QQp3i,I.ev);

    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true),bQQDi_feqF=new double [n*n];
    final double I_0_feqF=bbasis[ixH],dIF=I_0_feqF-I_0_feq;
    for(int k=n*n;--k>=0;) bQQDi_feqF[k]=I_0_feqF*bQQ0[k]+dV*bQQD0[k]+bQQDi[k];
    final double [] Dbbasis=M.EV.getGEV_Generalized(n,bQQDi_feqF,bQQi_feq,true);
    final DiffSkewness [] SK_rates=new DiffSkewness [n];
    for(int k=0;k<n;k++)
    {
	SK_rates[k]=DiffSkewness.getDiffSkewness_FromMonomials(
	    TMatr.getAMatrA(n,Dbbasis,n+k*n,bQQi,n),
	    TMatr.getAMatrA(n,Dbbasis,n+k*n,bQQpi,n),
	    TMatr.getAMatrA(n,Dbbasis,n+k*n,bQQp2i,n),
	    TMatr.getAMatrA(n,Dbbasis,n+k*n,bQQp3i,n),
	    dV*TMatr.getAMatrA(n,Dbbasis,n+k*n,bQQ0,n));
    }
    if(eps>0)
    {
	// preliminary checks
	final double [] bbQQDi_feqF=EVXData.convertMatrFromQtoB_Basis(n,bQQDi_feqF,bbasis);
	final double [] bDbbasis=EVXData.getSymEVDiagT(n,bbQQDi_feqF,TMatr.getDiagMatrix(java.util.Arrays.copyOf(bbasis,n)),M.EV);
	UnitTests.checkDoubleArrayRel1(java.util.Arrays.copyOf(Dbbasis,n),"EV Dbbasis",
				       java.util.Arrays.copyOf(bDbbasis,n),"EV bDbbasis",eps);
	// check orthogonality
	EVXData.test_orthogonality("PnLdI2Skewness.get_PnLdI2Skewness:I.ev",n,I.ev,mm.QQi,mm.QQt,eps); // old basis
	EVXData.test_orthogonality("PnLdI2Skewness.get_PnLdI2Skewness:bbasis",n,bbasis,bQQi_feq,TMatr.getUnitMatrix(n),eps);
	if(n>2) EVXData.test_orthogonality("PnLdI2Skewness.get_PnLdI2Skewness:Dbbasis",n,Dbbasis,bQQDi_feqF,bQQi_feq,eps);

	// Check values on |psi0>
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<pis0|QQI_feq|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,mm.QQi,n),"<pis0|QQI|psi0>",I.s0,"I.s0",eps);
	final double [] di=new double [n];
	// Check values on |psi>
	double spur_bQQDi_feqF=0;
	for(int k=0;k<n;k++)
	{
	    final double [] psi=EVXData.getPsiFromBasis(k,bbasis,n);
	    final double psiat0=TMatr.mulVV(psi,I.psik_at_x0),
		me=(I_0_feqF-bbasis[k])*Math.pow(psiat0,2);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,bQQDi_feqF,n),"<psik|bQQDi_feqF|psik>",me,"(I_f-I_k)*psik(x0)^2",eps);
	    spur_bQQDi_feqF+=me/bbasis[k];
	    final DiffSkewness s=SK_rates[k];
	    UnitTests.checkDoubleRel(s.di+s.sknodi.w1+s.sknodi.w2,"s.di+s.sknodi.w1+s.sknodi.w2",1,"1",eps);
	    di[k]=s.di;
	}
	UnitTests.checkDoubleRel(TMatr.sumElements(java.util.Arrays.copyOf(Dbbasis,n)),"spur(Dbbasis)",spur_bQQDi_feqF,"spur_bQQDi_feqF",eps);
	System.err.println("PnLdI2Skewness: I="+java.util.Arrays.toString(I.getEValues()));
	System.err.println("PnLdI2Skewness: If="+java.util.Arrays.toString(java.util.Arrays.copyOf(bbasis,n)));
	System.err.println("PnLdI2Skewness: rates="+java.util.Arrays.toString(java.util.Arrays.copyOf(Dbbasis,n))+" sum(rates)="+spur_bQQDi_feqF);
	System.err.println("PnLdI2Skewness: di="+java.util.Arrays.toString(di));
	System.err.println("PnLdI2Skewness: matrix test OK n="+n+" eps="+eps);
    }

    return new PnLdI2Skewness(I,mm.p_offset,p_last__with_p_offset_included,
			      TMatr.getAMatrA(I.psiH,mm.QQpi,n)/I.sH,I_0_feq,
			      SK_rates,java.util.Arrays.copyOf(Dbbasis,n),eps);
}

}
