package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.ValueCorrelation;
import com.polytechnik.utils.EVSolver;

/**
 * Skewness is back. Correlation added.
 *  @version $Id: PnLdICorr.java,v 1.52 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLdICorr {

/** Calculate {@literal <l^{(v)}_k>_t} using bv basis,
 * then calculate the rate {@literal <l^{(v)}_k>_v/<l^{(v)}_k>_t}
 */
static class RateRel {
    public final double r1,r2,r1_div,r2_div;
    RateRel(final Skewness bv,final Skewness bt,final double eps)
    {
	final double dbv=bv.x1-bv.x2,
	    tau1=Math.pow((bt.x1-bv.x2)/dbv,2)*bt.w1+Math.pow((bt.x2-bv.x2)/dbv,2)*bt.w2,
	    tau2=Math.pow((bt.x1-bv.x1)/dbv,2)*bt.w1+Math.pow((bt.x2-bv.x1)/dbv,2)*bt.w2;
	this.r1=bv.w1/tau1;
	this.r2=bv.w2/tau2;
	this.r1_div=bv.w1/bt.w1;
	this.r2_div=bv.w2/bt.w2;
	if(eps>0)
	{
	    final double
		vau1=Math.pow((bv.x1-bv.x2)/dbv,2)*bv.w1+Math.pow((bv.x2-bv.x2)/dbv,2)*bv.w2,
		vau2=Math.pow((bv.x1-bv.x1)/dbv,2)*bv.w1+Math.pow((bv.x2-bv.x1)/dbv,2)*bv.w2;
	    UnitTests.checkDoubleRel(vau1,"vau1",bv.w1,"bv.w1",eps);
	    UnitTests.checkDoubleRel(vau2,"vau2",bv.w2,"bv.w2",eps);
	}

    }
}

static class CrossCorr {
    public final Skewness SK_pv,SK_pt,SK_rv,SK_rt;
    public final double [] crosscorr_v,crosscorr_t;
    public final double corr_v;
    public final RateRel ipv;
    CrossCorr(final Skewness SK_pv,final Skewness SK_pt,
	      final Skewness SK_rv,final Skewness SK_rt,
	      final double prv,final double prt,final double eps)
    {
	this.SK_pv=SK_pv;this.SK_pt=SK_pt;this.SK_rv=SK_rv;this.SK_rt=SK_rt;
	this.crosscorr_v=getCrossCorr(SK_pv,SK_rv,prv);
	this.crosscorr_t=getCrossCorr(SK_pt,SK_rt,prt);
	this.corr_v=TMatr.mulVV(crosscorr_v,dir_signs)/TMatr.sumElements(crosscorr_v);
	if(eps>0)
	{
	    final double w=TMatr.sumElements(crosscorr_v);
	    UnitTests.checkDoubleRel(SK_pv.w1+SK_pv.w2,"SK_pv.w1+SK_pv.w2",w,"w",eps);
	    UnitTests.checkDoubleRel(SK_rv.w1+SK_rv.w2,"SK_rv.w1+SK_rv.w2",w,"w",eps);
	    UnitTests.checkDoubleArrayRel(crosscorr_v,"crosscorr_v",test_getCrossCorr(SK_pv,SK_rv,prv),"test crosscor",eps);
	    UnitTests.checkDoubleRel(getL4Corr(SK_pv,SK_rv,prv),"getL4Corr(SK_pv,SK_rv,prv)",corr_v,"corr_v",eps);
	    UnitTests.checkDoubleRel(crosscorr_v[0]+crosscorr_v[1],"crosscorr_v[0]+crosscorr_v[1]",SK_pv.w1,"SK_pv.w1",eps);
	    UnitTests.checkDoubleRel(crosscorr_v[2]+crosscorr_v[3],"crosscorr_v[2]+crosscorr_v[3]",SK_pv.w2,"SK_pv.w2",eps);
	    UnitTests.checkDoubleRel(crosscorr_v[0]+crosscorr_v[2],"crosscorr_v[0]+crosscorr_v[2]",SK_rv.w1,"SK_rv.w1",eps);
	    UnitTests.checkDoubleRel(crosscorr_v[1]+crosscorr_v[3],"crosscorr_v[1]+crosscorr_v[3]",SK_rv.w2,"SK_rv.w2",eps);
	    System.err.println("crosscorr_v="+java.util.Arrays.toString(crosscorr_v)+" ; SK_pv.w=("+SK_pv.w1+" ; "+SK_pv.w2+") SK_rv.w=("+SK_rv.w1+" ; "+SK_rv.w2+")");
	    final ValueCorrelation cv=new ValueCorrelation(w,prv,SK_pv,SK_rv);
	    UnitTests.checkDoubleRel(crosscorr_v[0],"crosscorr_v[0]",cv.w_pL_rL,"cv.w_pL_rL",eps);
	    UnitTests.checkDoubleRel(crosscorr_v[1],"crosscorr_v[1]",cv.w_pL_rH,"cv.w_pL_rH",eps);
	    UnitTests.checkDoubleRel(crosscorr_v[2],"crosscorr_v[2]",cv.w_pH_rL,"cv.w_pH_rL",eps);
	    UnitTests.checkDoubleRel(crosscorr_v[3],"crosscorr_v[3]",cv.w_pH_rH,"cv.w_pH_rH",eps);
	    UnitTests.checkDoubleRel(corr_v,"corr_v",cv.corr,"cv.corr",eps);
	    System.err.println("PnLdICorr.CrossCorr: test OK eps="+eps);
	}
	if(eps>0)
	{
	    UnitTests.checkDoubleRel(SK_pt.w1+SK_pt.w2,"SK_pv.w1+SK_pv.w2",SK_rt.w1+SK_rt.w2,"SK_rv.w1+SK_rv.w2",eps);
	}
	this.ipv=new RateRel(SK_pv,SK_pt,eps);
	if(eps>0)
	{
	    System.err.println("PnLdICorr.CrossCorr: test OK eps="+eps);
	}
    }
    private static final double [] dir_signs=new double []{1,-1,-1,1};
}

    public final EVXData I;
    public final double p_offset;
    public final CrossCorr IHf_cc,movaver_cc,spur_cc;
    PnLdICorr(final EVXData I,final double p_offset,
	      final CrossCorr IHf_cc,
	      final CrossCorr movaver_cc,
	      final CrossCorr spur_cc,
	      final double eps)
{
    this.I=I;
    this.p_offset=p_offset;
    this.IHf_cc=IHf_cc;
    this.movaver_cc=movaver_cc;
    this.spur_cc=spur_cc;
    if(eps>0)
    {
	System.err.println("PnLdICorr: test OK n="+I.neq+" eps="+eps);
    }
}

public static double [] getCrossCorr(final Skewness sp,final Skewness sr,final double pr)
{
    final double p12=sp.x1-sp.x2,r12=sr.x1-sr.x2,wp=sp.w1+sp.w2,pa=wp*sp.xa;
    return new double []{
	((pr-pa*sr.x2)/r12-sp.x2*sr.w1)/p12,
	-((pr-pa*sr.x1)/r12+sp.x2*sr.w2)/p12,
	-((pr-pa*sr.x2)/r12-sp.x1*sr.w1)/p12,
	((pr-pa*sr.x1)/r12+sp.x1*sr.w2)/p12
  };
}

public static double getL4Corr(final Skewness sp,final Skewness sr,final double pr_moment_not_scaled)
{
    final double w=sr.w1+sr.w2; // sp.w1+sp.w2 have to be the same
    return
	((pr_moment_not_scaled/w-sp.xa*sr.xa)+((sp.x1+sp.x2)/2-sp.xa)*((sr.x1+sr.x2)/2-sr.xa))/
	(0.25*(sp.x1-sp.x2)*(sr.x1-sr.x2));
}

static double [] test_getCrossCorr(final Skewness sp,final Skewness sr,final double pr)
{
    final double p1=sp.x1,p2=sp.x2,wp=sp.w1+sp.w2,
	r1=sr.x1,r2=sr.x2,wr=sr.w1+sr.w2,pr12=(p1-p2)*(r1-r2),
	pa=wp*sp.xa,ra=wr*sr.xa;
    return new double []{
	(pr-p2*ra-r2*pa+p2*r2*wr)/pr12,
	-(pr-p2*ra-r1*pa+p2*r1*wr)/pr12,
	-(pr-p1*ra-r2*pa+p1*r2*wr)/pr12,
	(pr-p1*ra-r1*pa+p1*r1*wr)/pr12
  };
}

/**
 * Solve {@literal <psi|I^f|psi>/<psi|psi>  -> max}
 * extremum problem, where I^{f} if execution flow
 * operator (with or without impact from the future).
 * Then consider skewness (calculated from quadratures built) on:
 * 1. {@literal <psi|p^m I|psi>} moments (price--space regular skewness).
 * 2. {@literal <psi|r^m I|psi>} moments (r=x0-x, Radau term. The distance to x0).
 * obtain p_{1,2} and r_{1,2} quadrature nodes.
 *
 * Then calculate cross--correlation matrix
 * {@literal <}l^{(p)})_j l^{(r)})_k I psi^2{@literal >}  can be considered, where
 * l^{(p)})_j -- Lagrange interpolating polynomial in p space,
 * l^{(r)})_k -- Lagrange interpolating polynomial in r space.
 * Try to obtain future information from p and r distribution skewness,
 * and, most hope, from cross--correlation matrix.
 */

public static PnLdICorr get_PnLdICorr(final int n,
				      final double p_offset,
				      final double [] momt,
				      final double [] mompt,
				      final double [] momp2t,
				      final double [] momp3t,
				      final double [] momi,
				      final double [] mompi,
				      final double [] momp2i,
				      final double [] momp3i,
				      final boolean add_dI,
				      final double p_last__with_p_offset_included,
				      final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				      final double eps)
{
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final double [] momri=M.getRadauMomentsFromMoments(momi),
	momr2i=M.getRadauMomentsFromMoments(momri),
	momr3i=M.getRadauMomentsFromMoments(momr2i),
	momrt=M.getRadauMomentsFromMoments(momt),
	momr2t=M.getRadauMomentsFromMoments(momrt),
	momr3t=M.getRadauMomentsFromMoments(momr2t),
	momrpi=M.getRadauMomentsFromMoments(mompi),
	momrpt=M.getRadauMomentsFromMoments(mompt);
    final double [] QQt=M.B.getQQMatr(n,momt),QQi=M.B.getQQMatr(n,momi);
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final double [] bQQi=TMatr.getDiagMatrix(I.getEValues()),
	bQQpi=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,mompi),I.ev),
	bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momp2i),I.ev),
	bQQp3i=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momp3i),I.ev),
	bQQpt=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,mompt),I.ev),
	bQQp2t=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momp2t),I.ev),
	bQQp3t=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momp3t),I.ev),
	bQQri=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momri),I.ev),
	bQQr2i=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momr2i),I.ev),
	bQQr3i=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momr3i),I.ev),
	bQQrt=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momrt),I.ev),
	bQQr2t=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momr2t),I.ev),
	bQQr3t=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momr3t),I.ev),
	bQQrpi=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momrpi),I.ev),
	bQQrpt=EVXData.convertMatrFromQtoB_Basis(n,M.B.getQQMatr(n,momrpt),I.ev);


    // Create operators with an impact from the future.
    if(add_dI)
    {
	final double pl=p_last__with_p_offset_included;
	final double [] bQQ0=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o));
	TMatr.sumP1(bQQi,bQQ0,1);
	TMatr.sumP1(bQQpi,bQQ0,pl);
	TMatr.sumP1(bQQp2i,bQQ0,Math.pow(pl,2));
	TMatr.sumP1(bQQp3i,bQQ0,Math.pow(pl,3));
	// 1. all r-matrix have 0 contribution at x0, no need to add dI.
	// 2. think later about bQQpt. (small, about =0).
    }
    if(eps>0)
    {
	EVXData.test_orthogonality("PnLdICorr.get_PnLdICorr:I.ev",n,I.ev,QQi,QQt,eps); // old basis
    }

    final int ixH=n-1;
    final double [] bbasis=M.EV.getEV(n,bQQi,true),psi=EVXData.getPsiFromBasis(ixH,bbasis,n),
	psiC=EVXData.convertPsiFromQtoB_Basis(n,java.util.Arrays.copyOf(new double []{1/Math.sqrt(QQt[0])},n),I.ev,QQt);
    if(eps>0)
    {
	EVXData.test_orthogonality("PnLdICorr.get_PnLdICorr:bbasis",n,bbasis,bQQi,TMatr.getUnitMatrix(n),eps); // new basis
	UnitTests.checkDoubleRel(bbasis[ixH],"bbasis[ixH]",TMatr.getAMatrA(psi,bQQi,n),"<psi|bQQi|psi>",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(psi,psi),"<psi|psi>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(psiC,psiC),"<psiC|psiC>",1,"1",eps);
    }
    return new PnLdICorr(I,p_offset,
			 new CrossCorr(Skewness.getSkewnessFromMonomialMoments(
					   bbasis[ixH],
					   TMatr.getAMatrA(psi,bQQpi,n),
					   TMatr.getAMatrA(psi,bQQp2i,n),
					   TMatr.getAMatrA(psi,bQQp3i,n)),
				       Skewness.getSkewnessFromMonomialMoments(
					   1,
					   TMatr.getAMatrA(psi,bQQpt,n),
					   TMatr.getAMatrA(psi,bQQp2t,n),
					   TMatr.getAMatrA(psi,bQQp3t,n)),
				       Skewness.getSkewnessFromMonomialMoments(
					   bbasis[ixH],
					   TMatr.getAMatrA(psi,bQQri,n),
					   TMatr.getAMatrA(psi,bQQr2i,n),
					   TMatr.getAMatrA(psi,bQQr3i,n)),
				       Skewness.getSkewnessFromMonomialMoments(
					   1,
					   TMatr.getAMatrA(psi,bQQrt,n),
					   TMatr.getAMatrA(psi,bQQr2t,n),
					   TMatr.getAMatrA(psi,bQQr3t,n)),
				       TMatr.getAMatrA(psi,bQQrpi,n),TMatr.getAMatrA(psi,bQQrpt,n),eps),
			 new CrossCorr(Skewness.getSkewnessFromMonomialMoments(
					   TMatr.getAMatrA(psiC,bQQi,n),
					   TMatr.getAMatrA(psiC,bQQpi,n),
					   TMatr.getAMatrA(psiC,bQQp2i,n),
					   TMatr.getAMatrA(psiC,bQQp3i,n)),
				       Skewness.getSkewnessFromMonomialMoments(
					   TMatr.mulVV(psiC,psiC),
					   TMatr.getAMatrA(psiC,bQQpt,n),
					   TMatr.getAMatrA(psiC,bQQp2t,n),
					   TMatr.getAMatrA(psiC,bQQp3t,n)),
				       Skewness.getSkewnessFromMonomialMoments(
					   TMatr.getAMatrA(psiC,bQQi,n),
					   TMatr.getAMatrA(psiC,bQQri,n),
					   TMatr.getAMatrA(psiC,bQQr2i,n),
					   TMatr.getAMatrA(psiC,bQQr3i,n)),
				       Skewness.getSkewnessFromMonomialMoments(
					   TMatr.mulVV(psiC,psiC),
					   TMatr.getAMatrA(psiC,bQQrt,n),
					   TMatr.getAMatrA(psiC,bQQr2t,n),
					   TMatr.getAMatrA(psiC,bQQr3t,n)),
				       TMatr.getAMatrA(psiC,bQQrpi,n),TMatr.getAMatrA(psiC,bQQrpt,n),eps),
			 new CrossCorr(Skewness.getSkewnessFromMonomialMoments(
					   TMatr.getSpurM(n,bQQi),
					   TMatr.getSpurM(n,bQQpi),
					   TMatr.getSpurM(n,bQQp2i),
					   TMatr.getSpurM(n,bQQp3i)),
				       Skewness.getSkewnessFromMonomialMoments(
					   n,
					   TMatr.getSpurM(n,bQQpt),
					   TMatr.getSpurM(n,bQQp2t),
					   TMatr.getSpurM(n,bQQp3t)),
				       Skewness.getSkewnessFromMonomialMoments(
					   TMatr.getSpurM(n,bQQi),
					   TMatr.getSpurM(n,bQQri),
					   TMatr.getSpurM(n,bQQr2i),
					   TMatr.getSpurM(n,bQQr3i)),
				       Skewness.getSkewnessFromMonomialMoments(
					   n,
					   TMatr.getSpurM(n,bQQrt),
					   TMatr.getSpurM(n,bQQr2t),
					   TMatr.getSpurM(n,bQQr3t)),
				       TMatr.getSpurM(n,bQQrpi),TMatr.getSpurM(n,bQQrpt),eps),eps);
}


static void matchPnLdICorr_CrossCorr(final PnLdICorr.CrossCorr a,final double a_p_offset,
				     final PnLdICorr.CrossCorr b,final double b_p_offset,
				     final double eps)
{
    Skewness.matchSkewness(a.SK_pv.getSameSkewnessXShifted(a_p_offset),b.SK_pv.getSameSkewnessXShifted(b_p_offset),eps);
    Skewness.matchSkewness(a.SK_pt.getSameSkewnessXShifted(a_p_offset),b.SK_pt.getSameSkewnessXShifted(b_p_offset),eps);
    Skewness.matchSkewness(a.SK_rv,b.SK_rv,eps);
    Skewness.matchSkewness(a.SK_rt,b.SK_rt,eps);
    UnitTests.checkDoubleArrayRel(a.crosscorr_v,"a.crosscorr_v",b.crosscorr_v,"b.crosscorr_v",eps);
    UnitTests.checkDoubleArrayRel(a.crosscorr_t,"a.crosscorr_t",b.crosscorr_t,"b.crosscorr_t",eps);

    System.err.println("PnLdICorr.matchPnLdICorr_CrossCorr test OK eps="+eps);
}


static void matchPnLdICorr(final PnLdICorr a,final PnLdICorr b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);

    matchPnLdICorr_CrossCorr(a.IHf_cc,a.p_offset,b.IHf_cc,b.p_offset,eps);
    matchPnLdICorr_CrossCorr(a.movaver_cc,a.p_offset,b.movaver_cc,b.p_offset,eps);
    matchPnLdICorr_CrossCorr(a.spur_cc,a.p_offset,b.spur_cc,b.p_offset,eps);

    System.err.println("PnLdICorr.matchPnLdICorr test OK n="+a.I.neq+" eps="+eps);
}

static void matchPnLdISkewness(final PnLdICorr a,final PnLdISkewness b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);

    Skewness.matchSkewness(a.IHf_cc.SK_pv.getSameSkewnessXShifted(a.p_offset),b.IHf_S.getSameSkewnessXShifted(b.p_offset),eps);
    Skewness.matchSkewness(a.movaver_cc.SK_pv.getSameSkewnessXShifted(a.p_offset),b.movaver_S.getSameSkewnessXShifted(b.p_offset),eps);
    Skewness.matchSkewness(a.spur_cc.SK_pv.getSameSkewnessXShifted(a.p_offset),b.spur_S.getSameSkewnessXShifted(b.p_offset),eps);

    System.err.println("PnLdICorr.matchPnLdISkewness test OK n="+a.I.neq+" eps="+eps);
}

static void matchPnLdIP(final PnLdICorr a,final PnLdIP b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);

    Skewness.matchSkewness(a.IHf_cc.SK_pv.getSameSkewnessXShifted(a.p_offset),b.IHf_S.getSameSkewnessXShifted(b.p_offset),eps);
    Skewness.matchSkewness(a.IHf_cc.SK_pt.getSameSkewnessXShifted(a.p_offset),b.IHf_St.getSameSkewnessXShifted(b.p_offset),eps);

    System.err.println("PnLdICorr.PnLdIP test OK n="+a.I.neq+" eps="+eps);
}
}
