package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLdID.java,v 1.30 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLdID {

    public final EVXData I;
    public final double p_offset,p_IH,I_0_feq,P_movaver,P_movaver_spur,var_pI_IH,P0_IHf,dp_IH,dp_orig_IH,dp_orig_0,dp_orig_IHf,P0_IHf_dp0,w_IHf_0,p_IHf;
    PnLdID(final EVXData I,final double p_offset,final double p_last__with_p_offset_included,
	   final double p_IH,final double I_0_feq,final double P_movaver,final double P_movaver_spur,final double var_pI_IH,final double P0_IHf,final double dp_orig_IH,final double dp_orig_0,final double dp_orig_IHf,final double w_IHf_0,final double p_IHf)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.I_0_feq=I_0_feq;
	this.P_movaver=P_movaver;
	this.P_movaver_spur=P_movaver_spur;
	this.var_pI_IH=var_pI_IH;
	this.P0_IHf=P0_IHf;
	this.dp_IH=-var_pI_IH/I.sH;
	this.dp_orig_IH=dp_orig_IH;
	this.dp_orig_0=dp_orig_0;
	this.dp_orig_IHf=dp_orig_IHf;
	this.w_IHf_0=w_IHf_0;
	this.P0_IHf_dp0=p_last__with_p_offset_included-(dp_orig_IHf/w_IHf_0)/I.o_o;
	this.p_IHf=p_IHf;
    }

/**
 * Solve {@literal <psi|I|psi>/<psi|psi>}=lambda extremum problem, find eigenstates &amp; eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-{@literal <psi0|I|psi0>} (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + {@literal |psi0>dI<psi0|}
 * Find the eigenstates of this new I^{f} matrix,
 * and the state psi^{IHf}, corresponding to the max of future I.
 * Then consider {@literal <}psi^{IHf}|dI/dt|psi^{IHf}{@literal >}=0 and {@literal <}psi^{IHf}|dpI/dt|psi^{IHf}{@literal >}=0.
 */
public static PnLdID get_PnLdID(final MomentsData mm,
				final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				final VolMatrix VOLATILITY_EDPsi,
				final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final int ixH=n-1; // the state with maximal I
    final double I_0=I.s0,I_0_feq=I.ev[ixH],dI=I_0_feq-I_0;

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o));
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];

    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev),
	bQQdp=EVXData.convertMatrFromQtoB_Basis(n,mm.QQdp,I.ev),
	bpsi_IH=new double [n];
    bpsi_IH[ixH]=1;

    // Diagonalize "future" I:
    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true),
	bpsi_IHf=EVXData.getPsiFromBasis(ixH,bbasis,n),
	Dbpsi_IHf=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,bpsi_IHf,I.ev)),I.ev,mm.QQt),
	Dbpsi_IH=EVXData.convertPsiFromQtoB_Basis(n,VOLATILITY_EDPsi.getDPsi(EVXData.convertPsiFromBtoQ_Basis(n,bpsi_IH,I.ev)),I.ev,mm.QQt);


    final double bpsi_IHf_x0_sq=2*TMatr.mulVV(bpsi_IHf,Dbpsi_IHf),
	dIr=2*dI*TMatr.mulVV(bpsi0,bpsi_IHf)*TMatr.mulVV(bpsi0,Dbpsi_IHf),
	pIf=2*TMatr.getAMatrB(bpsi_IHf,Dbpsi_IHf,bQQpi,n),
	iden=(bbasis[ixH]*bpsi_IHf_x0_sq-dIr),
	P0_IHf=pIf/iden,
	var_pI_IH=2*(TMatr.getAMatrB(bpsi_IH,Dbpsi_IH,bQQpi,n)-bQQpi[ixH+n*ixH]*TMatr.mulVV(bpsi_IH,Dbpsi_IH));

    if(eps>0)
    {
	EVXData.test_orthogonality("PnLdID.get_PnLdID:I.ev",n,I.ev,mm.QQi,mm.QQt,eps); // old basis
	EVXData.test_orthogonality("PnLdID.get_PnLdID:bbasis",n,bbasis,bQQi_feq,TMatr.getUnitMatrix(n),eps); // new basis

	// Check spur invariance.
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQi),"Spur(IQQt,QQi)",TMatr.sumElements(I.getEValues()),"sum(I.ev)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQpi),"Spur(IQQt,QQpi)",TMatr.getSpurM(n,bQQpi),"Spur(bQQpi)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQi_feq),"Spur(bQQi_feq)",TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),"sum(ev bbasis)",eps);

	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi_IHf,bQQi_feq,n),"<bpsi_IHf|bQQi_feq|bpsi_IHf>",bbasis[ixH],"bbasis[ixH]",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<bpsi0|bQQi_feq|bpsi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(iden,"iden",2*TMatr.mulVVV(bpsi_IHf,Dbpsi_IHf,I.getEValues()),"2*<bpsi_IHf|I|Dbpsi_IHf>",eps);
	UnitTests.checkDoubleRel(bpsi_IHf_x0_sq,"bpsi_IHf_x0_sq",Math.pow(TMatr.mulVV(bpsi_IHf,I.psik_at_x0),2),"bpsi_IHf(x0)^2",eps);
	UnitTests.checkDoubleRel(2*TMatr.mulVV(bpsi_IH,Dbpsi_IH),"2<bpsi_IH|Dbpsi_IH>",Math.pow(I.wH,2)*I.o_o,"I.wH^2*I.o_o",eps);
	System.err.println("PnLdID.get_PnLdID iden="+iden+" P0_IHf="+P0_IHf+" var_pI_IH="+var_pI_IH);

	System.err.println("PnLdID.get_PnLdID selftest OK n="+n+" eps="+eps);
    }
    final PnLdID r=new PnLdID(
	I,mm.p_offset,mm.p_last,
	bQQpi[ixH+n*ixH]/I.ev[ixH],I_0_feq,
	mm.QQpi[0]/mm.QQi[0],TMatr.getSpurM(n,bQQpi)/TMatr.sumElements(I.getEValues()),
	var_pI_IH,P0_IHf,
	TMatr.getAMatrA(bpsi_IH,bQQdp,n),
	TMatr.getAMatrA(bpsi0,bQQdp,n),
	TMatr.getAMatrA(bpsi_IHf,bQQdp,n),
	Math.pow(TMatr.mulVV(bpsi_IHf,bpsi0),2),
	TMatr.getAMatrA(bpsi_IHf,bQQpi,n)/TMatr.mulV2V(bpsi_IHf,I.getEValues()));
    if(eps>0)
    {
	final double [] psi_IHf=EVXData.convertPsiFromBtoQ_Basis(n,bpsi_IHf,I.ev),
	    QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	    QQdp_feq=TMatr.sumAcB(mm.QQdp,(r.P0_IHf_dp0-mm.p_last),QQ0);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_IHf,QQdp_feq,n),"<psi_IHf|QQdp_feq|psi_IHf>",0,"0",eps);
	System.err.println("PnLdID.get_PnLdID selftest2 OK n="+n+" eps="+eps);
    }
    return r;
}


}
