package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.DiffSkewness;
import com.polytechnik.utils.DerivativesCalculateable;

/**
 * Skewness is back. See also our old skewness approaches in RkiQVM,DynSki and QkQVM,EAssym.
 *  @version $Id: PnLdIDSk.java,v 1.136 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLdIDSk {


    public static class SKData {
	/** Skewness with additional di measure contribution
	 * at given support point, the di is in sk.di.
	 */
	public final DiffSkewness sk;
	/** w0={@literal <psi0|psi^f_k>}^2, ifuture=lambda^f_k, and di=sk.di=(I_H-I_0)*w0
	 */
	public final double ifuture;
	SKData(final DiffSkewness sk,final double ifuture,final double eps)
	{
	    this.sk=sk;
	    this.ifuture=ifuture; // =sk.di+sk.sknodi.w1+sk.sknodi.w2
	    if(eps>0){
		UnitTests.checkDoubleRel(sk.di+sk.sknodi.w1+sk.sknodi.w2,"sk.di+sk.sknodi.w1+sk.sknodi.w2",ifuture,"ifuture",eps);
		test_Self(eps); // more tests
	    }
	}

	/** Calculates skewness, Dskewness/Dp and D^2skewness/Dp^2.
	 * Typically not used, the sk.getSkewnessAt(p)
	 * is typically used.
	 */
	double [] getDSkewnessAt(final double p)
	{
	    final DerivativesCalculateable G=sk.getGamma();
	    return new double []{G.getF(p),G.getDF(p),G.getD2F(p)};
	}
	public Skewness getSkewnessAtPlast(final PnLdIDSk sd)
	{
	    return sk.getSkewnessAt(sd.p_last__with_p_offset_included);
	}


private void test_Self(final double eps)
{
    final Skewness st=sk.sknodi;
    final DerivativesCalculateable _pp[]=sk.getX1X2(),P1=_pp[0],P2=_pp[1];
    final double om1=P1.getD2F(sk.sknodi.x2), // d^2p1/dp^2|_{p2}
	om2=P2.getD2F(sk.sknodi.x1); // d^2p2/dp^2|_{p1}
    UnitTests.checkDoubleRel(P1.getF(st.x1),"P1(x1)",st.x1,"st.x1",eps);
    UnitTests.checkDoubleRel(P1.getF(st.x2),"P1(x2)",st.x1,"st.x1",eps);
    UnitTests.checkDoubleRel(P2.getF(st.x1),"P2(x1)",st.x2,"st.x2",eps);
    UnitTests.checkDoubleRel(P2.getF(st.x2),"P2(x2)",st.x2,"st.x2",eps);
    UnitTests.checkDoubleRel(P1.getDF(st.x2),"dp1/dp2",0,"0",eps);
    UnitTests.checkDoubleRel(P2.getDF(st.x1),"dp2/dp1",0,"0",eps);
    if(om1<0) throw new RuntimeException("om1="+om1);
    if(om2>0) throw new RuntimeException("om2="+om2);

    final double m0=st.w1+st.w2;
    UnitTests.checkDoubleRel(sk.di+m0,"sk.di+m0",ifuture,"ifuture",eps);
    if(sk.di/ifuture>eps)
    {
	// skewness limit at p=\infty is (m0-sk.di)/(m0+sk.di)
	if(Math.abs(st.w1+st.w2-sk.di)/ifuture>eps)
	UnitTests.checkDoubleRel(getDSkewnessAt(1/eps)[0],"limit",(st.w1+st.w2-sk.di)/ifuture,"(m0-sk.di)/ifuture",eps);
	// st.{x1,x2} (unperturbed) are skewness extemums, x1: max, x2: min, x1<x2.
	// max
	final double [] skmax=getDSkewnessAt(st.x1);
	UnitTests.checkDoubleRel(skmax[0],"Sk(p1)",(st.w1-st.w2+sk.di)/ifuture,"(st.w1-st.w2+sk.di)/ifuture",eps);
	UnitTests.checkDoubleRel(skmax[1],"d/dp_1 Sk",0,"0",eps);
	if(skmax[2]>0) throw new RuntimeException("st.x1+"+st.x1+" is not max");
	// min
	final double [] skmin=getDSkewnessAt(st.x2);
	UnitTests.checkDoubleRel(skmin[0],"Sk(p2)",(st.w1-st.w2-sk.di)/ifuture,"(st.w1-st.w2-sk.di)/ifuture",eps);
	UnitTests.checkDoubleRel(skmin[1],"d/dp_2 Sk",0,"0",eps);
	if(skmin[2]<0) throw new RuntimeException("st.x2+"+st.x2+" is not min");

	// midpoint p_mid=(p1+p2)/2 should be the same.
	UnitTests.checkDoubleRel(sk.getX1X2midpoint().getF(st.xa),"p_mid(xa)",sk.getX1X2midpoint().getF(st.x1),"p_mid(x1)",eps);
	UnitTests.checkDoubleRel(sk.getX1X2midpoint().getF(st.xa),"p_mid(xa)",sk.getX1X2midpoint().getF(st.x2),"p_mid(x2)",eps);
	// x2-x1 should be the same.
	//UnitTests.checkDoubleRel(sk.getX2minusX1().getF(st.xa),"x2x1(xa)",sk.getX2minusX1().getF(st.x1),"x2x1(x1)",eps);
	UnitTests.checkDoubleRel(sk.getX2minusX1().getF(st.x1),"x2x1(x1)",sk.getX2minusX1().getF(st.x2),"x2x1(x2)",eps);
    }
    System.err.println("PnLdIDSk:SKData:test_Self:p1="+st.x1+" D2P2(x1)="+sk.getX1X2()[1].getD2F(st.x1)+
		       "\tp2="+st.x2+" D2P1(x2)="+sk.getX1X2()[0].getD2F(st.x2));
} // END OF test_Self
private static void test_checkSKData(final SKData s,final double m0,final double m1,final double m2,final double m3,
				     final double di,final double p_last__with_p_offset_included,final double eps)
{
    UnitTests.checkDoubleRel(s.sk.getSkewnessAt(p_last__with_p_offset_included).sk,"Skewness(k)",s.getDSkewnessAt(p_last__with_p_offset_included)[0],"DSkewness[0]",eps);
    final double [] t[]=DiffSkewness.getNOM_SDENOM_FromMonomials(m0,m1,m2,m3,di);
    final double A=s.sk.M.B.getD0(t[6],p_last__with_p_offset_included),
	B=s.sk.M.B.getD0(t[7],p_last__with_p_offset_included),
	D=s.sk.M.B.getD0(t[8],p_last__with_p_offset_included),
	XA=s.sk.M.B.getD0(t[9],p_last__with_p_offset_included);
    final Skewness sl=s.sk.getSkewnessAt(p_last__with_p_offset_included);
    UnitTests.checkDoubleRel(sl.sk,"sl.sk",(-B-2*XA*D)/Math.sqrt(B*B-4*A*D),"(-B-2*XA*D)/Math.sqrt(B*B-4*A*D)",eps);
    UnitTests.checkDoubleRel(sl.xa,"sl.xa",XA,"XA",eps);
    UnitTests.checkDoubleRel(sl.x1,"sl.x1",(-B-Math.sqrt(B*B-4*A*D))/(2*D),"(-B-Math.sqrt(B*B-4*A*D))/(2*D)",eps);
    UnitTests.checkDoubleRel(sl.x2,"sl.x2",(-B+Math.sqrt(B*B-4*A*D))/(2*D),"(-B+Math.sqrt(B*B-4*A*D))/(2*D)",eps);
    UnitTests.checkDoubleRel(sl.w1,"sl.w1",(s.sk.sknodi.w1+s.sk.sknodi.w2+di)/(1+Math.pow(sl.x1-XA,2)/D),"w/(1+Math.pow(sl.x1-XA,2)/D)",eps);
    UnitTests.checkDoubleRel(sl.w2,"sl.w2",(s.sk.sknodi.w1+s.sk.sknodi.w2+di)/(1+Math.pow(sl.x2-XA,2)/D),"w/(1+Math.pow(sl.x2-XA,2)/D)",eps);

    final DerivativesCalculateable [] c=s.sk.getOrthogonalPolynomial2Coefs();
    UnitTests.checkDoubleRel(c[0].getF(p_last__with_p_offset_included),"c[0].getF()",sl.x1*sl.x2,"sl.x1*sl.x2",eps);
    UnitTests.checkDoubleRel(c[1].getF(p_last__with_p_offset_included),"c[1].getF()",-(sl.x1+sl.x2),"-(sl.x1+sl.x2)",eps);

    final double [] P2=s.sk.M.B.mult2Pol(new double []{-sl.x1,1},new double []{-sl.x2,1}),P22=s.sk.M.B.mult2Pol(P2,P2),PA2=s.sk.M.B.mult2Pol(new double []{-sl.xa,1},new double []{-sl.xa,1});
    final double [] ml={1,
			(sl.x1*sl.w1+sl.x2*sl.w2)/(sl.w1+sl.w2),
			(Math.pow(sl.x1,2)*sl.w1+Math.pow(sl.x2,2)*sl.w2)/(sl.w1+sl.w2),
			(Math.pow(sl.x1,3)*sl.w1+Math.pow(sl.x2,3)*sl.w2)/(sl.w1+sl.w2),
			0+(s.sk.di/s.ifuture)*Math.pow(p_last__with_p_offset_included,4)};
    UnitTests.checkDoubleRel(TMatr.mulVVn1(PA2,ml),"<PA2,ml>",s.sk.getL2Volatility().getF(p_last__with_p_offset_included),"L2",eps);
    UnitTests.checkDoubleRel(TMatr.mulVVn1(P2,ml),"<P2,ml>",0,"0",eps);
    UnitTests.checkDoubleRel(TMatr.mulVVn1(P22,ml),"<P22,ml>",s.sk.getL4Volatility().getF(p_last__with_p_offset_included),"L4",eps);
    //UnitTests.checkDoubleRel(0,"0",s.sk.getL4Volatility().getDF(sl.x1),"DL4|x1",eps);
    //UnitTests.checkDoubleRel(0,"0",s.sk.getL4Volatility().getDF(sl.x2),"DL4|x2",eps);

    UnitTests.checkDoubleRel((sl.x1+sl.x2)/2,"(sl.x1+sl.x2)/2",s.sk.getX1X2midpoint().getF(p_last__with_p_offset_included),"Midpoit",eps);
    UnitTests.checkDoubleRel(sl.xa,"sl.xa",s.sk.getXaverage().getF(p_last__with_p_offset_included),"Average",eps);
    UnitTests.checkDoubleRel((s.sk.di/s.ifuture),"Dsl.xa/DPf",s.sk.getXaverage().getDF(p_last__with_p_offset_included),"DAverage",eps);
    System.err.println("PnLdIDSk:SKData:test_checkSKData: test OK eps="+eps);
} // END OF test_checkSKData
} // END OF class SKData



public final EVXData I;
public final double p_offset,p_last__with_p_offset_included,p_IH,I_0_feq,dI,Pf_from_ptpi,deg_from_pt_true_pi,Pf_from_pt_true_pi,Pf_from_dyn,deg_from_dyn,Pf_from_f0,delta_Pdyn,delta_P_IHf,Pf_from_xx;
public final SKData SKs_spur,SKs_W0pi,SKs_W0truepi,SKs_W0pt,SKs[];
public final Skewness SK_movaver__nodI,SK_spur__nodI;
PnLdIDSk(final EVXData I,final double p_offset,final double p_last__with_p_offset_included,
	 final double p_IH,final double I_0_feq,
	 final Skewness SK_movaver__nodI,final Skewness SK_spur__nodI,
	 final SKData SKs_W0truepi,final SKData SKs_W0B,
	 final double [] psik_psi0,
	 final double [] psik_I_psik,final double [] psik_pI_psik,
	 final double [] psik_p2I_psik,final double [] psik_p3I_psik,final double [] bbasis,
	 final double Pf_from_xx,
	 final double delta_Pdyn,final double delta_P_IHf,
	 final double eps)
{
    this.I=I;
    this.p_offset=p_offset;
    this.p_last__with_p_offset_included=p_last__with_p_offset_included;
    this.p_IH=p_IH;
    this.I_0_feq=I_0_feq;
    this.SK_movaver__nodI=SK_movaver__nodI;
    this.SK_spur__nodI=SK_spur__nodI;
    final int n=I.neq;
    this.dI=I_0_feq-I.s0;
    this.SKs_W0truepi=SKs_W0truepi;
    this.Pf_from_xx=Pf_from_xx;
    this.delta_Pdyn=delta_Pdyn;
    this.delta_P_IHf=delta_P_IHf;

    final double [] w0k=TMatr.squareArrayElements(psik_psi0);
    this.SKs_spur=new SKData(DiffSkewness.getDiffSkewness_FromMonomials(
				 TMatr.sumElements(psik_I_psik),
				 TMatr.sumElements(psik_pI_psik),
				 TMatr.sumElements(psik_p2I_psik),
				 TMatr.sumElements(psik_p3I_psik),
				 dI),TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),eps);
    this.SKs_W0pi=new SKData(DiffSkewness.getDiffSkewness_FromMonomials(
				 TMatr.mulVV(w0k,psik_I_psik),
				 TMatr.mulVV(w0k,psik_pI_psik),
				 TMatr.mulVV(w0k,psik_p2I_psik),
				 TMatr.mulVV(w0k,psik_p3I_psik),
				 dI*TMatr.mulVV(w0k,w0k)),I_0_feq,eps);
    final double [] wIHIk=new double [n];
    for(int k=0;k<n;k++){wIHIk[k]=I_0_feq/bbasis[k];}
    this.SKs_W0pt=new SKData(DiffSkewness.getDiffSkewness_FromMonomials(
				 TMatr.mulVVV(w0k,wIHIk,psik_I_psik),
				 TMatr.mulVVV(w0k,wIHIk,psik_pI_psik),
				 TMatr.mulVVV(w0k,wIHIk,psik_p2I_psik),
				 TMatr.mulVVV(w0k,wIHIk,psik_p3I_psik),
				 dI*TMatr.mulV2V(w0k,wIHIk)),I_0_feq,eps);
    // For unit tests, to match DynVar2DIDP calculations.
    final double deg_from_ptpi=TMatr.mulVV(w0k,w0k)-TMatr.mulV2V(w0k,wIHIk); // must be >0
    this.Pf_from_ptpi=(TMatr.mulVVV(w0k,wIHIk,psik_pI_psik)-TMatr.mulVV(w0k,psik_pI_psik))/(dI*deg_from_ptpi);

    // The criteria of future directional information available, no info when deg_from_pt_true_pi=0.
    this.deg_from_pt_true_pi=1-TMatr.mulV2V(w0k,wIHIk);
    // Equilibrium price.
    final double pI_0=(SKs_W0truepi.sk.sknodi.xa*(SKs_W0truepi.ifuture-SKs_W0truepi.sk.di));
    this.Pf_from_pt_true_pi=(TMatr.mulVVV(w0k,wIHIk,psik_pI_psik)-pI_0)/(dI*deg_from_pt_true_pi);
    this.deg_from_dyn=1-TMatr.mulVV(w0k,w0k);
    Pf_from_dyn=(TMatr.mulVV(w0k,psik_pI_psik)-(SKs_W0B.sk.sknodi.xa*I.s0))/(dI*deg_from_dyn);
    Pf_from_f0=(TMatr.mulVV(w0k,psik_pI_psik)-pI_0)/(dI*deg_from_dyn);
    if(eps>0)
    {
	// check the norm
	UnitTests.checkDoubleRel((TMatr.mulVVV(w0k,wIHIk,psik_I_psik)-TMatr.mulVV(w0k,psik_I_psik))/(dI*deg_from_ptpi),"1",1,"1",eps);
	UnitTests.checkDoubleRel((TMatr.mulVVV(w0k,wIHIk,psik_I_psik)-I.s0)/(dI*deg_from_pt_true_pi),"1",1,"1",eps);
	UnitTests.checkDoubleRel(SKs_W0truepi.sk.sknodi.w1+SKs_W0truepi.sk.sknodi.w2,"I_0",I.s0,"I.s0",eps);
	// check i- and t- averages to match
	UnitTests.checkDoubleRel(SKs_W0pi.sk.getSkewnessAt(Pf_from_ptpi).xa,"SKs_W0pi xa",SKs_W0pt.sk.getSkewnessAt(Pf_from_ptpi).xa,"SKs_W0pt xa",eps);
	UnitTests.checkDoubleRel(SKs_W0truepi.sk.getSkewnessAt(Pf_from_pt_true_pi).xa,"SKs_W0truepi xa",SKs_W0pt.sk.getSkewnessAt(Pf_from_pt_true_pi).xa,"SKs_W0pt xa",eps);
    }
    this.SKs=new SKData [n];
    for(int k=0;k<n;k++)
    {
	SKs[k]=new SKData(DiffSkewness.getDiffSkewness_FromMonomials(psik_I_psik[k],psik_pI_psik[k],psik_p2I_psik[k],psik_p3I_psik[k],dI*w0k[k]),bbasis[k],eps);
	if(eps>0) SKData.test_checkSKData(SKs[k],psik_I_psik[k],psik_pI_psik[k],psik_p2I_psik[k],psik_p3I_psik[k],SKs[k].sk.di,p_last__with_p_offset_included,eps);
    }

    if(eps>0)
    {
	// spur invariance:
	UnitTests.checkDoubleRel(TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),"sum(ev bbasis)",TMatr.sumElements(I.getEValues())+dI,"sum(I.ev)+dI",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(psik_I_psik),"sum(psik_I_psik)",TMatr.sumElements(I.getEValues()),"sum(I.ev)",eps);
	// check I_0_feq
	UnitTests.checkDoubleRel(TMatr.mulV2V(psik_psi0,java.util.Arrays.copyOf(bbasis,n)),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(w0k,psik_I_psik)+dI*TMatr.mulVV(w0k,w0k),"sum(<psik|psi0>^2*I_k+<psik|psi0>^4*dI)",I_0_feq,"I_0_feq",eps);
	System.err.println("PnLdIDSk.PnLdIDSk selftest2 deg_from_pt_true_pi="+deg_from_pt_true_pi+">0 OK n="+n+" eps="+eps);
    }
}

/**
 * Solve {@literal <psi|I|psi>/<psi|psi>}=lambda extremum problem, find eigenstates &amp; eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-{@literal <psi0|I|psi0>} (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + {@literal |psi0>dI<psi0|}
 * Find the eigenstates of this new I^{f} matrix,
 * in dynamic impact approximation its eigenstates should be also the P^{f} operator eigenstates.
 *
 * For the moments of |psik> state calculate the price, manipulating with the skewness.
 */
public static PnLdIDSk get_PnLdIDSk(final MomentsData mm,
				    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				    final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final int ixH=n-1; // the state with maximal I
    final double I_0=I.s0,I_0_feq=I.ev[ixH],dI=I_0_feq-I_0;

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o));
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];

    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev),
	bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,mm.QQp2i,I.ev),
	bQQp3i=EVXData.convertMatrFromQtoB_Basis(n,mm.QQp3i,I.ev),
	bQQdp=EVXData.convertMatrFromQtoB_Basis(n,mm.QQdp,I.ev);

    // Diagonalize "future" I:
    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true),
	psik_psi0=new double [n],psik_I_psik=new double[n],psik_pI_psik=new double[n],
	psik_p2I_psik=new double[n],psik_p3I_psik=new double[n],psik_dp_psik=new double[n];
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,bbasis,n);

	psik_psi0[k]=TMatr.mulVV(psik,bpsi0);
	psik_I_psik[k]=TMatr.mulV2V(psik,I.getEValues());
	psik_pI_psik[k]=TMatr.getAMatrA(n,bbasis,n+k*n,bQQpi,n);
	psik_p2I_psik[k]=TMatr.getAMatrA(n,bbasis,n+k*n,bQQp2i,n);
	psik_p3I_psik[k]=TMatr.getAMatrA(n,bbasis,n+k*n,bQQp3i,n);
	psik_dp_psik[k]=TMatr.getAMatrA(n,bbasis,n+k*n,bQQdp,n);
    }

    if(eps>0)
    {
	EVXData.test_orthogonality("PnLdIDSk.get_PnLdIDSk:I.ev",n,I.ev,mm.QQi,mm.QQt,eps); // old basis
	EVXData.test_orthogonality("PnLdIDSk.get_PnLdIDSk:bbasis",n,bbasis,bQQi_feq,TMatr.getUnitMatrix(n),eps); // new basis
	// check some matrix elements
	UnitTests.checkDoubleRel(TMatr.mulVV(psik_psi0,psik_psi0),"<psi0|psi0>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(psik_psi0,java.util.Arrays.copyOf(bbasis,n)),"<psi0|If|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(bpsi0,bpsi0),"<psi0|psi0>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(bpsi0,I.getEValues()),"<psi0|I|psi0>",I.s0,"I.s0",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|I^f|psi0>",I_0_feq,"I_0_feq",eps);
	// check spur invariance
	UnitTests.checkDoubleRel(TMatr.sumElements(I.getEValues()),"sum(I.ev)",TMatr.sumElements(psik_I_psik),"sum(psik_I_psik)",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(I.getEValues())+dI,"sum(I.ev)+dI",TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),"sum(I^{f})",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQpi),"Spur(bQQpi)",TMatr.sumElements(psik_pI_psik),"sum(psik_pI_psik)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQp2i),"Spur(bQQp2i)",TMatr.sumElements(psik_p2I_psik),"sum(psik_p2I_psik)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQp3i),"Spur(bQQp3i)",TMatr.sumElements(psik_p3I_psik),"sum(psik_p3I_psik)",eps);
	System.err.println("PnLdIDSk.get_PnLdIDSk selftest OK n="+n+" eps="+eps);
    }

    final double [] w0B=TMatr.scaleVectorInPlace(TMatr.squareArrayElements(I.psik_at_x0),1/I.o_o),
	w0F=TMatr.squareArrayElements(psik_psi0);
    final int inds_w0F[]=com.polytechnik.utils.SortArrays.sortArrayIndexes(w0F),i0_e=inds_w0F[n-1],i0_p=inds_w0F[n-2];
    return new PnLdIDSk(I,mm.p_offset,mm.p_last,
			bQQpi[ixH+n*ixH]/I.ev[ixH],I_0_feq,
			Skewness.getSkewnessFromMonomialMoments(mm.QQi[0]/mm.QQt[0],mm.QQpi[0]/mm.QQt[0],mm.QQp2i[0]/mm.QQt[0],mm.QQp3i[0]/mm.QQt[0]),
			Skewness.getSkewnessFromMonomialMoments(TMatr.sumElements(I.getEValues()),TMatr.getSpurM(n,bQQpi),TMatr.getSpurM(n,bQQp2i),TMatr.getSpurM(n,bQQp3i)),
			new SKData(DiffSkewness.getDiffSkewness_FromMonomials(
				       TMatr.mulV2V(bpsi0,I.getEValues()),
				       TMatr.getAMatrA(bpsi0,bQQpi,n),
				       TMatr.getAMatrA(bpsi0,bQQp2i,n),
				       TMatr.getAMatrA(bpsi0,bQQp3i,n),
				       dI),I_0_feq,eps),
			new SKData(DiffSkewness.getDiffSkewness_FromMonomials(
				       TMatr.mulVV(w0B,I.getEValues()),
				       TMatr.mulVV(w0B,TMatr.getMatrixDiagElements(n,bQQpi)),
				       TMatr.mulVV(w0B,TMatr.getMatrixDiagElements(n,bQQp2i)),
				       TMatr.mulVV(w0B,TMatr.getMatrixDiagElements(n,bQQp3i)),
				       dI),I_0_feq,eps),
			psik_psi0,psik_I_psik,psik_pI_psik,psik_p2I_psik,psik_p3I_psik,bbasis,
			-TMatr.getAMatrB(n,bbasis,n+i0_e*n,bbasis,n+i0_p*n,bQQpi,n)/(psik_psi0[i0_e]*psik_psi0[i0_p]*dI),
			(TMatr.mulVV(w0F,psik_dp_psik)-TMatr.getAMatrA(bpsi0,bQQdp,n))/((1-TMatr.mulVV(w0F,w0F))*I.o_o),
			-psik_dp_psik[ixH]/(w0F[ixH]*I.o_o),
			eps);
}

static void matchDynVar2DIDP(final PnLdIDSk a,final DynVar2DIDP b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.SKs_spur.sk.getSkewnessAt(a.p_last__with_p_offset_included).xa,"spur(xa)",b.Pf_SPUR+p_offset_shift,"b.Pf_SPUR+p_offset_shift",eps);
     UnitTests.checkDoubleRel(a.Pf_from_ptpi,"a.Pf_from_ptpi",b.Pf_from_ptpi+p_offset_shift,"b.Pf_from_ptpi+p_offset_shift",eps);
     UnitTests.checkDoubleRel(a.Pf_from_pt_true_pi,"a.Pf_from_pt_true_pi",b.Pf_from_pt_true_pi+p_offset_shift,"b.Pf_from_pt_true_pi+p_offset_shift",eps);
     UnitTests.checkDoubleRel(a.Pf_from_f0,"a.Pf_from_f0",b.Pf_from_f0+p_offset_shift,"b.Pf_from_f0+p_offset_shift",eps);
     UnitTests.checkDoubleRel(a.Pf_from_dyn,"a.Pf_from_dyn",b.Pf_from_dyn+p_offset_shift,"b.Pf_from_dyn+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.delta_Pdyn,"a.delta_Pdyn",b.delta_Pdyn,"b.delta_Pdyn",eps);
    System.err.println("PnLdIDSk.matchDynVar2DIDP test OK n="+a.I.neq+" eps="+eps);
}


static void matchPnLdIpx(final PnLdIDSk a,final PnLdIpx b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.Pf_from_pt_true_pi,"a.Pf_from_pt_true_pi",b.Pft_from_f0+p_offset_shift,"b.Pft_from_f0+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.deg_from_pt_true_pi,"a.deg_from_pt_true_pi",b.degt_from_dyn,"b.degt_from_dyn",eps);
    UnitTests.checkDoubleRel(a.Pf_from_f0,"a.Pf_from_f0",b.Pf_from_f0+p_offset_shift,"b.Pf_from_f0+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.Pf_from_xx,"a.Pf_from_xx",b.Pf_from_xx+p_offset_shift,"b.Pf_from_xx+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.deg_from_dyn,"a.deg_from_dyn",b.deg_from_dyn,"b.deg_from_dyn",eps);
    UnitTests.checkDoubleRel(a.delta_Pdyn,"a.delta_Pdyn",b.delta_Pdyn_from_dyn,"b.delta_Pdyn_from_dyn",eps);
    UnitTests.checkDoubleRel(a.delta_P_IHf,"a.delta_P_IHf",b.delta_P_IHf,"b.delta_P_IHf",eps);
    System.err.println("PnLdIDSk.matchPnLdIpx test OK n="+a.I.neq+" eps="+eps);
}

static void matchPnLdISpur(final PnLdIDSk a,final PnLdISpur b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.Pf_from_pt_true_pi,"a.Pf_from_pt_true_pi",b.Pft_from_f0+p_offset_shift,"b.Pft_from_f0+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.deg_from_pt_true_pi,"a.deg_from_pt_true_pi",b.degt,"b.degt",eps);
    UnitTests.checkDoubleRel(a.Pf_from_f0,"a.Pf_from_f0",b.Pf_from_f0+p_offset_shift,"b.Pf_from_f0+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.deg_from_dyn,"a.deg_from_dyn",b.deg,"b.deg",eps);
    UnitTests.checkDoubleRel(a.Pf_from_xx,"a.Pf_from_xx",b.Pf_from_xx+p_offset_shift,"b.Pf_from_xx+p_offset_shift",eps);

    System.err.println("PnLdIDSk.matchPnLdISpur test OK n="+a.I.neq+" eps="+eps);
}


static void matchMicroPnLdI(final PnLdIDSk a,final MicroPnLdI b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);
    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.g_IH+p_offset_shift,"b.g_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.SKs_W0truepi.sk.sknodi.xa,"a.SKs_W0truepi.sk.sknodi.xa",b.g_0+p_offset_shift,"b.g_0+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.Pf_from_pt_true_pi,"a.Pf_from_pt_true_pi",b.gf_from_pt_true_pi+p_offset_shift,"b.gf_from_pt_true_pi+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.deg_from_pt_true_pi,"a.deg_from_pt_true_pi",b.deg_from_pt_true_pi,"b.deg_from_pt_true_pi",eps);
    UnitTests.checkDoubleRel(a.Pf_from_f0,"a.Pf_from_f0",b.gf_from_f0+p_offset_shift,"b.gf_from_f0+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.deg_from_dyn,"a.deg_from_dyn",b.deg_from_dyn,"b.deg_from_dyn",eps);

    System.err.println("PnLdIDSk.matchMicroPnLdI test OK n="+a.I.neq+" eps="+eps);
}
}
