package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLdIDyn.java,v 1.22 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLdIDyn {

    public final EVXData I;
    public final double p_offset,p_IH,I_0_feq,P_movaver,P_movaver_spur,p_E_IHf,p_E_0,P1_IHf,P2_IHf;
    PnLdIDyn(final EVXData I,final double p_offset,final double P_movaver,final double P_movaver_spur,
	     final double p_IH,final double I_0_feq,
	     final double p_E_IHf,final double p_E_0,
	     final double P1_IHf,final double P2_IHf)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.P_movaver=P_movaver;
	this.P_movaver_spur=P_movaver_spur;
	this.p_IH=p_IH;
	this.I_0_feq=I_0_feq;
	this.p_E_IHf=p_E_IHf;
	this.p_E_0=p_E_0;
	this.P1_IHf=P1_IHf;
	this.P2_IHf=P2_IHf;
    }

public static final int IMPACTAPPROXIMATION_NONE=561,
	IMPACTAPPROXIMATION_DYNAMIC_IH=562,
	IMPACTAPPROXIMATION_DYNAMIC_PURE=563;

/**
 * Solve {@literal <psi|I|psi>/<psi|psi>}=lambda extremum problem, find eigenstates &amp; eigenvalues.
 * find psi^[IH] and lambda^[IH]=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^[IH]-{@literal <psi0|I|psi0>} (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + {@literal |psi0>dI<psi0|}
 * Find the eigenstates of this new I^{f} matrix.
 * Then consider the pI operator.
 * In dynamic impact approximation in has to be modified as
 * 1. Pure Dynamic Impact: Diagonal in  {@literal |psi^[i]>} basis.
 * 2. Dynamic Impact: only {@literal <psi^[IH]|pI|psi^[i]>}=0 for i!=IH
 * 3. None: the pI is not modified.
 * Then variate obtained {@literal <psi^[IH]|pI|psi^[IH]>}
 * for {@literal |dpsi>} one may use {@literal |psi0>} or {@literal |psif>}, corresponding to max I^{f}.
 * Try different options.
 */
public static PnLdIDyn get_PnLdIDyn(final MomentsData mm,
			   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			   final double eps,
			   final int flag_IMPACTAPPROXIMATION)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final int ixH=n-1; // the state with maximal I

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o));
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];

    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev);
    final double pI_IH=bQQpi[ixH+n*ixH];
    if(flag_IMPACTAPPROXIMATION==IMPACTAPPROXIMATION_NONE){}
    else if(flag_IMPACTAPPROXIMATION==IMPACTAPPROXIMATION_DYNAMIC_PURE)
    {
	for(int k=0;k<n;k++) for(int l=0;l<n;l++) if(k!=l) bQQpi[k+n*l]=0;
    }
    else if(flag_IMPACTAPPROXIMATION==IMPACTAPPROXIMATION_DYNAMIC_IH)
    {
	for(int k=ixH,l=0;l<n;l++) if(k!=l) bQQpi[k+n*l]=bQQpi[l+n*k]=0;
    }
    else throw new RuntimeException("flag_IMPACTAPPROXIMATION="+flag_IMPACTAPPROXIMATION+" is not known");
    // Diagonalize "future" I:
    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true),
	bpsi_IHf=EVXData.getPsiFromBasis(ixH,bbasis,n),
	bpsi_IHf_projO0=TMatr.sumAcB(bpsi_IHf,-TMatr.mulVV(bpsi_IHf,bpsi0),bpsi0),
	bpsi_IH=new double [n];
    bpsi_IH[ixH]=1;
    final double pI_IHf=TMatr.getAMatrA(bpsi_IHf,bQQpi,n),I_IHf=TMatr.mulV2V(bpsi_IHf,I.getEValues()),
	pI_0=TMatr.getAMatrA(bpsi0,bQQpi,n),
	p_E_IHf=(pI_IHf-pI_IH)/(I_IHf-I.ev[ixH]),
	p_E_0=(pI_0-pI_IH)/(I.s0-I.ev[ixH]);
    if(eps>0)
    {
	EVXData.test_orthogonality("PnLdIDyn.get_PnLdIDyn:I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	// Check spur invariance
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQpi),"Spur(IQQt,QQpi)",TMatr.getSpurM(n,bQQpi),"Spur(bQQpi)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQi),"Spur(IQQt,QQi)",TMatr.sumElements(I.getEValues()),"sum(I.ev)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQi_feq),"Spur(bQQi_feq)",TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),"sum(ev bbasis)",eps);
	// Check few known matrix elements.
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|I^f|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi_IHf,bQQi_feq,n),"<psi_IHf|I^f|psi_IHf>",bbasis[ixH],"max(I^f)",eps);
	// Check the result match the one obtained from variations.
	final double [] O=new double [n*n],Oi=new double [n*n];
	for(int k=0;k<n;k++) {O[k+n*k]=1;Oi[k+n*k]=I.ev[k];}
	if(true){
	final double []
	    rqi=RayleighQuotient.getRQVariationValues(bpsi_IH,bpsi_IHf,null,Oi,O,n),
	    rqpi=RayleighQuotient.getRQVariationValues(bpsi_IH,bpsi_IHf,null,bQQpi,O,n);
	UnitTests.checkDoubleRel(rqi[0],"rqi[0]",I.ev[ixH],"I.ev[ixH]",eps);
	UnitTests.checkDoubleRel(rqi[1],"rqi[1]",0,"0",eps);
	UnitTests.checkDoubleRel(0.5*rqi[2],"0.5*rqi[2]",I_IHf-I.ev[ixH],"I_IHf-I.ev[ixH]",eps);
	UnitTests.checkDoubleRel(rqpi[0],"rqpi[0]",pI_IH,"pI_IH",eps);
	if(flag_IMPACTAPPROXIMATION!=IMPACTAPPROXIMATION_NONE){
	    UnitTests.checkDoubleRel(rqpi[1],"rqpi[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(0.5*rqpi[2],"0.5*rqpi[2]",pI_IHf-pI_IH,"pI_IHf-pI_IH",eps);
	}
	}
	if(true){
	final double []
	    rqi=RayleighQuotient.getRQVariationValues(bpsi_IH,bpsi0,null,Oi,O,n),
	    rqpi=RayleighQuotient.getRQVariationValues(bpsi_IH,bpsi0,null,bQQpi,O,n);
	UnitTests.checkDoubleRel(rqi[0],"rqi[0]",I.ev[ixH],"I.ev[ixH]",eps);
	UnitTests.checkDoubleRel(rqi[1],"rqi[1]",0,"0",eps);
	UnitTests.checkDoubleRel(0.5*rqi[2],"0.5*rqi[2]",I.s0-I.ev[ixH],"I.s0-I.ev[ixH]",eps);
	UnitTests.checkDoubleRel(rqpi[0],"rqpi[0]",pI_IH,"pI_IH",eps);
	if(flag_IMPACTAPPROXIMATION!=IMPACTAPPROXIMATION_NONE){
	    UnitTests.checkDoubleRel(rqpi[1],"rqpi[1]",0,"0",eps);
	    UnitTests.checkDoubleRel(0.5*rqpi[2],"0.5*rqpi[2]",pI_0-pI_IH,"pI_0-pI_IH",eps);
	}
	}
	UnitTests.checkDoubleRel(TMatr.mulVV(bpsi_IHf_projO0,bpsi0),"<psi_IHf_projO0|psi0>",0,"0",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(bpsi_IHf_projO0,bpsi_IHf),"<psi_IHf_projO0|bpsi_IHf>",TMatr.mulVV(bpsi_IHf_projO0,bpsi_IHf_projO0),"<psi_IHf_projO0|psi_IHf_projO0>",eps);
	System.err.println("PnLdIDyn.get_PnLdIDyn selftest OK n="+n+" eps="+eps);
    }

    return new PnLdIDyn(I,mm.p_offset,
		    mm.QQpi[0]/mm.QQi[0],TMatr.getSpurM(n,bQQpi)/TMatr.sumElements(I.getEValues()),
			bQQpi[ixH+n*ixH]/I.ev[ixH],I_0_feq,p_E_IHf,p_E_0,
			TMatr.getAMatrB(bpsi_IHf_projO0,bpsi_IHf,bQQpi,n)/
			TMatr.mulVVV(bpsi_IHf_projO0,bpsi_IHf,I.getEValues()),
			TMatr.getAMatrA(bpsi_IHf_projO0,bQQpi,n)/
			TMatr.mulV2V(bpsi_IHf_projO0,I.getEValues())
			);
}




}
