package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLdIdPRQ.java,v 1.29 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLdIdPRQ {

    public final EVXData I;
    public final double p_offset,p_IH,p_IHf,I_0_feq,P_movaver,P_movaver_spur,Pf_covolLow,Pf_covolHigh;
    PnLdIdPRQ(final EVXData I,final double p_offset,final double p_last__with_p_offset_included,
	      final double p_IH,final double p_IHf,final double I_0_feq,final double P_movaver,final double P_movaver_spur,final double Pf_covolLow,final double Pf_covolHigh)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.p_IHf=p_IHf;
	this.I_0_feq=I_0_feq;
	this.P_movaver=P_movaver;
	this.P_movaver_spur=P_movaver_spur;
	this.Pf_covolLow=Pf_covolLow;
	this.Pf_covolHigh=Pf_covolHigh;
    }

/**
 * Solve {@literal <psi|I|psi>/<psi|psi>}=lambda extremum problem, find eigenstates &amp; eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-{@literal <psi0|I|psi0>} (contribution from the future, of not yet executed, trades).
 * Then operator I^{f} (future I) is: I^{f}= I + {@literal |psi0>dI<psi0|}
 * Find the eigenstates of this new I^{f} matrix,
 * and the state psif^{IH}, corresponding to the max of future I.
 * Then consider co-volatility:
 * {@literal <}psif^{IH}|(p-Pf_covol)^2 I^{f}|psif^{IH}{@literal >}/{@literal <}psi^{IH}|(p-Pf_covol)^2 I|psi^{IH}{@literal >}
 * to find Pf_covol, corresponding to min/max.
 */
public static PnLdIdPRQ get_PnLdIdPRQ(final MomentsData mm,
				  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				  final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final int ixH=n-1; // the state with maximal I
    final double I_0=I.s0,I_0_feq=I.ev[ixH],dI=I_0_feq-I_0;

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o));
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];

    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev),
	bQQp2i=EVXData.convertMatrFromQtoB_Basis(n,mm.QQp2i,I.ev);

    // Diagonalize "future" I:
    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true),
	bpsif_IH=EVXData.getPsiFromBasis(ixH,bbasis,n);
    final double pfl=mm.p_last,dI_proj=dI*Math.pow(TMatr.mulVV(bpsi0,bpsif_IH),2),
	md0=I.ev[ixH],md1=bQQpi[ixH+n*ixH],md2=bQQp2i[ixH+n*ixH],
	mn0=bbasis[ixH],mn1=(TMatr.getAMatrA(bpsif_IH,bQQpi,n)+pfl*dI_proj),
	mn2=TMatr.getAMatrA(bpsif_IH,bQQp2i,n)+pfl*pfl*dI_proj;
    final double [] evcovol=M.EV.getGEV_Generalized(2,new double []{mn0,mn1,mn1,mn2},new double []{md0,md1,md1,md2},true);
    final double Pf_covolLow=-evcovol[2+0*2+0]/evcovol[2+0*2+1],Pf_covolHigh=-evcovol[2+1*2+0]/evcovol[2+1*2+1];

    if(eps>0)
    {
	EVXData.test_orthogonality("PnLdIdP.get_PnLdIdP:I.ev",n,I.ev,mm.QQi,mm.QQt,eps); // old basis
	EVXData.test_orthogonality("PnLdIdP.get_PnLdIdP:bbasis",n,bbasis,bQQi_feq,TMatr.getUnitMatrix(n),eps); // new basis

	// Check spur invariance.
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQi),"Spur(IQQt,QQi)",TMatr.sumElements(I.getEValues()),"sum(I.ev)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQpi),"Spur(IQQt,QQpi)",TMatr.getSpurM(n,bQQpi),"Spur(bQQpi)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQp2i),"Spur(IQQt,QQp2i)",TMatr.getSpurM(n,bQQp2i),"Spur(bQQp2i)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(n,bQQi_feq),"Spur(bQQi_feq)",TMatr.sumElements(java.util.Arrays.copyOf(bbasis,n)),"sum(ev bbasis)",eps);

	final double [] psif_IH=EVXData.convertPsiFromBtoQ_Basis(n,bpsif_IH,I.ev),
	    QQ0=TMatr.getQQaa(M.B.getX0Moments(n,I.x0)),
	    QQi_feq=TMatr.sumAcB(mm.QQi,dI/I.o_o,QQ0),
	    QQpi_feq=TMatr.sumAcB(mm.QQpi,pfl*dI/I.o_o,QQ0),
	    QQp2i_feq=TMatr.sumAcB(mm.QQp2i,pfl*pfl*dI/I.o_o,QQ0),
	    QQcovol_feq=new double [n*n],QQcovol=new double [n*n];
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psif_IH,QQi_feq,n),"<psif|QQi_feq|psif>",bbasis[ixH],"bbasis[ixH]",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsif_IH,bQQi_feq,n),"<bpsif|bQQi_feq|bpsif>",bbasis[ixH],"bbasis[ixH]",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(I.psi0,QQi_feq,n),"<psi0|QQi_feq|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<bpsi0|bQQi_feq|bpsi0>",I_0_feq,"I_0_feq",eps);

	UnitTests.checkDoubleRel(mn0,"mn0",TMatr.getAMatrA(psif_IH,QQi_feq,n),"<psif|QQi_feq|psif>",eps);
	UnitTests.checkDoubleRel(mn1,"mn1",TMatr.getAMatrA(psif_IH,QQpi_feq,n),"<psif|QQpi_feq|psif>",eps);
	UnitTests.checkDoubleRel(mn2,"mn2",TMatr.getAMatrA(psif_IH,QQp2i_feq,n),"<psif|QQp2i_feq|psif>",eps);
	UnitTests.checkDoubleRel(md0,"md0",TMatr.getAMatrA(I.psiH,mm.QQi,n),"<psiH|QQi|psiH>",eps);
	UnitTests.checkDoubleRel(md1,"md1",TMatr.getAMatrA(I.psiH,mm.QQpi,n),"<psiH|QQpi|psiH>",eps);
	UnitTests.checkDoubleRel(md2,"md2",TMatr.getAMatrA(I.psiH,mm.QQp2i,n),"<psiH|QQp2i|psiH>",eps);

	for(int j=0;j<2;j++)
	{
	    final double p=j<=0?Pf_covolLow:Pf_covolHigh;
	    for(int k=0;k<QQcovol.length;k++)
	    {
		QQcovol_feq[k]=QQp2i_feq[k]+p*(-2*QQpi_feq[k]+p*QQi_feq[k]);
		QQcovol[k]=mm.QQp2i[k]+p*(-2*mm.QQpi[k]+p*mm.QQi[k]);
	    }
	    final double lambda=TMatr.getAMatrA(psif_IH,QQcovol_feq,n)/TMatr.getAMatrA(I.psiH,QQcovol,n);
	    UnitTests.checkDoubleRel(lambda,"lambda",evcovol[j],"evcovol[j]",eps);
	}
	System.err.println("PnLdIdPRQ.get_PnLdIdPRQ selftest OK n="+n+" eps="+eps);

    }
    return new PnLdIdPRQ(I,mm.p_offset,mm.p_last,
			 md1/md0,mn1/mn0,I_0_feq,
			 mm.QQpi[0]/mm.QQi[0],TMatr.getSpurM(n,bQQpi)/TMatr.sumElements(I.getEValues()),
			 Pf_covolLow,Pf_covolHigh);
}


}
