package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLdIfdyn.java,v 1.23 2022/08/07 15:44:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLdIfdyn {

    public final EVXData I;
    public final double p_offset,p_IH,p_IHf,I_0_feq,Pf,A_Pb,Pa;
    PnLdIfdyn(final EVXData I,final double p_offset,final double p_IH,final double p_IHf,final double I_0_feq,
	    final double [] psik_psi0,final double [] psik_I_psi0,final double [] psik_pI_psi0,final double [] psik_I_psik,final double [] psik_pI_psik)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.p_IHf=p_IHf;
	this.I_0_feq=I_0_feq;
	final double dI=I_0_feq-I.s0;
	final int n=I.neq;
	double sPfdI=0,sPb=0;
	for(int k=0;k<n;k++)
	{
	    final double psik_psi0_sq=Math.pow(psik_psi0[k],2);
	    sPfdI+=(psik_pI_psik[k]*psik_psi0_sq-(psik_pI_psi0[k]*psik_psi0[k]))/(1-psik_psi0_sq);
	    final double pk=psik_pI_psik[k]/psik_I_psik[k];
	    sPb+=pk*psik_psi0_sq;
	}
	this.Pf=sPfdI/dI;
	this.A_Pb=sPb;
	this.Pa=TMatr.mulV2V(psik_psi0,psik_pI_psik)/TMatr.mulV2V(psik_psi0,psik_I_psik);
    }

/**
 * Similar to PnLdIf but assume pI &amp; I are diagonal simultaneously.
 */
public static PnLdIfdyn get_PnLdIfdyn(final MomentsData mm,
				      final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				      final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final int ixH=n-1; // the state with maximal I

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o));
    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev),ev_pI=new double [n];
    for(int k=0;k<n;k++){
	bQQi_feq[k+n*k]+=I.ev[k];
	ev_pI[k]=bQQpi[k+n*k];
    }

    // Diagonalize "future" I:
    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true),
	psik_psi0=new double [n],psik_I_psi0=new double [n],psik_pI_psi0=new double [n],
	psik_I_psik=new double[n],psik_pI_psik=new double[n];
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,bbasis,n);

	psik_psi0[k]=TMatr.mulVV(psik,bpsi0);
	psik_I_psi0[k]=TMatr.mulVVV(psik,bpsi0,I.getEValues());
	psik_pI_psi0[k]=TMatr.mulVVV(psik,bpsi0,java.util.Arrays.copyOf(ev_pI,n));
	psik_I_psik[k]=TMatr.mulV2V(psik,I.getEValues());
	psik_pI_psik[k]=TMatr.mulV2V(psik,java.util.Arrays.copyOf(ev_pI,n));
    }
    final double [] bpsiH=new double [n];
    bpsiH[ixH]=1;
    final double [] bpsiHf=EVXData.getPsiFromBasis(ixH,bbasis,n),
	bpsiHfProj=TMatr.sumAcB(bpsiHf,-TMatr.mulVV(bpsiHf,bpsiH),bpsiH);
    final double p_IHf=TMatr.mulV2V(bpsiHfProj,java.util.Arrays.copyOf(ev_pI,n))/TMatr.mulV2V(bpsiHfProj,I.getEValues());
    if(eps>0)
    {
	EVXData.test_orthogonality("PnLdIfdyn.get_PnLdIfdyn:I.ev",n,I.ev,mm.QQi,mm.QQt,eps);
	// Check spur invariance
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,mm.QQt);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQpi),"Spur(IQQt,QQpi)",TMatr.sumElements(ev_pI),"sum(ev_pI)",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurMM(n,IQQt,mm.QQi),"Spur(IQQt,QQi)",TMatr.sumElements(psik_I_psik),"sum(psik_I_psik)",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(I.getEValues()),"sum(I.ev)",TMatr.sumElements(psik_I_psik),"sum(psik_I_psik)",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(ev_pI),"sum(ev_pI)",TMatr.sumElements(psik_pI_psik),"sum(psik_pI_psik)",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(psik_psi0,psik_psi0),"sum(psik_psi0[k]^2)",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulV2V(psik_psi0,java.util.Arrays.copyOf(bbasis,n)),"sum(psik_psi0[k]^2 If_k)",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(psik_psi0,psik_pI_psi0),"sum(psik_psi0*psik_pI_psi0)",TMatr.mulV2V(bpsi0,java.util.Arrays.copyOf(ev_pI,n)),"<psi0|pI|psi0>",eps);
	UnitTests.checkDoubleRel(TMatr.mulVV(psik_psi0,psik_I_psi0),"sum(psik_psi0*psik_I_psi0)",TMatr.mulV2V(bpsi0,I.getEValues()),"<psi0|I|psi0>",eps);

	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQi_feq,n),"<psi0|I^f|psi0>",I_0_feq,"I_0_feq",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsi0,bQQpi,n),"<bpsi0|bQQpi|bpsi0>",TMatr.getAMatrA(I.psi0,mm.QQpi,n),"<psi0|pI|psi0>",eps);
	UnitTests.checkDoubleRel(I.s0,"I.s0",TMatr.mulV2V(bpsi0,I.getEValues()),"<psi0|I|psi0>",eps);

	System.err.println("PnLdIfdyn.get_PnLdIfdyn selftest OK n="+n+" eps="+eps);
    }

    return new PnLdIfdyn(I,mm.p_offset,ev_pI[ixH]/I.ev[ixH],p_IHf,I_0_feq,psik_psi0,psik_I_psi0,psik_pI_psi0,psik_I_psik,psik_pI_psik);
}

}
