package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2015
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.RayleighQuotient;
import com.polytechnik.utils.VolMatrix;

/**
 *  @version $Id: PnLfIHI0.java,v 1.26 2022/07/01 17:28:21 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PnLfIHI0 {

    private static final boolean FLAG_TRY_COMPENSATE_ROUNDINGERR=true;
    public final EVXData I;
    public final double p_offset,p_IH,I_0_feq,I_H_feq,Pf_psi0_dI,Pf_psi0,Pf_psiIH_dI,Pf_psiIH,Pf,Pfdynimp,A_PfdI,A_Pf,A_Pb;
    PnLfIHI0(final EVXData I,
	     final double p_offset,
	     final double p_IH,
	     final double I_0_feq,final double I_H_feq,
	     final double Pf_psi0_dI,final double Pf_psiIH_dI,final double Pf,final double Pfdynimp,final double A_PfdI,final double A_Pb)
    {
	this.I=I;
	this.p_offset=p_offset;
	this.p_IH=p_IH;
	this.I_0_feq=I_0_feq;
	this.I_H_feq=I_H_feq;
	this.Pf_psi0_dI=Pf_psi0_dI;
	this.Pf_psiIH_dI=Pf_psiIH_dI;
	final double dI=I_0_feq-I.s0;
	this.Pf_psi0=Pf_psi0_dI/dI;
	this.Pf_psiIH=Pf_psiIH_dI/dI;
	this.Pf=Pf;
	this.Pfdynimp=Pfdynimp;
	this.A_PfdI=A_PfdI;
	this.A_Pf=A_PfdI/dI;
	this.A_Pb=A_Pb;
    }




/**
 * Solve <psi|I|psi>/<psi|psi>=lambda extremum problem, find eigenstates & eigenvalues.
 * find psi^{IH} and lambda^{IH}=max lambda, used as an estimator of
 * the level of "future" execution.
 * Then dI=lambda^{IH}-<psi0|I|psi0> (contribution of future, not yet executed, trades).
 * Add to I the |psi0><psi0| operator in a way <psi0|Ix|psi0> to be equal lambda^{IH}.
 * Then If (future I) is: If= I + |psi0>dI<psi0|
 * Find the eigenstates of this new Ix matrix, the |psifH> corresponding to I_H_feq=max If.
 * Then variate pI+|psi0>Pf*dI<psi0| on max(Ix) eigenstate with several |\delta psi> variations.
 * Find Pf from the condition the variation equal to 0.
 */
public static PnLfIHI0 get_PnLfIHI0(final MomentsData mm,
				    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				    final double eps)
{
    final int n=mm.n;
    if(n<2) throw new RuntimeException("n="+n+"<2");
    final EVXData I=EVXData.getEVXDataFromQQ(n,mm.QQi,mm.QQt,M);
    final double I_0=I.s0,I_0_feq=I.sH,dI=I_0_feq-I_0;
    final int ixH=n-1; // the state with maximal I

    // Create bQQi_feq (operator of "future" I), with the |bpsi0>dI<bpsi0| contribution from the future:
    // bQQi_feq = bQQi + |bpsi0>dI<bpsi0|
    final double [] bQQi_feq=TMatr.scaleVectorInPlace(TMatr.getQQaa(I.psik_at_x0),dI/(I.o_o)),bpsi0=TMatr.getVectorScaled(I.psik_at_x0,1/Math.sqrt(I.o_o)),bpsiH=new double [n];
    for(int k=0;k<n;k++) bQQi_feq[k+n*k]+=I.ev[k];
    bpsiH[ixH]=1;

    final double [] bQQpi=EVXData.convertMatrFromQtoB_Basis(n,mm.QQpi,I.ev),C=new double [n];
    for(int k=0;k<n;k++) C[k]=bQQpi[k+n*k];
    final double C_IH=C[ixH]; // Element corresponding to IH (max I) state.
    // Diagonalize "future" I:
    final double [] bbasis=M.EV.getEV(n,bQQi_feq,true);
    final double [] bpsifH=EVXData.getPsiFromBasis(ixH,bbasis,n);
    final double I_H_feq=bbasis[ixH],
	bpsifH_pI_bpsifH=TMatr.getAMatrA(bpsifH,bQQpi,n),
	bpsifH_pI_bpsi0=TMatr.getAMatrB(bpsifH,bpsi0,bQQpi,n),
	bpsifH_bpsi0=TMatr.mulVV(bpsifH,bpsi0),
	Pf_psi0_dI=(bpsifH_pI_bpsifH-(bpsifH_pI_bpsi0/bpsifH_bpsi0))/(1-Math.pow(bpsifH_bpsi0,2)),
	_Pf_psi0_dI=PnLbI0.geteq_PfdI(n,bpsifH,bpsi0,bpsi0,bQQpi),
	bpsifH_pI_bpsiH=TMatr.getAMatrB(bpsifH,bpsiH,bQQpi,n),
	bpsifH_bpsiH=TMatr.mulVV(bpsifH,bpsiH),
	bpsiH_bpsi0=TMatr.mulVV(bpsiH,bpsi0),
	Pf_psiIH_dI=(bpsifH_pI_bpsifH-bpsifH_pI_bpsiH/bpsifH_bpsiH)/
	(bpsifH_bpsi0*bpsiH_bpsi0/bpsifH_bpsiH-Math.pow(bpsifH_bpsi0,2)),
	_Pf_psiIH_dI=PnLbI0.geteq_PfdI(n,bpsifH,bpsi0,bpsiH,bQQpi),
	Pf=(C_IH-bpsifH_pI_bpsifH)/(I.sH-TMatr.mulV2V(bpsifH,I.getEValues())),
	Pfdynimp=(C_IH-TMatr.mulV2V(bpsifH,java.util.Arrays.copyOf(C,n)))/(I.sH-TMatr.mulV2V(bpsifH,I.getEValues()));
    double sPfdI=0,sPb=0;
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,bbasis,n);
	final double bpsifk_bpsi0=TMatr.mulVV(psik,bpsi0),
	    bpsifk_pI_bpsi0=TMatr.getAMatrB(n,bbasis,n+k*n,bpsi0,0,bQQpi,n),
	    bpsifk_pI_bpsik=TMatr.getAMatrA(n,bbasis,n+k*n,bQQpi,n),
	    bpsifk_I_bpsik=TMatr.mulV2V(psik,I.getEValues()),
	    w=Math.pow(bpsifk_bpsi0,2),
	    PfkdI=(bpsifk_pI_bpsik-(bpsifk_pI_bpsi0/bpsifk_bpsi0))/(1-w);
	sPfdI+=w*PfkdI;
	sPb+=w*(bpsifk_pI_bpsik/bpsifk_I_bpsik);
    }
    if(eps>0)
    {
	final double Ib=(I_H_feq-dI*Math.pow(bpsifH_bpsi0,2));
	System.err.println("PnLfIHI0.get_PnLfIHI0 Pf="+Pf+" I.sH="+I.sH+" Ib="+Ib+" dir Ib="+TMatr.mulV2V(bpsifH,I.getEValues()));
	UnitTests.checkDoubleRel(Pf_psi0_dI,"Pf_psi0_dI",_Pf_psi0_dI,"_Pf_psi0_dI",eps);
	UnitTests.checkDoubleRel(Pf_psiIH_dI,"Pf_psiIH_dI",_Pf_psiIH_dI,"_Pf_psiIH_dI",eps);
	if(n>2) UnitTests.checkDoubleRel(Pf,"Pf",(C_IH-bpsifH_pI_bpsifH)/(I.sH-Ib),"Pf dir",eps);
	System.err.println("PnLfIHI0.get_PnLfIHI0 selftest OK n="+n+" eps="+eps);
    }


    return new PnLfIHI0(I,mm.p_offset,C_IH/I.sH,I_0_feq,I_H_feq,
			FLAG_TRY_COMPENSATE_ROUNDINGERR?_Pf_psi0_dI:Pf_psi0_dI,
			FLAG_TRY_COMPENSATE_ROUNDINGERR?_Pf_psiIH_dI:Pf_psiIH_dI,
			Pf,Pfdynimp,sPfdI,sPb);

}

public static void matchPnLdIf0(final PnLfIHI0 a,final PnLdIf0 b,final double eps)
{
    EVXData.matchEVXData(a.I,b.I,false,eps);
    UnitTests.checkDoubleRel(a.I_0_feq,"a.I_0_feq",b.I_0_feq,"b.I_0_feq",eps);

    final double p_offset_shift=b.p_offset-a.p_offset;
    UnitTests.checkDoubleRel(a.p_IH,"a.p_IH",b.p_IH+p_offset_shift,"b.p_IH+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.A_Pf,"a.A_Pf",b.A_Pf+p_offset_shift,"b.A_Pf+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.A_Pb,"a.A_Pb",b.A_Pb+p_offset_shift,"b.A_Pb+p_offset_shift",eps);
    UnitTests.checkDoubleRel(a.Pf_psi0_dI,"a.Pf_psi0_dI",b.match_Pf_psi0_dI_PnLfIHI0+p_offset_shift*(b.I_0_feq-b.I.s0),"b.match_Pf_psi0_dI_PnLfIHI0+p_offset_shift*(b.I_0_feq-b.I.s0)",eps);
    UnitTests.checkDoubleRel(a.A_PfdI,"a.A_PfdI",b.A_PfdI+p_offset_shift*(b.I_0_feq-b.I.s0),"b.A_PfdI+p_offset_shift*(b.I_0_feq-b.I.s0)",eps);

    System.err.println("PnLfIHI0.matchPnLdIf0 test OK n="+a.I.neq+" eps="+eps);
}

}
