package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.OrthogonalPolynomialsLegendreShiftedBasis;
import com.polytechnik.utils.OrthogonalPolynomialsLaguerreBasis;
import com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis;

/**
 *  @version $Id: PsiIntegratedByInterval.java,v 1.13 2022/07/28 22:30:50 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
abstract class PsiIntegratedByInterval {
    static class Intervals {
	/** Edge at 0-th element. */
	final double [] x,w;
	final int index_w_max_interval;
	final double x_w_max_interval,w_max_interval;
	Intervals(final double [] x,final double [] w)
	{
	    this.x=x;
	    this.w=w;
	    int ind_wmax=-1;
	    double wm=0;
	    for(int k=1;k<w.length;k++)
	    {
		final double ww=w[k-1]-w[k];
		if(ww>=wm){
		    wm=ww;
		    ind_wmax=k;
		}
	    }
	    this.index_w_max_interval=ind_wmax;
	    this.x_w_max_interval=ind_wmax>=0?x[ind_wmax]:Double.NaN;
	    this.w_max_interval=wm;
	}
    }


    abstract double getOmega(final double x);
    abstract double [] getPsiWIntegrated(final double [] psi2);
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
    final double xstart,xend;
    final boolean isEdgeAtRight;
    PsiIntegratedByInterval(final double xstart,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	this.M=M;
	this.xstart=xstart;
	final double [] ONE=M.B.getONE(),r=M.getRadauMultiplied(ONE);
	final double r0=M.B.getD0(r,0),dr0=M.B.getD1(r,0);
	this.isEdgeAtRight=dr0<0;
	this.xend=-r0/dr0;
	if(isEdgeAtRight?xstart>xend:xstart<xend) throw new IllegalArgumentException("isEdgeAtRight="+isEdgeAtRight+" start="+xstart+" xend="+xend);
    }
    double [] getXIntervals(final double [] x_sorted_asc)
    {
	final double [] xs=new double [x_sorted_asc.length+1];
	int n=0;
	xs[n++]=xend;
	if(isEdgeAtRight)
	{
	    for(int k=x_sorted_asc.length;--k>=0;)
	    {
		final double x=x_sorted_asc[k];
		if(xstart<=x && x<xend) xs[n++]=x;
	    }
	}
	else
	{
	    for(int k=0;k<x_sorted_asc.length;k++)
	    {
		final double x=x_sorted_asc[k];
		if(xstart>=x && x>xend) xs[n++]=x;
	    }
	}
	//System.err.println("xstart="+xstart+" xend="+xend+" isEdgeAtRight="+isEdgeAtRight);
	return  java.util.Arrays.copyOf(xs,n);
    }
    Intervals getIntervals(final double [] x_sorted_asc,final double [] psi2)
    {
	final double [] x=getXIntervals(x_sorted_asc),w=new double [x.length],
	    psiw2=getPsiWIntegrated(psi2);
	for(int k=0;k<x.length;k++)
	{
	    final double xk=x[k];
	    w[k]=getOmega(xk)*M.B.getD0(psiw2,xk);
	}
	return new Intervals(x,w);
    }
    double test_getWeight_Numerically(final double xl,final double xr,final double [] psi2)
    {
	final double abserr=1e-10,relerr=1e-9;
	final int nomax=50000;
	final com.polytechnik.utils.QUANC8.DoubleFunction
	    integrate=new com.polytechnik.utils.QUANC8.DoubleFunction(){
		    public double f(final double x){return M.B.getD0(psi2,x)*getOmega(x);}
		};
	final com.polytechnik.utils.QUANC8 i=
	    com.polytechnik.utils.QUANC8.integrate(integrate,
						   xl,
						   xr,
						   abserr,
						   relerr,
						   nomax);
	if(!(i.flag==0)) return Double.NaN;
	return i.integral;
    }

    static final PsiIntegratedByInterval INTR_QVMDataP=
	new PsiIntegratedByInterval(0,new OrthogonalPolynomialsLegendreShiftedBasis()){
	    double getOmega(final double x){return 1;}
	    double [] getPsiWIntegrated(final double [] psi2){return KkQVMLegendreShifted.getPsiWIntegrated(psi2);}
	},
	INTR_QVMDataL=new PsiIntegratedByInterval(Double.POSITIVE_INFINITY,new OrthogonalPolynomialsLaguerreBasis()){
		double getOmega(final double x){return Math.exp(-x);}
		double [] getPsiWIntegrated(final double [] psi2){return KkQVMLaguerre.getPsiWIntegrated(psi2);}
	    },
	INTR_QVMData=new PsiIntegratedByInterval(Double.NEGATIVE_INFINITY,new OrthogonalPolynomialsMonomialsBasis0()){
		double getOmega(final double x){return Math.exp(x);}
		double [] getPsiWIntegrated(final double [] psi2){return KkQVMMonomials.getPsiWIntegrated(psi2);}
	    };
}
