package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.LogPolynomialsRatio;
import com.polytechnik.realtimebook.*;
import com.polytechnik.outils.*;

/** Impact QVM calculations with kernel <Q(x)|Q(x')> polynomials.
 *  @version $Id: QQVM.java,v 1.65 2022/08/08 20:10:47 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class QQVM<DTYPE> {

final DTYPE odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;

final double [] kkv,kkt,kkfv,kkrts,kkrvs,kkpv,kkp2v,kkpt,kkp2t;
final TwoCloseRootsFinder.RootsData d;
final double xeq,K00t,K00v,A00v,KEQt,KEQv,AEQv,I0,Ia0eq,Ieq,Iaeq,SI0,SIeq;
final double ptE,pt0,pvE,pv0,sigma2_ptE,sigma2_pt0,sigma2_pvE,sigma2_pv0;

QQVM(final int neq,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();

    final int nmoms=2*neq-1;
    final double [] vmom=new double [nmoms],tmom=new double [nmoms],rtsmom=new double [nmoms],rvsmom=new double [nmoms],pvmom=new double [nmoms],p2vmom=new double [nmoms],ptmom=new double [nmoms],p2tmom=new double [nmoms];
    access.setM(vmom,0,access.M_I,odata,nmoms);
    access.setMT(tmom,nmoms);
    access.setM(rtsmom,0,access.M_RTS,odata,nmoms);
    access.setM(rvsmom,0,access.M_RVS,odata,nmoms);
    access.setM(pvmom,1,access.M_I,odata,nmoms);
    access.setM(p2vmom,2,access.M_I,odata,nmoms);
    access.setM(ptmom,1,access.M_T,odata,nmoms);
    access.setM(p2tmom,2,access.M_T,odata,nmoms);
    final double [][] QQ=access.getQQ_Matrs(neq,1,new double [][]{vmom,tmom,rtsmom,rvsmom,pvmom,p2vmom,ptmom,p2tmom});
    final double [] QQMv=QQ[0],QQMt=QQ[1],QQMrts=QQ[2],QQMrvs=QQ[3],
	QQMpv=QQ[4],QQMp2v=QQ[5],QQMpt=QQ[6],QQMp2t=QQ[7],
	IQQv=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQMv),
	IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQMt);
    final double [] m=new double [neq*neq];
    final double [] QQfv=ZQVM.getIQQ_F_IQQ(neq,IQQt,QQMv,m),
	QQrts=ZQVM.getIQQ_F_IQQ(neq,IQQt,QQMrts,m),
	QQrvs=ZQVM.getIQQ_F_IQQ(neq,IQQt,QQMrvs,m),
	QQpv=ZQVM.getIQQ_F_IQQ(neq,IQQv,QQMpv,m),
	QQp2v=ZQVM.getIQQ_F_IQQ(neq,IQQv,QQMp2v,m),
	QQpt=ZQVM.getIQQ_F_IQQ(neq,IQQt,QQMpt,m),
	QQp2t=ZQVM.getIQQ_F_IQQ(neq,IQQt,QQMp2t,m);
    final double [][] kkk=access.getKK(neq,new double [][] {IQQv,IQQt,QQfv,QQrts,QQrvs,QQpv,QQp2v,QQpt,QQp2t});
    kkv=kkk[0];
    kkt=kkk[1];
    kkfv=kkk[2];
    kkrts=kkk[3];
    kkrvs=kkk[4];
    kkpv=kkk[5];
    kkp2v=kkk[6];
    kkpt=kkk[7];
    kkp2t=kkk[8];

    this.d=getTR(kkv,kkt);
    this.xeq=(d.x1_type==d.TYPE_MAX?d.x1:(d.x2_type==d.TYPE_MAX?d.x2:Double.NaN));
    final double x0=getX0();
    K00t=access.fx(kkt,x0);
    K00v=access.fx(kkv,x0);
    A00v=access.fx(kkfv,x0);
    KEQt=access.fx(kkt,xeq);
    KEQv=access.fx(kkv,xeq);
    AEQv=access.fx(kkfv,xeq);

    I0=K00t/K00v;
    Ia0eq=A00v/K00t;
    Ieq=KEQt/KEQv;
    Iaeq=AEQv/KEQt;

    SI0=access.fx(kkrvs,x0)/K00t;
    SIeq=access.fx(kkrvs,xeq)/KEQt;

    final double p_offset=getPOffset(),
	dpvE=access.fx(kkpv,xeq)/KEQv,
	dpv0=access.fx(kkpv,x0)/K00v,
	dptE=access.fx(kkpt,xeq)/KEQt,
	dpt0=access.fx(kkpt,x0)/K00t;
    pvE=p_offset+dpvE;
    sigma2_pvE=access.fx(kkp2v,xeq)/KEQv-dpvE*dpvE;
    pv0=p_offset+dpv0;
    sigma2_pv0=access.fx(kkp2v,x0)/K00v-dpv0*dpv0;
    ptE=p_offset+dptE;
    sigma2_ptE=access.fx(kkp2t,xeq)/KEQt-dptE*dptE;
    pt0=p_offset+dpt0;
    sigma2_pt0=access.fx(kkp2t,x0)/K00t-dpt0*dpt0;

}


abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double getX0();
abstract double getPOffset();
abstract TwoCloseRootsFinder.RootsData getTR(final double [] kkv,final double [] kkt);

static abstract class QQVMTest<DTYPE> extends QTest {

abstract QQVM<DTYPE> getQQVM();
abstract ZQVM<DTYPE> getZQVM();
abstract ZVQVM<DTYPE> getZVQVM();
abstract BQVM<DTYPE> getBQVM();
abstract double [][] getMonomialMoments(DTYPE o);



QQVMTest(final String name,
	 final java.util.Random r,final double eps)
{
    super(name,r,eps);
}


abstract boolean isDescending();
void test_Monomials_BasisExpansion()
{
    final QQVM<DTYPE> q=getQQVM();
    final double [][] amoms=getMonomialMoments(q.odata);
    final double [] mom_v=amoms[0],mom_t=amoms[1],mom_rts=amoms[2],mom_rvs=amoms[3];

    final double [] kkt=OQVM.OQVMTest.getKK(neq,mom_t);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.kkt),"conv(q.kkt)",kkt,"kkt",eps);
    final double [] kkv=OQVM.OQVMTest.getKK(neq,mom_v);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.kkv),"conv(q.kkc)",kkv,"kkv",eps);
    final double [] QQMv=NQVM.NQVMTest.pol_getMM_Matr_scal(neq,mom_v,1),
	QQMt=NQVM.NQVMTest.pol_getMM_Matr_scal(neq,mom_t,1),
	QQMrts=NQVM.NQVMTest.pol_getMM_Matr_scal(neq,mom_rts,1),
	QQMrvs=NQVM.NQVMTest.pol_getMM_Matr_scal(neq,mom_rvs,1);

    final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQMt),
	IQQv=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQMv);
    final double []
	QQfv=TMatr.mulMMM(neq,IQQt,QQMv,IQQt),
	QQrts=TMatr.mulMMM(neq,IQQt,QQMrts,IQQt),
	QQrvs=TMatr.mulMMM(neq,IQQt,QQMrvs,IQQt);

    final double [] kkfv=OQVM.OQVMTest.getKK_fromQQ(neq,QQfv),kkrts=OQVM.OQVMTest.getKK_fromQQ(neq,QQrts),kkrvs=OQVM.OQVMTest.getKK_fromQQ(neq,QQrvs);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.kkfv),"conv(q.kkfv)",kkfv,"kkfv",eps);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.kkrvs),"conv(q.kkrvs)",kkrvs,"kkrvs",eps);
    UnitTests.checkDoubleArrayRel(convertToMonomials(q.kkrts),"conv(q.kkrts)",kkrts,"kkrts",eps);


    final LogPolynomialsRatio tr= new LogPolynomialsRatio(q.kkv,q.kkt,q.getABasis().B);
    for(int j=0;j<50;j++)
    {
	final double x=0.5*r.nextDouble();
	UnitTests.checkDoubleRel(tr.getF(x),"tr.getF(x)",Math.log(Polynomials.getD0(kkt,x)/Polynomials.getD0(kkv,x)),"F",eps);
	UnitTests.checkDoubleRel(tr.getDF(x),"tr.getDF(x)",Polynomials.getD1(kkt,x)/Polynomials.getD0(kkt,x)-Polynomials.getD1(kkv,x)/Polynomials.getD0(kkv,x),"dF",eps);
	final double ee=eps*1e0,d2f=(tr.getDF(x+ee)-tr.getDF(x-ee))/(2*ee);
	UnitTests.checkDoubleRel(tr.getD2F(x),"tr.getD2F(x)",d2f,"d2F",Math.max(eps,1e-2));
	UnitTests.checkDoubleRel(tr.getD2F(x),"tr.getD2F(x)",Polynomials.getD2(kkt,x)/Polynomials.getD0(kkt,x)-Polynomials.getD2(kkv,x)/Polynomials.getD0(kkv,x)-Math.pow(Polynomials.getD1(kkt,x)/Polynomials.getD0(kkt,x),2)+Math.pow(Polynomials.getD1(kkv,x)/Polynomials.getD0(kkv,x),2),"d2F",eps);
    }

    System.err.println(name+":QQVMTest:test_Monomials_BasisExpansion(): test OK neq="+neq+" eps="+eps);
}



void test_Common()
{
    final QQVM<DTYPE> q=getQQVM();
    final double x0=q.getX0();
    System.err.println("test_Common():x0="+q.d.x0+" F0="+q.d.F0+" dF0="+q.d.dF0+" d2F0="+q.d.d2F0);
    System.err.println("test_Common():x1="+q.d.x1+" F1="+q.d.F1+" dF1="+q.d.dF1+" d2F1="+q.d.d2F1+" x1_type="+q.d.x1_type);
    System.err.println("test_Common():x2="+q.d.x2+" F2="+q.d.F2+" dF2="+q.d.dF2+" d2F2="+q.d.d2F2+" x2_type="+q.d.x2_type);

    UnitTests.checkDoubleRel(q.d.F0,"q.d.F0",Math.log(q.I0),"log(q.I0)",eps);

    if(q.d.x1_type!=q.d.TYPE_UNKNOWN) UnitTests.checkDoubleRel(q.d.dF1,"q.d.dF1",0,"0",eps);
    if(q.d.x2_type!=q.d.TYPE_UNKNOWN) UnitTests.checkDoubleRel(q.d.dF2,"q.d.dF2",0,"0",eps);

    //if(!Double.isNaN(q.d.xmin)) throw new RuntimeException("q.d.xmin="+q.d.xmin);
    if(q.d.x2_type!=q.d.TYPE_UNKNOWN)
    {
	if(isDescending())
	{
	    if((q.d.dF0>0 && (q.d.x1_type!=q.d.TYPE_MIN || q.d.x2_type!=q.d.TYPE_MAX))||
	       (q.d.dF0<0 && (q.d.x1_type!=q.d.TYPE_MAX || q.d.x2_type!=q.d.TYPE_MIN)))
		throw new RuntimeException("q.d.dF0="+q.d.dF0+" q.d.x1_type="+q.d.x1_type+" q.d.x2_type="+q.d.x2_type+" isDescending()="+isDescending());
	}
	else
	{
	    if((q.d.dF0<0 && (q.d.x1_type!=q.d.TYPE_MIN || q.d.x2_type!=q.d.TYPE_MAX))||
	       (q.d.dF0>0 && (q.d.x1_type!=q.d.TYPE_MAX || q.d.x2_type!=q.d.TYPE_MIN)))
		throw new RuntimeException("q.d.dF0="+q.d.dF0+" q.d.x1_type="+q.d.x1_type+" q.d.x2_type="+q.d.x2_type+" isDescending()="+isDescending());
	}
    }

    System.err.println(name+":QQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}

void test_ZQVM()
{
    final QQVM<DTYPE> q=getQQVM();
    {
    final ZVQVM<DTYPE> s=getZVQVM();
    UnitTests.checkDoubleArrayRel(q.kkv,"q.kkv",s.kk,"s.kk",eps);
    }
    {
    final ZQVM<DTYPE> s=getZQVM();
    UnitTests.checkDoubleArrayRel(q.kkrts,"q.kkrts",s.kkrt,"s.kkrt",eps);
    UnitTests.checkDoubleArrayRel(q.kkt,"q.kkt",s.kk,"s.kk",eps);
    }

    System.err.println(name+":QQVMTest:test_ZQVM(): test OK neq="+neq+" eps="+eps);
}

void test_BQVM()
{
    final QQVM<DTYPE> q=getQQVM();
    final BQVM<DTYPE> s=getBQVM();
    UnitTests.checkDoubleRel(q.pv0,"q.pv0",s.pv0,"s.pv0",eps);
    UnitTests.checkDoubleRel(q.sigma2_pv0,"q.sigma2_pv0",s.sigma2_pv0,"s.sigma2_pv0",eps);
    UnitTests.checkDoubleRel(q.pt0,"q.pt0",s.pt0,"s.pt0",eps);
    UnitTests.checkDoubleRel(q.sigma2_pt0,"q.sigma2_pt0",s.sigma2_pt0,"s.sigma2_pt0",eps);

    System.err.println(name+":QQVMTest:test_BQVM(): test OK neq="+neq+" eps="+eps);
}

void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":QQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_ZQVM();
    test_BQVM();
    test_Monomials_BasisExpansion();

    System.err.println(name+":QQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":QQVMTest: starting tests; ver=$Id: QQVM.java,v 1.65 2022/08/08 20:10:47 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<118;i++)
	{
	    test_Calculations(1350+r.nextInt(1550),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":QQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: QQVM.java,v 1.65 2022/08/08 20:10:47 mal Exp $");
}
} // END OF class QQVMTest

}// END OF class QQVM
