package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.Laguerre;

/** The data in Laguerre basis.
 *  @version $Id: QVMDataBL.java,v 1.7 2022/06/27 13:32:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class QVMDataBL {

final long currenttime,tau,vol;
final double p_offset;

final int ntm,npm;

final double [] // here offset=p_offset and array dims are: ntm,npm.
    vmoments_Ltau_x_p // [k][l]=sum L_k(t_cur-t) (p-offset)^l dv_i exp()
    ;
final BookItem edge;


private QVMDataBL(final long currenttime,
		  final long tau,
		  final long vol,
		  final double p_offset,

		  final double [] vmoments_Ltau_x_p,

		  final int ntm,
		  final int npm,
		  final BookItem edge)
{
    this.currenttime=currenttime;
    this.tau=tau;
    this.vol=vol;
    this.p_offset=p_offset;
    this.vmoments_Ltau_x_p=vmoments_Ltau_x_p;
    this.ntm=ntm;
    this.npm=npm;
    this.edge=edge;
}


public static QVMDataBL calculateQVMDataBL(final int ntm,
					 final int npm,
					 final long currenttime,
					 final long tau,
					 final BookItem [] tb,
					 final int tsize)
{
    final double atau=tau;
    final double p_offset=(tsize>0?tb[0].price:0);



    final double []  // all 0 by default
	vmoments_Ltau_x_p=new double [ntm*npm];

    //System.err.println("===================");
    long vol=0;
    for(int i=0;i<tsize;i++)
    {
	final BookItem b=tb[i];
	//System.err.println("b["+i+"]="+b.price);
	final long prevt=b.time;
	vol+=b.shares;
	final double etau,ex;
	if(tau<=0){etau=0;ex=1;}
	else {etau=(currenttime-prevt)/atau;ex=Math.exp(-etau);}
	final double dp=b.price-p_offset,bwmw=ex*b.shares;

	double bte=0,ste=1;
	for(int k=0,j=0;k<ntm;++k,ste=((1-k)*bte+((2*k-1)-etau)*(bte=ste))/k) // Laguerre moments
	{
	    double wpe=ste;
	    for(int l=0;l<npm;l++,j++,wpe*=dp) // Polynomial moments
	    {
		vmoments_Ltau_x_p[j]+=wpe*bwmw;
	    }
	}
    }

    return new QVMDataBL(currenttime,
			 tau,
			 vol,
			 p_offset,
			 vmoments_Ltau_x_p,
			 ntm,
			 npm,
			 tsize>0?tb[0]:null);
}


private static void checkMatchingQVMDataL(final QVMDataBL r,final QVMDataL ref,final double eps)
{
    if(r.currenttime!=ref.currenttime)
    {
	throw new RuntimeException("r.currenttime="+r.currenttime+
				   " ref.currenttime="+ref.currenttime);
    }
    if(r.tau!=ref.tau)
    {
	throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
    }
    if(r.vol!=ref.vol)
    {
	throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
    }

    //UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);
    final double [] vmoments_Ltau_x_p_offsetp=new double [r.ntm*r.npm];
    final double doffset=(ref.p_offset-r.p_offset);

    for(int k=0,o=0;k<r.ntm;k++,o+=r.npm)
    {
	for(int m=0;m<r.npm;m++)
	{
	    final double [] p=new double [m+1];
	    p[m]=1;
	    final double [] po=Polynomials.getPolynomialShifted(p,doffset);
	    vmoments_Ltau_x_p_offsetp[o+m]=DeprecatedHelperTMatr.mulVVo(m+1,ref.vmoments_Ltau_x_p,o,po,0);
	}
    }

    UnitTests.checkDoubleArrayRel(r.vmoments_Ltau_x_p,"r.vmoments_Ltau_x_p",
				  vmoments_Ltau_x_p_offsetp,"vmoments_Ltau_x_p_offsetp",eps);

    System.err.println("QVMDataBL:checkMatching(): matching OK eps="+eps);
}



private static void checkCalculations(final int tsize,
				      final java.util.Random r,
				      final int ntm,
				      final int npm,
				      final double eps)
{
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMDataBL:checkCalculations(): test with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+" ntm="+ntm+" npm="+npm);

    final QVMDataBL rq=QVMDataBL.calculateQVMDataBL(ntm,npm,currenttime,tau,tb,tsize);
    final QVMDataL rs=QVMDataL.calculateQVMDataL(ntm,npm,currenttime,tau,tb,tsize,null);
    checkMatchingQVMDataL(rq,rs,eps);

    System.err.println("QVMDataBL:checkCalculations(): OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+" ntm="+ntm+" npm="+npm);
}


public static void main(String [] args)
{
    System.err.println("QVMDataBL::main(): starting tests; ver=$Id: QVMDataBL.java,v 1.7 2022/06/27 13:32:46 mal Exp $");

    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-79091);
    final double eps=1e-5;

    for(int ntm=1;ntm<=7;ntm++)
    for(int npm=1;npm<=7;npm++)
    {
	for(int i=0;i<100;i++)
	{
	    checkCalculations(100+r.nextInt(100),r,ntm,npm,eps);
	    checkCalculations(10010+r.nextInt(100010),r,ntm,npm,eps);
	}
    }

    final long t1=System.currentTimeMillis();
    System.err.println("QVMDataBL::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVMDataBL.java,v 1.7 2022/06/27 13:32:46 mal Exp $");
}


} // END of class QVMDataBL
