package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;

/** Book data in P-monomials basis.
 *  Simplified QVMDataL. No various moments. No recurrence.
 *  @version $Id: QVMDataBMP.java,v 1.15 2022/06/27 13:32:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class QVMDataBMP {

final long currenttime,tau,vol;
final double p_offset,p_cut;

final double []
    vmoments_x_p, // [k]=sum (p_i-offset)^k exp(-(t_cur-t_i)/tau) dv_i
    tvmoments_x_p,  // [k]=sum (p_i-offset)^k exp(-(t_cur-t_i)/tau) (t_cur-t_i)/tau dv_i
    etvmoments_x_p  // [k]=sum (p_i-offset)^k exp(-(t_cur-t_i)/tau) (1-exp(-(t_cur-t_i)/tau)) dv_i
    ;
final int npm;
final BookItem edge;

private QVMDataBMP(final long currenttime,
		   final long tau,
		   final long vol,
		   final double p_offset,

		   final double [] vmoments_x_p,
		   final double [] tvmoments_x_p,
		   final double [] etvmoments_x_p,
		   final double p_cut,
		   final int npm,

		   final BookItem edge)
{
    this.currenttime=currenttime;
    this.tau=tau;
    this.vol=vol;
    this.p_offset=p_offset;
    this.p_cut=p_cut;
    this.vmoments_x_p=vmoments_x_p;
    this.tvmoments_x_p=tvmoments_x_p;
    this.etvmoments_x_p=etvmoments_x_p;
    this.npm=npm;
    this.edge=edge;
}


public static QVMDataBMP calculateQVMDataBMP(final int npm,
					     final long currenttime,
					     final long tau,
					     final BookItem [] tb,
					     final int tsize,
					     final double p_cut)
{
    final double atau=tau,p_offset=(tsize>0?tb[0].price:0);

    final double []  // all 0 by default
	vmoments_x_p=new double [npm],
	tvmoments_x_p=new double [npm],
	etvmoments_x_p=new double [npm];

    long vol=0;
    for(int i=0;i<tsize;i++)
    {
	final BookItem b=tb[i];
	if(b.buy_sell_indicator==BookItem.BOOK_BUY){
	    if(b.price<p_cut) break;
	} else if(b.buy_sell_indicator==BookItem.BOOK_SELL){
	   if(b.price>p_cut) break;
	}
	final long prevt=b.time;
	vol+=b.shares;
	final double taum,ex;
	if(tau<=0){taum=currenttime-prevt;ex=1;}
	else {taum=(currenttime-prevt)/atau;ex=Math.exp(-taum);}
	final double dp=b.price-p_offset,bwmw=ex*b.shares;

	double wpe=1;
	for(int k=0;k<npm;++k,wpe*=dp) // Monomial Moments
	{
	    vmoments_x_p[k]+=wpe*bwmw;
	    tvmoments_x_p[k]+=wpe*bwmw*taum;
	    etvmoments_x_p[k]+=wpe*bwmw*(1-ex);
	}
    }

    return new QVMDataBMP(currenttime,
			  tau,
			  vol,
			  p_offset,
			  vmoments_x_p,
			  tvmoments_x_p,
			  etvmoments_x_p,
			  p_cut,
			  npm,
			  tsize>0?tb[0]:null);
}



private static void checkMatchingQVMDataL(final QVMDataBMP r,final QVMDataL ref,final QVMDataP refp,final double eps)
{
    if(r.currenttime!=ref.currenttime)
    {
	throw new RuntimeException("r.currenttime="+r.currenttime+
				   " ref.currenttime="+ref.currenttime);
    }
    if(r.tau!=ref.tau)
    {
	throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
    }
    if(r.vol!=ref.vol)
    {
	throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
    }


    //UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);
    final double [] vmoments_x_p=new double [r.npm],tvmoments_x_p=new double [r.npm],etvmoments_x_p=new double [r.npm];
    final double doffsetL=(ref.p_offset-r.p_offset),doffsetP=(refp.p_offset-r.p_offset);
    final DeprecatedHelperQVMDataLDirectAccessWithBasis accessL=new DeprecatedHelperQVMDataLDirectAccessWithBasis();
    final DeprecatedHelperQVMDataPDirectAccessWithBasis accessP=new DeprecatedHelperQVMDataPDirectAccessWithBasis();
    final double [] t=new double [2],vL=new double [r.npm],tvL=new double [r.npm],vP=new double [r.npm],tvP=new double [r.npm];
    for(int k=0;k<r.npm;k++)
    {
	accessL.setM(t,k,accessL.M_I,ref,2);
	vL[k]=t[0];
	tvL[k]=-(t[1]-t[0]); // x=-(L_1(x)-L_0(x))
	accessP.setM(t,k,accessL.M_I,refp,2);
	vP[k]=t[0];
	tvP[k]=-(t[1]-t[0])/2; // (1-x)=-(P_1(x)-P_0(x))/2
	final double [] p=new double [k+1];p[k]=1;
	final double [] poL=Polynomials.getPolynomialShifted(p,doffsetL);
	vmoments_x_p[k]=TMatr.mulVVn(k+1,vL,poL);
	tvmoments_x_p[k]=TMatr.mulVVn(k+1,tvL,poL);
	final double [] poP=Polynomials.getPolynomialShifted(p,doffsetP);
	etvmoments_x_p[k]=TMatr.mulVVn(k+1,tvP,poP);
    }


    UnitTests.checkDoubleArrayRel(r.vmoments_x_p,"r.vmoments_x_p",vmoments_x_p,"vmoments_x_p",eps);
    UnitTests.checkDoubleArrayRel(r.tvmoments_x_p,"r.tvmoments_x_p",tvmoments_x_p,"tvmoments_x_p",eps);
    UnitTests.checkDoubleArrayRel(r.etvmoments_x_p,"r.etvmoments_x_p",etvmoments_x_p,"etvmoments_x_p",eps);

    System.err.println("QVMDataBMP:checkMatchingQVMDataL(): matching OK eps="+eps);
}


private static void checkCalculations(final int tsize,
				      final java.util.Random r,
				      final int npm,
				      final double eps)
{
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/8,currenttime=tb[tb.length-1].time+2000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMDataBMP:checkCalculations(): test with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+" npm="+npm);

    final QVMDataBMP rq=QVMDataBMP.calculateQVMDataBMP(npm,currenttime,tau,tb,tsize,Double.NaN);
    final QVMDataL rsL=QVMDataL.calculateQVMDataL(2,npm,currenttime,tau,tb,tsize,null);
    final QVMDataP rsP=QVMDataP.calculateQVMDataP(2,npm,currenttime,tau,tb,tsize,null);
    checkMatchingQVMDataL(rq,rsL,rsP,eps);


    System.err.println("QVMDataBMP:checkCalculations(): OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+" npm="+npm);
}


public static void main(String [] args)
{
    System.err.println("QVMDataBMP::main(): starting tests; ver=$Id: QVMDataBMP.java,v 1.15 2022/06/27 13:32:46 mal Exp $");

    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-16803);
    final double eps=1e-5;

    for(int npm=1;npm<=7;npm++)
    {
	for(int i=0;i<100;i++)
	{
	    checkCalculations(100+r.nextInt(100),r,npm,eps);
	    checkCalculations(10010+r.nextInt(100010),r,npm,eps);
	}
    }

    final long t1=System.currentTimeMillis();
    System.err.println("QVMDataBMP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVMDataBMP.java,v 1.15 2022/06/27 13:32:46 mal Exp $");
}


} // END of class QVMDataBMP
