package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.outils.BissectionRootFinder;
import com.polytechnik.utils.PolynomialRootsFinder;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Quadrature;
/** Book in price space.
 *  @version $Id: QVMF.java,v 1.42 2020/09/27 20:33:20 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class QVMF<DTYPE> {


final double pbuy,psell,k0buy_pbuy,k0buy_psell,k0sell_psell,k0sell_pbuy,peq,
    k0buy_peq,k0sell_peq,peqx;
private final double [] Kbuy,Ksell;
private final double poffset_buy,poffset_sell;

final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;
private final DTYPE odata_buy,odata_sell;

QVMF(final int neq,final DTYPE odata_buy,final DTYPE odata_sell)
{
    this.odata_buy=odata_buy;
    this.odata_sell=odata_sell;
    this.access=getAccess();

    final int nmoms=2*neq-1;
    final double [] pmoms_buy=new double [nmoms],pmoms_sell=new double [nmoms];
    access.setM(pmoms_buy,0,access.M_I,this.odata_buy,nmoms);
    access.setM(pmoms_sell,0,access.M_I,this.odata_sell,nmoms);
    final double [][] QQ=access.getQQ_Matrs(neq,1,new double [][]{pmoms_buy,pmoms_sell});
    final double [] iQQbuy=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQ[0]),
	iQQsell=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQ[1]);
    final double [][] kkk=access.getKK(neq,new double [][] {iQQbuy,iQQsell});
    this.Kbuy=kkk[0];
    this.Ksell=kkk[1];

    this.pbuy=getPEdge(this.odata_buy);
    this.psell=getPEdge(this.odata_sell);

    poffset_buy=getPOffset(this.odata_buy);
    poffset_sell=getPOffset(this.odata_sell);
    this.k0buy_pbuy=access.fx(this.Kbuy,pbuy-poffset_buy);
    this.k0buy_psell=access.fx(this.Kbuy,psell-poffset_buy);
    this.k0sell_psell=access.fx(this.Ksell,psell-poffset_sell);
    this.k0sell_pbuy=access.fx(this.Ksell,pbuy-poffset_sell);
    this.peq=getPEQ();
    this.k0buy_peq=access.fx(this.Kbuy,peq-poffset_buy);
    this.k0sell_peq=access.fx(this.Ksell,peq-poffset_sell);
    this.peqx=new BissectionRootFinder(){
	public double getF(final double x)
	{
	    return (1/access.fx(Kbuy,x-poffset_buy)-1/k0buy_psell)-
		(1/access.fx(Ksell,x-poffset_sell)-1/k0sell_pbuy);
	}
	}.doBissection(pbuy,psell);
}

double getPEQ()
{
    //System.err.println("pbuy="+pbuy+" psell="+psell);
    if(!(psell>pbuy)) return Double.NaN;
    final double S=(psell-pbuy)/5;
    final int NS=250;
    double pxbuy=pbuy;
    for(int k=0;k<NS;k++)
    {
	//System.err.println("pxbuy: k="+k+" f="+(access.fx(this.Kbuy,pxbuy-poffset_buy)-access.fx(this.Ksell,pxbuy-poffset_sell)));
	if(access.fx(this.Kbuy,pxbuy-poffset_buy)<access.fx(this.Ksell,pxbuy-poffset_sell)){break;}
	else if(k>=NS-1) return Double.NaN;
	pxbuy-=S;
    }
    double pxsell=psell;
    for(int k=0;k<NS;k++)
    {
	if(access.fx(this.Kbuy,pxsell-poffset_buy)>access.fx(this.Ksell,pxsell-poffset_sell)){break;}
	else if(k>=NS-1) return Double.NaN;
	pxsell+=S;
    }
    return new BissectionRootFinder(){
	public double getF(final double x)
	{
	    return access.fx(Kbuy,x-poffset_buy)-access.fx(Ksell,x-poffset_sell);
	}
    }.doBissection(pxbuy,pxsell);
}

abstract double getPOffset(final DTYPE odata);
abstract double getPEdge(final DTYPE odata);
abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();



static abstract class QVMFTest<DTYPE> {

int neq;
long currenttime,tau;
Trade [] tbuy,tsell;
BookItem [] tbuy_byp,tsell_byp;

abstract QVMF<DTYPE> getQVMF();
abstract long getTau();

final String name;
final java.util.Random r;
final double eps;
QVMFTest(final String name,
	 final java.util.Random r,final double eps)
{
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Common()
{
    final QVMF<DTYPE> q=getQVMF();

    for(int k=tbuy.length;--k>=0;)
    {
	if(tbuy[k].price>q.pbuy) throw new RuntimeException("tbuy["+k+"].price="+tbuy[k].price+" q.pbuy="+q.pbuy);
    }
    for(int k=tsell.length;--k>=0;)
    {
	if(tsell[k].price<q.psell) throw new RuntimeException("tsell["+k+"].price="+tsell[k].price+" q.psell="+q.psell);
    }

    System.err.println(name+":QVMFTest:test_Common(): test OK neq="+neq+" eps="+eps);
}
void setData(final int buysize,final int sellsize,final int neq)
{
    this.neq=neq;
    this.tbuy=CalculateKendall.TradeEMoments.getRandomTradesQuantumPrice(buysize,r,-1,-2.2,9.98);
    tbuy_byp=tbuy.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tbuy_byp,0,tbuy_byp.length-1,Distribution.cmp_by_price_lowest_down);
    this.tsell=CalculateKendall.TradeEMoments.getRandomTradesQuantumPrice(sellsize,r,-1,2.2,10.1);
    tsell_byp=tsell.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tsell_byp,0,tsell_byp.length-1,Distribution.cmp_by_price_lowest_up);
    this.currenttime=Math.max(tbuy[tbuy.length-1].time,tsell[tsell.length-1].time);
    this.tau=getTau();
    //System.err.println("tau="+tau);
}

void test_QATradeIPBQVM()
{
    if(!(neq>=2) || !(neq<=7)) return;
    final QVMF<DTYPE> q=getQVMF();
    final CalculateKendall.QATradeIP
	sabuy=CalculateKendall.QATradeIP.calculateQATradeIP_V(
	    neq-1,
	    currenttime,
	    tau,
	    tbuy,
	    tbuy.length,
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG(),
	    null),
	sasell=CalculateKendall.QATradeIP.calculateQATradeIP_V(
	    neq-1,
	    currenttime,
	    tau,
	    tsell,
	    tsell.length,
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG(),
	    null);
    final double [] kQA_buy=sabuy.didv.clone(),kQA_sell=sasell.didv.clone();
    kQA_buy[0]+=1;kQA_sell[0]+=1; // to match standard definition
    for(int k=0;k<kQA_buy.length;k++)
    { // make unnormalized
	kQA_buy[k]/=sabuy.odata.moments_pv[0];
	kQA_sell[k]/=sasell.odata.moments_pv[0];
    }
    final double [] ref_sabuy=Polynomials.getPolynomialShifted(kQA_buy,-(sabuy.p_offset-q.poffset_buy)),
	ref_sasell=Polynomials.getPolynomialShifted(kQA_sell,-(sasell.p_offset-q.poffset_sell));

    UnitTests.checkDoubleArrayRel(q.Kbuy,"q.Kbuy",ref_sabuy,"ref_sabuy",eps);
    UnitTests.checkDoubleArrayRel(q.Ksell,"q.Ksell",ref_sasell,"ref_sasell",eps);
    UnitTests.checkDoubleRel(q.k0buy_pbuy,"q.k0buy_pbuy",Polynomials.getD0(kQA_buy,q.pbuy-sabuy.p_offset),"kQA_buy(pbuy)",eps);
    UnitTests.checkDoubleRel(q.k0buy_psell,"q.k0buy_psell",Polynomials.getD0(kQA_buy,q.psell-sabuy.p_offset),"kQA_buy(psell)",eps);
    UnitTests.checkDoubleRel(q.k0sell_psell,"q.k0sell_psell",Polynomials.getD0(kQA_sell,q.psell-sasell.p_offset),"kQA_sell(psell)",eps);
    UnitTests.checkDoubleRel(q.k0sell_pbuy,"q.k0sell_pbuy",Polynomials.getD0(kQA_sell,q.pbuy-sasell.p_offset),"kQA_sell(pbuy)",eps);

    System.err.println(name+":QVMFTest:test_QATradeIPBQVM(): test OK neq="+neq+" eps="+eps);
}

OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getOrtPolBasis(final double p0,final boolean flag_buy)
{
    return new com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis()
    {
	@Override
	public double [] getRadauMultiplied(final double [] nodes0)
	{
	    final double [] res=getxP(nodes0);
	    if(flag_buy){for(int k=0;k<=nodes0.length;k++) res[k]=(k>=nodes0.length?0:p0*nodes0[k])-res[k];}
	    else{for(int k=0;k<=nodes0.length;k++) res[k]=res[k]-(k>=nodes0.length?0:p0*nodes0[k]);}
	    return res;
	}
    };
}

void test_Pol(final BookItem [] tb,final boolean flag_buy,final double p0,final double k00,final double [] kk,final double kk_p_offset)
{
    final CalculateKendall.TradeIPData d=
	    CalculateKendall.TradeIPData.calculateTradeIPData(2*neq+1,
							      currenttime,
							      tau,
							      tb,
							      tb.length,
							      null,
							      true);
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getOrtPolBasis(p0-d.p_offset,flag_buy);
    final double [][] p=M.getOrthogonalPolynomialsFirstKind(neq+1,d.moments_pv);
    final Quadrature xx=M.getRadauQuadratureFromPolynomials(neq-1,p);
    final double [] x=xx.x,w=xx.w;
    if(w.length!=neq) throw new RuntimeException(" w.length="+w.length+" neq="+neq);
    final double w0=w[w.length-1],x0=x[x.length-1]+d.p_offset;
    final double [] K=new double [2*neq-1];
    for(int k=0;k<neq;k++)
    {
	DeprecatedHelperTMatr.sumP2(K,p[k]);
    }

    UnitTests.checkDoubleRel(w0,"w0",1/k00,"1/k00",eps);
    UnitTests.checkDoubleRel(x0,"x0",p0,"p0",eps);
    UnitTests.checkDoubleArrayRel(Polynomials.getPolynomialShifted(kk,-(kk_p_offset-d.p_offset)),"kk",K,"K",eps);

    System.err.println(name+":QVMFTest:test_Pol(): test OK neq="+neq+" eps="+eps);
}

void test_OrtPol()
{
    final QVMF<DTYPE> q=getQVMF();
    test_Pol(tbuy,true,q.pbuy,q.k0buy_pbuy,q.Kbuy,q.poffset_buy);
    test_Pol(tsell,false,q.psell,q.k0sell_psell,q.Ksell,q.poffset_sell);

    System.err.println("pbuy="+q.pbuy+" psell="+q.psell+" peq="+q.peq+
		       "\nk0buy_peq="+q.k0buy_peq+" k0sell_peq="+q.k0sell_peq);
    if(q.peq>q.pbuy) test_Pol(tbuy,true,q.peq,q.k0buy_peq,q.Kbuy,q.poffset_buy);
    else System.err.println("Skipping buy test:peq="+q.peq+" pbuy="+q.pbuy);
    if(q.peq<q.psell) test_Pol(tsell,false,q.peq,q.k0sell_peq,q.Ksell,q.poffset_sell);
    else System.err.println("Skipping sell test:peq="+q.peq+" psell="+q.psell);

    System.err.println(name+":QVMFTest:test_OrtPol(): test OK neq="+neq+" eps="+eps);
}

void test_Calculations(final int tbuysize,final int tsellsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":QVMFTest:test_Calculations(): test with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);

    setData(tbuysize,tsellsize,neq);
    test_Common();
    test_QATradeIPBQVM();
    test_OrtPol();

    System.err.println(name+":QVMFTest:test_Calculations(): test OK with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":QVMFTest: starting tests; ver=$Id: QVMF.java,v 1.42 2020/09/27 20:33:20 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(850),1950+r.nextInt(850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":QVMFTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: QVMF.java,v 1.42 2020/09/27 20:33:20 mal Exp $");
}
} // END OF class QVMFTest


} // END of class QVMF
