package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Quadrature;
import com.polytechnik.realtimebook.*;

/** Test Skewness.
 *  @version $Id: RkQVM.java,v 1.32 2022/06/27 10:50:04 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class RkQVM<DTYPE> {

final DTYPE odata;
final DataInterfaceToMoments<DTYPE> access;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final Skewness sk;
final EVXSpectrum PA;
final double plo,Pi,Pt;

RkQVM(final int n,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();
    this.M=getABasis();

    final int nmoms=2*n-1;
    final double [] p0imom=new double [nmoms],p1imom=new double [nmoms],p2imom=new double [nmoms],p3imom=new double [nmoms];
    access.setM(p0imom,0,access.M_I,odata,nmoms);
    access.setM(p1imom,1,access.M_I,odata,nmoms);
    access.setM(p2imom,2,access.M_I,odata,nmoms);
    access.setM(p3imom,3,access.M_I,odata,nmoms);

    final double [] p0tmom=new double [nmoms],p1tmom=new double [nmoms];
    access.setM(p0tmom,0,access.M_T,odata,nmoms);
    access.setM(p1tmom,1,access.M_T,odata,nmoms);


    final double [] P0imom=getFmom(p0imom),P1imom=getFmom(p1imom),
	P2imom=getFmom(p2imom),P3imom=getFmom(p3imom);

    this.PA=EVXSpectrum.getEVXSpectrumFromQQ(n,M.B.getQQMatr(n,P1imom),M.B.getQQMatr(n,P0imom),M);

    this.plo=getPLast()-getPOffset();

    //System.err.println("plo="+plo+" sL="+PA.sL+" sH="+PA.sH);
    final double [] psil=getPsi(this.PA,this.plo);

    final double m0=TMatr.mulVVn1(psil,P0imom),
	m1=TMatr.mulVVn1(psil,P1imom),
	m2=TMatr.mulVVn1(psil,P2imom),
	m3=TMatr.mulVVn1(psil,P3imom);
    //System.err.println("m0="+m0+" m1="+m1+" m2="+m2+" m3="+m3);
    this.sk=Skewness.getSkewnessFromMonomialMoments(m0,m1,m2,m3);

    final double [] P0tmom=getFmom(p0tmom),P1tmom=getFmom(p1tmom);
    this.Pi=TMatr.mulVVn1(psil,P1imom)/TMatr.mulVVn1(psil,P0imom);
    this.Pt=TMatr.mulVVn1(psil,P1tmom)/TMatr.mulVVn1(psil,P0tmom);

}


double [] getFmom(final double [] moms)
{
    final double [] fmoms=new double [moms.length];
    access.setFMoments(fmoms,moms,fmoms.length);
    TMatr.scaleVectorInPlace(fmoms,-1);
    return fmoms;
}

double [] getPsi(final EVXSpectrum e,final double lamx)
{
    //return e.getPsiLamX(lamx);
    //return e.getPsi2LamX(lamx);
    final double [] pp=e.getPsiLamX(lamx);
    return M.B.mult2Pol(pp,pp);
}


abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double getPOffset();
abstract double getPLast();



static abstract class RkQVMTest<DTYPE> extends QTest {

abstract RkQVM<DTYPE> getRkQVM();


RkQVMTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}

void test_quad()
{
    final RkQVM<DTYPE> q=getRkQVM();
    final int nmoms=2*q.PA.neq-1,n=q.PA.neq-1;
    final double [] sp0imom=new double [nmoms];
    q.access.setM(sp0imom,0,q.access.M_I,q.odata,nmoms);

    final double [] moms=q.getFmom(sp0imom);
    final double [] Xmoms=q.M.B.getXQkMomentsFromQkMoments(moms);
    final double [] XQQ=q.M.B.getQQMatr(n,Xmoms),QQ=q.M.B.getQQMatr(n,moms);
    final EVXSpectrum es=EVXSpectrum.getEVXSpectrumFromQQ(n,XQQ,QQ,q.M);
    final Quadrature Q=q.M.getGaussQuadratureFromMoments(n,moms);
    final double [] x=Q.x,w=Q.w;
    for(int k=0;k<n;k++)
    {
	{
	final double [] psi=es.getPsi(k);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psi,moms),"<psi_k>",es.wpsi[k],"es.wpsi[k]",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psi,Xmoms)/TMatr.mulVVn1(psi,moms),"<Xpsi_k>/<psi_k>",x[k],"x[k]",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,QQ,n),"<psi_xk|1|psi_xk>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,XQQ,n),"<psi_xk|X|psi_xk>",x[k],"x[k]",eps);
	}
	{
	final double [] psil=es.getPsiLamX(x[k]);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil,moms),"<psil_xk>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil,Xmoms),"<Xpsil_xk>",x[k],"x[k]",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psil,XQQ,n),"<psil_xk|X|psil_xk>",x[k]/w[k],"x[k]/w[k]",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psil,QQ,n),"<psil_xk|1|psil_xk>",1/w[k],"1/w[k]",eps);
	}
	{
	final double [] psil2=es.getPsi2LamX(x[k]);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil2,moms),"<psil2_xk>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil2,Xmoms),"<Xpsil2_xk>",x[k],"x[k]",eps);
	}
    }
    System.err.println(name+":RkQVMTest:test_quad(): test OK n="+n+" eps="+eps);
}

void test_F()
{
    final RkQVM<DTYPE> q=getRkQVM();
    final int nmoms=2*q.PA.neq-1,n=q.PA.neq;
    final double [] jmoms=new double [nmoms],jXmoms=new double [nmoms];
    q.access.setM(jmoms,0,q.access.M_I,q.odata,nmoms);
    q.access.setM(jXmoms,1,q.access.M_I,q.odata,nmoms);
    final double [] moms=q.getFmom(jmoms),Xmoms=q.getFmom(jXmoms);
    final double [] XQQ=q.M.B.getQQMatr(n,Xmoms),QQ=q.M.B.getQQMatr(n,moms);

    final EVXSpectrum es=q.PA;
    for(int k=0;k<n;k++)
    {
	final double ek=es.getE(k);
	{
	final double [] psi=es.getPsi(k);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psi,moms),"<psi_k>",es.wpsi[k],"es.wpsi[k]",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psi,Xmoms)/TMatr.mulVVn1(psi,moms),"<Xpsi_k>/<psi_k>",ek,"ek",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,QQ,n),"<psi_xk|1|psi_xk>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi,XQQ,n),"<psi_xk|X|psi_xk>",ek,"ek",eps);
	}
	{
	final double [] psil=es.getPsiLamX(ek);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil,moms),"<psil_xk>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil,Xmoms),"<Xpsil_xk>",ek,"ek",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psil,XQQ,n),"<psil_xk|X|psil_xk>",ek/Math.pow(es.wpsi[k],2),"ek/wpsi[k]^2",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psil,QQ,n),"<psil_xk|1|psil_xk>",1/Math.pow(es.wpsi[k],2),"1/wpsi[k]^2",eps);
	}
	{
	final double [] psil2=es.getPsi2LamX(ek);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil2,moms),"<psil2_xk>",1,"1",eps);
	UnitTests.checkDoubleRel(TMatr.mulVVn1(psil2,Xmoms),"<Xpsil2_xk>",ek,"ek",eps);
	}
    }

    final double [] psi_plo=q.PA.getPsiLamX(q.plo);
    if(q.PA.sL<=q.plo && q.plo<=q.PA.sH)  System.err.println("INSIDE");
    System.err.println("q.plo="+q.plo+"\t<psi(plo)|QQF|psi(plo)>/<psi(plo)|QQ|psi(plo)>="+(TMatr.getAMatrA(psi_plo,XQQ,n)/TMatr.getAMatrA(psi_plo,QQ,n))+
		       "\t<psi(plo)F>/<psi(plo)>="+(TMatr.mulVVn1(psi_plo,Xmoms)/TMatr.mulVVn1(psi_plo,moms))+
		       "\t<psi_plo>="+TMatr.mulVVn1(psi_plo,moms));
    System.err.println("q.sk.xa="+q.sk.xa+"\tq.sk.xm="+q.sk.xm);
    System.err.println(name+":RkQVMTest:test_F(): test OK n="+n+" eps="+eps);
}


void test_Common()
{
    final RkQVM<DTYPE> q=getRkQVM();
    final double o=q.getPOffset();
    System.err.println("q.plo="+q.plo+" q.PA.sL="+q.PA.sL+" q.PA.sH="+q.PA.sH);


    System.err.println(name+":RkQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}


void test_Extra(){}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":RkQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_quad();
    test_F();
    test_Extra();

    System.err.println(name+":RkQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":RkQVMTest: starting tests; ver=$Id: RkQVM.java,v 1.32 2022/06/27 10:50:04 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":RkQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: RkQVM.java,v 1.32 2022/06/27 10:50:04 mal Exp $");
}
} // END OF class RkQVMTest




}// END OF class RkQVM
