package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.PolynomialRootsFinder;
import com.polytechnik.utils.TMatr;

/**  Utilities for S.
 *  @version $Id: SM.java,v 1.35 2018/07/06 08:01:41 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class SM {

    static final int ATYPE_FMIN=20,ATYPE_FAVER=21,ATYPE_FLAST=22,ATYPE_FDP=23,ATYPE_FDPL=24,ATYPE_FMINA=25;
/** An option for alfa calculation type.
 */
static final int ALFA_TYPE=ATYPE_FMIN;

/** Minimal I (used for stability. */
static final double MIN_I_FACT=0.2;

/** Think later about <T>/<U> analogue.
 *  The value 2 correspond to scale equal to sqrt(sigma2_eq),
 *  1 corresponds to 0.5*sqrt(sigma2_eq).
 *  The QVMpi still has VIRIAL_FACTOR broken.
 */
static final double VIRIAL_FACTOR=2;

private static double getS(final double x,final double [] F,final double [] N,final double alfa)
{
    final double f=Polynomials.getD0(F,x);
    return f-alfa*(0+Polynomials.getD0(N,x)/f);
}
private static double [] getDSPol(final double [] F,final double [] N,final double alfa)
{
    final double a=F[2],b=F[1],c=F[0],aa=a*a,bb=b*b,ac=a*c;
    final double [] p=new double []{
	c*c*b,2*c*(bb+ac),b*(bb+6*ac),4*a*(bb+ac),5*aa*b,2*aa*a
    };
    final double n1=N[1],n0=N[0];
    p[0]-=alfa*(n1*c-b*n0);
    p[1]-=alfa*(-2*a*n0);
    p[2]-=alfa*(-n1*a);
    return p;
}
static double getPSmin(final double [] F,final double [] N,final double alfa,
		       final com.polytechnik.utils.PolynomialRootsFinder rf)
{
    final double [] p=getDSPol(F,N,alfa);
    final PolynomialRootsFinder.PRoots u=rf.findRoots(p);
    final double [] root_re=u.re_root,root_im=u.im_root;

    double smin=Double.POSITIVE_INFINITY;int ind_smin=-1;
    for(int k=0;k<u.nroots;k++)
    {
	final double sk=getS(root_re[k],F,N,alfa);
	if(sk<=smin)
	{
	    smin=sk;
	    ind_smin=k;
	}
    }
    if(ind_smin>=0)
    {
	return root_re[ind_smin];
    }
    return Double.NaN;
}
private static void test_S(final int ntest,final java.util.Random r,final double eps)
{
    final com.polytechnik.utils.PolynomialRootsFinder rf=
	new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();
    for(int j=0;j<ntest;j++)
    {
	final double x0=1-2*r.nextDouble();
	final double [] F=new double []{x0*x0+r.nextDouble()+0.13,-2*x0,1},N=new double []{(1-2*r.nextDouble()),(1-2*r.nextDouble())};
	final double alfa=r.nextDouble()+0.2;

	final double [] p=getDSPol(F,N,alfa);
	final double [] tp=
	    Polynomials.mult2Pol(Polynomials.getPolynomialDifferentiated(F),Polynomials.mult2Pol(F,F));
	TMatr.sumP1(tp,Polynomials.mult2Pol(Polynomials.getPolynomialDifferentiated(N),F),-alfa);
	TMatr.sumP1(tp,Polynomials.mult2Pol(Polynomials.getPolynomialDifferentiated(F),N),alfa);
	UnitTests.checkDoubleArrayRel(p,"p",tp,"tp",eps);

	final double pm=getPSmin(F,N,alfa,rf);
	UnitTests.checkDoubleRel(Polynomials.getD0(p,pm),"p(pm)",0,"0",eps);
	final double dSm=Polynomials.getD1(F,pm)-alfa*(Polynomials.getD1(N,pm)*Polynomials.getD0(F,pm)-Polynomials.getD1(F,pm)*Polynomials.getD0(N,pm))/Math.pow(Polynomials.getD0(F,pm),2);
	UnitTests.checkDoubleRel(dSm,"dSm",0,"0",eps);

	for(int k=0;k<100;k++)
	{
	    final double px=r.nextDouble();
	    final double S=getS(px,F,N,alfa);
	    final double tS=Polynomials.getD0(F,px)-alfa*(Polynomials.getD0(N,px)/Polynomials.getD0(F,px));
	    UnitTests.checkDoubleRel(S,"S",tS,"tS",eps);
	    final double dS=Polynomials.getD0(p,px)/Math.pow(Polynomials.getD0(F,px),2);
	    final double tdS=Polynomials.getD1(F,px)-alfa*(Polynomials.getD1(N,px)*Polynomials.getD0(F,px)-Polynomials.getD1(F,px)*Polynomials.getD0(N,px))/Math.pow(Polynomials.getD0(F,px),2);
	    UnitTests.checkDoubleRel(dS,"dS",tdS,"tdS",eps);
	}
    }
    System.err.println("SM:test_S: tests OK, ntest="+ntest+" eps="+eps);
}


public static void main(String [] args)
{
    System.err.println("SM::main(): starting ALL tests; ver=$Id: SM.java,v 1.35 2018/07/06 08:01:41 mal Exp $");
    final long t0=System.currentTimeMillis();

    final double eps=1e-7;
    final java.util.Random r=new java.util.Random(-1239L);

    test_S(1400000,r,eps);

    final long t1=System.currentTimeMillis();
    System.err.println("SM::main(): ALL tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: SM.java,v 1.35 2018/07/06 08:01:41 mal Exp $");
}

}

