package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.Laguerre;
import com.polytechnik.utils.PolynomialRootsFinder;
import com.polytechnik.realtimebook.*;

/** QVM calculations.
 *  @version $Id: SQVMLaguerre.java,v 1.44 2022/06/27 13:32:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class SQVMLaguerre extends SQVM<QVMDataL> {

    static final Laguerre B=new Laguerre();

SQVMLaguerre(final int neq,final QVMDataL odata)
{
    super(neq,odata);
}
DeprecatedHelperQVMDataLDirectAccessWithBasis getAccess(){return new DeprecatedHelperQVMDataLDirectAccessWithBasis();}
double getPOffset(){return odata.p_offset;}
public static SQVMLaguerre calculate(final int neq,
					    final long currenttime,
					    final long tau,
					    final Trade [] tb,
					    final int tsize,
					    final SQVMLaguerre ref)
{
    final int npm=3;
    final QVMDataL o=QVMDataL.calculateQVMDataL(2*neq,
						npm,
						currenttime,
						tau,
						tb,
						tsize,
						ref!=null?ref.odata:null);
    return new SQVMLaguerre(neq,o);
}

/** Calculate all nodes.
 */
double [] getXNodes(final double [] nodespolynomial)
{
    final int n=nodespolynomial.length-1;
    // /* Option 1: use Jenkins Traub in monomials space. Good enough for n<15. */
    //final double [] npol=Laguerre.convertToPolynomialFromLaguerre_Test(n,nodespolynomial);
    //final int nroots=new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG(n).findRoots(n,npol,x,npol);
    // /* Option 2: use Confederate matrix, very stable but slow. */
    //final int nroots=new com.polytechnik.utils.PolynomialRootsConfederateMatrix.ABasisLaguerre().findRoots(n,nodespolynomial,x,new double [n]);
    // /* Option 3: Newtonian iterations & synthetic division. Good enough for simple roots and n<20. */
    final PolynomialRootsFinder.PRoots u=
	new com.polytechnik.outils.LaguerrePolynomialRootsFinder().findRoots(nodespolynomial);
    final double [] x=u.re_root;
    if(u.nroots!=n) java.util.Arrays.fill(x,Double.NaN);
    com.polytechnik.utils.SortArrays.sortArrayWithLoad(x.length,x,null);
    return x;
}
double [] getWeights(final double [] x,final double [] nodespolynomial,final double [] bmoms)
{
    final double [] w=new double [x.length];
    final int n=nodespolynomial.length-1;
    for(int k=x.length;--k>=0;)
    {
	final double xk=x[k];
	final double [] tmp=B.sdiv1(nodespolynomial,xk).q;
	w[k]=TMatr.mulVVn1(tmp,bmoms)/fx(n,tmp,xk);
    }
    return w;
}
double [] getWeights2(final double [] x,final double [] nodespolynomial,final double [] bmoms)
{
    final double [] w=new double [x.length];
    final int n=nodespolynomial.length-1;
    for(int k=x.length;--k>=0;)
    {
	final double xk=x[k];
	final double [] tmp=B.sdiv1(nodespolynomial,xk).q;
	final double t0=fx(n,tmp,xk);
	for(int m=tmp.length;--m>=0;) tmp[m]/=t0;
	final double [] mul2=B.mult2Pol(tmp,tmp);
	w[k]=TMatr.mulVVn1(mul2,bmoms);
    }
    return w;
}
double fx(final int nm,final double [] m,final double x)
{
    if(m.length!=nm) throw new RuntimeException();
    return B.getD0(m,x);
}
static double [] getPData(final QVMData q,final int n,final double p_offset)
{
    return MQVMLaguerre.MQVMLaguerreTest.getMonomialMoments(q.ntm,q,n,QVMData.MOM_dV,p_offset);
}

public static void main(String [] args)
{
    new SQVMTest<QVMDataL>("SQVMLaguerre",1e-6,1e-3){
	void setData(final int tsize,final java.util.Random r,final int neq)
	{
	    this.neq=neq;
	    this.r=r;
	    this.tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
	    this.currenttime=tb[tb.length-1].time+1000;
	    this.tau=(tb[tb.length-1].time-tb[0].time)/24;
	}
	SQVMLaguerre getSQVM()
	{
	    return calculate(neq,currenttime,tau,tb,tb.length,null);
	}

	QVMData getData(final SQVMLaguerre q)
	{
	    final QVMData qdata=QVMData.calculateQVMData(1,q.odata.ntm+r.nextInt(4),4+r.nextInt(3),currenttime,tau,tb,tb.length,null);
	    return qdata;
	}

	void matchMoments()
	{
	    System.err.println(name+":SQVMLaguerreTest:matchMoments(): neq="+neq);
	    final SQVMLaguerre q=getSQVM();
	    final QVMData s=getData(q);
	    matchPt(q,getPData(s,0,s.p_offset),getPData(s,1,s.p_offset),getPData(s,2,s.p_offset),false,s.p_offset);


	    // TO REMOVE LATER
	    final double [] mi=getPData(s,0,s.p_offset);
	    final double [] mp=getPData(s,1,s.p_offset),mp2=getPData(s,2,s.p_offset);
	    final double [] rnodes=getNodesPolynomialBasis(neq,mi);
	    {
		final double [] l=B.convertFunctionTo_Basis_from_Monomials(rnodes);
		final double f=q.rnodes[q.rnodes.length-1]/l[l.length-1];
		for(int k=0;k<l.length;k++)l[k]*=f;
		UnitTests.checkDoubleArrayRel(q.rnodes,"q.rnodes",l,"l",eps);
	    }

	    checkXW(q,rnodes,mi,mp,mp2,s.p_offset,false);

	    System.err.println(name+":SQVMLaguerreTest:matchMoments(): ALL TESTS OK, neq="+neq+" eps="+eps);
	}
    }.runTests(6);
}


} // END OF class SQVMLaguerre
