package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.utils.EVXData;
import com.polytechnik.freemoney.WIntegrator;

/**
 * @version $Id: ScalpedMaxIProjection.java,v 1.195 2022/07/28 08:04:53 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public abstract class ScalpedMaxIProjection<DTYPE> {


/** Internal state.
 * It preserves the scalp--moments of the previous call
 * to recurrently speed up the calculations.
 */
public final ScalpedMaxIProjectionStateSaver sst;

/** "Regular", not scalp--moments. The object has
 * an internal state to recurrently speed up the calculations.
 */
final DTYPE odata;

public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getABasis();

/** The |I|psi>=lambda |psi> solution.
 */
public final EVXData I;

public final double
// Helper data and moving averages.
    p_offset,p_last,pi_average,pt_average,
// The https://arxiv.org/abs/1510.05510 results (the state of max I).
    dpdt_0,p_0,pt_0,p_IH,pt_IH,pV_IH,pT_IH,dpdt_IH,var1pI_IH,var1pI_IH_00,
// The https://arxiv.org/abs/1903.11530 results
    pmin_0_IH,pmax_0_IH,Skewness_0_IH,ProbabilityCorrelation_0_IH,
// Integration term
    Fdt,dIH,dp_IH,
// The https://arxiv.org/abs/1903.11530 results with scalp--function.
    DIR,DIRa,F_IH,Fa_IH;

final VolMatrix VOLATILITY_EDPsi=this::getEDPsi;

public static final int
    F_SAMPLE_DP_NOSCALP=351,
    F_SAMPLE_DP_SCALP=352,
    F_dpdt0_NOSCALP=353,
    F_dpdt0_SCALP=354,
    F_dpdtIH_SCALP=355,
    F_pipt0_SCALP=356,
    F_varpIH_SCALP=357,
    F_varpIH_divI_SCALP=358,
    F_varpIH_0_SCALP=359,
    F_varpIH_0_divI_SCALP=360,
    F_P0_PIH_SCALP=361,
    F_SKEWNESS_at_Pl_SCALP=362,
    F_PROBABILITYCORRELATION_SCALP=363,
    F_PROBABILITYCORRELATION_VOLUME_SCALP=364,
    F_SAMPLE_DPIH_SCALP=365,
    F_SAMPLE_DIgt0DPIH_SCALP=366,
    F_SAMPLE_DIgt0DPIH_NOSCALP=367,
    F_SAMPLE_DVDPIH_SCALP=368,
    F_SAMPLE_DVDPIH_NOSCALP=369;
/** Determines what Fl model to use.
 */
public final int dp_to_use=F_dpdt0_SCALP;

/** Not worth a function or constructor argument.
 */
private static double EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0,EPS_KLUDGY_WAY_TEST_DP=0;

/**
* Constructor.
* @param n Basis dimension.
* @param odata Regular moments internal state.
* @param rst Previous tick internal state.
*/
ScalpedMaxIProjection(final int n,final DTYPE odata,final ScalpedMaxIProjectionStateSaver rst)
{

    final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING; // turn on selftests if required
    this.odata=odata;
    final SMomentsData moms=getSMoments(2*n-1),mm=moms.asQQMatrices(n,M);
    // mm has the matrices (not moments) stored.
    final double [] QQt=mm.tmom,QQi=mm.imom,QQpi=mm.pimom,QQdpdt=mm.dpmom,QQa=mm.amom,
	QQp1t=mm.p1tmom,QQp0t=mm.p0tmom,
	/* they have negative sign convention. */
	QQT=TMatr.getVectorScaled(mm.Tmom,-1),
	QQp1T=TMatr.getVectorScaled(mm.p1Tmom,-1),QQp0T=TMatr.getVectorScaled(mm.p0Tmom,-1),
	QQV=TMatr.getVectorScaled(mm.Vmom,-1),QQpV=TMatr.getVectorScaled(mm.pVmom,-1);
    p_offset=mm.p_offset;
    p_last=mm.p_last;
    pi_average=QQpi[0]/QQi[0];
    pt_average=QQp1t[0]/QQp0t[0];
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    double [] psiH=I.getPsi(n-1/* max I */),psi0=I.psi0;
    p_0=TMatr.getAMatrA(psi0,QQpi,n)/TMatr.getAMatrA(psi0,QQi,n);
    pt_0=TMatr.getAMatrA(psi0,QQp1t,n)/TMatr.getAMatrA(psi0,QQp0t,n);
    dpdt_0=TMatr.getAMatrA(psi0,QQdpdt,n);

    final double pI_IH=TMatr.getAMatrA(psiH,QQpi,n);
    p_IH=pI_IH/TMatr.getAMatrA(psiH,QQi,n);
    pt_IH=TMatr.getAMatrA(psiH,QQp1t,n)/TMatr.getAMatrA(psiH,QQp0t,n);
    pV_IH=TMatr.getAMatrA(psiH,QQpV,n)/TMatr.getAMatrA(psiH,QQV,n);
    pT_IH=TMatr.getAMatrA(psiH,QQp1T,n)/TMatr.getAMatrA(psiH,QQp0T,n);
    dpdt_IH=TMatr.getAMatrA(psiH,QQdpdt,n);
    final double [] DpsiH=VOLATILITY_EDPsi.getDPsi(psiH);
    final double psiH_DpsiH=TMatr.getAMatrB(psiH,DpsiH,QQt,n),psiH_psi0=TMatr.getAMatrB(psiH,psi0,QQt,n),
	psiH_pI_psi0=TMatr.getAMatrB(psiH,psi0,QQpi,n),var1pI_IH_0=2*(psiH_pI_psi0-pI_IH*psiH_psi0);
    var1pI_IH_00=var1pI_IH_0*psiH_psi0;
    var1pI_IH=2*(TMatr.getAMatrB(psiH,DpsiH,QQpi,n)-pI_IH*psiH_DpsiH);
    if(eps>0){
	final double [] rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(psiH,DpsiH,null,QQpi,QQt,n);
	UnitTests.checkDouble(rq[1],"rq[1]",var1pI_IH,"var1pI_IH",eps);
	UnitTests.checkDouble(Math.pow(M.B.getD0(psiH,I.x0),2),"psiH(x0)^2",2*psiH_DpsiH,"2*psiH_DpsiH",eps);
	UnitTests.checkDouble(M.B.getD0(psiH,I.x0),"psiH(x0)",psiH_psi0*Math.sqrt(I.o_o),"psiH_psi0*sqrt(o_o)",eps);
	final double [] rq0=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(psiH,psi0,null,QQpi,QQt,n);
	UnitTests.checkDouble(rq0[1],"rq0[1]",var1pI_IH_0,"var1pI_IH_0",eps);
	UnitTests.checkDouble(psiH_psi0,"psiH_psi0",I.wH,"I.wH",eps);
	//if(I.wH<0) throw new RuntimeException("test: projection sign may be negative I.wH="+I.wH+", take the square");
    }

    // Price eigenvalues in the basis of {|psi0>, |psiH>}
    final int d=2;
    final double [] Mpi={p_0*I.s0,psiH_pI_psi0,psiH_pI_psi0,pI_IH},
	Mi={I.s0,I.sH*psiH_psi0,I.sH*psiH_psi0,I.sH},
	evp=M.EV.getGEV_Generalized(d,Mpi,Mi,true),
	_phiPL=EVXData.getPsiFromBasis(0,evp,d),_phiPH=EVXData.getPsiFromBasis(d-1,evp,d),_phi0={M.B.getD0(psi0,I.x0),M.B.getD0(psiH,I.x0)};
    pmin_0_IH=evp[0];
    pmax_0_IH=evp[d-1];
    // Skewness like characteristics in {|psi0>, |psiH>} basis.
    Skewness_0_IH=(2*p_last-pmin_0_IH-pmax_0_IH)/(pmin_0_IH-pmax_0_IH);
    final double phiPL2_atx0=Math.pow(TMatr.mulVV(_phi0,_phiPL),2),
	phiPH2_atx0=Math.pow(TMatr.mulVV(_phi0,_phiPH),2);
    // Probability correlation in {|psi0>, |psiH>} basis: P & distance to x0.
    ProbabilityCorrelation_0_IH=(phiPH2_atx0-phiPL2_atx0)/(phiPH2_atx0+phiPL2_atx0);

    if(eps>0 && rst!=null && rst.seq>10){
	final double [] phiPL=TMatr.sumaAbB(_phiPL[0],psi0,_phiPL[1],psiH),
	    phiPH=TMatr.sumaAbB(_phiPH[0],psi0,_phiPH[1],psiH);
	UnitTests.checkDouble(TMatr.getAMatrA(_phiPL,Mi,2),"<phiPL|I|phiPL>",1,"1",eps);
	UnitTests.checkDouble(TMatr.getAMatrA(_phiPH,Mi,2),"<phiPH|I|phiPH>",1,"1",eps);
	UnitTests.checkDouble(TMatr.getAMatrA(phiPL,QQpi,n),"<phiPL|pI|phiPL>",evp[0],"evp[0]",eps);
	UnitTests.checkDouble(TMatr.getAMatrA(phiPH,QQpi,n),"<phiPH|pI|phiPH>",evp[d-1],"evp[d-1]",eps);
	UnitTests.checkDouble(Math.pow(M.B.getD0(phiPL,I.x0),2),"phiPL(x0)^2",phiPL2_atx0,"phiPL2_atx0",eps);
	UnitTests.checkDouble(Math.pow(M.B.getD0(phiPH,I.x0),2),"phiPH(x0)^2",phiPH2_atx0,"phiPH2_atx0",eps);
    }

    // Preserve the state, to obtain the "scalped" moments.
    // For "regular" moments the DTYPE also has an internal state :(
    this.sst=new ScalpedMaxIProjectionStateSaver(rst,mm.currenttime,p_last+p_offset /* All prices are relatively p_offset */,mm.totalvolume-(rst!=null?rst.allshares:0),p_IH+p_offset,I.sH,Math.pow(I.wH,2),this::getEmptyWIntegrator);
    final double dt,dv;
    if(rst==null) {dt=Fdt=dIH=dp_IH=0;dv=sst.allshares-0;}
    else {
	dv=(sst.allshares-rst.allshares);
	if(rst.time<=0) dt=Fdt=dIH=dp_IH=0;
	else {
	    dt=(sst.time-rst.time)/sst.wi.atau;
	    dIH=sst.IH-rst.IH;
	    dp_IH=sst.p_IH-rst.p_IH;
	    // I.o_o is a constant, some Fdt values are multiplied by it for proper scale
	    if(dp_to_use==F_SAMPLE_DP_NOSCALP){
		Fdt=sst.P-rst.P; // dP without a scalp--function
	    } else if(dp_to_use==F_SAMPLE_DP_SCALP){
		Fdt=(sst.P-rst.P)*sst.scalp; // dP*sst.scalp
	    } else if(dp_to_use==F_PROBABILITYCORRELATION_VOLUME_SCALP){
		Fdt=dv * /*Math.abs(sst.P-rst.P)*/ ProbabilityCorrelation_0_IH*sst.scalp;
	    } else if(dp_to_use==F_SAMPLE_DPIH_SCALP){
		Fdt=(dp_IH)*sst.scalp;
	    } else if(dp_to_use==F_SAMPLE_DIgt0DPIH_SCALP){
		Fdt=(dIH>0?1:0)*(dp_IH)*sst.scalp;
	    } else if(dp_to_use==F_SAMPLE_DIgt0DPIH_NOSCALP){
		Fdt=(dIH>0?1:0)*(dp_IH);
	    } else if(dp_to_use==F_SAMPLE_DVDPIH_SCALP){
		Fdt=dv*(dIH>0?1:0)*(dp_IH)*sst.scalp;
	    } else if(dp_to_use==F_SAMPLE_DVDPIH_NOSCALP){
		Fdt=dv*(dIH>0?1:0)*(dp_IH);
	    } else {
		Fdt=dt*getFlFromRegularMoments();
	    }
	}
    }
    sst.updateState(sst.time,dt*sst.scalp,!Double.isFinite(Fdt)?0:Fdt,dv*sst.scalp,dv);

    final double [] psiWeigth=getPsi2WIntegratedDt(/* I.psiH by default. */getPsiToEstimateScalp(n,QQt,QQi,QQpi,M));
    this.DIR=TMatr.mulVVn1(psiWeigth,sst.getFMoments());
    this.DIRa=TMatr.mulVVn1(psiWeigth,sst.getAbsFMoments());
    this.F_IH=TMatr.getAMatrA(psiH,M.B.getQQMatr(n,sst.getFMoments()),n);
    this.Fa_IH=TMatr.getAMatrA(psiH,M.B.getQQMatr(n,sst.getAbsFMoments()),n);


    // Unit tests below. This approach is often more numerically stable, but more elaborate code.
    if(eps>0){
	// Obtain the scalped data.
	final DataInterfaceToMoments<DTYPE> access=getAccess();
	final double [] jFmoments=new double [sst.wi.ntm],jAbsFmoments=new double [sst.wi.ntm];
	// The Fmoments moments to match exactly the moms.dpmom for dp_to_use==F_SAMPLE_DP_NOSCALP
	// Integrate by parts, to obtain prices from dp/dt
	access.setFMoments(jFmoments,sst.getFMoments(),sst.wi.ntm);
	access.setFMoments(jAbsFmoments,sst.getAbsFMoments(),sst.wi.ntm);
	// Obtain the n x n matrices <QjQkF> from 2*n-1 moments <QmF>, m=0..2n-2.
	final double []
	    // sign convention
	    QQjF=TMatr.scaleVectorInPlace(M.B.getQQMatr(n,jFmoments),-1),
	    QQjaF=TMatr.scaleVectorInPlace(M.B.getQQMatr(n,jAbsFmoments),-1);
	// equals to p_last-pt_IH for dp_to_use==F_dpdt0
	final double _DIR=TMatr.getAMatrA(psiH,QQjF,n),
	    _DIRa=TMatr.getAMatrA(psiH,QQjaF,n);
	// match the result above with psi^2 integration directly
	UnitTests.checkDoubleRel(DIR,"DIR",_DIR,"_DIR",eps);
	UnitTests.checkDoubleRel(DIRa,"DIRa",_DIRa,"_DIRa",eps);
    }

    if(eps>0){
	UnitTests.checkDoubleRel(mm.totalvolume,"mm.totalvolume",sst.getSumdVorig(),"sst.getSumdVorig()",eps);
	UnitTests.checkDoubleRel(mm.totalvolume,"mm.totalvolume",sst.allshares,"sst.allshares",eps);
    }

    final double eps2=EPS_KLUDGY_WAY_TEST_DP;
    if(eps2>0){
	// imom to match exactly, tmom has left/right sum difference, test it later
	UnitTests.checkDoubleArrayRel(moms.imom,"moms.imom",sst.getdVOriginalMoments(),"sst.getdVOriginalMoments()",eps);
	if(dp_to_use==F_SAMPLE_DP_NOSCALP){
	    UnitTests.checkDoubleArrayRel(moms.dpmom,"moms.dpmom",sst.getFMoments(),"sst.getFMoments()",eps2);
	    UnitTests.checkDoubleArrayRel(moms.amom,"moms.amom",sst.getAbsFMoments(),"sst.getAbsFMoments()",eps2);
	    UnitTests.checkDoubleRel(DIR,"DIR",p_last-pt_IH,"p_last-pt_IH",eps2);
	    System.err.println("ScalpedMaxIProjection: OK n="+n+" eps2="+eps2);
	} else {
	    System.err.println("ScalpedMaxIProjection: dp_to_use="+dp_to_use+" skipping dP test n="+n+" eps2="+eps2);
	}

    }
    if(eps>0){
	System.err.println("ScalpedMaxIProjection: OK n="+n+" eps="+eps);
    }
}

public double getFlFromRegularMoments(){
    final double scalp=sst.scalp;
    if(dp_to_use==F_dpdt0_NOSCALP){
	// Note! the dpdt_0 is calculated above in full sample, no missed price ticks.
	return dpdt_0; // like dP
    } else if(dp_to_use==F_dpdt0_SCALP){
	return dpdt_0*scalp; // like dP*scalp
    } else if(dp_to_use==F_dpdtIH_SCALP){
	return dpdt_IH*scalp; //almost the same as: dpdt_0*scalp
    } else if(dp_to_use==F_pipt0_SCALP){
	return  (p_0-pt_0)*scalp*I.o_o; // similar to: dpdt_0*scalp
    } else if(dp_to_use==F_varpIH_SCALP){
	return var1pI_IH*scalp;  // not a good one
    } else if(dp_to_use==F_varpIH_divI_SCALP){
	return (var1pI_IH/I.sH)*scalp; // also not a good one
    } else if(dp_to_use==F_varpIH_0_SCALP){
	final double scalp_power=1; // oscillates, not a good one for scalp_power=0
	return (var1pI_IH_00)*Math.pow(scalp,scalp_power);
    } else if(dp_to_use==F_varpIH_0_divI_SCALP){
	final double scalp_power=1; // not a good one for scalp_power=0
	return (var1pI_IH_00/I.sH)*Math.pow(scalp,scalp_power);
    } else if(dp_to_use==F_P0_PIH_SCALP){
	return (p_0-p_IH)*scalp*I.o_o;
    } else if(dp_to_use==F_SKEWNESS_at_Pl_SCALP){
	return -(pmax_0_IH-pmin_0_IH)*Skewness_0_IH*scalp;
    } else if(dp_to_use==F_PROBABILITYCORRELATION_SCALP){
	return (pmax_0_IH-pmin_0_IH)*ProbabilityCorrelation_0_IH*scalp;
    } else {
	if(false) throw new IllegalArgumentException("dp_to_use="+dp_to_use);
	return Double.NaN;
    }
}

/** Returns I.psiH by default.
 */
public double [] getPsiToEstimateScalp(final int n,final double [] QQt,final double [] QQi,final double [] QQpi,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M){
    // The default
    if(true) return I.psiH; // unconstrained

    final double [] QQconstraint=TMatr.sumAcB(QQpi,-p_last,QQi);
    // Constrained attempt. May be rather slow.
    if(false)
	return new com.polytechnik.utils.IstatesConditional(n,QQt,QQi,QQconstraint,M,sst.seq>10?EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING:0).psi_M;
    // Localized solution.
    return new com.polytechnik.utils.IstatesConditionalLocalized(n,QQt,QQi,QQconstraint,M,/*(rt)->rt.i,*/sst.seq>10?EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING:0).psi_M;
}

public SMomentsData getSMoments(final int nmoms)
{
    // Whether to use sampled or integrated by parts t-moments.
    final boolean flag_use_samped_t_moments=false;
    return SMomentsData.get_MomentsData(nmoms,odata,getAccess(),flag_use_samped_t_moments);
}
// Basis-dependence hooks.
abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double [] getEDPsi(final double [] psi);
/** Square the wavefunction, then integrate by parts.
 */
abstract double [] getPsi2WIntegratedDt(final double [] psi);
abstract WIntegrator getEmptyWIntegrator();

/** A class for unit test.
*/
static abstract class ScalpedMaxIProjectionTest<DTYPE,P extends ScalpedMaxIProjection<DTYPE>> extends QTest {

P getScalpedMaxIProjection(){
    // requires a call for every tick for internal WIntegration
    P b=getScalpedMaxIProjectionWithThisRef(0,0,null);
    for(int k=0;k<tb.length;k++) b=getScalpedMaxIProjectionWithThisRef(k+1,tb[k].time,b);
    return getScalpedMaxIProjectionWithThisRef(tb.length,currenttime,b);
}
abstract P getScalpedMaxIProjectionWithThisRef(final int tsize_to,final long currenttime,P ref);

ScalpedMaxIProjectionTest(final String name,final java.util.Random r,final double eps)
{
    super(name,r,eps);
}

void test_Sequence_Calculations(final int tsize,final int neq)
{
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":ScalpedMaxIProjectionTest:test_Sequence_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);
    setData(tsize,neq);
    test_Sequence();
    System.err.println(name+":ScalpedMaxIProjectionTest.test_Sequence_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void test_Sequence()
{
    P b=getScalpedMaxIProjectionWithThisRef(0,0L,null);
    final DataInterfaceToMoments<DTYPE> access=b.getAccess();
    for(int tk=0;tk<tb.length;tk++){
	final int tsize_to=tk+1;

	// Test Integrator
	// direct match from partial sample
	final P bu=getScalpedMaxIProjectionWithThisRef(tsize_to,tb[tk].time,null);
	final int nmoms=bu.sst.wi.ntm;
	final double [] dporig=new double [nmoms],aorig=new double [nmoms],imomorig=new double [nmoms],tmomorig=new double [nmoms];
	access.setM(dporig,1,access.M_DP,bu.odata,nmoms);
	access.setM(aorig,0,access.M_A,bu.odata,nmoms);
	access.setM(imomorig,0,access.M_I,bu.odata,nmoms);
	access.setM(tmomorig,0,access.M_T,bu.odata,nmoms);
	final WIntegrator wi=bu.getEmptyWIntegrator(),wtm=bu.getEmptyWIntegrator();
	wi.updateSavedMomentsToNewTimeReference(tb[tk].time);
	wtm.updateSavedMomentsToNewTimeReference(tb[tk].time);
	for(int j=tk;j>=0;j--){
	    final double dv=tb[j].shares;
	    if(j>=1){
		final double dt=(tb[j].time-tb[j-1].time)/wi.atau,dp=tb[j].price-tb[j-1].price;
		wi.addSingleObservation(tb[tk].time,tb[j].time,new double []{dt,dp,dv,Math.abs(dp),dv});
		wtm.addSingleObservation(tb[tk].time,tb[j-1].time,new double []{dt,dp,dv,Math.abs(dp),dv});
	    } else {
		wi.addSingleObservation(tb[tk].time,tb[j].time,new double []{0,0,dv,0,dv});
	    }
	}
	final double [] Fmoments=wi.integration_moments[1],AbsFmoments=wi.integration_moments[3];
	UnitTests.checkDoubleArrayRel(dporig,"dporig",Fmoments,"Fmoments",eps);
	UnitTests.checkDoubleArrayRel(aorig,"aorig",AbsFmoments,"AbsFmoments",eps);
	UnitTests.checkDoubleArrayRel(imomorig,"imomorig",wi.integration_moments[2],"wi.integration_moments[2]",eps);
	UnitTests.checkDoubleArrayRel(imomorig,"imomorig",wi.integration_moments[4],"wi.integration_moments[4]",eps);
	UnitTests.checkDoubleArrayRel(tmomorig,"tmomorig",wtm.integration_moments[0],"wtm.integration_moments[0]",eps);

	// test the ScalpedMaxIProjection
	EPS_KLUDGY_WAY_TEST_DP=eps;
	b=getScalpedMaxIProjectionWithThisRef(tsize_to,tb[tk].time,b);
	UnitTests.checkDoubleArrayRel(b.sst.getdVOriginalMoments(),"b.sst.getdVOriginalMoments()",imomorig,"imomorig",eps);
	if(b.dp_to_use==F_SAMPLE_DP_NOSCALP){
	    UnitTests.checkDoubleArrayRel(b.sst.getFMoments(),"b.sst.getFMoments()",Fmoments,"Fmoments",eps);
	    UnitTests.checkDoubleArrayRel(b.sst.getAbsFMoments(),"b.sst.getAbsFMoments()",AbsFmoments,"AbsFmoments",eps);
	}
	EPS_KLUDGY_WAY_TEST_DP=0;

	//System.err.println("test_Sequence()="+tk+" tb[tk].price="+tb[tk].price+" b.sst.P="+b.sst.P);
    }
    b=getScalpedMaxIProjectionWithThisRef(tb.length,currenttime,b);
    final P bo=getScalpedMaxIProjection();
    matchScalpedMaxIProjection(b,bo,eps);
}

void test_Extra(){}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":ScalpedMaxIProjectionTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Extra();

    System.err.println(name+":ScalpedMaxIProjectionTest.test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_min,final int neq_max)
{
    System.err.println(name+":ScalpedMaxIProjectionTest: starting tests; ver=$Id: ScalpedMaxIProjection.java,v 1.195 2022/07/28 08:04:53 mal Exp $");
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=eps;
    final long t0=System.currentTimeMillis();
    for(int neq=neq_min;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	    // make at smaller size, this test is slow O(n^2)
	    test_Sequence_Calculations(10+r.nextInt(100),neq);

	}
    }
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;
    final long t1=System.currentTimeMillis();
    System.err.println(name+":ScalpedMaxIProjectionTest.main: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: ScalpedMaxIProjection.java,v 1.195 2022/07/28 08:04:53 mal Exp $");
}
} // END OF class ScalpedMaxIProjectionTest


static void matchScalpedMaxIProjection(final ScalpedMaxIProjection<?> a,final ScalpedMaxIProjection<?> b,final double eps){

    EVXData.matchEVXData(a.I,b.I,false,eps);

    UnitTests.checkDoubleRel(a.dpdt_IH,"a.dpdt_IH",b.dpdt_IH,"b.dpdt_IH",eps);
    UnitTests.checkDoubleRel(a.var1pI_IH,"a.var1pI_IH",b.var1pI_IH,"b.var1pI_IH",eps);
    UnitTests.checkDoubleRel(a.var1pI_IH_00,"a.var1pI_IH_00",b.var1pI_IH_00,"b.var1pI_IH_00",eps);
    UnitTests.checkDoubleRel(a.DIR,"a.DIR",b.DIR,"b.DIR",eps);
    UnitTests.checkDoubleRel(a.DIRa,"a.DIRa",b.DIRa,"b.DIRa",eps);

    System.err.println("ScalpedMaxIProjection:matchScalpedMaxIProjection test OK eps="+eps);
}

}
