package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2020
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.EVXData;

/** Save a state.
 * @version $Id: SimpleStateSaver.java,v 1.1 2020/12/30 09:20:21 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class SimpleStateSaver {

    public final long seq,time,allshares;
    /** The price, it propagates directly from constructor argument.
     */
    public final double P;
    /** The eigenvectors/eigenvalues of the dV/dt operator.
     */
    public final EVXData I;

    /** The operator I=dV/dt eigenvector index, that has the maximal
     * {@literal <psi_k|psi0>}<sup>2</sup>. The value equals to 0
     * means the dV/dt is very small. The value equals to I.neq-1 means
     * the dV/dt is very large. The value equals to -1 means eigenproblem solution failed.
     * See "<a href="https://www.researchgate.net/publication/324171914">The Spectral Approach to Timeserie Bursts Analysis</a>" article, DOI:10.24411/0131-5226-2018-10010.
     */
    public final int index_max_psi_projection;

    final EnterPoint lastExecution;


    /** Initial state.
     */
    public SimpleStateSaver(){
	// first call, no previous state available
	this.time=0;
	this.P=Double.NaN;
	this.allshares=0;
	this.I=null;
	this.index_max_psi_projection=-1;
	this.seq=0;
	this.lastExecution=null;
    }

    /** Recurrently updated state.
     */
    public SimpleStateSaver(final SimpleStateSaver prev,
			    final long time,final double P,final long dV,
			    final EVXData I)
    {
	this.time=time;
	this.P=P;
	this.allshares=prev.allshares+dV;
	this.I=I;
	this.seq=1+prev.seq;

	// find the index of the maximal <psi0|psik>^2 projection
	double wmax=0,sw=0;int indwmax=-1;
	for(int k=0;k<I.neq;k++){
	    final double wk=Math.pow(I.psik_at_x0[k],2)/I.o_o;
	    sw+=wk;
	    if(wk>=wmax){
		wmax=wk;
		indwmax=k;
	    }
	}
	this.index_max_psi_projection=indwmax;

	if(P>0 && dV>0){
	    lastExecution=new EnterPoint(time,P,I.sH);
	} else {
	    lastExecution=prev.lastExecution;
	}
    }

} // END OF class SimpleStateSaver
