package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2020
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.*;
import com.polytechnik.freemoney.*;

/**
 * A simple support price calculation.
 * @version $Id: SimpleSupportPrice.java,v 1.22 2022/07/28 07:29:40 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public abstract class SimpleSupportPrice {
    public final int n;
    public final long tau;
    private final double [] QQt;
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;

    /**
     * @param n Basis dimention.
     * @param tau Time scale (in nanoseconds).
     */
    public SimpleSupportPrice(final int n,final long tau){
	this.n=n;
	this.tau=tau;
	this.M=getABasis();
	this.QQt=M.B.getQQMatr(n,getTMoments(2*n-1));
    }


/** Given (time,price,shares traded) triple produce the next state.
 * @param currenttime Time.
 * @param P Price.
 * @param dV Execution shares, set to 0 for limit order book events. If dV=0,
 * then the {@link SupportPriceStateSaver SupportPriceStateSaver} state
 * is not updated, except the P propagates to {@link SupportPriceStateSaver#P SupportPriceStateSaver.P}. For {@literal dV>0}
 * The {@link SupportPriceStateSaver#P_of_last_spike_in_I P_of_last_spike_in_I},
 * {@link SupportPriceStateSaver#P_of_last_execution P_of_last_execution},
 * {@link SupportPriceStateSaver#time_of_last_spike_in_I time_of_last_spike_in_I},
 * and {@link SupportPriceStateSaver#time_of_last_execution time_of_last_execution}
 * get updated according to {@link SupportPriceStateSaver constructor logic}.
 * @param prev Previous state.
 * @return The state updated according to (currenttime,P,dV) posted.
 */
public SimpleSupportPriceStateSaver getNext(
    final long currenttime,
    final double P,
    final int dV,
    final SimpleSupportPriceStateSaver prev){

    // separate integrator for dv/dt
    final WIntegrator wiI=prev!=null?prev.wiI:getEmptyWIntegrator(1);
    // add dvorig ahead of time to calculate dV/dt
    wiI.updateSavedMomentsToNewTimeReference(currenttime);
    wiI.addSingleObservation(currenttime,currenttime,new double []{dV});
    final double [] QQi=M.B.getQQMatr(n,wiI.integration_moments[0]);
    final EVXData I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);

    // Preserve the state, to obtain the "scalped" moments.
    final SupportPriceStateSaver sst=prev!=null?
	new SupportPriceStateSaver(prev.sst,currenttime,P,dV,I)
	:
	new SupportPriceStateSaver(getEmptyWIntegrator(5/* 5 fields for SupportPriceStateSaver */),currenttime,P,dV,I);

    final double dt,dv,dP;
    if(prev==null) {dt=dP=0;dv=sst.allshares-0;}
    else {
	dv=(sst.allshares-prev.sst.allshares);
	if(prev.sst.time<=0) dt=dP=0;
	else {
	    dt=(sst.time-prev.sst.time)/sst.wi.atau;
	    dP=sst.P_of_last_spike_in_I-prev.sst.P_of_last_spike_in_I;
	}
    }
    // The moments are calculated recurrently
    sst.updateState(sst.time,dt,Double.isFinite(dP)?dP:0,dv,dv);

    // Note: the wiI.integration_moments[0] and sst.getDvorigMoments() have to match exactly

    //  =========== In |psi> state, corresponding to the time of last spike
    // dPs/dt in the time of last spike
    final double x_at_last_spike_in_I=sst.wi.getXFromTime(currenttime,sst.time_of_last_spike_in_I);
    final double []
	 // like dp/dt, but for d/dt sst.P_of_last_spike_in_I
	QQdPsdt=M.B.getQQMatr(n,sst.getDpMoments()),
	// obtain Qk(x_at_last_spike_in_I)
	Q_at_last_spike_in_I=M.B.getX0Moments(n,x_at_last_spike_in_I),
	// obtain sum_k psik(x)psik(x_at_last_spike_in_I) in Q-basis
	psi_x_at_last_spike_in_I=EVXData.convertPsiFromBtoQ_Basis(n,I.getPsiAt(Q_at_last_spike_in_I),I.ev);
    // normalize to 1
    TMatr.scaleVectorInPlace(psi_x_at_last_spike_in_I,1/Math.sqrt(TMatr.mulVV(psi_x_at_last_spike_in_I,Q_at_last_spike_in_I)));
    final double dPsdt_at_last_spike_in_I=TMatr.getAMatrA(psi_x_at_last_spike_in_I,QQdPsdt,n),
	PsChange_at_last_spike_in_I=TMatr.mulVV(getPsi2WIntegratedDt(psi_x_at_last_spike_in_I),sst.getDpMoments());
    final double w_at_last_spike_in_I=1/Math.pow(TMatr.mulVV(psi_x_at_last_spike_in_I,Q_at_last_spike_in_I),2);

    final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING;
    if(eps>0){
	UnitTests.checkDoubleArrayRel1(wiI.integration_moments[0],"wiI.integration_moments[0]",sst.getDvorigMoments(),"sst.getDvorigMoments()",eps);
	UnitTests.checkDoubleRel(wiI.sum[0],"wiI.sum[0]",sst.allshares,"sst.allshares",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_x_at_last_spike_in_I,QQt,n),"psi_x",1,"1",eps);
	UnitTests.checkDoubleRel(w_at_last_spike_in_I,"w_at_last_spike_in_I",Math.pow(TMatr.getAMatrB(psi_x_at_last_spike_in_I,M.B.getONE(),QQt,n),2),"<psix|1>^2",eps);
	UnitTests.checkDoubleRel(w_at_last_spike_in_I,"w_at_last_spike_in_I",1/TMatr.sumElements(TMatr.squareArrayElements(I.getPsiAt(Q_at_last_spike_in_I))),"K(x_at_last_spike_in_I)",eps);
	if(0==Double.compare(I.x0,x_at_last_spike_in_I)) UnitTests.checkDoubleArrayRel1(psi_x_at_last_spike_in_I,"psi_x_at_last_spike_in_I",I.psi0,"I.psi0",eps);
	System.err.println("SimpleSupportPrice: OK n="+n+" x_at_last_spike_in_I="+x_at_last_spike_in_I+" eps="+eps);
    }

    return new SimpleSupportPriceStateSaver(sst,wiI,
					    dPsdt_at_last_spike_in_I,
					    x_at_last_spike_in_I,
					    w_at_last_spike_in_I,
					    PsChange_at_last_spike_in_I);
}
private static final double EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=true?0:1e-5;

// ========= Basis hooks
/** Polynomaial basis to use.
*/
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
/** The {@literal <Qk>} moments in the measure used.
 */
abstract double [] getTMoments(final int nmoms);
/** Square a wavefunction, then integrate by parts for the measure used in the basis chosen.
 */
abstract double [] getPsi2WIntegratedDt(final double [] psi);
/** Create a recurrent integrator for the measure used.
 * @param nfields The number of timeseries to be integrated.
 */
abstract WIntegrator getEmptyWIntegrator(final int nfields);
}
