package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.realtimebook.*;

/** Test John Pierpont Morgan D+ version.
 *  @version $Id: SkQVM.java,v 1.14 2022/07/28 12:27:19 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class SkQVM<DTYPE> {

final DTYPE odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final EjpComm dptI,vstI,DdptI,DvstI;
final EjaComm ja_sam,Rja_sam,RRja_sam;

SkQVM(final int n,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();
    this.M=getABasis();

    final int nmoms=2*n-1;

    final double [] dpmom=new double [nmoms],rvsmom=new double [nmoms],
	tmom=new double [nmoms],imom=new double [nmoms];
    access.setM(rvsmom,0,access.M_RVS,odata,nmoms);
    access.setM(dpmom,1,access.M_DP,odata,nmoms);
    access.setM(imom,0,access.M_I,odata,nmoms);
    access.setMT(tmom,nmoms);

    this.dptI=new EjpComm(n,imom,tmom,dpmom,M);
    this.vstI=new EjpComm(n,imom,tmom,rvsmom,M);
    this.DdptI=new EjpComm(n,imom,tmom,getDFmom(dpmom),M);
    this.DvstI=new EjpComm(n,imom,tmom,getDFmom(rvsmom),M);

    final double [] tmom_sam=new double [nmoms],ptmom_sam=new double [nmoms],pimom=new double [nmoms];
    access.setM(pimom,1,access.M_I,odata,nmoms);
    access.setM(tmom_sam,0,access.M_T,odata,nmoms);
    access.setM(ptmom_sam,1,access.M_T,odata,nmoms);
    this.ja_sam=new EjaComm(n-1,imom,tmom_sam,pimom,ptmom_sam,M);
    this.Rja_sam=new EjaComm(n-1,M.getRadauMomentsFromMoments(imom),M.getRadauMomentsFromMoments(tmom_sam),M.getRadauMomentsFromMoments(pimom),M.getRadauMomentsFromMoments(ptmom_sam),M);
    this.RRja_sam=new EjaComm(n-1,M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(imom)),M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(tmom_sam)),M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(pimom)),M.getRadauMomentsFromMoments(M.getRadauMomentsFromMoments(ptmom_sam)),M);

}
double [] getDFmom(final double [] moms)
{
    final double [] fmoms=new double [moms.length];
    access.setDFMoments(fmoms,moms,fmoms.length);
    TMatr.scaleVectorInPlace(fmoms,-1);
    return fmoms;
}

abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();


static abstract class SkQVMTest<DTYPE> extends QTest {

abstract SkQVM<DTYPE> getSkQVM();


SkQVMTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}


void test_O()
{
    final SkQVM<DTYPE> q=getSkQVM();

    for(int jj=0;jj<2;jj++)
    {
	final EjpComm y=jj<=0?q.dptI:q.vstI;
	final int nmoms=y.sfmom.length,n=y.F.neq+1;
	final double [] o=q.M.B.getX0Moments(nmoms,y.F.x0),io=TMatr.sumAcB(y.simom,2.12e2*y.I_H,o),
	    fdpo=TMatr.sumAcB(y.sfmom,1.53e2*(Math.abs(y.F.sL)+Math.abs(y.F.sH)),o);
	final EjpComm r=new EjpComm(n,io,y.stmom,q.getDFmom(fdpo),q.M);
	final EjpComm d=jj<=0?q.DdptI:q.DvstI;
	EVXData.matchEVXData(r.F,d.F,eps);
	ACorr.matchACorr(r.acorr,d.acorr,eps);
    }
    System.err.println(name+":SkQVMTest:test_O(): test OK neq="+neq+" eps="+eps);
}


void test_Common()
{
    final SkQVM<DTYPE> q=getSkQVM();
    q.dptI.test_Self(eps,q.access);
    q.vstI.test_Self(eps,q.access);
    q.DdptI.test_Self(eps,q.access);
    q.DvstI.test_Self(eps,q.access);


    q.ja_sam.test_Self(eps);
    q.Rja_sam.test_Self(eps);
    q.RRja_sam.test_Self(eps);

    System.err.println(name+":SkQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}


void test_Extra(){}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":SkQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_O();
    test_Extra();

    System.err.println(name+":SkQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":SkQVMTest: starting tests; ver=$Id: SkQVM.java,v 1.14 2022/07/28 12:27:19 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":SkQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: SkQVM.java,v 1.14 2022/07/28 12:27:19 mal Exp $");
}
} // END OF class SkQVMTest




}// END OF class SkQVM
