package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2016
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Skewness;
import com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.SortArrays;
import com.polytechnik.utils.DiffSkewness;
import com.polytechnik.utils.DerivativesCalculateable;
import com.polytechnik.outils.SimpleRootsFinder;


/** Find the p, corresponding to mim/max skewness.
 *  @version $Id: SkewnessMinMax.java,v 1.46 2020/02/06 10:02:50 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class SkewnessMinMax {
    final double [] re_roots=new double [3],im_roots=new double [3];
    public final Skewness SKinit;
    public final double sk_thr,p1,sk1,p2,sk2,p_M,p_L,p_R,p1_sk2,p2_sk1;
    public final int n_SK_roots;

    /** Given the measure with
     * (m0,m1,m2,m3) moments find the p,
     * such the measure with the moments
     * (m0+dm,m1+dm*x,m2+dm*x^2,m3+dm*x^3) (initial measure plus delta function with dm)
     * have min/max skewness.
     * if eps>0, then unit test with this eps.
     */
    SkewnessMinMax(final double m0,final double m1,final double m2,final double m3,
		   final double dm,final double eps)
    {
	this.SKinit=Skewness.getSkewnessFromMonomialMoments(m0,m1,m2,m3);
	final double [] tt[]=DiffSkewness.getNOM_SDENOM_FromMonomials(m0,m1,m2,m3,dm),
	    NOM=tt[0],SDENOM=tt[1];
	final double m=m0+dm,b=dm/m,p3_coeff=b*(1-b)*(1-2*b);
	final OrthogonalPolynomialsMonomialsBasis MB=DiffSkewness.MB;

	// Skewness:
	// SK=((x1+x2)/2-aver(x))/(x2-x1)=NOM(p)/sqrt(SDENOM(p))
	final DiffSkewness F=DiffSkewness.getDiffSkewness_FromMonomials(m0,m1,m2,m3,dm);
	final DerivativesCalculateable G=F.getGamma();

	// Find skewness roots.
	com.polytechnik.utils.PolynomialRootsElementary.root3(NOM,re_roots,im_roots);
	if(im_roots[0]==0 && im_roots[1]==0 && im_roots[2]==0){n_SK_roots=3;}
	else if(Math.abs(im_roots[1])>0){n_SK_roots=1;}
	else {n_SK_roots=0;}

	// Threshold (p=+- \infty)
	this.sk_thr=Math.abs(NOM[NOM.length-1])/Math.sqrt(SDENOM[SDENOM.length-1]);
	final double [] D_NOM=MB.B.getDifferentiated(NOM),
	    D_SDENOM=MB.B.getDifferentiated(SDENOM),
	    Dp=MB.B.mult2Pol(D_NOM,SDENOM);
	TMatr.sumP1(Dp,MB.B.mult2Pol(D_SDENOM,NOM),-0.5);
	final com.polytechnik.utils.PolynomialRootsFinder.PRoots u=//MB.getPolynomialRootsFinderInBasis().findRoots(Dp);
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG().findRoots(Dp);
	final double [] xre=u.re_root,xim=u.im_root;
	final double scale=SKinit.x2-SKinit.x1;
	double skmin=Double.NaN,pskmin=Double.NaN,skmax=Double.NaN,pskmax=Double.NaN;
	for(int k=0;k<u.nroots;k++)
	{
	    if(Math.abs(xim[k]/scale)>EPS_im_roots) continue; // complex roots
	    final double p=xre[k];
	    final double d2sk=scale*scale*G.getD2F(p);
	    if(Math.abs(d2sk)<=EPS_im_roots) continue; // near edges
	    final double sk=G.getF(p);
	    if(d2sk<0){	if(sk>=skmax || Double.isNaN(pskmax)){skmax=sk;pskmax=p;} } // max sk
	    else { if(sk<=skmin || Double.isNaN(pskmin)){skmin=sk;pskmin=p;} } // min sk
	    if(eps>0)
	    {
		System.err.println("SkewnessMinMax:minmax:"+k+":p="+p+" imp="+xim[k]+" sk="+G.getF(p)+" Dsk="+G.getDF(p)+" D2sk="+G.getD2F(p));
	    }
	}
	if(pskmin<=pskmax){this.p1=pskmin;this.sk1=skmin;this.p2=pskmax;this.sk2=skmax;}
	else {this.p2=pskmin;this.sk2=skmin;this.p1=pskmax;this.sk1=skmax;}

	final double scale_dp=20*(SKinit.x2-SKinit.x1);
	final TargetSkewness sk_init=new TargetSkewness(F,(SKinit.w1-SKinit.w2)/m);
	this.p_M=SimpleRootsFinder.findRoots(sk_init,Double.NaN,MAXITERPERSTEP,true,SKinit.x1,SKinit.x2,EPS_im_roots);
	this.p_L=SimpleRootsFinder.findRoots(sk_init,Double.NaN,MAXITERPERSTEP,true,SKinit.x1-scale_dp,SKinit.x1,EPS_im_roots);
	this.p_R=SimpleRootsFinder.findRoots(sk_init,Double.NaN,MAXITERPERSTEP,true,SKinit.x2,SKinit.x2+scale_dp,EPS_im_roots);
	final TargetSkewness sk_t1=new TargetSkewness(F,(dm+SKinit.w1-SKinit.w2)/m),
	    sk_t2=new TargetSkewness(F,(-dm+SKinit.w1-SKinit.w2)/m);
	this.p1_sk2=SimpleRootsFinder.findRoots(sk_t2,Double.NaN,MAXITERPERSTEP,true,SKinit.x1-scale_dp,SKinit.x1,EPS_im_roots);
	this.p2_sk1=SimpleRootsFinder.findRoots(sk_t1,Double.NaN,MAXITERPERSTEP,true,SKinit.x2,SKinit.x2+scale_dp,EPS_im_roots);
	if(eps>0)
	{
	  final com.polytechnik.outils.FunctionCalculateable
	      SkXp=new com.polytechnik.outils.FunctionCalculateable(){
		  public double getF(final double x)
		      {return MB.B.getD0(NOM,x)/Math.sqrt(MB.B.getD0(SDENOM,x));}
	      },
	      SkXDp_=new com.polytechnik.outils.FunctionCalculateable(){
		      public double getF(final double x)
		      {return MB.B.getD0(Dp,x)/Math.pow(MB.B.getD0(SDENOM,x),1.5);}
	      },
	      SkXDp=new com.polytechnik.outils.FunctionCalculateable(){
		      public double getF(final double x)
		      { final double dnom=MB.B.getD0(SDENOM,x),sqdnom=Math.sqrt(dnom);
			return MB.B.getD0(D_NOM,x)/sqdnom-0.5*(MB.B.getD0(NOM,x)/sqdnom)*(MB.B.getD0(D_SDENOM,x)/dnom);
		      }
		  },
	      SkX=new com.polytechnik.outils.FunctionCalculateable(){
		  public double getF(final double x)
		  {return DiffSkewness.getDiffSkewness_FromMonomials(m0,m1,m2,m3,dm).getSkewnessAt(x).sk;}
	      },
	      PMed=new com.polytechnik.outils.FunctionCalculateable(){
		  public double getF(final double x)
		  {final Skewness sm=DiffSkewness.getDiffSkewness_FromMonomials(m0,m1,m2,m3,dm).getSkewnessAt(x);
		      return 0.5*(sm.x1+sm.x2);
		  }
		  },
	      Pd=new com.polytechnik.outils.FunctionCalculateable(){
		  public double getF(final double x)
		  {final Skewness sm=DiffSkewness.getDiffSkewness_FromMonomials(m0,m1,m2,m3,dm).getSkewnessAt(x);
		      return (sm.x2-sm.x1);
		  }
		  },
	       V4=new com.polytechnik.outils.FunctionCalculateable(){
		  public double getF(final double x)
		  {final Skewness sm=DiffSkewness.getDiffSkewness_FromMonomials(m0,m1,m2,m3,dm).getSkewnessAt(x);
		      double a4=0;
		      return a4+(dm/(m0+dm))*Math.pow(x,4)-
			  (sm.w1*Math.pow(sm.x1,4)+sm.w2*Math.pow(sm.x2,4))/(sm.w1+sm.w2);
		  }
	          },
	     SkP1=new com.polytechnik.outils.FunctionCalculateable(){
		  public double getF(final double x)
		  {return DiffSkewness.getDiffSkewness_FromMonomials(m0,m1,m2,m3,dm).getSkewnessAt(x).x1;}
		 },
	     SkP2=new com.polytechnik.outils.FunctionCalculateable(){
		  public double getF(final double x)
		  {return DiffSkewness.getDiffSkewness_FromMonomials(m0,m1,m2,m3,dm).getSkewnessAt(x).x2;}
		 };
	  UnitTests.checkDoubleRel(p3_coeff,"p3_coeff",NOM[NOM.length-1],"NOM[NOM.length-1]",eps);
	  UnitTests.checkDoubleArrayRel(NOM,"NOM",tt[2],"tt[2]",eps);

	  System.err.println("SkewnessMinMax:matchf: p1="+p1+" SkX(p1)="+SkX.getF(p1)+" SkXDp(p1)="+SkXDp.getF(p1)+" p2="+p2+" SkX(p2)="+SkX.getF(p2)+" SkXDp(p2)="+SkXDp.getF(p2)+
			     " n_SK_roots="+n_SK_roots+
			     " re_roots="+java.util.Arrays.toString(re_roots)+
			     " im_roots="+java.util.Arrays.toString(im_roots)+
			     " SkX(r0)="+SkX.getF(re_roots[0])+" SkX(r1)="+SkX.getF(re_roots[1])+" SkX(r2)="+SkX.getF(re_roots[2])+" sk_thr="+sk_thr);
	  //UnitTests.checkDoubleRel(F.getX1X2()[0].getF(1/(eps*1e-2)),"p1(inf)",F.sknodi.xa,"F.sknodi.xa",eps);
	  //UnitTests.checkDoubleRel(F.getX1X2()[1].getF(-1/(eps*1e-2)),"p2(-inf)",F.sknodi.xa,"F.sknodi.xa",eps);

	  UnitTests.checkDoubleRel(SkX.getF(p1),"SkX.getF(p1)",sk1,"sk1",eps);
	  UnitTests.checkDoubleRel(SkX.getF(p2),"SkX.getF(p2)",sk2,"sk2",eps);
	  UnitTests.checkDoubleRel(SkXp.getF(p1),"SkXp.getF(p1)",sk1,"sk1",eps);
	  UnitTests.checkDoubleRel(SkXp.getF(p2),"SkXp.getF(p2)",sk2,"sk2",eps);
	  if(!Double.isNaN(p1)) UnitTests.checkDoubleRel(SkXDp.getF(p1),"SkXDp.getF(p1)",0,"0",eps);
	  if(!Double.isNaN(p2)) UnitTests.checkDoubleRel(SkXDp.getF(p2),"SkXDp.getF(p2)",0,"0",eps);

	  // Match skewness calculation
	  final double sgm=Math.sqrt(m2/m0-Math.pow(m1/m0,2));
	  for(double sf:new double []{-100,-3,-2.2,-1.1,-0.3,-0.1,0,0.1,0.3,1.1,2.2,3,100})
	  {
	      final double p=m1/m+sf*sgm;
	      System.err.println("SkewnessMinMax:matchskewness:p="+p+" sk="+SkX.getF(p)+" Dsk="+SkXDp.getF(p));
	      UnitTests.checkDoubleRel(SkX.getF(p),"SkX.getF(p)",SkXp.getF(p),"SkXp.getF(p)",eps);
	      final double dpe=eps,Dsk_diff=(SkX.getF(p+dpe)-SkX.getF(p-dpe))/(2*dpe),
		  D2sk_diff=(SkXDp.getF(p+dpe)-SkXDp.getF(p-dpe))/(2*dpe);

	      System.err.println("SkewnessMinMax:difftest:p="+p+" SkXDp.getF(p)(analytical)="+SkXDp.getF(p)+" SkXDp_.getF(p)(analytical)="+SkXDp_.getF(p)+" Dsk_diff="+Dsk_diff+" D2sk(analytical)="+G.getD2F(p)+" D2sk_diff="+D2sk_diff);
	      UnitTests.checkDoubleRel(SkXDp_.getF(p),"SkXDp_.getF(p)",SkXDp.getF(p),"SkXDp.getF(p)",eps);
	      UnitTests.checkDoubleRel(G.getDF(p),"G.getDF(p)",SkXDp.getF(p),"SkXDp.getF(p)",eps);

	      // check F.getGamma() derivatives
	      checkDeriv_vs_Finitediff(p,F.getGamma(),SkX,eps,eps);

	      // check F.getX1X2midpoint() derivatives
	      checkDeriv_vs_Finitediff(p,F.getX1X2midpoint(),PMed,eps,2*eps);

	      // check F.getX2minusX1() derivatives
	      checkDeriv_vs_Finitediff(p,F.getX2minusX1(),Pd,eps,2*eps);

	      // check F.getL4Volatility() derivatives
	      checkDeriv_vs_Finitediff(p,F.getL4Volatility(),V4,eps,2*eps);

	      // check nodes derivatives
	      checkDeriv_vs_Finitediff(p,F.getX1X2()[0],SkP1,eps,2*eps);
	      checkDeriv_vs_Finitediff(p,F.getX1X2()[1],SkP2,eps,2*eps);

	  }
	  if(true)
	  { // Match Skewness0 roots.
	      final Skewness0 s0=new Skewness0(m0,m1,m2,m3,dm,eps);
	      UnitTests.checkDoubleArrayRel(re_roots,"re_roots",s0.re_roots,"s0.re_roots",eps);
	      UnitTests.checkDoubleArrayRel(im_roots,"im_roots",s0.im_roots,"im.re_roots",eps);
	  }

	  if(dm/m>eps)
	  {
	      System.err.println("SkewnessMinMax:sktarget:sk_target="+sk_init.sk_target+" SkX(x1)="+SkX.getF(SKinit.x1)+" SkX(x2)="+SkX.getF(SKinit.x2)+" dm/m="+(dm/m));
	      UnitTests.checkDoubleRel(SkX.getF(p_M),"SkX.getF(p_M)",sk_init.sk_target,"sk_init.sk_target",eps);
	      UnitTests.checkDoubleRel(SkX.getF(p_L),"SkX.getF(p_L)",sk_init.sk_target,"sk_init.sk_target",eps);
	      UnitTests.checkDoubleRel(SkX.getF(p_R),"SkX.getF(p_R)",sk_init.sk_target,"sk_init.sk_target",eps);
	      UnitTests.checkDoubleRel(SkX.getF(SKinit.x2),"SkX.getF(SKinit.x2)",sk_t2.sk_target,"sk_t2.sk_target",eps);
	      UnitTests.checkDoubleRel(SkX.getF(p1_sk2),"SkX.getF(p1_sk2)",sk_t2.sk_target,"sk_t2.sk_target",eps);
	      UnitTests.checkDoubleRel(SkX.getF(SKinit.x1),"SkX.getF(SKinit.x1)",sk_t1.sk_target,"sk_t1.sk_target",eps);
	      UnitTests.checkDoubleRel(SkX.getF(p2_sk1),"SkX.getF(p2_sk1)",sk_t1.sk_target,"sk_t1.sk_target",eps);
	      System.err.println("SkewnessMinMax:sktarget:p_M="+p_M+" p_L="+p_L+" p_R="+p_R+" (SKinit.x1+SKinit.x2)/2="+((SKinit.x1+SKinit.x2)/2)+" SKinit.xa="+SKinit.xa+" p1_sk2="+p1_sk2+" p2_sk1="+p2_sk1);
	  }

	  System.err.println("SkewnessMinMax:final: p1="+p1+" sk1="+sk1+" p2="+p2+" sk2="+sk2+
			     " Dsk1="+SkXDp.getF(p1)+" Dsk2="+SkXDp.getF(p2));
	  System.err.println("SkewnessMinMax:final:SKinit.x1="+SKinit.x1+" p2_sk1="+p2_sk1+" sk_t1.sk_target="+sk_t1.sk_target);
	  System.err.println("SkewnessMinMax:final:p1_sk2="+p1_sk2+" SKinit.x2="+SKinit.x2+" sk_t2.sk_target="+sk_t2.sk_target);
	}
    }

    static void checkDeriv_vs_Finitediff(
	final double p,
	final DerivativesCalculateable F,
	final com.polytechnik.outils.FunctionCalculateable f,
	final double eps,final double epsthr)
    {
	final double dpe=eps;
	final double Df_diff=(f.getF(p+dpe)-f.getF(p-dpe))/(2*dpe),
	    D2F_diff=(F.getDF(p+dpe)-F.getDF(p-dpe))/(2*dpe);
	UnitTests.checkDoubleRel(F.getF(p),"F.getF(p)",f.getF(p),"f.getF(p)",eps);
	if(Math.abs(Df_diff)>epsthr) UnitTests.checkDoubleRel(F.getDF(p),"F.getDF(p)",Df_diff,"Df_diff",eps);
	if(Math.abs(D2F_diff)>epsthr) UnitTests.checkDoubleRel(F.getD2F(p),"F.getD2F(p)",D2F_diff,"D2F_diff",eps);
	System.err.println("SkewnessMinMax:checkDeriv_vs_Finitediff: match OK p="+p+" eps="+eps);
    }


    static final double EPS_im_roots=1e-8;
    static final int MAXITERPERSTEP=100;
}
