package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.freemoney.CommonlyUsedMoments;
import com.polytechnik.freemoney.IandDM;

/** A class to split dI/dt like operator on positive and negative subspaces.
 * @see com.polytechnik.trading.IntervalSincePsiM
 * @version $Id: SplitdIHdt.java,v 1.21 2022/09/16 21:30:21 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class SplitdIHdt {

    public final double
	PdIHdt_spurM,
	dIHdt_posdIH_spurM,dIHdt_negdIH_spurM,
	pF_posdIH_spurM,pF_negdIH_spurM,
	dpdt_spurM,
	dpdt_posdIH_spurM,dpdt_negdIH_spurM;

    public final double dIHdt_spurM;
    /** Time to "now" (in tau units) in selected state.
     */
    public final double T_M,Tv_M;
    public final double pv_M,I_M;
    public final EVXData IH;
    /** The state of the spike.
     */
    final double [] psiM;

    SplitdIHdt(final EVXData IH,
	       final double [] psiM,
	       final double p_offset,
	       final double p_last,
	       final double [] bQQdIHdt,
	       final double [] bQQpdIHdt,
	       final double [] bQQDensityMatrix,
	       final double [] bQQdp,
	       final double T_M,
	       final double Tv_M,
	       final double I_M,
	       final double pv_M,
	       // basis operations
	       final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	       // for unit tests
	       final double eps){
	this.IH=IH;
	this.psiM=psiM;
	this.T_M=T_M;
	this.Tv_M=Tv_M;
	this.I_M=I_M;
	this.pv_M=pv_M;
	final int n=IH.neq;
	// Equals to zero if proper "impact from the future" is used.
	dIHdt_spurM= TMatr.getSpurMM(n,bQQdIHdt,bQQDensityMatrix);
	final double []
	    // we assume omega(x0)=1
	    bQQ0=TMatr.getQQaa(IH.psik_at_x0),
	    // Force set the impact from the future to the value
	    // that provides: Spur(bQQdIHdtF*bQQDensityMatrix)=0
	    bQQdIHdtF=TMatr.sumaAbB(-dIHdt_spurM,bQQ0,1,bQQdIHdt),
	    bQQpdIHdtF=TMatr.sumaAbB(-p_last*dIHdt_spurM,bQQ0,1,bQQpdIHdt);
	PdIHdt_spurM=TMatr.getSpurMM(n,bQQpdIHdtF,bQQDensityMatrix);
	dpdt_spurM=TMatr.getSpurMM(n,bQQdp,bQQDensityMatrix);
	final double [] ev_bQQdIHdtF=M.EV.getEV(n,bQQdIHdtF,true),
	    bbQQ_selPosdIHdtF=new double [n*n],
	    bbQQ_selNegdIHdtF=new double [n*n];
	// Set projection operators in |dIH/dt|psi>=lambda|psi> basis.
	for(int k=0;k<n;k++){
	    final int diagk=k*n+k;
	    if(ev_bQQdIHdtF[k]>0){
		bbQQ_selPosdIHdtF[diagk]=1;
	    } else if(ev_bQQdIHdtF[k]<=0){
		bbQQ_selNegdIHdtF[diagk]=1;
	    }
	}
	final double [] UNIT=TMatr.getUnitMatrix(n),
	    // convert back to b-basis
	    bQQ_selPosdIHdtF=EVXData.convertMatrFromBtoQ_Basis(n,bbQQ_selPosdIHdtF,ev_bQQdIHdtF,UNIT),
	    bQQ_selNegdIHdtF=EVXData.convertMatrFromBtoQ_Basis(n,bbQQ_selNegdIHdtF,ev_bQQdIHdtF,UNIT);
	dIHdt_posdIH_spurM=TMatr.getSpurMMM(n,bQQdIHdtF,bQQ_selPosdIHdtF,bQQDensityMatrix);
	dIHdt_negdIH_spurM=TMatr.getSpurMMM(n,bQQdIHdtF,bQQ_selNegdIHdtF,bQQDensityMatrix);
	final double
	    PdIHdt_posdIH_spurM=TMatr.getSpurMMM(n,bQQpdIHdtF,bQQ_selPosdIHdtF,bQQDensityMatrix),
	    PdIHdt_negdIH_spurM=TMatr.getSpurMMM(n,bQQpdIHdtF,bQQ_selNegdIHdtF,bQQDensityMatrix);
	pF_posdIH_spurM=p_offset+PdIHdt_posdIH_spurM/dIHdt_posdIH_spurM;
	pF_negdIH_spurM=p_offset+PdIHdt_negdIH_spurM/dIHdt_negdIH_spurM;
	// just for info
	dpdt_posdIH_spurM=TMatr.getSpurMMM(n,bQQdp,bQQ_selPosdIHdtF,bQQDensityMatrix);
	dpdt_negdIH_spurM=TMatr.getSpurMMM(n,bQQdp,bQQ_selNegdIHdtF,bQQDensityMatrix);

	if(eps>0){
	    UnitTests.checkDoubleRel1(0,"0",TMatr.getSpurMM(n,bQQdIHdtF,bQQDensityMatrix),"spur(bQQdIHdtF*bQQDensityMatrix)",eps);
	    UnitTests.checkDoubleRel1(0,"0",dIHdt_posdIH_spurM+dIHdt_negdIH_spurM,"dIHdt_posdIH_spurM+dIHdt_negdIH_spurM",eps);
	    UnitTests.checkDoubleRel1(PdIHdt_spurM,"PdIHdt_spurM",PdIHdt_posdIH_spurM+PdIHdt_negdIH_spurM,"PdIHdt_posdIH_spurM+PdIHdt_negdIH_spurM",eps);
	    UnitTests.checkDoubleRel1(dpdt_spurM,"dpdt_spurM",dpdt_posdIH_spurM+dpdt_negdIH_spurM,"dpdt_posdIH_spurM+dpdt_negdIH_spurM",eps);


	    // basis invariace checks
	    final double []
		// Convert the matrices to the basis of ev_bQQdIHdtF eigenstates
		// as projection operators have a very simple form in this basis.
		bbQQdIHdtF=TMatr.getDiagMatrix(java.util.Arrays.copyOf(ev_bQQdIHdtF,n)),
		bbQQDensityMatrix=EVXData.convertMatrFromQtoB_Basis(n,bQQDensityMatrix,ev_bQQdIHdtF),
		// actual dp/dt
		bbQQdp=EVXData.convertMatrFromQtoB_Basis(n,bQQdp,ev_bQQdIHdtF),
		// actual p dIH/dt
		bbQQpdIHdtF=EVXData.convertMatrFromQtoB_Basis(n,bQQpdIHdtF,ev_bQQdIHdtF);

	    UnitTests.checkDoubleRel1(TMatr.getSpurMM(n,bbQQdIHdtF,bbQQDensityMatrix),"spur(bbQQdIHdtF*bbQQDensityMatrix)",0,"0",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMM(n,bbQQpdIHdtF,bbQQDensityMatrix),"spur(bbQQpdIHdtF*bbQQDensityMatrix)",PdIHdt_spurM,"PdIHdt_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMM(n,bbQQdp,bbQQDensityMatrix),"spur(bbQQdp*bbQQDensityMatrix)",dpdt_spurM,"dpdt_spurM",eps);

	    // check in the basis of dIH/dt operator
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQdIHdtF,bbQQ_selPosdIHdtF,bbQQDensityMatrix),"spur(bbQQdIHdtF*bbQQ_selPosdIHdtF*bbQQDensityMatrix)",dIHdt_posdIH_spurM,"dIHdt_posdIH_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQdIHdtF,bbQQ_selNegdIHdtF,bbQQDensityMatrix),"spur(bbQQdIHdtF*bbQQ_selNegdIHdtF*bbQQDensityMatrix)",dIHdt_negdIH_spurM,"dIHdt_negdIH_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQpdIHdtF,bbQQ_selPosdIHdtF,bbQQDensityMatrix),"spur(bbQQpdIHdtF*bbQQ_selPosdIHdtF*bbQQDensityMatrix)",PdIHdt_posdIH_spurM,"PdIHdt_posdIH_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQpdIHdtF,bbQQ_selNegdIHdtF,bbQQDensityMatrix),"spur(bbQQpdIHdtF*bbQQ_selNegdIHdtF*bbQQDensityMatrix)",PdIHdt_negdIH_spurM,"PdIHdt_negdIH_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQdp,bbQQ_selPosdIHdtF,bbQQDensityMatrix),"spur(bbQQdp*bbQQ_selPosdIHdtF*bbQQDensityMatrix)",dpdt_posdIH_spurM,"dpdt_posdIH_spurM",eps);
	    UnitTests.checkDoubleRel1(TMatr.getSpurMMM(n,bbQQdp,bbQQ_selNegdIHdtF,bbQQDensityMatrix),"spur(bbQQdp*bbQQ_selNegdIHdtF*bbQQDensityMatrix)",dpdt_negdIH_spurM,"dpdt_negdIH_spurM",eps);

	    System.err.println("SplitdIHdt: test OK EV="+java.util.Arrays.toString(java.util.Arrays.copyOf(ev_bQQdIHdtF,n))+" n="+n+" eps="+eps);
	}
    }


    /** Obtain the result from moments (possibly of secondary sampling).
     * The "execution flow" is extracted internally from its derivative's {@code dIHdtmom} moments
     * using integration by parts. This allows the moments of secondary sampling
     * to be used for modelling execution flow.
     * For regular execution flow {@code I=dV/dt} use
     * {@code sst.getDFMomentsFromF(TMatr.sumaAbB(0*Ifuture,tmom,-1,imom))};
     * the {@code Ifuture} may be set to 0, as it is properly set
     * inside {@link SplitdIHdt#SplitdIHdt constructor} anyway.
     * @param n Problem dimension.
     * @param dIHdtmom Input moments of dI/dt -"style".
     * @param pdIHdtmom Input moments of p*dI/dt -"style". One may use the
     * moments of \(p\frac{dI}{dt}\) and \(\frac{dpI}{dt}\) types.
     * @param dpmom Some moments to test other ideas. Propagates to
     * {@link SplitdIHdt#dpdt_spurM}, {@link SplitdIHdt#dpdt_posdIH_spurM}, and
     * {@link SplitdIHdt#dpdt_negdIH_spurM}.
     * @param sst Basis operations.
     * @param eps If positive -- run unit test with this eps.
     */
    public static SplitdIHdt getFromdIHdt(
	final int n,
	// input moment, possibly of secondary sampling
	final double [] dIHdtmom,
	final double [] pdIHdtmom,
	// regular moments for testing other concepts
	final double [] dpmom,
	final CommonlyUsedMoments sst,
	// for unit tests
	final double eps){

	final double p_offset=sst.getPOffset(),
	    p_last=sst.getPLast_withoutPOffset()-p_offset;

	final var M=sst.getABasis();
	// The result is correct but different within a constant; this does not change anything.
	// The getFMomentsFromDF sets the boundary value=0.
	final double [] IHmom=sst.getFMomentsFromDF(dIHdtmom),
	    // Gram matrix
	    QQt=M.B.getQQMatr(n,sst.getdtMoments());

	final double [] QQIH=M.B.getQQMatr(n,IHmom),
	    QQdIHdt=M.B.getQQMatr(n,dIHdtmom),
	    QQpdIHdt=M.B.getQQMatr(n,pdIHdtmom),
	    // to test old ideas
	    QQdp=M.B.getQQMatr(n,dpmom),
	    // used only for testings/diag
	    QQT=TMatr.scaleVectorInPlace(M.B.getQQMatr(n,sst.getTdtMoments()),-1),
	    QQTi=TMatr.scaleVectorInPlace(M.B.getQQMatr(n,sst.getTdVMoments()),-1),
	    QQi=M.B.getQQMatr(n,sst.getdVMoments()),
	    QQpi=M.B.getQQMatr(n,sst.getPdVMoments());

	final var IHDM=new IandDM(n,QQt,QQIH,M,sst::getPolynomialWIntegratedDt,eps);
	final double
	    I_M=TMatr.getAMatrA(IHDM.psiM,QQi,n),
	    pv_M=p_offset+TMatr.getAMatrA(IHDM.psiM,QQpi,n)/I_M;
	final var r=new SplitdIHdt(IHDM.I,
				   IHDM.psiM,
				   p_offset,
				   p_last,
				   EVXData.convertMatrFromQtoB_Basis(n,QQdIHdt,IHDM.I.ev),
				   EVXData.convertMatrFromQtoB_Basis(n,QQpdIHdt,IHDM.I.ev),
				   EVXData.convertMatrFromQtoB_Basis(n,IHDM.QQDensityMatrix,IHDM.I.ev),
				   EVXData.convertMatrFromQtoB_Basis(n,QQdp,IHDM.I.ev),
				   TMatr.getAMatrA(IHDM.psiM,QQT,n),
				   TMatr.getAMatrA(IHDM.psiM,QQTi,n)/I_M,
				   I_M,
				   pv_M,
				   M,eps);
	if(eps>0){
	    UnitTests.checkDoubleRel1(TMatr.getSpurMM(n,TMatr.mulMatrices(n,new double [][]{IHDM.IQQt,QQdIHdt,IHDM.IQQt}),IHDM.QQDensityMatrix),"spur(IQQt*QQdIHdt*IQQt*IHDM.QQDensityMatrix)",(0-IHDM.I.sH),"0-IH.sH",eps);
	    UnitTests.checkDoubleRel1(r.dIHdt_spurM,"r.dIHdt_spurM",0-IHDM.I.sH,"0-IH.sH",eps);

	    // Check the moments converted back are correct
	    final double [] dIHmom=TMatr.scaleVectorInPlace(sst.getDFMomentsFromF(IHmom),-1);
	    UnitTests.checkDoubleArrayRel1(dIHmom,"dIHmom",dIHdtmom,"dIHdtmom",eps);
	    System.err.println("SplitdIHdt:getFromdIHdt: Test OK IH.EV="+java.util.Arrays.toString(IHDM.I.getEValues())+" n="+n+" eps="+eps);
	}
	return r;
    }

    /** For unit test.
     */
    public static void match(final SplitdIHdt a,final SplitdIHdt b,final double eps){
	if(false){
	    // The eignevalues potentially may differ on a constant, then this fails.
	    EVXData.matchEVXData(a.IH,b.IH,false,eps);
	} else {
	    final double [] aev=a.IH.getEValues(),bev=b.IH.getEValues();
	    final double evOffset=a.IH.sL-b.IH.sL;
	    for(int k=0;k<aev.length;k++){
		UnitTests.checkDoubleRel1(aev[k],"aev[k]",bev[k]+evOffset,"bev[k]+evOffset",eps);
	    }
	}

	UnitTests.checkDoubleRel(a.T_M,"a.T_M",b.T_M,"b.T_M",eps);
	UnitTests.checkDoubleRel(a.Tv_M,"a.Tv_M",b.Tv_M,"b.Tv_M",eps);
	UnitTests.checkDoubleRel(a.I_M,"a.I_M",b.I_M,"b.I_M",eps);
	UnitTests.checkDoubleRel(a.pv_M,"a.pv_M",b.pv_M,"b.pv_M",eps);
	UnitTests.checkDoubleRel(a.PdIHdt_spurM,"a.PdIHdt_spurM",b.PdIHdt_spurM,"b.PdIHdt_spurM",eps);
	UnitTests.checkDoubleRel(a.dpdt_spurM,"a.dpdt_spurM",b.dpdt_spurM,"b.dpdt_spurM",eps);
	UnitTests.checkDoubleRel(a.dIHdt_posdIH_spurM,"a.dIHdt_posdIH_spurM",b.dIHdt_posdIH_spurM,"b.dIHdt_posdIH_spurM",eps);
	UnitTests.checkDoubleRel(a.dIHdt_negdIH_spurM,"a.dIHdt_negdIH_spurM",b.dIHdt_negdIH_spurM,"b.dIHdt_negdIH_spurM",eps);
	UnitTests.checkDoubleRel(a.pF_posdIH_spurM,"a.pF_posdIH_spurM",b.pF_posdIH_spurM,"b.pF_posdIH_spurM",eps);
	UnitTests.checkDoubleRel(a.pF_negdIH_spurM,"a.pF_negdIH_spurM",b.pF_negdIH_spurM,"b.pF_negdIH_spurM",eps);
	UnitTests.checkDoubleRel(a.dpdt_posdIH_spurM,"a.dpdt_posdIH_spurM",b.dpdt_posdIH_spurM,"b.dpdt_posdIH_spurM",eps);
	UnitTests.checkDoubleRel(a.dpdt_negdIH_spurM,"a.dpdt_negdIH_spurM",b.dpdt_negdIH_spurM,"b.dpdt_negdIH_spurM",eps);

	System.err.println("SplitdIHdt:match test OK n="+a.IH.neq+" eps="+eps);
    }
}
