package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.VolMatrix;
import com.polytechnik.freemoney.WIntegrator;

/**
 * Further development of ScalpedMaxIProjection.java. Now try scalped time, scalped price, scalped shares traded.
 * @version $Id: SupportPrice.java,v 1.125 2022/07/28 08:04:53 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public abstract class SupportPrice<DTYPE> {


public final SupportPriceStateSaver sst;

/** "Regular", not scalp--moments. The object has
 * an internal state to recurrently speed up the calculations.
 */
final DTYPE odata;

final VolMatrix VOLATILITY_EDPsi=this::getEDPsi;

public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getABasis();

/** The |I|psi>=lambda |psi> solution.
 */
public final EVXData I;

public final double
// Helper data and moving averages.
    p_offset,p_last,pi_average,pt_average,
// The https://arxiv.org/abs/1510.05510 results (the state of max I).
    dpdt_0,p_IH,pt_IH,dpdt_IH,var1pI_IH,var1pI_x
// The https://arxiv.org/abs/1903.11530 results

    ;
public final double
    w_psiH,
    dpdt_at_last_spike_in_I,
    dPsdt_at_last_spike_in_I,
    PsChange_at_last_spike_in_I,
    x_at_last_spike_in_I,
    w_at_last_spike_in_I,
    DIRdPs;

/** Not worth a function or constructor argument.
 */
private static double EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;

/**
* Constructor.
* @param n Basis dimension.
* @param odata Regular moments internal state.
* @param rst Previous tick internal state.
*/
SupportPrice(final int n,final DTYPE odata,final SupportPriceStateSaver rst)
{

    final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING; // turn on selftests if required
    this.odata=odata;
    final SMomentsData mm=getSMoments(2*n-1).asQQMatrices(n,M);
    // mm has the matrices (not moments) stored.
    final double [] QQt=mm.tmom,QQi=mm.imom,QQpi=mm.pimom,QQdpdt=mm.dpmom,QQp1t=mm.p1tmom,QQp0t=mm.p0tmom;
    p_offset=mm.p_offset;
    p_last=mm.p_last;
    pi_average=QQpi[0]/QQi[0];
    pt_average=QQp1t[0]/QQp0t[0];
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] psiH=I.getPsi(n-1/* max I */),psi0=I.psi0,
	DpsiH=VOLATILITY_EDPsi.getDPsi(psiH);
    // Lebesgue weight for |psiH>
    w_psiH=Math.pow(TMatr.getAMatrB(psiH,M.B.getONE(),QQt,n),2);
    dpdt_0=TMatr.getAMatrA(psi0,QQdpdt,n);
    final double pI_IH=TMatr.getAMatrA(psiH,QQpi,n),
        psiH_DpsiH=TMatr.getAMatrB(psiH,DpsiH,QQt,n);
    p_IH=pI_IH/TMatr.getAMatrA(psiH,QQi,n);
    pt_IH=TMatr.getAMatrA(psiH,QQp1t,n)/TMatr.getAMatrA(psiH,QQp0t,n);
    dpdt_IH=TMatr.getAMatrA(psiH,QQdpdt,n);
    var1pI_IH=2*(TMatr.getAMatrB(psiH,DpsiH,QQpi,n)-pI_IH*psiH_DpsiH);
    if(eps>0){
	final double [] QQpmp_IH=TMatr.sumAcB(QQpi,-p_IH,QQi),
	    rq=com.polytechnik.utils.RayleighQuotient.getRQVariationValues(psiH,DpsiH,null,QQpmp_IH,QQt,n);
	if(Double.isFinite(rq[0])){
	    UnitTests.checkDoubleRel(rq[0],"rq[0]",0,"0",eps);
	    UnitTests.checkDouble(rq[1],"rq[1]",var1pI_IH,"var1pI_IH",eps);
	}
    }

    // Preserve the state, to obtain the "scalped" moments.
    // For "regular" moments the DTYPE also has an internal state :(
    this.sst=rst!=null?
	new SupportPriceStateSaver(rst,mm.currenttime,p_last+p_offset /* All prices are relatively p_offset */,mm.totalvolume-rst.allshares,I)
	:
	new SupportPriceStateSaver(getEmptyWIntegrator(),mm.currenttime,p_last+p_offset /* All prices are relatively p_offset */,mm.totalvolume,I);

    final double dt,dv,dP;
    if(rst==null) {dt=dP=0;dv=sst.allshares-0;}
    else {
	dv=(sst.allshares-rst.allshares);
	if(rst.time<=0) dt=dP=0;
	else {
	    dt=(sst.time-rst.time)/sst.wi.atau;
	    dP=sst.P_of_last_spike_in_I-rst.P_of_last_spike_in_I;
	}
    }
    sst.updateState(sst.time,dt,Double.isFinite(dP)?dP:0,dv,dv);


    //  =========== In |psi> state, corresponding to the time of last spike
    // dPs/dt in the time of last spike
    this.x_at_last_spike_in_I=sst.wi.getXFromTime(mm.currenttime,sst.time_of_last_spike_in_I);
    final double []
	 // like dp/dt, but for d/dt sst.P_of_last_spike_in_I
	QQdPsdt=M.B.getQQMatr(n,sst.getDpMoments()),
	// obtain Qk(x_at_last_spike_in_I)
	Q_at_last_spike_in_I=M.B.getX0Moments(n,x_at_last_spike_in_I),
	// obtain sum_k psik(x)psik(x_at_last_spike_in_I) in Q-basis
	psi_x_at_last_spike_in_I=EVXData.convertPsiFromBtoQ_Basis(n,I.getPsiAt(Q_at_last_spike_in_I),I.ev);
    // normalize to 1
    TMatr.scaleVectorInPlace(psi_x_at_last_spike_in_I,1/Math.sqrt(TMatr.mulVV(psi_x_at_last_spike_in_I,Q_at_last_spike_in_I)));
    this.dpdt_at_last_spike_in_I=TMatr.getAMatrA(psi_x_at_last_spike_in_I,QQdpdt,n);
    this.dPsdt_at_last_spike_in_I=TMatr.getAMatrA(psi_x_at_last_spike_in_I,QQdPsdt,n);
    this.w_at_last_spike_in_I=1/Math.pow(TMatr.mulVV(psi_x_at_last_spike_in_I,Q_at_last_spike_in_I),2);
    this.PsChange_at_last_spike_in_I=TMatr.mulVVn1(getPsi2WIntegratedDt(psi_x_at_last_spike_in_I),sst.getDpMoments());
    this.DIRdPs=dPsdt_at_last_spike_in_I*w_at_last_spike_in_I;

    final double
	psiHa[]=psiH,
	psia[]=true?psi_x_at_last_spike_in_I:M.B.getONE(),
	pI_Ha=TMatr.getAMatrA(psiHa,QQpi,n),
	psiHa_psia=TMatr.getAMatrB(psiHa,psia,QQt,n);
    var1pI_x=2*(TMatr.getAMatrB(psiHa,psia,QQpi,n)-pI_Ha*psiHa_psia)*psiHa_psia;


    if(eps>0){
	final SMomentsData moms=getSMoments(2*n-1);
	UnitTests.checkDoubleArrayRel1(moms.imom,"moms.imom",sst.getDvorigMoments(),"sst.getDvorigMoments()",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_x_at_last_spike_in_I,QQt,n),"psi_x",1,"1",eps);
	final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	    psixm=TMatr.getMatr_x_A(n,IQQt,Q_at_last_spike_in_I);
	TMatr.scaleVectorInPlace(psixm,1/Math.sqrt(TMatr.getAMatrA(psixm,QQt,n)));
	UnitTests.checkDoubleArrayRel1(psi_x_at_last_spike_in_I,"psi_x_at_last_spike_in_I",psixm,"psixm",eps);
	UnitTests.checkDoubleRel(w_at_last_spike_in_I,"w_at_last_spike_in_I",Math.pow(TMatr.getAMatrB(psi_x_at_last_spike_in_I,M.B.getONE(),QQt,n),2),"<psix|1>^2",eps);
	UnitTests.checkDoubleRel(w_at_last_spike_in_I,"w_at_last_spike_in_I",1/TMatr.sumElements(TMatr.squareArrayElements(I.getPsiAt(Q_at_last_spike_in_I))),"K(x_at_last_spike_in_I)",eps);
	if(0==Double.compare(I.x0,x_at_last_spike_in_I)) UnitTests.checkDoubleArrayRel1(psi_x_at_last_spike_in_I,"psi_x_at_last_spike_in_I",I.psi0,"I.psi0",eps);
	UnitTests.checkDoubleRel(mm.totalvolume,"mm.totalvolume",sst.getSumdVorig(),"sst.getSumdVorig()",eps);
	UnitTests.checkDoubleRel(mm.totalvolume,"mm.totalvolume",sst.allshares,"sst.allshares",eps);
	System.err.println("SupportPrice: OK n="+n+" x_at_last_spike_in_I="+x_at_last_spike_in_I+" eps="+eps);
    }
}


public SMomentsData getSMoments(final int nmoms)
{
    // Whether to use sampled or integrated by parts t-moments.
    final boolean flag_use_samped_t_moments=false;
    return SMomentsData.get_MomentsData(nmoms,odata,getAccess(),flag_use_samped_t_moments);
}
// Basis-dependence hooks.
abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double [] getEDPsi(final double [] psi);
/** Square a wavefunction, then integrate by parts.
 */
abstract double [] getPsi2WIntegratedDt(final double [] psi);
abstract WIntegrator getEmptyWIntegrator();

/** A class for unit test.
*/
static abstract class SupportPriceTest<DTYPE,P extends SupportPrice<DTYPE>> extends QTest {

P getSupportPrice(){
    // requires a call for every tick for internal WIntegration
    P b=getSupportPriceWithThisRef(0,0,null);
    for(int k=0;k<tb.length;k++) b=getSupportPriceWithThisRef(k+1,tb[k].time,b);
    return getSupportPriceWithThisRef(tb.length,currenttime,b);
}
abstract P getSupportPriceWithThisRef(final int tsize_to,final long currenttime,P ref);

SupportPriceTest(final String name,final java.util.Random r,final double eps)
{
    super(name,r,eps);
}



void test_Sequence_Calculations(final int tsize,final int neq)
{
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":ScalpedSupportPriceTest:test_Sequence_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);
    setData(tsize,neq);
    test_Sequence();
    System.err.println(name+":ScalpedSupportPriceTest.test_Sequence_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void test_Sequence()
{
    P b=getSupportPriceWithThisRef(0,0L,null);
    final DataInterfaceToMoments<DTYPE> access=b.getAccess();
    for(int tk=0;tk<tb.length;tk++){
	final int tsize_to=tk+1;

	// test the SupportPrice
	b=getSupportPriceWithThisRef(tsize_to,tb[tk].time,b);
	final int nmoms=b.sst.wi.ntm;
	final double [] imomorig=new double [nmoms],tmomorig=new double [nmoms];
	access.setM(imomorig,0,access.M_I,b.odata,nmoms);
	access.setM(tmomorig,0,access.M_T,b.odata,nmoms);

	UnitTests.checkDoubleArrayRel(b.sst.getDvorigMoments(),"b.sst.getDvorigMoments()",imomorig,"imomorig",eps);
	// different dt interval,
	// see test_Sequence() in ScalpedMaxIProjection.java
	//UnitTests.checkDoubleArrayRel(b.sst.getDtMoments(),"b.sst.getDtMoments()",tmomorig,"tmomorig",eps);
	UnitTests.checkDoubleArrayRel(b.sst.getDvMoments(),"b.sst.getDvMoments()",imomorig,"imomorig",eps);

	//System.err.println("test_Sequence()="+tk+" tb[tk].price="+tb[tk].price+" b.sst.P="+b.sst.P);
    }
    b=getSupportPriceWithThisRef(tb.length,currenttime,b);
    final P bo=getSupportPrice();
    matchSupportPrice(b,bo,eps);
}


void test_Extra(){}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":SupportPriceTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Extra();

    System.err.println(name+":SupportPriceTest.test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_min,final int neq_max)
{
    System.err.println(name+":SupportPriceTest: starting tests; ver=$Id: SupportPrice.java,v 1.125 2022/07/28 08:04:53 mal Exp $");
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=eps;
    final long t0=System.currentTimeMillis();
    for(int neq=neq_min;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	    // make at smaller size, this test is slow O(n^2)
	    test_Sequence_Calculations(10+r.nextInt(100),neq);

	}
    }
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;
    final long t1=System.currentTimeMillis();
    System.err.println(name+":SupportPriceTest.main: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: SupportPrice.java,v 1.125 2022/07/28 08:04:53 mal Exp $");
}
} // END OF class SupportPriceTest


static void matchSupportPrice(final SupportPrice<?> a,final SupportPrice<?> b,final double eps){

    EVXData.matchEVXData(a.I,b.I,false,eps);

    UnitTests.checkDoubleRel(a.dpdt_IH,"a.dpdt_IH",b.dpdt_IH,"b.dpdt_IH",eps);
    UnitTests.checkDoubleRel(a.sst.P_of_last_spike_in_I,"a.sst.P_of_last_spike_in_I",b.sst.P_of_last_spike_in_I,"b.sst.P_of_last_spike_in_I",eps);
    UnitTests.checkDoubleRel(a.PsChange_at_last_spike_in_I,"a.PsChange_at_last_spike_in_I",b.PsChange_at_last_spike_in_I,"b.PsChange_at_last_spike_in_I",eps);
    UnitTests.checkDoubleRel(a.dPsdt_at_last_spike_in_I,"a.dPsdt_at_last_spike_in_I",b.dPsdt_at_last_spike_in_I,"b.dPsdt_at_last_spike_in_I",eps);
    UnitTests.checkDoubleRel(a.dpdt_at_last_spike_in_I,"a.dpdt_at_last_spike_in_I",b.dpdt_at_last_spike_in_I,"b.dpdt_at_last_spike_in_I",eps);

    System.err.println("SupportPrice:matchSupportPrice test OK eps="+eps);
}

}
