package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2020
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.EVXData;
import com.polytechnik.freemoney.*;

/** Save the state. The class used in {@link SupportPrice SupportPrice}
 * and {@link SimpleSupportPrice SimpleSupportPrice}.
 * @version $Id: SupportPriceStateSaver.java,v 1.27 2022/07/10 10:59:04 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class SupportPriceStateSaver {

    public final long seq,time,allshares;
    /** The price, it propagates directly from constructor argument.
     */
    public final double P;
    /** The eigenvectors/eigenvalues of the dV/dt operator.
     */
    public final EVXData I;
    /** The operator I=dV/dt eigenvector index, that has the maximal
     * {@literal <psi_k|psi0>}<sup>2</sup>. The value equals to 0
     * means the dV/dt is very small. The value equals to I.neq-1 means
     * the dV/dt is very large. The value equals to -1 means eigenproblem solution failed.
     * See "<a href="https://www.researchgate.net/publication/324171914">The Spectral Approach to Timeserie Bursts Analysis</a>" article, DOI:10.24411/0131-5226-2018-10010.
     */
    public final int index_max_psi_projection;
    /** Support price; It is calculated as the price of last spike in I=dV/dt.
     */
    public final double P_of_last_spike_in_I,P_of_last_execution;
    public final long time_of_last_spike_in_I,time_of_last_execution;
    /** If true --- the "last spike" event was saved,
     * if false -- it was carried out from previous observation.
     */
    public final boolean flag_last_spike_event_was_saved;
    /** Recurrent integrator.
     */
    final WIntegrator wi;

    /** Initial state.
     */
    SupportPriceStateSaver(final WIntegrator wi,
			   final long time,final double P,final long dV,
			   final EVXData I){
	// first call, no previous state available
	this.time=time;
	this.P=P;
	this.allshares=dV;
	this.I=I;
	this.wi=wi;if(wi.nobservables!=5/* five time sequences to save */) throw new IllegalArgumentException();
	this.index_max_psi_projection=-1;

	this.seq=0;
	if(P>0 && dV>0){
	    P_of_last_spike_in_I=P_of_last_execution=P;
	    time_of_last_spike_in_I=time_of_last_execution=time;
	    flag_last_spike_event_was_saved=true;
	} else {
	    P_of_last_spike_in_I=P_of_last_execution=Double.NaN;
	    time_of_last_spike_in_I=time_of_last_execution=0;
	    flag_last_spike_event_was_saved=false;
	}
    }

    /** Recurrently updated state.
     */
    SupportPriceStateSaver(final SupportPriceStateSaver prev,
			   final long time,final double P,final long dV,
			   final EVXData I)
    {
	this.time=time;
	this.P=P;
	this.allshares=prev.allshares+dV;
	this.I=I;
	this.wi=prev.wi;

	this.seq=1+prev.seq;

	// find the index of the maximal <psi0|psik>^2 projection
	double wmax=0,sw=0;int indwmax=-1;
	for(int k=0;k<I.neq;k++){
	    final double wk=Math.pow(I.psik_at_x0[k],2)/I.o_o;
	    sw+=wk;
	    if(wk>=wmax){
		wmax=wk;
		indwmax=k;
	    }
	}
	this.index_max_psi_projection=indwmax;

	if(P>0 && dV>0 && index_max_psi_projection==I.neq-1 && I.sH>prev.I.sH ){
	    // this is a spike in I=dV/dt and the price is valid and executed volime dV>0
	    P_of_last_spike_in_I=P;
	    time_of_last_spike_in_I=time;
	    flag_last_spike_event_was_saved=true;
	} else {
	    P_of_last_spike_in_I=prev.P_of_last_spike_in_I;
	    time_of_last_spike_in_I=prev.time_of_last_spike_in_I;
	    flag_last_spike_event_was_saved=false;
	}
	if(P>0 && dV>0){
	    P_of_last_execution=P;
	    time_of_last_execution=time;
	} else {
	    P_of_last_execution=prev.P_of_last_execution;
	    time_of_last_execution=prev.time_of_last_execution;
	}
    }


void updateState(long currenttime,double dt,double dp,double dv,double dvorig){
    final long t=currenttime;
    // five arguments, the order: dt:[0], dp:[1], dv:[2], abs dp:[3], dv original [4]
    wi.updateSavedMomentsToNewTimeReference(currenttime);
    wi.addSingleObservation(currenttime,t,new double []{dt/*[0]*/,dp/*[1]*/,dv/*[2]*/,Math.abs(dp)/*[3]*/,dvorig/*[4]*/});
}


    public double getSumDp(){return wi.sum[1/* dp */];}
    double getSumdVorig(){return wi.sum[4/* dvorig */];}

    double [] getDtMoments(){return wi.integration_moments[0/* dt */];}
    double [] getDpMoments(){return wi.integration_moments[1/* dp */];}
    double [] getDvMoments(){return wi.integration_moments[2/* dv */];}
    double [] getAbsDpMoments(){return wi.integration_moments[3/* abs dp */];}
    double [] getDvorigMoments(){return wi.integration_moments[4/* dvorig */];}

} // END OF class SupportPriceStateSaver
