package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.realtimebook.*;

/** TQVM Test QVM.
 *  @version $Id: TQVM.java,v 1.46 2022/08/08 20:10:47 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class TQVM<DTYPE> {

final DTYPE odata;
final DataInterfaceToMoments<DTYPE> access;
final RurikTQVM vs,ts,bvs,bts,dpv,brvs;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> O;

TQVM(final int neq,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();
    this.O=getABasis();

    // +2: Two last moments are not actually used. Just to use the code from OrthogonalPolynomialsBasisFunctionsCalculatable<?extends BasisFunctionsCalculatable>
    final int nmoms=2*neq-1+2;

    final double [] tmoms=new double [nmoms],vmoms=new double [nmoms],
	vsmoms=new double [nmoms],tsmoms=new double [nmoms],
	rvsmoms=new double [nmoms],rtsmoms=new double [nmoms],dpmoms=new double [nmoms];
    access.setMT(tmoms,nmoms);
    access.setM(vmoms,0,access.M_I,odata,nmoms);
    access.setM(vsmoms,0,access.M_VS,odata,nmoms);
    access.setM(tsmoms,0,access.M_TS,odata,nmoms);
    access.setM(rvsmoms,0,access.M_RVS,odata,nmoms);
    access.setM(rtsmoms,0,access.M_RTS,odata,nmoms);
    access.setM(dpmoms,1,access.M_DP,odata,nmoms);
    this.vs=getRurikTQVM(neq,vmoms,tsmoms);
    this.ts=getRurikTQVM(neq,tmoms,vsmoms);
    bvs=getRurikTQVM(neq,vmoms,rvsmoms);
    bts=getRurikTQVM(neq,tmoms,rtsmoms);
    dpv=getRurikTQVM(neq,vmoms,dpmoms);
    brvs=getRurikTQVM(neq,tmoms,rvsmoms);
}

/** A class to store calculated data. */
static class RurikTQVM {
final double w0,X0,K00,nf0,sigma2_nf0,nf0B,sigma2_nf0B,
    of0,sigma2_of0,af0,sigma_af0;
    final double [] xnodes,wnodes,fnodes,ffnodes;
RurikTQVM(final double w0,final double X0,
	  final double K00,final double [] xnodes,final double [] wnodes,final double [] fnodes,final double [] ffnodes,
	  final double nf0,final double sigma2_nf0,
	  final double of0,final double sigma2_of0,
	  final double nf0B,final double sigma2_nf0B,
	  final double af0,final double sigma_af0)
{
    this.w0=w0;
    this.X0=X0;
    this.nf0=nf0;
    this.sigma2_nf0=sigma2_nf0;
    this.nf0B=nf0B;
    this.sigma2_nf0B=sigma2_nf0B;
    this.of0=of0;
    this.sigma2_of0=sigma2_of0;
    this.af0=af0;
    this.sigma_af0=sigma_af0;
    this.K00=K00;
    this.xnodes=xnodes;
    this.wnodes=wnodes;
    this.fnodes=fnodes;
    this.ffnodes=ffnodes;
}
} // END OF class RurikTQVM

RurikTQVM getRurikTQVM(final int n,final double [] vmom,final double [] fmom)
{
    final double [] xnodes=O.getRadauQuadratureFromPolynomials(n-1,O.getOrthogonalPolynomialsFirstKind(n+1,vmom)).x;
    if(xnodes.length!=n) throw new RuntimeException("xnodes.length="+xnodes.length+" n="+n);
    final double [] QQv=O.B.getQQMatr(n,vmom),QQf=O.B.getQQMatr(n,fmom),
	IQQv=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQv),
	IQQf=TMatr.mulMMM(n,IQQv,QQf,IQQv);
    final double [] kkv=O.B.getKK(n,IQQv),kkf=O.B.getKK(n,IQQf),kka=TMatr.getMatr_x_A(n,IQQv,java.util.Arrays.copyOf(fmom,n));
    double sigma2_of0=0,sigma2_nf0B=0,sigma_af0=0,of0=0,nf0=0,nf0B=0,K00=Double.NaN,X0=Double.NaN;
    final double [] wnodes=new double [n],fnodes=new double [n],ffnodes=new double [n];
    for(int k=0;k<n;k++)
    {
	final double x=xnodes[k],K=O.B.getD0(kkv,x),w=wnodes[k]=1/K,
	    fo=fnodes[k]=O.B.getD0(kka,x),fn=ffnodes[k]=O.B.getD0(kkf,x)/K;
	if(k<(n-1))
	{
	    sigma2_of0+=(fo*fo)*w;
	    sigma2_nf0B+=(fn*fn)*w;
	    sigma_af0+=Math.abs(fn)*w;
	}
	else // the last point is Radau edge node
	{
	    final double sqrtK00=Math.sqrt(K);
	    of0=fo/sqrtK00;nf0=fn;nf0B=nf0/sqrtK00;K00=K;
	    X0=x;
	}
    }
    return new RurikTQVM(vmom[0],X0,K00,xnodes,wnodes,fnodes,ffnodes,
			 nf0,TMatr.getSpurM(n,TMatr.mulMM(n,QQf,IQQf))-nf0*nf0,of0,sigma2_of0,
			 nf0B,sigma2_nf0B,nf0/K00,sigma_af0);
}

abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();



static abstract class TQVMTest<DTYPE> {

int neq;
long currenttime,tau;
Trade [] tb;

abstract void setData(final int tsize,final int neq);
abstract TQVM<DTYPE> getTQVM();
abstract ZQVM<DTYPE> getZQVM();
abstract ZVQVM<DTYPE> getZVQVM();

final String name;
final java.util.Random r;
final double eps;
TQVMTest(final String name,
	 final java.util.Random r,final double eps)
{
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Common()
{
    final TQVM<DTYPE> q=getTQVM();

    System.err.println(name+":TQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}
void match_to_RurikTQVM(final int n,final RurikTQVM rurik,final double X0,final double K00,
			final double nf0,final double sigma2_nf0,final double of0,final double sigma2_of0,
			final double nf0B,final double sigma2_nf0B,final double af0,final double sigma_af0)
{
    UnitTests.checkDoubleRel(rurik.X0,"rurik.X0",X0,"X0",eps);
    UnitTests.checkDoubleRel(rurik.K00,"rurik.K00",K00,"K00",eps);
    UnitTests.checkDoubleRel(rurik.w0,"rurik.w0",TMatr.sumElements(rurik.wnodes),"sum(rurik.wnodes)",eps);

    UnitTests.checkDoubleRel(rurik.nf0,"rurik.nf0",nf0,"nf0",eps);
    UnitTests.checkDoubleRel(rurik.sigma2_nf0,"rurik.sigma2_nf0",sigma2_nf0,"sigma2_nf0",eps);
    UnitTests.checkDoubleRel(rurik.of0,"rurik.of0",of0,"of0",eps);
    UnitTests.checkDoubleRel(rurik.sigma2_of0,"rurik.sigma2_of0",sigma2_of0,"sigma2_of0",eps);

    if(!Double.isNaN(nf0B)) UnitTests.checkDoubleRel(rurik.nf0B,"rurik.nf0B",nf0B,"nf0B",eps);
    if(!Double.isNaN(sigma2_nf0B))UnitTests.checkDoubleRel(rurik.sigma2_nf0B,"rurik.sigma2_nf0B",sigma2_nf0B,"sigma2_nf0B",eps);
    if(!Double.isNaN(af0)) UnitTests.checkDoubleRel(rurik.af0,"rurik.af0",af0,"af0",eps);
    if(!Double.isNaN(sigma_af0)) UnitTests.checkDoubleRel(rurik.sigma_af0,"rurik.sigma_af0",sigma_af0,"sigma_af0",eps);

    System.err.println(name+":TQVMTest:match_to_RurikTQVM(): test OK neq="+neq+" eps="+eps);
}

void test_ZQVM()
{
    final TQVM<DTYPE> q=getTQVM();
    {
    final ZQVM<DTYPE> s=getZQVM();
    match_to_RurikTQVM(neq,q.ts,s.getX0(),s.K00,s.ntsvs0,s.sigma2T_ntsvs0,s.otsvs0,s.sigma2A_otsvs0,
		       Double.NaN,Double.NaN,Double.NaN,Double.NaN);
    match_to_RurikTQVM(neq,q.bts,s.getX0(),s.K00,s.ntrts0,s.sigma2T_ntrts0,s.otrts0,s.sigma2A_otrts0,
		       Double.NaN,Double.NaN,Double.NaN,Double.NaN);
    }
    {
    final ZVQVM<DTYPE> s=getZVQVM();
    match_to_RurikTQVM(neq,q.vs,s.getX0(),s.K00,s.nvsts0,s.sigma2T_nvsts0,s.ovsts0,s.sigma2A_ovsts0,
		       Double.NaN,Double.NaN,Double.NaN,Double.NaN);
    match_to_RurikTQVM(neq,q.bvs,s.getX0(),s.K00,s.nvrvs0,s.sigma2T_nvrvs0,s.ovrvs0,s.sigma2A_ovrvs0,
		       Double.NaN,Double.NaN,Double.NaN,Double.NaN);
    match_to_RurikTQVM(neq,q.dpv,s.getX0(),s.K00,s.nvdp0,s.sigma2T_nvdp0,s.ovdp0,s.sigma2A_ovdp0,
		       Double.NaN,Double.NaN,Double.NaN,Double.NaN);
    }

    System.err.println(name+":TQVMTest:test_ZQVM(): test OK neq="+neq+" eps="+eps);
}

void test_Lastpoint()
{
    final TQVM<DTYPE> q=getTQVM();
    final Trade [] tb=new Trade [this.tb.length+1];
    System.arraycopy(this.tb,0,tb,0,this.tb.length);
    final double pl=130.0+2*r.nextDouble();
    tb[this.tb.length]=new Trade(pl,
				 10000,
				 null,
				 r.nextInt(2)==1?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
				 currenttime,
				 -1,-1,-1,
				 null,Trade.TRADE_TYPE_REGULAR,
				 Trade.getDummyTrade(pl-10*r.nextDouble(),100,BookItem.BOOK_BUY,currenttime),
				 Trade.getDummyTrade(pl+10*r.nextDouble(),100,BookItem.BOOK_SELL,currenttime),
				 -1);
    this.tb=tb;
    final TQVM<DTYPE> s=getTQVM();

    UnitTests.checkDoubleRel(q.ts.sigma2_nf0,"q.ts.sigma2_nf0",s.ts.sigma2_nf0,"s.ts.sigma2_nf0",eps);
    UnitTests.checkDoubleRel(q.ts.sigma2_nf0B,"q.ts.sigma2_nf0B",s.ts.sigma2_nf0B,"s.ts.sigma2_nf0B",eps);
    UnitTests.checkDoubleRel(q.ts.sigma2_of0,"q.ts.sigma2_of0",s.ts.sigma2_of0,"s.ts.sigma2_of0",eps);
    UnitTests.checkDoubleRel(q.ts.sigma_af0,"q.ts.sigma_af0",s.ts.sigma_af0,"s.ts.sigma_af0",eps);

    //UnitTests.checkDoubleRel(q.vs.sigma2_nf0,"q.vs.sigma2_nf0",s.vs.sigma2_nf0,"s.vs.sigma2_nf0",eps); // does not match
    UnitTests.checkDoubleRel(q.vs.sigma2_nf0B,"q.vs.sigma2_nf0B",s.vs.sigma2_nf0B,"s.vs.sigma2_nf0B",eps);
    UnitTests.checkDoubleRel(q.vs.sigma2_of0,"q.vs.sigma2_of0",s.vs.sigma2_of0,"s.vs.sigma2_of0",eps);
    UnitTests.checkDoubleRel(q.vs.sigma_af0,"q.vs.sigma_af0",s.vs.sigma_af0,"s.vs.sigma_af0",eps);

    System.err.println(name+":TQVMTest:test_Lastpoint(): test OK neq="+neq+" eps="+eps);
}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":TQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_ZQVM();
    if(neq>1) test_Lastpoint();

    System.err.println(name+":TQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":TQVMTest: starting tests; ver=$Id: TQVM.java,v 1.46 2022/08/08 20:10:47 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(950+r.nextInt(850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":TQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: TQVM.java,v 1.46 2022/08/08 20:10:47 mal Exp $");
}
} // END OF class TQVMTest







}// END OF class TQVM
