package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.Polynomials;

/** Simplified two level models.
 *  @version $Id: TSTwoLevels.java,v 1.14 2022/07/28 21:03:03 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
class TSTwoLevels {
final double I0a,I0b,f0a,f0b,IL,IH,fL,fH;
TSTwoLevels(final double I0a,final double I0b,final double f0a,final double f0b,final double IL,final double IH,final double fL,final double fH)
{
    this.I0a=I0a;
    this.I0b=I0b;
    this.f0a=f0a;
    this.f0b=f0b;
    this.IL=IL;
    this.IH=IH;
    this.fL=fL;
    this.fH=fH;
}
static TSTwoLevels getTSTwoLevelsDiagInd(final int n,final double [] QQi,final double [] QQf,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final int ia,final int ib)
{
    final double f0a=QQf[ia+n*ia],f0b=QQf[ib+n*ib],f0ab=QQf[ia+n*ib],
	I0a=QQi[ia+n*ia],I0b=QQi[ib+n*ib],I0ab=QQi[ia+n*ib];
    final double [] diag=new double []{I0a,I0b},offdiag=new double []{I0ab};
    final double [] ev=M.EV.getEV_Thridiagonal(2,diag,0,offdiag,0,true);

    return new TSTwoLevels(I0a,I0b,f0a,f0b,
		  ev[0],
		  ev[1],
		  f0a*Math.pow(ev[(2)+0],2)+f0ab*(2*ev[(2)+0]*ev[(2)+1])+f0b*Math.pow(ev[(2)+1],2),
		  f0a*Math.pow(ev[(2+2)+0],2)+f0ab*(2*ev[(2+2)+0]*ev[(2+2)+1])+f0b*Math.pow(ev[(2+2)+1],2));
}

static TSTwoLevels test_getTSTwoLevels(final double [] imom,final double [] tmom,final double [] fmom,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,final double [] psia,final double [] psib)
{
    final int n=Math.max(psia.length,psib.length);
    final double [] QQt=M.B.getQQMatr(n,tmom),QQi=M.B.getQQMatr(n,imom),QQf=M.B.getQQMatr(n,fmom);
    final double [] pn=new double [] {
	TMatr.getAMatrA(psia,QQi,n),
	2*TMatr.getAMatrB(psia,psib,QQi,n),
	TMatr.getAMatrA(psib,QQi,n)
    },
    pdn=new double [] {
	TMatr.getAMatrA(psia,QQt,n),
	2*TMatr.getAMatrB(psia,psib,QQt,n),
	TMatr.getAMatrA(psib,QQt,n)
    };
    final double [] _d=TMatr.sumAcB(
	Polynomials.mult2Pol(Polynomials.getPolynomialDifferentiated(pn),pdn),
	-1,
        Polynomials.mult2Pol(Polynomials.getPolynomialDifferentiated(pdn),pn)),
	d=java.util.Arrays.copyOf(_d,pn.length+pdn.length-1-1-1);
    final com.polytechnik.utils.PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(d);
    final double [] xr=u.re_root,xi=u.im_root;
    if(u.nroots!=2) xr[0]=xr[1]=Double.NaN;
    final double [] psiL=psia.clone(),psiH=psia.clone();
    if(Polynomials.getD1(d,xr[0])>0)
    {
	TMatr.sumP1(psiL,psib,xr[0]);
	TMatr.sumP1(psiH,psib,xr[1]);
    }
    else
    {
	TMatr.sumP1(psiL,psib,xr[1]);
	TMatr.sumP1(psiH,psib,xr[0]);
    }
    TMatr.scaleVectorInPlace(psiL,1/Math.sqrt(TMatr.getAMatrA(psiL,QQt,n)));
    TMatr.scaleVectorInPlace(psiH,1/Math.sqrt(TMatr.getAMatrA(psiH,QQt,n)));

    final double na=TMatr.getAMatrA(psia,QQt,n),
	nb=TMatr.getAMatrA(psib,QQt,n);

    return new TSTwoLevels(
	TMatr.getAMatrA(psia,QQi,n)/na,
	TMatr.getAMatrA(psib,QQi,n)/nb,
	TMatr.getAMatrA(psia,QQf,n)/na,
	TMatr.getAMatrA(psib,QQf,n)/nb,
	TMatr.getAMatrA(psiL,QQi,n),
	TMatr.getAMatrA(psiH,QQi,n),
	TMatr.getAMatrA(psiL,QQf,n),
	TMatr.getAMatrA(psiH,QQf,n)
		      );
}

void match_Values(final TSTwoLevels q,final double eps)
{
    System.err.println("I0a="+I0a+" I0b="+I0b+" IL="+IL+" IH="+IH);
    UnitTests.checkDoubleRel(I0a,"I0a",q.I0a,"q.I0a",eps);
    UnitTests.checkDoubleRel(I0b,"I0b",q.I0b,"q.I0b",eps);
    UnitTests.checkDoubleRel(f0a,"f0a",q.f0a,"q.f0a",eps);
    UnitTests.checkDoubleRel(f0b,"f0b",q.f0b,"q.f0b",eps);
    UnitTests.checkDoubleRel(IL,"IL",q.IL,"q.IL",eps);
    UnitTests.checkDoubleRel(IH,"IH",q.IH,"q.IH",eps);
    UnitTests.checkDoubleRel(fL,"fL",q.fL,"q.fL",eps);
    UnitTests.checkDoubleRel(fH,"fH",q.fH,"q.fH",eps);
    System.err.println("TSTwoLevels:match_Values: test OK eps="+eps);
}


} // END OF class TSTwoLevels
