package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.broker.*;
import com.polytechnik.outils.*;

/** A class that performs actual trading.
 *  @version $Id: Trader.java,v 1.234 2022/05/29 19:17:55 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Trader implements TraderInterface {
    /** The number of pre-set threads. */
    private static final int N_WORKERS=10;


    private static final boolean PRG_DEBUG=true;
    /** Whether not to wait for explicit 'status OK' message.
     */
    static final boolean DEFAULT_ALLOW_TO_TRADE=false;

    private final BrokerInterface BROKER;

    private final DataAnalyzable EXTERNALANALYZIS;
    private final WorkersPool WORKERSPOOL;
    private final ModelParameters mparameters;
    final int dump_trading_diagnostics;
    final boolean flag_realtime;
    final String description;

    private final FindOpportunityTask findopportunity;
    public Trader(final DataAnalyzable EXTERNALANALYZIS,
		  final int dump_trading_diagnostics,
		  final ModelParameters mparameters,
		  final BrokerInterface BROKER,
		  final boolean flag_singlethread,
		  final boolean flag_realtime,
		  final String description)
    {
	// single threaded for training
	this.WORKERSPOOL=new WorkersPool((flag_singlethread||EXTERNALANALYZIS!=null)?0:N_WORKERS,false);
	this.EXTERNALANALYZIS=EXTERNALANALYZIS;
	this.dump_trading_diagnostics=dump_trading_diagnostics;
	this.mparameters=mparameters;
	this.BROKER=BROKER;
	this.findopportunity=new FindOpportunityTask(EXTERNALANALYZIS!=null);
	this.flag_realtime=flag_realtime;
	this.description=description;
    }




    private final CalculateAntecedents calc_antecedents=new CalculateAntecedents();
    private final CalculateConsequents calc_consequents=new CalculateConsequents();

    private final StatusData getStatusData(final OrdersAndTrades ot,final long currenttime)
    {
	StatusData h=ot.getData();
	if(h==null)
	{
	    this.calc_antecedents.clearData();
	    this.calc_consequents.clearData();
	    final ModelParameters.ModelParametersData mp=
		this.mparameters.getParameters(ot.ticker);
	    h=new StatusData(ot.ticker,
			     this.flag_realtime,
			     new MoneyManager(mp.MAX_LONG,mp.MAX_SHORT),
			     this,
			     mp,
			     this.calc_antecedents,
			     this.calc_consequents,
			     ot,
			     this.EXTERNALANALYZIS!=null);
	    ot.setData(h);
	    ot.setActive(h.mparametersdata.flag_active); // do not store trades for unused stocks
	    if(this.EXTERNALANALYZIS!=null)
	    {
		if(h.mparametersdata.flag_active)
		{
		    // active ticker, init arrays
		    this.EXTERNALANALYZIS.setTicker(ot.ticker,
						    this.calc_antecedents,
						    this.calc_consequents,
						    h);
		}
	    }
	    else
	    {
		if(h.mparametersdata.flag_active)
		{
		    BROKER.readInitialPosition(h); // initial position from broker
		    h.DIAG.setTicker();
		}
	    }
	    this.calc_antecedents.clearData();
	    this.calc_consequents.clearData();
	}
	if(h.ticker!=ot.ticker || h.ot!=ot)
	{
	    throw new RuntimeException("Ticker or object mismatch:"+
				       " ticker1="+ot.ticker+
				       " and ticker2="+h.ticker+
				       " ot1="+h.ot+" ot2="+ot);
	}
	h.setTraderCurrentTime(currenttime);
	return h;
    }


    boolean allow_trading=DEFAULT_ALLOW_TO_TRADE;
    public void allowTrading(final long currenttime,final boolean on_off)
    {
	this.allow_trading=on_off;
    }


    public void allowTrading(final OrdersAndTrades ot,final long currenttime,final int value,final String text)
    {
	final StatusData sd=getStatusData(ot,currenttime);
	System.err.println("Trader:allowTrading: for ticker="+sd.ticker+" value="+value+" msg="+text);
	if(value==TraderInterface.SEC_TRADING_OK) sd.setAllowTrading(true);
	else if(value==TraderInterface.SEC_TRADING_DISABLED) sd.setAllowTrading(false);
	// shorts can be handled separately
    }


    public void setMarketMaker(final OrdersAndTrades ot,
			       final long currenttime,
			       final String mpid,
			       final boolean flag_primary,
			       final boolean flag_ok)
    {
	// FIXME, not implemented yet
    }


    public void setInstrument(final OrdersAndTrades ot,
			      final long currenttime,
			      final boolean flag_statusOK,
			      final int round_lot_shares,
			      final boolean round_lot_only,
			      final String market)
    {
	//System.err.println("TICKER="+ot.ticker);
	getStatusData(ot,currenttime).setStatus(flag_statusOK,
				    round_lot_shares,
				    round_lot_only,
				    market);
    }


    private class FindOpportunityTask implements WorkerTask {
	StatusData sd=null;
	long id=-1;

	private final boolean flag_training;
	private final TradeOpportunityChecker [] opp;
	FindOpportunityTask(final boolean flag_training)
	{
	    this.flag_training=flag_training;
	    this.opp=new TradeOpportunityChecker [N_WORKERS];
	    for(int i=this.opp.length;i-- >0;)
	    {
		this.opp[i]=new TradeOpportunityChecker(EXTERNALANALYZIS);
	    }
	}
	public void initializeTask(final int tasknum)
	{
	    // called synchroneously, no race can occur.
	    this.sd.processstarted=true;
	    final long currenttime=this.sd.getTraderCurrentTime();
	    this.sd.time_processstarted=currenttime;

	    // copy all data to trade opportunity checker
	    this.opp[tasknum].setData(currenttime,this.sd,this.id);
	    // clear the data
	    this.sd=null;
	    this.id=-1;
	}
	public void cleanupTask(final int tasknum){}

	public void performTask(final int tasknum)
	{
	    final TradeOpportunityChecker checker=this.opp[tasknum];
	    final long magicid=checker.getMagicID();
	    boolean sameidok=false;
	    final StatusData sd=checker.tinfo;
	    long tau=-1L;

	    try {
		if(!sd.processstarted)
		{
		    throw new RuntimeException("Something is terrible wrong for ticker="+sd.ticker+" time="+checker.currenttime+" processstarted="+sd.processstarted);
		}

		// a caller makes sure that only SINGLE
		// processor for a given ticker is running.
		// No race can occur
		final CalculationData cd;
		// prepare internal structures
		checker.prepareBook();
		// do calculations
		if(this.flag_training)
		{
		    // calculate and save consequent
		    checker.training_CalculateConsequent(checker.calculateRef());
		    // calculate and save antecedent
		    // we perform antecedent calculations at the end because
		    // this poison buffers in checker.
		    cd=checker.calculateEverything();

		    // since Oct 15 2008 we do not use past CalculationData
		    if(sd.pastcalculations.training_last_data_processed!=null)
		    {
			sd.pastcalculations.clearBefore(
			    sd.pastcalculations.training_last_data_processed.getTime(),null);
		    }
		    tau=checker.currenttime+sd.mparametersdata.tau_training_quant;
		}
		else
		{

		    BROKER.hookForDebugging(new Object []{sd.ticker,
							  checker.savedtrades,
							  Double.valueOf(checker.p_buy),
							  Double.valueOf(checker.p_sell)},checker.currenttime);

		    // actual real time trading may require more fine-grained calculation
		    // so we do lazy calculations, when we need.
		    cd=checker.calculateNothing();

		    // actual fine-grained calculations in predictor
		    sd.PREDICTOR.updateInternalState(cd);

		    // trade if opportunity is good.
		    tau=sd.PREDICTOR.tradeIfOpportunityIsGood(BROKER,
							      cd.getTime(),
							      cd.p_buy,
							      cd.p_sell,
							      cd.vol_p_buy,
							      cd.vol_p_sell);

		    sd.DIAG.dumpTickData(cd);

		    if(sd.pastcalculations.getSize()!=0)
		    {
			throw new RuntimeException("The size of past reference for "+sd.ticker+
						   "="+sd.pastcalculations.getSize()+
						   "; must be =0.");
		    }
		}
	    } finally {
		// Clear opportunity checker data.
		sameidok=(magicid==checker.clear());
		// Clear predictor data
		if(sd.PREDICTOR!=null)
		{
		    sd.PREDICTOR.clear();
		}
		// race cannot occur if no two processors of a singe stock are run at once
		sd.time_processstarted=0;
		if(tau>=0)
		{
		    sd.setNextCheck(tau);
		}
		sd.processstarted=false;
	    }
	    if(tau<0L)
	    {
		throw new RuntimeException("Invalid tau="+tau+" for ticker="+sd.ticker);
	    }
	    if(!sameidok)
	    { // no exception & id mismatch
		throw new RuntimeException("Something is terrible wrong with threading for ticker="+sd.ticker+" time="+checker.currenttime+" processstarted="+sd.processstarted);
	    }
	}
    }


    private long id=0;
    private long getID()
    {
	return ++this.id;
    }


    public void update(final OrdersAndTrades ot,
		       final long currenttime,
		       final int eventid)
    {

	if(false && (eventid==TraderInterface.UPD_EXECUTIONORDER_TRADE ||
		     eventid==TraderInterface.UPD_EXECUTIONOTH_TRADE ||
		     eventid==TraderInterface.UPD_BROKENTRADE_TRADE))
	{
	    // skip notifications from exe,
	    // because they may cummulate
	    return;
	}

	final StatusData sd=getStatusData(ot,currenttime);
	/*
	if(sd.mparametersdata.flag_active)
	{
	    System.err.println("eventid="+eventid+
			       " ticker="+ot.ticker+
			       " currenttime="+BookItem.getTimeDiag(currenttime)+
			       " tradebook.getSize()="+ot.tradebook.getSize()+
			       " orderbook.getSize()="+ot.orderbook.getSize());
	}
	*/

	if(sd.processstarted)
	{
	    if(PRG_DEBUG && currenttime>sd.mparametersdata.tau_training_quant+sd.time_processstarted)
	    {
		   System.err.println(
		       "Trader: work skipped because a one for ticker="+
		       ot.ticker+" is already running started at t="+
		       BookItem.getTimeDiag(sd.time_processstarted)+
		       " now="+BookItem.getTimeDiag(currenttime)+
		       " diff="+(currenttime-sd.time_processstarted));
	    }
	    return; // pending excecution, skip this work
	}
	if(sd.ot!=ot)
	{
	    throw new RuntimeException("Something is terrible wrong with OrdersAndTrades: "+
				       " ot1="+ot+" ot2="+sd.ot+
				       " ticker="+sd.ticker+" time="+currenttime);
	}


	long tclear=-1;
	if(!sd.mparametersdata.flag_active)
	{
	    if(ot.tradebook.getSize()>0)
	    {
		throw new RuntimeException("There should be no trades here "+
					   " ticker="+ot.ticker+
					   " ot.tradebook.getSize()="+ot.tradebook.getSize());
	    }
	}
	else if(sd.shouldCheck(currenttime))
	{
	    // perform trade opportunity check
	    // set data available to workers
	    this.findopportunity.sd=sd;
	    this.findopportunity.id=getID();
	    // perform work.
	    final boolean work_started=this.WORKERSPOOL.performWork(this.findopportunity);
	    if(PRG_DEBUG)
	    {
		if(!work_started)
		{
		   System.err.println("Trader: work skipped becauseus too "+
				      "many tasks are already started; ticker="+
				      ot.ticker+" time="+currenttime);
		}
	    }

	    if(sd.mparametersdata.tau_readdata_clearBefore_delay>=0)
	    {
		// positive sd.mparametersdata.tau_readdata_clearBefore_delay
		// means do clean old trades.
		if(this.EXTERNALANALYZIS!=null)
		{
		    // when we are here, no execution is pending, no race possible.
		    final CalculationDataRef training_last_processed=sd.pastcalculations.training_last_data_processed;
		    tclear=training_last_processed!=null?(training_last_processed.getTime()-sd.mparametersdata.tau_readdata_clearBefore_delay):-1L;
		}
		else
		{
		    tclear=currenttime-sd.mparametersdata.tau_readdata_clearBefore_delay;
		}
	    }
	    if(!work_started)
	    {
		tclear=-1; // do not clear old data if some work is skipped
	    }
	}

	// clear old trades
	if(tclear>=0L)
	{
	    /*
	    System.err.println("CLEAR ticker="+sd.ticker+
			       " lastprocessed="+sd.pastcalculations.training_last_data_processed+
			       " tlast="+(sd.pastcalculations.training_last_data_processed!=null?
					  sd.pastcalculations.training_last_data_processed.getTime():-1)+
			       " dt="+((currenttime-tclear)/1e9));
	    */
	    ot.clearBefore(tclear);
	}

    }

}
