package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.EVXData;
import com.polytechnik.freemoney.*;

/**
 * Further development of SupportPrice.java.
 * @version $Id: TstPnLILowIHigh.java,v 1.43 2022/07/28 08:04:53 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
*/
public abstract class TstPnLILowIHigh<DTYPE> {


public final TstPnLILowIHighStateSaverWithWI sst;

/** "Regular", not scalp--moments. The object has
 * an internal state to recurrently speed up the calculations.
 */
final DTYPE odata;

public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getABasis();

/** The |I|psi>=lambda |psi> solution.
 */
public final EVXData I;

public final double
// Helper data and moving averages.
    p_offset,p_last,pi_average,pt_average,
// The https://arxiv.org/abs/1510.05510 results (the state of max I).
    dpdt_0,p_IH,pt_IH,std_p_IH,std_pt_IH
    ;

public final EVXData dIHpnl;
/** The <a href="https://arxiv.org/abs/1903.11530">scalp price</a>
 * style results.
 */
public final double
    DP_since_dIH_H,AbsDP_since_dIH_H,
    P_at_dIH_H,DP_at_dIH_H,AbsDP_at_dIH_H;


/** Not worth a function or constructor argument.
 */
private static double EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;

/**
* Constructor.
* @param n Basis dimension.
* @param odata Regular moments internal state.
* @param rst Previous tick internal state.
*/
TstPnLILowIHigh(final int n,final DTYPE odata,TstPnLILowIHighStateSaverWithWI rst)
{
    final double eps=EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING; // turn on selftests if required
    this.odata=odata;
    final SMomentsData mm=getSMoments(2*n-1).asQQMatrices(n,M);
    // mm has the matrices (not moments) stored.
    final double [] QQt=mm.tmom,QQi=mm.imom,QQpi=mm.pimom,QQdpdt=mm.dpmom,QQp1t=mm.p1tmom,QQp0t=mm.p0tmom,QQp2t=mm.p2tmom,QQp2i=mm.p2imom;
    p_offset=mm.p_offset;
    p_last=mm.p_last;
    pi_average=QQpi[0]/QQi[0];
    pt_average=QQp1t[0]/QQp0t[0];
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    final double [] psiH=I.getPsi(n-1/* max I */),psi0=I.psi0;
    dpdt_0=TMatr.getAMatrA(psi0,QQdpdt,n);
    p_IH=TMatr.getAMatrA(psiH,QQpi,n)/TMatr.getAMatrA(psiH,QQi,n);
    pt_IH=TMatr.getAMatrA(psiH,QQp1t,n)/TMatr.getAMatrA(psiH,QQp0t,n);
    std_p_IH=Math.sqrt(TMatr.getAMatrA(psiH,TMatr.sumAcBdE(QQp2i,-2*p_IH,QQpi,p_IH*p_IH,QQi),n)/TMatr.getAMatrA(psiH,QQi,n));
    std_pt_IH=Math.sqrt(TMatr.getAMatrA(psiH,TMatr.sumAcBdE(QQp2t,-2*pt_IH,QQp1t,pt_IH*pt_IH,QQp0t),n)/TMatr.getAMatrA(psiH,QQp0t,n));


    if(rst==null){
	rst=new TstPnLILowIHighStateSaverWithWI(getEmptyWIntegrator(5));
    }
    // Preserve the state, to obtain the "scalped" moments.
    // For "regular" moments the DTYPE also has an internal state :(
    this.sst=new TstPnLILowIHighStateSaverWithWI(rst,mm.currenttime,p_last+p_offset /* All prices are relatively p_offset */,mm.totalvolume-rst.allshares,I);
    sst.wi.updateSavedMomentsToNewTimeReference(sst.time);


    double dt=rst.time<=0?0:(sst.time-rst.time)/sst.wi.atau,
	dv=(sst.allshares-rst.allshares),
	dIH=rst.I==null?sst.I.sH:sst.I.sH-rst.I.sH,
	dP=sst.P//(sst.P-rst.P)
	;
    if(!(dIH>0 && Double.isFinite(dP))){dIH=dP=0;}

    // Alternavive integration, here a more general function can be used (e.g. depending on I).
    // The sst.getDvMoments() have to match getSMoments(2*n-1).imom exactly as unit test.
    sst.addObservation(/*currenttime=*/sst.time,dt,dv,dP*dIH,dP*dIH,dIH);

    final double [] QQdIHpnl=M.B.getQQMatr(n,sst.getDIHpnlMoments());
    this.dIHpnl=EVXData.getEVXDataFromQQ(n,QQdIHpnl,QQt,M);
    final double [] psiHdIH2Integrated=getPsi2WIntegratedDt(dIHpnl.psiH);
    this.DP_since_dIH_H=TMatr.mulVV(psiHdIH2Integrated,sst.getPcDIHpnlMoments());
    //TMatr.mulVV(psiHdIH2Integrated,sst.getDIHpnlMoments());
    this.AbsDP_since_dIH_H=TMatr.mulVV(psiHdIH2Integrated,sst.getAbsPcDIHpnlMoments());
    //TMatr.mulVV(psiHdIH2Integrated,sst.getDIHpnlMoments());


    // d/dt at |psiH>
    final double [] QQPcdIHpnl=M.B.getQQMatr(n,sst.getPcDIHpnlMoments()),
	QQAbsPcdIHpnl=M.B.getQQMatr(n,sst.getAbsPcDIHpnlMoments()),
	dIHpnl_DpsiH=getEDPsi(dIHpnl.psiH);
    this.P_at_dIH_H=TMatr.getAMatrA(dIHpnl.psiH,QQPcdIHpnl,n)/dIHpnl.sH;
    this.DP_at_dIH_H=2*(TMatr.getAMatrB(dIHpnl.psiH,dIHpnl_DpsiH,QQPcdIHpnl,n)/dIHpnl.sH-P_at_dIH_H*TMatr.getAMatrB(dIHpnl.psiH,dIHpnl_DpsiH,QQt,n));
    this.AbsDP_at_dIH_H=TMatr.getAMatrA(dIHpnl.psiH,M.B.getQQMatr(n,sst.getAbsPcDIHpnlMoments()),n)/dIHpnl.sH;

    if(eps>0){
	final double [] psiH2Integrated=getPsi2WIntegratedDt(psiH);
	final SMomentsData moms=getSMoments(2*n-1);
	UnitTests.checkDoubleArrayRel1(moms.imom,"moms.imom",sst.getDvMoments(),"sst.getDvMoments()",eps);
	UnitTests.checkDoubleRel(mm.totalvolume,"mm.totalvolume",sst.getSumDv(),"sst.getSumDv()",eps);
	UnitTests.checkDoubleRel(mm.totalvolume,"mm.totalvolume",sst.allshares,"sst.allshares",eps);
	// check integration by parts
	UnitTests.checkDoubleRel(TMatr.mulVV(psiH2Integrated,sst.getDvMoments()),"int dV Integral(psi2)",-TMatr.getAMatrA(psiH,mm.Vmom,n),"-<psiH|V|psiH>",eps);
	System.err.println("TstPnLILowIHigh: OK n="+n+" eps="+eps);
     }
}


public SMomentsData getSMoments(final int nmoms)
{
    // Whether to use sampled or integrated by parts t-moments.
    final boolean flag_use_samped_t_moments=false;
    return SMomentsData.get_MomentsData(nmoms,odata,getAccess(),flag_use_samped_t_moments);
}
// Basis-dependence hooks.
abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
/** Square a wavefunction, then integrate by parts.
 */
abstract double [] getPsi2WIntegratedDt(final double [] psi);
abstract WIntegrator getEmptyWIntegrator(int dim);
abstract double [] getEDPsi(final double [] psi);

/** A class for unit test.
*/
static abstract class PnLILowIHighTest<DTYPE,P extends TstPnLILowIHigh<DTYPE>> extends QTest {

P getTstPnLILowIHigh(){
    // requires a call for every tick for internal WIntegration
    P b=getTstPnLILowIHighWithThisRef(0,0,null);
    for(int k=0;k<tb.length;k++){
	b=getTstPnLILowIHighWithThisRef(k+1,tb[k].time,b);
	var state=b.sst;
	//System.err.println("REG:state.PriceChangePnL="+state.PriceChangePnL+" TPV=("+tb[k].time+","+tb[k].price+","+tb[k].shares+") IH="+state.I.sH);
    }
    return getTstPnLILowIHighWithThisRef(tb.length,currenttime,b);
}
    PnLILowIHighTPVFactory.PnLILowIHighTPV getTPV(){
	final PnLILowIHighTPVFactory f=getTPVFactory();
	PnLILowIHighStateSaverWithWI state=null;
	for(int k=0;k<tb.length;k++) {
	    state=f.getNext(tb[k].time,tb[k].price,tb[k].shares,state).sst;
	    //System.err.println("TPV:state.PriceChangePnL="+state.PriceChangePnL+" TPV=("+tb[k].time+","+tb[k].price+","+tb[k].shares+") IH="+state.I.sH);
	}
	return f.getNext(currenttime,tb[tb.length-1].price,0,state);
    }

    /** This class unit test.
     */
    abstract P getTstPnLILowIHighWithThisRef(final int tsize_to,final long currenttime,P ref);
    /** One more unit test. */
    abstract PnLILowIHighTPVFactory getTPVFactory();

PnLILowIHighTest(final String name,final java.util.Random r,final double eps)
{
    super(name,r,eps);
}



void test_Sequence_Calculations(final int tsize,final int neq)
{
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":PnLILowIHighTest:test_Sequence_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);
    setData(tsize,neq);
    test_Sequence();
    System.err.println(name+":PnLILowIHighTest.test_Sequence_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void test_Sequence()
{
    P b=getTstPnLILowIHighWithThisRef(0,0L,null);
    final DataInterfaceToMoments<DTYPE> access=b.getAccess();
    for(int tk=0;tk<tb.length;tk++){
	final int tsize_to=tk+1;

	// test the PnLILowIHigh
	b=getTstPnLILowIHighWithThisRef(tsize_to,tb[tk].time,b);
	final int nmoms=b.sst.wi.ntm;
	final double [] imomorig=new double [nmoms],tmomorig=new double [nmoms];
	access.setM(imomorig,0,access.M_I,b.odata,nmoms);
	access.setM(tmomorig,0,access.M_T,b.odata,nmoms);

	// different dt interval,
	// see test_Sequence() in ScalpedMaxIProjection.java
	//UnitTests.checkDoubleArrayRel(b.sst.getDtMoments(),"b.sst.getDtMoments()",tmomorig,"tmomorig",eps);
	UnitTests.checkDoubleArrayRel(b.sst.getDvMoments(),"b.sst.getDvMoments()",imomorig,"imomorig",eps);

	//System.err.println("test_Sequence()="+tk+" tb[tk].price="+tb[tk].price+" b.sst.P="+b.sst.P);
    }
    b=getTstPnLILowIHighWithThisRef(tb.length,currenttime,b);
    final P bo=getTstPnLILowIHigh();
    matchTstPnLILowIHigh(b,bo,eps);
    matchTPV(bo,getTPV(),eps);
}


void test_Extra(){}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":PnLILowIHighTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Extra();

    System.err.println(name+":PnLILowIHighTest.test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_min,final int neq_max)
{
    System.err.println(name+":PnLILowIHighTest: starting tests; ver=$Id: TstPnLILowIHigh.java,v 1.43 2022/07/28 08:04:53 mal Exp $");
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=eps;
    final long t0=System.currentTimeMillis();
    for(int neq=neq_min;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	    // make at smaller size, this test is slow O(n^2)
	    test_Sequence_Calculations(50+r.nextInt(100),neq);

	}
    }
    EPS_KLUDGY_WAY_TO_TURN_ON_THE_DEBUGGING=0;
    final long t1=System.currentTimeMillis();
    System.err.println(name+":PnLILowIHighTest.main: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: TstPnLILowIHigh.java,v 1.43 2022/07/28 08:04:53 mal Exp $");
}
} // END OF class PnLILowIHighTest


static void matchTstPnLILowIHigh(final TstPnLILowIHigh<?> a,final TstPnLILowIHigh<?> b,final double eps){

    EVXData.matchEVXData(a.I,b.I,false,eps);

    UnitTests.checkDoubleRel(a.dpdt_0,"a.dpdt_0",b.dpdt_0,"b.dpdt_0",eps);
    UnitTests.checkDoubleRel(a.p_IH+a.p_offset,"a.p_IH+a.p_offset",b.p_IH+b.p_offset,"b.p_IH+b.p_offset",eps);
    UnitTests.checkDoubleRel(a.pt_IH+a.p_offset,"a.pt_IH+a.p_offset",b.pt_IH+b.p_offset,"b.pt_IH+b.p_offset",eps);
    UnitTests.checkDoubleRel(a.std_p_IH,"a.std_p_IH",b.std_p_IH,"b.std_p_IH",eps);
    UnitTests.checkDoubleRel(a.std_pt_IH,"a.std_pt_IH",b.std_pt_IH,"b.std_pt_IH",eps);

    // match the sums
    UnitTests.checkDoubleRel(a.sst.getSumDv(),"a.sst.getSumDv()",b.sst.getSumDv(),"b.sst.getSumDv()",eps);
    UnitTests.checkDoubleRel(a.sst.getSumPcDIH(),"a.sst.getSumPcDIH()",b.sst.getSumPcDIH(),"b.sst.getSumPcDIH()",eps);
    // try dIHpnl
    EVXData.matchEVXData(a.dIHpnl,b.dIHpnl,false,eps);
    //UnitTests.checkDoubleRel(a.dIHpnl.sH,"a.dIHpnl.sH",b.dIHpnl.sH,"b.dIHpnl.sH",eps);
    UnitTests.checkDoubleRel(a.DP_since_dIH_H,"a.DP_since_dIH_H",b.DP_since_dIH_H,"b.DP_since_dIH_H",eps);
    UnitTests.checkDoubleRel(a.AbsDP_since_dIH_H,"a.AbsDP_since_dIH_H",b.AbsDP_since_dIH_H,"b.AbsDP_since_dIH_H",eps);
    UnitTests.checkDoubleRel(a.P_at_dIH_H,"a.P_at_dIH_H",b.P_at_dIH_H,"b.P_at_dIH_H",eps);
    UnitTests.checkDoubleRel(a.DP_at_dIH_H,"a.DP_at_dIH_H",b.DP_at_dIH_H,"b.DP_at_dIH_H",eps);
    UnitTests.checkDoubleRel(a.AbsDP_at_dIH_H,"a.AbsDP_at_dIH_H",b.AbsDP_at_dIH_H,"b.AbsDP_at_dIH_H",eps);

    System.err.println("TstPnLILowIHigh:matchTstPnLILowIHigh test OK eps="+eps);
}


static void matchTPV(final TstPnLILowIHigh<?> a,final PnLILowIHighTPVFactory.PnLILowIHighTPV b,final double eps){

    EVXData.matchEVXData(a.I,b.I,false,eps);

    UnitTests.checkDoubleRel(a.sst.getSumPcDIH(),"a.sst.getSumPcDIH()",b.sst.getSumPcDIH(),"b.sst.getSumPcDIH()",eps);

    UnitTests.checkDoubleRel(a.p_IH+a.p_offset,"a.p_IH+a.p_offset",b.p_IH,"b.p_IH",eps);

    // match the sums
    UnitTests.checkDoubleRel(a.sst.getSumDv(),"a.sst.getSumDv()",b.sst.getSumDv(),"b.sst.getSumDv()",eps);
    UnitTests.checkDoubleRel(a.sst.getSumPcDIH(),"a.sst.getSumPcDIH()",b.sst.getSumPcDIH(),"b.sst.getSumPcDIH()",eps);
    // try dIHpnl
    EVXData.matchEVXData(a.dIHpnl,b.dIHpnl,false,eps);
    //UnitTests.checkDoubleRel(a.dIHpnl.sH,"a.dIHpnl.sH",b.dIHpnl.sH,"b.dIHpnl.sH",eps);
    UnitTests.checkDoubleRel(a.DP_since_dIH_H,"a.DP_since_dIH_H",b.DP_since_dIH_H,"b.DP_since_dIH_H",eps);
    UnitTests.checkDoubleRel(a.AbsDP_since_dIH_H,"a.AbsDP_since_dIH_H",b.AbsDP_since_dIH_H,"b.AbsDP_since_dIH_H",eps);
    UnitTests.checkDoubleRel(a.P_at_dIH_H,"a.P_at_dIH_H",b.P_at_dIH_H,"b.P_at_dIH_H",eps);
    UnitTests.checkDoubleRel(a.DP_at_dIH_H,"a.DP_at_dIH_H",b.DP_at_dIH_H,"b.DP_at_dIH_H",eps);
    UnitTests.checkDoubleRel(a.AbsDP_at_dIH_H,"a.AbsDP_at_dIH_H",b.AbsDP_at_dIH_H,"b.AbsDP_at_dIH_H",eps);

    System.err.println("TstPnLILowIHigh:matchTPV test OK eps="+eps);
}

}
