package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.realtimebook.*;

/** Directional Liquidity Deficit with Total Market Capitalization.
 *  @version $Id: UMdQVM.java,v 1.41 2022/06/27 07:19:16 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class UMdQVM<DTYPE> {
public static final String VERSION="$Id: UMdQVM.java,v 1.41 2022/06/27 07:19:16 mal Exp $";
final DTYPE [] idata;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=getABasis();;
final DataInterfaceToMoments<DTYPE> access=getAccess();
final double lam_0,lam_L,lam_H,si_H,si_0;
final State st;

UMdQVM(final int n,final DTYPE [] idata)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");

    this.idata=idata;

    final QQHedges QQH=getQQHedgeMatrs(n);
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQH.QQc,QQH.QQt,M);
    final int i0=0,iL=1,iH=n-1;
    this.lam_0=basis[i0];
    this.lam_L=basis[iL];
    this.lam_H=basis[iH];

    final double [] psi=EVXData.getPsiFromBasis(iH,basis,n),
	psi0=EVXData.getPsiFromBasis(i0,basis,n);
    this.si_H=TMatr.getAMatrA(psi,QQH.QQl,n);
    this.si_0=TMatr.getAMatrA(psi0,QQH.QQl,n);
    this.st=new State(QQH.getJHedges(psi,QQH.H_i),QQH.getJHedges(psi,QQH.H_c),QQH.pH_last,true);

}

// =============================================================
/** Store total capital & execution flow matrices.
 */
static class QQHedges {
    final double [] QQt,QQc,QQl;
    final double [][] QQiHedges,QQcHedges;
    final double [] pH_last;
    final int dim;

    QQHedges(final double [] QQt,final double [] QQc,final double [] QQl,
	     final double [][] QQiHedges,final double [][] QQcHedges,final int dim,
	     final double [] pH_last)
    {
	this.QQt=QQt;
	this.QQc=QQc;
	this.QQl=QQl;
	this.QQiHedges=QQiHedges;
	this.QQcHedges=QQcHedges;
	this.dim=dim;
	this.pH_last=pH_last;
    }
    static final int H_i=56,H_c=57;
    double [] getJHedges(final double [] psi,final int what)
    {
	final double [][] Q;
	if(what==H_i) Q=QQiHedges;
	else if(what==H_c) Q=QQcHedges;
	else throw new IllegalArgumentException("what="+what+" is not supported");
	final double [] j=new double [Q.length];
	for(int k=0;k<j.length;k++)
	{
	    j[k]=TMatr.getAMatrA(psi,Q[k],dim);
	}
	return j;
    }
}
/** Calculates C-moments sum pI and Last-moments sum (p-p_l)I.
 */
QQHedges getQQHedgeMatrs(final int n)
{
    final int nmoms=2*n-1;
    final double [] momc=new double [nmoms],moml=new double [nmoms],
	mom0=new double [nmoms],mom1=new double [nmoms];
    final double [][] QQiHedges=new double [idata.length][],
	QQcHedges=new double [idata.length][];
    final double [] pH_last=new double [idata.length];

    long tends=0;
    for(int k=0;k<idata.length;k++)
    {
	final DTYPE o=idata[k];
	final long tend=getTEnd(o);
	if(k!=0 && tend!=tends) throw new IllegalArgumentException("tend["+k+"]="+tend+" mistmatch tends="+tends);
	tends=tend;

	final double po=getPOffset(o),pl=getPLast(o),dpo=po-pl;
	access.setM(mom0,0,access.M_I,o,nmoms);
	access.setM(mom1,1,access.M_I,o,nmoms);
	QQiHedges[k]=M.B.getQQMatr(n,mom0);
	pH_last[k]=pl;
	for(int l=0;l<nmoms;l++)
	{
	    moml[l]+=mom1[l]+dpo*mom0[l];
	    momc[l]+=(mom1[l]=mom1[l]+po*mom0[l]);
	}
	QQcHedges[k]=M.B.getQQMatr(n,mom1);
    }
    access.setMT(mom1,nmoms);
    return new QQHedges(M.B.getQQMatr(n,mom1),M.B.getQQMatr(n,momc),M.B.getQQMatr(n,moml),QQiHedges,QQcHedges,n,pH_last);
}
// END OF class QQHedges related code.
// =============================================================

static class State {
    final double aCap,aSuml;
    final double [] i,cap,dcap;
    final boolean flag_normalize_C;
    final int n_positions;
    final double [] posH,pH_last;

    State(final double [] i,final double [] cap,final double [] pH_last,final boolean flag_normalize_C)
    {
	if(i.length!=pH_last.length||cap.length!=pH_last.length) throw new IllegalArgumentException("dimension mismatch");
	this.i=i;
	this.cap=cap;
	this.pH_last=pH_last;
	this.flag_normalize_C=flag_normalize_C;
	this.posH=new double [i.length];
	this.dcap=new double [i.length];
	double ac=0,as=0;
	for(int k=0;k<i.length;k++)
	{
	    posH[k]=i[k];
	    ac+=i[k]*pH_last[k];
	    as+=(dcap[k]=cap[k]-pH_last[k]*i[k]);
	}
	this.aCap=ac;
	this.aSuml=as;
	final double dirSum=Math.signum(aSuml);
	int npos=0;
	for(int k=0;k<i.length;k++)
	{
	    if(!(Math.signum(dcap[k])!=dirSum)) posH[k]=0;
	    else {npos++;posH[k]*=dirSum;}
	}
	this.n_positions=npos;
	if(flag_normalize_C && npos>0)
	{
	    TMatr.scaleVectorInPlace(posH,1/Math.abs(TMatr.mulVV(pH_last,posH)));
	}
    }
    double getPosC(final double [] cur_pH_last)
    {
	return TMatr.mulVV(posH,cur_pH_last);
    }
    double getPosCOrig(){return getPosC(pH_last);}
    double getPnL(final double [] exit_pH_last)
    {
	return getPosC(exit_pH_last)-getPosCOrig();
    }
    double getPnL(final State exitPrices)
    {
	return getPnL(exitPrices.pH_last);
    }
}
// END OF class State

abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double getPLast(final DTYPE o);
abstract double getPOffset(final DTYPE o);
abstract long getTEnd(final DTYPE o);

static abstract class UMdQVMTest<DTYPE> extends QTest {

abstract UMdQVM<DTYPE> getUMdQVM();
abstract UMnQVM<DTYPE> getUMnQVM();


UMdQVMTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}


void test_UMnQVM()
{
    final UMdQVM<DTYPE> q=getUMdQVM();
    final UMnQVM<DTYPE> s=getUMnQVM();
    UnitTests.checkDoubleRel(q.lam_0,"q.lam_0",s.I_0,"s.I_0",eps);
    UnitTests.checkDoubleRel(q.lam_L,"q.lam_L",s.I_L,"s.I_L",eps);
    UnitTests.checkDoubleRel(q.lam_H,"q.lam_H",s.I_H,"s.I_H",eps);
    UnitTests.checkDoubleRel(q.si_H,"q.si_H",s.pnl_I_H,"s.pnl_I_H",eps);

    System.err.println(name+":UMdQVMTest:test_UMnQVM(): test OK neq="+neq+" eps="+eps);
}


void test_Common()
{
    final UMdQVM<DTYPE> q=getUMdQVM();
    UnitTests.checkDoubleRel(q.si_H,"q.si_H",TMatr.sumElements(q.st.dcap),"TMatr.sumElements(q.st.dcap)",eps);

    final QQHedges QQH=q.getQQHedgeMatrs(neq);

    System.err.println("Total CAP: L-S="+q.st.getPosCOrig());
    System.err.println(name+":UMdQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}


void test_Extra(){}

Trade [][] tdata;
int [] tdata_size;
void setDataS(final int nstocks,final int tsize,final int neq)
{
    this.tdata=new Trade [nstocks][];
    long ct=Long.MIN_VALUE;
    for(int i=0;i<nstocks;i++)
    {
	setData(tsize+r.nextInt(tsize)/7,neq); // sets tau as a sideeffect
	this.tdata[i]=tb;
	final long t=tb[tb.length-1].time;
	//System.err.println("t="+t);
	if(t>=ct){ct=t;}
	tb=null;
    }
    tb=null;

    currenttime=ct+330; // maximal last time plus some small delay
    // tau has the value of latest generated tb
    tdata_size=new int [tdata.length];
    for(int k=0;k<tdata.length;k++)
    {
	tdata_size[k]=tdata[k].length;
    }
}



void test_Calculations(final int nstocks,final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":UMdQVMTest:test_Calculations(): test with N points="+tsize+
		       " nstocks="+nstocks+" neq="+neq+" eps="+eps);

    setDataS(nstocks,tsize,neq);
    test_Common();
    test_UMnQVM();
    test_Extra();

    System.err.println(name+":UMdQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":UMdQVMTest: starting tests; ver=$Id: UMdQVM.java,v 1.41 2022/06/27 07:19:16 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int nstocks=1;nstocks<=5;nstocks++)
	{
	    for(int i=0;i<119;i++)
	    {
		test_Calculations(nstocks,1950+r.nextInt(1850),neq);
	    }
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":UMdQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: UMdQVM.java,v 1.41 2022/06/27 07:19:16 mal Exp $");
}
} // END OF class UMdQVMTest




}// END OF class UMdQVM
