package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.realtimebook.*;

/** Multi stock version of UnQVM.
 *  @version $Id: UMnQVM.java,v 1.24 2020/09/27 20:33:22 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class UMnQVM<DTYPE> {

final DTYPE [] idata;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final double I_0,I_L,I_H,p_I_H,pnl_I_H,w_piL_IH,w_piH_IH,r_piL_IH,r_piH_IH,pnl_piL,pnl_piH,pnlE;
final State sst;

static class State {
    final MMaxStateSaver [] rimF;
    final double [] p_last;
    State(final MMaxStateSaver [] rimF,final double [] p_last)
    {
	this.rimF=rimF;
	this.p_last=p_last;
	if(rimF.length!=p_last.length) throw new IllegalArgumentException("dimension mismatch");
    }
}

UMnQVM(final int n,final DTYPE [] idata,final State rst)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");

    this.idata=idata;
    this.M=getABasis();

    final double [][] QQ=getQQMatrs(n);
    final double [] QQt=QQ[0],QQi=QQ[1],QQpi=QQ[2];
    final int nd=n-1;
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M);
    final int i0=0,iL=1,iH=n-1;
    this.I_0=basis[i0];
    this.I_L=basis[iL];
    this.I_H=basis[iH];

    final double [] RQQpi;
    if(i0==0) RQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,basis,nd,n+n+0);
    else throw new RuntimeException("Not supported i0="+i0);
    this.pnl_I_H=RQQpi[(nd-1)*nd+(nd-1)];
    this.p_I_H=this.pnl_I_H/this.I_H;

    final double [] ev=M.EV.getEV(nd,RQQpi,true),tmp=new double [nd*nd];
    this.w_piL_IH=ev[nd+(nd-1)];
    this.w_piH_IH=ev[nd+(nd-1)*nd+(nd-1)];
    for(int k=0;k<nd;k++) for(int l=0;l<nd;l++) tmp[k*nd+l]=(k==l?basis[k+1]:0);
    this.pnl_piL=ev[0];
    this.r_piL_IH=this.pnl_piL/TMatr.getAMatrA(EVXData.getPsiFromBasis(0,ev,nd),tmp,nd);
    this.pnl_piH=ev[nd-1];
    this.r_piH_IH=this.pnl_piH/TMatr.getAMatrA(EVXData.getPsiFromBasis(nd-1,ev,nd),tmp,nd);

    if(this.pnl_piL<0&&this.pnl_piH<0){this.pnlE=this.pnl_piL;}
    else if(this.pnl_piL>0&&this.pnl_piH>0){this.pnlE=this.pnl_piH;}
    else{this.pnlE=Math.abs(this.pnl_piL)>Math.abs(this.pnl_piH)?this.pnl_piL:this.pnl_piH;}


    final MMaxStateSaver [] simF=new MMaxStateSaver [idata.length];
    final double [] p_last=new double [simF.length];
    long tends=0;
    for(int k=0;k<simF.length;k++)
    {
	p_last[k]=getPLast(idata[k]);
	final long tend=getTEnd(idata[k]);
	if(k!=0 && tend!=tends) throw new IllegalArgumentException("tend["+k+"]="+tend+" mistmatch tends="+tends);
	tends=tend;
	if(I_0<I_L){
	    simF[k]=MMaxStateSaver.getTaintedUpdate(rst!=null?rst.rimF[k]:null);
	} else if(I_0>I_H){
	    simF[k]=MMaxStateSaver.getMaxUpdate(I_0,p_last[k],tend,I_0,w_piL_IH,w_piH_IH,getVolume(idata[k]),0,0,rst!=null?rst.rimF[k]:null);
	} else {
	    simF[k]=MMaxStateSaver.getNothingUpdate(rst!=null?rst.rimF[k]:null);
	}
    }

    this.sst=new State(simF,p_last);

}



double [][] getQQMatrs(final int n)
{
    final int nmoms=2*n-1;
    final DataInterfaceToMoments<DTYPE> access=getAccess();
    final double [] mom=new double [nmoms],pmom=new double [nmoms],
	mom0=new double [nmoms],mom1=new double [nmoms];

    for(int k=0;k<idata.length;k++)
    {
	final double po=getPOffset(idata[k]),pl=getPLast(idata[k]),dpo=po-pl;
	access.setM(mom0,0,access.M_I,idata[k],nmoms);
	access.setM(mom1,1,access.M_I,idata[k],nmoms);
	for(int l=0;l<nmoms;l++)
	{
	    mom[l]+=mom1[l]+po*mom0[l];
	    pmom[l]+=mom1[l]+dpo*mom0[l];
	}
    }

    final double [] QQi=M.B.getQQMatr(n,mom),QQpi=M.B.getQQMatr(n,pmom);
    access.setMT(mom,nmoms);
    final double [] QQt=M.B.getQQMatr(n,mom);

    return new double [][]{QQt,QQi,QQpi};
}

abstract DataInterfaceToMoments<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double getPLast(final DTYPE o);
abstract double getPOffset(final DTYPE o);
abstract long getTEnd(final DTYPE o);
abstract long getVolume(final DTYPE o);

static abstract class UMnQVMTest<DTYPE> extends QTest {

abstract UMnQVM<DTYPE> getUMnQVM();


UMnQVMTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}




void test_Common()
{
    System.err.println(name+":UMnQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}


void test_Extra(){}

abstract NkQVM<?> getNkQVM();
void test_NkQVM()
{
    final UMnQVM<DTYPE> q=getUMnQVM();
    final NkQVM<?> s=getNkQVM();

    UnitTests.checkDoubleRel(q.I_0,"q.I_0",s.I.s0,"s.I.s0",eps);
    //UnitTests.checkDoubleRel(q.I_L,"q.I_L",s.E.Px[1],"s.E.Px[1]",eps);
    //UnitTests.checkDoubleRel(q.I_H,"q.I_H",s.E.Px[s.E.Px.length-1],"s.E.Px[s.E.Px.length-1]",eps);

    System.err.println(name+":UMnQVMTest:test_NkQVM(): test OK neq="+neq+" eps="+eps);
}


Trade [][] tdata;
int [] tdata_size;
void setDataS(final int nstocks,final int tsize,final int neq)
{
    this.tdata=new Trade [nstocks][];
    long ct=Long.MIN_VALUE;
    for(int i=0;i<nstocks;i++)
    {
	setData(tsize+r.nextInt(tsize),neq); // sets tau as a sideeffect
	this.tdata[i]=tb;
	final long t=tb[tb.length-1].time;
	if(t>=ct){ct=t;}
	tb=null;
    }
    currenttime=ct+330; // maximal last time plus some small delay
    // tau has the value of latest generated tb
    tdata_size=new int [tdata.length];
    for(int k=0;k<tdata.length;k++)
    {
	tdata_size[k]=tdata[k].length;
    }
}


void test_Calculations(final int nstocks,final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":UMnQVMTest:test_Calculations(): test with N points="+tsize+
		       " nstocks="+nstocks+" neq="+neq+" eps="+eps);

    setDataS(nstocks,tsize,neq);
    test_Common();
    test_NkQVM();
    test_Extra();

    System.err.println(name+":UMnQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":UMnQVMTest: starting tests; ver=$Id: UMnQVM.java,v 1.24 2020/09/27 20:33:22 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int nstocks=1;nstocks<=5;nstocks++)
	{
	    for(int i=0;i<119;i++)
	    {
		test_Calculations(nstocks,1950+r.nextInt(1850),neq);
	    }
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":UMnQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: UMnQVM.java,v 1.24 2020/09/27 20:33:22 mal Exp $");
}
} // END OF class UMnQVMTest




}// END OF class UMnQVM
