package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;

/** Book in time/price space with Radau measure.
 *  @version $Id: UkbQVM.java,v 1.9 2022/06/27 13:32:46 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class UkbQVM<DTYPE> {


final DataInterfaceDirectBookPrices<DTYPE> access;
private final DTYPE odata_buy,odata_sell;
final double p_buy,tau_buy,p_sell,tau_sell;
static final int MOM_TYPE=DataInterfaceDirectBookPrices.M_V;
UkbQVM(final int neq,final DTYPE odata_buy,final DTYPE odata_sell)
{
    this.access=getAccess();
    final int nmoms=2*neq-1;

    this.odata_buy=odata_buy;
    this.p_buy=getPEdge(this.odata_buy);
    this.tau_buy=getTauEdge(neq,this.odata_buy,MOM_TYPE);
    this.odata_sell=odata_sell;
    this.p_sell=getPEdge(this.odata_sell);
    this.tau_sell=getTauEdge(neq,this.odata_sell,MOM_TYPE);
}

double getTauEdge(final int n,final DTYPE odata,final int mom)
{
    final int nmoms=2*n-1;
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M=this.access.getABasis();

    final double p_edge=getPEdge(odata),p_offset=getPOffset(odata);
    final double [] vmoms_p=new double [nmoms],tvmoms_p=new double [nmoms];
    access.setM(vmoms_p,0,mom,odata,nmoms);
    access.setM(tvmoms_p,1,mom,odata,nmoms);
    //System.err.println("vmoms_p="+java.util.Arrays.toString(vmoms_p)+"\n"+"tvmoms_p="+java.util.Arrays.toString(tvmoms_p));

    final double [] QQv=M.B.getQQMatr(n,vmoms_p),QQtv=M.B.getQQMatr(n,tvmoms_p);
    final double px=p_edge-p_offset;
    final double [] p0=M.B.getX0Moments(n,px),matr=new double [n*n],Gm1p0=new double [n];
    System.arraycopy(QQv,0,matr,0,QQv.length);
    System.arraycopy(p0,0,Gm1p0,0,p0.length);
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,n,n,ipvt);
    if(pinfo!=0){java.util.Arrays.fill(Gm1p0,Double.NaN);}
    else{com.polytechnik.utils.Linsystems.dgesl_m1(matr,n,n,ipvt,Gm1p0,1);}
    final double i=TMatr.getAMatrA(Gm1p0,QQtv,n)/TMatr.mulVVn1(p0,Gm1p0);
    //System.err.println("px="+px+"\te.s0="+EVXData.getEVXDataFromQQ(n,QQtv,QQv,M).s0+"\ti="+i+"\tp_edge="+p_edge+"\tp_offset="+p_offset);
    return i;
}


abstract double getPOffset(final DTYPE odata);
abstract double getPEdge(final DTYPE odata);
abstract DataInterfaceDirectBookPrices<DTYPE> getAccess();


static abstract class UkbQVMTest<DTYPE> {

int neq;
long currenttime,tau;
Trade [] tbuy,tsell;
BookItem [] tbuy_byp,tsell_byp;

abstract UkbQVM<DTYPE> getUkbQVM();
abstract long getTau();

final String name;
final java.util.Random r;
final double eps;
UkbQVMTest(final String name,
	 final java.util.Random r,final double eps)
{
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Extra(){}
void test_Common()
{
    final UkbQVM<DTYPE> q=getUkbQVM();

    for(int k=tbuy.length;--k>=0;)
    {
	if(tbuy[k].price>q.p_buy) throw new RuntimeException("tbuy["+k+"].price="+tbuy[k].price+" q.p_buy="+q.p_buy);
    }
    for(int k=tsell.length;--k>=0;)
    {
	if(tsell[k].price<q.p_sell) throw new RuntimeException("tsell["+k+"].price="+tsell[k].price+" q.p_sell="+q.p_sell);
    }

    System.err.println(name+":UkbQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}
void setData(final int buysize,final int sellsize,final int neq)
{
    this.neq=neq;
    this.tbuy=CalculateKendall.TradeEMoments.getRandomTrades(buysize,r);
    tbuy_byp=tbuy.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tbuy_byp,0,tbuy_byp.length-1,Distribution.cmp_by_price_lowest_down);
    this.tsell=CalculateKendall.TradeEMoments.getRandomTrades(sellsize,r);
    tsell_byp=tsell.clone();
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tsell_byp,0,tsell_byp.length-1,Distribution.cmp_by_price_lowest_up);
    this.currenttime=Math.max(tbuy[tbuy.length-1].time,tsell[tsell.length-1].time);
    this.tau=getTau();
    //System.err.println("tau="+tau);
}

void test_Calculations(final int tbuysize,final int tsellsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":UkbQVMTest:test_Calculations(): test with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);

    setData(tbuysize,tsellsize,neq);
    test_Common();
    test_Extra();

    System.err.println(name+":UkbQVMTest:test_Calculations(): test OK with N points:"+
		       " buy="+tbuysize+" sell="+tsellsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":UkbQVMTest: starting tests; ver=$Id: UkbQVM.java,v 1.9 2022/06/27 13:32:46 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(850),1950+r.nextInt(850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":UkbQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: UkbQVM.java,v 1.9 2022/06/27 13:32:46 mal Exp $");
}
} // END OF class UkbQVMTest



} // END of class UkbQVM
