package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.EVXData;
import com.polytechnik.utils.OrthogonalPolynomialsBasisFunctionsCalculatable;
import com.polytechnik.utils.BasisFunctionsCalculatable;
import com.polytechnik.realtimebook.*;

/**
 *  @version $Id: UnQVM.java,v 1.29 2020/09/27 20:33:22 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class UnQVM<DTYPE> {

final DTYPE odata;
final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
final double I_0,I_L,I_H,p_I_H,w_piL_IH,w_piH_IH,p_PnL;
final State sst;

static class State {
    final MMaxStateSaver rimF;
    State(final MMaxStateSaver rimF)
    {
	this.rimF=rimF;
    }
}

UnQVM(final int n,final DTYPE odata,final State rst)
{
    if(n<=1) throw new RuntimeException("n="+n+" must be >1");

    this.odata=odata;
    this.M=getABasis();

    final double [][] QQ=getQQMatrs(n);
    final double [] QQt=QQ[0],QQi=QQ[1],QQpi=QQ[2];
    final int nd=n-1;
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,M);
    final int i0=0,iL=1,iH=n-1;
    this.I_0=basis[i0];
    this.I_L=basis[iL];
    this.I_H=basis[iH];

    final double [] RQQpi;
    if(i0==0) RQQpi=EVXData.convertMatrFromQtoB_Basis(n,QQpi,basis,nd,n+n+0);
    else throw new RuntimeException("Not supported i0="+i0);
    this.p_I_H=RQQpi[(nd-1)*nd+(nd-1)]/this.I_H;

    final double [] ev=M.EV.getEV(nd,RQQpi,true);
    double ilih=0;for(int k=0;k<nd;k++){ ilih+=(Math.pow(ev[nd+0*nd+k],2)+Math.pow(ev[nd+(nd-1)*nd+k],2))*basis[k+1];}
    this.p_PnL=(ev[0]+ev[nd-1])/ilih;
    this.w_piL_IH=ev[nd+(nd-1)];
    this.w_piH_IH=ev[nd+(nd-1)*nd+(nd-1)];

    final MMaxStateSaver simF;
    if(I_0<I_L){
	simF=MMaxStateSaver.getTaintedUpdate(rst!=null?rst.rimF:null);
    } else if(I_0>I_H){
	simF=MMaxStateSaver.getMaxUpdate(I_0,getPLast(),getTLast(),I_0,w_piL_IH,w_piH_IH,getVolume(),(p_I_H+getPLast())*I_H,p_I_H*I_H,rst!=null?rst.rimF:null);
    } else {
	simF=MMaxStateSaver.getNothingUpdate(rst!=null?rst.rimF:null);
    }


    this.sst=new State(simF);

}



double [][] getQQMatrs(final int n)
{
    final int nmoms=2*n-1;
    final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access=getAccess();
    final double [] mom=new double [nmoms];

    access.setMT(mom,nmoms);
    final double [] QQt=M.B.getQQMatr(n,mom);
    access.setM(mom,0,access.M_I,odata,nmoms);
    final double [] QQi=M.B.getQQMatr(n,mom);
    final double [] pmom=new double [nmoms];
    access.setM(pmom,1,access.M_I,odata,nmoms);
    TMatr.sumP1(pmom,mom,getPOffset()-getPLast());

    final double [] QQpi=M.B.getQQMatr(n,pmom);
    return new double [][]{QQt,QQi,QQpi};
}

abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getABasis();
abstract double [] getEDPsi(final double [] psi);
abstract double getPLast();
abstract double getPOffset();
abstract long getTLast();
abstract long getVolume();

static abstract class UnQVMTest<DTYPE> extends QTest {

abstract UnQVM<DTYPE> getUnQVM();


UnQVMTest(final String name,
	  final java.util.Random r,final double eps)
{
    super(name,r,eps);
}




void test_Common()
{
    final int n=neq;
    final UnQVM<DTYPE> q=getUnQVM();
    final double [][] QQ=q.getQQMatrs(n);
    final double [] QQt=QQ[0],QQi=QQ[1],QQpi=QQ[2];
    final double [] basis=EdComm.getIRadauBasisPsi0FromQQ(n,QQi,QQt,q.M);
    final DynPnL s=new DynPnL(n,basis,n,new int []{0},QQpi,QQi,QQt,0,q.M);

    System.err.println("q.p_PnL="+q.p_PnL+" s.pEQ="+s.pEQ);
    UnitTests.checkDoubleRel(q.p_PnL,"q.p_PnL",s.pEQ,"s.pEQ",eps);

    System.err.println(name+":UnQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}


void test_Extra(){}

abstract UkQVM<?> getUkQVM();
void test_UkQVM()
{
    final UnQVM<DTYPE> q=getUnQVM();
    final UkQVM<?> s=getUkQVM();

    UnitTests.checkDoubleRel(q.I_0,"q.I_0",s._I_0,"s._I_0",eps);
    UnitTests.checkDoubleRel(q.I_L,"q.I_L",s._I_L,"s._I_L",eps);
    UnitTests.checkDoubleRel(q.I_H,"q.I_H",s._I_H,"s._I_H",eps);
    UnitTests.checkDoubleRel(q.p_I_H,"q.p_I_H",s._p_I_H,"s._p_I_H",eps);

    UnitTests.checkDoubleRel(Math.abs(q.w_piL_IH),"q.w_piL_IH",Math.abs(s._w_pIL_IH),"s._w_pIL_IH",eps);
    UnitTests.checkDoubleRel(Math.abs(q.w_piH_IH),"q.w_piH_IH",Math.abs(s._w_pIH_IH),"s._w_pIH_IH",eps);

    System.err.println(name+":UnQVMTest:test_UkQVM(): test OK neq="+neq+" eps="+eps);
}


void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":UnQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_UkQVM();
    test_Extra();

    System.err.println(name+":UnQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":UnQVMTest: starting tests; ver=$Id: UnQVM.java,v 1.29 2020/09/27 20:33:22 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=2;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(1950+r.nextInt(1850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":UnQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: UnQVM.java,v 1.29 2020/09/27 20:33:22 mal Exp $");
}
} // END OF class UnQVMTest




}// END OF class UnQVM
