package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2012
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;
import com.polytechnik.realtimebook.*;

/** ZVQVM and a set of measures.
 *  @version $Id: ZVQVM.java,v 1.25 2022/08/07 15:44:04 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class ZVQVM<DTYPE> {

final DTYPE odata;
final DeprecatedHelperDataDirectAccessWithBasis<DTYPE> access;

/** Rurik evolution. */
final double K00,
    nvrts0,sigma2T_nvrts0,ovrts0,sigma2A_ovrts0,
    nvsts0,sigma2T_nvsts0,ovsts0,sigma2A_ovsts0,
    nvrvs0,sigma2T_nvrvs0,ovrvs0,sigma2A_ovrvs0,
    nvdp0,sigma2T_nvdp0,ovdp0,sigma2A_ovdp0;
final double [] kk,kks,kkr,kkrv,kkdp;

ZVQVM(final int neq,final DTYPE odata)
{
    this.odata=odata;
    this.access=getAccess();

    final int nmoms=2*neq-1;

    final double [] vmom=new double [nmoms],rtsmom=new double [nmoms],dtsmom=new double [nmoms],tsmom=new double [nmoms],rvsmom=new double [nmoms],dpmom=new double [nmoms];
    access.setM(vmom,0,access.M_I,odata,nmoms);
    access.setM(tsmom,0,access.M_TS,odata,nmoms);
    access.setM(rtsmom,0,access.M_RTS,odata,nmoms);
    access.setM(rvsmom,0,access.M_RVS,odata,nmoms);
    access.setM(dpmom,1,access.M_DP,odata,nmoms);
    final double [][] QQ=access.getQQ_Matrs(neq,1,new double [][]{vmom,tsmom,rtsmom,rvsmom,dpmom});
    final double [] QQv=QQ[0],QQsts=QQ[1],QQrts=QQ[2],QQrvs=QQ[3],QQdp=QQ[4],
	IQQ=com.polytechnik.utils.Linsystems.getInvertedMatrix(neq,neq,QQv);
    final double [] tmp=new double [neq*neq];
    final double [] QQs=ZQVM.getIQQ_F_IQQ(neq,IQQ,QQsts,tmp);
    final double [] QQr=ZQVM.getIQQ_F_IQQ(neq,IQQ,QQrts,tmp);
    final double [] QQrv=ZQVM.getIQQ_F_IQQ(neq,IQQ,QQrvs,tmp);
    final double [] QQdpk=ZQVM.getIQQ_F_IQQ(neq,IQQ,QQdp,tmp);
    final double [][] kkk=access.getKK(neq,new double [][] {IQQ,QQs,QQr,QQrv,QQdpk});

    kk=kkk[0];
    kks=kkk[1];
    kkr=kkk[2];
    kkrv=kkk[3];
    kkdp=kkk[4];

    final double x0=getX0();
    K00=access.fx(kk,x0);
    final double sqrtK00=Math.sqrt(K00);

    nvsts0=access.fx(kks,x0)/K00;
    sigma2T_nvsts0=TMatr.getSpurMM(neq,QQsts,QQs)-nvsts0*nvsts0;
    double [] r=TMatr.getMatr_x_A(neq,IQQ,java.util.Arrays.copyOf(tsmom,neq));
    ovsts0=access.fx(r,x0)/sqrtK00;
    sigma2A_ovsts0=TMatr.mulVVn1(r,tsmom)-ovsts0*ovsts0;

    nvrts0=access.fx(kkr,x0)/K00;
    sigma2T_nvrts0=TMatr.getSpurMM(neq,QQrts,QQr)-nvrts0*nvrts0;
    r=TMatr.getMatr_x_A(neq,IQQ,java.util.Arrays.copyOf(rtsmom,neq));
    ovrts0=access.fx(r,x0)/sqrtK00;
    sigma2A_ovrts0=TMatr.mulVVn1(r,rtsmom)-ovrts0*ovrts0;

    nvrvs0=access.fx(kkrv,x0)/K00;
    sigma2T_nvrvs0=TMatr.getSpurMM(neq,QQrvs,QQrv)-nvrvs0*nvrvs0;
    r=TMatr.getMatr_x_A(neq,IQQ,java.util.Arrays.copyOf(rvsmom,neq));
    ovrvs0=access.fx(r,x0)/sqrtK00;
    sigma2A_ovrvs0=TMatr.mulVVn1(r,rvsmom)-ovrvs0*ovrvs0;

    nvdp0=access.fx(kkdp,x0)/K00;
    sigma2T_nvdp0=TMatr.getSpurMM(neq,QQdp,QQdpk)-nvdp0*nvdp0;
    r=TMatr.getMatr_x_A(neq,IQQ,java.util.Arrays.copyOf(dpmom,neq));
    ovdp0=access.fx(r,x0)/sqrtK00;
    sigma2A_ovdp0=TMatr.mulVVn1(r,dpmom)-ovdp0*ovdp0;
}

abstract DeprecatedHelperDataDirectAccessWithBasis<DTYPE> getAccess();
abstract double getX0();

static abstract class ZVQVMTest<DTYPE> {

int neq;
long currenttime,tau;
Trade [] tb;

abstract void setData(final int tsize,final int neq);
abstract ZVQVM<DTYPE> getZVQVM();
abstract XQVM<DTYPE> getXQVM();

final String name;
final java.util.Random r;
final double eps;
ZVQVMTest(final String name,
	 final java.util.Random r,final double eps)
{
    this.name=name;
    this.r=r;
    this.eps=eps;
}
void test_Common()
{
    final ZVQVM<DTYPE> q=getZVQVM();

    System.err.println(name+":ZVQVMTest:test_Common(): test OK neq="+neq+" eps="+eps);
}

void test_XQVM()
{
    final ZVQVM<DTYPE> q=getZVQVM();
    final XQVM<DTYPE> s=getXQVM();

    UnitTests.checkDoubleRel(q.K00,"q.K00",s.ovK00,"s.ovK00",eps);

    UnitTests.checkDoubleRel(q.ovsts0,"q.ovsts0",s.ovdts0,"s.ovdts0",eps);
    if(neq>1) UnitTests.checkDoubleRel(q.sigma2A_ovsts0,"q.sigma2A_ovsts0",s.sigma2A_ovdts0a,"s.sigma2A_ovdts0a",eps);
    UnitTests.checkDoubleRel(q.nvsts0,"q.nvsts0",s.nvdts0,"s.nvdts0",eps);
    if(neq>1) UnitTests.checkDoubleRel(q.sigma2T_nvsts0,"q.sigma2T_nvsts0",s.sigma2T_nvdts0a,"s.sigma2T_nvdts0a",eps);

    UnitTests.checkDoubleRel(q.ovrts0,"q.ovrts0",s.ovrts0,"s.ovrts0",eps);
    if(neq>1) UnitTests.checkDoubleRel(q.sigma2A_ovrts0,"q.sigma2A_ovrts0",s.sigma2A_ovrts0,"s.sigma2A_ovrts0",eps);
    UnitTests.checkDoubleRel(q.nvrts0,"q.nvrts0",s.nvrts0,"s.nvrts0",eps);
    //if(neq>1) UnitTests.checkDoubleRel(q.sigma2T_nvrts0,"q.sigma2T_nvrts0",s.sigma2T_nvrts0a,"s.sigma2T_nvrts0a",eps);



    System.err.println(name+":ZVQVMTest:test_XQVM(): test OK neq="+neq+" eps="+eps);
}

void test_Lastpoint()
{
    final ZVQVM<DTYPE> q=getZVQVM();
    final Trade [] tb=new Trade [this.tb.length+1];
    System.arraycopy(this.tb,0,tb,0,this.tb.length);
    final double pl=130.0+2*r.nextDouble();
    tb[this.tb.length]=new Trade(pl,
				 10000,
				 null,
				 r.nextInt(2)==1?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
				 currenttime,
				 -1,-1,-1,
				 null,Trade.TRADE_TYPE_REGULAR,
				 Trade.getDummyTrade(pl-10*r.nextDouble(),100,BookItem.BOOK_BUY,currenttime),
				 Trade.getDummyTrade(pl+10*r.nextDouble(),100,BookItem.BOOK_SELL,currenttime),
				 -1);
    this.tb=tb;
    final ZVQVM<DTYPE> s=getZVQVM();

    //UnitTests.checkDoubleRel(q.K00,"q.K00",s.K00,"s.K00",eps);
    //UnitTests.checkDoubleRel(q.sigma2T_nvsts0,"q.sigma2T_nvsts0",s.sigma2T_nvsts0,"s.sigma2T_nvsts0",eps);
    UnitTests.checkDoubleRel(q.sigma2A_ovsts0,"q.sigma2A_ovsts0",s.sigma2A_ovsts0,"s.sigma2A_ovsts0",eps);
    //UnitTests.checkDoubleRel(q.sigma2T_nvrts0,"q.sigma2T_nvrts0",s.sigma2T_nvrts0,"s.sigma2T_nvrts0",eps);
    UnitTests.checkDoubleRel(q.sigma2A_ovrts0,"q.sigma2A_ovrts0",s.sigma2A_ovrts0,"s.sigma2A_ovrts0",eps);

    System.err.println(name+":ZVQVMTest:test_Lastpoint(): test OK neq="+neq+" eps="+eps);
}

void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":ZVQVMTest:test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,neq);
    test_Common();
    test_XQVM();
    if(neq>1) test_Lastpoint();

    System.err.println(name+":ZVQVMTest:test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":ZVQVMTest: starting tests; ver=$Id: ZVQVM.java,v 1.25 2022/08/07 15:44:04 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<119;i++)
	{
	    test_Calculations(950+r.nextInt(850),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":ZVQVMTest: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: ZVQVM.java,v 1.25 2022/08/07 15:44:04 mal Exp $");
}
} // END OF class ZVQVMTest






}// END OF class ZVQVM
