package com.polytechnik.utils;

/** Polynomial basis functions calculatable at x.
 * The most important one --- multiply p*(ma*x+mb).
 * This interface is typically a part of {@link BasisPolynomials BasisPolynomials} functionality,
 * however it can be used to representa a more general
 * than {@link RecurrenceAB three term recurrence} type of polynomials.
 * @version $Id: BasisFunctionsCalculatable.java,v 1.9 2021/11/27 00:50:14 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public interface BasisFunctionsCalculatable extends BasisFunctionsMultipliable,ConfederateMatrixCalculatable {
    /** Multiply p*(ma*x+mb) in the basis.
     * Do not mistake with a and b recurrence coefficients
     * of {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getAB(double [][]) getAB()}.
     * @param p Input polynomial.
     * @return p*(ma*x+mb), an array of the length 1+p.length.
     */
    double [] getaxbP(final double [] p,final double ma,final double mb);

    /** Calculate 1 in basis.
     * @return An array of 1 length such that getD0(getONE(),x)=1.
     */
    default double [] getONE(){
	final double [] ONE={1};
	ONE[0]/=getD0(ONE,0);
	return ONE;
    }


    /**
     * Calculation basis polynomial value.
     * @param n T index.
     * @param x Argument.
     * @return The value of the polynomial Tn at x.
     */
    double calculate(final int n,final double x);


    /** Calculates \( \sum_k Q_k(x) Qcoefs[k] \).
     * @param Qcoefs Coefficients.
     * @param x Point to estimate the sum.
     * @return The sum.
     */
    double getD0(final double [] Qcoefs,final double x);

    /** Calculates \( d/dx \sum_k Q_k(x) Qcoefs[k] \).
     * @param Qcoefs Coefficients.
     * @param x Point to estimate the sum.
     * @return The derivative.
     */
    double getD1(final double [] Qcoefs,final double x);

    /** Calculates \( d^2/dx^2 \sum_k Q_k(x) Qcoefs[k] \).
     * @param Qcoefs Coefficients.
     * @param x Point to estimate the sum.
     * @return The second derivative.
     */
    double getD2(final double [] Qcoefs,final double x);

    /** Differentiate a function in basis,
     *  @param Qcoefs : \( F(x)=\sum_k Q_k(x) Qcoefs[k] \).
     *  @return \( dF/dx=\sum_k Q'_k(x) Qcoefs[k] \).
     */
    double [] getDifferentiated(final double [] Qcoefs);

    /** Integrate a function in basis.
     *  @param Qcoefs : \( F(x)=\sum_k Q_k(x) Qcoefs[k] \).
     *  @return \( \int F(x)dx=\sum_k Qcoefs[k] \int Q_k(x) dx \).
     *  The integral is typically normalized as I(0)=0.
     */
    double [] getIntegrated(final double [] Qcoefs);


    /** Given observations x and weights w calculate the 0..nmoms-1 moments in the basis.
     * @param nmoms The number of moments to calculate.
     * @param x observation samples.
     * @param w weights.
     * @param nelsinsample How many elements to sample.
     * @return nelsinsample generalzied moments, \( k=0\dots nmoms-1 \):
     * \( \sum_{j=0}^{nelsinsample-1} Q_k(x[j])*w[j] \).
     */
    double [] getMomentsFromSample(final int nmoms,final double [] x,final double [] w,final int nelsinsample);

    default double [] getMomentsFromSample(final int nmoms,final double [] x,final double [] w)
    {
	if(w!=null && w.length!=x.length) throw new IllegalArgumentException();
	return getMomentsFromSample(nmoms,x,w,x.length);
    }

    /** All weights are taken=1.
     */
    default double [] getMomentsFromSample(final int nmoms,final double [] x)
    {
	return getMomentsFromSample(nmoms,x,null);
    }

    /** Calcuate \( Q_k(x0) \) moments (single value calculation (at {@code x0}) of all basis functions).
     * @param nmoms The number of moments, k=0...nmoms-1.
     * @param x0 At what point to calculate.
     * @return \( Q_k(x0) \) array.
     */
    default double [] getX0Moments(final int nmoms,final double x0)
    {
	return getMomentsFromSample(nmoms,new double []{x0});
    }


    /** Synthetically divide  p by (x-x0).
     * Calculations in basis.
     * @param p Input polynomial;
     * @param x0 To divide on (x-x0)
     * @return Quotient q and remainder r in class SDivRes.
     */
    default SDivRes sdiv1(double [] p,final double x0){
	return sdiv(p,getaxbP(getONE(),1,-x0));
    }


    /** Calculate {@literal <F|xQk>} moments from {@literal <F|Qk>}
     * The method should be overloaded if optimal implementation is required.
     *  @param moments {@literal <F|Qk>}.
     *  @return {@literal <F|xQk>}.
     */
    default double [] getXQkMomentsFromQkMoments(final double [] moments){
	return getPQkMomentsFromQkMoments(getaxbP(getONE(),1,0),moments);
    }



    /** A helper method to convert this basis to PBASIS.
Usage example: <pre>{@code
new Chebyshev().convertBasisToPBASIS(7,new Monomials())
={ { 1.0 },
   { 0.0, 1.0 },
   { -1.0, 0.0, 2.0 },
   { 0.0, -3.0, 0.0, 4.0 },
   { 1.0, 0.0, -8.0, 0.0, 8.0 },
   { 0.0, 5.0, 0.0, -20.0, 0.0, 16.0 },
   { -1.0, 0.0, 18.0, 0.0, -48.0, 0.0, 32.0 }
 }
}</pre> are <a href="https://en.wikipedia.org/wiki/Chebyshev_polynomials#First_kind">Chebyshev polynomials</a>. Or convert {@link Chebyshev Chebyshev} to {@link Laguerre Laguerre}  basis and back: <pre>{@code
new Chebyshev().convertBasisToPBASIS(7,new Laguerre())
={ { 1.0 },
   { 1.0, -1.0 },
   { 3.0, -8.0, 4.0 },
   { 21.0, -69.0, 72.0, -24.0 },
   { 177.0, -736.0, 1136.0, -768.0, 192.0 },
   { 1805.0, -9245.0, 18840.0, -19080.0, 9600.0, -1920.0 },
   { 21923.0, -133704.0, 338724.0, -456192.0, 344448.0, -138240.0, 23040.0 }
}

new Laguerre().convertBasisToPBASIS(7,new Chebyshev())
={ { 1.0 },
   { 1.0, -1.0 },
   { 1.25, -2.0, 0.25 },
   { 1.7500000000000002, -3.125, 0.75, -0.041666666666666664 },
   { 2.5156250000000004, -4.5, 1.5208333333333333, -0.16666666666666666, 0.005208333333333333 },
   { 3.578125000000001, -6.255208333333334, 2.6041666666666665, -0.41927083333333337, 0.026041666666666664, -5.208333333333333E-4 },
   { 4.9848090277777795, -8.53125, 4.063151041666666, -0.8489583333333333, 0.07838541666666667, -0.0031249999999999997, 4.340277777777777E-5 }
}
}</pre>
     * @param PBASIS The basis in which we obtain this basis polynomials.
     * @return expansion coefficients.
     */
    double [][] convertBasisToPBASIS(final int num,
				     final BasisFunctionsCalculatable PBASIS);

    /**
     * Convert from basis to PBASIS basis.
     * @param Qcoefs Coefficients.
     * @param PBASIS The basis to convert to.
     * @return Polynomial in PBASIS.
     */
    double [] convertFunctionTo_PBASIS_from_Basis(final double [] Qcoefs,
						  final BasisFunctionsCalculatable PBASIS);

    /**
     * Convert from PBASIS to this basis. Equivalent to
     * {@link #convertFunctionTo_PBASIS_from_Basis(double [] Qcoefs,BasisFunctionsCalculatable PBASIS) PBASIS.convertFunctionTo_PBASIS_from_Basis(coefs,this)}.
     * @param coefs Polynomial coefficients.
     * @return Polynomial in basis.
     */
    default double [] convertFunctionTo_Basis_from_PBASIS(final double [] coefs,
							  final BasisFunctionsCalculatable PBASIS)
    {
	return PBASIS.convertFunctionTo_PBASIS_from_Basis(coefs,this);
    }


    /**
     * Convert from basis to polynomial basis. Used for testing.
     * Equivalent to {@link #convertFunctionTo_PBASIS_from_Basis convertFunctionTo_PBASIS_from_Basis(Qcoefs,new Monomials())}.
     * @param Qcoefs Coefficients.
     * @return Polynomial in regular basis.
     */
    default double [] convertFunctionTo_Monomials_from_Basis(final double [] Qcoefs)
    {
	return convertFunctionTo_PBASIS_from_Basis(Qcoefs,Monomials.BASIS);
    }

    /**
     * Convert from polynomials to basis. Used for testing.
     * Equivalent to {@link #convertFunctionTo_Basis_from_PBASIS convertFunctionTo_Basis_from_PBASIS(coefs,new Monomials())}.
     * @param coefs Polynomial coefficients.
     * @return Polynomial in basis.
     */
    default double [] convertFunctionTo_Basis_from_Monomials(final double [] coefs)
    {
	return convertFunctionTo_Basis_from_PBASIS(coefs,Monomials.BASIS);
    }
}
