package com.polytechnik.utils;

/** Multiplication--related basis functions.
 *
 * @version $Id: BasisFunctionsMultipliable.java,v 1.24 2022/11/19 11:08:58 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
@FunctionalInterface
public interface BasisFunctionsMultipliable {

    /** Multiply basis functions: expand \( Q_l(x)Q_m(x)=\sum_{k=0}^{l+m} s_k Q_k(x) \).
     *  @param l l.
     *  @param m m.
     *  @param s A buffer of dimension at least {@code l+m+1} to save the expansion coeficients.
     */
    void setQlQmExpansionCoefs(final int l,final int m,final double [] s);


    /** Multiply two polynomials in basis using {@link #setQlQmExpansionCoefs(int,int,double [])} operator.
     * @param p1 Firts polynomial.
     * @param p2 Second polynomial.
     * @return  p*q. For example: <pre>{@code
new Chebyshev().mult2Pol(new double[]{0,0,1},new double[]{0,0,0,0,0,1})
={ 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.5 }
}</pre>.
     */
    default double [] mult2Pol(final double [] p1,final double [] p2)
    {
	final int n1=p1.length-1,n2=p2.length-1;
	final double [] res=new double [n1+n2+1],tmp=new double [n1+n2+1];
	for(int k=n1;k>=0;k--)
	{
	    final double c1=p1[k];
	    for(int l=n2;l>=0;l--)
	    {
		setQlQmExpansionCoefs(k,l,tmp);
		TMatr.sumP1(res,tmp,l+k+1,c1*p2[l]);
	    }
	}
	return res;
    }

     /**
     * Multiply Q_l, l=0..numQl-1, by a polynomial.
     * @param numQl The number of Ql*P to calculate.
     * @param pol Polynomial to multiply.
     * @return An array of numQl dimension, elements: Q0*pol, Q1*pol, ... Q_{numQl-1}*pol. For example <pre>{@code
new Chebyshev().getBasisFunctionsMultipliedByPolynomial(3,new double[]{0,0,0,0,0,1})
={ { 0.0, 0.0, 0.0, 0.0, 0.0, 1.0 },
   { 0.0, 0.0, 0.0, 0.0, 0.5, 0.0, 0.5 },
   { 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.5 }
}
}</pre>
     */
    default double [][] getBasisFunctionsMultipliedByPolynomial(final int numQl,double [] pol){
	final double [][] res=new double [numQl][];
	final double [] tmp=new double [pol.length -1+ numQl-1+1];
	for(int k=0;k<numQl;k++){
	    res[k]=new double [pol.length+k];
	    for(int j=0;j<pol.length;j++){
		setQlQmExpansionCoefs(j,k,tmp);
		TMatr.sumP1(res[k],tmp,j+k+1,pol[j]);
	    }
	}
	return res;
    }

    /** Calculate {@literal <P(x) Q_k(x)>} moments from {@literal <Q_k(x)>} moments.
     *  @param moments {@literal <F|Qk>}.
     *  @return {@literal <F|P(x)Qk>}.
     */
    default double [] getPQkMomentsFromQkMoments(final double [] P,final double [] moments)
    {
	final double [] r=new double [moments.length-(P.length-1)],tmp=new double [moments.length];
	for(int j=0;j<r.length;j++)
	{
	    double s=0;
	    for(int k=0;k<P.length;k++)
	    {
		setQlQmExpansionCoefs(j,k,tmp);
		s+=P[k]*TMatr.mulVVn(1+j+k,moments,tmp);
	    }
	    r[j]=s;
	}
	return r;
    }


    /** Synthetically divide  p by d, such that p=q*d+r.
     * Calculations in basis. This implementation requires {@literal p.length>=d.length}.
     * @param p Polynomial to divide.
     * @param d Divider.
     * @return Quotient q and remainder r in class SDivRes.
     */
    default SDivRes sdiv(final double [] p,final double [] d)
    {
	return _sdiv(p,p.length-1,d,d.length-1);
    }

/** Divide  p by d, such that p=q*d+r.
 * Calculations in basis.
 * @param p Polynomial to divide.
 * @param np Max index of p, dim=np+1.
 * @param d Divider.
 * @param nd Max index of d, dim=nd+1.
 * Note: in contrast with class Polynomials, where the size of p &amp; d is passed,
 * here we pass max index (1 less than the number of elements).
 */
private SDivRes _sdiv(final double [] p,final int np,final double [] d,final int nd)
{
    // q Quotent, 0..np-nd elements are set, dim=np-nd+1.
    // r Remainder, 0..nd-1 elements are set, dim=nd.
    if(nd<0 || np-nd+1<0) throw new IllegalArgumentException("dim(p)="+(np+1)+" dim(d)="+(nd+1));
    final double [] q=new double [np-nd+1],r=new double [nd];
    final int n=np+1;
    final double [] matr=new double [n*n],tmp=new double [n];
    for(int k=np;k>=0;k--)
    {
	if(k>=nd)
	{
	    /** For {@link RecurrenceAB RecurrenceAB} can be optimized using
	     *  {@link RecurrenceA#getBasisFunctionsMultipliedByPolynomial}.
	     */
	    for(int i=nd,ki=k-nd;i>=0;i--)
	    {
		setQlQmExpansionCoefs(ki,i,tmp);
		final double c=d[i];
		for(int l=ki+i,o=k*n+l;l>=0;l--,o--)
		{
		    matr[o]+=c*tmp[l];
		}
	    }
	}
	else
	{
	    matr[k*n+k]+=1;
	}
    }
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,n,n,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(q,0,n-nd,Double.NaN);
	java.util.Arrays.fill(r,0,nd,Double.NaN);
    }
    else
    {
	System.arraycopy(p,0,tmp,0,n);
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,n,n,ipvt,tmp,0);
	System.arraycopy(tmp,nd,q,0,n-nd);
	System.arraycopy(tmp,0,r,0,nd);
    }
    return new SDivRes(q,r);
}


/** Calculate {@literal <Qj|Qk>} matrix from {@literal <Qm>}
 * moments using {@link #setQlQmExpansionCoefs setQlQmExpansionCoefs}
 * basis functions multiplication operator.
 * @param n Matrix dimension.
 * @param moments {@literal <Qk>} Moments, at least 2n-1 elements should present (0..2n-2 moments).
 * @return {@literal <Q_i|Q_j>} matrix. For example <pre>{@code
new Chebyshev().getQQMatr(5,new double []{1,0,0,0,0,0,0,0,0})
={ 1.0, 0.0, 0.0, 0.0, 0.0,
   0.0, 0.5, 0.0, 0.0, 0.0,
   0.0, 0.0, 0.5, 0.0, 0.0,
   0.0, 0.0, 0.0, 0.5, 0.0,
   0.0, 0.0, 0.0, 0.0, 0.5
 }
}</pre>
*/
default double [] getQQMatr(final int n,final double [] moments)
{
    if(2*n-1>moments.length) throw new IllegalArgumentException();

    final double [] matr=new double [n*n],e=n<=0?null:new double [2*n-1];
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<=k;l++)
	{
	    setQlQmExpansionCoefs(l,k,e);
	    matr[k+l*n]=matr[l+k*n]=TMatr.mulVVn(l+k+1,e,moments);
	}
    }
    return matr;
}

/** Calculate Q(x) MM Q(x) polynomial of 2*(n-1) degree.
 *  @param n the dimension, MM.length=n*n.
 *  @param MM The MM matrix, dim=n*n.
 *  @return Q(x) MM Q(x) polynomial, an array of dim=2*n-1.
 */
default double [] getKK(final int n,final double [] MM)
{
    final int nm=n;
    if(n!=nm||n*n!=MM.length) throw new IllegalArgumentException();
    final double [] r=new double [2*n-1];
    final double [] tmp=new double [2*n-1];
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    setQlQmExpansionCoefs(l,k,tmp);
	    TMatr.sumP1(r,tmp,1+k+l,MM[k*nm+l]);
	}
    }
    // Q(x) MM Q(x) is now formed
    return r;
}


/** Generate a measure, that, when {@link #getKK getKK(n,QQmu)} is applied, produce given polynomial P_toproduce.
 * The moments are obtained solving linear system problem.
 * @param n the dimension.
 * @param P_toproduce a polynomial to produce, must have 0..2n-2, total 2n-1 elements.
 * @return Measure moments {@code mu} 0..2n-2.
 */
default double [] getMomentsOfMeasureProducingPolynomialInKK(final int n,final double [] P_toproduce)
{
    final int nm=2*n-1;
    if(P_toproduce.length!=nm) throw new IllegalArgumentException();
    final double [] matr=new double [nm*nm],P=java.util.Arrays.copyOf(P_toproduce,nm),tmp=new double [nm];
    if(false)
    {
	for(int j=0;j<n;j++) for(int k=0;k<n;k++) {
	   final int njk=j+k;
	   setQlQmExpansionCoefs(j,k,tmp);
	   for(int l=0;l<=njk;l++) for(int m=0;m<=njk;m++){
	      matr[l*nm+m]+=tmp[l]*tmp[m];
	   }
	 }
    } else { // optimized version, does it worth it?
	for(int j=0;j<n;j++) for(int k=0;k<=j;k++) {
	   final int njk=j+k;
	   setQlQmExpansionCoefs(j,k,tmp);
	   final double s=j!=k?2:1;
	   for(int l=0;l<=njk;l++) for(int m=0;m<=l;m++){
	      matr[l*nm+m]+=s*tmp[l]*tmp[m];
	   }
	}
	for(int l=0;l<nm;l++) for(int m=0;m<l;m++){
	   matr[m*nm+l]=matr[l*nm+m];
	}
    }
    final int [] ipvt=new int [nm];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,nm,nm,ipvt);
    if(pinfo!=0){java.util.Arrays.fill(P,Double.NaN);}
    else {com.polytechnik.utils.Linsystems.dgesl_m1(matr,nm,nm,ipvt,P,1);}
    return P;
}

/** Generate a measure, that, when {@link #getKK getKK(n,M*QQmu*M)} is applied, produce given polynomial P_toproduce.
 * The moments are obtained solving linear system problem.
 * @param n the dimension.
 * @param P_toproduce a polynomial to produce, must have 0..2n-2, total 2n-1 elements.
 * @param M a symmetric matrix to multiply QQmu (typically G^{-1} is used).
 * @return Measure moments {@code mu} 0..2n-2.
 */
default double [] getMomentsOfMeasureProducingPolynomialInKK_MQQM(final int n,final double [] P_toproduce,final double [] M)
{
    final int nm=2*n-1;
    if(P_toproduce.length!=nm||M.length!=n*n) throw new IllegalArgumentException();
    final double [] matr=new double [nm*nm],P=java.util.Arrays.copyOf(P_toproduce,nm),tmp=new double [nm];
    if(false){
	final double [] tmpst=new double [nm];
	for(int j=0;j<n;j++) for(int k=0;k<n;k++) {
		final int njk=j+k;
		setQlQmExpansionCoefs(j,k,tmp);
		for(int s=0;s<n;s++) for(int t=0;t<n;t++)  {
			setQlQmExpansionCoefs(s,t,tmpst);
			final int nst=s+t;
			for(int m=0;m<=njk;m++) for(int l=0;l<=nst;l++){
				matr[m*nm+l]+=tmp[m]*M[j*n+s]*tmpst[l]*M[t*n+k];
			}
		}
	}
    } else {
	final double [] sumjk=new double [nm],sumsk=new double [nm];
	for(int j=0;j<n;j++) for(int s=0;s<n;s++)
	{
	    java.util.Arrays.fill(sumjk,0);
	    java.util.Arrays.fill(sumsk,0);
	    for(int k=0;k<n;k++)
	    {
		setQlQmExpansionCoefs(j,k,tmp);
		TMatr.sumP1(sumjk,tmp,j+k+1,M[k*n+s]);
		setQlQmExpansionCoefs(k,s,tmp);
		TMatr.sumP1(sumsk,tmp,k+s+1,M[k*n+j]);
	    }
	    for(int m=0;m<n+j;m++) for(int l=0;l<n+s;l++)
	    {
		matr[l*nm+m]+=sumjk[m]*sumsk[l];
	    }
	}
    }
    final int [] ipvt=new int [nm];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,nm,nm,ipvt);
    if(pinfo!=0){java.util.Arrays.fill(P,Double.NaN);}
    else {com.polytechnik.utils.Linsystems.dgesl_m1(matr,nm,nm,ipvt,P,1);}
    return P;
}


/** A conversion of two quadratic forms product.
Given two quadratic forms \(Q_j(x) A_{jk} Q_k(x) \) ; \(j,k=0\dots \mathtt{nA}-1\)
and  \(Q_u(x) B_{uv} Q_v(x) \) ; \(u,v=0\dots \mathtt{nB}-1\)
construct a matrix \( C_{gh} \) such that {@literal
\[
\begin{align}
\sum_{g,h=0}^{\mathtt{nA}+\mathtt{nB}-2} Q_g(x) C_{gh} Q_h(x) &=
\sum_{j,k=0}^{\mathtt{nA}-1} Q_j(x) A_{jk} Q_k(x)
\cdot
\sum_{u,v=0}^{\mathtt{nB}-1} Q_u(x) B_{uv} Q_v(x)
\end{align}
\]
} The conversion is performed by applying
* multiplication operator {@link #setQlQmExpansionCoefs}.
* While the polynomial {@link #getKK getKK(nA+nB-1,C)}=\(\sum_{g,h=0}^{\mathtt{nA}+\mathtt{nB}-2} Q_g(x) C_{gh} Q_h(x)\) is unique, obtained solution (the matrix \( C_{gh} \) ) is non-inique.
* It is often advantageous to use the matrix generated by the measure
* from {@link #getMomentsOfMeasureProducingPolynomialInKK_MQQM}.
* @return The matrix {@code C} of the dimension {@code (nA+nB-1)*(nA+nB-1)}.
*/
default public double [] getTwoQuadraticFormsProductAsQuadraticForm(final int nA,final double [] A,final int nB,final double [] B){
    if(A.length!=nA*nA) throw new IllegalArgumentException();
    if(B.length!=nB*nB) throw new IllegalArgumentException();
    final int d=nA-1+nB-1+1;
    final double [] C=new double [d*d],
	mul=new double [d],tmp=new double [d*nA*nB],tmpMul=new double [d*nA*nB];
    for(int j=0;j<nA;j++){
	for(int u=0;u<nB;u++){
	    setQlQmExpansionCoefs(j,u,mul);
	    System.arraycopy(mul,0,tmpMul,(j*nB+u)*d,j+u+1);
	    for(int k=0;k<nA;k++){
		for(int v=0;v<nB;v++) {
		    final double f=A[j*nA+k]*B[u*nB+v];
		    for(int s=0,se=j+u,o=(k*nB+v)*d;s<=se;s++){
			tmp[o+s]+=mul[s]*f;
		    }
		}
	    }
	}
    }

    for(int k=0;k<nA;k++){
	for(int v=0;v<nB;v++) {
	    if(false){
		final double eps=1e-6;
		setQlQmExpansionCoefs(k,v,mul);
		UnitTests.checkDoubleArrayRel1(java.util.Arrays.copyOf(mul,k+v+1),"mul",java.util.Arrays.copyOfRange(tmpMul,(k*nB+v)*d,(k*nB+v)*d + k+v+1),"tmpMul",eps);
	    }
	    for(int s=0,o=(k*nB+v)*d,skve=k+v;s<d;s++){
		for(int skv=0;skv<=skve;skv++){
		    C[s*d+skv]+=tmp[o+s]*tmpMul[o+skv];
		}
	    }
	}
    }
    return TMatr.selectHermitianPartInPlace(d,C);
}


}
