package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */

/** Basis of orthogonal polynomials.
 * Contains a few extra methods not available in
 * {@link SimpleBasisPolynomials SimpleBasisPolynomials}.
 * Typical implementation is a classic basis such as
 {@link Chebyshev},
 {@link Legendre},
 {@link Laguerre},
 {@link HermiteE},
 {@link LegendreShifted}, and
 {@link Monomials};
arbitrary three-term recurrence {@link RecurrenceAB} is also available.
 * @version $Id: BasisPolynomials.java,v 1.110 2022/08/09 17:17:28 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public interface BasisPolynomials extends SimpleBasisPolynomials {


@Override default double [] getNext(final int n,
			  final double [] Qnm1,final double [] Qnm2,
			  final BasisFunctionsCalculatable PBASIS){
    return getNext(n,Qnm1,Qnm2,false,PBASIS);
}


/** Calculates n-th basis polynomial of this basis in PBASIS basis.
 * @param n T index.
 * @param Qnm1 Input \( Q_{n-1} \).
 * @param Qnm2 Input \( Q_{n-2} \).
 * @param flag_type2 Whether to use type 2 polynomials. If type 2 polynomials are not available --- throw an exception. If not implemented -- an exception will be thrown. This method is typically not the proper way to obtain type 2 polynomials, use the {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getOrthogonalPolynomialsSecondKindFromAB getOrthogonalPolynomialsSecondKindFromAB} instead.
 * @param PBASIS The basis in which to calculte, form this basis we only need P*(a*x+b) operation, that is performed relatively this bais recurrence.
 * @return \( Q_{n} \) polynomial of this basis in PBASIS basis, array dimension n+1.
 */
double [] getNext(final int n,
		  final double [] Qnm1,final double [] Qnm2,
		  final boolean flag_type2,
		  final BasisFunctionsCalculatable PBASIS);




/** Calculates \( Q_n(a*(x+x_0)) \) using 3 term recurrence.
 * Calculations in  basis.
 * The \( b_k^{n} \) are defined as:
 * \( Q_n(a*(x+x_0))=\sum_{k=0}^{k=n}b_k^{n} Q_k(x) \)
 * The method should be applied consequently to get \( b_k^{n} \) for n=0,1,2,3,... (in-place calculation).
 * This is a generalization of
 * Newton Binomial ({@link Monomials monomials} case):
 * \( (a*(x+x_0))^n=a^n\sum x^k x_0^{n-k} C_{n}^{k} \);
 * \( b_k^n= a^n x_0^{n-k} C_{n}^{k} \).
 * This method calculates "Newton Binomial" in arbitrary basis:
 * \( Q_n(a*(x+x_0))=\sum_{k=0}^{k=n}b_k^{n} Q_k(x) \).
 * Actual implementation is special, highly optimized version of
 * {@link SimpleBasisPolynomials#getBasisFunctionsOnPolynomialArgument getBasisFunctionsOnPolynomialArgument} with \( pol(x)=a*(x+x_0) \). The default implementation is very crude and slow (calls {@link SimpleBasisPolynomials#getBasisFunctionsOnPolynomialArgument getBasisFunctionsOnPolynomialArgument}),
 * it is overloaded in actual implementations for speed.
 * <br><br>
 * Typical usage: realtime data moments calculation speed-up by recurrent calculation according to time--shift, see {@link com.polytechnik.trading.QVMDataP QVMDataP}, {@link com.polytechnik.trading.QVMDataL QVMDataL}, and {@link com.polytechnik.trading.QVMData QVMData}.
 * @param n The max dimension of b.
 * @param b On input: \( b_k^{n-1} \), on output \( b_k^{n} \).
 * @param q On input: \( b_k^{n-2} \), on output: \( b_k^{n-1} \).
 */
default void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0){
	final double [][] Qlpolx=getBasisFunctionsOnPolynomialArgument(n+1,convertFunctionTo_Basis_from_Monomials(new double []{a*x0,a}));
	System.arraycopy(Qlpolx[n],0,b,0,Qlpolx[n].length);
	if(n>0){
	    System.arraycopy(Qlpolx[n-1],0,q,0,Qlpolx[n-1].length);
	}
}


} // END of class BasisPolynomials
