package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */

/** A class to store read sample data.
  * @version $Id: DataReadObservationVectorXVectorF.java,v 1.24 2024/04/05 09:06:40 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
 */
public class DataReadObservationVectorXVectorF {
    public final java.util.List<ObservationVectorXVectorF> data;
    public final java.util.List<String> xHeader,fHeader;
    public final DataStat datastat;

    public DataReadObservationVectorXVectorF(final java.util.List<ObservationVectorXVectorF> data,
					     final java.util.List<String> xHeader,
					     final java.util.List<String> fHeader)
    {
	this.data=data;
	this.xHeader=xHeader;
	this.fHeader=fHeader;
	this.datastat=new DataStat(this);
    }

    public void printData()
    {
	for(ObservationVectorXVectorF o:data)
	{
	    System.err.println("x="+java.util.Arrays.toString(o.x)+"\tf="+java.util.Arrays.toString(o.f)+"\tweight="+o.weight);
	}
    }

    /** Calculate &lt;observationToFun1(o)[j]|observationToFun2(o)[k]&gt;
     * averages. For unit test.
     * @param observationToFun1 a function to obtain first arg to average.
     * @param observationToFun2 a function to obtain second arg to average.
     * @param observationCommonFactor a function with common factor.
     * @return  The matrix [j][k] of Fun1.length rows and Fun2.length columns
     * of product averages.
     */
    public double [] getFun1Fun2Averaged(
	final java.util.function.Function<ObservationVectorXVectorF,double []> observationToFun1,
	final java.util.function.Function<ObservationVectorXVectorF,double []> observationToFun2,
	final java.util.function.ToDoubleFunction<ObservationVectorXVectorF>observationCommonFactor){

	final var ot=new ObservationVectorXVectorF(null,new double [xHeader.size()],new double [fHeader.size()],1);
	final int nFun1Values=observationToFun1.apply(ot).length,
	    nFun2Values=observationToFun2.apply(ot).length;
	final double [] scalprods=new double [nFun1Values*nFun2Values];
	for(ObservationVectorXVectorF o:data)
	{
	    final double [] x=observationToFun1.apply(o);
	    if(x.length!=nFun1Values) throw new IllegalArgumentException();
	    final double [] y=observationToFun2.apply(o);
	    if(y.length!=nFun2Values) throw new IllegalArgumentException();
	    final double cfactor=observationCommonFactor.applyAsDouble(o)*o.weight;
	    for(int j=0,of=0;j<x.length;j++){
		final double jfact=x[j]*cfactor;
		for(int k=0;k<y.length;k++,of++){
		    scalprods[of]+=y[k]*jfact;
		}
	    }
	}
	return scalprods;
    }



    /** For unit test. */
    DataReadObservationVectorXF getXBasisSample_OLDSCALARCODE(){
	final var dataXF=new java.util.ArrayList<ObservationVectorXF>(10+data.size());
	int l=0;
	for(var o:data){
	    // observation number as f
	    dataXF.add(new ObservationVectorXF(o.label,o.x,++l,o.weight));
	}
	return new DataReadObservationVectorXF(dataXF,xHeader);
    }

    /** For unit test. */
    DataReadObservationVectorXF getFBasisSample_OLDSCALARCODE(){
	final var dataXF=new java.util.ArrayList<ObservationVectorXF>(10+data.size());
	int l=0;
	for(var o:data){
	    // observation number as f
	    dataXF.add(new ObservationVectorXF(o.label,o.f,++l,o.weight));
	}
	return new DataReadObservationVectorXF(dataXF,fHeader);
    }



    /**  Regularization parameter.
     */
    static final double eps_rel=1e-5,eps_abs=1e-6;


    public static class DataStat {
	public final double totalweight;
	final double [] xorig_average,forig_average;
	DataStat(DataReadObservationVectorXVectorF data)
	{
	    double wtotal=0;
	    final int n=data.xHeader.size();
	    final int m=data.fHeader.size();
	    final double [] xa=new double [n];
	    final double [] fa=new double [m];
	    for(ObservationVectorXVectorF o:data.data)
	    {
		if(o.x.length!=n) throw new IllegalArgumentException("o.x.length="+o.x.length+" n="+n);
		TMatr.sumP1(xa,o.x,o.weight);
		if(o.f.length!=m) throw new IllegalArgumentException("o.f.length="+o.f.length+" m="+m);
		TMatr.sumP1(fa,o.f,o.weight);
		wtotal+=o.weight;
	    }
	    this.xorig_average=TMatr.getVectorScaled(xa,1/wtotal);
	    this.forig_average=TMatr.getVectorScaled(fa,1/wtotal);
	    this.totalweight=wtotal;
	}
    }

/** Fields separator.
 */
public static final String SEPARATOR=DataReadObservationVectorXF.SEPARATOR;
/** Read observations sequence from a file.
 * The lines starting with "|" are considered as comments.
 * If the first line starts with "|#", then it considered as attributes names.
 */
public static DataReadObservationVectorXVectorF getObservations(
    final String filename,
    final int columns_number,
    final int [] indcol_x,
    final int [] indcol_f,
    final int indcol_weight,
    final int indcol_label)
	throws java.io.IOException
    {
	final var data=new java.util.ArrayList<ObservationVectorXVectorF>();
	final java.util.function.Function<String,String> postprocess=(str)->str.trim();
	System.err.print("Opening file="+filename);
	final var br=new java.io.BufferedReader(new java.io.FileReader(filename));
        final java.io.StreamTokenizer tk=ReadWriteUtils.getTokenizer(SEPARATOR,br);
        tk.quoteChar('\"');

	if(/*indcol_weight<0||*/indcol_weight>=columns_number) throw new IllegalArgumentException("indcol_weight="+indcol_weight+" is invalid, must be <"+columns_number);

	final var xHeader=new java.util.ArrayList<String>(indcol_x.length);
	// default header
	for(int c=0;c<indcol_x.length;c++){
	    xHeader.add("x["+c+"]:ind="+indcol_x[c]);
	    if(indcol_x[c]<0||indcol_x[c]>=columns_number) throw new IllegalArgumentException("indcol_x["+c+"]="+indcol_x[c]+" is invalid, must be >=0 and <"+columns_number);
	}

	final var fHeader=new java.util.ArrayList<String>(indcol_f.length);
	// default header
	for(int c=0;c<indcol_f.length;c++){
	    fHeader.add("f["+c+"]:ind="+indcol_f[c]);
	    if(indcol_f[c]<0||indcol_f[c]>=columns_number) throw new IllegalArgumentException("indcol_f["+c+"]="+indcol_f[c]+" is invalid, must be >=0 and <"+columns_number);
	}

	int l=0;
	for(;;)
	{
	    final String [] s=ReadWriteUtils.readOneLine(tk);
	    if(s==null) break; // end of file
	    // trim spaces
	    for(int c=0;c<s.length;c++) s[c]=postprocess.apply(s[c]);
	    l++;
	    if(l==1){
		// First line is possibly the header
		if(s.length>0 && s[0].startsWith("|#")){
		    if(s.length!=columns_number) throw new RuntimeException("File="+filename+" line="+l+" records number is not "+columns_number+" from string="+java.util.Arrays.toString(s));
		    s[0]=s[0].substring(2); // remove |#
		    xHeader.clear();
		    // replace by provided header
		    for(int c=0;c<indcol_x.length;c++){
			xHeader.add("x["+c+"]"+s[indcol_x[c]]+":ind="+indcol_x[c]);
		    }
		    fHeader.clear();
		    // replace by provided header
		    for(int c=0;c<indcol_f.length;c++){
			fHeader.add("f["+c+"]"+s[indcol_f[c]]+":ind="+indcol_f[c]);
		    }
		    continue;
		}
	    }
	    //System.err.println("line="+l+" s.length="+s.length+" s="+java.util.Arrays.toString(s));
	    if(s.length<=0||s[0].startsWith("|")) continue; // comments
	    if(s.length!=columns_number) throw new RuntimeException("File="+filename+" line="+l+" records number="+s.length+" is not "+columns_number+" from string="+java.util.Arrays.toString(s));
	    final double [] x=new double [indcol_x.length];
	    for(int c=0;c<indcol_x.length;c++){
		x[c]=Double.valueOf(s[indcol_x[c]]);
	    }
	    final double [] f=new double [indcol_f.length];
	    for(int c=0;c<indcol_f.length;c++){
		f[c]=Double.valueOf(s[indcol_f[c]]);
	    }
	    data.add(new ObservationVectorXVectorF(indcol_label<0?null:s[indcol_label],x,f,indcol_weight<0?1:Double.valueOf(s[indcol_weight])));
	}
	br.close();
	System.err.println(" read "+l+" lines , nrecords="+data.size());
	return new DataReadObservationVectorXVectorF(data,xHeader,fHeader);
    }


     /**
     * @param regularization_methodX regularizatuion method of {@code x}.
     * @param regularization_methodF regularizatuion method of {@code f}.
     * @return Regularized data.
     */
     public RegularizedVectorXVectorF getDataRegularized_OLDSCALARCODE(
	 final RegularizationMethod regularization_methodX,
	 final RegularizationMethod regularization_methodF){

	 final RegularizedVectorXF
	     sX_old=RN.getDataRegularizedAccordingToConfig(getXBasisSample_OLDSCALARCODE(),regularization_methodX),
	     sF_old=RN.getDataRegularizedAccordingToConfig(getFBasisSample_OLDSCALARCODE(),regularization_methodF);
	 return new RegularizedVectorXVectorF(
	     sX_old.getDataWithoutF(),
	     sF_old.getDataWithoutF());
     }

    /** An adapter to use old scalar-f code.
     */
    static DataReadObservationVectorXVectorF getVectorXVectorFromScalars(
	final java.util.function.Function<DataReadObservationVectorXVectorF,DataReadObservationVectorXF> toX,
	final java.util.function.Function<DataReadObservationVectorXVectorF,DataReadObservationVectorXF> toF,
	final DataReadObservationVectorXVectorF orig_sample){

	final DataReadObservationVectorXF prodX=toX.apply(orig_sample),prodF=toF.apply(orig_sample);
	// merge two scalar samples into one vector sample

	final int no=orig_sample.data.size();
	if(prodX.data.size()!=no||prodF.data.size()!=no) throw new IllegalArgumentException();
	final var dataProd=new java.util.ArrayList<ObservationVectorXVectorF>(10+no);
	final var isoX=prodX.data.iterator();
	final var isoF=prodF.data.iterator();
	for(var o:orig_sample.data){
	    final ObservationVectorXF soX=isoX.next(),soF=isoF.next();
	    dataProd.add(new ObservationVectorXVectorF(o.label,soX.x,soF.x,o.weight));
	}
	return new DataReadObservationVectorXVectorF(
	    dataProd,
	    prodX.xHeader,
	    prodF.xHeader);
    }


} // END OF class DataReadObservationVectorXVectorF
