package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */

/** A container to store eigenvalue problem solution.
 *  @see LebesgueQuadratureWithEVData
 *  @version $Id: EVXData.java,v 1.134 2022/11/22 10:53:05 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class EVXData {
    public final double sL,sH,wL,wH,s0,o_o,x0;
    public final double [] psiL,psiH,psi0,ev,psik_at_x0;
    public final int neq;

    public EVXData(final int neq,
		   final double wL,final double sL,final double [] psiL,
		   final double wH,final double sH,final double [] psiH,
		   final double [] psi0,final double [] ev,final double [] psik_at_x0,
		   final double s0,final double o_o,final double x0)
    {
	this.neq=neq;
	this.wL=wL;
	this.sL=sL;
	this.psiL=psiL;
	this.wH=wH;
	this.sH=sH;
	this.psiH=psiH;
	this.psi0=psi0;
	this.ev=ev;
	this.psik_at_x0=psik_at_x0;
	this.s0=s0;
	this.o_o=o_o;
	this.x0=x0;
    }

/** Return given eigenvector.
 */
public double [] getPsi(final int ind_lev)
{
    if(ind_lev<0||ind_lev>=neq)
	throw new RuntimeException("ind_lev="+ind_lev+" neq="+neq);
    return getPsiFromBasis(ind_lev,this.ev,neq);
}

/** Return all eigenvalues.
 * @see LebesgueQuadratureWithEVData#f
 */
public double [] getEValues(){
    return java.util.Arrays.copyOf(this.ev,this.neq);
}

/** Return a single diagonal matrix element
 * {@literal <psi_{ind_lev}|FQQ|psi_{ind_lev}>}.
 * @return FQQ diagonal matrix element.
 */
public double getFPsi(final int ind_lev,final double [] FQQ)
{
    return getFPsiPsi(ind_lev,ind_lev,FQQ);
}

/** Calculates {@literal <psik|FQQ|psik>} for all eigenvectors.
 * @param FQQ The matrix.
 * @return FQQ diagonal matrix elements.
 */
public double [] getFPsi(final double [] FQQ)
{
    final double [] fpsi=new double [neq];
    for(int k=0;k<neq;k++) fpsi[k]=getFPsi(k,FQQ);
    return fpsi;
}

/** Return {@literal lambda[k]*<psik|FQQ|psik>} for all eigenvectors.
 * @param FQQ The matrix.
 * @return FQQ diagonal matrix elements multiplied by eigenvalues.
 */
public double [] getLambdaFPsi(final double [] FQQ)
{
    return TMatr.getVAB(neq,getFPsi(FQQ),ev);
}

/** Calculates {@literal <psi_{ind_lev}|FQQ|psi>}
 * for a given eigenvector {@code ind_lev} and wavefunction {@literal |psi>}.
 */
public double getFPsiPsi(final int ind_lev,final double [] psi,final double [] FQQ)
{
    final int ldFQQ=neq;
    if(psi.length>neq||ind_lev<0||ind_lev>=neq||ldFQQ<neq) throw new IllegalArgumentException();
    return TMatr.getAMatrB(neq,this.ev,neq+ind_lev*neq,psi.length,psi,0,FQQ,ldFQQ);
}

/** Calculates {@literal <psi_i|FQQ|psi_j>}.
 * @return A single matrix element.
 */
public double getFPsiPsi(final int i,final int j,final double [] FQQ)
{
    final int ldFQQ=neq;
    if(i<0||i>=neq||j<0||j>=neq||ldFQQ<neq) throw new IllegalArgumentException();
    return TMatr.getAMatrB(neq,this.ev,neq+i*neq,neq,this.ev,neq+j*neq,FQQ,ldFQQ);
}

/** Calculate all eigenvectors at given distribution of basis functions.
 * Regular values are {@code getPsiAt(M.B.getX0Moments(this.neq,x))}.
 * @see LebesgueQuadratureWithEVData#getPsiAt
 */
public double [] getPsiAt(final double [] Qbasis_distribution)
{
    if(Qbasis_distribution.length!=neq) throw new IllegalArgumentException();
    final double [] psikat=new double [neq];
    for(int k=0;k<neq;k++)
    {
	psikat[k]=TMatr.mulVV(neq,Qbasis_distribution,0,ev,neq+k*neq);
    }
    return psikat;
}

/** Calculate RN interpolation at given distribution of basis functions.
 * Regular RN interpolation is {@code getRNat(M.B.getX0Moments(this.neq,x))}.
 * @see LebesgueQuadratureWithEVData#getEvaluationAt
 */
public double getRNat(final double [] Qbasis_distribution)
{
    if(Qbasis_distribution.length!=neq) throw new IllegalArgumentException();
    final double [] psikat=getPsiAt(Qbasis_distribution);
    // The average of weighted eigenvalues.
    double sf=0,s=0;
    for(int k=0;k<neq;k++)
    {
	final double wproj=Math.pow(psikat[k],2);
	s+=wproj;
	sf+=ev[k]*wproj;
    }
    return sf/s;
}
/** Calculates {@literal <psik|QQ|1>}.
 * @see LebesgueQuadratureWithEVData#psiaverage
 */
public double [] getPsiAveraged(final double [] QQ,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int n=neq;
    if(n*n!=QQ.length) throw new IllegalArgumentException();
    final double [] psiaverage=new double [n],C=M.B.getONE(); // make C=1 in basis.
    for(int k=0;k<n;k++) psiaverage[k]=TMatr.getAMatrB(C.length,C,0,n,ev,n+k*n,QQ,n);
    return psiaverage;
}


/** Make the highest order coefficient of sign=1.
 * Used for unit tests to match eigenvectors within +- sign.
 */
public static double [] getMSign(final double [] a)
{
    return getMSign(a,1);
}
/** Make the highest order coefficient of specific sign.
 * Used for unit tests to match eigenvectors within +- sign.
 */
public static double [] getMSign(final double [] a,final double sign)
{
    final double [] b=a.clone();
    if(Math.signum(b[b.length-1])==Math.signum(sign)) return b;
    TMatr.scaleVectorInPlace(b,-1);
    return b;
}

/** Test orthogonality unit test.
 */
public static void test_orthogonality(
    final String txt,
    final int n,final double [] basis,final double [] QQi,final double [] QQt,
    final double eps)
{
    for(int k=0;k<n;k++)
    {
	final double [] psik=EVXData.getPsiFromBasis(k,basis,n);
	for(int l=0;l<n;l++)
	{
	    final double [] psil=EVXData.getPsiFromBasis(l,basis,n);
	    if(k==l){
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQt,n),"<psik|psik>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQi,n),"<psik|I|psik>",basis[k],"basis[k]",eps);
	    }
	    else {
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQt,n),"<psik|psik>",0,"0",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrB(psik,psil,QQi,n),"<psik|I|psil>",0,"0",eps);
	    }
	}
    }
    if(txt!=null) System.err.println(txt+":test_orthogonality test OK n="+n+" eps="+eps);
}


public static void matchEVXData(final EVXData a,final EVXData b,final double eps)
{
    matchEVXData(a,b,true,eps);
}
public static void matchEVXData(final EVXData a,final EVXData b,final boolean flag_psicheck,final double eps)
{
    matchEVXData(a,b,flag_psicheck,true,eps);
}
public static void matchEVXData(final EVXData a,final EVXData b,final boolean flag_psicheck,final boolean flag_check_valuesAt_x0,final double eps)
{
    UnitTests.checkDoubleRel(a.neq,"a.neq",b.neq,"b.neq",0);
    UnitTests.checkDoubleArrayRel1(a.getEValues(),"a.getEValues()",b.getEValues(),"b.getEValues()",eps);
    UnitTests.checkDoubleRel(a.sL,"a.sL",b.sL,"b.sL",eps);
    UnitTests.checkDoubleRel(a.sH,"a.sH",b.sH,"b.sH",eps);
    if(flag_check_valuesAt_x0)
    {
	UnitTests.checkDoubleRel(a.o_o,"a.o_o",b.o_o,"b.o_o",eps);
	UnitTests.checkDoubleRel(a.s0,"a.s0",b.s0,"b.s0",eps);
	UnitTests.checkDoubleRel1(Math.abs(a.wL),"|a.wL|",Math.abs(b.wL),"|b.wL|",eps);
	UnitTests.checkDoubleRel1(Math.abs(a.wH),"|a.wH|",Math.abs(b.wH),"|b.wH|",eps);
	for(int k=0;k<a.neq;k++){
	    UnitTests.checkDoubleRel1(Math.abs(a.psik_at_x0[k]),"abs(a.psik_at_x0[k])",Math.abs(b.psik_at_x0[k]),"abs(b.psik_at_x0[k])",eps);
	}
    }
    if(flag_psicheck)
    {
	UnitTests.checkDoubleArrayRel1(a.psiL,"a.psiL",getMSign(b.psiL,a.psiL[a.neq-1]),"b.psiL",eps);
	UnitTests.checkDoubleArrayRel1(a.psiH,"a.psiH",getMSign(b.psiH,a.psiH[a.neq-1]),"b.psiH",eps);
	UnitTests.checkDoubleArrayRel1(a.psi0,"a.psi0",getMSign(b.psi0,a.psi0[a.neq-1]),"b.psi0",eps);
	for(int k=0;k<a.neq;k++){
	    final double [] apsik=a.getPsi(k),bpsik=b.getPsi(k);
	    UnitTests.checkDoubleArrayRel1(apsik,"apsik",getMSign(bpsik,apsik[a.neq-1]),"bpsik",eps);
	}
    }
    System.err.println("EVXData:matchEVXData test OK neq="+a.neq+" eps="+eps);
}

/** A warapper to convert {@link LebesgueQuadratureWithEVData} to {@link EVXData} object.
 */
public static EVXData getEVXDataFromLebesgueQuadrature(final LebesgueQuadratureWithEVData q)
{
    return _getEVXDataFromEVStates(q.f.length,q.ev,q.M,q.M.getRadauRoot());
}
/** Calculate EVXData from two matrices.
 */
public static EVXData getEVXDataFromQQ(final int n,final double [] QQf,final double [] QQa,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    return _getEVXDataFromEVStates(n,M.EV.getGEV_Generalized(n,QQf,QQa,true),M,M.getRadauRoot());
}
/** Calculate EVXData from two matrices, assuming the QQa is diagonal.
 */
public static EVXData getEVXDataFromQQfDiagQQa(final int n,final double [] QQf,final double [] QQa,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    return _getEVXDataFromEVStates(n,getSymEVDiagT(n,QQf,QQa,M.EV),M,M.getRadauRoot());
}

private static EVXData _getEVXDataFromEVStates(final int n,final double [] evS,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
					 final double x0)
{
    if(evS.length!=n+n*n) throw new IllegalArgumentException();
    final double [] psiL=new double [n],psiH=new double [n],psi0=new double [n],psik_at_x0=new double [n];
    final double [] r=psiL; // use as a temporary array
    double s_s_o_o=0,ss0=0,s_wL=0,s_wH=0;
    for(int k=0;k<n;k++)
    {
	System.arraycopy(evS,n+(k)*n,r,0,n);
	final double fkx0=M.B.getD0(r,x0),f2kx0=fkx0*fkx0;
	psik_at_x0[k]=fkx0;
	TMatr.sumP1(psi0,r,fkx0);
	if(k<=0) s_wL=fkx0;
	if(k>=n-1) s_wH=fkx0;
	s_s_o_o+=f2kx0;
	ss0+=f2kx0*evS[k];
    }
    final double sL=evS[0],sH=evS[n-1];
    System.arraycopy(evS,n+(0)*n,psiL,0,n);
    System.arraycopy(evS,n+(n-1)*n,psiH,0,n);
    final double sqrt_s_s_o_o=Math.sqrt(s_s_o_o);
    final double wL=s_wL/sqrt_s_s_o_o,wH=s_wH/sqrt_s_s_o_o,s0=ss0/s_s_o_o;
    for(int k=0;k<n;k++) psi0[k]/=sqrt_s_s_o_o;

    return new EVXData(n,wL,sL,psiL,wH,sH,psiH,psi0,evS,psik_at_x0,s0,s_s_o_o,x0);
}

/** Extract psi from basis. The basis is n eigenvalues, followed by n eigenvectors.
 */
public static double [] getPsiFromBasis(final int ind,final double [] basis,final int n)
{
    if(ind<0||ind>=n||n*n+n!=basis.length) throw new IllegalArgumentException();
    return getPsiFromBasis(ind,basis,n,n);
}
/** Extract psi from basis. Eigenvectors start at given offset, n elements each.
 */
public static double [] getPsiFromBasis(final int ind,final double [] basis,final int n,final int basis_offset)
{
    //if(ind<0||ind>=n||n*n+basis_offset!=basis.length) throw new IllegalArgumentException();
    final double [] psi=new double [n];
    System.arraycopy(basis,basis_offset+ind*n,psi,0,n);
    return psi;
}


/**
 * Given b_i=\sum_k eVB_{ik} Q_k and {@literal <b_i b_j>=1}, {@literal <Q_k Q_l>=G_{lk}},
 * then Q_k = \sum_l {@literal <Q_k Q_l>} eVB_lk and given a matrix {@literal matrB=<b_i|f|b_j>}
 * the matrix {@literal <Q_k|f|Q_s>} = \sum_{l,i,j,m} G_{kl} eVB_il {@literal <b_i|f|b_j>} eVB_jm G_{ms}.
 * The method calculates G eV^T B eV G^T.
 * @param n Basis dimension.
 * @param matrB The matrix in B basis.
 * @param eVB Conversion of Q basis to B basis. An array of n+n*n elements, firts n elements are eigenvalues (ignored) then n eigenvectors follow.
 * @param G Gram matrix in Q basis.
 * @return The matrix converted to Q basis.
 */
public static double [] convertMatrFromBtoQ_Basis(final int n,final double [] matrB,final double [] eVB,final double [] G)
{
    if(n*n!=matrB.length||matrB.length!=G.length||n+n*n!=eVB.length) throw new IllegalArgumentException();
    final double [] GeVT=new double [n*n];
    final int basis_offset=n;
    TMatr.multiply_MA_x_MBTrans(GeVT,n,G,n,0+0,n,eVB,n,n,basis_offset+0*n);
    return convertMatrFromQtoB_Basis(n,matrB,GeVT,n,0+0);
}


/** Calculate eV matrQ eV^T.
 * @param n Basis dimension.
 * @param matrQ The matrix in Q basis.
 * @param eVB Conversion of Q basis to B basis. An array of n+n*n elements, firts n elements are eigenvalues (ignored) then n eigenvectors follow.
 * @return The matrix converted to B basis.
 */
public static double [] convertMatrFromQtoB_Basis(final int n,final double [] matrQ,final double [] eVB)
{
    //case n=nd, when the number
    return convertMatrFromQtoB_Basis(n,matrQ,eVB,n,n+0);
}

/** Transorm quadratic form matrix to new basis.
 *  @param n Original basis dimesion.
 *  @param matrQ Matrix in original basis, n x n matrix lD=n.
 *  @param basis New basis, nd x n matrix, lD=n, possible data offset basis_offset.
 *  @param nd New basis dimension.
 *  @param basis_offset Offset of the baisis.
 *  @return Output, nd x nd matrix, res=basis x matrQ x basis^T.
 */
public static double [] convertMatrFromQtoB_Basis(final int n,final double [] matrQ,final double [] basis,final int nd,final int basis_offset)
{
    final double [] tmp=new double [nd*n],res=new double [nd*nd];
    // matrQ: n x n matrix in original Q basis.
    // basis: nd x n matrix original Q basis.
    // res: nd x nd matrix in new basis
    TMatr.multiply_MA_x_MBTrans(tmp,n,matrQ,n,0,n,basis,nd,n,basis_offset+0*n); // tmp: n x nd
    TMatr.multiply_MA_x_MB(res,nd,basis,n,basis_offset+0*n,n,tmp,nd,nd,0); // res: nd x nd
    return res;
}

/** Calculates eVB (ndxldQQ) x QQT (ldQQxldQQ) x psiQ.
 */
public static double [] convertPsiFromQtoB_Basis(final double [] psiQ,final double [] eVB,final int nd,final int basis_offset,final double [] QQt,final int ldQQ)
{
    if(eVB.length!=basis_offset+nd*ldQQ||QQt.length!=ldQQ*ldQQ||psiQ.length>ldQQ) throw new IllegalArgumentException();
    final double [] tmp=new double [ldQQ];
    TMatr.multiply_MA_x_MB(tmp,ldQQ,QQt,ldQQ,0,psiQ.length,psiQ,1,1,0); // tmp: 1xldQQ
    final double [] psiB=new double [nd];
    TMatr.multiply_MA_x_MB(psiB,nd,eVB,ldQQ,basis_offset,ldQQ,tmp,1,1,0); // psiB: nd x 1
    return psiB;
}
/** Calculate eVB G psiQ.
 */
public static double [] convertPsiFromQtoB_Basis(final int n,final double [] psiQ,final double [] eVB,final double [] QQt)
{
    if(n!=psiQ.length) throw new IllegalArgumentException();
    return convertPsiFromQtoB_Basis(psiQ,eVB,n,n,QQt,n);
}
public static double [] convertPsiFromBtoQ_Basis(final int n,final double [] psiB,final double [] eVB)
{
    return convertPsiFromBtoQ_Basis(n,psiB,n,eVB,n+0);
}
/** Calculate  l=0..nq-1: psiQ[l]=sum_{k=0..nb-1} psiB[k]*basis[l+k*nq+basis_offset]
 */
public static double [] convertPsiFromBtoQ_Basis(final int nb,final double [] psiB,final int nq,final double [] basis,final int basis_offset)
{
    if(basis.length!=basis_offset+nb*nq||psiB.length!=nb) throw new IllegalArgumentException();
    final double [] psiQ=new double [nq];
    //TMatr.multiply_MATrans_x_MB(psiQ,nq,basis,nq,basis_offset,nb,psiB,1,1,0);
    TMatr.multiply_MA_x_MB(psiQ,1,psiB,nb,0,nb,basis,nq,nq,basis_offset); // same thing
    return psiQ;
}

/** Calculate basisA QQ basisB^T.
 * @return {@literal matrix[i*n+j]= <psiA[i]|QQ|psiB[j]>}
 */
public static double [] getBasisABProjections(final int n,final double [] basisA,final double [] QQ,final double [] basisB)
{
    if(basisA.length!=n*(n+1)||basisB.length!=basisA.length||QQ.length!=n*n) throw new IllegalArgumentException();
    final double [] tmp=new double [n*n],res=new double [n*n];
    // matrQ: n x n matrix in original Q basis.
    TMatr.multiply_MA_x_MBTrans(tmp,n,QQ,n,0,n,basisB,n,n,n+0*n);
    TMatr.multiply_MA_x_MB(res,n,basisA,n,n+0*n,n,tmp,n,n,0);
    return res;
}

/** Solve generalised eigenvalues problem,
 *  assuming QQt is diagonal.
 */
public static double [] getSymEVDiagT(final int n,final double [] QQv,final double [] QQt,final EVSolver MEV)
{
    final double [] qq=new double [n*n],qqn=new double [n];
    // here we assume the QQt is diagonal.
    for(int k=0;k<n;k++)
    {
	final int ok=k*n;
	final double diagk=QQt[ok+k];
	if(!(diagk>0))
	{
	    final double [] qqr=new double [n*n+n];
	    java.util.Arrays.fill(qqr,Double.NaN);
	    return qqr;
	}
	final double s=qqn[k]=Math.sqrt(diagk);
	for(int l=0;l<=k;l++)
	{
	    qq[ok+l]=qq[l*n+k]=QQv[ok+l]/(l==k?diagk:s*qqn[l]);
	}
    }
    final double [] ev=MEV.getEV(n,qq,true);
    // convert back to our basis
    for(int k=0;k<n;k++)
    {
	for(int l=0,o=(k+1)*n;l<n;l++,o++) ev[o]=ev[o]/(qqn[l]);
    }
    return ev;
}


} // END OF class EVXData
