package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */

/**
 * Perform simplified version of
 * <a href="https://www.netlib.org/utk/papers/factor/node7.html">LU Factorization</a>
 * by checking max pivot at rows and columns.
 * Expand {@code m=conditions.length} constraint equations, \( i=0\dots m-1 \)
 * for \(\mathbf{x}\) vector of the dimension {@code n=conditions[i].length=nC*nX}.
$$ 0=\sum\limits_{j=0}^{n-1} x[j]*\mathrm{conditions}[i][j] $$
to the form
$$x[s]=\sum\limits_{j=0}^{n-1} x[j]*\mathrm{expansion}[i_s][j]$$
 * the \(  \mathrm{expansion}[i\leq i_{s}][s]=0 \) thus the \( x[s] \)
 * can be eliminated by consequent application of the expansion
 * starting from the last one.
 * This allows to build a linear expansion
 * $$ x[j]=\sum\limits_{i=0}^{\mathrm{dimXnew}-1} M_{ji} V_i $$
 * of original \( \mathbf{x} \) of the dimension \(n\)
 * via reduced size \( \mathbf{V} \) of the dimension
 * \( n-\mathrm{rank}(\mathrm{conditions}) \).
 * The method {@link fromXnewToXorigAsMatrix} creates this \( M_{ji} \) matrix.
 * @version $Id: EliminateLinearConstraints_HomegrownLUFactorization.java,v 1.35 2024/05/12 22:45:04 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class EliminateLinearConstraints_HomegrownLUFactorization {

    /** Original problem dimension. */
    public final int dimXorig;
    /** The number of eliminated variables, equals to the rank of
     * {@code conditions} matrix.
     */
    public final int nEliminated;
    /** New problem dimension {@code dimXnew=dimXorig-nEliminated}. */
    public final int dimXnew;


    /** Eliminated variables. */
    private final java.util.ArrayList<EliminateVector > elim;

    /** Not a large size, does not worth a HashMap.
     */
    private EliminateVector  getByIndX(final int indX){
	for(var e:elim){
	    if(e.indX==indX) return e;
	}
	return null;
    }

    /** Constructor.
     * @param dimXorig Original problem dimension.
     * @param conditions Linear constraints, a matrix of nEq x dimXorig.
     * Obtained {@link nEliminated} is the rank of this matrix.
     * @param condMaxToFail Minimal value of maximal absolute element
     * to decide the remaining constraints are degenerated and
     * elimintation process should be stopped, {@code 1e-10} is a typical value.
     */
    public EliminateLinearConstraints_HomegrownLUFactorization(
	final int dimXorig,
	final double [][] conditions,
	final double condMaxToFail){

	this.dimXorig=dimXorig;
	this.elim=new java.util.ArrayList<EliminateVector >();
	double [][] curCond=conditions;
	for(int i=0;i<conditions.length;i++){
	    if(conditions[i].length!=dimXorig) throw new IllegalArgumentException();
	    final var e=EliminateVector.expandSingleElimination(curCond,condMaxToFail);
	    if(e!=null){
		// success
		curCond=e.conditionsLeft;
		// check no dups
		if(null!=getByIndX(e.indX)) throw new IllegalArgumentException();
		elim.add(e);
	    } else break;
	}
	this.nEliminated=elim.size();
	this.dimXnew=dimXorig-this.nEliminated;
    }

    /** Discard some components.
     */
    public double [] fromXorigToXnew(double [] xorig){
	if(xorig.length!=dimXorig) throw new IllegalArgumentException();
	final double [] xnew=new double [dimXnew];
	int o=0;
	for(int j=0;j<xorig.length;j++){
	    if(null==getByIndX(j)){
		xnew[o++]=xorig[j];
	    }
	}
	if(o!=dimXnew) throw new IllegalArgumentException();
	return xnew;
    }

    /** From new to old, recover previously eliminated
     * components based on linear constraints.
     */
    public double [] fromXnewToXorig(double [] xnew){
	//System.err.println("dimXnew="+dimXnew+" xnew.length="+xnew.length+" nEliminated="+nEliminated);
	if(xnew.length!=dimXnew) throw new IllegalArgumentException();
	final double [] xorig=new double [dimXorig];
	int o=0;
	for(int j=0;j<xorig.length;j++){
	    if(null==getByIndX(j)){
		xorig[j]=xnew[o++];
	    }
	}
	if(o!=dimXnew) throw new IllegalArgumentException();

	// inverse elimination
	for(int j=elim.size(); --j>=0;){
	    final var e=elim.get(j);
	    xorig[e.indX] += TMatr.mulVV(e.xExpansion,xorig);
	}
	return xorig;
    }

    /** From new to old, recover components based on linear constraints,
      * the result is presented as a matrix of dimXorig * dimXnew size.
      */
    public double [] fromXnewToXorigAsMatrix(){
	final double [] conv=new double [dimXorig*dimXnew];

	int o=0;
	for(int i=0;i<dimXorig;i++){
	    if(null==getByIndX(i)){
		conv[i*dimXnew+(o++)]=1;
	    }
	}
	if(o!=dimXnew) throw new IllegalArgumentException();

	// inverse elimination
	final double [][] mElim=new double [elim.size()][];
	for(int i=elim.size(); --i>=0;){
	    mElim[i]=elim.get(i).xExpansion.clone();
	}
	for(int i=elim.size(); --i>=0;){
	    final var e=elim.get(i);
	    final double [] et=mElim[i],c=fromXorigToXnew(et);
	    System.arraycopy(c,0,conv,e.indX*dimXnew,c.length);
	    for(int j=i-1;j>=0;j--){
		TMatr.sumP1(mElim[j],et,mElim[j][e.indX]);
	    }
	}
	return conv;
    }

    private static class EliminateVector  {
	final int indX;
	final int dimXorig;
	final double [] xExpansion;
	final double [][] conditionsLeft;
	final double condNum;
	EliminateVector(final int indX,
			final int dimXorig,
			final double [] xExpansion,
			final double [][] conditionsLeft,
			final double condNum){
	    this.indX=indX;
	    this.dimXorig=dimXorig;
	    this.xExpansion=xExpansion;
	    this.conditionsLeft=conditionsLeft;
	    this.condNum=condNum;
	}


	/** Single condition elimination.
	 */
	static EliminateVector expandSingleElimination(
	    final double [][] conditions,
	    final double condMaxToFail){

	    int nprev=-1,rowSel=-1,colSel=-1;
	    double maxEl=0;
	    for(int i=0;i<conditions.length;i++){
		final double [] ci=conditions[i];
		if(nprev>=0 && nprev!=ci.length) throw new IllegalArgumentException();
		nprev=ci.length;
		for(int j=0;j<ci.length;j++){
		    if(Math.abs(ci[j]) > Math.abs(maxEl)){
			maxEl=ci[j];
			rowSel=i;
			colSel=j;
		    }
		}
	    }
	    if(nprev<0 || colSel<0 || !(Math.abs(maxEl)>condMaxToFail)) return null;

	    // create x[colSel] expansion
	    final double [] xExpansion=TMatr.getVectorScaled(conditions[rowSel],-1/conditions[rowSel][colSel]);
	    xExpansion[colSel]=0;

	    // remaining
	    final double [][] conditionsNew=new double [conditions.length-1][];
	    for(int i=0,o=0;i<conditions.length;i++){
		if(i!=rowSel){
		    final double [] r=conditionsNew[o++]=conditions[i].clone();
		    r[colSel]=0;
		    TMatr.sumP1(r,xExpansion,conditions[i][colSel]);
		}
	    }

	    return new EliminateVector(colSel,nprev,xExpansion,conditionsNew,Math.abs(maxEl));
	}
    } // END OF class EliminateVector



    static EliminateLinearConstraints_HomegrownLUFactorization testOne(
	final String name,
	final int nX,
	final double [][] eqs,
	final java.util.Random r,
	final double eps){

	final double cond=1e-12;
	final var t=new EliminateLinearConstraints_HomegrownLUFactorization(nX,eqs,cond);

	System.err.println(name+":"+(t.elim.stream().map(e->"(col="+e.indX+" condNum="+e.condNum+")").toList()));
	for(int ts=0;ts<100;ts++){
	    final double [] xnew=new RandomSample(-1,1,t.dimXnew,r).sample,
		xorig=t.fromXnewToXorig(xnew),
		matr=t.fromXnewToXorigAsMatrix(),
		xorigMatr=TMatr.mul_MA_x_MB(t.dimXorig,matr,t.dimXnew,xnew,1);
	    UnitTests.checkDoubleArrayRel1(xnew,"xnew",t.fromXorigToXnew(xorig),"xnew(xorig)",eps);
	    UnitTests.checkDoubleArrayRel1(xorig,"xorig",t.fromXnewToXorig(t.fromXorigToXnew(xorig)),"xorig(xnew)",eps);
	    UnitTests.checkDoubleArrayRel1(xorig,"xorig",xorigMatr,"xorigMatr",eps);

	    // check every condition is satisfied
	    for(int j=0;j<eqs.length;j++){
		UnitTests.checkDoubleRel1(TMatr.mulVV(eqs[j],xorig),"eqs*xorig",0,"0",eps);
	    }
	}
	return t;
    }

    static double [][] getDegeneratedSample(final int nEq,
					    final int nX,
					    final java.util.Random r,
					    final double [][] eqs){

	final int nToAddEq=r.nextInt(2*nEq)+1;
	final int [][] xInd=new int [nToAddEq][nEq];
	final double [][] xCoef=new double [nToAddEq][nEq];
	for(int i=0;i<nToAddEq;i++){
	    for(int j=0;j<nEq;j++) {
		xInd[i][j]=r.nextInt(eqs.length);
		xCoef[i][j]=5.2*(1-2*r.nextDouble());
	    }
	}

	final double [][] eqsDeg=java.util.Arrays.copyOf(eqs,eqs.length+nToAddEq);
	for(int i=0;i<nToAddEq;i++){
	    final double [] e=eqsDeg[i+eqs.length]=new double [nX];
	    for(int j=0;j<nEq;j++){
		TMatr.sumP1(e,eqs[xInd[i][j]],xCoef[i][j]);
	    }
	}
	return eqsDeg;
    }


    static void testBasic(final int nEq,final int nX,final java.util.Random r,final double eps){
	final double [][] eqs=new double [nEq][];
	for(int j=0;j<eqs.length;j++){
	    eqs[j]=new RandomSample(-1,1,nX,r).sample;
	}
	final var t=testOne("nonDeg",nX,eqs,r,eps);
	if(t.nEliminated!=eqs.length) throw new RuntimeException("dimension mismatch");

	// test degenerated
	for(int tg=0;tg<10;tg++){
	    final double [][] eqsDeg=getDegeneratedSample(nEq,nX,r,eqs);
	    final var tDeg=testOne("deg",nX,eqsDeg,r,eps);
	    if(tDeg.nEliminated!=eqs.length) throw new RuntimeException("dimension mismatch");
	}

	System.err.println("EliminateLinearConstraints_HomegrownLUFactorization:testBasic: tests OK nEq="+nEq+" nX="+nX+" eps="+eps);
    }

    static void runMultiTest(final String name,
			     final int nTest,
			     final java.util.Random r,
			     final double eps){
	System.err.println(name+":EliminateLinearConstraints_HomegrownLUFactorization:starting tests; ver=$Id: EliminateLinearConstraints_HomegrownLUFactorization.java,v 1.35 2024/05/12 22:45:04 mal Exp $");
	final long t0=System.currentTimeMillis();
	for(int n=1;n<=nTest;n++){
	    for(var a:new int [][]{{2,5},{4,10},{8,10},{16,20},{7,20}}){
		testBasic(a[0],a[1],r,eps);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println(name+":EliminateLinearConstraints_HomegrownLUFactorization: tests OK done in "+(t1-t0)+" milliseconds. nTest="+nTest+" ver=$Id: EliminateLinearConstraints_HomegrownLUFactorization.java,v 1.35 2024/05/12 22:45:04 mal Exp $");
    }



    /** A unit test. */
    public static void main(String [] args){
	runMultiTest("lin",100,new java.util.Random(123),1e-5);
    }
}
