package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */

/**
 * Given an input {@literal |psi>} state
 * adjust it to be a subject to the constraint: {@literal <psi|pI|psi>/<psi|I|psi> - <psi|pJ|psi>/<psi|J|psi> =0},
 * where {@literal I} an {@literal J} are two positive measures.
 * Variations along all {@literal tryVariations} directions are tried and quartic
 * equation is solved for every try using {@link PolynomialRootsElementary#root4} method,
 * all versions of adjsuted state are saved to {@link #psiConstrained} to select the best one later.
 * @version $Id: FindPsiConstrainedSameAverage.java,v 1.19 2022/06/27 07:19:16 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
class FindPsiConstrainedSameAverage {

    private static final double PSINORMMIN=1e-6;

    final java.util.ArrayList<PsiConstrained> psiConstrained=new java.util.ArrayList<>();
    final double orig_constraint,orig_bpsi[];

    FindPsiConstrainedSameAverage(final double [] bpsi,
				  final double [] bQQI,
				  final double [] bQQpI,
				  final double [] bQQJ,
				  final double [] bQQpJ,
				  final java.util.List<double []>tryVariations,
				  final double eps){
	final int n=bpsi.length;
	this.orig_bpsi=bpsi;
	final double
	    cpI=TMatr.getAMatrA(bpsi,bQQpI,n),
	    cI=TMatr.getAMatrA(bpsi,bQQI,n),
	    cpJ=TMatr.getAMatrA(bpsi,bQQpJ,n),
	    cJ=TMatr.getAMatrA(bpsi,bQQJ,n);
	orig_constraint=cpI/cI-cpJ/cJ;
	if(eps>0){
	    UnitTests.checkDoubleRel(TMatr.mulVV(bpsi,bpsi),"<bpsi|bpsi>",1,"1",eps);
	}

	for(int j=0,nVars=tryVariations.size();j<nVars;j++){
	    final double [] varbpsi=tryVariations.get(j);
	    if(eps>0){
		UnitTests.checkDoubleRel(TMatr.mulVV(varbpsi,varbpsi),"<varbpsi|varbpsi>",1,"1",eps);
	    }
	    if(varbpsi.length!=n) throw new IllegalArgumentException();
	    if(!(1-PSINORMMIN > Math.abs(TMatr.mulVV(bpsi,varbpsi)))) continue;
	    final double []
		pol_pI={cpI,2*TMatr.getAMatrB(bpsi,varbpsi,bQQpI,n),TMatr.getAMatrA(varbpsi,bQQpI,n)},
		pol_I={cI,2*TMatr.getAMatrB(bpsi,varbpsi,bQQI,n),TMatr.getAMatrA(varbpsi,bQQI,n)},
		pol_pJ={cpJ,2*TMatr.getAMatrB(bpsi,varbpsi,bQQpJ,n),TMatr.getAMatrA(varbpsi,bQQpJ,n)},
		pol_J={cJ,2*TMatr.getAMatrB(bpsi,varbpsi,bQQJ,n),TMatr.getAMatrA(varbpsi,bQQJ,n)};
	    // solve
	    // 0=pol_pI(x)/pol_I(x) - pol_pJ(x)/pol_J(x)
	    // obtain quartic equation
	    // 0=P4*x^4+P3*x^3+P2*x^2+P1*x+P0
	    double [] coefs=TMatr.sumAcB(
		Polynomials.mult2Pol(pol_pI,pol_J),
		-1,
		Polynomials.mult2Pol(pol_pJ,pol_I));
	    final double MINHIGHORDERCOEF=1e-5;
	    // check the highest order coefficient is not too small
	    for(int i=coefs.length;--i>=2;){
		//System.err.println("i="+i+" coefs[i]="+coefs[i]);
		if(!(Math.abs(coefs[i])>MINHIGHORDERCOEF)){
		    //System.err.println("remove "+coefs[i]+" coefs="+java.util.Arrays.toString(coefs));
		    coefs=java.util.Arrays.copyOf(coefs,i);
		} else break;
	    }
	    final double MINCONSTRAINTERR=1e-8;
	    // solve quartic equation
	    final var r=PolynomialRootsElementary.ROOTSFINDER.findRoots(coefs);
	    if(r.nroots!=coefs.length-1) continue; // failed to find root
	    for(int k=0;k<r.nroots;k++){
		if(!(r.im_root[k]==0)) continue; // skip complex root
		// some polynomial coefficients can be very small, this gives a rounding error
		final double rr=r.re_root[k];
		if(!(MINCONSTRAINTERR>=
		     Math.abs(Polynomials.getD0(pol_pI,rr)/Polynomials.getD0(pol_I,rr)
			      -Polynomials.getD0(pol_pJ,rr)/Polynomials.getD0(pol_J,rr)))){
		    continue; // skip root
		}


		final double [] bpsiC=TMatr.sumAcB(bpsi,rr,varbpsi);
		// normalize
		final double nrC=Math.sqrt(TMatr.mulVV(bpsiC,bpsiC));
		if(!(nrC>PSINORMMIN)) continue; // skip degenerated
		TMatr.scaleVectorInPlace(bpsiC,1/nrC);
		psiConstrained.add(new PsiConstrained(bpsiC,1000+j*10+k /* some number just to debug */));
		if(eps>0){
		    //System.err.println("ROOT:rr["+k+"]="+rr+" pI="+(Polynomials.getD0(pol_pI,rr)/Polynomials.getD0(pol_I,rr))+" pJ="+(Polynomials.getD0(pol_pJ,rr)/Polynomials.getD0(pol_J,rr))+" p4(rr)="+Polynomials.getD0(coefs,rr)+" coefs="+java.util.Arrays.toString(coefs));
		    UnitTests.checkDoubleRel(TMatr.mulVV(bpsiC,bpsiC),"<bpsiC|bpsiC>",1,"1",eps);
		    UnitTests.checkDoubleRel(getConstraint(bpsiC,bQQI,bQQpI,bQQJ,bQQpJ),"getConstraint(bpsiC)",0,"0",eps);
		}
	    }
	}

	if(eps>0){
	    System.err.println("FindPsiConstrainedSameAverage: found "+psiConstrained.size()+" constraint adjusted psi n="+n+" eps="+eps+" types="+psiConstrained.stream().map(x->"(adjtype="+x.adjustment_type+",I/J="+(TMatr.getAMatrA(x.bpsi,bQQI,n)/TMatr.getAMatrA(x.bpsi,bQQJ,n))+")").toList());
	    for(PsiConstrained bpsiC:psiConstrained){
		UnitTests.checkDoubleRel(TMatr.mulVV(bpsiC.bpsi,bpsiC.bpsi),"<bpsiC|bpsiC>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.mulVV(bpsi,bpsi),"<bpsi|bpsi>",1,"1",eps);
		UnitTests.checkDoubleRel(getConstraint(bpsiC.bpsi,bQQI,bQQpI,bQQJ,bQQpJ),"getConstraint(bpsiC)",0,"0",eps);
	    }
	}
    }


    /** A variation of {@literal <psi|pI|psi>/<psi|I|psi> - <psi|pJ|psi>/<psi|J|psi> =0}.
     */
    static double [] getAveragesVariation_bQQc(
	final double [] bpsi,
	final double [] bQQI,
	final double [] bQQpI,
	final double [] bQQJ,
	final double [] bQQpJ,
	final double eps){

	final int n=bpsi.length;
	final double pI=TMatr.getAMatrA(bpsi,bQQpI,n),
	    I=TMatr.getAMatrA(bpsi,bQQI,n),
	    pii=pI/I/I,
	    pJ=TMatr.getAMatrA(bpsi,bQQpJ,n),
	    J=TMatr.getAMatrA(bpsi,bQQJ,n),
	    pjj=pJ/J/J;
	if(eps>0){
	    UnitTests.checkDoubleRel(TMatr.mulVV(bpsi,bpsi),"<bpsi|bpsi>",1,"1",eps);
	    // should never be called with constraint unsatisfied
	    UnitTests.checkDoubleRel(pI/I-pJ/J,"pI/I-pJ/J",0,"0",eps);
	}
	final double [] bQQc=new double [n*n];
	for(int k=0;k<bQQc.length;k++){
	    bQQc[k]=(bQQpI[k]/I-pii*bQQI[k])-(bQQpJ[k]/J-pjj*bQQJ[k]);
	}
	return bQQc;
    }

    static double getConstraint(final double [] bpsi,
				final double [] bQQI,
				final double [] bQQpI,
				final double [] bQQJ,
				final double [] bQQpJ){
	final int n=bpsi.length;
	return TMatr.getAMatrA(bpsi,bQQpI,n)/TMatr.getAMatrA(bpsi,bQQI,n)
	    - TMatr.getAMatrA(bpsi,bQQpJ,n)/TMatr.getAMatrA(bpsi,bQQJ,n);
    }

}
