package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */

/**
 * Given an input {@literal |psi>} state
 * adjust it to be a subject to the constraint: {@literal <psi|C|psi>=0}.
 * Several strategies are used and in {@link #psiConstrained}
 * several versions of adjsuted state are saved.
 * A helper class.
 * @version $Id: FindPsiConstrainedSingleQuadratic0.java,v 1.29 2024/05/13 07:10:29 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
class FindPsiConstrainedSingleQuadratic0 {


    final java.util.ArrayList<PsiConstrained> psiConstrained=new java.util.ArrayList<>();
    final double orig_constraint,orig_bpsi[];
    final boolean flag_solution_exists;

    FindPsiConstrainedSingleQuadratic0(final double [] bpsi,
				       final double [] bQQc,
				       final double [] ev_bQQc,
				       final double eps){
	final int n=bpsi.length;
	this.orig_constraint=TMatr.getAMatrA(bpsi,bQQc,n);
	this.orig_bpsi=bpsi;
	this.flag_solution_exists=ev_bQQc[0]<0 && ev_bQQc[n-1]>0;
	if(eps>0){
	    UnitTests.checkDoubleRel(TMatr.mulVV(bpsi,bpsi),"<bpsi|bpsi>",1,"1",eps);
	}
	for(int trytype=0;trytype<=1 && flag_solution_exists;trytype++){
	    if(trytype==0){
		// TRY: quadratic equation
		addPsiQuadraticEquationConstraint(psiConstrained,bpsi,bQQc,eps);
	    } else {
		// TRY: ev_bQQc basis expansion
		if(eps>0){
		    final double [] bbpsi=EVXData.convertPsiFromQtoB_Basis(n,bpsi,ev_bQQc,TMatr.getUnitMatrix(n));
		    UnitTests.checkDouble(TMatr.mulV2Vn1(bbpsi,ev_bQQc),"<bbpsi|bQQc|bbpsi>",orig_constraint,"orig_constraint",eps);
		}
		addPsiAdjustEVSigned(psiConstrained,bpsi,ev_bQQc,eps);
	    }
	}

	if(eps>0){
	    System.err.println("FindPsiConstrainedSingleQuadratic0: found "+psiConstrained.size()+" constraint adjusted psi n="+n+" eps="+eps+" types="+psiConstrained.stream().mapToInt(x->x.adjustment_type).boxed().toList());
	    for(PsiConstrained bpsiC:psiConstrained){
		UnitTests.checkDoubleRel(TMatr.getAMatrA(bpsiC.bpsi,bQQc,n),"<bpsiC|bQQc|bpsiC>",0,"0",eps);
		UnitTests.checkDoubleRel(TMatr.mulVV(bpsiC.bpsi,bpsiC.bpsi),"<bpsiC|bpsiC>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.mulVV(bpsi,bpsi),"<bpsi|bpsi>",1,"1",eps);
	    }
	}
    }


    private static final double PSINORMMIN=1e-7;

    /** Quadratic equation to solve the constraint.
     */
    private static void addPsiQuadraticEquationConstraint(
	java.util.ArrayList<PsiConstrained> storePsi,
	final double [] bpsi,
	final double [] bQQc,
	final double eps){

	final int n=bpsi.length;

	// TRY: Quadratic equation
	final double [] Cbpsi=TMatr.getMatr_x_A(n,bQQc,bpsi);
	// Remove bpsi projection upfront to simplify the formulas below
	TMatr.sumP1(Cbpsi,bpsi,-TMatr.mulVV(Cbpsi,bpsi));
	final double nrC=Math.sqrt(TMatr.mulVV(Cbpsi,Cbpsi));
	//System.err.println("bpsi="+java.util.Arrays.toString(bpsi)+" rem1="+java.util.Arrays.toString(Cbpsi)+" nrC="+nrC);
	if(!(nrC>PSINORMMIN)) return; // skip degenerated
	TMatr.scaleVectorInPlace(Cbpsi,1/nrC);
	if(eps>0){
	    UnitTests.checkDoubleRel(TMatr.mulVV(bpsi,bpsi),"<bpsi|bpsi>",1,"1",eps);
	    UnitTests.checkDoubleRel(TMatr.mulVV(bpsi,Cbpsi),"<bpsi|Cbpsi>",0,"0",eps);
	    UnitTests.checkDoubleRel(TMatr.mulVV(Cbpsi,Cbpsi),"<Cbpsi|Cbpsi>",1,"1",eps);
	}
	// quadratic equation coefficients
	final double
	    a=TMatr.getAMatrA(Cbpsi,bQQc,n),
	    b=2*TMatr.getAMatrB(Cbpsi,bpsi,bQQc,n),
	    c=TMatr.getAMatrA(bpsi,bQQc,n);
	if(a!=0){
	    final PolynomialRootsFinder.PRoots r=PolynomialRootsElementary.ROOTSFINDER.findRoots(new double []{c,b,a});
	    if(r.nroots==2 && Math.abs(r.im_root[0])<=0 && Math.abs(r.im_root[1])<=0){
		final double [] psi1=TMatr.sumAcB(bpsi,r.re_root[0],Cbpsi),
		    psi2=TMatr.sumAcB(bpsi,r.re_root[1],Cbpsi);
		final double nr1=TMatr.mulVV(psi1,psi1);
		if(nr1>0){
		    TMatr.scaleVectorInPlace(psi1,1/Math.sqrt(nr1));
		    storePsi.add(new PsiConstrained(psi1,-2));
		}
		final double nr2=TMatr.mulVV(psi2,psi2);
		if(nr2>0){
		    TMatr.scaleVectorInPlace(psi2,1/Math.sqrt(nr2));
		    storePsi.add(new PsiConstrained(psi2,-2));
		}
		if(eps>0){
		    System.err.println("FindPsiConstrainedSingleQuadratic0: found EV[-2] two quadratic equation constraint adjusted psi n="+n+" eps="+eps);
		}
	    }
	} else {
	    final PolynomialRootsFinder.PRoots r=PolynomialRootsElementary.ROOTSFINDER.findRoots(new double []{c,b});
	    if(r.nroots==1){
		final double [] psi=TMatr.sumAcB(bpsi,r.re_root[0],Cbpsi);
		final double nrP=TMatr.mulVV(psi,psi);
		if(!(nrP>0)) return;
		TMatr.scaleVectorInPlace(psi,1/Math.sqrt(nrP));
		storePsi.add(new PsiConstrained(psi,-3));
		if(eps>0){
		    System.err.println("FindPsiConstrainedSingleQuadratic0: found EV[-3] single quadratic equation constraint adjusted psi n="+n+" eps="+eps);
		}
	    }
	}
    }


/** Adjust eigenstates projections to satisfy the constraint.
 */
    private void addPsiAdjustEVSigned(final java.util.ArrayList<PsiConstrained> storePsi,
				      final double [] bpsi,
				      final double [] ev_bQQc,
				      final double eps){
	final int n=bpsi.length;
	// convert to bQQc basis to simplify algebra
	final double [] _bbpsi=EVXData.convertPsiFromQtoB_Basis(n,bpsi,ev_bQQc,TMatr.getUnitMatrix(n));

	for(int kadjust=-1;kadjust<n;kadjust++){
	    final double [] bbpsi=_bbpsi.clone();
	    if(kadjust<0) {
		// TRY: adjust all +/- eigenvector projections proportionally
		double swplus=0,swminus=0;
		for(int k=0;k<n;k++){
		    final double bqqclam=ev_bQQc[k],w=Math.pow(bbpsi[k],2);
		    if(bqqclam>0) {swplus+=w*bqqclam;}
		    else if(bqqclam<0) {swminus-=w*bqqclam;}
		}
		if(!(swplus>0 && swminus>0)) continue; // no solution
		// adjust +/- projections to have <bpsiC|bQQc|bpsiC>=0
		final double factplus=Math.sqrt(swminus/(swplus+swminus)),factminus=Math.sqrt(swplus/(swplus+swminus));
		for(int k=0;k<n;k++){
		    if(ev_bQQc[k]>0) {bbpsi[k]*=factplus;}
		    else if(ev_bQQc[k]<0) {bbpsi[k]*=factminus;}
		}
	    } else {
		// TRY: adjust individual eigenvector
		final double orig_constraint=TMatr.mulV2Vn1(bbpsi,ev_bQQc);
		if(!Double.isFinite(ev_bQQc[kadjust]) ||
		   !( // check for a failure
		       (ev_bQQc[kadjust]>0 && orig_constraint<0)
		       || (ev_bQQc[kadjust]<0 && orig_constraint>0)
		      )) continue; // no solution
		// adjust the single projection to have <bpsiC|bQQc|bpsiC>=0
		bbpsi[kadjust]=0;
		final double fn2=-TMatr.mulV2Vn1(bbpsi,ev_bQQc)/ev_bQQc[kadjust];
		if(fn2>0) bbpsi[kadjust]=Math.sqrt(fn2); else continue;
	    }
	    // normalize on success
	    if(TMatr.mulVV(bbpsi,bbpsi)>0){
		// convert back to bQQi basis
		final double [] bpsiC=EVXData.convertPsiFromBtoQ_Basis(n,bbpsi,ev_bQQc);
		final double nr2=TMatr.mulVV(bpsiC,bpsiC);
		if(nr2>0){
		    TMatr.scaleVectorInPlace(bpsiC,1/Math.sqrt(nr2));
		    storePsi.add(new PsiConstrained(bpsiC,kadjust));
		    if(eps>0){
			System.err.println("FindPsiConstrainedSingleQuadratic0: found EV["+kadjust+"] constraint adjusted psi n="+n+" eps="+eps);
		    }
		}
	    }
	}
    }

}
