package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Hermite(E) polynomials basis,
 * using  \( H_k=x*H_{k-1}-(k-1)H_{k-2} \).
 *  @version $Id: HermiteE.java,v 1.72 2022/07/16 10:04:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class HermiteE implements BasisPolynomials {


/** Calculates n-th HermiteE polynomial
 * using  \( H_n=x*H_{n-1}-(n-1)H_{n-2} \).
 * @param n H index.
 * @param Hnm1 Input \( H_{n-1} \).
 * @param Hnm2 Input \( H_{n-2} \).
 * @param flag_type2 Whether to use type 2 polynomials.
 * @param PBASIS The basis in which to calculte.
 * @return \( H_{n} \) polynomial, array dimension n+1.
 */
public double [] getNext(final int n,
			 final double [] Hnm1,final double [] Hnm2,
			 final boolean flag_type2,
			 final BasisFunctionsCalculatable PBASIS)
{
    if(n<0) throw new IllegalArgumentException();
    if(n<=0) return PBASIS.getONE();
    if(Hnm1.length!=n) throw new IllegalArgumentException();
    final double [] res=PBASIS.getaxbP(Hnm1,1,0);
    if(n<=1) return res;
    if(Hnm2.length!=n-1) throw new IllegalArgumentException();
    TMatr.sumP1(res,Hnm2,-(!flag_type2?n-1:n));
    return res;
}

public double [][] getBasisFunctionsOnPolynomialArgument(
    final int numQl,
    final double [] pol,
    final BasisFunctionsCalculatable PBASIS){

    final double [][] res=new double [numQl][];
    double [] fprev={},f=res[0]=PBASIS.getONE();
    for(int n=1;n<numQl;n++)
    {
	final double [] fmul=PBASIS.mult2Pol(pol,f);
	TMatr.sumP1(fmul,fprev,-(n-1));
	fprev=f;f=res[n]=fmul;
    }
    return res;
}

/**
 * Calculation of the value of n-Hermite polynomial at given x.
 * @param n H index.
 * @param x Argument.
 * @return The value of the Hermite polynomial Hn at x.
 */
public double calculate(final int n,final double x)
{
    if(n<0) throw new IllegalArgumentException("n="+n);
    double b=0,s=1;
    for(int i=1;i<=n;++i)
    {
	s=(1-i)*b+(x)*(b=s);
    }
    return s;
}

public double [] getMomentsFromSample(final int nmoments,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [nmoments];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=x[k],ww=w!=null?w[k]:1;
	double b=0,s=1;
	for(int i=0;i<nmoments;++i)
	{
	    if(i<=0) m[i]+=s*ww;
	    else {s=(1-i)*b+(xx)*(b=s);m[i]+=s*ww;}
	}
    }
    return m;
}

/** Calculates \( sum_k H_k(x) Hcoefs[k] \).
 * This version is copied from http://www.alglib.net/.
 * @param Hcoefs Coefficients.
 * @param x Point to estimate the sum.
 * @return The sum.
 */
public double getD0(final double [] Hcoefs,final double x)
{
    final int n=Hcoefs.length-1;
    if(n<0) throw new IllegalArgumentException("n="+n);
    double result = 0,b1 = 0,b2 = 0;

    for(int i=n; i>=0; i--)
    {
        result = (x*b1-(i+1)*b2)+Hcoefs[i];
        b2 = b1;
        b1 = result;
    }
    return result;

}

/** Calculates \( d/dx \sum_k H_k(x) Hcoefs[k] \).
 * @param Hcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double getD1(final double [] Hcoefs,final double x)
{
    final int n=Hcoefs.length-1;
    if(n<0) throw new IllegalArgumentException("n="+n);
    double result = 0,b1 = 0,b2 = 0;

    for(int i=n-1; i>=0; i--)
    {
        result = (x*b1-(i+1)*b2)+(i+1)*Hcoefs[i+1];
        b2 = b1;
        b1 = result;
    }
    return result;
}


/** Calculates \( d^2/dx^2 \sum_k H_k(x) Hcoefs[k] \).
 * @param Hcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The second derivative.
 */
public double getD2(final double [] Hcoefs,final double x)
{
    final int n=Hcoefs.length-1;
    if(n<0) throw new IllegalArgumentException("n="+n);
    double result = 0,b1 = 0,b2 = 0;

    for(int i=n-2; i>=0; i--)
    {
        result = (x*b1-(i+1)*b2)+((i+1)*(i+2))*Hcoefs[i+2];
        b2 = b1;
        b1 = result;
    }
    return result;
}



/** Differentiate a function in Hermite basis,
 *  F(x) - a polynomial of n-th order.
 *  See Gradstein Ryzik, 1963, formulae 8.952, page 1047.
 *  @param Hcoefs Input \( F(x)=\sum_k H_k(x)*Hcoefs[k] \).
 *  @return  \( dF/dx=\sum_k H'_k(x)*Hcoefs[k] \).
 */
public double [] getDifferentiated(final double [] Hcoefs)
{
    final int n=Hcoefs.length-1;
    // He_k(x)'=k He_{k-1}(x)
    final double [] d=new double [n];
    for(int k=n;--k>=0;)
    {
	d[k]=(k+1)*Hcoefs[k+1];
    }
    return d;
}

/** Integrate in Hermite basis,
 *  See Gradstein Ryzik, 1963, formulae 8.952, page 1047.
 */
public double [] getIntegrated(final double [] Hcoefs)
{
    final int n=Hcoefs.length-1;
    if(n<0) throw new IllegalArgumentException("n="+n);
    final double [] i=new double [n+2];
    // \int He_k(x) dx =He_{k+1}(x)/(k+1)
    for(int k=n;k>=0;--k)
    {
	i[k+1]=Hcoefs[k]/(k+1);
    }
    i[0]-=getD0(i,0); // normalize integral to f(0)=0
    return i;
}



/** Expand \( H_l(z)H_m(z)=\sum_{k=0}^{l+m} s_k H_k(z) \),
 * formulae from
 * L. Carlitz. Monatshefte fur Mathematik. Volume 66 issue 5 1962.
 * The product of several Hermite or Laguerre polynomials.
 * And see the refences to original work by N. Nielsen therein.
 * This formulae is also available from
 * Prudnikov, Brychkov, Marichev, volume 2, 2003, formulae 4.5.1.11 page 569.
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] s)
{
    final int e=Math.min(l,m),mpl=m+l;
    java.util.Arrays.fill(s,0,mpl+1,0);
    long cl=1,cm=1,cr=1;
    for(int k=0;k<=e;)
    {
	s[mpl-2*k]=cl*((double)cr)*cm;
	cl*=(l-k);
	cm*=(m-k);
	++k;
	cl/=k;
	cm/=k;
	cr*=k;
    }
}


/** Multiply p by ma*x+mb.
 * Calculations in Hermite basis.
 * Do not mistake with a and b recurrence coefficiemts
 * of {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getAB(double [][]) getAB()}.
 * @param p Input polynomial.
 */
@Override
public double [] getaxbP(final double [] p,final double ma,final double mb)
{
    final int n=p.length-1;
    final double [] res=new double [n+2];
    double r=p[n];
    res[n]=res[n+1]=0;
    for(int k=n;k>=0;k--)
    {
	final double ar=ma*r;
	res[k+1]+=ar;
	res[k]+=mb*r;
	if(k>0)
	{
	    final double t=p[k-1];
	    res[k-1]=k*ar;
	    r=t;
	}
    }
    return res;
}


/** Divide  p by (x-x0), returns quotient and remainder.
 * Calculations in Hermite basis.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return Quotient and remainder.
 */
@Override
public SDivRes sdiv1(final double [] p,final double x0)
{
    final int n=p.length-1;
    final double [] res=new double [n];
    double r=p[n];
    //System.err.println("pi="+java.util.Arrays.toString(p));
    if(n>0) res[n-1]=p[n-1];
    for(int k=n;--k>=0;)
    {
	final double c=r,b=(res[k]+=c*(x0));
	if(k>0)res[k-1]=p[k-1]-(k)*c;
	res[k]=c;
	r=b;
    }
    //System.err.println("p="+java.util.Arrays.toString(p));
    //System.err.println("res="+java.util.Arrays.toString(res));
    return new SDivRes(res,new double []{r});
}


/** Calculates H_n(a(x+x0)) using 3 term recurrence.
 * Calculations in Hermite basis.
 * The \( b_k^{n} \) are defined as:
 * \( H_n(a(x+x0))=\sum_{k=0}^{k=n}b_k^{n} H_k(x) \)
 * The method should be applied consequently to get \( b_k^{n} \) for n=0,1,2,3,...
 * @param n The max dimension of b.
 * @param b On input: \( b_k^{n-1} \), on output \( b_k^{n} \).
 * @param q On input: \( b_k^{n-2} \), on output: \( b_k^{n-1} \).
 */
@Override
public void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    if(n<=0)
    {
	b[n]=1;
	return;
    }
    final int n1=n-1;
    double r=b[n1];
    b[n1]=b[n]=0;
    for(int k=n;--k>=0;)
    {
	final double ar=r*a;
	b[k+1]+=ar;
	b[k]+=x0*ar;
	if(k<n1) b[k]-=n1*q[k];
	q[k]=r;
	if(k>0)
	{
	    final double bb=b[k-1];
	    b[k-1]=k*ar;
	    r=bb;
	}
    }
}


/** Calculate {@literal <F|xHk>} moments for k=0..n-1 from {@literal <F|Hk>}, k=0..n
 * @param moments {@literal <F|Hk>}.
 * @return {@literal <F|xHk>}.
 */
@Override
public double [] getXQkMomentsFromQkMoments(final double [] moments)
{
    final int n=moments.length-1;
    final double [] d=new double [n];
    for(int k=n;--k>=0;)
    {
	d[k]=(moments[k+1]+(k<=0?0:k*moments[k-1]));
    }
    return d;
}


/** Calculates confederate matrix to solve
 * polynomial equations in Hermite basis directly.
 * We use the three term recurrence:
 * \[ H_k=x*H_{k-1}-(k-1)H_{k-2} \]
 * Use {@link PolynomialRootsConfederateMatrix#ABASISHERMITEE ABASISHERMITEE}
 * to calculate roots.
 */
public double [] getConfederateMatrix(final double [] coefs)
{
    final int n=coefs.length-1;
    final double [] matr=new double [n*n];
    for(int m=0;m<n;m++)
    {
	final int k=m+1;
	matr[m*n+m]=0;                // delta/alfa
	if(m>0) matr[(m-1)*n+m]=k-1;  // gamma/alfa
	if(m<n-1) matr[(m+1)*n+m]=1;  // 1/alfa
	else
	{
	    final double f=coefs[n];  // p[n]*alfa
	    for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	}
    }
    return matr;
}

/** A unit test.
 */
public static void main(String [] args)
{
    final double eps=1e-5;
    /* for(long initrandom=1;initrandom<100;initrandom++){try { */
    final long initrandom=88;
    new TestBasisPolynomials("HermiteE",new HermiteE(),new java.util.Random(initrandom),eps){
	@Override double [] getMonomialsMoments(final int n)
	{
	    final double [] pmom=new double [n];
	    for(int k=0;k<pmom.length;k++){  // monomials moments
		// Gradstein Ryzik, 1963, 3.461.{2,3}, page 351.
		if(k%2!=0) {pmom[k]=0;}
		else {pmom[k]=Math.sqrt(2*Math.PI)*Polynomials.fact2(k-1);}
	    }
	    return pmom;
	}
	@Override OrthogonalPolynomialsHermiteEBasis getOrtPol()
	{
	    return new OrthogonalPolynomialsHermiteEBasis();
	}
	@Override double [][] getExact()
	{
	    return new double [][]{
		{1},
		{0,1},
		{-1,0,1},
		{0,-3,0,1},
		{3,0,-6,0,1},
		{0,15,0,-10,0,1},
		{-15,0,45,0,-15,0,1},
		{0,-105,0,105,0,-21,0,1},
		{105,0,-420,0,210,0,-28,0,1},
		{0,945,0,-1260,0,378,0,-36,0,1},
		{-945,0,4725,0,-3150,0,630,0,-45,0,1}
	    };
	}
    }.runTests("$Id: HermiteE.java,v 1.72 2022/07/16 10:04:33 mal Exp $");
    /* }catch(final RuntimeException e) {
		e.printStackTrace(System.err);
		System.err.println("initrandom="+initrandom+" failed");
		continue;
	}
	System.err.println("initrandom="+initrandom+" OK");
	break;
    } */
}

} // END of class HermiteE
