package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */

/**
 * Find the {@literal <psi|I|psi> ->} max state
 * subject to the constraint: {@literal <psi|C|psi>=0}.
 * Here we use self--concordant approach to solve Rayleigh quotient
 * maximization with the quadratic constraint:{@literal <psi|C|psi>=0}.
 * There is an updated and improved version {@link IstatesConditionalV2}
 * and especially {@link IstatesConditionalSubspaceLinearConstraints}.
 * @see IstatesConditionalLocalized
 * @see IstatesConditionalV2
 * @see IstatesConditionalSubspaceLinearConstraints
 * @see MaximizeStandardDeviation
 * @version $Id: IstatesConditional.java,v 1.128 2024/05/13 22:16:15 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class IstatesConditional {
    private static final int N_iterations=12;


    public final double i_M,lambda_i_M,LagrangeMultiplier_M,wr0_M,orig_constraint;

    /** The state of maxI conditional to {@literal <psi|QQc|psi>=0}.
     */
    public final double [] psi_M;
    /** Whether a solution exists.
     */
    public final boolean flag_solution_exists;

    /** The {@literal |I|psi>=lambda |psi>} solution.
     */
    public final EVXData I;
public IstatesConditional(final int n,
			  final double [] QQt,final double [] QQi,
			  final double [] QQc,
			  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			  final double eps)
{
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);

    final double [] bQQi=TMatr.getDiagMatrix(n,I.ev),
	bQQc=getbQQcFromQQc(QQc), // may be overloaded for more complex models.
	// to check whether a solution exist
	ev_bQQc=M.EV.getEV(n,bQQc,true);
    // if the bQQc simultaneously has positive and negative eigenvalues the solution exists
    this.flag_solution_exists=ev_bQQc[0]<0 && ev_bQQc[n-1]>0;

    // conditional optimization
    final double [] bQQconditional=new double [n*n];
    GetMaxIState r=null;
    for(int it=0;it<N_iterations && flag_solution_exists;it++)
    {
	// construct a matrix with Lagrange multiplier r.m,
	// |bQQconditional| = |I|+r.m*|C|
	final double lm=r==null?(it>0?0:getInitialLagrangeMultiplier(n,QQt,QQi,QQc,M,bQQi,bQQc)):r.m;
	for(int k=bQQconditional.length;--k>=0;) {
	    bQQconditional[k]=bQQi[k]+lm*bQQc[k];
	}
	// we specifically converted to the basis of |I| to have unit right hand side
	final double [] ev=M.EV.getEV(n,bQQconditional,true);
	// find the state of maximal I. because of bpsi_M adjustment the vector may be different from max eigenvalue
	r=new GetMaxIState(n,ev,bQQc,bQQi,ev_bQQc,eps);
	if(r.bpsi_M==null){
	    r=null;
	    if(eps>0)
	    {
		// convergence diag
		System.err.println("IstatesConditional: it="+it+" FAILURE flag_solution_exists="+flag_solution_exists);
	    }
	    break;
	}
	if(eps>0)
	{
	    // convergence diag
	    System.err.println("IstatesConditional: it="+it+" r.i="+r.i+" r.m="+r.m+" constraint="+TMatr.getAMatrA(r.bpsi_M.bpsi,bQQc,n)+" =0 on success r.orig_constraint="+r.orig_constraint+"=0 on good convergence r.adjustment_type="+r.bpsi_M.adjustment_type+" flag_solution_exists="+flag_solution_exists);
	}
    }
    if(eps>0){
	if(r!=null) System.err.println("IstatesConditional: found max r.i="+r.i+" r.m="+r.m+" constraint="+TMatr.getAMatrA(r.bpsi_M.bpsi,bQQc,n)+" =0 on success r.orig_constraint="+r.orig_constraint+" r.adjustment_type="+r.bpsi_M.adjustment_type+" flag_solution_exists="+flag_solution_exists);
    }
    // bpsi_M is in the basis of bQQi eigenvectors I.ev, convert it to regular basis.
    this.psi_M=r==null?new double [n]:EVXData.convertPsiFromBtoQ_Basis(n,r.bpsi_M.bpsi,I.ev);
    // the <psi_M|psi0>^2 as a kind of "distance" to "now"
    this.wr0_M=r==null?Double.NaN:Math.pow(TMatr.getAMatrB(psi_M,I.psi0,QQt,n),2);
    this.LagrangeMultiplier_M=r==null?Double.NaN:r.m;
    this.i_M=TMatr.getAMatrA(psi_M,QQi,n);
    this.lambda_i_M=r==null?Double.NaN:r.lambda;
    if(r!=null){
	orig_constraint=r.orig_constraint;
	if(eps>0){
	    final double [] bx=TMatr.getMatr_x_A(n,bQQi,r.bpsi_M.bpsi),
		cx=TMatr.getMatr_x_A(n,bQQc,r.bpsi_M.bpsi);
	    TMatr.sumP1(bx,r.bpsi_M.bpsi,-TMatr.mulVV(bx,r.bpsi_M.bpsi));
	    UnitTests.checkDoubleRel(TMatr.mulVV(bx,r.bpsi_M.bpsi),"bx*bpsi_M",0,"0",eps);
	    UnitTests.checkDoubleRel(TMatr.mulVV(cx,r.bpsi_M.bpsi),"cx*bpsi_M",0,"0",eps);
	    UnitTests.checkDoubleRel(-TMatr.mulVV(bx,cx)/TMatr.mulVV(cx,cx),"-bx*cx/(cx*cx)",r.m,"r.m",eps);
	}
    } else orig_constraint=Double.NaN;

    if(eps>0){
	System.err.println("IstatesConditional: r.orig_constraint="+(r!=null?r.orig_constraint:Double.NaN)+" r.adjustment_type="+(r!=null?String.valueOf(r.bpsi_M.adjustment_type):"??")+" flag_solution_exists="+flag_solution_exists);
	if(r!=null){
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_M,QQt,n),"<psi_M|psi_M>",1,"1",eps);
	    UnitTests.checkDoubleRel(i_M,"i_M",r.i,"r.i",eps);
	}
    }
    if(eps>0 && flag_solution_exists){
	// compare with localized solution
	final IstatesConditionalLocalized localized=new IstatesConditionalLocalized(n,QQt,QQi,QQc,M,0*eps);
	IstatesConditionalLocalized.printDiag(
	    "IstatesConditional: test OK i_M="+i_M+" wr0_M="+wr0_M+" ",
	    localized,
	    "; compare: <psi_M|localized.psi_M>="+TMatr.getAMatrB(psi_M,localized.psi_M,QQt,n)+" DiffLoc i_M="+(i_M-localized.i_M)+
	    " flag_solution_exists="+flag_solution_exists+" eps="+eps);
    }
}

double getInitialLagrangeMultiplier(
    final int n,
    final double [] QQt,final double [] QQi,final double [] QQc,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
    final double [] bQQi,final double [] bQQc){
    if(false){
	// try localized as an initial state
	final IstatesConditionalLocalized localized=new IstatesConditionalLocalized(n,QQt,QQi,QQc,M,0);
	final double [] bpsi_M_localized=EVXData.convertPsiFromQtoB_Basis(n,localized.psi_M,I.ev,QQt);
	return GetMaxIState.getLagrangeMultiplier(bpsi_M_localized,bQQi,bQQc);
    }
    return 0;
}

double [] getbQQcFromQQc(final double [] QQc){
    return EVXData.convertMatrFromQtoB_Basis(I.neq,QQc,I.ev);
}




/** Try all eigenvectors available.
 * Select the one, providing maximal I
 * AFTER the constraint {@literal <psi|bQQc|psi>=0} is satisfied.
 */
private class GetMaxIState {
    final double i,orig_constraint;
    final PsiConstrained bpsi_M;
    final double m,lambda;
    GetMaxIState(final int n,
		 final double [] basis,
		 final double [] bQQc,
		 final double [] bQQi,
		 final double [] ev_bQQc,
		 final double eps)
{
    double iMax=0,orig_constraint=Double.NaN;
    PsiConstrained bpsi_M=null;
    int ksel=-1;
    for(int k=0;k<n;k++){
	final FindPsiConstrainedSingleQuadratic0 f=new FindPsiConstrainedSingleQuadratic0(EVXData.getPsiFromBasis(k,basis,n),bQQc,ev_bQQc,eps);
	for(PsiConstrained b: f.psiConstrained){
	    final double ib=TMatr.getAMatrA(b.bpsi,bQQi,n);
	    if(Double.isFinite(ib) && (bpsi_M==null || ib>=iMax)){
		ksel=k;
		iMax=ib;
		bpsi_M=b;
		orig_constraint=f.orig_constraint;
	    }
	}
    }
    this.i=iMax;
    this.orig_constraint=orig_constraint;
    this.bpsi_M=bpsi_M;
    this.lambda=ksel<0?Double.NaN:basis[ksel];
    if(eps>0){
	if(this.bpsi_M!=null){
	    System.err.println("IstatesConditional:GetMaxIState: selected ksel="+ksel+" i="+this.i+" orig_constraint="+orig_constraint+" adjustment_type="+this.bpsi_M.adjustment_type);
	} else {
	    System.err.println("IstatesConditional:GetMaxIState: no selected ksel="+ksel);
	}
    }
    this.m=getLagrangeMultiplier(this.bpsi_M.bpsi,bQQi,bQQc);
}

    private static double getLagrangeMultiplier(final double [] bpsi_M,final double [] bQQi,final double [] bQQc){
	if(bpsi_M!=null){
	    final int n=bpsi_M.length;
	    final double [] Cx=TMatr.getMatr_x_A(n,bQQc,bpsi_M),
		Ix=TMatr.getMatr_x_A(n,bQQi,bpsi_M);
	    final double cc=TMatr.mulVV(Cx,Cx);
	    // Lagrange multiplier
	    if(cc>0) return -TMatr.mulVV(Cx,Ix)/cc;
	}
	return 0;
    }
}

/** For unit test.
 */
public static void matchIstatesConditional(final IstatesConditional a,final IstatesConditional b,final double eps){
    EVXData.matchEVXData(a.I,b.I,false,eps);

    UnitTests.checkDoubleRel(a.i_M,"a.i_M",b.i_M,"b.i_M",eps);
    UnitTests.checkDoubleRel(a.LagrangeMultiplier_M,"a.LagrangeMultiplier_M",b.LagrangeMultiplier_M,"b.LagrangeMultiplier_M",eps);

    System.err.println("IstatesConditional:matchIstatesConditional test OK eps="+eps);
}

static void printDiag(IstatesConditional m){
    // print some diag
    System.err.println("IstatesConditional.main:I EV="+java.util.Arrays.toString(m.I.getEValues()));
    System.err.println("IstatesConditional.main:i_M="+m.i_M+" lambda_i_M="+m.lambda_i_M+" psi_M="+java.util.Arrays.toString(m.psi_M)+" flag_solution_exists="+m.flag_solution_exists+" n="+m.I.neq);
}

/** Unit test. */
public static void main(String [] args){
    final int [] testCounter=new int [3];
    TestOptimizationPsi.runMultiTest("IstatesConditional",15,600,
		 new java.util.Random(18301L),
		 1e-5,
		 testC-> {
			   final var pc=0.2; // some arbitrary value
			   final var m=new IstatesConditional(
			       testC.n,
			       testC.QQt,
			       testC.QQi,
			       TMatr.sumAcB(testC.QQpi,-pc,testC.QQi), // QQc
			       testC.M,
			       testC.eps);
			  // print some diag
			  printDiag(m);

			  // try V2 on the same data
			  final var m2=new IstatesConditionalV2(
			       testC.n,
			       testC.QQt,
			       testC.QQi,
			       TMatr.sumAcB(testC.QQpi,-pc,testC.QQi), // QQc
			       testC.M,
			       testC.eps*0/* disable internal tests for IstatesConditionalV2*/);
			  // print some diag
			  IstatesConditionalV2.printDiag(m2);

			  if(!m.flag_solution_exists) testCounter[2]++;
			  testCounter[0]++;

			  // potentantially may fail, but for this data it is mostly OK.
			  if(
			      // I diffrent
			      UnitTests.checkDoubleRelNoFail(m.i_M,m2.i_M,testC.eps) ||
			      // one program found and another not
			      m.flag_solution_exists!=m2.flag_solution_exists ||
			      // different psi_M found
			      (m2.flag_solution_exists && UnitTests.checkDoubleRelNoFail(Math.abs(TMatr.getAMatrB(m.psi_M,m2.psi_M,testC.QQt,testC.n)),1,testC.eps))
			     ){
			      testCounter[1]++;
			      System.err.println(
				  "!!! IstatesConditional is different from IstatesConditionalV2: m.i_M="+m.i_M+" m2.i_M="+m2.i_M+
				  " <m.psi_M|m.psi_M>="+TMatr.getAMatrA(m.psi_M,testC.QQt,testC.n)+
				  " <m2.psi_M|m2.psi_M>="+TMatr.getAMatrA(m2.psi_M,testC.QQt,testC.n)+
				  " <m.psi_M|m2.psi_M>="+TMatr.getAMatrB(m.psi_M,m2.psi_M,testC.QQt,testC.n)+
				  " n="+testC.n+" eps="+testC.eps);
			  }
			}
		 );
    System.err.println("IstatesConditional run "+testCounter[0]+" times."+
		       " Different from IstatesConditionalV2 "+testCounter[1]+" times."+
		       " Solution does not exist "+testCounter[2]+" times.");
}

}
