package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */

/**
 * Find the {@literal <psi|I|psi> -> max} state among LOCALIZED psi(x)=Q(x)G^{-1}Q(yloc)
 * subject to the constraint: {@literal <psi|C|psi>=0}.
 * @see IstatesConditionalV2
 * @see MaximizeStandardDeviation
 * @version $Id: IstatesConditionalLocalized.java,v 1.34 2022/08/08 20:10:48 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class IstatesConditionalLocalized {
    static final double EPS_IMROOT=1e-6;

    public final int n_real_roots;
    public final double y_M,i_M;

    /** The {@literal <psi_M|psi0>^2}, a kind of "distance" to "now";  {@code y} is actual distance.
     */
    public final double wr0_M;

    /** The state of {@literal max <psi|I|psi>} subject to {@literal <psi|C|psi>=0}.
     */
    public final double [] psi_M;

    /** All localized states corresponding to {@literal <psi|C|psi>=0}.
     */
    public final PriceMatchStorage [] matches;

    /** A class to store one {@literal <psi|C|psi>=0} localized at {@code y} match.
     */
    public static class PriceMatchStorage {
	public final double y,i,wr0;
	public final double [] psi_M;
	PriceMatchStorage(final double y,
			  final double i,
			  final double wr0,
			  final double [] psi_M){
	    this.y=y;
	    this.i=i;
	    this.wr0=wr0;
	    this.psi_M=psi_M;
	}
    } // END OF class PriceMatchStorage


public IstatesConditionalLocalized(final int n,
				   final double [] QQt,final double [] QQi,
				   final double [] QQc,
				   final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				   final double eps)
    {
    // by default use (rt)->rt.i
    final java.util.function.ToDoubleFunction<PriceMatchStorage> max_selector=(rt)->rt.i;

    final double [] IQQt=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQt),
	pol=M.B.getKK(n,TMatr.mulMMM(n,IQQt,QQc,IQQt));
    final PolynomialRootsFinder.PRoots u=M.getPolynomialRootsFinderInBasis().findRoots(pol);
    final java.util.ArrayList<PriceMatchStorage> pm=new java.util.ArrayList<>();
    final double [] psi0=TMatr.getMatr_x_A(n,IQQt,M.B.getX0Moments(n,M.getRadauRoot()));
    TMatr.scaleVectorInPlace(psi0,1/Math.sqrt(TMatr.getAMatrA(psi0,QQt,n)));
    for(int r=0;r<u.nroots;r++){
	if(Math.abs(u.im_root[r])<=EPS_IMROOT){
	    // real root, save localized at x=u.re_root[r] state:
	    final double [] psi=TMatr.getMatr_x_A(n,IQQt,M.B.getX0Moments(n,u.re_root[r]));
	    TMatr.scaleVectorInPlace(psi,1/Math.sqrt(TMatr.getAMatrA(psi,QQt,n)));
	    pm.add(new PriceMatchStorage(
		       u.re_root[r], // localization point
		       TMatr.getAMatrA(psi,QQi,n), // <psi|I|psi>
		       Math.pow(TMatr.getAMatrB(psi,psi0,QQt,n),2), // projection to |psi0>: <psi|psi0>^2
		       psi // the wavefunction localized at localization point
					 ));
	    if(eps>0){
		final PriceMatchStorage xr=pm.get(pm.size()-1);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(xr.psi_M,QQt,n),"<psi_M|psi_M>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(xr.psi_M,QQc,n),"<psi_M|QQc|psi_M>",0,"0",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(xr.psi_M,QQi,n),"<psi_M|I|psi_M>",xr.i,"xr.i",eps);
		System.err.println("IstatesConditionalLocalized: real root"+r+" y="+xr.y+" added; i="+xr.i);
	    }
	}
	if(eps>0){
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi0,QQt,n),"<psi0|psi0>",1,"1",eps);
	    System.err.println("IstatesConditionalLocalized: ROOT"+r+"=("+u.re_root[r]+","+u.im_root[r]+") pol(re root)="+M.B.getD0(pol,u.re_root[r])+" nroots="+u.nroots);
	}
    }
    pm.sort((o1,o2)->Double.compare(max_selector.applyAsDouble(o2),max_selector.applyAsDouble(o1)));
    this.matches=pm.toArray(PriceMatchStorage[]::new);
    // the max is at index 0
    final PriceMatchStorage pmselected=matches.length<=0?null:matches[0];
    if(pmselected!=null){
	psi_M=pmselected.psi_M;
	y_M=pmselected.y;
	i_M=pmselected.i;
	n_real_roots=pm.size();
	wr0_M=pmselected.wr0;
    } else {
	psi_M=new double [n];
	y_M=Double.NaN;
	i_M=Double.NaN;
	n_real_roots=0;
	wr0_M=0;
    }
    if(eps>0){
	System.err.println("IstatesConditionalLocalized: n="+n+" Nrealroots="+pm.size()+" Ntotalroots="+u.nroots+" max i="+(pmselected!=null?pmselected.i:Double.NaN)+" y="+(pmselected!=null?pmselected.y:Double.NaN)+" eps="+eps);
    }
}

static void printDiag(IstatesConditionalLocalized m){printDiag("",m,"");}
static void printDiag(String txtBefore,IstatesConditionalLocalized m,String txtAfter){
    // print some diag
    System.err.println(
	txtBefore+"IstatesConditionalLocalized.main:i_M="+m.i_M+
	" n_real_roots="+m.n_real_roots+" y_M="+m.y_M+" i_M="+m.i_M+" wr0_M="+m.wr0_M+
	" ALL MATCHES="+java.util.Arrays.stream(m.matches).map(o->"(y="+o.y+",i="+o.i+",wr0="+o.wr0+")").toList()+
	" psi_M="+java.util.Arrays.toString(m.psi_M)+" n="+m.psi_M.length+txtAfter);
}

/** Unit test. */
public static void main(String [] args){
    TestOptimizationPsi.runMultiTest("IstatesConditionalLocalized",20,600,
		 new java.util.Random(18301L),
		 1e-5,
		 testC-> {var m=new IstatesConditionalLocalized(
		                testC.n,
				testC.QQt,
				testC.QQi,
				TMatr.sumAcB(testC.QQpi,-0.2,testC.QQi), // QQc
				testC.M,
				testC.eps);
			  printDiag(m);
			  }
		 );
}


}
