package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */

/**
  * Find {@literal \( \left|\psi\right> \) } state such that
  * execution rate
  * {@literal \[
  \frac{\left<\psi|I| \psi\right>}{\left<\psi|\psi\right> }
  \xrightarrow[\psi]{\quad }\max
  \] }
  * is maximal, subject to constraints:
{@literal
\begin{align}
0&=\frac{\left<\psi|pI|\psi\right>}{\left<\psi|I|\psi\right>}
 -\,  \frac{\left<\psi|pJ|\psi\right>}{\left<\psi|J|\psi\right>} \\
1&=\left<\psi|\psi\right>
\end{align}
}.
As \( J \) typically time--averge is used,
 the constraint means time- and volume- averaged prices to be identical
in  {@literal \( \left|\psi\right> \)} state:
{@literal \(
\frac{\left<\psi|pI|\psi\right>}{\left<\psi|I|\psi\right>}
 =  \frac{\left<\psi|p|\psi\right>}{\left<\psi|\psi\right>}
\)}.
  * @see MaximizeStandardDeviation
  * @see IstatesConditionalV2
  * @version $Id: IstatesConditionalSameAverage.java,v 1.22 2022/07/19 08:36:32 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
  */
public class IstatesConditionalSameAverage {
    private static final int N_iterations=12;


    /** The  {@literal <psi_M|I|psi_M>}.
     */
    public final double i_M;

    /** Last value of eigenvalue, in case of algorithm convergence
     * matches to {@link #i_M}.
     */
    public final double lambda_i_M;

    /** Lagrange multiplier value. */
    public final double LagrangeMultiplier_M;

    /** The {@literal <psi_M|psi0>^2}, a kind of "distance" to "now".
     */
    public final double wr0_M;

    /** The state of maxI conditional to
     * {@literal <psi|QQpI|psi>/<psi|QQI|psi> - <psi|QQpJ|psi>/<psi|QQJ|psi> = 0}.
     */
    public final double [] psi_M;


    /** The {@literal \( \left|I|\psi\right>=\lambda \left|\psi\right> \)} solution.
     */
    public final EVXData I;

    /** Prices (volume- and time- averaged) in initial {@literal |psiH>} state.
     */
    public final double
	pI_IH,pJ_IH;

    /** Prices (volume- and time- averaged) in {@literal |psi_M>} state;
     * must be identical according to constraint.
     */
    public final double
	pI_M,pJ_M;

    public final boolean flag_solution_found;

    public IstatesConditionalSameAverage(
	final int n,
	final double [] QQt,
	final double [] QQi,

	final double [] QQI,
	final double [] QQpI,
	final double [] QQJ,
	final double [] QQpJ,
	final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
	final double eps)
{
    // solve generalized eignenvalue problem, find I->max state
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);
    pI_IH=TMatr.getAMatrA(I.psiH,QQpI,n)/TMatr.getAMatrA(I.psiH,QQI,n);
    pJ_IH=TMatr.getAMatrA(I.psiH,QQpJ,n)/TMatr.getAMatrA(I.psiH,QQJ,n);


    // convert the matrices to |I|psik>=lambda|psik> basis to simplify algebra
    final double []
	bQQi=TMatr.getDiagMatrix(I.neq,I.ev),
	bQQI=EVXData.convertMatrFromQtoB_Basis(I.neq,QQI,I.ev), // I if QQp0i=QQi
	bQQpI=EVXData.convertMatrFromQtoB_Basis(I.neq,QQpI,I.ev),
	bQQJ=EVXData.convertMatrFromQtoB_Basis(I.neq,QQJ,I.ev), // unit if QQp0t=QQt
	bQQpJ=EVXData.convertMatrFromQtoB_Basis(I.neq,QQpJ,I.ev);


    // conditional optimization
    final double [] bQQconditional=new double [n*n];

     // Try several initial values. Note: they are in |I|psi>=lambda|psi> basis, not in the original basis.
    final java.util.List<GetMaxIState> bpsiInit=getSeqOfInitFunctions(
	n,
	IstatesConditionalV2.getInitialTriesCommonVersion(n), // staring functions
	bQQi,bQQI,bQQpI,bQQJ,bQQpJ,
	IstatesConditionalV2.getInitialTriesCommonVersion(n), // use the same variations as staring functions
	eps);

    double Icur_saved=0,mu_saved=0,lambda_Icur_saved=0,bpsi_saved[]=null;
    int ind_lambda_Icur_saved=-1,iTry_saved=-1;
    for(int iTry=0;iTry<bpsiInit.size();iTry++){
	GetMaxIState mbpsi=bpsiInit.get(iTry);
	double Icur=Double.NaN,mu=0,lambda_Icur=0;int ind_lambda_Icur=-1;
    eNoSave:{for(int it=0;it<N_iterations;it++){
	    final double [] bQQc=FindPsiConstrainedSameAverage.getAveragesVariation_bQQc(
		mbpsi.bpsi_M.bpsi,bQQI,bQQpI,bQQJ,bQQpJ,eps);
	    mu=getLagrangeMultiplier(mbpsi.bpsi_M,bQQi,bQQc);
	    for(int k=bQQconditional.length;--k>=0;) bQQconditional[k]=bQQi[k]+mu*bQQc[k];
	    Icur=mbpsi.i;
	    if(eps>0){
		// check constraint
		UnitTests.checkDoubleRel(TMatr.mulVV(mbpsi.bpsi_M.bpsi,mbpsi.bpsi_M.bpsi),"<bpsi|bpsi>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(mbpsi.bpsi_M.bpsi,bQQi,n),"<bpsi|I|bpsi>",mbpsi.i,"mbpsi.i",eps);
		UnitTests.checkDoubleRel(FindPsiConstrainedSameAverage.getConstraint(mbpsi.bpsi_M.bpsi,bQQI,bQQpI,bQQJ,bQQpJ),"getConstraint(mbpsi.bpsi_M.bpsi)",0,"0",eps);
		// convergence diag
		System.err.println("IstatesConditionalSameAverage: iTry="+iTry+" it="+it+" mbpsi.i="+mbpsi.i+" lambda_Icur="+lambda_Icur+" mu="+mu+" constraint="+FindPsiConstrainedSameAverage.getConstraint(mbpsi.bpsi_M.bpsi,bQQI,bQQpI,bQQJ,bQQpJ)+" =0 on success mbpsi.bpsiInit.orig_constraint="+mbpsi.bpsiInit.orig_constraint+"=0 on good convergence mbpsi.adjustment_type="+mbpsi.bpsi_M.adjustment_type);
	    }

	    if(it>=N_iterations-1) break;
	    final EVSelection s=new EVSelection(n,bQQconditional,bQQi,bQQI,bQQpI,bQQJ,bQQpJ,M,eps);
	    if(s.sNext==null){
		// iteration failed
		break eNoSave;
	    }
	    mbpsi=s.sNext;
	    lambda_Icur=s.lambda_i;
	    ind_lambda_Icur=s.ind_lambda_i;
	}
	if(Double.isFinite(Icur) && (bpsi_saved==null || Icur>=Icur_saved)){
	    // Save this try. No "herd" and genetic algorithms for now.
	    Icur_saved=Icur;
	    mu_saved=mu;
	    lambda_Icur_saved=lambda_Icur;
	    ind_lambda_Icur_saved=ind_lambda_Icur;
	    iTry_saved=iTry;
	    bpsi_saved=mbpsi.bpsi_M.bpsi;
	}}
	if(eps>0){
	    System.err.println("IstatesConditionalSameAverage: iTry="+iTry+" result: with:ind_lambda_Icur="+ind_lambda_Icur+" mbpsi.i="+mbpsi.i+" lambda_Icur="+lambda_Icur+" mu="+mu+" mbpsi.bpsi_M.bpsi="+java.util.Arrays.toString(mbpsi.bpsi_M.bpsi)+" constraint="+FindPsiConstrainedSameAverage.getConstraint(mbpsi.bpsi_M.bpsi,bQQI,bQQpI,bQQJ,bQQpJ));
	}
    }
    this.i_M=Icur_saved;
    this.lambda_i_M=lambda_Icur_saved;
    this.LagrangeMultiplier_M=mu_saved;
    this.flag_solution_found=bpsi_saved!=null;
    // convert found |psi> back to regular basis.
    this.psi_M=!flag_solution_found?new double [n]:EVXData.convertPsiFromBtoQ_Basis(n,bpsi_saved,I.ev);
    this.wr0_M=Math.pow(TMatr.getAMatrB(psi_M,I.psi0,QQt,n),2);

    pI_M=TMatr.getAMatrA(psi_M,QQpI,n)/TMatr.getAMatrA(psi_M,QQI,n);
    pJ_M=TMatr.getAMatrA(psi_M,QQpJ,n)/TMatr.getAMatrA(psi_M,QQJ,n);

    // test in original basis
    if(eps>0){
	System.err.println("IstatesConditionalSameAverage result: with:ind_lambda_Icur_saved="+ind_lambda_Icur_saved+" iTry_saved="+iTry_saved+" i_M="+i_M+" lambda_i_M="+lambda_i_M+" I.sH(max)="+I.sH+" wr0_M="+wr0_M+" LagrangeMultiplier_M="+LagrangeMultiplier_M+" flag_solution_found="+flag_solution_found+" n="+n);
	if(bpsi_saved!=null){
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_M,QQt,n),"<psi_M|psi_M>",1,"1",eps);
	    UnitTests.checkDouble(TMatr.getAMatrA(psi_M,QQi,n),"<psi_M|I|psi_M>",i_M,"i_M",eps);
	    UnitTests.checkDoubleRel(pI_M-pJ_M,"pI_M-pJ_M",0,"0",eps);
	}
    }
}


/** Given a sequence of input {@code bpsiInit}
 * modify each of them to satisfy the constraint then
 * apply {@link GetMaxIState} to find the state with maximal
 * {@literal <psi|bQQi|psi>}.
 */
private static java.util.List<GetMaxIState> getSeqOfInitFunctions(
    final int n,
    final java.util.List<double []> bpsiInit,
    final double [] bQQi,
    final double [] bQQI,
    final double [] bQQpI,
    final double [] bQQJ,
    final double [] bQQpJ,
    final java.util.List<double []> tryVariations,
    final double eps){

    final int nInitApprox=5;
    final var l=bpsiInit.stream()
        .map(bpsi->new GetMaxIState(n,
				    new FindPsiConstrainedSameAverage(bpsi,
								      bQQI,
								      bQQpI,
								      bQQJ,
								      bQQpJ,
								      tryVariations,
								      eps),
				    bQQi,eps))
	.filter(s->s.bpsi_M!=null) // take only valid solutions
        .sorted((p1,p2)->Double.compare​(p2.i,p1.i)).toList();

    return IstatesConditionalV2.selectSomeNotVerySimilarInitFunctions(n,nInitApprox,l,x->x.bpsi_M.bpsi);
}


/** Given an input bQQconditional=bQQi+mu*bQQc matrix
 * find it's eigenvectors and modify then to satisfy the constraint
 * to obtain new {@literal |psi>} candidates in several
 * variants as per {@link FindPsiConstrainedSameAverage}. Then test all available
 * states to find the one with maximal {@literal <psi|bQQi|psi>}.
 * This will be a new iteration.
 */
private static class EVSelection {
    final double [] ev_bQQconditional;
    /** The value of eigenvalue corresponding to eigenvector selected.
     */
    final double lambda_i;
    /** Index of lambda. */
    final int ind_lambda_i;
    /** Next state iteration. */
    final GetMaxIState sNext;
    EVSelection(final int n,
		final double [] bQQconditional,
		final double [] bQQi,
		final double [] bQQI,
		final double [] bQQpI,
		final double [] bQQJ,
		final double [] bQQpJ,
		final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		final double eps){
	// solve eigenvalue problem
	ev_bQQconditional=M.EV.getEV(n,bQQconditional,true);

	final java.util.ArrayList<double []> ev_states=new java.util.ArrayList<>();
	for(int s=0;s<n;s++){ev_states.add(EVXData.getPsiFromBasis(s,ev_bQQconditional,n));}
	final java.util.ArrayList<double []> tryVariations=new java.util.ArrayList<>(ev_states);
	tryVariations.addAll(IstatesConditionalV2.getInitialTriesCommonVersion(n));

	// do it with a cycle as we need to propagate eigenstate index for diag.
	int indS=-1;
	GetMaxIState res=null;
	for(int s=0;s<n;s++){
	    final java.util.List<GetMaxIState> t=getSeqOfInitFunctions(
		n,java.util.List.of(ev_states.get(s)),
		bQQi,bQQI,bQQpI,bQQJ,bQQpJ,tryVariations,eps);
	    if(t.size()>0 && (indS<0 || (t.get(0).i>res.i))){
		indS=s;
		res=t.get(0);
	    }
	}
	this.ind_lambda_i=indS;
	this.sNext=res;
	this.lambda_i=this.ind_lambda_i>=0?ev_bQQconditional[this.ind_lambda_i]:Double.NaN;
	if(res!=null && eps>0){
	    System.err.println("IstatesConditionalSameAverage:EVSelection: ind_lambda_i="+ind_lambda_i+" res.i="+res.i+" orig_constraint="+res.bpsiInit.orig_constraint+" adjustment_type="+res.bpsi_M.adjustment_type);
	}
    }
}


/**
 * Given an input {@link FindPsiConstrainedSameAverage} class it takes
 * all it's adjusted {@link FindPsiConstrainedSameAverage#psiConstrained} states
 * and select the one with maximal {@literal <psi|I|psi>} among all the
 * provided {@literal |psi>} with the constraint satisfied.
 */
private static class GetMaxIState {
    final double i;
    /** The state of maximal {@literal <psi|I|psi>}.
     * It is {@code null} if nothing available.
     */
    final PsiConstrained bpsi_M;
    final FindPsiConstrainedSameAverage bpsiInit;
    GetMaxIState(final int n,
		 final FindPsiConstrainedSameAverage bpsiInit,
		 final double [] bQQi,
		 final double eps)
{
    this.bpsiInit=bpsiInit;
    double iMax=Double.NaN;
    PsiConstrained bpsi_M=null;
    // Find the state with maximal I among all candidates found.
    // It may be a better option to create a "herd" in a genetic algorithm style.
    for(PsiConstrained b:bpsiInit.psiConstrained){
	final double ib=TMatr.getAMatrA(b.bpsi,bQQi,n);
	if(Double.isFinite(ib) && (bpsi_M==null || ib>=iMax)){
	    iMax=ib;
	    bpsi_M=b;
	}
	if(eps>0){
	    // check constraint
	    UnitTests.checkDoubleRel(TMatr.mulVV(b.bpsi,b.bpsi),"<b.bpsi|b.bpsi>",1,"1",eps);
	}
    }
    this.i=iMax;
    this.bpsi_M=bpsi_M;
    if(eps>0){
	if(this.bpsi_M!=null){
	    System.err.println("IstatesConditionalSameAverage:GetMaxIState: selected i="+this.i+" orig_constraint="+bpsiInit.orig_constraint+" adjustment_type="+this.bpsi_M.adjustment_type+" of total input states "+bpsiInit.psiConstrained.size());
	} else {
	    System.err.println("IstatesConditionalSameAverage:GetMaxIState: no selected psi");
	}
    }
}
} // END OF class GetMaxIState


/** Calculate Lagrange multiplier.
 */
private static double getLagrangeMultiplier(
    final PsiConstrained bpsi_M,
    final double [] bQQi,
    final double [] bQQc){

    double mu=0;
    if(bpsi_M!=null){
	final int n=bpsi_M.bpsi.length;
	final double [] Cx=TMatr.getMatr_x_A(n,bQQc,bpsi_M.bpsi),
	    Ix=TMatr.getMatr_x_A(n,bQQi,bpsi_M.bpsi);
	final double cc=TMatr.mulVV(Cx,Cx);
	// Lagrange multiplier
	if(cc>0) mu=-TMatr.mulVV(Cx,Ix)/cc;
	if(!Double.isFinite(mu)) mu=0;
    }
    return mu;
}



static void printDiag(IstatesConditionalSameAverage m){
    // print some diag
    System.err.println("IstatesConditionalSameAverage.main:i_M="+m.i_M+" lambda_i_M="+m.lambda_i_M+" psi_M="+java.util.Arrays.toString(m.psi_M)+" flag_solution_found="+m.flag_solution_found+" n="+m.I.neq);
}

/** Unit test. */
public static void main(String [] args){
    TestOptimizationPsi.runMultiTest("IstatesConditionalSameAverage",15,600,
		 new java.util.Random(18301L),
		 1e-5,
		 testC-> {final var m=new IstatesConditionalSameAverage(
			    testC.n,
			    testC.QQt,
			    testC.QQi,

			    testC.QQi,testC.QQpi,
			    testC.QQt,testC.QQpt,

			    testC.M,
			    testC.eps);
			   printDiag(m);
			  }
		 );
}
}
