package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2024
 * This file is under GPL version 3.
 *
 */

/**
 * Find the {@literal <psi|I|psi> ->} max state
 * subject to the constraint: {@literal <psi|C|psi>=0}.
 * Here we use self--concordant approach to solve Rayleigh quotient
 * maximization with the quadratic constraint:{@literal <psi|C|psi>=0}.
 * This class uses per-iteration linear constraint to improve convergence.
 * On each iteration instead of commnonly used pair (solution,Lagrange multiplier) we use
 * a triple: (solution,Lagrange multiplier,Linear constraint).
 * @see IstatesConditionalV2
 * @version $Id: IstatesConditionalSubspaceLinearConstraints.java,v 1.13 2024/05/14 05:23:16 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class IstatesConditionalSubspaceLinearConstraints {
    private static final int N_iterations=15;
    private static final boolean FORCE_DIAG_DUMP=true;

    public final double i_M,lambda_i_M,LagrangeMultiplier_M,wr0_M,orig_constraint;

    /** The state of maxI conditional to {@literal <psi|QQc|psi>=0}.
     */
    public final double [] psi_M;

    /** Whether a solution exists.
     */
    public final boolean flag_solution_exists;

    /** The {@literal |I|psi>=lambda |psi>} solution.
     */
    public final EVXData I;

public IstatesConditionalSubspaceLinearConstraints(
    final int n,
    final double [] QQt,
    final double [] QQi,
    final double [] QQc,
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
    final double eps)
{
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);

    final double [] bQQi=TMatr.getDiagMatrix(n,I.ev),
	bQQc=getbQQcFromQQc(QQc), // may be overloaded for more complex models.
	// to check whether a solution exist
	ev_bQQc=M.EV.getEV(n,bQQc,true);
    // if the bQQc simultaneously has positive and negative eigenvalues the solution exists
    this.flag_solution_exists=ev_bQQc[0]<0 && ev_bQQc[n-1]>0;

    final double [] bQQconditional=new double [n*n];
    final java.util.function.DoubleFunction<double []> adjWithLM=lm->{
	// construct the matrix adjusted with Lagrange multiplier
	// |bQQconditional| = |I|+lm*|C|
	for(int k=bQQconditional.length;--k>=0;) {
	    bQQconditional[k]=bQQi[k]+lm*bQQc[k];
	}
	return bQQconditional;
    };


    class CreateSubspaceOrtU {

	/** A basis with linear constraints eliminated,
	 * a matrix of newDim*n size.
	 */
	final double [] fromSubspaceToOrig;

	/** Problem dimension after elimination of linear constraints.
	 */
	final int newDim;

	/**
	 * @param u Solution current iteration.
	 */
	CreateSubspaceOrtU(final double [] u){

	    // BEGIN create linear constraint
	    final double [][] linearConstrains=new double [][]{TMatr.getMatr_x_A(n,bQQc,u)};
	    // END create linear constraints

	    // Gaussian elimination object from linear constraints.
	    final var t=new EliminateLinearConstraints_HomegrownLUFactorization(n,linearConstrains,condMaxToFail_LinearSystem);

	    // BEGIN create a basis for eigenvalue problem.
	    this.newDim=t.dimXnew;
	    this.fromSubspaceToOrig=TMatr.getMatrTrans(t.dimXorig,t.dimXnew,t.fromXnewToXorigAsMatrix());
	    // END create a basis for eigenvalue problem.

	    // BEGIN unit test
	    if(eps>0){
		final double []
		    mSKs=EVXData.convertMatrFromQtoB_Basis(n,bQQi,fromSubspaceToOrig,newDim,0),
		    mQs=TMatr.mul_MA_x_MBTrans(newDim,fromSubspaceToOrig,n,fromSubspaceToOrig,newDim);

		UnitTests.checkDoubleRel1(TMatr.getAMatrA(u,bQQc,n),"<u|bQQc|u>",0,"0",eps);
		for(int p=0;p<newDim;p++){
		    final double [] vbpX=EVXData.getPsiFromBasis(p,fromSubspaceToOrig,n,0);
		    // check the constraint is OK for every vector in basis
		    UnitTests.checkDoubleRel1(TMatr.getAMatrB(u,vbpX,bQQc,n),"<u|bQQc|vbpX>",0,"0",eps);

		    // check basis conversion
		    final double [] xnewpZ=new double [newDim];xnewpZ[p]=1;
		    final double []
			xnewpXbfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewpZ,1),
			xnewpXb=t.fromXnewToXorig(xnewpZ);
		    UnitTests.checkDoubleArrayRel1(xnewpXb,"xnewpXb",vbpX,"vbpX",eps);
		    UnitTests.checkDoubleArrayRel1(xnewpXb,"xnewpXb",xnewpXbfm,"xnewpXbfm",eps);

		    // check matrices
		    for(int i=0;i<newDim;i++){
			final double [] xnewiZ=new double [newDim];xnewiZ[i]=1;
			final double [] xnewiXbfm=TMatr.mul_MATrans_x_MB(n,fromSubspaceToOrig,newDim,xnewiZ,1);
			// test matrix conversion formulas
			UnitTests.checkDoubleRel1(TMatr.getAMatrB(xnewpXbfm,xnewiXbfm,bQQi,n),"<xnewpXbfm|bQQi|xnewiXbfmx>",TMatr.getAMatrB(xnewpZ,xnewiZ,mSKs,newDim),"<xnewpZ|mSKs|xnewiZ>",eps);
			UnitTests.checkDoubleRel1(TMatr.mulVV(xnewpXbfm,xnewiXbfm),"<xnewpXbfm|xnewiXbfm>",TMatr.getAMatrB(xnewpZ,xnewiZ,mQs,newDim),"<xnewpZ|mQs|xnewiZ>",eps);
		    }
		}
		System.err.println("CreateSubspaceOrtU: nEliminated="+t.nEliminated+" newDim="+newDim+" n="+n);
	    }
	    // END unit test
	}
    } // END OF class CreateSubspaceOrtU


    class LocalMax {
	final double Forig;
	final EVSelection e;
	final int itNum;
	final double [] u_orig;
	final GetMaxIStateAdj adj;
	LocalMax(final EVSelection e,int itNum){
	    this.e=e;
	    this.u_orig=TMatr.getVectorScaled(e.eVector,Math.sqrt(1/TMatr.mulVV(e.eVector,e.eVector)));
	    this.Forig=TMatr.getAMatrA(u_orig,bQQi,n);
	    this.itNum=itNum;
	    this.adj=new GetMaxIStateAdj(n,u_orig,bQQc,bQQi,ev_bQQc,eps);
	}
	boolean isValid(){return Double.isFinite(adj.i);}
	boolean isArgLocalBetter(LocalMax s){
	    // for now
	    return s.adj.i>=this.adj.i;
	}
	boolean isArgGlobalBetter(LocalMax s){
	    return s.adj.i>=this.adj.i;
	}
    } // END OF class LocalMax

    // conditional optimization
    int nA=n;
    double [] T=TMatr.getUnitMatrix(nA);
    double mu=0; // Lagrange multiplier by <u|bQQc|u>
    LocalMax globMax=null,prevLocMax=null;
    for(int it=0;it<N_iterations && flag_solution_exists;it++)
    {
	// solve generalized regular eigenproblem
	final double [] evConstr=M.EV.getGEV_Generalized(
		nA,
		/*SKs=*/EVXData.convertMatrFromQtoB_Basis(n,adjWithLM.apply(mu),T,nA,0),
		/*Qs=*/TMatr.mul_MA_x_MBTrans(nA,T,n,T,nA),
		true);

	LocalMax locMax=null;
	boolean flagGlobalMax=false;
	int nNegLambda=0,nFromTopLambda=0;
	// select eigenvalue, start with the maximal
	for(int p=nA;--p>=0;){
	    final double eLambda=evConstr[p];
	    nFromTopLambda++;if(eLambda<=0) nNegLambda++;
	    // Currently we test only the maximal eigenvalue, add heuristics & tries later.
	    if(it<=0 || nFromTopLambda <=1 || locMax==null){
		final var e=new EVSelection(
		    p,
		    TMatr.mul_MATrans_x_MB(n,T,nA,EVXData.getPsiFromBasis(p,evConstr,nA),1),
		    eLambda);
		final var s=new LocalMax(e,it);
		if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(s)){
		    // found global max (across iterations)
		    globMax=locMax=s;
		    flagGlobalMax=true;
		}
		else if(!flagGlobalMax && (
			    locMax==null || !locMax.isValid() || locMax.isArgLocalBetter(s)
					   )){
		    // some local max
		    locMax=s;
		}
	    }
	}
	// END select the eigenvector among several
	if(FORCE_DIAG_DUMP||eps>0){
	    System.err.println("it="+it+":Selected ind="+locMax.e.ind+" ("+nA+") eLambda="+locMax.e.eLambda+" mu="+locMax.adj.m+"; Forig="+locMax.Forig+" Fadj="+locMax.adj.i+" orig_constraint="+locMax.adj.orig_constraint+" n="+n);
	}


	// BEGIN calculate new values of Lagrange multiplier and constraint satisfying basis
	final var l=new CreateSubspaceOrtU(locMax.adj.bpsi_M.bpsi);
	// Lagrange multiplier
	mu=locMax.adj.m;
	// linear constraints subspace
	nA=l.newDim;
	T=l.fromSubspaceToOrig;
	// END calculate new values of Lagrange multiplier and constraint satisfying basis

	prevLocMax=locMax;
	if(globMax==null || !globMax.isValid() || globMax.isArgGlobalBetter(locMax)) {
	    globMax=locMax;
	}
    }

    if(eps>0){
	if(globMax!=null) System.err.println("IstatesConditionalSubspaceLinearConstraints: Selected found max globMax.adj.i="+globMax.adj.i+" eLambda="+globMax.e.eLambda+" mu="+globMax.adj.m+"; Forig="+globMax.Forig+" Fadj="+globMax.adj.i+" constraint="+TMatr.getAMatrA(globMax.adj.bpsi_M.bpsi,bQQc,n)+" =0 on success orig_constraint="+globMax.adj.orig_constraint+" =0 in iteration convergence;  adjustment_type="+globMax.adj.bpsi_M.adjustment_type+" flag_solution_exists="+flag_solution_exists);
    }
    // bpsi_M is in the basis of bQQi eigenvectors I.ev, convert it to regular basis.
    this.psi_M=globMax==null?new double [n]:EVXData.convertPsiFromBtoQ_Basis(n,globMax.adj.bpsi_M.bpsi,I.ev);
    // the <psi_M|psi0>^2 as a kind of "distance" to "now"
    this.wr0_M=globMax==null?Double.NaN:Math.pow(TMatr.getAMatrB(psi_M,I.psi0,QQt,n),2);
    this.LagrangeMultiplier_M=globMax==null?Double.NaN:globMax.adj.m;
    this.i_M=TMatr.getAMatrA(psi_M,QQi,n);
    this.lambda_i_M=globMax==null?Double.NaN:globMax.e.eLambda;
    if(globMax!=null){
	this.orig_constraint=globMax.adj.orig_constraint;
	if(eps>0){
	     UnitTests.checkDoubleRel1(i_M,"i_M",globMax.adj.i,"globMax.adj.i",eps);
	    final double [] bx=TMatr.getMatr_x_A(n,bQQi,globMax.adj.bpsi_M.bpsi),
		cx=TMatr.getMatr_x_A(n,bQQc,globMax.adj.bpsi_M.bpsi);
	    TMatr.sumP1(bx,globMax.adj.bpsi_M.bpsi,-TMatr.mulVV(bx,globMax.adj.bpsi_M.bpsi));
	    UnitTests.checkDoubleRel(TMatr.mulVV(bx,globMax.adj.bpsi_M.bpsi),"bx*bpsi_M",0,"0",eps);
	    UnitTests.checkDoubleRel(TMatr.mulVV(cx,globMax.adj.bpsi_M.bpsi),"cx*bpsi_M",0,"0",eps);
	    UnitTests.checkDoubleRel1(-TMatr.mulVV(bx,cx)/TMatr.mulVV(cx,cx),"-bx*cx/(cx*cx)",globMax.adj.m,"mu",eps);
	}
    } else this.orig_constraint=Double.NaN;

    if(eps>0){
	System.err.println("IstatesConditionalSubspaceLinearConstraints: i_M="+i_M+" orig_constraint="+orig_constraint+" r.adjustment_type="+(globMax!=null?String.valueOf(globMax.adj.bpsi_M.adjustment_type):"??")+" flag_solution_exists="+flag_solution_exists);
	if(globMax!=null){
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_M,QQt,n),"<psi_M|psi_M>",1,"1",eps);
	    UnitTests.checkDoubleRel(i_M,"i_M",globMax.adj.i,"globMax.adj.i",eps);
	}
    }
    if(eps>0 && flag_solution_exists){
	// compare with localized solution
	final var localized=new IstatesConditionalLocalized(n,QQt,QQi,QQc,M,0*eps);
	IstatesConditionalLocalized.printDiag(
	    "IstatesConditionalSubspaceLinearConstraints: test OK i_M="+i_M+" wr0_M="+wr0_M+" ",
	    localized,
	    "; compare: <psi_M|localized.psi_M>="+TMatr.getAMatrB(psi_M,localized.psi_M,QQt,n)+" DiffLoc i_M="+(i_M-localized.i_M)+
	    " flag_solution_exists="+flag_solution_exists+" eps="+eps);
    }
}


double [] getbQQcFromQQc(final double [] QQc){
    return EVXData.convertMatrFromQtoB_Basis(I.neq,QQc,I.ev);
}



private static class EVSelection {
    /** Index of original eigenvector, \(ind \in [0\dots n-1] \).
     */
    final int ind;
    /** Original eigenvector, an operator is extracted from it row by row.
     */
    final double [] eVector;
    /** The eigenvalue of this state.
     */
    final double eLambda;
    EVSelection(final int ind,
		final double [] eVector,
		final double eLambda){
	this.ind=ind;
	this.eVector=eVector;
	this.eLambda=eLambda;
    }
}// END OF class EVSelection


/** For a given {@code u} try a number of adjustment methods,
 * select the one, providing the maximal I
 * after the constraint {@literal <psi|bQQc|psi>=0} is satisfied.
 */
private static class GetMaxIStateAdj {
    final double i,orig_constraint;
    final PsiConstrained bpsi_M;
    /** Lagrange multiplier. */
    final double m;
    GetMaxIStateAdj(final int n,
		    final double [] u,
		    final double [] bQQc,
		    final double [] bQQi,
		    final double [] ev_bQQc,
		    final double eps)
{
    double iMax=0,orig_constraint=Double.NaN;
    PsiConstrained bpsi_M=null;
    final var f=new FindPsiConstrainedSingleQuadratic0(u,bQQc,ev_bQQc,eps);
    for(PsiConstrained b: f.psiConstrained){
	final double ib=TMatr.getAMatrA(b.bpsi,bQQi,n);
	if(Double.isFinite(ib) && (bpsi_M==null || ib>=iMax)){
	    iMax=ib;
	    bpsi_M=b;
	    orig_constraint=f.orig_constraint;
	}
    }
    this.bpsi_M=bpsi_M;
    if(bpsi_M==null){
	this.i=this.orig_constraint=Double.NaN;
	this.m=0;
    } else {
	this.i=iMax;
	this.orig_constraint=orig_constraint;
	this.m=getLagrangeMultiplier(this.bpsi_M.bpsi,bQQi,bQQc);
    }
    if(eps>0){
	if(this.bpsi_M!=null){
	    System.err.println("IstatesConditionalSubspaceLinearConstraints:GetMaxIState: selected i="+this.i+" orig_constraint="+orig_constraint+" adjustment_type="+this.bpsi_M.adjustment_type);
	} else {
	    System.err.println("IstatesConditionalSubspaceLinearConstraints:GetMaxIState: no selected solution");
	}
    }
}

private static double getLagrangeMultiplier(final double [] bpsi_M,final double [] bQQi,final double [] bQQc){
    if(bpsi_M!=null){
	final int n=bpsi_M.length;
	final double [] Cx=TMatr.getMatr_x_A(n,bQQc,bpsi_M),
	    Ix=TMatr.getMatr_x_A(n,bQQi,bpsi_M);
	final double cc=TMatr.mulVV(Cx,Cx);
	// Lagrange multiplier
	if(cc>0) return -TMatr.mulVV(Cx,Ix)/cc;
    }
    return 0;
}
} // END OF class GetMaxIStatewAdj






/** For unit test.
 */
public static void matchIstatesConditionalSubspaceLinearConstraints(final IstatesConditionalSubspaceLinearConstraints a,final IstatesConditionalSubspaceLinearConstraints b,final double eps){
    EVXData.matchEVXData(a.I,b.I,false,eps);

    UnitTests.checkDoubleRel(a.i_M,"a.i_M",b.i_M,"b.i_M",eps);
    UnitTests.checkDoubleRel(a.LagrangeMultiplier_M,"a.LagrangeMultiplier_M",b.LagrangeMultiplier_M,"b.LagrangeMultiplier_M",eps);

    System.err.println("IstatesConditionalSubspaceLinearConstraints:matchIstatesConditionalSubspaceLinearConstraints test OK eps="+eps);
}

static void printDiag(IstatesConditionalSubspaceLinearConstraints m){
    // print some diag
    System.err.println("IstatesConditionalSubspaceLinearConstraints.main:I EV="+java.util.Arrays.toString(m.I.getEValues()));
    System.err.println("IstatesConditionalSubspaceLinearConstraints.main:i_M="+m.i_M+" lambda_i_M="+m.lambda_i_M+" psi_M="+java.util.Arrays.toString(m.psi_M)+" flag_solution_exists="+m.flag_solution_exists+" n="+m.I.neq);
}

/** Unit test. */
public static void main(String [] args){
    final int [] testCounter=new int [3];
    TestOptimizationPsi.runMultiTest("IstatesConditionalSubspaceLinearConstraints",15,600,
		 new java.util.Random(18301L),
		 1e-5,
		 testC-> {
			   final var pc=0.2; // some arbitrary value
			   final var m=new IstatesConditionalSubspaceLinearConstraints(
			       testC.n,
			       testC.QQt,
			       testC.QQi,
			       TMatr.sumAcB(testC.QQpi,-pc,testC.QQi), // QQc
			       testC.M,
			       testC.eps);
			  // print some diag
			  printDiag(m);

			  // try V2 on the same data
			  final var m2=new IstatesConditionalV2(
			       testC.n,
			       testC.QQt,
			       testC.QQi,
			       TMatr.sumAcB(testC.QQpi,-pc,testC.QQi), // QQc
			       testC.M,
			       testC.eps*0/* disable internal tests for IstatesConditionalSubspaceLinearConstraintsV2*/);
			  // print some diag
			  IstatesConditionalV2.printDiag(m2);

			  if(!m.flag_solution_exists) testCounter[2]++;
			  testCounter[0]++;

			  // potentantially may fail, but for this data it is mostly OK.
			  if(
			      // I diffrent
			      UnitTests.checkDoubleRelNoFail(m.i_M,m2.i_M,testC.eps) ||
			      // one program found and another not
			      m.flag_solution_exists!=m2.flag_solution_exists ||
			      // different psi_M found
			      (m2.flag_solution_exists && UnitTests.checkDoubleRelNoFail(Math.abs(TMatr.getAMatrB(m.psi_M,m2.psi_M,testC.QQt,testC.n)),1,testC.eps))
			     ){
			      testCounter[1]++;
			      System.err.println(
				  "!!! IstatesConditionalSubspaceLinearConstraints is different from IstatesConditionalSubspaceLinearConstraintsV2: m.i_M="+m.i_M+" m2.i_M="+m2.i_M+
				  //" <m.psi_M|m.psi_M>="+TMatr.getAMatrA(m.psi_M,testC.QQt,testC.n)+
				  //" <m2.psi_M|m2.psi_M>="+TMatr.getAMatrA(m2.psi_M,testC.QQt,testC.n)+
				  " <m.psi_M|m2.psi_M>="+TMatr.getAMatrB(m.psi_M,m2.psi_M,testC.QQt,testC.n)+
				  " m.lambda_i_M="+m.lambda_i_M+" m2.lambda_i_M="+m2.lambda_i_M+
				  " m.LagrangeMultiplier_M="+m.LagrangeMultiplier_M+" m2.LagrangeMultiplier_M="+m2.LagrangeMultiplier_M+
				  " n="+testC.n+" eps="+testC.eps);
			  }
			}
		 );
    System.err.println("IstatesConditionalSubspaceLinearConstraints run "+testCounter[0]+" times."+
		       " Different from IstatesConditionalSubspaceLinearConstraintsV2 "+testCounter[1]+" times."+
		       " Solution does not exist "+testCounter[2]+" times.");
}


    /** Minimal norm for Gaussian elimination.
     */
    private static final double condMaxToFail_LinearSystem=1e-10;
}
