package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2022
 * This file is under GPL version 3.
 *
 */

/**
  * Find {@literal \( \left|\psi\right> \) } state such that
  * {@literal \[
  \frac{\left<\psi|I| \psi\right>}{\left<\psi|\psi\right> }
  \xrightarrow[\psi]{\quad }\max
  \] }
  * subject to:
{@literal \[
\left<\psi|C|\psi\right>=0
\] }.
 * Similar to {@link IstatesConditional} but has an updated
 * algorithm (minor improvements). A different approach
 * in {@link IstatesConditionalSubspaceLinearConstraints} seems to provide
 * better results.
 * @see IstatesConditionalSubspaceLinearConstraints
 * @see IstatesConditionalLocalized
 * @see MaximizeStandardDeviation
 * @version $Id: IstatesConditionalV2.java,v 1.49 2024/05/14 05:23:16 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class IstatesConditionalV2 {
    private static final int N_iterations=12;

    /** The  {@literal <psi_M|I|psi_M>}.
     */
    public final double i_M;

    /** Last value of eigenstate, in case of algorithm convergence
     * matches to {@link #i_M}.
     */
    public final double lambda_i_M;

    /** Lagrange multiplier value. */
    public final double LagrangeMultiplier_M;

    /** The {@literal <psi_M|psi0>^2}, a kind of "distance" to "now".
     */
    public final double wr0_M;

    /** The state of maxI conditional to {@literal <psi|QQc|psi>=0}.
     */
    public final double [] psi_M;
    /** Whether a solution exists.
     */
    public final boolean flag_solution_exists;

    /** The {@literal |I|psi>=lambda |psi>} solution.
     */
    public final EVXData I;
    public IstatesConditionalV2(final int n,
				final double [] QQt,final double [] QQi,
				final double [] QQc,
				final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				final double eps)
{
    this.I=EVXData.getEVXDataFromQQ(n,QQi,QQt,M);

    final double [] bQQi=TMatr.getDiagMatrix(n,I.ev),
	bQQc=getbQQcFromQQc(QQc), // may be overloaded for more complex models.
	// to check whether a solution exist
	ev_bQQc=M.EV.getEV(n,bQQc,true);
    // if the bQQc simultaneously has positive and negative eigenvalues the solution exists
    this.flag_solution_exists=ev_bQQc[0]<0 && ev_bQQc[n-1]>0;


    // conditional optimization
    final double [] bQQconditional=new double [n*n];

    // Try several initial values. Note: they are in |I|psi>=lambda|psi> basis, not in the original basis.
    final java.util.List<GetMaxIState> bpsiInit=getSeqOfInitFunctions(
	n,
	// Every |I|psi>=lambda|psi> state as init.
	getInitialTriesCommonVersion(n),
	/*java.util.stream.Stream.of(
	    // Every |I|psi>=lambda|psi> state as init.
	    getInitialTriesCommonVersion(n).stream()
	    // Localized solutions as init. Provide little value and can be probably removed for speed.
	    //,java.util.Arrays.stream(new IstatesConditionalLocalized(n,QQt,QQi,QQc,M,eps).matches).map(x->EVXData.convertPsiFromQtoB_Basis(n,x.psi_M,I.ev,QQt))
	    ).flatMap(x -> x).toList(),*/
	bQQi,
	bQQc,
	ev_bQQc,
	eps);
    if(eps>0){
	System.err.println("IstatesConditionalV2: initStates="+bpsiInit.stream().map(s->"(i="+s.i+" adjustment_type="+s.bpsi_M.adjustment_type+" bpsi="+java.util.Arrays.toString(s.bpsiInit.orig_bpsi)+")").toList()+" flag_solution_exists="+flag_solution_exists);
    }

    double Icur_saved=0,mu_saved=0,lambda_Icur_saved=0,bpsi_saved[]=null;
    int ind_lambda_Icur_saved=-1,iTry_saved=-1;
    for(int iTry=0;flag_solution_exists && iTry<bpsiInit.size();iTry++){
	GetMaxIState mbpsi=bpsiInit.get(iTry);
	double Icur=Double.NaN,mu=0,lambda_Icur=0;int ind_lambda_Icur=-1;
    eNoSave:{for(int it=0;it<N_iterations;it++){
	    mu=getLagrangeMultiplier(mbpsi.bpsi_M,bQQi,bQQc);
	    for(int k=bQQconditional.length;--k>=0;) bQQconditional[k]=bQQi[k]+mu*bQQc[k];
	    Icur=mbpsi.i;
	    if(eps>0){
		// check constraint
		UnitTests.checkDoubleRel(TMatr.mulVV(mbpsi.bpsi_M.bpsi,mbpsi.bpsi_M.bpsi),"<bpsi|bpsi>",1,"1",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(mbpsi.bpsi_M.bpsi,bQQi,n),"<bpsi|I|bpsi>",mbpsi.i,"mbpsi.i",eps);
		UnitTests.checkDoubleRel(TMatr.getAMatrA(mbpsi.bpsi_M.bpsi,bQQc,n),"<bpsi|C|bpsi>",0,"0",eps);
		// convergence diag
		System.err.println("IstatesConditionalV2: iTry="+iTry+" it="+it+" mbpsi.i="+mbpsi.i+" lambda_Icur="+lambda_Icur+" mu="+mu+" constraint="+TMatr.getAMatrA(mbpsi.bpsi_M.bpsi,bQQc,n)+" =0 on success mbpsi.bpsiInit.orig_constraint="+mbpsi.bpsiInit.orig_constraint+"=0 on good convergence mbpsi.adjustment_type="+mbpsi.bpsi_M.adjustment_type);
	    }

	    if(it>=N_iterations-1) break;
	    final EVSelection s=new EVSelection(n,bQQconditional,bQQi,bQQc,ev_bQQc,M,eps);
	    if(s.sNext==null){
		// iteration failed
		break eNoSave;
	    }
	    mbpsi=s.sNext;
	    lambda_Icur=s.lambda_i;
	    ind_lambda_Icur=s.ind_lambda_i;
	}
	if(Double.isFinite(Icur) && (bpsi_saved==null || Icur>=Icur_saved)){
	    // Save this try. No "herd" and genetic algorithms for now.
	    Icur_saved=Icur;
	    mu_saved=mu;
	    lambda_Icur_saved=lambda_Icur;
	    ind_lambda_Icur_saved=ind_lambda_Icur;
	    iTry_saved=iTry;
	    bpsi_saved=mbpsi.bpsi_M.bpsi;
	}}
	if(eps>0){
	    System.err.println("IstatesConditionalV2: iTry="+iTry+" result: with:ind_lambda_Icur="+ind_lambda_Icur+" mbpsi.i="+mbpsi.i+" lambda_Icur="+lambda_Icur+" mu="+mu+" constraint="+TMatr.getAMatrA(mbpsi.bpsi_M.bpsi,bQQc,n));
	}
    }
    this.i_M=Icur_saved;
    this.lambda_i_M=lambda_Icur_saved;
    this.LagrangeMultiplier_M=mu_saved;
    // convert found |psi> back to regular basis.
    this.psi_M=bpsi_saved==null?new double [n]:EVXData.convertPsiFromBtoQ_Basis(n,bpsi_saved,I.ev);
    this.wr0_M=Math.pow(TMatr.getAMatrB(psi_M,I.psi0,QQt,n),2);
    // test in original basis
    if(eps>0){
	System.err.println("IstatesConditionalV2 result: with:ind_lambda_Icur_saved="+ind_lambda_Icur_saved+" iTry_saved="+iTry_saved+" i_M="+i_M+" lambda_i_M="+lambda_i_M+" wr0_M="+wr0_M+" LagrangeMultiplier_M="+LagrangeMultiplier_M+" flag_solution_exists="+flag_solution_exists+" n="+n);
	if(bpsi_saved!=null){
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_M,QQt,n),"<psi_M|psi_M>",1,"1",eps);
	    UnitTests.checkDouble(TMatr.getAMatrA(psi_M,QQi,n),"<psi_M|I|psi_M>",i_M,"i_M",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(psi_M,QQc,n),"<psi_M|C|psi_M>",0,"0",eps);
	}
    }

    if(eps>0 && flag_solution_exists){
	// compare with localized solution
	final IstatesConditionalLocalized localized=new IstatesConditionalLocalized(n,QQt,QQi,QQc,M,0*eps);
	IstatesConditionalLocalized.printDiag(
	    "IstatesConditionalV2: test OK i_M="+i_M+" wr0_M="+wr0_M+" ",
	    localized,
	    "; compare: <psi_M|localized.psi_M>="+TMatr.getAMatrB(psi_M,localized.psi_M,QQt,n)+" DiffLoc i_M="+(i_M-localized.i_M)+
	    " flag_solution_exists="+flag_solution_exists+" eps="+eps);
    }
}

double [] getbQQcFromQQc(final double [] QQc){
    return EVXData.convertMatrFromQtoB_Basis(I.neq,QQc,I.ev);
}


static java.util.List<double[]> getInitialTriesCommonVersion(final int n){
    return
	// Every |I|psi>=lambda|psi> state as init and all projections the same
	java.util.stream.IntStream.range(-1,n).mapToObj(k->{
		final var d=new double [n];
		// an eigenstate or all projections the same init.
		if(k<0){java.util.Arrays.fill(d,1/Math.sqrt(n));}else d[k]=1;
		return d;
	    }).toList();
}

/** Given a sequence of input {@code bpsiInit}
 * modify each of them to satisfy the constraint then
 * apply {@link GetMaxIState} to find the state with maximal
 * {@literal <psi|bQQi|psi>}.
 */
private static java.util.List<GetMaxIState> getSeqOfInitFunctions(
    final int n,
    final java.util.List<double []> bpsiInit,
    final double [] bQQi,
    final double [] bQQc,
    final double [] ev_bQQc,
    final double eps){

    final int nInitApprox=5;
    final var l=bpsiInit.stream()
        .map(bpsi->new GetMaxIState(n,new FindPsiConstrainedSingleQuadratic0(bpsi,bQQc,ev_bQQc,eps),bQQi,bQQc,eps))
	.filter(s->s.bpsi_M!=null) // take only valid solutions
        .sorted((p1,p2)->Double.compare​(p2.i,p1.i)).toList();

    return selectSomeNotVerySimilarInitFunctions(n,nInitApprox,l,x->x.bpsi_M.bpsi);
}


/** Select given number of initial states removing too similar.
 */
static <T> java.util.List<T> selectSomeNotVerySimilarInitFunctions(
    final int n,
    final int nInitApprox,
    final java.util.List<T> inSorted,
    final java.util.function.Function<T,double []> getPsi){

    // too few available
    if(inSorted.size()<=nInitApprox) return inSorted;

    final double wMin=1.0-1.0/(nInitApprox+n);
    // Select at most nInitApprox skipping too similar
    final java.util.ArrayList<T>
	a=new java.util.ArrayList<>(),al=new java.util.ArrayList<>();
    al.addAll(inSorted);

    // take only not very similar
    while(al.size()>0 && a.size()<Math.min(nInitApprox,inSorted.size())){
	//System.err.println("a.size  before:"+al.stream().map(x->String.valueOf(Math.abs(TMatr.mulVV(getPsi.apply(x),getPsi.apply(al.get(0)))))).toList()+" n="+n);
	//System.err.print("inSorted.size="+inSorted.size()+" wMin="+wMin+" a.size="+a.size()+" al.size="+al.size());
	final var pCur=al.remove(0);
	a.add(pCur);
	final var pCur_psi=getPsi.apply(pCur);
	if(n!=pCur_psi.length) throw new IllegalArgumentException();
	al.removeIf(x-> wMin <= Math.abs(TMatr.mulVV(getPsi.apply(x),pCur_psi)));
	//System.err.println(": after remove a.size="+a.size()+" al.size="+al.size());
	//System.err.println("a.size  after:"+al.stream().map(x->String.valueOf(Math.abs(TMatr.mulVV(getPsi.apply(x),getPsi.apply(pCur))))).toList()+" n="+n);
    }
    return a;
}

/** Given an input bQQconditional=bQQi+mu*bQQc matrix
 * find it's eigenvectors and modify then to satisfy the constraint
 * {@literal <psi|bQQc|psi>=0} to obtain new {@literal |psi>} candidates in several
 * variants as per {@link FindPsiConstrainedSingleQuadratic0}. Then test all available
 * states to find the one with maximal {@literal <psi|bQQi|psi>}.
 * This will be a new iteration.
 */
private static class EVSelection {
    final double [] ev_bQQconditional;
    /** The value of eigenvalue corresponding to eigenvector selected.
     */
    final double lambda_i;
    /** Index of lambda. */
    final int ind_lambda_i;
    /** Next state iteration. */
    final GetMaxIState sNext;
    EVSelection(final int n,
		final double [] bQQconditional,
		final double [] bQQi,
		final double [] bQQc,
		final double [] ev_bQQc,
		final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
		final double eps){
	// solve eigenvalue problem
	ev_bQQconditional=M.EV.getEV(n,bQQconditional,true);

	// do it with a cycle as we need to propagate eigenstate index for diag.
	int indS=-1;
	GetMaxIState res=null;
	for(int s=0;s<n;s++){
	    final java.util.List<GetMaxIState> t=getSeqOfInitFunctions(
		n,java.util.List.of(EVXData.getPsiFromBasis(s,ev_bQQconditional,n)),
		bQQi,bQQc,ev_bQQc,eps);
	    if(t.size()>0 && (indS<0 || (t.get(0).i>res.i))){
		indS=s;
		res=t.get(0);
	    }
	}
	this.ind_lambda_i=indS;
	this.sNext=res;
	this.lambda_i=this.ind_lambda_i>=0?ev_bQQconditional[this.ind_lambda_i]:Double.NaN;
	if(res!=null && eps>0){
	    System.err.println("IstatesConditionalV2:EVSelection: ind_lambda_i="+ind_lambda_i+" selected res.i="+res.i+" orig_constraint="+res.bpsiInit.orig_constraint+" adjustment_type="+res.bpsi_M.adjustment_type);
	}
    }
}


/**
 * Given an input {@link FindPsiConstrainedSingleQuadratic0} class it takes
 * all it's adjusted {@link FindPsiConstrainedSingleQuadratic0#psiConstrained} states
 * and select the one with maximal {@literal <psi|I|psi>} among all the
 * provided {@literal |psi>} with the constraint {@literal <psi|bQQc|psi>=0}
 * satisfied.
 */
private static class GetMaxIState {
    final double i;
    /** The state of maximal {@literal <psi|I|psi>}.
     * It is {@code null} if nothing available.
     */
    final PsiConstrained bpsi_M;
    final FindPsiConstrainedSingleQuadratic0 bpsiInit;
    GetMaxIState(final int n,
		 final FindPsiConstrainedSingleQuadratic0 bpsiInit,
		 final double [] bQQi,
		 final double [] bQQc,
		 final double eps)
{
    this.bpsiInit=bpsiInit;
    double iMax=Double.NaN;
    PsiConstrained bpsi_M=null;
    // Find the state with maximal I among all candidates found.
    // It may be a better option to create a "herd" in a genetic algorithm style.
    for(PsiConstrained b:bpsiInit.psiConstrained){
	final double ib=TMatr.getAMatrA(b.bpsi,bQQi,n);
	if(Double.isFinite(ib) && (bpsi_M==null || ib>=iMax)){
	    iMax=ib;
	    bpsi_M=b;
	}
	if(eps>0){
	    // check constraint
	    UnitTests.checkDoubleRel(TMatr.mulVV(b.bpsi,b.bpsi),"<b.bpsi|b.bpsi>",1,"1",eps);
	    UnitTests.checkDoubleRel(TMatr.getAMatrA(b.bpsi,bQQc,n),"<b.bpsi|bQQc|b.bpsi>",0,"0",eps);
	}
    }
    this.i=iMax;
    this.bpsi_M=bpsi_M;
    if(eps>0){
	if(this.bpsi_M!=null){
	    System.err.println("IstatesConditionalV2:GetMaxIState: selected i="+this.i+" orig_constraint="+bpsiInit.orig_constraint+" adjustment_type="+this.bpsi_M.adjustment_type+" of total input states "+bpsiInit.psiConstrained.size());
	} else {
	    System.err.println("IstatesConditionalV2:GetMaxIState: no selected psi");
	}
    }
}
}

/** Calculate Lagrange multiplier.
 */
private static double getLagrangeMultiplier(
    final PsiConstrained bpsi_M,
    final double [] bQQi,
    final double [] bQQc){

    double mu=0;
    if(bpsi_M!=null){
	final int n=bpsi_M.bpsi.length;
	final double [] Cx=TMatr.getMatr_x_A(n,bQQc,bpsi_M.bpsi),
	    Ix=TMatr.getMatr_x_A(n,bQQi,bpsi_M.bpsi);
	final double cc=TMatr.mulVV(Cx,Cx);
	// Lagrange multiplier
	if(cc>0) mu=-TMatr.mulVV(Cx,Ix)/cc;
	if(!Double.isFinite(mu)) mu=0;
    }
    return mu;
}

static void printDiag(IstatesConditionalV2 m){
    // print some diag
    System.err.println("IstatesConditionalV2.main:I EV="+java.util.Arrays.toString(m.I.getEValues()));
    System.err.println("IstatesConditionalV2.main:i_M="+m.i_M+" lambda_i_M="+m.lambda_i_M+" psi_M="+java.util.Arrays.toString(m.psi_M)+" flag_solution_exists="+m.flag_solution_exists+" n="+m.I.neq);
}


/** Unit test. */
public static void main(String [] args){
    TestOptimizationPsi.runMultiTest("IstatesConditionalV2",15,600,
		 new java.util.Random(18301L),
		 1e-5,
		 testC-> {final var m=new IstatesConditionalV2(
		                testC.n,
				testC.QQt,
				testC.QQi,
				TMatr.sumAcB(testC.QQpi,-0.2,testC.QQi), // QQc
				testC.M,
				testC.eps);
			  printDiag(m);
			  }
		 );
}

}
