package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */

/** Store Lebesgue quadrature.
 *  @version $Id: LebesgueQuadratureJointDistribution.java,v 1.14 2022/11/08 09:32:37 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class LebesgueQuadratureJointDistribution {
    public final LebesgueQuadratureWithEVData ff,gg;
    public LebesgueQuadratureJointDistribution(final LebesgueQuadratureWithEVData ff,final LebesgueQuadratureWithEVData gg){
	this.ff=ff;
	this.gg=gg;
	// also need to make sure the matrices ff.QQ & gg.QQ are the same
	if(ff.f.length!=gg.f.length) throw new IllegalArgumentException();
	// disable because of NaN
	if(false && !java.util.Arrays.equals(ff.QQ,gg.QQ)) throw new IllegalArgumentException();
    }

    /**
     * @param DM Density matrix;
     * If DM=QQ the result is probability correlation;
     * if DM={@literal |1><1|} the result is Value correlation.
     * @return The matrix {@literal [i*n+j]=<ff.psi^{[i]}|gg.psi^{[j]}><ff.psi^{[i]}|DM|gg.psi^{[j]}>}.
     */
    public double [] getDensityMatrixCorrelation(final double [] DM)
    {
	return get_M_M_Correlation(ff.QQ,DM);
    }
    /** Most general case
     * matrix {@literal [i*n+j]=<ff.psi^{[i]}|M1|gg.psi^{[j]}>*<ff.psi^{[i]}|M2|gg.psi^{[j]}>}.
     */
    public double [] get_M_M_Correlation(final double [] M1,final double [] M2)
    {
	final int n=ff.f.length;
	if(M1.length!=M2.length || M1.length!=n*n) throw new IllegalArgumentException();
	final double [] m1=EVXData.getBasisABProjections(n,ff.ev,M1,gg.ev),
	    m2=EVXData.getBasisABProjections(n,ff.ev,M2,gg.ev);
	for(int i=0;i<n;i++) for(int j=0;j<n;j++){
		m1[i*n+j]*=m2[i*n+j];
	}
	return m1;
    }
    /** Calculate corr of f,g eigenvalues (correlation multiplied by eigenvalues).
     * @param corr The coor[i*n+j] is ff.f[i] and gg.f[j] correlation.
     * @return ff.f[i]*coor[i*n+j]*gg.f[j].
     */
    public double [] getFGWeightedWithCorrelation(final double [] corr)
    {
	final int n=ff.f.length;
	if(corr.length!=n*n) throw new IllegalArgumentException();
	final double [] fgcorr=new double [n*n];
	for(int i=0;i<n;i++) for(int j=0;j<n;j++){
		fgcorr[i*n+j]=ff.ev[i]*corr[i*n+j]*gg.ev[j];
	}
	return fgcorr;
    }

    /**
     * Equal to getDensityMatrixCorrelation with DM=QQ.
     * @return matrix {@literal [i*n+j]=<ff.psi^{[i]}|gg.psi^{[j]}>^2}
     */
    public double [] getProbabilityCorrelation()
    {
	if(false){
	    // same thing
	    return getDensityMatrixCorrelation(ff.QQ);
	}
	final int n=ff.f.length;
	//<ff.psi[i]|gg.psi[j]>= ff.ev * QQ * (gg.ev)^T
	final double [] mm=EVXData.getBasisABProjections(n,ff.ev,ff.QQ,gg.ev);
	//System.err.println("mm="+java.util.Arrays.toString(mm));
	//return TMatr.squareArrayElements(mm);
	for(int i=0;i<n;i++) for(int j=0;j<n;j++) mm[i*n+j]*=mm[i*n+j];
	return mm;
    }
    /**
     * Equal to getDensityMatrixCorrelation with {@literal DM=|1><1|}.
     * @return matrix {@literal [i*n+j]=<ff.psi^{[i]}><ff.psi^{[i]}|gg.psi^{[j]}><gg.psi^{[j]}>}
     */
    public double [] getValueCorrelation()
    {
	final int n=ff.f.length;
	if(false){
	    // same thing
	    final double [] QA=TMatr.scaleVectorInPlace(java.util.Arrays.copyOf(ff.QQ,n),1/ff.M.B.getD0(new double []{1},0));
	    return getDensityMatrixCorrelation(TMatr.getQQaa(QA));
	}
	//<ff.psi[i]|gg.psi[j]>= ff.ev * QQ * (gg.ev)^T
	final double [] mm=EVXData.getBasisABProjections(n,ff.ev,ff.QQ,gg.ev);
	for(int i=0;i<n;i++) for(int j=0;j<n;j++) mm[i*n+j]*=ff.psiaverage[i]*gg.psiaverage[j];
	return mm;
    }
}
