% $Id: LebesgueQuadratureJointDistribution.m,v 1.4 2020/02/11 09:49:24 mal Exp $

%/**
% * (C) Vladislav Malyshkin 2018
% * This file is under GPL version 3.
% *
% * Joint distribution from two Lebesgue quadratures.
% * and https://arxiv.org/abs/1807.08197
% *
% * @version $Id: LebesgueQuadratureJointDistribution.m,v 1.4 2020/02/11 09:49:24 mal Exp $
% * @author Vladislav Malyshkin mal@gromco.com
% */

classdef LebesgueQuadratureJointDistribution
  properties
    ff;gg;
  endproperties
  methods
  % Lebesgue quadrature joint distrinution constructor.
  function self = LebesgueQuadratureJointDistribution(ff,gg)
     n=size(ff.QQ,1);
     if(n!=size(ff.QQ,1)|| n!=size(gg.QQ,1))
       error (["invalid matrices size " mat2str(n)]);
     endif
     self.ff = ff;
     self.gg = gg;
  endfunction

  % ==================== getFGWeightedWithCorrelation
  function FG_weighted = getFGWeightedWithCorrelation(self,corr)
    n=size(self.ff.QQ,1);
    if(size(corr,1)!=n || size(corr,2)!=n)
      error (["invalid matrices size " mat2str(n)]);
    endif
    FG_weighted=zeros(n,n);
    for j=1:n
      for k=1:n
	FG_weighted(j,k)=self.ff.f(j)*corr(j,k)*self.gg.f(k);
      endfor
    endfor
  endfunction
  % ==================== END getFGWeightedWithCorrelation


  % ==================== get_M_M_Correlation
  % Corelation most general case (with two matrices).
  function Prob_fg = get_M_M_Correlation(self,M1,M2)
    n=size(self.ff.QQ,1);
    if(size(M1,1)!=n || size(M1,2)!=n || size(M2,1)!=n || size(M2,2)!=n)
      error (["invalid matrices size " mat2str(n)]);
    endif
    Prob_fg = LebesgueQuadratureWithEVData.getEVProj(self.ff.ev,M1,self.gg.ev) .* LebesgueQuadratureWithEVData.getEVProj(self.ff.ev,M2,self.gg.ev);
  endfunction
  % ==================== END get_M_M_Correlation

  % ==================== getDensityMatrixCorrelation
  function Prob_fg = getDensityMatrixCorrelation(self,DM)
    Prob_fg=self.get_M_M_Correlation(self.ff.QQ,DM);
  endfunction
  % ==================== END getDensityMatrixCorrelation

  % ==================== getProbabilityCorrelation
  % Prob_fg(j,k)=<psif[j]|psig[k]>^2
  % sum(Prob_fg)=n
  function Prob_fg = getProbabilityCorrelation(self)
    % ff.QQ & gg.QQ to be the same!
    MM=LebesgueQuadratureWithEVData.getEVProj(self.ff.ev,self.ff.QQ,self.gg.ev);
    Prob_fg=MM .* MM ;
  endfunction
  % ==================== END getProbabilityCorrelation

  % ==================== getValueCorrelation
  % Val_fg(j,k)=<psif[j]><psif[j]|psig[k]><psig[k]>
  % sum(Val_fg)=\int dmu=<1>
  function Prob_fg = getValueCorrelation(self)
    if(0)
      % method to calculate #1
      % assume first basis function is 1
      A=self.ff.QQ(:,1);
      Prob_fg=self.getDensityMatrixCorrelation(A*A');
    else
      % method to calculate #2
      % ff.QQ & gg.QQ to be the same!
      MM=LebesgueQuadratureWithEVData.getEVProj(self.ff.ev,self.ff.QQ,self.gg.ev);
      n=size(self.ff.QQ,1);
      Prob_fg=zeros(n,n);
      for j=1:n
	for k=1:n
	  Prob_fg(j,k)=self.ff.psiaverage(j)*MM(j,k)*self.gg.psiaverage(k);
	endfor
      endfor
    endif
  endfunction
  % ==================== END getValueCorrelation

  endmethods

endclassdef
