package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2018
 * This file is under GPL version 3.
 *
 */

/** Store Lebesgue quadrature.
 *  @version $Id: LebesgueQuadratureWithEVData.java,v 1.32 2022/07/08 15:40:49 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class LebesgueQuadratureWithEVData {
    public final double [] f,w,psiaverage,ev,QQf,QQ;
    public final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M;
    public LebesgueQuadratureWithEVData(final int n,final double [] QQf,final double [] QQ,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
    {
	if(QQ.length!=n*n || QQf.length!=n*n) throw new IllegalArgumentException();
	this.M=M;
	this.psiaverage=new double[n];
	this.ev=M.EV.getGEV_Generalized(n,QQf,QQ,true);
	final double [] C=M.B.getONE(); // make C=1 in basis.
	for(int k=0;k<n;k++) psiaverage[k]=TMatr.getAMatrB(C.length,C,0,n,ev,n+k*n,QQ,n);
	this.w=TMatr.squareArrayElements(psiaverage);
	// first n elements are eigenvalues, then n eigenvectors follow (n*n elements).
	this.f=java.util.Arrays.copyOf(ev,n);
	this.QQf=QQf;
	this.QQ=QQ;
    }

    /** Calculates psik at Qbasis_moments_to_check_PSI.
     */
    public double [] getPsiAt(final double [] Qbasis_moments_to_check_PSI)
    {
	final int n=f.length;
	if(Qbasis_moments_to_check_PSI.length!=n) throw new IllegalArgumentException();
	final double [] psikat=new double [n];
	for(int k=0;k<n;k++)
	{
	    psikat[k]=TMatr.mulVV(n,Qbasis_moments_to_check_PSI,0,ev,n+k*n);
	}
	return psikat;
    }
    /** For all eigenvectors calculate psik(x0).
     * @return psik(x0) vector, k=0..n-1.
     */
    public double [] getPsiAtX0(final double x0)
    {
	return getPsiAt(M.B.getX0Moments(f.length,x0));
    }

    /** Calculate Radon--Nikodym, Least Squares and 1/Christoffel function all at once.
     */
    public RNPointAccess getEvaluationAt(final double [] Qbasis_moments_to_check)
    {
	final int n=f.length;
	if(Qbasis_moments_to_check.length!=n) throw new IllegalArgumentException();
	double sflsq=0,sf=0,s=0;
	for(int k=0;k<n;k++)
	{
	    sflsq+=f[k]*TMatr.mulVV(n,Qbasis_moments_to_check,0,ev,n+k*n)*psiaverage[k];
	    final double wproj=Math.pow(TMatr.mulVV(n,Qbasis_moments_to_check,0,ev,n+k*n),2);
	    s+=wproj;
	    sf+=f[k]*wproj;
	}
	final double _sflsq=sflsq,_sf=sf,_s=s;
	return new RNPointAccess(){
	    public double getRN(){return _sf/_s;}
	    public double getChristoffelOatX(){return _s;}
	    public double getLS(){return _sflsq;}
	};
    }
    /**
     * Calculate Radon--Nikodym, Least Squares and 1/Christoffel function at x0.
     * The 1/Christoffel is the same thing as M.B.getD0(M.getChristoffelOFunctionFromMatrix(n,QQ),x0).
     */
    public RNPointAccess getEvaluationAtX0(final double x0)
    {
	return getEvaluationAt(M.B.getX0Moments(f.length,x0));
    }
    /**
     * Calculate MM matrix in all eigenstates.
     * @param MM The matrix.
     * @return {@literal <psik|MM|psik>}, k=0..n-1
     */
    public double [] getMatrixDiagonalsInPsiStates(final double [] MM)
    {
	final int n=f.length;
	return TMatr.getMatrixDiagElementsInBasis(n,ev,MM);
    }
    /**
     * Calculate lambda*MM matrix in all eigenstates.
     * @param MM The matrix.
     * @return {@literal lambda[k]*<psik|MM|psik>}, k=0..n-1.
     */
    public double [] getLambdaMatrixDiagonalsInPsiStates(final double [] MM)
    {
	return TMatr.getVAB(getMatrixDiagonalsInPsiStates(MM),f);
    }
    /**
     * Calculate sum(lambda*MM) in all eigenstates,
     * typically used for spur calculation, the answer actually is: Spur(G^{-1}*QQf*G^{-1}*MM).
     * @param MM The matrix.
     * @return {@literal sum(lambda[k]*<psik|MM|psik>)}.
     */
    public double sumLambdaMatrixDiagonalsInPsiStates(final double [] MM)
    {
	return TMatr.sumElements(getLambdaMatrixDiagonalsInPsiStates(MM));
    }
}
