% $Id: LebesgueQuadratureWithEVData.m,v 1.56 2020/02/11 10:50:49 mal Exp $

%/**
% * (C) Vladislav Malyshkin 2018
% * This file is under GPL version 3.
% *
% * Slow and simple test of Lebesgue quadratures.
% *
% * Note that Lebesgue quadratures for f(x)=x is Gauss quadrature.
% * See the paper https://arxiv.org/abs/1807.06007 for Lebesgue quadratures description.
% *
% * @version $Id: LebesgueQuadratureWithEVData.m,v 1.56 2020/02/11 10:50:49 mal Exp $
% * @author Vladislav Malyshkin mal@gromco.com
% */

classdef LebesgueQuadratureWithEVData
  properties
    f;w;psiaverage;ev;QQf;QQ;basisfun;
  endproperties
  methods
  % Lebesgue quadrature constructor.
  % See https://arxiv.org/abs/1807.06007 for details.
  function self = LebesgueQuadratureWithEVData(QQf,QQ,basisfun)
     n=size(QQ,1);
     if(n!=size(QQ,1)|| n!=size(QQ,2) || n!=size(QQf,1)|| n!=size(QQf,2))
       error (["invalid matrices size " mat2str(n)]);
     endif
     self.QQ = QQ;
     self.QQf = QQf;
     self.basisfun=basisfun;

     [lambda,V] = LebesgueQuadratureWithEVData.getGeneralizedEV(QQf,QQ);
     self.psiaverage=zeros(n,1);
     self.f=lambda;
     self.ev=V;
     c=basisfun(0,0.0);
     if (c!=1 || basisfun(0,1.2223)!=c || basisfun(0,0.44223)!=c || basisfun(0,-0.73)!=c)
       error (["First (0-th) basis function must be equal to 1"]);
     endif
     for k=1:n
       self.psiaverage(k)=(QQ(1,:)*V(:,k))/c; % if 0-th basis finction is not 1.
     endfor
     self.w = self.psiaverage .* self.psiaverage;
  endfunction

  % ==================== getRNat
  % Calculate Radon-Nikodym for given qk moments.
  function RN = getRNat(self,q)
    n=size(q,1);
    if(n!=size(self.f,1)|| 1!=size(self.f,2) ||
       n!=size(q,1) || 1!=size(q,2) ||
       n!=size(self.ev,1) || n!=size(self.ev,2))
      error (["invalid matrices size " mat2str(n)]);
    endif
    s=0;sf=0;
    for k=1:n
      wproj=power(sum(self.ev(:,k) .* q),2);
      s+=wproj;
      sf+=self.f(k)*wproj;
    endfor
    RN=sf/s;
  endfunction % getRNat
  % ==================== END getRNat

  % ==================== getX0Moments
  function qk = getX0Moments(self,x0)
    n=size(self.QQ,1);
    qk=zeros(n,1);
    for k=1:n
      qk(k)=self.basisfun(k-1,x0);
    endfor
  endfunction
  % ==================== END getX0Moments

  % ==================== getRNatX0
  % Radon-Nikodym eigenvalues lambda weighted with (psik(x0))^2
  % https://arxiv.org/abs/1510.05510 page 37 (D4):
  % RN(x0)=sum_k (lambda(k)(psik(x0))^2)/ sum_k (psik(x0))^2
  function RN = getRNatX0(self,x0)
    RN=self.getRNat(self.getX0Moments(x0));
  endfunction
  % ==================== END getRNatX0

  % ==================== getLSatX0
  % Least squares answer: https://arxiv.org/abs/1510.05510 page 37 (D3):
  % LS(x0)=sum_k (lambda(k)<psik>psik(x0))
  function LS = getLSatX0(self,x0)
    LS=sum(self.f .* self.psiaverage .* self.getPSIatX0(x0) );
  endfunction
  % ==================== END getLSatX0

  % ==================== getPSIatX0
  % Calculate all psik(x0)
  % Note that getRNatX0(x0) is equal to sum(PSIx0 .* f .* PSIx0)/sum(PSIx0 .* PSIx0)
  % and getLSatX0(self,x0) is equal to sum(psiaverage .* f .* PSIx0)
  function PSIx0 = getPSIatX0(self,x0)
    q=self.getX0Moments(x0);
    n=size(self.QQ,1);
    PSIx0=zeros(n,1);
    for k=1:n
      PSIx0(k)=sum(self.ev(:,k) .* q);
    endfor
  endfunction
  % ==================== END getPSIatX0

  % ==================== getChristoffelOatX0
  function Kx0 = getChristoffelOatX0(self,x0)
    PSIx0 = self.getPSIatX0(x0);
    % 1/Christoffel functions is getRNatX0 denominator, the nominator is sum(PSIx0 .* self.f .* PSIx0)
    Kx0=sum(PSIx0 .* PSIx0);
  endfunction
  % ==================== END getChristoffelOatX0

  % ==================== getEvaluationAtX0
  % To simplify unit test
  function rneval = getEvaluationAtX0(self,x0)
    rneval=RNPointAccess(self.getLSatX0(x0),self.getRNatX0(x0),self.getChristoffelOatX0(x0));
  endfunction
  % ==================== END getEvaluationAtX0

  % ==================== getMatrixDiagonalsInPsiStates
  % Calculate MM matrix diagonal elements in eigenstates.
  function psikMMpsik = getMatrixDiagonalsInPsiStates(self,MM)
    n=size(self.QQ,1);
    if(n!=size(MM,1)|| n!=size(MM,2))
      error (["invalid matrix size " mat2str(n)]);
    endif
    psikMMpsik=zeros(n,1);
    for k=1:n
      psikMMpsik(k)=self.ev(:,k)' * MM * self.ev(:,k);
    endfor
  endfunction
  % ==================== END getMatrixDiagonalsInPsiStates

  % ==================== getLambdaMatrixDiagonalsInPsiStates
  % Calculate lambda * MM matrix diagonal elements in eigenstates.
  function LambdapsikMMpsik = getLambdaMatrixDiagonalsInPsiStates(self,MM)
    LambdapsikMMpsik=self.f .* self.getMatrixDiagonalsInPsiStates(MM);
  endfunction
  % ==================== END getLambdaMatrixDiagonalsInPsiStates
  endmethods

  methods (Static = true)

    %=======================================
    % Some useful basis functions
    % Monomials basis finction num = 0 .. n-1
    function y = basisfun_monomials (num,x)
      if(num<0 || !(floor(num)==num))
	error ( ["num=" mat2str(num) " is invalid"] );
      endif
      y = power(x,num);
    endfunction

    % Chebyshev polynomial basis function num = 0 .. n-1
    function y = basisfun_chebyshev (num,x)
      if(num<0 || !(floor(num)==num))
	error ( ["num=" mat2str(num) " is invalid"] );
      endif
      if (num==0)
	y=x;
	y(:)=1;
      else
	b=1;s=x;
	for i=2:num
	  s=-b+2*(x .* (b=s));
	endfor
	y=s;
      endif
    endfunction


    % Legendre polynomial basis function num = 0 .. n-1
    function y = basisfun_legendre (num,x)
      if(num<0 || !(floor(num)==num))
	error ( ["num=" mat2str(num) " is invalid"] );
      endif
      b=0;s=x;s(:)=1;
      for k=1:num
	s=(-(k-1)*b+(2*k-1)*(x .* (b=s)))/k;
      endfor
      y=s;
    endfunction

    % Fourie basis functions num = 0 .. n-1
    % numeration:
    % 0 -> 1 ; 1 -> sin(pi*x) ; 2 -> cos(pi*x) ; 3 -> sin(2*pi*x) ; 4 -> cos(2*pi*x)
    function y = basisfun_fourie (num,x)
      if(num<0 || !(floor(num)==num))
	error ( ["num=" mat2str(num) " is invalid"] );
      endif
      if (num==0)
	y=x;
	y(:)=1;
      else
	xx=x*(1+idivide (num-1, 2, "floor"))*pi;
	if (0==mod(num+1,2))
	  y=sin(xx);
	else
	  y=cos(xx);
	endif
      endif
    endfunction

  % ==================== getGaussQuadratureLegendreNodesAndWeights.
  % Obtain Legendre measure Gauss quadrature nodes and weights, [-1:1] interval, dmu=dx.
  % This is usage demonstration example.
  function [x,w]=getGaussQuadratureLegendreNodesAndWeights(n)
    % octave bug http://savannah.gnu.org/bugs/?46659 workaround
    % basisfun=@LebesgueQuadratureWithEVData.basisfun_chebyshev
    basisfun=@(num,x) (LebesgueQuadratureWithEVData.basisfun_chebyshev(num,x));
    QQ=LebesgueQuadratureWithEVData.getMatrixByAnalyticalIntegration(basisfun,@(x) (1),-1,1,n);
    XQQ=LebesgueQuadratureWithEVData.getMatrixByAnalyticalIntegration(basisfun,@(x) (x),-1,1,n);
    lq=LebesgueQuadratureWithEVData(XQQ,QQ,basisfun);
    x=lq.f;
    w=lq.w;
  endfunction
  % ==================== END getGaussQuadratureLegendreNodesAndWeights.

  % ==================== getGeneralizedEV
  % A function to solve generalized eigenvalues problem.
  function [lambda,V] =getGeneralizedEV(AM,BM)
  n=size(AM,1);
  if(n!=size(AM,1)|| n!=size(AM,2) || n!=size(BM,1)|| n!=size(BM,2))
    error (["getGeneralizedEV:invalid matrices size " mat2str(n)]);
  endif

  % generalized eigenvalues %lambda=qz(AM,BM);
  [AA, BB, Q, Z, V, W, lambda] = qz (AM, BM);
  % normalize eigenvectors, octave does not normalize
  for k=1:n
    vk=V(:,k);
    % octave does not normalize
    V(:,k)=vk/sqrt(vk' * BM *vk);
  endfor
  [lambda,V] = LebesgueQuadratureWithEVData.sortEVAccordingLambdas(lambda,V);
  LebesgueQuadratureWithEVData.selftestEV(lambda,V,AM,BM,LebesgueQuadratureWithEVData.getEpsSelftest());
  endfunction
  % ==================== END getGeneralizedEV

  % ==================== sortEVAccordingLambdas
  function [lambda,V] = sortEVAccordingLambdas(lambdai,Vi)
  n=size(lambdai,1);
  if(n!=size(lambdai,1)|| 1!=size(lambdai,2) ||
     n!=size(Vi,1) || n!=size(Vi,2))
    error (["invalid matrices size " mat2str(n)]);
  endif
  % octave does not sort
  s=sortrows([lambdai,Vi']);
  lambda=s(:,1);
  V=s(:,2:n+1)';
  endfunction
  % ==================== END sortEVAccordingLambdas

  % ==================== getEVProj
  % Project Vf and Vg eigenvalues, MM(j,k)=<psif[j]|psig[k]> is unitary.
  function MM = getEVProj(Vf,QQ,Vg)
    n=size(QQ,1);
    if(n!=size(QQ,1)|| n!=size(QQ,2) || n!=size(Vf,1)|| n!=size(Vf,2) || n!=size(Vg,1)|| n!=size(Vg,2))
      error (["invalid matrices size " mat2str(n)]);
    endif
    MM=zeros(n,n);
    for j=1:n
      for k=1:n
	MM(j,k)=Vf(:,j)'*QQ*Vg(:,k);
      endfor
    endfor
    LebesgueQuadratureWithEVData.matchVectors(MM(:),"MM",(Vf'*QQ*Vg)(:),"Vf'*QQ*Vg",LebesgueQuadratureWithEVData.getEpsSelftest(),0);
  endfunction
  % ==================== END getEVProj


  % ==================== getMatrixBySampleSum
  % A function to calculate <Qj|Qk> by sum(w(m)Qj(x(m))Qk(x(m))), m=1:Nsample ; j,k=0:n-1
  % For <Qj|f|Qk> matrix use the (f .* w) as argument w.
  % no multiplication operator used, direct nxn multiplication
  function QQ = getMatrixBySampleSum (basisfun,w,x,n)
    QQ=zeros(n,n);
    for j=1:n
      for k=1:j
	QQ(j,k)=QQ(k,j)=sum(w .* basisfun(j-1,x) .* basisfun(k-1,x) );
      endfor
    endfor
  endfunction
  % ==================== END getMatrixBySampleSum

% ==================== getMatrixByAnalyticalIntegration
% Obtain the <Qj Qk> moments of analytically known function fun(x)
function QQ = getMatrixByAnalyticalIntegration(basisfun,fun,a,b,n)
  QQ=zeros(n,n);
  % create <fun Qj Qk> matrix directly, no multiplication operator used
  for j=1:n
    for k=1:j
      QQ(j,k)=QQ(k,j)=quad(@(x) (fun(x) .* basisfun(j-1,x) .* basisfun(k-1,x)),a,b);
    endfor
  endfor
endfunction
% ==================== END getMatrixByAnalyticalIntegration

  % ==================== selftestEV
  function selftestEV(lambda,V,QQf,QQ,xxeps)
  if (xxeps>=0)
    n=size(QQ,1);
    if(n!=size(QQ,1)|| n!=size(QQ,2) || n!=size(QQf,1)|| n!=size(QQf,2))
      error (["invalid matrices size " mat2str(n)]);
    endif
    for j=1:n
      for k=1:n
	fjk=V(:,j)'*QQf*V(:,k);
	bjk=V(:,j)'*QQ*V(:,k);
	if(j!=k)
	  if(!( abs(fjk)<=xxeps ))
            error(["QQf offdiag is incorrect " mat2str([j,k,fjk])]);
	  endif
	  if(!( abs(bjk)<=xxeps ))
            error(["QQ offdiag is incorrect " mat2str([j,k,bjk])]);
	  endif
	else
	  if(!( abs(fjk-lambda(k))<=xxeps ))
            error(["QQf diag is incorrect " mat2str([j,k,fjk,lambda(k)])]);
	  endif
	  if(!( abs(bjk-1)<=xxeps ))
            error(["QQ diag is incorrect " mat2str([j,k,bjk,1.0])]);
	  endif
	endif
	if LebesgueQuadratureWithEVData.shouldPrintDebug() %================== DEBUG
          if(j==k)
            disp(["getGeneralizedEV:Lambda(" mat2str(j) ") " mat2str(lambda(j))]);
          endif
          disp(["getGeneralizedEV:<Vj|AM|Vk> matr element: " mat2str([j,k,V(:,j)'*QQf*V(:,k)])]);
          disp(["getGeneralizedEV:<Vj|BM|Vk> matr element: " mat2str([j,k,V(:,j)'*QQ*V(:,k)])]);
	endif                 %================== END DEBUG
      endfor
    endfor
    disp(["lambda EV test OK n=" mat2str(n) " xxeps=" mat2str(xxeps)]);
  endif
  endfunction
  % ==================== END selftestEV

  % ==================== matchVectors
  % Check two 1D matrices to match
  function matchVectors(m1,m1name,m2,m2name,xxeps,flag_fail_on_error)
    if(length(m1)!=length(m2))
      error(["matchVectors" m1name " and " m2name " size mismatch"]);
    endif
    if(xxeps>=0)
      for k=1:length(m1)
	if(!(abs(m1(k)-m2(k))<=xxeps))
	    disp(["-----matchVectors " m1name "(" mat2str(k) ")=" mat2str(m1(k)) " and " m2name "(" mat2str(k) ")=" mat2str(m2(k)) " diff=" mat2str(abs(m1(k)-m2(k))) " xxeps=" mat2str(xxeps)]);
	    if(flag_fail_on_error)
	      error("FAIL");
	    endif
	endif
      endfor
    endif
  endfunction
  % ==================== END matchVectors

  function xxeps = getEpsSelftest()
    xxeps=1e-6; % set to negative to disable selftest
  endfunction
  function y = shouldPrintDebug()
    y=0; % disabled by default
  endfunction
  endmethods

endclassdef
