% $Id: LebesgueQuadratures_call_java_example.m,v 1.74 2020/02/11 09:47:59 mal Exp $

%/**
% * (C) Vladislav Malyshkin 2018
% * This file is under GPL version 3.
% *
% * An example of octave/matlab interface to java code, performing:
% * 1. polynomial calculations in arbitrary basis
% * 2. Gauss--type quadrature calculation.
% * 3. Polynomial roots.
% * 4. Lebesgue quadratures calculations.
% * 5. Radon--Nikodim in x-localized states.
% *
% * Note that Lebesgue quadratures for f(x)=x is Gauss quadrature.
% * See the paper https://arxiv.org/abs/1807.06007 for details.
% *
% * @version $Id: LebesgueQuadratures_call_java_example.m,v 1.74 2020/02/11 09:47:59 mal Exp $
% * @author Vladislav Malyshkin mal@gromco.com
% */
%THIS FILE CALLS java implementation polynomials_abasis.jar FOR DEMONSTRATION.

% To install and run:
%
% 0. Download, save, unzip the file:
% http://www.ioffe.ru/LNEPS/malyshkin/code_polynomials_quadratures.zip
%
% 1. Compile java and create .jar
% javac -g com/polytechnik/*/*.java
% jar 0cvf polynomials_abasis.jar com/polytechnik/*/*.class
%
% 2. add "polynomials_abasis.jar" library to octave path (once per session)
%>> javaaddpath (["/path/to/dir/FIXME", "/polynomials_abasis.jar"]);
%
% 3. run the code below to test:
%
format long;
xxeps=1e-6 % selftest, eps is often clobbered, use xxeps
% Obtain a sample: x,w,f,g are arrays of Nsample=1000 elements
[x,w,f,g,sumW,Nsample,xrn]=LebesgueQuadratures_TestHelpers.getSampleWeUseForTests(1);
%
% Quadrature dimension
% Can be large if x-interval is properly scaled to selected basis, otherwise use 5.
n=5


% The bases to try (the six bases implemented in "polynomials_abasis.jar").
% For best numerical stability x interval should be scaled to basis interval.
% e.g. Laguerre has [0:inf] interval with exp(-x) weight, is not so good for x=[-1:1] in this test,
% what can produce numerical instability.
names={"com.polytechnik.utils.OrthogonalPolynomialsHermiteEBasis" ,
       "com.polytechnik.utils.OrthogonalPolynomialsChebyshevBasis" ,
       "com.polytechnik.utils.OrthogonalPolynomialsLegendreBasis" ,
       "com.polytechnik.utils.OrthogonalPolynomialsLegendreShiftedBasis" ,
       "com.polytechnik.utils.OrthogonalPolynomialsHermiteEBasis" ,
       "com.polytechnik.utils.OrthogonalPolynomialsLaguerreBasis" ,
       "com.polytechnik.utils.OrthogonalPolynomialsMonomialsBasis" };

% Try them all, the results have to be identical.
for o=1:length(names) % try all basises available
 javaclassname=names{o};
 % fix java classpath if fails:
 % javaaddpath (["/path/to/dir/FIXME", "/polynomials_abasis.jar"]);
 M=javaObject(javaclassname);
 flag_fail_on_error=1;
 if(0!=strcmp(javaclassname,"com.polytechnik.utils.OrthogonalPolynomialsLaguerreBasis"))
   disp(["Set flag_fail_on_error for " javaclassname " class "]);
   flag_fail_on_error=0 % LaguerreBasis is not very good on this [-1:1] interval
 endif
 %
 % M.B -> basis functions operations (multiply, divide, integrate, etc.): class com.polytechnik.utils.BasisPolynomials
 % M -> Orthogonal polynomials/Quadratures/Lebesgue/Radon--Nikodym/etc. that are constructed in M.B basis: class com.polytechnik.utils.OrthogonalPolynomialsABasis
 %
 % This way one can easily construct Gauss/Lebesgue quadratures or calculate Radon--Nikodym derivative
 % using the moments in Qk(x) basis, not only in usual x^k.
 %
 % Below the moments=<Qk(x)> and fmoments=<fQk(x)> are such generalized moments.
 % The transition from x^k to Qk(x) basis give mathematically identical result,
 % but greatly increase numerical stability of calculations.
 %
 disp("");
 disp(["name=" javaclassname ]);
 %
 % Create <Qk> moments from (x,w)-sample in Qk basis: obtain them as:
 % <Qk>=sum(w(m)Qk(x(m))), m=1:Nsample ; k=0:2n-1
 moments=M.B.getMomentsFromSample(2*n,x,w);

 % ========== Example 1
 % Christoffel function is 1/K(x)
 K=M.getChristoffelOFunctionFromMoments(n,moments); % the coefficients are Qk(x) expansion coefs
 % n-th orthogonal polynomial
 P=M.getSingleOrthogonalPolynomial(n,moments); % the coefficients are Qk(x) expansion coefs
 % Multiplication and synthetic division in basis, see
 % OrthogonalPolynomialsABasis.java for other available methods.
 PmulP=M.B.mult2Pol(P,P); % multiplication in Qk basis.
 disp(["Java:K="  mat2str(K) " multiplication P*P=" mat2str(PmulP)]);
 % synthetic division
 sdivKP=M.B.sdiv(K,P);
 disp(["Java:K/P: quotent=" mat2str(sdivKP.q) " remainder=" mat2str(sdivKP.r) " K=qp+r=" mat2str(M.B.mult2Pol(P,sdivKP.q)+resize(sdivKP.r,length(K),1)) ]);
 % synthetic division by x-0.41
 sdivKP1=M.B.sdiv1(K,0.41);

 %
 % ========== Example 2
 % Gauss quadrature
 q=M.getGaussQuadratureFromMoments(n,moments);
 disp(["Java:Gauss:X=" mat2str(q.x) " W=" mat2str(q.w) " sum=" mat2str([sum(q.w),sumW])]);

 % ========== Example 3
 % Polynomial roots in Qk basis. K(x) has no real roots, P(x) has all roots real:
 Kroots=M.getPolynomialRootsFinderInBasis().findRoots(K);
 Proots=M.getPolynomialRootsFinderInBasis().findRoots(P);
 % not sorted
 disp(["Java:K roots: re=" mat2str(Kroots.re_root) " im=" mat2str(Kroots.im_root)]);
 disp(["Java:P roots (to be equal to Gauss nodes): re=" mat2str(Proots.re_root) " im=" mat2str(Proots.im_root)]);

 % ========== Example 4
 % Gauss quadrature weight is equal to 1/K(x)
 aqw=q.w;aqx=q.x; % convert java arrays for easy octave access
 for k=1:n
 if(!(abs(aqw(k)-1/M.B.getD0(K,aqx(k)))<=xxeps))
   disp(["----Christoffel Function mismatch weight, diff=" mat2str([aqw(k)-1/M.B.getD0(K,aqx(k)) ; aqw(k) ; 1/M.B.getD0(K,aqx(k))]) " this often happen with not properly scaled basis interval or too small xxesp=" mat2str(xxeps)]);
 endif
 endfor

 % ========== Example 5
 % a & b recurrence coefficients for orthogonal polynomials built from sampled moments in Qk basis.
 % Jacobi matrix is: xp_{k}=a_{k+1}*p_{k+1}+b_{k}p_{k}+a_{k}p_{k-1}
 ab=M.getAB(M.getOrthogonalPolynomialsFirstKind(n,moments));
 disp(["Java:A=" mat2str(ab.getAllAk())]);
 disp(["Java:B=" mat2str(ab.getAllBk())]);

 % ========== Example 6
 % Lebesgue quadrature for f(x)=x is Gauss quadrature:
 qgl=M.getLebesgueQuadratureFromMoments(n,M.B.getXQkMomentsFromQkMoments(moments),moments);
 disp(["Java:Gauss from Lebesgue: X=" mat2str(qgl.f) " W=" mat2str(qgl.w) " sum=" mat2str([sum(qgl.w),sumW])]);

 % ========== Example 7
 % Lebesgue quadrature general case.
 % Create <fQk> moments from (f,x,w)-sample in Qk basis: obtain them as:
 % <fQk(x)>=sum(f(m)w(m)Qk(x(m))), m=1:Nsample ; k=0:2n-1
 fmoments=M.B.getMomentsFromSample(length(moments),x,f .* w);
 % calculate the quadrature
 qlf=M.getLebesgueQuadratureFromMoments(n,fmoments,moments);
 disp(["Java:Lebesgue: F=" mat2str(qlf.f) " W=" mat2str(qlf.w) " sum=" mat2str([sum(qlf.w),sumW])]);

 % ========== Example 8 (second Lebesgue quadrature)
 gmoments=M.B.getMomentsFromSample(length(moments),x,g .* w);
 qlg=M.getLebesgueQuadratureFromMoments(n,gmoments,moments);
 disp(["Java:Lebesgue: G=" mat2str(qlg.f) " W=" mat2str(qlg.w) " sum=" mat2str([sum(qlg.w),sumW])]);

 % ========== Example 9
 % value- and probability- correlations
 jcorr_fg=javaObject("com.polytechnik.utils.LebesgueQuadratureJointDistribution",qlf,qlg);
 disp(["Java:Value Correlation(f,g) sum=" mat2str([sum(jcorr_fg.getValueCorrelation()), sumW])]);
 jcorr_fg.getValueCorrelation()
 disp(["Java:Probability Correlation(f,g) sum=" mat2str([sum(jcorr_fg.getProbabilityCorrelation()), n])]);
 jcorr_fg.getProbabilityCorrelation()

 % ========== Example 10: Radon-Nikodym & Nevai operator.
 % Radom--Nikodym x-localized example (without spectrum):
 rn=javaObject("com.polytechnik.utils.NevaiOperator").getNevaiOperator(n,moments,fmoments,M.B);
 % Try Radon--Nikodym at a few different x (from xrn array):
 for x0=xrn
   % instead of localized moments M.B.getX0Moments(n,xx) a Qk distribution can be optionally used.
   % Localized Qk(x0) (single support point measure) give the same as rn.getF(x0):
     disp(["Java:(EV): at x=" mat2str(x0) " RN f=" mat2str([rn.getF(x0),qlf.getEvaluationAtX0(x0).getRN()]) " LS f=" mat2str(qlf.getEvaluationAtX0(x0).getLS())]);
 endfor
 disp(["RN & LS for these x,f to match the results of LebesgueQuadrature.m"]);

% ========== Example 11: Similar calculations (without java) using LebesgueQuadratureWithEVData.m octave class
 ocorr_fg=LebesgueQuadratures_TestHelpers.printLebesgueOctaveNatively(x,f,g,w,n,xrn);
 % match java to octave
 LebesgueQuadratures_TestHelpers.testJavaAndOctaveJointDistributionsMatch(javaclassname,jcorr_fg,ocorr_fg,xrn,xxeps,flag_fail_on_error);

endfor % END try all basises available
disp(["All tests OK, plotting a chart of Christoffel function"]);

q=M.getGaussQuadratureFromMoments(n,moments);
% plot Christoffel function
% qlf.getChristoffelOatX0(xa(x)) and M.B.getD0(K,xa(k)) is the same
LebesgueQuadratures_TestHelpers.printChristoffelChart(qlf,-1.2,0.01,1.2,q.x);
disp([" n=" mat2str(n) " moments(0)=" mat2str(moments(1+0)) " to match the weight; Nodes and Weights: q.x=" mat2str(q.x) " q.w=" mat2str(q.w)]);



