% $Id: LebesgueQuadratures_selftest.m,v 1.64 2020/02/11 11:02:16 mal Exp $

%/**
% * (C) Vladislav Malyshkin 2018
% * This file is under GPL version 3.
% *
% * Slow and simple test of Lebesgue quadratures.
% * A very small subset of java implementation "polynomials_abasis.jar"
% * is actually implemented in octave natively.
% *
% * There are 3 tests:
% * 1. printExampleRungeFunctuonInterpolation()
% *    interpolate Runge function=1/(1+25*x*x) by Radon--Nikodym and Least Squares,
% *    the Fig. 1 of https://arxiv.org/abs/1511.01887
% *
% * 2. printExampleQuadraturesCompareToGauss()
% *    Obtain Gauss quadrature from Lebesgue quadratures by putting f(x)=x.
% *    see https://arxiv.org/abs/1807.06007 for details.
% *
% * 3. printExampleLebesgueQuadratures()
% *    Calculate and test two Lebesgue quadratures,
% *    obtain Value- and Probability- correlation.
% *    see https://arxiv.org/abs/1807.08197 for details.
% *
% * @version $Id: LebesgueQuadratures_selftest.m,v 1.64 2020/02/11 11:02:16 mal Exp $
% * @author Vladislav Malyshkin mal@gromco.com
% */

%THIS FILE CALLS  LebesgueQuadratureWithEVData.m FOR DEMONSTRATION.

ss=1; % filename kludge
format long;

% Chebyshev measure
function y = omega_chebyshev (x)
    y= 1 / sqrt (1.0 - x .* x);
endfunction

% Chebyshev 2 kind measure
function y = omega_chebyshev2 (x)
    y= sqrt (1.0 - x .* x);
endfunction

% Legendre measure
function y = omega_legendre (x)
  y= x;
  y(:)=1;
endfunction

% ==================== getGaussQuadratureByIntegration
% Obtains Gauss quadrature from Lebesgue quadratures.
function [x,w] = getGaussQuadratureByIntegration(basisfun,omegafun,a,b,n)
  % we can calculate the matrices by integration
  QQ=LebesgueQuadratureWithEVData.getMatrixByAnalyticalIntegration(basisfun,@(x) (omegafun(x)),a,b,n);
  XQQ=LebesgueQuadratureWithEVData.getMatrixByAnalyticalIntegration(basisfun,@(x) (x .* omegafun(x)),a,b,n);
  lq=LebesgueQuadratureWithEVData(XQQ,QQ,basisfun);
  x=lq.f;
  w=lq.w;
endfunction
% ==================== END getGaussQuadratureByIntegration


% ========== Quadratures Example printExampleQuadraturesCompareToGauss()
function printExampleQuadraturesCompareToGauss()
  n=5;
  % octave bug http://savannah.gnu.org/bugs/?46659 workaround
  % basisfun=@LebesgueQuadratureWithEVData.basisfun_chebyshev
    basisfun=@(num,x) (LebesgueQuadratureWithEVData.basisfun_chebyshev(num,x));
    basisfun_monomials=@(num,x) (LebesgueQuadratureWithEVData.basisfun_monomials(num,x));
  % ========== Example 1
  % Gauss quadrature for dmu=sqrt(1-x*x)dx measure:
  [x,w] = getGaussQuadratureByIntegration(basisfun, @omega_chebyshev2,-1,1,n)
  sum(w)
  % exact roots of chebyshev2
  disp (["exact x for chebyshev2"]);
  cos((1 : n)*pi/(n+1))'

  % ========== Example 2
  % Gauss quadrature for dmu=dx measure:
  % the results to be identical for two different bases
  [x,w] = getGaussQuadratureByIntegration(basisfun_monomials, @omega_legendre,-1,1,n)
  sum(w)
  [x,w] = getGaussQuadratureByIntegration(basisfun, @omega_legendre,-1,1,n)
  sum(w)
  % call buit-in example
  disp(["Built-in test"]);
  [x,w] = LebesgueQuadratureWithEVData.getGaussQuadratureLegendreNodesAndWeights(n)
  sum(w)

  % ========== Example 3
  % Lebesgue quadrature for sin(x)dx
  % Obtain matrices by integration. second option is LebesgueQuadratureWithEVData.getMatrixBySampleSum()
  QQf=LebesgueQuadratureWithEVData.getMatrixByAnalyticalIntegration(basisfun,@(x) (sin(x).*omega_legendre(x)),-1,1,n);
  QQ=LebesgueQuadratureWithEVData.getMatrixByAnalyticalIntegration(basisfun,@(x) (omega_legendre(x)),-1,1,n);
  % call LebesgueQuadratureWithEVData that buld the quadrature
  ql=LebesgueQuadratureWithEVData(QQf,QQ,basisfun);
  disp(["Lebesgue f=sinx: f=" mat2str(ql.f) " w=" mat2str(ql.w) " sumw=" mat2str(sum(ql.w))]);
endfunction
% ========== END printExampleQuadraturesCompareToGauss()

% ========== Print Value and Probability correlation for Lebesgue quadratures
function [qlf,qlg]=printExampleLebesgueQuadratures()
n=5;
% random sample: x,w,f,g are arrays of Nsample elements
[x,w,f,g,sumW,Nsample,xrn]=LebesgueQuadratures_TestHelpers.getSampleWeUseForTests(1);
% calculate and print quadratures
corr_fg=LebesgueQuadratures_TestHelpers.printLebesgueOctaveNatively(x,f,g,w,n,xrn);
qlf=corr_fg.ff;qlg=corr_fg.gg;
endfunction % printExample2()
% ========== END Print Value and Probability correlation for Lebesgue quadratures

% ========== Runge function test, see Fig. 1 of https://arxiv.org/abs/1511.01887
function [xorig,forig,fRN,fLS]=printExampleRungeFunctuonInterpolation(n,nPointsIntegration,nPointsOutput)
[xorig,forig,fRN,fLS,ql]=LebesgueQuadratures_TestHelpers.getFunRNandLSinterpolated(n,@(x) 1/(1+25*x*x),-1,1,@omega_legendre,nPointsIntegration,nPointsOutput);
disp([ "Runge Spectrum\t" mat2str(ql.f) ]);
disp([ "Runge Weights\t" mat2str(ql.w) " sum W=" mat2str(sum(ql.w))]);
disp([ "|X"  "\t" "Runge(x)" "\t" "ARN(x)" "\t" "ALS(x)" ]);
for k=1:nPointsOutput
disp([ mat2str(xorig(k)) "\t" mat2str(forig(k)) "\t" mat2str(fRN(k)) "\t" mat2str(fLS(k)) ]);
endfor
LebesgueQuadratureWithEVData.matchVectors(ql.f,"Ql.f",ql.getMatrixDiagonalsInPsiStates(ql.QQf),"ql.getMatrixDiagonalsInPsiStates(ql.QQf)",LebesgueQuadratureWithEVData.getEpsSelftest(),0);
endfunction
% ========== END Runge function test



% run a few examples and tests

% Gauss & Lebesgue quadratures
printExampleQuadraturesCompareToGauss();
% Value and Probability correlations of two Lebesgue quadratures
[qlf,qlg]=printExampleLebesgueQuadratures();
% Obtain Fig. 1 of https://arxiv.org/abs/1511.01887
[xorig,forig,fRN,fLS]=printExampleRungeFunctuonInterpolation(7,10000,100);

disp(["Tests are done, printing two charts"]);
figure (1);LebesgueQuadratures_TestHelpers.printChristoffelChart(qlf,-1.2,0.01,1.2,[]);
figure (2);plot(xorig,forig,"linewidth", 3, xorig,fRN,"linewidth", 3, xorig,fLS, "linewidth", 3) ;title ("Runge(x)");
