package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Legendre polynomials in [-1:1] interval.
 *  @version $Id: Legendre.java,v 1.61 2022/07/16 10:04:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Legendre implements BasisPolynomials {

/** Calculates n-th Legendre polynomial.
 *  Uses: \( nP_n=x*(2n-1)*P_{n-1}-(n-1)*P_{n-2} \).
 * @param n P index.
 * @param Pnm1 Input \( P_{n-1} \).
 * @param Pnm2 Input \( P_{n-2} \).
 * @param flag_type2 Whether to use type 2 polynomials.
 * @param PBASIS The basis in which to calculte.
 * @return \( P_{n} \) polynomial, array dimension n+1.
 */
public double [] getNext(final int n,
			 final double [] Pnm1,final double [] Pnm2,
			 final boolean flag_type2,
			 final BasisFunctionsCalculatable PBASIS)
{
    if(n<0) throw new IllegalArgumentException();
    if(n<=0) return PBASIS.getONE();
    if(Pnm1.length!=n) throw new IllegalArgumentException();
    final double onn=1.0/(!flag_type2?n:n+1);
    final double [] res=PBASIS.getaxbP(Pnm1,2-onn,0);
    if(n<=1) return res;
    if(Pnm2.length!=n-1) throw new IllegalArgumentException();
    TMatr.sumP1(res,Pnm2,onn-1.0);
    return res;
}

public double [][] getBasisFunctionsOnPolynomialArgument(
    final int numQl,
    final double [] pol,
    final BasisFunctionsCalculatable PBASIS){

    final double [][] res=new double [numQl][];
    double [] fprev={},f=res[0]=PBASIS.getONE();
    for(int n=1;n<numQl;n++)
    {
	final double onn=1.0/n;
	final double [] fmul=TMatr.scaleVectorInPlace(PBASIS.mult2Pol(pol,f),2-onn);
	TMatr.sumP1(fmul,fprev,onn-1.0);
	fprev=f;f=res[n]=fmul;
    }
    return res;
}

/**
 * Calculation of the value of n-th Legendre polynomial at given {@code x}.
 * Uses: \( nP_n=x *(2n-1)P_{n-1}-(n-1)P_{n-2} \).
 * @param n P index.
 * @param x Argument.
 * @return The value of the Legendre polynomial \( P_n \) at {@code x}.
 */
public double calculate(final int n,final double x)
{
    double b=0,s=1;
    for(int k=1;k<=n;++k)
    {
	final double a=b;
	b=s;
	s=((2*k-1)*b*x-(k-1)*a)/k;
    }
    return s;
}


public double [] getMomentsFromSample(final int nmoments,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [nmoments];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=x[k],ww=w!=null?w[k]:1;
	double b=0,s=1;
	for(int i=0;i<nmoments;++i)
	{
	    if(i<=0) m[i]+=s*ww;
	    else {s=((1-i)*b+(2*i-1)*(b=s)*(xx))/i;m[i]+=s*ww;}
	}
    }
    return m;
}


/** Calculates \( \sum_k P_k(x) Pcoefs[k] \)
 *  using Clenshaw recurrence formula.
 * @param Pcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The sum.
 */
public double getD0(final double [] Pcoefs,final double x)
{
    final int n=Pcoefs.length-1;
    double b1=0,b2=0,s=0;
    for(int i=n;i>=0;--i)
    {
	s=((2*i+1)*x)*b1/(i+1)-(i+1)*b2/(i+2)+Pcoefs[i];
	b2=b1;
	b1=s;
    }
    return s;
}

/** Calculates \( d/dx \sum_k P_k(x) Pcoefs[k] \)
 *  using Clenshaw recurrence formula.
 * @param Pcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double getD1(final double [] Pcoefs,final double x)
{
    final int n=Pcoefs.length-1;
    double b1=0,b2=0,s=0,se=0,so=0;
    for(int k=n;--k>=0;)
    {
	final double f;
	if(k%2==0) f=(2*k+1)*(so+=Pcoefs[k+1]);
	else f=(2*k+1)*(se+=Pcoefs[k+1]);
	s=((2*k+1)*x)*b1/(k+1)-(k+1)*b2/(k+2)+f;
	b2=b1;
	b1=s;
    }
    return s;
}

/** Calculates \( d^2/dx^2 \sum_k P_k(x) Pcoefs[k] \)
 *  using Clenshaw recurrence formula.
 * @param Pcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The second derivative.
 */
public double getD2(final double [] Pcoefs,final double x)
{
    final int n=Pcoefs.length-1;
    double b1=0,b2=0,s=0,se=0,so=0,se2=0,so2=0;
    for(int k=n-1;--k>=0;)
    {
	final double f;
	if(k%2==0) f=(2*k+1)*(so2+=(2*k+3)*(se+=Pcoefs[k+2]));
	else f=(2*k+1)*(se2+=(2*k+3)*(so+=Pcoefs[k+2]));
	s=((2*k+1)*x)*b1/(k+1)-(k+1)*b2/(k+2)+f;
	b2=b1;
	b1=s;
    }
    return s;
}

/** Differentiate in Legendre basis.
 *  formulae 8.915.2, A(9036), page 1040, Gradshtein &amp; Ryzhik 1963.
 */
public double [] getDifferentiated(final double [] Pcoefs)
{
    final int n=Pcoefs.length-1;
    final double [] d=new double [n];
    double se=0,so=0;
    for(int k=n;--k>=0;)
    {
	final double f;
	if(k%2==0) f=(so+=Pcoefs[k+1]);
	else f=(se+=Pcoefs[k+1]);
	d[k]=f*(2*k+1);
    }
    return d;
}

/** Integrate Legendre polynomial.
 * Prudnikov, Brychkov, Marichev, volume 2, 2003,
 * formulae 1.14.1.1 page 46.
 */
public double [] getIntegrated(final double [] Pcoefs)
{
    final int n=Pcoefs.length-1;
    final double [] i=new double [n+2];
    //java.util.Arrays.fill(i,0,n+2,0);
    for(int k=n;k>=1;--k)
    {
	final double f=Pcoefs[k]/(2*k+1);
	i[k+1]+=f;
	i[k-1]-=f;
    }
    i[1]+=Pcoefs[0];
    i[0]-=getD0(i,0); // normalize integral to f(0)=0
    return i;
}



/** Expand \( P_l(z)P_m(z)=\sum_{k=0}^{l+m} s_k P_k(z) \),
 *  formulae 8.915.5, A(9036), page 1040, Gradshtein &amp; Ryzhik 1963.
 *  The formulae of Neumann-Adams, see W. N. Bailey,
 *  Generalized hypergeometric series, p.101, Cambridge, 1962.
 *  Second option is to use Gaunt's formula:
 *  see John C. Slater Quantum Theory of Atomic Structure,
 *  McGraw-Hill (New York, 1960), Volume I, page 309,
 *  which cites the original work of J. A. Gaunt,
 *  Philosophical Transactions of the Royal Society of London, A228:151 (1929).
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] s)
{
    final int nx,mx;
    if(m<=l) {nx=l;mx=m;}
    else {nx=m;mx=l;}
    final int nmx=nx+mx,na=2*nx,ma=2*mx,nma=na+ma;

    double cfact=1;
    for(int k=0;k<mx;k++)
    {
	cfact=(cfact*((nmx-k)*(2*k+1)))/((k+1)*(2*(k+1+nx)-1));
    }
    for(int k=0;k<=nmx;k++) s[k]=0;
    for(int k=0;k<=mx;k++)
    {
	final int k2=2*k,k21=k2+1;
	s[nmx-k2]=(cfact*(nma-2*k2+1))/(nma-k2+1);
	if(false)
	{ // unit test with 8.915, page 1040, Gradshtein & Ryzhik 1963.
	    final double c=(((Polynomials.fact2(2*(mx-k)-1)/(double)Polynomials.fact(mx-k))*
			     (Polynomials.fact2(2*k-1)/(double)Polynomials.fact(k))*
			     (Polynomials.fact2(2*(nx-k)-1)/(double)Polynomials.fact(nx-k)))/
			    (Polynomials.fact2(2*(nx+mx-k)-1)/(double)Polynomials.fact(nx+mx-k)))*
		(2*nx+2*mx-4*k+1)/(2*nx+2*mx-2*k+1);
	    UnitTests.checkDoubleRel(c,"c",s[nmx-k2],"s[nmx-k2]{l="+l+" m="+m+" k="+k+"}",1e-7);
	}
	if(k<mx)
	{
	    cfact=(cfact*(((long)mx-k)*(nx-k)*(nma-k21)*k21))/(((long)nmx-k)*(k+1)*(na-k21)*(ma-k21));
	}
    }
}


/** Multiply p by ma*x+mb.
 * Calculations in  Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * Do not mistake with a and b recurrence coefficiemts
 * of {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getAB(double [][]) getAB()}.
 * @param p Input polynomial.
 */
@Override
public double [] getaxbP(final double [] p,final double ma,final double mb)
{
    final int n=p.length-1;
    final double [] res=new double [n+2];
    double r=p[n];
    res[n]=res[n+1]=0;
    for(int k=n;k>=0;k--)
    {
	final double ar=ma*r;
	final int k21=(2*k+1);
	res[k+1]+=((k+1)*ar)/k21;
	res[k]+=0*ar+mb*r;
	if(k>0)
	{
	    final double t=p[k-1];
	    res[k-1]=(k*ar)/k21;
	    r=t;
	}
    }
    return res;
}



/** Divide  p by (x-x0), returns quotient and remainder.
 * Calculations in Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return Quotient and remainder.
 */
@Override
public SDivRes sdiv1(final double [] p,final double x0)
{
    final int n=p.length-1;
    final double [] res=new double [n];
    double r=p[n];
    //System.err.println("pi="+java.util.Arrays.toString(p));
    if(n>0) res[n-1]=p[n-1];
    for(int k=n;--k>=0;)
    {
	final int k1=k+1;
	final double c=((2*k+1)*r)/k1,b=(res[k]+=c*x0);
	if(k>0)res[k-1]=p[k-1]-(r*k)/k1;
	res[k]=c;
	r=b;
    }
    //System.err.println("p="+java.util.Arrays.toString(p));
    //System.err.println("res="+java.util.Arrays.toString(res));
    return new SDivRes(res,new double []{r});
}



/** Calculates P_n(a(x+x0)) using 3 term recurrence
 * Calculations in  Legendre basis.
 * The b_k are defined as:
 * \( P_n(a(x+x0))=\sum_{k=0}^{k=n}b_k^{n} P_k(x) \)
 * The method should be applied consequently to get \( b_k^{n} \) for n=0,1,2,3,...
 * Uses Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * @param n The max dimension of b.
 * @param b On input: \( b_k^{n-1} \), on output \( b_k^{n} \).
 * @param q On input: \( b_k^{n-2} \), on output: \( b_k^{n-1} \).
 */
@Override
public void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    if(n<=0)
    {
	b[n]=1;
	return;
    }
    final int n1=n-1;
    double r=b[n1];
    b[n1]=b[n]=0;
    final double fn=(2*n-1)/((double)n),fnashift=(a*(x0))*fn,afn=a*fn,fn1=(n-1)/((double)n);
    for(int k=n;--k>=0;)
    {
	final double f=afn/(2*k+1);
	b[k+1]+=((k+1)*r)*f;
	b[k]+=r*fnashift;
	if(k<n1) b[k]-=fn1*q[k];
	q[k]=r;
	if(k>0)
	{
	    final double bb=b[k-1];
	    b[k-1]=(k*r)*f;
	    r=bb;
	}
    }
}


/** Calculate {@literal <F|xPk>} moments for k=0..n-1 from {@literal <F|Pk>}, k=0..n
 *  @param moments {@literal <F|Pk>}.
 *  @return {@literal <F|xPk>}
 */
@Override
public double [] getXQkMomentsFromQkMoments(final double [] moments)
{
    final int n=moments.length-1;
    final double [] d=new double [n];
    for(int k=0;k<n;++k)
    {
	d[k]=((k+1)*moments[k+1]+(k>0?k*moments[k-1]:0))/(2*k+1);
    }
    return d;
}


/** Calculates confederate matrix to solve
 * polynomial equations in Legendre basis directly.
 * We use the three term recurrence:
 * \[ P_k=(x*(2*k-1)/k)*P_{k-1}-((k-1)/k)*P_{k-2} \]
 * Use {@link PolynomialRootsConfederateMatrix#ABASISLEGENDRE ABASISLEGENDRE}
 * to calculate roots.
 */
public double [] getConfederateMatrix(final double [] coefs)
{
    final int n=coefs.length-1;
    final double [] matr=new double [n*n];
    for(int m=0;m<n;m++)
    {
	final int k=m+1;
	final double k21=2*k-1;
	matr[m*n+m]=0;                    // delta/alfa
	if(m>0) matr[(m-1)*n+m]=(k-1)/k21;  // gamma/alfa
	if(m<n-1) matr[(m+1)*n+m]=k/k21;    // 1/alfa
	else
	{
	    final double f=k21*coefs[n]/k;  // p[n]*alfa
	    for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	}
    }
    return matr;
}


static final Legendre BASIS=new Legendre();


/** A unit test.
 */
public static void main(String [] args)
{
    final double eps=1e-6;
    new TestBasisPolynomials("Legendre",new Legendre(),new java.util.Random(-790991),eps){
	@Override double [] getMonomialsMoments(final int n)
	{
	    final double [] pmom=new double [n];
	    for(int k=0;k<pmom.length;k++){
		// monomials moments
		if(k%2==0){pmom[k]=2.0/(k+1);}
		else {pmom[k]=0;}
	    }
	    return pmom;
	}
	@Override OrthogonalPolynomialsLegendreBasis getOrtPol()
	{
	    return new OrthogonalPolynomialsLegendreBasis();
	}
	@Override double [][] getExact()
	{
	    return new double [][]{
		{1},
		{0,1},
		TMatr.getVectorScaled(new double []{-1,0,3},1.0/2),
		TMatr.getVectorScaled(new double []{0,-3,0,5},1.0/2),
		TMatr.getVectorScaled(new double []{3,0,-30,0,35},1.0/8),
		TMatr.getVectorScaled(new double []{0,15,0,-70,0,63},1.0/8),
		TMatr.getVectorScaled(new double []{-5,0,105,0,-315,0,231},1.0/16),
		TMatr.getVectorScaled(new double []{0,-35,0,315,0,-693,0,429},1.0/16),
		TMatr.getVectorScaled(new double []{35,0,-1260,0,6930,0,-12012,0,6435},1.0/128),
		TMatr.getVectorScaled(new double []{0,315,0,-4620,0,18018,0,-25740,0,12155},1.0/128),
		TMatr.getVectorScaled(new double []{-63,0,3465,0,-30030,0,90090,0,-109395,0,46189},1.0/256),
		TMatr.getVectorScaled(new double []{0,-693,0,15015,0,-90090,0,218790,0,-230945,0,88179},1.0/256)
	    };
	}
    }.runTests("$Id: Legendre.java,v 1.61 2022/07/16 10:04:33 mal Exp $");
}



} // END of class Legendre

