package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Legendre polynomials shifted to [0:1] interval.
 *  @version $Id: LegendreShifted.java,v 1.238 2022/08/31 07:48:52 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class LegendreShifted implements BasisPolynomials {

/** Calculates n-th Legendre polynomial shifted to [0,1].
 *  Uses: \[ nP_n=2*(2n-1)xP_{n-1}-(2n-1)P_{n-1}-(n-1)P_{n-2} \]
 * @param n P index.
 * @param Pnm1 Input \( P_{n-1} \).
 * @param Pnm2 Input \( P_{n-2} \).
 * @param flag_type2 Whether to use type 2 polynomials.
 * @param PBASIS The basis in which to calculte.
 * @return \( P_{n} \) polynomial, array dimension n+1.
 */
public double [] getNext(final int n,
			 final double [] Pnm1,final double [] Pnm2,
			 final boolean flag_type2,
			 final BasisFunctionsCalculatable PBASIS)
{
    if(n<0) throw new IllegalArgumentException();
    if(n<=0) return TMatr.scaleVectorInPlace(PBASIS.getONE(),!flag_type2?1:2);
    if(Pnm1.length!=n) throw new IllegalArgumentException();
    final double onn=1.0/(!flag_type2?n:n+1),s=(2-onn);
    final double [] res=PBASIS.getaxbP(Pnm1,2*s,-s);
    if(n<=1) return res;
    if(Pnm2.length!=n-1) throw new IllegalArgumentException();
    TMatr.sumP1(res,Pnm2,onn-1.0);
    return res;
}

public double [][] getBasisFunctionsOnPolynomialArgument(
    final int numQl,
    final double [] pol,
    final BasisFunctionsCalculatable PBASIS){

    final double [][] res=new double [numQl][];
    double [] fprev={},f=res[0]=PBASIS.getONE();
    for(int n=1;n<numQl;n++)
    {
	final double onn=1.0/n,s=(2-onn);
	final double [] fmul=TMatr.scaleVectorInPlace(PBASIS.mult2Pol(pol,f),2*s);
	TMatr.sumP1(fmul,f,-s);
	TMatr.sumP1(fmul,fprev,onn-1.0);
	fprev=f;f=res[n]=fmul;
    }
    return res;
}


/**
 * Calculation of the value of n-th Legendre polynomial shifted to [0,1].
 * @param n P index.
 * @param x Argument.
 * @return The value of the Legendre shifted polynomial Pn at x.
 */
public double calculate(final int n,final double x)
{
    return Legendre.BASIS.calculate(n,2*x-1);
}


public double [] getMomentsFromSample(final int nmoments,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [nmoments];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=2*x[k]-1,ww=w!=null?w[k]:1;
	double b=0,s=1;
	for(int i=0;i<nmoments;++i)
	{
	    if(i<=0) m[i]+=s*ww;
	    else {s=((1-i)*b+(2*i-1)*(b=s)*(xx))/i;m[i]+=s*ww;}
	}
    }
    return m;
}


/** Calculates \( \sum_k P_k(x) Pcoefs[k] \)
 *  using Clenshaw recurrence formula.
 * @param Pcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The sum.
 */
public double getD0(final double [] Pcoefs,final double x)
{
    return Legendre.BASIS.getD0(Pcoefs,2*x-1);
}

/** Calculates \( d/dx \sum_k P_k(x) Pcoefs[k] \)
 *  using Clenshaw recurrence formula.
 * @param Pcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double getD1(final double [] Pcoefs,final double x)
{
    return 2*Legendre.BASIS.getD1(Pcoefs,2*x-1);
}


/** Calculates \( d^2/dx^2 \sum_k P_k(x) Pcoefs[k] \)
 *  using Clenshaw recurrence formula.
 * @param Pcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double getD2(final double [] Pcoefs,final double x)
{
    return 4*Legendre.BASIS.getD2(Pcoefs,2*x-1);
}

public double [] getDifferentiated(final double [] Pcoefs)
{
    final double [] d=Legendre.BASIS.getDifferentiated(Pcoefs);
    for(int k=0;k<d.length;k++) d[k]*=2;
    return d;
}

/** Integrate shifted Legendre polynomial.
 * Prudnikov, Brychkov, Marichev, volume 2, 2003,
 * formulae 1.14.1.1 page 46.
 */
public double [] getIntegrated(final double [] Pcoefs)
{
    final double [] i=Legendre.BASIS.getIntegrated(Pcoefs);
    for(int k=0;k<i.length;k++) i[k]*=0.5;
    i[0]-=getD0(i,0); // normalize integral to f(0)=0
    return i;
}



/** Expand \( P_l(z)P_m(z)=\sum_{k=0}^{l+m} s_k P_k(z) \),
 *  formulae 8.915.5, A(9036), page 1040, Gradshtein &amp; Ryzhik 1963.
 *  The formulae of Neumann-Adams, see W. N. Bailey,
 *  Generalized hypergeometric series, p.101, Cambridge, 1962.
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] s)
{
    Legendre.BASIS.setQlQmExpansionCoefs(l,m,s);
}


/** Multiply p by ma*x+mb.
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * Do not mistake with a and b recurrence coefficiemts
 * of {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getAB(double [][]) getAB()}.
 * @param p Input polynomial.
 */
@Override
public double [] getaxbP(final double [] p,final double ma,final double mb)
{
    return Legendre.BASIS.getaxbP(p,0.5*ma,mb+0.5*ma);
}



/** Divide  p by (x-x0), returns quotient and remainder.
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return Quotent and remainder.
 */
@Override
public SDivRes sdiv1(final double [] p,final double x0)
{
    final SDivRes sdr=Legendre.BASIS.sdiv1(p,2*x0-1);
    for(int k=sdr.q.length;--k>=0;){sdr.q[k]*=2;}
    return sdr;
}



/** Calculates P_n(a(x+x0)) using 3 term recurrence
 * Calculations in Shifted Legendre basis.
 * The b_k are defined as:
 * \( P_n(a(x+x0))=\sum_{k=0}^{k=n}b_k^{n} P_k(x) \)
 * The method should be applied consequently to get \( b_k^{n} \) for n=0,1,2,3,...
 * Uses Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * @param n The max dimension of b.
 * @param b On input: \( b_k^{n-1} \), on output \( b_k^{n} \)
 * @param q On input: \( b_k^{n-2} \), on output: \( b_k^{n-1} \).
 */
@Override
public void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    Legendre.BASIS.setNewtonBinomialLikeCoefs(n,b,q,a,1+2*x0-1/a);
}


/** Calculate {@literal <F|xPk>} moments for k=0..n-1 from {@literal <F|Pk>}, k=0..n.
 * @param moments {@literal <F|Pk>}.
 * @return {@literal <F|xPk>}.
 */
@Override
public double [] getXQkMomentsFromQkMoments(final double [] moments)
{
    final int n=moments.length-1;
    final double [] d=new double [n];
    double pk1=0,pk=moments[0];
    for(int k=0;k<n;++k)
    {
	final double f=2*(2*k+1);
	d[k]=(k/f)*pk1+0.5*(pk1=pk)+((k+1)/f)*(pk=moments[k+1]);
    }
    return d;
}


/** Calculates confederate matrix to solve
 * polynomial equations in Legendre basis directly.
 * We use the three term recurrence:
 * \[ P_k=(x*2*(2*k-1)/k-(2*k-1)/k)*P_{k-1}-((k-1)/k)*P_{k-2} \]
 * Use {@link PolynomialRootsConfederateMatrix#ABASISLEGENDRESHIFTED ABASISLEGENDRESHIFTED}
 * to calculate roots.
 */
public double [] getConfederateMatrix(final double [] coefs)
{
    final int n=coefs.length-1;
    final double [] matr=new double [n*n];
    for(int m=0;m<n;m++)
    {
	final int k=m+1;
	final double k42=4*k-2;
	matr[m*n+m]=0.5;                    // delta/alfa
	if(m>0) matr[(m-1)*n+m]=(k-1)/k42;  // gamma/alfa
	if(m<n-1) matr[(m+1)*n+m]=k/k42;    // 1/alfa
	else
	{
	    final double f=k42*coefs[n]/k;  // p[n]*alfa
	    for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	}
    }
    return matr;
}


private static final LegendreShifted BASIS=new LegendreShifted();


public static double get2LegendreAveraged(final double [] p1,final double [] p2)
{
    if(p1.length!=p2.length) throw new IllegalArgumentException();
    final int n=p1.length-1;
    double s=0;
    for(int k=0;k<=n;k++)
    {
	s+=p1[k]*p2[k]/(2*k+1);
    }
    return s;
}

/** Generating expansion coefficients
 *  xP'n+Pn=(n+1)Pn+sum_{k=0}^{k=n-1}(2k+1)P_k
 */
private static double [] getXPnprimeplusPnCoefs(final int Pn_number)
{
    final int n=Pn_number;
    final double [] s=new double [n+1];
    s[n]=(2*n+1)-n;
    for(int k=n;--k>=0;)
    {
	s[k]=(2*k+1);
    }
    return s;
}
/** Calculates xF'+F.
 * @param Pcoefs,input,
 * @return xF'+F
 */
public static double [] getXDiffFplusF(final double [] Pcoefs)
{
    final int n=Pcoefs.length-1;
    final double [] d=new double [n+1];
    double s=0;
    for(int k=n;k>=0;--k)
    {
	final double x;
	d[k]=s*(2*k+1)+(k+1)*(x=Pcoefs[k]);
	s+=x;
    }
    return d;
}

/** Calculate {@literal <F|Pn+xP'n>} moments. Same as {@link #getXPnprimeplusPnMomentsFromPnMoments}
 *  but uses extra (2k+1) normalizing factor.
 *  @param Pcoefs Input, Pcoefs[k]=(2*k+1){@literal <F|Pk>}, k=0..n.
 *  @return {@literal (2*k+1)<F|Pk+xP'k>}, k=0..n.
 */
private static double [] getXPnprimeplusPnMoments(final double [] Pcoefs)
{
    final int n=Pcoefs.length-1;
    final double [] d=new double [n+1];
    double s=0;
    for(int k=0;k<=n;++k)
    {
	final double x;
	d[k]=s*(2*k+1)+(k+1)*(x=Pcoefs[k]);
	s+=x;
    }
    return d;
}

/** Calculate {@literal <F|Pn+xP'n>} moments.
 *  @param moments Input, moments[k]={@literal <F|Pk>}, k=0..n.
 *  @return {@literal <F|Pk+xP'k>}, k=0..n.
 */
public static double [] getXPnprimeplusPnMomentsFromPnMoments(final double [] moments)
{
    final int n=moments.length-1;
    final double [] d=new double [n+1];
    double s=0;
    for(int k=0;k<=n;++k)
    {
	final double x;
	d[k]=s+(k+1)*(x=moments[k]);
	s+=x*(2*k+1);
    }
    return d;
}


/** Calculate {@literal <F|(1-x)Pk>} moments for k=0..n-1 from {@literal <F|Pk>}, k=0..n
 *  @param moments Input, moments[k]={@literal <F|Pk>}, k=0..n;
 *  @return {@literal <F|(1-x)Pk>}, k=0..n-1;
 */
public static double [] get1mXPnMomentsFromPnMoments(final double [] moments)
{
    final int n=moments.length-1;
    final double [] d=new double [n];
    double pk1=0,pk=moments[0];
    for(int k=0;k<n;++k)
    {
	final double f=2*(2*k+1);
	d[k]=-(k/f)*pk1+0.5*(pk1=pk)-((k+1)/f)*(pk=moments[k+1]);
    }
    return d;
}


/** Calculate xPcoefs'(x) in Legendre shifted basis.
 * @param Pcoefs.
 * @return xPcoefs'.
 */
public static double [] getXDiffF(final double [] Pcoefs)
{
    final int n=Pcoefs.length-1;
    final double [] d=new double [n+1];
    double s=0;
    for(int k=n;k>=0;--k)
    {
	final double x;
	d[k]=s*(2*k+1)+k*(x=Pcoefs[k]);
	s+=x;
    }
    return d;
}

/** The same as setQlQmExpansionCoefs,
 *  but uses 3 term recursion instead of hypergeometric function expansion.
 *  It is rather slow at large l,m. Used for unit test.
 */
public static void setQlQmExpansionCoefs_test(final int l,final int m,final double [] s)
{
    for(int k=l+m;k>=0;k--) s[k]=0;
    final int ll,mm;
    if(l<=m) {ll=l;mm=m;}
    else {ll=m;mm=l;}
    sumQlQmExpansionCoefs(ll,mm,s,1);
}
private static void sumQlQmExpansionCoefs(final int l,final int m,final double [] s,final double c)
{
    if(l>m) throw new IllegalArgumentException("l="+l+">m="+m);
    if(l<=0) {s[m]+=c; return;}
    final double ol=1.0/l;
    if(l>=2) sumQlQmExpansionCoefs(l-2,m,s,c*(-1+ol));
    final double ol2m1=c*(2-ol)/(2*m+1);
    sumQlQmExpansionCoefs(l-1,m-1,s,ol2m1*m);
    sumQlQmExpansionCoefs(l-1,m+1,s,ol2m1*(m+1));
}



/** Multiply p by x.
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * @param p Input polynomial.
 * @return x*p
 */
public static double [] getxP(final double [] p)
{
    return BASIS.getaxbP(p,1,0);
}


/** Multiply p by (1-x).
 * Calculations in Shifted Legendre basis.
 * Use Legendre recurrence, see
 * formulae 8.914.1, UVII99, page 1040, Gradshtein &amp; Ryzhik 1963.
 * @param p Input polynomial.
 * @return (1-x)*p
 */
public static double [] get1mxP(final double [] p)
{
    return BASIS.getaxbP(p,-1,1);
}



/** A different way to calculate roots. Use Jacobi matrix.
 * @param n The Laguerre polynomial order.
 * @param x0 Radau edge, ignored  if flag_radau=false;
 * @param flag_radau Whether to use Radau polynomial.
 * @return The roots.
 */
static double [] getRoots(final int n,final double x0,final boolean flag_radau)
{
    final double [] a=new double [n+1],b=new double [n];
    a[0]=1;
    for(int k=0;k<n;k++)
    {
	b[k]=0.5;
	a[k+1]=0.25*(k+1)/Math.sqrt((k+1)*(k+1)-0.25);
    }
    if(flag_radau)
    {
	b[n-1]+=a[n]*(BASIS.calculate(n,x0)/BASIS.calculate(n-1,x0))*Math.sqrt((2*n+1)/((double)(2*n-1)));
    }
    return EVSolver.getDefaultEVSolver().getEV_Thridiagonal(n,b,0,a,1,false);
}
public static double [] getRoots(final int n)
{
    return getRoots(n,Double.NaN,false);
}
public static double [] getRadauRoots(final int n)
{
    return getRoots(n,1,true);
}
/** Uses Hildebrand, F. B. "Introduction to Numerical Analysis."
 *  New York: McGraw-Hill, pp. 323-325, 1956.
 */
public static double [] getWeights(final double [] xx)
{
    final int n=xx.length;
    final double [] w=new double [n];
    final double f=4.0/((n+1)*(n+1));
    for(int k=n;--k>=0;)
    {
	final double x=xx[k];
	w[k]=f*x*(1-x)/Math.pow(BASIS.calculate(n+1,x),2);
    }
    return w;
}
public static double [] getNodesPolynomial(final int n)
{
    final double [] nodes=new double [n+1];
    nodes[n]=1;
    return nodes;
}

/** Radau exact weights from
 * Walter Gautschi, "Gauss-Radau formulae for Jacobi and Laguerre weight functions".
 */
public static double [] getRadauWeights(final double [] xr)
{
    final int n=xr.length;
    final double [] w=new double [n];
    final double f=0.5/(n*n);
    for(int k=n;--k>=0;)
    {
	final double x=xr[k];
	// 1.8 . sign changed in (1-t) because our node is at the right
	w[k]=f*(1+(2*x-1))/Math.pow(BASIS.calculate(n-1,x),2);
    }
    return w;
}
public static double [] getRadauNodesPolynomial(final int n)
{
    final double [] nodes0=new double [n];nodes0[0]=1;
    final double [] nodes=getXPnprimeplusPnMoments(nodes0); // get delta function by differentiating a step function
    return get1mxP(nodes);
}
/** Christoffel function. */
public static double [] getChristoffelKernel(final int n)
{
    final double [] c=new double [2*n+1],tmp=new double [2*n+1];
    for(int k=0;k<=n;k++)
    {
	BASIS.setQlQmExpansionCoefs(k,k,tmp);
	TMatr.sumP1(c,tmp,2*k+1,2*k+1);
    }
    return c;
}


/** A class for unit test. */
private static class TestBasisLegendreShifted extends TestBasisPolynomials {

TestBasisLegendreShifted()
{
    this(new java.util.Random(-751431),1e-5);
}

TestBasisLegendreShifted(final java.util.Random r,final double eps)
{
    super("LegendreShifted",BASIS,r,eps);
}

@Override double [] getMonomialsMoments(final int n)
{
    final double [] pmom=new double [n];
    for(int k=0;k<pmom.length;k++) pmom[k]=1.0/(k+1); // monomials moments
    return pmom;
}
@Override OrthogonalPolynomialsLegendreShiftedBasis getOrtPol()
{
    return new OrthogonalPolynomialsLegendreShiftedBasis();
}
@Override double [][] getExact()
{
    return new double [][]{
	{1},
	{-1,2},
	{1,-6,6},
	{-1,12,-30,20},
	{1,-20,90,-140,70},
	{-1,30,-210,560,-630,252},
	{1,-42,420,-1680,3150,-2772,924},
	{-1,56,-756,4200,-11550,16632,-12012,3432},
	{1,-72,1260,-9240,34650,-72072,84084,-51480,12870},
	{-1,90,-1980,18480,-90090,252252,-420420,411840,-218790,48620}
    };
}


private void printAllQuadratures(final int nmax)
{
    final PolynomialRootsFinder rf=new PolynomialRootsFinder(){
	    public PolynomialRootsFinder.PRoots findRoots(final double [] p){
		return new PolynomialRootsJenkinsTraubORIG().findRoots(BASIS.convertFunctionTo_Monomials_from_Basis(p));
	    }};
    for(int n=1;n<=nmax;n++)
    {
	printTestRoots(rf,n);
	printQuadratures(rf,n);
	printRadauQuadratures2(rf,n,r);
	printDeltaFunctionRoots(rf,n);
    }
}

void printQuadratures(final PolynomialRootsFinder rf,final int n)
{
    System.err.println("=========LegendreShifted: Gauss quadratures for n="+n);
    final double eps=1e-4;

    final double [] pol=new double [n+1];pol[n]=1;
    final PolynomialRootsFinder.PRoots u=rf.findRoots(pol);
    final double [] root_re=u.re_root,root_im=u.im_root;
    if(u.nroots!=n) throw new RuntimeException("u.nroots="+u.nroots+" n="+n);
    SortArrays.sortArrayWithLoadDesc(n,root_re,root_im);
    final double [] xr=getRoots(n);
    SortArrays.sortArrayWithLoadDesc(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",root_re,"root_re",eps);
    final double [] wh=getWeights(xr),ww=new double [xr.length];
    final double [] kk=getChristoffelKernel(n),wk=new double [ww.length];

    System.err.println("X\tW\tS");
    double s=0,S=0;
    for(int k=0;k<n;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double w=2/(n*BASIS.calculate(n-1,root_re[k])*BASIS.getD1(pol,root_re[k]));
	s+=w;
	S-=w*Math.log(w);
	ww[k]=w;
	wk[k]=1.0/BASIS.getD0(kk,root_re[k]);
	System.err.println(String.valueOf(root_re[k])+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    UnitTests.checkDoubleArrayRel1(ww,"ww",wh,"wh",eps);
    UnitTests.checkDoubleArrayRel1(wk,"wk",wh,"wh",eps);
    System.err.println("s="+s+" S="+S);
    System.err.println();
}
void printTestRoots(final PolynomialRootsFinder rf,final int n)
{
    final double [] pol=new double [n+1];pol[n]=1;
    PolynomialRootsFinder.PRoots u=rf.findRoots(pol);
    final double [] root_re=u.re_root,root_im=u.im_root;
    if(u.nroots!=n) throw new RuntimeException("u.nroots="+u.nroots+" n="+n);
    SortArrays.sortArrayWithLoadDesc(n,root_re,root_im);
    final double [] xr=getRoots(n);
    SortArrays.sortArrayWithLoadDesc(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",root_re,"root_re",eps);
    System.err.println("=========LegendreShifted: TestRoots for n="+n);
    System.err.println("Xm\tFm\tXf\tFf");
    for(int k=0;k<n;k++)
    {
	System.err.println(String.valueOf(xr[k])+"\t"+BASIS.calculate(n,xr[k])+"\t"+
			   root_re[k]+"\t"+BASIS.calculate(n,root_re[k]));
    }
    System.err.println();
}

void printRadauQuadratures2(final PolynomialRootsFinder rf,final int neq,final java.util.Random r)
{
    System.err.println("=========Legendre-Radau: Gauss quadratures for neq="+neq);
    final double eps=1e-4;

    final double [] mom=new double [2*neq+1];
    mom[0]=1;

    final double [] matr=new double [neq*neq],
	pol=new double [neq+1];
    for(int k=0;k<neq;k++)
    {
	final double [] ee=new double [neq+k+1];
	BASIS.setQlQmExpansionCoefs(neq,k,ee);
	final double [] e=get1mxP(ee);
	if(e.length!=neq+k+2) throw new IllegalArgumentException();
	//System.err.println("e="+java.util.Arrays.toString(e));
	pol[k]=-TMatr.mulVVn1(e,mom);
	for(int l=0;l<=k;l++)
	{
	    final double [] ee2=new double [l+k+1];
	    BASIS.setQlQmExpansionCoefs(l,k,ee2);
	    final double [] e2=get1mxP(ee2);
	    if(e2.length!=l+k+2) throw new IllegalArgumentException();
	    matr[k+l*neq]=matr[l+k*neq]=TMatr.mulVVn1(e2,mom);
	}
    }
    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0) throw new RuntimeException("pinfo="+pinfo);
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,pol,1);
    pol[neq]=1;
    //System.err.println("pol="+java.util.Arrays.toString(pol));
    final PolynomialRootsFinder.PRoots u=rf.findRoots(pol);
    if(u.nroots!=neq) throw new RuntimeException("u.nroots="+u.nroots+" neq="+neq);
    final double [] xs=new double [neq+1],root_im=new double [neq+1];
    System.arraycopy(u.re_root,0,xs,0,neq);System.arraycopy(u.im_root,0,root_im,0,neq);
    xs[neq]=1;root_im[neq]=0; // radau node
    final double [] rpol=get1mxP(pol);
    if(true)
    {
	final double [] nrr=getRadauNodesPolynomial(rpol.length-1);
	for(int k=0;k<rpol.length;k++) nrr[k]*=rpol[rpol.length-1]/nrr[rpol.length-1];
	UnitTests.checkDoubleArrayRel(nrr,"nrr",rpol,"rpol",eps);
    }
    final double [] ws=new double [neq+1],tmp2=new double [neq+1];
    for(int k=0;k<=neq;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double x=xs[k];
	if(Double.compare(x,1.0)==0)
	{ // radau node
	    ws[k]=TMatr.mulVVn1(pol,mom)/BASIS.getD0(pol,x);
	}
	else
	{
	    final SDivRes sdr=BASIS.sdiv1(rpol,x);
	    ws[k]=TMatr.mulVVn1(sdr.q,mom)/BASIS.getD0(sdr.q,x);
	}
    }
    SortArrays.sortArrayWithLoadDesc(neq+1,xs,ws);

    final double [] xr=getRadauRoots(neq+1);
    SortArrays.sortArrayWithLoadDesc(xr.length,xr,null);
    UnitTests.checkDoubleArrayRel(xr,"xr",xs,"xs",eps);

    System.err.println("X\tW\tS");
    double s=0,S=0;
    for(int k=0;k<=neq;k++)
    {
	final double x=xs[k],w=ws[k];
	s+=w;
	S-=w*Math.log(w);
	System.err.println(String.valueOf(x)+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    System.err.println("s="+s+" S="+S);
    System.err.println();

    // test radau exact integral, polynomial of 2*neq power
    for(int t=0;t<1000;t++)
    {
    final double [] pt=TMatr.getRandomVector(2*neq +1,r,1,0);
    pt[0]+=0.6+r.nextDouble()+r.nextDouble();
    double i=0;
    for(int k=0;k<xs.length;k++)
    {
	i+=ws[k]*BASIS.getD0(pt,xs[k]);
    }
    final double ia=TMatr.mulVVn1(pt,mom);
    UnitTests.checkDoubleRel(i,"i",ia,"ia",eps);
    }

    // check Radau exact weights.
    if(neq<=12)
    checkRadauExactWeights(xs,ws,eps);
}


/** Check Radau exact weights from
 * Walter Gautschi, "Gauss-Radau formulae for Jacobi and Laguerre weight functions".
 */
void checkRadauExactWeights(final double [] xs,final double [] ws,final double eps)
{
    final int neq=xs.length-1;
    UnitTests.checkDoubleRel(ws[0],"ws[0]",1.0/((neq+1)*(neq+1)),"1.0/((neq+1)*(neq+1))",eps);  // 1.8
    UnitTests.checkDoubleArrayRel(ws,"ws",getRadauWeights(xs),"getRadauWeights(xs)",eps);

    System.err.println("checkRadauExactWeights: Radau exact weights OK, neq="+neq);
}

void printDeltaFunctionRoots(final PolynomialRootsFinder rf,final int neq)
{
    System.err.println("=========LegendreShifted: Quadratures on (1-x)delta(1-x) roots for neq="+neq);
    final double eps=1e-5;
    final double [] o0=new double [neq+1];o0[0]=1;
    final double [] o=getXPnprimeplusPnMoments(o0); // get delta function by differentiating a step function

    //System.err.println("o="+java.util.Arrays.toString(o));

    final double [] mom=new double [2*neq+1];
    mom[0]=1;

    final double [] wnodes=BASIS.mult2Pol(o,new double []{-0.5,0.5});
    final PolynomialRootsFinder.PRoots u=rf.findRoots(wnodes);
    if(u.nroots!=neq+1) throw new RuntimeException("u.nroots="+u.nroots+" neq+1="+(neq+1));
    final double [] xs=u.re_root,root_im=u.im_root,ws=new double [neq+1],tmp2=new double [neq+1];
    for(int k=0;k<=neq;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double x=xs[k];
	final SDivRes sdr=BASIS.sdiv1(wnodes,x);
	ws[k]=TMatr.mulVVn1(sdr.q,mom)/BASIS.getD0(sdr.q,x);
    }

    SortArrays.sortArrayWithLoadDesc(neq+1,xs,ws);
    System.err.println("X\tW\tS");
    double s=0;
    for(int k=0;k<=neq;k++)
    {
	final double x=xs[k],w=ws[k];
	s+=w;
	System.err.println(String.valueOf(x)+"\t"+w+"\t"+s);
    }
    UnitTests.checkDoubleRel(s,"s",1,"1",eps);
    System.err.println();
    if(neq<=12)
    checkRadauExactWeights(xs,ws,eps);
}



@Override void test_setQlQmExpansionCoefs(final int n_tests,final int nmax)
{
    super.test_setQlQmExpansionCoefs(n_tests,nmax);

    for(int t=0;t<n_tests;t++)
    {

    for(int l=0;l<=nmax;l++)
    for(int m=0;m<=nmax;m++)
    {
	final double [] res=TMatr.getRandomVector(l+m+1,r),resa=TMatr.getRandomVector(l+m+1,r);
	BASIS.setQlQmExpansionCoefs(l,m,res);
	setQlQmExpansionCoefs_test(l,m,resa);
	UnitTests.checkDoubleArrayRel(res,"res",resa,"resa",eps);
    }
    }
    System.err.println("LegendreShifted:test_setQlQmExpansionCoefs(): EXTRA test OK with nmax="+nmax+" eps="+eps);
}


void test_getLegendreFromPByIntegration(final int n_tests)
{
    for(int t=0;t<n_tests;t++)
    {
	final int n=1+r.nextInt(6);
	final double [] p=TMatr.getRandomVector(n+1,r,1,0);
	//System.err.println("p="+java.util.Arrays.toString(p));
	final double [] res=new double [n+1];
	double [] ra=null,rb=null;
	for(int k=0;k<=n;k++)
	{
	    final double [] x=BASIS.getNext(k,ra,rb,Monomials.BASIS);
	    rb=ra;ra=x;
	    final double [] i=Polynomials.getPolynomialIntegrated(Polynomials.mult2Pol(java.util.Arrays.copyOf(x,k+1),p));
	    //System.err.println("p x="+java.util.Arrays.toString(Polynomials.getMult2Pol(x,k+1,p,n+1)));

	    //System.err.println("i="+java.util.Arrays.toString(i));
	    //System.err.println("x["+k+"]="+java.util.Arrays.toString(x));
	    res[k]=(Polynomials.getD0(i,1)-Polynomials.getD0(i,0))*(2*k+1); // [0,1] interval, (2*k+1) normalizing coefficient.
	}
	//System.err.println("res="+java.util.Arrays.toString(res));
	UnitTests.checkDoubleArrayRel1(p,"p",BASIS.convertFunctionTo_Monomials_from_Basis(res),"convertFunctionTo_Monomials_from_Basis(res)",eps);
	UnitTests.checkDoubleArrayRel1(BASIS.convertFunctionTo_Basis_from_Monomials(p),"convertFunctionTo_Basis_from_Monomials",res,"res",eps);
    }
    System.err.println("LegendreShifted:test_getLegendreFromPByIntegration() norm (2n+1) test OK with n_tests="+n_tests+" eps="+eps);
}


void test_getLegendreAveraged(final int n_tests)
{
    for(int t=0;t<n_tests;t++)
    {
	final double [] p1=TMatr.getRandomVector(1+r.nextInt(6),r,1,0),
	    p2=TMatr.getRandomVector(p1.length,r,1,0);
	final double [] i=Polynomials.getPolynomialIntegrated(Polynomials.mult2Pol(BASIS.convertFunctionTo_Monomials_from_Basis(p1),BASIS.convertFunctionTo_Monomials_from_Basis(p2)));
	UnitTests.checkDoubleRel(get2LegendreAveraged(p1,p2),"<p1 x p2>",Polynomials.getD0(i,1)-Polynomials.getD0(i,0),"int pols",eps);
    }
    System.err.println("LegendreShifted:test_getLegendreAveraged() test OK with n_tests="+n_tests+" eps="+eps);
}


void test_getXPnplusPn(final int n_tests)
{
    // testing getXPnprimeplusPn
    for(int n=1;n<16;n++)
    {
	final double [] pol=new double [n+1];pol[n]=1;
	final double [] xpplusp=getXPnprimeplusPnCoefs(n);
	//System.err.println("P"+n+"="+java.util.Arrays.toString(POLS[n]));
	//System.err.println("xpplusp"+n+"="+java.util.Arrays.toString(xpplusp));

	final double [] xxp=BASIS.convertFunctionTo_Monomials_from_Basis(pol),
	    xx=TMatr.sumAcB(xxp,1,Polynomials.getxP(Polynomials.getPolynomialDifferentiated(xxp)));
	UnitTests.checkDoubleArrayRel(BASIS.convertFunctionTo_Basis_from_Monomials(xx),"xx",xpplusp,"xpplusp",eps);
	UnitTests.checkDoubleArrayRel(xx,"xx",BASIS.convertFunctionTo_Monomials_from_Basis(xpplusp),"xpplusp",eps);
	//System.err.println("xx"+n+"="+java.util.Arrays.toString(xx));

	final double [] d=getXDiffFplusF(pol);
	UnitTests.checkDoubleArrayRel(d,"d",xpplusp,"xpplusp",eps);
    }

    for(int t=0;t<n_tests;t++)
    {
	final double [] pol=TMatr.getRandomVector(1+r.nextInt(7),r,1,0);
	final int n=pol.length-1;
	if(true) // Test getXDiffF.
	{
	    final double [] p=BASIS.convertFunctionTo_Monomials_from_Basis(pol),
		xdp=Polynomials.getxP(Polynomials.getPolynomialDifferentiated(p));

	    final double [] d=getXDiffF(pol),dfp=TMatr.sumAcB(getXDiffFplusF(pol),-1,pol);
	    UnitTests.checkDoubleArrayRel(d,"d",BASIS.convertFunctionTo_Basis_from_Monomials(xdp),"x pol'",eps);
	    UnitTests.checkDoubleArrayRel(BASIS.convertFunctionTo_Monomials_from_Basis(d),"d",xdp,"xdp",eps);
	    UnitTests.checkDoubleArrayRel(d,"d",dfp,"dfp'",eps);
	}

	if(true) // Test getXDiffFplusF agains direct calculations.
	{
	    final double [] d=getXDiffFplusF(pol);
	    final double [] xxp=BASIS.convertFunctionTo_Monomials_from_Basis(pol),
		xx=TMatr.sumAcB(Polynomials.getxP(Polynomials.getPolynomialDifferentiated(xxp)),1,xxp);
	    UnitTests.checkDoubleArrayRel(BASIS.convertFunctionTo_Basis_from_Monomials(xx),"xx",d,"d",eps);
	    UnitTests.checkDoubleArrayRel(xx,"xx",BASIS.convertFunctionTo_Monomials_from_Basis(d),"d",eps);
	    UnitTests.checkDoubleArrayRel(TMatr.sumAcB(getXDiffF(pol),1,pol),"xF'+F",d,"d",eps);
	}

	if(true) // Test getXPnprimeplusPnMoments.
	{
	    final double [] d=getXPnprimeplusPnMoments(pol);
	    if(true)
	    {
		final double [] pp=new double [n+1];
		for(int k=0;k<=n;k++) pp[k]=pol[k]/(2*k+1);
		final double [] dd=getXPnprimeplusPnMomentsFromPnMoments(pp);
		final double [] ddt=new double [n+1];
		for(int k=0;k<=n;k++) ddt[k]=dd[k]*(2*k+1);
		UnitTests.checkDoubleArrayRel(ddt,"ddt",d,"d",eps);

		for(int k=0;k<=n;k++)
		{
		    final double [] tmp=new double [k+1],rtmp=new double [k+1];
		    tmp[k]=1;
		    if(k>0) {
			TMatr.sumP1(rtmp,getxP(BASIS.getDifferentiated(tmp)),1);
		    }
		    TMatr.sumP1(rtmp,tmp,k+1,1);
		    ddt[k]=TMatr.mulVVn1(rtmp,pp);
		}
		UnitTests.checkDoubleArrayRel(ddt,"ddt",dd,"dd",eps);
		final double [] ddt2=getXPnprimeplusPnMomentsFromPnMoments(pp);
		UnitTests.checkDoubleArrayRel(ddt2,"ddt2",dd,"dd",eps);
	    }


	    if(true) // Test against getXDiffF.
	    {
		final double f1=BASIS.getD0(pol,1);
		final double [] dd=new double [n+1];
		for(int k=0;k<=n;k++){dd[k]=(2*k+1)*f1-d[k];}
		final double [] xdd=getXDiffF(pol);
		UnitTests.checkDoubleArrayRel(dd,"dd",xdd,"xdd",eps);
	    }

	    if(true) // Test against direct calculation.
	    {
		final double [] xd=new double [n+1];
		double [] ra=null,rb=null;
		for(int k=0;k<=n;k++)
		{
		    final double [] l=BASIS.getNext(k,ra,rb,Monomials.BASIS);
		    rb=ra;ra=l;
		    final double [] xx=new double [k+1];
		    xx[0]=l[0];for(int j=0;j<k;j++) {xx[j+1]=(j+1+1)*l[j+1];} // P+xP'
		    final double [] xxl=BASIS.convertFunctionTo_Basis_from_Monomials(xx);
		    xd[k]=(2*k+1)*get2LegendreAveraged(xxl,java.util.Arrays.copyOf(pol,xxl.length));
		}
		UnitTests.checkDoubleArrayRel(d,"d",xd,"xd",eps);
	    }

	    if(true) // Test against getXPnprimeplusPnCoefs.
	    {
		final double [] sd=new double [n+1];
		for(int k=0;k<=n;k++)
		{
		    final double [] temp=getXPnprimeplusPnCoefs(k);
		    if(temp.length!=k+1) throw new RuntimeException();
		    sd[k]=(2*k+1)*get2LegendreAveraged(temp,java.util.Arrays.copyOf(pol,k+1)); // (2*k+1) to convert moments to coefficients
		}
		UnitTests.checkDoubleArrayRel(d,"d",sd,"sd",eps);
	    }
	}
	if(true)
	{
	    // save moments in pp.
	    final double [] pp=new double [n+1];
	    for(int k=0;k<=n;k++) pp[k]=pol[k]/(2*k+1);
	    final double [] Xpp=new double [n],_1mXpp=new double [n];
	    for(int k=0;k<n;k++)
	    {
		final double [] tmp=new double [k+1];tmp[k]=1;
		final double [] rtmp=getxP(tmp);
		Xpp[k]=TMatr.mulVVn1(rtmp,pp);
		final double [] rtmp2=get1mxP(tmp);
		_1mXpp[k]=TMatr.mulVVn1(rtmp2,pp);
	    }
	    {
	    final double [] sXpp=BASIS.getXQkMomentsFromQkMoments(pp);
	    UnitTests.checkDoubleArrayRel(sXpp,"sXpp",Xpp,"Xpp",eps);
	    final double [] sXpp_same=BASIS.getXQkMomentsFromQkMoments(pp);
	    UnitTests.checkDoubleArrayRel(sXpp_same,"sXpp_same",Xpp,"Xpp",eps);
	    }
	    {
	    final double [] s1mXpp=get1mXPnMomentsFromPnMoments(pp);
	    UnitTests.checkDoubleArrayRel(s1mXpp,"s1mXpp",_1mXpp,"_1mXpp",eps);
	    final double [] s1mXpp_same=get1mXPnMomentsFromPnMoments(pp);
	    UnitTests.checkDoubleArrayRel(s1mXpp_same,"s1mXpp_same",_1mXpp,"_1mXpp",eps);
	    }
	}
    }

    // Testing an expansion using getXPnprimeplusPnCoefs.
    for(int t=0;t<n_tests;t++)
    {
	final double [] pol=TMatr.getRandomVector(1+r.nextInt(7),r,1,0);
	final int n=pol.length-1;
	final double [] p=BASIS.convertFunctionTo_Monomials_from_Basis(pol),
	    xdp=Polynomials.getxP(Polynomials.getPolynomialDifferentiated(p)); // xP'
	//System.err.println("p="+java.util.Arrays.toString(p));
	//System.err.println("xdp="+java.util.Arrays.toString(xdp));
	//System.err.println("pol="+java.util.Arrays.toString(pol));

	final double [] pk1=new double [n+1];
	final double [] pk2=new double [n+1];
	double [] ra=null,rb=null;
	for(int k=0;k<=n;k++)
	{
	    final double [] pleg=BASIS.getNext(k,ra,rb,Monomials.BASIS);
	    rb=ra;ra=pleg;
	    final double [] i=Polynomials.getPolynomialIntegrated(Polynomials.mult2Pol(xdp,pleg));
	    //System.err.println("i"+k+"="+java.util.Arrays.toString(i));
	    pk1[k]=(Polynomials.getD0(i,1)-Polynomials.getD0(i,0)); // [0,1] interval.

	    final double [] temp=getXPnprimeplusPnCoefs(k);
	    //System.err.println("temp"+k+"="+java.util.Arrays.toString(temp));
	    double s=0;
	    for(int l=0;l<=k;l++){s+=temp[l]*pol[l]/(2*l+1);} // 1/(2*l+1) is Pl^2 integral value.
	    //System.err.println("legendreSum("+n+",pol,1)="+legendreSum(n,pol,1)+" s="+s);
	    pk2[k]=BASIS.getD0(pol,1)-s;
	}

	UnitTests.checkDoubleArrayRel1(pk1,"pk1",pk2,"pk2",eps);

	final double [] x_dpol=getXDiffF(pol);
	final double [] x_dpol_frompk=new double [n+1];
	for(int k=0;k<=n;k++)
	{
	    x_dpol_frompk[k]=pk1[k]*(2*k+1); // normalizing
	}

	UnitTests.checkDoubleArrayRel1(x_dpol,"x_dpol",x_dpol_frompk,"x_dpol_frompk",eps);
	UnitTests.checkDoubleArrayRel1(xdp,"xdp",BASIS.convertFunctionTo_Monomials_from_Basis(x_dpol),"x_dpol",eps);

	//System.err.println("pk1="+java.util.Arrays.toString(pk1)+"\npk2="+java.util.Arrays.toString(pk2)+"\ndpol="+java.util.Arrays.toString(x_dpol)+"\n"+"DD2(xdp)="+java.util.Arrays.toString(convertToLegendreFromPolynomial_Test(xdp)));



    }
    System.err.println("LegendreShifted:test_getXPnplusPn() test OK with eps="+eps);
}

@Override void print_Extra(final int nmax)
{
    printAllQuadratures(15);
}

@Override void test_Extra(final int n_tests,final int nmax)
{
    test_getLegendreFromPByIntegration(n_tests);
    test_getLegendreAveraged(n_tests);
    test_getXPnplusPn(n_tests);
}

} // END OF class TestBasisLegendreShifted


/** A unit test.
 */
public static void main(String [] args)
{
    new TestBasisLegendreShifted().runTests("$Id: LegendreShifted.java,v 1.238 2022/08/31 07:48:52 mal Exp $");
}


} // END of class LegendreShifted

