package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Monomials polynomials basis,
 *  using  \( T_n = x*T_{n-1} \).
 *  @version $Id: Monomials.java,v 1.54 2022/07/16 10:04:33 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class Monomials implements BasisPolynomials {


/** Calculates n-th Monomials polynomial
 * using  \( T_k = x*T_{k-1} \)
 * @param n T index.
 * @param Tnm1 Input \( T_{n-1} \).
 * @param Tnm2 Input \( T_{n-2} \).
 * @param flag_type2 Whether to use type 2 polynomials.
 * @param PBASIS The basis in which to calculte.
 * @return \( T_{n} \) polynomial, array dimension n+1.
 */
public double [] getNext(final int n,
			 final double [] Tnm1,final double [] Tnm2,
			 final boolean flag_type2,
			 final BasisFunctionsCalculatable PBASIS)
{
    if(flag_type2) throw new IllegalArgumentException("flag_type2=true is not supported by Monomials");
    if(n<0) throw new IllegalArgumentException();
    if(n<=0) return PBASIS.getONE();
    if(Tnm1.length!=n) throw new IllegalArgumentException();
    if(n>1 && Tnm2.length!=n-1) throw new IllegalArgumentException();
    return PBASIS.getaxbP(Tnm1,1,0);
}

public double [][] getBasisFunctionsOnPolynomialArgument(
    final int numQl,
    final double [] pol,
    final BasisFunctionsCalculatable PBASIS){

    final double [][] res=new double [numQl][];
    double [] fprev={},f=res[0]=PBASIS.getONE();
    for(int n=1;n<numQl;n++)
    {
	final double [] fmul=PBASIS.mult2Pol(pol,f);
	fprev=f;f=res[n]=fmul;
    }
    return res;
}


/**
 * Calculation of the value of n-th Monomials polynomial \( x^n \) at given x.
 * @param n T index.
 * @param x Argument.
 * @return The value of the Monomials polynomial Tn at x.
 */
public double calculate(final int n,final double x)
{
    if(n<0) throw new IllegalArgumentException("n="+n);
    double s=1;
    if(n==0) return s;
    for(int i=1;i<=n;++i)
    {
	s=s*x;
    }
    return s;
}

public double [] getMomentsFromSample(final int nmoments,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [nmoments];
    for(int k=0;k<nelsinsample;k++)
    {
	final double xx=x[k],ww=w!=null?w[k]:1;
	double s=1;
	for(int i=0;i<nmoments;++i)
	{
	    m[i]+=s*ww;
	    s=s*xx;
	}
    }
    return m;
}

/** Calculates \( \sum_k x^k Tcoefs[k] \).
 *  Two versions: original and the one copied from http://www.alglib.net/.
 * @param Tcoefs Coefficients.
 * @param x Point to estimate the sum.
 * @return The sum.
 */
public double getD0(final double [] Tcoefs,final double x)
{
    return Polynomials.getD0(Tcoefs,x);
}

/** Calculates \( d/dx \sum_k x^k Tcoefs[k] \).
 * @param Tcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The derivative.
 */
public double getD1(final double [] Tcoefs,final double x)
{
    return Polynomials.getD1(Tcoefs,x);
}


/** Calculates \( d^2/dx^2 \sum_k x^k Tcoefs[k] \).
 * @param Tcoefs Coefficients.
 * @param x Pont to estimate the sum.
 * @return The second derivative.
 */
public double getD2(final double [] Tcoefs,final double x)
{
    return Polynomials.getD2(Tcoefs,x);
}




public double [] getDifferentiated(final double [] Tcoefs)
{
    return Polynomials.getPolynomialDifferentiated(Tcoefs);
}


public double [] getIntegrated(final double [] Tcoefs)
{
    return Polynomials.getPolynomialIntegrated(Tcoefs);
}



/** Expand \( z^l z^m=\sum_{k=0}^{l+m} s_k z^k \).
 *  @param l l.
 *  @param m m.
 *  @param s coefs.
 */
public void setQlQmExpansionCoefs(final int l,final int m,final double [] s)
{
    java.util.Arrays.fill(s,0,l+m+1,0);
    s[l+m]=1;
}


/** Multiply p by ma*x+mb.
 * Calculations in Monomials.
 * Do not mistake with a and b recurrence coefficiemts
 * of {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getAB(double [][]) getAB()}.
 * @param p Input polynomial.
 * @param ma Factor at {@code x} to multiply.
 * @param mb Factor to multiply.
 * @return {@code p*(ma*x+mb)} in basis.
 */
@Override
public double [] getaxbP(final double [] p,final double ma,final double mb)
{
    final int n=p.length-1;
    final double [] res=new double [n+2];
    double r=p[n];
    res[n]=res[n+1]=0;
    for(int k=n;k>=0;k--)
    {
	res[k+1]+=ma*r;
	res[k]+=mb*r;
	if(k>0)
	{
	    final double t=p[k-1];
	    res[k-1]=0;
	    r=t;
	}
    }
    return res;
}


/** Divide  p by (x-x0), returns quotient and remainder.
 * Calculations in Monomials basis.
 * @param p Input polynomial.
 * @param x0 To divide on (x-x0)
 * @return Quotient and remainder.
 */
@Override
public SDivRes sdiv1(final double [] p,final double x0)
{
    return Polynomials.sdiv1(p,x0);
}


/** Calculates \( T_n(a(x+x0)) \) using 3 term recurrence.
 * Calculations in Monomials basis (Newton Binomial).
 * The \( b_k^{n} \) are defined as:
 * \[ T_n(a(x+x0))=\sum_{k=0}^{k=n}b_k^{n} T_k(x) \]
 * The method should be applied consequently to get \( b_k^{n} \) for n=0,1,2,3,...
 * @param n The max dimension of b.
 * @param b On input: \( b_k^{n-1} \), on output \( b_k^{n} \).
 * @param q On input: \( b_k^{n-2} \), on output: \( b_k^{n-1} \).
 */
@Override
public void setNewtonBinomialLikeCoefs(final int n,final double [] b,final double [] q,final double a,final double x0)
{
    if(n<=0)
    {
	b[n]=1;
	return;
    }
    System.arraycopy(b,0,q,0,n);
    final double [] bn=getaxbP(java.util.Arrays.copyOf(q,n),a,a*x0);
    System.arraycopy(bn,0,b,0,bn.length);
}



/** Calculate {@literal <F|xTk>} moments from {@literal <F|Tk>}.
 *  @param moments {@literal <F|Tk>}.
 *  @return {@literal <F|xTk>}.
 */
@Override
public double [] getXQkMomentsFromQkMoments(final double [] moments)
{
    final int n=moments.length-1;
    final double [] d=new double [n];
    for(int k=n;--k>=0;)
    {
	d[k]=moments[k+1];
    }
    return d;
}

/** Calculates confederate matrix to solve
 * polynomial equations in Monomials basis (companion matrix).
 * We use the three term recurrence:
 * \[ T_k=x*T_{k-1} \]
 * Use {@link PolynomialRootsConfederateMatrix#ABASISMONOMIALS ABASISMONOMIALS}
 * to calculate roots.
 */
public double [] getConfederateMatrix(final double [] coefs)
{
    return Polynomials.getConfederateMatrix(coefs);
}

static final Monomials BASIS=new Monomials();

/** A unit test.
 */
public static void main(String [] args)
{
    final double eps=1e-6;
    new TestBasisPolynomials("Monomials",new Monomials(),new java.util.Random(-79099),eps){
	@Override double [] getMonomialsMoments(final int n){return null;}
	@Override OrthogonalPolynomialsMonomialsBasis getOrtPol(){return null;}
	@Override double [][] getExact()
	{
	    return new double [][]{
		{1},
		{0,1},
		{0,0,1},
		{0,0,0,1},
		{0,0,0,0,1},
		{0,0,0,0,0,1},
		{0,0,0,0,0,0,1},
		{0,0,0,0,0,0,0,1},
		{0,0,0,0,0,0,0,0,1}
	    };
	}
    }.runTests("$Id: Monomials.java,v 1.54 2022/07/16 10:04:33 mal Exp $");
}

} // END of class Monomials
