package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Construct orthogonal polynomials in a given basis.
 * @version $Id: OrthogonalPolynomialsBasisFunctionsCalculatable.java,v 1.311 2022/07/10 23:08:38 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class OrthogonalPolynomialsBasisFunctionsCalculatable<T extends BasisFunctionsCalculatable> {


/**
 * Basis functions.
 */
public final T B;

/**
 * @param B The basis in which to construct orthogonal polynomials. We construct some other orthogonal polynomials in the basis {@code B}.
 */
public OrthogonalPolynomialsBasisFunctionsCalculatable(final T B){this.B=B;}

/** Override it for a different {@link EVSolver EVSolver} implementation (e.g. {@link EVSolver_JNI_lapacke EVSolver_JNI_lapacke}, direct JNI call to http://www.netlib.org/lapack/lapacke.html )
 * because my com/polytechnik/lapack may have problems.
 */
protected EVSolver getEVSolver(){return EVSolver.getDefaultEVSolver();}
public final EVSolver EV=getEVSolver();


/** For a given polynomial P_toproduce generate an operator (density matrix), such that
 * P_toproduce(x)=sum_i lambda^{[i]} (psi^{[i]}(x))^2
 * Such an approach allows to use Spur(f|rho) instead of {@literal <f|P_toproduce>},
 * matrix spur allows further generalization, @see
 * <a href="https://arxiv.org/abs/1807.06007">https://arxiv.org/abs/1807.06007</a>.
 * @param n the dimension.
 * @param P_toproduce a polynomial to produce, must have 0..2n-2, total 2n-1 array elements.
 * @param QQ scalar product matrix.
 * @return Lebsgue quadrature, eigenvalues/eigenvectors of which produce given polynomial.
 */
public LebesgueQuadratureWithEVData getDensityMatrixProducingGivenPolynomial(final int n,final double [] P_toproduce,final double [] QQ)
{
    if(n*n!=QQ.length) throw new IllegalArgumentException();
    return getLebesgueQuadratureFromMatrices(n,B.getQQMatr(n,B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,P_toproduce,com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQ))),QQ);
}

/** Calculates Christoffel function 1/K(x) as {@literal K(x)=Q(x)(<QQ>)^{-1}Q(x)} polynomial from {@literal <Qk>} moments.
 *  The method constructs QQ using {@link BasisFunctionsMultipliable#getQQMatr getQQMatr}, inverts is, and calls {@link BasisFunctionsMultipliable#getKK getKK}.
 *  @param n the dimension.
 *  @param moments The {@literal <Qk>} moments, {@literal dim>=2*n-1}.
 *  @return The polynomial {@literal Q(x)(<QQ>)^{-1}Q(x)}.
 */
public double [] getChristoffelOFunctionFromMoments(final int n,final double [] moments)
{
    return getChristoffelOFunctionFromMatrix(n,B.getQQMatr(n,moments));
}
public double [] getChristoffelOFunctionFromMatrix(final int n,final double [] QQ)
{
    if(n*n!=QQ.length) throw new IllegalArgumentException();
    return B.getKK(n,com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQ));
}
public double [] getChristoffelOFunctionFromPolynomials(final int n,final double [][] p)
{
    if(n>p.length) throw new IllegalArgumentException();
    final double [] kk=new double [2*n-1];
    for(int k=0;k<n;k++) TMatr.sumP1(kk,B.mult2Pol(p[k],p[k]),1);
    return kk;
}

/** Calculate first n orhogonal polynomials (0..n-1)
 * from first 2n-1 moments.
 * All we need is {@literal <Q_l Q_m>} matrix which is internally
 * constructed from mom[k]={@literal <Q_k>} by {@link BasisFunctionsMultipliable#getQQMatr getQQMatr()} using {@link BasisFunctionsMultipliable#setQlQmExpansionCoefs setQlQmExpansionCoefs()}.
 * @param n The number of polynomials to generate.
 * @param moments Moments, must be at least 2n-1 elements (0..2n-2 moments).
 * @return Orthogonal Polynomials.
 */
public double [][] getOrthogonalPolynomialsFirstKind(final int n, final double [] moments)
{
    if(moments.length<2*n-1) throw new IllegalArgumentException();
    final double [][] p;
    // Two options to calculate the same thing:
    if(false /*true*/) {
	p=getOrthogonalized(n,B.getQQMatr(n,moments));
    } else {
	p=getGramSchmidtOrthogonalized(n,B.getQQMatr(n,moments));
    }
    for(int k=0;k<p.length;k++)
    {
	TMatr.scaleVectorInPlace(p[k],getSignToMatch(k));
    }
    return p;
}

/** Calculates a single orthogonal polynomial by solving a linear system. Useful for unit tests.
 *  @param n Order of the polynomial.
 *  @param moments Moments in basis, should have at least 2*n elements in array (0..2n-1 moments).
 *  @return Polynomial, an array of n+1 length, it is normalized as P[n]=1.
 */
public double [] getSingleOrthogonalPolynomial(final int n,final double [] moments)
{
    if(moments.length<2*n) throw new IllegalArgumentException();

    final double [] matr=B.getQQMatr(n,moments),P=new double [n+1],tmp=new double [2*n];
    for(int k=0;k<n;k++)
    {
	B.setQlQmExpansionCoefs(n,k,tmp);
	P[k]=-TMatr.mulVVn(1+n+k,moments,tmp);
    }
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,n,n,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(P,Double.NaN);
	return P;
    }
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,n,n,ipvt,P,1);
    P[n]=1;
    return P;
}

/** The same as {@link #getGramSchmidtOrthogonalized getGramSchmidtOrthogonalized}, slower, but hopefully? more stable. The method is equivalent to repeatedly calling {@link #getSingleOrthogonalPolynomial getSingleOrthogonalPolynomial} with subsequent normalizing.
 * Calculations can be performed in various basises Q_k.
 * @param n The number of basis vectors to build.
 * @param QQ The {@literal <Q_i Q_j>} matrix, i,j=0..n-1, dim=n*n.
 */
public static double [][] getOrthogonalized(final int n,final double [] QQ)
{
    final int nm=n;
    if(nm!=n||n*n!=QQ.length) throw new IllegalArgumentException();

    final double [] e=new double [nm*n];
    final int [] ipvt=new int [n-1];
    // it may be more stable to solve linear system all at once several times
    // rather than solving it incrementally
    final double [][] p=new double [n][];
    for(int k=0;k<n;k++)
    {
	System.arraycopy(QQ,0,e,0,k*nm);
	final double [] pk=p[k]=new double [k+1];
	pk[k]=1;
	for(int l=0;l<k;l++) pk[l]=-QQ[k*nm+l];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(e,k,nm,ipvt);
	if(pinfo!=0) java.util.Arrays.fill(pk,Double.NaN);
	else com.polytechnik.utils.Linsystems.dgesl_m1(e,k,nm,ipvt,pk,1);
	// normalize
	final double anorm=Math.sqrt(TMatr.getAMatrA(pk,QQ,nm));
	for(int l=0;l<=k;l++) pk[l]/=anorm;
    }
    return p;
}
/** Orthogonal polynomials are defined within a factor of +-1 multiplied by.
 *  This method return this factor to match standard definition,
 *  corresponding positive sign of the coefficient with maximal x^n,
 *  when alfa_n Q_n(x) is converted to monomials.
 * @return +1 or -1.
 */
public double getSignToMatch(final int k){return 1;}



/** Calculate a and b recurrence coefficients from a Gauss quadrature.
 *  Used for unit test.
 */
public RecurrenceAB getABFromQuadrature(final double [] x,final double [] w)
{
    return getAB(getOrthogonalPolynomialsFirstKind(x.length,B.getMomentsFromSample(2*x.length,x,w)));
}

/** Calculate a and b recurrence coefficients from orthogonal polynomials.
 * The definitions as in
 * Walter Gautschi, "Orthogonal Polynomials:Computation and approximation", page 12.
 * Here a[k]=sqrt(beta_k) and b[k]=alfa_k.
 * And the recurrence is: <pre>{@code x*p_{k}=a_{k+1}*p_{k+1}+b_{k}p_{k}+a_{k}p_{k-1}}</pre> Usage example obtain three term recurrence coefficients
 * from generalized moments (modified moments)
 * in {@link OrthogonalPolynomialsChebyshevBasis Chebyshev} basis<pre>{@code
var M=new OrthogonalPolynomialsChebyshevBasis();
var ab=M.getAB(M.getOrthogonalPolynomialsFirstKind(7,new double []{1,0,0,0,0,0,0,0,0,0,0,0,0}))
ab.getAllAk() ={ 1.0, 0.7071067811865476, 0.5, 0.5, 0.5, 0.5, 0.5 }
ab.getAllBk() ={ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 }
// first kind
M.getOrthogonalPolynomialsFirstKindFromAB(ab)
={ { 1.0 },
   { 0.0, 1.414213562373095 },
   { 0.0, 0.0, 1.414213562373095 },
   { 0.0, 0.0, 0.0, 1.414213562373095 },
   { 0.0, 0.0, 0.0, 0.0, 1.414213562373095 },
   { 0.0, 0.0, 0.0, 0.0, 0.0, 1.414213562373095 },
   { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.414213562373095 }
}
// second kind
var p2=M.getOrthogonalPolynomialsSecondKindFromAB(ab)
={ { 1.414213562373095 },
   { 0.0, 2.82842712474619 },
   { 1.414213562373095, 0.0, 2.82842712474619 },
   { 0.0, 2.82842712474619, 0.0, 2.82842712474619 },
   { 1.414213562373095, 0.0, 2.82842712474619, 0.0, 2.82842712474619 },
   { 0.0, 2.82842712474619, 0.0, 2.82842712474619, 0.0, 2.82842712474619 }
}
var ab2=M.getAB(p2)
ab2.getAllAk()={ 0.7071067811865476, 0.5, 0.5, 0.5, 0.5, 0.5 }
ab2.getAllBk()={ 0.0, 0.0, 0.0, 0.0, 0.0 }
 }</pre>
 * @param p Orthogonal polynomials.
 * @return Three--term recurrece coefficients.
 */
public RecurrenceAB getAB(final double [][] p)
{
    final int n=p.length,ne=n-1;
    if(n<=0) return new RecurrenceAB(new double []{},new double []{});
    final double [] a=new double [n],b=new double [ne];
    for(int k=ne;--k>=0;)
    {
	final double [] pk=p[k],pk1=p[k+1],xpk=getxP(pk);
	b[k]=(xpk[k]-(a[k+1]=xpk[k+1]/pk1[k+1])*pk1[k])/pk[k];
	//System.err.println("k="+k+" a[k+1]="+a[k+1]+" b[k]="+b[k]);
    }
    a[0]=1/fx(p[0],0);
    return new RecurrenceAB(a,b);
}

/** Modify b[n-2] to have Jacobi matrix
 * generate (n-1)-th Radau polynomial exactly.
 * See "Journal of Computational and applied Mathematics",
 * Volume 127, Numbers 1|2, 15 January 2001,
 * D.P. Laurie, "Computation of Gauss-type quadrature formulas", p.201-217.
 */
public RecurrenceAB getABRadauModified(final int n,final RecurrenceAB ab_in)
{
    if(n>ab_in.numAk()) throw new IllegalArgumentException("n="+n+" ab_in.numAk()="+ab_in.numAk());
    if(n<=0) return new RecurrenceAB(new double []{},new double []{});
    else if(n<=1) return new RecurrenceAB(new double []{ab_in.getAk(0)},new double []{});
    final double [] a=new double [n],b=new double [n-1];
    for(int k=0;k<n;k++){
	a[k]=ab_in.getAk(k);
	if(k<n-1) b[k]=ab_in.getBk(k);
    }
    final RecurrenceAB ab=new RecurrenceAB(a,b);
    final double x0=getRadauRoot();

    //System.err.println("getABRadauModified:n="+n+" a.length="+a.length+" b.length="+b.length+" x0="+x0);
    b[n-2]+=a[n-1]*ab.calculate(n-1,x0)/ab.calculate(n-2,x0);
    return new RecurrenceAB(a,b);
}


/** Build orthogonal polynomials from a and b recurrence coefficients.
 * May not always be stable, but useful for second kind polynomials generation.
 * @param kind, 0-first kind, 1-second kind, etc.
 * @param ab recurrence coefficients.
 * @return An array of orthogonal polynomials, the number of polynomials calculated is ab.numAk()-kind.
 */
public double [][] getOrthogonalPolynomialsFromAB(final int kind,final RecurrenceAB ab)
{
    if(ab.numAk()<kind) throw new IllegalArgumentException();
    if(ab.numAk()<=kind) return new double [][]{};
    // generate p - kind-type polynomials.
    final double [][] p=new double [ab.numAk()-kind][];
    p[0]=new double []{1};p[0][0]/=(fx(p[0],0)*ab.getAk(kind));
    for(int k=1;k<p.length;k++)
    {
	final int km1=k-1;
	final double [] pk1=p[km1],pk2=k>1?p[k-2]:null,pk=p[k]=getxP(pk1);// save p_{k} to p[k];
	final double ak=ab.getAk(k+kind),bk1=ab.getBk(km1+kind),ak1=ab.getAk(km1+kind);
	pk[k]/=ak;
	pk[km1]=(pk[km1]-bk1*pk1[km1])/ak;
	for(int l=km1;--l>=0;) pk[l]=(pk[l]-bk1*pk1[l]-ak1*pk2[l])/ak;
    }
    return p;
}

public double [][] getOrthogonalPolynomialsSecondKindFromAB(final RecurrenceAB ab)
{
    return getOrthogonalPolynomialsFromAB(1,ab);
}
public double [][] getOrthogonalPolynomialsFirstKindFromAB(final RecurrenceAB ab)
{
    return getOrthogonalPolynomialsFromAB(0,ab);
}
public double [][] getSecondKindOrthogonalPolynomialsFromFirstKind(final double [][] p)
{
    return getOrthogonalPolynomialsSecondKindFromAB(getAB(p));
}

/** Recover 0 .. 2n-1 moments from 0 .. n  orthogonal polynomials.
 *  The 2n-th moment is the normalizing factor of p[n] and cannot be recovered.
 *  Used for unit tests. For production use {@link #getMomentsFromAB getMomentsFromAB}.
 */
public double [] getMomentsFromPolynomials(final double [][] p)
{
    final int n=p.length-1;
    if(n<=0) throw new IllegalArgumentException("p.length="+p.length+"<=1");
    final double [] xs=getXNodes(p[n]).re_root,ws=getWeightsChristoffel(xs,p);
    return B.getMomentsFromSample(2*xs.length,xs,ws);
}


/** Recover 0 .. 2n-1 moments from a[0..n] b[0..n-1].
 *  Used for unit tests and to recover moments for second kind
 *  orthogonal polynomials measure.
 */
public double [] getMomentsFromAB(final RecurrenceAB ab)
{
    final int n=ab.numAk()-1;
    if(n<=0) throw new IllegalArgumentException("ab.numAk()="+ab.numAk()+"<=1");
    final Quadrature nd=getGaussQuadratureFromAB(n,ab);
    return B.getMomentsFromSample(2*nd.x.length,nd.x,nd.w);
}


/** Weights by integration.
 *  Integration is replaced by synthetic division by basis[] polynomials.
 *  The coefficient by p[0] give the value of integral.
 *  The weights are normalized as {@literal sum(weights)=<1>}.
 *  Require all orthogonal polynomials to be available.
 */
public double [] getWeights(final double [] nodespolynomial,final double [] x,final double [][] p)
{
    final double [] w=new double [x.length];
    final double f=fx(p[0],0);
    for(int k=x.length;--k>=0;)
    {
	final double [] q=B.sdiv1(nodespolynomial,x[k]).q;
	final double t0=fx(q,x[k]);
	final double [] tmp=TMatr.expandOverLowerDiagBasis(q,p);
	w[k]=(tmp[0]/t0)/f;
    }
    return w;
}
/** Weights using Christoffel function.
 *  Here x are the roots of p[x.length], and
 *  the weights are normalized as {@literal sum(weights)=<1>}.
 *  Require two orthogonal polynomials to be available: p[n] &amp; p[n-1], and p[0] for normalizing.
 */
public double [] getWeightsChristoffel(final double [] x,final double [][] p)
{
    final int n=x.length;
    if(p.length<=n) throw new IllegalArgumentException();
    if(n<=0) return new double []{};
    else if(n<=1) return new double []{1/Math.pow(fx(p[0],0),2)};
    final double [] nodespolynomial=p[n],pp=p[n-1];
    final double [] ws=B.sdiv1(nodespolynomial,0).q;
    final double f=ws[n-1]/pp[n-1];
    for(int k=0;k<n;k++)
    {
	// xx is nodespolynomial root
	final double xx=x[k];
	ws[k]=f/(fx(pp,xx)*Dfx(nodespolynomial,xx));
    }
    return ws;
}
/** Weights by integration.
 *  Here the x[] are the roots of nodespolynomial polynomial.
 *  Synthetically divide nodespolynomial by (x-x[k]), then integrate Legendre interpolatory polynomial directly using known values of {@literal moments[k]=<Q_k(x)>}.
 *  Typically stable (except for exponentially small weights).
 *  The weights are normalized as sum(weights)={@literal <1>}.
 * @param nodespolynomial Nodes polynomial.
 * @param x Points.
 * @param moments Moments, at least nodespolynomial.length-1 elements must present.
 * @return The weights.
 */
public double [] getWeightsByIntegration(final double [] nodespolynomial,final double [] x,final double [] moments)
{
    final double [] w=new double [x.length];
    final int n=nodespolynomial.length-1;
    if(x.length<n || moments.length<n) throw new IllegalArgumentException();
    for(int k=x.length;--k>=0;)
    {
	final double [] q=B.sdiv1(nodespolynomial,x[k]).q;
	w[k]=TMatr.mulVVn1(q,moments)/fx(q,x[k]);
    }
    return w;
}

/** Find roots by finding the eigenvalues of Jacobi matrix, {@literal n<ab.numAk()}.
 */
public double [] getXNodesFromAB(final int n,final RecurrenceAB ab)
{
    if(ab.numAk()<=n) throw new IllegalArgumentException("n="+n+">= ab.numAk()="+ab.numAk());
    //System.err.println("a="+java.util.Arrays.toString(ab.a)+" b="+java.util.Arrays.toString(ab.b)+" nodes="+java.util.Arrays.toString(nodes));
    return EV.getEV_Thridiagonal(n,ab.getAllBk(),0,ab.getAllAk(),1,false);
}

/** Find the roots by solving polynomial equation nodespolynomial(X)=0.
 * The method calls the roots finded, then sorts the nodes
 * according to real part of the root. The result of PolynomialRootsFinder
 * may be unsorted.
 * @param nodespolynomial A polynomial in basis {@link #B B}.
 * @return The result of getXNodes: all roots are found (otherwise NaN) and then are sorted.
 */
public PolynomialRootsFinder.PRoots getXNodes(final double [] nodespolynomial)
{
    final PolynomialRootsFinder.PRoots u=getPolynomialRootsFinderInBasis().findRoots(nodespolynomial);
    if(u.nroots!=nodespolynomial.length-1) {
	final double [] re=new double [nodespolynomial.length-1],im=new double [nodespolynomial.length-1];
	java.util.Arrays.fill(re,Double.NaN);
	java.util.Arrays.fill(im,Double.NaN);
	return new PolynomialRootsFinder.PRoots(0,u.Pcoefs,re,im);
    } else {
	SortArrays.sortArrayWithLoad(u.re_root.length,u.re_root,u.im_root);
    }
    return u;
}


/**
 *  Corresponding to quadratures built on p[n] polynomial.
 *  The weights are normalized as {@literal sum(weights)=<1>}.
 *  Used for unit tests.
 */
public Quadrature.GaussQuadratureWithNodesPolynomialData getGaussQuadratureFromPolynomials(final int n,final double [][] p)
{
    if(p.length<=n) throw new IllegalArgumentException("n="+n+">= p.length="+p.length);
    final double [] nodes=p[n];
    final double [] xs=getXNodes(nodes).re_root,ws=getWeightsChristoffel(xs,p);
    return new Quadrature.GaussQuadratureWithNodesPolynomialData(xs,ws,nodes.clone());
}


public double [][] getRadauPolynomials(final double [][] p)
{
    final int n=p.length-1;
    if(n<=0) return new double [][]{};
    final double [] am=new double [n],cm=new double [n],bm=new double [n-1];
    for(int k=0;k<n;k++)
    {
	final double [] rmp=getRadauMultiplied(p[k]);
	final double a=rmp[k+1]/p[k+1][k+1],b=(rmp[k]-a*p[k+1][k])/p[k][k];
	cm[k]=b;               // <Q_k Q_k *radau>
	if(k<n-1) bm[k]=a;     // <Q_k Q_{k+1} *radau>
	if(k>0) am[k]=bm[k-1]; // <Q_k Q_{k-1} *radau>
    }
    am[0]=Double.NaN;

    final int [] in1=new int [n-1],in2=new int [n-1];
    final double [][] qq=new double [n][],rr=qq;
    final double [] tmp=new double [n-1];
    for(int k=0;k<n;k++)
    {
	final double [] qk=qq[k]=new double [k+1];
	qk[k]=1;
	if(k>0) qk[k-1]=-am[k]; // <Q_k Q_{k-1} *radau>
	com.polytechnik.utils.Progonka3.eliminate(k,am,cm,bm,qk,qk,tmp,in1,in2);
	// normalize
	double s=0;
	for(int l=0;l<=k;l++) {s+=qk[l]*(qk[l]*cm[l]+(l>0?qk[l-1]*am[l]:0)+(l<k?qk[l+1]*bm[l]:0));}
	//System.err.println("sum s="+s);
	s=Math.sqrt(s);for(int l=0;l<=k;l++) qk[l]/=s;
	//System.err.println("s="+s);

	// Multiply matrices. Because they are lower diagonal, we can do it in place.
	final double [] rk=rr[k];
	for(int l=0;l<=k;l++)
	{
	    final double [] bl=p[l];final double qkl=qk[l];
	    for(int m=0;m<l;m++) rk[m]+=qkl*bl[m];
	    rk[l]=qkl*bl[l];
	}
    }
    return rr;
}

/**
 *  Corresponding to quadratures built on Radau polynomial, constructed from p (measue: (x-x0)dmu).
 *  The weights are normalized as {@literal sum(weights)=<1>}.
 *  This approach may be less stable than direct calculation from moments.
 *  but it allows us to calculate quadratures for second kind polynomials.
 * @param n The order of generated Radau quadrature is n+1.
 * @param p Orthogonal polynomials, there are should be at least n+2 polynomials.
 * @return nodes weighs and nodes polynomial
 */
public Quadrature.GaussQuadratureWithNodesPolynomialData getRadauQuadratureFromPolynomials(final int n,final double [][] p)
{
    if(n+1>=p.length) throw new IllegalArgumentException("(n="+n+")+1 >= p.length="+p.length);
    // Calculate rr - basis polynomials for Radau measure.
    final double [][] rr=getRadauPolynomials(p);

    final double [] nodes=rr[n];
    final RecurrenceAB ab=getAB(rr);
    final double [] xr=getXNodesFromAB(n,ab),wr=true?getWeightsChristoffel(xr,rr):getWeights(nodes,xr,rr);
    // convert weigths back to out basis
    final double [] tmp1=new double []{1};tmp1[0]/=fx(tmp1,0);
    final double [] tmp=getRadauMultiplied(tmp1);
    final double x0=-fx(tmp,0)/Dfx(tmp,0);
    final double [] xs=new double [n+1],ws=new double [n+1];
    for(int k=0;k<n;k++){ws[k]=wr[k]/fx(tmp,xr[k]);xs[k]=xr[k];}
    xs[n]=x0;
    final double [] nnodes=TMatr.expandOverLowerDiagBasis(nodes,p);
    final double p00=fx(p[0],0);
    ws[n]=(nnodes[0]/fx(nodes,x0))/p00;

    return new Quadrature.GaussQuadratureWithNodesPolynomialData(xs,ws,getRadauMultiplied(nodes));
}

/** Generate Radau {@literal <(x-x0)Q_k>} moments from {@literal <Q_k>} moments.
 */
public double [] getRadauMomentsFromMoments(final double [] moments)
{
    final double [] r=new double [moments.length-1];
    for(int k=0;k<r.length;k++)
    {
	final double [] tmp=new double [k+1];tmp[k]=1;
	r[k]=TMatr.mulVVn1(getRadauMultiplied(tmp),moments);
    }
    return r;
}

/** Calculate Lebesgue quadrature.
 * If QQf=QQX the result is Gaussian quadrature,
 * otherwise it can be considered as Lebesgue quadratures.
 * @param n Points number.
 * @param QQf {@literal <fQjQk>} matrix.
 * @param QQ {@literal <QjQk>} matrix.
 * @return value-nodes, weights, eigenvaluies/eigenvectors.
 */
public LebesgueQuadratureWithEVData getLebesgueQuadratureFromMatrices(final int n,final double [] QQf,final double [] QQ)
{
    return new LebesgueQuadratureWithEVData(n,QQf.clone(),QQ.clone(),this);
}
/** Calculate Lebesgue quadrature from moments.
 * If fmoments=Xmoments the result is Gaussian quadrature,
 * otherwise it can be considered as Lebesgue quadratures.
 * @param n Points number.
 * @param fmoments {@literal <fQj>} moments.
 * @param moments {@literal <Qj>} moemnts.
 * @return value-nodes (eigenvalues), weights, eigenvaluies/eigenvectors array,QQf,QQ.
 */
public LebesgueQuadratureWithEVData getLebesgueQuadratureFromMoments(final int n,final double [] fmoments,final double [] moments)
{
    return getLebesgueQuadratureFromMatrices(n,B.getQQMatr(n,fmoments),B.getQQMatr(n,moments));
}

/** Calculate Gauss quadrature from moments in Qk(x) basis.
 *  Weights are normalized as {@literal sum(weights)=<1>}.
 *  Uses generalized eigenvalues expansion. Most stable.
 * @param n Order of the quadrature.
 * @param moments Moments in basis, the array must have at least 2n elements (0..2n-1 moments).
 * @return nodes, weights, eigenvaluies/eigenvectors array,XQQ,QQ.
 */
public Quadrature.GaussQuadratureWithEVData getGaussQuadratureFromMoments(final int n,final double [] moments)
{
    final LebesgueQuadratureWithEVData l=
	getLebesgueQuadratureFromMoments(n,B.getXQkMomentsFromQkMoments(moments),moments);
    return new Quadrature.GaussQuadratureWithEVData(l.f,l.w,l.ev,l.QQf,l.QQ);
}


/** Similar to {@link #getGaussQuadratureFromMoments getGaussQuadratureFromMoments}
 * but calculates quadratures directly: find n-th orthogonal polynomial,
 * then find roots, and finally weights.
 * Used for accuracy estimation of {@link #getGaussQuadratureFromMoments getGaussQuadratureFromMoments}.
 * @param n Ponts in the quadrature.
 * @param moments Moments in basis, must be at least 2n elements (0..2n-1 moments).
 * @return nodes, weights, and nodes polynomial.
 */
public Quadrature.GaussQuadratureWithNodesPolynomialData getGaussQuadratureFromMoments_Direct(final int n,final double [] moments)
{
    final double [] P=getSingleOrthogonalPolynomial(n,moments);
    final double [] x=getXNodes(P).re_root;
    final double [] w=getWeightsByIntegration(P,x,moments);
    return new Quadrature.GaussQuadratureWithNodesPolynomialData(x,w,P);
}



/**
 *  Corresponding to quadratures built on n-th polynomial.
 *  Weights are normalized as {@literal sum(weights)=<1>}.
 *  Uses eigenvectors expansion in orthogonal polynomials basis.
 *  Most stable.
 * @param n Quadrature order.
 * @param ab Recurrence coefficient, must have at least n+1 element,
 * e.g. obtained by {@link #getAB(double [][]) getAB()} method.
 * @return The quadrature.
 */
public Quadrature getGaussQuadratureFromAB(final int n,final RecurrenceAB ab)
{
    final double [] ev=EV.getEV_Thridiagonal(n,ab.getAllBk(),0,ab.getAllAk(),1,true);
    final double [] x=new double [n],w=new double [n],psi=new double [n];
    for(int k=0;k<n;k++)
    {
	x[k]=ev[k];
	System.arraycopy(ev,n+k*n,psi,0,n);
	w[k]=1/Math.pow(ab.getD0(psi,x[k]),2);
    }
    return new Quadrature(x,w);
}


/**
 *  Corresponding to quadratures built on n+1 -th Radau polynomial.
 *  Weights are normalized as {@literal sum(weights)=<1>}.
 *  Uses generalized eigenvalues expansion. Most stable.
 * @param n The order of generated Radau quadrature is n+1.
 * @param moments Moments, should be at least 2n+1 elements.
 * @return x-nodes, weights, nodes polynomial.
 */
public Quadrature.GaussQuadratureWithNodesPolynomialData getRadauQuadratureFromMoments(final int n,final double [] moments)
{
    final Quadrature.GaussQuadratureWithEVData nn=getGaussQuadratureFromMoments(n,getRadauMomentsFromMoments(moments));
    double [] xr=nn.x,wr=nn.w,nodes={1};
    if(n>0)
    {
	final int inda=0;
	final double [] psi0=EVXData.getPsiFromBasis(inda,nn.ev,n);
	TMatr.sumP1(nodes=getxP(psi0),psi0,-xr[inda]); // create nodes polynomial
    }
    final double [] tmp1=new double []{1};tmp1[0]/=fx(tmp1,0);
    final double [] tmp=getRadauMultiplied(tmp1);
    final double x0=-fx(tmp,0)/Dfx(tmp,0);
    final double [] xs=new double [n+1],ws=new double [n+1];
    for(int k=0;k<n;k++){ws[k]=wr[k]/fx(tmp,xr[k]);xs[k]=xr[k];}
    xs[n]=x0;
    ws[n]=TMatr.mulVVn1(nodes,moments)/fx(nodes,x0);
    return new Quadrature.GaussQuadratureWithNodesPolynomialData(xs,ws,getRadauMultiplied(nodes));
}



/** Given {@literal <Q_i|Q_j>} matrix calculate first n orthonormal vectors
 *  using Gram Schmidt orthogonalization.
 *  In polynomial basis gives first n orthogonal polynomials.
 *  Note: may be numerically unstable.
 *  @param n The number of basis vectors to generate.
 *  @param QQ The {@literal <Q_i|Q_j>} matrix, generated by {@link BasisFunctionsMultipliable#getQQMatr getQQMatr()} method from moments, i,j=0..n-1, dim=n*n.
 *  @return Othogonal basis obtained by applying Gram Schmidt orthogonalization
 *  to basis vectors in original order.
 */
public static double [][] getGramSchmidtOrthogonalized(final int n,final double [] QQ)
{
    final int nm=n;
    if(nm!=n||n*n!=QQ.length) throw new IllegalArgumentException();

    final double [][] basis=new double [n][];
    for(int i=0;i<n;i++)
    {
	final double [] b=basis[i]=new double [i+1];
	b[i]=1;
	for(int k=i,inm=i*nm;--k>=0;)
	{
	    TMatr.sumP1(b,basis[k],-TMatr.mulVV(k+1,QQ,inm,basis[k],0));
	}
	// normalize
	final double anorm=Math.sqrt(TMatr.getAMatrA(i+1,b,QQ,nm));
	for(int k=i;k>=0;--k){b[k]/=anorm;}
    }
    return basis;
}


/** Calculates single Type II multiple orthogonal polynomial.
 *  @param multiindex Multiindex. Typically all elements are the same.
 *  @param measures Moment measures {@literal <Q_k>_mu_s}. Each s-th measure should have at least sum(multiindex)+multiindex[s] elements.
 *  @return Type II multiple orthogonal polynomial.
 */
public double [] getMultipleOrthogonalPolynomial_TypeII(final int [] multiindex,final double [][] measures)
{
    int N=0;
    for(int k=multiindex.length;--k>=0;) {N+=multiindex[k];}
    final double [] P=new double [N+1],matr=new double [N*N],tmp=new double [N*2+1];
    for(int j=0;j<=N;j++)
    {
	for(int i=0,m=0;m<multiindex.length;m++)
	{
	    final int im=multiindex[m];
	    for(int k=0;k<im;k++,i++)
	    {
		B.setQlQmExpansionCoefs(j,k,tmp);
		final double mom=TMatr.mulVVn(1+j+k,measures[m],tmp);
		if(j<N)
		{
		    matr[i*N+j]=mom;
		}
		else
		{
		    P[i]=-mom;
		}
	    }
	}
    }
    final int [] ipvt=new int [N];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,N,N,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(P,Double.NaN);
	return P;
    }
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,N,N,ipvt,P,1);
    P[N]=1;
    return P;
}

/** Calculates Gaussian quadratures for multiple orthogonal polynimials.
 *  @param multiindex Multiindex, typically all elements are the same.
 *  @param measures Moment measures {@literal <Q_k>_mu_s}. Each s-th measure should have at least sum(multiindex)+multiindex[s] elements.
 *  @return Quadratures. of
 * x nodes.
 * imag(x), all equal to 0 on success, complex roots indicate failure to find proper quadrature.
 * weight[s], s=0..multiindex.length-1;
 *  multiindex.length-1+3-th element: the nodes polynomial in basis.
 *  The results tested to match exactly the
 *  Carlos F. Borges,Numer. Math. 67: 271-288 (1994),
 *  On a class of Gauss-like quadrature rules.
 *  However the results are not stable for arbitrary measures,
 *  the method {@link TestOrthogonalPolynomialsBasisFunctionsCalculatable#test_MultipleOrthogonality TestOrthogonalPolynomialsBasisFunctionsCalculatable.test_MultipleOrthogonality() }fails for some randomly generated measures.
 */
public Quadrature.MultipleOrthogonalityQuadrature getQuadraturesForMultipleOrthogonalPolynomial(final int [] multiindex,final double [][] measures)
{
    final double [] P=getMultipleOrthogonalPolynomial_TypeII(multiindex,measures);
    final int N=P.length-1;
    final PolynomialRootsFinder.PRoots u=getPolynomialRootsFinderInBasis().findRoots(P);
    final double [] rex=u.re_root,imx=u.im_root;
    if(u.nroots!=N)
    {
	java.util.Arrays.fill(rex,Double.NaN);
    }
    else
    {
	SortArrays.sortArrayWithLoad(rex.length,rex,imx);
    }

    final double [][] w=new double [multiindex.length][];
    for(int m=0;m<multiindex.length;m++)
    {
	w[m]=getWeightsByIntegration(P,rex,measures[m]);
    }
    return new Quadrature.MultipleOrthogonalityQuadrature(rex,imx,P,w);
}



/** Calculates Gauss-Kronrod quadratures.
 *  @param n Order or original Gauss Quadrature, final quadrature has 2n+1 order.
 *  @param moments Moments, must have at least 3n+2 elements (0..3n+1 moments).
 *  @return Quadratures. The 6 elements of {@link Quadrature.KronrodQuadrature Quadrature.KronrodQuadrature} returned:
 *  Gauss x nodes.
 *  imag(x) of Gauss nodes, all equal to 0 on success.
 *  weights for Gauss nodes.
 *  Kronrod x nodes.
 *  imag(x) of Kronrod nodes, all equal to 0 on success.
 *  weights for Kronrod nodes.
 */
public Quadrature.KronrodQuadrature getKronrodQuadrature(final int n,final double [] moments)
{
    final PolynomialRootsFinder rf=getPolynomialRootsFinderInBasis();

    // Gauss nodes
    final double [] P=getSingleOrthogonalPolynomial(n,moments);
    final PolynomialRootsFinder.PRoots ug=rf.findRoots(P);
    final double [] gxre=ug.re_root,gxim=ug.im_root;
    if(ug.nroots!=n)
    {
	java.util.Arrays.fill(gxre,Double.NaN);
    }
    else
    {
	SortArrays.sortArrayWithLoad(n,gxre,gxim);
    }

    // Kronrod nodes
    final double [] momentsP=B.getPQkMomentsFromQkMoments(P,moments);
    final double [] K=getSingleOrthogonalPolynomial(n+1,momentsP);
    final PolynomialRootsFinder.PRoots uk=rf.findRoots(K);
    final double [] kxre=uk.re_root,kxim=uk.im_root;
    if(uk.nroots!=(n+1))
    {
	java.util.Arrays.fill(kxre,Double.NaN);
    }
    else
    {
	SortArrays.sortArrayWithLoad(n+1,kxre,kxim);
    }

    // Weights for both: Gauss & Kronrod nodes.
    final double [] momentsK=B.getPQkMomentsFromQkMoments(K,moments);
    final double [] wg=getWeightsByIntegration(P,gxre,momentsK),
	wk=getWeightsByIntegration(K,kxre,momentsP);
    for(int k=0;k<gxre.length;k++){wg[k]/=fx(K,gxre[k]);}
    for(int k=0;k<kxre.length;k++){wk[k]/=fx(P,kxre[k]);}

    return new Quadrature.KronrodQuadrature(gxre,wg,gxim,kxre,wk,kxim);
}

// Hooks to BasisPolynomilas in case we need to override (without basis available).


/** Multiply P(x)=\sum_{k=0}^{k=n} p[k] Q_k(x) by x in Q_k(x) basis.
 */
public double [] getxP(final double [] p)
{
    return B.getaxbP(p,1,0);
}


/** Calculate f(x)=\sum_{k=0}^{k=n} m[k] Q_k(x).
 */
private double fx(final double [] m,final double x)
{
    return B.getD0(m,x);
}

/** Calculate f'(x)=\sum_{k=0}^{k=n} m[k] Q'_k(x).
 */
private double Dfx(final double [] m,final double x)
{
    return B.getD1(m,x);
}


/** Polynomial root finder in original basis.
 * Used for unit testing the quadratures.
 */
public PolynomialRootsFinder getPolynomialRootsFinderInBasis()
{
    return new PolynomialRootsConfederateMatrix(B);
}


/** Multiply nodes0 polynomial, P(x)=\sum_{k=0}^{k=n} nodes0[k] Q_k(x),
 *  by (x-x0) or (x0-x) , whatever is positive
 *  to obtain Radau nodes polynomial.
 *  The method has to be implemented if Radau quadrature is required,
 *  the default version throws an exception.
 */
public double [] getRadauMultiplied(final double [] nodes0)
{
    throw new IllegalArgumentException("The getRadauMultiplied() must be implemented if getRadauQuadratureFromPolynomials() is used");
}

/** Calculate the root of getRadauMultiplied([1]).
 */
public double getRadauRoot()
{
    final double [] rtmp=getRadauMultiplied(new double []{1});
    return -fx(rtmp,0)/Dfx(rtmp,0);
}



/** Used for diagnostics. */
public static void printQuadratures(final String txt,final int n,final double [] moments,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> q)
{
    final Quadrature qg=q.getGaussQuadratureFromMoments(n,moments),
	qr=q.getRadauQuadratureFromMoments(n-1,moments); // Note: Radu nodes have one extra.
    final double [] xg=qg.x,wg=qg.w,xr=qr.x,wr=qr.w;
    SortArrays.sortArrayWithLoad(xg.length,xg,wg);
    SortArrays.sortArrayWithLoad(xr.length,xr,wr);
    if(xr.length!=xg.length) throw new IllegalArgumentException("Mismatch");

    System.err.println("========="+txt+": Gauss quadratures for n_nodes="+n);
    System.err.println("X\tW\tS");
    double s=0;
    for(int k=0;k<n;k++)
    {
	s+=wg[k];
	System.err.println(String.valueOf(xg[k])+"\t"+wg[k]+"\t"+s);
    }
    //UnitTests.checkDoubleRel(s,"s",1,"1",eps);

    System.err.println("========="+txt+": Radau quadratures for n_nodes="+n);
    System.err.println("X\tW\tS");
    s=0;
    for(int k=0;k<n;k++)
    {
	s+=wr[k];
	System.err.println(String.valueOf(xr[k])+"\t"+wr[k]+"\t"+s);
    }
    //UnitTests.checkDoubleRel(s,"s",1,"1",eps);

    System.err.println();
    System.err.println();
}

/** Used for diagnostics. */
public static void printAB(final String header,RecurrenceAB ab){
    System.err.println();
    final double [] a=ab.getAllAk(),b=ab.getAllBk();
    System.err.println(header);
    System.err.println("K\tA\tB");
    for(int k=0;k<Math.max(a.length,b.length);k++)
    {
	System.err.println(String.valueOf(k)+"\t"+(k<a.length?a[k]:"N/A")+"\t"+(k<b.length?b[k]:"N/A"));
    }
}

}
