package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Polynomial roots in arbitrary basis.
 *  @version $Id: PolynomialRootsConfederateMatrix.java,v 1.47 2022/07/22 07:33:43 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class PolynomialRootsConfederateMatrix implements PolynomialRootsFinder {

    private final ConfederateMatrixCalculatable B;

    public PolynomialRootsConfederateMatrix(final ConfederateMatrixCalculatable B){this.B=B;}

/** Override it for a different EVSolver implementation (e.g. direct JNI call to
<a href="http://www.netlib.org/lapack/lapacke.html">http://www.netlib.org/lapack/lapacke.html</a>) as in JNI version
{@link com.polytechnik.lapack.Eigenvalues_JNI_lapacke  com.polytechnik.lapack.Eigenvalues_JNI_lapacke}
 * because my com/polytechnik/lapack may have problems.
 */
protected EVSolver getEVSolver(){return EVSolver.getDefaultEVSolver();}


    public PRoots findRoots(final double [] op)
    {
	final int n=op.length-1;
	final double [] re_root=new double [n],im_root=new double [n],polynomial_coefs=op.clone();
	final double a=polynomial_coefs[n];
	if(!(a!=0)) // think of NaN
	{
	    java.util.Arrays.fill(re_root,0,n,Double.NaN);
	    java.util.Arrays.fill(im_root,0,n,Double.NaN);
	    return new PRoots(0,polynomial_coefs,re_root,im_root);
	}
	final double [] matr=B.getConfederateMatrix(polynomial_coefs);

	// note, that fortran notation is a transponse (first index change most often)
	// to java one, which we use (last index change most often),
	// but eigenvalues are the same and we do not need eigenvectors here.
	//final int info=com.polytechnik.lapack.Eigenvalues.getNonSymEV_dgees(n,matr,n,re_root,im_root)
	final int info=getEVSolver().getNonSymEV(n,matr,re_root,im_root);
	if(info!=0)
	{
	    java.util.Arrays.fill(re_root,0,n,Double.NaN);
	    java.util.Arrays.fill(im_root,0,n,Double.NaN);
	    return new PRoots(0,polynomial_coefs,re_root,im_root);
	}
	return new PRoots(n,polynomial_coefs,re_root,im_root);
    }

public static final PolynomialRootsFinder
ABASISMONOMIALS=new PolynomialRootsConfederateMatrix(new Monomials()),
ABASISLAGUERRE=new PolynomialRootsConfederateMatrix(new Laguerre()),
ABASISLEGENDRESHIFTED=new PolynomialRootsConfederateMatrix(new LegendreShifted()),
ABASISCHEBYSHEV=new PolynomialRootsConfederateMatrix(new Chebyshev()),
ABASISLEGENDRE=new PolynomialRootsConfederateMatrix(new Legendre()),
ABASISHERMITEE=new PolynomialRootsConfederateMatrix(new HermiteE());


private static void test_Solvers(final int n,final int n_tests,final java.util.Random ra,final double eps)
{
    for(int kk=0;kk<n_tests;kk++)
    {

    final double [] pol=new double [n+1];
    for(int k=pol.length;--k>=0;) pol[k]=(1-2*ra.nextDouble())/(k+1);
    pol[n]+=3.2;

    PRoots r=null;
    r=new PolynomialRootsJenkinsTraubORIG().findRoots(pol);
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    final double [] root_re=r.re_root,root_im=r.im_root;

    r=ABASISMONOMIALS.findRoots(pol);
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    UnitTests.checkDoubleArrayRel1(r.re_root,"r.re_root",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(r.im_root,"r.im_root",root_im,"root_im",eps);


    r=ABASISLAGUERRE.findRoots(new Laguerre().convertFunctionTo_Basis_from_Monomials(pol));
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    UnitTests.checkDoubleArrayRel1(r.re_root,"r.re_root",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(r.im_root,"r.im_root",root_im,"root_im",eps);


    r=ABASISLEGENDRESHIFTED.findRoots(new LegendreShifted().convertFunctionTo_Basis_from_Monomials(pol));
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    UnitTests.checkDoubleArrayRel1(r.re_root,"r.re_root",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(r.im_root,"r.im_root",root_im,"root_im",eps);


    r=ABASISCHEBYSHEV.findRoots(new Chebyshev().convertFunctionTo_Basis_from_Monomials(pol));
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    UnitTests.checkDoubleArrayRel1(r.re_root,"r.re_root",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(r.im_root,"r.im_root",root_im,"root_im",eps);

    r=ABASISLEGENDRE.findRoots(new Legendre().convertFunctionTo_Basis_from_Monomials(pol));
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    UnitTests.checkDoubleArrayRel1(r.re_root,"r.re_root",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(r.im_root,"r.im_root",root_im,"root_im",eps);

    r=ABASISHERMITEE.findRoots(new HermiteE().convertFunctionTo_Basis_from_Monomials(pol));
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    UnitTests.checkDoubleArrayRel1(r.re_root,"r.re_root",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(r.im_root,"r.im_root",root_im,"root_im",eps);


    r=ABASISMONOMIALS.findRoots(new Monomials().convertFunctionTo_Basis_from_Monomials(pol));
    if(r.nroots!=n) throw new RuntimeException("r.nroots="+r.nroots+" n="+n);
    SortArrays.sortArrayWith2Load(r.nroots,TMatr.sumAB(r.nroots,r.re_root,r.im_root),r.re_root,r.im_root);
    UnitTests.checkDoubleArrayRel1(r.re_root,"r.re_root",root_re,"root_re",eps);
    UnitTests.checkDoubleArrayRel1(r.im_root,"r.im_root",root_im,"root_im",eps);

    }

    System.err.println("PolynomialRootsConfederateMatrix:test_Solvers() test OK for N="+n+" n_tests="+n_tests+" with eps="+eps);

}

private static void printAllQuadratures(final int nmax,final java.util.Random r)
{
    for(int n=1;n<=nmax;n++)
    {
	for(var o:java.util.List.of(
		new OrthogonalPolynomialsLegendreShiftedBasis(),
		new OrthogonalPolynomialsLaguerreBasis(),
		new OrthogonalPolynomialsLegendreBasis(),
		new OrthogonalPolynomialsChebyshevBasis(),
		new OrthogonalPolynomialsHermiteEBasis())){
	    final double [] moments=new double [2*n];
	    moments[0]=1;
	    OrthogonalPolynomialsBasisFunctionsCalculatable.printQuadratures(o.getClass().getSimpleName(),n,moments,o);
	}
    }
}


    public static void main(String [] args)
    {
	System.err.println("PolynomialRootsConfederateMatrix: starting ALL tests; ver=$Id: PolynomialRootsConfederateMatrix.java,v 1.47 2022/07/22 07:33:43 mal Exp $");
	final long t0=System.currentTimeMillis();

	final java.util.Random r=new java.util.Random(8631);

	printAllQuadratures(80,r);

	final double eps=1e-4;
	for(int n=1;n<=8;n++)
	{
	    test_Solvers(n,110000,r,eps);
	}
	final long t1=System.currentTimeMillis();
	System.err.println("PolynomialRootsConfederateMatrix: tests OK done in "+
			   (t1-t0)+" milliseconds. ver=$Id: PolynomialRootsConfederateMatrix.java,v 1.47 2022/07/22 07:33:43 mal Exp $");
    }
}
