package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2014
 * This file is under GPL version 3.
 *
 */

/** Probability(or Value) Correlation of p and i.
 * Actually is the case d=2 of LebesgueQuadratureWithEVData,
 * the corr is the difference of diagonal and off-diagonal elemets: (diag-offdiag)/(diag+offdiag) elemets,
 * depnding of flag_probabiliy_correlation flag,
 * of the matrix returned by either getProbabilityCorrelation or getValueCorrelation
 * of LebesgueQuadratureWithEVData.
 * @version $Id: ProbabilityCorrelation.java,v 1.55 2022/06/27 08:50:40 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class ProbabilityCorrelation {
    /** Min/Max eigenvalue.
     */
    public final double pL,pH,iL,iH;
/** W=w_pL_iL+w_pL_iH+w_pH_iL+w_pH_iH.
 *  corr=(w_pL_iL-w_pL_iH-w_pH_iL+w_pH_iH)/W.
 */
public final double W,w_pL_iL,w_pL_iH,w_pH_iL,w_pH_iH,corr;
/** If true, probability correlation is calculated,
 * otherwise value correlation.
 */
public final boolean flag_probabiliy_correlation;

/** Obtain probability correlation from three 2x2 matrices:
 * https://arxiv.org/abs/1709.06759,
 * Appendix C: "Probability Correlation of Variables".
 * For probability correlation (flag_probabiliy_correlation=true),
 * no basis functions required, all we need is eigenvalues problem solver M.EV,
 * use a standard one. This is not the case for value correlation.
 */
public ProbabilityCorrelation(final double [] QQ,final double [] QQp,final double [] QQi){
    this(QQ,QQp,QQi,Skewness.MB,true);
}


/** Probaility Correlation (from monomials moments).
 */
public ProbabilityCorrelation(final double x0,final double x1,final double x2,
			      final double px0,final double px1,final double px2,
			      final double ix0,final double ix1,final double ix2,
			      final boolean flag_probabiliy_correlation)
{
    this(x0,x1,x2,px0,px1,px2,ix0,ix1,ix2,Skewness.MB,flag_probabiliy_correlation);
}

/** Probaility Correlation (from the moments in arbitrary basis, the Q_0(x) is assumed to be a constant!).
 */
public ProbabilityCorrelation(final double x0,final double x1,final double x2,
			      final double px0,final double px1,final double px2,
			      final double ix0,final double ix1,final double ix2,
			      final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			      final boolean flag_probabiliy_correlation)
{
    this(x0,x1,x2,px0,px1,px2,ix0,ix1,ix2,M,flag_probabiliy_correlation,0);
}

/** Probaility Correlation (from the moments in arbitrary basis, the Q_0(x) is assumed to be a constant!).
 */
protected ProbabilityCorrelation(final double x0,final double x1,final double x2,
				 final double px0,final double px1,final double px2,
				 final double ix0,final double ix1,final double ix2,
				 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				 final boolean flag_probabiliy_correlation,
				 final double eps_if_selftest_required)
{
    this(M.B.getQQMatr(2,new double []{x0,x1,x2}),
	 M.B.getQQMatr(2,new double []{px0,px1,px2}),
	 M.B.getQQMatr(2,new double []{ix0,ix1,ix2}),M,flag_probabiliy_correlation,eps_if_selftest_required);
}

/** From 2x2 matrices in arbitraty basis,
 * the Q_0(x) is assumed to be a constant, otherwise averages will be incorrect.
 * Unit selftest is disabled.
 */
public ProbabilityCorrelation(final double [] QQ,final double [] QQp,final double [] QQi,
			      final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
			      final boolean flag_probabiliy_correlation)
{
    this(QQ,QQp,QQi,M,flag_probabiliy_correlation,0);
}


/** From 2x2 matrices in arbitraty basis,
 *  if {@literal eps_if_selftest_required>0}, then unit self-test with this eps is run.
 */
protected ProbabilityCorrelation(final double [] QQ,final double [] QQp,final double [] QQi,
				 final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M,
				 final boolean flag_probabiliy_correlation,
				 final double eps_if_selftest_required)
{
    final int d=2;
    if(QQ.length!=d*d || QQp.length!=d*d || QQi.length!= d*d)
	throw new IllegalArgumentException("Matrix dimension is not 2");
    this.flag_probabiliy_correlation=flag_probabiliy_correlation;
    final double [] ev_MiM=M.EV.getGEV_Generalized(d,QQi,QQ,true),
	bQQp= /* Transform the basis to optimize calculations. */EVXData.convertMatrFromQtoB_Basis(d,QQp,ev_MiM);
    iL=ev_MiM[0];
    iH=ev_MiM[d-1];

    final double [] bev_MpMt=M.EV.getEV(d,bQQp,true);
    pL=bev_MpMt[0];
    pH=bev_MpMt[d-1];
    if(flag_probabiliy_correlation)
    {
	this.W=d;
	w_pL_iL=Math.pow(bev_MpMt[d+0*d+0],2);
	w_pL_iH=Math.pow(bev_MpMt[d+0*d+1],2);
	w_pH_iL=Math.pow(bev_MpMt[d+1*d+0],2);
	w_pH_iH=Math.pow(bev_MpMt[d+1*d+1],2);
	corr=(w_pL_iL-w_pL_iH-w_pH_iL+w_pH_iH)/d;
    } else {
	final double [] ONE=M.B.getONE();
	this.W=TMatr.getAMatrA(ONE,QQ,d);
	final double bONE[]=EVXData.convertPsiFromQtoB_Basis(d,java.util.Arrays.copyOf(ONE,d),ev_MiM,QQ),
	    pL_a=TMatr.mulVV(d,bONE,0,bev_MpMt,d+0*d),
	    pH_a=TMatr.mulVV(d,bONE,0,bev_MpMt,d+1*d),
	    iL_a=TMatr.getAMatrB(ONE.length,ONE,0,d,ev_MiM,d+0*d,QQ,d),
	    iH_a=TMatr.getAMatrB(ONE.length,ONE,0,d,ev_MiM,d+1*d,QQ,d);
	w_pL_iL=pL_a*bev_MpMt[d+0*d+0]*iL_a;
	w_pL_iH=pL_a*bev_MpMt[d+0*d+1]*iH_a;
	w_pH_iL=pH_a*bev_MpMt[d+1*d+0]*iL_a;
	w_pH_iH=pH_a*bev_MpMt[d+1*d+1]*iH_a;
	corr=(w_pL_iL-w_pL_iH-w_pH_iL+w_pH_iH)/(w_pL_iL+w_pL_iH+w_pH_iL+w_pH_iH);
    }
    if(eps_if_selftest_required>0 && flag_probabiliy_correlation)
    {
	final double eps=eps_if_selftest_required;
	// Test normalizing
	UnitTests.checkDoubleRel(w_pL_iL+w_pL_iH,"w_pL_iL+w_pL_iH",1,"1",eps);
	UnitTests.checkDoubleRel(w_pH_iL+w_pH_iH,"w_pH_iL+w_pH_iH",1,"1",eps);
	UnitTests.checkDoubleRel(w_pL_iL+w_pH_iL,"w_pL_iL+w_pH_iL",1,"1",eps);
	UnitTests.checkDoubleRel(w_pL_iH+w_pH_iH,"w_pL_iH+w_pH_iH",1,"1",eps);
	// Test calculations in original basis
	final double [] ev_MpM=M.EV.getGEV_Generalized(d,QQp,QQ,true);
	UnitTests.checkDoubleRel(w_pL_iL,"w_pL_iL",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+0*d,ev_MiM,d+0*d,QQ,d),2),"w_pL_iL",eps);
	UnitTests.checkDoubleRel(w_pL_iH,"w_pL_iH",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+0*d,ev_MiM,d+1*d,QQ,d),2),"w_pL_iH",eps);
	UnitTests.checkDoubleRel(w_pH_iL,"w_pH_iL",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+1*d,ev_MiM,d+0*d,QQ,d),2),"w_pH_iL",eps);
	UnitTests.checkDoubleRel(w_pH_iH,"w_pH_iH",Math.pow(TMatr.getAMatrB(d,ev_MpM,d+1*d,ev_MiM,d+1*d,QQ,d),2),"w_pH_iH",eps);
	final double corr_test=(TMatr.getAMatrA(d,ev_MiM,d+0*d,QQp,d)-TMatr.getAMatrA(d,ev_MiM,d+1*d,QQp,d))
	    /(ev_MpM[0]-ev_MpM[1]);
	UnitTests.checkDoubleRel(corr,"corr",corr_test,"corr_test",eps);
	System.err.println("ProbabilityCorrelation:(<phi_IL|p|<phi_IL> - <phi_IH|p|<phi_IH>)/(p_min-p_max)="+corr_test+" corr="+corr);
    }

    if(eps_if_selftest_required>0)
    {
	final double eps=eps_if_selftest_required;
	final double [] ONE=M.B.getONE();
	final double _W=TMatr.getAMatrA(ONE,QQ,d);
	final LebesgueQuadratureWithEVData
	    i=new LebesgueQuadratureWithEVData(d,QQi,QQ,M),
	    p=new LebesgueQuadratureWithEVData(d,QQp,QQ,M);
	final LebesgueQuadratureJointDistribution jdist=new LebesgueQuadratureJointDistribution(p,i);
	final double [] corr_matr=flag_probabiliy_correlation?
	    jdist.getProbabilityCorrelation():jdist.getValueCorrelation();
	UnitTests.checkDoubleArrayRel1(corr_matr,"corr_matr",new double []{w_pL_iL,w_pL_iH,w_pH_iL,w_pH_iH},"{w_pL_iL,w_pL_iH,w_pH_iL,w_pH_iH}",eps);
	UnitTests.checkDoubleRel(this.W,"this.W",TMatr.sumElements(corr_matr),"sum(corr_matr)",eps);
	UnitTests.checkDoubleRel(_W,"_W",TMatr.sumElements(i.w),"sum(i.w)",eps);
	UnitTests.checkDoubleRel(_W,"_W",TMatr.sumElements(p.w),"sum(p.w)",eps);
	UnitTests.checkDoubleRel(corr,"corr",(corr_matr[0]-corr_matr[1]-corr_matr[2]+corr_matr[3])/TMatr.sumElements(corr_matr),"(diag-offdiag)/(diag+offdiag)",eps);
	System.err.println("ProbabilityCorrelation:flag_probabiliy_correlation="+flag_probabiliy_correlation+" matched LebesgueQuadratureWithEVData OK eps="+eps);
    }
}


public static void matchProbabilityCorrelation(final ProbabilityCorrelation a,final ProbabilityCorrelation b,final double eps)
{
    matchProbabilityCorrelationWithOffset(a,0,b,0,eps);
}
public static void matchProbabilityCorrelationWithOffset(final ProbabilityCorrelation a,final double off_pa,final ProbabilityCorrelation b,final double off_pb,final double eps)
{
    final double poffset_shift=off_pb-off_pa;
    UnitTests.checkDoubleRel(a.pL,"a.pL",b.pL+poffset_shift,"b.pL+poffset_shift",eps);
    UnitTests.checkDoubleRel(a.pH,"a.pH",b.pH+poffset_shift,"b.pH+poffset_shift",eps);
    UnitTests.checkDoubleRel(a.iL,"a.iL",b.iL,"b.iL",eps);
    UnitTests.checkDoubleRel(a.iH,"a.iH",b.iH,"b.iH",eps);
    UnitTests.checkDoubleRel(a.w_pL_iL,"a.w_pL_iL",b.w_pL_iL,"b.w_pL_iL",eps);
    UnitTests.checkDoubleRel(a.w_pL_iH,"a.w_pL_iH",b.w_pL_iH,"b.w_pL_iH",eps);
    UnitTests.checkDoubleRel(a.w_pH_iL,"a.w_pH_iL",b.w_pH_iL,"b.w_pH_iL",eps);
    UnitTests.checkDoubleRel(a.w_pH_iH,"a.w_pH_iH",b.w_pH_iH,"b.w_pH_iH",eps);
    UnitTests.checkDoubleRel(a.W,"a.W",b.W,"b.W",eps);

    System.err.println("ProbabilityCorrelation:matchProbabilityCorrelation test OK eps="+eps);
}
}
