package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */

/** A simple ML example of the Spectral Radon-Nikodym approach. A "driver" class.
  * @version $Id: RN.java,v 1.80 2024/05/01 09:06:58 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
  */
public class RN {


/** Print RN model.
 * @param rn_data_file_to_build_model_from The filename the model is built from (used for diag).
 * @param xHeader Columns header (attributes names).
 * @param rn The model.
 * @param wK Christoffel weights, an alternative to Lebesgue weights rn.w  (obtained as rn.getChristoffelWeights(d)).
 * @param flag_print_verbosity Whether to print the psi_f[k](x) for every point or original x expansion in psi_f[k].
 * @param evaluation_data_file The filename on which to evaluate (used for diag).
 * @param getoutputfilename A function to generate output file name (typically evaluation_data_file+suffix).
 */
public static void printRNEvaluation(
    final String [] comand_line_arguments,
    final String rn_data_file_to_build_model_from,
    final java.util.List<String> xHeader,
    final RadonNikodymSpectralModel rn,
    final double [] wK,
    final double [][] getxorigprojectedtopsik,
    final int flag_print_verbosity,
    final java.util.List<ObservationVectorXF> evaluation_data,
    final String evaluation_data_file,
    final java.util.function.Function<String,String> getoutputfilename) throws java.io.IOException{

    final String s=DataReadObservationVectorXF.SEPARATOR;
    final String fout=getoutputfilename.apply(new java.io.File(evaluation_data_file).getName());
    System.err.println("Writing to "+fout+" evaluation_data "+evaluation_data_file+" model is built from "+rn_data_file_to_build_model_from);
    final java.io.PrintWriter out=new java.io.PrintWriter(fout);
    out.println("|@RN_command_line_arguments:"+s+java.util.Arrays.toString(comand_line_arguments));
    out.println("|@version:"+s+"$Id: RN.java,v 1.80 2024/05/01 09:06:58 mal Exp $");
    out.println("|");
    out.println("|MODEL");
    out.println("|rn_data_file_to_build_model_from"+s+rn_data_file_to_build_model_from);
    out.println("|totalweight"+s+TMatr.sumElements(rn.w));
    out.println("|aver_f_FromLebesgue"+s+(TMatr.mulVV(rn.f,rn.w)/TMatr.sumElements(rn.w)));
    out.println("|noriginal"+s+rn.noriginal);
    out.println("|nX"+s+rn.nX);
    double sumf=0,sumW=0,sumWk=0;
    for(int k=0;k<rn.nX;k++){
	out.println("|Lebesgue_f["+k+"]"+s+rn.f[k]+s+"Lebesgue_w["+k+"]"+s+rn.w[k]+s+"wK["+k+"]"+s+wK[k]);
	sumW+=rn.w[k];
	sumWk+=wK[k];
	sumf+=rn.f[k];
    }
    out.println("|SUM"+s+sumf+s+s+sumW+s+s+sumWk);
    if(flag_print_verbosity>=3){
	out.println("|Projections of xorig to psi_f[k] (inverse transform)");
	out.println("|Matches exactly ONLY if psi_f[k] is full basis");
	out.print("|");
	for(int k=0;k<rn.nX;k++){out.print(s+"psi_f["+k+"]");}
	out.println();
	for(int j=0;j<rn.noriginal;j++){
	    out.print("|"+xHeader.get(j));
	    for(int k=0;k<rn.nX;k++){out.print(s+getxorigprojectedtopsik[j][k]);}
	    out.println();
	}
    }
    out.println("|");
    out.println("|EVALUATION");
    out.println("|rn_data_file_to_build_model_from"+s+rn_data_file_to_build_model_from);
    out.println("|evaluation_data_file"+s+evaluation_data_file);
    out.println("|sample stat:");
    double we=0,fsum=0,f2sum=0,fmflssum=0;
    for(ObservationVectorXF o:evaluation_data){
	we+=o.weight;
	fsum+=o.f*o.weight;
	f2sum+=Math.pow(o.f,2)*o.weight;
	fmflssum+=Math.pow(rn.getRNatXoriginal(o.x).getLS()-o.f,2)*o.weight;
    }
    out.println("|f_average"+s+(fsum/we));
    out.println("|f^2 average"+s+(f2sum/we));
    out.println("|(f-fLS)^2 average"+s+(fmflssum/we));
    out.println("|totalweight"+s+we);
    out.println("|points"+s+evaluation_data.size());
    out.println("|");
    out.print("|");
    out.print("Observation Label"+s);
    for(String h:xHeader){
	out.print(h);out.print(s);
    }
    out.print("f (Class Label)"+s+"weight"+s+"f_RN"+s+"f_LS"+s+"Christoffel"+s+"f_RNW"+s+"Coverage");
    if(flag_print_verbosity>=2){
	for(int k=0;k<rn.nX;k++){out.print(s+"psi_f["+k+"](x)");}
    }
    out.println();

    // print points
    for(ObservationVectorXF o:evaluation_data){
	if(o.x.length!=xHeader.size()) throw new RuntimeException("Header/Data length mismatch");
	out.print((o.label!=null?o.label:"??")+s);
	for(double x:o.x){
	    out.print(x);out.print(s);
	}
	final RadonNikodymSpectralModel.RNPointEvaluation e=rn.getRNatXoriginal(o.x);
	final double [] psikAtX=e.getPsikAtX();

	out.print(String.valueOf(o.f)+s+o.weight+
		  s+e.getRN()+
		  s+e.getLS()+
		  s+String.valueOf(1/e.getChristoffelOatX())+
		  s+e.getRNW()+
		  s+TMatr.sumElements(e.getPosteriorDistributionWeights()));
	if(flag_print_verbosity>=2){
	    for(int k=0;k<rn.nX;k++){out.print(s+psikAtX[k]);}
	}
	out.println();
    }
    out.close();
}

/** The driver for Radon--Nikodym.
 */
public static void main(final String [] args) throws java.io.IOException
{
    final RNConfig c=new RNConfig(args);
    System.err.println("description\n"+c.getDescription());
    final java.util.List<String> errs=c.validateOptions();
    if(errs.size()>0)
    {
	System.err.println("Invalid or not set arguments:\n"+String.join("\n",errs));
	System.exit(1);
    }

    // Read the data from c.data_file_to_build_model_from and test data from c.data_file_evaluation
    DataReadObservationVectorXF d=
	DataReadObservationVectorXF.getObservations(c.data_file_to_build_model_from,c.columns_number,c.getXIndexesSequence(),c.indcol_f,c.indcol_weight,c.indcol_label),
	d_test[]=new DataReadObservationVectorXF[c.data_file_evaluation.length];
    for(int k=0;k<c.data_file_evaluation.length;k++){
	d_test[k]=DataReadObservationVectorXF.getObservations(c.data_file_evaluation[k],c.columns_number,c.getXIndexesSequence(),c.indcol_f,c.indcol_weight,c.indcol_label);
	if(!d.xHeader.equals(d_test[k].xHeader))
	    throw new RuntimeException("The files data_file_to_build_model_from="+c.data_file_to_build_model_from+" and data_file_evaluation="+c.data_file_evaluation[k]+" have different headers:"+d.xHeader+" vs. "+d_test[k].xHeader);
    }

    if(c.max_multiindex>0){
	// Multiindex products are required.
	// We can get either from regularized data or from RN model.
	// The result is the same, but numerical stability seems to be better for Radon-Nikodym basis.
	final AttributesProductsMultiIndexed dr;
	// These three options to give identical results:
	if(c.regularization_method==RegularizationMethod.NONE){
	    // this one to fail for degenerated data or a data without const in basis
	    dr=d.getRadonNikodymSpectralModel_WithoutRegularization();
	} else if(c.flag_replace_f_by_christoffel_function||c.flag_assume_f_is_diagonal_in_christoffel_function_basis){
	    // minimal calculations, also for RegularizationMethod.LIN
	    // even no eigenproblem solution is required!
	    dr=getDataRegularizedAccordingToConfig(d,c.regularization_method);
	} else {
	    // seems to be more stable
	    dr=getDataRegularizedAccordingToConfig(d,c.regularization_method).getRadonNikodymSpectralModel();
	}
	d=dr.getAttribsProductsMultiIndexed(c.max_multiindex,d);
	for(int k=0;k<c.data_file_evaluation.length;k++){
	    d_test[k]=dr.getAttribsProductsMultiIndexed(c.max_multiindex,d_test[k]);
	    if(!d.xHeader.equals(d_test[k].xHeader))
		throw new RuntimeException("The files data_file_to_build_model_from="+c.data_file_to_build_model_from+" and data_file_evaluation="+c.data_file_evaluation[k]+" have different headers:"+d.xHeader+" vs. "+d_test[k].xHeader);
	}
    }

    RadonNikodymSpectralModel rn;
    // Construct Radon--Nikodym spectral model out of data in d
    if(c.flag_assume_f_is_diagonal_in_christoffel_function_basis){
	// Christoffel function spectrum is required.
	// Replace f by a one diagonal in the Christoffel function basis
	rn=// for non--degenerated data the result is identical:
	    (c.regularization_method==RegularizationMethod.NONE?
	     d.getRadonNikodymSpectralModel_WithoutRegularization()
	     :
	     getDataRegularizedAccordingToConfig(d,c.regularization_method)
	     ).getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis(d);
    }
    else if(c.flag_replace_f_by_christoffel_function){
	// Christoffel function spectrum is required.
	// Replace f by the Christoffel function.
	final RadonNikodymSpectralModel _rn=rn=// for non--degenerated data the result is identical:
	    (c.regularization_method==RegularizationMethod.NONE?
	     d.getRadonNikodymSpectralModel_WithoutRegularization()
	    :
	     getDataRegularizedAccordingToConfig(d,c.regularization_method)
	     ).getRadonNikodymSpectralModelWithChristoffelFunctionAsF(d);

	// replace f by Christoffel function in learning and in test data.
	final java.util.function.Function<ObservationVectorXF,ObservationVectorXF> replacementF=
	    (o)->new ObservationVectorXF(o.label,o.x,1/_rn.getRNatXoriginal(o.x).getChristoffelOatX(),o.weight);
	d=d.getReplacedData(replacementF,h->h);
	for(int k=0;k<c.data_file_evaluation.length;k++){
	    d_test[k]=d_test[k].getReplacedData(replacementF,h->h);
	}
    } else {
	// regular model, use the read f
	rn=// for non--degenerated data the result is identical:
	    (c.regularization_method==RegularizationMethod.NONE?
	    d.getRadonNikodymSpectralModel_WithoutRegularization()
	    :
	    getDataRegularizedAccordingToConfig(d,c.regularization_method).getRadonNikodymSpectralModel()
	    );
    }

    if(c.clusters_number>0){
	// Optimal clustering is required.
	System.err.println("RN:The clusters_number="+c.clusters_number+" trying to redice basis dimension from "+rn.nX+" to "+c.clusters_number+". Original x dimension="+rn.noriginal);
	rn=rn.reduceBasisSize(c.clusters_number);
    }
    // Evaluate and print the data on the model built
    printRNEvaluation(args,
		      c.data_file_to_build_model_from,
		      d.xHeader,
		      rn,
		      rn.getChristoffelWeights(d),
		      rn.getAllXorigProjectedToPsik(d),
		      c.flag_print_verbosity,
		      d.data,
		      c.data_file_to_build_model_from,
		      (fn)->c.output_files_prefix+fn+".RN.csv");
    // Evaluate on test data.
    for(int k=0;k<c.data_file_evaluation.length;k++){
	printRNEvaluation(args,
			  c.data_file_to_build_model_from,
			  d.xHeader,
			  rn,
			  rn.getChristoffelWeights(d),
			  rn.getAllXorigProjectedToPsik(d),
			  c.flag_print_verbosity,
			  d_test[k].data,
			  c.data_file_evaluation[k],
			  (fn)->c.output_files_prefix+fn+".RN.csv");
    }
}


    public static RegularizedVectorXF getDataRegularizedAccordingToConfig(final DataReadObservationVectorXF d,final RegularizationMethod regularization_method){
	final var t0=System.currentTimeMillis();
	final RegularizedVectorXF reg=switch(regularization_method){
	   case EV -> d.getDataRegularized_EV();
	   case LIN -> d.getDataRegularized_LIN();
	   default -> throw new IllegalArgumentException("Unknow regularization method "+regularization_method);
	};
	final var t1=System.currentTimeMillis();System.err.println("RN:getDataRegularizedAccordingToConfig "+regularization_method+" noriginal="+reg.noriginal+" n="+reg.nX+"; regularization done in "+(t1-t0));
	return reg;
    }
}
