package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */

/** Radon-Nikodym Spectral Model
<a href="https://arxiv.org/abs/1906.00460">https://arxiv.org/abs/1906.00460</a>.
  * @version $Id: RadonNikodymSpectralModel.java,v 1.90 2022/11/02 16:54:51 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
 */
public class RadonNikodymSpectralModel implements AttributesProductsMultiIndexed,RNPointEvaluatable,FtoChristoffelFunctionTransformable {
    public final int nX,noriginal;
    /** Input matrix. */
    public final double [] XX,XfX;
    /** Lebesgue quadrature. See <a href="https://arxiv.org/abs/1807.06007">https://arxiv.org/abs/1906.00460</a>.
     */
    public final double [] f,w,psiaverage;
    final java.util.function.Function<double [],double []> convertxFromOriginalToX;
    final double [] ev;

    /** Construct the model from
{@literal <X|X>} and {@literal <X|f|X>} matrices.
Typically not used directly,
see {@link DataReadObservationVectorXF#getDataRegularized_EV() getDataRegularized_EV()}
and {@link DataReadObservationVectorXF#getDataRegularized_LIN() getDataRegularized_LIN()}
methods that obtain the data of {@link RegularizedVectorXF} type.
    */
    public RadonNikodymSpectralModel(
	final int noriginal,final int nX,
	final double [] XfX,final double [] XX,final double [] ev,
	final java.util.function.Function<double [],double []> convertxFromOriginalToX,
	final java.util.function.ToDoubleFunction<double []> takeAverageOfFunctionInBasisX)
    {
	this.noriginal=noriginal;
        this.nX=nX;
        this.XfX=XfX;
        this.XX=XX;
	this.convertxFromOriginalToX=convertxFromOriginalToX;
	// Lebesgue quadrature https://arxiv.org/abs/1807.06007
        this.ev=ev; //=Skewness.MB.EV.getGEV_Generalized(nX,XfX,XX,true)
	if(EPS_selftest>0){
	    // Test ev if already calculated is used.
	    EVXData.test_orthogonality("RadonNikodymSpectralModel:ev",nX,ev,XfX,XX,EPS_selftest);
	}
	this.f=java.util.Arrays.copyOf(ev,nX);
        this.psiaverage=new double [nX];
        for(int k=0;k<nX;k++){
	    psiaverage[k]=takeAverageOfFunctionInBasisX.applyAsDouble(EVXData.getPsiFromBasis(k,ev,nX));
	}
        this.w=TMatr.squareArrayElements(psiaverage);
    }

    /** Used internally. For public exposure use getRNatXoriginal(xorig).getPsikAtX().
     */
    double [] getPsikAtXorig(double [] xorig){
	final double [] X=convertxFromOriginalToX.apply(xorig),psik_at_X=new double [nX];
	// The X is coordinates (not psi), converted directly as: ev X
	TMatr.multiply_MA_x_MB(psik_at_X,nX,ev,nX,nX,nX,X,1,1,0);
	return psik_at_X;
    }

    /** One point output.
     */
    public class RNPointEvaluation implements RNPointAccess {
	final double [] psik_at_X;
	RNPointEvaluation(double [] xorig){
	    this.psik_at_X=getPsikAtXorig(xorig);
	}
	 /** An evaluation of Radon-Nikodym derivative.
	  */
	public double getRN(){return true?
		TMatr.mulV2V(psik_at_X,f)/TMatr.mulVV(psik_at_X,psik_at_X)
		: // unoptimized
		TMatr.mulVV(getProjections(),f);
	}
	/** Inverse to Christoffel function 1/K(x), a polynomial on x.
	 */
	public double getChristoffelOatX(){
	    return TMatr.mulVV(psik_at_X,psik_at_X);
	}
	public double [] getPsikAtX(){return psik_at_X.clone();}
	public double [] getPsikSquaredAtX(){return TMatr.squareArrayElements(psik_at_X);}
	public double [] getProjections(){
	    final double [] projections=getPsikSquaredAtX();
	    return TMatr.scaleVectorInPlace(projections,1/TMatr.sumElements(projections));
	}
	public double getLS(){return TMatr.mulVVV(psik_at_X,f,psiaverage);}

	/** With different weights.
	 */
	public double getRNW(){
	    final double [] wk=getPosteriorDistributionWeights();
	    return TMatr.mulVV(wk,f)/TMatr.sumElements(wk);
	}
	public double [] getPriorDistributionWeights(){return w.clone();}
	public double [] getPosteriorDistributionWeights(){return TMatr.getVAB(getProjections(),getPriorDistributionWeights());}
    } // END OF class RNPointEvaluation

    public RNPointEvaluation getRNatXoriginal(final double [] xorig){return new RNPointEvaluation(xorig);}

    /** Expansion coefficients of the original attributes in psik.
     * @return The matrix[j][k]=&lt;xorig[j]*psi[k]&gt;.
     */
    public double [][] getAllXorigProjectedToPsik(final DataReadObservationVectorXF orig_sample){
	return orig_sample.getFun1Fun2Averaged(o->o.x,o->this.getPsikAtXorig(o.x)).getXXAsArrayOfArrays();
    }

    /** Calculate Christoffel function from original sample.
     * The result is in {@literal |psi>} basis.
     */
    XXMatrs getChristoffelFunctionDensityMatrixInEvBasis(final DataReadObservationVectorXF orig_sample){
	return orig_sample.getXXandXfX(this::getPsikAtXorig,(X)->TMatr.mulVV(X,X));
    }

    public double [] getChristoffelWeights(final DataReadObservationVectorXF orig_sample){
	return TMatr.getMatrixDiagElements(nX,getChristoffelFunctionDensityMatrixInEvBasis(orig_sample).XX);
    }

    /** Replace XfX by Christoffel function operator.
     * @param orig_sample Original sample the model is built from.
     * @return A model with Christoffel function operator as f.
     */
    public RadonNikodymSpectralModel getRadonNikodymSpectralModelWithChristoffelFunctionAsF(final DataReadObservationVectorXF orig_sample){
	final double [] XKX_in_RN=getChristoffelFunctionDensityMatrixInEvBasis(orig_sample).XX,
	    evK=Skewness.MB.EV.getEV(nX,XKX_in_RN,true);
	return new RadonNikodymSpectralModel(
	    noriginal,nX,
	    XKX_in_RN,TMatr.getUnitMatrix(nX),evK,
	    this::getPsikAtXorig,(psiX)->TMatr.mulVV(psiaverage,psiX));
    }

    /** Convert XfX to Christoffel function operator bases, take only diagonal elements, then convert back.
     * @param orig_sample Original sample the model is built from.
     * @return A model with adjusted f to be diagonal in Christoffel function operator basis.
     */
    public RadonNikodymSpectralModel getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis(final DataReadObservationVectorXF orig_sample){

	final double [] XKX_in_RN=getChristoffelFunctionDensityMatrixInEvBasis(orig_sample).XX, // in |psi>
	    evK=Skewness.MB.EV.getEV(nX,XKX_in_RN,true),
	    diagFelds=new double [nX];
	for(int k=0;k<nX;k++){
	    diagFelds[k]=TMatr.mulVVV(nX,evK,nX+k*nX,evK,nX+k*nX,f,0);
	}
	// assume diag elements are now the eigenvalues
	System.arraycopy(diagFelds,0,evK,0,nX);
	final double [] Unit=TMatr.getUnitMatrix(nX),
	    XfXdiagK=EVXData.convertMatrFromBtoQ_Basis(nX,TMatr.getDiagMatrix(diagFelds),evK,Unit);
	return new RadonNikodymSpectralModel(
	    noriginal,nX,
	    XfXdiagK,Unit,evK,
	    this::getPsikAtXorig,(psiX)->TMatr.mulVV(psiaverage,psiX));
    }

    /** Reduce the number of basis functions X performing
	<a href="https://arxiv.org/abs/1906.00460">optimal clustering</a>.
     * A D-point Gaussian quadrature is built first (with f[i] nodes and w[i] weights, i=0..nX-1),
     * then the obtained eigenfuctions are converted from f to X basis,
     * and new RadonNikodymSpectralModel is created.
     * @param D Clusters number.
     * @return Optimal model of D clusters.
     */
    public RadonNikodymSpectralModel reduceBasisSize(final int D){
	if(D>nX) throw new IllegalArgumentException();

	// convert from f[0] : f[nX-1] to [-1:1] for numerical stability.
	final double fmed=(f[nX-1]+f[0])/2,fdist2=(f[nX-1]-f[0])/2;
	final java.util.function.DoubleUnaryOperator
	    toQbasis=(forig)-> (forig-fmed)/fdist2,
	    fromQbasis=(F)-> fmed+F*fdist2;
	// A simple form of basis functions multiplication operator
	// formulae 22.7.24 Abramowitz and Stegun 1972, p. 872.
	// makes the Chebyshev basis the most stable.
	// The OrthogonalPolynomialsLegendreBasis also works OK.
	final OrthogonalPolynomialsChebyshevBasis Q=new OrthogonalPolynomialsChebyshevBasis();
	// Convert to [-1:1] interval for the best numerical stability in Q basis.
	final double [] fQ=new double [nX];
	for(int k=0;k<nX;k++) fQ[k]=toQbasis.applyAsDouble(f[k]);
	// Build D-point Gaussian quadrature out of Lebesgue integral weights.
	final Quadrature.GaussQuadratureWithEVData g=
	    Q.getGaussQuadratureFromMoments(D,Q.B.getMomentsFromSample(2*D,fQ,w));
	// A simple unit test.
	if(EPS_selftest>0){
	    UnitTests.checkDoubleRel(TMatr.sumElements(w),"sum(w)",TMatr.sumElements(g.w),"sum(g.w)",EPS_selftest);
	    for(int k=0;k<nX;k++) {
		UnitTests.checkDouble(fromQbasis.applyAsDouble(fQ[k]),"fromQbasis.applyAsDouble(fQ[k])",f[k],"f[k]",EPS_selftest);
	    }
	    // check the scale
	    UnitTests.checkDouble(fQ[0],"fQ[0]",-1,"-1",EPS_selftest);
	    UnitTests.checkDouble(fQ[nX-1],"fQ[nX-1]",1,"1",EPS_selftest);
	    // check normalizaing
	    UnitTests.checkDoubleRel(TMatr.sumElements(g.w),"sum(g.w)",TMatr.sumElements(w),"sum(w)",EPS_selftest);
	    System.err.println("RadonNikodymSpectralModel.reduceBasisSize:sum(w)="+TMatr.sumElements(w)+" sum(g.w)="+TMatr.sumElements(g.w)+" g.f="+java.util.Arrays.toString(g.x)+" D="+D+" nX="+nX+" f="+java.util.Arrays.toString(f)+" fQ="+java.util.Arrays.toString(fQ));
	}

	// Convert Lagrange interpolating polynomials psiG(f) (q.ev is different from it in just a constant): from Qf to X basis:
	final double [][] basisfun=new double [D][nX];
	final double [] psiG_average=new double [D];
	for(int i=0;i<D;i++){
	    final double [] psiG=EVXData.getPsiFromBasis(i,g.ev,D);
	    double psiG_a=0;
	    for(int k=0;k<nX;k++){
		final double psiGatfk=Q.B.getD0(psiG,fQ[k]);
		TMatr.sumP1(basisfun[i],ev,nX+k*nX,nX,psiaverage[k]*psiGatfk);
		psiG_a+=psiGatfk*w[k];
	    }
	    psiG_average[i]=psiG_a;
	}
	// The answer is known in Gaussian quadrature basis.
	final double [] evG_constructed=new double [D+D*D],XgXg_constructed=new double [D*D],XgfXg_constructed=new double [D*D];
	for(int i=0;i<D;i++){
	    evG_constructed[i]=XgfXg_constructed[D*i+i]=fromQbasis.applyAsDouble(g.x[i]);
	    evG_constructed[D+D*i+i]=1;
	    XgXg_constructed[D*i+i]=1;
	}

	// Unit test the "constructed" answer
	if(EPS_selftest>0){
	    // total weight to match
	    UnitTests.checkDoubleRel(TMatr.sumElements(w),"sum(w)",TMatr.sumElements(g.w),"sum(g.w)",EPS_selftest);
            // Copy the basis to the form acceptable by EVXData.convertMatrFromQtoB_Basis
	    final double [] basisfunflatmap=new double [D*nX];
	    for(int i=0;i<D;i++){
		System.arraycopy(basisfun[i],0,basisfunflatmap,i*nX,nX);
	    }
	    // Make sure the "constructed" basis is correct
	    // The XgXg and XgfXg to be diagonal
	    final double [] XgXg=EVXData.convertMatrFromQtoB_Basis(nX,XX,basisfunflatmap,D,0),
		XgfXg=EVXData.convertMatrFromQtoB_Basis(nX,XfX,basisfunflatmap,D,0);
	    EVXData.test_orthogonality("RadonNikodymSpectralModel.reduceBasisSize:evG_constructed",D,evG_constructed,XgfXg,XgXg,EPS_selftest);
	    UnitTests.checkDoubleArrayRel1(XgXg,"XgXg",XgXg_constructed,"XgXg_constructed",EPS_selftest);
	    UnitTests.checkDoubleArrayRel1(XgfXg,"XgfXg",XgfXg_constructed,"XgfXg_constructed",EPS_selftest);
	    for(int i=0;i<D;i++){
		System.err.println("gauss["+i+"]="+fromQbasis.applyAsDouble(g.x[i])+" XgfXg diag="+XgfXg[i*D+i]);
		UnitTests.checkDouble(fromQbasis.applyAsDouble(g.x[i]),"fromQbasis.applyAsDouble(g.x[i])",XgfXg[i*D+i],"XgfXg[i*D+i]",EPS_selftest);
	    }
	    // print them
	    System.err.println(" XgXg="+java.util.Arrays.toString(XgXg));
	    System.err.println("XgfXg="+java.util.Arrays.toString(XgfXg));
	}

	// New model with D basis function (reducted from nX).
	final RadonNikodymSpectralModel reducedDimension=new RadonNikodymSpectralModel(
	    noriginal,D,
	    XgfXg_constructed,XgXg_constructed,evG_constructed,
	    (xorig)->{
		final double [] X=convertxFromOriginalToX.apply(xorig),Xg=new double [D];
		for(int i=0;i<D;i++) Xg[i]=TMatr.mulVV(basisfun[i],X);
		return Xg;
	    },
	    // The new basis is already selected as psiG, the average is:
	    (psiXg)->TMatr.mulVV(psiG_average,psiXg)
					     );
	// Unit test the weights and nodes
	if(EPS_selftest>0){
	    UnitTests.checkDoubleRel(TMatr.sumElements(w),"sum(w)",TMatr.sumElements(reducedDimension.w),"sum(reducedDimension.w)",EPS_selftest);
	    UnitTests.checkDoubleArrayRel1(reducedDimension.w,"reducedDimension.w",g.w,"g.w",EPS_selftest);
	    for(int i=0;i<D;i++){
		UnitTests.checkDouble(reducedDimension.f[i],"reducedDimension.f[i]",fromQbasis.applyAsDouble(g.x[i]),"fromQbasis.applyAsDouble(g.x[i])",EPS_selftest);
	    }
	    System.err.println("RadonNikodymSpectralModel.reduceBasisSize: basis reduction is OK old nX="+nX+" new nX="+D);
	}
	return reducedDimension;
    }

    /** Create attributes products of a multi--index not greater than d.
     * The method products the eigenfunctions, not the original basis.
     * @param d The degree.
     * @param orig_sample Original data.
     * @return producted {@link DataReadObservationVectorXF}.
     */
    public DataReadObservationVectorXF getAttribsProductsMultiIndexed(
	final int d,
	final DataReadObservationVectorXF orig_sample){
	return AttributesProductsMultiIndexed.getAttribsProductsMultiIndexed(d,orig_sample,this::getPsikAtXorig);
    }

    /** No "final", a kludgy way to make a test from {@link TestRN}.
     * An alternative can be constructor's argument, a factory, or an injection.
     */
    static double EPS_selftest=0;//1e-6;
}
