package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */

// javadoc with math formulas
// javadoc --allow-script-in-comments -header '<script type="text/javascript" src="http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"></script>'  -Xdoclint:none -d /tmp/ com.polytechnik.utils

/**
 * A normalized basis
 * constructed from given three--term recurrence coefficients.
 * The basis uses three--term recurrence coefficients obtained
 * analytically or from sampled data by
 * e.g  {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getAB(double [][]) getAB()} method.
 * A typical usage: given three--term recurrence coefficients
 * obtain corresponding orthogonal polynomial
 * in other polynomials basis, see {@link SimpleBasisPolynomials#convertBasisToPBASIS convertBasisToPBASIS}
 * as an example
 * <pre>
 * {@code
 * var ab=new RecurrenceAB(a,b);
 * double [][] polynomials_in_PBASIS=ab.convertBasisToPBASIS(ab.numAk(),PBASIS);
 * // now polynomials_in_PBASIS[k] is k-th orthogobal polynomial with (a,b) recurrence.
 * }
 * </pre> A common case of PBASIS
 * is
 * {@link Monomials Monomials()},
 * {@link Chebyshev Chebyshev()}, etc.
 * Another example: assume we have two polynomial bases known
 * only through their three--term recurrence coefficients (a1,b1) and (a2,b2),
 * then we can transform
 * corresponding orthogonal polynomials through each other.
 * See such an example in the unit test {@link TestOrthogonalPolynomialsBasisFunctionsCalculatable#test_AB_conversion_PBASIS(int) TestOrthogonalPolynomialsBasisFunctionsCalculatable.test_AB_conversion_PBASIS()} method.
 *<pre>
 * {@code
 * // create random recurrence coefficients a1 b1, a2 b2 from random measures.
 * var r=new java.util.Random(1234);
 * var n=10;
 * var MC=new OrthogonalPolynomialsLegendreBasis();
 * var tmp1=MC.getAB(MC.getOrthogonalPolynomialsFirstKind(n,MC.B.getMomentsFromSample(2*n-1,IntStream.range(0,1000).mapToDouble(o-> -1+2*r.nextDouble()).toArray())));
 * var a1=tmp1.getAllAk();
 * var b1=tmp1.getAllBk();
 * var tmp2=MC.getAB(MC.getOrthogonalPolynomialsFirstKind(n,MC.B.getMomentsFromSample(2*n-1,IntStream.range(0,1000).mapToDouble(o-> -1+2*r.nextDouble()).toArray())));
 * var a2=tmp2.getAllAk();
 * var b2=tmp2.getAllBk();
 * }</pre>
 * Then run the code:
 *<pre>
 * {@code
 * RecurrenceAB ab1=new RecurrenceAB(a1,b1),ab2=new RecurrenceAB(a2,b2);
 * // the pols2_in_pols1Basis_direct[k] is the k-th orthogonal polynomial of ab2 in the basis of ab1.
 * double [][] pols2_in_pols1Basis_direct=ab2.convertBasisToPBASIS(ab2.numAk(),ab1);
 *
 * // same polynomial can be obtained via an intermediate, e.g. the basis
 * var M=new OrthogonalPolynomialsChebyshevBasis();
 * double [][] pols1=M.getOrthogonalPolynomialsFirstKindFromAB(ab1), // ab1 polynomials in Chebyshev basis
 *             pols2=M.getOrthogonalPolynomialsFirstKindFromAB(ab2); // ab2 polynomials in Chebyshev basis
 *
 * for(int k=0;k<pols2.length;k++){
 * // this is the k-th orthogonal polynomial of ab2 in the basis of ab1.
 * double [] pols2k_in_pols1_via_intermediate=TMatr.expandOverLowerDiagBasis(pols2[k],pols1);
 * // it matches exactly the pols2_in_pols1Basis_direct[k].
 *  System.err.println("direct="+java.util.Arrays.toString(pols2_in_pols1Basis_direct[k]));
 *  System.err.println("via intermediate="+java.util.Arrays.toString(pols2k_in_pols1_via_intermediate));
 * }
 * }
 *</pre>
 * @version $Id: RecurrenceAB.java,v 1.94 2022/08/09 15:13:34 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public class RecurrenceAB implements BasisPolynomials,ThreeTermRecurrenceCalculatable {
    private final double [] a,b;
    /** a_k */
    public double getAk(int k){return a[k];}
    /** b_k */
    public double getBk(int k){return b[k];}
    /** The number of a_k available. */
    public int numAk(){return a.length;}

    /** Convenience method. */
    public double [] getAllAk(){return a.clone();}
    /** Convenience method. */
    public double [] getAllBk(){return b.clone();}

    public RecurrenceAB(final double [] a,final double [] b){this.a=a;this.b=b;}


    /** Calculate \( Q_n(x) \) using three term recurrence at given x.
     * @param n Polynomial index (starts with 0).
     * @param x The point to calculate the polynomial.
     */
    public double calculate(final int n,final double x)
    {
	if(numAk()<=n) throw new IllegalArgumentException("index n="+n+" should be lower numAk()="+numAk());
	double f=1/getAk(0),fprev=0;
	for(int k=0;k<n;k++)
	{
	    f=(-getAk(k)*(fprev)+(x-getBk(k))*(fprev=f))/getAk(k+1);
	}
	return f;
    }


    /** Calculate psi_jQ_j(x) using three term recurrence and Clenshaw recurrence formula.
     * See L. Fox, I. B. Parker (1968), Chebyshev Polynomials in Numerical Analysis, page 56.
     */
    public double getD0(final double [] psi,final double x)
    {
	if(psi.length>numAk()) throw new IllegalArgumentException("psi.length="+psi.length+" should be lower equal numAk()="+numAk());
	final int n=psi.length-1;
	if(n<0) return 0;

	// Two options to calculate: direct and  Clenshaw. Which one is more stable?
	if(true){
	    double f=0,fprev=0,s=0;
	    for(int k=-1;k<n;)
	    {
		if(k<0) {f=1/getAk(0);fprev=0;}
		else {f=(-getAk(k)*(fprev)+(x-getBk(k))*(fprev=f))/getAk(k+1);}
		s+=f*psi[++k];
	    }
	    return s;
	} else {
	    // Clenshaw rule.
	    double sprev=0,s=0;
	    for(int i=n,n1=n-1; i>=0;--i)
	    {
		s=sprev*((i>=n1)?0:-getAk(i+1)/getAk(i+2))+(sprev=s)*(i>=n?0:(x-getBk(i))/getAk(i+1))+psi[i];
	    }
	    // Take into account that Q_{-1}=0 and this lead to
	    // second factor in formulae (44) (L. Fox, I. B. Parker, p.56) being equal to 0.
	    return s/getAk(0);
	}
    }

    /** First derivative at x.
     */
    public double getD1(final double [] psi,final double x)
    {
	if(psi.length>numAk()) throw new IllegalArgumentException("psi.length="+psi.length+" should be lower equal numAk()="+numAk());
	final int n=psi.length-1;
	if(n<0) return 0;

	double f=0,fprev=0,s=0,df=0,dfprev=0;
	for(int k=0;k<=n;++k)
	{
	    if(k <= 0) {f=1/getAk(0);fprev=df=dfprev=0;}
	    else {
		final double xxmb=x-getBk(k-1);
		df=(-getAk(k-1)*dfprev + xxmb*(dfprev=df) + f )/getAk(k);
		f= (-getAk(k-1)*fprev + xxmb*(fprev=f))/getAk(k);
	    }
	    s+=df*psi[k];
	}
	return s;
    }

    /** Second derivative at x.
     */
    public double getD2(final double [] psi,final double x)
    {
	if(psi.length>numAk()) throw new IllegalArgumentException("psi.length="+psi.length+" should be lower equal numAk()="+numAk());
	final int n=psi.length-1;
	if(n<0) return 0;

	double f=0,fprev=0,s=0,df=0,dfprev=0,d2f=0,d2fprev=0;
	for(int k=0;k<=n;++k)
	{
	    if(k <= 0) {f=1/getAk(0);fprev=df=dfprev=d2f=d2fprev=0;}
	    else {
		final double xxmb=x-getBk(k-1);
		d2f=(-getAk(k-1)*d2fprev + xxmb*(d2fprev=d2f) + 2*df )/getAk(k);
		df=(-getAk(k-1)*dfprev + xxmb*(dfprev=df) + f )/getAk(k);
		f= (-getAk(k-1)*fprev + xxmb*(fprev=f))/getAk(k);
	    }
	    s+=d2f*psi[k];
	}
	return s;
    }


    /**
     * Calculate the derivative of basis functions.
     * Used by {@link #getDifferentiated getDifferentiated} and
     * {@link #getIntegrated getIntegrated}.
     * @param numQl The number of basis functions to calculate the derivative.
     * @return An array of Q'_0, Q'_1,... Q'_{numQl-1}.
     * For an implementation with internal caching the method may return
     * more functions than the numQl.
     */
    protected double [][] getBasisFunctionsDifferentiated(final int numQl){
	double [][] res=new double [numQl][];
	if(numQl<=0) return res;
	double [] dfprev={},df=res[0]=dfprev;
	for(int k=0,ke=numQl-1;k<ke;k++)
	{
	    final double [] dfmul=getaxbP(df,1/getAk(k+1),-getBk(k)/getAk(k+1));
	    dfmul[k]+=1/getAk(k+1);
	    TMatr.sumP1(dfmul,dfprev,-getAk(k)/getAk(k+1));
	    dfprev=df;df=res[k+1]=dfmul;
	}
	return res;
    }

    @Override public double [] getDifferentiated(final double [] Tcoefs){
	final int n=Tcoefs.length-1;
	double [] res=new double [n],dPol[]=getBasisFunctionsDifferentiated(n+1);
	for(int k=0;k<=n;k++)
	{
	    TMatr.sumP1(res,dPol[k],Tcoefs[k]);
	}
	return res;
    }

    /** Calculate the integral as inverse to differentiation operation.
     */
    @Override public double [] getIntegrated(final double [] Tcoefs){
	double [] dPol[]=getBasisFunctionsDifferentiated(Tcoefs.length+1),
	    dPoltmp[]=new double [Tcoefs.length][];
	// remove Q0 from basis
	System.arraycopy(dPol,1,dPoltmp,0,dPoltmp.length);
	final double [] resE=TMatr.expandOverLowerDiagBasis(Tcoefs,dPoltmp),
	    res=new double [Tcoefs.length+1];
	// insert Q0 back to basis
	System.arraycopy(resE,0,res,1,Tcoefs.length);
	// normalize to Integral(0)=0;
	TMatr.sumP1(res,getONE(),-getD0(res,0));
	return res;
    }

    /** Do not mistake ma and mb multipliers with a and b recurrence
     * coefficients.
     * @return p*(ma*x+mb).
     */
    @Override public double [] getaxbP(final double [] p,final double ma,final double mb){
	if(p.length>=numAk()) throw new IllegalArgumentException("p.length="+p.length+" should be lower numAk()="+numAk()+" returned 1+p.length size array");
	final double [] res=new double [p.length+1];
	for(int k=p.length;--k>=0;){
	    final double mapk=p[k]*ma;
	    res[k+1]+=mapk*getAk(k+1);
	    res[k]+=mapk*getBk(k)+mb*p[k];
	    if(k>0) res[k-1]+=mapk*getAk(k);
	}
	return res;
    }

    /** Obtain next polynomial in PBASIS according to
     * three term recurrence coefficients.
     */
    @Override public double [] getNext(final int n,
			     final double [] Tnm1,final double [] Tnm2,
			     final BasisFunctionsCalculatable PBASIS){
	if(n<0) throw new IllegalArgumentException();
	if(n<=0) return TMatr.scaleVectorInPlace(PBASIS.getONE(),1/getAk(0));
	if(n>=numAk()) throw new IllegalArgumentException("too large n="+n+" should be lower numAk()="+numAk()+" returned n+1 size array");
	if(Tnm1.length!=n) throw new IllegalArgumentException();
	final double [] res=PBASIS.getaxbP(Tnm1,1/getAk(n),-getBk(n-1)/getAk(n));
	if(n<=1) return res;
	if(Tnm2.length!=n-1) throw new IllegalArgumentException();
	TMatr.sumP1(res,Tnm2,-getAk(n-1)/getAk(n));
	return res;
    }

/** A mockup non-implemented method from {@link BasisPolynomials}, it is put there
 * with other non-optimized default methods from {@link BasisPolynomials}.
 * For type 2 polynomials it is recommended to create a new
 * instance of {@link #RecurrenceAB RecurrenceAB(double [] a,double [] b)}
 * with {@code a} and {@code b} arrays having {@code 0}-th element removed.
 * If the arguments used to create an instance are not availble --
 * one can get the values from convenience methods {@link getAllAk} and {@link getAllBk}.
 */
    public double [] getNext(final int n,
			     final double [] Tnm1,final double [] Tnm2,
			     final boolean flag_type2,
			     final BasisFunctionsCalculatable PBASIS){
	if(flag_type2) throw new IllegalArgumentException("Not implemented");
	return getNext(n,Tnm1,Tnm2,PBASIS);
    }

    @Override public double [][] getBasisFunctionsOnPolynomialArgument(
	final int numQl,
	final double [] pol,
	final BasisFunctionsCalculatable PBASIS){

    	final double [][] res=new double [numQl][];
	double [] fprev={},f=res[0]=TMatr.scaleVectorInPlace(PBASIS.getONE(),1/getAk(0));
	for(int n=1;n<numQl;n++)
	{
	    final double an=getAk(n);
	    final double [] fmul=TMatr.scaleVectorInPlace(PBASIS.mult2Pol(pol,f),1/an);
	    TMatr.sumP1(fmul,f,-getBk(n-1)/an);
	    TMatr.sumP1(fmul,fprev,-getAk(n-1)/an);
	    fprev=f;f=res[n]=fmul;
	}
	return res;
    }

    /** Confederate matrix used in
     * {@link PolynomialRootsConfederateMatrix PolynomialRootsConfederateMatrix}.
     * Corresponds to
     *
     \[ p_{k}=(x/a_{k}-b_{k-1}/a_{k})*p_{k-1} - (a_{k-1}/a_{k})*p_{k-2} \]
     recurrence.
     * @param coefs Polynomial coefficients.
     * @return confederate matrix.
     */
    @Override public double [] getConfederateMatrix(final double [] coefs){
	final int n=coefs.length-1;
	final double [] matr=new double [n*n];
	for(int m=0;m<n;m++)
	{
	    final int k=m+1;
	    matr[m*n+m]=getBk(k-1);               // delta/alfa
	    if(m>0) matr[(m-1)*n+m]=getAk(k-1);   // gamma/alfa
	    if(m<n-1) matr[(m+1)*n+m]=getAk(k);   // 1/alfa
	    else
	    {
		final double f=coefs[n]/getAk(k); // p[n]*alfa
		for(int j=0;j<n;j++) matr[j*n+m]-=coefs[j]/f;
	    }
	}
	return matr;
    }

    /**
     * Calls {@link #getBasisFunctionsMultipliedByPolynomial getBasisFunctionsMultipliedByPolynomial},
     * the {@link BasisPolynomials#setQlQmExpansionCoefs BasisPolynomials#setQlQmExpansionCoefs}
     * typically uses analytic expression; this implementation
     * instead uses three term recurrence implemented in {@link #getBasisFunctionsMultipliedByPolynomial getBasisFunctionsMultipliedByPolynomial}.
     */
    @Override public void setQlQmExpansionCoefs(final int l,final int m,final double [] s){
	if(l>m) {
	    setQlQmExpansionCoefs(m,l,s);
	    return;
	}
	final double [] p0=new double [m+1];p0[m]=1;
	// may be optimized
	final double [] p=getBasisFunctionsMultipliedByPolynomial(l+1,p0)[l];
	System.arraycopy(p,0,s,0,p.length);
    }

    /** Re-implement base interface method not to use
     * {@link BasisFunctionsCalculatable#mult2Pol PBASIS.mult2Pol} becasuse for {@link RecurrenceAB}
     * this method is the foundation of multiplication operation; instead, we now repeatedly apply the
     * {@link RecurrenceAB recurrence} formula to array.
     * @param numQl The number of {@code Ql*pol} to calculate.
     * @param pol Polynomial to multiply, it is given in basis {@code PBASIS}.
     * @param PBASIS The basis in which {@code pol} is given and the result is returned.
     * @return An array of {@code numQl} dimension, elements: {@code Q0*pol, Q1*pol, ... Q_{numQl-1}*pol }
     * of this basis functions {@code Ql} multiplied by {@code pol}; the result is returned in basis {@code PBASIS}.
     */
    @Override public double [][] getBasisFunctionsMultipliedByPolynomial(
	final int numQl,
	final double [] pol,
	final BasisFunctionsCalculatable PBASIS){

	double [][] res=new double [numQl][];
	if(numQl<=0) return res;
	double [] fprev={},f=res[0]=TMatr.getVectorScaled(pol,1/getAk(0));
	for(int k=0,ke=numQl-1;k<ke;k++)
	{
	    final double ak1=getAk(k+1);
	    final double [] fmul=PBASIS.getaxbP(f,1/ak1,-getBk(k)/ak1);
	    TMatr.sumP1(fmul,fprev,-getAk(k)/ak1);
	    fprev=f;f=res[k+1]=fmul;
	}
	return res;
    }

    /** Used as multiplication foundation when an analytic formula for
     * {@link #setQlQmExpansionCoefs}
     * is not available, calls {@link #getBasisFunctionsMultipliedByPolynomial(int, double [],BasisFunctionsCalculatable) getBasisFunctionsMultipliedByPolynomial(numQ,pol,this)}
     */
    @Override public double [][] getBasisFunctionsMultipliedByPolynomial(
	final int numQl,
	final double [] pol){
	return getBasisFunctionsMultipliedByPolynomial(numQl,pol,this);
    }

    @Override public double [] getPQkMomentsFromQkMoments(final double [] P,final double [] moments){
	final double [] r=new double [moments.length-(P.length-1)],
	    mulP[]=getBasisFunctionsMultipliedByPolynomial(r.length,P);
	for(int k=0;k<r.length;k++)
	{
	    r[k]=TMatr.mulVVn1(mulP[k],moments);
	}
	return r;
    }

    @Override public double [] mult2Pol(final double [] p1,final double [] p2)
    {
	if(p1.length>p2.length) return mult2Pol(p2,p1);
	final double [] res=new double [p1.length+p2.length-1],
	    mulP2[]=getBasisFunctionsMultipliedByPolynomial(p1.length,p2);
	for(int k=0;k<p1.length;k++){
	    TMatr.sumP1(res,mulP2[k],p1[k]);
	}
	return res;
    }

    @Override public SDivRes sdiv1(final double [] p,final double x0)
    {
	final int n=p.length-1;
	final double [] res=new double [n];
	double r=p[n];
	if(n>0) res[n-1]=p[n-1];
	for(int k=n;--k>=0;)
	{
	    final double c=r/getAk(k+1),rnext=(res[k]+=c*(x0-getBk(k)));
	    if(k>0)res[k-1]=p[k-1]-c*getAk(k);
	    res[k]=c;
	    r=rnext;
	}
	return new SDivRes(res,new double []{r});
    }

    @Override public double [] getMomentsFromSample(final int nmoments,final double [] x,final double [] w,final int nelsinsample)
{
    final double [] m=new double [nmoments];
    for(int j=0;j<nelsinsample;j++)
    {
	final double xx=x[j],ww=w!=null?w[j]:1;
	double f=0,fprev=0;
	for(int k=-1,ke=nmoments-1;k<ke;)
	{
	    if(++k <= 0) {f=1/getAk(0);fprev=0;}
	    else {f=(-getAk(k-1)*(fprev)+(xx-getBk(k-1))*(fprev=f))/getAk(k);}
	    m[k]+=f*ww;
	}
    }
    return m;
}



/** Unit test a general 3 term recurrence basis.
 */
public static void main(String [] args)
{
    final double eps=1e-6;
    // create random sample moments
    final OrthogonalPolynomialsChebyshevBasis.TestChebyshevInMonomials TM=new OrthogonalPolynomialsChebyshevBasis.TestChebyshevInMonomials();

    final int neq_max=50, // a large number of basis functions.
	n_sample=100000;
    // from sample
    final RecurrenceAB ab=TM.getRecurrenceABFromBasisRandomMoments(neq_max,n_sample);

    // Use the ab as a basis.
    // Create a unit test from recurrence ab and run it:
    new TestBasisPolynomials("RecurrenceAB",ab,new java.util.Random(-79099),eps){
	@Override double [] getMonomialsMoments(final int n){return null;}
	@Override OrthogonalPolynomialsRecurrenceABBasis getOrtPol(){return new OrthogonalPolynomialsRecurrenceABBasis(ab){
		@Override public double [] getRadauMultiplied(final double [] nodes0)
		{
		    return B.getaxbP(nodes0,-1,TM.x_max);
		}
	    };}
	@Override double [][] getExact(){return null;}
    }.runTests("$Id: RecurrenceAB.java,v 1.94 2022/08/09 15:13:34 mal Exp $");
}

}


