package com.polytechnik.utils;

/** A basis (possibly non--normalized) with three term recurrence.
 * This interface is typically a part of {@link BasisPolynomials BasisPolynomials} functionality,
 * however when we have a basis with only recurrence coefficients
 * available this interface can be used directly to convert to/from it using
 * {@link #convertFunctionTo_PBASIS_from_Basis(double [] coefs,BasisFunctionsCalculatable  PBASIS) convertFunctionTo_PBASIS_from_Basis}
 * and
 * {@link #convertFunctionTo_Basis_from_PBASIS(double [] coefs,BasisFunctionsCalculatable  PBASIS) convertFunctionTo_Basis_from_PBASIS}.
 *
 * @version $Id: SimpleBasisPolynomials.java,v 1.87 2021/12/06 07:44:52 mal Exp $
 * @author Vladislav Malyshkin mal@gromco.com
 */
public interface SimpleBasisPolynomials extends BasisFunctionsCalculatable {



    /** Calculates n-th basis polynomial of this basis in PBASIS basis.
     * @param n Q index.
     * @param Qnm1 Input \( Q_{n-1} \).
     * @param Qnm2 Input \( Q_{n-2} \).
     * @param PBASIS The basis in which to calculte.
     * @return \( Q_{n} \) polynomial of this basis in PBASIS basis, array dimension n+1.
     */
    double [] getNext(final int n,
		      final double [] Qnm1,final double [] Qnm2,
		      final BasisFunctionsCalculatable PBASIS);

    /** This method calculates \( Q_l(pol(x))\), where
     * pol(x) is a given polynomial and \( Q_l \) is basis finction of this (not {@code PBASIS}!) basis.
     * This method may not be stable for a large dimensions of {@code pol.length}.
     * Equivalent to
     * numQl applications of {@link #getNext} to {@code pol} array.
     * The method can be used to calculate e.g. \( T_n(P_k(x)) \)
     * or \( T_n(P_k(L_m(x^j))) \). For example  <pre>{@code
new Legendre().getBasisFunctionsOnPolynomialArgument(4,new double []{0,0,0,0,1},new Chebyshev())
={
{ 1.0 },
{ 0.0, 0.0, 0.0, 0.0, 1.0 },
{ 0.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.75 },
{ 0.0, 0.0, 0.0, 0.0, 0.375, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.625 }
}
}</pre> means
<br>\( P_0(T_4(x))=T_0 \)
<br>\( P_1(T_4(x))=T_4(x) \)
<br>\( P_2(T_4(x))=0.25*T_0+0.75*T_8(x) \)
<br>\( P_3(T_4(x))=0.375*T_4(x)+0.625*T_{12}(x) \).
     * @param numQl The number of basis functions to calculate.
     * @param pol The polynomial to use in \( Q_l(pol(x))\), it is given in the {@code PBASIS} basis.
     * @param PBASIS the basis in which {@code pol} is given and the result is returned.
     * @return An array of numQl dimension, elements: \( Q_0(pol(x)) \),
     * \( Q_1(pol(x)) \), ... \( Q_{numQl-1}(pol(x)) \) in the basis {@code PBASIS}.
     */
    double [][] getBasisFunctionsOnPolynomialArgument(final int numQl,
						      final double [] pol,
						      final BasisFunctionsCalculatable PBASIS);

    /** The same as {@link #getBasisFunctionsOnPolynomialArgument(int,double [],BasisFunctionsCalculatable) getBasisFunctionsOnPolynomialArgument} with both {@code pol} and result in this basis.
    */
    default double [][] getBasisFunctionsOnPolynomialArgument(final int numQl,
						      double [] pol){
	return getBasisFunctionsOnPolynomialArgument(numQl,pol,this);
    }


    /** A helper method to calculate a product of this basis functions
     * by a polynomial in given basis.
     * The method can be used to calculate e.g. \( P_k(x) T_j(x) \)
     * or  \( P_k(x) T_j(x) L_i(x) \).
     * For example  <pre>{@code
new Legendre().getBasisFunctionsMultipliedByPolynomial(4,new double []{0,0,0,0,1},new Chebyshev())
={
{ 0.0, 0.0, 0.0, 0.0, 1.0 },
{ 0.0, 0.0, 0.0, 0.5, 0.0, 0.5 },
{ 0.0, 0.0, 0.375, 0.0, 0.25, 0.0, 0.375 },
{ 0.0, 0.3125, 0.0, 0.1875, 0.0, 0.1875, 0.0, 0.3125 }
}
}</pre> means
<br>\( P_0(x)*T_4(x)=T_4(x) \)
<br>\( P_1(x)*T_4(x)=0.5*T_3(x)+0.5*T_5(x) \)
<br>\( P_2(x)*T_4(x)=0.375*T_2(x)+0.25*T_4(x)+0.375*T_6(x) \)
<br>\( P_3(x)*T_4(x)=0.3125*T_1(x)+0.1875*T_{3}(x)+0.1875*T_5(x)+0.3125*T_7(x) \).
     */
    default double [][] getBasisFunctionsMultipliedByPolynomial(
	final int numQl,
	final double [] pol,
	final BasisFunctionsCalculatable PBASIS){

	final double [][] res=new double [numQl][];
	double [] f={},fprev={};
	for(int n=0;n<numQl;n++)
	{
	    final double [] fnext=getNext(n,f,fprev,PBASIS);
	    res[n]=PBASIS.mult2Pol(pol,fnext);
	    fprev=f;f=fnext;
	}
	return res;
    }


    /** A helper method to convert this basis to PBASIS.
     * A typical usage is e.g. to convert {@link Chebyshev} to
     * {@link Legendre} or to a general
     * {@link RecurrenceAB RecurrenceAB} basis. To convert to monomials
     * use {@link Monomials PBASIS=new Monomials()}.
     * The method applies {@link #getNext(int,double [],double [],BasisFunctionsCalculatable) getNext()} in a serie.
     * For example  <pre>{@code
new Legendre().convertBasisToPBASIS(4,new Chebyshev())
={
{ 1.0 },
{ 0.0, 1.0 },
{ 0.25, 0.0, 0.75 },
{ 0.0, 0.375, 0.0, 0.625 }
}
}</pre> means
<br>\( P_0(x)=T_0(x) \)
<br>\( P_1(x)=T_1(x) \)
<br>\( P_2(x)=0.25*T_0(x)+0.75*T_2(x) \)
<br>\( P_3(x)=0.375*T_1(x)+0.625*T_3(x) \).
     * @param num The number of elements in serie.
     * @param PBASIS The basis in which we obtain this basis polynomials.
     * @return {@link #getNext} applied to {@code n=0,1,2,...num-1} basis functions; output array
     * elements are this basis polynomials in the basis PBASIS.
    */
    @Override default double [][] convertBasisToPBASIS(final int num,
					     final BasisFunctionsCalculatable PBASIS){
	double [] ru=null,rv=null,res []=new double [num][];
	for(int k=0;k<num;k++){
	    final double [] l=res[k]=getNext(k,ru,rv,PBASIS);rv=ru;ru=l;
	}
	return res;
    }

    /**
     * Convert from basis to PBASIS basis.
     * @param Qcoefs Coefficients.
     * @param PBASIS The basis to convert to.
     * @return Polynomial in PBASIS.
     */
    @Override default double [] convertFunctionTo_PBASIS_from_Basis(final double [] Qcoefs,
							  final BasisFunctionsCalculatable PBASIS)
    {
	double [] ru=null,rv=null,sum=new double [Qcoefs.length];
	for(int k=0;k<Qcoefs.length;k++)
	{
	    double [] l=getNext(k,ru,rv,PBASIS);rv=ru;ru=l;
	    TMatr.sumP1(sum,l,k+1,Qcoefs[k]);
	}
	return sum;
    }


}
