package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Simple (and inefficient) matrix multiplication and linear systems.
 *  Use for testing only.
 *  @version $Id: TMatr.java,v 1.142 2024/04/02 01:45:23 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class TMatr {

public static double sumElements(final double [] pin)
{
    return sumElements(pin.length,pin);
}

private static double sumElements(final int n,final double [] pin)
{
    double s=0;
    for(int k=n;--k>=0;) s+=pin[k];
    return s;
}

public static double [] squareArrayElements(final double [] m)
{
    return getVAB(m.length,m,m);
}

public static double [] expandOverLowerDiagBasis(final double [] p,final double [][] basis)
{
    if(p.length>basis.length) throw new IllegalArgumentException();
    final double [] res=p.clone();
    for(int k=p.length; --k>=0;)
    {
	final double [] b=basis[k];
	if(b.length!=k+1) throw new IllegalArgumentException();
	final double e=res[k]/b[k];
	sumP1(res,b,k,-e);
	res[k]=e;
    }
    return res;
}


public static void sumP1(final double [] res,final double [] pin,final double w)
{
    sumP1(res,pin,pin.length,w);
}
static void sumP1(final double [] res,final double [] pin,final int nels,final double w)
{
    sumP1(res,pin,0,nels,w);
}
static void sumP1(final double [] res,final double [] pin,final int pin_offset,final int nels,final double w)
{
    for(int k=0;k<nels;k++)
    {
	res[k]+=w*pin[k+pin_offset];
    }
}
static double [] sumAB(final int n,final double [] A,final double [] B)
{
    final double [] r=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=A[k]+B[k];
    }
    return r;
}
public static double [] sumAcBdE(final double [] A,final double c,final double [] B,final double d,final double [] E)
{
    if(A.length!=B.length || B.length!=E.length) throw new IllegalArgumentException();
    final int n=A.length;
    final double [] r=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=A[k]+c*B[k]+d*E[k];
    }
    return r;
}
public static double [] sumAcB(final double [] A,final double c,final double [] B)
{
    if(A.length!=B.length) throw new IllegalArgumentException();
    return sumAcB(A.length,A,c,B);
}
private static double [] sumAcB(final int n,final double [] A,final double c,final double [] B)
{
    final double [] r=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=A[k]+c*B[k];
    }
    return r;
}
public static double [] sumaAbB(final double a,final double [] A,final double b,final double [] B)
{
    if(A.length!=B.length) throw new IllegalArgumentException();
    return sumaAbB(A.length,a,A,b,B);
}
private static double [] sumaAbB(final int n,final double a,final double [] A,final double b,final double [] B)
{
    final double [] r=new double [n];
    for(int k=0;k<n;k++)
    {
	r[k]=a*A[k]+b*B[k];
    }
    return r;
}

/** Scale a vector and return it. In-place calculation.
 * @return the argument {@code v} scaled.
 */
private static double [] scaleVectorInPlace(final int n,final double [] v,final double w)
{
    for(int k=n;--k>=0;){v[k]*=w;}
    return v;
}
/** Scale a vector and return it. In-place calculation.
 * @return the argument {@code v} scaled.
 */
public static double [] scaleVectorInPlace(final double [] v,final double w)
{
    return scaleVectorInPlace(v.length,v,w);
}
/** Scale a copy of a vector. The argument is intact.
 * @return scaled input vector.
 */
public static double [] getVectorScaled(final double [] pin,final double w)
{
    return scaleVectorInPlace(pin.clone(),w);
}


/**  Scalar product of two arrays of equal dimension. */
public static double mulVV(final double [] a,final double [] b)
{
    if(a.length!=b.length) throw new IllegalArgumentException();
    return mulVVn1(a,b);
}
/**  Scalar product of two arrays of equal dimension,
 * the number of elements is taken from first array dimension:
 * sum a[k]*b[k] ; k=0 to the length of a.
 */
public static double mulVVn1(final double [] a,final double [] b)
{
    return mulVV(a.length,a,0,b,0);
}


/** Calculates spur of matrix.
 */
public static double getSpurM(final int n,final double [] matr)
{
    if(n*n!=matr.length) throw new IllegalArgumentException();
    final int nm=n;
    double s=0;
    for(int l=0;l<n;l++)
    {
	s+=matr[l*nm+l];
    }
    return s;
}

/** Calculates the spur of A*B*C.
 */
public static double getSpurMMM(final int n,final double [] A,final double [] B,final double [] C){
    return getSpurMM(n,mulMM(n,A,B),C);
}

/** Calculates the spur of A*B.
 */
public static double getSpurMM(final int n,final double [] A,final double [] B)
{
    if(n*n!=A.length || A.length!=B.length) throw new IllegalArgumentException();
    final int na=n,nb=n;
    double s=0;
    for(int k=0;k<n;k++)
    {
	double ss=0;
	for(int l=0;l<n;l++){ss+=A[k*na+l]*B[l*nb+k];}
	s+=ss;
    }
    return s;
}
public static double getAMatrB(final double [] A,final double [] B,final double [] matr,final int nm)
{
    return getAMatrB(A.length,A,0,B.length,B,0,matr,nm);
}
public static double getAMatrB(final int n,final double [] A,final double [] B,final double [] matr,final int nm)
{
    return getAMatrB(n,A,0,B,0,matr,nm);
}
public static double getAMatrB(final int n,final double [] A,final int A_offset,final double [] B,final int B_offset,final double [] matr,final int nm)
{
    return getAMatrB(n,A,A_offset,n,B,B_offset,matr,nm,0);
}
public static double getAMatrB(final int na,final double [] A,final int A_offset,final int nb,final double [] B,final int B_offset,final double [] matr,final int nm)
{
    return getAMatrB(na,A,A_offset,nb,B,B_offset,matr,nm,0) ;
}

/** Calculates A_i matr_ij B_j
 */
private static double getAMatrB(final int na,final double [] A,final int A_offset,final int nb,final double [] B,final int B_offset,final double [] matr,final int nm,final int off_matr)
{
    if(matr.length!=nm*nm+off_matr) throw new IllegalArgumentException();
    double s=0;
    for(int l=0;l<na;l++)
    {
	s+=A[l+A_offset]*mulVV(nb,B,B_offset,matr,off_matr+l*nm);
    }
    return s;
}
/** Calculates A_i matr_ij A_j
 */
public static double getAMatrA(final double [] A,final double [] matr,final int nm)
{
    return getAMatrA(A.length,A,matr,nm);
}
public static double getAMatrA(final int n,final double [] A,final double [] matr,final int nm)
{
    return getAMatrB(n,A,A,matr,nm);
}
public static double getAMatrA(final int n,final double [] A,final int A_offset,final double [] matr,final int nm)
{
    return getAMatrB(n,A,A_offset,n,A,A_offset,matr,nm);
}
/** Calculates matr_ij A_j
 */
public static double [] getMatr_x_A(final int n,final double [] matr,final double [] A)
{
    if(n*n!=matr.length||n!=A.length)throw new IllegalArgumentException();
    final int nm=n;
    final double [] r=new double [n];
    if(false) multiply_MA_x_MB(r,n,matr,n,0,n,A,1,1,0);
    else for(int l=0;l<n;l++) {r[l]=mulVV(n,A,0,matr,l*nm);}
    return r;
}

public static double mulVV(final int n,final double [] a,final int off_a,final double [] b,final int off_b)
{
    double s=0;
    for(int i=n;--i>=0;)
    {
	s+=a[i+off_a]*b[i+off_b];
    }
    return s;
}
/** @return The sum a[k]*b[k]; k=0 to n-1.
 */
public static double mulVVn(final int n,final double [] a,final double [] b)
{
    return mulVV(n,a,0,b,0);
}

/** @return The sum a[k]*b[k]*c[k].
 */
public static double mulVVV(final double [] a,final double [] b,final double [] c)
{
    if(a.length!=b.length || b.length!=c.length) throw new IllegalArgumentException();
    return mulVVV(a.length,a,0,b,0,c,0);
}
static double mulVVV(final int n,final double [] a,final int off_a,final double [] b,final int off_b,final double [] c,final int off_c)
{
    double s=0;
    for(int i=n;--i>=0;)
    {
	s+=a[i+off_a]*b[i+off_b]*c[i+off_c];
    }
    return s;
}
/** @return The sum a[k]*a[k]*b[k].
 */
public static double mulV2V(final double [] a,final double [] b)
{
    return mulVVV(a,a,b);
}
/** @return The sum a[k]*a[k]*b[k]; k=0 to the length of a.
 */
static double mulV2Vn1(final double [] a,final double [] b)
{
    return mulVVV(a.length,a,0,a,0,b,0);
}

/** Multiply 2 matrices.
 */
public static double [] mulMM(final int n,final double [] matrA,final double [] matrB)
{
    if(matrA.length!=n*n||matrB.length!=n*n) throw new IllegalArgumentException();
    final double [] r=new double [n*n];
    multiply_MA_x_MB(r,n,matrA,n,matrB,n);
    return r;
}

/** Multiply 3 matrices.
 */
public static double [] mulMMM(final int n,final double [] A,final double [] B,final double [] C)
{
    return mulMM(n,A,mulMM(n,B,C));
}

/** Get matrix copy, according to leading dimension,
 *  then transponse it.
 */
public static double [] getMatrTrans(final int n,final double [] matr){
    return getMatrTrans(n,n,matr);
}

/** Get matrix copy, according to leading dimension and offset,
 *  then transponse it.
 */
public static double [] getMatrTrans(final int nr,final int nc,final double [] matr)
{
    if(matr.length!=nr*nc) throw new IllegalArgumentException();
    final double [] m=new double [nc*nr];
    for(int k=0;k<nr;k++)
    {
	for(int l=0;l<nc;l++)
	{
	    m[l*nr+k]=matr[k*nc+l];
	}
    }
    return m;
}

/** Get matrix copy, according to leading dimension.
 */
private static double [] getMatrNoTrans(final int n,final double [] matr,final int nmA){
    return getMatrNoTrans(n,matr,nmA,0);
}

/** Get matrix copy, according to leading dimension and offset.
 */
private static double [] getMatrNoTrans(final int n,final double [] matr,final int nmA,final int off)
{
    final double [] m=new double [n*n];
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<n;l++)
	{
	    m[k*n+l]=matr[k*nmA+l+off];
	}
    }
    return m;
}

/** Multiply several matrices res=matrs[0] x matrs[1] x matrs[2]  ...
 *  Used for unit tests.
 */
public static double [] mulMatrices(final int n,final double [][] matrs)
{
    if(matrs.length<=0) return null;
    if(matrs[0].length!=n*n) throw new IllegalArgumentException();
    double [] tmp1=getMatrNoTrans(n,matrs[0],n,0),tmp2=new double [n*n];
    for(int k=1;k<matrs.length;k++)
    {
	if(matrs[k].length!=n*n) throw new IllegalArgumentException();
	multiply_MA_x_MB(tmp2,n,tmp1,n,matrs[k],n);
	final double [] xx=tmp2;
	tmp2=tmp1;
	tmp1=xx;
    }
    return tmp1;
}


public static double [] mul_MA_x_MB(final int sA,final double [] A,final int s,final double [] B,final int sB){
    final double [] r=new double [sA*sB];
    multiply_MA_x_MB(r,sA,A,s,B,sB);
    return r;
}
/** r[i*sB+j]=\(\sum_{k=0}^{k=s-1} A[i*s+k]*B[k*sB+j] ; i=[0\dots sA-1], j=[0\dots sB-1]\);
    A: sA x s matrix, B: s x sB matrix, r: sA x sB matrix.
 */
public static void multiply_MA_x_MB(final double [] r,final int sA,final double [] A,final int s,final double [] B,final int sB){
    if(A.length!=sA*s||B.length!=s*sB) throw new IllegalArgumentException();
    multiply_MA_x_MB(r,sA,A,s,0,s,B,sB,sB,0);
}

/** r[i*sB+j]=\sum_{k=0}^{k=s-1} A[i*ldA+k]*B[k*ldB+j] ; i=[0..sA-1], j=[0..sB-1].
    A: sA x s matrix, B: s x sB matrix, r: sA x sB matrix, ld(r)=sB.
 */
static void multiply_MA_x_MB(final double [] r,final int sA,final double [] A,final int ldA,final int off_A,final int s,final double [] B,final int sB,final int ldB,final int off_B)
{
    if(s>ldA||sB>ldB) throw new IllegalArgumentException();
    for(int i=0;i<sA;i++)
    {
	for(int j=0;j<sB;j++)
	{
	    double sum=0;
	    for(int k=0;k<s;k++)
	    {
		sum+=A[off_A+i*ldA+k]*B[off_B+k*ldB+j];
	    }
	    r[i*sB+j]=sum;
	}
    }
}

public static double [] mul_MA_x_MBTrans(final int sA,final double [] A,final int s,final double [] B,final int sB){
    final double [] r=new double [sA*sB];
    multiply_MA_x_MBTrans(r,sA,A,s,B,sB);
    return r;
}
/** r[i*sB+j]=\(\sum_{k=0}^{k=s-1} A[i*s+k]*B[j*s+k] ; i=[0\dots sA-1], j=[0\dots sB-1]\);
    A: sA x s matrix, B: sB x s matrix, r: sA x sB matrix.
 */
public static void multiply_MA_x_MBTrans(final double [] r,final int sA,final double [] A,final int s,final double [] B,final int sB){
    if(A.length!=sA*s||B.length!=sB*s) throw new IllegalArgumentException();
    multiply_MA_x_MBTrans(r,sA,A,s,0,s,B,sB,s,0);
}

/** r[i*sB+j]=\sum_{k=0}^{k=s-1} A[i*ldA+k]*B[j*ldB+k] ; i=[0..sA-1], j=[0..sB-1].
    A: sA x s matrix, B: sB x s matrix, r: sA x sB matrix, ld(r)=sB.
 */
static void multiply_MA_x_MBTrans(final double [] r,final int sA,final double [] A,final int ldA,final int off_A,final int s,final double [] B,final int sB,final int ldB,final int off_B)
{
    if(s>ldA||s>ldB) throw new IllegalArgumentException();
    for(int i=0;i<sA;i++)
    {
	for(int j=0;j<sB;j++)
	{
	    double sum=0;
	    for(int k=0;k<s;k++)
	    {
		sum+=A[off_A+i*ldA+k]*B[off_B+j*ldB+k];
	    }
	    r[i*sB+j]=sum;
	}
    }
}

public static double [] mul_MATrans_x_MB(final int sA,final double [] A,final int s,final double [] B,final int sB){
    final double [] r=new double [sA*sB];
    multiply_MATrans_x_MB(r,sA,A,s,B,sB);
    return r;
}
/** r[i*sB+j]=\( \sum_{k=0}^{k=s-1} A[k*sA+i]*B[k*sB+j] ; i=[0\dots sA-1], j=[0\dots sB-1] \);
    A: s x sA matrix, B: s x sB matrix, r: sA x sB matrix.
 */
public static void multiply_MATrans_x_MB(final double [] r,final int sA,final double [] A,final int s,final double [] B,final int sB){
    if(A.length!=s*sA||B.length!=s*sB) throw new IllegalArgumentException();
    multiply_MATrans_x_MB(r,sA,A,sA,0,s,B,sB,sB,0);
}
/** r[i*sB+j]=\sum_{k=0}^{k=s-1} A[k*ldA+i]*B[k*ldB+j] ; i=[0..sA-1], j=[0..sB-1].
    A: s x sA matrix, B: s x sB matrix, r: sA x sB matrix, ld(r)=sB.
 */
static void multiply_MATrans_x_MB(final double [] r,final int sA,final double [] A,final int ldA,final int off_A,final int s,final double [] B,final int sB,final int ldB,final int off_B)
{
    if(sA>ldA||sB>ldB) throw new IllegalArgumentException();
    for(int i=0;i<sA;i++)
    {
	for(int j=0;j<sB;j++)
	{
	    double sum=0;
	    for(int k=0;k<s;k++)
	    {
		sum+=A[off_A+k*ldA+i]*B[off_B+k*ldB+j];
	    }
	    r[i*sB+j]=sum;
	}
    }
}

public static double [] mul_MATrans_x_MBTrans(final int sA,final double [] A,final int s,final double [] B,final int sB){
    final double [] r=new double [sA*sB];
    multiply_MATrans_x_MBTrans(r,sA,A,s,B,sB);
    return r;
}
/** r[i*sB+j]=\( \sum_{k=0}^{k=s-1} A[k*sA+i]*B[j*s+k] ; i=[0\dots sA-1], j=[0\dots sB-1] \);
    A: s x sA matrix, B: sB x s matrix, r: sA x sB matrix.
 */
public static void multiply_MATrans_x_MBTrans(final double [] r,final int sA,final double [] A,final int s,final double [] B,final int sB){
    if(A.length!=s*sA||B.length!=sB*s) throw new IllegalArgumentException();
    multiply_MATrans_x_MBTrans(r,sA,A,sA,0,s,B,sB,s,0);
}
/** r[i*sB+j]=\sum_{k=0}^{k=s-1} A[k*ldA+i]*B[j*ldB+k] ; i=[0..sA-1], j=[0..sB-1].
    A: s x sA matrix, B: sB x s matrix, r: sA x sB matrix, ld(r)=sB.
 */
static void multiply_MATrans_x_MBTrans(final double [] r,final int sA,final double [] A,final int ldA,final int off_A,final int s,final double [] B,final int sB,final int ldB,final int off_B)
{
    if(sA>ldA||s>ldB) throw new IllegalArgumentException();
    for(int i=0;i<sA;i++)
    {
	for(int j=0;j<sB;j++)
	{
	    double sum=0;
	    for(int k=0;k<s;k++)
	    {
		sum+=A[off_A+k*ldA+i]*B[off_B+j*ldB+k];
	    }
	    r[i*sB+j]=sum;
	}
    }
}

/** Calculates 0.5*(QQ+QQ^T).
 * In-place calculation.
 * @return the argument {@code QQ}, updated to hermitian part.
*/
public static double [] selectHermitianPartInPlace(final int n,final double [] QQ){
    return scaleVectorInPlace(_selectHermitianPart(n,QQ,false),0.5);
}

/** Calculates 0.5*(QQ-QQ^T).
 *  In-place calculation.
 * @return the argument {@code QQ}, updated to anti-hemitian part.
 */
public static double [] selectAntiHermitianPartInPlace(final int n,final double [] QQ){
    return scaleVectorInPlace(_selectHermitianPart(n,QQ,true),0.5);
}

/** Calculates QQ+QQ^T or QQ-QQ^T
 * depending on flag_antiHermitian. In-place calculation. No 1/2 factor used!
 * @return the argument {@code QQ}, updated to hermitian or anti-hemitian part.
*/
private static double [] _selectHermitianPart(final int n,final double [] QQ,final boolean flag_antiHermitian)
{
    if(n*n!=QQ.length) throw new IllegalArgumentException();
    final int offQQ=0,ldQQ=n;
    for(int k=0;k<n;k++)
    {
	for(int l=0;l<=k;l++)
	{
	    if(flag_antiHermitian)
	    {
		final double d=QQ[offQQ+k*ldQQ+l]-QQ[offQQ+k+l*ldQQ];
		QQ[offQQ+k+l*ldQQ]=-d;
		QQ[offQQ+k*ldQQ+l]=d;
	    }
	    else
	    {
		final double d=QQ[offQQ+k*ldQQ+l]+QQ[offQQ+k+l*ldQQ];
		QQ[offQQ+k+l*ldQQ]=QQ[offQQ+k*ldQQ+l]=d;
	    }
	}
    }
    return QQ;
}

/** Calculates a[i]*a[j] matrix.
 */
public static double [] getQQaa(final double [] a)
{
    return getQQab(a.length,a,a);
}


/** Calculates 0.5*(a[i]*b[j]+b[i]*a[j]) matrix.
 */
public static double [] getQQab(final int n,final double [] a,final double [] b)
{
    final double [] QQ=new double [n*n];
    for(int i=0;i<n;i++){
	final double ai=a[i],bi=b[i];
	for(int j=0,in=i*n;j<n;j++) QQ[in+j]=ai*b[j]+bi*a[j];
    }
    return scaleVectorInPlace(QQ,0.5);
}

/** Calculates a[i]*b[j] matrix.
 */
public static double [] dyadicMatrix(final double [] a,final double [] b)
{
    final double [] m=new double [a.length*b.length];
    for(int of=0,j=0;j<a.length;j++){
	for(int k=0;k<b.length;k++,of++){
	    m[of]=a[j]*b[k];
	}
    }
    return m;
}

/** Calculates c[i]=a[i]*b[i].
 */
static double [] getVAB(final int n,final double [] a,final double [] b)
{
    final double [] v=new double [n];
    for(int i=0;i<n;i++){
	v[i]=a[i]*b[i];
    }
    return v;
}
/** Calculates c[i]=a[i]*b[i].
 */
public static double [] getVAB(final double [] a,final double [] b)
{
    if(a.length!=b.length) throw new IllegalArgumentException();
    return getVAB(a.length,a,b);
}
static double [] getConstArray(final int n,final double c){
    final double [] a=new double [n];
    java.util.Arrays.fill(a,c);
    return a;
}
public static double [] getUnitMatrix(final int n)
{
    final double [] O=new double [n*n];
    for(int k=0;k<n;k++) O[k+n*k]=1;
    return O;
}
public static double [] getDiagMatrix(final double [] v)
{
    return getDiagMatrix(v.length,v);
}
static double [] getDiagMatrix(final int n,final double [] v)
{
    final double [] O=new double [n*n];
    for(int k=0;k<n;k++) O[k+n*k]=v[k];
    return O;
}

public static double [] getMatrixDiagElements(final int n,final double [] Matrix)
{
    if(Matrix.length!=n*n) throw new IllegalArgumentException();
    final double [] dM=new double [n];
    for(int k=0;k<n;k++) dM[k]=Matrix[k+n*k];
    return dM;
}
public static double [] getMatrixDiagElementsInBasis(final int n,final double [] basis,final double [] Matrix)
{
    return getMatrixDiagElementsInBasis(n,basis,n,Matrix);
}
private static double [] getMatrixDiagElementsInBasis(final int n,final double [] basis,final int basis_offset,final double [] Matrix)
{
    if(basis.length!=basis_offset+n*n||Matrix.length!=n*n) throw new IllegalArgumentException();
    final double [] m=new double [n];
    for(int k=0;k<n;k++) m[k]=getAMatrA(n,basis,basis_offset+k*n,Matrix,n);
    return m;
}

static double [] getRandomVector(final int nels,final java.util.Random r)
{
    return getRandomVector(nels,r,-2,1);
}
static double [] getRandomVector(final int nels,final java.util.Random r,final double mul,final double add)
{
    final double [] q=new double [nels];
    for(int k=0;k<q.length;k++) q[k]=add+mul*r.nextDouble();
    return q;
}

static void test_mulMatrices(final java.util.Random r,final double eps)
{
    final int n_tests=5000;
    for(int t=0;t<n_tests;t++)
    {
    for(int n=1;n<10;n++)
    {
	for(int d=1;d<6;d++)
	{
	    final double [][] matrices=new double [d][];
	    for(int kk=0;kk< matrices.length;kk++)
	    {
		 matrices[kk]=getRandomVector(n*n,r);
	    }

	    final double [] mmul=mulMatrices(n,matrices);
	    double [] mt=matrices[0];
	    for(int k=1;k<matrices.length;k++)
	    {
		mt=mulMM(n,mt,matrices[k]);
	    }
	    UnitTests.checkDoubleArrayRel(mmul,"mmul",mt,"mt",eps);
	}
    }
    }
    System.err.println("TMatr:test_mulMatrices: OK with n_tests="+n_tests+" eps="+eps);
}
static void test_Mul(final java.util.Random r,final double eps)
{
    final int n_tests=5000;
    for(int t=0;t<n_tests;t++)
    {
	for(int n=1;n<10;n++)
	{
	    for(int ld1=n;ld1<=n+3;ld1++)
	    {
		final int o1=r.nextInt(3),o2=r.nextInt(30),ld2=n+r.nextInt(ld1-n+1);
		final double [] m1=getRandomVector(o1+n*ld1,r),m2=getRandomVector(o2+n*ld2,r);
		final double [] res=new double [n*n];
		multiply_MA_x_MB(res,n,m1,ld1,o1,n,m2,n,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"multiply_MA_x_MB",mulMatrices(n,new double [][]{getMatrNoTrans(n,m1,ld1,o1),getMatrNoTrans(n,m2,ld2,o2)}),"mul",eps);

		multiply_MA_x_MBTrans(res,n,m1,ld1,o1,n,m2,n,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"multiply_MA_x_MBTrans",mulMatrices(n,new double [][]{getMatrNoTrans(n,m1,ld1,o1),getMatrTrans(n,getMatrNoTrans(n,m2,ld2,o2))}),"mul",eps);

		multiply_MATrans_x_MB(res,n,m1,ld1,o1,n,m2,n,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"multiply_MATrans_x_MB",mulMatrices(n,new double [][]{getMatrTrans(n,getMatrNoTrans(n,m1,ld1,o1)),getMatrNoTrans(n,m2,ld2,o2)}),"mul",eps);

		multiply_MATrans_x_MBTrans(res,n,m1,ld1,o1,n,m2,n,ld2,o2);
		UnitTests.checkDoubleArrayRel(res,"multiply_MatrTrans_x_MatrTrans",mulMatrices(n,new double [][]{getMatrTrans(n,getMatrNoTrans(n,m1,ld1,o1)),getMatrTrans(n,getMatrNoTrans(n,m2,ld2,o2))}),"mul",eps);
	    }
	}

	for(int n1=1;n1<10;n1++)
	{
	    for(int n2=1;n2<10;n2++)
	    {
		for(int id=1;id<20;id++)
		{
		    final int d=r.nextInt(n1+n2);
		    final double [] m1=getRandomVector(n1*d,r),m2=getRandomVector(n2*d,r);
		    final double []
			resMM=new double [n1*n2],
			_resMM=new double [n1*n2],
			resMMT=new double [n1*n2],
			_resMMT=new double [n1*n2],
			resMTM=new double [n1*n2],
			_resMTM=new double [n1*n2],
			resMTMT=new double [n1*n2],
			_resMTMT=new double [n1*n2];
		    for(int j=0;j<n1;j++){
			for(int k=0;k<n2;k++){
			    for(int s=0;s<d;s++){
				resMM[j*n2+k]+=m1[j*d+s]*m2[s*n2+k];
				resMMT[j*n2+k]+=m1[j*d+s]*m2[k*d+s];
				resMTMT[j*n2+k]+=m1[s*n1+j]*m2[k*d+s];
				resMTM[j*n2+k]+=m1[s*n1+j]*m2[s*n2+k];
			    }
			}
		    }
		    multiply_MA_x_MB(_resMM,n1,m1,d,m2,n2);
		    UnitTests.checkDoubleArrayRel(_resMM,"_resMM",resMM,"resMM",eps);
		    multiply_MA_x_MBTrans(_resMMT,n1,m1,d,m2,n2);
		    UnitTests.checkDoubleArrayRel(_resMMT,"_resMMT",resMMT,"resMMT",eps);
		    multiply_MATrans_x_MBTrans(_resMTMT,n1,m1,d,m2,n2);
		    UnitTests.checkDoubleArrayRel(_resMTMT,"_resMTMT",resMTMT,"resMTMT",eps);
		    multiply_MATrans_x_MB(_resMTM,n1,m1,d,m2,n2);
		    UnitTests.checkDoubleArrayRel(_resMTM,"_resMTM",resMTM,"resMTM",eps);
		}
	    }
	}
    }
    System.err.println("TMatr:test_Mul: OK with n_tests="+n_tests+" eps="+eps);
}



public static void main(String [] args)
{
    final java.util.Random r=new java.util.Random(662);
    final double eps=1e-7;

    System.err.println("TMatr::main(): starting ALL tests; ver=$Id: TMatr.java,v 1.142 2024/04/02 01:45:23 mal Exp $");
    final long t0=System.currentTimeMillis();
    test_mulMatrices(r,eps);
    test_Mul(r,eps);
    final long t1=System.currentTimeMillis();
    System.err.println("TMatr::main(): ALL tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: TMatr.java,v 1.142 2024/04/02 01:45:23 mal Exp $");
}


}
