package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Basis of orthogonal polynomials.
 *  @version $Id: TestBasisPolynomials.java,v 1.44 2024/04/15 06:15:34 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public abstract class TestBasisPolynomials extends TestBasisFunctionsCalculatable {

abstract double [] getMonomialsMoments(final int n);
abstract OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> getOrtPol();
abstract double [][] getExact();

/** Calculates n-th type 2 basis polynomial (the result is in monomials basis).
 * May not always be available.
 * This mononials basis calculation is used for unit test only.
 */
double [] getNextType2(final int n,final double [] Tnm1,final double [] Tnm2)
{
    return B.getNext(n,Tnm1,Tnm2,true/* Explicit type 2 */,Monomials.BASIS);
}

/** Mask B with a different type, a kludge to avoid generic. */
final BasisPolynomials B;


TestBasisPolynomials(final String txt,final BasisPolynomials B,final java.util.Random r,final double eps)
{
    super(txt,B,r,eps);
    this.B=B;
}

void test_ExactValues(final double eps)
{
    final double [][] P=getExact();
    if(P==null) return;
    double [] ru=null,rv=null;
    for(int k=0;k<P.length;k++)
    {
	final double [] Pk=B.getNext(k,ru,rv,Monomials.BASIS);rv=ru;ru=Pk;
	UnitTests.checkDoubleArrayRel(Pk,"Pk",P[k],"P[k]",eps);
    }
    System.err.println(txt+":test_ExactValues: test OK with nmax="+P.length+" eps="+eps);
}

void printPolynomialsType2(final int nmax)
{
    System.err.println("========="+txt+": Type 2 polynomials");
    final double [] pmom=getMonomialsMoments(nmax+1);
    if(pmom==null)
    {
	System.err.println("========="+txt+": Type 2 polynomials pmom=null, skipping");
	return;
    }

    double [] ru=null,rv=null;
    for(int n=0;n<=nmax;n++)
    {
	final double [] L=new double [n+2];L[n+1]=1;
	final double [] t2=Polynomials.getSecondKindOrthogonalPolynomialFromFirstKind(B.convertFunctionTo_Monomials_from_Basis(L),pmom);
	System.err.print("n="+n);
	for(int j=0;j<t2.length;j++)
	{
	    System.err.print("\t"+t2[j]);
	}
	System.err.println();

	final double [] t2x=getNextType2(n,ru,rv);rv=ru;ru=t2x;
	UnitTests.checkDoubleArrayRel(t2,"t2",t2x,"t2x",eps);
    }

    // test calculations in various bases
    for(BasisPolynomials PBASIS : java.util.stream.Stream.concat(
	    java.util.stream.Stream.of(B),
	    java.util.Arrays.stream(PBASIS_CONVERSION_TEST))
	    .toList())
    {
	ru=rv=null;
	// limit to 6, sometiles fails for Laguerre
	for(int n=0;n<=Math.min(nmax,6);n++)
	{
	    final double [] L=new double [n+2];L[n+1]=1;
	    final double [] t2=Polynomials.getSecondKindOrthogonalPolynomialFromFirstKind(B.convertFunctionTo_Monomials_from_Basis(L),pmom);
	    final double [] x=B.getNext(n,ru,rv,true/* explicit Type 2 flag */,PBASIS);rv=ru;ru=x;
	    final double [] t2x=PBASIS.convertFunctionTo_Monomials_from_Basis(x);
	    UnitTests.checkDoubleArrayRel(t2,"t2",t2x,"t2x",eps);
	}
    }
}


void test_getBasisFunctionsOnPolynomialArgument(final int n_tests,final int nmax)
{
 int n_tests_performed=0;
 for(BasisPolynomials PBASIS : java.util.stream.Stream.concat(
	 java.util.stream.Stream.of(B),
	 java.util.Arrays.stream(PBASIS_CONVERSION_TEST_WITHOUT_Laguerre_AND_HermiteE))
	 .toList())
 {

  for(int t=0;t<n_tests;t++)
  {
    for(int nn=0;nn<=nmax;nn++)
    {
	// a random polynomial in basis PBASIS
	final double [] pol=TMatr.getRandomVector(1+r.nextInt(nmax+1),r);
	final double [][] Ql_polX=B.getBasisFunctionsOnPolynomialArgument(nn+1,pol,PBASIS);
	final double FE=1e-3;
	for(double x:TMatr.getRandomVector(50,r,0.5,0))
	{
	    final double polX=PBASIS.getD0(pol,x);
	    if(Math.abs(polX)< -1.5 || Math.abs(polX)>3){
		// if the argument is too large -- skip the test as the result may be
		// unstable e.g. for Laguerre large negative arguments
		continue;
	    }
	    n_tests_performed++;
	    for(int n=0;n<Ql_polX.length;n++)
	    {
		final double Qn_polX=B.calculate(n,polX), // Qn(pol(x))
		    Qn_polX_expanded=PBASIS.getD0(Ql_polX[n],x) // linearized Qn(pol(x))
		    ;
		// match the values calculated in basis B and in basis PBASIS.
		if(!(Math.abs(Qn_polX)<=FE))
		{ // check non-small values only
		    if(Math.abs(Qn_polX)<1)
			UnitTests.checkDouble(Qn_polX,"pol.len="+pol.length+" (x="+x+" polX="+polX+") n="+n+" Qn_polX",Qn_polX_expanded,"length="+Ql_polX[n].length+" Qn_polX_expanded in PBASIS="+PBASIS,eps);
		    else
			UnitTests.checkDoubleRel(Qn_polX,"pol.len="+pol.length+" (x="+x+" polX="+polX+") n="+n+" Qn_polX",Qn_polX_expanded,"length="+Ql_polX[n].length+" Qn_polX_expanded in PBASIS="+PBASIS,eps);
		}
	    }
	}
    }
  }
 }
 System.err.println(txt+":test_getBasisFunctionsOnPolynomialArgument(): test OK with n_tests_performed="+n_tests_performed+" nmax="+nmax+" eps="+eps);
}

void test_Sum(final int n_tests,final int nmax)
{
  for(int t=0;t<n_tests;t++)
  {
    for(int nn=0;nn<=nmax;nn++)
    {
	// a random polynomial in basis B
	double [] coefs=TMatr.getRandomVector(nn+1,r);
	double [] ru=null,rv=null,res=new double [nn+1];
	for(int n=0;n<=nn;n++)
	{
	    double [] l=B.getNext(n,ru,rv,Monomials.BASIS);rv=ru;ru=l;
	    TMatr.sumP1(res,l,coefs[n]);
	}
	final double [] sum_test=B.convertFunctionTo_Monomials_from_Basis(coefs);
	UnitTests.checkDoubleArrayRel(sum_test,"sum_test",res,"res",eps);
	UnitTests.checkDoubleArrayRel(coefs,"coefs",B.convertFunctionTo_Basis_from_Monomials(sum_test),"sum_test",eps);

	// differentiation
	UnitTests.checkDoubleArrayRel(B.convertFunctionTo_Monomials_from_Basis(B.getDifferentiated(coefs)),"DBasis",
				      Polynomials.getPolynomialDifferentiated(B.convertFunctionTo_Monomials_from_Basis(coefs)),"DP",eps);
	UnitTests.checkDoubleArrayRel(B.getDifferentiated(coefs),"DBasis",
				      B.convertFunctionTo_Basis_from_Monomials(Polynomials.getPolynomialDifferentiated(B.convertFunctionTo_Monomials_from_Basis(coefs))),"DP",eps);


	// integration
	UnitTests.checkDoubleArrayRel(B.convertFunctionTo_Monomials_from_Basis(B.getIntegrated(coefs)),"IBasis",
				      Polynomials.getPolynomialIntegrated(B.convertFunctionTo_Monomials_from_Basis(coefs)),"IP",eps);
	UnitTests.checkDoubleArrayRel1(B.getIntegrated(coefs),"IBasis",
				      B.convertFunctionTo_Basis_from_Monomials(Polynomials.getPolynomialIntegrated(B.convertFunctionTo_Monomials_from_Basis(coefs))),"IBasis",eps);


	final double FE=1e-3;
	for(double x:TMatr.getRandomVector(50,r,1,0))
	{
	    if(!(Math.abs(Polynomials.getD0(ru,x))<=FE))
		UnitTests.checkDoubleRel(Polynomials.getD0(ru,x),"Ln",B.calculate(nn,x),"calculate",eps);
	    if(!(Math.abs(Polynomials.getD0(res,x))<=FE))
		UnitTests.checkDoubleRel(Polynomials.getD0(res,x),"D0sum",B.getD0(coefs,x),"getD0",eps);
	    if(!(Math.abs(Polynomials.getD1(res,x))<=FE))
		UnitTests.checkDoubleRel(Polynomials.getD1(res,x),"D1sum L",B.getD1(coefs,x),"getD1",eps);
	    if(!(Math.abs(Polynomials.getD2(res,x))<=FE))
		UnitTests.checkDoubleRel(Polynomials.getD2(res,x),"D2sum L",B.getD2(coefs,x),"getD2",eps);
	}
    }
  }
  System.err.println(txt+":test_Sum(): test OK with n_tests="+n_tests+" nmax="+nmax+" eps="+eps);
}
void test_Sum_PBASIS(final int n_tests,final int nmax)
{
  for(int t=0;t<n_tests;t++)
  {
    for(int nn=0;nn<=nmax;nn++)
    {
	// Create a random polynomial in basis B.
	// select random vector for unit test stability.
	final double [] coefs=B.convertFunctionTo_Basis_from_PBASIS(TMatr.getRandomVector(nn+1,r,-2,1.1), new Monomials() );
	for(BasisPolynomials PBASIS : java.util.stream.Stream.concat(
		java.util.stream.Stream.of(B),
		java.util.Arrays.stream(PBASIS_CONVERSION_TEST))
		.toList())
	{
	    double [] ru=null,rv=null,res=new double [nn+1];
	    for(int n=0;n<=nn;n++)
	    {
		double [] l=B.getNext(n,ru,rv,PBASIS);rv=ru;ru=l;
		TMatr.sumP1(res,l,coefs[n]);
	    }
	    final double [] sum_test=B.convertFunctionTo_PBASIS_from_Basis(coefs,PBASIS);
	    UnitTests.checkDoubleArrayRel1(sum_test,"sum_test",res,"res",eps);
	    UnitTests.checkDoubleArrayRel1(coefs,"coefs",B.convertFunctionTo_Basis_from_PBASIS(sum_test,PBASIS),"sum_test",eps);

	    // coefs: basis B
	    // res: basis PBASIS
	    // differentiation
	    UnitTests.checkDoubleArrayRel1(B.convertFunctionTo_PBASIS_from_Basis(B.getDifferentiated(coefs),PBASIS),"DBasis",
					  PBASIS.getDifferentiated(B.convertFunctionTo_PBASIS_from_Basis(coefs,PBASIS)),"DP",eps);
	    UnitTests.checkDoubleArrayRel1(B.getDifferentiated(coefs),"DBasis",
					  B.convertFunctionTo_Basis_from_PBASIS(PBASIS.getDifferentiated(B.convertFunctionTo_PBASIS_from_Basis(coefs,PBASIS)),PBASIS),"DP",eps);


	    // integration
	    UnitTests.checkDoubleArrayRel1(B.convertFunctionTo_PBASIS_from_Basis(B.getIntegrated(coefs),PBASIS),"IBasis",
					  PBASIS.getIntegrated(B.convertFunctionTo_PBASIS_from_Basis(coefs,PBASIS)),"IP",eps);
	    UnitTests.checkDoubleArrayRel1(B.getIntegrated(coefs),"IBasis",
					   B.convertFunctionTo_Basis_from_PBASIS(PBASIS.getIntegrated(B.convertFunctionTo_PBASIS_from_Basis(coefs,PBASIS)),PBASIS),"IBasis",eps);


	    final double FE=1e-3;
	    for(double x:TMatr.getRandomVector(50,r,1,0))
	    {
		if(!(Math.abs(PBASIS.getD0(ru,x))<=FE))
		    UnitTests.checkDoubleRel(PBASIS.getD0(ru,x),"Ln",B.calculate(nn,x),"calculate",eps);
		if(!(Math.abs(PBASIS.getD0(res,x))<=FE))
		    UnitTests.checkDoubleRel(PBASIS.getD0(res,x),"D0sum",B.getD0(coefs,x),"getD0",eps);
		if(!(Math.abs(PBASIS.getD1(res,x))<=FE))
		    UnitTests.checkDoubleRel(PBASIS.getD1(res,x),"D1sum L",B.getD1(coefs,x),"getD1",eps);
		if(!(Math.abs(PBASIS.getD2(res,x))<=FE))
		    UnitTests.checkDoubleRel(PBASIS.getD2(res,x),"D2sum L",B.getD2(coefs,x),"getD2",eps);
	    }
	}
    }
  }
  System.err.println(txt+":test_Sum_PBASIS(): test OK with n_tests="+n_tests+" nmax="+nmax+" eps="+eps);
}

void test_setNewtonBinomialLikeCoefs(final int n_tests,final int nmax)
{
    for(int t=0;t<n_tests;t++)
    {
	final int n=r.nextInt(nmax+1);

	double [] ru=null,rv=null;
   	for(int k=0;k<=n;k++)
	{
	    final double [] l=B.getNext(k,ru,rv,Monomials.BASIS);rv=ru;ru=l;

	    final double [] p=java.util.Arrays.copyOf(l,k+1);
	    final double x0=r.nextDouble(),a=r.nextDouble();
	    final double [] psbasis=B.convertFunctionTo_Basis_from_Monomials(Polynomials.getPolynomialShifted(Polynomials.getPolynomialArgumentScaled(p,a),x0));

	    final double [] b=TMatr.getRandomVector(k+1,r),q=TMatr.getRandomVector(k,r);
	    for(int m=0;m<b.length;m++) B.setNewtonBinomialLikeCoefs(m,b,q,a,x0);
	    UnitTests.checkDoubleArrayRel1(b,"b",psbasis,"psbasis",eps);
	}
    }
    System.err.println(txt+":test_setNewtonBinomialLikeCoefs: test OK with nmax="+nmax+" eps="+eps);
}



void test_getBasisFunctionsMultipliedByPolynomial(final int n_tests,final int nmax)
{
 int n_tests_performed=0;
 for(BasisPolynomials PBASIS : java.util.stream.Stream.concat(
	 java.util.stream.Stream.of(B),
	 java.util.Arrays.stream(PBASIS_CONVERSION_TEST))
	 .toList())
 {
  for(int t=0;t<n_tests;t++)
  {
    for(int nn=0;nn<=nmax;nn++)
    {
	// a random polynomial in basis PBASIS
	final double [] pol=TMatr.getRandomVector(1+r.nextInt(nmax+1),r);
	final double [][] Ql_mult_polX=B.getBasisFunctionsMultipliedByPolynomial(nn+1,pol,PBASIS);
	final double FE=1e-3;
	for(double x:TMatr.getRandomVector(50,r,0.5,0))
	{
	    n_tests_performed++;
	    for(int n=0;n<Ql_mult_polX.length;n++)
	    {
		final double
		    Qn_multiply_polX=B.calculate(n,x)*PBASIS.getD0(pol,x),
		    Qn_multiply_polX_expanded=PBASIS.getD0(Ql_mult_polX[n],x) // linearized
		    ;
		// match the values calculated in basis B and in basis PBASIS.
		if(!(Math.abs(Qn_multiply_polX)<=FE))
		{ // check non-small values only
		    if(Math.abs(Qn_multiply_polX)<1)
			UnitTests.checkDouble(Qn_multiply_polX,"Qn_multiply_polX",Qn_multiply_polX_expanded,"Qn_multiply_polX_expanded",eps);
		    else
			UnitTests.checkDoubleRel(Qn_multiply_polX,"Qn_multiply_polX",Qn_multiply_polX_expanded,"Qn_multiply_polX_expanded",eps);
		}
	    }
	}
    }
  }
 }
 System.err.println(txt+":test_getBasisFunctionsMultipliedByPolynomial(): test OK with n_tests_performed="+n_tests_performed+" nmax="+nmax+" eps="+eps);
}




void print_Extra(final int nmax){}
void test_Extra(final int n_tests,final int nmax){}

@Override void runSingleTestInSerie(final int n_tests,final int nmax){
    test_Sum(n_tests,nmax); // must match almost identically
    test_Sum_PBASIS(11+n_tests/10,Math.min(nmax,6));
    super.runSingleTestInSerie(n_tests,nmax);
    test_getBasisFunctionsOnPolynomialArgument(21+n_tests/50,Math.min(nmax,4));
    test_getBasisFunctionsMultipliedByPolynomial(11+n_tests/30,Math.min(nmax,5));
    test_setNewtonBinomialLikeCoefs(n_tests,nmax);
    test_Extra(n_tests,nmax);
}

@Override void runSingleTest(final int nmax){
    super.runSingleTest(nmax);
    printPolynomialsType2(nmax);
    print_Extra(nmax);
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> o=getOrtPol();
    if(o!=null)
    {
	for(int k=1,ne=25;k<=ne;k++)
	{
	    final double [] moments=new double [2*k];
	    moments[0]=1;
	    OrthogonalPolynomialsBasisFunctionsCalculatable.printQuadratures(txt,k,moments,o);
	    if(k==ne){
		OrthogonalPolynomialsBasisFunctionsCalculatable.printAB("Three term recurrence coefficients",o.getAB(o.getOrthogonalPolynomialsFirstKind(k,moments)));
	    }
	}
    }
    test_ExactValues(1e-14);
}

void runTests(final String id)
{
    runTests(id,10000,7,15);
}


}
