package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

/** Test OrthogonalPolynomialsBasisFunctionsCalculatable in monomials.
 *  @version $Id: TestOrthogonalPolynomialsBasisFunctionsCalculatable.java,v 1.266 2024/04/15 06:15:34 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

class TestOrthogonalPolynomialsBasisFunctionsCalculatable {
    final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> O;
    final String name;
    final double eps;
    final java.util.Random r;
    final double x_min,x_max;
    TestOrthogonalPolynomialsBasisFunctionsCalculatable(final String name,
		  final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> O,
		  final double eps,
		  final java.util.Random r,
		  final double x_min,
		  final double x_max)
    {
	this.name=name;
	this.O=O;
	this.eps=eps;
	this.r=r;
	this.x_min=x_min;
	this.x_max=x_max;
    }

    double [] convertToMonomials(final double [] p)
    {
	return O.B.convertFunctionTo_Monomials_from_Basis(p);
    }

    /** Multiply (in monomials basis) nodes by (x-x0) or (x0-x).
     *  Similar method {@link OrthogonalPolynomialsBasisFunctionsCalculatable#getRadauMultiplied} does this
     *  in some basis, and this method does this in monomials basis.
     */
    double [] getRadauMultiplied(final double [] nodes){
	final double [] tmp1=O.B.getONE();
	final double x0=O.getRadauRoot();
	final double [] tmp=O.getRadauMultiplied(tmp1);
	if(O.B.getD1(tmp,x0)>0)
	{
	    return Polynomials.mult2Pol(nodes,new double [] {-x0,1});
	}
	else
	{
	    return Polynomials.mult2Pol(nodes,new double [] {x0,-1});
	}
    }


static double [] getPolynomialRandomMoments(final int n,final double x_min,final double x_max,final int size,final java.util.Random r)
{
    final double [] mom=new double [n];
    for(int k=size;--k>=0;)
    {
	final double x=x_min+(x_max-x_min)*r.nextDouble();
	//System.err.println("x="+x);
	double s=1;
	for(int m=0;m<mom.length;m++) {mom[m]+=s;s*=x;}
    }
    return mom;
}
double [] getBasisRandomMoments(final int n,final double x_min,final double x_max,final int size,final java.util.Random r)
{
    final double [] mom=new double [n];
    final double [][] f=new double [mom.length][];
    for(int m=0;m<mom.length;m++)
    {
	f[m]=new double [m+1];
	f[m][m]=1;
    }
    for(int k=size;--k>=0;)
    {
	final double x=x_min+(x_max-x_min)*r.nextDouble();
	//System.err.println("x="+x);
	for(int m=0;m<mom.length;m++) {mom[m]+=O.B.getD0(f[m],x);}
    }
    return mom;
}
/** This method: creates random sample, calculates it's moments (in basis),
finds orthogonal polynomials, and finally calculates the
{@link RecurrenceAB RecurrenceAB} from them.
* @param n The number or polynomials to construct.
* @param tsize Sample size.
* @return RecurrenceAB for the measure corresponding to the sample.
*/
RecurrenceAB getRecurrenceABFromBasisRandomMoments(final int n,final int tsize){
   final double [] moms_b=getBasisRandomMoments(2*n-1,x_min,x_max,tsize,r);
    // create orhogonal polynomials
   final double [][] pols=O.getOrthogonalPolynomialsFirstKind(n,moms_b);
   // obtain ab from sampled moments moms_b:
   return O.getAB(pols);
}

double [] mom_p,mom_b;
void setData(final int tsize,final int n_moments)
{
    mom_p=getPolynomialRandomMoments(n_moments,this.x_min,this.x_max,tsize,this.r);
    mom_b=getBasisMomentsConvertedFromMonomials(mom_p);
}

Quadrature getXWDiscrete(final int n)
{
    final double [] x=new double [n],w=new double [n];
    final double delta=(x_max-x_min)/(n+1);
    for(int k=0;k<n;k++)
    {
	x[k]=x_min+k*delta+0.9*delta*r.nextDouble();
	w[k]=3.1*(r.nextDouble()+r.nextDouble());
    }
    sortNodes(x,w);
    //System.err.println("getXWDiscrete: x="+java.util.Arrays.toString(x));
    return new Quadrature(x,w);
}
/**
 *  x nodes on left argument, xnodes on right argument.
 *  Uses generalized eigenvalues expansion.
 *  The method is used to test discrete measures.
 */
double [][] getD2QuadratureFromMoments_EV(final int n,final double [] Lmoments,final double [] Rmoments)
{
    final double [] QQL=O.B.getQQMatr(n,Lmoments),QQR=O.B.getQQMatr(n,Rmoments);
    final double [] ev=O.EV.getGEV_Generalized(n,QQL,QQR,true);
    final double [] xQQL=O.B.getQQMatr(n,O.B.getXQkMomentsFromQkMoments(Lmoments)),xQQR=O.B.getQQMatr(n,O.B.getXQkMomentsFromQkMoments(Rmoments));
    final double [] values=new double [n],xl=new double [n],xr=new double [n];
    for(int k=0;k<n;k++)
    {
	final double lam=values[k]=ev[k];
	final int o=n+k*n;
	xl[k]=TMatr.getAMatrB(n,ev,o,ev,o,xQQL,n)/lam;
	xr[k]=TMatr.getAMatrB(n,ev,o,ev,o,xQQR,n);
    }
    return new double [][]{values,xl,xr};
}


double [] getBasisMomentsConvertedFromMonomials(final double [] monomial_moments)
{
    final int n=monomial_moments.length;
    final double [] basis_moments=new double [n];
    for(int k=0;k<n;k++)
    {
	final double [] mo=new double [k+1];mo[k]=1;
	basis_moments[k]=TMatr.mulVVn1(convertToMonomials(mo),monomial_moments);
    }
    return basis_moments;
}

static void sortNodes(final double [] x,final double [] w)
{
    SortArrays.sortArrayWithLoad(x.length,x,w);
}


double [][] test_NodesPolynomial(final double [] nodes,final double [] mom,final int maxdegree)
{
    final int n=nodes.length-1;
    final PolynomialRootsFinder.PRoots u=new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG().findRoots(nodes);
    if(u.nroots!=n) throw new RuntimeException("n="+n+" u.nroots="+u.nroots);
    final double [] xs=u.re_root,root_im=u.im_root,ws=new double [n],ws2=new double [n];
    for(int k=0;k<n;k++)
    {
	UnitTests.checkDoubleRel(root_im[k],"root_im[k]",0,"0",eps);
	final double [] q=Polynomials.sdiv1(nodes,xs[k]).q;
	final double t0=Polynomials.getD0(q,xs[k]);
	for(int m=q.length;--m>=0;) q[m]/=t0;
	ws[k]=TMatr.mulVVn1(q,mom);
	ws2[k]=TMatr.mulVVn1(Polynomials.mult2Pol(q,q),mom);
    }
    UnitTests.checkDoubleRel(TMatr.sumElements(ws),"sum(ws)",mom[0],"mom[0]",eps);

    final double [] pt=TMatr.getRandomVector(maxdegree+1,r,1,0);
    pt[0]+=0.6+r.nextDouble();
    double i=0;
    for(int k=0;k<xs.length;k++)
    {
	i+=ws[k]*Polynomials.getD0(pt,xs[k]);
    }
    final double ia=TMatr.mulVVn1(pt,mom);
    UnitTests.checkDoubleRel(i,"i",ia,"ia",eps);

    System.err.println(name+":test_NodesPolynomial(): test OK with n="+n+" maxdegree="+maxdegree+" eps="+eps);

    return new double [][]{xs,ws,ws2};
}

void test_QuadratureBasis(final double [] x,final double [] w,final double [] moments_b,final int maxdegree,final double eps)
{
    final double [] pt=TMatr.getRandomVector(maxdegree+1,r,1,0);
    double i=0;
    for(int k=0;k<x.length;k++)
    {
	i+=w[k]*O.B.getD0(pt,x[k]);
    }
    final double ia=TMatr.mulVVn1(pt,moments_b);
    UnitTests.checkDoubleRel(i,"i",ia,"ia",eps);

    System.err.println(name+":test_QuadratureBasis(): test OK with n_points="+x.length+" maxdegree="+maxdegree+" eps="+eps);
}

/** Out of 0 .. 2*n moments (n=p.length-1) used to construct p
 * only 0 .. 2*n-1 match. The last one is used
 * for normalization of p[n] and cannot be recovered.
 */
static double [] getMomentsFromPolynomials(final double [][] p)
{
    final int n=p.length;
    if(n<=0) throw new RuntimeException();
    final double [] mom=new double [2*n-1];
    // Obtain all moments from p. Use Gauss quadrature integration.
    final double [] xs=getXNodes(p[n-1]).re_root,ws=getWeightsChristoffel(xs,p);
    for(int k=mom.length;--k>=0;)
    {
	// monomial
	final double [] t=new double [k+1];t[k]=1;
	// integrate
	double s=0;
	for(int l=0;l<xs.length;l++) s+=ws[l]*Polynomials.getD0(t,xs[l]);
	mom[k]=s;
    }
    return mom;
}

static double [] getPolynomial(final int n,final double [] imom)
{
    final double [] rp=new double [n+1];
    for(int k=n;--k>=0;){rp[k]=-imom[n+k];}
    final double [] QQ=Polynomials.getQQMatr(n,imom);
    final int [] ipvt=new int [n];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(QQ,n,n,ipvt);
    if(pinfo!=0) java.util.Arrays.fill(rp,Double.NaN);
    else com.polytechnik.utils.Linsystems.dgesl_m1(QQ,n,n,ipvt,rp,1);
    rp[rp.length-1]=1;
    // normalize
    final double norm=Math.sqrt(TMatr.mulVVn1(Polynomials.mult2Pol(rp,rp),imom));
    for(int l=rp.length;--l>=0;){rp[l]/=norm;}
    return rp;
}

static double [][] getOrthogonalPolynomialsFirstKind(final int n, final double [] mom)
{
    final double [][] p=new double [n][];
    for(int k=0;k<n;k++)
    {
	final double [] pk=p[k]=getPolynomial(k,mom);
    }
    return p;
}
/** Like Polynomials.getSecondKindOrthogonalPolynomialFromFirstKind,
 *  but scaled to orthonormal basis.
 */
static double [] getSecondKindOrthogonalPolynomialFromFirstKindAndMoments(final double [] p,final double [] mom)
{
    final double [] q=Polynomials.getSecondKindOrthogonalPolynomialFromFirstKind(p,mom);
    // normalize to compensate extra sqrt(<1>) divided in getSecondKindOrthogonalPolynomialFromFirstKind.
    final double normtocompensate=Math.sqrt(mom[0]);
    for(int l=q.length;--l>=0;){q[l]*=normtocompensate;}
    return q;
}
static double [][] getSecondKindOrthogonalPolynomialsFromFirstKindAndMoments(final double [][] p, final double [] mom)
{
    final int n=p.length-1;
    final double [][] q=new double [n][];
    for(int k=0;k<n;k++)
    {
	q[k]=getSecondKindOrthogonalPolynomialFromFirstKindAndMoments(p[k+1],mom);
    }
    return q;
}
static PolynomialRootsFinder.PRoots getXNodes(final double [] nodes)
{
    final int n=nodes.length-1;
    final PolynomialRootsFinder.PRoots u=new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG().findRoots(nodes);
    if(u.nroots!=n) throw new RuntimeException("u.nroots="+u.nroots+" n="+n);
    return u;
}
static double [] getWeightsChristoffel(final double [] x,final double [][] p)
{
    final int n=x.length;
    if(n<=0) throw new RuntimeException();
    else if(n<=1)  return new double []{1/(p[0][0]*p[0][0])};
    final double [] ws=new double [n];
    final double f=(p[n][n]/p[n-1][n-1]);
    for(int k=0;k<n;k++)
    {
	final double xx=x[k];
	ws[k]=f/((Polynomials.getD0(p[n-1],xx)*Polynomials.getD1(p[n],xx)));
    }
    return ws;
}
static double [] getWeigthsSDiv(final double [] nodes,final double [] x,final double [][] p)
{
    final int n=nodes.length-1;
    final double [] ws=new double [x.length];
    final double [] quot=new double [1];
    // note, the integral dM =1/p[0][0]^2
    for(int k=0;k<x.length;k++)
    {
	final double xx=x[k];
	double [] tmp=Polynomials.sdiv1(nodes,xx).q;
	final double t0=Polynomials.getD0(tmp,xx);
	//System.err.println("f("+xx+"_"+k+"="+Polynomials.getD0(nodes,xx)+" tmp.length="+tmp.length);
	for(int m=tmp.length;--m>=0;) tmp[m]/=t0;
	for(int m=tmp.length;--m>=0;)
	{
	    if(p[m].length!=m+1||tmp.length!=m+1) throw new RuntimeException("m="+m);
	    final SDivRes psdr=Polynomials.sdiv_stable(tmp,p[m]);
	    tmp=psdr.r;
	    quot[0]=psdr.q[0];
	    //System.err.println("k="+k+" m="+m+" quot="+quot[0]+" tmp[0]="+tmp[0]+" "+(1/quot[0]));
	}
	ws[k]=(quot[0]/p[0][0]);
	//ws[k]=tmp[0];
    }
    return ws;
}
static double [] getWeightsByIntegration(final double [] nodes,final double [] x,final double [] moms)
{
    final double [] w=new double [x.length];
    for(int k=0;k<x.length;k++)
    {
	final double [] q=Polynomials.sdiv1(nodes,x[k]).q;
	final double t0=Polynomials.getD0(q,x[k]);
	for(int m=q.length;--m>=0;) q[m]/=t0;
	w[k]=TMatr.mulVVn1(q,moms);
    }
    return w;
}


double [] getRadauMomentsFromMoments(final double [] mom)
{
    final double [] rmom=new double [mom.length-1];
    for(int k=0;k<rmom.length;k++)
    {
	final double [] p=new double [k+1];
	p[k]=1;
	rmom[k]=TMatr.mulVVn1(getRadauMultiplied(p),mom);
    }
    return rmom;
}
double getX0()
{
    final double [] radau=getRadauMultiplied(new double []{1});
    final double x0=-radau[0]/radau[1];
    return x0;
}
Quadrature.GaussQuadratureWithNodesPolynomialData getRadauQuadratureFromPolynomials(final int n,final double [][] p)
{
    if(n+1>=p.length) throw new RuntimeException("n="+n+" p.length="+p.length);
    final int nn=n+1;
    final double [] matr=new double [nn*nn];
    final double [] radau=getRadauMultiplied(new double []{1}),quot=new double [1];
    final double x0=getX0();
    //System.err.println("x0="+x0);
    for(int k=0;k<nn;k++)
    {
	for(int l=0;l<=k;l++)
	{
	    double [] mul=Polynomials.mult2Pol(radau,p[l]);
	    quot[0]=0;
	    for(int j=l+1;j>=k;j--)
	    {
		if(mul.length!=j+1 || p[j].length!=j+1) throw new RuntimeException();
		final SDivRes psdr=Polynomials.sdiv_stable(mul,p[j]);
		mul=psdr.r;
		quot[0]=psdr.q[0];
	    }
	    matr[k+nn*l]=matr[l+nn*k]=quot[0]*1;
	}
    }

    final double [] matr2=new double [nn*nn];
    final int [] ipvt=new int [n];
    final double [][] qp=new double [nn][],qq=new double [nn][];
    for(int k=0;k<nn;k++)
    {
	System.arraycopy(matr,0,matr2,0,k*nn);
	final double [] rq=qq[k]=new double [k+1];
	rq[k]=1;
	for(int l=0;l<k;l++) rq[l]=-matr[k*nn+l];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr2,k,nn,ipvt);
	if(pinfo!=0) java.util.Arrays.fill(rq,Double.NaN);
	else com.polytechnik.utils.Linsystems.dgesl_m1(matr2,k,nn,ipvt,rq,1);
	// normalize
	final double s=Math.sqrt(TMatr.getAMatrA(k+1,rq,matr,nn));
	for(int l=0;l<=k;l++) rq[l]/=s;

	final double [] rp=qp[k]=new double [k+1];
	for(int l=0;l<=k;l++)
	{
	    TMatr.sumP1(rp,p[l],rq[l]);
	}
	//System.err.println("qp["+k+"]="+java.util.Arrays.toString(rp));
    }

    final double [] rnodes=qp[n],xr=getXNodes(rnodes).re_root,wr=getWeigthsSDiv(rnodes,xr,qp);
    final double [] xs=new double [n+1],ws=new double [n+1];
    for(int k=0;k<n;k++)
    {
	xs[k]=xr[k];
	ws[k]=wr[k]/Polynomials.getD0(radau,xs[k]);
    }
    xs[n]=x0;
    double [] tmp=new double [rnodes.length];
    System.arraycopy(rnodes,0,tmp,0,rnodes.length);
    final double t0=Polynomials.getD0(tmp,x0);
    for(int m=tmp.length;--m>=0;) tmp[m]/=t0;
    quot[0]=0;
    for(int k=tmp.length;--k>=0;)
    {
	if(p[k].length!=k+1||tmp.length!=k+1) throw new RuntimeException();
	final SDivRes psdr=Polynomials.sdiv_stable(tmp,p[k]);
	tmp=psdr.r;
	quot[0]=psdr.q[0];
    }
    ws[n]=(quot[0]/p[0][0]);

    return new Quadrature.GaussQuadratureWithNodesPolynomialData(xs,ws,getRadauMultiplied(rnodes));
}

static RecurrenceAB getAB(final double [][] p)
{
    final int n=p.length,ne=n-1;
    if(n<=0) return new RecurrenceAB(new double []{},new double []{});
    final double [] a=new double [n],b=new double [ne];
    a[0]=1/p[0][0];
    for(int k=0;k<ne;k++)
    {
	final double [] pk=p[k],pk1=p[k+1],tmp=Polynomials.getxP(pk);
	a[k+1]=(tmp[k+1]/pk1[k+1]);
	b[k]=(tmp[k]-a[k+1]*pk1[k])/pk[k];
    }
    //b[ne]=Double.NaN;
    // a & b are now generated
    return new RecurrenceAB(a,b);
}
static double [][] getOrthogonalPolynomialsFromAB(final int kind,final RecurrenceAB ab)
{
    if(kind>=ab.numAk()) return new double [][]{};
    // generate p - kind-type polynomials.
    final double [][] p=new double [ab.numAk()-kind][];
    p[0]=new double []{1/ab.getAk(kind)};
    for(int k=1;k<p.length;k++)
    {
	final double ak=ab.getAk(k+kind),bk1=ab.getBk(k+kind-1);
	final double [] pk1=p[k-1],pk=p[k]=Polynomials.getxP(pk1);// save p_{k} to p[k];
	pk[k]/=ak;
	for(int l=k;--l>=0;) pk[l]=(pk[l]-bk1*pk1[l])/ak;
	if(k>1) TMatr.sumP1(pk,p[k-2],-ab.getAk(k+kind-1)/ak);
    }
    return p;
}

static double [][] getSecondKindOrthogonalPolynomialsFromFirstKind(final double [][] p)
{
    final RecurrenceAB ab=getAB(p);
    return getOrthogonalPolynomialsSecondKindFromAB(ab);
}
static double [][] getOrthogonalPolynomialsSecondKindFromAB(final RecurrenceAB ab)
{
    return getOrthogonalPolynomialsFromAB(1,ab);
}
static double [][] getOrthogonalPolynomialsFirstKindFromAB(final RecurrenceAB ab)
{
    return getOrthogonalPolynomialsFromAB(0,ab);
}

void test_RegularNodes(final int neq)
{
    final double const_aver=mom_p[0];
    // regular nodes
    final double [][] p=getOrthogonalPolynomialsFirstKind(neq+1,mom_p);
    //System.err.println("mom_p="+java.util.Arrays.toString(mom_p));
    final double [][] b=O.getOrthogonalPolynomialsFirstKind(neq+1,java.util.Arrays.copyOf(mom_b,2*(neq+1)-1));
    //System.err.println("mom_b="+java.util.Arrays.toString(mom_b));
    for(int k=0;k<p.length;k++)
    {
	UnitTests.checkDoubleArrayRel(p[k],"p["+k+"]",convertToMonomials(b[k]),"conv(b["+k+"])",eps);
	final double [] bp=O.getSingleOrthogonalPolynomial(k,java.util.Arrays.copyOf(mom_b,2*k));
	UnitTests.checkDoubleArrayRel(bp,"bp",TMatr.getVectorScaled(b[k],1/b[k][b[k].length-1]),"b[k]",eps);
    }
    final double [] nodes=p[neq];
    final double [][] ndp=test_NodesPolynomial(nodes,mom_p,2*neq-1);
    final double [] xt=ndp[0],wt=ndp[1];
    sortNodes(xt,wt);
    if(true)
    {
	final RecurrenceAB ab=O.getAB(b);
	final Quadrature nab=O.getGaussQuadratureFromAB(neq,ab);
	final double [] xab=nab.x,wab=nab.w; // nodes polynomial is not direcly available
	sortNodes(xab,wab);
	UnitTests.checkDoubleArrayRel(xt,"xt",xab,"xab",eps);
	UnitTests.checkDoubleArrayRel(wt,"wt",wab,"wab",eps);


	final Quadrature.GaussQuadratureWithNodesPolynomialData nr=O.getGaussQuadratureFromPolynomials(neq,b);
	final double [] xm=nr.x,wm=nr.w,bnodes=nr.nodespolynomial;
	sortNodes(xm,wm);
	UnitTests.checkDoubleArrayRel(xt,"xt",xm,"xm",eps);
	UnitTests.checkDoubleArrayRel(wt,"wt",wm,"wm",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(wm),"sum(wm)",mom_p[0],"mom_p[0]",eps);

	final Quadrature.GaussQuadratureWithEVData ne=O.getGaussQuadratureFromMoments(neq,java.util.Arrays.copyOf(mom_b,2*neq));
	final double [] xe=ne.x,we=ne.w; // nodes polynomial is not direcly available
	sortNodes(xe,we);
	UnitTests.checkDoubleArrayRel(xe,"xe",xm,"xm",eps);
	UnitTests.checkDoubleArrayRel(we,"we",wm,"wm",eps);
	test_QuadratureBasis(xe,we,mom_b,2*neq-1,eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(we),"sum(we)",mom_p[0],"mom_p[0]",eps);

	final Quadrature.GaussQuadratureWithNodesPolynomialData nd=O.getGaussQuadratureFromMoments_Direct(neq,java.util.Arrays.copyOf(mom_b,2*neq));
	final double [] xd=nd.x,wd=nd.w,bnodesd=nd.nodespolynomial;
	sortNodes(xd,wd);
	UnitTests.checkDoubleArrayRel(xd,"xd",xm,"xm",eps);
	UnitTests.checkDoubleArrayRel(wd,"wd",wm,"wm",eps);
	test_QuadratureBasis(xd,wd,mom_b,2*neq-1,eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(wd),"sum(wd)",mom_p[0],"mom_p[0]",eps);
	for(int k=0;k<xm.length;k++)
	{
	    UnitTests.checkDoubleRel(O.B.getD0(bnodes,xm[k]),"bnodes(xm[k])",0,"0",eps);
	    UnitTests.checkDoubleRel(O.B.getD0(bnodesd,xd[k]),"bnodesd(xd[k])",0,"0",eps);
	}

	// match Christoffel function directly
	final double [] ChristoffelFun=O.getChristoffelOFunctionFromMoments(neq,mom_b),
	    ChristoffelFun_b=O.getChristoffelOFunctionFromPolynomials(neq,b);
	UnitTests.checkDoubleArrayRel(ChristoffelFun,"ChristoffelFun",ChristoffelFun_b,"ChristoffelFun_b",eps);
	for(int k=0;k<neq;k++){
	    UnitTests.checkDoubleRel(wm[k],"wm[k]",1/O.B.getD0(ChristoffelFun,xm[k]),"1/O.B.getD0(ChristoffelFun,xm[k])",eps);
	}

	final PolynomialRootsFinder.PRoots ub=O.getPolynomialRootsFinderInBasis().findRoots(bnodes);
	final double [] root_re=ub.re_root,root_im=ub.im_root;
	if(ub.nroots!=xm.length) throw new RuntimeException("xm.length="+xm.length+" ub.nroots="+ub.nroots);
	sortNodes(root_re,root_im);
	UnitTests.checkDoubleArrayRel(root_re,"root_re",xm,"xm",eps);
	UnitTests.checkDoubleArrayRel(root_im,"root_im",new double [xm.length],"0",eps);
    }

    final double [] xs=getXNodes(nodes).re_root;
    sortNodes(xs,null);
    final double [] ws=getWeightsChristoffel(xs,p);
    UnitTests.checkDoubleArrayRel(xt,"xt",xs,"xs",eps);
    UnitTests.checkDoubleArrayRel(wt,"wt",ws,"ws",eps);

    final double [] ws2=getWeigthsSDiv(nodes,xs,p);
    UnitTests.checkDoubleArrayRel(ws,"ws",ws2,"ws2",eps);

    //System.err.println("mom_p="+java.util.Arrays.toString(mom_p));
    final double [] bnodes=b[neq],bnx=bnodes.clone();
    final RecurrenceAB ab=O.getAB(b);
    final double [] ax=ab.getAllAk(),bx=ab.getAllBk();
    final double [] xb=O.getXNodesFromAB(neq,ab),xbp=O.getXNodes(bnodes).re_root;
    sortNodes(xb,null);sortNodes(xbp,null);
    UnitTests.checkDoubleArrayRel(xb,"xb",xbp,"xbp",eps);
    // to match exactly
    UnitTests.checkDoubleArrayRel(bnodes,"bnodes",bnx,"bnx",0*eps);
    UnitTests.checkDoubleArrayRel(ab.getAllAk(),"ab.getAllAk()",ax,"ax",0*eps);
    UnitTests.checkDoubleArrayRel(ab.getAllBk(),"ab.getAllBk()",bx,"bx",0*eps);

    final double [] xbx=O.getXNodesFromAB(neq,ab);
    sortNodes(xbx,null);
    UnitTests.checkDoubleArrayRel(xs,"xs",xb,"xb",eps);
    UnitTests.checkDoubleArrayRel(xbx,"xbx",xb,"xb",eps*0);
    final double bb00=Math.pow(O.B.getD0(b[0],0),2);
    UnitTests.checkDoubleRel(1/bb00,"1/bb00",const_aver,"const_aver",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(ws),"sum(ws)",1/bb00,"1/bb00",eps);
    UnitTests.checkDoubleArrayRel(ws,"ws",O.getWeights(bnodes,xb,b),"wb",eps);
    UnitTests.checkDoubleArrayRel(ws,"ws",O.getWeightsByIntegration(bnodes,xb,java.util.Arrays.copyOf(mom_b,bnodes.length-1)),"intwb",eps);
    UnitTests.checkDoubleArrayRel(ws,"ws",O.getWeightsChristoffel(xb,b),"Christoffel",eps);


    final Quadrature xwab=O.getGaussQuadratureFromAB(neq,new RecurrenceAB(java.util.Arrays.copyOf(ab.getAllAk(),neq+1),java.util.Arrays.copyOf(ab.getAllBk(),neq)));
    final double [] xab=xwab.x,wab=xwab.w;
    sortNodes(xab,wab);
    UnitTests.checkDoubleArrayRel(xab,"xab",xs,"xs",eps);
    UnitTests.checkDoubleArrayRel(wab,"wab",ws,"ws",eps);

    // one less a & b element available
    final RecurrenceAB abq=O.getABFromQuadrature(xab,wab);
    UnitTests.checkDoubleArrayRel(abq.getAllAk(),"abq.getAllAk()",java.util.Arrays.copyOf(ab.getAllAk(),abq.numAk()),"ab.getAllAk()",eps);
    UnitTests.checkDoubleArrayRel(abq.getAllBk(),"abq.getAllBk()",java.util.Arrays.copyOf(ab.getAllBk(),abq.numAk()-1),"ab.getAllBk()",eps);
    /* 1 less element to check
    final double [][] qabq=O.getGaussQuadratureFromAB(neq,abq[0],abq[1]);
    sortNodes(qabq[0],qabq[1]);
    UnitTests.checkDoubleArrayRel(qabq[0],"qabq[0]",xs,"xs",eps);
    UnitTests.checkDoubleArrayRel(qabq[1],"qabq[1]",ws,"ws",eps);
    */


    System.err.println(name+":test_RegularNodes(): test OK"+" neq="+neq+" eps="+eps);
}


void test_WExtremal(final int n)
{
    if(O.B==null) return;

    final Quadrature.GaussQuadratureWithNodesPolynomialData nr=O.getGaussQuadratureFromPolynomials(n,O.getOrthogonalPolynomialsFirstKind(n+1,mom_b));
    final double [] x=nr.x,w=nr.w,nodes=nr.nodespolynomial;
    final double [] QQ=O.B.getQQMatr(n,mom_b);

    for(int k=0;k<x.length;k++)
    {
	final double xk=x[k];
	final double [] psi=O.B.sdiv1(nodes,xk).q;
	final double fp=O.B.getD0(psi,xk),fdp=O.B.getD1(psi,xk);
	final double df=fdp/fp,wk=TMatr.getAMatrA(psi,QQ,n)/(fp*fp);
	UnitTests.checkDoubleRel(wk,"wk",w[k],"w[k]",eps);

	final double [] dpsi=O.B.getDifferentiated(psi);
	final double ddm=TMatr.getAMatrB(psi,dpsi,QQ,n)/(fp*fp),dd=ddm-wk*df;
	System.err.println("test_WExtremal: x["+k+"]="+x[k]+" wk="+wk+" ddm="+ddm+" df="+df+" dd="+dd);
	UnitTests.checkDoubleRel(dd,"dd",0,"0",eps);
    }

    System.err.println(name+":test_WExtremal(): test OK"+" neq="+n+" eps="+eps);
}

void test_RadauNodes(final int neq)
{
    // Radau nodes
    final double [][] pradau=getOrthogonalPolynomialsFirstKind(neq+1,getRadauMomentsFromMoments(mom_p));
    final double [][] nd=test_NodesPolynomial(getRadauMultiplied(pradau[neq]),mom_p,2*neq);
    final double [] xt=nd[0],wt=nd[1];
    sortNodes(xt,wt);

    final Quadrature.GaussQuadratureWithNodesPolynomialData nc=getRadauQuadratureFromPolynomials(neq,getOrthogonalPolynomialsFirstKind(neq+2,mom_p));
    final double [] xr=nc.x,wr=nc.w;
    UnitTests.checkDoubleRel(TMatr.sumElements(wr),"sum(wr)",mom_p[0],"mom_p[0]",eps);
    sortNodes(xr,wr);
    UnitTests.checkDoubleArrayRel(xr,"xr",xt,"xt",eps);
    UnitTests.checkDoubleArrayRel(wr,"wr",wt,"wt",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(wr),"sum(wr)",mom_p[0],"mom_p[0]",eps);

    final Quadrature.GaussQuadratureWithNodesPolynomialData nb=O.getRadauQuadratureFromPolynomials(neq,O.getOrthogonalPolynomialsFirstKind(neq+2,java.util.Arrays.copyOf(mom_b,2*(neq+2)-1)));
    final double [] xb=nb.x,wb=nb.w,bnodes=nb.nodespolynomial;
    UnitTests.checkDoubleRel(TMatr.sumElements(wb),"sum(wb)",mom_p[0],"mom_p[0]",eps);
    sortNodes(xb,wb);
    UnitTests.checkDoubleArrayRel(xr,"xr",xb,"xb",eps);
    UnitTests.checkDoubleArrayRel(wr,"wr",wb,"wb",eps);
    UnitTests.checkDoubleRel(TMatr.sumElements(wb),"sum(wb)",mom_p[0],"mom_p[0]",eps);


    final Quadrature.GaussQuadratureWithNodesPolynomialData ne=O.getRadauQuadratureFromMoments(neq,java.util.Arrays.copyOf(mom_b,2*neq+1));
    final double [] xe=ne.x,we=ne.w,bnodese=ne.nodespolynomial;
    UnitTests.checkDoubleRel(TMatr.sumElements(we),"sum(we)",mom_p[0],"mom_p[0]",eps);
    sortNodes(xe,we);
    UnitTests.checkDoubleArrayRel(xe,"xe",xr,"xr",eps);
    UnitTests.checkDoubleArrayRel(we,"we",wr,"wr",eps);

    for(int k=0;k<xe.length;k++)
    {
	UnitTests.checkDoubleRel(O.B.getD0(bnodes,xb[k]),"bnodes(xb[k])",0,"0",eps);
	UnitTests.checkDoubleRel(O.B.getD0(bnodese,xe[k]),"bnodese(xe[k])",0,"0",eps);
    }
    for(double [] no:new double [][]{bnodes,bnodese})
    {
	final PolynomialRootsFinder.PRoots ub=O.getPolynomialRootsFinderInBasis().findRoots(no);
	final double [] root_re=ub.re_root,root_im=ub.im_root;
	if(ub.nroots!=xb.length) throw new RuntimeException("xb.length="+xb.length+" ub.nroots="+ub.nroots);
	sortNodes(root_re,root_im);
	UnitTests.checkDoubleArrayRel(root_re,"root_re",xb,"xb",eps);
	UnitTests.checkDoubleArrayRel(root_re,"root_re",xe,"xe",eps);
	UnitTests.checkDoubleArrayRel(root_im,"root_im",new double [xb.length],"0",eps);
    }

    System.err.println(name+":test_RadauNodes(): test OK"+" neq="+neq+" eps="+eps);
}

void test_Type2(final int neq)
{
    final double [][] p=getOrthogonalPolynomialsFirstKind(neq+1,mom_p);
    final double [][] qC=getSecondKindOrthogonalPolynomialsFromFirstKindAndMoments(p,mom_p),
	qP=getSecondKindOrthogonalPolynomialsFromFirstKind(p);
    final double [][] qB=O.getSecondKindOrthogonalPolynomialsFromFirstKind(O.getOrthogonalPolynomialsFirstKind(neq+1,mom_b));
    for(int k=0;k<qC.length;k++)
    {
	UnitTests.checkDoubleArrayRel(qC[k],"qC["+k+"]",qP[k],"qP["+k+"]",eps);
	UnitTests.checkDoubleArrayRel(qC[k],"qC["+k+"]",convertToMonomials(qB[k]),"conv(qB["+k+"])",eps);
    }

    System.err.println(name+":test_Type2(): test OK"+" neq="+neq+" eps="+eps);
}
void test_TypeK(final int neq)
{
    double [][] p=getOrthogonalPolynomialsFirstKind(neq+1,mom_p);
    final RecurrenceAB ab=O.getAB(O.getOrthogonalPolynomialsFirstKind(neq+1,mom_b));
    double [] ra=ab.getAllAk(),rb=ab.getAllBk();
    for(int kind=1,n=ab.numAk();kind<ab.numAk();kind++,n--)
    {
	final double [] mom=getMomentsFromPolynomials(p);
	p=getSecondKindOrthogonalPolynomialsFromFirstKindAndMoments(p,mom);
	final double [][] b=O.getOrthogonalPolynomialsFromAB(kind,ab);
	ra=java.util.Arrays.copyOfRange(ra,1,ra.length);
	rb=java.util.Arrays.copyOfRange(rb,1,rb.length);
	final double [][] rpb=O.getOrthogonalPolynomialsFirstKindFromAB(new RecurrenceAB(ra,rb));
	for(int k=0;k<b.length;k++)
	{
	    UnitTests.checkDoubleArrayRel(p[k],"p["+k+"]",convertToMonomials(b[k]),"conv(b["+k+"])",eps);
	    UnitTests.checkDoubleArrayRel(rpb[k],"rpb["+k+"]",b[k],"b["+k+"]",eps);
	}
    }
    System.err.println(name+":test_TypeK(): test OK"+" neq="+neq+" eps="+eps);
}
void test_Polynomials(final int neq)
{
    final double [][] p=getOrthogonalPolynomialsFirstKind(neq,mom_p);
    // out of 2*neq-1 moments returned by getMomentsFromPolynomials
    // only 2*neq-2 match.
    if(p.length>1)UnitTests.checkDoubleArrayRel(java.util.Arrays.copyOf(mom_p,2*neq-2),"mom_p",java.util.Arrays.copyOf(getMomentsFromPolynomials(p),2*neq-2),"mom_p(p)",eps);
    final double [][] pb=O.getOrthogonalPolynomialsFirstKind(neq,java.util.Arrays.copyOf(mom_b,2*neq-1));
    if(pb.length>1)
    {
	final double [] mom_bb=O.getMomentsFromPolynomials(pb);
	UnitTests.checkDoubleArrayRel(mom_bb,"mom_b(pb)",java.util.Arrays.copyOf(mom_b,mom_bb.length),"mom_b",eps);
	final double [] momab_b=O.getMomentsFromAB(O.getAB(pb));
	UnitTests.checkDoubleArrayRel(mom_bb,"mom_b(pb)",momab_b,"momab_b",eps);
    }

    for(int k=0;k<pb.length;k++)
    {
	UnitTests.checkDoubleArrayRel(p[k],"p["+k+"]",convertToMonomials(pb[k]),"conv(pb["+k+"])",eps);
    }
    final double [][] qp=getSecondKindOrthogonalPolynomialsFromFirstKind(p),
	qpc=getSecondKindOrthogonalPolynomialsFromFirstKindAndMoments(p,mom_p),
	qpb=O.getSecondKindOrthogonalPolynomialsFromFirstKind(pb);
    for(int k=0;k<qpb.length;k++)
    {
	UnitTests.checkDoubleArrayRel(qp[k],"qp["+k+"]",convertToMonomials(qpb[k]),"conv(qpb["+k+"])",eps);
	UnitTests.checkDoubleArrayRel(qpc[k],"qpc["+k+"]",convertToMonomials(qpb[k]),"conv(qpb["+k+"])",eps);
    }
    System.err.println(name+":test_Polynomials(): test OK, neq="+neq+" eps="+eps);
}
void test_PolynomialsGramSchmidt(final int neq)
{
    final double [][] pb=OrthogonalPolynomialsBasisFunctionsCalculatable.getOrthogonalized(neq,O.B.getQQMatr(neq,java.util.Arrays.copyOf(mom_b,2*neq-1))),
	pb_grammschmidt=OrthogonalPolynomialsBasisFunctionsCalculatable.getGramSchmidtOrthogonalized(neq,O.B.getQQMatr(neq,mom_b));
    if(pb.length!=pb_grammschmidt.length) throw new RuntimeException("pb.length="+pb.length+" pb_grammschmidt.length="+pb_grammschmidt.length);

    for(int k=0;k<pb.length;k++)
    {
	final double [] pk=pb[k],pgsk=pb_grammschmidt[k].clone();
	if(Math.signum(pgsk[pgsk.length-1])!=Math.signum(pk[pk.length-1]))
	{
	    TMatr.scaleVectorInPlace(pgsk,-1);
	}
	UnitTests.checkDoubleArrayRel(pk,"pk",pgsk,"pgsk",eps);
    }
    System.err.println(name+":test_PolynomialsGramSchmidt(): test OK, neq="+neq+" eps="+eps);
}
void test_AB(final int neq)
{
    final double [][] p=getOrthogonalPolynomialsFirstKind(neq,mom_p),
	b=O.getOrthogonalPolynomialsFirstKind(neq,java.util.Arrays.copyOf(mom_b,2*neq-1));
    for(int k=0;k<p.length;k++)
    {
	UnitTests.checkDoubleArrayRel(p[k],"p["+k+"]",convertToMonomials(b[k]),"conv(b["+k+"])",eps);
    }
    final RecurrenceAB ab_p=getAB(p),ab_b=O.getAB(b);
    UnitTests.checkDoubleArrayRel(ab_p.getAllAk(),"ab_p.getAllAk()",ab_b.getAllAk(),"ab_b.getAllAk()",eps);
    UnitTests.checkDoubleArrayRel(ab_p.getAllBk(),"ab_p.getAllBk()",ab_b.getAllBk(),"ab_b.getAllBk()",eps);
    if(true)
    {
	final double [][] vp=getOrthogonalPolynomialsFirstKindFromAB(ab_p),
	    vb=O.getOrthogonalPolynomialsFirstKindFromAB(ab_b);
	for(int k=0;k<p.length;k++)
	{
	    UnitTests.checkDoubleArrayRel(p[k],"p["+k+"]",vp[k],"vp[k]",eps);
	    UnitTests.checkDoubleArrayRel(b[k],"b["+k+"]",vb[k],"vb[k]",eps);
	}
    }
    if(true)
    {
	final double [][] vp=getOrthogonalPolynomialsSecondKindFromAB(ab_p),
	    vc=getSecondKindOrthogonalPolynomialsFromFirstKindAndMoments(p,mom_p),
	    vb=O.getOrthogonalPolynomialsSecondKindFromAB(ab_b);
	for(int k=0;k<vp.length;k++)
	{
	    UnitTests.checkDoubleArrayRel(vp[k],"vp["+k+"]",convertToMonomials(vb[k]),"conv(vb["+k+"])",eps);
	    UnitTests.checkDoubleArrayRel(vc[k],"vc["+k+"]",convertToMonomials(vb[k]),"conv(vb["+k+"])",eps);
	}
	final RecurrenceAB ab2_p=getAB(vp),ab2_b=O.getAB(vb);
	UnitTests.checkDoubleArrayRel(ab2_p.getAllAk(),"ab2_p.getAllAk()",ab2_b.getAllAk(),"ab2_b.getAllAk()",eps);
	UnitTests.checkDoubleArrayRel(ab2_p.getAllBk(),"ab2_p.getAllBk()",ab2_b.getAllBk(),"ab2_b.getAllBk()",eps);

	final int no=ab_b.numAk();
	if(no>0)
	    UnitTests.checkDoubleArrayRel(ab2_b.getAllAk(),"ab2_b.getAllAk()",java.util.Arrays.copyOfRange(ab_b.getAllAk(),1,no),"rem0(ab_b.getAllAk())",eps);
	if(no>1)
	    UnitTests.checkDoubleArrayRel(ab2_b.getAllBk(),"ab2_b.getAllBk()",java.util.Arrays.copyOfRange(ab_b.getAllBk(),1,no-1),"rem0(ab_b.getAllBk())",eps);
    }

    System.err.println(name+":test_AB(): test OK neq="+neq+" eps="+eps);
}

void test_AB_as_Basis(final int neq){
    if(O.B==null){
	System.err.println(name+":test_AB_as_Basis(): no O.B, skipping test neq="+neq+" eps="+eps);
	return;
    }
    // the number of basis function should be at least the number of moments
    final int nbasisfun=mom_b.length;
    final double [][] pols=O.getOrthogonalPolynomialsFirstKind(nbasisfun,getBasisRandomMoments(2*nbasisfun-1,this.x_min,this.x_max,1000+nbasisfun*nbasisfun,this.r));
    final RecurrenceAB ab=O.getAB(pols);
    OrthogonalPolynomialsBasisFunctionsCalculatable.printAB("test_AB_as_Basis:Selected working basis:",ab);

    // run basis test for recurrence coefficients basis
    // use max degree neq-1, we have sufficient number of basis functions.
    new TestBasisFunctionsCalculatable("AB basis in "+name,ab,r,eps).runTestsOneN("Basis",1,neq-1);

    System.err.println(name+":test_AB_as_Basis(): test OK neq="+neq+" eps="+eps);
}

/** This method test a construction
 * of orthogonal polynomials identified by the recurrence coefficients
 * in any basis defined by {@link RecurrenceAB RecurrenceAB}.
 */
void test_AB_conversion_PBASIS(final int neq)
{
    if(O.B==null){
	System.err.println(name+":test_AB_conversion_PBASIS(): no O.B, skipping test neq="+neq+" eps="+eps);
	return;
    }
    final double [][] pols=O.getOrthogonalPolynomialsFirstKind(neq,java.util.Arrays.copyOf(mom_b,2*neq-1));
    final RecurrenceAB ab=O.getAB(pols);
    // A random polynomial in the basis of pols.
    final double [] rpol_basisAB=TMatr.getRandomVector(pols.length,r);
    // convert the basis
    final double [] rpol_basisO=O.B.convertFunctionTo_Basis_from_PBASIS(rpol_basisAB,ab),
	_rpol_basisAB=O.B.convertFunctionTo_PBASIS_from_Basis(rpol_basisO,ab);
    // match conversion
    UnitTests.checkDoubleArrayRel(rpol_basisAB,"rpol_basisAB",_rpol_basisAB,"_rpol_basisAB",eps);
    // check recurrence
    double [] ru=null,rv=null;
    for(int k=0;k<rpol_basisAB.length;k++){
	final double [] l=ab.getNext(k,ru,rv,ab);rv=ru;ru=l;
	final double [] lt=new double [k+1];lt[k]=1;
	UnitTests.checkDoubleArrayRel(l,"l",lt,"lt",eps);
	UnitTests.checkDoubleArrayRel(ab.convertFunctionTo_Basis_from_PBASIS(pols[k],O.B),"pols[k]",lt,"lt",eps);
	UnitTests.checkDoubleArrayRel(ab.convertFunctionTo_PBASIS_from_Basis(l,O.B),"l",pols[k],"pols[k]",eps);
    }

    // check recurrence in two bases
    final double [][] pols_in_B=ab.convertBasisToPBASIS(pols.length,O.B),
	pols_in_Monomial=ab.convertBasisToPBASIS(pols.length,Monomials.BASIS);
    for(int k=0;k<pols.length;k++){
	// check recurrence in O.B basis
	UnitTests.checkDoubleArrayRel(pols_in_B[k],"pols_in_B[k]",pols[k],"pols[k]",eps);
	// check recurrence in monomials basis
	UnitTests.checkDoubleArrayRel(pols_in_Monomial[k],"pols_in_Monomial[k]",O.B.convertFunctionTo_Monomials_from_Basis(pols[k]),"pols[k]",eps);
    }

    // Test conversion of two bases defined by the three--term recurrence: ab and ab2.
    // Compare direct conversion with intermediate throuh O.B basis.
    final double [][] pols2=O.getOrthogonalPolynomialsFirstKind(neq,getBasisRandomMoments(2*neq-1,this.x_min,this.x_max,1000+neq*neq,this.r));
    final RecurrenceAB ab2=O.getAB(pols2);
    final double [][] pols2_in_polsBasis_direct=ab2.convertBasisToPBASIS(pols2.length,ab);
    for(int k=0;k<pols2.length;k++){
	// direct expansion of ab2 in ab basis.
	final double [] pol2k_via_pols_direct=pols2_in_polsBasis_direct[k];
	// an expansion with an intermediate basis O.B
	final double [] pol2k_via_pols_intermediate=TMatr.expandOverLowerDiagBasis(pols2[k],pols);
	// match them
	UnitTests.checkDoubleArrayRel(pol2k_via_pols_direct,"pol2k_via_pols_direct",pol2k_via_pols_intermediate,"pol2k_via_pols_intermediate",eps);
    }

    System.err.println(name+":test_AB_conversion_PBASIS(): test OK neq="+neq+" eps="+eps);
}
void test_getQxFromAB(final int neq)
{
    final double [][] p=getOrthogonalPolynomialsFirstKind(neq,mom_p),
	b=O.getOrthogonalPolynomialsFirstKind(neq,java.util.Arrays.copyOf(mom_b,2*neq-1));
    final RecurrenceAB _ab=O.getAB(b);
    for(int k=0;k<p.length;k++)
    {
	final double [] Pb=getBasisRandomMoments(k+1,this.x_min,this.x_max,2*p.length+2,this.r);
	final RecurrenceAB _AB=new RecurrenceAB(java.util.Arrays.copyOf(_ab.getAllAk(),k+1),java.util.Arrays.copyOf(_ab.getAllBk(),k));
	for(int t=0;t<10000;t++)
	{
	    final double x=-r.nextDouble();
	    final double bx=O.B.getD0(b[k],x),px=Polynomials.getD0(p[k],x),
		bx_ab=_AB.calculate(k,x);
	    UnitTests.checkDoubleRel(bx,"bx",px,"px",eps);
	    UnitTests.checkDoubleRel(bx,"bx",bx_ab,"bx_ab",eps);
	    final double f=_AB.getD0(Pb,x);
	    double sf=0;
	    for(int j=0;j<Pb.length;j++){sf+=Pb[j]*_AB.calculate(j,x);}
	    UnitTests.checkDoubleRel(f,"f",sf,"sf",eps);
	}
    }

    System.err.println(name+":test_getQxFromAB(): test OK neq="+neq+" eps="+eps);
}

void test_getABRadauModified(final int neq)
{
    if(neq<=1) return;
    // test Radau AB
    for(int n=neq,nmax=neq+2;n<=nmax;n++)
    {
	final double [][] rr=getOrthogonalPolynomialsFirstKind(nmax-1,getRadauMomentsFromMoments(mom_p));

	final double [][] bb=O.getOrthogonalPolynomialsFirstKind(nmax,mom_b);
	final RecurrenceAB ab_bb=O.getAB(bb),abr=O.getABRadauModified(n,ab_bb);
	final double [][] bradau=O.getOrthogonalPolynomialsFirstKindFromAB(abr);
	final double [] rn=getRadauMultiplied(rr[n-2]),bn=convertToMonomials(bradau[n-1]);
	final double f=rn[rn.length-1]/bn[bn.length-1];for(int k=0;k<bn.length;k++) bn[k]*=f;
	UnitTests.checkDoubleArrayRel(rn,"rn",bn,"bn",eps);
	for(int k=0;k<bradau.length;k++)
	{
	    if(k<bradau.length-1)
	    {
		UnitTests.checkDoubleArrayRel(bradau[k],"bradau[k]",bb[k],"bb[k]",eps);
	    }
	    else
	    {
		final double [] brad=O.getRadauPolynomials(bb)[k-1],bmulrad=O.getRadauMultiplied(brad);
		final double ff=bradau[k][k]/bmulrad[k];
		for(int j=0;j<bmulrad.length;j++) bmulrad[j]*=ff;
		UnitTests.checkDoubleArrayRel(bradau[k],"bradau[k]",bmulrad,"bmulrad",eps);
	    }
	    //System.err.println("k="+k+" fx("+getX0()+")="+O.B.getD0(bradau[k],getX0()));
	}
	final int nq=bradau[bradau.length-1].length-1;
	System.err.println("x0="+getX0()+" n="+n+" neq="+neq+" F_end="+O.B.getD0(bradau[n-1],getX0())+" nq="+nq+" (abr.numAk()-1)="+(abr.numAk()-1));
	//UnitTests.checkDoubleArrayRel(rr[n-1],"rr[n-1]",convertToMonomials(bradau[n-1]),"conv(bradau[n-1])",eps);

	final Quadrature qd_ab=O.getGaussQuadratureFromAB(nq,abr),
	    qd_pol=O.getRadauQuadratureFromPolynomials(nq-1,bb);
	sortNodes(qd_ab.x,qd_ab.w);
	sortNodes(qd_pol.x,qd_pol.w);
	UnitTests.checkDoubleArrayRel(qd_ab.x,"qd_ab.x",qd_pol.x,"qd_pol.x",eps);
	UnitTests.checkDoubleArrayRel(qd_ab.w,"qd_ab.w",qd_pol.w,"qd_pol.w",eps);
    }
    System.err.println(name+":test_getABRadauModified(): test OK neq="+neq+" eps="+eps);
}

void test_RadauMoments_directly_Bbasis(final double [] bnodes,final double [] rbnodes){}
void test_RadauMoments(final int neq)
{
    final double [] rp=getRadauMomentsFromMoments(mom_p);
    final double [] rb=O.getRadauMomentsFromMoments(mom_b);
    final double [] rb_fromrp=getBasisMomentsConvertedFromMonomials(rp);
    // may be rather unstable
    if(rp.length<=10) UnitTests.checkDoubleArrayRel(rb,"rb",rb_fromrp,"rb_fromrp",eps);
    final double [][] brad=O.getRadauPolynomials(O.getOrthogonalPolynomialsFirstKind(neq+1,mom_b));
    final double [][] bradc=O.getOrthogonalPolynomialsFirstKind(neq,rb);
    for(int k=0;k<brad.length;k++)
    {
	UnitTests.checkDoubleArrayRel(brad[k],"brad[k]",bradc[k],"bradc[k]",eps);
    }
    test_RadauMoments_directly_Bbasis(mom_b,rb);
    System.err.println(name+":test_RadauMoments(): test OK neq="+neq+" eps="+eps);
}

void test_WeightsAndNodes(final int neq)
{
    final double [][] p=getOrthogonalPolynomialsFirstKind(neq+1,mom_p);
    final double [][] pb=O.getOrthogonalPolynomialsFirstKind(neq+1,mom_b);
    final double [] pnodes=p[neq],pbnodes=pb[neq];
    final RecurrenceAB pab=getAB(p),ab=O.getAB(pb);
    UnitTests.checkDoubleArrayRel(ab.getAllAk(),"ab.getAllAk()",pab.getAllAk(),"pab.getAllAk()",eps);
    UnitTests.checkDoubleArrayRel(ab.getAllBk(),"ab.getAllBk()",pab.getAllBk(),"pab.getAllBk()",eps);

    final double [] xp=getXNodes(pnodes).re_root,xb=O.getXNodesFromAB(neq,ab),xbx=O.getXNodesFromAB(neq,ab),xbp=O.getXNodes(pbnodes).re_root;
    sortNodes(xp,null);
    sortNodes(xb,null);
    sortNodes(xbx,null);
    sortNodes(xbp,null);
    UnitTests.checkDoubleArrayRel(xp,"xp",xb,"xb",eps);
    UnitTests.checkDoubleArrayRel(xbx,"xbx",xb,"xb",eps*0);
    UnitTests.checkDoubleArrayRel(xbp,"xbp",xb,"xb",eps);

    final double [] wp=getWeigthsSDiv(pnodes,xp,p),
	wpin=getWeightsByIntegration(pnodes,xp,mom_p),
	wb=O.getWeights(pbnodes,xb,pb),
	wbin=O.getWeightsByIntegration(pbnodes,xb,java.util.Arrays.copyOf(mom_b,pbnodes.length-1));
    UnitTests.checkDoubleArrayRel(wp,"wp",wb,"wb",eps);
    UnitTests.checkDoubleArrayRel(wpin,"wpin",wb,"wb",eps);
    UnitTests.checkDoubleArrayRel(wbin,"wbin",wb,"wb",eps);

    System.err.println(name+":test_WeightsAndNodes(): test OK neq="+neq+" eps="+eps);
}

void test_LebesgueQuadrature(final int n)
{
 final int nmoms=2*n-1,tsize=1000+n*n*10;
 final double [] ml=getBasisRandomMoments(nmoms,this.x_min,this.x_max,tsize,this.r);
 final LebesgueQuadratureWithEVData l=O.getLebesgueQuadratureFromMoments(n,ml,mom_b);
 EVXData.test_orthogonality("test_LebesgueQuadrature:l",n,l.ev,l.QQf,l.QQ,eps);
 final double [] mm=getBasisRandomMoments(nmoms,this.x_min,this.x_max,tsize,this.r);
 final LebesgueQuadratureWithEVData m=O.getLebesgueQuadratureFromMoments(n,mm,mom_b);
 EVXData.test_orthogonality("test_LebesgueQuadrature:m",n,m.ev,m.QQf,m.QQ,eps);
 final double [] DM=O.B.getQQMatr(n,getBasisRandomMoments(nmoms,this.x_min,this.x_max,tsize,this.r));
 final LebesgueQuadratureJointDistribution jdist=new LebesgueQuadratureJointDistribution(l,m);
 final double [] corr_prob=jdist.getProbabilityCorrelation(),corr_value=jdist.getValueCorrelation(),
     corr_DM=jdist.getDensityMatrixCorrelation(DM),
     lm_withDMdistrib=jdist.getFGWeightedWithCorrelation(corr_DM);
 final double [] proj_lm=EVXData.getBasisABProjections(n,l.ev,l.QQ,m.ev),
     proj_ml=EVXData.getBasisABProjections(n,m.ev,l.QQ,l.ev);
 System.err.println("test corr_value="+java.util.Arrays.toString(corr_value));
 UnitTests.checkDoubleArrayRel(l.QQ,"l.QQ",m.QQ,"m.QQ",0*eps); // to match exactly
 for(int i=0;i<n;i++)
 {
     final double [] l_psi_i=EVXData.getPsiFromBasis(i,jdist.ff.ev,n);
     double svl=0,svm=0;
     for(int j=0;j<n;j++)
     {
	 final double [] m_psi_j=EVXData.getPsiFromBasis(j,jdist.gg.ev,n);
	 UnitTests.checkDoubleRel(l.psiaverage[i],"l.psiaverage[i]",TMatr.mulVVn1(l_psi_i,mom_b),"<l_psi_i>",eps);
	 UnitTests.checkDoubleRel(m.psiaverage[j],"m.psiaverage[j]",TMatr.mulVVn1(m_psi_j,mom_b),"<m_psi_j>",eps);
	 final double l_psi_i__m_psi_j=TMatr.getAMatrB(l_psi_i,m_psi_j,m.QQ,n),
	     vcorr=l.psiaverage[i]*l_psi_i__m_psi_j*m.psiaverage[j],pcorr=Math.pow(l_psi_i__m_psi_j,2),
	     DMcorr=l_psi_i__m_psi_j*TMatr.getAMatrB(l_psi_i,m_psi_j,DM,n),
	     lm_DMcorr=jdist.ff.f[i]*DMcorr*jdist.gg.f[j];
	 UnitTests.checkDoubleRel(proj_lm[n*i+j],"proj_lm[n*i+j]",l_psi_i__m_psi_j,"l_psi_i__m_psi_j",eps);
	 UnitTests.checkDoubleRel(proj_ml[n*j+i],"proj_ml[n*j+i]",l_psi_i__m_psi_j,"l_psi_i__m_psi_j",eps);

	 UnitTests.checkDoubleRel(pcorr,"pcorr",corr_prob[i*n+j],"corr_prob[i+n*j]",eps);
	 UnitTests.checkDoubleRel(vcorr,"vcorr",corr_value[i*n+j],"corr_value[i+n*j]",eps);
	 UnitTests.checkDoubleRel(DMcorr,"DMcorr",corr_DM[i*n+j],"corr_DM[i+n*j]",eps);
	 UnitTests.checkDoubleRel(lm_DMcorr,"lm_DMcorr",lm_withDMdistrib[i*n+j],"lm_withDMdistrib[i+n*j]",eps);
	 svl+=corr_value[i*n+j];
	 svm+=corr_value[j*n+i];
     }
     UnitTests.checkDoubleRel(svl,"svl",l.w[i],"l.w[i]",eps);
     UnitTests.checkDoubleRel(svm,"svm",m.w[i],"m.w[i]",eps);
 }
 UnitTests.checkDoubleRel(TMatr.sumElements(corr_prob),"sum(corr_prob)",n,"n",eps);
 UnitTests.checkDoubleRel(TMatr.sumElements(corr_value),"sum(corr_value)",TMatr.sumElements(l.w),"sum(l.w)",eps);
 UnitTests.checkDoubleRel(TMatr.sumElements(corr_value),"sum(corr_value)",TMatr.sumElements(m.w),"sum(m.w)",eps);
 System.err.println(name+":test_LebesgueQuadrature(): test OK n="+n+" eps="+eps);
}

void test_PsiAver_and_LebesgueQuadrature(final int n)
{
    final int nmoms=2*n-1,tsize=1000+n*n*10;
    final double [] ml=getBasisRandomMoments(nmoms,this.x_min,this.x_max,tsize,this.r);
    final double [] Rev=O.EV.getGEV_Generalized(n,O.B.getQQMatr(n,ml),O.B.getQQMatr(n,mom_b),true),psi=new double [n];
    final LebesgueQuadratureWithEVData l=O.getLebesgueQuadratureFromMoments(n,ml,mom_b);
    for(int k=0;k<n;k++)
    {
	System.arraycopy(Rev,n+k*n,psi,0,n);
	final double [] psi2=O.B.mult2Pol(psi,psi); // multiply without O.B available
	final double s1=TMatr.mulVVn1(psi,ml),w1=TMatr.mulVVn1(psi,mom_b),
	    s2=TMatr.mulVVn1(psi2,ml),w2=TMatr.mulVVn1(psi2,mom_b);
	UnitTests.checkDoubleRel(w2,"w2",1,"1",eps);
	UnitTests.checkDoubleRel(s2/w2,"s2/w2",Rev[k],"Rev[k]",eps);
	UnitTests.checkDoubleRel(s1/w1,"s1/w1",Rev[k],"Rev[k]",eps);
	UnitTests.checkDoubleRel(Math.pow(w1,2),"w1^2",l.w[k],"l.w[k]",eps);
	UnitTests.checkDoubleRel(Rev[k],"Rev[k]",l.f[k],"l.f[k]",eps);
    }

    System.err.println(name+":test_PsiAver_and_LebesgueQuadrature(): test OK n="+n+" eps="+eps);
}

void test_GenerateMeasureProducingPolynomial(final int n)
{
    final double epsS=1e-4;
    for(int j=0;j<2;j++)
    {
	final double [] Pb_toproduce=j<=0?
	    getBasisRandomMoments(2*n-1,this.x_min,this.x_max,2*n+10,this.r)
	    :
	    O.getChristoffelOFunctionFromMoments(n,mom_b);
	final double [] QQ=O.B.getQQMatr(n,mom_b),IQQ=com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQ),
	    Pb_back=O.B.getKK(n,O.B.getQQMatr(n,O.B.getMomentsOfMeasureProducingPolynomialInKK(n,Pb_toproduce))),
	    PbM_back=O.B.getKK(n,TMatr.mulMMM(n,IQQ,O.B.getQQMatr(n,O.B.getMomentsOfMeasureProducingPolynomialInKK_MQQM(n,Pb_toproduce,IQQ)),IQQ));
	UnitTests.checkDoubleArrayRel1(Pb_toproduce,"Pb_toproduce",Pb_back,"Pb_back",Math.max(eps,epsS));
	UnitTests.checkDoubleArrayRel1(Pb_toproduce,"Pb_toproduce",PbM_back,"PbM_back",Math.max(eps,epsS));
	// Test Pb obtained using density matrix
	final LebesgueQuadratureWithEVData lk=O.getDensityMatrixProducingGivenPolynomial(n,Pb_toproduce,QQ);
	final double [] Pb_back2=new double [Pb_toproduce.length];
	for(int k=0;k<n;k++)
	{
	    final double [] psi=EVXData.getPsiFromBasis(k,lk.ev,n);
	    TMatr.sumP1(Pb_back2,O.B.mult2Pol(psi,psi),lk.f[k]);
	}
	UnitTests.checkDoubleArrayRel1(Pb_toproduce,"Pb_toproduce",Pb_back2,"Pb_back2",Math.max(eps,epsS));
	System.err.println(name+":test_GenerateMeasureProducingPolynomial():density matrix spectrum="+java.util.Arrays.toString(lk.f));
    }

    System.err.println(name+":test_GenerateMeasureProducingPolynomial(): test OK n="+n+" eps="+eps);
}


void test_TwoQuadraticFormsProduct(final int n)
{
    final double [] mom_A=getBasisRandomMoments(2*n-1,this.x_min,this.x_max,1000+n*n,this.r);
    for(int nA=1;nA<=n;nA++){
	final double [] QQA=O.B.getQQMatr(nA,mom_A);
	for(int nB=1;nB<=n;nB++){
	    final double [] QQB=O.B.getQQMatr(nB,mom_b);

	    final double [] ProdTwoFormsDirect=O.B.mult2Pol(O.B.getKK(nA,QQA),O.B.getKK(nB,QQB)),
		MQQA_MQQB_as_Product=O.B.getTwoQuadraticFormsProductAsQuadraticForm(nA,QQA,nB,QQB),
		Prod=O.B.getKK(nA+nB-1,MQQA_MQQB_as_Product);
	    UnitTests.checkDoubleArrayRel1(ProdTwoFormsDirect,"ProdTwoFormsDirect",Prod,"Prod",eps);
	}
    }

    System.err.println(name+":test_TwoQuadraticFormsProduct(): test OK n="+n+" eps="+eps);
}

void test_ChristoffelFunction(final int n)
{
    final double [] QQp=Polynomials.getQQMatr(n,mom_p);
    final double [] kkb_christoffel=O.getChristoffelOFunctionFromMoments(n,mom_b),
	kkp_christoffel_fromB=convertToMonomials(kkb_christoffel),
	kkp_christoffel=Polynomials.getKK(n,com.polytechnik.utils.Linsystems.getInvertedMatrix(n,n,QQp));
    UnitTests.checkDoubleArrayRel(kkp_christoffel_fromB,"kkp_christoffel_fromB",kkp_christoffel,"kkp_christoffel",eps);
    // check with orthogonal polynomials
    final double [][] b=O.getOrthogonalPolynomialsFirstKind(n,java.util.Arrays.copyOf(mom_b,2*n-1));
    final double [] kkbP_christoffel=O.getChristoffelOFunctionFromPolynomials(n,b);
    UnitTests.checkDoubleArrayRel(kkb_christoffel,"kkb_christoffel",kkbP_christoffel,"kkbP_christoffel",eps);

    System.err.println(name+":test_ChristoffelFunction(): test OK n="+n+" eps="+eps);
}


void test_DistributionSkewness()
{
    final Skewness sk=Skewness.getSkewness(mom_b[0],mom_b[1],mom_b[2],mom_b[3],O),
	sk2=Skewness.getSkewness2(mom_b[0],mom_b[1],mom_b[2],mom_b[3],O),
	psk=Skewness.getSkewnessFromMonomialMoments(mom_p[0],mom_p[1],mom_p[2],mom_p[3]),
	psk2=Skewness.getSkewnessFromMonomialMoments2(mom_p[0],mom_p[1],mom_p[2],mom_p[3]);
    Skewness.matchSkewness(sk,sk2,eps);
    Skewness.matchSkewness(sk,psk,eps);
    Skewness.matchSkewness(sk,psk2,eps);
    final double [] Xmom_b=O.B.getXQkMomentsFromQkMoments(mom_b);
    final Skewness sk3=Skewness.getGSkewness(mom_b[0],mom_b[1],mom_b[2],Xmom_b[0],Xmom_b[1],Xmom_b[2],O);
    Skewness.matchSkewness(sk,sk3,eps);

    // Test explicit skewness expression
    UnitTests.checkDoubleRel(sk.sk,"sk.sk",(2*sk.xa-sk.x1-sk.x2)/(sk.x1-sk.x2),"(2*sk.xa-sk.x1-sk.x2)/(sk.x1-sk.x2)",eps);
    UnitTests.checkDoubleRel(sk.sk,"sk.sk",(sk.w1-sk.w2)/(sk.w1+sk.w2),"(sk.w1-sk.w2)/(sk.w1+sk.w2)",eps);
    final ProbabilityCorrelation pc=
	new ProbabilityCorrelation(mom_p[0],mom_p[1],mom_p[2],
				   mom_p[1],mom_p[2],mom_p[3],
				   mom_p[1],mom_p[2],mom_p[3],true);
    //UnitTests.checkDoubleRel(sk.sk,"sk.sk",pc.skewnessLike_p,"pc.skewnessLike_p",eps);
    //UnitTests.checkDoubleRel(sk.sk,"sk.sk",pc.skewnessLike_i,"pc.skewnessLike_i",eps);
    UnitTests.checkDoubleRel(1,"1",pc.corr,"pc.corr",eps);


    final double [] momI_p=getPolynomialRandomMoments(3,this.x_min,this.x_max,100,this.r),
	momI_b=getBasisMomentsConvertedFromMonomials(momI_p),
	momP_p=getPolynomialRandomMoments(3,this.x_min,this.x_max,100,this.r),
	momP_b=getBasisMomentsConvertedFromMonomials(momP_p);
    for(boolean correlationtype:new boolean[]{true,false})
    {
	final ProbabilityCorrelation c=new ProbabilityCorrelation(mom_p[0],mom_p[1],mom_p[2],
								  momP_p[0],momP_p[1],momP_p[2],
								  momI_p[0],momI_p[1],momI_p[2],correlationtype),
	    bc=new ProbabilityCorrelation(mom_b[0],mom_b[1],mom_b[2],
					  momP_b[0],momP_b[1],momP_b[2],
					  momI_b[0],momI_b[1],momI_b[2],
					  O,correlationtype,eps);
	ProbabilityCorrelation.matchProbabilityCorrelation(c,bc,eps);
    }

    final Skewness gP=Skewness.getGSkewnessFromMonomialMoments(mom_p[0],mom_p[1],mom_p[2],momP_p[0],momP_p[1],momP_p[2]),
	bgP=Skewness.getGSkewness(mom_b[0],mom_b[1],mom_b[2],momP_b[0],momP_b[1],momP_b[2],O);
    Skewness.matchSkewness(gP,bgP,eps);

    System.err.println(name+":test_DistributionSkewness(): test OK eps="+eps);
}

void test_Discrete(final int n)
{
    final Quadrature xw=getXWDiscrete(n);
    final double [] bmoms=O.B.getMomentsFromSample(2*n,xw.x,xw.w);

    if(true)
    { // discrete measure to match quadrature exactly
	final Quadrature xwq=O.getGaussQuadratureFromMoments(n,java.util.Arrays.copyOf(bmoms,2*n));
	UnitTests.checkDoubleArrayRel(xw.x,"xw.x",xwq.x,"xwq.x",eps);
	UnitTests.checkDoubleArrayRel(xw.w,"xw.w",xwq.w,"xwq.w",eps);
    }

    final double [] w0=new double [n],wc=xw.w.clone(),xc=xw.x.clone();
    java.util.Arrays.fill(w0,1);
    sortNodes(wc,xc);

    final double [][] q2=getD2QuadratureFromMoments_EV(n,bmoms,O.B.getMomentsFromSample(2*n,xw.x,w0));
    final double [] wq=q2[0],xv=q2[1],xt=q2[2];

    UnitTests.checkDoubleArrayRel(wq,"wq",wc,"wc",eps);
    UnitTests.checkDoubleArrayRel(xv,"xv",xt,"xt",eps);
    UnitTests.checkDoubleArrayRel(xc,"xc",xt,"xt",eps);

    if(true)
    { // minimal eigenvalue psi(x) to have only simple real distinct roots, not necessary on the support.
	final double [] QQbv=O.B.getQQMatr(n+1,O.B.getMomentsFromSample(2*n+1,xw.x,xw.w)),QQbt=O.B.getQQMatr(n+1,mom_b);
	final double [] ev=O.EV.getGEV_Generalized(n+1,QQbv,QQbt,true);
	final double [] psi=EVXData.getPsiFromBasis(0,ev,n+1);
	final PolynomialRootsFinder.PRoots ub=O.getPolynomialRootsFinderInBasis().findRoots(psi);
	final double [] xre=ub.re_root,xim=ub.im_root;
	if(ub.nroots!=n) throw new RuntimeException("ub.nroots="+ub.nroots+" n="+n);
	sortNodes(xre,xim);
	UnitTests.checkDoubleArrayRel(xw.x,"xw.x",xre,"xre",eps);
	UnitTests.checkDoubleArrayRel(new double [n],"0",xim,"xim",eps);
	final EVXData J=EVXData.getEVXDataFromQQ(n+1,QQbv,QQbt,O);
	UnitTests.checkDoubleRel(J.s0,"J.s0",J.getRNat(O.B.getMomentsFromSample(n+1,new double []{J.x0},new double []{1})),"J at x0",eps);
	final LebesgueQuadratureWithEVData lq=
	    O.getLebesgueQuadratureFromMatrices(n+1,QQbv,QQbt);
	UnitTests.checkDoubleArrayRel(lq.f,"lq.f",J.getEValues(),"J.ev",eps);
    }

    System.err.println(name+":test_Discrete(): test OK neq="+n+" eps="+eps);
}

void test_ECOND(final int n)
{
    final double [] momt=mom_b;
    setData(200+30*momt.length,momt.length);
    final double [] momv=mom_b;
    test_ONEECOND(n,momv,momt,"ORD");
    test_ONEECOND(n,O.getRadauMomentsFromMoments(momv),O.getRadauMomentsFromMoments(momt),"RAD");
}

/** Print the results of
Carlos F. Borges,Numer. Math. 67: 271-288 (1994),
On a class of Gauss-like quadrature rules.
x1 = 0.791263; x2 = 0.237959; x3 = 0.427455; x4 = 0.887481
Q1f = PI*(0.079050 f(x1) + 0.204235 f(x2) + 0.176469 f(x3) + 0.040245 f(x4)).
Q2f = 0.082275 f(x1) + 0.507998 f(x2) + 0.866157 f(x3) + 0.543569 f(x4).
The results match exactly.
*/
void test_Borges(final int neq)
{
    final int [] multiindex=new int []{neq,neq};
    final int N=multiindex[0]+multiindex[1];
    final int nmoms=N+neq;
    final double [] pmom1=new double [nmoms],pmom2=new double [nmoms];
    for(int k=0;k<nmoms;k++)
    {
	final boolean e=k%2==0;
	// measure sqrt(1-x^2)dx
	// Gradstein Ryzik, 1963, 3.248.{2,3}, page 308 and integration by parts.
	pmom1[k]=e?Math.PI*Polynomials.fact2(2+k-1)/Polynomials.fact2(2+k)/(k+1):0;
	// measure (1+x)dx
	pmom2[k]=e?2.0/(k+1):2.0/(k+2);
    }
    final double [][] measures_b=new double [][]{
	getBasisMomentsConvertedFromMonomials(pmom1),
	getBasisMomentsConvertedFromMonomials(pmom2)
    };
    final Quadrature.MultipleOrthogonalityQuadrature quad=
	O.getQuadraturesForMultipleOrthogonalPolynomial(multiindex,measures_b);
    final double [] xe=quad.x,imxe=quad.imx;
    System.err.println("test_Borges:N="+N+" multiindex="+java.util.Arrays.toString(multiindex));
    System.err.println("test_Borges:xe="+java.util.Arrays.toString(xe));
    System.err.println("test_Borges:imxe="+java.util.Arrays.toString(imxe));
    for(int k=0;k<multiindex.length;k++)
    {
	System.err.println("test_Borges:W["+k+"]="+java.util.Arrays.toString(quad.w[k]));
	test_QuadratureBasis(xe,quad.w[k],measures_b[k],N-1+multiindex[k],eps);
    }
    System.err.println(name+":test_Borges(): test OK"+" neq="+neq+" eps="+eps);
}

/** Test multiple orthogonality on the same interval.
 * Because the measures not always form Chebyshev system
 * the unit test often fails.
 */
void test_MultipleOrthogonality(final int neq)
{
    final int [] multiindex=new int [neq];
    int N=0;
    for(int k=0;k<multiindex.length;k++)
    {
	multiindex[k]=r.nextInt(5);
	N+=multiindex[k];
    }
    System.err.println("N="+N+" multiindex="+java.util.Arrays.toString(multiindex)+" x_min="+x_min+" x_max="+x_max);
    final double [][] measures_b=new double [multiindex.length][];
    final int tsize=N*N+1001;
    for(int k=0;k<multiindex.length;k++)
    {
	if(true)
	{
	    //measures_b[k]=getBasisRandomMoments(N+multiindex[k],this.x_min,this.x_max,tsize,this.r);
	    measures_b[k]=getBasisMomentsConvertedFromMonomials(getPolynomialRandomMoments(N+multiindex[k],this.x_min,this.x_max,tsize,this.r));
	}
	else
	{
	    final Quadrature xw=getXWDiscrete(2*N+100);
	    measures_b[k]=O.B.getMomentsFromSample(N+multiindex[k],xw.x,xw.w);
	}
    }

    final Quadrature.MultipleOrthogonalityQuadrature quad=
	O.getQuadraturesForMultipleOrthogonalPolynomial(multiindex,measures_b);
    final double [] xe=quad.x,imxe=quad.imx;
    for(int k=0;k<imxe.length;k++)
    {
	if(!(imxe[k]==0))
	{
	    System.err.println("N="+N+" xe="+java.util.Arrays.toString(xe)+"\nimxe="+java.util.Arrays.toString(imxe)+" skipping as im part is not 0");
	    return;
	}
    }
    final double epsS=1e-3; // rather large, the quadratures are often not that good.
    System.err.println("N="+N+" xe="+java.util.Arrays.toString(xe)+"\nim="+java.util.Arrays.toString(imxe));
    for(int k=0;k<multiindex.length;k++)
    {
	System.err.println("measure#"+k+" weights="+java.util.Arrays.toString(quad.w[k]));
	test_QuadratureBasis(xe,quad.w[k],measures_b[k],N-1+multiindex[k],epsS);
    }

    System.err.println(name+":test_MultipleOrthogonality(): test OK"+" neq="+neq+" eps="+eps+" epsS="+epsS);
}


/** Test multiple orthogonality on the disjoin intervals.
 */
void test_MultipleOrthogonality_Disjoin(final int neq)
{
    final int [] multiindex=new int [neq];
    int N=0;
    final double [] intervals=new double [multiindex.length+1];
    intervals[0]=this.x_min;
    final double delta=(this.x_max-this.x_min)/multiindex.length;
    for(int k=0;k<multiindex.length;k++)
    {
	intervals[k+1]=(k+1)*delta;
	multiindex[k]=1+r.nextInt(3);
	N+=multiindex[k];
    }
    System.err.println("N="+N+" multiindex="+java.util.Arrays.toString(multiindex)+" x_min="+x_min+" x_max="+x_max);
    final double [][] measures_b=new double [multiindex.length][];
    final int tsize=N*N+1001;
    for(int k=0;k<multiindex.length;k++)
    {
	measures_b[k]=getBasisRandomMoments(N+multiindex[k],intervals[k],intervals[k+1],tsize,this.r);
    }

    final Quadrature.MultipleOrthogonalityQuadrature quad=O.getQuadraturesForMultipleOrthogonalPolynomial(multiindex,measures_b);
    final double [] xe=quad.x,imxe=quad.imx;

    System.err.println("N="+N+" xe="+java.util.Arrays.toString(xe)+"\nim="+java.util.Arrays.toString(imxe));
    final double epsS=1e-3; // rather large, the quadratures are often not that good.
    for(int k=0;k<multiindex.length;k++)
    {
	int nintk=0;
	for(int j=0;j<xe.length;j++)
	{
	    if(intervals[k]<=xe[j] && xe[j]<=intervals[k+1])
	    {
		nintk++;
	    }
	}
	// the number of roots on k-th interval must be equal to multiindex.
	if(nintk!=multiindex[k])
	{
	    throw new RuntimeException("N="+N+" k="+k+" nintk="+k+" multiindex["+k+"]="+multiindex[k]);
	}
	// often not perfect, because support intervals are disjoin
	System.err.println("interval#"+k+" ["+intervals[k]+";"+intervals[k+1]+"] weights="+java.util.Arrays.toString(quad.w[k]));
	// many points are outside of the support of k-th measure
	// weights are ofthen negative & unstable for large number of intervals.
	if(false)
	{
	test_QuadratureBasis(xe,quad.w[k],measures_b[k],N-1+multiindex[k],epsS);
	}
    }

    System.err.println(name+":test_MultipleOrthogonality_Disjoin(): test OK"+" neq="+neq+" eps="+eps+" epsS="+epsS);
}




void test_Kronrod(final int n,final double [] moms,final boolean flag_print)
{
    final Quadrature.KronrodQuadrature quad=O.getKronrodQuadrature(n,java.util.Arrays.copyOf(moms,3*n+2));
    if(flag_print)
    {
	System.err.println("test_Kronrod: n="+n+" n_points="+(quad.xg.length+quad.xk.length));
	System.err.println("test_Kronrod:G:x="+java.util.Arrays.toString(quad.xg));
	System.err.println("test_Kronrod:G:im(x)="+java.util.Arrays.toString(quad.imxg));
	System.err.println("test_Kronrod:G:w="+java.util.Arrays.toString(quad.wg));
	System.err.println("test_Kronrod:K:x="+java.util.Arrays.toString(quad.xk));
	System.err.println("test_Kronrod:K:im(x)="+java.util.Arrays.toString(quad.imxk));
	System.err.println("test_Kronrod:K:w="+java.util.Arrays.toString(quad.wk));
    }

    final Quadrature K=quad.getGKCombined();

    test_QuadratureBasis(K.x,K.w,moms,3*n+1,eps);

    System.err.println(name+":test_Kronrod(): test OK"+" n="+n+" eps="+eps);
}


void test_Kronrod_random(final int n)
{
    final int tsize=1000+n*n*10;
    final double [] mom_b=getBasisRandomMoments(3*n+2,this.x_min,this.x_max,tsize,this.r);
    test_Kronrod(n,mom_b,true);
}


void test_ONEECOND(final int n,final double [] momv,final double [] momt,final String txt)
{

    final double [] QQbv=O.B.getQQMatr(n,momv),QQbt=O.B.getQQMatr(n,momt);
    final double [] ev=O.EV.getGEV_Generalized(n,QQbv,QQbt,true);

    final double [] psi=new double [n];
    final double [] x0=new double [n-1];
    for(int k=0;k<n;k++)
    {
	final double lam=ev[k];
	System.arraycopy(ev,n+k*n,psi,0,n);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(n,psi,QQbv,n),"<psi|QQbv|psi>",lam,"lam",eps);
	UnitTests.checkDoubleRel(TMatr.getAMatrA(n,psi,QQbt,n),"<psi|QQbt|psi>",1,"1",eps);
	final double apsi_v=TMatr.mulVVn1(psi,momv),apsi_t=TMatr.mulVVn1(psi,momt);

	final PolynomialRootsFinder.PRoots ub=O.getPolynomialRootsFinderInBasis().findRoots(psi);
	final double [] xre=ub.re_root,xim=ub.im_root;
	sortNodes(xre,xim);
	System.err.println("k="+k+" lam="+lam+" xre="+java.util.Arrays.toString(xre)+" xim="+java.util.Arrays.toString(xim)+"\tapsi_v="+apsi_v+" apsi_t="+apsi_t);
	if(ub.nroots!=(n-1)) throw new RuntimeException("ub.nroots="+ub.nroots+" n="+n);
	if(n>=2 && (k==0||k==(n-1)) )
	{
	    /*
	    final double [] xmom=new double [momt.length];
	    final int sign=k<=0?1:-1;
	    for(int j=0;j<momt.length;j++)
	    {
		xmom[j]=sign*(momv[j]-ev[k]*momt[j]);
	    }
	    final double [][] qa=O.getQuadratureFromMoments_EV(n-1,xmom);
	    sortNodes(qa[0],qa[1]);
	    final double [] x_qa=qa[0],w_qa=qa[1];
	    System.err.println("k="+k+" x_qa="+java.util.Arrays.toString(x_qa)+" w_qa="+java.util.Arrays.toString(w_qa));
	    UnitTests.checkDoubleArrayRel(xre,"xre",x_qa,"x_qa",eps);
	    */
	    for(int m=0;m<xre.length;m++)
	    {
		final SDivRes sdr=O.B.sdiv1(psi,xre[m]);
		final double [] psd=sdr.q;
		final double rem=sdr.r[0],
		    bt=TMatr.getAMatrB(psi.length,psi,0,psd.length,psd,0,QQbt,n),
		    bv_lam=TMatr.getAMatrB(psi.length,psi,0,psd.length,psd,0,QQbv,n)/lam;
		System.err.println("bt="+bt+" bv/lam="+(bv_lam)+" rem="+rem);
		UnitTests.checkDoubleRel(bt,"bt",bv_lam,"bv_lam",eps);
		//UnitTests.checkDoubleRel(rem,"rem",0,"0",eps);
	    }

	    if(xre[0]<=x_min||xre[n-2]>=x_max){
		System.err.println("test_ONEECOND:"+txt+": At least ONE root outside of the interval: xre[0]="+xre[0]+
				   "<x_min="+x_min+" xre["+(n-2)+"]="+xre[n-2]+">x_max="+x_max);
	    }
	    if(n>2&&xre[0]<=x_min&&xre[n-2]>=x_max){
		System.err.println("test_ONEECOND:"+txt+": At TWO roots outside of the interval: xre[0]="+xre[0]+
				   "<x_min="+x_min+" xre["+(n-2)+"]="+xre[n-2]+">x_max="+x_max);
	    }
	}
	if(k==0||k==(n-1))
	{
	    UnitTests.checkDoubleArrayRel(xim,"xim",x0,"0",eps);
	}
    }
    System.err.println(name+":test_ONEECOND():"+txt+": test OK neq="+n+" eps="+eps);
}

int getMaxNeqCut(final int neq){return neq<=4?neq:4;}

void test_Calculations(final int tsize,final int neq)
{
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println(name+":test_Calculations(): test with N points="+tsize+
		       " neq="+neq+" eps="+eps);

    setData(tsize,2*neq+5);
    test_RegularNodes(neq);
    test_WExtremal(neq);
    test_RadauNodes(neq);
    test_Type2(neq);
    test_TypeK(neq);
    test_Polynomials(neq);
    test_PolynomialsGramSchmidt(neq);
    test_AB(neq);
    test_AB_conversion_PBASIS(neq);
    test_AB_as_Basis(neq);
    test_getQxFromAB(neq);
    test_getABRadauModified(neq);
    test_RadauMoments(neq);
    test_WeightsAndNodes(neq);
    if(mom_b.length>=4) test_DistributionSkewness();
    test_Discrete(neq);
    test_PsiAver_and_LebesgueQuadrature(neq);
    test_LebesgueQuadrature(neq);
    test_GenerateMeasureProducingPolynomial(neq>5?5:neq);
    test_TwoQuadraticFormsProduct(neq>5?5:neq);
    test_ChristoffelFunction(neq);
    test_Kronrod_random(neq);
    test_Borges(getMaxNeqCut(neq));
    test_MultipleOrthogonality_Disjoin(getMaxNeqCut(neq));
    if(false) test_MultipleOrthogonality(neq); // may fail, no guarantee of roots.
    test_ECOND(neq); // must be called the last, clobber moments

    System.err.println(name+":test_Calculations(): test OK with N points="+tsize+
		       " neq="+neq+" eps="+eps);
}
void runTests(final int neq_max)
{
    System.err.println(name+":TestOrthogonalPolynomialsBasisFunctionsCalculatable:starting tests; ver=$Id: TestOrthogonalPolynomialsBasisFunctionsCalculatable.java,v 1.266 2024/04/15 06:15:34 mal Exp $");
    final long t0=System.currentTimeMillis();
    for(int neq=1;neq<=neq_max;neq++)
    {
	for(int i=0;i<118;i++)
	{
	    test_Calculations(2210+r.nextInt(2210),neq);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println(name+":TestOrthogonalPolynomialsBasisFunctionsCalculatable: tests OK done in "+(t1-t0)+" milliseconds. neq_max="+neq_max+" ver=$Id: TestOrthogonalPolynomialsBasisFunctionsCalculatable.java,v 1.266 2024/04/15 06:15:34 mal Exp $");
}
} // END OF class TestOrthogonalPolynomialsBasisFunctionsCalculatable
