package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2019
 * This file is under GPL version 3.
 *
 */

/** A class for {@link RadonNikodymSpectralModel} unit test.
  * @version $Id: TestRN.java,v 1.145 2022/11/19 10:48:52 mal Exp $
  * @author Vladislav Malyshkin mal@gromco.com
 */
public class TestRN {

    final java.util.Random r;
    final int size;
    final double eps;
    TestRN(final java.util.Random r,final int size,final double eps){
	this.r=r;
	this.size=size;
	this.eps=eps;
    }

    void testNonDegenerated1(final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method){
	final DataReadObservationVectorXF
	    ro=TestObservationSequenceFromSampleVectorXF.getNoNDegenerated(r,size);
	final RadonNikodymSpectralModel rn=regularization_method.apply(ro).getRadonNikodymSpectralModel();
	// match dr & rn
	matchDataRegualizedAndRN(regularization_method,regularization_method.apply(ro),rn,ro);
	// regular test
	testChristoffel(regularization_method,"Regular",rn,ro);
	testSampled("Regular",rn,ro);
	testXorigExpansionInPsik(rn,ro); // only full basis matches

	// Test non-diagonal XX
	final RadonNikodymSpectralModel rnxx=ro.getRadonNikodymSpectralModel_WithoutRegularization();
	testChristoffel(regularization_method,"Non-Diag XX",rnxx,ro);
	testSampled("Non-Diag XX",rnxx,ro);
	testXorigExpansionInPsik(rnxx,ro); // only full basis matches
	matchTwoRN(rn,ro,rnxx,ro);

	// Reduced basis test
	final RadonNikodymSpectralModel rnred=rn.reduceBasisSize(rn.nX/2+1);
	testSampled("Clustered",rnred,ro);
	matchTwoRN(rnred,ro,rnxx.reduceBasisSize(rn.nX/2+1),ro);
	// apply twice
	testSampled("Clustered2",rnred.reduceBasisSize(rnred.nX/2+1),ro);

	// test getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis
	final RadonNikodymSpectralModel dr_cfdiagF=regularization_method.apply(ro).getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis(ro),
	    rn_cfdiagF=rn.getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis(ro),
	    rnxx_cfdiagF=rnxx.getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis(ro);
	testFDiagInChristoffelFunctionBasis("Regular from DR",dr_cfdiagF,ro);
	testFDiagInChristoffelFunctionBasis("Regular",rn_cfdiagF,ro);
	testFDiagInChristoffelFunctionBasis("Non-Diag XX",rnxx_cfdiagF,ro);
	matchTwoRN(dr_cfdiagF,ro,rn_cfdiagF,ro);
	matchTwoRN(rn_cfdiagF,ro,rnxx_cfdiagF,ro);

    }

    void matchTwoRN(final RadonNikodymSpectralModel rnro,final DataReadObservationVectorXF ro,
		    final RadonNikodymSpectralModel rndro,final DataReadObservationVectorXF dro){
	UnitTests.checkDoubleRel(TMatr.sumElements(rnro.w),"sum(wLeb)",ro.data.stream().mapToDouble(b->b.weight).sum(),"sum(w)",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(rndro.w),"sum(wLeb)",dro.data.stream().mapToDouble(b->b.weight).sum(),"sum(w)",eps);
	System.err.println("sum(rndro.w)="+TMatr.sumElements(rndro.w)+" sum(rnro.w)="+TMatr.sumElements(rnro.w));
	UnitTests.checkDoubleArrayRel1(rnro.f,"rnro.f",rndro.f,"rndro.f",eps);
	UnitTests.checkDoubleArrayRel1(rnro.w,"rnro.w",rndro.w,"rndro.w",eps);
	UnitTests.checkDoubleArrayRel1(rnro.getChristoffelWeights(ro),"rnro.getChristoffelWeights(ro)",rndro.getChristoffelWeights(dro),"rndro.getChristoffelWeights(dro)",eps);

	// evaluate every point
	matchRNValues(rnro,ro,rndro,dro);

	System.err.println("matchTwoRN(): two RadonNikodymSpectralModel matched OK eps="+eps);
    }

    void matchRNValues(final RadonNikodymSpectralModel rnro,final DataReadObservationVectorXF ro,
		       final RadonNikodymSpectralModel rndro,final DataReadObservationVectorXF dro){
	UnitTests.checkDoubleRel(TMatr.sumElements(rnro.w),"sum(wLeb)",ro.data.stream().mapToDouble(b->b.weight).sum(),"sum(w)",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(rndro.w),"sum(wLeb)",dro.data.stream().mapToDouble(b->b.weight).sum(),"sum(w)",eps);

	if(ro.data.size()!=dro.data.size()) throw new RuntimeException();

	// evaluate every point
	for(int l=0,M=ro.data.size();l<M;l++)
	{
	    final RadonNikodymSpectralModel.RNPointEvaluation
		ernro=rnro.getRNatXoriginal(ro.data.get(l).x),
		edrnro=rndro.getRNatXoriginal(dro.data.get(l).x);
	    // match Radon--Nikodym derivatives (of two kind) for every observation
	    UnitTests.checkDoubleRel(ernro.getRN(),"ernro.getRN()",edrnro.getRN(),"edrnro.getRN()",eps);
	    UnitTests.checkDoubleRel(ernro.getRNW(),"ernro.getRNW()",edrnro.getRNW(),"edrnro.getRNW()",eps);
	    // match least squares
	    UnitTests.checkDoubleRel(ernro.getLS(),"ernro.getLS()",edrnro.getLS(),"edrnro.getLS()",eps);
	    // Christoffel function match
	    UnitTests.checkDoubleRel(ernro.getChristoffelOatX(),"ernro.getChristoffelOatX()",edrnro.getChristoffelOatX(),"edrnro.getChristoffelOatX()",eps);
	    UnitTests.checkDoubleArrayRel1(ernro.getPosteriorDistributionWeights(),"ernro.getPosteriorDistributionWeights()",edrnro.getPosteriorDistributionWeights(),"edrnro.getPosteriorDistributionWeights()",eps);
	    UnitTests.checkDoubleArrayRel1(ernro.getPriorDistributionWeights(),"ernro.getPriorDistributionWeights()",edrnro.getPriorDistributionWeights(),"edrnro.getPriorDistributionWeights()",eps);
	}
	System.err.println("matchRNValues(): two RadonNikodymSpectralModel matched OK eps="+eps);
    }

    /** Test the xorig expansion in psik. The match possible only for psik being the full basis.
     */
    void testXorigExpansionInPsik(final RadonNikodymSpectralModel rn,final DataReadObservationVectorXF ro){
	final double [][] XorigInPsikExpansion=rn.getAllXorigProjectedToPsik(ro);

	// evaluate every point
	for(ObservationVectorXF o:ro.data)
	{
	    final double [] xorig=o.x,psik=rn.getRNatXoriginal(xorig).getPsikAtX();
	    for(int j=0;j<xorig.length;j++) {
		final double xorig_j=TMatr.mulVV(XorigInPsikExpansion[j],psik);
		// if psik form the full basis they will match
		UnitTests.checkDoubleRel(xorig_j,"xorig_j",xorig[j],"xorig[j]",eps);
	    }
	}
	System.err.println("testXorigExpansionInPsik(): xorig in psik matched OK eps="+eps);
    }


    void matchTwoDataSets(
	final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method,
	final DataReadObservationVectorXF ro,
	final DataReadObservationVectorXF dro){
	matchTwoRN(regularization_method.apply(ro).getRadonNikodymSpectralModel(),ro,
		   regularization_method.apply(dro).getRadonNikodymSpectralModel(),dro);

    }

    void testDegenerated1(final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method){
	final var ro=TestObservationSequenceFromSampleVectorXF.getDegenerated_VectorX(17,r,size);
	final RadonNikodymSpectralModel rnts=regularization_method.apply(ro.ndeg).getRadonNikodymSpectralModel(),
	    rndts=regularization_method.apply(ro.deg).getRadonNikodymSpectralModel();
	testChristoffel(regularization_method,"Non-Degenerated",rnts,ro.ndeg);
	testSampled("Non-Degenerated",rnts,ro.ndeg);
	testChristoffel(regularization_method,"Degenerated",rndts,ro.deg);
	testSampled("Degenerated",rndts,ro.deg);
	matchTwoRN(rnts,ro.ndeg,rndts,ro.deg);
	// match dr & rn (degenerated case).
	matchDataRegualizedAndRN(regularization_method,regularization_method.apply(ro.deg),rndts,ro.deg);
	final int nXnew=Math.min(rnts.nX,4);
	matchTwoRN(rnts.reduceBasisSize(nXnew),ro.ndeg,rndts.reduceBasisSize(nXnew),ro.deg);
	testXorigExpansionInPsik(rndts,ro.deg); // only full basis matches


	// test getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis
	final RadonNikodymSpectralModel rnts_cfdiagF=rnts.getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis(ro.ndeg),
	    rndts_cfdiagF=rndts.getRadonNikodymSpectralModelWithFassumedDiagonalInChristoffelFunctionBasis(ro.deg);
	testFDiagInChristoffelFunctionBasis("Non-Degenerated",rnts_cfdiagF,ro.ndeg);
	testFDiagInChristoffelFunctionBasis("Degenerated",rndts_cfdiagF,ro.deg);
	matchTwoRN(rnts_cfdiagF,ro.ndeg,rndts_cfdiagF,ro.deg);

	System.err.println("testDegenerated1(): OK eps="+eps);
    }

    void testRegularizations(final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method1,
			     final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method2){
	final var ro=TestObservationSequenceFromSampleVectorXF.getDegenerated_VectorX(17,r,size);

	final RadonNikodymSpectralModel
	    rnts1=regularization_method1.apply(ro.ndeg).getRadonNikodymSpectralModel(),
	    rnts2=regularization_method2.apply(ro.ndeg).getRadonNikodymSpectralModel(),
	    rndts1=regularization_method1.apply(ro.deg).getRadonNikodymSpectralModel(),
	    rndts2=regularization_method2.apply(ro.deg).getRadonNikodymSpectralModel();
	testSampled("Non-Degenerated",rnts1,ro.ndeg);
	testSampled("Non-Degenerated",rnts2,ro.ndeg);
	testSampled("Degenerated",rndts1,ro.deg);
	testSampled("Degenerated",rndts2,ro.deg);
	matchTwoRN(rnts1,ro.ndeg,rnts2,ro.ndeg);
	matchTwoRN(rndts1,ro.deg,rndts2,ro.deg);
	matchTwoRN(rnts1,ro.ndeg,rndts2,ro.deg);
	matchTwoRN(rnts1,ro.ndeg,rndts1,ro.deg);
	matchTwoRN(rnts2,ro.ndeg,rndts1,ro.deg);
	matchTwoRN(rnts2,ro.ndeg,rndts1,ro.deg);

	System.err.println("testRegularizations(): OK eps="+eps);
    }

    double evaluateStringArray(String xstr,double [] x){
	// a quick and dirty way to evaluate a string: use javascript
	final String s="var x="+java.util.Arrays.toString(x)+";"+xstr+";";
	try{
	    javax.script.ScriptEngine engine = new javax.script.ScriptEngineManager().getEngineByName("JavaScript");
	    return java.lang.Double.parseDouble(engine.eval(s).toString());
	} catch(Exception e){
	    throw new RuntimeException("failed to parse s="+s);
	}
    }
    void testAttributesTextualRepresentation(final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method){
	final var rod=TestObservationSequenceFromSampleVectorXF.getDegenerated_VectorX(17,r,size);
	for(var ro:java.util.List.of(rod.ndeg,rod.deg)){
	    final var dr=regularization_method.apply(ro);
	    final java.util.List<String> allXdescr=
		java.util.stream.IntStream.range(0, dr.nX).mapToObj(dr::getRegulaizedBasisXDescription).toList();
	    //System.err.println("allXdescr="+allXdescr);
	    for(ObservationVectorXF o:ro.data)
	    {
		final double [] X=dr.getRegularizedBasisXFromOriginalX(o.x),
		 evalXdescr=allXdescr.stream().mapToDouble((xstr)-> evaluateStringArray(xstr,o.x)).toArray();
		UnitTests.checkDoubleArrayRel1(X,"X",evalXdescr,"evalXdescr",eps);
	    }
	}
	System.err.println("testAttributesTextualRepresentation(): OK eps="+eps);
    }


    void testChristoffel(final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method,
			 final String testname,
			 final RadonNikodymSpectralModel rn,final DataReadObservationVectorXF ro){
	// test ChristoffelFunction
	final DataReadObservationVectorXF roK=ro.getReplacedData((o)->new ObservationVectorXF(o.label,o.x,1/rn.getRNatXoriginal(o.x).getChristoffelOatX(),o.weight),h->h);
	final RadonNikodymSpectralModel rnK=rn.getRadonNikodymSpectralModelWithChristoffelFunctionAsF(ro);
	testSampled("Christoffel "+testname,rnK,roK);
	matchTwoRN(rn.getRadonNikodymSpectralModelWithChristoffelFunctionAsF(ro),ro,regularization_method.apply(roK).getRadonNikodymSpectralModel(),roK);

	System.err.println("testChristoffel(): test "+testname+" OK noriginal="+rn.noriginal+" rn.nX="+rn.nX+" eps="+eps);
    }

    void testSampled(final String testname,final RadonNikodymSpectralModel rn,final DataReadObservationVectorXF ro){
	System.err.println("=======testSampled:"+testname+" TEST START ===========================");
	System.err.println("totalweight="+TMatr.sumElements(rn.w)+" noriginal="+rn.noriginal+" rn.nX="+rn.nX+" sample points="+ro.data.size()+" eps="+eps);

	final XXMatrs rs=ro.getXXandXfX(rn.convertxFromOriginalToX);
	if(false){
	    System.err.println("rs.XX="+java.util.Arrays.toString(rs.XX));
	    System.err.println("rn.XX="+java.util.Arrays.toString(rn.XX));
	    System.err.println("rs.XfX="+java.util.Arrays.toString(rs.XfX));
	    System.err.println("rn.XfX="+java.util.Arrays.toString(rn.XfX));
	}
	// check the matrix
	UnitTests.checkDoubleArrayRel1(rs.XX,"rs.XX",rn.XX,"rn.XX",eps);
	UnitTests.checkDoubleArrayRel1(rs.XfX,"rs.XfX",rn.XfX,"rn.XfX",eps);
	final double [] xorig_average=ro.getFun1Fun2Averaged(o -> new double []{1},o->o.x).XX,
	    fxmxorig_average=ro.getFun1Fun2Averaged(o -> new double []{o.f},o->TMatr.sumAcB(o.x,-1,ro.datastat.xorig_average)).XX,
	    fxmxorig_average2=ro.getFun1Fun2Averaged(o -> new double []{1},o->TMatr.sumAcB(o.x,-1,ro.datastat.xorig_average)).XfX;
	UnitTests.checkDoubleArrayRel1(TMatr.getVectorScaled(xorig_average,1/ro.datastat.totalweight),"xorig_average",ro.datastat.xorig_average,"ro.datastat.xorig_average",eps);
	UnitTests.checkDoubleArrayRel1(TMatr.getVectorScaled(fxmxorig_average,1/ro.datastat.totalweight),"fxmxorig_average",ro.datastat.fxmxorig_average,"ro.datastat.fxmxorig_average",eps);
	UnitTests.checkDoubleArrayRel1(TMatr.getVectorScaled(fxmxorig_average2,1/ro.datastat.totalweight),"fxmxorig_average2",ro.datastat.fxmxorig_average,"ro.datastat.fxmxorig_average",eps);

	// check averages
	final XXMatrs sampleaverage=
	    TestObservationSequenceFromSampleVectorXF.getPsiAverageFromSample(rn.convertxFromOriginalToX,rn.ev,ro);
	UnitTests.checkDoubleArrayRel1(sampleaverage.XX,"sampleaverage.XX",rn.psiaverage,"rn.psiaverage",eps);
	UnitTests.checkDoubleArrayRel1(sampleaverage.XfX,"sampleaverage.XfX",TMatr.getVAB(rn.psiaverage,rn.f),"TMatr.getVAB(rn.psiaverage,rn.f)",eps);

	// check eigenfunctions orthogonality
	final XXMatrs sampleev=ro.getXXandXfX(rn::getPsikAtXorig);
	UnitTests.checkDoubleArrayRel1(sampleev.XX,"sampleev.XX",TMatr.getUnitMatrix(rn.nX),"1",eps);
	UnitTests.checkDoubleArrayRel1(sampleev.XfX,"sampleev.XfX",TMatr.getDiagMatrix(rn.f),"f",eps);

	// check Lebesgue quadrature weights
	final double Wtotal=ro.data.stream().mapToDouble(b->b.weight).sum();
	System.err.println("f="+java.util.Arrays.toString(rn.f)+" w="+java.util.Arrays.toString(rn.w)+" sum(w)="+TMatr.sumElements(rn.w)+" Wtotal="+Wtotal);
	UnitTests.checkDoubleRel(TMatr.sumElements(rn.w),"sum(rn.w)",Wtotal,"Wtotal",eps);
	UnitTests.checkDoubleRel(TMatr.sumElements(rn.getChristoffelWeights(ro)),"sum(wK)",Wtotal,"Wtotal",eps);

	// test PCA expansion
	final double [] std_sample=TestObservationSequenceFromSampleVectorXF.getStdevExpansion(rn,ro),
	    fm=TestObservationSequenceFromSampleVectorXF.getfMoments(ro);
	final double stdev2_sample=std_sample[2],average_sample=std_sample[1],weight_sample=std_sample[0],
	    stddev2_PCA_expansion=fm[2]-TMatr.mulV2V(rn.f,rn.w),fa=fm[1]/fm[0],fev_shifted_to_aver[]=new double [rn.nX];
	for(int k=0;k<rn.nX;k++) fev_shifted_to_aver[k]=rn.f[k]-fa;
	final double stddev2_PCA_expansion_=(fm[2]-fa*fa*fm[0])-TMatr.mulV2V(fev_shifted_to_aver,rn.w);
	UnitTests.checkDoubleRel(weight_sample,"weight_sample",TMatr.sumElements(rn.w),"sum(rn.w)",eps);
	UnitTests.checkDoubleRel(average_sample,"average_sample",0,"0",eps);
	UnitTests.checkDoubleRel(stdev2_sample,"stdev2_sample",stddev2_PCA_expansion,"stddev2_PCA_expansion",eps);
	UnitTests.checkDoubleRel(0,"0",TMatr.mulVV(fev_shifted_to_aver,rn.w),"TMatr.mulVV(fev_shifted_to_aver,rn.w)",eps);
	UnitTests.checkDoubleRel(stdev2_sample,"stdev2_sample",stddev2_PCA_expansion_,"stddev2_PCA_expansion_",eps);

	System.err.println("testSampled(): test OK noriginal="+rn.noriginal+" rn.nX="+rn.nX+" sample points="+ro.data.size()+" eps="+eps);
	System.err.println("=======testSampled:"+testname+" TEST END OK ===========================");
    }

    void testFDiagInChristoffelFunctionBasis(final String testname,final RadonNikodymSpectralModel rn,final DataReadObservationVectorXF ro){
	System.err.println("=======testFDiagInChristoffelFunctionBasis:"+testname+" TEST START ===========================");
	System.err.println("totalweight="+TMatr.sumElements(rn.w)+" noriginal="+rn.noriginal+" rn.nX="+rn.nX+" sample points="+ro.data.size()+" eps="+eps);

	// direct sample in the basis of rn eigenvectors
	final XXMatrs rs=ro.getXXandXfX(rn::getPsikAtXorig);
	// check the matrix
	UnitTests.checkDoubleArrayRel1(rs.XX,"rs.XX",TMatr.getUnitMatrix(rn.nX),"Unit",eps); // Unit
	// only diag elements match for f
	final double [] rn_XfX_in_basis=EVXData.convertMatrFromQtoB_Basis(rn.nX,rn.XfX,rn.ev);
	UnitTests.checkDoubleArrayRel1(TMatr.getMatrixDiagElements(rs.dim2,rs.XfX),"diag rs.XfX",TMatr.getMatrixDiagElements(rn.nX,rn_XfX_in_basis),"diag in basis: rn.XfX",eps);
	// off-diag elements must be 0
	for(int j=0;j<rn.nX;j++){
	    for(int k=0;k<rn.nX;k++){
		if(j!=k){
		    UnitTests.checkDoubleRel(rn_XfX_in_basis[j*rn.nX+k],"off-diag",0,"0",eps);
		} else {
		    UnitTests.checkDoubleRel(rn_XfX_in_basis[j*rn.nX+k],"diag rn",rs.XfX[j*rn.nX+k],"diag sample",eps);
		}
	    }
	}

	System.err.println("testFDiagInChristoffelFunctionBasis(): test OK noriginal="+rn.noriginal+" rn.nX="+rn.nX+" sample points="+ro.data.size()+" eps="+eps);
	System.err.println("=======testFDiagInChristoffelFunctionBasis:"+testname+" TEST END OK ===========================");
    }

    void matchDataRegualizedAndRN(final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method,
				  final RegularizedVectorXF dr,
				  final RadonNikodymSpectralModel rn,
				  final DataReadObservationVectorXF ro){
	// evaluate every point
	for(ObservationVectorXF o:ro.data){
	    final RNPointAccess orn=rn.getRNatXoriginal(o.x),odr=dr.getRNatXoriginal(o.x);
	    // match Radon--Nikodym derivative for every observation
	    UnitTests.checkDoubleRel(odr.getRN(),"odr.getRN()",orn.getRN(),"orn.getRN()",eps);
	    // match least squares
	    UnitTests.checkDoubleRel(odr.getLS(),"odr.getLS()",orn.getLS(),"orn.getLS()",eps);
	    // Christoffel function match
	    UnitTests.checkDoubleRel(odr.getChristoffelOatX(),"odr.getChristoffelOatX()",orn.getChristoffelOatX(),"orn.getChristoffelOatX()",eps);
	}

	// Christoffel function match
	final double [] XKX_fromDR_in_RN_basis=EVXData.convertMatrFromQtoB_Basis(rn.nX,dr.getChristoffelFunctionDensityMatrixInXBasis(ro).XX,rn.ev),
	    XKX_in_RN_basis=rn.getChristoffelFunctionDensityMatrixInEvBasis(ro).XX;
	UnitTests.checkDoubleArrayRel1(XKX_fromDR_in_RN_basis,"XKX_fromDR_in_RN_basis",XKX_in_RN_basis,"XKX_in_RN_basis",eps);
	UnitTests.checkDoubleRel(TMatr.getSpurM(rn.nX,XKX_in_RN_basis),"Spur(rhoK)",ro.data.stream().mapToDouble(b->b.weight).sum(),"sum(w)",eps);

	matchTwoRN(rn.getRadonNikodymSpectralModelWithChristoffelFunctionAsF(ro),ro,
		   dr.getRadonNikodymSpectralModelWithChristoffelFunctionAsF(ro),ro);
	final DataReadObservationVectorXF roK=ro.getReplacedData((o)->new ObservationVectorXF(o.label,o.x,1/dr.getRNatXoriginal(o.x).getChristoffelOatX(),o.weight),h->h);
	matchTwoRN(rn.getRadonNikodymSpectralModelWithChristoffelFunctionAsF(ro),ro,
		   regularization_method.apply(roK).getRadonNikodymSpectralModel(),roK);

	System.err.println("matchDataRegualizedAndRN(): RegularizedVectorXF and RadonNikodymSpectralModel matched OK on "+ro.data.size()+" sample points eps="+eps);
    }

    void testAttributesProducts(final java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF> regularization_method){
	final var ro=TestObservationSequenceFromSampleVectorXF.getDegenerated_VectorX(7,r,size);
	// Match obtained from two bases
	final var dr=regularization_method.apply(ro.ndeg);
	final RadonNikodymSpectralModel rn=dr.getRadonNikodymSpectralModel();
	// match d=1 case
	matchTwoDataSets(regularization_method,ro.ndeg,dr.getAttribsProductsMultiIndexed(1,ro.ndeg));
	matchTwoDataSets(regularization_method,ro.ndeg,rn.getAttribsProductsMultiIndexed(1,ro.ndeg));

	// try d>1
	int d=3;
	final DataReadObservationVectorXF ro_fromDR=dr.getAttribsProductsMultiIndexed(d,ro.ndeg),
	    ro_fromRN=rn.getAttribsProductsMultiIndexed(d,ro.ndeg);
	matchTwoDataSets(regularization_method,ro_fromRN,ro_fromDR);
	matchTwoDataSets(regularization_method,ro_fromRN,ro.ndeg.getAttribsProductsMultiIndexed_WithoutRegularization(d));

	// match degenerated: regularize, then product.
	matchTwoDataSets(regularization_method,ro_fromRN,regularization_method.apply(ro.deg).getAttribsProductsMultiIndexed(d,ro.deg));
	matchTwoDataSets(regularization_method,ro_fromRN,regularization_method.apply(ro.deg).getRadonNikodymSpectralModel().getAttribsProductsMultiIndexed(d,ro.deg));
	// match degenerated: product, then regularized.
	matchTwoDataSets(regularization_method,ro_fromRN,ro.deg.getAttribsProductsMultiIndexed_WithoutRegularization(d));

	// match Chebyshev. here we may have an instability.
	d=2; // for d>2 we get a numerical instability with Chebyshev
	final BasisFunctionsCalculatable B=new Chebyshev();
	final DataReadObservationVectorXF ro_d=rn.getAttribsProductsMultiIndexed(d,ro.ndeg);

	matchTwoDataSets(regularization_method,ro_d,AttributesProductsMultiIndexed.getAttribsProductsMultiIndexed(
			     d,ro.ndeg,rn::getPsikAtXorig,(order,x)->B.calculate(order,x)));
	matchTwoDataSets(regularization_method,ro_d,AttributesProductsMultiIndexed.getAttribsProductsMultiIndexed(
			     d,ro.ndeg,dr::getRegularizedBasisXFromOriginalX));
	matchTwoDataSets(regularization_method,ro_d,AttributesProductsMultiIndexed.getAttribsProductsMultiIndexed(
			     d,ro.ndeg,dr::getRegularizedBasisXFromOriginalX,(order,x)->B.calculate(order,x)));

	System.err.println("testAttributesProducts(): matched OK eps="+eps);
    }

    static void runAllTests(final java.util.Random r,final double eps){
	System.err.println("TestRN:starting tests; ver=$Id: TestRN.java,v 1.145 2022/11/19 10:48:52 mal Exp $");
	final long t0=System.currentTimeMillis();

	for(int stest=0;stest<10;stest++){
	    final TestRN t=new TestRN(r,1300+r.nextInt(1000),eps);

	    // test two methods of regularization
	    for(var regularization_method:
		    java.util.List.<java.util.function.Function<DataReadObservationVectorXF,RegularizedVectorXF>>of(
			// May fail, depends on EV engine quality.
			// Works OK with EVSolver_LapackOLD
			// sometimes fail with EVSolver_Lapack310 when USE_dsyevr=false
			(x)->x.getDataRegularized_EV(),
			// This is OK
			(x)->x.getDataRegularized_LIN()
			    )){
		t.testNonDegenerated1(regularization_method);
		t.testDegenerated1(regularization_method);
		t.testAttributesProducts(regularization_method);
		if(false) // requires Nashorn engine and too slow.
		    t.testAttributesTextualRepresentation(regularization_method);
		t.testRegularizations((x)->x.getDataRegularized_EV(),(x)->x.getDataRegularized_LIN());

	    }
	}

	final long t1=System.currentTimeMillis();
	System.err.println("TestRN: tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: TestRN.java,v 1.145 2022/11/19 10:48:52 mal Exp $");
    }

    public static void main(String [] args){

	final double eps=1e-5;
	// force a selftest
	RadonNikodymSpectralModel.EPS_selftest=eps;
	// run the selftest
	runAllTests(new java.util.Random(123),eps);
	// disable selftest
	RadonNikodymSpectralModel.EPS_selftest=0;

    }
}
