package com.polytechnik.utils;
/*
 * (C) Vladislav Malyshkin 2013
 * This file is under GPL version 3.
 *
 */

/** Volatility calculation.
 *  Unit test is in VkQVM, BkQVM and XkVM.XkQVMTest.test_Matrices()
 *  @version $Id: VolMatrix.java,v 1.24 2022/11/19 11:08:58 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
@FunctionalInterface public interface VolMatrix {

/** Calculates D(psi).
 */
double [] getDPsi(final double [] psi);

/** Volatility matrix. D^T QQ D
 */
default double [] getVolMatrix(final int n,final double [] QQ)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final double [] DQQD=new double [n*n];
    final double [][] bb=new double [n][];
    for(int k=0;k<n;k++)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] dpsik=bb[k]=getDPsi(pp);
	if(dpsik.length>pp.length) throw new RuntimeException("invalid getDPsi");
	TMatr.multiply_MA_x_MB(pp,pp.length,QQ,ldQQ,0,dpsik.length,dpsik,1,1,0);
	for(int l=0;l<=k;l++)
	{
	    DQQD[n*k+l]=DQQD[n*l+k]=TMatr.mulVVn1(bb[l],pp);
	}
    }
    return DQQD;
}

/** Differentiated matrix. QQ D + D^T QQ^T.
 */
default double [] getDiffMatrix(final int n,final double [] QQ)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final double [] QQD=new double [n*n];
    for(int k=n;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] dpsik=getDPsi(pp);
	TMatr.multiply_MA_x_MB(QQD,n,QQ,ldQQ,0,dpsik.length,dpsik,1,1,0);
	if(k>0) System.arraycopy(QQD,0,QQD,k*n,n);
    }
    return TMatr.scaleVectorInPlace(TMatr.selectHermitianPartInPlace(n,QQD),2); // factor of 2 is here
}

/** D2 Differentiated matrix. QQ DD + DD^T QQ^T + D^T QQ D + D^T QQ^T D
 */
default double [] getDiff2Matrix(final int n,final double [] QQ)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final double [] QQD2=new double [n*n],pps=new double [n];
    final double [][] bb=new double [n][];
    for(int k=n;--k>=0;){final double [] ppk=new double [k+1];ppk[k]=1;bb[k]=getDPsi(ppk);}
    for(int k=n;--k>=0;)
    {
	final double [] dpsik=bb[k],ddpsik=getDPsi(dpsik);
	TMatr.multiply_MA_x_MB(QQD2,n,QQ,ldQQ,0,ddpsik.length,ddpsik,1,1,0);
	TMatr.multiply_MA_x_MB(pps,n,QQ,ldQQ,0,dpsik.length,dpsik,1,1,0);
	for(int l=n,kn=k*n;--l>=0;){QQD2[kn+l]=QQD2[l]+TMatr.mulVVn1(bb[l],pps);}
    }
    return TMatr.scaleVectorInPlace(TMatr.selectHermitianPartInPlace(n,QQD2),2); // factor of 2 is here
}

/** A part of D2: QQ DD + DD^T QQ^T.
 */
default double [] getDiff2PartMatrix(final int n,final double [] QQ)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final double [] QQPD2=new double [n*n];
    for(int k=n;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] ddpsik=getDPsi(getDPsi(pp));
	TMatr.multiply_MA_x_MB(QQPD2,n,QQ,ldQQ,0,ddpsik.length,ddpsik,1,1,0);
	if(k>0) System.arraycopy(QQPD2,0,QQPD2,k*n,n);
    }
    return TMatr.scaleVectorInPlace(TMatr.selectHermitianPartInPlace(n,QQPD2),2); // factor of 2 is here
}

/** D- matrix. If Dpsi=getDPsi(psi), then Dpsi_k=\sum_{l}D_{kl}psi_l.
 *
 */
default double [] getDMatrix(final int n)
{
    final double [] D=new double [n*n];
    for(int k=n;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] dpsik=getDPsi(pp);
	for(int l=dpsik.length;--l>=0;) D[n*l+k]=dpsik[l];
    }
    return D;
}

/** Hermitian QQ D + D^T QQ^T.
 */
public static double [] getHAB(final int n,final double [] QQ,final double [] D)
{
    if(QQ.length!=n*n||D.length!=QQ.length) throw new IllegalArgumentException();
    final double [] HAB=TMatr.mulMM(n,QQ,D);
    return TMatr.scaleVectorInPlace(TMatr.selectHermitianPartInPlace(n,HAB),2); // factor of 2 is here
}

/** Calculates {@literal <r*Q_i Q_j>}.
 * @param QQ, the matrix n*n
 * @return rQQ, the matrix (n-1)*(n-1)
 */
public static double [] getRadauMatrix(final int n,final double [] QQ,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final int nd=n-1;
    final double [] rQQ=new double [nd*nd];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] rpp=M.getRadauMultiplied(pp);
	TMatr.multiply_MA_x_MB(rQQ,nd,QQ,ldQQ,0,rpp.length,rpp,1,1,0);
	if(k>0) System.arraycopy(rQQ,0,rQQ,k*nd,nd);
    }
    return rQQ;
}

/** Calculates {@literal <D(r*Q_i Q_j)>}.
 * @param QQ, the matrix n*n
 * @return D(rQQ), the matrix (n-1)*(n-1)
 */
default double [] getDRadauMatrix(final int n,final double [] QQ,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final int nd=n-1;
    final double [] rQQ=new double [nd*nd],tmp1=new double [nd],tmp2=new double [nd];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] Drpp=getDPsi(M.getRadauMultiplied(pp)),rDpp=M.getRadauMultiplied(getDPsi(pp));
	TMatr.multiply_MA_x_MB(tmp1,nd,QQ,ldQQ,0,Drpp.length,Drpp,1,1,0);
	TMatr.multiply_MA_x_MB(tmp2,1,rDpp,n,0,rDpp.length,QQ,nd,ldQQ,0);
	for(int j=nd;--j>=0;) {rQQ[k*nd+j]+=tmp1[j];rQQ[j*nd+k]+=tmp2[j];}
    }
    return rQQ;
}

/** Calculates {@literal <r^2*Q_i Q_j>}.
 * @param QQ, the matrix n*n
 * @return r^2*QQ, the matrix (n-1)*(n-1)
 */
public static double [] getRadau2Matrix(final int n,final double [] QQ,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final int nd=n-1;
    final double [] trans=new double [n*nd],rQQ=new double [nd*nd],tmp=new double [n*nd];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] rpp=M.getRadauMultiplied(pp);
	if(rpp.length!=pp.length+1) throw new IllegalArgumentException();
	System.arraycopy(rpp,0,trans,k*n,rpp.length);
    }
    // calculate trans QQ trans^T, trans: nd x n.
    TMatr.multiply_MA_x_MBTrans(tmp,n,QQ,ldQQ,0,n,trans,nd,n,0); // tmp: n x nd
    TMatr.multiply_MA_x_MB(rQQ,nd,trans,n,0,n,tmp,nd,nd,0);
    return rQQ;
}

/** Calculates {@literal <D(r^2*Q_i Q_j)>}.
 * @param QQ, the matrix n*n
 * @return D(r^2*QQ), the matrix (n-1)*(n-1)
 */
default double [] getDRadau2Matrix(final int n,final double [] QQ,final OrthogonalPolynomialsBasisFunctionsCalculatable<? extends BasisFunctionsCalculatable> M)
{
    final int ldQQ=n;
    if(QQ.length!=n*n) throw new IllegalArgumentException();
    final int nd=n-1;
    final double [] rQQ=new double [nd*nd],transR=new double [nd*n],transD=new double [nd*n],tmp=new double [nd*n];
    for(int k=nd;--k>=0;)
    {
	final double [] pp=new double [k+1];pp[k]=1;
	final double [] rpp=M.getRadauMultiplied(pp),Drpp=getDPsi(rpp);
	if(rpp.length!=pp.length+1||Drpp.length>n) throw new IllegalArgumentException();
	System.arraycopy(rpp,0,transR,k*n,rpp.length);
	System.arraycopy(Drpp,0,transD,k*n,Drpp.length);
    }
    // transR: nd x n ; transD: nd x n
    // now calculate transR QQ transD^T + transD QQ transR^T
    TMatr.multiply_MA_x_MBTrans(tmp,n,QQ,ldQQ,0,n,transD,nd,n,0); // tmp: n x nd
    TMatr.multiply_MA_x_MB(rQQ,nd,transR,n,0,n,tmp,nd,nd,0);
    if(false){
	// QQ is assumed to be symmetric
	TMatr.scaleVectorInPlace(TMatr.selectHermitianPartInPlace(nd,rQQ),2);
    } else {
	TMatr.multiply_MA_x_MBTrans(tmp,n,QQ,ldQQ,0,n,transR,nd,n,0); // tmp: n x nd
	TMatr.multiply_MA_x_MB(transR,nd,transD,n,0,n,tmp,nd,nd,0);
	TMatr.sumP1(rQQ,transR,nd*nd,1);
    }
    return rQQ;
}


}
