package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2008
 * This file is under GPL version 3.
 *
 */
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.SortArrays;
import com.polytechnik.utils.PolynomialRootsFinder;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.outils.*;
import com.polytechnik.realtimebook.*;

/** Kendall calculations.
 *  @version $Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */

class CalculateKendall {
    static final boolean PRINT_DATA_DEBUG=false;


    /** Result. */
    final long currenttime,vi,v_cons,v_discons,v_zero;
    CalculateKendall(final long currenttime,
		     final long vi,
		     final long v_cons,
		     final long v_discons,
		     final long v_zero)
    {
	this.currenttime=currenttime;
	this.vi=vi;
	this.v_cons=v_cons;
	this.v_discons=v_discons;
	this.v_zero=v_zero;
	if(PRINT_DATA_DEBUG)
	{
	    System.err.println("================="+getClass().getName());
	    System.err.println("N="+this.vi);
	    System.err.println("TAU="+(getRank()));
	    System.err.println("N_PAIRS="+(this.v_cons+this.v_discons+this.v_zero));
	    //System.err.println("SLTAU="+this.sltau);
	    System.err.println("S="+(this.v_cons-this.v_discons));
	    //System.err.println("VARS="+this.var);
	    System.err.println("DENOM="+(Math.sqrt(this.v_cons+this.v_discons)*Math.sqrt(this.v_cons+this.v_discons+this.v_zero)));
	    System.err.println("v_cons="+this.v_cons);
	    System.err.println("v_discons="+this.v_discons);
	    System.err.println("v_zero="+this.v_zero);
	}
    }

    long getN()
    {
	return this.v_cons+this.v_discons+this.v_zero;
    }


    double getRank()
    {
	final long n=getN();
	if(n<=0)
	{
	    return Double.NaN;
	}
	final double denom;
	if(this.v_zero<=0)
	{
	    denom=n;
	}
	else
	{
	    denom=Math.sqrt(n)*Math.sqrt(this.v_cons+this.v_discons);
	}
	return (this.v_cons-this.v_discons)/denom;
    }

    double getRankA()
    {
	final long n=getN();
	if(n<=0)
	{
	    return Double.NaN;
	}
	return (this.v_cons-this.v_discons)/((double)n);
    }

    double getRankAStdev()
    {
	final long n=getN();
	if(n<=0)
	{
	    return Double.NaN;
	}
	final double ra=getRankA();
	return Math.sqrt((this.v_cons+this.v_discons)/((double)n)-ra*ra); // biased
    }


    /**
       Algorithm AS66 Applied Statistics (1973) vol22 no.3
       Evaluates the tail area of the standardised normal curve
       from x to infinity if upper is true or
       from minus infinity to x if upper is false
    */
    private static double alnorm(final double x,final boolean upper)
    {
	final double
	    zero=0.0e0, half=0.5e0,
	    ltone=7.0e0,utzero=18.66e0,
	    con=1.28e0,
	    p=0.398942280444e0,q=0.39990348504e0,r=0.398942280385e0,
	    a1=5.75885480458e0,a2=2.62433121679e0,a3=5.92885724438e0,
	    b1=-29.8213557807e0,b2=48.6959930692e0,
	    c1=-3.8052e-8,c2=3.98064794e-4,c3=-0.151679116635e0,
	    c4=4.8385912808e0,c5=0.742380924027e0,c6=3.99019417011e0,
	    d1=1.00000615302e0,d2=1.98615381364e0,d3=5.29330324926e0,
	    d4=-15.1508972451e0,d5=30.789933034e0;

	final boolean up;
	final double z,ialnorm;
	if(x>=zero)
	{
	    up=upper;
	    z=x;
	}
	else
	{
	    up=!upper;
	    z=-x;
	}

	if(z<=ltone || (up && z<=utzero))
	{
	    final double y=half*z*z;
	    if(z>con)
	    {
		ialnorm=r*Math.exp(-y)/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))));
	    }
	    else
	    {
		ialnorm=half-z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))));
	    }
	}
	else
	{
	    ialnorm=0;
	}
	final double alnorm=up?ialnorm:1.0-ialnorm;
	return alnorm;
    }


    private static int cmpt(final long x1,final long x2)
    {
	return x1<x2?-1:(x1>x2?1:0);
    }
    private static int cmpi(final int x1,final int x2)
    {
	return x1<x2?-1:(x1>x2?1:0);
    }

    private static int cmpp(final double x1,final double x2)
    {
	return x1<x2?-1:(x1>x2?1:0);
    }

    /** Kendal tau for orders of one side.
     */
static CalculateKendall calculateBookKendallRankOneSide(
    final long currenttime,
    final int n_orders,
    final Distribution<BookItem> book)
    {
	final BookItem [] bb=book.getBuffer();
	long vi=0,v_cons=0,v_discons=0,v_zero=0;
	for(int i=0;i<n_orders;i++)
	{
	    final BookItem ti=bb[i];
	    final long vvi=ti.shares;
	    for(int j=i+1;j<n_orders;j++)
	    {
		final BookItem tj=bb[j];
		final int c=BookItem.cmpByTime(ti,tj)*cmpp(ti.price,tj.price);
		if(c>0)
		{
		    v_cons+=vvi*tj.shares;
		}
		else if(c<0)
		{
		    v_discons+=vvi*tj.shares;
		}
		else
		{
		    // time scale is nanoseconds, asume there is no degeneracy in time,
		    // only in prices
		    v_zero+=vvi*tj.shares;
		}
	    }
	    vi+=vvi;
	}


	return new CalculateKendall(currenttime,
				    vi,
				    v_cons,
				    v_discons,
				    v_zero);
    }

    static void check_calculateBookKendallRankOneSide(
	 final long currenttime,
	 final int n_orders,
	 final Distribution<BookItem> book,
	 final CalculateKendall r,
	 final Distribution<BookItem> tmpordersbuf)
    {
	tmpordersbuf.clearAll();
	final BookItem [] bb=tmpordersbuf.getBuffer();
	System.arraycopy(book.getBuffer(),0,bb,0,n_orders);

	// the buffer is considered poisoned after this.
	com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(bb,0,n_orders-1,CompareItems.cmp_by_time);
	long vi=0,v_cons=0,v_discons=0,v_zero=0,prevt=Long.MIN_VALUE;
	for(int i=0;i<n_orders;i++)
	{
	    final BookItem ti=bb[i];
	    if(ti.time<prevt)
	    {
		throw new RuntimeException("Ordering is broken");
	    }
	    prevt=ti.time;
	    final long vvi=ti.shares;
	    final double pi=ti.price;
	    for(int j=i+1;j<n_orders;j++)
	    {
		final BookItem tj=bb[j];
		final long vs=vvi*tj.shares;
		final double pj=tj.price;
		if(pj>pi)
		{
		    v_cons+=vs;
		}
		else if(pj<pi)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
	    }
	    vi+=vvi;
	}
	checkKendalls("Results mismatch:",
		      r,
		      new CalculateKendall(currenttime,
					   vi,
					   v_cons,
					   v_discons,
					   v_zero));
	tmpordersbuf.setInvalidState();
    }


    /** Kendal tau for trades in ptr_t_range_cut to tsize
     *  interval. Trades origin version.
     */
static CalculateKendall [] calculateTradesKendallRankTOrigDiff(
    final long currenttime,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	long vi=0,v_cons_buy=0,v_discons_buy=0,v_zero_buy=0,v_cons_sell=0,v_discons_sell=0,v_zero_sell=0,v_cons_bs=0,v_discons_bs=0,v_zero_bs=0,v_cons_self=0,v_discons_self=0,v_zero_self=0,v_cons_dir=0,v_discons_dir=0,v_zero_dir=0,v_cons_cd=0,v_discons_cd=0,v_zero_cd=0;
	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares,ti_time=ti.time,ti_tau=ti_time-ti_orig;
	    final byte bs=ti.buy_sell_indicator;
	    for(int j=i+1;j<tsize;j++)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		final long tj_tau=tj.time-tj_orig;
		final int c_buy,c_sell,c_bs1,c_bs2,c_dir,c_cd;
		if(bs==BookItem.BOOK_BUY)
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c_buy=-cmpt(ti_orig,tj_orig);
			c_sell=-1;
			c_bs1=-1;
			c_bs2=-cmpt(ti_time,tj_orig);
			c_dir=cmpt(ti_tau,tj_tau);
			c_cd=tj_orig!=ti_orig && tj_orig<ti_time?-1:0;
		    }
		    else
		    {
			c_buy=1;
			c_sell=cmpt(ti_time,tj_orig);
			c_bs1=cmpt(ti_orig,tj_orig);
			c_bs2=1;
			c_dir=-cmpt(ti_tau,tj_tau);
			c_cd=tj_orig<ti_time?1:0;
			//c_dir=1;
		    }
		}
		else
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c_buy=-cmpt(ti_time,tj_orig);
			c_sell=-1;
			c_bs1=-cmpt(ti_orig,tj_orig);
			c_bs2=-1;
			c_dir=cmpt(ti_tau,tj_tau);
			c_cd=tj_orig<ti_time?-1:0;
			//c_dir=-1;
		    }
		    else
		    {
			c_buy=1;
			c_sell=cmpt(ti_orig,tj_orig);
			c_bs1=1;
			c_bs2=cmpt(ti_time,tj_orig);
			c_dir=-cmpt(ti_tau,tj_tau);
			c_cd=tj_orig!=ti_orig && tj_orig<ti_time?1:0;
		    }
		}
		final long vs=vvi*tj.shares;
		if(c_buy>0)
		{
		    v_cons_buy+=vs;
		}
		else if(c_buy<0)
		{
		    v_discons_buy+=vs;
		}
		else
		{
		    v_zero_buy+=vs;
		}
		if(c_sell>0)
		{
		    v_cons_sell+=vs;
		}
		else if(c_sell<0)
		{
		    v_discons_sell+=vs;
		}
		else
		{
		    v_zero_sell+=vs;
		}
		if(c_bs1>0)
		{
		    v_cons_bs+=vs;
		}
		else if(c_bs1<0)
		{
		    v_discons_bs+=vs;
		}
		else
		{
		    v_zero_bs+=vs;
		}
		if(c_bs2>0)
		{
		    v_cons_bs+=vs;
		}
		else if(c_bs2<0)
		{
		    v_discons_bs+=vs;
		}
		else
		{
		    v_zero_bs+=vs;
		}

		if(c_dir>0)
		{
		    v_cons_dir+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons_dir+=vs;
		}
		else
		{
		    v_zero_dir+=vs;
		}
		if(c_cd>0)
		{
		    v_cons_cd+=vs;
		}
		else if(c_cd<0)
		{
		    v_discons_cd+=vs;
		}
		else
		{
		    v_zero_cd+=vs;
		}

	    }
	    vi+=vvi;
	    if(bs==BookItem.BOOK_SELL)
	    {
		v_cons_self+=vvi*vvi;
	    }
	    else if(bs==BookItem.BOOK_BUY)
	    {
		v_discons_self+=vvi*vvi;
	    }
	    else
	    {
		throw new RuntimeException("Orders problem");
	    }
	}


	final CalculateKendall [] r=new CalculateKendall[]{
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_buy+v_cons_sell,
				 v_discons_buy+v_discons_sell,
				 v_zero_buy+v_zero_sell),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_buy,
				 v_discons_buy,
				 v_zero_buy),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_sell,
				 v_discons_sell,
				 v_zero_sell),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_bs,
				 v_discons_bs,
				 v_zero_bs),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_self,
				 v_discons_self,
				 v_zero_self),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_bs+v_cons_self,
				 v_discons_bs+v_discons_self,
				 v_zero_bs+v_zero_self),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_dir,
				 v_discons_dir,
				 v_zero_dir),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_cd,
				 v_discons_cd,
				 v_zero_cd)
	};

	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{ // for expensive constrains

	    final CalculateKendall [] rr=calculateTradesKendallRankTOrigDiffXY(
		currenttime,
		trades,
		ptr_t_range_cut,
		tsize);
	    checkKendalls("Results mismatch:",rr[0],r[5]);
	    checkKendalls("Results mismatch:",rr[1],r[4]);

	    final CalculateKendall [] rrr=calculateTradesKendallRankTPOrigDiff(
		currenttime,
		trades,
		ptr_t_range_cut,
		tsize);
	    checkKendalls("Results mismatch:",rrr[1],r[1]);
	    checkKendalls("Results mismatch:",rrr[2],r[2]);
	}


	return r;
    }


    /** Kendal tau for trades in ptr_t_range_cut to tsize
     *  interval. Trades XY version.
     */
static CalculateKendall [] calculateTradesKendallRankTOrigDiffXY(
    final long currenttime,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	long vi=0,v_cons=0,v_discons=0,v_zero=0,v_cons_self=0,v_discons_self=0,v_zero_self=0;
	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares,ti_time=ti.time;
	    final byte bs=ti.buy_sell_indicator;
	    for(int j=ptr_interval;j<tsize;j++)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		final int c;
		if(bs==BookItem.BOOK_BUY)
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c=cmpt(ti_orig,tj.time);
		    }
		    else
		    {
			c=cmpt(ti_orig,tj_orig);
		    }
		}
		else
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c=cmpt(ti_time,tj.time);
		    }
		    else
		    {
			c=cmpt(ti_time,tj_orig);
		    }
		}
		final long vs=vvi*tj.shares;
		if(c>0)
		{
		    v_cons+=vs;
		}
		else if(c<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
		if(i==j)
		{
		   if(c>0)
		   {
		       v_cons_self+=vs;
		   }
		   else if(c<0)
		   {
		       v_discons_self+=vs;
		   }
		   else
		   {
		       v_zero_self+=vs;
		   }
		}
	    }
	    vi+=vvi;
	}


	final CalculateKendall [] r=new CalculateKendall []{
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons,
				 v_discons,
				 v_zero),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_self,
				 v_discons_self,
				 v_zero_self)
	};
	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{ // for expensive constrains

	    //check_calculateTradesKendallRankTOrigDiff(currenttime,trades,ptr_t_range_cut,r);
	}
	return r;
    }



    /** Kendal tau for  book, take time into account.
     */
static CalculateKendall calculateBookKendallRankXY(
    final long currenttime,
    final int n_buy,
    final Distribution<BookItem> buy,
    final int n_sell,
    final Distribution<BookItem> sell)
    {
	final BookItem [] bbuy=buy.getBuffer(),bsell=sell.getBuffer();
	long v_cons=0,v_discons=0,v_zero=0;
	for(int i=0;i<n_buy;i++)
	{
	    final BookItem ti=bbuy[i];
	    final long vvi=ti.shares,ti_time=ti.time;
	    for(int j=0;j<n_sell;j++)
	    {
		final BookItem tj=bsell[j];
		final int c=-cmpt(ti_time,tj.time);
		final long vs=vvi*tj.shares;
		if(c>0)
		{
		    v_cons+=vs;
		}
		else if(c<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
	    }
	}

	return new CalculateKendall(currenttime,
				    0,
				    v_cons,
				    v_discons,
				    v_zero);
    }




    /** Kendal tau for book, take into account both: price & time.
     *  interval. Trades XY version.
     */
static CalculateKendall calculateBookKendallCumm(
    final long currenttime,
    final int n_buy,
    final Distribution<BookItem> buy,
    final int n_sell,
    final Distribution<BookItem> sell,
    final BookItem [] bb)
    {
	System.arraycopy(buy.getBuffer(),0,bb,0,n_buy);
	System.arraycopy(sell.getBuffer(),0,bb,n_buy,n_sell);
	final int bsize=n_buy+n_sell;
	long vi=0,v_cons=0,v_discons=0,v_zero=0;
	for(int i=0;i<bsize;i++)
	{
	    final BookItem ti=bb[i];
	    final long vvi=ti.shares;
	    final double ti_p=ti.price;
	    final byte bs=ti.buy_sell_indicator;
	    for(int j=i+1;j<bsize;j++)
	    {
		final BookItem tj=bb[j];
		final int dt=BookItem.cmpByTime(ti,tj);
		final int c;
		if(bs==BookItem.BOOK_BUY)
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c=cmpp(ti_p,tj.price)*dt;
		    }
		    else
		    {
			c=dt;
		    }
		}
		else
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c=-dt;
		    }
		    else
		    {
			c=cmpp(ti_p,tj.price)*dt;
		    }
		}

		final long vs=vvi*tj.shares;
		if(c>0)
		{
		    v_cons+=vs;
		}
		else if(c<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
	    }
	    vi+=vvi;
	}

	return new CalculateKendall(currenttime,
				    vi,
				    v_cons,
				    v_discons,
				    v_zero);
    }




    /** Kendal tau for book, take into account both: price & time.
     *  interval. Trades XY version.
     */
static CalculateKendall [] calculateBookPTradesKendall(
    final long currenttime,
    final int n_buy,
    final Distribution<BookItem> buy,
    final int n_sell,
    final Distribution<BookItem> sell,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut)
    {

	final int tsize=trades.getSize();
	final Trade [] tb=trades.getBuffer();
	final BookItem [] bbuy=buy.getBuffer(),bsell=sell.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	long vi=0,v_cons_buy=0,v_discons_buy=0,v_zero_buy=0,v_cons_sell=0,v_discons_sell=0,v_zero_sell=0;

	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares;
	    for(int j=0;j<n_buy;j++)
	    {
		final BookItem tj=bbuy[j];
		final int c_plts=tj.time<ti_orig?-cmpp(ti.price,tj.price):0;
		final long vs=vvi*tj.shares;
		if(c_plts>0)
		{
		    v_cons_buy+=vs;
		}
		else if(c_plts<0)
		{
		    v_discons_buy+=vs;
		}
		else
		{
		    v_zero_buy+=vs;
		}
	    }
	    for(int j=0;j<n_sell;j++)
	    {
		final BookItem tj=bsell[j];
		final int c_plts=tj.time<ti_orig?-cmpp(ti.price,tj.price):0;
		final long vs=vvi*tj.shares;
		if(c_plts>0)
		{
		    v_cons_sell+=vs;
		}
		else if(c_plts<0)
		{
		    v_discons_sell+=vs;
		}
		else
		{
		    v_zero_sell+=vs;
		}
	    }
	    vi+=vvi;
	}

	return new CalculateKendall[]{
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_buy,
				 v_discons_buy,
				 v_zero_buy),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_sell,
				 v_discons_sell,
				 v_zero_sell)
	};
    }




    /** Kendal tau for trades in ptr_t_range_cut to tsize
     *  interval. Trades origin version.
     */
static CalculateKendall [] calculateTradesKendallRankTPOrigDiff(
    final long currenttime,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	long vi=0,v_cons_buy=0,v_discons_buy=0,v_zero_buy=0,v_cons_sell=0,v_discons_sell=0,v_zero_sell=0,v_cons_dir=0,v_discons_dir=0,v_zero_dir=0,v_cons_ddir=0,v_discons_ddir=0,v_zero_ddir=0;
	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares,ti_time=ti.time;
	    final byte bs=ti.buy_sell_indicator;
	    for(int j=i+1;j<tsize;j++)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		final int c_buy,c_sell,c_dir,c_ddir;
		if(bs==BookItem.BOOK_BUY)
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c_sell=-1;
			c_buy=-cmpt(ti_orig,tj_orig);
			if(tj_orig!=ti_orig && tj_orig<=ti_time)
			{
			    c_dir=-1;
			    c_ddir=-cmpp(ti.price,tj.price);
			}
			else
			{
			    c_dir=c_ddir=0;
			}
		    }
		    else
		    {
			c_buy=1;
			c_sell=cmpt(ti_time,tj_orig);
			if(tj_orig<=ti_time)
			{
			    c_dir=1;
			    c_ddir=-cmpp(ti.price,tj.price);
			}
			else
			{
			    c_dir=c_ddir=0;
			}
		    }
		}
		else
		{
		    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		    {
			c_buy=-cmpt(ti_time,tj_orig);
			c_sell=-1;
			if(tj_orig<=ti_time)
			{
			    c_dir=-1;
			    c_ddir=-cmpp(ti.price,tj.price);
			}
			else
			{
			    c_dir=c_ddir=0;
			}
		    }
		    else
		    {
			c_buy=1;
			c_sell=cmpt(ti_orig,tj_orig);
			if(tj_orig!=ti_orig && tj_orig<=ti_time)
			{
			    c_dir=1;
			    c_ddir=-cmpp(ti.price,tj.price);
			}
			else
			{
			    c_dir=c_ddir=0;
			}
		    }
		}
		final long vs=vvi*tj.shares;
		if(c_buy>0)
		{
		    v_cons_buy+=vs;
		}
		else if(c_buy<0)
		{
		    v_discons_buy+=vs;
		}
		else
		{
		    v_zero_buy+=vs;
		}
		if(c_sell>0)
		{
		    v_cons_sell+=vs;
		}
		else if(c_sell<0)
		{
		    v_discons_sell+=vs;
		}
		else
		{
		    v_zero_sell+=vs;
		}
		if(c_dir>0)
		{
		    v_cons_dir+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons_dir+=vs;
		}
		else
		{
		    v_zero_dir+=vs;
		}
		if(c_ddir>0)
		{
		    v_cons_ddir+=vs;
		}
		else if(c_ddir<0)
		{
		    v_discons_ddir+=vs;
		}
		else
		{
		    v_zero_ddir+=vs;
		}
	    }
	    vi+=vvi;
	}



	final CalculateKendall [] r=new CalculateKendall[]{
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_buy+v_cons_sell,
				 v_discons_buy+v_discons_sell,
				 v_zero_buy+v_zero_sell),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_buy,
				 v_discons_buy,
				 v_zero_buy),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_sell,
				 v_discons_sell,
				 v_zero_sell),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_dir,
				 v_discons_dir,
				 v_zero_dir),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_ddir,
				 v_discons_ddir,
				 v_zero_ddir)
	};
	return r;
    }



private static void check_calculateBookSideTradesKendall(
    final long currenttime,
    final int n_orders,
    final BookItem [] orders,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final CalculateKendall rcoupling,
    final CalculateKendall rcut_book,
    final CalculateKendall rcut_trades)
{
    final int tsize=trades.getSize();
    final Trade [] tb=trades.getBuffer();
    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;

    long vi=0,v_cons_dir=0,v_discons_dir=0,v_zero_dir=0,v_cons_bdir=0,v_discons_bdir=0,v_zero_bdir=0,v_cons_tdir=0,v_discons_tdir=0,v_zero_tdir=0;
    for(int i=0;i<n_orders;i++)
    {
	final BookItem t=orders[i];
	if(t.time<t_range_cut)
	{
	    v_cons_bdir+=t.shares;
	}
	else
	{
	    v_discons_bdir+=t.shares;
	}
    }

    for(int i=ptr_interval;i<tsize;i++)
    {
	final Trade t=tb[i];
	if(t.origtime<=0)
	{
	    continue;
	}
	for(int j=0;j<n_orders;j++)
	{
	    final BookItem b=orders[j];
	    final long b_orig=b.time;
	    if(b_orig<=0)
	    {
		System.err.println("Warning: b_orig="+b_orig);
		continue;
	    }
	    final int c_dir;
	    if(t.time<b_orig)
	    {
		c_dir=0;
	    }
	    else
	    {
		c_dir=(b_orig!=t.origtime||b.buy_sell_indicator!=t.buy_sell_indicator)?(b.buy_sell_indicator==BookItem.BOOK_BUY?-1:1):0;
	    }

	    final long vs=b.shares*t.shares;
	    if(c_dir>0)
	    {
		v_cons_dir+=vs;
	    }
	    else if(c_dir<0)
	    {
		v_discons_dir+=vs;
	    }
	    else
	    {
		v_zero_dir+=vs;
	    }
	}

	if(t.origtime<t_range_cut)
	{
	    v_cons_tdir+=t.shares;
	}
	else
	{
	    v_discons_tdir+=t.shares;
	}
	vi+=t.shares;
    }


    checkKendalls("Results mismatch:",rcut_book,new CalculateKendall(currenttime,
								     v_cons_bdir+v_discons_bdir,
								     v_cons_bdir,
								     v_discons_bdir,
								     v_zero_bdir));

    checkKendalls("Results mismatch:",rcut_trades,new CalculateKendall(currenttime,
								       vi,
								       v_cons_tdir,
								       v_discons_tdir,
								       v_zero_tdir));

    checkKendalls("Results mismatch: Vbook="+(v_cons_bdir+v_discons_bdir)+
		  " Vtrades="+vi,
		  rcoupling,
		  new CalculateKendall(currenttime,
				       vi,
				       v_cons_dir,
				       v_discons_dir,
				       v_zero_dir));

}


static class DCalculateKendall extends CalculateKendall {

    final double dv_cons,dv_discons,dv_zero,diff;
    DCalculateKendall(final long currenttime,
		      final long vi,
		      final double v_cons,
		      final double v_discons,
		      final double v_zero,
		      final double diff)
    {

	super(currenttime,vi,0,0,0);
	this.dv_cons=v_cons;
	this.dv_discons=v_discons;
	this.dv_zero=v_zero;
	this.diff=diff;
    }

    long getN()
    {
	throw new RuntimeException("Not supported DCalculateKendall");
    }

    double getND()
    {
	return this.dv_cons+this.dv_discons+this.dv_zero;
    }

    double getRank()
    {
	final double n=getND();
	if(n<=0)
	{
	    return Double.NaN;
	}
	final double denom;
	if(this.dv_zero<=0)
	{
	    denom=n;
	}
	else
	{
	    denom=Math.sqrt(n)*Math.sqrt(this.dv_cons+this.dv_discons);
	}
	// use diff to reduce rounding error
	return this.diff /*(this.dv_cons-this.dv_discons)*/ /denom;
    }

    double getRankA()
    {
	final double n=getND();
	if(n<=0)
	{
	    return Double.NaN;
	}
	return (this.dv_cons-this.dv_discons)/n;
    }
}



    /** Kendal tau for trades in ptr_t_range_cut to tsize
     *  interval. Trades origin version.
     */
static CalculateKendall [] calculateTradesKendallRankTP3OrigDiff(
    final long currenttime,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	long va_cons_dir=0,va_discons_dir=0,v_cons_dir=0,v_discons_dir=0,v_zero_dir=0,v_cons_ddir=0,v_discons_ddir=0,v_zero_ddir=0;
	long n0_cons_dir=0,n0_discons_dir=0,n_cons_dir=0,n_discons_dir=0,n_zero_dir=0,n_cons_dir3=0,n_discons_dir3=0,n_zero_dir3=0;
	double v_cons_dir3=0,v_discons_dir3=0,v_zero_dir3=0,diff_dir3=0,v_cons_ddir3=0,v_discons_ddir3=0,v_zero_ddir3=0,diff_ddir3=0;
	long v_cons_dbs=0,v_discons_dbs=0,v_zero_dbs=0;

	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares,ti_time=ti.time;
	    final byte bsi=ti.buy_sell_indicator;
	    final int c_dbsi=bsi==BookItem.BOOK_BUY?-1:1;
	    for(int j=i+1;j<tsize;j++)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		final byte bsj=tj.buy_sell_indicator;
		final int c_ddir,c_dbsj=bsj==BookItem.BOOK_BUY?-1:1;
		final long c_dir;
		if((tj_orig!=ti_orig || bsi!=bsj) && tj_orig<=ti_time)
		{
		    c_dir=(bsj==BookItem.BOOK_BUY?-1:1)/* *(ti_time-tj_orig+1) */;
		    c_ddir=-cmpp(ti.price,tj.price);
		}
		else
		{
		    c_dir=c_ddir=0;
		}

		final long vs=vvi*tj.shares;
		if(c_dir>0)
		{
		    v_cons_dir+=vs;
		    n_cons_dir+=1;
		}
		else if(c_dir<0)
		{
		    v_discons_dir+=vs;
		    n_discons_dir+=1;
		}
		else
		{
		    v_zero_dir+=vs;
		    n_zero_dir+=1;
		}
		if(c_ddir>0)
		{
		    v_cons_ddir+=vs;
		}
		else if(c_ddir<0)
		{
		    v_discons_ddir+=vs;
		}
		else
		{
		    v_zero_ddir+=vs;
		}
		if(c_dbsj>c_dbsi)
		{
		    v_cons_dbs+=vs;
		}
		else if(c_dbsj<c_dbsi)
		{
		    v_discons_dbs+=vs;
		}
		else
		{
		    v_zero_dbs+=vs;
		}
		for(int k=j+1;k<tsize;k++)
		{
		    final Trade tk=tb[k];
		    final long tk_orig=tk.origtime;
		    if(tk_orig<=0)
		    {
			continue;
		    }
		    final byte bsk=tk.buy_sell_indicator;
		    final int cc_ddir;
		    final long cc_dir;
		    if((tk_orig!=tj_orig || bsj!=bsk) && tk_orig<=tj.time)
		    {
			cc_dir=(bsk==BookItem.BOOK_BUY?-1:1)/* *(tj.time-tk_orig+1) */;
			cc_ddir=-cmpp(tj.price,tk.price);
		    }
		    else
		    {
			cc_dir=cc_ddir=0;
		    }
		    final double vsk=vs*tk.shares;
		    if(cc_dir>c_dir)
		    {
			v_cons_dir3+=vsk;
			diff_dir3+=vsk;
			n_cons_dir3+=1;
		    }
		    else if(cc_dir<c_dir)
		    {
			v_discons_dir3+=vsk;
			diff_dir3-=vsk;
			n_discons_dir3+=1;
		    }
		    else
		    {
			v_zero_dir3+=vsk;
			n_zero_dir3+=1;
		    }
		    if(cc_ddir>c_ddir)
		    {
			v_cons_ddir3+=vsk;
			diff_ddir3+=vsk;
		    }
		    else if(cc_ddir<c_ddir)
		    {
			v_discons_ddir3+=vsk;
			diff_ddir3-=vsk;
		    }
		    else
		    {
			v_zero_ddir3+=vsk;
		    }
		}
	    }

	    if(bsi==BookItem.BOOK_BUY)
	    {
		va_discons_dir+=vvi;
		n0_cons_dir+=1;
	    }
	    else if(bsi==BookItem.BOOK_SELL)
	    {
		va_cons_dir+=vvi;
		n0_discons_dir+=1;
	    }
	    else
	    {
		throw new RuntimeException("Problem with trades");
	    }
	}
	final long vi=va_discons_dir+va_cons_dir;


	final CalculateKendall [] r=new CalculateKendall[]{
	    new CalculateKendall(currenttime,
				 vi,
				 va_cons_dir,
				 va_discons_dir,
				 0),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_dir,
				 v_discons_dir,
				 v_zero_dir),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_ddir,
				 v_discons_ddir,
				 v_zero_ddir),
	    new DCalculateKendall(currenttime,
				  vi,
				  v_cons_dir3,
				  v_discons_dir3,
				  v_zero_dir3,
				  diff_dir3),
	    new DCalculateKendall(currenttime,
				  vi,
				  v_cons_ddir3,
				  v_discons_ddir3,
				  v_zero_ddir3,
				  diff_ddir3),
	    new CalculateKendall(currenttime,
				 vi,
				 n0_cons_dir,
				 n0_discons_dir,
				 0),
	    new CalculateKendall(currenttime,
				 vi,
				 n_cons_dir,
				 n_discons_dir,
				 n_zero_dir),
	    new CalculateKendall(currenttime,
				 vi,
				 n_cons_dir3,
				 n_discons_dir3,
				 n_zero_dir3),
	    new CalculateKendall(currenttime,
				 vi,
				 v_cons_dbs,
				 v_discons_dbs,
				 v_zero_dbs)
	};

	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{ // for expensive constrains

	    final CalculateKendall [] rr=calculateTradesKendallRankTPOrigDiff(
		currenttime,
		trades,
		ptr_t_range_cut,
		tsize);
	    final CalculateKendall dir=rr[3],ddir=rr[4];
	    checkKendalls("Results mismatch:",dir,r[1]);
	    checkKendalls("Results mismatch:",ddir,r[2]);
	}


	return r;
    }



    static void checkKendalls(final String txt,
			      final CalculateKendall k1,
			      final CalculateKendall k2)
    {
	if(k2.vi!=k1.vi||
	   k2.v_cons!=k1.v_cons||
	   k2.v_discons!=k1.v_discons||
	   k2.v_zero!=k1.v_zero||
	   k2.currenttime!=k1.currenttime)
	{
	    throw new RuntimeException(
		"Results mismatch: "+txt+
		" k2.vi="+k2.vi+" k1.vi="+k1.vi+
		" k2.v_cons="+k2.v_cons+" k1.v_cons="+k1.v_cons+
		" k2.v_discons="+k2.v_discons+" k1.v_discons="+k1.v_discons+
		" k2.v_zero="+k2.v_zero+" k1.v_zero="+k1.v_zero+
		" k2.currenttime="+k2.currenttime+" k1.currenttime="+k1.currenttime);
	}
    }

    static void checkVKendalls(final String txt,
			       final ValCalculateKendall k1,
			       final ValCalculateKendall k2)
    {
	final double eps=1e-9;
	checkKendalls(txt,k1,k2);
	final double vn1=k1.getN(),vn2=k2.getN();
	final double val_diff1=k1.val_diff/vn1,val_squared1=k1.val_squared/vn1;
	final double val_diff2=k2.val_diff/vn2,val_squared2=k2.val_squared/vn2;
	if(Double.isNaN(val_diff1)!=Double.isNaN(val_diff2)||
	   (!Double.isNaN(val_diff1) && !(eps>=Math.abs(val_diff1-val_diff2)))||
	   Double.isNaN(val_squared1)!=Double.isNaN(val_squared2)||
	   (!Double.isNaN(val_squared1) && !(eps>=Math.abs(val_squared1-val_squared2))))
	{
	    throw new RuntimeException(
		"Values mismatch:"+
		" val_diff1="+val_diff1+" val_diff2="+val_diff2+
		" val_squared1="+val_squared1+" val_squared2="+val_squared2);
	}
    }



    static class KendallMultipleStats {
	final long t_range_cut,currenttime;
	final double p_aver,p_stddev;
	final CalculateKendall krank1,krank2,krank2taubuy,krank2tausell,krank2cd,krank3;
	KendallMultipleStats(final CalculateKendall krank1,
			     final CalculateKendall krank2,
			     final CalculateKendall krank2taubuy,
			     final CalculateKendall krank2tausell,
			     final CalculateKendall krank2cd,
			     final CalculateKendall krank3,
			     final double p_aver,
			     final double p_stddev,
			     final long t_range_cut,
			     final long currenttime)
	{
	    this.krank1=krank1;
	    this.krank2=krank2;
	    this.krank2taubuy=krank2taubuy;
	    this.krank2tausell=krank2tausell;
	    this.krank3=krank3;
	    this.krank2cd=krank2cd;
	    this.p_aver=p_aver;
	    this.p_stddev=p_stddev;
	    this.t_range_cut=t_range_cut;
	    this.currenttime=currenttime;
	}
    }



    /** Kendal tau for trades in ptr_t_range_cut to tsize
     *  interval. Trades origin version.
     */
static KendallMultipleStats [] calculateTradesKendallRankMultipleTP3OrigDiff(
    final long currenttime,
    final Distribution<Trade> trades,
    final int [] ptr_t_range_cut,
    final boolean skip_d3)
    {
	final int tsize=trades.getSize(),nt=ptr_t_range_cut.length;
	final Trade [] tb=trades.getBuffer();
	final KendallMultipleStats[] r=new KendallMultipleStats [nt];

	class CMP implements com.polytechnik.utils.IntPTRCompare {
	    final int [] px=com.polytechnik.outils.TMPQSortAlgorithm.sortArrayIndexes(nt,this);
	    public int compare(final int p1,final int p2)
	    {
		final int v1=ptr_t_range_cut[p1],v2=ptr_t_range_cut[p2];
		if(v1==v2)
		{
		    return 0;
		}
		if(v1<0)
		{
		    return -1;
		}
		if(v2<0)
		{
		    return 1;
		}
		return v1<v2?1:-1; // highest up
	    }
	}

	long va_cons_dir=0,va_discons_dir=0,v_cons_dir=0,v_discons_dir=0,v_zero_dir=0,v_cons_taubuy=0,v_discons_taubuy=0,v_zero_taubuy=0,v_cons_tausell=0,v_discons_tausell=0,v_zero_tausell=0,v_cons_cd=0,v_discons_cd=0,v_zero_cd=0;
	double v_cons_dir3=0,v_discons_dir3=0,v_zero_dir3=0,diff_dir3=0;
	double p_sum=0,p2_sum=0;

	final CMP ind=new CMP();
	final int ptr_interval=ptr_t_range_cut[ind.px[nt-1]]<0?tsize:ptr_t_range_cut[ind.px[nt-1]];
	int ptr_save=0,ptr_i_end=tsize;
	for(;ptr_save<nt;)
	{
	    ptr_i_end=ptr_t_range_cut[ind.px[ptr_save]]<0?tsize:ptr_t_range_cut[ind.px[ptr_save]];
	    if(ptr_i_end<ptr_interval)
	    {
		throw new RuntimeException("Sorting problem");
	    }
	    if(ptr_i_end>=tsize)
	    {
		final long vi=va_discons_dir+va_cons_dir;
		r[ind.px[ptr_save]]=new KendallMultipleStats(
		    new CalculateKendall(currenttime,
					 vi,
					 va_cons_dir,
					 va_discons_dir,
					 0),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_dir,
					 v_discons_dir,
					 v_zero_dir),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_taubuy,
					 v_discons_taubuy,
					 v_zero_taubuy),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_tausell,
					 v_discons_tausell,
					 v_zero_tausell),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_cd,
					 v_discons_cd,
					 v_zero_cd),
		    new DCalculateKendall(currenttime,
					  vi,
					  v_cons_dir3,
					  v_discons_dir3,
					  v_zero_dir3,
					  diff_dir3),
		    Double.NaN,
		    Double.NaN,
		    -1,
		    currenttime);
		ptr_save++;
	    }
	    else
	    {
		break;
	    }
	}

	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }

	    final long vvi=ti.shares,ti_time=ti.time,ti_tau=ti_time-ti_orig;
	    final boolean bsi=ti.buy_sell_indicator==BookItem.BOOK_BUY;
	    for(int j=i+1;j<tsize;j++)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		final long tj_tau=tj.time-tj_orig;
		final boolean bsj=tj.buy_sell_indicator==BookItem.BOOK_BUY;
		final int c_dir,c_tau_buy,c_tau_sell,cd_buy,cd_sell;
		if(bsi)
		{
		    if(bsj)
		    {
			cd_buy=tj_orig!=ti_orig && tj_orig<ti_time?-1:0;
			cd_sell=0;
		    }
		    else
		    {
			cd_buy=0;
			cd_sell=tj_orig<ti_time?1:0;
		    }
		}
		else
		{
		    if(bsj)
		    {
			cd_buy=tj_orig<ti_time?-1:0;
			cd_sell=0;
		    }
		    else
		    {
			cd_buy=0;
			cd_sell=tj_orig!=ti_orig && tj_orig<ti_time?1:0;
		    }
		}

		final boolean dtrades=(tj_orig!=ti_orig || bsi!=bsj);
		if(dtrades && tj_orig<=ti_time)
		{
		    c_dir=(bsj?-1:1);
		    if(bsi)
		    {
			if(bsj)
			{
			    c_tau_buy=cmpt(ti_tau,tj_tau);
			    c_tau_sell=0;
			}
			else
			{
			    c_tau_buy=1;
			    c_tau_sell=1;
			}
		    }
		    else
		    {
			if(bsj)
			{
			    c_tau_buy=-1;
			    c_tau_sell=-1;
			}
			else
			{
			    c_tau_buy=0;
			    c_tau_sell=-cmpt(ti_tau,tj_tau);
			}
		    }
		}
		else
		{
		    // identical orders or not coupled pairs
		    c_tau_buy=c_tau_sell=c_dir=0;
		}

		final long vs=vvi*tj.shares;
		if(c_dir>0)
		{
		    v_cons_dir+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons_dir+=vs;
		}
		else
		{
		    v_zero_dir+=vs;
		}
		if(c_tau_buy>0)
		{
		    v_cons_taubuy+=vs;
		}
		else if(c_tau_buy<0)
		{
		    v_discons_taubuy+=vs;
		}
		else
		{
		    v_zero_taubuy+=vs;
		}
		if(c_tau_sell>0)
		{
		    v_cons_tausell+=vs;
		}
		else if(c_tau_sell<0)
		{
		    v_discons_tausell+=vs;
		}
		else
		{
		    v_zero_tausell+=vs;
		}
		if(cd_buy>0||cd_sell>0)
		{
		    v_cons_cd+=vs;
		}
		else if(cd_buy<0||cd_sell<0)
		{
		    v_discons_cd+=vs;
		}
		else
		{
		    v_zero_cd+=vs;
		}

		if(skip_d3)
		{
		    continue;
		}
		for(int k=j+1;k<tsize;k++)
		{
		    final Trade tk=tb[k];
		    final long tk_orig=tk.origtime;
		    if(tk_orig<=0)
		    {
			continue;
		    }
		    final boolean bsk=tk.buy_sell_indicator==BookItem.BOOK_BUY;
		    final long cc_dir;
		    if((tk_orig!=tj_orig || bsj!=bsk) && tk_orig<=tj.time)
		    {
			cc_dir=(bsk?-1:1);
		    }
		    else
		    {
			cc_dir=0;
		    }
		    final double vsk=vs*tk.shares;
		    if(cc_dir>c_dir)
		    {
			v_cons_dir3+=vsk;
			diff_dir3+=vsk;
		    }
		    else if(cc_dir<c_dir)
		    {
			v_discons_dir3+=vsk;
			diff_dir3-=vsk;
		    }
		    else
		    {
			v_zero_dir3+=vsk;
		    }
		}
	    }

	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		va_discons_dir+=vvi;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		va_cons_dir+=vvi;
	    }
	    else
	    {
		throw new RuntimeException("Problem with trades");
	    }
	    p_sum+=ti.price*vvi;
	    p2_sum+=(ti.price*ti.price)*vvi;

	    for(;i<=ptr_i_end;)
	    {
		if(i!=ptr_i_end)
		{
		    throw new RuntimeException("Pointers problem");
		}
		final long vi=va_discons_dir+va_cons_dir;
		final double pa=p_sum/vi;
		r[ind.px[ptr_save]]=new KendallMultipleStats(
		    new CalculateKendall(currenttime,
					 vi,
					 va_cons_dir,
					 va_discons_dir,
					 0),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_dir,
					 v_discons_dir,
					 v_zero_dir),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_taubuy,
					 v_discons_taubuy,
					 v_zero_taubuy),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_tausell,
					 v_discons_tausell,
					 v_zero_tausell),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_cd,
					 v_discons_cd,
					 v_zero_cd),
		    new DCalculateKendall(currenttime,
					  vi,
					  v_cons_dir3,
					  v_discons_dir3,
					  v_zero_dir3,
					  diff_dir3),
		    pa,
		    Math.sqrt(p2_sum/vi-pa*pa), // biased
		    tb[ptr_t_range_cut[ind.px[ptr_save]]].time,
		    currenttime);
		ptr_save++;
		if(ptr_save<nt)
		{
		    final int prev_ptr_i_end=ptr_i_end;
		    ptr_i_end=ptr_t_range_cut[ind.px[ptr_save]]<0?tsize:ptr_t_range_cut[ind.px[ptr_save]];
		    if(ptr_i_end>prev_ptr_i_end || ptr_i_end<ptr_interval)
		    {
			throw new RuntimeException("Ordering problem");
		    }
		}
		else
		{
		    if(i!=ptr_interval)
		    {
			throw new RuntimeException("Pointers problem");
		    }
		    break;
		}
	    }
	}
	if(ptr_save!=nt)
	{
	    throw new RuntimeException("Pointers count problem ptr_save="+ptr_save+" nt="+nt);
	}


	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{
	    for(int i=0;i<nt;i++)
	    {
		final KendallMultipleStats cr=r[i];
		final long t_range_cut=ptr_t_range_cut[i]<0?-1:tb[ptr_t_range_cut[i]].time;
		if(t_range_cut!=cr.t_range_cut)
		{
		    throw new RuntimeException("Sorting priblem IND="+i+
					       " t_range_cut="+t_range_cut+
					       " cr.t_range_cut="+cr.t_range_cut+
					       " currenttime="+currenttime);
		}
		final CalculateKendall [] rr3=
		    CalculateKendall.calculateTradesKendallRankTP3OrigDiff(
			currenttime,
			trades,
			ptr_t_range_cut[i],
			trades.getSize());

		checkKendalls("K1 Results mismatch: for IND="+i,rr3[0],cr.krank1);
		checkKendalls("K2 Results mismatch: for IND="+i,rr3[1],cr.krank2);
		final CalculateKendall k3=rr3[3];
		if(!skip_d3 && (
		    Double.isNaN(cr.krank3.getRank())!=Double.isNaN(k3.getRank())||
		   (!Double.isNaN(cr.krank3.getRank()) &&
		    Math.abs(cr.krank3.getRank()-k3.getRank())>1e-9)))
		{
		   throw new RuntimeException(
		       "Results mismatch: for IND="+i+
		       " cr.krank3.getRank()="+cr.krank3.getRank()+
		       " k3.getRank()="+k3.getRank()+
		       " currenttime="+currenttime);
		}
		final CalculateKendall [] rr2=calculateTradesKendallRankTOrigDiff(
		    currenttime,
		    trades,
		    ptr_t_range_cut[i],
		    trades.getSize());
		checkKendalls("K2 Results mismatch: for IND="+i,rr2[7],cr.krank2cd);
	    }
	}

	return r;
    }





    /** Kendal tau for trades in given tau interval.
     */
static KendallMultipleStats calculateTradesKendallRankMultipleTP3OrigTau(
    final long currenttime,
    final Distribution<Trade> trades,
    final long tau)
    {
	final int tsize=trades.getSize();
	final Trade [] tb=trades.getBuffer();
	final long t_cut=currenttime-tau;

	long va_cons_dir=0,va_discons_dir=0,v_cons_dir=0,v_discons_dir=0,v_zero_dir=0,v_cons_taubuy=0,v_discons_taubuy=0,v_zero_taubuy=0,v_cons_tausell=0,v_discons_tausell=0,v_zero_tausell=0;
	double p_sum=0,p2_sum=0;

	for(int i=tsize;i-->0;)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares,ti_time=ti.time,ti_tau=ti_time-ti_orig,t_cut_j=ti_time-tau;
	    if(ti_time<t_cut)
	    {
		break;
	    }
	    final boolean bsi=ti.buy_sell_indicator==BookItem.BOOK_BUY;
	    for(int j=i;j-->0;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		if(tj.time<t_cut_j)
		{
		    break;
		}
		final long tj_tau=tj.time-tj_orig;

		final boolean bsj=tj.buy_sell_indicator==BookItem.BOOK_BUY;
		final int c_dir,c_tau_buy,c_tau_sell;
		final boolean dtrades=(tj_orig!=ti_orig || bsi!=bsj);
		if(dtrades && tj_orig<=ti_time)
		{
		    c_dir=(bsj?-1:1);
		    if(bsi)
		    {
			if(bsj)
			{
			    c_tau_buy=cmpt(ti_tau,tj_tau);
			    //c_tau_buy=-cmpt(ti_orig,tj_orig);
			    c_tau_sell=0;
			}
			else
			{
			    c_tau_buy=1;
			    c_tau_sell=1;
			    //c_tau_buy=c_tau_sell=0;
			}
		    }
		    else
		    {
			if(bsj)
			{
			    c_tau_buy=-1;
			    c_tau_sell=-1;
			    //c_tau_buy=c_tau_sell=0;
			}
			else
			{
			    c_tau_buy=0;
			    c_tau_sell=-cmpt(ti_tau,tj_tau);
			    //c_tau_sell=cmpt(ti_orig,tj_orig);
			}
		    }
		}
		else
		{
		    // identical orders or not coupled pairs
		    c_tau_buy=c_tau_sell=c_dir=0;
		}

		final long vs=vvi*tj.shares;
		if(c_dir>0)
		{
		    v_cons_dir+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons_dir+=vs;
		}
		else
		{
		    v_zero_dir+=vs;
		}
		if(c_tau_buy>0)
		{
		    v_cons_taubuy+=vs;
		}
		else if(c_tau_buy<0)
		{
		    v_discons_taubuy+=vs;
		}
		else
		{
		    v_zero_taubuy+=vs;
		}
		if(c_tau_sell>0)
		{
		    v_cons_tausell+=vs;
		}
		else if(c_tau_sell<0)
		{
		    v_discons_tausell+=vs;
		}
		else
		{
		    v_zero_tausell+=vs;
		}
	    }

	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		va_discons_dir+=vvi;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		va_cons_dir+=vvi;
	    }
	    else
	    {
		throw new RuntimeException("Problem with trades");
	    }
	    p_sum+=ti.price*vvi;
	    p2_sum+=(ti.price*ti.price)*vvi;
	}

	final long vi=va_discons_dir+va_cons_dir;
	final double pa=p_sum/vi;
	return new KendallMultipleStats(
		    new CalculateKendall(currenttime,
					 vi,
					 va_cons_dir,
					 va_discons_dir,
					 0),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_dir,
					 v_discons_dir,
					 v_zero_dir),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_taubuy,
					 v_discons_taubuy,
					 v_zero_taubuy),
		    new CalculateKendall(currenttime,
					 vi,
					 v_cons_tausell,
					 v_discons_tausell,
					 v_zero_tausell),
		    new DCalculateKendall(currenttime,
					  vi,
					  0,
					  0,
					  0,
					  0),
		    new DCalculateKendall(currenttime,
					  vi,
					  0,
					  0,
					  0,
					  0),
		    pa,
		    Math.sqrt(p2_sum/vi-pa*pa), // biased
		    t_cut,
		    currenttime);
    }



    /** Kendal D3.
     */
static CalculateKendall calculateBookTradesKendallD3(
    final long currenttime,
    final int n_buy,
    final Distribution<BookItem> buy,
    final int n_sell,
    final Distribution<BookItem> sell,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final BookItem [] bb)
    {
	System.arraycopy(buy.getBuffer(),0,bb,0,n_buy);
	System.arraycopy(sell.getBuffer(),0,bb,n_buy,n_sell);
	final int bsize=n_buy+n_sell,tsize=trades.getSize(),ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final Trade [] tb=trades.getBuffer();
	long vi=0;
	for(int i=0;i<n_buy;i++)
	{
	    vi+=bb[i].shares;
	}
	final double scale_book_buy=vi>0?vi:Double.NaN;
	vi=0;
	for(int i=n_buy;i<bsize;i++)
	{
	    vi+=bb[i].shares;
	}
	final double scale_book_sell=vi>0?vi:Double.NaN;
	vi=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    vi+=ti.shares;
	}
	final long v_trades=vi;

	double v_cons_dir3=0,v_discons_dir3=0,v_zero_dir3=0,diff_dir3=0;
	for(int k=0;k<bsize;k++)
	{
	    final BookItem tk=bb[k];
	    final long tk_orig=tk.time;
	    final byte bsk=tk.buy_sell_indicator;
	    final double vvk=tk.shares/(bsk==BookItem.BOOK_BUY?scale_book_buy:scale_book_sell);
	    for(int j=tsize;j-->ptr_interval;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		final byte bsj=tj.buy_sell_indicator;
		final long tj_time=tj.time,cc_dir;
		final double vs=vvk*tj.shares;
		if(tk_orig<=tj_time)
		{
		    if(tj_orig!=tk_orig || bsj!=bsk)
		    {
			cc_dir=(bsk==BookItem.BOOK_BUY?-1:1)/* *(tj_time-tk_orig+1) */;
		    }
		    else
		    {
			cc_dir=0;
		    }
		}
		else
		{
		    // optimization
		    break;
		}

		vi=0;
		for(int i=j;i-->ptr_interval;)
		{
		    final Trade ti=tb[i];
		    final long ti_orig=ti.origtime;
		    if(ti_orig<=0)
		    {
			continue;
		    }
		    final byte bsi=ti.buy_sell_indicator;
		    final long ti_time=ti.time,c_dir;
		    final double vsk;
		    if(tj_orig<=ti_time)
		    {
			if(ti_orig!=tj_orig || bsi!=bsj)
			{
			    c_dir=(bsj==BookItem.BOOK_BUY?-1:1)/* *(ti_time-tj_orig+1) */;
			}
			else
			{
			    c_dir=0;
			}
			vsk=vs*ti.shares;
		    }
		    else
		    {
			 // optimization
			break;
		    }

		    if(c_dir!=0 && cc_dir!=0)
		    {
			if(cc_dir>c_dir)
			{
			    v_cons_dir3+=vsk;
			    diff_dir3+=vsk;
			}
			else if(cc_dir<c_dir)
			{
			    v_discons_dir3+=vsk;
			    diff_dir3-=vsk;
			}
			else
			{
			    v_zero_dir3+=vsk;
			}
		    }

		    vi+=ti.shares;
		}
	    }
	}

	return new DCalculateKendall(currenttime,
				     v_trades,
				     v_cons_dir3,
				     v_discons_dir3,
				     v_zero_dir3,
				     diff_dir3);
    }





private static long findVAfter(final long t,
			       final Trade [] tb,
			       final int tsize,
			       final int ptr_interval)
{
    long v=0;
    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade ti=tb[i];
	if(ti.origtime<=0)
	{
	    continue;
	}
	if(ti.time<t)
	{
	    break;
	}
	v+=ti.shares;
    }
    return v;
}

    /** Kendal D2Diff unoptimized version.
     */
static CalculateKendall calculateBookTradesKD2Diff(
    final long currenttime,
    final int n_orders,
    final Distribution<BookItem> orders,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut)
    {
	final BookItem [] bb=orders.getBuffer();
	long v_book=0;
	int n_pricelevels=0;
	double pprev=Double.NaN;
	for(int i=0;i<n_orders;i++)
	{
	    final BookItem b=bb[i];
	    if(b.price!=pprev)
	    {
		pprev=b.price;
		n_pricelevels++;
	    }
	    v_book+=b.shares;
	}
	// temporary arrays, re-use them in optimized version
	final long [] v_sumcompare=new long [n_pricelevels],
	    v_level=new long [n_pricelevels];

	final int tsize=trades.getSize(),ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final Trade [] tb=trades.getBuffer();

	pprev=Double.NaN;
	for(int i=0,level=-1;i<n_orders;i++)
	{
	    final BookItem b=bb[i];
	    if(b.price!=pprev)
	    {
		pprev=b.price;
		level++;
	    }
	    v_level[level]+=b.shares;
	    v_sumcompare[level]+=b.shares*findVAfter(b.time,tb,tsize,ptr_interval);
	}

	long v_cons=0,v_discons=0,v_zero=0;
	for(int i=0;i<n_pricelevels;i++)
	{
	    final long vi=v_level[i];
	    final double vsi=v_sumcompare[i]/((double)vi);
	    for(int j=i+1;j<n_pricelevels;j++)
	    {
		final int c=-cmpp(vsi,v_sumcompare[j]/((double)v_level[j]));
		final long vs=vi*v_level[j];
		if(c>0)
		{
		    v_cons+=vs;
		}
		else if(c<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
	    }
	}
	return new CalculateKendall(currenttime,
				    v_book,
				    v_cons,
				    v_discons,
				    v_zero);
    }






    /** Kendal Book-Trades unoptimized coupled version.
     */
static CalculateKendall calculateBookTradesCoupledDiff(
    final long currenttime,
    final int n_orders,
    final Distribution<BookItem> orders,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut)
    {

	final int tsize=trades.getSize();
	final Trade [] tb=trades.getBuffer();
	final BookItem [] bb=orders.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;

	long v_book=0;
	double v_cons=0,v_discons=0,v_zero=0,v_diff=0;

	for(int i=0;i<n_orders;i++)
	{
	    final BookItem b=bb[i];
	    final long b_orig=b.time,bshares=b.shares;
	    final byte btype=b.buy_sell_indicator;
	    long sv_cons=0,sv_discons=0,sv_zero=0,vi=0;
	    for(int j=tsize;j--> /* 0 */ ptr_interval;)
	    {
		final Trade t=tb[j];
		final long t_orig=t.origtime;
		if(t_orig<=0 || t.buy_sell_indicator!=btype)
		{
		    continue;
		}
		if(t.time<b_orig)
		{
		    break;
		}
		// only coupled trades of the same type left
		final int c_dir=cmpt(b_orig,t_orig);
		final long vs=t.shares;
		if(c_dir>0)
		{
		    sv_cons+=vs;
		}
		else if(c_dir<0)
		{
		    sv_discons+=vs;
		}
		else
		{
		    sv_zero+=vs;
		}
		vi+=vs;
	    }
	    if(vi>0)
	    {
		final double dvi=1;//vi;
		v_cons+=(bshares*sv_cons)/dvi;
		v_discons+=(bshares*sv_discons)/dvi;
		v_zero+=(bshares*sv_zero)/dvi;
		v_diff+=(bshares*(sv_cons-sv_discons))/dvi;
	    }
	    v_book+=bshares;
	}
	return new DCalculateKendall(currenttime,
				     v_book,
				     v_cons,
				     v_discons,
				     v_zero,
				     v_diff);
    }



    /** Kendal D2Sum.
     */
static CalculateKendall calculateBookTradesKendallD2Sum(
    final long currenttime,
    final int n_buy,
    final Distribution<BookItem> buy,
    final int n_sell,
    final Distribution<BookItem> sell,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final BookItem [] bb)
    {
	System.arraycopy(buy.getBuffer(),0,bb,0,n_buy);
	System.arraycopy(sell.getBuffer(),0,bb,n_buy,n_sell);
	final int bsize=n_buy+n_sell,tsize=trades.getSize(),ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final Trade [] tb=trades.getBuffer();

	long vt_buy=0,vt_sell=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		vt_buy+=ti.shares;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		vt_sell+=ti.shares;
	    }
	    else
	    {
		throw new RuntimeException("Orders problem");
	    }
	}
	final long v_trades=vt_buy+vt_sell;

	long vi=0;
	for(int i=0;i<n_buy;i++)
	{
	    vi+=bb[i].shares;
	}
	final double scale_book_buy=vi>0 && vt_buy>0?vi*vt_buy:Double.NaN;
	vi=0;
	for(int i=n_buy;i<bsize;i++)
	{
	    vi+=bb[i].shares;
	}
	final double scale_book_sell=vi>0 && vt_sell>0?vi*vt_sell:Double.NaN;

	double v_cons_dir2=0,v_discons_dir2=0,v_zero_dir2=0,diff_dir2=0;
	for(int k=0;k<bsize;k++)
	{
	    final BookItem tk=bb[k];
	    final long tk_orig=tk.time;
	    final byte bsk=tk.buy_sell_indicator;
	    final double vvk=tk.shares/(bsk==BookItem.BOOK_BUY?scale_book_buy:scale_book_sell);
	    for(int j=tsize;j-->ptr_interval;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		final byte bsj=tj.buy_sell_indicator;
		if(tj_orig<=0 || bsj!=bsk)
		{
		    continue;
		}
		final long tj_time=tj.time,cc_dir;
		final double vsk=vvk*tj.shares;
		if(tk_orig<=tj_time)
		{
		    if(tj_orig!=tk_orig)
		    {
			cc_dir=(bsk==BookItem.BOOK_BUY?-1:1)*cmpt(tj_orig,tk_orig);
		    }
		    else
		    {
			cc_dir=0;
		    }
		}
		else
		{
		    cc_dir=0;
		    // optimization
		    //break;
		}
		if(cc_dir>0)
		{
		    v_cons_dir2+=vsk;
		    diff_dir2+=vsk;
		}
		else if(cc_dir<0)
		{
		    v_discons_dir2+=vsk;
		    diff_dir2-=vsk;
		}
		else
		{
		    v_zero_dir2+=vsk;
		}
	    }
	}

	return new DCalculateKendall(currenttime,
				     v_trades,
				     v_cons_dir2,
				     v_discons_dir2,
				     v_zero_dir2,
				     diff_dir2);
    }



    /** TP2 optimized version.
     */
static KendallMultipleStats calculateTradesKendallRankTP2OrigDiff(
    final long currenttime,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	long va_cons_dir=0,va_discons_dir=0,v_cons=0,v_discons=0,v_zero=0;
	double p_sum=0;
	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    final int vvi=ti.shares;
	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		va_discons_dir+=vvi;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		va_cons_dir+=vvi;
	    }
	    else
	    {
		throw new RuntimeException("Problem with trades");
	    }
	    p_sum+=ti.price*vvi;
	}
	final long v_trades=va_discons_dir+va_cons_dir;
	final double pa=(v_trades>0?p_sum/v_trades:Double.NaN);

	double p2_sum=0;

	long vi=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares;
	    vi+=vvi;
	    final byte bsi=ti.buy_sell_indicator;
	    final int c_i=bsi==BookItem.BOOK_BUY?-1:1;
	    long vsj=v_trades-vi;
	    for(int j=i;j-->ptr_interval;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		if(tj.time<ti_orig)
		{
		    v_zero+=vvi*vsj;
		    break; // optimization
		}
		final int c_dir=(tj_orig!=ti_orig || bsi!=tj.buy_sell_indicator)?c_i:0;
		final long vs=vvi*tj.shares;
		if(c_dir>0)
		{
		    v_cons+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
		vsj-=tj.shares;
	    }
	    p2_sum+=(ti.price-pa)*(ti.price-pa)*vvi;
	}

	final KendallMultipleStats r=new KendallMultipleStats(
	    new CalculateKendall(currenttime,
				 v_trades,
				 va_cons_dir,
				 va_discons_dir,
				 0),
	    new CalculateKendall(currenttime,
				 v_trades,
				 v_cons,
				 v_discons,
				 v_zero),
	    null,
	    null,
	    null,
	    null,
	    pa,
	    Math.sqrt(p2_sum/v_trades), // biased
	    ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time,
	    currenttime);

	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{
	    final CalculateKendall [] rr3=
		CalculateKendall.calculateTradesKendallRankTP3OrigDiff(
		    currenttime,
		    trades,
		    ptr_t_range_cut,
		    tsize);

		checkKendalls("K1 Results mismatch: ",rr3[0],r.krank1);
		checkKendalls("K2 Results mismatch: ",rr3[1],r.krank2);
	}
	return r;
    }

   static class KendallBookTradesCoupling {
       final long t_range_cut,currenttime;
       final CalculateKendall krank_coupling,krank_forward_cut,krank_back_cut;
       KendallBookTradesCoupling(final CalculateKendall krank_coupling,
				 final CalculateKendall krank_forward_cut,
				 final CalculateKendall krank_back_cut,
				 final long t_range_cut,
				 final long currenttime)
	{
	    this.krank_coupling=krank_coupling;
	    this.krank_forward_cut=krank_forward_cut;
	    this.krank_back_cut=krank_back_cut;
	    this.t_range_cut=t_range_cut;
	    this.currenttime=currenttime;
	}
    }




/** Kendal tau for book-trades coupling.
 */
static KendallBookTradesCoupling calculateBookSideTradesKendall(
    final long currenttime,
    final int n_orders,
    final BookItem [] orders,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut)
    {
	final int tsize=trades.getSize();
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
	long v_cons=0,v_discons=0,v_zero=0,v_cons_bdir=0,v_discons_bdir=0,v_cons_tdir=0,v_discons_tdir=0;

	for(int j=ptr_interval;j<tsize;j++)
	{
	    final Trade tj=tb[j];
	    if(tj.origtime<=0)
	    {
		continue;
	    }
	    if(tj.origtime<t_range_cut)
	    {
		v_cons_tdir+=tj.shares;
	    }
	    else
	    {
		v_discons_tdir+=tj.shares;
	    }
	}
	final long v_trades=v_cons_tdir+v_discons_tdir;

	for(int i=0;i<n_orders;i++)
	{
	    final BookItem ti=orders[i];
	    final long ti_orig=ti.time;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares;
	    final byte bsi=ti.buy_sell_indicator;
	    final int c_i=bsi==BookItem.BOOK_BUY?-1:1;
	    long vsj=v_trades;
	    for(int j=tsize;j-->ptr_interval;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		if(tj.time<ti_orig)
		{
		    v_zero+=vvi*vsj;
		    break; // optimization
		}
		final int c_dir=(tj_orig!=ti_orig || bsi!=tj.buy_sell_indicator)?c_i:0;
		final long vs=vvi*tj.shares;
		if(c_dir>0)
		{
		    v_cons+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
		vsj-=tj.shares;
	    }


	    if(ti_orig<t_range_cut)
	    {
		v_cons_bdir+=vvi;
	    }
	    else
	    {
		v_discons_bdir+=vvi;
	    }
	}

	final KendallBookTradesCoupling r=
	    new KendallBookTradesCoupling(
		new CalculateKendall(currenttime,
				     v_trades,
				     v_cons,
				     v_discons,
				     v_zero),
		new CalculateKendall(currenttime,
				     v_cons_bdir+v_discons_bdir,
				     v_cons_bdir,
				     v_discons_bdir,
				     0),
		new CalculateKendall(currenttime,
				     v_trades,
				     v_cons_tdir,
				     v_discons_tdir,
				     0),
		t_range_cut,
		currenttime);


	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{ // for expensive constrains
	   check_calculateBookSideTradesKendall(
	       currenttime,
	       n_orders,
	       orders,
	       trades,
	       ptr_t_range_cut,
	       r.krank_coupling,
	       r.krank_forward_cut,
	       r.krank_back_cut);
	}
	return r;
    }


/** Kendal tau for trades-trades coupling.
 */
static KendallBookTradesCoupling calculateTradesTradesCouplingKendall(
    final long currenttime,
    final Trade [] tb,
    final int ptr_start,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final int ptr_interval=ptr_t_range_cut<0||ptr_start<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0||ptr_start<0?-1:tb[ptr_t_range_cut].time;
	final int ptr_start_interval=ptr_start<0?tsize:ptr_start;
	if(ptr_start_interval>ptr_interval)
	{
	    throw new IllegalArgumentException(" ptr_start_interval="+ptr_start_interval+
					       " ptr_interval="+ptr_interval+
					       " tsize="+tsize);
	}
	long v_cons=0,v_discons=0,v_zero=0,v_cons_bdir=0,v_discons_bdir=0,v_cons_tdir=0,v_discons_tdir=0;

	for(int j=ptr_start_interval;j<ptr_interval;j++)
	{
	    final Trade tj=tb[j];
	    if(tj.origtime<=0)
	    {
		continue;
	    }
	    if(tj.origtime<t_range_cut)
	    {
		v_cons_tdir+=tj.shares;
	    }
	    else
	    {
		v_discons_tdir+=tj.shares;
	    }
	}
	final long v_trades=v_cons_tdir+v_discons_tdir;

	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares;
	    final byte bsi=ti.buy_sell_indicator;
	    final int c_i=bsi==BookItem.BOOK_BUY?-1:1;
	    long vsj=v_trades;
	    for(int j=ptr_interval;j-->ptr_start_interval;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		if(tj.time<ti_orig)
		{
		    v_zero+=vvi*vsj;
		    break; // optimization
		}
		final int c_dir=(tj_orig!=ti_orig || bsi!=tj.buy_sell_indicator)?c_i:0;
		final long vs=vvi*tj.shares;
		if(c_dir>0)
		{
		    v_cons+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
		vsj-=tj.shares;
	    }


	    if(ti_orig<t_range_cut)
	    {
		v_cons_bdir+=vvi;
	    }
	    else
	    {
		v_discons_bdir+=vvi;
	    }
	}

	final KendallBookTradesCoupling r=
	    new KendallBookTradesCoupling(
		new CalculateKendall(currenttime,
				     v_trades,
				     v_cons,
				     v_discons,
				     v_zero),
		new CalculateKendall(currenttime,
				     v_cons_bdir+v_discons_bdir,
				     v_cons_bdir,
				     v_discons_bdir,
				     0),
		new CalculateKendall(currenttime,
				     v_trades,
				     v_cons_tdir,
				     v_discons_tdir,
				     0),
		t_range_cut,
		currenttime);

	return r;
    }





static class ValCalculateKendall extends CalculateKendall {
    final double val_diff,val_squared;
    ValCalculateKendall(final long currenttime,
			final long vi,
			final long v_cons,
			final long v_discons,
			final long v_zero,
			final double val_diff,
			final double val_squared)
    {

	super(currenttime,vi,v_cons,v_discons,v_zero);
	this.val_diff=val_diff;
	this.val_squared=val_squared;
    }

    double getValA()
    {
	final long n=getN();
	if(n<=0)
	{
	    return Double.NaN;
	}
	// use diff to reduce rounding error
	return this.val_diff/((double)n);
    }

    double getValAA()
    {
	final long n1=this.v_cons+this.v_discons;
	if(n1<=0)
	{
	    return Double.NaN;
	}
	return Math.sqrt(((double)getN())/n1)*getValA();
    }

    double getValAStdev()
    {
	final long n=getN();
	if(n<=0)
	{
	    return Double.NaN;
	}
	final double va=getValA();
	return Math.sqrt(this.val_squared/n-va*va); // biased
    }
}


    static class KendallMStats {
	final double price0_average,price0_stded,p0_direction;
	final long t_range_cut;
	final ValCalculateKendall trades_price,trades_time;
	final CalculateKendall volume_growth;
	KendallMStats(final ValCalculateKendall trades_price,
		      final ValCalculateKendall trades_time,
		      final double price0_average,
		      final double price0_stded,
		      final long t_range_cut,
		      final double p0_direction,
		      final CalculateKendall volume_growth)
	{
	    this.trades_price=trades_price;
	    this.trades_time=trades_time;
	    this.price0_average=price0_average;
	    this.price0_stded=price0_stded;
	    this.t_range_cut=t_range_cut;
	    this.p0_direction=p0_direction;
	    this.volume_growth=volume_growth;
	}
	double getTauRel()
	{
	    if(this.t_range_cut<=0)
	    {
		return Double.NaN;
	    }
	    return this.trades_time.getValA()/(this.trades_time.currenttime-this.t_range_cut);
	}
    }



/** TP2 in price space (slightly optimized version).
 */
static KendallMStats calculateTradesKendallRankTP2OrigPriceDiff(
    final long currenttime,
    final Distribution<Trade> trades,
    final int ptr_t_range_cut,
    final int tsize,
    final double pcurrent_buy,
    final double pcurrent_sell)
    {
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
	long va_sell=0,va_buy=0;
	double ps=0,dps=0;
	for(int i=ptr_interval;i<tsize;i++)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    final int vvi=ti.shares;
	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		va_buy+=vvi;
		dps+=(pcurrent_buy-ti.price)*vvi;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		va_sell+=vvi;
		dps+=(pcurrent_sell-ti.price)*vvi;
	    }
	    else
	    {
		throw new RuntimeException("Problem with trades");
	    }
	    ps+=ti.price*vvi;
	}
	final long v_trades=va_buy+va_sell;
	final double p0_average,p0_dir;
	if(v_trades>0)
	{
	    p0_average=ps/v_trades;
	    p0_dir=dps/v_trades;
	}
	else
	{
	    p0_average=p0_dir=Double.NaN;
	}

	long v_p_cons=0,v_p_discons=0,v_p_zero=0;
	double p_cons=0,p_discons=0,p_diff=0,p_squared=0;
	long v_dir_cons=0,v_dir_discons=0,v_dir_zero=0;
	double tau_cons=0,tau_discons=0,tau_diff=0;
	long v_growth_cons=0,v_growth_discons=0,v_growth_zero=0;
	ps=0;
	long vi=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares;
	    vi+=vvi;
	    final byte bsi=ti.buy_sell_indicator;
	    final int c_i=bsi==BookItem.BOOK_BUY?-1:1;
	    final double p_i=ti.price;
	    final double tdvi=(currenttime-ti.time)/((double)vi);
	    long vsj=v_trades-vi,vj=vi;
	    for(int j=i;j-->ptr_interval;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		if(tj.time<ti_orig)
		{
		    final long v0=vvi*vsj;
		    v_p_zero+=v0;
		    v_dir_zero+=v0;
		    v_growth_zero+=v0;
		    break; // optimization
		}
		final boolean same_trade_type=bsi==tj.buy_sell_indicator;
		final long vs=vvi*tj.shares;
		final double dp=p_i-tj.price,dpabs=(dp<0?-dp:dp),pvs=dpabs*vs;
		p_squared+=dpabs*pvs;
		if(dp>0)
		{
		    v_p_cons+=vs;
		    p_cons+=pvs;
		    p_diff+=pvs;
		    if(false && bsi!=BookItem.BOOK_SELL)
		    {
			System.err.println("calculateTradesKendallRankTP2OrigPriceDiff:"+
					   " positive dp="+dp+
					   " for ti.buy_sell_indicator="+bsi+
					   " this should not happen"+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
		    }
		}
		else if(dp<0)
		{
		    v_p_discons+=vs;
		    p_discons+=pvs;
		    p_diff-=pvs;
		    if(false && bsi!=BookItem.BOOK_BUY)
		    {
			System.err.println("calculateTradesKendallRankTP2OrigPriceDiff:"+
					   " negative dp="+dp+
					   " for ti.buy_sell_indicator="+bsi+
					   " this should not happen"+
					   " currenttime="+BookItem.getTimeDiag(currenttime));
		    }
		}
		else
		{
		    v_p_zero+=vs;
		}
		final double dtau=ti.time-tj.time,tvs=dtau*vs;
		final int c_dir=(tj_orig!=ti_orig || !same_trade_type)?c_i:0;
		if(c_dir>0)
		{
		    v_dir_cons+=vs;
		    tau_cons+=tvs;
		    tau_diff+=tvs;
		}
		else if(c_dir<0)
		{
		    v_dir_discons+=vs;
		    tau_discons+=tvs;
		    tau_diff-=tvs;
		}
		else
		{
		    v_dir_zero+=vs;
		}
		vj+=tj.shares;
		final double tdvj=(currenttime-tj.time)/((double)vj);
		final int c_vdir=c_dir==0?0:(tdvi<tdvj?1:(tdvi>tdvj?-1:0));
		if(c_vdir>0)
		{
		    v_growth_cons+=vs;
		}
		else if(c_vdir<0)
		{
		    v_growth_discons+=vs;
		}
		else
		{
		    v_growth_zero+=vs;
		}

		vsj-=tj.shares;
	    }
	    final double dpa=ti.price-p0_average;
	    ps+=dpa*(dpa*vvi);
	}
	final double p0_stdev=Math.sqrt(ps/v_trades); // biased

	final KendallMStats r=new KendallMStats(
	    new ValCalculateKendall(currenttime,
				    v_trades,
				    v_p_cons,
				    v_p_discons,
				    v_p_zero,
				    p_diff,
				    p_squared),
	    new ValCalculateKendall(currenttime,
				    v_trades,
				    v_dir_cons,
				    v_dir_discons,
				    v_dir_zero,
				    tau_diff,
				    Double.NaN),
	    p0_average,
	    p0_stdev,
	    t_range_cut,
	    p0_dir,
	    new CalculateKendall(currenttime,
				 v_trades,
				 v_growth_cons,
				 v_growth_discons,
				 v_growth_zero));

	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{
	    final CalculateKendall [] rr3=
		CalculateKendall.calculateTradesKendallRankTP3OrigDiff(
		    currenttime,
		    trades,
		    ptr_t_range_cut,
		    tsize);
	    checkKendalls("K2 Results mismatch: ",rr3[1],r.trades_time);
	}
	return r;
    }



    /** Kendal tau for  book, take time into account.
     */
static CalculateKendall calculateBookKendallRankV(
     final long currenttime,
     final Distribution<Trade> trades,
     final int ptr_t_range_cut,
     final int tsize)
    {
	final Trade [] tb=trades.getBuffer();
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;

	long v_growth_cons=0,v_growth_discons=0,v_growth_zero=0;
	long vi=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares;
	    final byte bsi=ti.buy_sell_indicator;
	    final int c_i=bsi==BookItem.BOOK_BUY?-1:1;
	    vi+=vvi;
	    long vj=vi;
	    final double tdvi=(currenttime-ti.time)/((double)vi);
	    for(int j=i;j-->ptr_interval;)
	    {
		final Trade tj=tb[j];
		final long tj_orig=tj.origtime;
		if(tj_orig<=0)
		{
		    continue;
		}
		final long vs=vvi*tj.shares;
		vj+=tj.shares;
		final double tdvj=(currenttime-tj.time)/((double)vj);

		final int c_dir=(tj_orig!=ti_orig || bsi!=tj.buy_sell_indicator)?c_i:0;
		final int c_vdir=c_dir==0?0:(tdvi<tdvj?1:(tdvi>tdvj?-1:0));
		if(c_vdir>0)
		{
		    v_growth_cons+=vs;
		}
		else if(c_vdir<0)
		{
		    v_growth_discons+=vs;
		}
		else
		{
		    v_growth_zero+=vs;
		}
	    }
	}
	return new CalculateKendall(currenttime,
				    vi,
				    v_growth_cons,
				    v_growth_discons,
				    v_growth_zero);
    }



static void check_calculateTradesKendallEndPriceDiff(final long currenttime,
						     final Trade [] tb,
						     final int ptr_t_range_cut,
						     final int tsize,
						     final double pcurrent_buy,
						     final double pcurrent_sell,
						     final ValCalculateKendall r,
						     final double pi_aver_sum)
{
    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final double midp=0.5*(pcurrent_buy+pcurrent_sell);

    long v_p_cons=0,v_p_discons=0,v_p_zero=0;
    double p_cons=0,p_discons=0,p_diff=0,p_squared=0;
    long vi=0;
    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade ti=tb[i];
	if(ti.buy_sell_indicator!=BookItem.BOOK_BUY &&
	   ti.buy_sell_indicator!=BookItem.BOOK_SELL)
	{
	    continue;
	}
	final long vvi=ti.shares;
	vi+=vvi;
	for(int j=tsize;j-->i;)
	{
	    final Trade tj=tb[j];
	    final double dp;
	    final int c_dir;
	    if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		c_dir=cmpp(pcurrent_buy,tj.price);
		//dp=pcurrent_buy-tj.price;
	    }
	    else if(tj.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		c_dir=cmpp(pcurrent_sell,tj.price);
		//dp=pcurrent_sell-tj.price;
	    }
	    else
	    {
		continue;
	    }
	    //c_dir=cmpp(midp,tj.price);
	    dp=midp-tj.price;
	    final long n=vvi;
	    final long vs=n*tj.shares;
	    final double dpcontrib=vs*dp,absdpcontrib=dp<0?-dpcontrib:dpcontrib;
	    p_diff+=dpcontrib;
	    p_squared+=dpcontrib*dp;
	    if(c_dir>0)
	    {
		v_p_cons+=vs;
		p_cons+=absdpcontrib;
	    }
	    else if(c_dir<0)
	    {
		v_p_discons+=vs;
		p_discons+=absdpcontrib;
	    }
	    else
	    {
		v_p_zero+=vs;
	    }
	}
    }


    final ValCalculateKendall rc=new ValCalculateKendall(
	currenttime,
	vi,
	v_p_cons,
	v_p_discons,
	v_p_zero,
	p_diff,
	p_squared);

    checkVKendalls("Results mismatch:",r,rc);
}



static class MSAverageKendall {
    final long t_range_cut,currenttime;
    final double dp_delayed_average,dp_buy,dp_sell,dp_underestimate,dp_overestimate,dp_end,dt_tau,dt_tau_cut,dtd_tau,dtd_tau_cut;
    final ValCalculateKendall trades_krank_forward;
    MSAverageKendall(final long t_range_cut,
		     final long currenttime,
		     final ValCalculateKendall trades_krank_forward,
		     final double dp_delayed_average,
		     final double dp_buy,
		     final double dp_sell,
		     final double dp_underestimate,
		     final double dp_overestimate,
		     final double dp_end,
		     final double dt_tau,
		     final double dt_tau_cut,
		     final double dtd_tau,
		     final double dtd_tau_cut)
    {
	this.t_range_cut=t_range_cut;
	this.currenttime=currenttime;
	this.trades_krank_forward=trades_krank_forward;
	this.dp_delayed_average=dp_delayed_average;
	this.dp_buy=dp_buy;
	this.dp_sell=dp_sell;
	this.dp_underestimate=dp_underestimate;
	this.dp_overestimate=dp_overestimate;
	this.dp_end=dp_end;
	this.dt_tau=dt_tau;
	this.dt_tau_cut=dt_tau_cut;
	this.dtd_tau=dtd_tau;
	this.dtd_tau_cut=dtd_tau_cut;
    }

    double getDPEstimate()
    {
	return this.dp_overestimate;//+0.5*(this.dp_overestimate-this.dp_delayed_average);
    }
}

/** TP2 in price space at the end.
 */
static MSAverageKendall calculateTradesKendallEndPriceDiff(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize,
    final double pcurrent_buy,
    final double pcurrent_sell)
    {
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
	final long tau_cut=currenttime-t_range_cut;
	final double midp=0.5*(pcurrent_buy+pcurrent_sell);

	long v_p_cons=0,v_p_discons=0,v_p_zero=0,v_d_cons=0,v_d_discons=0,v_d_zero=0;
	double p_diff=0,p_squared=0,pi_diff=0,p_diff_buy=0,p_diff_sell=0,p_diff_underestimate=0,p_diff_overestimate=0,t_aver=0,t_aver_cut=0,td_aver=0,td_aver_cut=0;
	long vi=0;
	double dp_aver=0,dp_aver_buy=0,dp_aver_sell=0,dp_underestimate_aver=0,dp_overestimate_aver=0,dpsq_aver=0,spi_aver=0,dt_aver=0,dt_aver_cut=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    final long vvi=ti.shares;

	    final long dt,dt_cut;
	    if(ti.origtime<=0)
	    {
		dt=dt_cut=0;
	    }
	    else
	    {
		dt=ti.time-ti.origtime;
		dt_cut=dt>tau_cut?tau_cut:dt;
	    }
	    final int c_dir;
	    final long xdt,xdt_cut;
	    final double p=ti.price,vvid=vvi,dp,dpu,dpo;
	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		c_dir=cmpp(pcurrent_buy,p);
		dpu=pcurrent_sell-p;
		dpo=pcurrent_buy-p;
		xdt=-dt;
		xdt_cut=-dt_cut;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		c_dir=cmpp(pcurrent_sell,p);
		dpu=pcurrent_buy-p;
		dpo=pcurrent_sell-p;
		xdt=dt;
		xdt_cut=dt_cut;
	    }
	    else
	    {
		continue;
	    }

	    vi+=vvi;
	    dp=midp-p;

	    final double dpvvi=dp*vvid;
	    dp_aver+=dpvvi;
	    dpsq_aver+=dp*dpvvi;
	    dp_aver_buy+=(pcurrent_buy-p)*vvid;
	    dp_aver_sell+=(pcurrent_sell-p)*vvid;
	    dp_underestimate_aver+=dpu*vvid;
	    dp_overestimate_aver+=dpo*vvid;
	    spi_aver+=p*vvid;
	    dt_aver+=xdt*vvid;
	    dt_aver_cut+=xdt_cut*vvid;


	    if(c_dir>0)
	    {
		v_d_cons+=vvi;
	    }
	    else if(c_dir<0)
	    {
		v_d_discons+=vvi;
	    }
	    else
	    {
		v_d_zero+=vvi;
	    }

	    final long n=vvi;
	    final double nd=n;
	    p_diff+=nd*dp_aver;
	    p_squared+=nd*dpsq_aver;
	    pi_diff+=nd*(spi_aver-vi*p);
	    p_diff_buy+=nd*dp_aver_buy;
	    p_diff_sell+=nd*dp_aver_sell;
	    p_diff_underestimate+=nd*dp_underestimate_aver;
	    p_diff_overestimate+=nd*dp_overestimate_aver;
	    t_aver+=nd*dt_aver;
	    t_aver_cut+=nd*dt_aver_cut;
	    td_aver+=nd*(dt_aver-vi*((double)xdt));
	    td_aver_cut+=nd*(dt_aver_cut-vi*((double)xdt_cut));

	    v_p_cons+=n*v_d_cons;
	    v_p_discons+=n*v_d_discons;
	    v_p_zero+=n*v_d_zero;
	}

	final ValCalculateKendall kr=new ValCalculateKendall(
	    currenttime,
	    vi,
	    v_p_cons,
	    v_p_discons,
	    v_p_zero,
	    p_diff,
	    p_squared);
	final long n=kr.getN();
	final double norm=n>=0?n:Double.NaN;
	final double pa=p_diff/norm;
	final MSAverageKendall r=new MSAverageKendall(
	    t_range_cut,
	    currenttime,
	    kr,
	    pi_diff/norm,
	    p_diff_buy/norm,
	    p_diff_sell/norm,
	    p_diff_underestimate/norm,
	    p_diff_overestimate/norm,
	    pa-(midp-(tsize>0?tb[tsize-1].price:Double.NaN)),
	    t_aver/norm,
	    t_aver_cut/norm,
	    td_aver/norm,
	    td_aver_cut/norm);
	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{
	    check_calculateTradesKendallEndPriceDiff(currenttime,
						     tb,
						     ptr_t_range_cut,
						     tsize,
						     pcurrent_buy,
						     pcurrent_sell,
						     r.trades_krank_forward,
						     pi_diff);
	}
	return r;
    }


static class DAvergeCalculateKendall  {
    final long t_range_cut,currenttime,tau_V;
    final double norma,norma_unadj,v_aver,
	time_to_aver,time_to_aver_2,time_to_aver_2f,
	ttype_dir_average,ttype_dir_average2,ttype_dir_average2f,dttype_dir,
	tau_average,tau_average2,tau_average2f,dtau,
	dp_average,dp_average2,dp_average2f,ddp,
	dptau_average,dptau_average2,dptau_average2f,ddptau,
	dpv_average,dpv_average2,dpv_average2f,ddpv,
	ptauaver_normp;
    final double dpdtau_aver,dpdatau_aver,dpdt_aver,dtaudt_aver,dvdt_aver,dpdv_aver,dtaudv_aver,dttypedt_aver,dttypedp_aver,dpdvsigned_aver,paver2,taver2,tauaver2,vaver2,ttype2,vsignedaver2;
    DAvergeCalculateKendall(final long t_range_cut,
			    final long currenttime,
			    final double norma,
			    final double norma_unadj,
			    final double v_aver,
			    final long tau_V,

			    final double time_to_aver,
			    final double time_to_aver_2,
			    final double time_to_aver_2f,

			    final double ttype_dir_average,
			    final double ttype_dir_average2,
			    final double ttype_dir_average2f,
			    final double dttype_dir,

			    final double tau_average,
			    final double tau_average2,
			    final double tau_average2f,
			    final double dtau,

			    final double dp_average,
			    final double dp_average2,
			    final double dp_average2f,
			    final double ddp,

			    final double dptau_average,
			    final double dptau_average2,
			    final double dptau_average2f,
			    final double ddptau,

			    final double dpv_average,
			    final double dpv_average2,
			    final double dpv_average2f,
			    final double ddpv,

			    final double dpdtau_aver,
			    final double dpdatau_aver,
			    final double dpdt_aver,
			    final double dtaudt_aver,
			    final double dvdt_aver,
			    final double dpdv_aver,
			    final double dtaudv_aver,
			    final double dttypedt_aver,
			    final double dttypedp_aver,
			    final double dpdvsigned_aver,
			    final double paver2,
			    final double taver2,
			    final double tauaver2,
			    final double vaver2,
			    final double ttype2,
			    final double vsignedaver2,

			    final double ptauaver_normp)
    {
	this.t_range_cut=t_range_cut;
	this.currenttime=currenttime;
	this.norma=norma;
	this.norma_unadj=norma_unadj;
	this.v_aver=v_aver;
	this.tau_V=tau_V;

	this.time_to_aver=time_to_aver;
	this.time_to_aver_2=time_to_aver_2;
	this.time_to_aver_2f=time_to_aver_2f;

	this.ttype_dir_average=ttype_dir_average;
	this.ttype_dir_average2=ttype_dir_average2;
	this.ttype_dir_average2f=ttype_dir_average2f;
	this.dttype_dir=dttype_dir;

	this.tau_average=tau_average;
	this.tau_average2=tau_average2;
	this.tau_average2f=tau_average2f;
	this.dtau=dtau;

	this.dp_average=dp_average;
	this.dp_average2=dp_average2;
	this.dp_average2f=dp_average2f;
	this.ddp=ddp;

	this.dptau_average=dptau_average;
	this.dptau_average2=dptau_average2;
	this.dptau_average2f=dptau_average2f;
	this.ddptau=ddptau;

	this.dpv_average=dpv_average;
	this.dpv_average2=dpv_average2;
	this.dpv_average2f=dpv_average2f;
	this.ddpv=ddpv;

	this.dpdtau_aver=dpdtau_aver;
	this.dpdatau_aver=dpdatau_aver;
	this.dpdt_aver=dpdt_aver;
	this.dtaudt_aver=dtaudt_aver;
	this.dvdt_aver=dvdt_aver;
	this.dpdv_aver=dpdv_aver;
	this.dtaudv_aver=dtaudv_aver;
	this.dttypedt_aver=dttypedt_aver;
	this.dttypedp_aver=dttypedp_aver;
	this.dpdvsigned_aver=dpdvsigned_aver;
	this.paver2=paver2;
	this.taver2=taver2;
	this.tauaver2=tauaver2;
	this.vaver2=vaver2;
	this.ttype2=ttype2;
	this.vsignedaver2=vsignedaver2;

	this.ptauaver_normp=ptauaver_normp;
    }

    double getTauForward()
    {
	return this.tau_average+0.5*this.dtau;
    }

    double getTau()
    {
	return this.tau_average;
    }
}

/** TP2 with scaled volume.
 */
static DAvergeCalculateKendall calculateTradesKendallEndPriceDiffVScaled(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize,
    final double pcurrent_buy,
    final double pcurrent_sell,
    final long vol_scale)
    {
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
	final double midp=0.5*(pcurrent_buy+pcurrent_sell);

	final long norm_fact=vol_scale;

	long vi=0,vbs=0;
	double ws=0,ns=0,w_aver=0,w_aver_unadj=0;
	double ttype_aver=0,ttype_aver_2=0,ttype_aver_2f=0,dttype_diff=0,
	    tau_aver=0,tau_aver_2=0,tau_aver_2f=0,dtau_diff=0,
	    dp_aver=0,dp_aver_2=0,dp_aver_2f=0,ddp_diff=0,
	    dptau_aver=0,dptau_aver_2=0,dptau_aver_2f=0,ddptau_diff=0,
	    dpv_aver=0,dpv_aver_2=0,dpv_aver_2f=0,ddpv_diff=0,
	    time_to_aver=0,time_to_aver_2=0,time_to_aver_2f=0,
	    s_p=0,s_tau=0,s_ptau=0,s_ataup=0,s_p2=0,s_tau2=0,s_pt=0,s_t=0,s_t2=0,s_ttau=0,
	    s_v=0,s_v2=0,s_vt=0,s_pv=0,s_pvs=0,s_vs=0,s_vs2=0,s_tauv=0,
	    s_ttype=0,s_ttype2=0,s_ttypet=0,s_ttypep=0,
	    dvdtp_aver=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares,dt=ti.time-ti.origtime,timetotrade=currenttime-ti.time,xdt,vsigned;
	    final double ttype;
	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		xdt=-dt;
		ttype=-1;
		vsigned=-vvi;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		xdt=dt;
		ttype=1;
		vsigned=vvi;
	    }
	    else
	    {
		continue;
	    }
	    vi+=vvi;

	    final double p=ti.price,w=vvi/((double)(norm_fact+vi)),w_unadj=vvi/((double)vi);
	    final double dp=midp-p,dpw=dp*w,dpv=dp/vi;
	    dttype_diff+=w*(ttype_aver-ttype*w_aver);
	    dtau_diff+=w*(tau_aver-xdt*w_aver);
	    ddp_diff+=w*(dp_aver-dp*w_aver);
	    ddptau_diff+=w*(dptau_aver-xdt*p*w_aver);
	    ddpv_diff+=w*(dpv_aver-dpv*w_aver);
	    ns+=w*w_aver;

	    w_aver+=w;
	    w_aver_unadj+=w_unadj;
	    ttype_aver+=ttype*w;
	    tau_aver+=xdt*w;
	    dp_aver+=dpw;
	    dptau_aver+=xdt*p*w;
	    dpv_aver+=dpv*w;
	    time_to_aver+=timetotrade*w;
	    vbs+=vsigned;
	    dvdtp_aver+=vbs*w/vi;

	    final double nd=w;
	    ws+=nd*w_aver;
	    ttype_aver_2+=nd*ttype*w_aver;
	    ttype_aver_2f+=nd*ttype_aver;
	    tau_aver_2+=nd*xdt*w_aver;
	    tau_aver_2f+=nd*tau_aver;
	    time_to_aver_2+=nd*timetotrade*w_aver;
	    time_to_aver_2f+=nd*time_to_aver;
	    dp_aver_2+=nd*dp*w_aver;
	    dp_aver_2f+=nd*dp_aver;
	    dptau_aver_2+=nd*xdt*p*w_aver;
	    dptau_aver_2f+=nd*dptau_aver;
	    dpv_aver_2+=nd*dpv*w_aver;
	    dpv_aver_2f+=nd*dpv_aver;

	    final double px=p-midp,tx=-timetotrade; // just to reduce rounding errors
	    s_p+=px*w;
	    s_tau+=xdt*w;
	    s_ptau+=px*(xdt*w);
	    s_ataup+=px*(dt*w);
	    s_p2+=px*(px*w);
	    s_pt+=px*(tx*w);
	    s_t+=tx*w;
	    s_t2+=tx*(tx*w);
	    s_ttau+=xdt*(tx*w);
	    s_tau2+=xdt*(xdt*w);
	    s_v+=vvi*w;
	    s_v2+=vvi*(vvi*w);
	    s_vt+=vvi*(tx*w);
	    s_pv+=px*(vvi*w);
	    s_pvs+=px*(vsigned*w);
	    s_vs+=vsigned*w;
	    s_vs2+=vsigned*(vsigned*w);
	    s_tauv+=xdt*(vvi*w);
	    s_ttype+=ttype*w;
	    s_ttype2+=ttype*(ttype*w);
	    s_ttypet+=ttype*(tx*w);
	    s_ttypep+=ttype*(px*w);
	}

	final double normd=ns>0?ns:Double.NaN,norma=w_aver>0?w_aver:Double.NaN,norm2=ws>0?ws:Double.NaN;
	final double pa=s_p/norma,pa2=Math.sqrt(s_p2/norma-pa*pa),
	    taua=s_tau/norma,taua2=Math.sqrt(s_tau2/norma-taua*taua),
	    ta=s_t/norma,ta2=Math.sqrt(s_t2/norma-ta*ta),
	    va=s_v/norma,va2=Math.sqrt(s_v2/norma-va*va),
	    ttypea=s_ttype/norma,ttype2=Math.sqrt(s_ttype2/norma-ttypea*ttypea),
	    vsigneda=s_vs/norma,vsigneda2=Math.sqrt(s_vs2/norma-vsigneda*vsigneda);
	final double dpdtau_aver=s_ptau/norma-pa*taua,
	    dpdatau_aver=s_ataup/norma-pa*taua,
	    dpdt_aver=s_pt/norma-pa*ta,
	    dtaudt_aver=s_ttau/norma-taua*ta,
	    dvdt_aver=s_vt/norma-va*ta,
	    dpdv_aver=s_pv/norma-pa*va,
	    dtaudv_aver=s_tauv/norma-taua*va,
	    dttypedt_aver=s_ttypet/norma-ttypea*ta,
	    dttypedp_aver=s_ttypep/norma-ttypea*pa,
	    dpdvsigned_aver=s_pvs/norma-vsigneda*pa;
	final int ptr_to_v=CalculateLiquidityDeficit9.findPtrToV(currenttime,(long)va,tsize,tb);
	final long tau_V=ptr_to_v>=0?currenttime-tb[ptr_to_v].time:-1;
	return new DAvergeCalculateKendall(t_range_cut,
					   currenttime,
					   norma,
					   w_aver_unadj>0?w_aver_unadj:Double.NaN,
					   va,
					   tau_V,

					   time_to_aver/norma,
					   time_to_aver_2/norm2,
					   time_to_aver_2f/norm2,

					   ttype_aver/norma,
					   ttype_aver_2/norm2,
					   ttype_aver_2f/norm2,
					   dttype_diff/normd,

					   tau_aver/norma,
					   tau_aver_2/norm2,
					   tau_aver_2f/norm2,
					   dtau_diff/normd,

					   dp_aver/norma,
					   dp_aver_2/norm2,
					   dp_aver_2f/norm2,
					   ddp_diff/normd,

					   dptau_aver/norma,
					   dptau_aver_2/norm2,
					   dptau_aver_2f/norm2,
					   ddptau_diff/normd,

					   dpv_aver/norma,
					   dpv_aver_2/norm2,
					   dpv_aver_2f/norm2,
					   ddpv_diff/normd,

					   dpdtau_aver,
					   dpdatau_aver,
					   dpdt_aver,
					   dtaudt_aver,
					   dvdt_aver,
					   dpdv_aver,
					   dtaudv_aver,
					   dttypedt_aver,
					   dttypedp_aver,
					   dpdvsigned_aver,
					   pa2,
					   ta2,
					   taua2,
					   va2,
					   ttype2,
					   vsigneda2,

					   dpdtau_aver/pa2);
    }



static class DDiffCalculateKendall  {
    final long t_range_cut,currenttime;
    final double dpdv,dpdt,dvdtp;
    DDiffCalculateKendall(final long t_range_cut,
			  final long currenttime,
			  final double dpdv,
			  final double dpdt,
			  final double dvdtp)
    {
	this.t_range_cut=t_range_cut;
	this.currenttime=currenttime;
	this.dpdv=dpdv;
	this.dpdt=dpdt;
	this.dvdtp=dvdtp;
    }
}

/** TP2 with scaled volume.
 */
static DDiffCalculateKendall calculateTradesKendallDDiffVScaled(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize,
    final double pcurrent_buy,
    final double pcurrent_sell)
    {
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;

	final double norm_fact=0;

	long vi=0,vi_buy=0,vi_sell=0;
	double ws=0;
	double dpdv=0,dpdt=0,dvdtp=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares;
	    vi+=vvi;
	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		vi_buy+=vvi;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		vi_sell+=vvi;
	    }
	    else
	    {
		continue;
	    }

	    final double pi=ti.price,wi=vvi/(norm_fact+((double)vi));
	    long vj=0,vj_buy=0,vj_sell=0;
	    for(int j=tsize-1;j>i;j--)
	    {
		final Trade tj=tb[j];
		if(tj.origtime<=0)
		{
		    continue;
		}
		vj+=tj.shares;
		if(tj.buy_sell_indicator==BookItem.BOOK_BUY)
		{
		    vj_buy+=tj.shares;
		}
		else if(tj.buy_sell_indicator==BookItem.BOOK_SELL)
		{
		    vj_sell+=tj.shares;
		}
		else
		{
		    continue;
		}
		final double pj=tj.price,wj=tj.shares/(norm_fact+((double)vj)),ww=wi*wj;
		dpdv+=ww*(pj-pi)/(vi-vj);
		dpdt+=ww*(pj-pi)/(tj.time-ti.time);
		dvdtp+=(ww*((vi_sell-vj_sell)-(vi_buy-vj_buy)))/(vi-vj);
		ws+=ww;
		//System.err.println("i="+i+" j="+j+" pi="+pi+" pj="+pj+" dpdv="+dpdv+" vi="+vi+" vj="+vj+" ws="+ws);
	    }
	}
	//System.err.println("Ws="+ws+" dpdv="+dpdv+" dpdt="+dpdt);

	final double norm=ws>0?ws:Double.NaN;
	return new DDiffCalculateKendall(t_range_cut,
					 currenttime,
					 dpdv/norm,
					 dpdt/norm,
					 dvdtp/norm);
    }





static class DPTCalculateKendall  {
    final long t_range_cut,currenttime;
    final double ptaver,sptaver,tau_aver;
    final CalculateKendall pt;
    DPTCalculateKendall(final long t_range_cut,
			final long currenttime,
			final CalculateKendall pt,
			final double ptaver,
			final double sptaver,
			final double tau_aver)
    {
	this.t_range_cut=t_range_cut;
	this.currenttime=currenttime;
	this.pt=pt;
	this.ptaver=ptaver;
	this.sptaver=sptaver;
	this.tau_aver=tau_aver;
    }
}

/** PT2 with scaled volume.
 */
static DPTCalculateKendall calculateTradesKendallPT(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;

	long vi=0,v_cons=0,v_discons=0,v_zero=0;
	double stv_diff=0;
	double vp=0,vt=0,vpt=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    final long vvi=ti.shares,ti_tau=(ti.buy_sell_indicator==BookItem.BOOK_SELL?ti.time-ti.origtime:(-(ti.time-ti.origtime)));
	    vi+=vvi;
	    for(int j=tsize;j-->ptr_interval;)
	    {
		final Trade tj=tb[j];
		if(tj.origtime<=0)
		{
		    continue;
		}
		final int cp=cmpp(tj.price,ti.price);
		if(cp==0)
		{
		    continue;
		}
		final long tj_tau=(tj.buy_sell_indicator==BookItem.BOOK_SELL?tj.time-tj.origtime:(-(tj.time-tj.origtime)));
		final int c_dir=cp*cmpt(tj_tau,ti_tau);
		final long vs=tj.shares*vvi;
		stv_diff+=vs*cp*(tj_tau-ti_tau);
		if(c_dir>0)
		{
		    v_cons+=vs;
		}
		else if(c_dir<0)
		{
		    v_discons+=vs;
		}
		else
		{
		    v_zero+=vs;
		}
	    }
	    final double vvid=vvi;
	    vp+=ti.price*vvid;
	    vt+=ti_tau*vvid;
	    vpt+=ti_tau*ti.price*vvid;
	}

	final double ptaver=vi>0?2*(vpt*vi-vp*vt)/vi/vi:Double.NaN;
	final long sv=v_cons+v_discons+v_zero;
	final double sptaver=sv>0?stv_diff/sv:Double.NaN,vt_aver=vi>0?vt/vi:Double.NaN;
	return new DPTCalculateKendall(t_range_cut,
				       currenttime,
				       new CalculateKendall(currenttime,
							    vi,
							    v_cons/2,
							    v_discons/2,
							    v_zero/2),
				       ptaver,
				       sptaver,
				       vt_aver);
    }


static class DPPCalculateKendall  {
    static final double DEV_FACT=0.5;
    final double DP_AHEAD=0.5;
    // the value (0.29) that makes K3 vanish, but it is small enough
    //final double DP_AHEAD=1.0/Math.sqrt(12);

    final long t_range_cut,currenttime,vi;
    final double normaK1,normaK2,normaK3,
	DK1dp_end,DK1dp_end_var2,DK2dp_end,DK3dp_end,
	DK1dp_mm,DK1dp_mm_var2,DK2dp_mm,DK3dp_mm,
	p_aver,p_aver_var2,DK2p,DK2p_var2,DK3p,DK3p_var2,DK1v,DK2v,DK3v,DK3v_var2,DK1vs,DK2vs,DK3vs,nik1,nik1_var2sq,nik2,nik3,pcpo,pcdp_end,po_var2;
    DPPCalculateKendall(final long t_range_cut,
			final long currenttime,
			final long vi,
			final double normaK1,
			final double normaK2,
			final double normaK3,
			final double DK1dp_end,
			final double DK1dp_end_var2,
			final double DK2dp_end,
			final double DK3dp_end,
			final double DK1dp_mm,
			final double DK1dp_mm_var2,
			final double DK2dp_mm,
			final double DK3dp_mm,
			final double p_aver,
			final double p_aver_var2,
			final double DK2p,
			final double DK2p_var2,
			final double DK3p,
			final double DK3p_var2,
			final double DK1v,
			final double DK2v,
			final double DK3v,
			final double DK3v_var2,
			final double DK1vs,
			final double DK2vs,
			final double DK3vs,
			final double nik1,
			final double nik1_var2sq,
			final double nik2,
			final double nik3,
			final double pcpo,
			final double pcdp_end,
			final double po_var2)
    {
	this.t_range_cut=t_range_cut;
	this.currenttime=currenttime;
	this.vi=vi;
	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;
	this.DK1dp_end=DK1dp_end;
	this.DK1dp_end_var2=DK1dp_end_var2;
	this.DK2dp_end=DK2dp_end;
	this.DK3dp_end=DK3dp_end;
	this.DK1dp_mm=DK1dp_mm;
	this.DK1dp_mm_var2=DK1dp_mm_var2;
	this.DK2dp_mm=DK2dp_mm;
	this.DK3dp_mm=DK3dp_mm;
	this.p_aver=p_aver;
	this.p_aver_var2=p_aver_var2;
	this.DK2p=DK2p;
	this.DK2p_var2=DK2p_var2;
	this.DK3p=DK3p;
	this.DK3p_var2=DK3p_var2;
	this.DK1v=DK1v;
	this.DK2v=DK2v;
	this.DK3v=DK3v;
	this.DK3v_var2=DK3v_var2;
	this.DK1vs=DK1vs;
	this.DK2vs=DK2vs;
	this.DK3vs=DK3vs;
	this.nik1=nik1;
	this.nik1_var2sq=nik1_var2sq;
	this.nik2=nik2;
	this.nik3=nik3;
	this.pcpo=pcpo;
	this.pcdp_end=pcdp_end;
	this.po_var2=po_var2;
    }

    double getDPForward()
    {
	final double dr=DP_AHEAD;
	final double alpha=3*this.DK2dp_end,
	    beta=10*this.DK3dp_end,
	    c0=this.DK1dp_end-beta/12 // k1 renormalization
	    ;
	return c0+dr*(alpha+beta*dr); // second order petrubation theory
    }
    double getDPForward_Overestimated()
    {
	final double dr=DP_AHEAD;
	final double alpha=3*this.DK2dp_mm,
	    beta=10*this.DK3dp_mm,
	    c0=this.DK1dp_mm-beta/12 // k1 renormalization
	    ;
	return c0+dr*(alpha+beta*dr); // second order petrubation theory
    }

    double getLExcess()
    {
	final double rfact=0;  // think later about exact factors.
	if(this.pcpo<rfact)
	{ // liquidity deficit exists
	    if(this.DK1dp_end<0)
	    {
		if(this.DK2dp_end<0 && this.DK3dp_end<0)
		{
		    return this.DK1dp_end;
		}
	    }
	    if(this.DK1dp_end>0)
	    {
		if(this.DK2dp_end>0 && this.DK3dp_end>0)
		{
		    return this.DK1dp_end;
		}
	    }
	}
	return Double.NaN;
    }
}

private static final boolean WEIGHT_HAAR=false;
private static final boolean NO_MM_CALCULATIONS=true;
private static final boolean PRINT_DATA_WARNINGS=false;

/** TP2 with time converted to price.
 */
static DPPCalculateKendall calculateTradesKendallEndDKPPDiffVScaled(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize,
    final long vol_scale)
    {
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;

	final long norm_fact=vol_scale;

	long vi=0,vis=0,prevt=Long.MAX_VALUE;
	double ns1=0,ns2=0,ns3=0;
	double s1dp_end=0,s1dp_end_var2=0,s2dp_end=0,s3dp_adj_end=0,s3dp_end=0;
	double s1dp_mm=0,s1dp_mm_var2=0,s2dp_mm=0,s3dp_adj_mm=0,s3dp_mm=0;
	double sk1nn=0,sk2nn=0,sk3nn_adj=0,sk3nn=0,sk1nn2=0;
	double sp_aver=0,sp_aver_var2=0,s2p=0,s2p_var2=0,s3p_adj=0,s3p=0,s3p_var2=0,s3p_var2_a=0,s3p_var2_b=0;
	double svk1=0,svk2=0,svk3_adj=0,svk3_var2=0,svk3_var2_a=0,svk3_var2_b=0,svk_aver_var2=0,svk3=0,svsk1=0,svsk2=0,svsk3_adj=0,svsk3=0,spp=0,spo=0,spcdp_end=0,spo_var2=0;
	int np=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    if(ti.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted");
	    }
	    prevt=ti.time;
	    final long ti_orig=ti.origtime;
	    if(ti_orig<=0)
	    {
		continue;
	    }
	    final double p_i=ti.price,p_end,p_min,p_max;
	    if(NO_MM_CALCULATIONS)
	    {
		final int ptr_end=find_ptr_t0(ti_orig,tb,0,i+1);
		p_end=ptr_end<0?p_i:tb[ptr_end].price;
		p_min=p_max=Double.NaN;
	    }
	    else
	    {
		double px_end=p_i,px_min=p_i,px_max=p_i;
		for(int j=i;j>=0;j--)
		{
		    final Trade tj=tb[j];
		    if(ti_orig<=tj.time)
		    {
			px_end=tj.price;
			if(!(px_end<px_max))
			{
			    px_max=px_end;
			}
			if(!(px_end>px_min))
			{
			    px_min=px_end;
			}
		    }
		    else
		    {
			break;
		    }
		}
		p_end=px_end;
		p_min=px_min;
		p_max=px_max;
	    }

	    final long vvi=ti.shares;
	    final double dp=p_i-p_end,dp_end,dp_mm;
	    if(ti.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		if(dp<=0)
		{
		    dp_end=dp;
		}
		else
		{
		    dp_end=0;
		    if(PRINT_DATA_WARNINGS||OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
		    {
			System.err.println(
			    "CalculateKendall:calculateTradesKendallEndDKPPDiffVScaled"+
			    " dp="+dp+" ; incorrect sign,"+
			    " possible incorrect ti.origtime="+ti_orig+
			    " for trade="+Trade.getTradeDescription(ti)+
			    " p_end="+p_end+
			    " currenttime="+BookItem.getTimeDiag(currenttime));
		    }
		}
		dp_mm=p_i-p_max;
		vis-=vvi;
	    }
	    else if(ti.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		if(dp>=0)
		{
		    dp_end=dp;
		}
		else
		{
		    dp_end=0;
		    if(PRINT_DATA_WARNINGS||OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
		    {
			System.err.println(
			    "CalculateKendall:calculateTradesKendallEndDKPPDiffVScaled"+
			    " dp="+dp+" ; incorrect sign,"+
			    " possible incorrect ti.origtime="+ti_orig+
			    " for trade="+Trade.getTradeDescription(ti)+
			    " p_end="+p_end+
			    " currenttime="+BookItem.getTimeDiag(currenttime));
		    }
		}
		dp_mm=p_i-p_min;
		vis+=vvi;
	    }
	    else
	    {
		throw new RuntimeException("origtime problem");
	    }
	    vi+=vvi;

	    final double w;
	    if(WEIGHT_HAAR)
	    {
		w=vvi/((double)(norm_fact+vi));
	    }
	    else
	    {
		w=vvi;
	    }
	    s2dp_end+=w*(s1dp_end-dp_end*ns1);
	    s3dp_adj_end+=w*(s1dp_end-2*dp_end*ns1);
	    s2dp_mm+=w*(s1dp_mm-dp_mm*ns1);
	    s3dp_adj_mm+=w*(s1dp_mm-2*dp_mm*ns1);
	    sk2nn+=w*(sk1nn-(tsize-i)*ns1);
	    sk3nn_adj+=w*(sk1nn2-2*(tsize-i)*((tsize-i)*ns1));
	    s2p+=w*(sp_aver-p_i*ns1);
	    s2p_var2+=w*(p_i*(p_i*ns1-sp_aver)+(sp_aver_var2-p_i*sp_aver));

	    //System.err.println("i="+i+" p_i="+p_i+" sp_aver="+sp_aver+" sp_aver_var2="+sp_aver_var2+" ns1="+ns1+" w="+w+" s2p_var2="+s2p_var2);
	    s3p_adj+=w*(sp_aver-2*p_i*ns1);
	    svk2+=w*(svk1-vi*ns1);
	    svk3_adj+=w*(svk1-2*vi*ns1);
	    svsk2+=w*(svsk1-vis*ns1);
	    svsk3_adj+=w*(svsk1-2*vis*ns1);

	    final double prev_ns1=ns1,prev_s1dp_end=s1dp_end,prev_s1dp_mm=s1dp_mm,prev_sk1nn2=sk1nn2,prev_sp_aver=sp_aver,prev_svk1=svk1,prev_svsk1=svsk1,prev_sp_aver_var2=sp_aver_var2,prev_ns2=ns2,prev_svk_aver_var2=svk_aver_var2;

	    ns2+=w*ns1;
	    ns1+=w;
	    ns3+=w*prev_ns1*ns1;

	    s1dp_end+=dp_end*w;
	    s1dp_end_var2+=dp_end*(dp_end*w);
	    s1dp_mm+=dp_mm*w;
	    s1dp_mm_var2+=dp_mm*(dp_mm*w);
	    sk1nn+=(tsize-i)*w;
	    sk1nn2+=(tsize-i)*((tsize-i)*w);
	    sp_aver+=p_i*w;
	    sp_aver_var2+=p_i*(p_i*w);
	    svk1+=vi*w;
	    svk_aver_var2+=vi*(vi*w);
	    svsk1+=vis*w;
	    spcdp_end+=w*dp_end*p_i;
	    final double po=p_i-dp_end;
	    spp+=p_i*w*po;
	    spo+=po*w;
	    spo_var2+=po*w*po;
	    np++;
	    //System.err.println("S: vi="+vi+" w="+w+" i="+i+" po="+(p_i-dp_end)+" p="+p_i+" spp="+spp);

	    s3dp_end+=w*(prev_s1dp_end*ns1+prev_ns1*s1dp_end-2*dp_end*prev_ns1*ns1);
	    s3dp_mm+=w*(prev_s1dp_mm*ns1+prev_ns1*s1dp_mm-2*dp_mm*prev_ns1*ns1);
	    sk3nn+=w*(prev_sk1nn2*ns1+prev_ns1*sk1nn2-2*(tsize-i)*((tsize-i)*prev_ns1)*ns1);
	    s3p+=w*(prev_sp_aver*ns1+prev_ns1*sp_aver-2*p_i*prev_ns1*ns1);
	    s3p_var2+=w*(prev_ns2*p_i*p_i+s3p_var2_a+p_i*s3p_var2_b);
	    //System.err.println("i="+i+" p_i="+p_i+" sp_aver="+sp_aver+" sp_aver_var2="+sp_aver_var2+" ns1="+ns1+" w="+w+" s3p_var2="+s3p_var2);
	    svk3+=w*(prev_svk1*ns1+prev_ns1*svk1-2*vi*prev_ns1*ns1);
	    svk3_var2+=w*((prev_ns2*vi)*vi+svk3_var2_a+vi*svk3_var2_b);

	    svsk3+=w*(prev_svsk1*ns1+prev_ns1*svsk1-2*vis*prev_ns1*ns1);

	    s3p_var2_a+=w*(prev_sp_aver_var2+4*p_i*p_i*prev_ns1-4*p_i*prev_sp_aver);
	    s3p_var2_b+=w*(2*prev_sp_aver-4*p_i*prev_ns1);
	    svk3_var2_a+=w*(prev_svk_aver_var2+4*vi*(vi*prev_ns1)-4*vi*prev_svk1);
	    svk3_var2_b+=w*(2*prev_svk1-4*vi*prev_ns1);
	}
	ns3=ns1*ns2-ns3;
	s3dp_end=ns1*s3dp_adj_end+s1dp_end*ns2-s3dp_end;
	s3dp_mm=ns1*s3dp_adj_mm+s1dp_mm*ns2-s3dp_mm;
	sk3nn=ns1*sk3nn_adj+sk1nn2*ns2-sk3nn;
	s3p=ns1*s3p_adj+sp_aver*ns2-s3p;
	svk3=ns1*svk3_adj+svk1*ns2-svk3;
	svsk3=ns1*svsk3_adj+svsk1*ns2-svsk3;

	final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN;
	final double DK1dp_end=s1dp_end/normaK1,DK1dp_end_var2sq=s1dp_end_var2/normaK1-DK1dp_end*DK1dp_end,DK1dp_end_var2=Math.sqrt(DK1dp_end_var2sq),
	    DK1dp_mm=s1dp_mm/normaK1,DK1dp_mm_var2=Math.sqrt(s1dp_mm_var2/normaK1-DK1dp_mm*DK1dp_mm),
	    norm_np=np>1?np-1:Double.NaN,i_aver=sk1nn/normaK1,i_aver_var2sq=sk1nn2/normaK1-i_aver*i_aver,
	    p_aver=sp_aver/normaK1,p_aver_var2_sq=sp_aver_var2/normaK1-p_aver*p_aver,p_aver_var2=Math.sqrt(p_aver_var2_sq),
	    po_aver=spo/normaK1,
	    pc_cov_po=spp/normaK1-p_aver*po_aver,
	    pcpo=pc_cov_po,
	    pcdp_end=spcdp_end/normaK1-p_aver*DK1dp_end,
	    DK2p=s2p/normaK2,DK3p=s3p/normaK3,DK3v=svk3/normaK3,
	    po_var2=Math.sqrt(spo_var2/normaK1-po_aver*po_aver);
	//System.err.println("S: svk3_var2/normaK3-DK3v*DK3v="+(svk3_var2/normaK3-DK3v*DK3v));

	final DPPCalculateKendall r=
	    new DPPCalculateKendall(t_range_cut,
				    currenttime,
				    vi,
				    normaK1,
				    normaK2,
				    normaK3,
				    DK1dp_end,
				    DK1dp_end_var2,
				    s2dp_end/normaK2,
				    s3dp_end/normaK3,
				    DK1dp_mm,
				    DK1dp_mm_var2,
				    s2dp_mm/normaK2,
				    s3dp_mm/normaK3,
				    p_aver,
				    p_aver_var2,
				    DK2p,
				    Math.sqrt(s2p_var2/normaK2-DK2p*DK2p),
				    DK3p,
				    Math.sqrt(s3p_var2/normaK3-DK3p*DK3p),
				    svk1/normaK1,
				    svk2/normaK2,
				    svk3/normaK3,
				    Math.sqrt(svk3_var2/normaK3-DK3v*DK3v),
				    svsk1/normaK1,
				    svsk2/normaK2,
				    svsk3/normaK3,
				    i_aver/norm_np, // approx 1/2
				    i_aver_var2sq/norm_np/norm_np,
				    -sk2nn/normaK2/norm_np, // approx 1/3
				    sk3nn/normaK3/norm_np/norm_np, // approx 1/10
				    pcpo,
				    pcdp_end,
				    po_var2);

	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{
	    check_DPPCalculateKendall(
		currenttime,
		tb,
		ptr_t_range_cut,
		tsize,
		r);
	}
	return r;
    }


private static int find_ptr_t0(final long t0,
			       final Trade [] trades,
			       final int fromIndex,
			       final int toIndex)
{
    int low=fromIndex;
    int high=toIndex-1;

    for(;low<high;)
    {
	final int mid=(low+high)/2;
	if(t0<=trades[mid].time)
	{
	    high=mid;
	}
	else
	{
	    low=mid+1;
	}
    }
    for(;low<toIndex;)
    {
	if(t0<=trades[low].time)
	{
	    break;
	}
	else
	{
	    low++;
	}
    }
    return low<toIndex?low:-1;
}


private static void check_DPPCalculateKendall(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize,
    final DPPCalculateKendall res)
{
    if(WEIGHT_HAAR)
    {
	throw new RuntimeException("HAAR WEIGHT IS NOT SUPPORTED IN CHECK");
    }
    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
    double ns1=0,ns2=0,ns3=0,ta1=0,ta2=0,ta3=0,sk1_var2=0,sp_aver=0,s2p=0,s3p=0,s2p_var2=0,s3p_var2=0,sp_aver_var2=0,s_i=0,s_ii=0,s_ij=0,s_ijk2=0,s1v=0,s2v=0,s3v=0,s1vs=0,s2vs=0,s3vs=0,spp=0,spo=0,spcdp_end=0,spo_var2=0,svk3_var2=0;
    long vi=0,vis=0;
    int np=0;

    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade ti=tb[i];
	if(ti.origtime<=0)
	{
	    continue;
	}
	final double dp_end_i=finddp(ti.origtime,tb,0,i+1,true);
	final double wi=ti.shares;
	final long vvsi=ti.buy_sell_indicator==BookItem.BOOK_BUY?-ti.shares:ti.shares;
	vi+=ti.shares;
	ns1+=wi;
	ta1+=wi*dp_end_i;
	sk1_var2+=dp_end_i*(dp_end_i*wi);
	sp_aver+=ti.price*wi;
	sp_aver_var2+=ti.price*(ti.price*wi);
	np++;
	s_i+=(tsize-i)*wi;
	s_ii+=(tsize-i)*wi*(tsize-i);
	vis+=vvsi;
	s1v+=vi*wi;
	s1vs+=vis*wi;
	spcdp_end+=wi*dp_end_i*ti.price;
	spp+=wi*ti.price*(ti.price-dp_end_i);
	spo+=wi*(ti.price-dp_end_i);
	spo_var2+=wi*(ti.price-dp_end_i)*(ti.price-dp_end_i);
	//System.err.println("C: vi="+vi+" wi="+wi+" i="+i+" po="+(ti.price-dp_end_i));
	long dvj=-ti.shares,dvsj=-vvsi;
	for(int j=i+1;j<tsize;j++)
	{
	    final Trade tj=tb[j];
	    if(tj.origtime<=0)
	    {
		continue;
	    }
	    final double dp_end_j=finddp(tj.origtime,tb,0,j+1,false);
	    final double wj=tj.shares;
	    final long vvsj=tj.buy_sell_indicator==BookItem.BOOK_BUY?-tj.shares:tj.shares;
	    ns2+=wi*wj;
	    ta2+=wi*wj*(dp_end_j-dp_end_i);
	    s_ij+=wi*wj*(j-i);
	    s2p+=wi*wj*(tj.price-ti.price);
	    s2p_var2+=wi*wj*(tj.price-ti.price)*(tj.price-ti.price);
	    s2v+=wi*wj*dvj;
	    s2vs+=wi*wj*dvsj;
	    long dvk=-tj.shares,dvsk=-vvsj;
	    for(int k=j+1;k<tsize;k++)
	    {
		final Trade tk=tb[k];
		if(tk.origtime<=0)
		{
		    continue;
		}
		final double dp_end_k=finddp(tk.origtime,tb,0,k+1,false);
		final double wk=tk.shares;
		final long vvsk=tk.buy_sell_indicator==BookItem.BOOK_BUY?-tk.shares:tk.shares;
		ns3+=wi*wj*wk;
		ta3+=wi*wj*wk*(dp_end_k+dp_end_i-2*dp_end_j);
		s_ijk2+=wi*(wj*(wk*((tsize-k)*(tsize-k)-2*(tsize-j)*(tsize-j)+(tsize-i)*(tsize-i))));
		s3p+=wi*wj*wk*(tk.price+ti.price-2*tj.price);
		s3p_var2+=wi*wj*wk*(tk.price+ti.price-2*tj.price)*(tk.price+ti.price-2*tj.price);
		s3v+=wi*wj*wk*(dvk-dvj);
		svk3_var2+=(wi*wj*wk*(dvk-dvj))*(dvk-dvj);
		s3vs+=wi*wj*wk*(dvsk-dvsj);
		dvk-=tk.shares;
		dvsk-=vvsk;
	    }
	    dvj-=tj.shares;
	    dvsj-=vvsj;
	}
    }
    final double norm_np=np>1?np-1:Double.NaN;
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN;
    final double k1=ta1/normaK1,k1_var2=Math.sqrt(sk1_var2/normaK1-k1*k1),
	p_aver=sp_aver/normaK1,p_aver_var2=Math.sqrt(sp_aver_var2/normaK1-p_aver*p_aver),
	pcpo=spp/normaK1-p_aver*(p_aver-k1),
	pcdp_end=spcdp_end/normaK1-p_aver*k1,
	po_var2=Math.sqrt(spo_var2/normaK1-(spo/normaK1)*(spo/normaK1)),
	k2p=s2p/normaK2,k3p=s3p/normaK3,
	k2p_var2sq=s2p_var2/normaK2-k2p*k2p,k3p_var2sq=s3p_var2/normaK3-k3p*k3p,
	k2p_var2=k2p_var2sq<0?0:Math.sqrt(k2p_var2sq),k3p_var2=k3p_var2sq<0?0:Math.sqrt(k3p_var2sq);
    //System.err.println("s2p_var2/normaK2-k2p*k2p="+(s2p_var2/normaK2-k2p*k2p));
    final double k2=ta2/normaK2,k3=ta3/normaK3,k1v=s1v/normaK1,k2v=s2v/normaK2,k3v=s3v/normaK3,k3v_var2=Math.sqrt(svk3_var2/normaK3-k3v*k3v),k1vs=s1vs/normaK1,k2vs=s2vs/normaK2,k3vs=s3vs/normaK3;
    final double nnk1=s_i/normaK1,nik1=nnk1/norm_np,nik1_var2sq=(s_ii/normaK1-nnk1*nnk1)/norm_np/norm_np,
	nik2=s_ij/normaK2/norm_np,nik3=s_ijk2/normaK3/norm_np/norm_np;
    if(t_range_cut!=res.t_range_cut||vi!=res.vi)
    {
	throw new RuntimeException("Results mismatch:"+
				   " t_range_cut="+t_range_cut+
				   " res.t_range_cut="+res.t_range_cut+
				   " vi="+vi+
				   " res.vi="+res.vi);
    }
    final double eps=1e-8,eps_var=1e-4;
    UnitTests.checkDouble(normaK1,"normaK1",res.normaK1,"res.normaK1");
    UnitTests.checkDouble(k1,"k1",res.DK1dp_end,"res.DK1dp_end");
    UnitTests.checkDouble(k1_var2,"k1_var2",res.DK1dp_end_var2,"res.DK1dp_end_var2");
    UnitTests.checkDouble(p_aver,"p_aver",res.p_aver,"res.p_aver");
    UnitTests.checkDouble(pcdp_end,"pcdp_end",res.pcdp_end,"res.pcdp_end");
    UnitTests.checkDouble(p_aver_var2,"p_aver_var2",res.p_aver_var2,"res.p_aver_var2");
    UnitTests.checkDouble(pcpo,"pcpo",res.pcpo,"res.pcpo",eps);
    UnitTests.checkDouble(po_var2,"po_var2",res.po_var2,"res.po_var2",eps);
    UnitTests.checkDouble(k1v,"k1v",res.DK1v,"res.DK1v");
    UnitTests.checkDouble(k1vs,"k1vs",res.DK1vs,"res.DK1vs");

    UnitTests.checkDoubleRel(normaK2,"normaK2",res.normaK2,"res.normaK2",eps);
    UnitTests.checkDoubleRel(normaK3,"normaK3",res.normaK3,"res.normaK3",eps);
    UnitTests.checkDouble(k2,"k2",res.DK2dp_end,"res.DK2dp_end",eps);
    UnitTests.checkDouble(k3,"k3",res.DK3dp_end,"res.DK3dp_end",eps);
    UnitTests.checkDouble(k2p,"k2p",res.DK2p,"res.DK2p",eps);
    UnitTests.checkDouble(k2p_var2,"k2p_var2",Double.isNaN(res.DK2p_var2)&&!Double.isNaN(normaK2)?0:res.DK2p_var2,"res.DK2p_var2",eps_var);
    UnitTests.checkDouble(k3p,"k3p",res.DK3p,"res.DK3p",eps);
    UnitTests.checkDouble(k3p_var2,"k3p_var2",Double.isNaN(res.DK3p_var2)&&!Double.isNaN(normaK3)?0:res.DK3p_var2,"res.DK3p_var2",eps_var);
    UnitTests.checkDoubleRel(k2v,"k2v",res.DK2v,"res.DK2v",eps);
    UnitTests.checkDoubleRel(k3v,"k3v",res.DK3v,"res.DK3v",eps);
    UnitTests.checkDoubleRel(k3v_var2,"k3v_var2",res.DK3v_var2,"res.DK3v_var2",eps_var);
    UnitTests.checkDoubleRel(k2vs,"k2vs",res.DK2vs,"res.DK2vs",eps);
    UnitTests.checkDoubleRel(k3vs,"k3vs",res.DK3vs,"res.DK3vs",eps);
    UnitTests.checkDouble(nik1,"nik1",res.nik1,"res.nik1");
    UnitTests.checkDouble(nik1_var2sq,"nik1_var2sq",res.nik1_var2sq,"res.nik1_var2sq");
    UnitTests.checkDouble(nik2,"nik2",res.nik2,"res.nik2",eps);
    UnitTests.checkDouble(nik3,"nik3",res.nik3,"res.nik3",eps);
}

private static double finddp(final long t0,
			     final Trade [] trades,
			     final int fromIndex,
			     final int toIndex,
			     final boolean allow_warnings)
{
    final Trade t=trades[toIndex-1];
    final int ptr=find_ptr_t0(t0,trades,fromIndex,toIndex);
    final double dp=ptr<0?0:t.price-trades[ptr].price;
    if(t.buy_sell_indicator==BookItem.BOOK_BUY)
    {
	if(dp<=0)
	{
	    return dp;
	}
    }
    else if(t.buy_sell_indicator==BookItem.BOOK_SELL)
    {
	if(dp>=0)
	{
	    return dp;
	}
    }
    else
    {
	//return 0;
	throw new RuntimeException("origtime problem");
    }
    if(allow_warnings)
    {
	System.err.println(
	    "CalculateKendall:finddp"+
	    " dp="+dp+" ; incorrect sign,"+
	    " possible incorrect t.origtime="+t.origtime+
	    " for trade="+Trade.getTradeDescription(t)+
	    " p_end="+(ptr<0?Double.NaN:trades[ptr].price)+
	    " t_back="+(ptr<0?"N/A":Trade.getTradeDescription(trades[ptr])));
    }
    return 0;
}


private static double finddp0(final long t0,
			     final Trade [] trades,
			     final int fromIndex,
			     final int toIndex,
			     final boolean allow_warnings)
{
    final Trade t=trades[toIndex-1];
    final int ptr=find_ptr_t0(t0,trades,fromIndex,toIndex);
    final double dp=ptr<0?0:t.price-trades[ptr].price;
    return dp;
}



static class DPPVCalculateKendall  {
    final long t_range_cut,currenttime,vi;
    final double normaK1,normaK2,normaK3,
	dp_end,dp_end_var2,DdpDv_end,D2dpDv2_end,
	p_aver,p_aver_var2;
    DPPVCalculateKendall(final long t_range_cut,
			 final long currenttime,
			 final long vi,
			 final double normaK1,
			 final double normaK2,
			 final double normaK3,
			 final double dp_end,
			 final double dp_end_var2,
			 final double DdpDv_end,
			 final double D2dpDv2_end,
			 final double p_aver,
			 final double p_aver_var2)
    {
	this.t_range_cut=t_range_cut;
	this.currenttime=currenttime;
	this.vi=vi;
	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;
	this.dp_end=dp_end;
	this.dp_end_var2=dp_end_var2;
	this.DdpDv_end=DdpDv_end;
	this.D2dpDv2_end=D2dpDv2_end;
	this.p_aver=p_aver;
	this.p_aver_var2=p_aver_var2;
    }
    double getDPForward()
    {
	final double dv=0.5*this.vi;
	return this.dp_end+dv*this.DdpDv_end+0.5*dv*(dv*this.D2dpDv2_end);
    }
}

/** DPPV.
 */
static DPPVCalculateKendall calculateTradesKendallDPPV(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize)
    {
	final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
	final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;

	long vi=0;
	double ns1=0,ns2=0,ns3=0;
	double sdp_end=0,sdp_end_var2=0,sDdpDv_end=0,sD2dpDv2_end=0;
	double sp_aver=0,sp_aver_var2=0;
	for(int i=tsize;i-->ptr_interval;)
	{
	    final Trade ti=tb[i];
	    if(ti.origtime<=0)
	    {
		continue;
	    }
	    final double dp_end_i=finddp(ti.origtime,tb,0,i+1,PRINT_DATA_WARNINGS||OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS);
	    final double wi=ti.shares;
	    vi+=ti.shares;
	    ns1+=wi;
	    sdp_end+=wi*dp_end_i;
	    sdp_end_var2+=dp_end_i*(dp_end_i*wi);
	    sp_aver+=ti.price*wi;
	    sp_aver_var2+=ti.price*(ti.price*wi);

	    long dvj=0;
	    for(int j=i+1;j<tsize;j++)
	    {
		final Trade tj=tb[j];
		if(tj.origtime<=0)
		{
		    continue;
		}
		dvj+=tj.shares;
		final double dp_end_j=finddp(tj.origtime,tb,0,j+1,false);
		final double wj=tj.shares,dj=(dp_end_j-dp_end_i)/dvj;
		ns2+=wi*wj;
		sDdpDv_end+=wi*(wj*dj);
		long dvk=0;
		for(int k=j+1;k<tsize;k++)
		{
		    final Trade tk=tb[k];
		    if(tk.origtime<=0)
		    {
			continue;
		    }
		    dvk+=tk.shares;
		    final double dp_end_k=finddp(tk.origtime,tb,0,k+1,false);
		    final double wk=tk.shares,dk=(dp_end_k-dp_end_j)/dvk;
		    ns3+=wi*wj*wk;
		    sD2dpDv2_end+=0.5*(wi/(dvj+dvk))*wj*wk*(dk-dj);
		}
	    }
	}
	final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN;
	final double dp_end=sdp_end/normaK1,dp_end_var2=Math.sqrt(sdp_end_var2/normaK1-dp_end*dp_end),
	    p_aver=sp_aver/normaK1,p_aver_var2=Math.sqrt(sp_aver_var2/normaK1-p_aver*p_aver);
	return new DPPVCalculateKendall(t_range_cut,
					currenttime,
					vi,
					normaK1,
					normaK2,
					normaK3,
					dp_end,
					dp_end_var2,
					sDdpDv_end/normaK2,
					sD2dpDv2_end/normaK3,
					p_aver,
					p_aver_var2);
    }


static class BookTradesKStats {
    final long currenttime,vbook,tau_book;
    final double
	p0,
	p_stop,
	pbook_average,
	ptrades_end_book_average,
	ptrades_mm_book_average,
	pbook_average_var2,
	ptrades_end_book_average_var2,
	ptrades_mm_book_average_var2,
	pbook_x_ptrades_end,
	pbook_x_ptrades_mm,
	k2_ptrades_end,
	k2_ptrades_mm,
	k3_ptrades_end,
	k3_ptrades_mm,
	normaK1,
	normaK2,
	normaK3,
	ptrades_end_book_median,
	ptrades_mm_book_median,
	ptrades_end_book_tmax,
	ptrades_mm_book_tmax,
	ptrades_end_book_tmin,
	ptrades_mm_book_tmin,
	k1_ptrades_po,
	k2_ptrades_po,
	k3_ptrades_po;
    final int n_book_levels;

    BookTradesKStats(final long currenttime,
		     final long vbook,
		     final long tau_book,
		     final double p0,
		     final double p_stop,
		     final int n_book_levels,

		     final double pbook_average,
		     final double ptrades_end_book_average,
		     final double ptrades_mm_book_average,

		     final double pbook_average_var2,
		     final double ptrades_end_book_average_var2,
		     final double ptrades_mm_book_average_var2,

		     final double pbook_x_ptrades_end,
		     final double pbook_x_ptrades_mm,
		     final double k2_ptrades_end,
		     final double k2_ptrades_mm,
		     final double k3_ptrades_end,
		     final double k3_ptrades_mm,
		     final double normaK1,
		     final double normaK2,
		     final double normaK3,

		     final double ptrades_end_book_median,
		     final double ptrades_mm_book_median,
		     final double ptrades_end_book_tmax,
		     final double ptrades_mm_book_tmax,
		     final double ptrades_end_book_tmin,
		     final double ptrades_mm_book_tmin,

		     final double k1_ptrades_po,
		     final double k2_ptrades_po,
		     final double k3_ptrades_po)
    {
	this.currenttime=currenttime;
	this.vbook=vbook;
	this.tau_book=tau_book;
	this.p0=p0;
	this.p_stop=p_stop;
	this.n_book_levels=n_book_levels;

	this.pbook_average=pbook_average;
	this.ptrades_end_book_average=ptrades_end_book_average;
	this.ptrades_mm_book_average=ptrades_mm_book_average;
	this.pbook_average_var2=pbook_average_var2;
	this.ptrades_end_book_average_var2=ptrades_end_book_average_var2;
	this.ptrades_mm_book_average_var2=ptrades_mm_book_average_var2;
	this.pbook_x_ptrades_end=pbook_x_ptrades_end;
	this.pbook_x_ptrades_mm=pbook_x_ptrades_mm;
	this.k2_ptrades_end=k2_ptrades_end;
	this.k2_ptrades_mm=k2_ptrades_mm;
	this.k3_ptrades_end=k3_ptrades_end;
	this.k3_ptrades_mm=k3_ptrades_mm;
	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;

	this.ptrades_end_book_median=ptrades_end_book_median;
	this.ptrades_mm_book_median=ptrades_mm_book_median;
	this.ptrades_end_book_tmax=ptrades_end_book_tmax;
	this.ptrades_mm_book_tmax=ptrades_mm_book_tmax;
	this.ptrades_end_book_tmin=ptrades_end_book_tmin;
	this.ptrades_mm_book_tmin=ptrades_mm_book_tmin;

	this.k1_ptrades_po=k1_ptrades_po;
	this.k2_ptrades_po=k2_ptrades_po;
	this.k3_ptrades_po=k3_ptrades_po;
    }

    void checkTauAndV(final long v,final long tau,final int n_levels)
    {
	if(v!=this.vbook||tau!=this.tau_book||n_levels!=this.n_book_levels)
	{
	    throw new RuntimeException("Results mismatch:"+
				       " v="+v+
				       " vbook="+this.vbook+
				       " tau="+tau+
				       " tau_book="+this.tau_book+
				       " n_levels="+n_levels+
				       " n_book_levels="+this.n_book_levels+
				       " currenttime="+this.currenttime);
	}
    }
}

static BookTradesKStats calculateBookTradesKStats(final long currenttime,
						  final BookItem [] orders,
						  final int bsize,
						  final Trade [] trades,
						  final int tsize,
						  final BookItem [] tmpbuffer)
{
    final boolean LOOK_ZONE_EDGE=true;

    final BookItem [] bb=tmpbuffer;
    System.arraycopy(orders,0,bb,0,bsize);
    final double p0,p_stop;
    final byte btype;
    if(bsize>0)
    {
	p0=bb[0].price;
	p_stop=bb[bsize-1].price;
	btype=bb[0].buy_sell_indicator;
    }
    else
    {
	p0=p_stop=Double.NaN;
	btype=BookItem.BOOK_UNKNOWN;
    }
    long vi_book=0;
    int n_book_levels=0;
    double ppb=Double.NaN;
    for(int i=0;i<bsize;i++)
    {
	final BookItem b=bb[i];
	vi_book+=b.shares;
	if(b.buy_sell_indicator!=btype)
	{
	    throw new RuntimeException("Order type mismatch.");
	}
	if(btype==BookItem.BOOK_BUY)
	{
	    if(b.price>ppb)
	    {
		throw new RuntimeException("Ordering is broken.");
	    }
	}
	else if(btype==BookItem.BOOK_SELL)
	{
	    if(b.price<ppb)
	    {
		throw new RuntimeException("Ordering is broken.");
	    }
	}
	else
	{
	    throw new RuntimeException("Book type problem.");
	}
	if(b.price!=ppb)
	{
	    n_book_levels++;
	}
	ppb=b.price;
    }
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(bb,0,bsize-1,CompareItems.cmp_by_time_latest_at_0);

    double p=p0,p_min=p0,p_max=p0,ptrades_end_book_median=p0,ptrades_mm_book_median=p0,
	ptrades_end_book_tmax=p0,ptrades_mm_book_tmax=p0,ptrades_end_book_tmin=p0,ptrades_mm_book_tmin=p0;
    double s_pbook_average=0,s_ptrades_end_book_average=0,s_ptrades_mm_book_average=0,
	s_pbook_average_var2=0,s_ptrades_end_book_average_var2=0,s_ptrades_mm_book_average_var2=0,
	s_pbook_x_ptrades_end=0,s_pbook_x_ptrades_mm=0,s_k2_ptrades_end=0,s_k2_ptrades_mm=0,s_k3_ptrades_end=0,s_k3_ptrades_mm=0,s_k3_ptrades_end_adj=0,s_k3_ptrades_mm_adj=0,s_k1_ptrades_po=0,s_k2_ptrades_po=0,s_k3_ptrades_po_adj=0,s_k3_ptrades_po=0,ns1=0,ns2=0,ns3=0;
    long vi=0,v_med=vi_book/2,v0=0,prevtb=Long.MAX_VALUE,prevtt=Long.MAX_VALUE,stau=0;
    for(int i=0,ptr_trades=tsize-1,ntb=0;i<bsize;i++)
    {
	final BookItem b=bb[i];
	if(b.time>prevtb)
	{
	    throw new RuntimeException("Book ordering by time is broken");
	}
	prevtb=b.time;
	if(LOOK_ZONE_EDGE && ptr_trades>=0 && ptr_trades<(tsize-1) && trades[ptr_trades].buy_sell_indicator!=btype)
	{
	    throw new RuntimeException("pointer error ptr_trades="+ptr_trades+" tsize="+tsize);
	}
	for(;ptr_trades>=0;)
	{ // find relevant trades
	    final Trade t=trades[ptr_trades];
	    if(t.time>prevtt)
	    {
		throw new RuntimeException("Trades ordering by time is broken");
	    }
	    prevtt=t.time;
	    if(LOOK_ZONE_EDGE)
	    {
		if(t.buy_sell_indicator!=btype)
		{ // look at zone edge
		    ptr_trades--;
		    continue;
		}
	    }
	    if(prevtb<=prevtt)
	    {
		p=t.price;
		if(!(p<p_max))
		{
		    p_max=p;
		}
		if(!(p>p_min))
		{
		    p_min=p;
		}
		ptr_trades--;
		ntb++;
	    }
	    else
	    {
		break;
	    }
	} // end find relevant trades

	final long s=b.shares;
	vi+=s;
	stau+=s*((currenttime-prevtb)/ModelParameters.T_SUMM_SCALE);
	final double w=b.shares; //  think later about HAAR weight

	final double p_end,p_mm,p_book=b.price;
	if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	{
	    p_end=p>=p0?p:p0;
	    p_mm=p_max>=p0?p_max:p0;
	}
	else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	{
	    p_end=p<=p0?p:p0;
	    p_mm=p_min<=p0?p_min:p0;
	}
	else
	{
	    throw new RuntimeException("order time problem");
	}

	final double px=p;//ntb>0?p:p_book;


	s_k2_ptrades_end+=w*(s_ptrades_end_book_average-p_end*ns1);
	s_k3_ptrades_end_adj+=w*(s_ptrades_end_book_average-2*p_end*ns1);
	s_k2_ptrades_mm+=w*(s_ptrades_mm_book_average-p_mm*ns1);
	s_k3_ptrades_mm_adj+=w*(s_ptrades_mm_book_average-2*p_mm*ns1);
	s_k2_ptrades_po+=w*(s_k1_ptrades_po-px*ns1);
	s_k3_ptrades_po_adj+=w*(s_k1_ptrades_po-2*px*ns1);

	final double prev_ns1=ns1,
	    prev_s_ptrades_end_book_average=s_ptrades_end_book_average,
	    prev_s_ptrades_mm_book_average=s_ptrades_mm_book_average,
	    prev_s_k1_ptrades_po=s_k1_ptrades_po;
	ns2+=w*ns1;
	ns1+=w;
	ns3+=w*prev_ns1*ns1;
	//System.err.println("AA I="+i+" w="+w+" ns3="+ns3+" p_end="+p_end+" p_mm="+p_mm+" ptr_trades="+ptr_trades+" b.time="+b.time+" t.time="+(ptr_trades>=0?trades[ptr_trades].time:-1));

	s_pbook_average+=w*p_book;
	s_pbook_average_var2+=(w*p_book)*p_book;
	s_ptrades_end_book_average+=w*p_end;
	s_ptrades_end_book_average_var2+=(w*p_end)*p_end;
	s_ptrades_mm_book_average+=w*p_mm;
	s_ptrades_mm_book_average_var2+=(w*p_mm)*p_mm;
	s_k1_ptrades_po+=px*w;
	s_pbook_x_ptrades_end+=(w*p_book)*p_end;
	s_pbook_x_ptrades_mm+=(w*p_book)*p_mm;

	if(vi>=v_med)
	{
	    ptrades_end_book_median=p_end;
	    ptrades_mm_book_median=p_mm;
	    v_med=Long.MAX_VALUE; // catch first match
	}
	if(vi>=v0)
	{
	    ptrades_end_book_tmin=p_end;
	    ptrades_mm_book_tmin=p_mm;
	    v0=Long.MAX_VALUE; // catch first match
	}
	ptrades_end_book_tmax=p_end;
	ptrades_mm_book_tmax=p_mm;

	s_k3_ptrades_end+=w*(prev_s_ptrades_end_book_average*ns1+prev_ns1*s_ptrades_end_book_average-2*p_end*ns1*prev_ns1);
	s_k3_ptrades_mm+=w*(prev_s_ptrades_mm_book_average*ns1+prev_ns1*s_ptrades_mm_book_average-2*p_mm*ns1*prev_ns1);
	s_k3_ptrades_po+=w*(prev_s_k1_ptrades_po*ns1+prev_ns1*s_k1_ptrades_po-2*px*ns1*prev_ns1);
    }
    if(vi_book!=vi)
    {
	throw new RuntimeException("Volume mismatch");
    }
    ns3=ns1*ns2-ns3;
    s_k3_ptrades_end=ns1*s_k3_ptrades_end_adj+s_ptrades_end_book_average*ns2-s_k3_ptrades_end;
    s_k3_ptrades_mm=ns1*s_k3_ptrades_mm_adj+s_ptrades_mm_book_average*ns2-s_k3_ptrades_mm;
    s_k3_ptrades_po=ns1*s_k3_ptrades_po_adj+s_k1_ptrades_po*ns2-s_k3_ptrades_po;
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN,
	pbook_average=s_pbook_average/normaK1,
	ptrades_end_book_average=s_ptrades_end_book_average/normaK1,
	ptrades_mm_book_average=s_ptrades_mm_book_average/normaK1;

    final BookTradesKStats r=new BookTradesKStats(
	    currenttime,
	    vi,
	    vi>0?(stau/vi)*ModelParameters.T_SUMM_SCALE:-1,
	    p0,
	    p_stop,
	    n_book_levels,

	    pbook_average,
	    ptrades_end_book_average,
	    ptrades_mm_book_average,

	    Math.sqrt(s_pbook_average_var2/normaK1-pbook_average*pbook_average),
	    Math.sqrt(s_ptrades_end_book_average_var2/normaK1-ptrades_end_book_average*ptrades_end_book_average),
	    Math.sqrt(s_ptrades_mm_book_average_var2/normaK1-ptrades_mm_book_average*ptrades_mm_book_average),

	    s_pbook_x_ptrades_end/normaK1-pbook_average*ptrades_end_book_average,
	    s_pbook_x_ptrades_mm/normaK1-pbook_average*ptrades_mm_book_average,
	    s_k2_ptrades_end/normaK2,
	    s_k2_ptrades_mm/normaK2,
	    s_k3_ptrades_end/normaK3,
	    s_k3_ptrades_mm/normaK3,
	    normaK1,
	    normaK2,
	    normaK3,

	    ptrades_end_book_median,
	    ptrades_mm_book_median,
	    ptrades_end_book_tmax,
	    ptrades_mm_book_tmax,
	    ptrades_end_book_tmin,
	    ptrades_mm_book_tmin,

	    s_k1_ptrades_po/normaK1,
	    s_k2_ptrades_po/normaK2,
	    s_k3_ptrades_po/normaK3);

    if(!LOOK_ZONE_EDGE && OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // constrains
	if(LOOK_ZONE_EDGE)
	{
	    throw new RuntimeException("Constrains would fail if LOOK_ZONE_EDGE="+LOOK_ZONE_EDGE);
	}
	final BookPrevTradesStats rc=calculateTradesBack(currenttime,
							 orders,
							 bsize,
							 trades,
							 tsize);
	rc.checkTauAndV(r.vbook,r.tau_book);
	final double eps=1e-9,eps_var=5e-5;
	UnitTests.checkDouble(r.p0,"r.p0",rc.p0,"rc.p0");
	UnitTests.checkDouble(r.pbook_average,"r.pbook_average",rc.pbook_average,"rc.pbook_average",eps);
	UnitTests.checkDouble(r.ptrades_end_book_average,"r.ptrades_end_book_average",rc.ptrades_end_book_average,"rc.ptrades_end_book_average",eps);
	UnitTests.checkDouble(Double.isNaN(r.pbook_average_var2)&&!Double.isNaN(normaK1)?0:r.pbook_average_var2,"r.pbook_average_var2",Double.isNaN(rc.pbook_average_var2)&&!Double.isNaN(normaK1)?0:rc.pbook_average_var2,"rc.pbook_average_var2",eps_var);
	UnitTests.checkDouble(Double.isNaN(r.ptrades_end_book_average_var2)&&!Double.isNaN(normaK1)?0:r.ptrades_end_book_average_var2,"r.ptrades_end_book_average_var2",Double.isNaN(rc.ptrades_end_book_average_var2)&&!Double.isNaN(normaK1)?0:rc.ptrades_end_book_average_var2,"rc.ptrades_end_book_average_var2",eps_var);
	UnitTests.checkDoubleRel(r.normaK1,"r.normaK1",rc.normaK1,"rc.normaK1",eps);
	UnitTests.checkDouble(r.pbook_x_ptrades_end,"r.pbook_x_ptrades_end",rc.pbook_x_ptrades_end,"rc.pbook_x_ptrades_end",eps);
	if(!NO_MM_CALCULATIONS)
	{
	    UnitTests.checkDouble(r.ptrades_mm_book_average,"r.ptrades_mm_book_average",rc.ptrades_mm_book_average,"rc.ptrades_mm_book_average",eps);
	    UnitTests.checkDouble(Double.isNaN(r.ptrades_mm_book_average_var2)&&!Double.isNaN(normaK1)?0:r.ptrades_mm_book_average_var2,"r.ptrades_mm_book_average_var2",Double.isNaN(rc.ptrades_mm_book_average_var2)&&!Double.isNaN(normaK1)?0:rc.ptrades_mm_book_average_var2,"rc.ptrades_mm_book_average_var2",eps_var);
	    UnitTests.checkDouble(r.pbook_x_ptrades_mm,"r.pbook_x_ptrades_mm",rc.pbook_x_ptrades_mm,"rc.pbook_x_ptrades_mm",eps);
	}
	check_BookTradesKStatsk1k2k3(currenttime,
				     bb,
				     bsize,
				     trades,
				     tsize,
				     p0,
				     r);
    }

    return r;
}


static void check_BookTradesKStatsk1k2k3(final long currenttime,
					 final BookItem [] orders_sorted,
					 final int bsize,
					 final Trade [] trades,
					 final int tsize,
					 final double p0,
					 final BookTradesKStats res)
{
    final BookItem [] bb=orders_sorted;
    double s1_end=0,s2_end=0,s3_end=0,ns1=0,ns2=0,ns3=0;
    long vi=0,prevt=Long.MAX_VALUE;
    for(int i=0;i<bsize;i++)
    {
	final BookItem bi=bb[i];
	if(bi.time>prevt)
	{
	    throw new RuntimeException("sorting problem");
	}
	prevt=bi.time;
	final int ptr=find_ptr_t0(bi.time,trades,0,tsize);
	final double pi=ptr<0?p0:trades[ptr].price,wi=bi.shares,pi_end;
	if(bi.buy_sell_indicator==BookItem.BOOK_BUY)
	{
	    pi_end=pi>=p0?pi:p0;
	}
	else
	{
	    pi_end=pi<=p0?pi:p0;
	}
	vi+=bi.shares;
	ns1+=wi;
	s1_end+=wi*pi_end;
	//System.err.println("CHECK: I="+i+" wi="+wi+" pi_end="+pi_end+" s1_end="+s1_end+" ptr="+ptr+" bi.time="+bi.time+" t.time="+(ptr>=0?trades[ptr].time:-1)+" tsize="+tsize);
	for(int j=i-1;j>=0;j--)
	{
	    final BookItem bj=bb[j];
	    final int ptrj=find_ptr_t0(bj.time,trades,0,tsize);
	    final double pj=ptrj<0?p0:trades[ptrj].price,wj=bj.shares,pj_end;
	    if(bj.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		pj_end=pj>=p0?pj:p0;
	    }
	    else
	    {
		pj_end=pj<=p0?pj:p0;
	    }
	    ns2+=wj*wi;
	    s2_end+=wj*wi*(pj_end-pi_end);
	    for(int k=j-1;k>=0;k--)
	    {
		final BookItem bk=bb[k];
		final int ptrk=find_ptr_t0(bk.time,trades,0,tsize);
		final double pk=ptrk<0?p0:trades[ptrk].price,wk=bk.shares,pk_end;
		if(bk.buy_sell_indicator==BookItem.BOOK_BUY)
		{
		    pk_end=pk>=p0?pk:p0;
		}
		else
		{
		    pk_end=pk<=p0?pk:p0;
		}
		ns3+=wk*wj*wi;
		s3_end+=wk*wj*wi*(pk_end+pi_end-2*pj_end);
	    }
	}
    }
    if(vi!=res.vbook)
    {
	throw new RuntimeException("vi="+vi+" res.vbook="+res.vbook);
    }
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN;
    final double k1=s1_end/normaK1,k2=s2_end/normaK2,k3=s3_end/normaK3;
    final double eps=1e-7;
    UnitTests.checkDouble(normaK1,"normaK1",res.normaK1,"res.normaK1");
    UnitTests.checkDouble(k1,"k1",res.ptrades_end_book_average,"res.ptrades_end_book_average");
    UnitTests.checkDoubleRel(normaK2,"normaK2",res.normaK2,"res.normaK2",eps);
    UnitTests.checkDoubleRel(normaK3,"normaK3",res.normaK3,"res.normaK3",eps);
    UnitTests.checkDouble(k2,"k2",res.k2_ptrades_end,"res.k2_ptrades_end",eps);
    UnitTests.checkDouble(k3,"k3",res.k3_ptrades_end,"res.k3_ptrades_end",eps);
}



    static class BookPrevTradesStats {
	final long currenttime,vbook,tau_book;
	final double
	    p0,
	    pbook_average,
	    ptrades_end_book_average,
	    ptrades_mm_book_average,

	    pbook_x_ptrades_end,
	    pbook_x_ptrades_mm,

	    pbook_average_var2,
	    ptrades_end_book_average_var2,
	    ptrades_mm_book_average_var2,

	    pbook_ww,
	    ptrades_end_book_ww,
	    ptrades_mm_book_ww,

	    normaK1,
	    normaK2;

	BookPrevTradesStats(final long currenttime,
			    final long vbook,
			    final long tau_book,
			    final double p0,

			    final double pbook_average,
			    final double ptrades_end_book_average,
			    final double ptrades_mm_book_average,

			    final double pbook_x_ptrades_end,
			    final double pbook_x_ptrades_mm,

			    final double pbook_average_var2,
			    final double ptrades_end_book_average_var2,
			    final double ptrades_mm_book_average_var2,

			    final double pbook_ww,
			    final double ptrades_end_book_ww,
			    final double ptrades_mm_book_ww,

			    final double normaK1,
			    final double normaK2)
	{
	    this.currenttime=currenttime;
	    this.vbook=vbook;
	    this.tau_book=tau_book;
	    this.p0=p0;

	    this.pbook_average=pbook_average;
	    this.ptrades_end_book_average=ptrades_end_book_average;
	    this.ptrades_mm_book_average=ptrades_mm_book_average;

	    this.pbook_x_ptrades_end=pbook_x_ptrades_end;
	    this.pbook_x_ptrades_mm=pbook_x_ptrades_mm;

	    this.pbook_average_var2=pbook_average_var2;
	    this.ptrades_end_book_average_var2=ptrades_end_book_average_var2;
	    this.ptrades_mm_book_average_var2=ptrades_mm_book_average_var2;

	    this.pbook_ww=pbook_ww;
	    this.ptrades_end_book_ww=ptrades_end_book_ww;
	    this.ptrades_mm_book_ww=ptrades_mm_book_ww;

	    this.normaK1=normaK1;
	    this.normaK2=normaK2;
	}
	void checkTauAndV(final long v,final long tau)
	{
	    if(v!=this.vbook||tau!=this.tau_book)
	    {
		throw new RuntimeException("Results mismatch:"+
					   " v="+v+
					   " vbook="+this.vbook+
					   " tau="+tau+
					   " tau_book="+this.tau_book+
					   " currenttime="+this.currenttime);
	    }
	}
	double getDPForward_side()
	{
	    return (this.p0-this.ptrades_end_book_average)+(1.0/2)*this.ptrades_end_book_ww;
	}
	double getDPForward_Overestimated_side()
	{
	    return (this.p0-this.ptrades_mm_book_average)+(1.0/2)*this.ptrades_mm_book_ww;
	}
	static double getDP0(final BookPrevTradesStats buy,final BookPrevTradesStats sell)
	{
	    return buy.p0-buy.ptrades_end_book_average+sell.p0-sell.ptrades_end_book_average;
	}
    }

    static BookPrevTradesStats calculateTradesBack(final long currenttime,
						   final BookItem [] orders,
						   final int bsize,
						   final Trade [] tb,
						   final int tsize)
    {
	final double p0=bsize>0?orders[0].price:Double.NaN;

	double s_pbook_average=0,s_ptrades_end_book_average=0,s_ptrades_mm_book_average=0,
	    s_pbook_x_p_end=0,s_pbook_x_p_mm=0,
	    s_pbook_average_var2=0,s_ptrades_end_book_average_var2=0,s_ptrades_mm_book_average_var2=0,
	    ns1=0,ns2=0,s2_p_book=0,s2_p_end=0,s2_p_mm=0;
	long vi=0,stau=0;
	for(int i=0;i<bsize;i++)
	{
	    final BookItem b=orders[i];
	    final long ti_orig=b.time;
	    final double pe,p_min,p_max;
	    if(NO_MM_CALCULATIONS)
	    {
		final int ptr_end=find_ptr_t0(ti_orig,tb,0,tsize);
		pe=ptr_end<0?p0:tb[ptr_end].price;
		p_min=p_max=Double.NaN;
	    }
	    else
	    {
		long prevt=Long.MAX_VALUE;
		double px_end=p0,px_min=p0,px_max=p0;
		for(int j=tsize;j-->0;)
		{
		    final Trade tj=tb[j];
		    if(tj.time>prevt)
		    {
			throw new RuntimeException("Ordering is broken");
		    }
		    prevt=tj.time;
		    if(ti_orig<=tj.time)
		    {
			px_end=tj.price;
			if(!(px_end<px_max))
			{
			    px_max=px_end;
			}
			if(!(px_end>px_min))
			{
			    px_min=px_end;
			}
		    }
		    else
		    {
			break;
		    }
		}
		pe=px_end;
		p_min=px_min;
		p_max=px_max;
	    }

	    final long s=b.shares;
	    vi+=s;
	    stau+=s*((currenttime-ti_orig)/ModelParameters.T_SUMM_SCALE);
	    final double w=b.shares;

	    final double p_mm,p_end,p_book=b.price;
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		p_mm=p_max>=p0?p_max:p0;
		p_end=pe>=p0?pe:p0;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		p_mm=p_min<=p0?p_min:p0;
		p_end=pe<=p0?pe:p0;
	    }
	    else
	    {
		throw new RuntimeException("order type problem");
	    }

	    s2_p_book+=w*(s_pbook_average-p_book*ns1);
	    s2_p_end+=w*(s_ptrades_end_book_average-p_end*ns1);
	    s2_p_mm+=w*(s_ptrades_mm_book_average-p_mm*ns1);

	    ns2+=w*ns1;
	    ns1+=w;

	    s_pbook_average+=w*p_book;
	    s_pbook_average_var2+=(w*p_book)*p_book;
	    s_ptrades_end_book_average+=w*p_end;
	    s_ptrades_end_book_average_var2+=(w*p_end)*p_end;
	    s_ptrades_mm_book_average+=w*p_mm;
	    s_ptrades_mm_book_average_var2+=(w*p_mm)*p_mm;
	    s_pbook_x_p_end+=(w*p_book)*p_end;
	    s_pbook_x_p_mm+=(w*p_book)*p_mm;

	}
	final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN;
	final double
	    pbook_average=s_pbook_average/normaK1,
	    ptrades_end_book_average=s_ptrades_end_book_average/normaK1,
	    ptrades_mm_book_average=s_ptrades_mm_book_average/normaK1;

	return new BookPrevTradesStats(currenttime,
				       vi,
				       vi>0?(stau/vi)*ModelParameters.T_SUMM_SCALE:-1,
				       p0,

				       pbook_average,
				       ptrades_end_book_average,
				       ptrades_mm_book_average,

				       s_pbook_x_p_end/normaK1-pbook_average*ptrades_end_book_average,
				       s_pbook_x_p_mm/normaK1-pbook_average*ptrades_mm_book_average,

				       Math.sqrt(s_pbook_average_var2/normaK1-pbook_average*pbook_average),
				       Math.sqrt(s_ptrades_end_book_average_var2/normaK1-ptrades_end_book_average*ptrades_end_book_average),
				       Math.sqrt(s_ptrades_mm_book_average_var2/normaK1-ptrades_mm_book_average*ptrades_mm_book_average),

				       s2_p_book/normaK2,
				       s2_p_end/normaK2,
				       s2_p_mm/normaK2,

				       normaK1,
				       normaK2);
    }




static class BookTradesBKStats {
    final long currenttime,vbook,tau_book;
    final double
	p0,
	p_stop,
	pbook_average,
	pbook_average_var2,
	pbook_average_k2,
	pbook_average_k3,
	pbook_x_ptrades_po,
	normaK1,
	normaK2,
	normaK3,
	k1_ptrades_po,
	k1_ptrades_po_var2,
	k2_ptrades_po,
	k3_ptrades_po,
	s_k1_ptrades_po,
	s_k1_ptrades_po_var2,
	s_k2_ptrades_po,
	s_k3_ptrades_po,
	s_k2_pbook_average,
	s_k3_pbook_average,
	ns1,
	ns2,
	ns3;
    final int n_book_levels;

    BookTradesBKStats(final long currenttime,
		      final long vbook,
		      final long tau_book,
		      final double p0,
		      final double p_stop,
		      final int n_book_levels,

		      final double pbook_average,
		      final double pbook_average_var2,
		      final double pbook_average_k2,
		      final double pbook_average_k3,
		      final double pbook_x_ptrades_po,

		      final double normaK1,
		      final double normaK2,
		      final double normaK3,

		      final double k1_ptrades_po,
		      final double k1_ptrades_po_var2,
		      final double k2_ptrades_po,
		      final double k3_ptrades_po,

		      final double s_k1_ptrades_po,
		      final double s_k1_ptrades_po_var2,
		      final double s_k2_ptrades_po,
		      final double s_k3_ptrades_po,
		      final double s_k2_pbook_average,
		      final double s_k3_pbook_average,

		      final double ns1,
		      final double ns2,
		      final double ns3)
    {
	this.currenttime=currenttime;
	this.vbook=vbook;
	this.tau_book=tau_book;
	this.p0=p0;
	this.p_stop=p_stop;
	this.n_book_levels=n_book_levels;

	this.pbook_average=pbook_average;
	this.pbook_average_var2=pbook_average_var2;
	this.pbook_average_k2=pbook_average_k2;
	this.pbook_average_k3=pbook_average_k3;
	this.pbook_x_ptrades_po=pbook_x_ptrades_po;

	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;

	this.k1_ptrades_po=k1_ptrades_po;
	this.k1_ptrades_po_var2=k1_ptrades_po_var2;
	this.k2_ptrades_po=k2_ptrades_po;
	this.k3_ptrades_po=k3_ptrades_po;


	this.s_k1_ptrades_po=s_k1_ptrades_po;
	this.s_k1_ptrades_po_var2=s_k1_ptrades_po_var2;
	this.s_k2_ptrades_po=s_k2_ptrades_po;
	this.s_k3_ptrades_po=s_k3_ptrades_po;
	this.s_k2_pbook_average=s_k2_pbook_average;
	this.s_k3_pbook_average=s_k3_pbook_average;

	this.ns1=ns1;
	this.ns2=ns2;
	this.ns3=ns3;
    }

    void checkTauAndV(final long v,final long tau,final int n_levels)
    {
	if(v!=this.vbook||tau!=this.tau_book||n_levels!=this.n_book_levels)
	{
	    throw new RuntimeException("Results mismatch:"+
				       " v="+v+
				       " vbook="+this.vbook+
				       " tau="+tau+
				       " tau_book="+this.tau_book+
				       " n_levels="+n_levels+
				       " n_book_levels="+this.n_book_levels+
				       " currenttime="+this.currenttime);
	}
    }
}

private static final boolean P_ORIG_LAST_BookTradesBKStats=true;

static BookTradesBKStats calculateBookTradesBKStats(final long currenttime,
						    final BookItem [] orders,
						    final int b_start,
						    final int bsize,
						    final Trade [] trades,
						    final int tsize)
{
    final BookItem [] bb=orders;
    final double p0,p_stop;
    final byte btype;
    if(bsize>0)
    {
	p0=bb[0].price;
	p_stop=bb[bsize-1].price;
	btype=bb[0].buy_sell_indicator;
    }
    else
    {
	p0=p_stop=Double.NaN;
	btype=BookItem.BOOK_UNKNOWN;
    }


    double s_pbook_average=0,s_pbook_average_var2=0,s_pbook_x_ptrades_po=0,
	s_k2_pbook_average=0,s_k3_pbook_average_adj=0,s_k3_pbook_average=0,
	s_k1_ptrades_po=0,s_k1_ptrades_po_var2=0,s_k2_ptrades_po=0,s_k3_ptrades_po_adj=0,s_k3_ptrades_po=0,
	ns1=0,ns2=0,ns3=0;
    long vi=0,stau=0,prevtb=-1;
    int n_book_levels=0;
    double p_book=Double.NaN;
    for(int i=b_start;i<bsize;i++)
    {
	final BookItem b=bb[i];
	if(b.buy_sell_indicator!=btype)
	{
	    throw new RuntimeException("Order type mismatch.");
	}
	if(b.price!=p_book)
	{
	    n_book_levels++;
	}
	else if(b.time<prevtb)
	{
	    throw new IllegalArgumentException("Book time ordering is broken.");
	}
	prevtb=b.time;

	if(btype==BookItem.BOOK_BUY)
	{
	    if(b.price>p_book)
	    {
		throw new RuntimeException("Ordering is broken.");
	    }
	}
	else if(btype==BookItem.BOOK_SELL)
	{
	    if(b.price<p_book)
	    {
		throw new RuntimeException("Ordering is broken.");
	    }
	}
	else
	{
	    throw new RuntimeException("Book type problem.");
	}
	p_book=b.price;

	final double p;
	if(P_ORIG_LAST_BookTradesBKStats)
	{
	    p=findporigwithtype(prevtb,p_book,btype,p0,0,tsize,trades,false);
	    //System.err.println("BKOP: i="+i+" p0="+p0+" p_book="+p_book+" btype="+btype+" p="+p);
	}
	else
	{
	    long sw=0;
	    double sp=0;
	    for(int j=tsize;j-->0;)
	    {
		final Trade tj=trades[j];
		if(prevtb<=tj.time)
		{
		    sw+=tj.shares;
		    sp+=tj.price*tj.shares;
		}
		else
		{
		    break;
		}
	    }
	    p=sw>0?sp/sw:p_book;
	}

	final long s=b.shares;
	vi+=s;
	stau+=s*((currenttime-prevtb)/ModelParameters.T_SUMM_SCALE);
	final double w=s; //  think later about HAAR weight

	s_k2_ptrades_po+=w*(s_k1_ptrades_po-p*ns1);
	s_k3_ptrades_po_adj+=w*(s_k1_ptrades_po-2*p*ns1);
	s_k2_pbook_average+=w*(s_pbook_average-p_book*ns1);
	s_k3_pbook_average_adj+=w*(s_pbook_average-2*p_book*ns1);

	final double prev_ns1=ns1,prev_s_k1_ptrades_po=s_k1_ptrades_po,prev_s_pbook_average=s_pbook_average;
	ns2+=w*ns1;
	ns1+=w;
	ns3+=w*prev_ns1*ns1;
	//System.err.println("BKStat: i="+i+" p0="+p0+" w="+w+" btype="+btype+" b.price="+b.price+" p_orig="+p+" b.time="+b.time+" b_start="+b_start+" bsize="+bsize);

	s_pbook_average+=w*p_book;
	s_pbook_average_var2+=(w*p_book)*p_book;
	s_k1_ptrades_po+=p*w;
	s_k1_ptrades_po_var2+=(p*w)*p;

	s_pbook_x_ptrades_po+=(w*p_book)*p;
	s_k3_ptrades_po+=w*((prev_ns1+ns1)*(prev_s_k1_ptrades_po-p*prev_ns1));
	s_k3_pbook_average+=w*((prev_ns1+ns1)*(prev_s_pbook_average-p_book*prev_ns1));
    }
    ns3=ns1*ns2-ns3;
    s_k3_ptrades_po=ns1*s_k3_ptrades_po_adj+s_k1_ptrades_po*ns2-s_k3_ptrades_po;
    s_k3_pbook_average=ns1*s_k3_pbook_average_adj+s_pbook_average*ns2-s_k3_pbook_average;
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN,
	pbook_average=s_pbook_average/normaK1,k1_ptrades_po=s_k1_ptrades_po/normaK1;
    //System.err.println(" s_k2_ptrades_po="+s_k2_ptrades_po+" ns2="+ns2+" b_start="+b_start+" bsize="+bsize);

    final BookTradesBKStats r=new BookTradesBKStats(
	    currenttime,
	    vi,
	    vi>0?(stau/vi)*ModelParameters.T_SUMM_SCALE:-1,
	    p0,
	    p_stop,
	    n_book_levels,

	    pbook_average,
	    Math.sqrt(s_pbook_average_var2/normaK1-pbook_average*pbook_average),
	    s_k2_pbook_average/normaK2,
	    s_k3_pbook_average/normaK3,
	    s_pbook_x_ptrades_po/normaK1-pbook_average*k1_ptrades_po,

	    normaK1,
	    normaK2,
	    normaK3,

	    k1_ptrades_po,
	    Math.sqrt(s_k1_ptrades_po_var2/normaK1-k1_ptrades_po*k1_ptrades_po),
	    s_k2_ptrades_po/normaK2,
	    s_k3_ptrades_po/normaK3,

	    s_k1_ptrades_po,
	    s_k1_ptrades_po_var2,
	    s_k2_ptrades_po,
	    s_k3_ptrades_po,
	    s_k2_pbook_average,
	    s_k3_pbook_average,

	    ns1,
	    ns2,
	    ns3);

    return r;

}



static BookTradesBKStats calculateBookTradesBKStatsIntraLevelRemoved(final long currenttime,
								     final BookItem [] orders,
								     final int [] ptr_book,
								     final int nl,
								     final Trade [] trades,
								     final int tsize)
{
    int nend=0;
    for(int i=0;i<nl;i++)
    {
	if(ptr_book[i]>=0)
	{
	    nend=1+ptr_book[i];
	}
    }

    final BookTradesBKStats ks=calculateBookTradesBKStats(currenttime,
							  orders,
							  0,
							  nend,
							  trades,
							  tsize);
    double ns2=ks.ns2,ns3=ks.ns3,
	s_k2_ptrades_po=ks.s_k2_ptrades_po,
	s_k3_ptrades_po=ks.s_k3_ptrades_po,
	s_k2_pbook_average=ks.s_k2_pbook_average,
	s_k3_pbook_average=ks.s_k3_pbook_average;

    int l=0;
    for(int i=0;i<nl;i++)
    {
	// remove intra-order contributions.
	if(ptr_book[i]>=0)
	{
	    final BookTradesBKStats kl=calculateBookTradesBKStats(currenttime,
								  orders,
								  l,
								  1+ptr_book[i],
								  trades,
								  tsize);
	    //System.err.println(" CRR:i="+i+" l="+l+" kl.normaK2="+kl.normaK2);
	    l=1+ptr_book[i];
	    ns2-=kl.ns2;
	    ns3-=kl.ns3;
	    s_k2_ptrades_po-=kl.s_k2_ptrades_po;
	    s_k3_ptrades_po-=kl.s_k3_ptrades_po;
	    s_k2_pbook_average-=kl.s_k2_pbook_average;
	    s_k3_pbook_average-=kl.s_k3_pbook_average;
	}
	//System.err.println(" RR:i="+i+" l="+l+" ns2="+ns2+" ptr_book[i]="+ptr_book[i]);
    }
    final double normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN;

    final BookTradesBKStats r=new BookTradesBKStats(
	    ks.currenttime,
	    ks.vbook,
	    ks.tau_book,
	    ks.p0,
	    ks.p_stop,
	    ks.n_book_levels,

	    ks.pbook_average,
	    ks.pbook_average_var2,
	    s_k2_pbook_average/normaK2,
	    Double.NaN/normaK3, // s_k3_pbook_average/normaK3, // need more triples to remove
	    ks.pbook_x_ptrades_po,

	    ks.normaK1,
	    normaK2,
	    Double.NaN, // normaK3, // need more triples to remove

	    ks.k1_ptrades_po,
	    ks.k1_ptrades_po_var2,
	    s_k2_ptrades_po/normaK2,
	    Double.NaN/normaK3, // s_k3_ptrades_po/normaK3, // need more triples to remove

	    ks.s_k1_ptrades_po,
	    ks.s_k1_ptrades_po_var2,
	    s_k2_ptrades_po,
	    Double.NaN, // s_k3_ptrades_po, // need more triples to remove
	    s_k2_pbook_average,
	    Double.NaN, // s_k3_pbook_average,  // need more triples to remove

	    ks.ns1,
	    ns2,
	    Double.NaN);

    if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // constrains
	check_calculateBookTradesBKStatsIntraLevelRemoved(currenttime,
							  orders,
							  nend,
							  trades,
							  tsize,
							  r);
    }

    return r;
}


private static void check_calculateBookTradesBKStatsIntraLevelRemoved(final long currenttime,
								      final BookItem [] orders,
								      final int bsize,
								      final Trade [] trades,
								      final int tsize,
								      final BookTradesBKStats res)
{
    final double p0,p_stop;
    final byte btype;
    if(bsize>0)
    {
	p0=orders[0].price;
	p_stop=orders[bsize-1].price;
	btype=orders[0].buy_sell_indicator;
    }
    else
    {
	p0=p_stop=Double.NaN;
	btype=BookItem.BOOK_UNKNOWN;
    }

    double ppb=Double.NaN;
    int n_book_levels=0;

    double s_pbook_average=0,s_pbook_average_var2=0,s_pbook_x_ptrades_po=0;
    double ns1=0,ns2=0,ns3=0,sk1=0,sk1_var2=0,sk2=0,sk3=0,sk2_aver=0,sk3_aver=0;
    long vbook=0,stau=0;
    for(int i=0;i<bsize;i++)
    {
	final BookItem bi=orders[i];
	final double poi=findporigwithtype(bi.time,bi.price,btype,p0,0,tsize,trades,true);
	final double wi=bi.shares;
	//System.err.println("OP: i="+i+" p0="+p0+" wi="+wi+" btype="+btype+" bi.price="+bi.price+" poi="+poi+" bi.time="+bi.time);
	vbook+=bi.shares;
	stau+=bi.shares*((currenttime-bi.time)/ModelParameters.T_SUMM_SCALE);
	s_pbook_average+=wi*bi.price;
	s_pbook_average_var2+=(wi*bi.price)*bi.price;
	s_pbook_x_ptrades_po+=(wi*bi.price)*poi;
	ns1+=wi;
	sk1+=poi*wi;
	sk1_var2+=poi*(poi*wi);
	if(ppb!=bi.price)
	{
	    n_book_levels++;
	}
	ppb=bi.price;

	for(int j=i;j-->0;)
	{
	    final BookItem bj=orders[j];
	    if(bi.price==bj.price)
	    {
		continue;
	    }
	    final double poj=findporigwithtype(bj.time,bj.price,btype,p0,0,tsize,trades,false);
	    final double wj=bj.shares;
	    ns2+=wi*wj;
	    sk2+=wi*wj*(poj-poi);
	    sk2_aver+=wi*wj*(bj.price-bi.price);
	    //System.err.println("i="+i+" j="+j+" wi="+wi+" wj="+wj+" ns2="+ns2+" bi.price="+bi.price+" bj.price="+bj.price+" poi="+poi+" poj="+poj+" contrib="+(wi*wj*(poj-poi)));

	    for(int k=j;k-->0;)
	    {
		final BookItem bk=orders[k];
		if(bj.price==bk.price)
		{
		    continue;
		}
		final double pok=findporigwithtype(bk.time,bk.price,btype,p0,0,tsize,trades,false);
		final double wk=bk.shares;

		ns3+=wi*wj*wk;
		sk3+=wi*wj*wk*(pok-poj+(poi-poj));
		sk3_aver+=wi*wj*wk*(bk.price-bj.price+(bi.price-bj.price));
	    }
	}
    }

    final long tau=vbook>0?(stau/vbook)*ModelParameters.T_SUMM_SCALE:-1;
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN;
    final double k1=sk1/normaK1,k1_var2=Math.sqrt(sk1_var2/normaK1-k1*k1),
	pbook_average=s_pbook_average/normaK1,
	pbook_average_var2=Math.sqrt(s_pbook_average_var2/normaK1-pbook_average*pbook_average),
	pbook_x_ptrades_po=s_pbook_x_ptrades_po/normaK1-pbook_average*k1,
	k2=sk2/normaK2,k3=sk3/normaK3,k2_pbook_average=sk2_aver/normaK2,k3_pbook_average=sk3_aver/normaK3;
    res.checkTauAndV(vbook,tau,n_book_levels);

    final double eps=1e-8;
    UnitTests.checkDouble(normaK1,"normaK1",res.normaK1,"res.normaK1");
    UnitTests.checkDouble(pbook_average,"pbook_average",res.pbook_average,"res.pbook_average");
    UnitTests.checkDouble(pbook_average_var2,"pbook_average_var2",res.pbook_average_var2,"res.pbook_average_var2");
    UnitTests.checkDouble(pbook_x_ptrades_po,"pbook_x_ptrades_po",res.pbook_x_ptrades_po,"res.pbook_x_ptrades_po");
    UnitTests.checkDouble(k1,"k1",res.k1_ptrades_po,"res.k1_ptrades_po");
    UnitTests.checkDouble(k1_var2,"k1_var2",res.k1_ptrades_po_var2,"res.k1_ptrades_po_var2");
    UnitTests.checkDouble(p_stop,"p_stop",res.p_stop,"res.p_stop");


    UnitTests.checkDoubleRel(normaK2,"normaK2",Double.isNaN(res.normaK2)&&!Double.isNaN(normaK2)?0:res.normaK2,"res.normK2",eps);
    UnitTests.checkDouble(k2,"k2",Double.isNaN(res.k2_ptrades_po)&&!Double.isNaN(normaK2)?0:res.k2_ptrades_po,"res.k2_ptrades_po",eps);
    UnitTests.checkDouble(k2_pbook_average,"k2_pbook_average",Double.isNaN(res.pbook_average_k2)&&!Double.isNaN(normaK2)?0:res.pbook_average_k2,"res.pbook_average_k2",eps);

    // the k3 & normaK3 are not calculated yet in calculateBookTradesBKStatsIntraLevelRemoved
    // skip checks for now
    // UnitTests.checkDouble(k3,"k3",res.k3_ptrades_po,"res.k3_ptrades_po",eps);
    // UnitTests.checkDouble(k3_pbook_average,"k3_pbook_average",res.pbook_average_k3,"res.pbook_average_k3",eps);
    // UnitTests.checkDoubleRel(normaK3,"normaK3",res.normaK3,"res.normK3",eps);

}


private static double findporigwithtype(final long t0,
					final double p_order,
					final byte btype,
					final double p0_book,
					final int fromIndex,
					final int toIndex,
					final Trade [] trades,
					final boolean allow_warnings)
{
    int ptr=find_ptr_t0(t0,trades,fromIndex,toIndex);
    if(ptr<0)
    {
	return p0_book;
    }
    double p_orig=Double.NaN;
    boolean btype_reached=false;
    if(btype==BookItem.BOOK_BUY)
    {
	for(;ptr<toIndex;ptr++)
	{
	    if(!(trades[ptr].price>p_orig))
	    {
		p_orig=trades[ptr].price;

	    }
	    if(BookItem.BOOK_BUY==trades[ptr].buy_sell_indicator)
	    {
		btype_reached=true;
		break;
	    }
	}
	if(!btype_reached)
	{
	    p_orig=Double.NaN;
	}
	else if(p_orig>=p_order)
	{
	    return p_orig;
	}
    }
    else if(btype==BookItem.BOOK_SELL)
    {
	for(;ptr<toIndex;ptr++)
	{
	    if(!(trades[ptr].price<p_orig))
	    {
		p_orig=trades[ptr].price;

	    }
	    if(BookItem.BOOK_SELL==trades[ptr].buy_sell_indicator)
	    {
		btype_reached=true;
		break;
	    }
	}
	if(!btype_reached)
	{
	    p_orig=Double.NaN;
	}
	else if(p_orig<=p_order)
	{
	    return p_orig;
	}
    }
    else
    {
	throw new RuntimeException("btype="+btype);
    }

    if(Double.isNaN(p_orig))
    {
	// no matching
	return p0_book;
    }

    if(allow_warnings)
    {
	System.err.println(
	    "CalculateKendall:findporigwithtype"+
	    " incorrect sign, for btype="+btype+
	    " possible incorrect origtime="+t0+
	    " p_orig="+p_orig+
	    " p_order="+p_order+
	    " p0_book="+p0_book+
	    " ptr="+ptr+
	    " fromIndex="+fromIndex+
	    " toIndex="+toIndex);
    }
    return p_order;
}



static class BookKStats {
    final long currenttime,vbook_buy,tau_book_buy,vbook_sell,tau_book_sell;
    final double
	p0_buy,p0_sell,
	p_stop_buy,p_stop_sell,
	normaK1,
	normaK2,
	normaK3,
	k1,
	k1_var2,
	k2,
	k3,
	fact_med;
    final int n_book_levels_buy,n_book_levels_sell;

    BookKStats(final long currenttime,
	       final long vbook_buy,
	       final long tau_book_buy,
	       final double p0_buy,
	       final double p_stop_buy,
	       final int n_book_levels_buy,
	       final long vbook_sell,
	       final long tau_book_sell,
	       final double p0_sell,
	       final double p_stop_sell,
	       final int n_book_levels_sell,

	       final double normaK1,
	       final double normaK2,
	       final double normaK3,

	       final double k1,
	       final double k1_var2,
	       final double k2,
	       final double k3,
	       final double fact_med)
    {
	this.currenttime=currenttime;
	this.vbook_buy=vbook_buy;
	this.tau_book_buy=tau_book_buy;
	this.p0_buy=p0_buy;
	this.p_stop_buy=p_stop_buy;
	this.n_book_levels_buy=n_book_levels_buy;
	this.vbook_sell=vbook_sell;
	this.tau_book_sell=tau_book_sell;
	this.p0_sell=p0_sell;
	this.p_stop_sell=p_stop_sell;
	this.n_book_levels_sell=n_book_levels_sell;

	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;

	this.k1=k1;
	this.k1_var2=k1_var2;
	this.k2=k2;
	this.k3=k3;
	this.fact_med=fact_med;
    }

    void checkTauAndV(final long v_buy,final long tau_buy,final int n_levels_buy,
		      final long v_sell,final long tau_sell,final int n_levels_sell)
    {
	if(v_buy!=this.vbook_buy||tau_buy!=this.tau_book_buy||n_levels_buy!=this.n_book_levels_buy||
	   v_sell!=this.vbook_sell||tau_sell!=this.tau_book_sell||n_levels_sell!=this.n_book_levels_sell)
	{
	    throw new RuntimeException("Results mismatch:"+
				       " v_buy="+v_buy+
				       " vbook_buy="+this.vbook_buy+
				       " tau_buy="+tau_buy+
				       " tau_book_buy="+this.tau_book_buy+
				       " n_levels_buy="+n_levels_buy+
				       " n_book_levels_buy="+this.n_book_levels_buy+
				       " v_sell="+v_sell+
				       " vbook_sell="+this.vbook_sell+
				       " tau_sell="+tau_sell+
				       " tau_book_sell="+this.tau_book_sell+
				       " n_levels_sell="+n_levels_sell+
				       " n_book_levels_sell="+this.n_book_levels_sell+
				       " currenttime="+this.currenttime);
	}
    }
}

static BookKStats calculateBookKStats(final long currenttime,
				      final BookItem [] bbuy,
				      final int bsize_bbuy,
				      final BookItem [] bsell,
				      final int bsize_bsell,
				      final BookItem [] tmp)
{
    final BookItem [] bb=tmp;
    final double p0_buy,p_stop_buy;
    if(bsize_bbuy>0)
    {
	p0_buy=bbuy[0].price;
	p_stop_buy=bbuy[bsize_bbuy-1].price;
    }
    else
    {
	p0_buy=p_stop_buy=Double.NaN;
    }
    final double p0_sell,p_stop_sell;
    if(bsize_bsell>0)
    {
	p0_sell=bsell[0].price;
	p_stop_sell=bsell[bsize_bsell-1].price;
    }
    else
    {
	p0_sell=p_stop_sell=Double.NaN;
    }
    long vbook_buy=0;
    int n_book_levels_buy=0;
    final long tau_book_buy;
    final double wadj_buy;
    if(true)
    {
	long stau=0;
	double p_book=Double.NaN;
	for(int i=0;i<bsize_bbuy;i++)
	{
	    final BookItem b=bbuy[i];
	    vbook_buy+=b.shares;
	    stau+=b.shares*((currenttime-b.time)/ModelParameters.T_SUMM_SCALE);
	    if(b.price!=p_book)
	    {
		n_book_levels_buy++;
	    }
	    p_book=b.price;
	}
	if(vbook_buy>0)
	{
	    tau_book_buy=(stau/vbook_buy)*ModelParameters.T_SUMM_SCALE;
	    wadj_buy=1.0/vbook_buy;
	    //wadj_buy=1;
	}
	else
	{
	    tau_book_buy=-1;
	    wadj_buy=Double.NaN;
	}
    }
    long vbook_sell=0;
    int n_book_levels_sell=0;
    final long tau_book_sell;
    final double wadj_sell;
    if(true)
    {
	long stau=0;
	double p_book=Double.NaN;
	for(int i=0;i<bsize_bsell;i++)
	{
	    final BookItem b=bsell[i];
	    vbook_sell+=b.shares;
	    stau+=b.shares*((currenttime-b.time)/ModelParameters.T_SUMM_SCALE);
	    if(b.price!=p_book)
	    {
		n_book_levels_sell++;
	    }
	    p_book=b.price;
	}
	if(vbook_sell>0)
	{
	    tau_book_sell=(stau/vbook_sell)*ModelParameters.T_SUMM_SCALE;
	    wadj_sell=1.0/vbook_sell;
	    //wadj_sell=1.0;
	}
	else
	{
	    tau_book_sell=-1;
	    wadj_sell=Double.NaN;
	}
    }

    // copy data to temporary array
    System.arraycopy(bbuy,0,bb,0,bsize_bbuy);
    System.arraycopy(bsell,0,bb,bsize_bbuy,bsize_bsell);
    final int bsize=bsize_bbuy+bsize_bsell;
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(bb,0,bsize-1,CompareItems.cmp_by_time_latest_at_0);
    //final double spread=p0_sell-p0_buy;

    double s_k1=0,s_k1_var2=0,s_k2=0,s_k3_adj=0,s_k3=0,ns1=0,ns2=0,ns3=0;
    long vo_buy=0,vo_sell=0;
    double w_med=0.5*(vbook_buy*wadj_buy+vbook_sell*wadj_sell),fact_med=Double.NaN;
    for(int i=bsize;i-->0;)
    {
	final BookItem b=bb[i];
	final double dp,w;
	if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	{
	    dp=-1;//-Math.exp((b.price-p0_buy)/spread); //-1;//b.price/* -p0_buy */;
	    w=b.shares*wadj_buy;
	    vo_buy+=b.shares;
	}
	else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	{
	    dp=1;//Math.exp(-(b.price-p0_sell)/spread);//1;//b.price/* -p0_sell */;
	    w=b.shares*wadj_sell;
	    vo_sell+=b.shares;
	}
	else
	{
	    throw new RuntimeException("Book type problem.");
	}

	s_k2+=w*(s_k1-dp*ns1);
	s_k3_adj+=w*(s_k1-2*dp*ns1);

	final double prev_ns1=ns1,prev_s_k1=s_k1;
	ns2+=w*ns1;
	ns1+=w;
	ns3+=w*prev_ns1*ns1;

	s_k1+=dp*w;
	s_k1_var2+=(dp*w)*dp;

	s_k3+=w*(prev_s_k1*ns1+prev_ns1*s_k1-2*dp*ns1*prev_ns1);
	if(ns1>=w_med)
	{
	    fact_med=(vo_sell*wadj_sell-vo_buy*wadj_buy)/ns1;
	    w_med=Double.NaN;
	}
    }
    ns3=ns1*ns2-ns3;
    s_k3=ns1*s_k3_adj+s_k1*ns2-s_k3;
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN,
	k1=s_k1/normaK1;

    if(vo_buy!=vbook_buy||vo_sell!=vbook_sell)
    {
	throw new RuntimeException("Volume mismatch");
    }

    final BookKStats r=new BookKStats(
	currenttime,
	vbook_buy,
	tau_book_buy,
	p0_buy,
	p_stop_buy,
	n_book_levels_buy,
	vbook_sell,
	tau_book_sell,
	p0_sell,
	p_stop_sell,
	n_book_levels_sell,
	normaK1,
	normaK2,
	normaK3,
	k1,
	Math.sqrt(s_k1_var2/normaK1-k1*k1),
	s_k2/normaK2,
	s_k3/normaK3,
	fact_med);


    return r;

}





static class TradeKStats {
    final long currenttime,t_range_cut,vt_buy,vt_sell,t_pmax,vafter_max,t_pmin,vafter_min;
    final double
	p_last,
	pmax,
	paver_after_max,
	pmin,
	paver_after_min,

	normaK1,
	normaK2,
	normaK3,
	k1,
	k1_var2,
	k2,
	k3,

	k1_dp,
	k1_dp_var2,
	k2_dp,
	k3_dp,

	k1_dp_buy,
	k1_dp_buy_var2,
	k2_dp_buy,
	k3_dp_buy,
	k3d2_dp_buy,
	k3d1_dp_buy,

	k1_dp_sell,
	k1_dp_sell_var2,
	k2_dp_sell,
	k3_dp_sell,
	k3d2_dp_sell,
	k3d1_dp_sell,

	k1_dpa,
	k1_dpa_var2,
	k2_dpa,
	k3_dpa,

	k1v,
	k2v,
	k3v,

	k1tau,
	k2tau,
	k3tau,
	k3d2tau,
	k3d1tau,

	lk1_dp_buy,
	lk2_dp_buy,
	lk3_dp_buy,

	lk1_dp_sell,
	lk2_dp_sell,
	lk3_dp_sell,

	pt_lk1,
	pt_lk2,
	pt_lk3,

	normaCoupled,
	dpt_coupled,

	pt_x_dp_buy,
	pt_x_dp_sell,
	pt_aver,
	pt_aver_var2,
	pt_aver_k2,
	pt_aver_k3,
	pt_aver_k3d2,
	pt_aver_k3d1;

    TradeKStats(final long currenttime,
		final long t_range_cut,
		final long vt_buy,
		final long vt_sell,

		final double normaK1,
		final double normaK2,
		final double normaK3,

		final double p_last,

		final double k1,
		final double k1_var2,
		final double k2,
		final double k3,

		final double k1_dp,
		final double k1_dp_var2,
		final double k2_dp,
		final double k3_dp,

		final double k1_dp_buy,
		final double k1_dp_buy_var2,
		final double k2_dp_buy,
		final double k3_dp_buy,
		final double k3d2_dp_buy,
		final double k3d1_dp_buy,

		final double k1_dp_sell,
		final double k1_dp_sell_var2,
		final double k2_dp_sell,
		final double k3_dp_sell,
		final double k3d2_dp_sell,
		final double k3d1_dp_sell,

		final double k1_dpa,
		final double k1_dpa_var2,
		final double k2_dpa,
		final double k3_dpa,

		final double k1v,
		final double k2v,
		final double k3v,

		final double k1tau,
		final double k2tau,
		final double k3tau,
		final double k3d2tau,
		final double k3d1tau,

		final double normaCoupled,
		final double dpt_coupled,

		final double pt_x_dp_buy,
		final double pt_x_dp_sell,
		final double pt_aver,
		final double pt_aver_var2,
		final double pt_aver_k2,
		final double pt_aver_k3,
		final double pt_aver_k3d2,
		final double pt_aver_k3d1,

		final double lk1_dp_buy,
		final double lk2_dp_buy,
		final double lk3_dp_buy,

		final double lk1_dp_sell,
		final double lk2_dp_sell,
		final double lk3_dp_sell,

		final double pt_lk1,
		final double pt_lk2,
		final double pt_lk3,

		final double pmin,
		final double paver_after_min,
		final long vafter_min,
		final long t_pmin,

		final double pmax,
		final double paver_after_max,
		final long vafter_max,
		final long t_pmax)
    {
	this.currenttime=currenttime;
	this.t_range_cut=t_range_cut;
	this.vt_buy=vt_buy;
	this.vt_sell=vt_sell;

	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;

	this.p_last=p_last;

	this.k1=k1;
	this.k1_var2=k1_var2;
	this.k2=k2;
	this.k3=k3;

	this.k1_dp=k1_dp;
	this.k1_dp_var2=k1_dp_var2;
	this.k2_dp=k2_dp;
	this.k3_dp=k3_dp;

	this.k1_dp_buy=k1_dp_buy;
	this.k1_dp_buy_var2=k1_dp_buy_var2;
	this.k2_dp_buy=k2_dp_buy;
	this.k3_dp_buy=k3_dp_buy;
	this.k3d2_dp_buy=k3d2_dp_buy;
	this.k3d1_dp_buy=k3d1_dp_buy;

	this.k1_dp_sell=k1_dp_sell;
	this.k1_dp_sell_var2=k1_dp_sell_var2;
	this.k2_dp_sell=k2_dp_sell;
	this.k3_dp_sell=k3_dp_sell;
	this.k3d2_dp_sell=k3d2_dp_sell;
	this.k3d1_dp_sell=k3d1_dp_sell;

	this.k1_dpa=k1_dpa;
	this.k1_dpa_var2=k1_dpa_var2;
	this.k2_dpa=k2_dpa;
	this.k3_dpa=k3_dpa;

	this.k1v=k1v;
	this.k2v=k2v;
	this.k3v=k3v;

	this.k1tau=k1tau;
	this.k2tau=k2tau;
	this.k3tau=k3tau;
	this.k3d2tau=k3d2tau;
	this.k3d1tau=k3d1tau;

	this.normaCoupled=normaCoupled;
	this.dpt_coupled=dpt_coupled;

	this.pt_x_dp_buy=pt_x_dp_buy;
	this.pt_x_dp_sell=pt_x_dp_sell;
	this.pt_aver=pt_aver;
	this.pt_aver_var2=pt_aver_var2;
	this.pt_aver_k2=pt_aver_k2;
	this.pt_aver_k3=pt_aver_k3;
	this.pt_aver_k3d2=pt_aver_k3d2;
	this.pt_aver_k3d1=pt_aver_k3d1;


	this.lk1_dp_buy=lk1_dp_buy;
	this.lk2_dp_buy=lk2_dp_buy;
	this.lk3_dp_buy=lk3_dp_buy;

	this.lk1_dp_sell=lk1_dp_sell;
	this.lk2_dp_sell=lk2_dp_sell;
	this.lk3_dp_sell=lk3_dp_sell;

	this.pt_lk1=pt_lk1;
	this.pt_lk2=pt_lk2;
	this.pt_lk3=pt_lk3;

	this.pmin=pmin;
	this.paver_after_min=paver_after_min;
	this.vafter_min=vafter_min;
	this.t_pmin=t_pmin;

	this.pmax=pmax;
	this.paver_after_max=paver_after_max;
	this.vafter_max=vafter_max;
	this.t_pmax=t_pmax;
    }


    double getDbuy()
    {
	return this.k2_dp_buy+0.5*this.k3_dp_buy;
    }
    double getDsell()
    {
	return this.k2_dp_sell+0.5*this.k3_dp_sell;
    }
}


private static final boolean KTT_MM_CALCULATIONS=false;

static TradeKStats calculateTradesKendallTT(final long currenttime,
					    final Trade [] tb,
					    final int ptr_t_range_cut,
					    final int tsize)
{

    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
    final double p_last=tsize>0?tb[tsize-1].price:Double.NaN;

    double s_k1=0,s_k1_var2=0,s_k2=0,s_k3_adj=0,s_k3=0,s_k1_dp=0,s_k1_dp_var2=0,s_k2_dp=0,s_k3_dp_adj=0,s_k3_dp=0,s_k1_dpa=0,s_k1_dpa_var2=0,s_k2_dpa=0,s_k3_dpa_adj=0,s_k3_dpa=0,s_k1_dp_buy=0,s_k1_dp_buy_var2=0,s_k2_dp_buy=0,s_k3_dp_buy_adj=0,s_k3_dp_buy=0,s_k3d2_dp_buy_adj=0,s_k3d2_dp_buy=0,s_k3d1_dp_buy=0,s_k3d1_dp_buy_adj=0,s_k1_dp_sell=0,s_k1_dp_sell_var2=0,s_k2_dp_sell=0,s_k3_dp_sell_adj=0,s_k3_dp_sell=0,s_k3d2_dp_sell_adj=0,s_k3d2_dp_sell=0,s_k3d1_dp_sell_adj=0,s_k3d1_dp_sell=0,ns1=0,ns2=0,ns3=0;
    double s_pt_x_dp_buy=0,s_pt_x_dp_sell=0,s_pt_aver=0,s_pt_aver_var2=0,s_pt_k2=0,s_pt_k3_adj=0,s_pt_k3=0,s_pt_k3d2_adj=0,s_pt_k3d2=0,s_pt_k3d1_adj=0,s_pt_k3d1=0,
	s_dp_coupled=0,s_dp_weight=0,sv_k1=0,sv_k2=0,sv_k3_adj=0,sv_k3=0,stau_k1=0,stau_k2=0,stau_k3_adj=0,stau_k3=0,stau_k3d2_adj=0,stau_k3d2=0,stau_k3d1_adj=0,stau_k3d1=0;
    double sl_k1_dp_buy=0,sl_k2_dp_buy=0,sl_k3_dp_buy_adj=0,sl_k3_dp_buy=0,sl_k1_dp_sell=0,sl_k2_dp_sell=0,sl_k3_dp_sell_adj=0,sl_k3_dp_sell=0,sl_pt_k1=0,sl_pt_k2=0,sl_pt_k3_adj=0,sl_pt_k3=0,nsl1=0,nsl2=0;

    double pmax=Double.NaN,pmin=Double.NaN,s_pt_aver_after_max=0,s_pt_aver_after_min=0;
    long vto_buy=0,vto_sell=0,t_pmax=-1,t_pmin=-1,vafter_max=0,vafter_min=0;

    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade b=tb[i];
	if(b.origtime<=0)
	{
	    continue;
	}
	final byte btype=b.buy_sell_indicator;
	final double p0=b.price;
	final double w=b.shares;
	double p_orig=p0;
	final double dpa;
	if(KTT_MM_CALCULATIONS)
	{
	    final long b_orig=b.origtime;
	    long sw=0;
	    double sp=0;
	    for(int j=i;j-->0;)
	    {
		final Trade tj=tb[j];
		if(b_orig<=tj.time)
		{
		    sw+=tj.shares;
		    sp+=(p0-tj.price)*tj.shares;
		    if(tj.buy_sell_indicator!=btype)
		    {
			continue;
		    }
		    if(btype==BookItem.BOOK_BUY)
		    {
			// max old buy
			if(tj.price>=p_orig)
			{
			    p_orig=tj.price;
			}
		    }
		    else if(btype==BookItem.BOOK_SELL)
		    {
			// min old sell
			if(tj.price<=p_orig)
			{
			    p_orig=tj.price;
			}
		    }
		}
		else
		{
		    break;
		}
	    }
	    s_dp_coupled+=w*sp;
	    s_dp_weight+=w*sw;
	    dpa=sw>0?sp/sw:0;
	}
	else
	{
	    int ptr_orig=find_ptr_t0(b.origtime,tb,0,i+1);
	    if(ptr_orig>=0)
	    {
		for(;ptr_orig<=i;ptr_orig++)
		{
		    if(btype==tb[ptr_orig].buy_sell_indicator)
		    {
			p_orig=tb[ptr_orig].price;
			break;
		    }
		}
	    }
	    dpa=Double.NaN;
	}

	if(!(p0<=pmax))
	{
	    pmax=p0;
	    s_pt_aver_after_max=s_pt_aver;
	    vafter_max=vto_buy+vto_sell;
	    t_pmax=b.time;
	}
	if(!(p0>=pmin))
	{
	    pmin=p0;
	    s_pt_aver_after_min=s_pt_aver;
	    vafter_min=vto_buy+vto_sell;
	    t_pmin=b.time;
	}

	final double dpo,dpo_buy,dpo_sell,tdp;
	if(btype==BookItem.BOOK_BUY)
	{
	    vto_buy+=b.shares;
	    tdp=-1;
	    dpo_buy=dpo=p_orig>=p0?p0-p_orig:0;
	    dpo_sell=0;
	}
	else if(btype==BookItem.BOOK_SELL)
	{
	    vto_sell+=b.shares;
	    tdp=1;
	    dpo_sell=dpo=p_orig<=p0?p0-p_orig:0;
	    dpo_buy=0;
	}
	else
	{
	    continue;
	}
	final long vi=vto_buy+vto_sell;
	final double tau=currenttime-b.time,dpl=p_last-p0;
	final double prev_ns1=ns1,prev_s_k1=s_k1,prev_s_k1_dp=s_k1_dp,prev_s_k1_dpa=s_k1_dpa,prev_s_k1_dp_buy=s_k1_dp_buy,prev_s_k1_dp_sell=s_k1_dp_sell,prev_s_pt_aver=s_pt_aver,prev_sv_k1=sv_k1,prev_stau_k1=stau_k1,
	    prev_nsl1=nsl1,prev_sl_k1_dp_buy=sl_k1_dp_buy,prev_sl_k1_dp_sell=sl_k1_dp_sell,prev_sl_pt_k1=sl_pt_k1;
	ns1+=w;
	nsl1+=dpl*w;
	ns2+=w*prev_ns1;
	nsl2+=w*prev_nsl1;
	ns3+=(w*prev_ns1)*ns1;

	s_k1+=tdp*w;
	s_k1_var2+=(tdp*w)*tdp;
	s_k1_dp+=dpo*w;
	s_k1_dp_var2+=(dpo*w)*dpo;
	s_k1_dpa+=dpa*w;
	s_k1_dpa_var2+=(dpa*w)*dpa;
	s_k1_dp_buy+=dpo_buy*w;
	s_k1_dp_buy_var2+=(dpo_buy*w)*dpo_buy;
	s_k1_dp_sell+=dpo_sell*w;
	s_k1_dp_sell_var2+=(dpo_sell*w)*dpo_sell;
	s_pt_x_dp_buy+=p0*(dpo_buy*w);
	s_pt_x_dp_sell+=p0*(dpo_sell*w);
	s_pt_aver+=p0*w;
	s_pt_aver_var2+=p0*(p0*w);
	sv_k1+=vi*w;
	stau_k1+=tau*w;

	sl_k1_dp_buy+=dpl*(dpo_buy*w);
	sl_k1_dp_sell+=dpl*(dpo_sell*w);
	sl_pt_k1+=dpl*(p0*w);


	s_k2+=w*(prev_s_k1-tdp*prev_ns1);
	s_k3_adj+=w*(prev_s_k1-2*(tdp*prev_ns1));
	s_k2_dp+=w*(prev_s_k1_dp-dpo*prev_ns1);
	s_k3_dp_adj+=w*(prev_s_k1_dp-2*(dpo*prev_ns1));
	s_k2_dpa+=w*(prev_s_k1_dpa-dpa*prev_ns1);
	s_k3_dpa_adj+=w*(prev_s_k1_dpa-2*(dpa*prev_ns1));
	s_k2_dp_buy+=w*(prev_s_k1_dp_buy-dpo_buy*prev_ns1);
	s_k3_dp_buy_adj+=w*(prev_s_k1_dp_buy-2*(dpo_buy*prev_ns1));
	s_k3d2_dp_buy_adj+=w*prev_s_k1_dp_buy;
	s_k3d1_dp_buy_adj+=w*(prev_s_k1_dp_buy+dpo_buy*prev_ns1);
	s_k2_dp_sell+=w*(prev_s_k1_dp_sell-dpo_sell*prev_ns1);
	s_k3_dp_sell_adj+=w*(prev_s_k1_dp_sell-2*(dpo_sell*prev_ns1));
	s_k3d2_dp_sell_adj+=w*prev_s_k1_dp_sell;
	s_k3d1_dp_sell_adj+=w*(prev_s_k1_dp_sell+dpo_sell*prev_ns1);
	s_pt_k2+=w*(prev_s_pt_aver-p0*prev_ns1);
	s_pt_k3_adj+=w*(prev_s_pt_aver-2*(p0*prev_ns1));
	s_pt_k3d2_adj+=w*prev_s_pt_aver;
	s_pt_k3d1_adj+=w*(prev_s_pt_aver+p0*prev_ns1);
	sv_k2+=w*(prev_sv_k1-vi*prev_ns1);
	sv_k3_adj+=w*(prev_sv_k1-(2*vi)*prev_ns1);
	stau_k2+=w*(prev_stau_k1-tau*prev_ns1);
	stau_k3_adj+=w*(prev_stau_k1-2*(tau*prev_ns1));
	stau_k3d2_adj+=w*prev_stau_k1;
	stau_k3d1_adj+=w*(prev_stau_k1+tau*prev_ns1);

	sl_k2_dp_buy+=w*(prev_sl_k1_dp_buy-dpo_buy*prev_nsl1);
	sl_k3_dp_buy_adj+=w*(prev_sl_k1_dp_buy-2*(dpo_buy*prev_nsl1));
	sl_k2_dp_sell+=w*(prev_sl_k1_dp_sell-dpo_sell*prev_nsl1);
	sl_k3_dp_sell_adj+=w*(prev_sl_k1_dp_sell-2*(dpo_sell*prev_nsl1));
	sl_pt_k2+=w*(prev_sl_pt_k1-p0*prev_nsl1);
	sl_pt_k3_adj+=w*(prev_sl_pt_k1-2*(p0*prev_nsl1));

	s_k3+=w*((ns1+prev_ns1)*(prev_s_k1-tdp*prev_ns1));
	s_k3_dp+=w*((ns1+prev_ns1)*(prev_s_k1_dp-dpo*prev_ns1));
	s_k3_dpa+=w*((ns1+prev_ns1)*(prev_s_k1_dpa-dpa*prev_ns1));
	s_k3_dp_buy+=w*((ns1+prev_ns1)*(prev_s_k1_dp_buy-dpo_buy*prev_ns1));
	s_k3d2_dp_buy+=w*w*(prev_s_k1_dp_buy-dpo_buy*prev_ns1);
	s_k3d1_dp_buy+=w*(prev_s_k1_dp_buy*ns1+prev_ns1*s_k1_dp_buy+dpo_buy*prev_ns1*ns1);
	s_k3_dp_sell+=w*((ns1+prev_ns1)*(prev_s_k1_dp_sell-dpo_sell*prev_ns1));
	s_k3d2_dp_sell+=w*w*(prev_s_k1_dp_sell-dpo_sell*prev_ns1);
	s_k3d1_dp_sell+=w*(prev_s_k1_dp_sell*ns1+prev_ns1*s_k1_dp_sell+dpo_sell*prev_ns1*ns1);
	s_pt_k3+=w*((ns1+prev_ns1)*(prev_s_pt_aver-p0*prev_ns1));
	s_pt_k3d2+=w*w*(prev_s_pt_aver-p0*prev_ns1);
	s_pt_k3d1+=w*(prev_s_pt_aver*ns1+prev_ns1*s_pt_aver+p0*prev_ns1*ns1);
	sv_k3+=w*((ns1+prev_ns1)*(prev_sv_k1-vi*prev_ns1));
	stau_k3+=w*((ns1+prev_ns1)*(prev_stau_k1-tau*prev_ns1));
	stau_k3d2+=w*w*(prev_stau_k1-tau*prev_ns1);
	stau_k3d1+=w*(prev_stau_k1*ns1+prev_ns1*stau_k1+tau*prev_ns1*ns1);

	sl_k3_dp_buy+=w*(ns1*(prev_sl_k1_dp_buy-dpo_buy*prev_nsl1)+prev_nsl1*(s_k1_dp_buy-dpo_buy*ns1));
	sl_k3_dp_sell+=w*(ns1*(prev_sl_k1_dp_sell-dpo_sell*prev_nsl1)+prev_nsl1*(s_k1_dp_sell-dpo_sell*ns1));
	sl_pt_k3+=w*(ns1*(prev_sl_pt_k1-p0*prev_nsl1)+prev_nsl1*(s_pt_aver-p0*ns1));
    }
    ns3=ns1*ns2-ns3;
    s_k3=ns1*s_k3_adj+s_k1*ns2-s_k3;
    s_k3_dp=ns1*s_k3_dp_adj+s_k1_dp*ns2-s_k3_dp;
    s_k3_dpa=ns1*s_k3_dpa_adj+s_k1_dpa*ns2-s_k3_dpa;
    s_k3_dp_buy=ns1*s_k3_dp_buy_adj+s_k1_dp_buy*ns2-s_k3_dp_buy;
    s_k3d2_dp_buy=ns1*s_k3d2_dp_buy_adj-s_k1_dp_buy*ns2-s_k3d2_dp_buy;
    s_k3d1_dp_buy=ns1*s_k3d1_dp_buy_adj+s_k1_dp_buy*ns2-s_k3d1_dp_buy;
    s_k3_dp_sell=ns1*s_k3_dp_sell_adj+s_k1_dp_sell*ns2-s_k3_dp_sell;
    s_k3d2_dp_sell=ns1*s_k3d2_dp_sell_adj-s_k1_dp_sell*ns2-s_k3d2_dp_sell;
    s_k3d1_dp_sell=ns1*s_k3d1_dp_sell_adj+s_k1_dp_sell*ns2-s_k3d1_dp_sell;
    s_pt_k3=ns1*s_pt_k3_adj+s_pt_aver*ns2-s_pt_k3;
    s_pt_k3d2=ns1*s_pt_k3d2_adj-s_pt_aver*ns2-s_pt_k3d2;
    s_pt_k3d1=ns1*s_pt_k3d1_adj+s_pt_aver*ns2-s_pt_k3d1;
    sv_k3=ns1*sv_k3_adj+sv_k1*ns2-sv_k3;
    stau_k3=ns1*stau_k3_adj+stau_k1*ns2-stau_k3;
    stau_k3d2=ns1*stau_k3d2_adj-stau_k1*ns2-stau_k3d2;
    stau_k3d1=ns1*stau_k3d1_adj+stau_k1*ns2-stau_k3d1;
    sl_k3_dp_buy=ns1*sl_k3_dp_buy_adj+s_k1_dp_buy*nsl2-sl_k3_dp_buy;
    sl_k3_dp_sell=ns1*sl_k3_dp_sell_adj+s_k1_dp_sell*nsl2-sl_k3_dp_sell;
    sl_pt_k3=ns1*sl_pt_k3_adj+s_pt_aver*nsl2-sl_pt_k3;

    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN,
	normaCoupled=s_dp_weight>0?s_dp_weight:Double.NaN,
	k1=s_k1/normaK1,k1_dp=s_k1_dp/normaK1,k1_dpa=s_k1_dpa/normaK1,k1_dp_buy=s_k1_dp_buy/normaK1,k1_dp_sell=s_k1_dp_sell/normaK1,pt_aver=s_pt_aver/normaK1;



    final TradeKStats r=new TradeKStats(currenttime,
					t_range_cut,
					vto_buy,
					vto_sell,

					normaK1,
					normaK2,
					normaK3,

					p_last,

					k1,
				       	Math.sqrt(s_k1_var2/normaK1-k1*k1),
					s_k2/normaK2,
					s_k3/normaK3,

					k1_dp,
				       	Math.sqrt(s_k1_dp_var2/normaK1-k1_dp*k1_dp),
					s_k2_dp/normaK2,
					s_k3_dp/normaK3,

					k1_dp_buy,
				       	Math.sqrt(s_k1_dp_buy_var2/normaK1-k1_dp_buy*k1_dp_buy),
					s_k2_dp_buy/normaK2,
					s_k3_dp_buy/normaK3,
					0.5*s_k3d2_dp_buy/normaK3,
					(1.0/3)*s_k3d1_dp_buy/normaK3,

					k1_dp_sell,
				       	Math.sqrt(s_k1_dp_sell_var2/normaK1-k1_dp_sell*k1_dp_sell),
					s_k2_dp_sell/normaK2,
					s_k3_dp_sell/normaK3,
					0.5*s_k3d2_dp_sell/normaK3,
					(1.0/3)*s_k3d1_dp_sell/normaK3,

					k1_dpa,
				       	Math.sqrt(s_k1_dpa_var2/normaK1-k1_dpa*k1_dpa),
					s_k2_dpa/normaK2,
					s_k3_dpa/normaK3,

					sv_k1/normaK1,
					sv_k2/normaK2,
					sv_k3/normaK3,

					stau_k1/normaK1,
					stau_k2/normaK2,
					stau_k3/normaK3,
					0.5*stau_k3d2/normaK3,
					(1.0/3)*stau_k3d1/normaK3,

					normaCoupled,
					s_dp_coupled/normaCoupled,

					s_pt_x_dp_buy/normaK1,
					s_pt_x_dp_sell/normaK1,
					pt_aver,
					Math.sqrt(s_pt_aver_var2/normaK1-pt_aver*pt_aver),
					s_pt_k2/normaK2,
					s_pt_k3/normaK3,
					0.5*s_pt_k3d2/normaK3,
					(1.0/3)*s_pt_k3d1/normaK3,


					sl_k1_dp_buy/normaK1,
					sl_k2_dp_buy/normaK2,
					sl_k3_dp_buy/normaK3,

					sl_k1_dp_sell/normaK1,
					sl_k2_dp_sell/normaK2,
					sl_k3_dp_sell/normaK3,

					sl_pt_k1/normaK1,
					sl_pt_k2/normaK2,
					sl_pt_k3/normaK3,

					pmin,
					s_pt_aver_after_min/(vafter_min>0?vafter_min:Double.NaN),
					vafter_min,
					t_pmin,

					pmax,
					s_pt_aver_after_max/(vafter_max>0?vafter_max:Double.NaN),
					vafter_max,
					t_pmax);

    if(!KTT_MM_CALCULATIONS && OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // constrains
	check_TradeKStats(currenttime,tb,ptr_t_range_cut,tsize,r);
    }

    return r;

}


private static double finddpwithtype(final Trade [] trades,
				     final int fromIndex,
				     final int toIndex,
				     final boolean allow_warnings)
{
    final Trade t=trades[toIndex-1];
    final double p0=t.price;
    final byte btype=t.buy_sell_indicator;
    int ptr=find_ptr_t0(t.origtime,trades,fromIndex,toIndex);
    double p_orig=p0;
    if(ptr>=0)
    {
	for(;ptr<toIndex;ptr++)
	{
	    if(btype==trades[ptr].buy_sell_indicator)
	    {
		p_orig=trades[ptr].price;
		break;
	    }
	}
    }

    final double dp=p0-p_orig;
    if(btype==BookItem.BOOK_BUY)
    {
	if(dp<=0)
	{
	    return dp;
	}
    }
    else if(btype==BookItem.BOOK_SELL)
    {
	if(dp>=0)
	{
	    return dp;
	}
    }
    else
    {
	throw new RuntimeException("origtime problem");
	//return 0;
    }
    if(allow_warnings)
    {
	System.err.println(
	    "CalculateKendall:finddpwithtype"+
	    " dp="+dp+" ; incorrect sign,"+
	    " possible incorrect t.origtime="+t.origtime+
	    " for trade="+Trade.getTradeDescription(t)+
	    " fromIndex="+fromIndex+
	    " toIndex="+toIndex+
	    " p0="+p0+
	    " p_orig="+p_orig);
    }
    return 0;
}



private static void check_TradeKStats(
    final long currenttime,
    final Trade [] tb,
    final int ptr_t_range_cut,
    final int tsize,
    final TradeKStats res)
{
    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
    final double p_last=tsize>0?tb[tsize-1].price:Double.NaN;

    double ns1=0,ns2=0,ns3=0,sk1=0,sk1_var2=0,sk2=0,sk3=0,sk1_dp=0,sk1_dp_var2=0,sk2_dp=0,sk3_dp=0,sk3d2_dp=0,sk3d1_dp=0;
    double spt_k1=0,spt_k2=0,spt_k3=0,spt_k3d2=0,spt_k3d1=0;
    double s1v=0,s2v=0,s3v=0,s1tau=0,s2tau=0,s3tau=0,s3d2tau=0,s3d1tau=0;
    double sl_k1_dp=0,sl_k2_dp=0,sl_k3_dp=0,sl_pt_k1=0,sl_pt_k2=0,sl_pt_k3=0;

    long vi=0;
    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade ti=tb[i];
	if(ti.origtime<=0)
	{
	    continue;
	}
	final double tdpi=ti.buy_sell_indicator==BookItem.BOOK_BUY?-1:1;
	final double dp_end_i=finddpwithtype(tb,0,i+1,true),taui=currenttime-ti.time;
	final double wi=ti.shares;
	vi+=ti.shares;
	ns1+=wi;
	sk1+=tdpi*wi;
	sk1_var2+=tdpi*(tdpi*wi);
	sk1_dp+=wi*dp_end_i;
	sk1_dp_var2+=dp_end_i*(dp_end_i*wi);
	spt_k1+=wi*ti.price;
	s1v+=vi*wi;
	s1tau+=taui*wi;

	sl_k1_dp+=(p_last-ti.price)*(wi*dp_end_i);
	sl_pt_k1+=(p_last-ti.price)*(wi*ti.price);

	long dvj=-ti.shares;
	for(int j=i+1;j<tsize;j++)
	{
	    final Trade tj=tb[j];
	    if(tj.origtime<=0)
	    {
		continue;
	    }
	    final double tdpj=tj.buy_sell_indicator==BookItem.BOOK_BUY?-1:1;
	    final double dp_end_j=finddpwithtype(tb,0,j+1,false),tauj=currenttime-tj.time;
	    final double wj=tj.shares;
	    ns2+=wi*wj;
	    sk2+=wi*wj*(tdpj-tdpi);
	    sk2_dp+=wi*wj*(dp_end_j-dp_end_i);
	    spt_k2+=wi*wj*(tj.price-ti.price);
	    s2v+=wi*wj*dvj;
	    s2tau+=wi*wj*(tauj-taui);

	    sl_k2_dp+=(p_last-tj.price)*wi*wj*(dp_end_j-dp_end_i);
	    sl_pt_k2+=(p_last-tj.price)*wi*wj*(tj.price-ti.price);

	    long dvk=-tj.shares;
	    for(int k=j+1;k<tsize;k++)
	    {
		final Trade tk=tb[k];
		if(tk.origtime<=0)
		{
		    continue;
		}
		final double tdpk=tk.buy_sell_indicator==BookItem.BOOK_BUY?-1:1;
		final double dp_end_k=finddpwithtype(tb,0,k+1,false),tauk=currenttime-tk.time;
		final double wk=tk.shares;

		ns3+=wi*wj*wk;
		sk3+=wi*wj*wk*(tdpk+tdpi-2*tdpj);
		sk3_dp+=wi*wj*wk*(dp_end_k+dp_end_i-2*dp_end_j);
		sk3d2_dp+=wi*wj*wk*(dp_end_k-dp_end_i);
		sk3d1_dp+=wi*wj*wk*(dp_end_k+dp_end_i+dp_end_j);
		s3v+=wi*wj*wk*(dvk-dvj);
		s3tau+=wi*wj*wk*(tauk-tauj+(taui-tauj));
		s3d2tau+=wi*wj*wk*(tauk-taui);
		s3d1tau+=wi*wj*wk*(tauk+taui+tauj);
		spt_k3+=wi*wj*wk*(tk.price-tj.price+(ti.price-tj.price));
		spt_k3d2+=wi*wj*wk*(tk.price-ti.price);
		spt_k3d1+=wi*wj*wk*(tk.price+ti.price+tj.price);

		sl_k3_dp+=(p_last-tk.price)*wi*wj*wk*(dp_end_k+dp_end_i-2*dp_end_j);
		sl_pt_k3+=(p_last-tk.price)*wi*wj*wk*(tk.price+ti.price-2*tj.price);

		dvk-=tk.shares;
	    }
	    dvj-=tj.shares;
	}
    }
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN;
    final double k1=sk1/normaK1,k1_var2=Math.sqrt(sk1_var2/normaK1-k1*k1),
	k1_dp=sk1_dp/normaK1,k1_dp_var2=Math.sqrt(sk1_dp_var2/normaK1-k1_dp*k1_dp),
	k2=sk2/normaK2,k2_dp=sk2_dp/normaK2,
	k3=sk3/normaK3,k3_dp=sk3_dp/normaK3,k3d2_dp=0.5*sk3d2_dp/normaK3,k3d1_dp=(1.0/3)*sk3d1_dp/normaK3,
	pt_aver=spt_k1/normaK1,pt_aver_k2=spt_k2/normaK2,pt_aver_k3=spt_k3/normaK3,pt_aver_k3d2=0.5*spt_k3d2/normaK3,pt_aver_k3d1=(1.0/3)*spt_k3d1/normaK3,
	k1v=s1v/normaK1,k2v=s2v/normaK2,k3v=s3v/normaK3,
	k1tau=s1tau/normaK1,k2tau=s2tau/normaK2,k3tau=s3tau/normaK3,k3d2tau=0.5*s3d2tau/normaK3,k3d1tau=(1.0/3)*s3d1tau/normaK3,
	lk1_dp=sl_k1_dp/normaK1,lk2_dp=sl_k2_dp/normaK2,lk3_dp=sl_k3_dp/normaK3,
	pt_lk1=sl_pt_k1/normaK1,pt_lk2=sl_pt_k2/normaK2,pt_lk3=sl_pt_k3/normaK3;

    if(t_range_cut!=res.t_range_cut||vi!=(res.vt_buy+res.vt_sell))
    {
	throw new RuntimeException("Results mismatch:"+
				   " t_range_cut="+t_range_cut+
				   " res.t_range_cut="+res.t_range_cut+
				   " vi="+vi+
				   " res.vt_buy+res.vt_sell="+(res.vt_buy+res.vt_sell));
    }
    final double eps=1e-8;//,eps_var=1e-4;
    UnitTests.checkDouble(normaK1,"normaK1",res.normaK1,"res.normaK1");
    UnitTests.checkDouble(k1,"k1",res.k1,"res.k1");
    UnitTests.checkDouble(k1_var2,"k1_var2",res.k1_var2,"res.k1_var2");
    UnitTests.checkDouble(k1_dp,"k1_dp",res.k1_dp,"res.k1_dp");
    UnitTests.checkDouble(k1_dp_var2,"k1_dp_var2",res.k1_dp_var2,"res.k1_dp_var2");
    UnitTests.checkDouble(pt_aver,"pt_aver",res.pt_aver,"res.pt_aver");
    UnitTests.checkDouble(k1v,"k1v",res.k1v,"res.k1v");
    UnitTests.checkDouble(k1tau,"k1tau",res.k1tau,"res.k1tau");
    UnitTests.checkDouble(pt_lk1,"pt_lk1",res.pt_lk1,"res.pt_lk1");


    UnitTests.checkDoubleRel(normaK2,"normaK2",res.normaK2,"res.normaK2",eps);
    UnitTests.checkDoubleRel(normaK3,"normaK3",res.normaK3,"res.normaK3",eps);
    UnitTests.checkDouble(k2,"k2",res.k2,"res.k2",eps);
    UnitTests.checkDouble(k3,"k3",res.k3,"res.k3",eps);
    UnitTests.checkDouble(k2_dp,"k2_dp",res.k2_dp,"res.k2_dp",eps);
    UnitTests.checkDouble(k3_dp,"k3_dp",res.k3_dp,"res.k3_dp",eps);

    UnitTests.checkDouble(pt_aver_k2,"pt_aver_k2",res.pt_aver_k2,"res.pt_aver_k2",eps);
    UnitTests.checkDouble(pt_aver_k3,"pt_aver_k3",res.pt_aver_k3,"res.pt_aver_k3",eps);
    UnitTests.checkDouble(pt_aver_k3d2,"pt_aver_k3d2",res.pt_aver_k3d2,"res.pt_aver_k3d2",eps);
    UnitTests.checkDouble(pt_aver_k3d1,"pt_aver_k3d1",res.pt_aver_k3d1,"res.pt_aver_k3d1",eps);

    UnitTests.checkDoubleRel(k2v,"k2v",res.k2v,"res.k2v",eps);
    UnitTests.checkDoubleRel(k3v,"k3v",res.k3v,"res.k3v",eps);
    UnitTests.checkDoubleRel(k2tau,"k2tau",res.k2tau,"res.k2tau",eps);
    UnitTests.checkDoubleRel(k3tau,"k3tau",res.k3tau,"res.k3tau",eps);
    UnitTests.checkDoubleRel(k3d2tau,"k3d2tau",res.k3d2tau,"res.k3d2tau",eps);
    UnitTests.checkDoubleRel(k3d1tau,"k3d1tau",res.k3d1tau,"res.k3d1tau",eps);
    UnitTests.checkDouble(pt_lk2,"pt_lk2",res.pt_lk2,"res.pt_lk2",eps);
    UnitTests.checkDouble(pt_lk3,"pt_lk3",res.pt_lk3,"res.pt_lk3",eps);


    UnitTests.checkDouble(k1_dp,"k1_dp",res.k1_dp_buy+res.k1_dp_sell,"res.k1_dp_buy+res.k1_dp_sell",eps);
    UnitTests.checkDouble(k1_dp_var2*k1_dp_var2,"k1_dp_var2^2",res.k1_dp_buy_var2*res.k1_dp_buy_var2+res.k1_dp_sell_var2*res.k1_dp_sell_var2-2*res.k1_dp_buy*res.k1_dp_sell,"res.k1_dp_buy_var2^2+res.k1_dp_sell_var2^2-2*res.k1_dp_buy*res.k1_dp_sell",eps);
    UnitTests.checkDouble(k2_dp,"k2_dp",res.k2_dp_buy+res.k2_dp_sell,"res.k2_dp_buy+res.k2_dp_sell",eps);
    UnitTests.checkDouble(k3_dp,"k3_dp",res.k3_dp_buy+res.k3_dp_sell,"res.k3_dp_buy+res.k3_dp_sell",eps);
    UnitTests.checkDouble(k3d2_dp,"k3d2_dp",res.k3d2_dp_buy+res.k3d2_dp_sell,"res.k3d2_dp_buy+res.k3d2_dp_sell",eps);
    UnitTests.checkDouble(k3d1_dp,"k3d1_dp",res.k3d1_dp_buy+res.k3d1_dp_sell,"res.k3d1_dp_buy+res.k3d1_dp_sell",eps);

    UnitTests.checkDouble(lk1_dp,"lk1_dp",res.lk1_dp_buy+res.lk1_dp_sell,"res.lk1_dp_buy+res.lk1_dp_sell",eps);
    UnitTests.checkDouble(lk2_dp,"lk2_dp",res.lk2_dp_buy+res.lk2_dp_sell,"res.lk2_dp_buy+res.lk2_dp_sell",eps);
    UnitTests.checkDouble(lk3_dp,"lk3_dp",res.lk3_dp_buy+res.lk3_dp_sell,"res.lk3_dp_buy+res.lk3_dp_sell",eps);

}




static class TradeCoupledStats {
    final long currenttime,t_range_cut,vt_buy,vt_sell;
    final double
	normaK1,
	normaK2,

	k1c,
	k2c,

	k1c_buy,
	k2c_buy,

	k1c_sell,
	k2c_sell,

	n1_buy,
	n2_buy,
	n1_sell,
	n2_sell;


    TradeCoupledStats(final long currenttime,
		      final long t_range_cut,
		      final long vt_buy,
		      final long vt_sell,

		      final double normaK1,
		      final double normaK2,

		      final double k1c,
		      final double k2c,

		      final double k1c_buy,
		      final double k2c_buy,

		      final double k1c_sell,
		      final double k2c_sell,

		      final double n1_buy,
		      final double n2_buy,
		      final double n1_sell,
		      final double n2_sell)
    {
	this.currenttime=currenttime;
	this.t_range_cut=t_range_cut;
	this.vt_buy=vt_buy;
	this.vt_sell=vt_sell;

	this.normaK1=normaK1;
	this.normaK2=normaK2;

	this.k1c=k1c;
	this.k2c=k2c;

	this.k1c_buy=k1c_buy;
	this.k2c_buy=k2c_buy;

	this.k1c_sell=k1c_sell;
	this.k2c_sell=k2c_sell;

	this.n1_buy=n1_buy;
	this.n2_buy=n2_buy;
	this.n1_sell=n1_sell;
	this.n2_sell=n2_sell;
    }
}

static TradeCoupledStats calculateTradeCoupledStats(final long currenttime,
						    final Trade [] tb,
						    final int ptr_t_range_cut,
						    final int tsize)
{

    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;

    double s_k1c_buy=0,s_k2c_buy=0,s_k1c_sell=0,s_k2c_sell=0,np1_buy=0,np1_sell=0,np2_buy=0,np2_sell=0,ns1=0,ns2=0;
    long vto_buy=0,vto_sell=0;
    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade b=tb[i];
	final double torigtime=b.origtime;
	if(torigtime<=0)
	{
	    continue;
	}
	final byte btype=b.buy_sell_indicator;
	final double w=b.shares;
	final double p0=b.price;

	// unoptimized version, we need extra storage to replace extra loop by binary search and taking difference
	double ws=0,wsdp=0;
	for(int j=i;j-->0;)
	{
	    final Trade tj=tb[j];
	    if(tj.time<torigtime)
	    {
		break;
	    }
	    if(false || btype==tj.buy_sell_indicator)
	    {
		final double wj=tj.shares;
		ws+=wj;
		wsdp+=wj*(p0-tj.price);
	    }
	}

	if(btype==BookItem.BOOK_BUY)
	{
	    vto_buy+=b.shares;

	    s_k1c_buy+=w*p0;
	    s_k2c_buy+=w*wsdp;
	    np1_buy+=w;
	    np2_buy+=w*ws;
	}
	else if(btype==BookItem.BOOK_SELL)
	{
	    vto_sell+=b.shares;

	    s_k1c_sell+=w*p0;
	    s_k2c_sell+=w*wsdp;
	    np1_sell+=w;
	    np2_sell+=w*ws;
	}
	else
	{
	    continue;
	}

	ns2+=w*ns1;
	ns1+=w;
    }

    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN;


    final TradeCoupledStats r=new TradeCoupledStats(currenttime,
						    t_range_cut,
						    vto_buy,
						    vto_sell,

						    normaK1,
						    normaK2,

						    (s_k1c_buy+s_k1c_sell)/normaK1,
						    (s_k2c_buy+s_k2c_sell)/normaK2,

						    s_k1c_buy/normaK1,
						    s_k2c_buy/normaK2,

						    s_k1c_sell/normaK1,
						    s_k2c_sell/normaK2,


						    np1_buy,
						    np2_buy,
						    np1_sell,
						    np2_sell);
    return r;

}




static class TradeKT2Stats {
    final long currenttime,t_range_cut,vt_buy,vt_sell;
    final double

	normaK1,
	normaK2,
	normaK3,

	k1_dp_buy,
	k1_dp_buy_var2,
	k2_dp_buy,
	k3_dp_buy,

	k1_dp_sell,
	k1_dp_sell_var2,
	k2_dp_sell,
	k3_dp_sell,

	pt_x_dp_buy,
	pt_x_dp_sell,
	pt_aver,
	pt_aver_var2,
	pt_aver_k2,
	pt_aver_k3,
	k1v,
	k2v,
	k3v;

    final double pmin,paver_after_min,pmax,paver_after_max;
    final long vafter_min,t_pmin,vafter_max,t_pmax;

    TradeKT2Stats(final long currenttime,
		  final long t_range_cut,
		  final long vt_buy,
		  final long vt_sell,

		  final double normaK1,
		  final double normaK2,
		  final double normaK3,

		  final double k1_dp_buy,
		  final double k1_dp_buy_var2,
		  final double k2_dp_buy,
		  final double k3_dp_buy,

		  final double k1_dp_sell,
		  final double k1_dp_sell_var2,
		  final double k2_dp_sell,
		  final double k3_dp_sell,

		  final double k1v,
		  final double k2v,
		  final double k3v,

		  final double pt_x_dp_buy,
		  final double pt_x_dp_sell,
		  final double pt_aver,
		  final double pt_aver_var2,
		  final double pt_aver_k2,
		  final double pt_aver_k3,

		  final double pmin,
		  final double paver_after_min,
		  final long vafter_min,
		  final long t_pmin,

		  final double pmax,
		  final double paver_after_max,
		  final long vafter_max,
		  final long t_pmax)
    {
	this.currenttime=currenttime;
	this.t_range_cut=t_range_cut;
	this.vt_buy=vt_buy;
	this.vt_sell=vt_sell;

	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;

	this.k1_dp_buy=k1_dp_buy;
	this.k1_dp_buy_var2=k1_dp_buy_var2;
	this.k2_dp_buy=k2_dp_buy;
	this.k3_dp_buy=k3_dp_buy;

	this.k1_dp_sell=k1_dp_sell;
	this.k1_dp_sell_var2=k1_dp_sell_var2;
	this.k2_dp_sell=k2_dp_sell;
	this.k3_dp_sell=k3_dp_sell;

	this.k1v=k1v;
	this.k2v=k2v;
	this.k3v=k3v;

	this.pt_x_dp_buy=pt_x_dp_buy;
	this.pt_x_dp_sell=pt_x_dp_sell;
	this.pt_aver=pt_aver;
	this.pt_aver_var2=pt_aver_var2;
	this.pt_aver_k2=pt_aver_k2;
	this.pt_aver_k3=pt_aver_k3;

	this.pmin=pmin;
	this.paver_after_min=paver_after_min;
	this.vafter_min=vafter_min;
	this.t_pmin=t_pmin;

	this.pmax=pmax;
	this.paver_after_max=paver_after_max;
	this.vafter_max=vafter_max;
	this.t_pmax=t_pmax;

    }
}


static TradeKT2Stats calculateTradesKendallTT2(final long currenttime,
					       final Trade [] tb,
					       final int ptr_t_range_cut,
					       final int tsize)
{

    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;

    double s_k1_dp_buy=0,s_k1_dp_buy_var2=0,s_k2_dp_buy=0,s_k3_dp_buy_adj=0,s_k3_dp_buy=0,
	s_k1_dp_sell=0,s_k1_dp_sell_var2=0,s_k2_dp_sell=0,s_k3_dp_sell_adj=0,s_k3_dp_sell=0,ns1=0,ns2=0,ns3=0;
    double s_pt_x_dp_buy=0,s_pt_x_dp_sell=0,s_pt_aver=0,s_pt_aver_var2=0,s_pt_k2=0,s_pt_k3_adj=0,s_pt_k3=0;
    double sv_k1=0,sv_k2=0,sv_k3_adj=0,sv_k3=0;

    double pmax=Double.NaN,pmin=Double.NaN,s_pt_aver_after_max=0,s_pt_aver_after_min=0;

    long vto_buy=0,vto_sell=0,t_pmax=-1,t_pmin=-1,vafter_max=0,vafter_min=0;
    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade b=tb[i];
	if(b.origtime<=0)
	{
	    continue;
	}
	final byte btype=b.buy_sell_indicator;
	final double p0=b.price;
	double p_orig=p0,p_last=p0;
	if(true)
	{
	    int ptr_orig=find_ptr_t0(b.origtime,tb,0,i+1);
	    // scan back
	    for(;ptr_orig-->0;)
	    {
		if(btype==tb[ptr_orig].buy_sell_indicator)
		{
		    p_orig=tb[ptr_orig].price;
		    break;
		}
	    }

	    final byte btlast=btype==BookItem.BOOK_BUY?BookItem.BOOK_SELL:BookItem.BOOK_BUY;
	    ptr_orig=i;
	    for(;ptr_orig-->0;)
	    {
		if(btlast==tb[ptr_orig].buy_sell_indicator)
		{
		    p_last=tb[ptr_orig].price;
		    break;
		}
	    }
	}
	final double w=b.shares;
	if(!(p0<=pmax))
	{
	    pmax=p0;
	    s_pt_aver_after_max=s_pt_aver;
	    vafter_max=vto_buy+vto_sell;
	    t_pmax=b.time;
	}
	if(!(p0>=pmin))
	{
	    pmin=p0;
	    s_pt_aver_after_min=s_pt_aver;
	    vafter_min=vto_buy+vto_sell;
	    t_pmin=b.time;
	}

	final double dpo_buy,dpo_sell;
	if(btype==BookItem.BOOK_BUY)
	{
	    vto_buy+=b.shares;
	    dpo_buy=p0-p_orig; // orig price
	    dpo_sell=p0-p_last; // spread
	}
	else if(btype==BookItem.BOOK_SELL)
	{
	    vto_sell+=b.shares;
	    dpo_sell=p0-p_orig; // orig price
	    dpo_buy=p0-p_last; // spread
	}
	else
	{
	    continue;
	}
	final double prev_ns1=ns1,prev_s_k1_dp_buy=s_k1_dp_buy,prev_s_k1_dp_sell=s_k1_dp_sell,prev_s_pt_aver=s_pt_aver,prev_sv_k1=sv_k1;
	final long vi=vto_buy+vto_sell;
	ns1+=w;
	ns2+=w*prev_ns1;
	ns3+=(w*prev_ns1)*ns1;

	s_k1_dp_buy+=dpo_buy*w;
	s_k1_dp_buy_var2+=(dpo_buy*w)*dpo_buy;
	s_k1_dp_sell+=dpo_sell*w;
	s_k1_dp_sell_var2+=(dpo_sell*w)*dpo_sell;
	sv_k1+=vi*w;
	s_pt_x_dp_buy+=p0*(dpo_buy*w);
	s_pt_x_dp_sell+=p0*(dpo_sell*w);
	s_pt_aver+=p0*w;
	s_pt_aver_var2+=p0*(p0*w);

	s_k2_dp_buy+=w*(prev_s_k1_dp_buy-dpo_buy*prev_ns1);
	s_k3_dp_buy_adj+=w*(prev_s_k1_dp_buy-2*(dpo_buy*prev_ns1));
	s_k2_dp_sell+=w*(prev_s_k1_dp_sell-dpo_sell*prev_ns1);
	s_k3_dp_sell_adj+=w*(prev_s_k1_dp_sell-2*(dpo_sell*prev_ns1));
	s_pt_k2+=w*(prev_s_pt_aver-p0*prev_ns1);
	s_pt_k3_adj+=w*(prev_s_pt_aver-2*(p0*prev_ns1));
	sv_k2+=w*(prev_sv_k1-vi*prev_ns1);
	sv_k3_adj+=w*(prev_sv_k1-(2*vi)*prev_ns1);

	s_k3_dp_buy+=w*(ns1*(prev_s_k1_dp_buy-dpo_buy*prev_ns1)+prev_ns1*(s_k1_dp_buy-dpo_buy*ns1));
	s_k3_dp_sell+=w*(ns1*(prev_s_k1_dp_sell-dpo_sell*prev_ns1)+prev_ns1*(s_k1_dp_sell-dpo_sell*ns1));
	s_pt_k3+=w*(ns1*(prev_s_pt_aver-p0*prev_ns1)+prev_ns1*(s_pt_aver-p0*ns1));
	sv_k3+=w*((ns1+prev_ns1)*(prev_sv_k1-vi*prev_ns1));
    }
    ns3=ns1*ns2-ns3;
    s_k3_dp_buy=ns1*s_k3_dp_buy_adj+s_k1_dp_buy*ns2-s_k3_dp_buy;
    s_k3_dp_sell=ns1*s_k3_dp_sell_adj+s_k1_dp_sell*ns2-s_k3_dp_sell;
    s_pt_k3=ns1*s_pt_k3_adj+s_pt_aver*ns2-s_pt_k3;
    sv_k3=ns1*sv_k3_adj+sv_k1*ns2-sv_k3;
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN,
	k1_dp_buy=s_k1_dp_buy/normaK1,k1_dp_sell=s_k1_dp_sell/normaK1,pt_aver=s_pt_aver/normaK1;



    final TradeKT2Stats r=new TradeKT2Stats(currenttime,
					    t_range_cut,
					    vto_buy,
					    vto_sell,

					    normaK1,
					    normaK2,
					    normaK3,

					    k1_dp_buy,
					    Math.sqrt(s_k1_dp_buy_var2/normaK1-k1_dp_buy*k1_dp_buy),
					    s_k2_dp_buy/normaK2,
					    s_k3_dp_buy/normaK3,

					    k1_dp_sell,
					    Math.sqrt(s_k1_dp_sell_var2/normaK1-k1_dp_sell*k1_dp_sell),
					    s_k2_dp_sell/normaK2,
					    s_k3_dp_sell/normaK3,

					    sv_k1/normaK1,
					    sv_k2/normaK2,
					    sv_k3/normaK3,

					    s_pt_x_dp_buy/normaK1,
					    s_pt_x_dp_sell/normaK1,
					    pt_aver,
					    Math.sqrt(s_pt_aver_var2/normaK1-pt_aver*pt_aver),
					    s_pt_k2/normaK2,
					    s_pt_k3/normaK3,

					    pmin,
					    s_pt_aver_after_min/(vafter_min>0?vafter_min:Double.NaN),
					    vafter_min,
					    t_pmin,

					    pmax,
					    s_pt_aver_after_max/(vafter_max>0?vafter_max:Double.NaN),
					    vafter_max,
					    t_pmax);
    if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // constrains
	// check_TradeKStats(currenttime,tb,ptr_t_range_cut,tsize,r);
    }

    return r;

}


static class BookTradesK2Stats {
    final long currenttime,vbook,tau_book;
    final double
	p0,
	p_stop,
	pbook_average,
	pbook_average_var2,
	d1,
	d2,
	d3,
	dk2,
	norma,
	k1v,
	k2v,
	k3v;

    BookTradesK2Stats(final long currenttime,
		      final long vbook,
		      final long tau_book,
		      final double p0,
		      final double p_stop,

		      final double pbook_average,
		      final double pbook_average_var2,

		      final double d1,
		      final double d2,
		      final double d3,
		      final double dk2,
		      final double norma,
		      final double k1v,
		      final double k2v,
		      final double k3v)
    {
	this.currenttime=currenttime;
	this.vbook=vbook;
	this.tau_book=tau_book;
	this.p0=p0;
	this.p_stop=p_stop;

	this.pbook_average=pbook_average;
	this.pbook_average_var2=pbook_average_var2;
	this.d1=d1;
	this.d2=d2;
	this.d3=d3;
	this.dk2=dk2;
	this.norma=norma;
	this.k1v=k1v;
	this.k2v=k2v;
	this.k3v=k3v;
    }

    void checkTauAndV(final long v,final long tau)
    {
	if(v!=this.vbook||tau!=this.tau_book)
	{
	    throw new RuntimeException("Results mismatch:"+
				       " v="+v+
				       " vbook="+this.vbook+
				       " tau="+tau+
				       " tau_book="+this.tau_book+
				       " currenttime="+this.currenttime);
	}
    }
}



static BookTradesK2Stats calculateBookTradesK2Stats(final long currenttime,
						    final BookItem [] orders,
						    final int bsize,
						    final Trade [] trades,
						    final int tsize,
						    final BookItem [] tmpbuffer)
{
    final BookItem [] bb=tmpbuffer;
    System.arraycopy(orders,0,bb,0,bsize);
    final double p0,p_stop;
    if(bsize>0)
    {
	p0=bb[0].price;
	p_stop=bb[bsize-1].price;
    }
    else
    {
	p0=p_stop=Double.NaN;
    }
    com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(bb,0,bsize-1,CompareItems.cmp_by_time);

    long vi_book=0,stau=0,prevtb=Long.MAX_VALUE,prevtt=Long.MAX_VALUE,vi=0;
    double s_pbook_average=0,s_pbook_average_var2=0,s_k1=0,s_k1_var2=0,s_k2=0,s_k3_adj=0,s_k3=0,ns1=0,ns2=0,ns3=0,sd1=0,sd2=0,sd3=0,sdk2=0,sd1v=0,sd2v=0,sd3v=0;
    double sv_k1=0,sv_k2=0,sv_k3_adj=0,sv_k3=0;
    for(int i=bsize,ptr_trades=tsize-1;i-->0;)
    {
	final BookItem b=bb[i];
	if(b.time>prevtb)
	{
	    throw new RuntimeException("Book ordering by time is broken");
	}
	prevtb=b.time;
	for(;ptr_trades>=0;)
	{ // find relevant trades
	    final Trade t=trades[ptr_trades];
	    if(t.time>prevtt)
	    {
		throw new RuntimeException("Trades ordering by time is broken");
	    }
	    prevtt=t.time;
	    if(prevtb<=prevtt)
	    {
		vi+=t.shares;
		final double dp=p0-t.price,w=t.shares;
		final double prev_ns1=ns1,prev_s_k1=s_k1,prev_sv_k1=sv_k1;

		ns1+=w;
		ns2+=w*prev_ns1;
		ns3+=(w*prev_ns1)*ns1;

		s_k1+=dp*w;
		s_k1_var2+=dp*(dp*w);
		sv_k1+=vi*w;

		s_k2+=w*(prev_s_k1-dp*prev_ns1);
		s_k3_adj+=w*(prev_s_k1-2*(dp*prev_ns1));
		sv_k2+=w*(prev_sv_k1-vi*prev_ns1);
		sv_k3_adj+=w*(prev_sv_k1-(2*vi)*prev_ns1);

		s_k3+=w*((ns1+prev_ns1)*(prev_s_k1-dp*prev_ns1));
		sv_k3+=w*((ns1+prev_ns1)*(prev_sv_k1-vi*prev_ns1));

		ptr_trades--;
	    }
	    else
	    {
		break;
	    }
	} // end find relevant trades
	final double k1=ns1>0?s_k1/ns1:0,
	    k1_var2=ns1>0?Math.sqrt(s_k1_var2/ns1-k1*k1):0,
	    k2=ns2>0?-s_k2/ns2:0,
	    ns3a=ns1*ns2-ns3,
	    k3=ns3a>0?-(ns1*s_k3_adj+s_k1*ns2-s_k3)/ns3a:0,
	    k1v=ns1>0?sv_k1/ns1:0,
	    k2v=ns2>0?sv_k2/ns2:0,
	    k3v=ns3a>0?(ns1*sv_k3_adj+sv_k1*ns2-sv_k3)/ns3a:0;

	if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
	{ // constrains
	    check_k1k2k3_aver(prevtb,p0,-k1,k1_var2,k2,k3,ns1,ns2,ns3a,k1v,k2v,k3v,trades,tsize);
	}

	final long s=b.shares;
	vi_book+=s;
	stau+=s*((currenttime-prevtb)/ModelParameters.T_SUMM_SCALE);
	final double p_book=b.price,w=b.shares; //  think later about HAAR weight
	sd1+=w*k1;
	sd2+=w*k2;
	if(ns2>0)
	{
	    sdk2+=((ns1*w)/(ns2+ns1*w))*(k1-k2)*w;
	}
	sd3+=w*k3;
	sd1v+=w*k1v;
	sd2v+=w*k2v;
	sd3v+=w*k3v;
	s_pbook_average+=w*p_book;
	s_pbook_average_var2+=(w*p_book)*p_book;
    }
    final double norma=vi_book>0?vi_book:Double.NaN,pbook_average=s_pbook_average/norma;

    final BookTradesK2Stats r=new BookTradesK2Stats(
	    currenttime,
	    vi_book,
	    vi_book>0?(stau/vi_book)*ModelParameters.T_SUMM_SCALE:-1,
	    p0,
	    p_stop,

	    pbook_average,
	    Math.sqrt(s_pbook_average_var2/norma-pbook_average*pbook_average),

	    sd1/norma,
	    sd2/norma,
	    sd3/norma,
	    sdk2/norma,
	    norma,
	    sd1v/norma,
	    sd2v/norma,
	    sd3v/norma);

    return r;
}


private static void check_k1k2k3_aver(final long t0,
				      final double offset,
				      final double k1,
				      final double k1_var2,
				      final double k2,
				      final double k3,
				      final double n1,
				      final double n2,
				      final double n3,
				      final double kv1,
				      final double kv2,
				      final double kv3,
				      final Trade [] trades,
				      final int tsize)
{
    double ns1=0,ns2=0,ns3=0,sk1=0,sk1_var2=0,sk2=0,sk3_adj=0,sk3=0;
    double svk1=0,svk2=0,svk3_adj=0,svk3=0;
    long vi=0;
    for(int i=tsize;i-->0;)
    {
	final Trade ti=trades[i];
	if(ti.time<t0)
	{
	    break;
	}
	vi+=ti.shares;
	final double w=ti.shares,dp=ti.price-offset;
	sk2+=w*(sk1-dp*ns1);
	sk3_adj+=w*(sk1-2*dp*ns1);
	svk2+=w*(svk1-vi*ns1);
	svk3_adj+=w*(svk1-2*vi*ns1);
	final double prev_sk1=sk1,prev_ns1=ns1,prev_svk1=svk1;
	ns2+=w*ns1;
	ns1+=w;
	ns3+=w*prev_ns1*ns1;
	sk1+=dp*w;
	sk1_var2+=dp*(dp*w);
	svk1+=vi*w;
	sk3+=w*(prev_sk1*ns1+prev_ns1*sk1-2*dp*prev_ns1*ns1);
	svk3+=w*(prev_svk1*ns1+prev_ns1*svk1-2*vi*prev_ns1*ns1);
    }
    ns3=ns1*ns2-ns3;
    sk3=ns1*sk3_adj+sk1*ns2-sk3;
    svk3=ns1*svk3_adj+svk1*ns2-svk3;
    final double ck1=ns1>0?sk1/ns1:0,ck1_var2=ns1>0?Math.sqrt(sk1_var2/ns1-ck1*ck1):0,
	ck2=ns2>0?sk2/ns2:0,ck3=ns3>0?sk3/ns3:0,
	cvk1=ns1>0?svk1/ns1:0,cvk2=ns2>0?svk2/ns2:0,cvk3=ns3>0?svk3/ns3:0;
    final double eps=1e-7;
    UnitTests.checkDoubleRel(ns1,"ns1",n1,"n1",eps);
    UnitTests.checkDoubleRel(ns2,"ns2",n2,"n2",eps);
    UnitTests.checkDoubleRel(ns3,"ns3",n3,"n3",eps);
    UnitTests.checkDouble(ck1,"ck1",k1,"k1",eps);
    UnitTests.checkDouble(ck1_var2,"ck1_var2",k1_var2,"k1_var2",eps);
    UnitTests.checkDouble(ck2,"ck2",k2,"k2",eps);
    UnitTests.checkDouble(ck3,"ck3",k3,"k3",eps);
    UnitTests.checkDoubleRel(cvk1,"cvk1",kv1,"kv1",eps);
    UnitTests.checkDoubleRel(cvk2,"cvk2",kv2,"kv2",eps);
    UnitTests.checkDoubleRel(cvk3,"cvk3",kv3,"kv3",eps);
}



static class TradesK1K2K3 {
    final long currenttime,t_range_cut,vi;
    final double
	offset,
	k1,
	k1_var2,
	k2,
	k3,
	normaK1,
	normaK2,
	normaK3,
	k1v,
	k2v,
	k3v;

    TradesK1K2K3(final long currenttime,
		 final long t_range_cut,
		 final long vi,

		 final double offset,
		 final double k1,
		 final double k1_var2,
		 final double k2,
		 final double k3,
		 final double normaK1,
		 final double normaK2,
		 final double normaK3,
		 final double k1v,
		 final double k2v,
		 final double k3v)
    {
	this.currenttime=currenttime;
	this.t_range_cut=t_range_cut;
	this.vi=vi;

	this.offset=offset;
	this.k1=k1;
	this.k1_var2=k1_var2;
	this.k2=k2;
	this.k3=k3;

	this.normaK1=normaK1;
	this.normaK2=normaK2;
	this.normaK3=normaK3;

	this.k1v=k1v;
	this.k2v=k2v;
	this.k3v=k3v;
    }
}



static TradesK1K2K3 calculateTradesK1K2K3(final long currenttime,
					  final Trade [] trades,
					  final int tsize,
					  final long t_range_cut,
					  final double offset)
{

    long vi=0,prevtt=Long.MAX_VALUE;
    double s_k1=0,s_k1_var2=0,s_k2=0,s_k3_adj=0,s_k3=0,ns1=0,ns2=0,ns3=0;
    double sv_k1=0,sv_k2=0,sv_k3_adj=0,sv_k3=0;
    for(int i=tsize;i-->0;)
    {
	final Trade t=trades[i];
	if(t.time>prevtt)
	{
	    throw new RuntimeException("Trades ordering by time is broken");
	}
	prevtt=t.time;
	if(t_range_cut<=prevtt)
	{
	    vi+=t.shares;
	    final double dp=t.price-offset,w=t.shares;
	    final double prev_ns1=ns1,prev_s_k1=s_k1,prev_sv_k1=sv_k1;

	    ns1+=w;
	    ns2+=w*prev_ns1;
	    ns3+=(w*prev_ns1)*ns1;

	    s_k1+=dp*w;
	    s_k1_var2+=dp*(dp*w);
	    sv_k1+=vi*w;

	    s_k2+=w*(prev_s_k1-dp*prev_ns1);
	    s_k3_adj+=w*(prev_s_k1-2*(dp*prev_ns1));
	    sv_k2+=w*(prev_sv_k1-vi*prev_ns1);
	    sv_k3_adj+=w*(prev_sv_k1-(2*vi)*prev_ns1);

	    s_k3+=w*((ns1+prev_ns1)*(prev_s_k1-dp*prev_ns1));
	    sv_k3+=w*((ns1+prev_ns1)*(prev_sv_k1-vi*prev_ns1));
	}
	else
	{
	    break;
	}
    }
    ns3=ns1*ns2-ns3;
    s_k3=ns1*s_k3_adj+s_k1*ns2-s_k3;
    sv_k3=ns1*sv_k3_adj+sv_k1*ns2-sv_k3;
    final double normaK1=ns1>0?ns1:Double.NaN,normaK2=ns2>0?ns2:Double.NaN,normaK3=ns3>0?ns3:Double.NaN,
	k1=s_k1/normaK1;
    final TradesK1K2K3 r=new TradesK1K2K3(
	    currenttime,
	    t_range_cut,
	    vi,
	    offset,
	    k1,
	    Math.sqrt(s_k1_var2/normaK1-k1*k1),
	    s_k2/normaK2,
	    s_k3/normaK3,
	    normaK1,
	    normaK2,
	    normaK3,
	    sv_k1/normaK1,
	    sv_k2/normaK2,
	    sv_k3/normaK3);

    if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // constrains
	check_k1k2k3_aver(r.t_range_cut,
			  r.offset,
			  ns1>0?r.k1:0,
			  ns1>0?r.k1_var2:0,
			  ns2>0?r.k2:0,
			  ns3>0?r.k3:0,
			  ns1,
			  ns2,
			  ns3,
			  ns1>0?r.k1v:0,
			  ns2>0?r.k2v:0,
			  ns3>0?r.k3v:0,
			  trades,
			  tsize);
    }
    return r;
}




static class TradeDStats {
    final long currenttime,t_range_cut,vi;
    final double
	h,
	p_last,

	pt_aver,
	pt_aver_var2,
	d1m,d1e,d1es,d1f,d2m,d2e,d2es,d2f,d3m,d3e,d3es,d3f,f0,
	d1v_e,d2v_e,d3v_e;
    final double [] fmoments;
    final int order;
    final double [][] all_fmoments;

    TradeDStats(final long currenttime,
		final long t_range_cut,
		final long vi,

		final double h,
		final double p_last,

		final double pt_aver,
		final double pt_aver_var2,

		final double d1m,
		final double d1e,
		final double d1es,
		final double d1f,

		final double d2m,
		final double d2e,
		final double d2es,
		final double d2f,

		final double d3m,
		final double d3e,
		final double d3es,
		final double d3f,

		final double f0,

		final double [] fmoments,

		final double d1v_e,
		final double d2v_e,
		final double d3v_e,

		final int order,
		final double [][] all_fmoments)
    {
	this.currenttime=currenttime;
	this.t_range_cut=t_range_cut;
	this.vi=vi;

	this.h=h;
	this.p_last=p_last;

	this.pt_aver=pt_aver;
	this.pt_aver_var2=pt_aver_var2;

	this.d1m=d1m;
	this.d1e=d1e;
	this.d1es=d1es;
	this.d1f=d1f;

	this.d2m=d2m;
	this.d2e=d2e;
	this.d2es=d2es;
	this.d2f=d2f;

	this.d3m=d3m;
	this.d3e=d3e;
	this.d3es=d3es;
	this.d3f=d3f;

	this.f0=f0;

	this.fmoments=fmoments;

	this.d1v_e=d1v_e;
	this.d2v_e=d2v_e;
	this.d3v_e=d3v_e;

	this.order=order;
	this.all_fmoments=all_fmoments;
    }
}


static final int FM_ORDER=5;//7;

static TradeDStats calculateTradesKendallDD(final long currenttime,
					    final Trade [] tb,
					    final int ptr_t_range_cut,
					    final int tsize)
{

    final int n=FM_ORDER;
    final VandermondeInterpolation v=new VandermondeInterpolation(n);

    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
    final double p_last=tsize>0?tb[tsize-1].price:Double.NaN;


    // estimate total volume to scale weight.
    long vi=0;
    double s_pt_aver=0,s_pt_aver_var2=0;
    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade b=tb[i];
	vi+=b.shares;
	s_pt_aver+=b.shares*b.price;
	s_pt_aver_var2+=b.shares*b.price*b.price;
    }

    final double wscale,pt_aver,pt_aver_var2;
    if(vi>0)
    {
	wscale=vi;
	pt_aver=s_pt_aver/wscale;
	pt_aver_var2=Math.sqrt(s_pt_aver_var2/wscale-pt_aver*pt_aver);
    }
    else
    {
	wscale=pt_aver=pt_aver_var2=Double.NaN;
    }


    // recurrent summs, exactly like in com.polytechnik.utils.FMoments
    final double [][] sums=new double [n+1][],sumv=new double [n+1][];
    for(int pn=n;pn>=0;pn--)
    {
	sums[pn]=new double [pn+1]; // all 0 by default
	sumv[pn]=new double [pn+1]; // all 0 by default
    }
    sums[0][0]=sumv[0][0]=1;

    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade b=tb[i];
	final double w=b.shares/wscale,wf=w*b.price,wt=w*(currenttime-b.time);
	for(int pn=n;pn>=1;pn--)
	{
	    final double [] spn=sums[pn],spn1=sums[pn-1],spv=sumv[pn],spv1=sumv[pn-1];
	    for(int pf=pn;pf>=2;pf--)
	    {
		spn[pf]+=w*spn1[pf-1];
		spv[pf]+=w*spv1[pf-1];
	    }
	    spn[1]+=wf*spn1[0];
	    spn[0]+=w*spn1[0];
	    spv[1]+=wt*spv1[0];
	    spv[0]+=w*spv1[0];
	}
    }
    // normalize F-moments
    for(int pn=n;pn>=1;pn--)
    {
	final double [] spn=sums[pn],spv=sumv[pn];
	UnitTests.checkDouble(spn[0],"spn[0]",spv[0],"spv[0]");
	final double norm=spn[0]>0?spn[0]:Double.NaN;
	for(int pf=pn;pf>=1;pf--)
	{
	    spn[pf]/=norm;
	    spv[pf]/=norm;
	}
    }

    final double h=1.0/(1+n);
    v.setFMoments(sumv[n]);
    final double fd1v_e=v.getDF(1-h),fd2v_e=v.getD2F(1-h),fd3v_e=v.getD3F(1-h);

    // fmoment to use
    final double [] fm=sums[n];
    v.setFMoments(fm);


    final TradeDStats r=new TradeDStats(currenttime,
					t_range_cut,
					vi,

					h,
					p_last,

					pt_aver,
					pt_aver_var2,

					v.getDF(0.5),
					v.getDF(1-h),
					v.getDF(1-2*h),
					v.getDF(1),

					v.getD2F(0.5),
					v.getD2F(1-h),
					v.getD2F(1-2*h),
					v.getD2F(1),

					v.getD3F(0.5),
					v.getD3F(1-h),
					v.getD3F(1-2*h),
					v.getD3F(1),

					Double.NaN,//v.getFConst(),

					fm,

					fd1v_e,
					fd2v_e,
					fd3v_e,
					n,
					sums);

    if(true && n==5)
    {
	final double d1m=(1*fm[n-4]-8*fm[n-3]       +8*fm[n-1]-1*fm[n])/(12*h),
	    d2m=(-1*fm[n-4]+16*fm[n-3]-30*fm[n-2]+16*fm[n-1]-1*fm[n])/(12*h*h),
	    d3m=(-fm[n-4]+2*fm[n-3]                 -2*fm[n-1]+fm[n])/(2*h*h*h);
	final double d1e=(3*fm[n-4]-16*fm[n-3]+36*fm[n-2]-48*fm[n-1]+25*fm[n])/(12*h),
	    d2e=(11*fm[n-4]-56*fm[n-3]+114*fm[n-2]-104*fm[n-1]+35*fm[n])/(12*h*h),
	    d3e=(3*fm[n-4]-14*fm[n-3]+24*fm[n-2]-18*fm[n-1]+5*fm[n])/(2*h*h*h);
	final double d1es=(-1*fm[n-4]+6*fm[n-3]-18*fm[n-2]+10*fm[n-1]+3*fm[n])/(12*h),
	    d2es=(-1*fm[n-4]+4*fm[n-3]+6*fm[n-2]-20*fm[n-1]+11*fm[n])/(12*h*h);

	final double eps=1e-7;
	UnitTests.checkDouble(d1m,"d1m",r.d1m,"r.d1m",eps);
	UnitTests.checkDouble(d2m,"d2m",r.d2m,"r.d2m",eps);
	UnitTests.checkDouble(d3m,"d3m",r.d3m,"r.d3m",eps);

	UnitTests.checkDouble(d1e,"d1e",r.d1e,"r.d1e",eps);
	UnitTests.checkDouble(d2e,"d2e",r.d2e,"r.d2e",eps);
	UnitTests.checkDouble(d3e,"d3e",r.d3e,"r.d3e",eps);

	UnitTests.checkDouble(d1es,"d1es",r.d1es,"r.d1es",eps);
	UnitTests.checkDouble(d2es,"d2es",r.d2es,"r.d2es",eps);
	//UnitTests.checkDouble(d3es,"d3es",r.d3es,"r.d3es",eps);
    }
    return r;
}





static class TradeDSStats {
    final long currenttime,t_range_cut,vi;
    final double
	p_last,
	pt_aver,
	pt_aver_var2,

	d1pbe,
	d1pb,
	d2pb,
	d3pb,

	dps_sum,dpsa,dpsdv;

    TradeDSStats(final long currenttime,
		final long t_range_cut,
		final long vi,

		final double p_last,

		final double pt_aver,
		final double pt_aver_var2,

		 final double d1pbe,
		 final double d1pb,
		 final double d2pb,
		 final double d3pb,

		final double dps_sum,
		final double dpsa,
		final double dpsdv)
    {
	this.currenttime=currenttime;
	this.t_range_cut=t_range_cut;
	this.vi=vi;

	this.p_last=p_last;

	this.pt_aver=pt_aver;
	this.pt_aver_var2=pt_aver_var2;

	this.d1pbe=d1pbe;
	this.d1pb=d1pb;
	this.d2pb=d2pb;
	this.d3pb=d3pb;

	this.dps_sum=dps_sum;
	this.dpsa=dpsa;
	this.dpsdv=dpsdv;
    }
}



private static final boolean DSSTAT_OPTIMIZATION=false;



static TradeDSStats calculateTradesKendallDS(final long currenttime,
					     final Trade [] tb,
					     final int ptr_t_range_cut,
					     final int tsize)
{
    final int n=FM_ORDER;
    final double h=1.0/(1+n);
    final VandermondeInterpolation v=new VandermondeInterpolation(n);
    final double [][] sums=new double [n+1][];
    if(true)
    {
	// recurrent summs, exactly like in com.polytechnik.utils.FMoments
	for(int pn=n;pn>=0;pn--)
	{
	    sums[pn]=new double [pn+1]; // all 0 by default
	}
	sums[0][0]=1;
    }

    final int ptr_interval=ptr_t_range_cut<0?tsize:ptr_t_range_cut;
    final long t_range_cut=ptr_t_range_cut<0?-1:tb[ptr_t_range_cut].time;
    final double p_last=tsize>0?tb[tsize-1].price:Double.NaN;

    // estimate total volume to scale weight.
    long vi=0;
    double s_pt_aver=0,s_pt_aver_var2=0;
    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade b=tb[i];
	final long b_orig=b.origtime;
	if(b_orig<=0)
	{
	    continue;
	}
	vi+=b.shares;
	s_pt_aver+=b.shares*b.price;
	s_pt_aver_var2+=b.shares*b.price*b.price;
    }
    final double wscale,pt_aver,pt_aver_var2;
    if(vi>0)
    {
	wscale=vi;
	pt_aver=s_pt_aver/wscale;
	pt_aver_var2=Math.sqrt(s_pt_aver_var2/wscale-pt_aver*pt_aver);
    }
    else
    {
	wscale=pt_aver=pt_aver_var2=Double.NaN;
    }

    final long vic=vi;
    double s_dps=0,s_dpsa=0;
    vi=0;

    for(int i=tsize;i-->ptr_interval;)
    {
	final Trade b=tb[i];
	final long b_orig=b.origtime;
	if(b_orig<=0)
	{
	    continue;
	}
	final long s=b.shares;
	final double p0=b.price,wi=s,dpa,dsw;
	if(!DSSTAT_OPTIMIZATION)
	{
	    long sw=0;
	    double sp=0;
	    for(int j=i;j>=0;j--)
	    {
		final Trade tj=tb[j];
		if(b_orig<=tj.time)
		{
		    sw+=tj.shares;
		    sp+=(p0-tj.price)*tj.shares;
		}
		else
		{
		    break;
		}
	    }
	    if(false && sw<=0)
	    {
		throw new RuntimeException(" i="+i+
					   " b.time="+b.time+
					   " b.origtime="+b.origtime+
					   " sw="+sw+" sp="+sp);
	    }
	    dsw=sw>0?sw:Double.NaN;
	    dpa=sp/dsw;
	}
	else
	{
	    final int ptr_orig=find_ptr_t0(b_orig,tb,0,i+1);
	    if(ptr_orig>=0)
	    {
		// FIXME, put difference code for optimization
		//dpa=(ptr_orig>0?sum_pw[ptr_orig-1]:0)-((i+1)<tsize?sum_pw[i+1]:0);
		//dpw=(ptr_orig>0?sum_w[ptr_orig-1]:0)-((i+1)<tsize?sum_w[i+1]:0);
	    }
	    else
	    {
		dpa=0;
		dsw=wi;
	    }
	    throw new RuntimeException("Not supported yet");
	}

	vi+=s;
	final double sensrel=dpa*(wi/dsw);
	s_dps+=sensrel;
	s_dpsa+=wi*sensrel;

	if(true)
	{
	    final double w=wi/wscale,wf=w*sensrel;
	    for(int pn=n;pn>=1;pn--)
	    {
		final double [] spn=sums[pn],spn1=sums[pn-1];
		for(int pf=pn;pf>=2;pf--)
		{
		    spn[pf]+=w*spn1[pf-1];
		}
		spn[1]+=wf*spn1[0];
		spn[0]+=w*spn1[0];
	    }
	}
    }
    if(vi!=vic)
    {
	throw new RuntimeException("Volume Mismatch");
    }

    if(true)
    {
	// normalize F-moments
	for(int pn=n;pn>=1;pn--)
	{
	    final double [] spn=sums[pn];
	    final double norm=spn[0]>0?spn[0]:Double.NaN;
	    for(int pf=pn;pf>=1;pf--)
	    {
		spn[pf]/=norm;
	    }
	}
	// fmoment to use
	v.setFMoments(sums[n]);
    }


    final TradeDSStats r=new TradeDSStats(currenttime,
					  t_range_cut,
					  vi,

					  p_last,

					  pt_aver,
					  pt_aver_var2,

					  v.getDF(1),
					  v.getDF(1-h),
					  v.getD2F(1-h),
					  v.getD3F(1-h),

					  s_dps,
					  s_dpsa/wscale,
					  s_dps/wscale);
    if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // constrains
	//check_TradeDSStats(currenttime,tb,ptr_t_range_cut,tsize,r);
    }

    return r;

}


/*
// from http://martin.ankerl.com/2007/02/11/optimized-exponential-functions-for-java/
// http://citeseer.ist.psu.edu/schraudolph98fast.html
// ftp://ftp.idsia.ch/pub/techrep/IDSIA-07-98.ps.gz
public static double exponentFastApproximation(double val) {
	    final long tmp = (long) (1512775 * val + 1072632447);
	    return Double.longBitsToDouble(tmp << 32);
}
*/



static class TradeEMoments {
    final long currenttime,tau;
    final double p_last,d1Q,
	d1_b1,d1_b2,d1_b3,d1_b4,d1_b5,
	d1e,d2e,d3e,d1f,d2f,d3f,d0a,d1a,d2a,d3a,d0aa,d1aa,d2aa,d3aa,
	dv1e,dv2e,dv3e,dv0a,dvsignedt0a,dvsignedv0a,ddpt0a,ddpv0a,
	dp1e,dp2e,dp3e,dp0a,dp1f,d0fa,dp0fa,
	d1e_iv,d1e_v,d1a_iv,d1a_v,
	d0a_vsigned,d1a_vsigned,d0a_v;
    final double [] emoments,emoments_pt,emoments_vt,emoments_vsignedt,emoments_vsignedv,emoments_dpt,emoments_dpv,
	iv;

    final double []
	sum_pv_emoments,
	sum_p2v_emoments,
	sum_pvtau_emoments,
	sum_pvsignedtau_emoments,
	sum_v_emoments,
	sum_pt_emoments,
	sum_p2t_emoments,
	sum_vt_emoments,
	sum_pvsigned_emoments,
	sum_vsigned_emoments,
	sum_ptsigned_emoments,
	sum_tsigned_emoments,
	sum_vsignedt_emoments,
	sum_vsignedv_emoments,
	sum_dpt_emoments,
	sum_dpv_emoments,
	sum_t_emoments,
	sum_n,
	sum_vtau,
	sum_vsignedtau;
    // references to previous calculations
    private final BookItem trade_last;

    TradeEMoments(final long currenttime,
		  final long tau,
		  final int np,

		  // references
		  final double [] sum_pv_emoments,
		  final double [] sum_p2v_emoments,
		  final double [] sum_pvtau_emoments,
		  final double [] sum_pvsignedtau_emoments,
		  final double [] sum_v_emoments,
		  final double [] sum_pt_emoments,
		  final double [] sum_p2t_emoments,
		  final double [] sum_vt_emoments,
		  final double [] sum_pvsigned_emoments,
		  final double [] sum_vsigned_emoments,
		  final double [] sum_ptsigned_emoments,
		  final double [] sum_tsigned_emoments,
		  final double [] sum_vsignedt_emoments,
		  final double [] sum_vsignedv_emoments,
		  final double [] sum_dpt_emoments,
		  final double [] sum_dpv_emoments,
		  final double [] sum_t_emoments,
		  final double [] sum_n,
		  final double [] sum_vtau,
		  final double [] sum_vsignedtau,
		  final BookItem trade_last)
{
    this.currenttime=currenttime;
    this.tau=tau;
    this.sum_pv_emoments=sum_pv_emoments;
    this.sum_p2v_emoments=sum_p2v_emoments;
    this.sum_pvtau_emoments=sum_pvtau_emoments;
    this.sum_pvsignedtau_emoments=sum_pvsignedtau_emoments;
    this.sum_v_emoments=sum_v_emoments;
    this.sum_pt_emoments=sum_pt_emoments;
    this.sum_p2t_emoments=sum_p2t_emoments;
    this.sum_vt_emoments=sum_vt_emoments;
    this.sum_pvsigned_emoments=sum_pvsigned_emoments;
    this.sum_vsigned_emoments=sum_vsigned_emoments;
    this.sum_ptsigned_emoments=sum_ptsigned_emoments;
    this.sum_tsigned_emoments=sum_tsigned_emoments;
    this.sum_vsignedt_emoments=sum_vsignedt_emoments;
    this.sum_vsignedv_emoments=sum_vsignedv_emoments;
    this.sum_dpt_emoments=sum_dpt_emoments;
    this.sum_dpv_emoments=sum_dpv_emoments;
    this.sum_t_emoments=sum_t_emoments;
    this.sum_n=sum_n;
    this.sum_vtau=sum_vtau;
    this.sum_vsignedtau=sum_vsignedtau;
    this.trade_last=trade_last;
    this.p_last=this.trade_last!=null?this.trade_last.price:Double.NaN;

    final int n=sum_t_emoments.length;
    this.emoments=new double [n];
    this.emoments_pt=new double [n];
    this.emoments_vt=new double [n];
    this.emoments_vsignedt=new double [n];
    this.emoments_vsignedv=new double [n];
    this.emoments_dpt=new double [n];
    this.emoments_dpv=new double [n];
    this.iv=new double [n];
    // normalize
    if(np>0)
    {
	for(int m=0;m<n;m++)
	{
	    final double vm=sum_v_emoments[m];
	    emoments[m]=sum_pv_emoments[m]/vm;
	    emoments_vsignedv[m]=sum_vsignedv_emoments[m]/vm;
	    emoments_dpv[m]=sum_dpv_emoments[m]/vm;
	    iv[m]=sum_vtau[m]/(vm*tau);

	    final double tm=sum_t_emoments[m];
	    if(tm>0)
	    {
		emoments_pt[m]=sum_pt_emoments[m]/tm;
		emoments_vt[m]=sum_vt_emoments[m]/tm;
		emoments_vsignedt[m]=sum_vsignedt_emoments[m]/tm;
		emoments_dpt[m]=sum_dpt_emoments[m]/tm;
	    }
	    else
	    {
		emoments_pt[m]=emoments_vt[m]=emoments_vsignedt[m]=emoments_dpt[m]=Double.NaN;
	    }
	}
    }
    else
    {
	java.util.Arrays.fill(emoments,Double.NaN);
	java.util.Arrays.fill(emoments_vt,Double.NaN);
	java.util.Arrays.fill(emoments_vsignedt,Double.NaN);
	java.util.Arrays.fill(emoments_vsignedv,Double.NaN);
	java.util.Arrays.fill(emoments_dpt,Double.NaN);
	java.util.Arrays.fill(emoments_dpv,Double.NaN);
	java.util.Arrays.fill(emoments_pt,Double.NaN);
    }



    if(n<5)
    {
	throw new RuntimeException("There should be at least 5 emoments emoments.length="+n);
    }

    // calculate all D's
    final double h=1;
    d1e=(3*emoments[4]-16*emoments[3]+36*emoments[2]-48*emoments[1]+25*emoments[0])/(12*h);
    d1f=(3*emoments[3]-16*emoments[2]+36*emoments[1]-48*emoments[0]+25*p_last)/(12*h);
    dv1e=(3*emoments_vt[4]-16*emoments_vt[3]+36*emoments_vt[2]-48*emoments_vt[1]+25*emoments_vt[0])/(12*h);
    dp1e=(3*emoments_pt[4]-16*emoments_pt[3]+36*emoments_pt[2]-48*emoments_pt[1]+25*emoments_pt[0])/(12*h);
    dp1f=(3*emoments_pt[3]-16*emoments_pt[2]+36*emoments_pt[1]-48*emoments_pt[0]+25*p_last)/(12*h);

    d2e=(11*emoments[4]-56*emoments[3]+114*emoments[2]-104*emoments[1]+35*emoments[0])/(12*h*h);
    d2f=(11*emoments[3]-56*emoments[2]+114*emoments[1]-104*emoments[0]+35*p_last)/(12*h*h);
    dv2e=(11*emoments_vt[4]-56*emoments_vt[3]+114*emoments_vt[2]-104*emoments_vt[1]+35*emoments_vt[0])/(12*h*h);
    dp2e=(11*emoments_pt[4]-56*emoments_pt[3]+114*emoments_pt[2]-104*emoments_pt[1]+35*emoments_pt[0])/(12*h*h);

    d3e=(3*emoments[4]-14*emoments[3]+24*emoments[2]-18*emoments[1]+5*emoments[0])/(2*h*h*h);
    d3f=(3*emoments[3]-14*emoments[2]+24*emoments[1]-18*emoments[0]+5*p_last)/(2*h*h*h);
    dv3e=(3*emoments_vt[4]-14*emoments_vt[3]+24*emoments_vt[2]-18*emoments_vt[1]+5*emoments_vt[0])/(2*h*h*h);
    dp3e=(3*emoments_pt[4]-14*emoments_pt[3]+24*emoments_pt[2]-18*emoments_pt[1]+5*emoments_pt[0])/(2*h*h*h);

    dv0a=(1*emoments_vt[4]-5*emoments_vt[3]+10*emoments_vt[2]-10*emoments_vt[1]+5*emoments_vt[0]);
    dvsignedt0a=(1*emoments_vsignedt[4]-5*emoments_vsignedt[3]+10*emoments_vsignedt[2]-10*emoments_vsignedt[1]+5*emoments_vsignedt[0]);
    dvsignedv0a=(1*emoments_vsignedv[4]-5*emoments_vsignedv[3]+10*emoments_vsignedv[2]-10*emoments_vsignedv[1]+5*emoments_vsignedv[0]);
    ddpt0a=(1*emoments_dpt[4]-5*emoments_dpt[3]+10*emoments_dpt[2]-10*emoments_dpt[1]+5*emoments_dpt[0]);
    ddpv0a=(1*emoments_dpv[4]-5*emoments_dpv[3]+10*emoments_dpv[2]-10*emoments_dpv[1]+5*emoments_dpv[0]);

    dp0a=(1*emoments_pt[4]-5*emoments_pt[3]+10*emoments_pt[2]-10*emoments_pt[1]+5*emoments_pt[0]);
    d0a=(1*emoments[4]-5*emoments[3]+10*emoments[2]-10*emoments[1]+5*emoments[0]);
    d0fa=( 1*emoments[3]   -5*emoments[2]   +10*emoments[1]   -10*emoments[0]   +5*p_last);
    dp0fa=(1*emoments_pt[3]-5*emoments_pt[2]+10*emoments_pt[1]-10*emoments_pt[0]+5*p_last);
    d1a=(25*emoments[4]-122*emoments[3]+234*emoments[2]-214*emoments[1]+77*emoments[0])/(12*h);
    d2a=(35*emoments[4]-164*emoments[3]+294*emoments[2]-236*emoments[1]+71*emoments[0])/(12*h*h);
    d3a=(5*emoments[4]-22*emoments[3]+36*emoments[2]-26*emoments[1]+7*emoments[0])/(2*h*h*h);

    d0aa=(5*emoments[4]-24*emoments[3]+45*emoments[2]-40*emoments[1]+15*emoments[0]);
    d1aa=(77*emoments[4]-360*emoments[3]+648*emoments[2]-536*emoments[1]+171*emoments[0])/(12*h);
    d2aa=(71*emoments[4]-320*emoments[3]+546*emoments[2]-416*emoments[1]+119*emoments[0])/(12*h*h);
    d3aa=(7*emoments[4]-30*emoments[3]+48*emoments[2]-34*emoments[1]+9*emoments[0])/(2*h*h*h);

    d1_b1=(3*emoments[4]-16*emoments[3]+36*emoments[2]-48*emoments[1]+25*emoments[0])/(12*h);
    d1_b2=(-1*emoments[4]+6*emoments[3]-18*emoments[2]+10*emoments[1]+3*emoments[0])/(12*h);
    d1_b3=(1*emoments[4]-8*emoments[3]+0*emoments[2]+8*emoments[1]-1*emoments[0])/(12*h);
    d1_b4=(-3*emoments[4]-10*emoments[3]+18*emoments[2]-6*emoments[1]+1*emoments[0])/(12*h);
    d1_b5=(-25*emoments[4]+48*emoments[3]-36*emoments[2]+16*emoments[1]-3*emoments[0])/(12*h);

    d1Q=(48*emoments[4]-228*emoments[3]+420*emoments[2]-360*emoments[1]+120*emoments[0])/(12*h);


    d1e_iv=(3*iv[4]-16*iv[3]+36*iv[2]-48*iv[1]+25*iv[0])/(12*h);
    d1a_iv=(25*iv[4]-122*iv[3]+234*iv[2]-214*iv[1]+77*iv[0])/(12*h);

    d1e_v=(3*sum_v_emoments[4]-16*sum_v_emoments[3]+36*sum_v_emoments[2]-48*sum_v_emoments[1]+25*sum_v_emoments[0])/(12*h);
    d1a_v=(25*sum_v_emoments[4]-122*sum_v_emoments[3]+234*sum_v_emoments[2]-214*sum_v_emoments[1]+77*sum_v_emoments[0])/(12*h);
    d0a_v=1*sum_v_emoments[4]-5*sum_v_emoments[3]+10*sum_v_emoments[2]-10*sum_v_emoments[1]+5*sum_v_emoments[0];


    d0a_vsigned=1*sum_vsigned_emoments[4]-5*sum_vsigned_emoments[3]+10*sum_vsigned_emoments[2]-10*sum_vsigned_emoments[1]+5*sum_vsigned_emoments[0];
    d1a_vsigned=(25*sum_vsigned_emoments[4]-122*sum_vsigned_emoments[3]+234*sum_vsigned_emoments[2]-214*sum_vsigned_emoments[1]+77*sum_vsigned_emoments[0])/(12*h);


    //System.err.println("d1e="+d1e);

}



    private static void checkValuesAtLimitCase(final java.util.Random r)
    {
	System.err.println("checkValuesAtLimitCase() specific values check.");
	final int n_test=1000000;
	final Trade [] tb=new Trade [n_test];
	final long tquant=TraderInterface.TIME_SCALE/10000;
	final int v0=1;
	long t=0;
	for(int i=0;i<n_test;i++)
	{
	    t=(n_test+i)*tquant;
	    final double p=(n_test-1-i)*tquant;
	    tb[i]=Trade.getDummyTrade(p,v0,BookItem.BOOK_UNKNOWN,t);
	}

	long tau=0;
	for(int it=0;it<20;it++)
	{
	    tau+=tquant;
	    final TradeEMoments em=calculateEMoments(t,tau,tb,n_test,null);
	    final double intdt=tquant/((double)tau);

	    System.err.println("======= Test with tau="+em.tau+" p_last="+em.p_last);
	    for(int m=0;m<em.emoments.length;m++)
	    {
		System.err.println("int v["+m+"]="+(em.sum_v_emoments[m]*intdt)+" should be="+v0+
				   " int pt["+m+"]="+(em.emoments_pt[m]/tau)+" should be="+(m+1)+
				   " int vt["+m+"]="+(em.emoments_vt[m])+" should be="+v0+
				   " int pv["+m+"]="+(em.emoments[m]/tau)+" should be="+(m+1));
	    }
	    checkDEM(em);
	}
    }


    /** Unit test.
     */
    public static void main(String [] args)
    {
	final java.util.Random r=new java.util.Random(-79091);
	for(int i=0;i<50;i++)
	{
	    checkCalculations(10010+r.nextInt(1000010),r);
	}
	checkValuesAtLimitCase(r);

    }

    private static void checkMatching(final TradeEMoments r,
				      final TradeEMoments ref)
    {
	final double eps=1e-5;
	for(int m=0;m<r.sum_t_emoments.length;m++)
	{
	    UnitTests.checkDoubleRel(r.sum_pv_emoments[m],"r.sum_pv_emoments[m]",ref.sum_pv_emoments[m],"ref.sum_pv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_p2v_emoments[m],"r.sum_p2v_emoments[m]",ref.sum_p2v_emoments[m],"ref.sum_p2v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pvtau_emoments[m],"r.sum_pvtau_emoments[m]",ref.sum_pvtau_emoments[m],"ref.sum_pvtau_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pvsignedtau_emoments[m],"r.sum_pvsignedtau_emoments[m]",ref.sum_pvsignedtau_emoments[m],"ref.sum_pvsignedtau_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_v_emoments[m],"r.sum_v_emoments[m]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pt_emoments[m],"r.sum_pt_emoments[m]",ref.sum_pt_emoments[m],"ref.sum_pt_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_p2t_emoments[m],"r.sum_p2t_emoments[m]",ref.sum_p2t_emoments[m],"ref.sum_p2t_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vt_emoments[m],"r.sum_vt_emoments[m]",ref.sum_vt_emoments[m],"ref.sum_vt_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pvsigned_emoments[m],"r.sum_pvsigned_emoments[m]",ref.sum_pvsigned_emoments[m],"ref.sum_pvsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsigned_emoments[m],"r.sum_vsigned_emoments[m]",ref.sum_vsigned_emoments[m],"ref.sum_vsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_ptsigned_emoments[m],"r.sum_ptsigned_emoments[m]",ref.sum_ptsigned_emoments[m],"ref.sum_ptsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_tsigned_emoments[m],"r.sum_tsigned_emoments[m]",ref.sum_tsigned_emoments[m],"ref.sum_tsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsignedt_emoments[m],"r.sum_vsignedt_emoments[m]",ref.sum_vsignedt_emoments[m],"ref.sum_vsignedt_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsignedv_emoments[m],"r.sum_vsignedv_emoments[m]",ref.sum_vsignedv_emoments[m],"ref.sum_vsignedv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_dpt_emoments[m],"r.sum_dpt_emoments[m]",ref.sum_dpt_emoments[m],"ref.sum_dpt_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_dpv_emoments[m],"r.sum_dpv_emoments[m]",ref.sum_dpv_emoments[m],"ref.sum_dpv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_t_emoments[m],"r.sum_t_emoments[m]",ref.sum_t_emoments[m],"ref.sum_t_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_n[m],"r.sum_n[m]",ref.sum_n[m],"ref.sum_n[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vtau[m],"r.sum_vtau[m]",ref.sum_vtau[m],"ref.sum_vtau[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsignedtau[m],"r.sum_vsignedtau[m]",ref.sum_vsignedtau[m],"ref.sum_vsignedtau[m]",eps);
	}
	System.err.println("checkMatching(): matching OK");
    }



    private static void checkMatchingTradeVX(final TradeEMoments ref,final TradeVX r)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	final double atau=1.0/ref.tau;
	final int n=r.sum_v_emoments.length;
	final double eps=1e-5;
	for(int m=0;m<n;m++)
	{
	    UnitTests.checkDoubleRel(r.sum_v_emoments[m],"r.sum_v_emoments[m]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsigned_emoments[m],"r.sum_vsigned_emoments[m]",ref.sum_vsigned_emoments[m],"ref.sum_vsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsignedv_emoments[m],"r.sum_vsignedv_emoments[m]",ref.sum_vsignedv_emoments[m],"ref.sum_vsignedv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pv_emoments[m],"r.sum_pv_emoments[m]",ref.sum_pv_emoments[m],"ref.sum_pv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.matr[m][0],"r.matr[m][0]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.matr[m][1],"r.matr[m][1]",ref.sum_vtau[m]*atau,"ref.sum_vtau[m]*atau",eps);
	    UnitTests.checkDoubleRel(r.pmatr[m][0],"r.pmatr[m][0]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.pmatr[m][1],"r.pmatr[m][1]",ref.sum_pv_emoments[m],"ref.sum_pv_emoments[m]",eps);
	}
	System.err.println("TradeEmoments:checkMatchingTradeVX(): matching OK");
    }


    private static void checkMatchingTradeIPData(final TradeEMoments ref,final TradeIPData r)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	final double atau=1.0/ref.tau;
	final double eps=1e-5;
	final double o=r.p_offset,pv=ref.sum_pv_emoments[0]/ref.sum_v_emoments[0],pt=ref.sum_pt_emoments[0]/ref.sum_t_emoments[0];

	UnitTests.checkDoubleRel(r.v_sum,"r.v_sum",ref.sum_v_emoments[0],"ref.sum_v_emoments[0]",eps);
	UnitTests.checkDoubleRel(r.pv_sum,"r.pv_sum",ref.sum_pv_emoments[0],"ref.sum_pv_emoments[0]",eps);
	final double m2_pv=ref.sum_p2v_emoments[0]+o*(-2*ref.sum_pv_emoments[0]+o*ref.sum_v_emoments[0]);
	UnitTests.checkDoubleRel(r.moments_pv[2],"r.moments_pv[2]",m2_pv,"m2_pv",eps);

	UnitTests.checkDoubleRel(r.moments_pv_tau[0],"r.moments_pv_tau[0]",ref.sum_vtau[0]*(-atau),"-atau*ref.sum_vtau[0]",eps);
	final double m1_pv_tau=ref.sum_pvtau_emoments[0]-o*ref.sum_vtau[0];
	UnitTests.checkDoubleRel(r.moments_pv_tau[1],"r.moments_pv_tau[1]",m1_pv_tau*(-atau),"-atau*m1_pv_tau",eps);


	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.sum_t_emoments[0],"ref.sum_t_emoments[0]",eps);
	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.sum_pt_emoments[0],"ref.sum_pt_emoments[0]",eps);
	final double m2_pt=ref.sum_p2t_emoments[0]+o*(-2*ref.sum_pt_emoments[0]+o*ref.sum_t_emoments[0]);
	UnitTests.checkDoubleRel(r.moments_pt[2],"r.moments_pt[2]",m2_pt,"m2_pt",eps);


	final double vsigned=r.moments_pv_sell[0]-r.moments_pv_buy[0];
	UnitTests.checkDoubleRel(vsigned,"vsigned",ref.sum_vsigned_emoments[0],"ref.sum_vsigned_emoments[0]",eps);
	final double pvsigned=r.moments_pv_sell[1]-r.moments_pv_buy[1]+r.p_offset*vsigned;
	UnitTests.checkDoubleRel(pvsigned,"pvsigned",ref.sum_pvsigned_emoments[0],"ref.sum_pvsigned_emoments[0]",eps);

	/* // The definition of moments_pt_{buy,sell} changed since Oct 27 2010.
	final double tsigned=r.moments_pt_sell[0]-r.moments_pt_buy[0];
	UnitTests.checkDoubleRel(tsigned,"tsigned",ref.sum_tsigned_emoments[0],",ref.sum_tsigned_emoments[0]",eps);
	final double ptsigned=r.moments_pt_sell[1]-r.moments_pt_buy[1]+r.p_offset*tsigned;
	UnitTests.checkDoubleRel(ptsigned,"ptsigned",ref.sum_ptsigned_emoments[0],"ref.sum_ptsigned_emoments[0]",eps);
	*/

	System.err.println("TradeEmoments:checkMatchingTradeIPData(): matching OK");
    }




    private static void checkMatchingTradePV(final TradeEMoments ref,final TradePV r)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	final double atau=1.0/ref.tau;
	final int n=r.sum_v_emoments.length;
	final double eps=1e-5;
	for(int m=0;m<n;m++)
	{
	    UnitTests.checkDoubleRel(r.sum_v_emoments[m],"r.sum_v_emoments[m]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsigned_emoments[m],"r.sum_vsigned_emoments[m]",ref.sum_vsigned_emoments[m],"ref.sum_vsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pv_emoments[m],"r.sum_pv_emoments[m]",ref.sum_pv_emoments[m],"ref.sum_pv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pt_emoments[m],"r.sum_pt_emoments[m]",r.ptmatr[m][1],"r.ptmatr[m][1]",eps);

	    UnitTests.checkDoubleRel(ref.sum_t_emoments[m],"ref.sum_t_emoments[m]",r.ptmatr[m][0],"r.pmatr[m][0]",eps);
	    UnitTests.checkDoubleRel(ref.sum_pt_emoments[m],"ref.sum_pt_emoments[m]",r.ptmatr[m][1],"r.ptmatr[m][1]",eps);

	    UnitTests.checkDoubleRel(r.matr[m][0],"r.matr[m][0]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.matr[m][1],"r.matr[m][1]",ref.sum_vtau[m]*atau,"ref.sum_vtau[m]*atau",eps);
	    UnitTests.checkDoubleRel(r.pvmatr[m][0],"r.pvmatr[m][0]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.pvmatr[m][1],"r.pvmatr[m][1]",ref.sum_pv_emoments[m],"ref.sum_pv_emoments[m]",eps);
	}
	System.err.println("TradeEmoments:checkMatchingTradePV(): matching OK");
    }





    private static void checkDEM(final TradeEMoments r)
    {
	final int o=5;
	final double eps=1e-7;
	final double h=1.0/(o+1);
	final VandermondeInterpolation v=new VandermondeInterpolation(o);
	final double [] fm=new double [o+1];
	for(int m=0;m<o;m++)
	{
	    fm[1+m]=r.emoments[o-1-m];
	}
	v.setFMoments(fm);
	UnitTests.checkDouble(h*((v.getF(1.0+h)-v.getF(1.0))/h),"h*((v.getF(1.0+h)-v.getF(1.0))/h)",r.d1Q,"r.d1Q",eps);
	UnitTests.checkDouble(h*v.getDF(1.0-h),"h*v.getDF(1.0-h)",r.d1e,"r.d1e",eps);
	UnitTests.checkDouble(h*h*v.getD2F(1.0-h),"h*h*v.getD2F(1.0-h)",r.d2e,"r.d2e",eps);
	UnitTests.checkDouble(h*h*h*v.getD3F(1.0-h),"h*h*h*v.getD3F(1.0-h)",r.d3e,"r.d3e",eps);
	UnitTests.checkDouble(v.getF(1.0),"v.getF(1.0)",r.d0a,"r.d0a",eps);
	UnitTests.checkDouble(h*v.getDF(1.0),"h*v.getDF(1.0)",r.d1a,"r.d1a",eps);
	UnitTests.checkDouble(h*h*v.getD2F(1.0),"h*h*v.getD2F(1.0)",r.d2a,"r.d2a",eps);
	UnitTests.checkDouble(h*h*h*v.getD3F(1.0),"h*h*h*v.getD3F(1.0)",r.d3a,"r.d3a",eps);
	UnitTests.checkDouble(v.getF(1.0+h),"v.getF(1.0+h)",r.d0aa,"r.d0aa",eps);
	UnitTests.checkDouble(h*v.getDF(1.0+h),"h*v.getDF(1.0+h)",r.d1aa,"r.d1aa",eps);
	UnitTests.checkDouble(h*h*v.getD2F(1.0+h),"h*h*v.getD2F(1.0+h)",r.d2aa,"r.d2aa",eps);
	UnitTests.checkDouble(h*h*h*v.getD3F(1.0+h),"h*h*h*v.getD3F(1.0+h)",r.d3aa,"r.d3aa",eps);

	UnitTests.checkDouble(h*v.getDF(1-1*h),"h*v.getDF(1-1*h)",r.d1_b1,"r.d1_b1",eps);
	UnitTests.checkDouble(h*v.getDF(1-2*h),"h*v.getDF(1-2*h)",r.d1_b2,"r.d1_b2",eps);
	UnitTests.checkDouble(h*v.getDF(1-3*h),"h*v.getDF(1-3*h)",r.d1_b3,"r.d1_b3",eps);
	UnitTests.checkDouble(h*v.getDF(1-4*h),"h*v.getDF(1-4*h)",r.d1_b4,"r.d1_b4",eps);
	UnitTests.checkDouble(h*v.getDF(1-5*h),"h*v.getDF(1-5*h)",r.d1_b5,"r.d1_b5",eps);

	for(int m=0;m<o-1;m++)
	{
	    fm[1+m]=r.emoments[o-2-m];
	}
	fm[o]=r.p_last;
	v.setFMoments(fm);
	UnitTests.checkDouble(h*v.getDF(1.0-h),"h*v.getDF(1.0-h)",r.d1f,"r.d1f",eps);
	UnitTests.checkDouble(h*h*v.getD2F(1.0-h),"h*h*v.getD2F(1.0-h)",r.d2f,"r.d2f",eps);
	UnitTests.checkDouble(h*h*h*v.getD3F(1.0-h),"h*h*h*v.getD3F(1.0-h)",r.d3f,"r.d3f",eps);

	for(int m=0;m<o;m++)
	{
	    fm[1+m]=r.emoments_vt[o-1-m];
	}
	v.setFMoments(fm);
	UnitTests.checkDouble(v.getF(1.0),"v.getF(1.0)",r.dv0a,"r.dv0a",eps);
	UnitTests.checkDouble(h*v.getDF(1.0-h),"h*v.getDF(1.0-h)",r.dv1e,"r.dv1e",eps);
	UnitTests.checkDouble(h*h*v.getD2F(1.0-h),"h*h*v.getD2F(1.0-h)",r.dv2e,"r.dv2e",eps);
	UnitTests.checkDouble(h*h*h*v.getD3F(1.0-h),"h*h*h*v.getD3F(1.0-h)",r.dv3e,"r.dv3e",eps);

	for(int m=0;m<o;m++)
	{
	    fm[1+m]=r.emoments_pt[o-1-m];
	}
	v.setFMoments(fm);
	UnitTests.checkDouble(v.getF(1.0),"v.getF(1.0)",r.dp0a,"r.dp0a",eps);
	UnitTests.checkDouble(h*v.getDF(1.0-h),"h*v.getDF(1.0-h)",r.dp1e,"r.dp1e",eps);
	UnitTests.checkDouble(h*h*v.getD2F(1.0-h),"h*h*v.getD2F(1.0-h)",r.dp2e,"r.dp2e",eps);
	UnitTests.checkDouble(h*h*h*v.getD3F(1.0-h),"h*h*h*v.getD3F(1.0-h)",r.dp3e,"r.dp3e",eps);
    }

    private static final java.util.Comparator<Trade>cmp_by_time=
	new java.util.Comparator<Trade>()
	{
	    public int compare(final Trade o1,final Trade o2)
	    {
		return o1.time<o2.time?-1:(o1.time>o2.time?1:(o1.orderreferencenumber<o2.orderreferencenumber?-1:(o1.orderreferencenumber>o2.orderreferencenumber?1:0)));
	    }
	};



    private static TradeEMoments getSum2(final TradeEMoments sa,final TradeEMoments sb)
    {
	final int n=sa.sum_v_emoments.length;
	final double []
	    sum_pv_emoments=new double [n],
	    sum_p2v_emoments=new double [n],
	    sum_pvtau_emoments=new double [n],
	    sum_pvsignedtau_emoments=new double [n],
	    sum_v_emoments=new double [n],
	    sum_pt_emoments=new double [n],
	    sum_p2t_emoments=new double [n],
	    sum_vt_emoments=new double [n],
	    sum_pvsigned_emoments=new double [n],
	    sum_vsigned_emoments=new double [n],
	    sum_ptsigned_emoments=new double [n],
	    sum_tsigned_emoments=new double [n],
	    sum_vsignedt_emoments=new double [n],
	    sum_vsignedv_emoments=new double [n],
	    sum_dpt_emoments=new double [n],
	    sum_dpv_emoments=new double [n],
	    sum_t_emoments=new double [n],
	    sum_n=new double [n],
	    sum_vtau=new double [n],
	    sum_vsignedtau=new double [n];

	for(int m=0;m<n;m++)
	{
	    sum_pv_emoments[m]=sa.sum_pv_emoments[m]+sb.sum_pv_emoments[m];
	    sum_p2v_emoments[m]=sa.sum_p2v_emoments[m]+sb.sum_p2v_emoments[m];
	    sum_pvtau_emoments[m]=sa.sum_pvtau_emoments[m]+sb.sum_pvtau_emoments[m];
	    sum_v_emoments[m]=sa.sum_v_emoments[m]+sb.sum_v_emoments[m];
	    sum_pt_emoments[m]=sa.sum_pt_emoments[m]+sb.sum_pt_emoments[m];
	    sum_p2t_emoments[m]=sa.sum_p2t_emoments[m]+sb.sum_p2t_emoments[m];
	    sum_vt_emoments[m]=sa.sum_vt_emoments[m]+sb.sum_vt_emoments[m];
	    sum_pvsigned_emoments[m]=sa.sum_pvsigned_emoments[m]+sb.sum_pvsigned_emoments[m];
	    sum_vsigned_emoments[m]=sa.sum_vsigned_emoments[m]+sb.sum_vsigned_emoments[m];
	    sum_ptsigned_emoments[m]=sa.sum_ptsigned_emoments[m]+sb.sum_ptsigned_emoments[m];
	    sum_tsigned_emoments[m]+=sa.sum_tsigned_emoments[m]+sb.sum_tsigned_emoments[m];
	    sum_vsignedt_emoments[m]=sa.sum_vsignedt_emoments[m]+sb.sum_vsignedt_emoments[m];
	    sum_vsignedv_emoments[m]=sa.sum_vsignedv_emoments[m]+sb.sum_vsignedv_emoments[m];
	    sum_dpt_emoments[m]=sa.sum_dpt_emoments[m]+sb.sum_dpt_emoments[m];
	    sum_dpv_emoments[m]=sa.sum_dpv_emoments[m]+sb.sum_dpv_emoments[m];
	    sum_t_emoments[m]=sa.sum_t_emoments[m]+sb.sum_t_emoments[m];
	    sum_n[m]=sa.sum_n[m]+sb.sum_n[m];
	    sum_vtau[m]=sa.sum_vtau[m]+sb.sum_vtau[m];
	    sum_vsignedtau[m]=sa.sum_vsignedtau[m]+sb.sum_vsignedtau[m];
	}

	return new TradeEMoments(sa.currenttime,
				 sa.tau,
				 1,
				 sum_pv_emoments,
				 sum_p2v_emoments,
				 sum_pvtau_emoments,
				 sum_pvsignedtau_emoments,
				 sum_v_emoments,
				 sum_pt_emoments,
				 sum_p2t_emoments,
				 sum_vt_emoments,
				 sum_pvsigned_emoments,
				 sum_vsigned_emoments,
				 sum_ptsigned_emoments,
				 sum_tsigned_emoments,
				 sum_vsignedt_emoments,
				 sum_vsignedv_emoments,
				 sum_dpt_emoments,
				 sum_dpv_emoments,
				 sum_t_emoments,
				 sum_n,
				 sum_vtau,
				 sum_vsignedtau,
				 null);
    }





    static final double DEFAULT_FACT=10,DEFAULT_P0=3;
    static double getRandomPrice(final java.util.Random r)
    {
	return getRandomPrice(r,-1);
    }

    static double getRandomPrice(final java.util.Random r,final int n_quants,final double fact,final double p0)
    {
	if(n_quants<=0)	return p0+fact*r.nextDouble();
	return p0+(fact*r.nextInt(n_quants))/n_quants;
    }
    static double getRandomPrice(final java.util.Random r,final int n_quants)
    {
	return getRandomPrice(r,n_quants,DEFAULT_FACT,DEFAULT_P0);
    }
    static Trade [] getRandomTrades(final int n,final java.util.Random r)
    {
	return getRandomTradesQuantumPrice(n,r,-1);
    }
    static Trade [] getRandomTradesQuantumPrice(final int n,final java.util.Random r,final int n_quants)
    {
	return getRandomTradesQuantumPrice(n,r,n_quants,DEFAULT_FACT,DEFAULT_P0);
    }
    static Trade [] getRandomTradesQuantumPrice(final int n,final java.util.Random r,final int n_quants,final double fact,final double p0)
    {
	final double PROB_NOORIG=-0.05; // have always orig
	final Trade [] tb=new Trade [n];
	long t=1000*1000L;
	for(int i=0;i<tb.length;i++)
	{
	    t+=1+r.nextInt(10000);
	    final double p=getRandomPrice(r,n_quants,fact,p0);
	    final int v=100+r.nextInt(1000);
	    final byte type=r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL;
	    final long torig=r.nextDouble()<PROB_NOORIG?0:t-r.nextInt(100000);
	    final double pbuy=p-(type==BookItem.BOOK_BUY?0.05*r.nextInt(2):0.2*r.nextDouble()),
		psell=p+(type==BookItem.BOOK_SELL?0.05*r.nextInt(2):0.2*r.nextDouble());
	    final BookItem best_buy=new BookItem(pbuy,
						 100+r.nextInt(1000),
						 null,
						 BookItem.BOOK_BUY,
						 t,
						 0,new BookItem(pbuy-0.1*r.nextDouble(),
								100+r.nextInt(1000),
								null,
								BookItem.BOOK_BUY,
								t,
								0,null)),
		best_sell=new BookItem(psell,
				       100+r.nextInt(1000),
				       null,
				       BookItem.BOOK_SELL,
				       t,
				       0,new BookItem(psell+0.1*r.nextDouble(),
						      100+r.nextInt(1000),
						      null,
						      BookItem.BOOK_SELL,
						      t,
						      0,null));
	    tb[i]=new Trade(p,
			    v,
			    null,
			    type,
			    t,
			    0,
			    0,
			    torig,
			    torig<=0?null:new BookItem(p,
					 100+r.nextInt(1000),
					 null,
					 type,
					 torig,
					 0,
					 type==BookItem.BOOK_BUY?new BookItem(p+0.2*r.nextDouble(),
									      100+r.nextInt(1000),
									      null,
									      BookItem.BOOK_SELL,
									      t-r.nextInt(100000),
									      0,null):
					 (type==BookItem.BOOK_SELL?new BookItem(p-0.2*r.nextDouble(),
										100+r.nextInt(1000),
										null,
										BookItem.BOOK_BUY,
										t-r.nextInt(100000),
										0,null):null)),
			    Trade.TRADE_TYPE_REGULAR,
			    best_buy,
			    best_sell,
			    i);
	}
	return tb;
    }


    private static Trade [] selectTradesOfSpecificType(final int n,final Trade [] tb,final byte type)
    {
	final Trade [] tb2=new Trade [n];
	int na=0;
	for(int i=0;i<n;i++)
	{
	    final Trade t=tb[i];
	    if(t.buy_sell_indicator==type)
	    {
		tb2[na++]=t;
	    }
	}
	final Trade [] tb3=new Trade [na];
	System.arraycopy(tb2,0,tb3,0,na);
	return tb3;
    }


    private static Trade [] getRandomTradesFixedTau(final int tsize,
						    final java.util.Random r,
						    final long tau)
    {
	final Trade [] tb=new Trade [tsize];
	final long dt=1;
	long t=1000*1000L;
	for(int i=0;i<tb.length;i++)
	{
	    t+=dt;
	    final double p=100+10*r.nextDouble();
	    final int v=100+r.nextInt(1000);
	    tb[i]=new Trade(p,
			    v,
			    null,
			    r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
			    t,
			    0,
			    0,
			    t-r.nextInt(100000),
			    null,
			    Trade.TRADE_TYPE_REGULAR,
			    null,
			    null,
			    i);
	}
	return tb;
    }

    static Trade [] getTradesOrigExeSwapped(final Trade [] trade,final int tsize)
    {
	final Trade [] tb2=new Trade [tsize];
	for(int i=0;i<tsize;i++)
	{
	    final Trade t=trade[i];
	    if(t.origtime<=0)
	    {
		throw new RuntimeException("t.origtime="+t.origtime);
	    }
	    tb2[i]=new Trade(t.price,
			     t.shares,
			     t.attribution,
			     t.buy_sell_indicator==BookItem.BOOK_BUY?BookItem.BOOK_SELL:BookItem.BOOK_BUY,
			     t.origtime,
			     t.matchnumber,
			     t.orderreferencenumber,
			     t.time,
			     null,
			     t.tradetype,
			     t.best_buy,
			     t.best_sell,
			     i);
	}
	com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tb2,0,tsize-1,cmp_by_time);
	return tb2;
}



    private static TradeEMoments calculateTestEMoments(final long currenttime,
						       final long tau,
						       final BookItem [] tb,
						       final int tsize)
    {
	final int n=N_EMOMENTS;
	final double []
	    sum_pv_emoments=new double [n],
	    sum_p2v_emoments=new double [n],
	    sum_pvtau_emoments=new double [n],
	    sum_pvsignedtau_emoments=new double [n],
	    sum_v_emoments=new double [n],
	    sum_pt_emoments=new double [n],
	    sum_p2t_emoments=new double [n],
	    sum_vt_emoments=new double [n],
	    sum_pvsigned_emoments=new double [n],
	    sum_vsigned_emoments=new double [n],
	    sum_ptsigned_emoments=new double [n],
	    sum_tsigned_emoments=new double [n],
	    sum_vsignedt_emoments=new double [n],
	    sum_vsignedv_emoments=new double [n],
	    sum_dpt_emoments=new double [n],
	    sum_dpv_emoments=new double [n],
	    sum_t_emoments=new double [n],
	    sum_n=new double [n],
	    sum_vtau=new double [n],
	    sum_vsignedtau=new double [n];
	final double atau=1.0/tau;
	long prevt=currenttime,tinterval=currenttime;
	for(int i=tb.length;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;
	    final double dt=(tinterval-prevt)*atau;
	    final double bwsigned=(b.buy_sell_indicator==BookItem.BOOK_BUY?-b.shares:(b.buy_sell_indicator==BookItem.BOOK_SELL?b.shares:0));
	    final double dtsigned=(b.buy_sell_indicator==BookItem.BOOK_BUY?-dt:(b.buy_sell_indicator==BookItem.BOOK_SELL?dt:0));
	    final double ttau=currenttime-b.time,ts=ttau*atau;
	    final double dp=Double.NaN;//finddpx(tb,0,i+1);
	    for(int m=0,fact=1;m<n;m++)
	    {
		final double e;
		if(EXP_WEIGHT)
		{
		    e=Math.exp(-ts/(m+1))/(m+1);
		}
		else
		{
		    e=(Math.exp(-ts)*Math.pow(ts,m))/fact;
		    fact*=(m+1);
		}
		sum_pv_emoments[m]+=e*b.shares*b.price;
		sum_p2v_emoments[m]+=e*b.shares*b.price*b.price;
		sum_pvtau_emoments[m]+=e*b.shares*b.price*ttau;
		sum_pvsignedtau_emoments[m]+=e*bwsigned*b.price*ttau;
		sum_v_emoments[m]+=e*b.shares;
		sum_pt_emoments[m]+=e*b.price*dt;
		sum_p2t_emoments[m]+=e*b.price*b.price*dt;
		sum_vt_emoments[m]+=e*b.shares*dt;
		sum_pvsigned_emoments[m]+=e*b.price*bwsigned;
		sum_vsigned_emoments[m]+=e*bwsigned;
		sum_ptsigned_emoments[m]+=e*b.price*dtsigned;
		sum_tsigned_emoments[m]+=e*dtsigned;
		sum_vsignedt_emoments[m]+=e*bwsigned*dt;
		sum_vsignedv_emoments[m]+=e*bwsigned*b.shares;
		sum_dpt_emoments[m]+=e*dp*dt;
		sum_dpv_emoments[m]+=e*dp*b.shares;
		sum_t_emoments[m]+=e*dt;
		sum_n[m]+=e;
		sum_vtau[m]+=e*ttau*b.shares;
		sum_vsignedtau[m]+=e*ttau*bwsigned;
	    }
	    tinterval=prevt;
	}
	return new TradeEMoments(currenttime,tau,1,
				 sum_pv_emoments,
				 sum_p2v_emoments,
				 sum_pvtau_emoments,
				 sum_pvsignedtau_emoments,
				 sum_v_emoments,
				 sum_pt_emoments,
				 sum_p2t_emoments,
				 sum_vt_emoments,
				 sum_pvsigned_emoments,
				 sum_vsigned_emoments,
				 sum_ptsigned_emoments,
				 sum_tsigned_emoments,
				 sum_vsignedt_emoments,
				 sum_vsignedv_emoments,
				 sum_dpt_emoments,
				 sum_dpv_emoments,
				 sum_t_emoments,
				 sum_n,
				 sum_vtau,
				 sum_vsignedtau,
				 tb[tb.length-1]);
    }



    private static void checkCalculations(final int tsize,final java.util.Random r)
    {
	final Trade [] tb=getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println("checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);


	final TradeEMoments ro=calculateEMoments(currenttime,tau,tb,tb.length,null);

	// check calculations
	final TradeEMoments temc=calculateTestEMoments(currenttime,tau,tb,tb.length);


	// check values
	checkMatching(ro,temc);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeEMoments em_ref=calculateEMoments(tb[n_split-1].time,tau,tb,n_split,null);



	checkDEM(em_ref);
	if(true)
	{
	    final TradeEMoments em=calculateEMoments(currenttime,tau,tb,tb.length,em_ref);
	    checkMatching(em,ro);
	    checkDEM(em);

	}


	if(true)
	{
	    // check with data cut
	    final int ncut=tb.length-n_split;
	    System.arraycopy(tb,n_split,tb,0,ncut);
	    final TradeEMoments emc=calculateEMoments(currenttime,tau,tb,ncut,em_ref);
	    checkMatching(emc,ro);

	}
	System.err.println("checkCalculations(): test random values OK N points="+tb.length);
    }
}

static final boolean EXP_WEIGHT=true;

static final int N_EMOMENTS=5;

static TradeEMoments calculateEMoments(final long currenttime,
				       final long tau,
				       final BookItem [] tb,
				       final int tsize,
				       final TradeEMoments ref)
{

    final int n=N_EMOMENTS;
    final double []  // all 0 by default
	sum_pv_emoments=new double [n],
	sum_p2v_emoments=new double [n],
	sum_pvtau_emoments=new double [n],
	sum_pvsignedtau_emoments=new double [n],
	sum_v_emoments=new double [n],
	sum_pt_emoments=new double [n],
	sum_p2t_emoments=new double [n],
	sum_vt_emoments=new double [n],
	sum_pvsigned_emoments=new double [n],
	sum_vsigned_emoments=new double [n],
	sum_ptsigned_emoments=new double [n],
	sum_tsigned_emoments=new double [n],
	sum_vsignedt_emoments=new double [n],
	sum_vsignedv_emoments=new double [n],
	sum_dpt_emoments=new double [n],
	sum_dpv_emoments=new double [n],
	sum_t_emoments=new double [n],
	sum_n=new double [n],
	sum_vtau=new double [n],
	sum_vsignedtau=new double [n];


    int np=0;
    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    long tinterval=currenttime;
    for(int i=tsize;--i>=0;np++)
    {
	final BookItem b=tb[i];
	if(b==bref)
	{ // optimization
	    break;
	}
	final double ttau=currenttime-b.time,ts=ttau*atau,dt=(tinterval-b.time)*atau;

	final double p=b.price,bw=b.shares;
	final double bwsigned=b.buy_sell_indicator==BookItem.BOOK_BUY?-bw:(b.buy_sell_indicator==BookItem.BOOK_SELL?bw:0),
	    dtsigned=b.buy_sell_indicator==BookItem.BOOK_BUY?-dt:(b.buy_sell_indicator==BookItem.BOOK_SELL?dt:0);
	final double dp=Double.NaN;//finddpx(tb,0,i+1);

	if(EXP_WEIGHT)
	{
	    for(int m=0;m<n;m++)
	    {
		final double w=Math.exp(-ts/(m+1))/(m+1),bwmw=bw*w,dtmw=dt*w;
		sum_pv_emoments[m]+=p*bwmw;
		sum_p2v_emoments[m]+=p*p*bwmw;
		sum_pvtau_emoments[m]+=p*bwmw*ttau;
		sum_pvsignedtau_emoments[m]+=p*bwsigned*w*ttau;
		sum_v_emoments[m]+=bwmw;
		sum_pt_emoments[m]+=p*dtmw;
		sum_p2t_emoments[m]+=p*p*dtmw;
		sum_vt_emoments[m]+=bw*dtmw;
		sum_pvsigned_emoments[m]+=bwsigned*w*p;
		sum_vsigned_emoments[m]+=bwsigned*w;
		sum_ptsigned_emoments[m]+=dtsigned*w*p;
		sum_tsigned_emoments[m]+=dtsigned*w;
		sum_vsignedt_emoments[m]+=bwsigned*dtmw;
		sum_vsignedv_emoments[m]+=bwsigned*bwmw;
		sum_dpt_emoments[m]+=dp*dtmw;
		sum_dpv_emoments[m]+=dp*bwmw;
		sum_t_emoments[m]+=dtmw;
		sum_n[m]+=w;
		sum_vtau[m]+=ttau*bwmw;
		sum_vsignedtau[m]+=ttau*(bwsigned*w);
	    }
	}
	else
	{
	    final double e=Math.exp(-ts);
	    double we=1;
	    for(int m=0;m<n;m++)
	    {
		final double w=we*e,bwmw=bw*w,dtmw=dt*w;
		sum_pv_emoments[m]+=p*bwmw;
		sum_p2v_emoments[m]+=p*p*bwmw;
		sum_pvtau_emoments[m]+=p*bwmw*ttau;
		sum_pvsignedtau_emoments[m]+=p*bwsigned*w*ttau;
		sum_v_emoments[m]+=bwmw;
		sum_pt_emoments[m]+=p*dtmw;
		sum_p2t_emoments[m]+=p*p*dtmw;
		sum_vt_emoments[m]+=bw*dtmw;
		sum_pvsigned_emoments[m]+=bwsigned*w*p;
		sum_vsigned_emoments[m]+=bwsigned*w;
		sum_ptsigned_emoments[m]+=dtsigned*w*p;
		sum_tsigned_emoments[m]+=dtsigned*w;
		sum_vsignedt_emoments[m]+=bwsigned*dtmw;
		sum_vsignedv_emoments[m]+=bwsigned*bwmw;
		sum_dpt_emoments[m]+=dp*dtmw;
		sum_dpv_emoments[m]+=dp*bwmw;
		sum_t_emoments[m]+=dtmw;
		sum_n[m]+=w;
		sum_vtau[m]+=ttau*bwmw;
		sum_vsignedtau[m]+=ttau*(bwsigned*w);
		we*=(ts/(m+1));
	    }
	}
	/*
	System.err.println("currenttime="+currenttime+"i="+i+" b.time="+b.time+
			   " price="+b.price);
	*/
	tinterval=b.time;
    }

    if(bref!=null)
    { // optimization
	// add missed half slice
	final double ts=(currenttime-bref.time)*atau,dt=(tinterval-ref.currenttime)*atau;
	final double bwsigned=(bref.buy_sell_indicator==BookItem.BOOK_BUY?-bref.shares:(bref.buy_sell_indicator==BookItem.BOOK_SELL?bref.shares:0)),
	    dtsigned=bref.buy_sell_indicator==BookItem.BOOK_BUY?-dt:(bref.buy_sell_indicator==BookItem.BOOK_SELL?dt:0);
	for(int m=0,fact=1;m<n;m++)
	{
	    final double e,dp=Double.NaN;
	    if(EXP_WEIGHT)
	    {
		e=Math.exp(-ts/(m+1))/(m+1);
	    }
	    else
	    {
		e=(Math.exp(-ts)*Math.pow(ts,m))/fact;
		fact*=(m+1);
	    }
	    sum_pt_emoments[m]+=e*bref.price*dt;
	    sum_p2t_emoments[m]+=e*bref.price*bref.price*dt;
	    sum_vt_emoments[m]+=e*bref.shares*dt;
	    sum_vsignedt_emoments[m]+=e*bwsigned*dt;
	    sum_dpt_emoments[m]+=e*dp*dt;
	    sum_t_emoments[m]+=e*dt;
	    sum_ptsigned_emoments[m]+=e*bref.price*dtsigned;
	    sum_tsigned_emoments[m]+=e*dtsigned;
	}


	final double dtr=currenttime-ref.currenttime,ref_adelay=dtr*atau;
	if(EXP_WEIGHT)
	{
	    for(int m=0;m<n;m++)
	    {
		final double exp_delay=Math.exp(-ref_adelay/(m+1));
		sum_pv_emoments[m]+=exp_delay*ref.sum_pv_emoments[m];
		sum_p2v_emoments[m]+=exp_delay*ref.sum_p2v_emoments[m];
		sum_pvtau_emoments[m]+=exp_delay*(ref.sum_pvtau_emoments[m]+dtr*ref.sum_pv_emoments[m]);
		sum_pvsignedtau_emoments[m]+=exp_delay*(ref.sum_pvsignedtau_emoments[m]+dtr*ref.sum_pvsigned_emoments[m]);
		sum_v_emoments[m]+=exp_delay*ref.sum_v_emoments[m];
		sum_pt_emoments[m]+=exp_delay*ref.sum_pt_emoments[m];
		sum_p2t_emoments[m]+=exp_delay*ref.sum_p2t_emoments[m];
		sum_vt_emoments[m]+=exp_delay*ref.sum_vt_emoments[m];
		sum_pvsigned_emoments[m]+=exp_delay*ref.sum_pvsigned_emoments[m];
		sum_vsigned_emoments[m]+=exp_delay*ref.sum_vsigned_emoments[m];
		sum_ptsigned_emoments[m]+=exp_delay*ref.sum_ptsigned_emoments[m];
		sum_tsigned_emoments[m]+=exp_delay*ref.sum_tsigned_emoments[m];
		sum_vsignedt_emoments[m]+=exp_delay*ref.sum_vsignedt_emoments[m];
		sum_vsignedv_emoments[m]+=exp_delay*ref.sum_vsignedv_emoments[m];
		sum_dpt_emoments[m]+=exp_delay*ref.sum_dpt_emoments[m];
		sum_dpv_emoments[m]+=exp_delay*ref.sum_dpv_emoments[m];
		sum_t_emoments[m]+=exp_delay*ref.sum_t_emoments[m];
		sum_n[m]+=exp_delay*ref.sum_n[m];
		sum_vtau[m]+=exp_delay*(ref.sum_vtau[m]+dtr*ref.sum_v_emoments[m]);
		sum_vsignedtau[m]+=exp_delay*(ref.sum_vsignedtau[m]+dtr*ref.sum_vsigned_emoments[m]);
	    }
	}
	else
	{
	    final double exp_delay=Math.exp(-ref_adelay);
	    for(int m=0;m<n;m++)
	    {
		double s_pv=0,s_pvtau=0,s_pvsignedtau=0,s_v=0,s_pt=0,s_vt=0,s_pvsigned=0,s_vsigned=0,s_ptsigned=0,s_tsigned=0,s_vsignedt=0,
		    s_p2v=0,s_p2t=0,
		    s_vsignedv=0,s_vtau=0,s_vsignedtau=0,s_dpt=0,s_dpv=0,s_t=0,s_n=0,we=1;
		// use Newton binomial expansion
		for(int ms=0;ms<=m;ms++)
		{
		    final int mr=m-ms;
		    s_pv+=we*ref.sum_pv_emoments[mr];
		    s_p2v+=we*ref.sum_p2v_emoments[mr];
		    s_pvtau+=we*ref.sum_pvtau_emoments[mr];
		    s_pvsignedtau+=we*ref.sum_pvsignedtau_emoments[mr];
		    s_v+=we*ref.sum_v_emoments[mr];
		    s_pt+=we*ref.sum_pt_emoments[mr];
		    s_p2t+=we*ref.sum_p2t_emoments[mr];
		    s_vt+=we*ref.sum_vt_emoments[mr];
		    s_pvsigned+=we*ref.sum_pvsigned_emoments[mr];
		    s_vsigned+=we*ref.sum_vsigned_emoments[mr];
		    s_ptsigned+=we*ref.sum_ptsigned_emoments[mr];
		    s_tsigned+=we*ref.sum_tsigned_emoments[mr];
		    s_vsignedt+=we*ref.sum_vsignedt_emoments[mr];
		    s_vsignedv+=we*ref.sum_vsignedv_emoments[mr];
		    s_dpt+=we*ref.sum_dpt_emoments[mr];
		    s_dpv+=we*ref.sum_dpv_emoments[mr];
		    s_t+=we*ref.sum_t_emoments[mr];
		    s_n+=we*ref.sum_n[mr];
		    s_vtau+=we*ref.sum_vtau[mr];
		    s_vsignedtau+=we*ref.sum_vsignedtau[mr];
		    we*=(ref_adelay/(ms+1));
		}
		sum_pv_emoments[m]+=exp_delay*s_pv;
		sum_p2v_emoments[m]+=exp_delay*s_p2v;
		sum_pvtau_emoments[m]+=exp_delay*(s_pvtau+dtr*s_pv);
		sum_pvsignedtau_emoments[m]+=exp_delay*(s_pvsignedtau+dtr*s_pvsigned);
		sum_v_emoments[m]+=exp_delay*s_v;
		sum_pt_emoments[m]+=exp_delay*s_pt;
		sum_p2t_emoments[m]+=exp_delay*s_p2t;
		sum_vt_emoments[m]+=exp_delay*s_vt;
		sum_pvsigned_emoments[m]+=exp_delay*s_pvsigned;
		sum_vsigned_emoments[m]+=exp_delay*s_vsigned;
		sum_ptsigned_emoments[m]+=exp_delay*s_ptsigned;
		sum_tsigned_emoments[m]+=exp_delay*s_tsigned;
		sum_vsignedt_emoments[m]+=exp_delay*s_vsignedt;
		sum_vsignedv_emoments[m]+=exp_delay*s_vsignedv;
		sum_dpt_emoments[m]+=exp_delay*s_dpt;
		sum_dpv_emoments[m]+=exp_delay*s_dpv;
		sum_t_emoments[m]+=exp_delay*s_t;
		sum_n[m]+=exp_delay*s_n;
		sum_vtau[m]+=exp_delay*(s_vtau+dtr*s_v);
		sum_vsignedtau[m]+=exp_delay*(s_vsignedtau+dtr*s_vsigned);
	    }
	}
	np++; // previous as one more point
    }


    /*
    for(int m=0;m<n;m++)
    {
	System.err.println("m="+m+
			   " sum_v_emoments[m]="+sum_v_emoments[m]+
			   " sum_pv_emoments[m]="+sum_pv_emoments[m]+
			   " sum_pt_emoments[m]="+sum_pt_emoments[m]);
    }
    */


    final TradeEMoments r=
	new TradeEMoments(currenttime,
			  tau,
			  np,

			  // references
			  sum_pv_emoments,
			  sum_p2v_emoments,
			  sum_pvtau_emoments,
			  sum_pvsignedtau_emoments,
			  sum_v_emoments,
			  sum_pt_emoments,
			  sum_p2t_emoments,
			  sum_vt_emoments,
			  sum_pvsigned_emoments,
			  sum_vsigned_emoments,
			  sum_ptsigned_emoments,
			  sum_tsigned_emoments,
			  sum_vsignedt_emoments,
			  sum_vsignedv_emoments,
			  sum_dpt_emoments,
			  sum_dpv_emoments,
			  sum_t_emoments,
			  sum_n,
			  sum_vtau,
			  sum_vsignedtau,
			  // last processed point
			  tsize>0?tb[tsize-1]:bref);

    if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // for expensive constrains
	TradeEMoments.checkDEM(r);
    }

    return r;
}



static class TradePTMoments {
    final long currenttime,tau;
    final double p_last;

    final double apt0,ap2t0,ap3t0,ap5t0;
    final double d1e,d2e,d3e,d0a;
    final double [][] pvmom;
    final double [] pmom_t0;

    final double [] ap,ap2,ap3,ap5;

    private final BookItem trade_last;

    TradePTMoments(final long currenttime,
		   final long tau,
		   final double [][] pvmom,

		   final BookItem trade_last)
{
    this.currenttime=currenttime;
    this.tau=tau;
    this.pvmom=pvmom;
    this.trade_last=trade_last;
    this.p_last=this.trade_last!=null?this.trade_last.price:Double.NaN;

    final int nt=pvmom.length;
    if(nt<5)
    {
	throw new RuntimeException("There should be at least 5 time points nt="+nt);
    }
    final int np=pvmom[0].length;
    if(np<6)
    {
	throw new RuntimeException("There should be at least 6 p moments np="+np);
    }

    this.pmom_t0=new double [np];
    for(int m=1;m<np;m++)
    {
	// moments forward interpolation
	this.pmom_t0[m]=(1*(pvmom[4][m]/pvmom[4][0])-5*(pvmom[3][m]/pvmom[3][0])+
			 10*(pvmom[2][m]/pvmom[2][0])-10*(pvmom[1][m]/pvmom[1][0])
			 +5*(pvmom[0][m]/pvmom[0][0]));
    }
    if(true)
    {
	final double p1=pmom_t0[1],p2=pmom_t0[2],p3=pmom_t0[3],p4=pmom_t0[4],p5=pmom_t0[5];
	apt0=p1;
	ap2t0=Math.sqrt(p2-p1*p1);
	final double pz3=p3+p1*(-3*p2+2*p1*p1);
	ap3t0=Math.signum(pz3)*Math.pow(Math.abs(pz3),1.0/3);
	final double pz5=p5+p1*(-5*p4+p1*(10*p3+p1*(-10*p2+4*p1*p1)));
	ap5t0=Math.signum(pz5)*Math.pow(Math.abs(pz5),1.0/5);
    }

    this.ap5=new double [nt];
    this.ap3=new double [nt];
    this.ap2=new double [nt];
    this.ap=new double [nt];
    for(int t=0;t<nt;t++)
    {
	final double [] pv=pvmom[t];
	final double p0=pv[0],p1=pv[1]/p0,p2=pv[2]/p0,p3=pv[3]/p0,p4=pv[4]/p0,p5=pv[5]/p0;
	this.ap[t]=p1;
	this.ap2[t]=Math.sqrt(p2-p1*p1);
	final double pz3=p3+p1*(-3*p2+2*p1*p1);
	this.ap3[t]=Math.signum(pz3)*Math.pow(Math.abs(pz3),1.0/3);
	final double pz5=p5+p1*(-5*p4+p1*(10*p3+p1*(-10*p2+4*p1*p1)));
	this.ap5[t]=Math.signum(pz5)*Math.pow(Math.abs(pz5),1.0/5);
    }


    // calculate all D's
    final double h=1;
    d0a=(1*ap[4]-5*ap[3]+10*ap[2]-10*ap[1]+5*ap[0]);
    d1e=(3*ap[4]-16*ap[3]+36*ap[2]-48*ap[1]+25*ap[0])/(12*h);
    d2e=(11*ap[4]-56*ap[3]+114*ap[2]-104*ap[1]+35*ap[0])/(12*h*h);
    d3e=(3*ap[4]-14*ap[3]+24*ap[2]-18*ap[1]+5*ap[0])/(2*h*h*h);


}


    private static void checkMatching(final TradePTMoments r,
				      final TradePTMoments ref,
				      final String txt)
    {
	final double eps=1e-5;
	if(r.tau!=ref.tau||
	   r.currenttime!=ref.currenttime||
	   r.pvmom.length!=ref.pvmom.length)
	{
	    throw new RuntimeException(
		"Mistamtch: "+txt+
		" r.tau="+r.tau+" ref.tau="+ref.tau+
		" r.currenttime="+r.currenttime+" ref.currenttime="+ref.currenttime+
		" r.pvmom.length="+r.pvmom.length+" ref.pvmom.length="+ref.pvmom.length);
	}
	for(int t=0;t<r.pvmom.length;t++)
	{
	    final double [] pv=r.pvmom[t],pvref=ref.pvmom[t];
	    if(pv.length!=pvref.length)
	    {
		throw new RuntimeException(
		    "Array size mismatch:"+
		    " r.pvmom["+t+"].length"+r.pvmom[t].length+" ref.pvmom["+t+"].length"+ref.pvmom[t].length);
	    }
	    for(int m=0;m<pv.length;m++)
	    {
		UnitTests.checkDoubleRel(pv[m],"pv[m]",pvref[m],"pvref[m]",eps);
	    }
	}
	System.err.println("checkMatching(): "+txt+" matching OK");
    }


    private static void checkCalculations(final java.util.Random r)
    {
	final Trade [] tb=new Trade [100000];
	long time=1000;
	for(int i=0;i<tb.length;i++)
	{
	    time+=r.nextInt(10000);
	    final double p=100+10*r.nextDouble();
	    final int v=100+r.nextInt(1000);
	    tb[i]=Trade.getDummyTrade(p,v,r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL,time);
	}

	final long tau=10000*tb.length/20,currenttime=time+20000;


	final TradePTMoments [] rrc=new TradePTMoments [3];
	int nc=0;
	for(byte order_type_to_check : new byte[]{BookItem.BOOK_BUY,BookItem.BOOK_SELL,0})
	{
	    final TradePTMoments ro=calculatePTMoments(currenttime,tau,order_type_to_check,tb,tb.length,null);
	    final int nt=ro.pvmom.length,npv=ro.pvmom[0].length;

	    // check calculations
	    final double [][] sum_pv=new double [nt][npv];
	    final double atau=1.0/tau;
	    for(int i=tb.length;--i>=0;)
	    {
		final Trade b=tb[i];
		if(order_type_to_check!=0 && order_type_to_check!=b.buy_sell_indicator)
		{
		    continue;
		}
		final double ts=(currenttime-b.time)*atau,log_p=Math.log(b.price);
		for(int t=0;t<nt;t++)
		{
		    final double e=b.shares*Math.exp(-ts/(t+1))/(t+1);
		    final double [] pv=sum_pv[t];
		    double log_pf=1;
		    for(int m=0;m<npv;m++)
		    {
			pv[m]+=e*log_pf;
			log_pf*=log_p;
		    }
		}
	    }
	    // check values
	    checkMatching(rrc[nc++]=new TradePTMoments(currenttime,tau,sum_pv,tb[tb.length-1]),
			  ro,
			  "alt calc order_type_to_check="+order_type_to_check);
	    System.err.println("checkCalculations(): test random values OK N points="+tb.length+
			       " order_type_to_check="+order_type_to_check);

	    if(true)
	    {
		// recurrent test
		final int n_split=8*(tb.length/10);
		final TradePTMoments em_ref=calculatePTMoments(tb[n_split-1].time,tau,order_type_to_check,tb,n_split,null);
		final TradePTMoments em=calculatePTMoments(currenttime,tau,order_type_to_check,tb,tb.length,em_ref);
		checkMatching(em,ro,"recurrent order_type_to_check="+order_type_to_check);

		// check with data cut
		final Trade [] tbc=tb.clone();
		final int ncut=tbc.length-n_split;
		System.arraycopy(tbc,n_split,tbc,0,ncut);
		final TradePTMoments emc=calculatePTMoments(currenttime,tau,order_type_to_check,tbc,ncut,em_ref);
		checkMatching(emc,ro," recurrent check");

	    }
	}
	if(true)
	{
	    // check the result by two order types
	    final int nt=rrc[2].pvmom.length,npv=rrc[2].pvmom[0].length;
	    final double [][] sum_pv=new double [nt][npv];
	    for(int t=0;t<nt;t++)
	    {
		for(int m=0;m<npv;m++)
		{
		    sum_pv[t][m]=rrc[0].pvmom[t][m]+rrc[1].pvmom[t][m];
		}
	    }
	    checkMatching(new TradePTMoments(currenttime,tau,sum_pv,tb[tb.length-1]),rrc[2],"sum components");
	}
    }

    private static void checkPV4(final java.util.Random r)
    {
	//final double eps=1e-9;

	final int n0=100000;
	final java.util.ArrayList<Trade> tl=new java.util.ArrayList<Trade>(1000);
	final long dt=10000;
	long time=1000;
	final int nadd=1,v0=100;
	double cpi=2.0/n0;
	for(int na=0;na<nadd;na++)
	{
	    for(int i=0;i<n0;i++)
	    {
		time+=dt;//+r.nextInt(10000);
		final double log_p=4 /*+ 0.2*r.nextDouble() */ +cpi*i;
		final int v=v0;
		if(v<0)
		{
		    throw new RuntimeException("v="+v);
		}
		//System.err.println("v="+v+" log_p="+log_p);
		tl.add(Trade.getDummyTrade(Math.exp(log_p),v,BookItem.BOOK_UNKNOWN,time));
	    }
	}

	final Trade [] tb=tl.toArray(Trade[]::new);
	final long tau=10000*tb.length/50,currenttime=time+20000;
	final TradePTMoments ro=calculatePTMoments(currenttime,tau,(byte)0,tb,tb.length,null);
	for(int t=0;t<ro.ap3.length;t++)
	{
	    final double tau_exp=tau*(1+t);
	    final double [] pv=ro.pvmom[t];
	    final double n=pv[0],c=cpi/dt;
	    final double mx1=pv[1]/n,mx2=pv[2]/n,mx3=pv[3]/n,mx4=pv[4]/n,mx5=pv[5]/n;
	    System.err.println("t:"+t+"\t"+
			       "m1="+(pv[0]*dt/tau/v0)+"(=1)\t"+
			       "m2="+(ro.ap2[t]*ro.ap2[t])+":"+
			       (mx2-mx1*mx1)+":"+(c*c*tau_exp*tau_exp)+"\t"+
			       "m3="+(ro.ap3[t]*ro.ap3[t]*ro.ap3[t])+":"+
			       (mx3-3*mx2*mx1+2*mx1*mx1*mx1)+":"+(2*c*c*c*tau_exp*tau_exp*tau_exp)+"\t"+
			       "m5="+(ro.ap5[t]*ro.ap5[t]*ro.ap5[t]*ro.ap5[t]*ro.ap5[t])+":"+
			       (mx5-5*mx4*mx1+10*mx3*mx1*mx1-10*mx2*mx1*mx1*mx1+4*mx1*mx1*mx1*mx1*mx1)+
			       ":"+(44*c*c*c*c*c*tau_exp*tau_exp*tau_exp*tau_exp*tau_exp)+"\t");
	}
	System.err.println("checkPV4(): matching done");
    }


   /** Unit test.
     */
    public static void main(String [] args)
    {
	final java.util.Random r=new java.util.Random(-719091L);
	for(int k=0;k<20;k++)
	{
	    checkCalculations(r);
	}
	checkPV4(r);
    }

}




static final int N_PTMOMENTS_P=6,N_PTMOMENTS_T=5;

static TradePTMoments calculatePTMoments(final long currenttime,
					 final long tau,
					 final byte order_type_to_check,
					 final BookItem [] tb,
					 final int tsize,
					 final TradePTMoments ref)
{

    final int n=N_PTMOMENTS_P,nt=N_PTMOMENTS_T;
    final double [][]  // all 0 by default
	sum_pv=new double [nt][n];


    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b==bref)
	{ // optimization
	    break;
	}
	if(order_type_to_check!=0 && order_type_to_check!=b.buy_sell_indicator)
	{
	    continue;
	}

	final double ts=(currenttime-b.time)*atau;
	final double log_p=Math.log(b.price),bw=b.shares;
	for(int t=0;t<nt;t++)
	{
	    final double e=bw*Math.exp(-ts/(t+1))/(t+1);
	    final double [] pv=sum_pv[t];
	    double log_pf=1;
	    for(int m=0,npl=pv.length;m<npl;m++)
	    {
		pv[m]+=log_pf*e;
		log_pf*=log_p;
	    }
	}
    }

    if(ref!=null)
    { // optimization
	for(int t=0;t<nt;t++)
	{
	    final double [] pv=sum_pv[t],pvold=ref.pvmom[t];
	    final double ref_adelay=(currenttime-ref.currenttime)*atau;
	    final double exp_delay=Math.exp(-ref_adelay/(t+1));
	    for(int m=0,npl=pv.length;m<npl;m++)
	    {
		pv[m]+=exp_delay*pvold[m];
	    }
	}
    }

    final TradePTMoments r=
	new TradePTMoments(currenttime,
			   tau,
			   sum_pv,
			   // last processed point
			   tsize>0?tb[tsize-1]:bref);

    if(OrdersAndTrades.CONSTRAINS_EXPENSIVE_CHECKS)
    { // for expensive constrains
	// TradeEMoments.checkDEM(r);
    }

    return r;
}









static class IntgrTPFun implements com.polytechnik.utils.QUANC8.DoubleFunction {
    private final double ampl,tau,t_now,omega;
    private final int ppower,mtau;
    IntgrTPFun(final double ampl,
	       final int ppower,
	       final double tau,
	       final int mtau,
	       final double t_now,
	       final double omega)
    {
	this.ampl=ampl;
	this.ppower=ppower;
	this.tau=tau;
	this.mtau=mtau;
	this.t_now=t_now;
	this.omega=omega;
    }
    double getP(final double t)
    {
	return ampl*Math.cos(omega*t);
    }

    public double f(final double t)
    {
	final double p=getP(t);
	double wpe=1;
	for(int mt=0;mt<ppower;mt++)
	{
	    wpe*=p/(mt+1);
	}
	return wpe*Math.exp(-(t_now-t)/((mtau+1)*tau))/(mtau+1);
    }

    static double getExactIntP0(final double t,
				final double tau,final double omega,final double amp,final double t_now)
    {
	return tau*Math.exp(-(t_now-t)/tau);
    }

    static double getExactIntP1(final double t,
				final double tau,final double omega,final double amp,final double t_now)
    {
	if(t>t_now)
	{
	    throw new RuntimeException("t="+t+" t_now="+t_now);
	}
	final double c1=Math.exp(-(t_now-t)/tau);
	final double c2=amp*(Math.cos(omega*t)/tau+omega*Math.sin(omega*t))/(omega*omega+1/(tau*tau));
	return c1*c2;
    }

    static double getExactIntPn(final double t,final int m,final int mt,
				final double tau,final double omega,final double amp,final double t_now)
    {
	final double a=1/(tau*(1+m)),b=omega;
	if(mt<=0)
	{
	    return 1/a;
	}
	else if(mt<=1)
	{
	    return (a*Math.cos(b*t)+b*Math.sin(b*t))/(a*a+b*b);
	}
	final double cosbt=Math.cos(b*t);
	return ((a*cosbt+mt*b*Math.sin(b*t))*Math.pow(cosbt,mt-1)+
		mt*(mt-1)*b*b*getExactIntPn(t,m,mt-2,tau,omega,amp,t_now))/(a*a+mt*mt*b*b);
    }
    static double getExactIntP(final double t,final int m,final int mt,
			       final double tau,final double omega,final double amp,final double t_now)
    {
	if(t>t_now)
	{
	    throw new RuntimeException("t="+t+" t_now="+t_now);
	}
	return (Math.exp(-(t_now-t)/(tau*(1+m)))/(1+m))*Math.pow(amp,mt)*getExactIntPn(t,m,mt,tau,omega,amp,t_now)/Polynomials.fact(mt);
    }

    static double [][] getMatrDefault(final int n,
				      final double t_from,
				      final double t_to,
				      final double tau,final double omega,final double amp,final double t_now)
    {
	final double [][] matr=new double [n][n];
	for(int m=0;m<n;m++)
	{
	    for(int mt=0;mt<n;mt++)
	    {
		matr[m][mt]=getExactIntP(t_to,m,mt,tau,omega,amp,t_now)-
		    getExactIntP(t_from,m,mt,tau,omega,amp,t_now);
	    }
	}
	return matr;
    }



    static double getExactIntP1Sin(final double t,final int m,
				   final double tau,final double omega,final double amp,final double t_now)
    {
	final double a=1/(tau*(1+m)),b=omega;
	final double c2=(a*Math.sin(b*t)-b*Math.cos(b*t))/(a*a+b*b);
	return (Math.exp(-(t_now-t)/(tau*(1+m)))/(1+m))*amp*c2;
    }

    static double [] getDpDTExact(final int n,
				      final double t_from,
				      final double t_to,
				      final double tau,final double omega,final double amp,final double t_now)
    {
	final double [] dpdt=new double [n];
	for(int m=0;m<n;m++)
	{
	    dpdt[m]=-omega*(getExactIntP1Sin(t_to,m,tau,omega,amp,t_now)-
			    getExactIntP1Sin(t_from,m,tau,omega,amp,t_now));
	}
	return dpdt;
    }


    static double [] getExactPk(final int n,
				final double t,
				final int m,
				final double tau,final double omega,final double amp,final double t_now)
    {
	final double [] r=new double [n];
	for(int k=0;k<n;k++)
	{
	    r[k]=getExactIntP(t,m,k,tau,omega,amp,t_now);
	}
	return r;
    }
}

static class IntgrTIFun implements com.polytechnik.utils.QUANC8.DoubleFunction {
    private final double ampl,tau,t_now,omega;
    private final int mtau;
    IntgrTIFun(final double ampl,
	       final double tau,
	       final int mtau,
	       final double t_now,
	       final double omega)
    {
	this.ampl=ampl;
	this.tau=tau;
	this.mtau=mtau;
	this.t_now=t_now;
	this.omega=omega;
    }

    double getI(final double t)
    {
	return -ampl*omega*Math.sin(omega*t);
    }

    public double f(final double t)
    {
	return getI(t)*Math.exp(-(t_now-t)/((mtau+1)*tau))/(mtau+1);
    }

    static double getExactIntI(final double t,final int m,
			       final double tau,final double omega,final double amp,final double t_now)
    {
	if(t>t_now)
	{
	    throw new RuntimeException("t="+t+" t_now="+t_now);
	}
	final double a=1/(tau*(1+m)),b=omega;
	final double c1=Math.exp(-(t_now-t)*a)/(1+m);
	return (-amp*omega)*c1*(a*Math.sin(b*t)-b*Math.cos(b*t))/(a*a+b*b);
    }
    static void checkExactIntegrals()
    {
	final double eps=1e-5;
	final double abserr=0,relerr=1e-8,AMPLITUDE=1.3;
	final int nomax=5000000;

	final int n_tests=300000;
	final java.util.Random r=new java.util.Random(-791091);
	for(int i=0;i<n_tests;i++)
	{
	    final double tau=0.83+r.nextDouble(),t_old=-2-5*r.nextDouble(),t_now=r.nextDouble(),omega=1+r.nextDouble(),
		amp=-1+2*r.nextDouble();
	    final int m=r.nextInt(10);
	    final com.polytechnik.utils.QUANC8 res=
		    com.polytechnik.utils.QUANC8.integrate(
			new IntgrTIFun(amp,tau,m,t_now,omega),
			t_old,
			t_now,
			abserr,
			relerr,
			nomax);
	    if(res.flag!=0)
	    {
		throw new RuntimeException("flag="+res.flag);
	    }
	    final double i_exact=getExactIntI(t_now,m,tau,omega,amp,t_now)-getExactIntI(t_old,m,tau,omega,amp,t_now);
	    UnitTests.checkDoubleRel(res.integral,"res.integral",i_exact,"i_exact",eps);
	}
	System.err.println("IntgrTIFun:checkExactIntegrals() "+n_tests+" OK");
    }

    public static void main(final String [] args)
    {
	checkExactIntegrals();
    }

}


static class TestPolynomialT {
    final double [] fcoefs;
    static final double abserr=0,relerr=1e-13,AMPLITUDE=1.3;
    static final int nomax=5000000;
    final double [][] pmatr;
    final double [] is;
    TestPolynomialT(final double [] fcoefs)
    {
	this.fcoefs=fcoefs;
	final int n=fcoefs.length;
	final double tau=0.83,t_old=-5,t_now=1,omega=1.3;
	this.pmatr=new double [n][n];
	this.is=new double [n];
	for(int m=0;m<n;m++)
	{
	    for(int k=0;k<n;k++)
	    {
		//System.err.println("m="+m+" k="+k);
		final com.polytechnik.utils.QUANC8 res=
		    com.polytechnik.utils.QUANC8.integrate(
			new IntgrTPFun(AMPLITUDE,
				       k,
				       tau,
				       m,
				       t_now,
				       omega),
			t_old,
			t_now,
			abserr,
			relerr,
			nomax);
		this.pmatr[m][k]=res.integral;
		if(res.flag!=0)
		{
		    throw new RuntimeException("flag="+res.flag);
		}

		final double intgr=IntgrTPFun.getExactIntP(t_now,m,k,tau,omega,AMPLITUDE,t_now)-
		    IntgrTPFun.getExactIntP(t_old,m,k,tau,omega,AMPLITUDE,t_now);
		//System.err.println("m="+m+" k="+k+" pmatr[m][k]="+pmatr[m][k]+" intgr="+intgr);
		UnitTests.checkDoubleRel(intgr,"intgr",pmatr[m][k],"pmatr["+m+"]["+k+"]",1e-6);
	    }

	    final int ms=m;
	    final com.polytechnik.utils.QUANC8 resi=
		com.polytechnik.utils.QUANC8.integrate(
		    new
		    IntgrTIFun(AMPLITUDE,
			       tau,
			       m,
			       t_now,
			       omega)
		    {
			final IntgrTPFun pf=new IntgrTPFun(AMPLITUDE,
							   -1,
							   tau,
							   ms,
							   t_now,
							   omega);
			double getI(final double t)
			{
			    // check polynomial
			    final double p=pf.getP(t);
			    double i=0;
			    for(int k=fcoefs.length;--k>=0;)
			    {
				i=i*p/(k+1)+fcoefs[k];
			    }
			    return i;
			}
		    },
		    t_old,
		    t_now,
		    abserr,
		    relerr,
		    nomax);
	    this.is[m]=resi.integral;
	    //System.err.println("m="+m+" I int="+resi.integral);
	    if(resi.flag!=0)
	    {
		throw new RuntimeException("flagi="+resi.flag);
	    }
	    final double [] i_new=IntgrTPFun.getExactPk(n,t_now,m,tau,omega,AMPLITUDE,t_now),
		i_old=IntgrTPFun.getExactPk(n,t_old,m,tau,omega,AMPLITUDE,t_now);
	    double si_exact=0;
	    for(int j=0;j<n;j++)
	    {
		si_exact+=fcoefs[j]*(i_new[j]-i_old[j]);
	    }
	    UnitTests.checkDoubleRel(si_exact,"si_exact",this.is[m],"this.is["+m+"]",1e-6);
	}
	//if(true) throw new RuntimeException();

	final double int_check0=
	    IntgrTPFun.getExactIntP0(t_now,tau,omega,AMPLITUDE,t_now)-
	    IntgrTPFun.getExactIntP0(t_old,tau,omega,AMPLITUDE,t_now),
	    int_check1=
	    IntgrTPFun.getExactIntP1(t_now,tau,omega,AMPLITUDE,t_now)-
	    IntgrTPFun.getExactIntP1(t_old,tau,omega,AMPLITUDE,t_now);
	UnitTests.checkDoubleRel(int_check0,"int_check0",pmatr[0][0],"pmatr[0][0]",1e-6);
	UnitTests.checkDoubleRel(int_check1,"int_check1",pmatr[0][1],"pmatr[0][1]",1e-6);


	/*
	System.err.println("PMATR:");
	for(int m=0;m<n;m++)
	{
	    for(int j=0;j<n;j++)
	    {
		System.err.print(pmatr[m][j]+"\t");
	    }
	    System.err.println("r\t"+is[m]);
	}
	*/
    }

    static void testPol(final int n_tests)
    {
	final double eps=1e-4;
	final java.util.Random r=new java.util.Random(-791091);
	for(int n=2;n<=4;n++)
	{
	    final double [] fcoefs=new double [n],dummy=new double [n];
	    final double [][] dummy2=new double [n][n];
	    for(int i=0;i<n_tests;i++)
	    {
		for(int j=0;j<n;j++)
		{
		    fcoefs[j]=-1+2*r.nextDouble();
		}
		final TestPolynomialT ts=new TestPolynomialT(fcoefs);
		final TradeVX vx=new TradeVX(0,
					     1,
					     dummy,
					     ts.is,
					     dummy,
					     dummy,
					     dummy,
					     dummy2,
					     ts.pmatr,
					     null);
		for(int m=0;m<fcoefs.length;m++)
		{
		    //System.err.println("fcoefs["+m+"]="+fcoefs[m]+" vx.fcoefs["+m+"]="+vx.fcoefs[m]);
		    UnitTests.checkDoubleRel(fcoefs[m],"fcoefs["+m+"]",vx.fcoefs[m],"vx.fcoefs[m]",eps);
		}
		//throw new RuntimeException();
	    }
	    System.err.println("TestPolynomialT:testPol() "+n_tests+" OK n="+n);
	}
    }

    static void testPolExact(final int n_tests)
    {
	final double tau=0.83,t_old=-9,t_now=1,omega=5.33,t_scale=t_now;

	final double eps=1e-6;
	final java.util.Random r=new java.util.Random(-91091);
	for(int n=2;n<=4;n++)
	{
	    final double [] fcoefs=new double [n],dummy=new double [n];
	    final double [][] dummy2=new double [n][n];
	    for(int i=0;i<n_tests;i++)
	    {
		for(int j=0;j<n;j++)
		{
		    fcoefs[j]=-1+2*r.nextDouble();
		}

		// set exact polynomial flow
		final double [] is=new double [n];
		for(int m=0;m<n;m++)
		{
		    final double [] i_new=IntgrTPFun.getExactPk(n,t_now,m,tau,omega,AMPLITUDE,t_scale),
			i_old=IntgrTPFun.getExactPk(n,t_old,m,tau,omega,AMPLITUDE,t_scale);
		    double s=0;
		    for(int j=0;j<n;j++)
		    {
			s+=fcoefs[j]*(i_new[j]-i_old[j]);
		    }
		    is[m]=s;
		}


		final TradeVX vx=new TradeVX(0,
					     1,
					     dummy,
					     is,
					     dummy,
					     dummy,
					     dummy,
					     dummy2,
					     IntgrTPFun.getMatrDefault(n,t_old,t_now,tau,omega,AMPLITUDE,t_scale),
					     null);
		for(int m=0;m<fcoefs.length;m++)
		{
		    //System.err.println("fcoefs["+m+"]="+fcoefs[m]+" vx.fcoefs["+m+"]="+vx.fcoefs[m]);
		    UnitTests.checkDoubleRel(fcoefs[m],"fcoefs["+m+"]",vx.fcoefs[m],"vx.fcoefs[m]",eps);
		}
		//throw new RuntimeException();
	    }
	    System.err.println("TestPolynomialT:testPolExact() "+n_tests+" OK n="+n);
	}
    }

    public static void main(final String[] args)
    {
	final int n_tests=2000;
	testPol(n_tests);
	testPolExact(100*n_tests);
    }
}


static class PrintRoots {
    static final double omega=2*Math.PI,amp=1,t_old=-20,tau=1.4;
    static final int N=5;

    static double [][] getTOscillatorRoots(final double t_now)
    {
	final double t_scale=t_now;
	final double [][] pmatr=IntgrTPFun.getMatrDefault(N,t_old,t_now,tau,omega,amp,t_scale);
	final double [] I=new double [N],dummy=new double [N];
	final double [][] dummy2=new double [N][N];
	for(int m=0;m<N;m++)
	{
	    I[m]=IntgrTIFun.getExactIntI(t_now,m,tau,omega,amp,t_scale)-
		IntgrTIFun.getExactIntI(t_old,m,tau,omega,amp,t_scale);
	}
	final TradeVX vx=new TradeVX(0,
				     1,
				     dummy,
				     I,
				     dummy,
				     dummy,
				     dummy,
				     dummy2,
				     pmatr,
				     null);
	return new double [][]{
	    vx.proot_re,
	    vx.proot_im,
	    new double []{new IntgrTPFun(amp,1,tau,-1,t_now,omega).getP(t_now),
			  new IntgrTIFun(amp,tau,-1,t_now,omega).getI(t_now)}};
    }


    static void printRoots()
    {
	System.err.print("#");
	int ncol=1;
	System.err.print("t("+(ncol++)+")\t");
	System.err.print("P("+(ncol++)+")\t");
	System.err.print("I("+(ncol++)+")\t");
	for(int j=0;j<N-1;j++)
	{
	    System.err.print("X"+j+".re("+(ncol++)+")\t");
	    System.err.print("X"+j+".im("+(ncol++)+")\t");
	}
	for(int j=0;j<N-1;j++)
	{
	    System.err.print("SX"+j+".re("+(ncol++)+")\t");
	}
	System.err.println();

	for(double t=-2;t<2;t+=0.001)
	{
	    final double [][] r=getTOscillatorRoots(t);
	    final double [] root_re=r[0],root_im=r[1],v=r[2];
	    final double [] root_re_sorted=root_re.clone();
	    com.polytechnik.outils.TMPQSortAlgorithm.doubleQuickSort(root_re_sorted,0,root_re_sorted.length-1);
	    System.err.print(t+"\t"+v[0]+"\t"+v[1]+"\t");
	    for(int j=0;j<root_re.length;j++)
	    {
		System.err.print(root_re[j]+"\t"+root_im[j]+"\t");
	    }
	    for(int j=0;j<root_re_sorted.length;j++)
	    {
		System.err.print(root_re_sorted[j]+"\t");
	    }
	    System.err.println();
	}
    }

    public static void main(final String [] args)
    {
	printRoots();
    }
}




static int selectRealRoots(final double [] save_reroots,
			   final int n_roots,
			   final double [] root_re,
			   final double [] root_im,
			   final double eps_im)
{
    int n_reroots=0;
    double rr=Double.NaN;
    for(int m=0;m<n_roots;m++)
    {
	if(rr>root_re[m])
	{
	    throw new RuntimeException("Not sorted");
	}
	rr=root_re[m];
	if(Math.abs(root_im[m])<=eps_im)
	{ // real root
	    save_reroots[n_reroots++]=rr;
	}
    }
    return n_reroots;
}


static final double EPS_IM=1e-5;

private static int getPtrMinDist(final double [] re,final double [] im,final double p_re)
{
    // find closest point to p_re
    double d=Double.POSITIVE_INFINITY;
    int ptr_c=-1;
    for(int i=0;i<re.length;i++)
    {
	final double d1=Math.abs(p_re-re[i]);
	if(Math.abs(im[i])<=EPS_IM && (ptr_c<0||d1<=d))
	{
	    ptr_c=i;
	    d=d1;
	}
    }
    return ptr_c;
}



static class TradeVX {
    final long currenttime,tau;

    private final double []
	sum_v_emoments,
	sum_pvsigned_emoments,
	sum_vsignedv_emoments,
	sum_pv_emoments;
    final double [] sum_vsigned_emoments;
    private final double [][] matr,pmatr;
    // references to previous calculations
    private final BookItem trade_last;

    final double p0,prob0,vsigned0,p1,prob1,vsigned1,px;
    final double [] pcoefs,fcoefs,fcoefsv,proot_re,proot_im,prootv_re,prootv_im;
    final double reroot_min,reroot_max;

    final double d0a_prob,d1a_prob,d0a_vsigned,d1a_vsigned,d0a_pvsigned,d1a_pvsigned,p_c,dp_c,d0a_v,d1a_v;


    TradeVX(final long currenttime,
	    final long tau,

	    // references
	    final double [] sum_v_emoments,
	    final double [] sum_vsigned_emoments,
	    final double [] sum_pvsigned_emoments,
	    final double [] sum_vsignedv_emoments,
	    final double [] sum_pv_emoments,

	    final double [][] matr,
	    final double [][] pmatr,
	    final BookItem trade_last)
	{
	    this.currenttime=currenttime;
	    this.tau=tau;
	    this.sum_v_emoments=sum_v_emoments;
	    this.sum_vsigned_emoments=sum_vsigned_emoments;
	    this.sum_pvsigned_emoments=sum_pvsigned_emoments;
	    this.sum_vsignedv_emoments=sum_vsignedv_emoments;
	    this.sum_pv_emoments=sum_pv_emoments;
	    this.matr=matr;
	    this.pmatr=pmatr;
	    this.trade_last=trade_last;


	    final int n=sum_v_emoments.length;
	    final double [][] msol=new double [n][];
	    for(int m=0;m<n;m++)
	    {
		msol[m]=matr[m].clone();
	    }
	    final int [] ipvt=new int [n];
	    final int info=com.polytechnik.utils.Linsystems.dgefa(msol,n,ipvt);
	    if(info!=0)
	    {
		this.px=Double.NaN;
		this.p0=this.prob0=this.vsigned0=this.p1=this.prob1=this.vsigned1=Double.NaN;
		this.pcoefs=new double [n];
		java.util.Arrays.fill(this.pcoefs,Double.NaN);
	    }
	    else
	    {
		this.px=-(sum_vsigned_emoments[0]*sum_pv_emoments[1]-sum_vsigned_emoments[1]*sum_pv_emoments[0])/(sum_vsigned_emoments[1]*sum_v_emoments[0]-sum_vsigned_emoments[0]*sum_v_emoments[1]);

		final double [] rpart=new double [n];
		System.arraycopy(sum_vsigned_emoments,0,rpart,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,rpart,1);
		this.prob0=rpart[0];
		this.prob1=-rpart[1];

		System.arraycopy(sum_vsignedv_emoments,0,rpart,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,rpart,1);
		this.vsigned0=rpart[0];
		this.vsigned1=-rpart[1];

		System.arraycopy(sum_pv_emoments,0,rpart,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,rpart,1);
		this.p0=rpart[0];
		this.p1=-rpart[1];
		// save result is saved to this.pcoefs, rpart will not be clobbered later on.
		this.pcoefs=rpart;
	    }


	    // pmatr
	    for(int m=0;m<n;m++)
	    {
		System.arraycopy(pmatr[m],0,msol[m],0,n);
	    }
	    final int finfo=com.polytechnik.utils.Linsystems.dgefa(msol,n,ipvt);
	    //System.err.println("finfo="+finfo+" msol[0][0]="+msol[0][0]);
	    if(finfo!=0)
	    {
		this.fcoefsv=this.fcoefs=new double [n];
		java.util.Arrays.fill(this.fcoefs,Double.NaN);
	    }
	    else
	    {
		final double [] rpart=new double [n];
		System.arraycopy(sum_vsigned_emoments,0,rpart,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,rpart,1);
		this.fcoefs=rpart;

		this.fcoefsv=new double [n];
		System.arraycopy(sum_v_emoments,0,this.fcoefsv,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.fcoefsv,1);
	    }

	    /*
	    System.err.println("VX PMATR: finfo="+finfo);
	    for(int m=0;m<n;m++)
	    {
		for(int j=0;j<n;j++)
		{
		    System.err.print(pmatr[m][j]+"\t");
		}
		System.err.println("r\t"+sum_vsigned_emoments[m]);
	    }
	    */

	    // find roots
	    final double [] pcoe=new double [n];
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0,w=1;k<n;k++)
	    {
		pcoe[k]=this.fcoefs[k]/w;
		w*=(k+1);
	    }
	    final PolynomialRootsFinder.PRoots ut=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    this.proot_re=ut.re_root;
	    this.proot_im=ut.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(n-1,this.proot_re,this.proot_im);
	    // find min/max real part roots
	    this.reroot_min=this.proot_re[0];
	    this.reroot_max=this.proot_re[this.proot_re.length-1];

	    // find closest point to p_last
	    final int ptr_c=getPtrMinDist(this.proot_re,this.proot_im,(trade_last!=null?trade_last.price:Double.NaN));
	    p_c=ptr_c<0?Double.NaN:this.proot_re[ptr_c];
	    dp_c=getIP(p_c);

	    // use scale=1 expected by PolynomialRoot
	    for(int k=0,w=1;k<n;k++)
	    {
		pcoe[k]=this.fcoefsv[k]/w;
		w*=(k+1);
	    }
	    final PolynomialRootsFinder.PRoots uv=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    this.prootv_re=uv.re_root;
	    this.prootv_im=uv.im_root;
	    SortArrays.sortArrayWithLoad(n-1,this.prootv_re,this.prootv_im);

	    if(false) // debug
	    {
		if(currenttime>10*3600*TraderInterface.TIME_SCALE)
		{
		    System.err.println("=========== test currenntime\t"+currenttime+
				       "\ttau\t"+tau+
				       "\tp_last\t"+(trade_last!=null?trade_last.price:"N/A")+
				       "\t$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
		    System.err.println("n\tcoef\tre(x)\tim(x)\tI(re(x))\tDIDp(re(x))");
		    for(int k=0;k<n;k++)
		    {
			System.err.println(String.valueOf(k)+"\t"+
					   fcoefs[k]+"\t"+
					   (k<n-1?proot_re[k]+"\t"+proot_im[k]+"\t"+getIP(proot_re[k])+"\t"+getDIDP(proot_re[k]):""));
		    }
		}
		if(currenttime>(200+10*3600)*TraderInterface.TIME_SCALE)
		{
		    throw new RuntimeException("EXIT");
		}
	    }

	    if(n==5)
	    {
		final double h=1;

		final double [] prob=new double [n];
		for(int m=0;m<n;m++)
		{
		    prob[m]=sum_vsigned_emoments[m]/sum_v_emoments[m];
		}
		d0a_prob=1*prob[4]-5*prob[3]+10*prob[2]-10*prob[1]+5*prob[0];
		d1a_prob=(25*prob[4]-122*prob[3]+234*prob[2]-214*prob[1]+77*prob[0])/(12*h);

		d0a_vsigned=1*sum_vsigned_emoments[4]-5*sum_vsigned_emoments[3]+10*sum_vsigned_emoments[2]-10*sum_vsigned_emoments[1]+5*sum_vsigned_emoments[0];
		d1a_vsigned=(25*sum_vsigned_emoments[4]-122*sum_vsigned_emoments[3]+234*sum_vsigned_emoments[2]-214*sum_vsigned_emoments[1]+77*sum_vsigned_emoments[0])/(12*h);
		d0a_pvsigned=1*sum_pvsigned_emoments[4]-5*sum_pvsigned_emoments[3]+10*sum_pvsigned_emoments[2]-10*sum_pvsigned_emoments[1]+5*sum_pvsigned_emoments[0];
		d1a_pvsigned=(25*sum_pvsigned_emoments[4]-122*sum_pvsigned_emoments[3]+234*sum_pvsigned_emoments[2]-214*sum_pvsigned_emoments[1]+77*sum_pvsigned_emoments[0])/(12*h);

		d0a_v=1*sum_v_emoments[4]-5*sum_v_emoments[3]+10*sum_v_emoments[2]-10*sum_v_emoments[1]+5*sum_v_emoments[0];
		d1a_v=(25*sum_v_emoments[4]-122*sum_v_emoments[3]+234*sum_v_emoments[2]-214*sum_v_emoments[1]+77*sum_v_emoments[0])/(12*h);
	    }
	    else
	    {
		d0a_prob=d1a_prob=d0a_vsigned=d1a_vsigned=d0a_pvsigned=d1a_pvsigned=d0a_v=d1a_v=Double.NaN;
	    }
	}

    double getIP(final double p)
    {
	return getD0(this.fcoefs,p);
    }
    double getDIDP(final double p)
    {
	return getD1(this.fcoefs,p);
    }


    private static TradeVX calculateTestTradeVX(final long currenttime,
						final long tau,
						final boolean flag_skipnoorigtime,
						final boolean flag_includeorig,
						final Trade [] tb,
						final int tsize)
    {
	final int n=N_EMOMENTS;
	final double []  // all 0 by default
	    sum_v_emoments=new double [n],
	    sum_vsigned_emoments=new double [n],
	    sum_pvsigned_emoments=new double [n],
	    sum_vsignedv_emoments=new double [n],
	    sum_pv_emoments=new double [n];
	final double [][] matr=new double [n][n],pmatr=new double [n][n];
	final int JE=flag_includeorig?2:1;
	final double atau=1.0/tau;
	long prevt=currenttime;
	for(int i=tsize;--i>=0;)
	{
	    final Trade b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;
	    final long origtime=b.origtime;
	    if(origtime<=0 && flag_skipnoorigtime)
	    { // skip trades with unknown origtime
		continue;
	    }

	    final int sign_market,sign_limit;
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		sign_market=-1;  // sell just came
		sign_limit=1; // buy came some time ago
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		sign_market=1; // buy just came
		sign_limit=-1; // sell came some time ago
	    }
	    else
	    {
		throw new RuntimeException("b.buy_sell_indicator="+b.buy_sell_indicator);
	    }

	    // now we need to run twice: for market & limit
	    for(int jj=0;jj<JE;jj++)
	    {
		final double bwsigned;
		final long dt;
		if(jj<=0)
		{ // market order
		    dt=currenttime-prevt;
		    bwsigned=b.shares*sign_market;
		}
		else
		{ // limit order
		    dt=currenttime-origtime;
		    bwsigned=b.shares*sign_limit;
		}
		final double ts=dt*atau;

		for(int m=0;m<n;m++)
		{
		    final double w=Math.exp(-ts/(m+1))/(m+1),bwmw=w*b.shares;
		    sum_v_emoments[m]+=bwmw;
		    sum_vsigned_emoments[m]+=bwsigned*w;
		    sum_pvsigned_emoments[m]+=bwsigned*w*b.price;
		    sum_vsignedv_emoments[m]+=bwsigned*bwmw;
		    sum_pv_emoments[m]+=b.price*bwmw;
		    final double [] sum_matr_vt=matr[m],sum_pmatr_vt=pmatr[m];
		    double we=1,wpe=1;
		    for(int mt=0;mt<n;mt++)
		    {
			sum_matr_vt[mt]+=we*w*b.shares;
			sum_pmatr_vt[mt]+=wpe*w*b.shares;
			we*=(ts/(mt+1));
			wpe*=(b.price/(mt+1));
		    }
		}
	    }// end jj loop
	}// end i loop
	return new TradeVX(currenttime,
			   tau,
			   sum_v_emoments,
			   sum_vsigned_emoments,
			   sum_pvsigned_emoments,
			   sum_vsignedv_emoments,
			   sum_pv_emoments,
			   matr,
			   pmatr,
			   // last processed point
			   tsize>0?tb[tsize-1]:null);
    }


    private static void checkMatching(final TradeVX r,final TradeVX ref)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	final int n=r.sum_v_emoments.length;
	final double eps=1e-5;
	for(int m=0;m<n;m++)
	{
	    UnitTests.checkDoubleRel(r.sum_v_emoments[m],"r.sum_v_emoments[m]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsigned_emoments[m],"r.sum_vsigned_emoments[m]",ref.sum_vsigned_emoments[m],"ref.sum_vsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pvsigned_emoments[m],"r.sum_pvsigned_emoments[m]",ref.sum_pvsigned_emoments[m],"ref.sum_pvsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsignedv_emoments[m],"r.sum_vsignedv_emoments[m]",ref.sum_vsignedv_emoments[m],"ref.sum_vsignedv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pv_emoments[m],"r.sum_pv_emoments[m]",ref.sum_pv_emoments[m],"ref.sum_pv_emoments[m]",eps);

	    UnitTests.checkDoubleRel(r.sum_v_emoments[m],"r.sum_v_emoments[m]",r.matr[m][0],"r.matr[m][0]",eps);
	    UnitTests.checkDoubleRel(r.pmatr[m][0],"r.pmatr[m][0]",r.matr[m][0],"r.matr[m][0]",eps);
	    UnitTests.checkDoubleRel(r.sum_pv_emoments[m],"r.sum_pv_emoments[m]",r.pmatr[m][1],"r.pmatr[m][1]",eps);

	    for(int mt=0;mt<n;mt++)
	    {
		UnitTests.checkDoubleRel(r.matr[m][mt],"r.matr[m][mt]",ref.matr[m][mt],"ref.matr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.pmatr[m][mt],"r.pmatr[m][mt]",ref.pmatr[m][mt],"ref.pmatr[m][mt]",eps);
	    }
	}
	System.err.println("TradeVX:checkMatching(): matching OK");
    }

    /*private*/ static long getRandomTradesPolynomialP(final Trade [] tb,
						   final java.util.Random r,
						   final long tau,
						   final double [] polp)
    {
	final long tmax=tb.length*10000L+1000*1000L;
	final long currenttime=tmax+r.nextInt(1000);
	final double atau=1.0/tau;

	long t=tmax;
	for(int i=tb.length;i-->0;)
	{
	    t-=1+r.nextInt(10000);
	    final double ts=(currenttime-t)*atau;
	    final double p=getD0(polp,ts); // p=sum_0^L polp[k] ts^k / k!
	    final int v=100+r.nextInt(1000);
	    tb[i]=new Trade(p,
			    v,
			    null,
			    r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
			    t,
			    0,
			    0,
			    t-r.nextInt(100000),
			    null,
			    Trade.TRADE_TYPE_REGULAR,
			    null,
			    null,
			    i);
	    //System.err.println("i="+i+" p="+p+" v="+v);
	}
	return currenttime;
    }

    private static void checkExactPolynomiansP(final int tsize,final java.util.Random r)
    {
	final boolean flag_skipnoorigtime=true,flag_includeorig=false;
	final double eps=1e-6;

	final double [] polp=new double [N_EMOMENTS];
	for(int m=0;m<polp.length;m++)
	{
	    polp[m]=-1+2*r.nextDouble();
	}
	final Trade [] tb=new Trade [tsize];
	final long tau=10000L*tb.length/20;
	final long currenttime=getRandomTradesPolynomialP(tb,r,tau,polp);

	final TradeVX ro=calculateTradeVX(currenttime,
					  tau,
					  flag_skipnoorigtime,
					  flag_includeorig, // false for exact match
					  tb,
					  tsize,
					  null);

	for(int m=0;m<polp.length;m++)
	{
	    System.err.println("polp["+m+"]="+polp[m]+" ro.pcoefs[m]="+ro.pcoefs[m]);
	    UnitTests.checkDoubleRel(polp[m],"polp["+m+"]",ro.pcoefs[m],"ro.pcoefs[m]",eps);
	}
	System.err.println("TradeVX::checkExactPolynomiansP(): test OK coefficients matched, N points="+tsize);
    }


   private static long getRandomTradesPolynomialVP(final Trade [] tb,
						   final java.util.Random r,
						   final long tau,
						   final double [] polvp)
    {
	if((tb.length/2)*2!=tb.length)
	{
	    throw new RuntimeException("tb.length="+tb.length);
	}
	final long tmax=tb.length*10000L+1000*1000L;
	final long currenttime=tmax+r.nextInt(1000);

	long t=tmax;
	for(int i=tb.length-1;i>=0;i-=2)
	{
	    t-=1+r.nextInt(10000);
	    final double p0=-1+2*r.nextDouble(),v=getD0(polvp,p0); // v=sum_0^L polvp[k] p^k / (k+1)!
	    if(Math.abs(v)>1)
	    {
		throw new RuntimeException("v="+v);
	    }
	    final double vscale=0.1e0*(1+r.nextDouble())*Integer.MAX_VALUE;
	    final int vplus=(int)Math.round(vscale*(1+v)),vminus=(int)Math.round(vscale*(1-v));
	    tb[i]=new Trade(p0,
			    vplus,
			    null,
			    BookItem.BOOK_SELL,
			    t,
			    0,
			    0,
			    t-r.nextInt(100000),
			    null,
			    Trade.TRADE_TYPE_REGULAR,
			    null,
			    null,
			    i);
	    tb[i-1]=new Trade(p0,
			      vminus,
			      null,
			      BookItem.BOOK_BUY,
			      t,
			      0,
			      0,
			      t-r.nextInt(100000),
			      null,
			      Trade.TRADE_TYPE_REGULAR,
			      null,
			      null,
			      i-1);
	    //System.err.println("i="+i+" p="+p+" v="+v);
	}
	return currenttime;
    }


    private static void checkExactPolynomiansVP(final int tsize,final java.util.Random r)
    {
	final boolean flag_skipnoorigtime=true,flag_includeorig=false;
	final double eps=1e-3; // lower eps because we convert to int

	final double [] polvp=new double [N_EMOMENTS];
	double scaleto1=0;
	for(int m=0,fact=1;m<polvp.length;m++)
	{
	    polvp[m]=-1+2*r.nextDouble();
	    // // make sure the values are large enough
	    // polvp[m]=(-1+2*r.nextInt(2))*(0.2+0.7999999*r.nextDouble());
	    scaleto1+=1.0/fact;
	    fact*=(m+1);
	}
	// bring to [-1..1] range
	for(int m=0;m<polvp.length;m++)
	{
	    polvp[m]/=scaleto1;
	}

	final Trade [] tb=new Trade [tsize];
	final long tau=10000L*tb.length/55;
	final long currenttime=getRandomTradesPolynomialVP(tb,r,tau,polvp);

	final TradeVX ro=calculateTradeVX(currenttime,
					  tau,
					  flag_skipnoorigtime,
					  flag_includeorig, // false for exact match
					  tb,
					  tsize,
					  null);
	UnitTests.checkDoubleArrayRel(polvp,"polvp",ro.fcoefs,"ro.fcoefs",eps);

	final double p0=-1+2*r.nextDouble(),ip=ro.getIP(p0);
	final double v=getD0(polvp,p0);	// v=sum_0^L polvp[k] p^k / (k+1)!
	//System.err.println("v="+v+" ip="+ip+" at p0="+p0);
	UnitTests.checkDoubleRel(v,"v",ip,"ip",eps);

	System.err.println("TradeVX::checkExactPolynomiansVP(): test OK coefficients matched, N points="+tsize);

    }


    static TradeVX getSum2(final TradeVX sa,final TradeVX sb,final boolean sign_minus)
    {
	final double s=sign_minus?-1:1;
	final int n=sa.sum_v_emoments.length;
	final double []  // all 0 by default
	    sum_v_emoments=new double [n],
	    sum_vsigned_emoments=new double [n],
	    sum_pvsigned_emoments=new double [n],
	    sum_vsignedv_emoments=new double [n],
	    sum_pv_emoments=new double [n];
	final double [][] matr=new double [n][n],pmatr=new double [n][n];
	for(int m=0;m<n;m++)
	{
	    sum_v_emoments[m]=sa.sum_v_emoments[m]+s*sb.sum_v_emoments[m];
	    sum_vsigned_emoments[m]=sa.sum_vsigned_emoments[m]+s*sb.sum_vsigned_emoments[m];
	    sum_pvsigned_emoments[m]=sa.sum_pvsigned_emoments[m]+s*sb.sum_pvsigned_emoments[m];
	    sum_vsignedv_emoments[m]=sa.sum_vsignedv_emoments[m]+s*sb.sum_vsignedv_emoments[m];
	    sum_pv_emoments[m]=sa.sum_pv_emoments[m]+s*sb.sum_pv_emoments[m];
	    for(int mt=0;mt<n;mt++)
	    {
		matr[m][mt]+=sa.matr[m][mt]+s*sb.matr[m][mt];
		pmatr[m][mt]+=sa.pmatr[m][mt]+s*sb.pmatr[m][mt];
	    }
	}

	return new TradeVX(sa.currenttime,
			   sa.tau,
			   sum_v_emoments,
			   sum_vsigned_emoments,
			   sum_pvsigned_emoments,
			   sum_vsignedv_emoments,
			   sum_pv_emoments,
			   matr,
			   pmatr,
			   sa.trade_last==sb.trade_last?sa.trade_last:null);
    }

    private static void matchWithTest(final TradeVX r,
				      final boolean flag_skipnoorigtime,
				      final boolean flag_includeorig,
				      final Trade [] tb,
				      final int tsize)
    {
	final TradeVX ref=calculateTestTradeVX(r.currenttime,
					       r.tau,
					       flag_skipnoorigtime,
					       flag_includeorig,
					       tb,
					       tsize);
	checkMatching(r,ref);

	final Trade [] tb2;
	final TradeEMoments  refem;
	if(flag_includeorig)
	{
	    tb2=TradeEMoments.getTradesOrigExeSwapped(tb,tsize);
	    refem=TradeEMoments.getSum2(
		calculateEMoments(r.currenttime,
				  r.tau,
				  tb,
				  tsize,
				  null),
		calculateEMoments(r.currenttime,
				  r.tau,
				  tb2,
				  tsize,
				  null));
	}
	else
	{
	    tb2=null;
	    refem=calculateEMoments(r.currenttime,
				    r.tau,
				    tb,
				    tsize,
				    null);
	}
	TradeEMoments.checkMatchingTradeVX(refem,ref);


	if(flag_includeorig)
	{
	    final TradeVX refemvx=getSum2(
		calculateTestTradeVX(r.currenttime,
				     r.tau,
				     flag_skipnoorigtime,
				     false,
				     tb,
				     tsize),
		calculateTestTradeVX(r.currenttime,
				     r.tau,
				     flag_skipnoorigtime,
				     false,
				     tb2,
				     tsize),
		false);
	    checkMatching(refemvx,ref);
	}
    }


    private static void checkCalculations(final int tsize,final java.util.Random r)
    {
	final boolean flag_skipnoorigtime=true;
	for(boolean flag_includeorig:new boolean[]{false,true})
	{

	    final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	    final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	    System.err.println("checkCalculations(): test with N points="+tb.length+
			       " currenttime="+currenttime+" tau="+tau+
			       " flag_includeorig="+flag_includeorig);
	    final TradeVX ro=calculateTradeVX(currenttime,
					      tau,
					      flag_skipnoorigtime,
					      flag_includeorig,
					      tb,
					      tsize,
					      null);
	    matchWithTest(ro,flag_skipnoorigtime,flag_includeorig,tb,tsize);

	    // recurrent test
	    final int n_split=8*(tb.length/10);
	    final TradeVX em_ref=calculateTradeVX(tb[n_split-1].time,tau,flag_skipnoorigtime,flag_includeorig,tb,n_split,null);
	    matchWithTest(em_ref,flag_skipnoorigtime,flag_includeorig,tb,n_split);

	    final TradeVX em_fromref=calculateTradeVX(currenttime,tau,flag_skipnoorigtime,flag_includeorig,tb,tsize,em_ref);
	    checkMatching(ro,em_fromref);
	    matchWithTest(em_fromref,flag_skipnoorigtime,flag_includeorig,tb,tsize);

	    // check with data cut
	    final int ncut=tb.length-n_split;
	    System.arraycopy(tb,n_split,tb,0,ncut);
	    final TradeVX emc_cut=calculateTradeVX(currenttime,tau,flag_skipnoorigtime,flag_includeorig,tb,ncut,em_ref);
	    checkMatching(ro,emc_cut);
	}
	System.err.println("TradeVX::checkCalculations(): test random values OK N points="+tsize);
    }


    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-13091L);
	for(int i=0;i<7;i++)
	{
	    checkExactPolynomiansP(100010+r.nextInt(1000010),r);
	    checkExactPolynomiansVP(2*(100010+r.nextInt(100000)),r);
	    checkCalculations(10010+r.nextInt(1000010),r);
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeVX::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }
}





static TradeVX calculateTradeVX(final long currenttime,
				final long tau,
				final boolean flag_skipnoorigtime,
				final boolean flag_includeorig,
				final Trade [] tb,
				final int tsize,
				final TradeVX ref)
{
    if(!flag_skipnoorigtime)
    {
	throw new RuntimeException("flag_skipnoorigtime="+flag_skipnoorigtime+
				   " is not supported yet");
    }

    final int n=N_EMOMENTS;
    final double []  // all 0 by default
	sum_v_emoments=new double [n],
	sum_vsigned_emoments=new double [n],
	sum_pvsigned_emoments=new double [n],
	sum_vsignedv_emoments=new double [n],
	sum_pv_emoments=new double [n];
    final double [][] matr=new double [n][n],pmatr=new double [n][n];


    final int JE=flag_includeorig?2:1;

    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    long prevt=currenttime;
    gloop:for(int i=tsize;--i>=0;)
    {
	final Trade b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break gloop;
	}
	prevt=b.time;
	final long origtime=b.origtime;
	if(origtime<=0 && flag_skipnoorigtime)
	{ // skip trades with unknown origtime
	    continue gloop;
	}

	final int sign_market,sign_limit;
	if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	{
	    sign_market=-1;  // sell just came
	    sign_limit=1; // buy came some time ago
	}
	else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	{
	    sign_market=1; // buy just came
	    sign_limit=-1; // sell came some time ago
	}
	else
	{
	    throw new RuntimeException("b.buy_sell_indicator="+b.buy_sell_indicator);
	}

    final double bw=b.shares,p=b.price;
    // now we need to run twice: for market & limit
    for(int jj=0;jj<JE;jj++)
    {
	final double bwsigned;
	final long dt;
	if(jj<=0)
	{ // market order
	    dt=currenttime-prevt;
	    bwsigned=bw*sign_market;
	}
	else
	{ // limit order
	    dt=currenttime-origtime;
	    bwsigned=bw*sign_limit;
	}
	final double ts=dt*atau;

	for(int m=0;m<n;m++)
	{
	    final double w=Math.exp(-ts/(m+1))/(m+1),bwmw=bw*w;
	    sum_v_emoments[m]+=bwmw;
	    sum_vsigned_emoments[m]+=bwsigned*w;
	    sum_pvsigned_emoments[m]+=bwsigned*w*p;
	    sum_vsignedv_emoments[m]+=bwsigned*bwmw;
	    sum_pv_emoments[m]+=p*bwmw;
	    final double [] sum_matr_vt=matr[m],sum_pmatr_vt=pmatr[m];
	    double we=1,wpe=1;
	    for(int mt=0;mt<n;mt++)
	    {
		sum_matr_vt[mt]+=bwmw*we;
		sum_pmatr_vt[mt]+=bwmw*wpe;
		final double mt1=mt+1;
		we*=ts/mt1;
		wpe*=p/mt1;
	    }
	}
    }// end jj loop
    }// end i loop

    if(bref!=null)
    { // optimization
	final double ref_adelay=(currenttime-ref.currenttime)*atau;
	for(int m=0;m<n;m++)
	{
	    final double exp_delay=Math.exp(-ref_adelay/(m+1));
	    sum_v_emoments[m]+=exp_delay*ref.sum_v_emoments[m];
	    sum_vsigned_emoments[m]+=exp_delay*ref.sum_vsigned_emoments[m];
	    sum_pvsigned_emoments[m]+=exp_delay*ref.sum_pvsigned_emoments[m];
	    sum_vsignedv_emoments[m]+=exp_delay*ref.sum_vsignedv_emoments[m];
	    sum_pv_emoments[m]+=exp_delay*ref.sum_pv_emoments[m];
	    final double [] sum_matr_vt=matr[m],ref_sum_matr_vt=ref.matr[m],
		sum_pmatr_vt=pmatr[m],ref_sum_pmatr_vt=ref.pmatr[m];
	    // use Newton binomial expansion
	    for(int mt=0;mt<n;mt++)
	    {
		double we=1,s_v=0;
		for(int ms=0;ms<=mt;ms++)
		{
		    s_v+=we*ref_sum_matr_vt[mt-ms];
		    we*=(ref_adelay/(ms+1));
		}
		sum_matr_vt[mt]+=exp_delay*s_v;
		sum_pmatr_vt[mt]+=exp_delay*ref_sum_pmatr_vt[mt];
	    }
	}
    }

    final TradeVX r=
	new TradeVX(currenttime,
		    tau,
		    sum_v_emoments,
		    sum_vsigned_emoments,
		    sum_pvsigned_emoments,
		    sum_vsignedv_emoments,
		    sum_pv_emoments,
		    matr,
		    pmatr,
		    // last processed point
		    tsize>0?tb[tsize-1]:bref);

    return r;
}



static double getD0(final double [] coefs,final double x)
{
    double i=0;
    // i=sum_0^L coefs[k] p^k / k!
    for(int k=coefs.length;--k>=0;)
    {
	i=i*x/(k+1)+coefs[k];
    }
    return i;
}

static double getD1(final double [] coefs,final double x)
{

    double i=0;
    // i=sum_1^L coefs[k] p^{k-1} / k!
    for(int k=coefs.length;--k>=1;)
    {
	i=i*x/k+coefs[k];
    }
    return i;
}

static double getD2(final double [] coefs,final double x)
{

    double i=0;
    // i=sum_1^L coefs[k] p^{k-2} / (k-1)!
    for(int k=coefs.length;--k>=2;)
    {
	i=i*x/(k-1)+coefs[k];
    }
    return i;
}

static double getD3(final double [] coefs,final double x)
{

    double i=0;
    // i=sum_1^L coefs[k] p^{k-3} / (k-2)!
    for(int k=coefs.length;--k>=3;)
    {
	i=i*x/(k-2)+coefs[k];
    }
    return i;
}


static double getD0WithScale(final double [] coefs,final double x,final double [] scale)
{

    double i=0;
    // i=sum_0^L coefs[k] p^k / scale[k]
    for(int k=coefs.length;--k>=0;)
    {
	i=i*x+coefs[k]/scale[k];
    }
    return i;
}

static double getD1WithScale(final double [] coefs,final double x,final double [] scale)
{

    double i=0;
    // i=sum_1^L coefs[k] k p^{k-1} / scale[k]
    for(int k=coefs.length;--k>=1;)
    {
	i=i*x+k*coefs[k]/scale[k];
    }
    return i;
}

static double getI1Integrated(final double [] coefs,final double x,final double [] scale)
{
    double i=0;
    for(int k=coefs.length;--k>=0;)
    {
	i=x*(i+(coefs[k]/scale[k])/(k+1));
    }
    return i;
}


static double getD2WithScale(final double [] coefs,final double x,final double [] scale)
{

    double i=0;
    // i=sum_2^L coefs[k] k(k-1) p^{k-2} / scale[k]
    for(int k=coefs.length;--k>=2;)
    {
	i=i*x+(k*(k-1))*coefs[k]/scale[k];
    }
    return i;
}

static double getD3WithScale(final double [] coefs,final double x,final double [] scale)
{

    double i=0;
    // i=sum_3^L coefs[k] k(k-1)(k-2) p^{k-3} / scale[k]
    for(int k=coefs.length;--k>=3;)
    {
	i=i*x+(k*(k-1)*(k-2))*coefs[k]/scale[k];
    }
    return i;
}


static double getPolRootNorm(final double [] coefs,final double rex,final double imx)
{
    double res=0,ims=0;
    for(int k=coefs.length;--k>=0;)
    {
	final double res1=(res*rex-ims*imx)+coefs[k];
	final double ims1=(ims*rex+res*imx);
	res=res1;
	ims=ims1;
    }
    return Math.abs(res)+Math.abs(ims);
}

static class TradePV {
    final long currenttime,tau;

    private final double []
	sum_v_emoments, // [m]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i
	sum_vsigned_emoments,
	sum_vsignedetau_emoments,
	sum_t_emoments, // [m]=sum exp(-(t0-t_i)/((m+1)*tau)) dt_i
	sum_tsigned_emoments,
	sum_pv_emoments, // [m]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i p_i
	sum_pt_emoments; // [m]=sum exp(-(t0-t_i)/((m+1)*tau)) dt_i p_i
    private final double [][]
	pvmatr, // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i p^k/k!
	ptmatr, // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) dt_i p^k/k!
	matr, // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i (t0-t_i)^k/k!
	tmatr, // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) dt_i (t0-t_i)^k/k!
	taumatr, // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i *tauo_i* (t0-t_i)^k/k!
	ptaumatr // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i *tauo_i* p^k/k!
	;
    // references to previous calculations
    private final BookItem trade_last;

    final double [] sum_dpdt,ptcoefs,vtcoefs,vtsignedcoefs,vsignedetaucoefs,optcoefs,ovtcoefs,vpcoefs,ovpcoefs,
	dptpcoefs,dpvpcoefs;

    //final double [] pvroot_re,pvroot_im;

    final double dpx,dpy,odpx,odpy,p_c,d0p_c,d1p_c,d2p_c;

    final double d0pdt,d1pdt,d2pdt,d0Idt,d1Idt,d2Idt,d0Idp,d1Idp,d2Idp,d0oIdp,d1oIdp,d2oIdp;
    final double d0Idt0,d1Idt0,d0pdt0,d1pdt0,d0oIdt0,d1oIdt0,d0opdt0,d1opdt0,
	d0Isigneddt0,d1Isigneddt0,d0Isignedetaudt0,d1Isignedetaudt0,
	d0dptdp,d1dptdp,d2dptdp,d0dpvdp,d1dpvdp,d2dpvdp;

    final double [] dpdpvroot_re,dpdpvroot_im;

    TradePV(final long currenttime,
	    final long tau,

	    // references
	    final double [] sum_v_emoments,
	    final double [] sum_vsigned_emoments,
	    final double [] sum_vsignedetau_emoments,
	    final double [] sum_t_emoments,
	    final double [] sum_tsigned_emoments,
	    final double [] sum_pv_emoments,
	    final double [] sum_pt_emoments,

	    final double [][] pvmatr,
	    final double [][] ptmatr,
	    final double [][] matr,
	    final double [][] tmatr,
	    final double [][] taumatr,
	    final double [][] ptaumatr,
	    final BookItem trade_last)
	{
	    this.currenttime=currenttime;
	    this.tau=tau;
	    this.sum_v_emoments=sum_v_emoments;
	    this.sum_vsigned_emoments=sum_vsigned_emoments;
	    this.sum_vsignedetau_emoments=sum_vsignedetau_emoments;
	    this.sum_t_emoments=sum_t_emoments;
	    this.sum_tsigned_emoments=sum_tsigned_emoments;
	    this.sum_pv_emoments=sum_pv_emoments;
	    this.sum_pt_emoments=sum_pt_emoments;
	    this.pvmatr=pvmatr;
	    this.ptmatr=ptmatr;
	    this.matr=matr;
	    this.tmatr=tmatr;
	    this.taumatr=taumatr;
	    this.ptaumatr=ptaumatr;
	    this.trade_last=trade_last;

	    this.odpx=pvmatr[0][1]/pvmatr[0][0]-pvmatr[1][1]/pvmatr[1][0];
	    this.odpy=(sum_tsigned_emoments[0]*pvmatr[1][1]-sum_tsigned_emoments[1]*pvmatr[0][1])/
		(sum_tsigned_emoments[0]*pvmatr[1][0]-sum_tsigned_emoments[1]*pvmatr[0][0]);
	    this.dpx=ptmatr[0][1]/ptmatr[0][0]-ptmatr[1][1]/ptmatr[1][0];
	    this.dpy=(sum_vsigned_emoments[0]*ptmatr[1][1]-sum_vsigned_emoments[1]*ptmatr[0][1])/
		(sum_vsigned_emoments[0]*ptmatr[1][0]-sum_vsigned_emoments[1]*ptmatr[0][0]);

//	    this.dpy=(sum_t_emoments[0]*pvmatr[1][1]-sum_t_emoments[1]*pvmatr[0][1])/
//		(sum_t_emoments[0]*pvmatr[1][0]-sum_t_emoments[1]*pvmatr[0][0]);


	    final int n=sum_v_emoments.length;
	    final double [][] msol=new double [n][];
	    for(int m=0;m<n;m++)
	    {
		msol[m]=tmatr[m].clone();
	    }
	    final int [] ipvt=new int [n];
	    if(true)
	    {
	    final double h=1+(n-1)/2.0;
	    final int info=com.polytechnik.utils.Linsystems.dgefa(msol,n,ipvt);
	    this.ptcoefs=new double [n];
	    this.vtcoefs=new double [n];
	    this.vtsignedcoefs=new double [n];
	    this.vsignedetaucoefs=new double [n];
	    if(info!=0)
	    {
		java.util.Arrays.fill(this.ptcoefs,Double.NaN);
		java.util.Arrays.fill(this.vtcoefs,Double.NaN);
		java.util.Arrays.fill(this.vtsignedcoefs,Double.NaN);
		java.util.Arrays.fill(this.vsignedetaucoefs,Double.NaN);
	    }
	    else
	    {
		System.arraycopy(sum_pt_emoments,0,this.ptcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.ptcoefs,1);
		System.arraycopy(sum_v_emoments,0,this.vtcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.vtcoefs,1);
		System.arraycopy(sum_vsigned_emoments,0,this.vtsignedcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.vtsignedcoefs,1);
		System.arraycopy(sum_vsignedetau_emoments,0,this.vsignedetaucoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.vsignedetaucoefs,1);
	    }
	    d0pdt=getD0(this.ptcoefs,h);
	    d1pdt=-getD1(this.ptcoefs,h);
	    d2pdt=getD2(this.ptcoefs,h);
	    d0Idt=getD0(this.vtcoefs,h);
	    d1Idt=-getD1(this.vtcoefs,h);
	    d2Idt=getD2(this.vtcoefs,h);

	    d0Idt0=getD0(this.vtcoefs,0);
	    d1Idt0=-getD1(this.vtcoefs,0);
	    d0pdt0=getD0(this.ptcoefs,0);
	    d1pdt0=-getD1(this.ptcoefs,0);

	    d0Isigneddt0=getD0(this.vtsignedcoefs,0);
	    d1Isigneddt0=-getD1(this.vtsignedcoefs,0);
	    d0Isignedetaudt0=getD0(this.vsignedetaucoefs,0);
	    d1Isignedetaudt0=-getD1(this.vsignedetaucoefs,0);
	    }


	    if(true)
	    {
	    for(int m=0;m<n;m++)
	    {
		System.arraycopy(matr[m],0,msol[m],0,n);
	    }
	    final int info=com.polytechnik.utils.Linsystems.dgefa(msol,n,ipvt);
	    this.ovtcoefs=new double [n];
	    this.optcoefs=new double [n];
	    if(info!=0)
	    {
		java.util.Arrays.fill(this.ovtcoefs,Double.NaN);
		java.util.Arrays.fill(this.optcoefs,Double.NaN);
	    }
	    else
	    {
		System.arraycopy(sum_t_emoments,0,this.ovtcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.ovtcoefs,1);
		System.arraycopy(sum_pv_emoments,0,this.optcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.optcoefs,1);
	    }

	    d0oIdt0=getD0(this.ovtcoefs,0);
	    d1oIdt0=-getD1(this.ovtcoefs,0);
	    d0opdt0=getD0(this.optcoefs,0);
	    d1opdt0=-getD1(this.optcoefs,0);
	    }


	    final double p_last=this.trade_last!=null?this.trade_last.price:Double.NaN;
	    this.sum_dpdt=new double [n];
	    for(int m=0;m<n;m++)
	    {
		this.sum_dpdt[m]=(p_last-sum_pt_emoments[m])/(m+1);
	    }

	    if(true)
	    {
	    // ptmatr
	    for(int m=0;m<n;m++)
	    {
		System.arraycopy(ptmatr[m],0,msol[m],0,n);
	    }
	    final int info=com.polytechnik.utils.Linsystems.dgefa(msol,n,ipvt);
	    this.vpcoefs=new double [n];
	    this.dptpcoefs=new double [n];
	    if(info!=0)
	    {
		java.util.Arrays.fill(this.vpcoefs,Double.NaN);
		java.util.Arrays.fill(this.dptpcoefs,Double.NaN);
	    }
	    else
	    {
		System.arraycopy(sum_vsigned_emoments,0,this.vpcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.vpcoefs,1);
		System.arraycopy(this.sum_dpdt,0,this.dptpcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.dptpcoefs,1);
	    }
	    d0Idp=getD0(this.vpcoefs,p_last);
	    d1Idp=getD1(this.vpcoefs,p_last);
	    d2Idp=getD2(this.vpcoefs,p_last);

	    d0dptdp=getD0(this.dptpcoefs,p_last);
	    d1dptdp=getD1(this.dptpcoefs,p_last);
	    d2dptdp=getD2(this.dptpcoefs,p_last);
	    }

	    if(true)
	    {
	    // pvmatr
	    for(int m=0;m<n;m++)
	    {
		System.arraycopy(pvmatr[m],0,msol[m],0,n);
	    }
	    final int info=com.polytechnik.utils.Linsystems.dgefa(msol,n,ipvt);
	    this.ovpcoefs=new double [n];
	    this.dpvpcoefs=new double [n];
	    if(info!=0)
	    {
		java.util.Arrays.fill(this.ovpcoefs,Double.NaN);
		java.util.Arrays.fill(this.dpvpcoefs,Double.NaN);
	    }
	    else
	    {
		//System.arraycopy(sum_t_emoments,0,this.ovpcoefs,0,n);
		System.arraycopy(sum_tsigned_emoments,0,this.ovpcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.ovpcoefs,1);
		System.arraycopy(this.sum_dpdt,0,this.dpvpcoefs,0,n);
		com.polytechnik.utils.Linsystems.dgesl(msol,n,ipvt,this.dpvpcoefs,1);
	    }
	    d0oIdp=getD0(this.ovpcoefs,p_last);
	    d1oIdp=getD1(this.ovpcoefs,p_last);
	    d2oIdp=getD2(this.ovpcoefs,p_last);

	    d0dpvdp=getD0(this.dpvpcoefs,p_last);
	    d1dpvdp=getD1(this.dpvpcoefs,p_last);
	    d2dpvdp=getD2(this.dpvpcoefs,p_last);

	    // find roots
	    final double [] pcoe=new double [n];
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0,w=1;k<n;k++)
	    {
		pcoe[k]=this.dpvpcoefs[k]/w;
		w*=(k+1);
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    this.dpdpvroot_re=u.re_root;
	    this.dpdpvroot_im=u.im_root;
	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(n-1,this.dpdpvroot_re,this.dpdpvroot_im);

	    // find closest point to p_last
	    final int ptr_c=getPtrMinDist(this.dpdpvroot_re,this.dpdpvroot_im,p_last);
	    p_c=ptr_c<0?Double.NaN:this.dpdpvroot_re[ptr_c];
	    d1p_c=getD1(this.dpvpcoefs,p_c);
	    d2p_c=getD2(this.dpvpcoefs,p_c);
	    d0p_c=getD0(this.dpvpcoefs,p_c);


	    }


	}

    private static TradePV calculateTestTradePV(final long currenttime,
						final long tau,
						final boolean flag_skipnoorigtime,
						final /*BookItem */ Trade [] tb,
						final int tsize)
    {
	final int n=N_EMOMENTS;
	final double []  // all 0 by default
	    sum_v_emoments=new double [n],
	    sum_vsigned_emoments=new double [n],
	    sum_vsignedetau_emoments=new double [n],
	    sum_t_emoments=new double [n],
	    sum_tsigned_emoments=new double [n],
	    sum_pv_emoments=new double [n],
	    sum_pt_emoments=new double [n];
	final double [][] pvmatr=new double [n][n],ptmatr=new double [n][n],matr=new double [n][n],tmatr=new double [n][n],
	    taumatr=new double [n][n],ptaumatr=new double [n][n];
	final double atau=1.0/tau;
	long prevt=currenttime,tinterval=currenttime;
	for(int i=tsize;--i>=0;)
	{
	    final /*BookItem*/ Trade b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;
	    final long tauo;
	    if(b.origtime<=0)
	    {
		if(flag_skipnoorigtime)
		{ // skip trades with unknown origtime
		    continue;
		}
		tauo=0;
	    }
	    else
	    {
		tauo=prevt-b.origtime;
	    }

	    final long dt=tinterval-b.time;
	    final double ts=(currenttime-prevt)*atau,dtr=dt*atau,tauor=tauo*atau;
	    final double bwsigned,dtrsigned;
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		bwsigned=-b.shares;  // sell just came
		dtrsigned=-dtr;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		bwsigned=b.shares; // buy just came
		dtrsigned=dtr;
	    }
	    else
	    {
		continue;
	    }


	    for(int m=0;m<n;m++)
	    {
		final double w=Math.exp(-ts/(m+1))/(m+1),bwmw=w*b.shares;
		sum_v_emoments[m]+=bwmw;
		sum_vsigned_emoments[m]+=bwsigned*w;
		sum_vsignedetau_emoments[m]+=bwsigned*w*(1.0-Math.exp(-tauor/(m+1)));
		sum_t_emoments[m]+=w*dtr;
		sum_tsigned_emoments[m]+=w*dtrsigned;
		sum_pv_emoments[m]+=b.price*bwmw;
		sum_pt_emoments[m]+=b.price*w*dtr;
		final double [] sum_pvmatr=pvmatr[m],sum_ptmatr=ptmatr[m],sum_matr=matr[m],sum_tmatr=tmatr[m],
		    sum_taumatr=taumatr[m],sum_ptaumatr=ptaumatr[m];
		double we=1,wpe=1;
		for(int mt=0;mt<n;mt++)
		{
		    sum_pvmatr[mt]+=wpe*bwmw;
		    sum_ptmatr[mt]+=wpe*w*dtr;
		    sum_matr[mt]+=we*bwmw;
		    sum_tmatr[mt]+=we*w*dtr;
		    sum_taumatr[mt]+=tauor*we*bwmw;
		    sum_ptaumatr[mt]+=tauor*wpe*bwmw;
		    we*=(ts/(mt+1));
		    wpe*=(b.price/(mt+1));
		}
	    }
	    if(true ||(i>0 && tb[i-1].time<prevt))
	    {
		tinterval=prevt;
	    }
	}// end i loop
	return new TradePV(currenttime,
			   tau,
			   sum_v_emoments,
			   sum_vsigned_emoments,
			   sum_vsignedetau_emoments,
			   sum_t_emoments,
			   sum_tsigned_emoments,
			   sum_pv_emoments,
			   sum_pt_emoments,
			   pvmatr,
			   ptmatr,
			   matr,
			   tmatr,
			   taumatr,
			   ptaumatr,
			   // last processed point
			   tsize>0?tb[tsize-1]:null);
    }



    private static void checkMatching(final TradePV r,final TradePV ref)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	final int n=r.sum_v_emoments.length;
	final double eps=1e-5;
	for(int m=0;m<n;m++)
	{
	    UnitTests.checkDoubleRel(r.sum_v_emoments[m],"r.sum_v_emoments[m]",ref.sum_v_emoments[m],"ref.sum_v_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsigned_emoments[m],"r.sum_vsigned_emoments[m]",ref.sum_vsigned_emoments[m],"ref.sum_vsigned_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_vsignedetau_emoments[m],"r.sum_vsignedetau_emoments[m]",ref.sum_vsignedetau_emoments[m],"ref.sum_vsignedetau_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_t_emoments[m],"r.sum_t_emoments[m]",ref.sum_t_emoments[m],"ref.sum_t_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_tsigned_emoments[m],"r.sum_tsigned_emoments[m]",ref.sum_tsigned_emoments[m],"ref.sum_tsigned_emoments[m]",eps);

	    UnitTests.checkDoubleRel(r.sum_pv_emoments[m],"r.sum_pv_emoments[m]",ref.sum_pv_emoments[m],"ref.sum_pv_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pt_emoments[m],"r.sum_pt_emoments[m]",ref.sum_pt_emoments[m],"ref.sum_pt_emoments[m]",eps);
	    UnitTests.checkDoubleRel(r.sum_pt_emoments[m],"r.sum_pt_emoments[m]",r.ptmatr[m][1],"r.ptmatr[m][1]",eps);
	    UnitTests.checkDoubleRel(r.sum_v_emoments[m],"r.sum_v_emoments[m]",r.pvmatr[m][0],"r.pvmatr[m][0]",eps);
	    UnitTests.checkDoubleRel(r.sum_t_emoments[m],"r.sum_t_emoments[m]",r.tmatr[m][0],"r.tmatr[m][0]",eps);
	    UnitTests.checkDoubleRel(r.sum_pv_emoments[m],"r.sum_pv_emoments[m]",r.pvmatr[m][1],"r.pvmatr[m][1]",eps);
	    UnitTests.checkDoubleRel(r.pvmatr[m][0],"r.pvmatr[m][0]",r.matr[m][0],"r.matr[m][0]",eps);
	    UnitTests.checkDoubleRel(r.ptmatr[m][0],"r.ptmatr[m][0]",r.tmatr[m][0],"r.tmatr[m][0]",eps);
	    for(int mt=0;mt<n;mt++)
	    {
		UnitTests.checkDoubleRel(r.pvmatr[m][mt],"r.pvmatr[m][mt]",ref.pvmatr[m][mt],"ref.pvmatr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.ptmatr[m][mt],"r.ptmatr[m][mt]",ref.ptmatr[m][mt],"ref.ptmatr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.matr[m][mt],"r.matr[m][mt]",ref.matr[m][mt],"ref.matr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.tmatr[m][mt],"r.tmatr[m][mt]",ref.tmatr[m][mt],"ref.tmatr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.taumatr[m][mt],"r.taumatr[m][mt]",ref.taumatr[m][mt],"ref.taumatr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.ptaumatr[m][mt],"r.ptaumatr[m][mt]",ref.ptaumatr[m][mt],"ref.ptaumatr[m][mt]",eps);
	    }
	}
	System.err.println("TradePV:checkMatching(): matching OK");
    }





    private static void checkMatchingTradeDPV(final TradeDPV r,final TradePV ref)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	final int n=r.volmatr.length;
	final double eps=1e-5;
	for(int m=0;m<n;m++)
	{
	    UnitTests.checkDoubleRel(r.volmatr[m][0],"r.volmatr[m][0]",ref.pvmatr[m][0],"ref.pvmatr[m][0]",eps);
	    UnitTests.checkDoubleRel(r.pvolmatr[m][0],"r.pvolmatr[m][0]",ref.pvmatr[m][1],"ref.pvmatr[m][1]",eps);
	    for(int mt=0;mt<n;mt++)
	    {
		UnitTests.checkDoubleRel(r.pvmatr[m][mt],"r.pvmatr[m][mt]",ref.pvmatr[m][mt],"ref.pvmatr[m][mt]",eps);
	    }
	}
	System.err.println("TradePV:checkMatchingTradeDPV(): matching OK");
    }



    private static Trade [] getHarmonicPTrades(final int n,
					       final double [] out,
					       final long tau,
					       final java.util.Random r)
    {

	final Trade [] tb=new Trade [n];
	final long dt=tau/1000;
	final double omega=10*Math.PI/dt/tb.length,amp=1.2;
	System.err.println("TradePV:getHarmonicPTrades dt*omega="+(dt*omega)+" dt/tau="+(((double)dt)/tau));
	out[0]=omega;
	out[1]=amp;
	final int v0=100;
	out[2]=v0;

	long t=10;
	for(int i=0;i<tb.length;i++)
	{
	    t+=dt+(long)(dt*0.6*(1-2*r.nextDouble()));
	    final double p=amp*Math.cos(omega*t);

	    tb[i]=new Trade(p,
			    v0,
			    null,
			    r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
			    t,
			    0,
			    0,
			    t-r.nextInt(100000),
			    null,
			    Trade.TRADE_TYPE_REGULAR,
			    null,
			    null,
			    i);
	}
	return tb;
    }



    private static void checkHarmonicP(final int tsize,final java.util.Random r)
    {
	final boolean flag_skipnoorigtime=false;
	final double eps=1e-3;
	final long tau=1000L*tsize/20;
	final double [] out=new double [3];
	final Trade [] tb=getHarmonicPTrades(tsize,out,tau,r);
	final double omega=out[0],amp=out[1],v0=out[2],atau=1.0/tau;
	final long currenttime=tb[tb.length-1].time,t0=tb[0].time,t_now=currenttime;
	System.err.println("TradePV:: check at currenttime="+currenttime+" tau="+tau);
	final TradePV ro=calculateTradePV(currenttime,tau,flag_skipnoorigtime,tb,tsize,null);


	final int n=ro.sum_v_emoments.length;
	final double [][] ptdef=IntgrTPFun.getMatrDefault(n,t0,t_now,tau,omega,amp,t_now);
	final double [] dpdtae=IntgrTPFun.getDpDTExact(n,t0,t_now,tau,omega,amp,t_now);
	System.err.println("omega*t_now="+(omega*t_now)+" (t_now-t0)/tau="+((t_now-t0)*atau));
	for(int m=0;m<n;m++)
	{
	    UnitTests.checkDoubleRel(ro.sum_dpdt[m],"ro.sum_dpd["+m+"]",dpdtae[m],"dpdtae[m]",eps);
	}

	for(int m=0;m<n;m++)
	{
	    for(int mt=0;mt<n;mt++)
	    {
		UnitTests.checkDoubleRel(ptdef[m][mt]/tau,"ptdef["+m+"]["+mt+"]/tau",ro.ptmatr[m][mt],"ro.ptmatr[m][mt]",eps);
		// because dt can vary and v0=const, the condition on pvmatr may not match
		//UnitTests.checkDoubleRel(ptdef[m][mt]*v0*tsize/(t_now-t0),"ptdef["+m+"]["+mt+"]*v0*tsize/(t_now-t0)",ro.pvmatr[m][mt],"ro.pvmatr[m][mt]",eps);
	    }
	}


	System.err.println("TradePV::checkHarmonic(): test OK coefficients matched, N points="+tsize);
    }



    private static Trade [] getExactPolynomialPTrades(final int tsize,
						      final double [] coefs,
						      final long tau,
						      final java.util.Random r,
						      final int test_num)
    {

	java.util.Arrays.fill(coefs,0);
	final Trade [] tb=new Trade [tsize];
	final long dt=tau/1000;
	final double omega=10*Math.PI/dt/tb.length;
	final long tmax=tb.length*2L*dt+1000*1000L,currenttime=tmax;
	System.err.println("TredePV:getExactPolynomialPTrades dt/tau="+(((double)dt)/tau));
	final double a;
	if(test_num==0)
	{
	    //a=-r.nextDouble();
	    a=1.0;//omega;
	    coefs[0]=a;
	}
	else if(test_num==1)
	{
	    a=5;//r.nextDouble();
	    coefs[1]=a;
	}
	else if(test_num==2)
	{
	    a=r.nextDouble();
	    coefs[2]=a;
	}
	else if(test_num==3)
	{
	    a=r.nextDouble();
	    coefs[3]=a;
	}
	else if(test_num==4)
	{
	    a=r.nextDouble();
	    coefs[4]=a;
	}
	else if(test_num==5)
	{
	    a=r.nextDouble();
	    coefs[0]=a;
	    coefs[2]=-2*a;
	}
	else
	{
	    throw new RuntimeException("test_num="+test_num);
	}


	final int v0=100;

	long t=tmax;
	for(int i=tb.length;i-->0;)
	{
	    final double p;
	    if(test_num==0)
	    {
		p=1+a*(t-currenttime)/tau;
	    }
	    else if(test_num==1)
	    {
		p=Math.exp(a*(t-currenttime)/tau);
	    }
	    else if(test_num==2)
	    {
		p=-1/(a/Polynomials.fact(2)*(t-currenttime));
	    }
	    else if(test_num==3)
	    {
		p=-Math.sqrt(-0.5/(a/Polynomials.fact(3)*(t-currenttime)));
	    }
	    else if(test_num==4)
	    {
		p=Math.pow(-1/(3*a/Polynomials.fact(4)*(t-currenttime)),1.0/3);
	    }
	    else if(test_num==5)
	    {
		p=Math.tanh(a*(t-currenttime));
	    }
	    else
	    {
		throw new RuntimeException("test_num="+test_num);
	    }

	    tb[i]=new Trade(p,
			    v0,
			    null,
			    r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
			    t,
			    0,
			    0,
			    t-r.nextInt(100000),
			    null,
			    Trade.TRADE_TYPE_REGULAR,
			    null,
			    null,
			    i);

	    t-=dt+(long)(dt*0.6*(1-2*r.nextDouble()));
	}
	return tb;
    }

    private static void checkExactPolynomialP(final int tsize,final java.util.Random r)
    {
	final boolean flag_skipnoorigtime=false;
	final double eps=1e-2;
	final long tau=2000L*tsize/20;
	final int n=N_EMOMENTS;
	for(int test_num=0;test_num<=5;test_num++)
	{
	    final double [] coefs=new double [n];
	    final Trade [] tb=getExactPolynomialPTrades(tsize,coefs,tau,r,test_num);
	    final long currenttime=tb[tb.length-1].time;
	    final TradePV ro=calculateTradePV(currenttime,tau,flag_skipnoorigtime,tb,tsize,null);

	    for(int m=0;m<n;m++)
	    {
		System.err.print(" p_last="+ro.trade_last.price+
				 " ro.sum_dpdt["+m+"]="+ro.sum_dpdt[m]+
				 " ro.sum_pt_emoments["+m+"]="+ro.sum_pt_emoments[m]+
				 " coef="+coefs[m]+
				 " ro.dptpcoefs="+ro.dptpcoefs[m]);
		for(int mt=0;mt<n;mt++)
		{
		    System.err.print("\tro.ptmatr["+m+"]["+mt+"]="+ro.ptmatr[m][mt]);
		}

		System.err.println();
	    }

	    for(int m=0;m<n;m++)
	    {
		// match coefficients
		System.err.println("TradePV:checkExactPolynomialP ro.dptpcoefs["+m+"]="+ro.dptpcoefs[m]+" coefs[m]="+coefs[m]+" test_num="+test_num+
				   " ro.sum_dpdt[m]="+ro.sum_dpdt[m]+" ro.ptmatr[m][0]="+ro.ptmatr[m][0]);
		UnitTests.checkDoubleRel(ro.dptpcoefs[m],"ro.dptpcoefs["+m+"]",coefs[m],"coefs[m]",eps);
	    }
	    System.err.println("TradePV::checkExactPolynomial(): test OK coefficients matched, N points="+tsize+" test_num="+test_num);
	}
    }




    private static void matchWithTest(final TradePV r,
				      final Trade [] tb,
				      final int tsize)
    {
	final TradeEMoments ref=calculateEMoments(r.currenttime,
						  r.tau,
						  tb,
						  tsize,
						  null);
	TradeEMoments.checkMatchingTradePV(ref,r);
	final TradePV rt=calculateTestTradePV(r.currenttime,
					      r.tau,
					      false,
					      tb,
					      tsize);
	checkMatching(r,rt);
    }



    private static void checkFMatr(final int tsize,
				   final java.util.Random r,
				   final long tau)
    {
	final boolean flag_skipnoorigtime=false;
	final double eps=1e-3;

	// equal distance sample
	final Trade [] tb=TradeEMoments.getRandomTradesFixedTau(tsize,r,tau);
	final long currenttime=tb[tsize-1].time;
	final TradePV ro=calculateTradePV(currenttime,
					  tau,
					  flag_skipnoorigtime,
					  tb,
					  tsize,
					  null);
	final int n=ro.sum_v_emoments.length;
	for(int m=0;m<n;m++)
	{
	    for(int k=0;k<n;k++)
	    {
		System.err.println("MT["+m+"]["+k+"]="+ro.tmatr[m][k]+
				   " should be="+Math.pow(m+1,k));

	    }
	}


	matchWithTest(ro,tb,tsize);
	if(n==5)
	{
	    // print inverted matrix
	    System.err.println("====inverted matrix");
	    final double [][] minv=com.polytechnik.utils.Linsystems.getInvertedMatrix(ro.tmatr);
	    for(int i=0;i<n;i++)
	    {
		System.err.print("r"+i+":\t");
		for(int j=0;j<n;j++)
		{
		    System.err.print(minv[i][j]+"\t");
		}
		System.err.println();
	    }
	    for(int i=0;i<n;i++)
	    {
		for(int j=0;j<n;j++)
		{
		    double s=0;
		    for(int k=0;k<n;k++)
		    {
			s+=ro.tmatr[i][k]*minv[k][j];
		    }
		    UnitTests.checkDoubleRel(s,"mxm-1",i==j?1:0,i==j?"1":"0",eps);
		}
	    }
	    System.err.println("==== end of inverted matrix");

	    final double h=1;
	    for(final double [][] dx:new double [][][]{
		    {ro.ptcoefs,ro.sum_pt_emoments}
		    /*,{ro.vtcoefs,ro.sum_v_emoments}*/
		})
	    {
		final double [] dr=dx[1],c=dx[0];
		// check coefficients
		final double f0a=c[0],d1fa=-c[1],d2fa=c[2],d3fa=-c[3];
		// standard interpolation formulas. the 1/m! factor in front
		// take into account the right part definition we use.
		final double
		    f0at=(1.0/Polynomials.fact(0))*(1*dr[4]-5*dr[3]+10*dr[2]-10*dr[1]+5*dr[0]),
		    d1fat=(1.0/Polynomials.fact(1))*(25*dr[4]-122*dr[3]+234*dr[2]-214*dr[1]+77*dr[0])/(12*h),
		    d2fat=(1.0/Polynomials.fact(2))*(35*dr[4]-164*dr[3]+294*dr[2]-236*dr[1]+71*dr[0])/(12*h*h),
		    d3fat=(1.0/Polynomials.fact(3))*(5*dr[4]-22*dr[3]+36*dr[2]-26*dr[1]+7*dr[0])/(2*h*h*h);

		/* // FIXME
		final double epsa=1e-2;//eps;
		UnitTests.checkDoubleRel(f0a,"f0a",f0at,"f0at",epsa);
		UnitTests.checkDoubleRel(d1fa,"d1fa",d1fat,"d1fat",epsa);
		UnitTests.checkDoubleRel(d2fa,"d2fa",d2fat,"d2fat",epsa);
		UnitTests.checkDoubleRel(d3fa,"d3fa",d3fat,"d3fat",epsa);
		*/

		final double f0ad=getD0(c,0),d1fad=-getD1(c,0),d2fad=getD2(c,0),d3fad=-getD3(c,0);
		UnitTests.checkDoubleRel(f0a,"f0a",f0ad,"f0ad",eps);
		UnitTests.checkDoubleRel(d1fa,"d1fa",d1fad,"d1fad",eps);
		UnitTests.checkDoubleRel(d2fa,"d2fa",d2fad,"d2fad",eps);
		UnitTests.checkDoubleRel(d3fa,"d3fa",d3fad,"d3fad",eps);

		/*
		final double f0t=(1*dr[4]-5*dr[3]+10*dr[2]-10*dr[1]+5*dr[0]),
		    df1t=(3*dr[4]-16*dr[3]+36*dr[2]-48*dr[1]+25*dr[0])/(12*h),

		UnitTests.checkDoubleRel(df1,"df1",df1t,"df1t",eps);
		UnitTests.checkDoubleRel(f0,"f0",f0t,"f0t",eps);
		*/
	    }
	}
	System.err.println("TradePV::checkFMatr(): test random values OK N points="+tsize);
    }



    private static void checkCalculations(final int tsize,final java.util.Random r)
    {
	final boolean flag_skipnoorigtime=false;

	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println("checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);
	final TradePV ro=calculateTradePV(currenttime,
					  tau,
					  flag_skipnoorigtime,
					  tb,
					  tsize,
					  null);
	matchWithTest(ro,tb,tsize);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradePV em_ref=calculateTradePV((tb[n_split-1].time+tb[n_split].time)/2,tau,flag_skipnoorigtime,tb,n_split,null);
	matchWithTest(em_ref,tb,n_split);

	final TradePV em_fromref=calculateTradePV(currenttime,tau,flag_skipnoorigtime,tb,tsize,em_ref);
	checkMatching(ro,em_fromref);
	matchWithTest(em_fromref,tb,tsize);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradePV emc_cut=calculateTradePV(currenttime,tau,flag_skipnoorigtime,tb,ncut,em_ref);
	checkMatching(ro,emc_cut);

	System.err.println("TradePV::checkCalculations(): test random values OK N points="+tsize);
    }


    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int i=0;i<10;i++)
	{
	    //checkHarmonicP(/* 4*/100010+r.nextInt(10000),r);
	    //checkExactPolynomialP(100010+r.nextInt(10000),r);
	    checkCalculations(100+r.nextInt(100),r);
	    checkCalculations(10010+r.nextInt(1000010),r);
	    checkFMatr(1000010+r.nextInt(1000010),r,10000);
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradePV::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }
}



static TradePV calculateTradePV(final long currenttime,
				final long tau,
				final boolean flag_skipnoorigtime,
				final /* BookItem */ Trade [] tb,
				final int tsize,
				final TradePV ref)
{

    final int n=N_EMOMENTS;
    final double []  // all 0 by default
	sum_v_emoments=new double [n],
	sum_vsigned_emoments=new double [n],
	sum_vsignedetau_emoments=new double [n],
	sum_t_emoments=new double [n],
	sum_tsigned_emoments=new double [n],
	sum_pv_emoments=new double [n],
	sum_pt_emoments=new double [n];
    final double [][] pvmatr=new double [n][n],ptmatr=new double [n][n],matr=new double [n][n],tmatr=new double [n][n],
	taumatr=new double [n][n],ptaumatr=new double [n][n];

    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    long prevt=currenttime,tinterval=currenttime,sumdt=0;
    gloop:for(int i=tsize;--i>=0;)
    {
	final /*BookItem*/ Trade b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break gloop;
	}
	prevt=b.time;
	final long tauo;
	if(b.origtime<=0)
	{
	    if(flag_skipnoorigtime)
	    { // skip trades with unknown origtime
		continue gloop;
	    }
	    tauo=0;
	}
	else
	{
	    tauo=prevt-b.origtime;
	}

	final double p=b.price,ts=(currenttime-prevt)*atau,dtr=(tinterval-prevt)*atau,tauor=tauo*atau;
	final double bw=b.shares,bwsigned,dtrsigned;
	if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	{
	    bwsigned=-bw;  // sell just came
	    dtrsigned=-dtr;
	}
	else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	{
	    bwsigned=bw; // buy just came
	    dtrsigned=dtr;
	}
	else
	{
	    continue gloop;
	}
	sumdt+=tinterval-prevt;

	for(int m=0;m<n;m++)
	{
	    final double w=Math.exp(-ts/(m+1))/(m+1),bwmw=bw*w,dtrmw=dtr*w;
	    //System.err.println("m="+m+" w="+w+" bwmw="+bwmw+" e="+(ts/(m+1))+" currenttime="+currenttime+" tau="+tau+" t="+prevt+" dt="+(tinterval-prevt));
	    sum_v_emoments[m]+=bwmw;
	    sum_vsigned_emoments[m]+=bwsigned*w;
	    sum_vsignedetau_emoments[m]+=bwsigned*w*(1.0-Math.exp(-tauor/(1+m)));
	    sum_t_emoments[m]+=dtrmw;
	    sum_tsigned_emoments[m]+=dtrsigned*w;
	    sum_pv_emoments[m]+=p*bwmw;
	    sum_pt_emoments[m]+=p*dtrmw;
	    final double [] sum_pvmatr=pvmatr[m],sum_ptmatr=ptmatr[m],sum_matr=matr[m],sum_tmatr=tmatr[m],
		sum_taumatr=taumatr[m],sum_ptaumatr=ptaumatr[m];
	    double we=1,wpe=1;
	    for(int mt=0;mt<n;mt++)
	    {
		sum_pvmatr[mt]+=bwmw*wpe;
		sum_ptmatr[mt]+=dtrmw*wpe;
		sum_matr[mt]+=bwmw*we;
		sum_tmatr[mt]+=dtrmw*we;
		sum_taumatr[mt]+=bwmw*we*tauor;
		sum_ptaumatr[mt]+=bwmw*wpe*tauor;
		we*=(ts/(mt+1));
		wpe*=p/(mt+1);
	    }
	}
	if(true || (i<=0 || tb[i-1].time<prevt))
	{
	    tinterval=prevt;
	}
    }// end i loop

    if(bref!=null)
    { // optimization
	if(true)
	{
	    // add missed half slice
	    sumdt+=tinterval-ref.currenttime;
	    final double dtr=(tinterval-ref.currenttime)*atau,ts=(currenttime-bref.time)*atau,
		dtrsigned=bref.buy_sell_indicator==BookItem.BOOK_BUY?-dtr:(bref.buy_sell_indicator==BookItem.BOOK_SELL?dtr:0);
	    for(int m=0;m<n;m++)
	    {
		final double w=Math.exp(-ts/(m+1))/(m+1),dtrmw=dtr*w;
		sum_t_emoments[m]+=dtrmw;
		sum_tsigned_emoments[m]+=dtrsigned*w;
		sum_pt_emoments[m]+=dtrmw*bref.price;
		final double [] sum_ptmatr=ptmatr[m],sum_tmatr=tmatr[m];
		double we=1,wpe=1;
		for(int mt=0;mt<n;mt++)
		{
		    sum_ptmatr[mt]+=dtrmw*wpe;
		    sum_tmatr[mt]+=dtrmw*we;
		    we*=(ts/(mt+1));
		    wpe*=bref.price/(mt+1);
		}
	    }
	}
	if(sumdt!=(currenttime-ref.currenttime))
	{
	    throw new RuntimeException("sumdt="+sumdt+" currenttime-ref.currenttime="+(currenttime-ref.currenttime));
	}

	// add previous interval
	final double ref_adelay=(currenttime-ref.currenttime)*atau;
	for(int m=0;m<n;m++)
	{
	    final double exp_delay=Math.exp(-ref_adelay/(m+1));
	    sum_v_emoments[m]+=exp_delay*ref.sum_v_emoments[m];
	    sum_vsigned_emoments[m]+=exp_delay*ref.sum_vsigned_emoments[m];
	    sum_vsignedetau_emoments[m]+=exp_delay*ref.sum_vsignedetau_emoments[m];
	    sum_t_emoments[m]+=exp_delay*ref.sum_t_emoments[m];
	    sum_tsigned_emoments[m]+=exp_delay*ref.sum_tsigned_emoments[m];
	    sum_pv_emoments[m]+=exp_delay*ref.sum_pv_emoments[m];
	    sum_pt_emoments[m]+=exp_delay*ref.sum_pt_emoments[m];
	    final double [] sum_pvmatr=pvmatr[m],ref_sum_pvmatr=ref.pvmatr[m],
		sum_ptmatr=ptmatr[m],ref_sum_ptmatr=ref.ptmatr[m],
		sum_matr=matr[m],ref_sum_matr=ref.matr[m],
		sum_tmatr=tmatr[m],ref_sum_tmatr=ref.tmatr[m],
		sum_taumatr=taumatr[m],ref_sum_taumatr=ref.taumatr[m],
		sum_ptaumatr=ptaumatr[m],ref_sum_ptaumatr=ref.ptaumatr[m];
	    // use Newton binomial expansion
	    for(int mt=0;mt<n;mt++)
	    {
		double we=1,s_v=0,s_vt=0,s_vtauo=0;
		for(int ms=0;ms<=mt;ms++)
		{
		    s_v+=we*ref_sum_matr[mt-ms];
		    s_vt+=we*ref_sum_tmatr[mt-ms];
		    s_vtauo+=we*ref_sum_taumatr[mt-ms];
		    we*=(ref_adelay/(ms+1));
		}
		sum_pvmatr[mt]+=exp_delay*ref_sum_pvmatr[mt];
		sum_ptmatr[mt]+=exp_delay*ref_sum_ptmatr[mt];
		sum_ptaumatr[mt]+=exp_delay*ref_sum_ptaumatr[mt];
		sum_matr[mt]+=exp_delay*s_v;
		sum_tmatr[mt]+=exp_delay*s_vt;
		sum_taumatr[mt]+=exp_delay*s_vtauo;
	    }
	}
    }

    final TradePV r=
	new TradePV(currenttime,
		    tau,
		    sum_v_emoments,
		    sum_vsigned_emoments,
		    sum_vsignedetau_emoments,
		    sum_t_emoments,
		    sum_tsigned_emoments,
		    sum_pv_emoments,
		    sum_pt_emoments,
		    pvmatr,
		    ptmatr,
		    matr,
		    tmatr,
		    taumatr,
		    ptaumatr,
		    // last processed point
		    tsize>0?tb[tsize-1]:bref);

    return r;
}












static class TradeDPV {
    final long currenttime,tau;


    private final double [][]
	volmatr, // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i vol^k/k!
	pvmatr, // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i p^k/k!
        pvolmatr // [m][k]=sum exp(-(t0-t_i)/((m+1)*tau)) v_i p vol^k/k!
	;
    private final double [] p1coefs,p2coefs,sum_shortaction,actioncoefs;

    // references to previous calculations
    private final BookItem trade_last;
    private final long vol;


    final double [] root_re,root_im,real_roots;
    final double p_low,p_high;

    TradeDPV(final long currenttime,
	     final long tau,

	     final double [][] volmatr,
	     final double [][] pvmatr,
	     final double [][] pvolmatr,

	     // references
	     final BookItem trade_last,
	     final long vol)
	{
	    this.currenttime=currenttime;
	    this.tau=tau;
	    this.volmatr=volmatr;
	    this.pvmatr=pvmatr;
	    this.pvolmatr=pvolmatr;
	    this.trade_last=trade_last;
	    this.vol=vol;


	    final int nv=volmatr.length,npv=nv;
	    final double [][] msol=new double [nv][];
	    for(int m=0;m<nv;m++)
	    {
		msol[m]=this.volmatr[m].clone();
	    }
	    final int [] ipvt=new int [nv];
	    final int vinfo=com.polytechnik.utils.Linsystems.dgefa(msol,nv,ipvt);

	    /* System.err.println("TradeDPV volmatr: vinfo="+vinfo);
	    for(int m=0;m<nv;m++)
	    {
		for(int j=0,np=volmatr[m].length;j<np;j++)
		{
		    System.err.print(volmatr[m][j]+"\t");
		}
		System.err.println("r\t"+pvmatr[m][1]+"\t"+pvmatr[m][2]);
	    } */

	    final double [][] ppowv=new double [npv][nv];
	    if(vinfo!=0)
	    {
		for(int m=0;m<npv;m++)
		{
		    java.util.Arrays.fill(ppowv[m],Double.NaN);
		}
	    }
	    else
	    {
		for(int m=0;m<npv;m++)
		{
		    final double [] ppow=ppowv[m];
		    for(int k=0;k<nv;k++)
		    {
			ppow[k]=pvmatr[k][m]; // integral of p^m/m! at time tau_k
		    }
		    com.polytechnik.utils.Linsystems.dgesl(msol,nv,ipvt,ppow,1);
		}
	    }
	    this.p1coefs=ppowv[1];
	    this.p2coefs=ppowv[2];

	    // short action
	    this.sum_shortaction=getShortAction(this.p1coefs,this.p2coefs);


	    this.actioncoefs=new double [npv];
	    if(true)
	    { // exact calculations
		for(int m=0;m<npv;m++)
		{
		    System.arraycopy(pvmatr[m],0,msol[m],0,npv);
		}
	    }
	    else
	    { // same basis calculations
		for(int m=0;m<npv;m++)
		{
		    for(int k=0;k<npv;k++)
		    {
			double s=0;
			for(int g=0;g<npv;g++)
			{
			    s+=ppowv[k][g]*this.volmatr[m][g];
			}
			// should match exactly to pvmatr[m][k]
			msol[m][k]=s;
			//System.err.println("msol["+m+"]["+k+"]="+msol[m][k]+" pvmatr[m][k]="+pvmatr[m][k]);
		    }
		}
	    }
	    final int pinfo=com.polytechnik.utils.Linsystems.dgefa(msol,npv,ipvt);
	    if(pinfo!=0)
	    {
		java.util.Arrays.fill(this.actioncoefs,Double.NaN);
	    }
	    else
	    {
		System.arraycopy(this.sum_shortaction,0,this.actioncoefs,0,npv);
		com.polytechnik.utils.Linsystems.dgesl(msol,npv,ipvt,this.actioncoefs,1);
	    }



	    // find roots
	    final double [] pcoe=new double [npv];
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0,w=1;k<npv;k++)
	    {
		pcoe[k]=this.actioncoefs[k]/w;
		w*=(k+1);
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    this.root_re=u.re_root;
	    this.root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(npv-1,this.root_re,this.root_im);

	    final double p_last=this.trade_last!=null?this.trade_last.price:Double.NaN;
	    // find real roots & the point closest to p_last
	    int ptr_close_to_p_last=-1;
	    if(true)
	    {
		int n_reroots=0;
		double drr=Double.POSITIVE_INFINITY;
		final double [] reroots=pcoe; // use pcoe as a temporary array
		for(int m=0;m<this.root_re.length;m++)
		{
		    if(Math.abs(this.root_im[m])<=EPS_IM)
		    {
			final double rr=this.root_re[m];
			final double drr1=Math.abs(p_last-rr);
			if(drr1<drr)
			{
			    drr=drr1;
			    ptr_close_to_p_last=n_reroots;
			}
			reroots[n_reroots++]=rr;
		    }
		}
		this.real_roots=java.util.Arrays.copyOf(reroots,n_reroots);
	    }

	    if(ptr_close_to_p_last<0)
	    {
		this.p_low=this.p_high=Double.NaN;
	    }
	    else
	    {
		final int nrrm=this.real_roots.length-1;
		final double p_close_to_p_last=this.real_roots[ptr_close_to_p_last],
		    daction_p_close_to_p_last=getD1(this.actioncoefs,p_close_to_p_last);
		if(p_close_to_p_last>=p_last)
		{
		    /*
		    if(-Math.signum(daction_p_close_to_p_last)!=Math.signum(getD0(this.actioncoefs,p_last)))
		    {
			throw new RuntimeException();
		    }
		    */
		    if(daction_p_close_to_p_last<=0)
		    {
			this.p_low=ptr_close_to_p_last>0?this.real_roots[ptr_close_to_p_last-1]:Double.NEGATIVE_INFINITY;
			this.p_high=p_close_to_p_last;
		    }
		    else
		    {
			this.p_low=p_close_to_p_last;
			this.p_high=ptr_close_to_p_last<nrrm?this.real_roots[ptr_close_to_p_last+1]:Double.POSITIVE_INFINITY;
		    }
		}
		else
		{
		    /*
		    if(Math.signum(daction_p_close_to_p_last)!=Math.signum(getD0(this.actioncoefs,p_last)))
		    {
			throw new RuntimeException();
		    }
		    */
		    if(daction_p_close_to_p_last>=0)
		    {
			this.p_low=p_close_to_p_last;
			this.p_high=ptr_close_to_p_last<nrrm?this.real_roots[ptr_close_to_p_last+1]:Double.POSITIVE_INFINITY;
		    }
		    else
		    {
			this.p_low=ptr_close_to_p_last>0?this.real_roots[ptr_close_to_p_last-1]:Double.NEGATIVE_INFINITY;
			this.p_high=p_close_to_p_last;
		    }
		}
	    }
	}

    static double [] getDcoe(final double [] p1coefs)
    {
	final int nv=p1coefs.length;
	// try direct (dp/dv)^2
	final double [] dcoe=new double [1+2*(nv-2)];
	for(int l=0,ndcoe=dcoe.length;l<ndcoe;l++)
	{
	    long f=1;
	    double s=0;
	    for(int k=0;k<=l;k++)
	    {
		//if(f!=Polynomials.fact(l)/(Polynomials.fact(k)*Polynomials.fact(l-k))){throw new RuntimeException();}
		if(k+1<nv && l-k+1<nv)
		{
		    s+=(p1coefs[k+1]*p1coefs[l-k+1])*f;
		}
		f*=(l-k);
		f/=(1+k);
	    }
	    dcoe[l]=s;
	}
	return dcoe;
    }


    double [] getShortAction(final double [] p1coefs,final double [] p2coefs)
    {
	final int nv=p1coefs.length;
	final double [] sum_shortaction=new double [nv];
	if(true)
	{
	    final double [] dcoe=getDcoe(p1coefs);
	    // direct calculation
	    for(int m=0;m<nv;m++)
	    {
		double s=0;
		for(int mt=0,npd=Math.min(dcoe.length,this.volmatr[m].length);mt<npd;mt++)
		{
		    s+=dcoe[mt]*this.volmatr[m][mt];
		}
		sum_shortaction[m]=s;
	    }
	}
	else
	{
	    // (p^2/2)''-pp''
	    for(int m=0;m<nv;m++)
	    {
		double s=0;
		for(int mt=2;mt<nv;mt++)
		{
		    final int mx=mt-2;
		    s+=p2coefs[mt]*this.volmatr[m][mx]-p1coefs[mt]*this.pvolmatr[m][mx];
		}
		sum_shortaction[m]=s;
	    }
	}
	if(false)
	{ // debug - try exact action in special case,
	    // the roots correspond to TestPlowPhigh test.
	    for(int m=0;m<nv;m++)
	    {
		// int of sin^2 for p=cos
		sum_shortaction[m]=pvmatr[m][0]*TestPlowPhigh.pamp*TestPlowPhigh.pamp-2*pvmatr[m][2];
	    }
	}
	return sum_shortaction;
    }


    private static TradeDPV calculateTestTradeDPV(final long currenttime,
						  final long tau,
						  final BookItem [] tb,
						  final int tsize)
    {
	final int n=N_EMOMENTS,np=2*n;
	final double [][]  // all 0 by default
	    volmatr=new double [n][np],
	    pvmatr=new double [n][np],
	    pvolmatr=new double [n][np];
	final double atau=1.0/tau;
	long prevt=currenttime,vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;

	    vol+=b.shares;
	    final double ts=(currenttime-prevt)*atau,price_i=b.price,vol_d=vol;

	    for(int m=0;m<n;m++)
	    {
		final double w=Math.exp(-ts/(m+1))/(m+1),bwmw=w*b.shares,pbwmw=bwmw*price_i;
		final double [] sum_volmatr=volmatr[m],sum_pvmatr=pvmatr[m],sum_pvolmatr=pvolmatr[m];
		double w_vol=1,wpe=1;
		for(int mt=0;mt<np;mt++)
		{
		    sum_volmatr[mt]+=w_vol*bwmw;
		    sum_pvmatr[mt]+=wpe*bwmw;
		    sum_pvolmatr[mt]+=w_vol*pbwmw;
		    w_vol*=(vol_d/(mt+1));
		    wpe*=(price_i/(mt+1));
		}
	    }
	}// end i loop
	return new TradeDPV(currenttime,
			    tau,
			    volmatr,
			    pvmatr,
			    pvolmatr,
			    // last processed point
			    tsize>0?tb[tsize-1]:null,
			    vol);
    }


    private static void checkMatching(final TradeDPV r,final TradeDPV ref)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}
	final int n=r.volmatr.length;
	final double eps=1e-5;
	for(int m=0;m<n;m++)
	{
	    UnitTests.checkDoubleRel(r.volmatr[m][0],"r.volmatr[m][0]",r.pvmatr[m][0],"r.pvmatr[m][0]",eps);
	    UnitTests.checkDoubleRel(r.pvolmatr[m][0],"r.pvolmatr[m][0]",r.pvmatr[m][1],"r.pvmatr[m][1]",eps);
	    for(int mt=0,np=r.volmatr[m].length;mt<np;mt++)
	    {
		UnitTests.checkDoubleRel(r.volmatr[m][mt],"r.volmatr[m][mt]",ref.volmatr[m][mt],"ref.volmatr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.pvmatr[m][mt],"r.pvmatr[m][mt]",ref.pvmatr[m][mt],"ref.pvmatr[m][mt]",eps);
		UnitTests.checkDoubleRel(r.pvolmatr[m][mt],"r.pvolmatr[m][mt]",ref.pvolmatr[m][mt],"ref.pvolmatr[m][mt]",eps);
	    }
	}
	System.err.println("TradeDPV:checkMatching(): matching OK");
    }




    static void checkMatchingTradeDPVR(final TradeDPVR r,final TradeDPV ref)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}
	final double eps=1e-5;
	if(r.p_x_volmatr[0].length!=ref.volmatr[0].length)
	{
	    throw new RuntimeException("r.p_x_volmatr[0].length"+r.p_x_volmatr[0].length+
				       " ref.volmatr[0].length="+ref.volmatr[0].length);
	}
	final int nvs=r.p_x_volmatr[0].length;
	for(int k=0;k<nvs;k++)
	{
	    UnitTests.checkDoubleRel(r.p_x_volmatr[0][k],"r.p_x_volmatr[0][k]",ref.volmatr[0][k],"ref.volmatr[0][k]",eps);
	    UnitTests.checkDoubleRel(r.p_x_volmatr[1][k],"r.p_x_volmatr[1][k]",ref.pvolmatr[0][k],"ref.pvolmatr[0][k]",eps);
	}
	for(int m=0;m<r.p_x_volmatr.length;m++)
	{
	    UnitTests.checkDoubleRel(r.p_x_volmatr[m][0],"r.p_x_volmatr[m][0]",ref.pvmatr[0][m],"ref.pvmatr[0][m]",eps);
	}
	System.err.println("TradeDPV:checkMatchingTradeDPVR(): matching OK");
    }


    static final Trade [] getRandomTradesPolynomialPfromV(
	final long [] out_val,
	final int tsize,
	final java.util.Random r,
	final double tmaxdivtau,
	final long tau,
	final double [] polp)
    {
	final Trade [] tb=new Trade [tsize];
	final long tmax=(long)(tmaxdivtau*tau);
	final long currenttime=tmax;
	final long dt=(long)(tau*(tmaxdivtau/tb.length));
	final double atau=1.0/tau;
	System.err.println("getRandomTradesPolynomialPfromV: tau="+tau+
			   " tau/dt="+(((double)tau)/dt)+
			   " tmax/tau="+(((double)tmax)/tau));

	long t=tmax,vol=0;
	for(int i=tb.length;i-->0;)
	{
	    t-=1+(long)(dt*(0.2+1.6*r.nextDouble()));
	    final int v=1+r.nextInt(3);
	    vol+=v;
	    // p=sum_0^L polp[k] vol^k / (k+1)!
	    final double p=getD0(polp,vol);
	    tb[i]=new Trade(p,
			    v,
			    null,
			    r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
			    t,
			    0,
			    0,
			    t-r.nextInt(100000),
			    null,
			    Trade.TRADE_TYPE_REGULAR,
			    null,
			    null,
			    i);
	}
	out_val[0]=currenttime;
	out_val[1]=vol;
	return tb;
    }

    static double [][] getExactShortAction_x_P(final double [] polp,
					       final Trade [] tb,
					       final long currenttime,
					       final long tau,
					       final int nvs)
    {
	final double atau=1.0/tau;
	// calculate short action directly
	final double [][] saction=new double [nvs][polp.length];
	long vol=0;
	for(int i=tb.length;--i>=0;)
	{
	    final Trade t=tb[i];
	    vol+=t.shares;
	    final double ts=(currenttime-t.time)*atau;
	    final double dp=getD1(polp,vol),p=getD0(polp,vol);
	    for(int m=0;m<polp.length;m++)
	    {
		double wpe=1;
		for(int k=0;k<nvs;k++)
		{
		    saction[k][m]+=wpe*dp*dp*t.shares*Math.exp(-ts/(m+1))/(m+1);
		    wpe*=p/(k+1);
		}
	    }
	}
	return saction;
    }


    private static void checkPolynomialP1P2Action(final int tsize,final java.util.Random r)
    {
	final double eps=1e-5;
	final int n=N_EMOMENTS;
	final double tmaxdivtau=10;
	final long tau=10000000L;
	final long [] out_val=new long [2];
	final double vscale=(tsize/tmaxdivtau)*2;

	final double [] polp1=new double [n];
	double vv=1;
	for(int m=0;m<(n+1)/2;m++)
	{
	    polp1[m]=(-1+2*r.nextDouble())/vv;
	    vv*=vscale;
	}
	final double [] polp2=new double [n]; // coefs near p^2/2
	for(int m=0;m<n;m++)
	{
	    for(int k=0;k<=m;k++)
	    {
		polp2[m]+=0.5*(polp1[k]*polp1[m-k]/(Polynomials.fact(k)*Polynomials.fact(m-k)))*Polynomials.fact(m);
	    }
	}

	final Trade [] tb=getRandomTradesPolynomialPfromV(out_val,tsize,r,tmaxdivtau,tau,polp1);
	final long currenttime=out_val[0];

	final TradeDPV ro=calculateTradeDPV(currenttime,
					    tau,
					    tb,
					    tsize,
					    null);
	// check p1 polynomial coefficients match
	for(int m=0;m<n;m++)
	{
	    System.err.println("ro.p1coefs["+m+"]="+ro.p1coefs[m]+" polp1["+m+"]="+polp1[m]);
	    UnitTests.checkDoubleRel(ro.p1coefs[m],"ro.p1coefs["+m+"]",polp1[m],"polp1["+m+"]",eps);
	}
	// check p2 polynomial coefficients match
	for(int m=0;m<n;m++)
	{
	    System.err.println("ro.p2coefs["+m+"]="+ro.p2coefs[m]+" polp2["+m+"]="+polp2[m]);
	    UnitTests.checkDoubleRel(ro.p2coefs[m],"ro.p2coefs["+m+"]",polp2[m],"polp2["+m+"]",eps);
	}


	final double [] saction=getExactShortAction_x_P(polp1,tb,currenttime,tau,1)[0];
	// match action
	for(int m=0;m<n;m++)
	{
	    System.err.println("saction["+m+"]="+saction[m]+" ro.sum_shortaction["+m+"]="+ro.sum_shortaction[m]);
	    UnitTests.checkDoubleRel(saction[m],"saction["+m+"]",ro.sum_shortaction[m],"ro.sum_shortaction[m]",eps);
	}


	System.err.println("TradeDPV::checkPolynomialP1P2Action(): test OK coefficients matched, N points="+tsize);
    }



    private static void checkPolynomialP(final int tsize,final java.util.Random r)
    {
	final double eps=1e-5,aeps=eps;
	final int n=N_EMOMENTS;
	final double tmaxdivtau=100;
	final long tau=10000000L;
	final long [] out_val=new long [2];
	final double vscale=(tsize/tmaxdivtau)*100;

	final double [] polp=new double [n];
	double vv=1;
	for(int m=0;m<n;m++)
	{
	    polp[m]=(-1+2*r.nextDouble())/vv;
	    vv*=vscale;
	}
	final Trade [] tb=getRandomTradesPolynomialPfromV(out_val,tsize,r,tmaxdivtau,tau,polp);
	final long currenttime=out_val[0];

	final TradeDPV ro=calculateTradeDPV(currenttime,
					    tau,
					    tb,
					    tsize,
					    null);
	// check polynomial coefficients match
	for(int m=0;m<n;m++)
	{
	    System.err.println("ro.p1coefs["+m+"]="+ro.p1coefs[m]+" polp["+m+"]="+polp[m]);
	    UnitTests.checkDoubleRel(ro.p1coefs[m],"ro.p1coefs["+m+"]",polp[m],"polp["+m+"]",eps);
	}

	// for not dcoe: action will not match exactly , because p^2/2 term would give high order terms.
	// for dcoe: match for  polinomials.
	final double [] saction=getExactShortAction_x_P(polp,tb,currenttime,tau,1)[0];
	// match action
	for(int m=0;m<n;m++)
	{
	    System.err.println("saction["+m+"]="+saction[m]+" ro.sum_shortaction["+m+"]="+ro.sum_shortaction[m]);
	    UnitTests.checkDoubleRel(saction[m],"saction["+m+"]",ro.sum_shortaction[m],"ro.sum_shortaction[m]",aeps);
	}

	System.err.println("TradeDPV::checkPolynomialP(): test OK coefficients matched, N points="+tsize);
    }




    private static void matchWithTest(final TradeDPV r,
				      final Trade [] tb,
				      final int tsize)
    {

	final TradePV ref=calculateTradePV(r.currenttime,
					   r.tau,
					   false,
					   tb,
					   tsize,
					   null);
	TradePV.checkMatchingTradeDPV(r,ref);

	final TradeDPV rt=calculateTestTradeDPV(r.currenttime,
						r.tau,
						tb,
						tsize);
	checkMatching(r,rt);
    }


    private static void checkCalculations(final int tsize,final java.util.Random r)
    {
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println("checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);
	final TradeDPV ro=calculateTradeDPV(currenttime,
					    tau,
					    tb,
					    tsize,
					    null);
	matchWithTest(ro,tb,tsize);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeDPV em_ref=calculateTradeDPV((tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest(em_ref,tb,n_split);

	final TradeDPV em_fromref=calculateTradeDPV(currenttime,tau,tb,tsize,em_ref);
	checkMatching(ro,em_fromref);
	matchWithTest(em_fromref,tb,tsize);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeDPV emc_cut=calculateTradeDPV(currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut);

	System.err.println("TradeDPV::checkCalculations(): test random values OK N points="+tsize);
    }


    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int i=0;i<10;i++)
	{
	    checkCalculations(100+r.nextInt(100),r);
	    checkCalculations(10010+r.nextInt(1000010),r);
	    checkPolynomialP1P2Action(10010+r.nextInt(1010)/* 8 */,r);
	    checkPolynomialP(1000+r.nextInt(1010) ,r);
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeDPV::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }


    static class TestPlowPhigh {
	final double omegatau,omega;
	final int N_steps_per_tau,N_periods;
	final long tau=10000000000L,dt,tmax;
	TestPlowPhigh(final double omegatau,
		      final int N_steps_per_tau,
		      final int N_periods)
	{
	    this.omegatau=omegatau;
	    this.N_steps_per_tau=N_steps_per_tau;
	    this.N_periods=N_periods;
	    this.dt=tau/this.N_steps_per_tau;
	    this.omega=omegatau/tau;
	    this.tmax=(long)(Math.PI*2*N_periods/this.omega);
	}
	static final double p0=0,pamp=1.5;
	final int v0=100;


	final Trade [] tb=new Trade [1];
	long t=0;
	double p=Double.NaN,tscaled=Double.NaN;
        boolean setNextP()
	{
	    t+=dt;
	    p=p0+pamp*Math.cos(omega*t);
	    tscaled=omega*t/(2*Math.PI);
	    tb[0]=Trade.getDummyTrade(p,v0,BookItem.BOOK_BUY,t);
	    if(t<tmax)
	    {
		return true;
	    }
	    return false;
	}


	void printValues()
	{
	    System.err.println("#tscaled\tp\tp_low\tp_high\ta(p_last)\tn_reroots\tshortaction_m\tre_roots\td/dp at reroot");
	    TradeDPV prev_ro=null;
	    for(;;)
	    {
		if(!setNextP())
		{
		    break;
		}
		final TradeDPV ro=calculateTradeDPV(t,
						    tau,
						    tb,
						    1,
						    prev_ro);
		prev_ro=ro;
		System.err.print(String.valueOf(tscaled)+"\t"+p+"\t"+ro.p_low+"\t"+ro.p_high+"\t"+ro.real_roots.length+
				 "\t"+getD0(ro.actioncoefs,ro.trade_last.price));
		for(int m=0;m<ro.sum_shortaction.length;m++)
		{
		    System.err.print("\t"+ro.sum_shortaction[m]/(omega*omega));
		}

		for(int m=0;m<ro.root_re.length;m++)
		{
		    if(m<ro.real_roots.length)
		    {
			System.err.print("\t"+ro.real_roots[m]);
		    }
		    else
		    {
			System.err.print("\t?");
		    }
		}

		for(int m=0;m<ro.root_re.length;m++)
		{
		    if(m<ro.real_roots.length)
		    {
			System.err.print("\t"+getD1(ro.actioncoefs,ro.real_roots[m]));
		    }
		    else
		    {
			System.err.print("\t?");
		    }
		}
		System.err.println();
	    }
	}

	public static void main(final String [] args)
	{
	    final TestPlowPhigh s=new TestPlowPhigh(0.2,
						    1000,
						    4);
	    s.printValues();
	}
    }
}




static TradeDPV calculateTradeDPV(final long currenttime,
				  final long tau,
				  final BookItem [] tb,
				  final int tsize,
				  final TradeDPV ref)
{

    final int n=N_EMOMENTS,np=2*n;
    final double [][]  // all 0 by default
	volmatr=new double [n][np],
	pvmatr=new double [n][np],
	pvolmatr=new double [n][np];
    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    long prevt=currenttime,vol=0;
    gloop:for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break gloop;
	}
	prevt=b.time;
	vol+=b.shares;
	final double p=b.price,ts=(currenttime-prevt)*atau,bw=b.shares,vol_d=vol;

	for(int m=0;m<n;m++)
	{
	    final double w=Math.exp(-ts/(m+1))/(m+1),bwmw=bw*w,pbwmw=p*bwmw;
	    final double [] sum_volmatr=volmatr[m],sum_pvmatr=pvmatr[m],sum_pvolmatr=pvolmatr[m];
	    double w_vol=1,wpe=1;
	    for(int mt=0;mt<np;mt++)
	    {
		sum_volmatr[mt]+=w_vol*bwmw;
		sum_pvmatr[mt]+=wpe*bwmw;
		sum_pvolmatr[mt]+=w_vol*pbwmw;
		w_vol*=(vol_d/(mt+1));
		wpe*=(p/(mt+1));
	    }
	}
    }// end i loop

    if(bref!=null)
    { // optimization

	// add previous interval
	final double ref_adelay=(currenttime-ref.currenttime)*atau,vol_d=vol;
	for(int m=0;m<n;m++)
	{
	    final double exp_delay=Math.exp(-ref_adelay/(m+1));
	    final double [] sum_volmatr=volmatr[m],sum_pvmatr=pvmatr[m],sum_pvolmatr=pvolmatr[m],
		ref_sum_volmatr=ref.volmatr[m],ref_sum_pvmatr=ref.pvmatr[m],ref_sum_pvolmatr=ref.pvolmatr[m];
	    for(int mt=0;mt<np;mt++)
	    {
		// use Newton binomial expansion
		double w_vol=1,s_volmatr=0,s_pvolmatr=0;
		for(int ms=0;ms<=mt;ms++)
		{
		    s_volmatr+=w_vol*ref_sum_volmatr[mt-ms];
		    s_pvolmatr+=w_vol*ref_sum_pvolmatr[mt-ms];
		    w_vol*=(vol_d/(ms+1));
		}
		sum_volmatr[mt]+=exp_delay*s_volmatr;
		sum_pvolmatr[mt]+=exp_delay*s_pvolmatr;
		sum_pvmatr[mt]+=exp_delay*ref_sum_pvmatr[mt];
	    }
	}
	vol+=ref.vol;
    }

    return new TradeDPV(currenttime,
			tau,
			volmatr,
			pvmatr,
			pvolmatr,
			// last processed point
			tsize>0?tb[tsize-1]:null,
			vol);
}
















static final boolean TradeDPVR_FLAG_VOL=true,TradeDPVR_FLAG_LOG=false;

static class TradeDPVR {
    final long currenttime,tau;

    private final double [][]
	p_x_volmatr // [m][k]=sum exp(-(t0-t_i)/tau) v_i p^m vol^k/(k!*m!)
	;
    private final double [] p1coefs,shortaction_pcoefs;

    private final double [] saction_x_p;

    // references to previous calculations
    private final BookItem trade_last;
    final long vol;


    final double [] root_re,root_im,real_roots;
    final double p_low,p_high,F;

    TradeDPVR(final long currenttime,
	      final long tau,

	      final double [][] p_x_volmatr,

	      // references
	      final BookItem trade_last,
	      final long vol,
	      final int np,
	      final int nv,
	      final int np_ev)
	{
	    this.currenttime=currenttime;
	    this.tau=tau;
	    this.p_x_volmatr=p_x_volmatr;
	    this.trade_last=trade_last;
	    this.vol=vol;

	    final double [][] msol=new double [Math.max(np,nv)][Math.max(np,nv)];
	    final double [] dcoe;
	    // vol expansion
	    this.p1coefs=setVolMatrixAndRightPart(msol,nv,p_x_volmatr);
	    if(this.p1coefs!=null)
	    {
		final int [] ipvt=new int [msol.length];
		solveLinsystem(msol,this.p1coefs,nv,ipvt);
		dcoe=TradeDPV.getDcoe(this.p1coefs);
	    }
	    else
	    {
		dcoe=null;
	    }

	    // short action
	    this.shortaction_pcoefs=new double [np];
	    setPriceMatrixAndRightPart(msol,this.shortaction_pcoefs,dcoe,p_x_volmatr);
	    this.saction_x_p=this.shortaction_pcoefs.clone(); // for debugging

	    if(np_ev==np)
	    {
		final int [] ipvt=new int [msol.length];
		// Linear system
		solveLinsystem(msol,this.shortaction_pcoefs,np,ipvt);
	    }
	    else
	    {
		// Eigenvalues
		solveLinsystemEigenvalues(msol,this.shortaction_pcoefs,np,np_ev);
	    }



	    // find roots
	    final double [] pcoe=new double [np]; // temporary array
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0,w=1;k<np;k++)
	    {
		pcoe[k]=this.shortaction_pcoefs[k]/w;
		w*=(k+1);
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    this.root_re=u.re_root;
	    this.root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-1,this.root_re,this.root_im);

	    final double p_last=this.trade_last!=null?
		(TradeDPVR_FLAG_LOG?Math.log(this.trade_last.price):this.trade_last.price)
		:Double.NaN;
	    this.F=getD1(this.shortaction_pcoefs,p_last);
	    // find real roots & the point closest to p_last
	    int ptr_close_to_p_last=-1;
	    if(true)
	    {
		int n_reroots=0;
		double drr=Double.POSITIVE_INFINITY,rr=Double.NaN;
		final double [] reroots=pcoe; // use pcoe as a temporary array
		for(int m=0;m<this.root_re.length;m++)
		{
		    if(rr>this.root_re[m])
		    {
			throw new RuntimeException("Not sorted");
		    }
		    rr=this.root_re[m];
		    if(Math.abs(this.root_im[m])<=EPS_IM)
		    {
			final double drr1=Math.abs(p_last-rr);
			if(drr1<drr)
			{
			    drr=drr1;
			    ptr_close_to_p_last=n_reroots;
			}
			reroots[n_reroots++]=rr;
		    }
		}
		this.real_roots=java.util.Arrays.copyOf(reroots,n_reroots);
	    }

	    if(ptr_close_to_p_last<0)
	    {
		this.p_low=this.p_high=Double.NaN;
	    }
	    else
	    {
		final double vp_low,vp_high;
		final int nrrm=this.real_roots.length-1;
		final double p_close_to_p_last=this.real_roots[ptr_close_to_p_last],
		    daction_p_close_to_p_last=getD1(this.shortaction_pcoefs,p_close_to_p_last);
		if(p_close_to_p_last>=p_last)
		{
		    /*
		    if(-Math.signum(daction_p_close_to_p_last)!=Math.signum(getD0(this.shortaction_pcoefs,p_last)))
		    {
			throw new RuntimeException();
		    }
		    */
		    if(daction_p_close_to_p_last<=0)
		    {
			vp_low=ptr_close_to_p_last>0?this.real_roots[ptr_close_to_p_last-1]:Double.NEGATIVE_INFINITY;
			vp_high=p_close_to_p_last;
		    }
		    else
		    {
			vp_low=p_close_to_p_last;
			vp_high=ptr_close_to_p_last<nrrm?this.real_roots[ptr_close_to_p_last+1]:Double.POSITIVE_INFINITY;
		    }
		}
		else
		{
		    /*
		    if(Math.signum(daction_p_close_to_p_last)!=Math.signum(getD0(this.shortaction_pcoefs,p_last)))
		    {
			throw new RuntimeException();
		    }
		    */
		    if(daction_p_close_to_p_last>=0)
		    {
			vp_low=p_close_to_p_last;
			vp_high=ptr_close_to_p_last<nrrm?this.real_roots[ptr_close_to_p_last+1]:Double.POSITIVE_INFINITY;
		    }
		    else
		    {
			vp_low=ptr_close_to_p_last>0?this.real_roots[ptr_close_to_p_last-1]:Double.NEGATIVE_INFINITY;
			vp_high=p_close_to_p_last;
		    }
		}
		if(TradeDPVR_FLAG_LOG)
		{
		    this.p_low=Math.exp(vp_low);
		    this.p_high=Math.exp(vp_high);
		}
		else
		{
		    this.p_low=vp_low;
		    this.p_high=vp_high;
		}
	    }
	}

    void setPriceMatrixAndRightPart(final double [][] matr,final double [] rpart,final double [] dcoe,final double [][] p_x_v)
    {
	final int np=rpart.length;
	for(int l=0;l<np;l++)
	{
	    long f=1;
	    for(int k=0;k<np;k++)
	    {
		//if(f!=Polynomials.fact(l+k)/(Polynomials.fact(l)*Polynomials.fact(k))) {throw new RuntimeException();}
		matr[l][k]=p_x_v[l+k][0]*f;
		f*=l+1+k;
		f/=1+k;
	    }
	    double s=0;
	    for(int k=dcoe.length;--k>=0;)
	    {
		s+=p_x_v[l][k]*dcoe[k];
	    }
	    rpart[l]=s;
	}
    }

    double [] setVolMatrixAndRightPart(final double [][]  matr,final int nv,final double [][] p_x_v)
    {
	final double [] rpart=new double [nv];
	for(int l=0;l<nv;l++)
	{
	    long f=1;
	    for(int k=0;k<nv;k++)
	    {
		//if(f!=Polynomials.fact(l+k)/(Polynomials.fact(l)*Polynomials.fact(k))) {throw new RuntimeException();}
		matr[l][k]=p_x_v[0][l+k]*f;
		f*=l+1+k;
		f/=1+k;
	    }
	    rpart[l]=p_x_v[1][l];
	}
	return rpart;
    }


    private static void solveLinsystem(final double [][] matr,final double [] coefs,final int n,final int [] ipvt)
    {
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa(matr,n,ipvt);
	if(pinfo!=0)
	{
	    java.util.Arrays.fill(coefs,Double.NaN);
	}
	else
	{
	    com.polytechnik.utils.Linsystems.dgesl(matr,n,ipvt,coefs,1);
	}
    }


    private static void solveLinsystemEigenvalues(final double [][] matr,final double [] coefs,final int n,final int np_ev)
    {
	if(DUMP_DIAG_DPVRN)
	{
	    System.err.println("TsetShortActionEigenvalues matr: np_ev="+np_ev+" n="+n);
	    for(int m=0;m<n;m++)
	    {
		System.err.print("m"+m+"\t");
		for(int j=0;j<n;j++)
		{
		    System.err.print(matr[m][j]+"\t");
		}
		System.err.print("r\t"+coefs[m]);
		System.err.println();
	    }
	}

	final double [] lambdas=new double [n],work=new double [1 + n*(6 + 2*n)];
	final int [] iwork=new int [3 + 5*n];
	if(true)
	{
	    throw new RuntimeException("org.netlib.lapack. is temporarily commented out");
	}
	/*
	final org.netlib.util.intW info = new org.netlib.util.intW(2);
	org.netlib.lapack.DSYEVD.DSYEVD("V","U",n,matr,lambdas,work,work.length,iwork,iwork.length,info);
	if(info.val!=0)
	{
	    java.util.Arrays.fill(coefs,Double.NaN);
	}
	else
	{
	    final double [] rightpart=work; // use as a temporary array
	    // copy the right part
	    System.arraycopy(coefs,0,rightpart,0,n);
	    java.util.Arrays.fill(coefs,0); // clear array
	    // sum highest np_ev eigenvalues
	    for(int e=n,ef=n-np_ev;--e>=ef;)
	    {
		double s=0;
		for(int k=0;k<n;k++)
		{
		    s+=rightpart[k]*matr[k][e]; // eigenvectors are in columns
		}
		s/=lambdas[e]; // divide by lambda, the eigenvectors are normalized, no need to normalize here
		for(int k=0;k<n;k++)
		{
		    coefs[k]+=s*matr[k][e];
		}
	    }
	}
	*/

	if(DUMP_DIAG_DPVRN)
	{
	    for(int m=0;m<n;m++)
	    {
		System.err.print("s"+m+"\t");
		System.err.print(coefs[m]+"\t"+lambdas[m]);
		System.err.println();
	    }
	    // eigenvectors are stored in columns
	    for(int j=0;j<n;j++)
	    {
		System.err.print("E"+j+"\t"+lambdas[j]);
		for(int k=0;k<n;k++)
		{
		    System.err.print("\t"+matr[k][j]);
		}
		System.err.println();
	    }
	}
    }


    private static TradeDPVR calculateTestTradeDPVR(final long currenttime,
						    final long tau,
						    final BookItem [] tb,
						    final int tsize)
    {
	final int np=N_EMOMENTS,nvs=2*np;
	final double [][]  // all 0 by default
	    p_x_volmatr=new double [nvs][nvs];
	final double atau=1.0/tau;
	long prevt=currenttime,vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;

	    vol+=b.shares;
	    final double ts=(currenttime-prevt)*atau,
		vol_d=TradeDPVR_FLAG_VOL?vol:ts,
		price_i=TradeDPVR_FLAG_LOG?Math.log(b.price):b.price;
	    final double bwmw=Math.exp(-ts)*b.shares;
	    double wpe=1;
	    for(int m=0;m<nvs;m++)
	    {
		final double [] sump=p_x_volmatr[m];
		double w_vol=wpe;
		for(int k=0;k<nvs;k++)
		{
		    sump[k]+=bwmw*w_vol;
		    w_vol*=(vol_d/(k+1));
		}
		wpe*=(price_i/(m+1));
	    }
	}// end i loop
	return new TradeDPVR(currenttime,
			     tau,
			     p_x_volmatr,
			     // last processed point
			     tsize>0?tb[tsize-1]:null,
			     vol,
			     np,
			     np,
			     np);
    }


    private static void checkMatching(final TradeDPVR r,final TradeDPVR ref)
    {
	final double eps=1e-5;
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}
	for(int m=0;m<r.p_x_volmatr.length;m++)
	{
	    for(int k=0,nvs=r.p_x_volmatr[m].length;k<nvs;k++)
	    {
		UnitTests.checkDoubleRel(r.p_x_volmatr[m][k],"r.p_x_volmatr[m][k]",ref.p_x_volmatr[m][k],"ref.p_x_volmatr[m][k]",eps);
	    }
	}
	System.err.println("TradeDPVR:checkMatching(): matching OK");
    }

    private static void checkPolynomialP(final int tsize,final java.util.Random r,final double eps)
    {
	final double aeps=eps;
	final int n=N_EMOMENTS;
	final double tmaxdivtau=100;
	final long tau=10000000L;
	final long [] out_val=new long [2];
	final double vscale=(tsize/tmaxdivtau)*100;

	final double [] polp=new double [n];
	double vv=1;
	for(int m=0;m<n;m++)
	{
	    polp[m]=(-1+2*r.nextDouble())/vv;
	    vv*=vscale;
	}
	final Trade [] tb=TradeDPV.getRandomTradesPolynomialPfromV(out_val,tsize,r,tmaxdivtau,tau,polp);
	final long currenttime=out_val[0];

	final TradeDPVR ro=calculateTradeDPVR(currenttime,
					      tau,
					      tb,
					      tsize,
					      null);
	// check polynomial coefficients match
	for(int m=0;m<n;m++)
	{
	    System.err.println("ro.p1coefs["+m+"]="+ro.p1coefs[m]+" polp["+m+"]="+polp[m]);
	    UnitTests.checkDoubleRel(ro.p1coefs[m],"ro.p1coefs["+m+"]",polp[m],"polp["+m+"]",eps);
	}

	// match for  polinomials.
	final double [][] saction=TradeDPV.getExactShortAction_x_P(ro.p1coefs,tb,currenttime,tau,2*n);
	// match action
	for(int k=0;k<ro.saction_x_p.length;k++)
	{
	    System.err.println("saction["+k+"][0]="+saction[k][0]+" ro.saction_x_p["+k+"]="+ro.saction_x_p[k]);
	    UnitTests.checkDoubleRel(saction[k][0],"saction[k][0]",ro.saction_x_p[k],"ro.saction_x_p[k]",aeps);
	}

	//UnitTests.checkDoubleRel(ro.dpdtA*ro.dpdtA*(ro.p_x_volmatr[0][0]*ro.p_x_volmatr[0][0]),"dpdtA^2*Norm",ro.saction_x_p[0],"ro.saction_x_p[0]",aeps);

	System.err.println("TradeDPVR::checkPolynomialP(): test OK coefficients matched, N points="+tsize);
    }




    private static void matchWithTest(final TradeDPVR r,
				      final Trade [] tb,
				      final int tsize)
    {
	if(TradeDPVR_FLAG_VOL && !TradeDPVR_FLAG_LOG)
	{
	    final TradeDPV ref=calculateTradeDPV(r.currenttime,
						 r.tau,
						 tb,
						 tsize,
						 null);
	    TradeDPV.checkMatchingTradeDPVR(r,ref);
	}

	final TradeDPVR rt=calculateTestTradeDPVR(r.currenttime,
						  r.tau,
						  tb,
						  tsize);
	checkMatching(r,rt);
    }


    private static void checkCalculations(final int tsize,final java.util.Random r)
    {
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println("checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);
	final TradeDPVR ro=calculateTradeDPVR(currenttime,
					      tau,
					      tb,
					      tsize,
					      null);
	matchWithTest(ro,tb,tsize);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeDPVR em_ref=calculateTradeDPVR((tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest(em_ref,tb,n_split);

	final TradeDPVR em_fromref=calculateTradeDPVR(currenttime,tau,tb,tsize,em_ref);
	checkMatching(ro,em_fromref);
	matchWithTest(em_fromref,tb,tsize);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeDPVR emc_cut=calculateTradeDPVR(currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut);

	System.err.println("TradeDPVR::checkCalculations(): test random values OK N points="+tsize);
    }


    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-107793L);
	final double epspol=1e-4;
	for(int i=0;i<13;i++)
	{
	    checkCalculations(100+r.nextInt(100),r);
	    checkCalculations(10010+r.nextInt(1000010),r);
	    checkPolynomialP(1000+r.nextInt(1010),r,epspol);
	    checkPolynomialP(10010+r.nextInt(10010),r,epspol);
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeDPVR::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

     static class TestPlowPhighR extends TradeDPV.TestPlowPhigh {

	 TestPlowPhighR(final double omegatau,
			final int N_steps_per_tau,
			final int N_periods)
	 {
	     super(omegatau,N_steps_per_tau,N_periods);
	 }


	 void printValues()
	 {
	     System.err.println("#tscaled\tp\tp_low\tp_high\ta(p_last)\tn_reroots\tshortaction_m\tre_roots\td/dp at reroot");
	     TradeDPVR prev_ro=null;
	     for(;;)
	    {
		if(!setNextP())
		{
		    break;
		}
		final TradeDPVR ro=calculateTradeDPVR(t,
						      tau,
						      tb,
						      1,
						      prev_ro);
		prev_ro=ro;
		System.err.print(String.valueOf(tscaled)+"\t"+p+"\t"+ro.p_low+"\t"+ro.p_high+"\t"+ro.real_roots.length+
				 "\t"+getD0(ro.shortaction_pcoefs,TradeDPVR_FLAG_LOG?Math.log(ro.trade_last.price):ro.trade_last.price));

		for(int m=0;m<ro.root_re.length;m++)
		{
		    if(m<ro.real_roots.length)
		    {
			System.err.print("\t"+ro.real_roots[m]);
		    }
		    else
		    {
			System.err.print("\t?");
		    }
		}

		for(int m=0;m<ro.root_re.length;m++)
		{
		    if(m<ro.real_roots.length)
		    {
			System.err.print("\t"+getD1(ro.shortaction_pcoefs,ro.real_roots[m]));
		    }
		    else
		    {
			System.err.print("\t?");
		    }
		}
		System.err.println();
	    }
	 }

	 public static void main(final String [] args)
	 {
	     final TestPlowPhighR s=new TestPlowPhighR(0.2,
						       1000,
						       4);
	     s.printValues();
	 }
     }
}





static TradeDPVR calculateTradeDPVR(final long currenttime,
				    final long tau,
				    final BookItem [] tb,
				    final int tsize,
				    final TradeDPVR ref)
{

    final int np=N_EMOMENTS,nvs=2*np;
    final double [][]  // all 0 by default
	p_x_volmatr=new double [nvs][nvs];
    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    long prevt=currenttime,vol=0;
    gloop:for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break gloop;
	}
	prevt=b.time;
	vol+=b.shares;
	final double ts=(currenttime-prevt)*atau,
	    vol_d=TradeDPVR_FLAG_VOL?vol:ts,
	    p=TradeDPVR_FLAG_LOG?Math.log(b.price):b.price;
	final double bwmw=Math.exp(-ts)*b.shares;

	double wpe=1;
	for(int m=0;m<nvs;m++)
	{
	    final double [] sump=p_x_volmatr[m];
	    double w_vol=wpe;
	    for(int k=0;k<nvs;k++)
	    {
		sump[k]+=bwmw*w_vol;
		w_vol*=(vol_d/(k+1));
	    }
	    wpe*=(p/(m+1));
	}
    }// end i loop

    if(bref!=null)
    { // optimization

	// add previous interval
	final double ts=(currenttime-ref.currenttime)*atau,exp_delay=Math.exp(-ts),vol_d=TradeDPVR_FLAG_VOL?vol:ts;
	for(int m=0;m<nvs;m++)
	{
	    final double [] sump=p_x_volmatr[m],ref_sump=ref.p_x_volmatr[m];
	    for(int k=0;k<nvs;k++)
	    {
		// use Newton binomial expansion
		double w_vol=1,s_pv=0;
		for(int ms=0;ms<=k;ms++)
		{
		    s_pv+=w_vol*ref_sump[k-ms];
		    w_vol*=(vol_d/(ms+1));
		}
		sump[k]+=exp_delay*s_pv;
	    }
	}
	vol+=ref.vol;
    }

    return new TradeDPVR(currenttime,
			 tau,
			 p_x_volmatr,
			 // last processed point
			 tsize>0?tb[tsize-1]:null,
			 vol,
			 np,
			 np,
			 np);
}







static TradeDPVR calculateTradeDPVR_dis(final long currenttime,
					final long tau,
					final BookItem [] trades,
					final int n_trades,
					final BookItem [] tmpbuf)
{
    final int np=N_EMOMENTS,nvs=2*np,np_ev=np;
    final double [][]  // all 0 by default
	p_x_volmatr=new double [nvs][nvs];
    final long [] out_vol=new long [3];

    final int ptr=CalculateLiquidityDeficit9.findPtrToT_Inside(currenttime,
							       currenttime-tau,
							       n_trades,
							       trades,
							       out_vol);
    final BookItem [] tb=tmpbuf;
    final int tsize;
    if(ptr>=0)
    {
	tsize=n_trades-ptr;
	System.arraycopy(trades,ptr,tmpbuf,0,tsize);
	com.polytechnik.outils.TMPQSortAlgorithm.objectQuickSortPointers(tb,0,tsize-1,Distribution.cmp_by_price_lowest_up);
    }
    else
    {
	tsize=0;
    }
    final double p_min,p_max;
    if(tsize>0)
    {
	p_min=tb[0].price;
	p_max=tb[tsize-1].price;
    }
    else
    {
	p_min=p_max=Double.NaN;
    }

    long vol=0,tsum=0,idsum=0;
    final long vol0=out_vol[0],tsum0=out_vol[1],idsum0=out_vol[2]; // permutation invaliants
    double prevp=Double.NaN;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	vol+=b.shares;
	tsum+=b.time;
	idsum+=b.orderreferencenumber;
	final double vol_d=vol,p=b.price,bwmw=b.shares;
	if(p>prevp)
	{
	    throw new RuntimeException("The data is not sorted: p="+p+" and prevp="+prevp+" i="+i+" tsize="+tsize+" currenttime="+currenttime);
	}
	prevp=p;

	double wpe=1;
	for(int m=0;m<nvs;m++)
	{
	    final double [] sump=p_x_volmatr[m];
	    double w_vol=wpe;
	    for(int k=0;k<nvs;k++)
	    {
		sump[k]+=bwmw*w_vol;
		w_vol*=(vol_d/(k+1));
	    }
	    wpe*=(p/(m+1));
	}
    }// end i loop
    if(vol!=vol0 || tsum!=tsum0 || idsum!=idsum0)
    {
	throw new RuntimeException("Invariant mismatch: vol="+vol+" vol0="+vol0+
				   " tsum="+tsum+" tsum0="+tsum0+
				   " idsum="+idsum+" idsum0="+idsum0+
				   " tsize="+tsize+" currenttime="+currenttime);
    }

    return new TradeDPVR(currenttime,
			 tau,
			 p_x_volmatr,
			 // last processed point
			 n_trades>0?trades[n_trades-1]:null,
			 vol,
			 np,
			 np,
			 np_ev)
	{ double [] setVolMatrixAndRightPart(final double [][]  matr,final int nv,final double [][] p_x_v) {return null;}
	    void setPriceMatrixAndRightPart(final double [][] matr,final double [] rpart,final double [] dcoe,final double [][] p_x_v) {
	    super.setPriceMatrixAndRightPart(matr,rpart,new double [0],p_x_v); // set the matr, which is the same as before
	    if(p_max<p_min)
	    {
		throw new RuntimeException("p_max="+p_max+" p_min="+p_min);
	    }
	    final int np=rpart.length;
	    double wl_min=1,wl_max=1;
	    for(int l=0;l<np;l++)
	    {
		wl_min*=p_min/(1+l);
		wl_max*=p_max/(1+l);
		rpart[l]=wl_max-wl_min;
	    }

//System.err.println("calculateTradeDPVR_dis:: p_max="+p_max+" p_min="+p_min);
//	System.err.println("setPriceMatrixAndRightPart matr: np="+np);
//	for(int m=0;m<np;m++)
//	{
//	    System.err.print("m"+m+"\t");
//	    for(int j=0;j<np;j++)
//	    {
//		System.err.print(matr[m][j]+"\t");
//	    }
//	    System.err.print("r\t"+rpart[m]);
//	    System.err.println();
//	}

	} }
	/* { void setPriceMatrixAndRightPart(final double [][] matr,final double [] rpart,final double [] dcoe,final double [][] p_x_v) {
	final int np=rpart.length;
	double wl_min=p_min,wl_max=p_max;
	for(int l=0;l<np;l++)
	{
	    double wmin=wl_min,wmax=wl_max;
	    for(int k=0;k<np;k++)
	    {
		matr[l][k]=(wmax-wmin)/(1+l+k);//  (p_max^{1+l+k}-p_min^{1+l+k})/(l!*k!*(1+l+k))
		wmin*=p_min/(1+k);
		wmax*=p_max/(1+k);
		//UnitTests.checkDoubleRel(matr[l][k],"matr[l][k]",(Math.pow(p_max,1+l+k)-Math.pow(p_min,1+l+k))/(Polynomials.fact(l)*Polynomials.fact(k)*(1+l+k)),"P^pow",1e-6);
	    }
	    wl_min*=p_min/(1+l);
	    wl_max*=p_max/(1+l);

	    double s=0;
	    for(int k=dcoe.length;--k>=0;)
	    {
		s+=p_x_v[l][k]*dcoe[k];
	    }
	    rpart[l]=s;
	}
	} } */
	;
}





    static class TestPlowPhigh2 {
	final double omegatau,omega;
	final int N_steps_per_tau,N_periods;
	final long tau=10000000000L,dt,tmax;
	Trade [] tb;
	TestPlowPhigh2(final double omegatau,
		       final int N_steps_per_tau,
		       final int N_periods)
	{
	    this.omegatau=omegatau;
	    this.N_steps_per_tau=N_steps_per_tau;
	    this.N_periods=N_periods;
	    this.dt=tau/this.N_steps_per_tau;
	    this.omega=omegatau/tau;
	    this.tmax=(long)(Math.PI*2*N_periods/this.omega);

	    this.tb=new Trade [this.N_steps_per_tau*3+100];
	    this.tmpb=new Trade [this.tb.length];
	}
	static final double p0=0,pamp=1.5;
	final int v0=100;

	long t=0;
	double p=Double.NaN,tscaled=Double.NaN;
	int ntrades=0;
	Trade [] tmpb;
        boolean setTrades()
	{
	    t+=dt;
	    p=p0+pamp*Math.cos(omega*t);
	    tscaled=omega*t/(2*Math.PI);
	    if(ntrades>=tb.length)
	    {
		final Trade [] tbnew=new Trade [2*ntrades+10];
		System.arraycopy(tb,0,tbnew,0,ntrades);
		this.tb=tbnew;
		this.tmpb=new Trade [this.tb.length];
	    }
	    tb[ntrades++]=Trade.getDummyTrade(p,v0,BookItem.BOOK_BUY,t);
	    if(t<tmax)
	    {
		return true;
	    }
	    return false;
	}


	void printValues()
	{
	    System.err.println("#tscaled\tp\tp_low\tp_high\tn_reroots\tp_min\tp_max\tre_roots");
	    for(;;)
	    {
		if(!setTrades())
		{
		    break;
		}
		final TradeDPVRN ro=calculateTradeDPVRN_dis(t,
							  tau,
							  tb,
							  ntrades/*,
								   this.tmpb*/);
		System.err.print(String.valueOf(tscaled)+"\t"+p+"\t"+ro.p_low+"\t"+ro.p_high+"\t"+ro.real_roots.length+"\t"+ro.p_min+"\t"+ro.p_max);

		for(int m=0;m<ro.root_re.length;m++)
		{
		    if(m<ro.real_roots.length)
		    {
			System.err.print("\t"+ro.real_roots[m]);
		    }
		    else
		    {
			System.err.print("\t?");
		    }
		}
		for(int m=0;m<ro.saction_x_m.length;m++)
		{
		    System.err.print("\t"+ro.saction_x_m[m]);
		}

		System.err.println();
	    }
	}

	public static void main(final String [] args)
	{
	    final TestPlowPhigh2 s=new TestPlowPhigh2(0.2,
						      1000,
						      4);
	    s.printValues();
	}
    }




static class TradeDPVRN {
    final long currenttime,tau;

    private final double []
	moments,   // [m]=sum (p-<p>)^m v_i
	normformoments // typically [m]=sum |p-<p>|^m v_i
	;
    private final double [] shortaction_mcoefs;

    private final double [] saction_x_m;

    // references to previos calculations
    private final BookItem trade_last;
    final double p_min,p_max,p_average;
    final long vol;

    private final double [] root_re,root_im;
    final double [] real_roots;
    final double p_low,p_high,F;

    TradeDPVRN(final long currenttime,
	       final long tau,

	       final double [] moments,
	       final double [] normformoments,

	       final double p_min,
	       final double p_max,
	       final double p_average,

	       final long vol,
	       final int np,
	       final int np_ev,

	       // references
	       final BookItem trade_last)
	{
	    this.currenttime=currenttime;
	    this.tau=tau;
	    this.moments=moments;
	    this.normformoments=normformoments;

	    this.p_min=p_min;
	    this.p_max=p_max;
	    this.p_average=p_average;
	    this.trade_last=trade_last;
	    this.vol=vol;

	    final double [][] msol=new double [np][np];
	    this.shortaction_mcoefs=new double [np];
	    setPriceMatrixAndRightPart(msol,this.shortaction_mcoefs,this.moments,this.normformoments,this.p_min,this.p_max,this.p_average,np);
	    this.saction_x_m=this.shortaction_mcoefs.clone(); // for debugging


	    if(np_ev==np)
	    {
		// Linear system
		TradeDPVR.solveLinsystem(msol,this.shortaction_mcoefs,np,new int [msol.length]);
	    }
	    else
	    {
		// Eigenvalues, the this.shortaction_mcoefs is overriden.
		TradeDPVR.solveLinsystemEigenvalues(msol,this.shortaction_mcoefs,np,np_ev);
	    }

	    // find roots
	    final double [] pcoe=new double [np]; // temporary array
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0;k<np;k++)
	    {
		pcoe[k]=this.shortaction_mcoefs[k]/this.normformoments[k];
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    this.root_re=u.re_root;
	    this.root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-1,this.root_re,this.root_im);

	    final double dp_last=this.trade_last!=null?this.trade_last.price-p_average:Double.NaN;
	    this.F=getD1WithScale(this.shortaction_mcoefs,dp_last,this.normformoments);
	    // find real roots & the point closest to dp_last
	    int ptr_close_to_p_last=-1;
	    if(true)
	    {
		int n_reroots=0;
		double drr=Double.POSITIVE_INFINITY,rr=Double.NaN;
		final double [] reroots=pcoe; // use pcoe as a temporary array
		for(int m=0;m<this.root_re.length;m++)
		{
		    if(rr>this.root_re[m])
		    {
			throw new RuntimeException("Not sorted");
		    }
		    rr=this.root_re[m];
		    if(Math.abs(this.root_im[m])<=EPS_IM)
		    {
			final double drr1=Math.abs(dp_last-rr);
			if(drr1<drr)
			{
			    drr=drr1;
			    ptr_close_to_p_last=n_reroots;
			}
			reroots[n_reroots++]=rr;
		    }
		    if(DUMP_DIAG_DPVRN)
		    {
			System.err.println("this.root_re["+m+"]="+this.root_re[m]+" this.root_im["+m+"]="+this.root_im[m]+" ptr_close_to_p_last="+ptr_close_to_p_last);
		    }
		}
		this.real_roots=java.util.Arrays.copyOf(reroots,n_reroots);
	    }

	    if(ptr_close_to_p_last<0)
	    {
		this.p_low=this.p_high=Double.NaN;
	    }
	    else
	    {
		final double dp_low,dp_high;
		final int nrrm=this.real_roots.length-1;
		final double dp_close_to_p_last=this.real_roots[ptr_close_to_p_last],
		    daction_dp_close_to_p_last=getD1WithScale(this.shortaction_mcoefs,dp_close_to_p_last,this.normformoments);
		if(dp_close_to_p_last>=dp_last)
		{
		    /*
		    if(-Math.signum(daction_dp_close_to_p_last)!=Math.signum(getD0WithScale(this.shortaction_mcoefs,dp_last,this.normformoments)))
		    {
			throw new RuntimeException();
		    }
		    */
		    if(daction_dp_close_to_p_last<=0)
		    {
			dp_low=ptr_close_to_p_last>0?this.real_roots[ptr_close_to_p_last-1]:Double.NEGATIVE_INFINITY;
			dp_high=dp_close_to_p_last;
		    }
		    else
		    {
			dp_low=dp_close_to_p_last;
			dp_high=ptr_close_to_p_last<nrrm?this.real_roots[ptr_close_to_p_last+1]:Double.POSITIVE_INFINITY;
		    }
		}
		else
		{
		    /*
		    if(Math.signum(daction_dp_close_to_p_last)!=Math.signum(getD0WithScale(this.shortaction_mcoefs,dp_last,this.normformoments)))
		    {
			throw new RuntimeException();
		    }
		    */
		    if(daction_dp_close_to_p_last>=0)
		    {
			dp_low=dp_close_to_p_last;
			dp_high=ptr_close_to_p_last<nrrm?this.real_roots[ptr_close_to_p_last+1]:Double.POSITIVE_INFINITY;
		    }
		    else
		    {
			dp_low=ptr_close_to_p_last>0?this.real_roots[ptr_close_to_p_last-1]:Double.NEGATIVE_INFINITY;
			dp_high=dp_close_to_p_last;
		    }
		}
		this.p_low=p_average+dp_low;
		this.p_high=p_average+dp_high;
	    }
	    // shift to regular price offset
	    for(int m=0;m<this.real_roots.length;m++)
	    {
		this.real_roots[m]+=p_average;
	    }
	}


    void setPriceMatrixAndRightPart(final double [][] matr,final double [] rpart,final double [] moments,final double [] norm,
				    final double p_min,final double p_max,final double p_average,final int np)
    {
	final double dp_min=p_min-p_average,dp_max=p_max-p_average;
	double wl_min=1,wl_max=1;
	for(int l=0;l<np;l++)
	{
	    final double norml=norm[l];
	    for(int k=0;k<np;k++)
	    {
		matr[l][k]=moments[l+k]/norml/norm[k];
	    }
	    wl_min*=dp_min;
	    wl_max*=dp_max;
	    rpart[l]=(wl_max-wl_min)/(l+1)/norml;
	}
	if(DUMP_DIAG_DPVRN)
	{
	    System.err.println("setPriceMatrixAndRightPart: p_min="+p_min+" p_max="+p_max+" p_average="+p_average+" np="+np);
	}
    }




    private static void matchWithTest(final TradeDPVRN r,
				      final Trade [] tb,
				      final int tsize)
    {
	final double eps=1e-5;
	final TradeDPVR ref=calculateTradeDPVR_dis(r.currenttime,
						   r.tau,
						   tb,
						   tsize,
						   new BookItem [tsize]);
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.moments[0],"r.moments[0]",ref.p_x_volmatr[0][0],"ref.p_x_volmatr[0][0]",eps);
	UnitTests.checkDoubleRel(r.p_average,"r.p_average",ref.p_x_volmatr[1][0]/ref.vol,"ref.p_x_volmatr[1][0]/ref.vol",eps);
	final double p_average=r.p_average;
	// high order moments
	for(int l=0;l<r.moments.length;l++)
	{
	    double wpe=1.0/Polynomials.fact(l),s=0;
	    for(int k=0;k<=l;k++)
	    {
		s+=r.moments[l-k]*wpe;
		wpe*=(p_average/(k+1))*(l-k);
	    }
	    UnitTests.checkDoubleRel(s,"s("+l+")",ref.p_x_volmatr[l][0],"ref.p_x_volmatr["+l+"][0]",eps);
	}
	// right part
	UnitTests.checkDoubleRel(r.saction_x_m[0]*r.normformoments[0],"r.saction_x_m[0]*r.normformoments[0]",ref.saction_x_p[0],"ref.saction_x_p[0]",eps);
	for(int l=0;l<r.saction_x_m.length;l++)
	{
	    double wpe=1.0/Polynomials.fact(l+1),s=0;
	    for(int k=0;k<=l;k++)
	    {
		s+=r.saction_x_m[l-k]*wpe*r.normformoments[l-k]*(l-k+1);
		wpe*=(p_average/(k+1))*(l-k+1);
	    }
	    UnitTests.checkDoubleRel(s,"sright("+l+")",ref.saction_x_p[l],"ref.saction_x_p["+l+"]",eps);
	}

	// polynomial coefficients check
	System.err.println("r.p_min="+r.p_min+" r.p_max="+r.p_max+" r.p_average="+r.p_average+" tsize="+tsize);
	// convert to scale=1
	final double [] coefsn=new double [r.shortaction_mcoefs.length],coefso=new double [r.shortaction_mcoefs.length];
	for(int l=0;l<r.shortaction_mcoefs.length;l++)
	{
	    coefsn[l]=r.shortaction_mcoefs[l]/r.normformoments[l];
	    coefso[l]=ref.shortaction_pcoefs[l]/Polynomials.fact(l);
	    System.err.println("cm"+l+"\t"+coefsn[l]+"\t"+coefso[l]);
	}
	final double [] coefsno=new double [r.shortaction_mcoefs.length];
	for(int l=0;l<r.shortaction_mcoefs.length;l++)
	{
	    double wpe=coefsn[l];
	    for(int k=0;k<=l;k++)
	    {
		coefsno[l-k]+=wpe;
		wpe*=(-p_average/(1+k))*(l-k);
	    }
	}
	for(int l=0;l<r.shortaction_mcoefs.length;l++)
	{
	    System.err.println("coefsno["+l+"]="+coefsno[l]+" coefso["+l+"]="+coefso[l]+" (coefs with scale=1 offset=0)");
	    UnitTests.checkDoubleRel(coefsno[l]*Polynomials.fact(l),"coefsno["+l+"]*Polynomials.fact(l)",ref.shortaction_pcoefs[l],"ref.shortaction_pcoefs["+l+"]",eps);
	}
	// check roots & derivative
	for(int l=0;l<r.real_roots.length;l++)
	{
	    UnitTests.checkDoubleRel(r.real_roots[l],"r.real_roots["+l+"]",ref.real_roots[l],"ref.real_roots["+l+"]",eps);
	}
	UnitTests.checkDoubleRel(r.F,"r.F",ref.F,"ref.F",eps);


	System.err.println("r.p_low="+r.p_low+"\t ref.p_low="+ref.p_low);
	System.err.println("r.p_high="+r.p_high+"\t ref.p_high="+ref.p_high);


	System.err.println("TradeDPVRN::matchWithTest matched "+tsize+" points OK");
    }


    static Trade [] getRandomTradesPolynomialPV(final java.util.Random r,
						final int tsize,
						final double [] polp)
    {
	final Trade [] tb=new Trade [tsize];
	final int vscale=10;
	long t=10,vol=0;
	double p=1;
	for(int i=0;i<tb.length;i++)
	{
	    t+=1+r.nextInt(10000);
	    final int dv=vscale;//2+r.nextInt(vscale);
	    final double dp=dv*getD0(polp,p);
	    p+=dp;
	    vol+=dv;
	    //p=polp[0]*vol;
	    tb[i]=new Trade(p,
			    dv,
			    null,
			    r.nextInt(2)<=0?BookItem.BOOK_BUY:BookItem.BOOK_SELL,
			    t,
			    0,
			    0,
			    t-r.nextInt(100000),
			    null,
			    Trade.TRADE_TYPE_REGULAR,
			    null,
			    null,
			    i);
	}
	return tb;
    }


    static void checkPolynomialPV(final java.util.Random r,final int tsize)
    {
	final double [] polp=new double [N_EMOMENTS];
	polp[0]=1e-4;
	final Trade [] tb=getRandomTradesPolynomialPV(r,tsize,polp);
	final long currenttime=tb[tsize-1].time,tau=currenttime-tb[0].time+1;
	final TradeDPVRN ro=calculateTradeDPVRN_dis(currenttime,
						    tau,
						    tb,
						    tsize);
	matchWithTest(ro,tb,tsize);
	//throw new RuntimeException();
    }



    private static void checkCalculations(final int tsize,final java.util.Random r)
    {
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long currenttime=tb[tb.length-1].time+20000,tau=currenttime-tb[tsize/2].time;
	System.err.println("checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);
	final TradeDPVRN ro=calculateTradeDPVRN_dis(currenttime,
						    tau,
						    tb,
						    tsize);
	matchWithTest(ro,tb,tsize);
    }


    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int i=0;i<10;i++)
	{
	    //checkPolynomialPV(r,10000);

	    checkCalculations(300000,r);
	    checkCalculations(100+r.nextInt(100),r);
	    checkCalculations(10010+r.nextInt(1000010),r);
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeDPVRN::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }
}



static TradeDPVRN calculateTradeDPVRN_dis(final long currenttime,
					  final long tau,
					  final BookItem [] tb,
					  final int tsize)
{
    final int np=N_EMOMENTS,nvs=2*np,np_ev=np;
    double p_min=Double.NaN,p_max=Double.NaN,s_paver=0;
    final long t_cut=currenttime-tau;
    long vol0=0,prevt=currenttime;

    // set pointer
    int ptr_t_range_cut=tsize;
    for(int i=tsize;--i>=0;)
    {
	final BookItem t=tb[i];
	if(t.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+
				       prevt+" and "+t.time+
				       " currenttime="+currenttime);
	}
	prevt=t.time;
	if(prevt<t_cut)
	{
	    break;
	}
	ptr_t_range_cut=i;
	vol0+=t.shares;
	final double p=t.price;
	if(!(p_min<=p))
	{
	    p_min=p;
	}
	if(!(p_max>=p))
	{
	    p_max=p;
	}
	s_paver+=t.shares*p;
    }
    final double p_average=vol0>0?s_paver/vol0:Double.NaN;

    final double [] moments=new double [nvs],norm=new double [nvs];
    // scan second time to avoid rounding errors
    long vol=0;
    for(int i=tsize;--i>=ptr_t_range_cut;)
    {
	final BookItem t=tb[i];
	vol+=t.shares;
	final double dp=t.price-p_average,adp=Math.abs(dp),bwmw=t.shares;
	double mpew=bwmw,ampew=bwmw;
	for(int m=0;m<nvs;m++)
	{
	    moments[m]+=mpew;
	    norm[m]+=ampew;
	    mpew*=dp;
	    ampew*=adp;
	}
    }// end i loop
    if(vol!=vol0)
    {
	throw new RuntimeException("Invariant mismatch: vol="+vol+" vol0="+vol0);
    }
    if(vol>0)
    {
	// rounding errors adjustments
	moments[0]=norm[0]=vol;
	moments[1]=0; // adjustment: <p-<p>> is always 0
	if(DUMP_DIAG_DPVRN)
	{
	    for(int k=0;k<nvs;k++)
	    {
		System.err.println("moments["+k+"]="+moments[k]+" norm["+k+"]="+norm[k]);
	    }
	}
    }
    else
    {
	java.util.Arrays.fill(moments,Double.NaN);
	java.util.Arrays.fill(norm,Double.NaN);
    }

    return new TradeDPVRN(currenttime,
			  tau,
			  moments,
			  norm,

			  p_min,
			  p_max,
			  p_average,

			  vol,
			  np,
			  np_ev,
			  tsize>0?tb[tsize-1]:null);


}

static final boolean DUMP_DIAG_DPVRN=false;




static class TradeIP {
    final TradeIPData tipdata;
    private final double [] normformoments_pt; // typically [m]=<(p-offset)^2>_t^{m/2}

    private final double [] pcoefs,pcoefs_buy,pcoefs_sell,pcoefs_diff;
    final double sigma,skewness,kurtosis;
    final double [] nscale_pcoefs,nscale_pbscoefs;

    final double [] vattractors,vrepellers;
    final double min_A,max_A,ex_A,i0_A,min_R,max_R,ex_R,i0_R;
    private final double [] vreal_roots;



    final double [] DIrpos_buy,DIrneg_buy,DIrpos_sell,DIrneg_sell,DIrpos_sum,DIrneg_sum,DIrpos_diff,DIrneg_diff;
    final double Emin_P_buy,Emax_P_buy,Emin_N_buy,Emax_N_buy,
	Emin_P_sell,Emax_P_sell,Emin_N_sell,Emax_N_sell,
	Emin_P_sum,Emax_P_sum,Emin_N_sum,Emax_N_sum,
	Emin_P_diff,Emax_P_diff,Emin_N_diff,Emax_N_diff;

    final double [] attractors,repellers;
    private final double [] real_roots;

    private final double [] interval_extremum;
    final double [] interval_min,interval_max;


    final double aIPlast,aIPaver;

    // candidates for removal
    final double p_forward,p_backward,p_forward_sd,I0,Iplast,pw_close,pw_stable;
    final double dirv,dirv_buy,dirv_sell,diff_dirv,diffw_dirv,diffw_dirv_dp;
    private final double [] root_re,root_im;
    final double [] p_down,p_up;
    final double dp,ddirp,dddirp,dirp0,ddirp0,dirv0;


    TradeIP(final TradeIPData tipdata)
    {
	this.tipdata=tipdata;
	final int np=this.tipdata.np;
	final double p_offset=this.tipdata.p_offset;

	this.normformoments_pt=TradeIPData.getPowersOfStdev(this.tipdata.moments_pt[2]/this.tipdata.t_sum,np);

	final double [] norms=this.normformoments_pt;
	final double [][] msol=new double [np][np];
	this.pcoefs=new double [np];
	this.pcoefs_buy=new double [np];
	this.pcoefs_sell=new double [np];
	this.pcoefs_diff=new double [np];
	setPriceMatrixAndRightPart(msol,this.pcoefs,this.pcoefs_buy,this.pcoefs_sell,
				   this.tipdata.moments_pt,norms,
				   this.tipdata.moments_pv,this.tipdata.moments_pv_buy,this.tipdata.moments_pv_sell,
				   np);
	// Linear system
	final int [] ipvt=new int [np];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa(msol,np,ipvt);
	if(pinfo!=0)
	{
	    java.util.Arrays.fill(this.pcoefs,Double.NaN);
	    java.util.Arrays.fill(this.pcoefs_buy,Double.NaN);
	    java.util.Arrays.fill(this.pcoefs_sell,Double.NaN);
	    java.util.Arrays.fill(this.pcoefs_diff,Double.NaN);
	}
	else
	{
	    com.polytechnik.utils.Linsystems.dgesl(msol,np,ipvt,this.pcoefs,1);
	    com.polytechnik.utils.Linsystems.dgesl(msol,np,ipvt,this.pcoefs_buy,1);
	    com.polytechnik.utils.Linsystems.dgesl(msol,np,ipvt,this.pcoefs_sell,1);
	    final double w_buy=this.tipdata.moments_pv_buy[0],w_sell=this.tipdata.moments_pv_sell[0];
	    for(int m=0;m<np;m++)
	    {
		this.pcoefs_diff[m]=(w_sell*this.pcoefs_sell[m]-w_buy*this.pcoefs_buy[m])/this.tipdata.v_sum;
	    }
	}

	this.sigma=norms.length>1?norms[1]:Double.NaN;
	this.skewness=this.tipdata.moments_pt[3]/this.tipdata.t_sum;
	this.kurtosis=this.tipdata.moments_pt[4]/this.tipdata.t_sum;

	this.nscale_pcoefs=new double [np];
	this.nscale_pbscoefs=new double [np];
	for(int k=0;k<np;k++)
	{
	    this.nscale_pcoefs[k]=this.pcoefs[k]/norms[k];
	    this.nscale_pbscoefs[k]=this.pcoefs_diff[k]/norms[k];
	}

	// one more approach with d|v|/dt
	if(true)
	{
	    final int n_elements_to_save=7;
	    final double [] pcoe=new double [np-1];
	    final double [][] arres=new double [4*n_elements_to_save][];
	    for(int jj=0,o=0;jj<4;jj++,o+=n_elements_to_save)
	    {
		final double [] cur_pcoefs=jj<=0?this.pcoefs_buy:(jj<=1?this.pcoefs_sell:(jj<=2?this.pcoefs:this.pcoefs_diff));
		// use scale=1 expected by PolynomialRoot
		for(int k=1;k<np;k++)
		{
		    pcoe[k-1]=k*cur_pcoefs[k]/norms[k];
		}
		final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
		double [] root_re=u.re_root,root_im=u.im_root;

		// insertion sort by real part with extra load
		SortArrays.sortArrayWithLoad(np-2,root_re,root_im);
		final double [] reroots=pcoe, // re-use temporary arrays
		    save_pos=root_re,save_neg=root_im;

		int n_reroots=0,n_pos=0,n_neg=0;
		double rr=Double.NaN;
		for(int m=0;m<np-2;m++)
		{
		    if(rr>root_re[m])
		    {
			throw new RuntimeException("Not sorted");
		    }
		    rr=root_re[m];
		    if(Math.abs(root_im[m])<=EPS_IM)
		    { // real root
			if(getD2WithScale(cur_pcoefs,rr,norms)>0)
			{ // min
			    save_pos[n_pos++]=rr;
			}
			else
			{ // max
			    save_neg[n_neg++]=rr;
			}
			reroots[n_reroots++]=rr;
		    }
		}
		// save up to n_elements_to_save elements
		final double [] pos=arres[o]=java.util.Arrays.copyOf(save_pos,n_pos);
		final double [] neg=arres[o+1]=java.util.Arrays.copyOf(save_neg,n_neg);
		arres[o+2]=java.util.Arrays.copyOf(reroots,n_reroots);
		arres[o+3]=getmmAR(cur_pcoefs,norms,pos,false);
		arres[o+4]=getmmAR(cur_pcoefs,norms,neg,true);
		for(int m=pos.length;--m>=0;pos[m]+=p_offset){}
		for(int m=neg.length;--m>=0;neg[m]+=p_offset){}

	    }
	    double [] sar;

	    this.DIrpos_buy=arres[0*n_elements_to_save+0];
	    this.DIrneg_buy=arres[0*n_elements_to_save+1];
	    sar=arres[0*n_elements_to_save+3];
	    this.Emin_P_buy=sar[0]+p_offset;
	    this.Emax_P_buy=sar[1]+p_offset;
	    sar=arres[0*n_elements_to_save+4];
	    this.Emin_N_buy=sar[0]+p_offset;
	    this.Emax_N_buy=sar[1]+p_offset;

	    this.DIrpos_sell=arres[1*n_elements_to_save+0];
	    this.DIrneg_sell=arres[1*n_elements_to_save+1];
	    sar=arres[1*n_elements_to_save+3];
	    this.Emin_P_sell=sar[0]+p_offset;
	    this.Emax_P_sell=sar[1]+p_offset;
	    sar=arres[1*n_elements_to_save+4];
	    this.Emin_N_sell=sar[0]+p_offset;
	    this.Emax_N_sell=sar[1]+p_offset;

	    this.DIrpos_sum=arres[2*n_elements_to_save+0];
	    this.DIrneg_sum=arres[2*n_elements_to_save+1];
	    sar=arres[2*n_elements_to_save+3];
	    this.Emin_P_sum=sar[0]+p_offset;
	    this.Emax_P_sum=sar[1]+p_offset;
	    sar=arres[2*n_elements_to_save+4];
	    this.Emin_N_sum=sar[0]+p_offset;
	    this.Emax_N_sum=sar[1]+p_offset;

	    this.DIrpos_diff=arres[3*n_elements_to_save+0];
	    this.DIrneg_diff=arres[3*n_elements_to_save+1];
	    sar=arres[3*n_elements_to_save+3];
	    this.Emin_P_diff=sar[0]+p_offset;
	    this.Emax_P_diff=sar[1]+p_offset;
	    sar=arres[3*n_elements_to_save+4];
	    this.Emin_N_diff=sar[0]+p_offset;
	    this.Emax_N_diff=sar[1]+p_offset;
	}




	//  I_sell-I_buy=0 equation
	if(true)
	{
	    final double [] pcoe=new double [np]; // temporary array
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0;k<np;k++)
	    {
		pcoe[k]=this.pcoefs_diff[k]/norms[k];
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    final double [] root_re=u.re_root,root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-1,root_re,root_im);
	    final double [] reroots=pcoe, // use pcoe as a temporary array
		save_attractors=new double [np],save_repellers=new double [np];
	    int n_reroots=0,n_attractors=0,n_repellers=0;
	    double rr=Double.NaN;
	    for(int m=0;m<root_re.length;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{ // real root
		    final int type;
		    if(getD1WithScale(this.pcoefs_diff,rr,norms)<0)
		    { // repeller
			save_repellers[n_repellers++]=p_offset+rr;
		    }
		    else
		    { // attractor
			save_attractors[n_attractors++]=p_offset+rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }

	    this.attractors=java.util.Arrays.copyOf(save_attractors,n_attractors);
	    this.repellers=java.util.Arrays.copyOf(save_repellers,n_repellers);
	    this.real_roots=java.util.Arrays.copyOf(reroots,n_reroots);

	}
	final double p_last=this.tipdata.trade_last!=null?this.tipdata.trade_last.price:Double.NaN, // last price
	    dp_last=p_last-p_offset;

	if(true)
	{
	    final double [] pcoe=new double [np-1]; // temporary array
	    for(int k=1;k<np;k++)
	    {
		pcoe[k-1]=k*this.pcoefs_diff[k]/norms[k];
	    }

	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    final double [] ext_re=u.re_root,ext_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-2,ext_re,ext_im);
	    final double [] ext_real=pcoe, // use pcoe as a temporary array
		ext_min=new double [np],ext_max=new double [np];
	    int n_ext=0,n_min=0,n_max=0;
	    double rr=Double.NaN;
	    for(int m=0;m<ext_re.length;m++)
	    {
		if(rr>ext_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=ext_re[m];
		if(Math.abs(ext_im[m])<=EPS_IM)
		{ // real root
		    if(getD2WithScale(this.pcoefs_diff,rr,norms)>0)
		    { // min
			ext_min[n_min++]=p_offset+rr;
		    }
		    else
		    { // max
			ext_max[n_max++]=p_offset+rr;
		    }
		    ext_real[n_ext++]=rr;
		}
	    }
	    this.interval_extremum=java.util.Arrays.copyOf(ext_real,n_ext);
	    this.interval_min=java.util.Arrays.copyOf(ext_min,n_min);
	    this.interval_max=java.util.Arrays.copyOf(ext_max,n_max);
	}


	// different approach, d|v|/dt->max equation
	if(true)
	{
	    final double [] pcoe=new double [np-1]; // temporary array
	    // use scale=1 expected by PolynomialRoot
	    for(int k=1;k<np;k++)
	    {
		pcoe[k-1]=k*this.pcoefs[k]/norms[k];
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    final double [] root_re=u.re_root,root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-2,root_re,root_im);
	    final double [] reroots=pcoe, // re-use as temporary arrays
		save_attractors=root_re,save_repellers=root_im;
	    int n_reroots=0,n_attractors=0,n_repellers=0;
	    double rr=Double.NaN;
	    for(int m=0;m<np-2;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{ // real root
		    if(getD2WithScale(this.pcoefs,rr,norms)>0)
		    { // repeller (min)
			save_repellers[n_repellers++]=rr;
		    }
		    else
		    { // attractor (max)
			save_attractors[n_attractors++]=rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.vattractors=java.util.Arrays.copyOf(save_attractors,n_attractors);
	    this.vrepellers=java.util.Arrays.copyOf(save_repellers,n_repellers);
	    this.vreal_roots=java.util.Arrays.copyOf(reroots,n_reroots);



	    if(true)
	    {
		final double [] sar=getmmAR(this.pcoefs,norms,this.vattractors,true);
		// shift to 0
		this.min_A=sar[0]+p_offset;
		this.max_A=sar[1]+p_offset;
		this.ex_A=sar[2]+p_offset;
		this.i0_A=sar[3];

		final double [] srr=getmmAR(this.pcoefs,norms,this.vrepellers,false);
		// shift to 0
		this.min_R=srr[0]+p_offset;
		this.max_R=srr[1]+p_offset;
		this.ex_R=srr[2]+p_offset;
		this.i0_R=srr[3];
	    }


	    // shift attractor & repeller offset to 0
	    for(int m=this.vattractors.length;--m>=0;this.vattractors[m]+=p_offset){}
	    for(int m=this.vrepellers.length;--m>=0;this.vrepellers[m]+=p_offset){}
	}


	// forward & backward prices, related to the equation I_sell-I_buy=0
	if(true)
	{
	    final double dp_init=0; // start with p_offset
	    final double [] reroots=this.real_roots;
	    final int n_reroots=reroots.length;
	    int ptr_close_to_init=-1;
	    double drr=Double.POSITIVE_INFINITY;
	    for(int m=0;m<n_reroots;m++)
	    {
		final double rr=reroots[m],drr1=Math.abs(dp_init-rr);
		if(drr1<drr)
		{
		    drr=drr1;
		    ptr_close_to_init=m;
		}
	    }

	    this.I0=getD0WithScale(this.pcoefs_diff,dp_init,norms);
	    if(ptr_close_to_init<0)
	    {
		this.p_forward=this.p_backward=this.p_forward_sd=this.pw_close=Double.NaN;
	    }
	    else
	    {
		this.pw_close=reroots[ptr_close_to_init]+p_offset;
		final double dp_low,dp_high;
		final int nrrm=n_reroots-1;
		final double dp_close_to_init=reroots[ptr_close_to_init];
		if(dp_close_to_init>=dp_init)
		{
		    dp_low=ptr_close_to_init>0?reroots[ptr_close_to_init-1]:Double.NEGATIVE_INFINITY;
		    dp_high=dp_close_to_init;
		    if(this.I0<=0)
		    {
			this.p_forward_sd=p_offset+(ptr_close_to_init>1?reroots[ptr_close_to_init-2]:Double.NEGATIVE_INFINITY);
		    }
		    else
		    {
			this.p_forward_sd=p_offset+(ptr_close_to_init<nrrm?reroots[ptr_close_to_init+1]:Double.POSITIVE_INFINITY);
		    }
		}
		else
		{
		    dp_low=dp_close_to_init;
		    dp_high=ptr_close_to_init<nrrm?reroots[ptr_close_to_init+1]:Double.POSITIVE_INFINITY;
		    if(this.I0<=0)
		    {
			this.p_forward_sd=p_offset+(ptr_close_to_init>0?reroots[ptr_close_to_init-1]:Double.NEGATIVE_INFINITY);
		    }
		    else
		    {
			this.p_forward_sd=p_offset+(ptr_close_to_init<nrrm-1?reroots[ptr_close_to_init+2]:Double.POSITIVE_INFINITY);
		    }
		}
		if(this.I0<=0)
		{
		    this.p_backward=p_offset+dp_high;
		    this.p_forward=p_offset+dp_low;
		}
		else
		{
		    this.p_backward=p_offset+dp_low;
		    this.p_forward=p_offset+dp_high;
		}
	    }
	    if(true)
	    {
		this.Iplast=getD0WithScale(this.pcoefs_diff,dp_last,norms);
		double dp_stable;
		if(this.Iplast<0)
		{ // take left
		    dp_stable=Double.NEGATIVE_INFINITY;
		    for(int m=n_reroots;m-->0;)
		    {
			if(reroots[m]<=dp_last)
			{
			    dp_stable=reroots[m];
			    break;
			}
		    }
		}
		else
		{ // take right
		    dp_stable=Double.POSITIVE_INFINITY;
		    for(int m=0;m<n_reroots;m++)
		    {
			if(reroots[m]>=dp_last)
			{
			    dp_stable=reroots[m];
			    break;
			}
		    }
		}
		this.pw_stable=p_offset+dp_stable;
	    }
	}

	// below are candidates for removal
	this.dirv=getD0WithScale(this.pcoefs,dp_last,norms);
	this.ddirp=getDDirP(p_last);
	this.dddirp=getDDDirP(p_last);
	this.dirv_buy=getD0WithScale(this.pcoefs_buy,dp_last,norms);
	this.dirv_sell=getD0WithScale(this.pcoefs_sell,dp_last,norms);
	this.diff_dirv=(this.dirv_sell-this.dirv_buy);
	this.diffw_dirv=getD0WithScale(this.pcoefs_diff,dp_last,norms);
	this.diffw_dirv_dp=this.sigma*getD1WithScale(this.pcoefs_diff,dp_last,norms);



	this.dp=this.tipdata.pv_sum/this.tipdata.v_sum-this.tipdata.pt_sum/this.tipdata.t_sum;
	this.dirp0=p_last-p_offset;
	this.ddirp0=this.dp/norms[1];
	this.dirv0=1+this.ddirp0*(this.dirp0/norms[1]);

	// old approaches: d|v|/dt=v0/tau0 , find real roots
	if(true)
	{
	    final double [] pcoe=new double [np]; // temporary array
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0;k<np;k++)
	    {
		pcoe[k]=this.pcoefs[k]/norms[k];
	    }
	    pcoe[0]-=1; // equation 1=sum_k pcoe_k * (p-p_offset)^k
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    this.root_re=u.re_root;
	    this.root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-1,root_re,root_im);

	    int n_reroots=0,n_p_down=0,n_p_up=0;
	    double rr=Double.NaN;
	    final double [] reroots=pcoe, // use pcoe as a temporary array
		save_p_down=new double [np],save_p_up=new double [np];
	    for(int m=0;m<root_re.length;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{
		    if(getD1WithScale(this.pcoefs,rr,norms)>0)
		    { // up
			save_p_up[n_p_up++]=p_offset+rr;
		    }
		    else
		    { // down
			save_p_down[n_p_down++]=p_offset+rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.p_up=java.util.Arrays.copyOf(save_p_up,n_p_up);
	    this.p_down=java.util.Arrays.copyOf(save_p_down,n_p_down);
	    this.aIPlast=getD0WithScale(this.pcoefs,dp_last,norms);
	    this.aIPaver=getD0WithScale(this.pcoefs,0,norms);
	}
    }

    static void getPointsCloseToEXmmAR(final double [] res,
				       final double [] coefs,
				       final double [] norm,
				       final double pex)
    {

	double pmin=Double.NaN,pmax=Double.NaN;
	for(double p:getSRoots(coefs,norm,pex))
	{
	    if(p<pex && !(p<pmin))
	    {
		pmin=p;
	    }
	    if(p>pex && !(p>pmax))
	    {
		pmax=p;
	    }
	}
	if(false)
	{
	System.err.println(
	    "pex="+pex+" I(pex)="+getD0WithScale(coefs,pex,norm)+" DI(pex)="+getD1WithScale(coefs,pex,norm)+
	    " D2I(pex)="+getD2WithScale(coefs,pex,norm)+
	    " pmin="+pmin+" I(pmin)="+getD0WithScale(coefs,pmin,norm)+" DI(pmin)="+getD1WithScale(coefs,pmin,norm)+
	    " pmax="+pmax+" I(pmax)="+getD0WithScale(coefs,pmax,norm)+" DI(pmax)="+getD1WithScale(coefs,pmax,norm));
	}
	res[0]=pmin;
	res[1]=pmax;
    }


    static double [] getSRoots(final double [] coefs,
			       final double [] norm,
			       final double pex)
    {
	final int nc=3,np=coefs.length,neq=np-(nc-1);
	if(neq<=1)
	{
	    return new double [0];
	}

	final double [] c=new double [nc];
	double wpe=1;
	for(int m=nc;--m>=0;wpe*=-(pex*m)/(nc-m))
	{
	    c[m]=wpe;
	}
	final double [][] matr=new double [neq][];
	final double [] pcoe=new double [neq];
	for(int r=nc-1;r<np;r++)
	{
	    final int ro=r-(nc-1);
	    final double [] matrro=matr[ro]=new double [neq];
	    for(int k=ro,ke=Math.min(r,neq-1);k<=ke;k++)
	    {
		matrro[k]=c[r-k];
	    }
	    pcoe[ro]=coefs[r]/norm[r];
	}
	TradeDPVR.solveLinsystem(matr,pcoe,neq,new int [neq]);
	final PolynomialRootsFinder.PRoots u=pcoe.length-1<=4?com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe):
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG().findRoots(pcoe);
	final double [] root_re=u.re_root,root_im=u.im_root;
	//System.err.println("neq="+neq+" pcoe="+java.util.Arrays.toString(pcoe)+" root_re="+java.util.Arrays.toString(root_re)+" root_im="+java.util.Arrays.toString(root_im));
	SortArrays.sortArrayWithLoad(neq-1,root_re,root_im);
	return java.util.Arrays.copyOf(pcoe,
	    selectRealRoots(pcoe,neq-1,root_re,root_im,EPS_IM));
    }

    /* For testing getSRoots. */
    private static double [] getSRootsTest(final double [] coefs,
					   final double [] norm,
					   final double pex)
    {
	final int np=coefs.length;
	final double [] pcoe=new double [np]; // temporary array
	for(int k=0;k<np;k++)
	{
	    pcoe[k]=coefs[k]/norm[k];
	}
	pcoe[0]-=getD0WithScale(coefs,pex,norm);
	final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	final double [] root_re=u.re_root,root_im=u.im_root,reroots=new double [u.re_root.length];
	SortArrays.sortArrayWithLoad(np-1,root_re,root_im);
	int n_reroots=selectRealRoots(reroots,np-1,root_re,root_im,EPS_IM);
	if(n_reroots==np-1)
	{
	    // throw out two roots closest to pex
	    for(int t=0;t<2;t++)
	    {
		double rr=Double.POSITIVE_INFINITY;
		int ptr=-1;
		for(int m=0;m<n_reroots;m++)
		{
		    final double d=Math.abs(reroots[m]-pex);
		    if(d<=rr)
		    {
			rr=d;
			ptr=m;
		    }
		}
		if(ptr>=0)
		{
		    for(int m=ptr+1;m<n_reroots;m++)
		    {
			reroots[m-1]=reroots[m];
		    }
		    n_reroots--;
		}
	    }
	}
	else
	{
	    final double [] rer=new double [np];
	    int nrer=0;
	    for(int m=0;m<n_reroots;m++)
	    {
		double rr=reroots[m];
		if(Math.abs(rr-pex)>=EPS_IM)
		{ // skip the root=pex
		    rer[nrer++]=rr;
		}
	    }
	    return java.util.Arrays.copyOf(rer,nrer);
	}
	return java.util.Arrays.copyOf(reroots,n_reroots);
    }

    static double [] getmmAR(final double [] coefs,
			     final double [] norm,
			     final double [] extremums,
			     final boolean max_min)
    {
	// max_min: true- max, false-min
	final double pex,i;
	if(extremums.length<=1)
	{
	    pex=extremums.length==1?extremums[0]:Double.NaN;
	    i=getD0WithScale(coefs,pex,norm);
	}
	else
	{
	    // at most 2 attractors/repellers for supported np
	    final double i0=getD0WithScale(coefs,extremums[0],norm),
		i1=getD0WithScale(coefs,extremums[1],norm);
	    if(max_min?i0<i1:i0>i1)
	    {
		pex=extremums[0];
		i=i0;
	    }
	    else
	    {
		pex=extremums[1];
		i=i1;
	    }
	}
	double pmin=Double.NaN,pmax=Double.NaN;
	for(double p:getSRoots(coefs,norm,pex))
	{
	    final double d1=getD1WithScale(coefs,p,norm);
	    if((max_min?d1<0:d1>0) && p<pex)
	    {
		pmin=p;
	    }
	    if((max_min?d1>0:d1<0) && p>pex)
	    {
		pmax=p;
	    }
	}
	return new double []{
	    pmin,
	    pmax,
	    pex,
	    i};
    }


    double getIV(final double p)
    {
	return getD0WithScale(this.pcoefs,p-this.tipdata.p_offset,this.normformoments_pt);
    }

    static final int D_BUY=72813,D_SELL=72814,D_SUMM=72815,D_DIFF=72816;

    double getD0T(final double p,final int t)
    {
	if(t==D_BUY)
	{
	    return getD0WithScale(this.pcoefs_buy,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else if(t==D_SELL)
	{
	    return getD0WithScale(this.pcoefs_sell,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else if(t==D_SUMM)
	{
	    return getD0WithScale(this.pcoefs,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else if(t==D_DIFF)
	{
	    return getD0WithScale(this.pcoefs_diff,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else
	{
	    throw new RuntimeException("t="+t);
	}
    }

    double getD1T(final double p,final int t)
    {
	if(t==D_BUY)
	{
	    return this.sigma*getD1WithScale(this.pcoefs_buy,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else if(t==D_SELL)
	{
	    return this.sigma*getD1WithScale(this.pcoefs_sell,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else if(t==D_SUMM)
	{
	    return this.sigma*getD1WithScale(this.pcoefs,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else if(t==D_DIFF)
	{
	    return this.sigma*getD1WithScale(this.pcoefs_diff,p-this.tipdata.p_offset,this.normformoments_pt);
	}
	else
	{
	    throw new RuntimeException("t="+t);
	}
    }

    void setPriceMatrixAndRightPart(final double [][] matr,final double [] rpart,final double [] rpart_buy,final double [] rpart_sell,
				    final double [] moments,final double [] norm,
				    final double [] rmoments,final double [] rmoments_buy,final double [] rmoments_sell,
				    final int np)
    {
	final double lscale=moments[0];
	for(int l=0;l<np;l++)
	{
	    final double norml=norm[l];
	    for(int k=0;k<np;k++)
	    {
		matr[l][k]=(moments[l+k]/lscale)/norml/norm[k];
	    }
	    rpart[l]=(rmoments[l]/rmoments[0])/norml;
	    rpart_buy[l]=(rmoments_buy[l]/rmoments_buy[0])/norml;
	    rpart_sell[l]=(rmoments_sell[l]/rmoments_sell[0])/norml;
	}
    }


    double getDDirP(final double p)
    {
	return getD1WithScale(this.pcoefs,p-this.tipdata.p_offset,this.normformoments_pt)*this.sigma;
    }

    double getDDDirP(final double p)
    {
	return getD2WithScale(this.pcoefs,p-this.tipdata.p_offset,this.normformoments_pt)*this.sigma*this.sigma;
    }


    private static TradeIP calculateTestTradeIP(final int N_EM,
						final long currenttime,
						final long tau,
						final BookItem [] tb,
						final int tsize)
    {
	final boolean CALCULATE_VT=true;
	return new TradeIP(TradeIPData.calculateTestTradeIPData(N_EM,
								currenttime,
								tau,
								tb,
								tsize,
								CALCULATE_VT));
    }


    private static void checkMatching(final TradeIP r,final TradeIP ref)
    {
	TradeIPData.checkMatching(r.tipdata,ref.tipdata);

	final double eps=1e-5;

	UnitTests.checkDoubleRel(r.tipdata.p_offset,"r.tipdata.p_offset",r.tipdata.pt_sum/r.tipdata.t_sum,"r.tipdata.pt_sum/r.tipdata.t_sum",eps);
	UnitTests.checkDoubleRel(r.sigma,"r.sigma",Math.sqrt(r.tipdata.moments_pt[2]/r.tipdata.moments_pt[0]),"sqrt(r.tipdata.moments_pt[2]/r.tipdata.moments_pt[0])",eps);
	UnitTests.checkDoubleRel(r.skewness,"r.skewness",r.tipdata.moments_pt[3]/r.tipdata.moments_pt[0],"r.tipdata.moments_pt[3]/r.tipdata.moments_pt[0]",eps);
	UnitTests.checkDoubleRel(r.kurtosis,"r.kurtosis",r.tipdata.moments_pt[4]/r.tipdata.moments_pt[0],"r.tipdata.moments_pt[4]/r.tipdata.moments_pt[0]",eps);

	final double dirv_check=(r.dirv_sell*r.tipdata.moments_pv_sell[0]+r.dirv_buy*r.tipdata.moments_pv_buy[0])/r.tipdata.moments_pv[0],
	    diffw_dirv_check=(r.dirv_sell*r.tipdata.moments_pv_sell[0]-r.dirv_buy*r.tipdata.moments_pv_buy[0])/r.tipdata.moments_pv[0];
	UnitTests.checkDoubleRel(r.dirv,"r.dirv",dirv_check,"dirv_check",eps);
	UnitTests.checkDoubleRel(r.diffw_dirv,"r.diffw_dirv",diffw_dirv_check,"diffw_dirv_check",eps);
	final double dp_last=r.tipdata.trade_last.price-r.tipdata.p_offset;
	final double ddirp_check=r.sigma*(getD1WithScale(r.pcoefs_sell,dp_last,r.normformoments_pt)*r.tipdata.moments_pv_sell[0]+getD1WithScale(r.pcoefs_buy,dp_last,r.normformoments_pt)*r.tipdata.moments_pv_buy[0])/r.tipdata.moments_pv[0];
	UnitTests.checkDoubleRel(r.ddirp,"r.ddirp",ddirp_check,"ddirp_check",eps);

	final double stdev_t=Math.sqrt(ref.tipdata.moments_pt[2]/ref.tipdata.t_sum);
	double wn_t=1;
	for(int m=0;m<r.tipdata.np;m++)
	{
	    UnitTests.checkDoubleRel(r.normformoments_pt[m],"r.normformoments_pt[m]",wn_t,"stdev_t^"+m,eps);
	    wn_t*=stdev_t;
	}

	TradeIPData.checkPCoefs(r.pcoefs,r.tipdata.moments_pt,r.normformoments_pt,r.tipdata.moments_pv,eps);
	TradeIPData.checkPCoefs(r.pcoefs_buy,r.tipdata.moments_pt,r.normformoments_pt,r.tipdata.moments_pv_buy,eps);
	TradeIPData.checkPCoefs(r.pcoefs_sell,r.tipdata.moments_pt,r.normformoments_pt,r.tipdata.moments_pv_sell,eps);
	for(int i=0;i<r.root_re.length;i++)
	{
	    checkRoots(r.root_re[i],r.root_im[i],r.normformoments_pt,r.tipdata.pt_sum/r.tipdata.t_sum,r.tipdata.pv_sum/r.tipdata.v_sum,r,r.tipdata.moments_pt,
		       r.tipdata.moments_pv,r.tipdata.moments_pv_buy,r.tipdata.moments_pv_sell,eps);
	}

	System.err.println("TradeIP:checkMatching(): matching OK");
    }


    private static void checkRoots(final double re,
				   final double im,
				   final double [] norm,
				   final double pleft,
				   final double pright,
				   final TradeIP r,
				   final double [] m_left,
				   final double [] m_right,
				   final double [] m_right_buy,
				   final double [] m_right_sell,
				   final double eps)
    {
	final int np=r.pcoefs.length;
	final double [] pcoe=new double [np];
	for(int k=np;--k>=0;)
	{
	    pcoe[k]=r.pcoefs[k]/norm[k];
	}
	pcoe[0]-=1;
	//System.err.println("re="+re+" im="+im);
	final double nv=getPolRootNorm(pcoe,re,im);
	UnitTests.checkDoubleRel(nv,"nv",0,"0",eps);
	// check norm0
	UnitTests.checkDoubleRel(norm[0],"norm[0]",1,"1",eps);
	// check p_offset
	UnitTests.checkDoubleRel(r.tipdata.p_offset,"r.tipdata.p_offset",pleft,"pleft",eps);
	// check norm
	UnitTests.checkDoubleRel(norm[1],"norm[1]",Math.sqrt(m_left[2]/m_left[0]),"sqrt(m_left[2]/m_left[0])",eps);

	for(int k=0;k<r.vattractors.length;k++)
	{
	    final double d=r.vattractors[k]-r.tipdata.p_offset;
	    UnitTests.checkDoubleRel(getD1WithScale(r.pcoefs,d,norm),
			   "attractor:dI/dp(root)",0,"0",eps);
	    UnitTests.checkDouble(Math.signum(getD2WithScale(r.pcoefs,d,norm)),
			"signum(d^2I/dp^2(root))",-1,"-1");
	    UnitTests.checkDoubleArrayRel(getSRoots(r.pcoefs,norm,d),"A pex "+d,
				getSRootsTest(r.pcoefs,norm,d),"ATest pex "+d,
				eps);
	}
	for(int k=0;k<r.vrepellers.length;k++)
	{
	    final double d=r.vrepellers[k]-r.tipdata.p_offset;
	    UnitTests.checkDoubleRel(getD1WithScale(r.pcoefs,d,norm),
			   "repeller:dI/dp(root)",0,"0",eps);
	    UnitTests.checkDouble(Math.signum(getD2WithScale(r.pcoefs,d,norm)),
			"signum(d^2I/dp^2(root))",1,"1");

	    UnitTests.checkDoubleArrayRel(getSRoots(r.pcoefs,norm,d),"R pex"+d,
				getSRootsTest(r.pcoefs,norm,d),"RTest pex "+d,
				eps);
	}
	System.err.println("TradeIP:checkRoots: getSRoots & getSRootsTest matched OK for "+
			   (r.vattractors.length+r.vrepellers.length)+" points.");
	for(int k=0;k<r.vreal_roots.length;k++)
	{
	    UnitTests.checkDoubleRel(getD1WithScale(r.pcoefs,r.vreal_roots[k],norm),
			   "dI/dp(root)",0,"0",eps);
	}
	for(int k=0;k<r.interval_extremum.length;k++)
	{
	    UnitTests.checkDoubleRel(getD1WithScale(r.pcoefs_diff,r.interval_extremum[k],norm),
			   "dsI/dp(root)",0,"0",eps);
	}

	if(np==2)
	{ // check special case
	    final double a0=r.pcoefs[0]/norm[0],a1=r.pcoefs[1]/norm[1];
	    UnitTests.checkDoubleRel(a1*re+a0,"a1*re+a0",1,"1",eps);

	    final double pcalc=r.tipdata.p_offset+re;
	    UnitTests.checkDoubleRel(pcalc,"pcalc",r.tipdata.p_offset,"r.tipdata.p_offset",eps);

	    // check a0
	    UnitTests.checkDoubleRel(r.pcoefs[0],"r.pcoefs[0]",1,"1",eps);
	    // check a1
	    UnitTests.checkDoubleRel(r.pcoefs[1],"r.pcoefs[1]",r.dp/norm[1],"r.dp/norm[1]",eps);
	    // check dirv
	    final double p_last=r.tipdata.trade_last.price;
	    UnitTests.checkDoubleRel(r.dirv,"r.dirv",r.pcoefs[0]/norm[0]+r.pcoefs[1]*(p_last-r.tipdata.p_offset)/norm[1],"dirv check 1",eps);
	    UnitTests.checkDoubleRel(r.dirv,"r.dirv",1+(p_last-r.tipdata.p_offset)*(pright-pleft)/norm[1]/norm[1],"dirv check 2",eps);

	    UnitTests.checkDoubleRel(r.dirv_buy,"r.dirv_buy",1+(p_last-r.tipdata.p_offset)*(m_right_buy[1]/m_right_buy[0])/norm[1]/norm[1],"dirv_buy check",eps);
	    UnitTests.checkDoubleRel(r.dirv_sell,"r.dirv_sell",1+(p_last-r.tipdata.p_offset)*(m_right_sell[1]/m_right_sell[0])/norm[1]/norm[1],"dirv_sell check",eps);


	    // extra checks
	    UnitTests.checkDoubleRel(r.dirv,"r.dirv",r.dirv0,"r.dirv0",eps);
	    UnitTests.checkDoubleRel(r.dirp0,"r.dirp0",p_last-r.tipdata.p_offset,"p_last-r.tipdata.p_offset",eps);

	    UnitTests.checkDoubleRel(r.ddirp,"r.ddirp",r.ddirp0,"r.ddirp0",eps);
	    UnitTests.checkDoubleRel(r.ddirp,"r.ddirp",r.dp/r.sigma,"r.dp/r.sigma",eps);

	    UnitTests.checkDoubleRel(r.dddirp,"r.dddirp",0,"0",eps);
	}
	else if(np==3)
	{
	    final double s2=norm[1]*norm[1],nl=m_left[0],
		s3=m_left[3]/nl,s4=m_left[4]/nl,
		r1=r.dp,r2=m_right[2]/m_right[0];
	    final double x00=-s3/s2,x01=s2,x10=s2-s4/s2,x11=s3,rx1=r1-s3/s2,rx2=r2-s4/s2;
	    final double det=x00*x11-x10*x01,d0=rx1*x11-x01*rx2,d1=x00*rx2-x10*rx1;
	    final double a0=d0/det,a1=d1/det,a2=(1-a0)/s2;
	    UnitTests.checkDoubleRel(r.pcoefs[0]/norm[0],"r.pcoefs[0]/norm[0]",a0,"a0",eps);
	    UnitTests.checkDoubleRel(r.pcoefs[1]/norm[1],"r.pcoefs[1]/norm[1]",a1,"a1",eps);
	    UnitTests.checkDoubleRel(r.pcoefs[2]/norm[2],"r.pcoefs[2]/norm[2]",a2,"a2",eps);
	    final double dpl=r.tipdata.trade_last.price-r.tipdata.p_offset;
	    final double dirv_check=a0+dpl*(a1+dpl*a2);
	    UnitTests.checkDoubleRel(r.dirv,"r.dirv",dirv_check,"dirv_check",eps);
	    UnitTests.checkDoubleRel(r.ddirp,"r.ddirp",(a1+dpl*(a2*2))*r.sigma,"(a1+dpl*(a2*2))*sigma",eps);

	    UnitTests.checkDoubleRel(r.dddirp,"r.dddirp",2*a2*r.sigma*r.sigma,"2*a2*r.sigma*r.sigma",eps);

	    for(double [][] da : new double [][][]{
		    {r.tipdata.moments_pv,r.pcoefs,new double []{r.dirv}},
		    {r.tipdata.moments_pv_buy,r.pcoefs_buy,new double []{r.dirv_buy}},
		    {r.tipdata.moments_pv_sell,r.pcoefs_sell,new double []{r.dirv_sell}},
		})
	    {
		final double [] rp=da[0],coefs=da[1];
		final double dir=da[2][0];
		final double rr1=rp[1]/rp[0]-s3/s2,rr2=rp[2]/rp[0]-s4/s2;
		final double detp0=rr1*x11-x01*rr2,detp1=x00*rr2-x10*rr1;
		final double ap0=detp0/det,ap1=detp1/det,ap2=(1-ap0)/s2;
		UnitTests.checkDoubleRel(coefs[0]/norm[0],"coefs[0]/norm[0]",ap0,"ap0",eps);
		UnitTests.checkDoubleRel(coefs[1]/norm[1],"coefs[1]/norm[1]",ap1,"ap1",eps);
		UnitTests.checkDoubleRel(coefs[2]/norm[2],"coefs[2]/norm[2]",ap2,"ap2",eps);
		UnitTests.checkDoubleRel(dir,"dir",ap0+dpl*(ap1+dpl*ap2),"ap0+dpl*(ap1+dpl*ap2)",eps);
	    }


	    // extra checks
	    final double rp0=1,rp1=pright-pleft,sigmav2=m_right[2]/m_right[0]-rp1*rp1,rp2=sigmav2+(pright-pleft)*(pright-pleft);
	    UnitTests.checkDoubleRel(rp0,"rp0",1,"r0",eps);
	    UnitTests.checkDoubleRel(rp1,"rp1",r1,"r1",eps);
	    UnitTests.checkDoubleRel(rp2,"rp2",r2,"r2",eps);
	    final double sigmat2=m_left[2]/m_left[0],xi3=m_left[3]/m_left[0],xi4=m_left[4]/m_left[0];
	    final double a_det=sigmat2*(xi4-sigmat2*sigmat2)-xi3*xi3;
	    final double a1_det=rp1*(xi4-sigmat2*sigmat2-xi3*rp1)-(sigmav2-sigmat2)*xi3;
	    final double a2_det=rp1*(rp1*sigmat2-xi3)+(sigmav2-sigmat2)*sigmat2;
	    final double aa1=a1_det/a_det,aa2=a2_det/a_det,aa0=rp0-sigmat2*aa2;

	    UnitTests.checkDoubleRel(a0,"a0",aa0,"aa0",eps);
	    UnitTests.checkDoubleRel(a1,"a1",aa1,"aa1",eps);
	    UnitTests.checkDoubleRel(a2,"a2",aa2,"aa2",eps);
	}
	else if(np==4)
	{
	    final double r0=1,r1=r.dp,r2=m_right[2]/m_right[0],r3=m_right[3]/m_right[0],nl=m_left[0];
	    final double x11=norm[1]*norm[1],x12=m_left[3]/nl,x13=m_left[4]/nl,
		x21=m_left[3]/nl,x22=m_left[4]/nl-x11*x11,x23=m_left[5]/nl-x11*x12,
		x31=m_left[4]/nl,x32=m_left[5]/nl-x11*x12,x33=m_left[6]/nl-x12*x12,
		rx1=r1,rx2=r2-x11*r0,rx3=r3-x12*r0;
	    final double det=x11*(x22*x33-x32*x23)+x21*(x13*x32-x12*x33)+x31*(x12*x23-x22*x13);
	    final double a1=(rx1*(x22*x33-x32*x23)+rx2*(x13*x32-x12*x33)+rx3*(x12*x23-x22*x13))/det;
	    final double a2=(x11*(rx2*x33-rx3*x23)+x21*(x13*rx3-rx1*x33)+x31*(rx1*x23-rx2*x13))/det;
	    final double a3=(x11*(x22*rx3-x32*rx2)+x21*(rx1*x32-x12*rx3)+x31*(x12*rx2-x22*rx1))/det;
	    final double a0=r0-x11*a2-x12*a3;

	    UnitTests.checkDoubleRel(r.pcoefs[0]/norm[0],"r.pcoefs[0]/norm[0]",a0,"a0",eps);
	    UnitTests.checkDoubleRel(r.pcoefs[1]/norm[1],"r.pcoefs[1]/norm[1]",a1,"a1",eps);
	    UnitTests.checkDoubleRel(r.pcoefs[2]/norm[2],"r.pcoefs[2]/norm[2]",a2,"a2",eps);
	    UnitTests.checkDoubleRel(r.pcoefs[3]/norm[3],"r.pcoefs[3]/norm[3]",a3,"a3",eps);
	    final double dpl=r.tipdata.trade_last.price-r.tipdata.p_offset;
	    final double dirv_check=a0+dpl*(a1+dpl*(a2+dpl*a3));
	    UnitTests.checkDoubleRel(r.dirv,"r.dirv",dirv_check,"dirv_check",eps);
	    UnitTests.checkDoubleRel(r.ddirp,"r.ddirp",(a1+dpl*(a2*2+dpl*a3*3))*r.sigma,"(a1+dpl*(a2*2+dpl*a3*3))*sigma",eps);
	    UnitTests.checkDoubleRel(r.dddirp,"r.dddirp",(2*a2+6*a3*dpl)*r.sigma*r.sigma,"(2*a2+6*a3*dpl)*r.sigma*r.sigma",eps);

/*	    for(int k=0;k<r.p_up.length+r.p_down.length;k++)
	    {
		final double pI1=k<r.p_up.length?r.p_up[k]:r.p_down[k-r.p_up.length],
		             dpr=pI1-r.tipdata.p_offset;
		System.err.println("pI1="+pI1);
		UnitTests.checkDoubleRel(1,"1",a0+dpr*(a1+dpr*(a2+dpr*a3)),"a0+dpr*(a1+dpr*(a2+dpr*a3))",eps);
	    }
*/

	}
    }



    private static void matchWithTest(final TradeIP r,
				      final Trade [] tb,
				      final int tsize)
    {
	final TradeIP rt=calculateTestTradeIP(r.pcoefs.length,
					      r.tipdata.currenttime,
					      r.tipdata.tau,
					      tb,
					      tsize);
	checkMatching(r,rt);
	final TradeEMoments ref=calculateEMoments(r.tipdata.currenttime,
						  r.tipdata.tau,
						  tb,
						  tsize,
						  null);
	TradeEMoments.checkMatchingTradeIPData(ref,r.tipdata);
    }


    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println("checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);
	final TradeIP ro=calculateTradeIP(N_EM,
					  currenttime,
					  tau,
					  tb,
					  tsize,
					  null);
	matchWithTest(ro,tb,tsize);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeIP em_ref=calculateTradeIP(N_EM,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest(em_ref,tb,n_split);

	final TradeIP em_fromref=calculateTradeIP(N_EM,currenttime,tau,tb,tsize,em_ref);
	checkMatching(ro,em_fromref);
	matchWithTest(em_fromref,tb,tsize);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeIP emc_cut=calculateTradeIP(N_EM,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut);

	System.err.println("TradeIP::checkCalculations(): test random values OK N points="+tsize+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	System.err.println("TradeIP::main(): starting tests ; ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(7903L);
	for(int N_EM :new int []{3,4,N_EMOMENTS,2,3,4,5})
	{
	    for(int i=0;i<55;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

}


static class QQTradeIP{

    private final double [] P0,P1,P2;
    final double p_offset,sigma;

    final double [] p_imin,p_imax;

    final int neq;
    private final double [] p_real_roots;
    private final TradeIPData odata;

    private QQTradeIP(final com.polytechnik.utils.PolynomialRootsFinder rf,
		      final TradeIPData odata)
    {
	this.odata=odata;
	final int np=getNP(odata.np);
	this.neq=np+1;

	this.p_offset=odata.p_offset;

	this.sigma=Math.sqrt(odata.moments_pt[2]/odata.moments_pt[0]-Math.pow(odata.moments_pt[1]/odata.moments_pt[0],2));

	final double [][] x=getPCoefs(np);
	this.P0=x[0];
	this.P1=x[1];
	this.P2=x[2];

	final double [][] xx=QATradeIP.getPolynomialMinMax(2*np,this.P0,this.p_offset,rf);
	this.p_imin=xx[0];
	this.p_imax=xx[1];
	this.p_real_roots=xx[2];
    }

    int getNP(final int np0){return np0-1;}
    double [][] getPCoefs(final int np)
    {
	return getPNoCondititionCoefs(this.odata.moments_pt,this.odata.moments_pt_tau,this.odata.moments_pt_tau2,np,0);
    }

    /*private*/ static double [][] getPNoCondititionCoefs(final double [] pt,
						      final double [] pt_tau,
						      final double [] pt_tau2,
						      final int np,
						      final int nf)
    {
	final double pt0=pt[0],tau0=pt_tau[0]/pt0;
	final int ne=np+1;
	final double [] norms=TradeIPData.getPowersOfStdev(pt[2]/pt0,ne);
	final double [][] matr=new double [ne][ne];
	final int nf2=(nf/2)*2;
	final double sf=(nf==nf2?(pt[nf]/pt0):((pt[nf2]/pt0)*Math.sqrt(pt[2]/pt0))),sf2=sf*sf;
	for(int k=1;k<=np;k++)
	{
	    final double normsk=norms[k],pk=pt[k+nf]/pt0;
	    matr[k][0]=matr[0][k]=(((pt_tau[k+nf]/pt0)-tau0*pk)/normsk/norms[0])/sf;
	    for(int l=1;l<=np;l++)
	    {
		matr[k][l]=(((pt[l+k+2*nf]/pt0)-(pt[l+nf]/pt0)*pk)/normsk/norms[l])/sf2;
	    }
	}
	matr[0][0]=((pt_tau2[0]/pt0)-(tau0*tau0))/(norms[0]*norms[0]);

	final int [] ipvt=new int [ne];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa(matr,ne,ipvt);
	final double [] P0=new double [2*np+1],P1=new double [np+1],P2=new double [1];
	if(pinfo!=0)
	{
	    java.util.Arrays.fill(P0,Double.NaN);
	    java.util.Arrays.fill(P1,Double.NaN);
	    java.util.Arrays.fill(P2,Double.NaN);
	    return new double [][]{P0,P1,P2};
	}

	final double [] rpart=new double [ne];
	for(int k=ne;--k>=0;)
	{
	    java.util.Arrays.fill(rpart,0);
	    rpart[k]=1/norms[k];
	    com.polytechnik.utils.Linsystems.dgesl(matr,ne,ipvt,rpart,1);
	    final double c0=rpart[0]/norms[0];
	    if(k<=0)
	    {
		P0[0]+=c0*tau0*tau0;
		P1[0]-=2*c0*tau0;
		P2[0]+=c0;
		for(int l=np;l>=1;l--)
		{
		    final double c=rpart[l]/norms[l],plc=c*((pt[l+nf]/pt0)/sf);
		    P0[0]+=tau0*plc;
		    P0[l]-=tau0*c;
		    P1[0]-=plc;
		    P1[l]+=c;
		}
	    }
	    else
	    {
		final double pk=(pt[k+nf]/pt0)/sf;
		P0[0]+=tau0*(c0*pk);
		P0[k]-=tau0*c0;
		P1[0]-=c0*pk;
		P1[k]+=c0;
		for(int l=np;l>=1;l--)
		{
		    final double c=rpart[l]/norms[l],plc=c*((pt[l+nf]/pt0)/sf);
		    P0[0]+=plc*pk;
		    P0[k]-=plc;
		    P0[l]-=pk*c;
		    P0[k+l]+=c;
		}
	    }
	}

	return new double [][] {P0,P1,P2};
    }

    public static QQTradeIP calculateQQTradeIP(final int N_EM,
					       final long currenttime,
					       final long tau,
					       final BookItem [] tb,
					       final int tsize,
					       final com.polytechnik.utils.PolynomialRootsFinder rf,
					       final QQTradeIP ref)
    {
	final TradeIPData o=TradeIPData.calculateTradeIPData(N_EM+1,
							     currenttime,
							     tau,
							     tb,
							     tsize,
							     ref!=null?ref.odata:null,
							     true);
	return new QQTradeIP(rf,o);
    }


    double getDIDPDT_TOT(final double p)
    {
	return Polynomials.getD1(this.P0,p-this.p_offset);
    }
    double getDIDT(final double p)
    {
	return Polynomials.getD0(this.P0,p-this.p_offset);
    }

    double getPGMIN(final double p)
    {
	return QATradeIP.getPGMIN(p,this.p_imin,this.P0,this.p_offset);
    }

    double getPLMIN(final double p)
    {
	return QATradeIP.getPLMIN(p,this.p_real_roots,this.p_offset);
    }

    double getIF(final double p)
    {
	return (Polynomials.getD0(this.P0,p-this.p_offset)-1)/(this.neq-1);
    }

    double getIDP(final double p)
    {
	final double dp=p-this.p_offset;
	return (Polynomials.getD0(this.P0,dp)-1)/Polynomials.getD1(this.P0,dp);
    }


    private static void checkDP(final java.util.Random r,final QQTradeIP ro,final double eps,final Trade [] tb,final int tsize)
    {

	if(true)
	{ // check 0-th moments estimate
	    System.err.println("ro.odata.moments_pt[0]="+ro.odata.moments_pt[0]+" (should be 1)");
	    System.err.println("ro.odata.moments_pt_tau[0]="+ro.odata.moments_pt_tau[0]+" (should be -1)");
	    System.err.println("ro.odata.moments_pt_tau2[0]="+ro.odata.moments_pt_tau2[0]+" (should be 2)");
	}


	if(true)
	{
	    final double pt0=ro.odata.moments_pt[0];
	    final double [][] pcoefs=new double [][]{ro.P0,ro.P1,ro.P2},
		mom=new double [][]{ro.odata.moments_pt,ro.odata.moments_pt_tau,ro.odata.moments_pt_tau2};

	    double s=0;
	    for(int j=mom.length;--j>=0;)
	    {
		final double [] m=mom[j],c=pcoefs[j];
		for(int k=c.length;--k>=0;)
		{
		    s+=c[k]*(m[k]/pt0);
		}
	    }
	    //System.err.println("dIaver="+s);
	    UnitTests.checkDoubleRel(s,"dIaver",ro.neq,"ro.neq",eps);
	}

	if(true)
	{
	    final double pt0=ro.odata.moments_pt[0],tau0=ro.odata.moments_pt_tau[0]/pt0;
	    final double [] p0coefs=new double [ro.P0.length];
	    double w=1;
	    for(double [] pcoefs : new double [][]{ro.P0,ro.P1,ro.P2})
	    {
		for(int k=pcoefs.length;--k>=0;)
		{
		    p0coefs[k]+=w*pcoefs[k];
		}
		w*=tau0;
	    }
	    final double [] p1coefs=new double [ro.P1.length];
	    for(int k=ro.P1.length;--k>=0;)
	    {
		p0coefs[k]-=0.5*ro.P1[k]*tau0;
		p1coefs[k]=0.5*ro.P1[k];
	    }

	    final double [][] pcoefs=new double [][]{p0coefs,p1coefs},
		mom=new double [][]{ro.odata.moments_pt,ro.odata.moments_pt_tau};
	    double s=0;
	    for(int j=mom.length;--j>=0;)
	    {
		final double [] m=mom[j],c=pcoefs[j];
		for(int k=c.length;--k>=0;)
		{
		    s+=c[k]*(m[k]/pt0);
		}
	    }
	    //System.err.println("dIaver-1="+s);
	    UnitTests.checkDoubleRel(s,"dIaver-1",ro.neq-1,"ro.neq-1",eps);
	}
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("QQTradeIP:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);

	final QQTradeIP ro=calculateQQTradeIP(N_EM,
					      currenttime,
					      tau,
					      tb,
					      tsize,
					      new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG(),
					      null);
	checkDP(r,ro,eps,tb,tsize);
	// more tests to add

	System.err.println("QQTradeIP:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{1,2,3,4,5,6,7,1,2,3,4,5,6,7,9,11})
	{
	    for(int i=0;i<100;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QQTradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

} // END OF class QQTradeIP




static TradeIP calculateTradeIP(final int N_EM,
					final long currenttime,
					final long tau,
					final BookItem [] tb,
					final int tsize,
					final TradeIP ref)
{
    final boolean CALCULATE_VT=true;
    return new TradeIP(TradeIPData.calculateTradeIPData(N_EM,
							currenttime,
							tau,
							tb,
							tsize,
							ref!=null?ref.tipdata:null,
							CALCULATE_VT));
}





static class QATradeIP {

    /*private */final double [] pcoefs,didv;
    final double p_offset,sigma;

    final double [] p_imin,p_imax;

    final int neq;
    private final double [] p_real_roots;
    /*private*/ final TradeIPData odata;

    private QATradeIP(final double [] moments_left,final double [] moments_right,
		      final com.polytechnik.utils.PolynomialRootsFinder rf,
		      final TradeIPData odata)
    {
	this.odata=odata;
	final int np=this.neq=getNP(odata.np);

	this.p_offset=odata.p_offset;

	this.sigma=Math.sqrt(moments_left[2]/moments_left[0]-Math.pow(moments_left[1]/moments_left[0],2));

	final double [][] x=getPCoefs(moments_left,moments_right,np);
	this.pcoefs=x[0];
	this.didv=x[1];

	if(this.pcoefs.length!=np+1||this.didv.length!=2*np+1)
	{
	    throw new RuntimeException("np="+np+" pcoefs.length="+this.pcoefs.length+" didv.length="+this.didv.length);
	}

	final double [][] xx=getPolynomialMinMax(2*np,this.didv,this.p_offset,rf);
	this.p_imin=xx[0];
	this.p_imax=xx[1];
	this.p_real_roots=xx[2];
    }

    int getNP(final int np0){return np0-1;}
    static double [][] getPCoefs(final double [] pt,final double [] pv,final int np)
    {
	return getPNoCondititionCoefs(pt,pv,np,0);
    }

    static double [][] getPNoCondititionCoefs(final double [] pt,final double [] pv,final int np,final int nf)
    {
	final double pt0=pt[0];
	final double [] norms=TradeIPData.getPowersOfStdev(pt[2]/pt0,np);
	final double [] matr=new double [np*np];
	final double [] ip=new double [np+1];
	for(int k=0,j=0;k<np;k++)
	{
	    final int k1=k+1;
	    final double normsk=norms[k];
	    for(int l=0;l<np;l++,j++)
	    {
		final int l1=l+1;
		matr[j]=((pt[l1+k1+2*nf]/pt0)-(pt[l1+nf]/pt0)*(pt[k1+nf]/pt0))/normsk/norms[l];
	    }
	    ip[k]=((pv[k1]/pv[0])-(pt[k1]/pt0))/normsk;
	}
	final int [] ipvt=new int [np];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,np,np,ipvt);
	final double [] didv=new double [2*np+1];
	if(pinfo!=0)
	{
	    java.util.Arrays.fill(ip,Double.NaN);
	    java.util.Arrays.fill(didv,Double.NaN);
	    return new double [][]{ip,didv};
	}
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,np,np,ipvt,ip,1);

	final int nf2=(nf/2)*2;
	final double sf=(nf==nf2?(pt[nf]/pt0):((pt[nf2]/pt0)*Math.sqrt(pt[2]/pt0))),sf2=sf*sf;
	double sv0=0;
	final double [] rpart=new double [np];
	for(int k=np;--k>=0;)
	{
	    final int k1=k+1;
	    final double dk=pt[k1+nf]/pt0,normsk=norms[k];
	    ip[k1]=ip[k]/normsk;
	    sv0+=ip[k1]*dk;
	    rpart[k]=1;
	    com.polytechnik.utils.Linsystems.dgesl_m1(matr,np,np,ipvt,rpart,1);
	    for(int m=np;m>=1;m--)
	    {
		final double c=rpart[m-1]/norms[m-1]/normsk,dmc=c*(pt[m+nf]/pt0);
		rpart[m-1]=0;
		didv[0]+=dmc*dk;
		didv[k1]-=dmc*sf;
		didv[m]-=dk*c*sf;
		didv[k1+m]+=c*sf2;
	    }
	}
	ip[0]=-sv0;

	return new double [][] {
	    ip,
	    didv
	};
    }


    public static QATradeIP calculateQATradeIP(final int N_EM,
					       final long currenttime,
					       final long tau,
					       final BookItem [] tb,
					       final int tsize,
					       final com.polytechnik.utils.PolynomialRootsFinder rf,
					       final QATradeIP ref)
    {
	// dv/dt
	final TradeIPData o=TradeIPData.calculateTradeIPData(N_EM+1,
							     currenttime,
							     tau,
							     tb,
							     tsize,
							     ref!=null?ref.odata:null,
							     true);
	return new QATradeIP(o.moments_pt,o.moments_pv,rf,o);
    }
    public static QATradeIP calculateQATradeIP_V(final int N_EM,
						 final long currenttime,
						 final long tau,
						 final BookItem [] tb,
						 final int tsize,
						 final com.polytechnik.utils.PolynomialRootsFinder rf,
						 final QATradeIP ref)
    {
	// dt/dv
	final TradeIPData o=TradeIPData.calculateTradeIPData(N_EM+1,
							     currenttime,
							     tau,
							     tb,
							     tsize,
							     ref!=null?ref.odata:null,
							     false);
	return new QATradeIP(o.moments_pv,o.moments_pt,rf,o);
    }


    /** For debugging. */
    double getDIDP(final double p)
    {
	return Polynomials.getD1(this.pcoefs,p-this.p_offset);
    }
    double getAI0(final double p)
    {
	return Polynomials.getD0(this.pcoefs,p-this.p_offset);
    }

    double getDIDPDT_TOT(final double p)
    {
	return Polynomials.getD1(this.didv,p-this.p_offset);
    }
    double getDIDT(final double p)
    {
	return Polynomials.getD0(this.didv,p-this.p_offset);
    }

    double getIF(final double p)
    {
	return (Polynomials.getD0(this.didv,p-this.p_offset)-this.neq)/this.neq;
    }

    double getIDP(final double p)
    {
	final double dp=p-this.p_offset;
	return Polynomials.getD0(this.didv,dp)/Polynomials.getD1(this.didv,dp);
    }



    double getPGMIN(final double p)
    {
	return getPGMIN(p,this.p_imin,this.didv,this.p_offset);
    }

    /*private*/ static double getPGMIN(final double p,final double [] pmins,
				   final double [] pcoefs,final double offset)
    {
	double pm=Double.NaN,im=Double.POSITIVE_INFINITY;
	for(double px:pmins)
	{
	    final double ix=Polynomials.getD0(pcoefs,px-offset);
	    if(ix<=im)
	    {
		im=ix;
		pm=px;
	    }
	}
	return pm;
    }
    double getPLMIN(final double p)
    {
	return getPLMIN(p,this.p_real_roots,this.p_offset);
    }

    /*private*/ static double getPLMIN(final double p,final double [] real_roots_no_offset,final double offset)
    {
	final double dp=p-offset;
	for(int ne=real_roots_no_offset.length,i=0;i<=ne;i++)
	{
	    if((i<=0||real_roots_no_offset[i-1]<=dp) && (i>=ne||dp<real_roots_no_offset[i]))
	    {
		final int ptr=(i/2)*2; // mins are on edges
		return ptr<ne?real_roots_no_offset[ptr]+offset:Double.NaN;
	    }
	}
	return Double.NaN;
    }

    private static double getPLMAX(final double p,final double [] real_roots_no_offset,final double offset)
    {
	final double dp=p-offset;
	for(int ne=real_roots_no_offset.length,i=0;i<=ne;i++)
	{
	    if((i<=0||real_roots_no_offset[i-1]<=dp) && (i>=ne||dp<real_roots_no_offset[i]))
	    {
		final int ptr=(i/2)*2==i?i-1:i; // max are in the middle
		return ptr<ne && ptr>=0?real_roots_no_offset[ptr]+offset:Double.NaN;
	    }
	}
	return Double.NaN;
    }


    /*private*/ static double [][] getPolynomialMinMax(final int porder,
						       final double [] pcoefs,
						       final double offset,
						       final com.polytechnik.utils.PolynomialRootsFinder rf)
    {
	final double [] pcoe=new double [porder]; // temporary arrays

	// use scale=1 expected by PolynomialRoot
	for(int k=porder;k>=1;k--)
	{
	    pcoe[k-1]=k*pcoefs[k];
	}

	final PolynomialRootsFinder.PRoots u;
	if(porder-1>4){
	    if(rf==null) {
		u=new PolynomialRootsFinder.PRoots(0,pcoe,new double [0],new double [0]);
	    } else {
		u=rf.findRoots(pcoe);
	    }
	} else {
	    u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	}
	final double [] root_re=u.re_root,root_im=u.im_root;


	// insertion sort by real part with extra load
	SortArrays.sortArrayWithLoad(u.nroots,root_re,root_im);

	int n_reroots=0,n_p_imin=0,n_p_imax=0;
	double rr=Double.NaN;
	final double [] reroots=pcoe, // re-use as temporary arrays
	    save_p_imin=root_re,save_p_imax=root_im;
	for(int m=0;m<u.nroots;m++)
	{
	    if(rr>root_re[m])
	    {
		throw new RuntimeException("Not sorted");
	    }
	    rr=root_re[m];
	    if(Math.abs(root_im[m])<=EPS_IM)
	    {
		if(Polynomials.getD2(pcoefs,rr)>0)
		{ // min
		    save_p_imin[n_p_imin++]=rr+offset;
		}
		else
		{ // max
		    save_p_imax[n_p_imax++]=rr+offset;
		}
		reroots[n_reroots++]=rr;
	    }
	}
	return new double [][]{
	    java.util.Arrays.copyOf(save_p_imin,n_p_imin),
	    java.util.Arrays.copyOf(save_p_imax,n_p_imax),
	    java.util.Arrays.copyOf(reroots,n_reroots)
	};
    }

    private static void checkDP(final java.util.Random r,final QATradeIP ro,final double eps,final Trade [] tb,final int tsize)
    {
	if(true)
	{
	    final double p=TradeEMoments.getRandomPrice(r),dp=p-ro.p_offset;
	    final double [] pv=new double [1+ro.pcoefs.length];
	    double w=1;
	    for(int k=0;k<pv.length;k++)
	    {
		pv[k]=w;
		w*=dp;
	    }
	    final QTradeIP qa=new QTradeIP(ro.odata.moments_pt,pv,ro.odata);

	    // cross check
	    UnitTests.checkDoubleRel(ro.getDIDT(p),"ro.getDIDT(p)",qa.getAI0(p),"qa.getAI0(p)",eps);
	    UnitTests.checkDoubleRel(ro.getDIDPDT_TOT(p),"ro.getDIDPDT_TOT(p)",2*qa.getDIDP(p),"2*qa.getDIDP(p)",eps);
	}

	final QTradeIP qq=new QTradeIP(ro.odata.moments_pt,ro.odata.moments_pv,ro.odata);
	for(int n=0;n<100;n++)
	{
	    final double p=TradeEMoments.getRandomPrice(r);
	    // regular check
	    UnitTests.checkDoubleRel(ro.getDIDP(p),"ro.getDIDP(p)",qq.getDIDP(p),"qq.getDIDP(p)",eps);
	    UnitTests.checkDoubleRel(ro.getAI0(p),"ro.getAI0(p)",qq.getAI0(p),"qq.getAI0(p)",eps);
	    UnitTests.checkDoubleRel(ro.getPLMIN(p),"ro.getPLMIN(p)",ro.test_getPLMIN(p),"ro.test_getPLMIN(p)",eps);
	}

	if(true)
	{
	    final double pt0=ro.odata.moments_pt[0];
	    double s=0;
	    for(int k=ro.didv.length;--k>=0;)
	    {
		s+=ro.didv[k]*(ro.odata.moments_pt[k]/pt0);
	    }
	    System.err.println("dIaver="+s+" neq="+ro.neq);
	    UnitTests.checkDoubleRel(s,"dIaver",ro.neq,"ro.neq",eps);
	}
    }
    private double test_getPLMIN(final double p)
    {
	final double dp_init=p-this.p_offset;
	int ptr_close_to_init=-1;
	double drr=Double.POSITIVE_INFINITY;
	for(int m=this.p_real_roots.length;--m>=0;)
	{
	    final double rr=this.p_real_roots[m],drr1=Math.abs(dp_init-rr);
	    if(drr1<drr)
	    {
		drr=drr1;
		ptr_close_to_init=m;
	    }
	}
	if(ptr_close_to_init<0)
	{
	    return Double.NaN;
	}
	final double rr=this.p_real_roots[ptr_close_to_init];
	if(Polynomials.getD2(this.didv,rr)>0)
	{
	    return rr+this.p_offset;
	}
	if(rr>dp_init)
	{
	    return ptr_close_to_init>0?this.p_real_roots[ptr_close_to_init-1]+this.p_offset:Double.NaN;
	}
	return ptr_close_to_init<this.p_real_roots.length-1?this.p_real_roots[ptr_close_to_init+1]+this.p_offset:Double.NaN;
    }


    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("QATradeIP:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);

	final QATradeIP ro=calculateQATradeIP(N_EM,
					      currenttime,
					      tau,
					      tb,
					      tsize,
					      null,//new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG(2*N_EM-1),
					      null);
	checkDP(r,ro,eps,tb,tsize);
	// more tests to add

	System.err.println("QATradeIP:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{1,2,3,4,5,6,7,1,2,3,4,5,6,7,9,11})
	{
	    for(int i=0;i<100;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QATradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

} // END OF class QATradeIP



/** The like QATradeIP but with QVM projection. */
static class QVTradeIP {

    private final double [] pcoefs;
    final double sigma,pa_offset;
    final double [] p_imin,p_imax;

    private final double i0,dIdIestimate;
    final int neq;

    private final double [] p_real_roots;
    private final TradeIPData odata;


private QVTradeIP(final double [] moments_left,final double [] moments_right,
		  final com.polytechnik.utils.PolynomialRootsFinder rf,
		  final int neq,TradeIPData odata)
    {
	this.odata=odata;
	this.neq=neq;

	final double pt0=moments_left[0];
	this.pa_offset=odata.p_offset;
	this.sigma=Math.sqrt(moments_left[2]/pt0-Math.pow(moments_left[1]/pt0,2));

	this.dIdIestimate=getdIdIestimate(odata);
	this.pcoefs=getPCoefs(this.dIdIestimate,neq,moments_left,moments_right,0);

	double si0=0;
	for(int m=this.pcoefs.length;--m>=0;){si0+=this.pcoefs[m]*(moments_left[m]/pt0);}
	this.i0=si0;

	final double [][] xx=CalculateKendall.QATradeIP.getPolynomialMinMax(this.pcoefs.length-1,this.pcoefs,this.odata.p_offset,rf);
	this.p_imin=xx[0];
	this.p_imax=xx[1];
	this.p_real_roots=xx[2];
    }

    private static double getdIdIestimate(final TradeIPData odata)
    {
	return 2.11; // FIXME  // think later, factor is not correct
    }

    private static double [] getPCoefs(final double dIdIestimate,final int neq,final double [] pt, final double [] pv,final double dI)
    {
	final int ne=neq+1;
	final double pt0=pt[0],pv0=pv[0];
	final double [] norms=TradeIPData.getPowersOfStdev(pt[2]/pt0,ne);
	final double [] matr=new double [ne*ne];
	for(int k=1;k<ne;k++)
	{
	    final int ok=k*ne;
	    final double normsk=norms[k],pk=pt[k]/pt0,d=pv[k]/pv0-pk;
	    matr[ok]=matr[k]=d/normsk/norms[0];
	    for(int l=1;l<ne;l++)
	    {
		matr[ok+l]=((pt[l+k]/pt0)-(pt[l]/pt0)*pk)/normsk/norms[l];
	    }
	}
	matr[0]=dIdIestimate/norms[0]/norms[0];

	final int [] ipvt=new int [ne];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,ne,ne,ipvt);
	final double [] pcoefs=new double [2*neq+1];
	if(pinfo!=0)
	{
	    java.util.Arrays.fill(pcoefs,Double.NaN);
	    return pcoefs;
	}

	final double [] rpart=new double [ne];
	for(int k=ne;--k>=0;)
	{
	    rpart[k]=1;
	    com.polytechnik.utils.Linsystems.dgesl_m1(matr,ne,ne,ipvt,rpart,1);
	    final double normsk=norms[k],c0=rpart[0]/norms[0]/normsk;
	    rpart[0]=0;
	    if(k<=0)
	    {
		pcoefs[0]+=c0*dI*dI;
		for(int l=ne;--l>=1;)
		{
		    final double c=rpart[l]/normsk/norms[l],plc=c*(pt[l]/pt0);
		    rpart[l]=0;
		    pcoefs[0]-=dI*plc;
		    pcoefs[l]+=dI*c;
		}
	    }
	    else
	    {
		final double pk=pt[k]/pt0;
		pcoefs[0]-=dI*(c0*pk);
		pcoefs[k]+=dI*c0;
		for(int l=ne;--l>=1;)
		{
		    final double c=rpart[l]/normsk/norms[l],plc=c*(pt[l]/pt0);
		    rpart[l]=0;
		    pcoefs[0]+=plc*pk;
		    pcoefs[k]-=plc;
		    pcoefs[l]-=pk*c;
		    pcoefs[k+l]+=c;
		}
	    }
	}
	return pcoefs;
    }



    public static QVTradeIP calculateQVTradeIP(final int neq,
					       final long currenttime,
					       final long tau,
					       final BookItem [] tb,
					       final int tsize,
					       final com.polytechnik.utils.PolynomialRootsFinder rf,
					       final QVTradeIP ref)
    {
	final TradeIPData o=TradeIPData.calculateTradeIPData(neq,
							     currenttime,
							     tau,
							     tb,
							     tsize,
							     ref!=null?ref.odata:null,
							     true);
	return new QVTradeIP(o.moments_pt,o.moments_pv,rf,neq,o);
    }

    private static void checkIProj(final TradeIPData odata,final int neq,final double [] pt, final double [] pv,final double eps)
    {
	final double dIdIEstimate=getdIdIestimate(odata);
	final double [] p0=getPCoefs(dIdIEstimate,neq,pt,pv,0),pp1=getPCoefs(dIdIEstimate,neq,pt,pv,1),pm1=getPCoefs(dIdIEstimate,neq,pt,pv,-1);

	final double [] p1=new double [2*neq+1],p2=new double [2*neq+1];
	for(int k=0;k<p0.length;k++)
	{
	    p1[k]=0.5*(pp1[k]-pm1[k]);
	    p2[k]=0.5*(pp1[k]+pm1[k])-p0[k];
	}

	final double pt0=pt[0],pv0=pv[0];
	double s=0;
	for(int k=p0.length;--k>=0;)
	{
	    s+=p0[k]*(pt[k]/pt0);
	    s+=p1[k]*(pv[k]/pv0-pt[k]/pt0);
	}
	s+=dIdIEstimate*p2[0];

	System.err.println("dIaver="+s+" neq="+neq+" ddIdIEstimate="+dIdIEstimate);
	UnitTests.checkDoubleRel(s,"dIaver",neq+1,"neq+1",eps);

	System.err.println("QVTradeIP:checkIProj: OK");
    }


    double getDIDPDT_TOT(final double p)
    {
	return Polynomials.getD1(this.pcoefs,p-this.odata.p_offset);
    }
    double getDIDT(final double p)
    {
	return Polynomials.getD0(this.pcoefs,p-this.odata.p_offset);
    }

    double getPGMIN(final double p)
    {
	return CalculateKendall.QATradeIP.getPGMIN(p,this.p_imin,this.pcoefs,this.odata.p_offset);
    }

    double getPLMIN(final double p)
    {
	return CalculateKendall.QATradeIP.getPLMIN(p,this.p_real_roots,this.odata.p_offset);
    }

    double getIF(final double p)
    {
	return (Polynomials.getD0(this.pcoefs,p-this.odata.p_offset)-this.i0)/this.i0;
    }

    double getIDP(final double p)
    {
	final double dp=p-this.odata.p_offset;
	return Polynomials.getD0(this.pcoefs,dp)/Polynomials.getD1(this.pcoefs,dp);
    }



    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("QATradeIP:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);

	final QVTradeIP ro=calculateQVTradeIP(N_EM,
					      currenttime,
					      tau,
					      tb,
					      tsize,
					      null,//new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG(2*N_EM-1),
					      null);
	checkIProj(ro.odata,ro.neq,ro.odata.moments_pt,ro.odata.moments_pv,eps);
	// more tests to add

	System.err.println("QVTradeIP:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);
    }


   public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{1,2,3,4,5,6,7,1,2,3,4,5,6,7,9,11})
	{
	    for(int i=0;i<100;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QVTradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

} // END of class QVTradeIP





static class QTradeIP {

    private final double [] normformoments; // typically [m]=<(p-offset)^2>_t^{m/2}

    private final double [] pcoefs;
    final double p_offset,sigma,ia0,dia0;

    final double [] p_up,p_down,dp_up,dp_down;;
    private final double [] p_real_roots;

    final PTVV ptopology;
    /* private */ final TradeIPData odata;

    private QTradeIP(final double [] moments_left,final double [] moments_right,final TradeIPData odata)
    {
	this.odata=odata;
	final int np=getNP(odata.np);
	if(np<1)
	{
	    throw new RuntimeException("np="+np+" is not supported");
	}
	final double p_offset=this.p_offset=odata.p_offset;

	this.normformoments=getNorms(moments_left,np);
	this.sigma=np>1?this.normformoments[1]:Double.NaN;
	this.pcoefs=getPCoefs(moments_left,this.normformoments,moments_right,np);
	if(this.pcoefs.length!=np)
	{
	    throw new RuntimeException("np="+np+" pcoefs.length="+this.pcoefs.length);
	}

	if(true)
	{
	    final double [] pcoe=new double [np]; // temporary arrays
	    final double [] norms=this.normformoments;
	    double s=0,sd=0;

	    // use scale=1 expected by PolynomialRoot
	    for(int k=0;k<np;k++)
	    {
		pcoe[k]=this.pcoefs[k]/norms[k];
		s+=(pcoe[k]/(k+1))*(moments_left[k+1]/moments_left[0]);
		sd+=pcoe[k]*(moments_left[k]/moments_left[0]);
	    }
	    this.ia0=s;
	    this.dia0=sd;

	    // equation 0=sum_k pcoe_k * (p-p_offset)^k
	    final PolynomialRootsFinder.PRoots u;
	    if(np-1>4){
		u=new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG().findRoots(pcoe);
	    }else{
		u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    }
	    final double [] root_re=u.re_root,root_im=u.im_root;


	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(u.nroots,root_re,root_im);

	    int n_reroots=0,n_p_down=0,n_p_up=0;
	    double rr=Double.NaN;
	    final double [] reroots=pcoe, // re-use as temporary arrays
		save_p_up=root_re,save_p_down=root_im;
	    for(int m=0;m<u.nroots;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{
		    if(getD1WithScale(this.pcoefs,rr,norms)<=0)
		    { // down
			save_p_down[n_p_down++]=rr;
		    }
		    else
		    { // up
			save_p_up[n_p_up++]=rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.p_up=java.util.Arrays.copyOf(save_p_up,n_p_up);
	    this.p_down=java.util.Arrays.copyOf(save_p_down,n_p_down);
	    this.p_real_roots=java.util.Arrays.copyOf(reroots,n_reroots);
	}


	if(true)
	{
	    final double [] pcoe=new double [np+1];
	    final double [] norms=this.normformoments;

	    // use scale=1 expected by PolynomialRoot
	    for(int k=1;k<=np;k++)
	    {
		pcoe[k]=(this.pcoefs[k-1]/norms[k-1])/k;
	    }
	    pcoe[0]=-this.ia0;

	    final PolynomialRootsFinder.PRoots u;
	    if(np>4){
		u=new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG().findRoots(pcoe);
	    }else{
		u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    }
	    final double [] root_re=u.re_root,root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(u.nroots,root_re,root_im);

	    int n_reroots=0,n_dp_down=0,n_dp_up=0;
	    double rr=Double.NaN;
	    final double [] reroots=pcoe, // re-use as temporary arrays
		save_p_up=root_re,save_p_down=root_im;
	    for(int m=0;m<u.nroots;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{
		    if(getD0WithScale(this.pcoefs,rr,norms)<=0)
		    { // down
			save_p_down[n_dp_down++]=rr+p_offset;
		    }
		    else
		    { // up
			save_p_up[n_dp_up++]=rr+p_offset;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.dp_up=java.util.Arrays.copyOf(save_p_up,n_dp_up);
	    this.dp_down=java.util.Arrays.copyOf(save_p_down,n_dp_down);
	}


	if(true)
	{ // calculate p_i_eq_imax_{low,high} and shift offset to 0

	    final double [] pcoe=new double [np+1],reslh=new double [np+1],norms=new double [np+1]; // re-use as temporary arrays
	    for(int k=1;k<=np;k++)
	    {
		pcoe[k]=this.pcoefs[k-1]/k;
		norms[k]=this.normformoments[k-1];
	    }
	    norms[0]=norms[1];
	    pcoe[0]=-this.ia0*norms[0];


	    final double [] i_p_down=new double [this.p_down.length],
		p_i_eq_imax_low=new double [this.p_down.length],
		p_i_eq_imax_high=new double [this.p_down.length];
	    for(int k=this.p_down.length;--k>=0;)
	    {
		TradeIP.getPointsCloseToEXmmAR(reslh,pcoe,norms,this.p_down[k]);
		p_i_eq_imax_low[k]=reslh[0]+p_offset;
		p_i_eq_imax_high[k]=reslh[1]+p_offset;
		i_p_down[k]=getD0WithScale(pcoe,this.p_down[k],norms);
		this.p_down[k]+=p_offset; // shift offset to 0
	    }
	    final double [] i_p_up=new double [this.p_up.length],
		p_i_eq_imin_low=new double [this.p_up.length],
		p_i_eq_imin_high=new double [this.p_up.length];
	    for(int k=this.p_up.length;--k>=0;)
	    {
		TradeIP.getPointsCloseToEXmmAR(reslh,pcoe,norms,this.p_up[k]);
		p_i_eq_imin_low[k]=reslh[0]+p_offset;
		p_i_eq_imin_high[k]=reslh[1]+p_offset;
		i_p_up[k]=getD0WithScale(pcoe,this.p_up[k],norms);
		this.p_up[k]+=p_offset; // shift offset to 0
	    }

	    this.ptopology=new PTVV(getD1WithScale(pcoe,0,norms),p_offset,this.sigma,
				    this.p_up,i_p_up,p_i_eq_imin_low,p_i_eq_imin_high,
				    this.p_down,i_p_down,p_i_eq_imax_low,p_i_eq_imax_high);
	}

    }

    int getNP(final int np0){return np0-1;}

    double [] getNorms(final double [] moments_left,final int np){
	return TradeIPData.getPowersOfStdev(moments_left[2]/moments_left[0],np);
    }
    double [] getPCoefs(final double [] pt,final double [] norms,final double [] pv,final int np)
    {
	return getPNoCondititionCoefs(pt,norms,pv,np);
    }

    private static double [] getPNoCondititionCoefs(final double [] pt,final double [] norms,final double [] pv,final int np)
    {
	final double [][] matr=new double [np][np];
	final double [] rpart=new double [np];
	final double pt0=pt[0];
	for(int k=0;k<np;k++)
	{
	    final int k1=k+1;
	    final double normsk=norms[k];
	    for(int l=0;l<np;l++)
	    {
		final int l1=l+1;
		matr[k][l]=((pt[l1+k1]/pt0)-(pt[l1]/pt0)*(pt[k1]/pt0))/normsk/norms[l];
	    }
	    //rpart[k]=((pv[k1]/pt0)-((pv[0]/pt0)*(pt[k1]/pt0)))/norms[k];
	    // normalize right part, overload the getPCoefs
	    // if non-normalized right part is needed.
	    rpart[k]=((pv[k1]/pv[0])-(pt[k1]/pt0))/normsk;
	}
	TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
	for(int k=0;k<np;k++)
	{
	    rpart[k]*=k+1;
	}
	return rpart;
    }


    public static QTradeIP [] calculateQTradeIP(final int N_EM,
					     final long currenttime,
					     final long tau,
					     final BookItem [] tb,
					     final int tsize,
					     final TradeIPData ref)
    {
	final TradeIPData o=TradeIPData.calculateTradeIPData(N_EM+1,
							     currenttime,
							     tau,
							     tb,
							     tsize,
							     ref /* ref!=null?ref.odata:null */,
							     true);
	final double [] pv_diff=new double [o.np];
	for(int k=pv_diff.length;--k>=0;)
	{
	    pv_diff[k]=o.moments_pv_sell[k]-o.moments_pv_buy[k];
	}
	return new QTradeIP []{
	    new QTradeIP(o.moments_pt,o.moments_pv,o),
	    new QTradeIP(o.moments_pt_buy,o.moments_pv_buy,o),
	    new QTradeIP(o.moments_pt_sell,o.moments_pv_sell,o),
	    new QTradeIP(o.moments_pt,pv_diff,o) // FIXME to remove
	};
    }


    /** For debugging. */
    double getDIDP(final double p)
    {
	return getD0WithScale(this.pcoefs,p-this.p_offset,this.normformoments);
    }

    double getAI0(final double p)
    {
	return getI1Integrated(this.pcoefs,p-this.p_offset,this.normformoments)-this.ia0;
    }

} // END OF class QTradeIP






static abstract class TradeIPRData {
    final long currenttime,tau,vol;
    final double p_offset;
    final int np;
    TradeIPRData(final long currenttime,
		 final long tau,
		 final double p_offset,
		 final long vol,
		 final int np)
    {
	this.currenttime=currenttime;
	this.tau=tau;
	this.p_offset=p_offset;
	this.vol=vol;
	this.np=np;
    }
}





static class TradeIPMData extends TradeIPRData {

    final double pt_sum,t_sum;
    private final double [][] // here offset=<p>_t
	moments_tau_x_p   // [m][k]=sum (t-t_cur)^m (p-offset)^k dt_i exp()
	;

    // references to previos calculations
    private final BookItem trade_last;

    private final int ntm;

    TradeIPMData(final long currenttime,
		 final long tau,

		 final double pt_sum,
		 final double t_sum,
		 final double p_offset,

		 final double [][] moments_tau_x_p,

		 final long vol,
		 final int ntm,
		 final int npm,

		 // references
		 final BookItem trade_last)
    {
	super(currenttime,tau,p_offset,vol,npm);
	this.ntm=ntm;

	this.pt_sum=pt_sum;
	this.t_sum=t_sum;

	this.moments_tau_x_p=moments_tau_x_p;

	this.trade_last=trade_last;
    }

    private static TradeIPMData calculateTestTradeIPMData(final int ntm,
							  final int npm,
							  final long currenttime,
							  final long tau,
							  final BookItem [] tb,
							  final int tsize)
    {
	final double atau=tau;

	// calculate 0th & 1st moments
	double pt_sum=0,t_sum=0;
	long prevt=currenttime,tinterval=currenttime,vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    //System.err.println("i="+i+" b.time="+b.time);
	    prevt=b.time;
	    final long dt=tinterval-prevt;

	    vol+=b.shares;
	    final double ex=Math.exp((prevt-currenttime)/atau),twmw=ex*(dt/atau);

	    pt_sum+=b.price*twmw;
	    t_sum+=twmw;

	    tinterval=prevt;
	}

	final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);


	// calculate other moments
	final double [][]  // all 0 by default
	    moments_tau_x_p=new double [ntm][npm];
	prevt=tinterval=currenttime;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;
	    final double etau=(prevt-currenttime)/atau,ex=Math.exp(etau);
	    final long dt=tinterval-prevt;
	    final double dp=b.price-p_offset,dtatau=dt/atau,twmw=ex*dtatau;

	    double wte=twmw;
	    for(int m=0;m<ntm;m++)
	    {
		final double [] momp=moments_tau_x_p[m];
		double wpe=wte;
		for(int k=0;k<npm;k++)
		{
		    momp[k]+=wpe;
		    wpe*=dp;
		}
		wte*=etau;
	    }
	    tinterval=prevt;
	}

	return new TradeIPMData(currenttime,
				tau,
				pt_sum,
				t_sum,
				p_offset,
				moments_tau_x_p,
				vol,
				ntm,
				npm,
				tsize>0?tb[tsize-1]:null);
    }

    private static void checkMatching(final TradeIPMData r,final TradeIPMData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",r.moments_tau_x_p[0][1]+r.p_offset*r.moments_tau_x_p[0][0],"r.moments_tau_x_p[0][1]+r.p_offset*r.moments_tau_x_p[0][0]",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",r.moments_tau_x_p[0][0],"r.moments_tau_x_p[0][0]",eps);

	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	for(int m=0;m<r.ntm;m++)
	{
	    UnitTests.checkDoubleArrayRel(r.moments_tau_x_p[m],"r.moments_tau_x_p["+m+"]",ref.moments_tau_x_p[m],"ref.moments_tau_x_p["+m+"]",eps);
	}
	System.err.println("TradeIPMData:checkMatching(): matching OK");
    }


    private static void checkMatchingTradeIPData(final TradeIPMData r,final TradeIPData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	UnitTests.checkDoubleArrayRel(r.moments_tau_x_p[0],"r.moments_tau_x_p[0]",java.util.Arrays.copyOf(ref.moments_pt,r.np),"ref.moments_pt",eps);
	if(r.ntm>=2)
	    UnitTests.checkDoubleArrayRel(r.moments_tau_x_p[1],"r.moments_tau_x_p[1]",java.util.Arrays.copyOf(ref.moments_pt_tau,r.np),"ref.moments_pt_tau",eps);
	if(r.ntm>=3)
	    UnitTests.checkDoubleArrayRel(r.moments_tau_x_p[2],"r.moments_tau_x_p[2]",java.util.Arrays.copyOf(ref.moments_pt_tau2,r.np),"ref.moments_pt_tau2",eps);

	System.err.println("TradeIPMData:checkMatchingTradeIPData(): matching OK");
    }


static TradeIPMData calculateTradeIPMData(final int ntm,
					  final int npm,
					  final long currenttime,
					  final long tau,
					  final BookItem [] tb,
					  final int tsize,
					  final TradeIPMData ref)
{
    final double atau=tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime||ref.ntm!=ntm||ref.np!=npm)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.ntm="+ref.ntm+" ntm="+ntm+
				       " ref.np="+ref.np+" npm="+npm+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    double pt_sum=0,t_sum=0;
    long prevt=currenttime,tinterval=currenttime,vol=0,sumdt=0;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	final long dt=tinterval-prevt;
	vol+=b.shares;
	sumdt+=dt;
	final double ex=Math.exp((prevt-currenttime)/atau),twmw=ex*(dt/atau);
	pt_sum+=b.price*twmw;
	t_sum+=twmw;

	tinterval=prevt;
    }
    final double ref_delay,exp_ref_delay,bref_delay,expdt_bref_delay;
    if(bref!=null)
    { // optimization
	// add missed half slice
	sumdt+=tinterval-ref.currenttime;
	if(sumdt!=(currenttime-ref.currenttime))
	{
	    throw new RuntimeException("sumdt="+sumdt+" currenttime-ref.currenttime="+(currenttime-ref.currenttime));
	}
	bref_delay=(bref.time-currenttime)/atau;
	expdt_bref_delay=Math.exp(bref_delay)*((tinterval-ref.currenttime)/atau);
	pt_sum+=expdt_bref_delay*bref.price;
	t_sum+=expdt_bref_delay;

	// previous interval
	ref_delay=(ref.currenttime-currenttime)/atau;
	exp_ref_delay=Math.exp(ref_delay);
	pt_sum+=exp_ref_delay*ref.pt_sum;
	t_sum+=exp_ref_delay*ref.t_sum;
	vol+=ref.vol;
    }
    else
    {
	ref_delay=exp_ref_delay=bref_delay=expdt_bref_delay=0;
    }
    final double p_offset;
    final boolean offset_eq_average;
    if(t_sum>0)
    {
	p_offset=pt_sum/t_sum;
	offset_eq_average=true;
    }
    else
    {
	p_offset=tsize>0?tb[tsize-1].price:0;
	offset_eq_average=false;
    }


    final double [][]  // all 0 by default
	moments_tau_x_p=new double [ntm][npm];
    prevt=tinterval=currenttime;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	final double etau=(prevt-currenttime)/atau,ex=Math.exp(etau),
	    dp=b.price-p_offset,twmw=ex*((tinterval-prevt)/atau);

	double wte=twmw;
	for(int m=0;m<ntm;m++)
	{
	    final double [] momp=moments_tau_x_p[m];
	    double wpe=wte;
	    for(int k=0;k<npm;k++)
	    {
		momp[k]+=wpe;
		wpe*=dp;
	    }
	    wte*=etau;
	}
	tinterval=prevt;
    }


    if(bref!=null)
    { // optimization

	// previous interval & half slice
	final double bref_dp=bref.price-p_offset,dpa=ref.p_offset-p_offset;
	double wtest=1;
	for(int m=0;m<ntm;m++)
	{
	    double wpest=wtest;
	    for(int k=0;k<npm;k++)
	    {
		// add missed half slice
		moments_tau_x_p[m][k]+=wpest*expdt_bref_delay;

		// convert old interval to new p_offest basis
		// D2 polynomial group transform
		double s_tau_x_p=0;long fm=1;
		for(int ms=0;ms<=m;fm*=m-ms,fm/=(++ms))
		{
		    double sk_tau_x_p=0;long fk=1;
		    for(int ks=0;ks<=k;fk*=k-ks,fk/=(++ks))
		    {
			sk_tau_x_p=dpa*sk_tau_x_p+fk*ref.moments_tau_x_p[ms][ks];
		    }
		    s_tau_x_p=ref_delay*s_tau_x_p+fm*sk_tau_x_p;
		}
		moments_tau_x_p[m][k]+=exp_ref_delay*s_tau_x_p;

		wpest*=bref_dp;
	    }
	    wtest*=bref_delay;
	}
    } // end optimization
    if(offset_eq_average && true)
    {
	// reduce rounding errors for known values
	// <p-<p>>=0
	moments_tau_x_p[0][1]=0;
    }

    return new TradeIPMData(currenttime,
			    tau,
			    pt_sum,
			    t_sum,
			    p_offset,
			    moments_tau_x_p,
			    vol,
			    ntm,
			    npm,
			    tsize>0?tb[tsize-1]:null);
}



   private static void matchWithTest(final String description,
				     final TradeIPMData r,
				     final Trade [] tb,
				     final int tsize,
				     final double eps)
    {
	System.err.println("TradeIPMData:matchWithTest():======================"+description);
	final int ntm=r.ntm,npm=r.np;

	final TradeIPMData rt=calculateTestTradeIPMData(ntm,
							npm,
							r.currenttime,
							r.tau,
							tb,
							tsize);
	checkMatching(r,rt,eps);

	final TradeIPData rtipd=TradeIPData.calculateTestTradeIPData(npm/2+3,
								     r.currenttime,
								     r.tau,
								     tb,
								     tsize,
								     true);
	checkMatchingTradeIPData(r,rtipd,eps);
	System.err.println("TradeIPMData:matchWithTest(): OK");
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int ntm,final int npm)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("TradeIPMData:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" ntm="+ntm+" npm="+npm);

	final TradeIPMData ro=calculateTradeIPMData(ntm,
						    npm,
						    currenttime,
						    tau,
						    tb,
						    tsize,
						    null);
	matchWithTest("INITIAL CALCULATION",ro,tb,tsize,eps);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeIPMData em_ref=calculateTradeIPMData(ntm,npm,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest("GET REFERENCE FOR RECURRENT",em_ref,tb,n_split,eps);

	final TradeIPMData em_fromref=calculateTradeIPMData(ntm,npm,currenttime,tau,tb,tsize,em_ref);
	checkMatching(em_fromref,ro,eps);
	matchWithTest("RECURRENT CALCULATIONS",em_fromref,tb,tsize,eps);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeIPMData emc_cut=calculateTradeIPMData(ntm,npm,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut,eps);

	// dump values to diag
	System.err.println("moments_tau_x_p[0][1]="+ro.moments_tau_x_p[0][1]+" ; should be = 0");
	for(int m=0;m<ro.moments_tau_x_p.length;m++)
	{
	    System.err.println("moments_tau_x_p["+m+"][0]="+ro.moments_tau_x_p[m][0]+
			       " ; should be = "+(((m/2)*2==m?1:-1)*Polynomials.fact(m)));
	}

	System.err.println("TradeIPMData:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" ntm="+ntm+" npm="+npm);
    }

    public static void main(String [] args)
    {
	System.err.println("TradeIPMData::main(): starting tests; ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int ntm :new int []{1,2,3,4,5,6,7,8,9,10})
	{
	    for(int npm :new int []{2,3,4,5,6,7,8,9,10})
	    {
		for(int i=0;i<11;i++)
		{
		    checkCalculations(100+r.nextInt(100),r,ntm,npm);
		    checkCalculations(10010+r.nextInt(100010),r,ntm,npm);
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeIPMData::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }


} // END OF class TradeIPMData



static class QMTradeIP {

    private final boolean FLAG_L_CONDITION;


    private final double [] pcoefs;
    private final double [][] P;
    final double p_offset,sigma;

    final double [] p_imin,p_imax;

    final int neq_t,neq_p;
    private final double i0;
    private final double [] p_real_roots;
    private final TradeIPMData odata;

    private QMTradeIP(final com.polytechnik.utils.PolynomialRootsFinder rf,
		      final int neq_t,final int neq_p,
		      final TradeIPMData odata,
		      final boolean FLAG_L_CONDITION)
    {
	this.FLAG_L_CONDITION=FLAG_L_CONDITION;
	this.odata=odata;
	this.neq_t=neq_t;
	this.neq_p=neq_p;

	this.p_offset=odata.p_offset;
	final double pt0=odata.moments_tau_x_p[0][0];

	this.sigma=Math.sqrt(odata.moments_tau_x_p[0][2]/pt0-Math.pow(odata.moments_tau_x_p[0][1]/pt0,2));

	this.P=getPCoefs(neq_t,neq_p);

	this.pcoefs=new double [2*neq_p+1];
	final double dta=true?0:(odata.moments_tau_x_p.length<=1?0:-1*(odata.moments_tau_x_p[1][0]/pt0));
	double wt=1;
	for(double [] c : this.P)
	{
	    for(int k=c.length;--k>=0;)
	    {
		this.pcoefs[k]+=wt*c[k];
	    }
	    wt*=dta;
	}
	wt=0;
	for(int k=this.pcoefs.length;--k>=0;)
	{
	    wt+=this.pcoefs[k]*(odata.moments_tau_x_p[0][k]/pt0);
	}
	this.i0=wt;

	final double [][] xx=QATradeIP.getPolynomialMinMax(this.pcoefs.length-1,this.pcoefs,this.p_offset,rf);
	this.p_imin=xx[0];
	this.p_imax=xx[1];
	this.p_real_roots=xx[2];
    }

    double [][] getPCoefs(final int neq_t,final int neq_p)
    {
	if(FLAG_L_CONDITION)
	{
	    return getPLCondititionCoefs(
		this.odata.moments_tau_x_p,
		0,
		this.odata.trade_last==null?Double.NaN:this.odata.trade_last.price-this.odata.p_offset,
		neq_t,
		neq_p);
	}
	return getPNoCondititionCoefs(this.odata.moments_tau_x_p,neq_t,neq_p);
    }

    /*private*/ static double [][] getPNoCondititionCoefs(final double [][] tp,
						      final int neq_t,
						      final int neq_p)
    {
	final double pt0=tp[0][0],nstdev=Math.sqrt(tp[0][2]/pt0);
	final int ne=neq_t+neq_p;
	final double [] norms=new double [ne];
	double wn=1;
	long fact=1;
	for(int k=0;k<ne;k++)
	{
	    if(k<neq_t)
	    {
		norms[k]=(fact*=k+1);
	    }
	    else
	    {
		norms[k]=(wn*=nstdev);
	    }
	}
	final double [][] P=new double [2*neq_t+1][];
	for(int k=P.length;--k>=0;)
	{
	    if(k<=0)
	    {
		P[k]=new double [2*neq_p+1];
	    }
	    else if(k<=neq_t)
	    {
		P[k]=new double [neq_p+1];
	    }
	    else
	    {
		P[k]=new double [1];
	    }
	}

	final double [] matr=new double [ne*ne];
	for(int k=1,j=0;k<=ne;k++)
	{
	    final int ik;
	    final double normsk=norms[k-1],pk;
	    if(k<=neq_t)
	    {
		ik=k;
		pk=tp[ik][0]/pt0;
	    }
	    else
	    {
		ik=k-neq_t;
		pk=tp[0][ik]/pt0;
	    }
	    for(int l=1;l<=ne;l++,j++)
	    {
		final int il;
		final double pkl,pl;
		if(l<=neq_t)
		{
		    il=l;
		    pl=tp[il][0]/pt0;
		    pkl=(k<=neq_t?tp[ik+il][0]:tp[il][ik])/pt0;
		}
		else
		{
		    il=l-neq_t;
		    pl=tp[0][il]/pt0;
		    pkl=(k<=neq_t?tp[ik][il]:tp[0][ik+il])/pt0;
		}
		matr[j]=(pkl-pl*pk)/normsk/norms[l-1];
	    }
	}

	final int [] ipvt=new int [ne];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,ne,ne,ipvt);
	if(pinfo!=0)
	{
	    for(double [] p:P)
	    {
		java.util.Arrays.fill(p,Double.NaN);
	    }
	    return P;
	}

	final double [] rpart=new double [ne];
	for(int k=ne;k>=1;--k)
	{
	    rpart[k-1]=1;
	    com.polytechnik.utils.Linsystems.dgesl_m1(matr,ne,ne,ipvt,rpart,1);
	    final int ik;
	    final double normsk=norms[k-1],pk;
	    if(k<=neq_t)
	    {
		ik=k;
		pk=tp[ik][0]/pt0;
	    }
	    else
	    {
		ik=k-neq_t;
		pk=tp[0][ik]/pt0;
	    }
	    for(int l=ne;l>=1;--l)
	    {
		final int il;
		final double c=rpart[l-1]/normsk/norms[l-1];
		rpart[l-1]=0;
		if(l<=neq_t)
		{
		    il=l;
		    final double plc=c*(tp[il][0]/pt0);
		    if(k<=neq_t)
		    {
			P[0][0]+=plc*pk;
			P[ik][0]-=plc;
			P[il][0]-=pk*c;
			P[il+ik][0]+=c;
		    }
		    else
		    {
			P[0][0]+=plc*pk;
			P[0][ik]-=plc;
			P[il][0]-=pk*c;
			P[il][ik]+=c;
		    }
		}
		else
		{
		    il=l-neq_t;
		    final double plc=c*(tp[0][il]/pt0);
		    if(k<=neq_t)
		    {
			P[0][0]+=plc*pk;
			P[ik][0]-=plc;
			P[0][il]-=pk*c;
			P[ik][il]+=c;
		    }
		    else
		    {
			P[0][0]+=plc*pk;
			P[0][ik]-=plc;
			P[0][il]-=pk*c;
			P[0][il+ik]+=c;
		    }
		}
	    }
	}

	return P;
    }

    private static double [][] getPLCondititionCoefs(final double [][] tp,
						     final double dt,
						     final double dp,
						     final int neq_t,
						     final int neq_p)
    {
	final double pt0=tp[0][0],nstdev=Math.sqrt(tp[0][2]/pt0);
	final int ne=neq_t+neq_p;
	final double [] norms=new double [ne];
	double wkt=1,wkp=1;
	for(int k=0;k<ne;k++)
	{
	    if(k<neq_t)
	    {
		norms[k]=(wkt*=k+1);
	    }
	    else
	    {
		norms[k]=(wkp*=nstdev);
	    }
	}
	final double [][] P=new double [2*neq_t+1][];
	for(int k=P.length;--k>=0;)
	{
	    if(k<=0)
	    {
		P[k]=new double [2*neq_p+1];
	    }
	    else if(k<=neq_t)
	    {
		P[k]=new double [neq_p+1];
	    }
	    else
	    {
		P[k]=new double [1];
	    }
	}

	final double [][] matr=new double [ne][];
	wkt=wkp=1;
	for(int k=1;k<=ne;++k)
	{
	    final int ik;
	    final double normsk=norms[k-1],pk,wk;
	    final double [] matrk=matr[k-1]=new double [ne];
	    if(k<=neq_t)
	    {
		ik=k;
		wk=(wkt*=dt);
		pk=tp[ik][0]/pt0;
	    }
	    else
	    {
		ik=k-neq_t;
		wk=(wkp*=dp);
		pk=tp[0][ik]/pt0;
	    }
	    double wlt=1,wlp=1;
	    for(int l=1;l<=ne;++l)
	    {
		final int il;
		final double pkl,pl,wl;
		if(l<=neq_t)
		{
		    il=l;
		    wl=(wlt*=dt);
		    pl=tp[il][0]/pt0;
		    pkl=(k<=neq_t?tp[ik+il][0]:tp[il][ik])/pt0;
		}
		else
		{
		    il=l-neq_t;
		    wl=(wlp*=dp);
		    pl=tp[0][il]/pt0;
		    pkl=(k<=neq_t?tp[ik][il]:tp[0][ik+il])/pt0;
		}
		matrk[l-1]=(pkl-wl*pk+wk*(wl-pl))/normsk/norms[l-1];
	    }
	}

	final int [] ipvt=new int [ne];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa(matr,ne,ipvt);
	if(pinfo!=0)
	{
	    for(double [] p:P)
	    {
		java.util.Arrays.fill(p,Double.NaN);
	    }
	    return P;
	}

	wkt=wkp=1;
	final double [] rpart=new double [ne];
	for(int k=1;k<=ne;++k)
	{
	    rpart[k-1]=1;
	    com.polytechnik.utils.Linsystems.dgesl(matr,ne,ipvt,rpart,1);
	    final int ik;
	    final double normsk=norms[k-1],wk;
	    if(k<=neq_t)
	    {
		ik=k;
		wk=(wkt*=dt);
	    }
	    else
	    {
		ik=k-neq_t;
		wk=(wkp*=dp);
	    }
	    double wlt=1,wlp=1;
	    for(int l=1;l<=ne;++l)
	    {
		final int il;
		final double c=rpart[l-1]/normsk/norms[l-1],cwl;
		rpart[l-1]=0;
		if(l<=neq_t)
		{
		    il=l;
		    cwl=c*(wlt*=dt);
		    if(k<=neq_t)
		    {
			P[0][0]+=cwl*wk;
			P[ik][0]-=cwl;
			P[il][0]-=wk*c;
			P[il+ik][0]+=c;
		    }
		    else
		    {
			P[0][0]+=cwl*wk;
			P[0][ik]-=cwl;
			P[il][0]-=wk*c;
			P[il][ik]+=c;
		    }
		}
		else
		{
		    il=l-neq_t;
		    cwl=c*(wlp*=dp);
		    if(k<=neq_t)
		    {
			P[0][0]+=cwl*wk;
			P[ik][0]-=cwl;
			P[0][il]-=wk*c;
			P[ik][il]+=c;
		    }
		    else
		    {
			P[0][0]+=cwl*wk;
			P[0][ik]-=cwl;
			P[0][il]-=wk*c;
			P[0][il+ik]+=c;
		    }
		}
	    }
	}

	return P;
    }

    private static void checkDPLCondition(final java.util.Random r,final QMTradeIP ro,final double eps,final Trade [] tb,final int tsize)
    {
	final double pt0=ro.odata.moments_tau_x_p[0][0];
	for(int ip=0;ip<100;ip++)
	{
	    double dp=TradeEMoments.getRandomPrice(r)-ro.p_offset;
	    for(int it=0;it<200;it++)
	    {
		double dt=r.nextDouble();
		final double [][] P;
		if(ip==0 && it==0 && ro.FLAG_L_CONDITION)
		{
		    dt=0;
		    dp=ro.odata.trade_last.price-ro.odata.p_offset;
		    P=ro.P;
		}
		else
		{
		    P=getPLCondititionCoefs(ro.odata.moments_tau_x_p,
					    dt,
					    dp,
					    ro.neq_t,
					    ro.neq_p);
		}
		double s=0,st=0,wjt=1;
		for(int j=0;j<P.length;j++)
		{
		    final double [] m=ro.odata.moments_tau_x_p[j],c=P[j];
		    double sp=0,wkp=1;
		    for(int k=0;k<c.length;k++)
		    {
			s+=c[k]*(m[k]/pt0);
			sp+=wkp*c[k];
			wkp*=dp;
		    }
		    st+=wjt*sp;
		    wjt*=dt;
		}
		UnitTests.checkDoubleRel(s,"dIaver",ro.neq_t+ro.neq_p,"ro.neq_t+ro.neq_p",eps);
		UnitTests.checkDoubleRel(st,"I(0)",0,"0",eps);
	    }
	}
	System.err.println("checkDPLCondition: tests OK");
    }



    public static QMTradeIP calculateQMTradeIP(final int neq_t,
					       final int neq_p,
					       final long currenttime,
					       final long tau,
					       final BookItem [] tb,
					       final int tsize,
					       final com.polytechnik.utils.PolynomialRootsFinder rf,
					       final QMTradeIP ref,
					       final boolean FLAG_L_CONDITION)
    {
	final TradeIPMData o=TradeIPMData.calculateTradeIPMData(2*neq_t+1,
								2*neq_p+1,
								currenttime,
								tau,
								tb,
								tsize,
								ref!=null?ref.odata:null);
	return new QMTradeIP(rf,neq_t,neq_p,o,FLAG_L_CONDITION);
    }


    double getDIDPDT_TOT(final double p)
    {
	return Polynomials.getD1(this.pcoefs,p-this.p_offset);
    }
    double getDIDT(final double p)
    {
	return Polynomials.getD0(this.pcoefs,p-this.p_offset);
    }

    double getPGMIN(final double p)
    {
	return QATradeIP.getPGMIN(p,this.p_imin,this.pcoefs,this.p_offset);
    }

    double getPLMIN(final double p)
    {
	return QATradeIP.getPLMIN(p,this.p_real_roots,this.p_offset);
    }

    double getIF(final double p)
    {
	return (Polynomials.getD0(this.pcoefs,p-this.p_offset)-this.i0)/this.i0;
    }

    double getIDP(final double p)
    {
	final double dp=p-this.p_offset;
	return Polynomials.getD0(this.pcoefs,dp)/Polynomials.getD1(this.pcoefs,dp);
    }



    private static void checkDPNoCondition(final java.util.Random r,final QMTradeIP ro,final double eps,final Trade [] tb,final int tsize)
    {

	final double pt0=ro.odata.moments_tau_x_p[0][0];
	double s=0;
	for(int j=ro.P.length;--j>=0;)
	{
	    final double [] m=ro.odata.moments_tau_x_p[j],c=ro.P[j];
	    for(int k=c.length;--k>=0;)
	    {
		s+=c[k]*(m[k]/pt0);
	    }
	}
	UnitTests.checkDoubleRel(s,"dIaver",ro.neq_t+ro.neq_p,"ro.neq_t+ro.neq_p",eps);
	System.err.println("QMTradeIP:checkDPNoCondition: OK dIaver="+s+" ro.neq_t+ro.neq_p="+(ro.neq_t+ro.neq_p));
    }

    private static void checkD0(final int np,
				final long currenttime,
				final long tau,
				final BookItem [] tb,
				final int tsize,
				final com.polytechnik.utils.PolynomialRootsFinder rf,
				final java.util.Random r,
				final double eps)
    {
	final QATradeIP rx=QATradeIP.calculateQATradeIP(np,
							currenttime,
							tau,
							tb,
							tsize,
							rf,
							null);
	final QMTradeIP rm=QMTradeIP.calculateQMTradeIP(0,
							np,
							currenttime,
							tau,
							tb,
							tsize,
							rf,null,false);

	UnitTests.checkDoubleRel(rx.sigma,"rx.sigma",rm.sigma,"rm.sigma",eps);
	UnitTests.checkDoubleRel(rx.p_offset,"rx.p_offset",rm.p_offset,"rm.p_offset",eps);
	UnitTests.checkDoubleRel(rm.i0,"rm.i0",rm.neq_p,"rm.neq_p",eps);

	UnitTests.checkDoubleArrayRel(rx.didv,"rx.didv",rm.P[0],"rm.P[0]",eps);

	System.err.println("QMTradeIP:checkD0(): QATradeIP OK with np="+np+" rm.i0="+rm.i0);
    }


    private static void checkD1(final int np,
				final long currenttime,
				final long tau,
				final BookItem [] tb,
				final int tsize,
				final com.polytechnik.utils.PolynomialRootsFinder rf,
				final java.util.Random r,
				final double eps)
    {
	final QQTradeIP rx=QQTradeIP.calculateQQTradeIP(np,
							currenttime,
							tau,
							tb,
							tsize,
							rf,
							null);
	final QMTradeIP rm=QMTradeIP.calculateQMTradeIP(1,
							np,
							currenttime,
							tau,
							tb,
							tsize,
							rf,null,false);

	UnitTests.checkDoubleRel(rx.sigma,"rx.sigma",rm.sigma,"rm.sigma",eps);
	UnitTests.checkDoubleRel(rx.p_offset,"rx.p_offset",rm.p_offset,"rm.p_offset",eps);

	UnitTests.checkDoubleArrayRel(rx.P0,"rx.P0",rm.P[0],"rm.P[0]",eps);
	UnitTests.checkDoubleArrayRel(rx.P1,"rx.P1",rm.P[1],"rm.P[1]",eps);
	UnitTests.checkDoubleArrayRel(rx.P2,"rx.P2",rm.P[2],"rm.P[2]",eps);

	System.err.println("QMTradeIP:checkD1(): QQTradeIP OK with np="+np);
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_t,final int neq_p)
    {
	final double eps=1e-6;

	final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();
	for(boolean FLAG_L_CONDITION : new boolean[] {true,false})
	{
	    	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
		final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
		System.err.println();
		System.err.println("------------------------------------------------------------------------");
		System.err.println("QMTradeIP:checkCalculations(): test with N points="+tsize+
				   " currenttime="+currenttime+" tau="+tau+" neq_t="+neq_t+" neq_p="+neq_p+
				   " FLAG_L_CONDITION="+FLAG_L_CONDITION);


		final QMTradeIP ro=calculateQMTradeIP(neq_t,
						      neq_p,
						      currenttime,
						      tau,
						      tb,
						      tsize,
						      rf,
						      null,
						      FLAG_L_CONDITION);

		if(true)
		{ // check 0-th moments estimate
		    System.err.println("moments_tau_x_p[0][1]="+ro.odata.moments_tau_x_p[0][1]+" ; should be = 0");
		    for(int m=0;m<ro.odata.moments_tau_x_p.length;m++)
		    {
			System.err.println("moments_tau_x_p["+m+"][0]="+ro.odata.moments_tau_x_p[m][0]+
					   " ; should be = "+(((m/2)*2==m?1:-1)*Polynomials.fact(m)));
		    }
		}


		checkDPLCondition(r,ro,eps,tb,tsize);
		checkDPNoCondition(r,ro,eps,tb,tsize);
		checkD0(neq_p,currenttime,tau,tb,tsize,rf,r,eps);
		checkD1(neq_p,currenttime,tau,tb,tsize,rf,r,eps);


		System.err.println("QMTradeIP:checkCalculations(): OK with N points="+tsize+
				   " currenttime="+currenttime+" tau="+tau+" neq_t="+neq_t+" neq_p="+neq_p+
				   " FLAG_L_CONDITION="+FLAG_L_CONDITION);
	}
    }

    public static void main(String [] args)
    {
	System.err.println("QMTradeIP::main(): starting tests; ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int neq_t=0;neq_t<=6;neq_t++)
	{
	    for(int neq_p=1;neq_p<=8;neq_p++)
	    {
		for(int i=0;i<3;i++)
		{
		    checkCalculations(100+r.nextInt(100),r,neq_t,neq_p);
		    checkCalculations(10010+r.nextInt(100010),r,neq_t,neq_p);
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QMTradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

} // END OF class QMTradeIP






static class TradeIPVMData extends TradeIPRData {

    /* private */final double pt_sum,t_sum,v_sum,v_scale;
    /* private */final double [][][] // here offset=<p>_t
	moments_vol_x_tau_x_p   // [k][l][m]=sum (vol-v_cur)^k (t-t_cur)^l (p-offset)^m dt_i exp()
	;

    // references to previos calculations
    private final BookItem trade_last;

    /* private */final int nvm,ntm;

    TradeIPVMData(final long currenttime,
		  final long tau,

		  final double pt_sum,
		  final double t_sum,
		  final double v_sum,
		  final double p_offset,
		  final double v_scale,

		  final double [][][] moments_vol_x_tau_x_p,

		  final long vol,
		  final int nvm,
		  final int ntm,
		  final int npm,

		  // references
		  final BookItem trade_last)
    {
	super(currenttime,tau,p_offset,vol,npm);
	this.nvm=nvm;
	this.ntm=ntm;

	this.pt_sum=pt_sum;
	this.t_sum=t_sum;
	this.v_sum=v_sum;
	this.v_scale=v_scale;

	this.moments_vol_x_tau_x_p=moments_vol_x_tau_x_p;

	this.trade_last=trade_last;
    }

    /* private */ static TradeIPVMData calculateTestTradeIPVMData(final int nvm,
							    final int ntm,
							    final int npm,
							    final long currenttime,
							    final long tau,
							    final BookItem [] tb,
							    final int tsize)
    {
	final double atau=tau;

	// calculate 0th & 1st moments
	double pt_sum=0,t_sum=0,v_sum=0;
	long prevt=currenttime,tinterval=currenttime,vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    //System.err.println("i="+i+" b.time="+b.time);
	    prevt=b.time;
	    final long dt=tinterval-prevt;

	    vol+=b.shares;
	    final double ex=Math.exp((prevt-currenttime)/atau),twmw=ex*(dt/atau),bwmw=ex*b.shares;

	    pt_sum+=b.price*twmw;
	    t_sum+=twmw;
	    v_sum+=bwmw;

	    tinterval=prevt;
	}

	final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0),v_scale=v_sum>0?v_sum:1;


	// calculate other moments
	final double [][][]  // all 0 by default
	    moments_vol_x_tau_x_p=new double [nvm][ntm][npm];
	long vx=0;
	prevt=tinterval=currenttime;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;
	    vx-=b.shares;
	    final double etau=(prevt-currenttime)/atau,ex=Math.exp(etau),evol=vx/v_scale;
	    final long dt=tinterval-prevt;
	    final double dp=b.price-p_offset,dtatau=dt/atau,twmw=ex*dtatau;

	    double wve=twmw;
	    for(int k=0;k<nvm;k++)
	    {
		double wte=wve;
		for(int l=0;l<ntm;l++)
		{
		    double wpe=wte;
		    for(int m=0;m<npm;m++)
		    {
			moments_vol_x_tau_x_p[k][l][m]+=wpe;
			wpe*=dp;
		    }
		    wte*=etau;
		}
		wve*=evol;
	    }
	    tinterval=prevt;
	}

	return new TradeIPVMData(currenttime,
				 tau,
				 pt_sum,
				 t_sum,
				 v_sum,
				 p_offset,
				 v_scale,
				 moments_vol_x_tau_x_p,
				 vol,
				 nvm,
				 ntm,
				 npm,
				 tsize>0?tb[tsize-1]:null);
    }

    private static void checkMatching(final TradeIPVMData r,final TradeIPVMData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",r.moments_vol_x_tau_x_p[0][0][1]+r.p_offset*r.moments_vol_x_tau_x_p[0][0][0],"r.moments_vol_x_tau_x_p[0][0][1]+r.p_offset*r.moments_vol_x_tau_x_p[0][0][0]",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",r.moments_vol_x_tau_x_p[0][0][0],"r.moments_vol_x_tau_x_p[0][0][0]",eps);
	UnitTests.checkDoubleRel(r.v_sum,"r.v_sum",ref.v_sum,"ref.v_sum",eps);

	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);
	UnitTests.checkDoubleRel(r.v_scale,"r.v_scale",ref.v_scale,"ref.v_scale",eps);

	for(int k=0;k<r.nvm;k++)
	{
	    for(int l=0;l<r.ntm;l++)
	    {
		UnitTests.checkDoubleArrayRel(r.moments_vol_x_tau_x_p[k][l],"r.moments_vol_x_tau_x_p["+k+"]["+l+"]",
				    ref.moments_vol_x_tau_x_p[k][l],"ref.moments_vol_x_tau_x_p["+k+"]["+l+"]",eps);
	    }
	}
	System.err.println("TradeIPVMData:checkMatching(): matching OK");
    }


    private static void checkMatchingTradeIPMData(final TradeIPVMData r,final TradeIPMData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	for(int l=0;l<r.ntm;l++)
	{
	    UnitTests.checkDoubleArrayRel(r.moments_vol_x_tau_x_p[0][l],"r.moments_vol_x_tau_x_p[0]["+l+"]",
				ref.moments_tau_x_p[l],"r.moments_tau_x_p["+l+"]",eps);
	}

	System.err.println("TradeIPVMData:checkMatchingTradeIPMData(): matching OK");
    }


static TradeIPVMData calculateTradeIPVMData(final int nvm,
					    final int ntm,
					    final int npm,
					    final long currenttime,
					    final long tau,
					    final BookItem [] tb,
					    final int tsize,
					    final TradeIPVMData ref)
{
    final double atau=tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime||
	   ref.nvm!=nvm||ref.ntm!=ntm||ref.np!=npm)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " ref.nvm="+ref.nvm+" nvm="+nvm+
				       " ref.ntm="+ref.ntm+" ntm="+ntm+
				       " ref.np="+ref.np+" npm="+npm+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    double pt_sum=0,t_sum=0,v_sum=0;
    long prevt=currenttime,tinterval=currenttime,vol=0,sumdt=0;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	final long dt=tinterval-prevt;
	vol+=b.shares;
	sumdt+=dt;
	final double ex=Math.exp((prevt-currenttime)/atau),twmw=ex*(dt/atau);
	pt_sum+=b.price*twmw;
	t_sum+=twmw;
	v_sum+=ex*b.shares;

	tinterval=prevt;
    }
    final double ref_delay,exp_ref_delay,bref_delay,expdt_bref_delay;
    if(bref!=null)
    { // optimization
	// add missed half slice
	sumdt+=tinterval-ref.currenttime;
	if(sumdt!=(currenttime-ref.currenttime))
	{
	    throw new RuntimeException("sumdt="+sumdt+" currenttime-ref.currenttime="+(currenttime-ref.currenttime));
	}
	bref_delay=(bref.time-currenttime)/atau;
	expdt_bref_delay=Math.exp(bref_delay)*((tinterval-ref.currenttime)/atau);
	pt_sum+=expdt_bref_delay*bref.price;
	t_sum+=expdt_bref_delay;

	// previous interval
	ref_delay=(ref.currenttime-currenttime)/atau;
	exp_ref_delay=Math.exp(ref_delay);
	pt_sum+=exp_ref_delay*ref.pt_sum;
	t_sum+=exp_ref_delay*ref.t_sum;
	v_sum+=exp_ref_delay*ref.v_sum;
	vol+=ref.vol;
    }
    else
    {
	ref_delay=exp_ref_delay=bref_delay=expdt_bref_delay=0;
    }
    final double p_offset,v_scale=v_sum>0?v_sum:1;
    final boolean offset_eq_average;
    if(t_sum>0)
    {
	p_offset=pt_sum/t_sum;
	offset_eq_average=true;
    }
    else
    {
	p_offset=tsize>0?tb[tsize-1].price:0;
	offset_eq_average=false;
    }


    final double [][][]  // all 0 by default
	moments_vol_x_tau_x_p=new double [nvm][ntm][npm];
    long vx=0;
    prevt=tinterval=currenttime;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	vx-=b.shares;
	final double etau=(prevt-currenttime)/atau,ex=Math.exp(etau),evol=vx/v_scale,
	    dp=b.price-p_offset,twmw=ex*((tinterval-prevt)/atau);

	double wve=twmw;
	for(int k=0;k<nvm;k++)
	{
	    final double [][] momtp=moments_vol_x_tau_x_p[k];
	    double wte=wve;
	    for(int l=0;l<ntm;l++)
	    {
		final double [] momp=momtp[l];
		double wpe=wte;
		for(int m=0;m<npm;m++)
		{
		    momp[m]+=wpe;
		    wpe*=dp;
		}
		wte*=etau;
	    }
	    wve*=evol;
	}
	tinterval=prevt;
    }


    if(bref!=null)
    { // optimization

	// previous interval & half slice
	final double bref_dp=bref.price-p_offset,bref_evol=(vx-bref.shares)/v_scale,
	    dpa=ref.p_offset-p_offset,dvol=(ref.vol-vol)/ref.v_scale,dscale=ref.v_scale/v_scale;
	double wvest=1,scalesk=1;
	for(int k=0;k<nvm;k++)
	{
	    double wtest=wvest;
	    for(int l=0;l<ntm;l++)
	    {
		double wpest=wtest;
		for(int m=0;m<npm;m++)
		{
		    // add missed half slice
		    moments_vol_x_tau_x_p[k][l][m]+=wpest*expdt_bref_delay;

		    // convert old interval to new (v_cur,t_cur,p_offest) basis
		    // D3 polynomial group transform
		    double sk=0;long fk=1;
		    for(int ks=0;ks<=k;fk*=k-ks,fk/=(++ks))
		    {
			double sl=0;long fl=1;
			for(int ls=0;ls<=l;fl*=l-ls,fl/=(++ls))
			{
			    double sm=0;long fm=1;
			    for(int ms=0;ms<=m;fm*=m-ms,fm/=(++ms))
			    {
				sm=dpa*sm+fm*ref.moments_vol_x_tau_x_p[ks][ls][ms];
			    }
			    sl=ref_delay*sl+fl*sm;
			}
			sk=dvol*sk+fk*sl;
		    }
		    moments_vol_x_tau_x_p[k][l][m]+=scalesk>1?((exp_ref_delay*scalesk)*sk):(exp_ref_delay*(scalesk*sk));
		    wpest*=bref_dp;
		}
		wtest*=bref_delay;
	    }
	    wvest*=bref_evol;
	    scalesk*=dscale;
	}
    } // end optimization
    if(offset_eq_average && true)
    {
	// reduce rounding errors for known values
	// <p-<p>>=0
	moments_vol_x_tau_x_p[0][0][1]=0;
    }

    return new TradeIPVMData(currenttime,
			     tau,
			     pt_sum,
			     t_sum,
			     v_sum,
			     p_offset,
			     v_scale,
			     moments_vol_x_tau_x_p,
			     vol,
			     nvm,
			     ntm,
			     npm,
			     tsize>0?tb[tsize-1]:bref);
}



   private static void matchWithTest(final String description,
				     final TradeIPVMData r,
				     final Trade [] tb,
				     final int tsize,
				     final double eps)
    {
	System.err.println("TradeIPVMData:matchWithTest():======================"+description);
	final int nvm=r.nvm,ntm=r.ntm,npm=r.np;

	final TradeIPVMData rt=calculateTestTradeIPVMData(nvm,
							  ntm,
							  npm,
							  r.currenttime,
							  r.tau,
							  tb,
							  tsize);
	checkMatching(r,rt,eps);

	final TradeIPMData rtipd=TradeIPMData.calculateTestTradeIPMData(ntm,
									npm,
									r.currenttime,
									r.tau,
									tb,
									tsize);
	checkMatchingTradeIPMData(r,rtipd,eps);
	System.err.println("TradeIPVMData:matchWithTest(): OK");
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int nvm,final int ntm,final int npm)
    {
	final double eps=1e-5;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("TradeIPVMData:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" nvm="+nvm+" ntm="+ntm+" npm="+npm);

	final TradeIPVMData ro=calculateTradeIPVMData(nvm,
						      ntm,
						      npm,
						      currenttime,
						      tau,
						      tb,
						      tsize,
						      null);
	matchWithTest("INITIAL CALCULATION",ro,tb,tsize,eps);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeIPVMData em_ref=calculateTradeIPVMData(nvm,ntm,npm,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest("GET REFERENCE FOR RECURRENT",em_ref,tb,n_split,eps);

	final TradeIPVMData em_fromref=calculateTradeIPVMData(nvm,ntm,npm,currenttime,tau,tb,tsize,em_ref);
	checkMatching(em_fromref,ro,eps);
	matchWithTest("RECURRENT CALCULATIONS",em_fromref,tb,tsize,eps);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeIPVMData emc_cut=calculateTradeIPVMData(nvm,ntm,npm,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut,eps);

	// dump values to diag
	for(int k=0;k<ro.nvm;k++)
	{
	    System.err.println("V00:moments_vol_x_tau_x_p["+k+"][0][0]="+ro.moments_vol_x_tau_x_p[k][0][0]
			       +" ; should be "+(k==0?"1":"??"));
	}
	for(int l=0;l<ro.ntm;l++)
	{
	    System.err.println("T00:moments_vol_x_tau_x_p[0]["+l+"][0]="+ro.moments_vol_x_tau_x_p[0][l][0]+
			       " ; should be = "+(((l/2)*2==l?1:-1)*Polynomials.fact(l)));
	}
	for(int m=0;m<ro.np;m++)
	{
	    System.err.println("P00:moments_vol_x_tau_x_p[0][0]["+m+"]="+ro.moments_vol_x_tau_x_p[0][0][m]
			       +" ; should be "+(m==0?"1":(m==1?"0":"??")));
	}

	System.err.println("TradeIPVMData:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" nvm="+nvm+" ntm="+ntm+" npm="+npm);
    }

    public static void main(String [] args)
    {
	System.err.println("TradeIPVMData::main(): starting tests; ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79093);
	for(int nvm :new int []{3,1,2,3,4,5,6,7})
	{
	    for(int ntm :new int []{1,2,3,4,5,6,7})
	    {
		for(int npm :new int []{2,3,4,5,6,7,8})
		{
		    for(int i=0;i<3;i++)
		    {
			checkCalculations(100+r.nextInt(100),r,nvm,ntm,npm);
			checkCalculations(10010+r.nextInt(100010),r,nvm,ntm,npm);
		    }
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeIPVMData::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }
} // END OF class TradeIPVMData



static class QVMTradeIP {
    private static final boolean LOOK_FORWARD=true;

    /*private*/ final double [] pcoefs;
    private final double [][][] abc;
    final double p_offset,sigma;

    final double [] p_imin,p_imax;

    final double f_dv,f_dt;
    final int neq_v,neq_t,neq_p;
    /*private*/ final double i0;
    private final double [] p_real_roots;
    /*private*/ final TradeIPVMData odata;

    private QVMTradeIP(final com.polytechnik.utils.PolynomialRootsFinder rf,
		       final int neq_v,final int neq_t,final int neq_p,
		       final TradeIPVMData odata)
    {
	this.odata=odata;
	this.neq_v=neq_v;
	this.neq_t=neq_t;
	this.neq_p=neq_p;

	this.p_offset=odata.p_offset;
	this.abc=odata.moments_vol_x_tau_x_p;
	final double pt0=this.abc[0][0][0];

	this.sigma=Math.sqrt(this.abc[0][0][2]/pt0-Math.pow(this.abc[0][0][1]/pt0,2));

	final double [][][] P=getPCoefs(neq_v,neq_t,neq_p,this.abc);

	this.f_dv=!LOOK_FORWARD?0:-1*(this.abc.length<=1?0:this.abc[1][0][0]/pt0);
	this.f_dt=!LOOK_FORWARD?0:-1*(this.abc[0].length<=1?0:this.abc[0][1][0]/pt0);
	this.pcoefs=getPProjected(P,this.f_dv,this.f_dt);

	double si0=0;
	for(int m=this.pcoefs.length;--m>=0;)
	{
	    si0+=this.pcoefs[m]*(this.abc[0][0][m]/pt0);
	}
	this.i0=si0;

	final double [][] xx=QATradeIP.getPolynomialMinMax(this.pcoefs.length-1,this.pcoefs,this.p_offset,rf);
	this.p_imin=xx[0];
	this.p_imax=xx[1];
	this.p_real_roots=xx[2];
    }

    double [][][] getPCoefs(final int neq_v,final int neq_t,final int neq_p,final double [][][] abc)
    {
	return getPLCondititionCoefs(
	    abc,
	    0,
	    0,
	    this.odata.trade_last==null?Double.NaN:this.odata.trade_last.price-this.odata.p_offset,
	    neq_v,
	    neq_t,
	    neq_p);
    }

    /*private*/ static double [][][] getPLCondititionCoefs(final double [][][] abc,
						       final double da,
						       final double db,
						       final double dc,
						       final int neq_a,
						       final int neq_b,
						       final int neq_c)
    {
	final double pt0=abc[0][0][0];
	final double [] darr=new double []{
	    neq_a<=0?Double.NaN:Math.sqrt(abc[2][0][0]/pt0),
	    neq_b<=0?Double.NaN:Math.sqrt(abc[0][2][0]/pt0),
	    neq_c<=0?Double.NaN:Math.sqrt(abc[0][0][2]/pt0)};
	final int [] ndeq=new int []{neq_a,neq_b,neq_c};
	final int nd=3,ne=neq_a+neq_b+neq_c;

	final double [] norms=new double [ne];
	for(int ik=0,ek=0;ek<nd;ek++)
	{
	    final double stdev=darr[ek];
	    double w=1;
	    for(int k=1,nk=ndeq[ek];k<=nk;k++,ik++)
	    {
		norms[ik]=(w*=(stdev*k));
		//System.err.println("QVMTradeIP: norms["+ik+"]="+norms[ik]);
	    }
	}
	final double [][][] P=new double [2*neq_a+1][][];
	for(int k=P.length;--k>=0;)
	{
	    if(k<=0)
	    {
		final double [][] s=P[k]=new double [2*neq_b+1][];
		for(int m=s.length;--m>=0;)
		{
		    s[m]=new double [m<=0?2*neq_c+1:(m<=neq_b?neq_c+1:1)];
		}
	    }
	    else if(k<=neq_a)
	    {
		final double [][] s=P[k]=new double [neq_b+1][];
		for(int m=s.length;--m>=0;)
		{
		    s[m]=new double [m<=0?neq_c+1:1];
		}
	    }
	    else
	    {
		P[k]=new double [1][1];
	    }
	}

	// re-use array
	darr[0]=da;
	darr[1]=db;
	darr[2]=dc;
	final double [][] matr=new double [ne][];
	final int [] indk=new int [nd],indl=new int [nd];
	for(int ek=0,ik=0;ek<nd;ek++)
	{
	    final double dwk=darr[ek];
	    double wk=1;
	    for(int k=1,nk=ndeq[ek];(indk[ek]=k)<=nk;k++,ik++)
	    {
		wk*=dwk;
		final double [] matrk=matr[ik]=new double [ne];
		final double normsk=norms[ik],pk=abc[indk[0]][indk[1]][indk[2]]/pt0;
		for(int el=0,il=0;el<nd;el++)
		{
		   final double dwl=darr[el];
		   double wl=1;
		   for(int l=1,nl=ndeq[el];(indl[el]=l)<=nl;l++,il++)
		   {
		       wl*=dwl;
		       final double pl=abc[indl[0]][indl[1]][indl[2]]/pt0,
			   pkl=abc[indk[0]+indl[0]][indk[1]+indl[1]][indk[2]+indl[2]]/pt0;
		       matrk[il]=(pkl-wl*pk+wk*(wl-pl))/normsk/norms[il];
		   }
		   indl[el]=0;
		}
	    }
	    indk[ek]=0;
	}

	final int [] ipvt=new int [ne];
	final int pinfo=com.polytechnik.utils.Linsystems.dgefa(matr,ne,ipvt);
	if(pinfo!=0)
	{
	    for(double [][] pp:P)
	    {
		for(double [] p:pp)
		{
		    java.util.Arrays.fill(p,Double.NaN);
		}
	    }
	    return P;
	}

	final double [] rpart=new double [ne];
	for(int ek=0,ik=0;ek<nd;ek++)
	{
	    final double dwk=darr[ek];
	    double wk=1;
	    for(int k=1,nk=ndeq[ek];(indk[ek]=k)<=nk;k++,ik++)
	    {
		wk*=dwk;
		rpart[ik]=1;
		com.polytechnik.utils.Linsystems.dgesl(matr,ne,ipvt,rpart,1);
		final double normsk=norms[ik];
		for(int el=0,il=0;el<nd;el++)
		{
		   final double dwl=darr[el];
		   double wl=1;
		   for(int l=1,nl=ndeq[el];(indl[el]=l)<=nl;l++,il++)
		   {
		       final double c=rpart[il]/normsk/norms[il],cwl=c*(wl*=dwl);
		       rpart[il]=0;
		       P[0][0][0]+=cwl*wk;
		       P[indk[0]][indk[1]][indk[2]]-=cwl;
		       P[indl[0]][indl[1]][indl[2]]-=wk*c;
		       P[indk[0]+indl[0]][indk[1]+indl[1]][indk[2]+indl[2]]+=c;
		   }
		   indl[el]=0;
		}
	    }
	    indk[ek]=0;
	}

	return P;
    }


    /*private*/ static double [] getPProjected(final double [][][] P,final double da,final double db)
    {
	final double [] pcoe=new double [P[0][0].length];
	double wa=1;
	for(double [][] ct : P)
	{
	    double wb=wa;
	    for(double [] cp : ct)
	    {
		for(int m=cp.length;--m>=0;)
		{
		    pcoe[m]+=wb*cp[m];
		}
		wb*=db;
	    }
	    wa*=da;
	}
	return pcoe;
    }


    private static void checkDPLCondition(final java.util.Random r,final QVMTradeIP ro,final double eps,final Trade [] tb,final int tsize)
    {
	final double pt0=ro.abc[0][0][0],stdevv=1.1,stdevt=1.03;
	for(int iv=0;iv<12;iv++)
	{
	    double dv=r.nextDouble()*stdevv;
	    for(int it=0;it<13;it++)
	    {
		double dt=r.nextDouble()*stdevt;
		for(int ip=0;ip<14;ip++)
		{
		    double dp=TradeEMoments.getRandomPrice(r)-ro.p_offset;
		    boolean roorig=false;
		    if(iv==0 && it==0 && ip==0)
		    {
			roorig=true;
			dv=0;
			dt=0;
			dp=ro.odata.trade_last.price-ro.odata.p_offset;
		    }
		    final double [][][] P=
			getPLCondititionCoefs(ro.abc,
					      dv,
					      dt,
					      dp,
					      ro.neq_v,
					      ro.neq_t,
					      ro.neq_p);
		    if(roorig && !ro.LOOK_FORWARD)
		    {
			UnitTests.checkDoubleArrayRel(ro.pcoefs,"ro.pcoefs",getPProjected(P,dv,dt),"getPProjected(P,dv,dt)",eps);
			System.err.println("checkDPLCondition: orig matched");
		    }


		    double s=0,sv=0,wkv=1;
		    for(int k=0;k<P.length;k++)
		    {
			final double [][] ctp=P[k];
			double st=0,wlt=1;
			for(int l=0;l<ctp.length;l++)
			{
			    final double [] cp=ctp[l];
			    double sp=0,wlp=1;
			    for(int m=0;m<cp.length;m++)
			    {
				s+=cp[m]*(ro.abc[k][l][m]/pt0);
				sp+=wlp*cp[m];
				wlp*=dp;
				if(cp[m]==0)
				{
				    System.err.println(
					"checkDPLCondition:"+
					"P["+k+"]["+l+"]["+m+"]="+cp[m]);
				}
			    }
			    st+=wlt*sp;
			    wlt*=dt;
			}
			sv+=wkv*st;
			wkv*=dv;
		    }
		    UnitTests.checkDoubleRel(s,"dIaver",ro.neq_v+ro.neq_t+ro.neq_p,"ro.neq_v+ro.neq_t+ro.neq_p",eps);
		    UnitTests.checkDoubleRel(sv,"I(0)",0,"0",eps);

		    final double ia0=Polynomials.getD0(getPProjected(P,dv,dt),dp);
		    UnitTests.checkDoubleRel(ia0,"ia0",0,"0",eps);
		}
	    }
	}
	System.err.println("checkDPLCondition: tests OK");
    }



    public static QVMTradeIP calculateQVMTradeIP(final int neq_v,
						 final int neq_t,
						 final int neq_p,
						 final long currenttime,
						 final long tau,
						 final BookItem [] tb,
						 final int tsize,
						 final com.polytechnik.utils.PolynomialRootsFinder rf,
						 final QVMTradeIP ref)
    {
	final TradeIPVMData o=TradeIPVMData.calculateTradeIPVMData(2*neq_v+1,
								   2*neq_t+1,
								   2*neq_p+1,
								   currenttime,
								   tau,
								   tb,
								   tsize,
								   ref!=null?ref.odata:null);
	return new QVMTradeIP(rf,neq_v,neq_t,neq_p,o);
    }


    double getDIDPDT_TOT(final double p)
    {
	return Polynomials.getD1(this.pcoefs,p-this.p_offset);
    }
    double getDIDT(final double p)
    {
	return Polynomials.getD0(this.pcoefs,p-this.p_offset);
    }

    double getPGMIN(final double p)
    {
	return QATradeIP.getPGMIN(p,this.p_imin,this.pcoefs,this.p_offset);
    }

    double getPLMIN(final double p)
    {
	return QATradeIP.getPLMIN(p,this.p_real_roots,this.p_offset);
    }

    double getIF(final double p)
    {
	return (Polynomials.getD0(this.pcoefs,p-this.p_offset)-this.i0)/this.i0;
    }

    double getIDP(final double p)
    {
	final double dp=p-this.p_offset;
	return Polynomials.getD0(this.pcoefs,dp)/Polynomials.getD1(this.pcoefs,dp);
    }

    private static void checkCalculationsQM(final java.util.Random r,
					    final int neq_t,final int neq_p,
					    final double p_offset,
					    final double [][][] abc,
					    final double eps)
    {
	for(int it=0;it<13;it++)
	{
	    double dt=r.nextDouble();
	    for(int ip=0;ip<14;ip++)
	    {
		double dp=TradeEMoments.getRandomPrice(r)-p_offset;
		final double [][][] PVM=getPLCondititionCoefs(abc,
							      0,
							      dt,
							      dp,
							      0,
							      neq_t,
							      neq_p);
		final double [][] PM=QMTradeIP.getPLCondititionCoefs(abc[0],
								     dt,
								     dp,
								     neq_t,
								     neq_p);
		for(int l=0;l<PM.length;l++)
		{
		    UnitTests.checkDoubleArrayRel1(PVM[0][l],"PVM[0]["+l+"]",PM[l],"PM["+l+"]",eps);
		}
	    }
	}
	System.err.println("QVMTradeIP:checkCalculationsQM(): matched OK");
    }


    private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_v,final int neq_t,final int neq_p)
    {
	final double eps=1e-5;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("QVMTradeIP:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);

	final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();

	final QVMTradeIP ro=calculateQVMTradeIP(neq_v,
						neq_t,
						neq_p,
						currenttime,
						tau,
						tb,
						tsize,
						rf,
						null);

	// regular check
	checkDPLCondition(r,ro,eps,tb,tsize);
	// QM check
	checkCalculationsQM(r,neq_t,neq_p,ro.p_offset,ro.abc,eps);


	System.err.println("QVMTradeIP:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
    }

    public static void main(String [] args)
    {
	System.err.println("QVMTradeIP::main(): starting tests; ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int neq_v=0;neq_v<=6;neq_v++)
	{
	    for(int neq_t=0;neq_t<=6;neq_t++)
	    {
		for(int neq_p=1;neq_p<=8;neq_p++)
		{
		    for(int i=0;i<3;i++)
		    {
			checkCalculations(100+r.nextInt(100),r,neq_v,neq_t,neq_p);
			checkCalculations(10010+r.nextInt(100010),r,neq_v,neq_t,neq_p);
		    }
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QVMTradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

} // END OF class QVMTradeIP






static class TradeIPData extends TradeIPRData {

    final double pt_sum,t_sum,pv_sum,v_sum;
    /*private*/ final double [] // here offset=<p>_t or <p>_v
	moments_pt,   // [m]=sum (p-offset)^m dt_i exp()
	moments_pv,   // [m]=sum (p-offset)^m v_i exp()
	moments_pt_tau,
	moments_pt_tau2,
	moments_pv_tau,
	moments_pv_tau2,

	moments_pv_buy,
	moments_pv_sell,
	moments_pt_buy,
	moments_pt_sell
	;

    // references to previos calculations
    private final BookItem trade_last,trade_last_buy,trade_last_sell;


    TradeIPData(final long currenttime,
		final long tau,

		final double pt_sum,
		final double t_sum,
		final double pv_sum,
		final double v_sum,
		final double p_offset,

		final double [] moments_pt,
		final double [] moments_pv,
		final double [] moments_pt_tau,
		final double [] moments_pt_tau2,
		final double [] moments_pv_tau,
		final double [] moments_pv_tau2,

		final double [] moments_pv_buy,
		final double [] moments_pv_sell,
		final double [] moments_pt_buy,
		final double [] moments_pt_sell,

		final long vol,
		final int np,

		// references
		final BookItem trade_last,
		final BookItem trade_last_buy,
		final BookItem trade_last_sell)
    {
	super(currenttime,tau,p_offset,vol,np);

	this.pt_sum=pt_sum;
	this.t_sum=t_sum;
	this.pv_sum=pv_sum;
	this.v_sum=v_sum;

	this.moments_pt=moments_pt;
	this.moments_pv=moments_pv;
	this.moments_pt_tau=moments_pt_tau;
	this.moments_pt_tau2=moments_pt_tau2;
	this.moments_pv_tau=moments_pv_tau;
	this.moments_pv_tau2=moments_pv_tau2;

	this.moments_pv_buy=moments_pv_buy;
	this.moments_pv_sell=moments_pv_sell;
	this.moments_pt_buy=moments_pt_buy;
	this.moments_pt_sell=moments_pt_sell;

	this.trade_last=trade_last;
	this.trade_last_buy=trade_last_buy;
	this.trade_last_sell=trade_last_sell;
    }

    private static TradeIPData calculateTestTradeIPData(final int N_EM,
							final long currenttime,
							final long tau,
							final BookItem [] tb,
							final int tsize,
							final boolean CALCULATE_VT)
    {
	final int np=N_EM,nvs=1+2*np;
	final double atau=tau;

	// calculate 0th & 1st moments
	double pt_sum=0,t_sum=0,pv_sum=0,v_sum=0;
	long prevt=currenttime,tinterval=currenttime,vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    //System.err.println("i="+i+" b.time="+b.time);
	    prevt=b.time;
	    final long dt=tinterval-prevt;

	    vol+=b.shares;
	    final double ex=Math.exp(-(currenttime-prevt)/atau),bwmw=ex*b.shares,twmw=ex*(dt/atau);

	    pt_sum+=b.price*twmw;
	    t_sum+=twmw;
	    pv_sum+=b.price*bwmw;
	    v_sum+=bwmw;

	    tinterval=prevt;
	}

	final double p_offset;
	if(CALCULATE_VT)
	{
	    p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);
	}
	else
	{
	    p_offset=v_sum>0?pv_sum/v_sum:(tsize>0?tb[tsize-1].price:0);
	}

	BookItem trade_last_buy=null,trade_last_sell=null;

	// calculate other moments
	final double []  // all 0 by default
	    moments_pt=new double [nvs],
	    moments_pv=new double [nvs],
	    moments_pt_tau=new double [nvs],
	    moments_pt_tau2=new double [nvs],
	    moments_pv_tau=new double [nvs],
	    moments_pv_tau2=new double [nvs],
	    moments_pv_buy=new double [nvs],
	    moments_pv_sell=new double [nvs],
	    moments_pt_buy=new double [nvs],
	    moments_pt_sell=new double [nvs];
	prevt=tinterval=currenttime;
	long tinterval_buy=currenttime,tinterval_sell=currenttime,dt_buy=0,dt_sell=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;
	    final double etau=-(currenttime-prevt)/atau,ex=Math.exp(etau),twmw_buy,twmw_sell;
	    final long dt=tinterval-prevt;
	    final double dp=b.price-p_offset,bwmw=ex*b.shares,dtatau=dt/atau,twmw=ex*dtatau;
	    dt_buy+=dt;
	    dt_sell+=dt;
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		if(dt_buy!=(tinterval_buy-prevt))
		{
		    throw new RuntimeException("dt_buy mismatch");
		}
		dt_buy=0;

		twmw_buy=ex*((tinterval_buy-prevt)/atau);
		twmw_sell=Double.NaN;
		if(trade_last_buy==null)
		{
		    trade_last_buy=b;
		}
		tinterval_buy=prevt;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		if(dt_sell!=(tinterval_sell-prevt))
		{
		    throw new RuntimeException("dt_sell mismatch");
		}
		dt_sell=0;

		twmw_buy=Double.NaN;
		twmw_sell=ex*((tinterval_sell-prevt)/atau);
		if(trade_last_sell==null)
		{
		    trade_last_sell=b;
		}
		tinterval_sell=prevt;
	    }
	    else
	    {
		twmw_buy=twmw_sell=Double.NaN;
	    }
	    double wpe=1;
	    for(int m=0;m<nvs;m++)
	    {
		moments_pt[m]+=wpe*twmw;
		moments_pv[m]+=wpe*bwmw;
		moments_pt_tau[m]+=wpe*twmw*etau;
		moments_pt_tau2[m]+=wpe*twmw*(etau*etau);
		moments_pv_tau[m]+=wpe*bwmw*etau;
		moments_pv_tau2[m]+=wpe*bwmw*(etau*etau);
		if(b.buy_sell_indicator==BookItem.BOOK_BUY)
		{
		    moments_pv_buy[m]+=wpe*bwmw;
		    moments_pt_buy[m]+=wpe*twmw_buy;
		}
		else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
		{
		    moments_pv_sell[m]+=wpe*bwmw;
		    moments_pt_sell[m]+=wpe*twmw_sell;
		}
		wpe*=dp;
	    }
	    tinterval=prevt;
	}

	return new TradeIPData(currenttime,
			       tau,

			       pt_sum,
			       t_sum,
			       pv_sum,
			       v_sum,
			       p_offset,

			       moments_pt,
			       moments_pv,
			       moments_pt_tau,
			       moments_pt_tau2,
			       moments_pv_tau,
			       moments_pv_tau2,

			       moments_pv_buy,
			       moments_pv_sell,
			       moments_pt_buy,
			       moments_pt_sell,

			       vol,
			       np,
			       tsize>0?tb[tsize-1]:null,
			       trade_last_buy,
			       trade_last_sell);
    }

    static double [] getPowersOfStdev(final double stdev_2,final int np)
    {
	final double [] norms=new double [np];
	final double stdev=Math.sqrt(stdev_2);
	double nr=1;
	for(int m=0,s=0;m<np;m++,s=1-s)
	{
	    if(s==0)
	    {
		norms[m]=nr;
	    }
	    else
	    {
		norms[m]=nr*stdev;
		nr*=stdev_2;
	    }
	}
	return norms;
    }


    static void setPriceMatrixAndRightPart(final double [][] matr,final double [] rpart,
					   final double [] moments,final double [] norm,
					   final double [] rmoments,
					   final int np)
    {
	final double lscale=moments[0],rscale=rmoments[0];
	for(int l=0;l<np;l++)
	{
	    final double norml=norm[l];
	    for(int k=0;k<np;k++)
	    {
		matr[l][k]=(moments[l+k]/lscale)/norml/norm[k];
	    }
	    rpart[l]=(rmoments[l]/rscale)/norml;
	}
    }

    private static void checkMatchingSubtype(final TradeIPData r,final TradeIPData ref,final byte st,final double eps)
    {
	final double [] subsel_t,subsel_v;
	if(st==BookItem.BOOK_BUY)
	{
	    subsel_t=r.moments_pt_buy;
	    subsel_v=r.moments_pv_buy;
	}
	else if(st==BookItem.BOOK_SELL)
	{
	    subsel_t=r.moments_pt_sell;
	    subsel_v=r.moments_pv_sell;
	}
	else
	{
	    throw new RuntimeException("st="+st);
	}
	final int nvs=ref.moments_pt.length;
	final double [] subsel_tm=new double [nvs],subsel_vm=new double [nvs];
	// shift offset to be equal to r.p_offset
	final double dpa=ref.p_offset-r.p_offset;
	for(int m=0;m<nvs;m++)
	{
	    double s_pt=0,s_pv=0,wpe=1;
	    long f=1;
	    for(int ms=m;ms>=0;ms--)
	    {
		final double a=f*wpe;
		s_pt+=a*ref.moments_pt[ms];
		s_pv+=a*ref.moments_pv[ms];
		wpe*=dpa;
		f*=ms;
		f/=1+m-ms;
	    }
	    subsel_tm[m]=s_pt;
	    subsel_vm[m]=s_pv;
	}

	// match moments given equal offsets
	for(int m=0;m<nvs;m++)
	{
	    UnitTests.checkDoubleRel(subsel_v[m],"subsel_v[m]",subsel_vm[m],"subsel_vm[m]",eps);
	    UnitTests.checkDoubleRel(subsel_t[m],"subsel_t[m]",subsel_tm[m],"subsel_tm[m]",eps);
	}
	System.err.println("TradeIPData:checkMatchingSubtype(): subset type="+st+" matching OK");
    }


    private static void checkMatching(final TradeIPData r,final TradeIPData ref)
    {
	final double eps=1e-5;

	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}
	if(r.trade_last!=ref.trade_last||
	   r.trade_last_buy!=ref.trade_last_buy||
	   r.trade_last_sell!=ref.trade_last_sell)
	{
	    throw new RuntimeException("references mismatch");
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",r.moments_pt[1]+r.p_offset*r.moments_pt[0],"r.moments_pt[1]+r.p_offset*r.moments_pt[0]",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",r.moments_pt[0],"r.moments_pt[0]",eps);

	UnitTests.checkDoubleRel(r.pv_sum,"r.pv_sum",ref.pv_sum,"ref.pv_sum",eps);
	UnitTests.checkDoubleRel(r.pv_sum,"r.pv_sum",r.moments_pv[1]+r.p_offset*r.moments_pv[0],"r.moments_pv[1]+r.p_offset*r.moments_pv[0]",eps);
	UnitTests.checkDoubleRel(r.v_sum,"r.v_sum",ref.v_sum,"ref.v_sum",eps);
	UnitTests.checkDoubleRel(r.v_sum,"r.v_sum",r.moments_pv[0],"r.moments_pv[0]",eps);

	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	final int nvs=r.moments_pt.length;
	for(int m=0;m<nvs;m++)
	{
	    UnitTests.checkDoubleRel(r.moments_pt[m],"r.moments_pt["+m+"]",ref.moments_pt[m],"ref.moments_pt["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pt_buy[m],"r.moments_pt_buy["+m+"]",ref.moments_pt_buy[m],"ref.moments_pt_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pt_sell[m],"r.moments_pt_sell["+m+"]",ref.moments_pt_sell[m],"ref.moments_pt_sell["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pt_tau[m],"r.moments_pt_tau["+m+"]",ref.moments_pt_tau[m],"ref.moments_pt_tau["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pt_tau2[m],"r.moments_pt_tau2["+m+"]",ref.moments_pt_tau2[m],"ref.moments_pt_tau2["+m+"]",eps);


	    // Since Oct 27 2010  changes in moments_pt_{buy,sell} definition the summ below no longer holds
	    //UnitTests.checkDoubleRel(r.moments_pt[m],"r.moments_pt[m]",r.moments_pt_buy[m]+r.moments_pt_sell[m],"r.moments_pt_buy[m]+r.moments_pt_sell[m]",eps);

	    UnitTests.checkDoubleRel(r.moments_pv[m],"r.moments_pv[m]",ref.moments_pv[m],"ref.moments_pv[m]",eps);
	    UnitTests.checkDoubleRel(r.moments_pv_buy[m],"r.moments_pv_buy[m]",ref.moments_pv_buy[m],"ref.moments_pv_buy[m]",eps);
	    UnitTests.checkDoubleRel(r.moments_pv_sell[m],"r.moments_pv_sell[m]",ref.moments_pv_sell[m],"ref.moments_pv_sell[m]",eps);
	    UnitTests.checkDoubleRel(r.moments_pv[m],"r.moments_pv[m]",r.moments_pv_buy[m]+r.moments_pv_sell[m],"r.moments_pv_buy[m]+r.moments_pv_sell[m]",eps);
	    UnitTests.checkDoubleRel(r.moments_pv_tau[m],"r.moments_pv_tau[m]",ref.moments_pv_tau[m],"ref.moments_pv_tau[m]",eps);
	    UnitTests.checkDoubleRel(r.moments_pv_tau2[m],"r.moments_pv_tau2[m]",ref.moments_pv_tau2[m],"ref.moments_pv_tau2[m]",eps);


	}
	System.err.println("TradeIPData:checkMatching(): matching OK");
    }


    static void checkMatchingTradeIPBS(final TradeIPBSData r,final TradeIPData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);
	final int np=r.np;
	for(int m=0;m<np;m++)
	{
	    UnitTests.checkDoubleRel(r.moments_p_buy_v_buy[m],"r.moments_p_buy_v_buy[m]",ref.moments_pv_buy[m],"ref.moments_pv_buy[m]",eps);
	    UnitTests.checkDoubleRel(r.moments_p_sell_v_sell[m],"r.moments_p_sell_v_sell[m]",ref.moments_pv_sell[m],"ref.moments_pv_sell[m]",eps);
	    for(int k=0;k<np;k++)
	    {
		UnitTests.checkDoubleRel(r.moments_p_buy_pt_buy[m][k],"r.moments_p_buy_pt_buy[m][k]",
			       ref.moments_pt_buy[m+k],"ref.moments_pt_buy[m+k]",eps);
		UnitTests.checkDoubleRel(r.moments_p_sell_pt_sell[m][k],"r.moments_p_sell_pt_sell[m][k]",
			       ref.moments_pt_sell[m+k],"ref.moments_pt_sell[m+k]",eps);
	    }
	}

	System.err.println("TradeIPData:checkMatchingTradeIPBS(): matching OK");
    }

    static void checkMatchingTradeIPS(final TradeIPSData r,final TradeIPData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);
	if(r.S_FACTOR_POWER==0)
	{
	    final int np=r.np;
	    for(int m=0;m<np;m++)
	    {
		UnitTests.checkDoubleRel(r.moments_pvS_p_buy_v_buy[m],"r.moments_pvS_p_buy_v_buy[m]",ref.moments_pv_buy[m],"ref.moments_pv_buy[m]",eps);
		UnitTests.checkDoubleRel(r.moments_pvS_p_sell_v_sell[m],"r.moments_pvS_p_sell_v_sell[m]",ref.moments_pv_sell[m],"ref.moments_pv_sell[m]",eps);
		UnitTests.checkDoubleRel(r.moments_pv[m],"r.moments_pv[m]",ref.moments_pv[m],"ref.moments_pv[m]",eps);
		for(int k=0;k<np;k++)
		{
		    if(false)
		    { // matches approximately
			UnitTests.checkDoubleRel(r.moments_ptSS_buy[m+k],"r.moments_ptSS_buy[m+k]",
				       ref.moments_pt_buy[m+k],"ref.moments_pt_buy[m+k]",eps);
			UnitTests.checkDoubleRel(r.moments_ptSS_sell[m+k],"r.moments_ptSS_sell[m+k]",
				       ref.moments_pt_sell[m+k],"ref.moments_pt_sell[m+k]",eps);
		    }
		    if(false)
		    { // off 1 term
			UnitTests.checkDoubleRel(r.moments_ptSS[m+k],"r.moments_ptSS["+(m+k)+"]",
				       ref.moments_pt[m+k],"ref.moments_pt["+(m+k)+"]",eps);
		    }
		}
	    }
	}
	else
	{
	    System.err.println("TradeIPData:checkMatchingTradeIPSData(): r.S_FACTOR_POWER="+r.S_FACTOR_POWER+" SKIPPING MOMENTS MATCH");
	}

	System.err.println("TradeIPData:checkMatchingTradeIPS(): matching OK");
    }

    static void checkMatchingTradeIPTBSData(final TradeIPTBSData r,final TradeIPData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);
	if(r.S_FACTOR_POWER==0)
	{
	    final int np=r.np;
	    for(int m=0;m<np;m++)
	    {

		UnitTests.checkDoubleRel(r.moments_p_buy_v_buy[m],"r.moments_p_buy_v_buy[m]",ref.moments_pv_buy[m],"ref.moments_pv_buy[m]",eps);
		UnitTests.checkDoubleRel(r.moments_p_sell_v_sell[m],"r.moments_p_sell_v_sell[m]",ref.moments_pv_sell[m],"ref.moments_pv_sell[m]",eps);
		UnitTests.checkDoubleRel(r.moments_pv[m],"r.moments_pv[m]",ref.moments_pv[m],"ref.moments_pv[m]",eps);
		for(int k=0;k<np;k++)
		{
		    if(m+k==1)
		    {
			// a kludge to avoid test failure
			UnitTests.checkDouble(r.moments_pt[m][k],"r.moments_pt[m][k]",ref.moments_pt[m+k],"ref.moments_pt[m+k]",eps);
		    }
		    else
		    {
			UnitTests.checkDoubleRel(r.moments_pt[m][k],"r.moments_pt[m][k]",ref.moments_pt[m+k],"ref.moments_pt[m+k]",eps);
		    }
		    /*
		      UnitTests.checkDoubleRel(r.moments_p_buy_pt_buy[m][k],"r.moments_p_buy_pt_buy[m][k]",
		      ref.moments_pt_buy[m+k],"ref.moments_pt_buy[m+k]",eps);
		      UnitTests.checkDoubleRel(r.moments_p_sell_pt_sell[m][k],"r.moments_p_sell_pt_sell[m][k]",
		      ref.moments_pt_sell[m+k],"ref.moments_pt_sell[m+k]",eps);
		    */
		}
	    }
	}
	else
	{
	    System.err.println("TradeIPData:checkMatchingTradeIPTBSData(): r.S_FACTOR_POWER="+r.S_FACTOR_POWER+" SKIPPING MOMENTS MATCH");
	}

	System.err.println("TradeIPData:checkMatchingTradeIPTBSData(): matching OK");
    }


    static void checkPCoefs(final double [] pcoefs,
			    final double [] moments,
			    final double [] norm,
			    final double [] rpart,
			    final double eps)
    {
	final int np=pcoefs.length;
	for(int m=0;m<np;m++)
	{
	    double s=0;
	    for(int k=0;k<np;k++)
	    {
		s+=pcoefs[k]*(moments[m+k]/moments[0])/norm[m]/norm[k];
	    }
	    final double r=(rpart[m]/rpart[0])/norm[m];
	    UnitTests.checkDoubleRel(s,"s."+m,r,"r."+m,eps);
	}
    }




static TradeIPData calculateTradeIPData(final int N_EM,
					final long currenttime,
					final long tau,
					final BookItem [] tb,
					final int tsize,
					final TradeIPData ref,
					final boolean CALCULATE_VT)
{
    final int np=N_EM;
    final double atau=tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime||np!=ref.np)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " np="+np+" ref.np="+ref.np+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }
    double pt_sum=0,t_sum=0,pv_sum=0,v_sum=0;
    long prevt=currenttime,tinterval=currenttime,vol=0,sumdt=0;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	final long dt=tinterval-prevt;
	vol+=b.shares;
	sumdt+=dt;
	final double ex=Math.exp(-(currenttime-prevt)/atau),bwmw=ex*b.shares,twmw=ex*(dt/atau);
	pt_sum+=b.price*twmw;
	t_sum+=twmw;
	pv_sum+=b.price*bwmw;
	v_sum+=bwmw;

	tinterval=prevt;
    }
    final double ref_exp_delay,ref_dtr_exp,tau_ref;
    if(bref!=null)
    { // optimization
	// add missed half slice
	sumdt+=tinterval-ref.currenttime;
	if(sumdt!=(currenttime-ref.currenttime))
	{
	    throw new RuntimeException("sumdt="+sumdt+" currenttime-ref.currenttime="+(currenttime-ref.currenttime));
	}
	ref_dtr_exp=Math.exp(-(currenttime-bref.time)/atau)*((tinterval-ref.currenttime)/atau);
	pt_sum+=ref_dtr_exp*bref.price;
	t_sum+=ref_dtr_exp;

	// previous interval
	tau_ref=-(currenttime-ref.currenttime)/atau;
	ref_exp_delay=Math.exp(tau_ref);
	pt_sum+=ref_exp_delay*ref.pt_sum;
	t_sum+=ref_exp_delay*ref.t_sum;
	pv_sum+=ref_exp_delay*ref.pv_sum;
	v_sum+=ref_exp_delay*ref.v_sum;
	vol+=ref.vol;
    }
    else
    {
	tau_ref=ref_exp_delay=ref_dtr_exp=0;
    }
    final double p_offset;
    final boolean offset_eq_average;
    if(CALCULATE_VT && t_sum>0)
    {
	p_offset=pt_sum/t_sum;
	offset_eq_average=true;
    }
    else if(!CALCULATE_VT && v_sum>0)
    {
	p_offset=pv_sum/v_sum;
	offset_eq_average=true;
    }
    else
    {
	p_offset=tsize>0?tb[tsize-1].price:0;
	offset_eq_average=false;
    }


    final int nvs=1+2*np;
    final double []  // all 0 by default
	moments_pt=new double [nvs],
	moments_pv=new double [nvs],
	moments_pt_tau=new double [nvs],
	moments_pt_tau2=new double [nvs],
	moments_pv_tau=new double [nvs],
	moments_pv_tau2=new double [nvs],
	moments_pv_buy=new double [nvs],
	moments_pv_sell=new double [nvs],
	moments_pt_buy=new double [nvs],
	moments_pt_sell=new double [nvs];
    prevt=tinterval=currenttime;
    BookItem trade_last_buy=null,trade_last_sell=null;
    long tinterval_buy=currenttime,tinterval_sell=currenttime;
    // calculate with given offset. using average as offset reduces rounding errors & degeneracy.
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	final int tp=b.buy_sell_indicator;
	final double etau=(prevt-currenttime)/atau,ex=Math.exp(etau),dp=b.price-p_offset,
	    bwmw=ex*b.shares,twmw=ex*((tinterval-prevt)/atau),twmw_buy,twmw_sell;
	if(tp==BookItem.BOOK_BUY)
	{
	    twmw_buy=ex*((tinterval_buy-prevt)/atau);
	    twmw_sell=Double.NaN;
	    if(trade_last_buy==null) trade_last_buy=b;
	    tinterval_buy=prevt;
	}
	else if(tp==BookItem.BOOK_SELL)
	{
	    twmw_buy=Double.NaN;
	    twmw_sell=ex*((tinterval_sell-prevt)/atau);
	    if(trade_last_sell==null) trade_last_sell=b;
	    tinterval_sell=prevt;
	}
	else
	{
	    twmw_buy=twmw_sell=Double.NaN;
	}
	double wpe=1;
	for(int m=0;m<nvs;m++)
	{
	    final double xa=bwmw*wpe;
	    moments_pv[m]+=xa;
	    moments_pt[m]+=twmw*wpe;
	    moments_pv_tau[m]+=xa*etau;
	    moments_pv_tau2[m]+=xa*(etau*etau);
	    moments_pt_tau[m]+=twmw*wpe*etau;
	    moments_pt_tau2[m]+=twmw*wpe*(etau*etau);
	    if(tp==BookItem.BOOK_BUY)
	    {
		moments_pv_buy[m]+=xa;
		moments_pt_buy[m]+=twmw_buy*wpe;
	    }
	    else if(tp==BookItem.BOOK_SELL)
	    {
		moments_pv_sell[m]+=xa;
		moments_pt_sell[m]+=twmw_sell*wpe;
	    }
	    wpe*=dp;
	}
	tinterval=prevt;
    }
    if(bref!=null)
    { // optimization

	// previous interval & half slice
	final double dpt=bref.price-p_offset,dpa=ref.p_offset-p_offset,dpt_buy,dpt_sell,ref_dtr_exp_buy,ref_dtr_exp_sell;
	if(ref.trade_last_buy!=null)
	{
	    ref_dtr_exp_buy=Math.exp(-(currenttime-ref.trade_last_buy.time)/atau)*((tinterval_buy-ref.currenttime)/atau);
	    dpt_buy=ref.trade_last_buy.price-p_offset;
	}
	else
	{
	    dpt_buy=ref_dtr_exp_buy=0;
	}
	if(ref.trade_last_sell!=null)
	{
	    ref_dtr_exp_sell=Math.exp(-(currenttime-ref.trade_last_sell.time)/atau)*((tinterval_sell-ref.currenttime)/atau);
	    dpt_sell=ref.trade_last_sell.price-p_offset;
	}
	else
	{
	    dpt_sell=ref_dtr_exp_sell=0;
	}
	final double etauref=(bref.time-currenttime)/atau;
	double wpest=1,wpest_buy=1,wpest_sell=1;
	for(int m=0;m<nvs;m++)
	{
	    // add missed half slice
	    moments_pt[m]+=wpest*ref_dtr_exp;
	    moments_pt_tau[m]+=wpest*ref_dtr_exp*etauref;
	    moments_pt_tau2[m]+=wpest*ref_dtr_exp*(etauref*etauref);
	    moments_pt_buy[m]+=wpest_buy*ref_dtr_exp_buy;
	    moments_pt_sell[m]+=wpest_sell*ref_dtr_exp_sell;

	    // convert old interval to new p_offest basis
	    long f=1;
	    double s_pt=0,s_pv=0,s_pv_buy=0,s_pv_sell=0,s_pt_buy=0,s_pt_sell=0,s_pt_tau=0,s_pt_tau2=0,s_pv_tau=0,s_pv_tau2=0;
	    for(int ms=0;ms<=m;f*=m-ms,f/=(++ms))
	    {
		final double a=f;
		s_pt=dpa*s_pt+a*ref.moments_pt[ms];
		s_pv=dpa*s_pv+a*ref.moments_pv[ms];
		s_pt_tau=dpa*s_pt_tau+a*ref.moments_pt_tau[ms];
		s_pt_tau2=dpa*s_pt_tau2+a*ref.moments_pt_tau2[ms];
		s_pv_tau=dpa*s_pv_tau+a*ref.moments_pv_tau[ms];
		s_pv_tau2=dpa*s_pv_tau2+a*ref.moments_pv_tau2[ms];
		s_pv_buy=dpa*s_pv_buy+a*ref.moments_pv_buy[ms];
		s_pv_sell=dpa*s_pv_sell+a*ref.moments_pv_sell[ms];
		s_pt_buy=dpa*s_pt_buy+a*ref.moments_pt_buy[ms];
		s_pt_sell=dpa*s_pt_sell+a*ref.moments_pt_sell[ms];
	    }
	    // add previous interval in new p_offest basis
	    moments_pt[m]+=ref_exp_delay*s_pt;
	    moments_pv[m]+=ref_exp_delay*s_pv;
	    moments_pt_tau[m]+=ref_exp_delay*(s_pt_tau+tau_ref*s_pt);
	    moments_pt_tau2[m]+=ref_exp_delay*(s_pt_tau2+tau_ref*(2*s_pt_tau+tau_ref*s_pt));
	    moments_pv_tau[m]+=ref_exp_delay*(s_pv_tau+tau_ref*s_pv);
	    moments_pv_tau2[m]+=ref_exp_delay*(s_pv_tau2+tau_ref*(2*s_pv_tau+tau_ref*s_pv));
	    moments_pv_buy[m]+=ref_exp_delay*s_pv_buy;
	    moments_pv_sell[m]+=ref_exp_delay*s_pv_sell;
	    moments_pt_buy[m]+=ref_exp_delay*s_pt_buy;
	    moments_pt_sell[m]+=ref_exp_delay*s_pt_sell;

	    wpest*=dpt;
	    wpest_buy*=dpt_buy;
	    wpest_sell*=dpt_sell;
	}
    }
    if(offset_eq_average)
    {
	// reduce rounding errors for known values
	// <p-<p>>=0
	if(CALCULATE_VT)
	{
	    moments_pt[1]=0;
	}
	else
	{
	    moments_pv[1]=0;
	}
    }

    return new TradeIPData(currenttime,
			   tau,

			   pt_sum,
			   t_sum,
			   pv_sum,
			   v_sum,
			   p_offset,

			   moments_pt,
			   moments_pv,
			   moments_pt_tau,
			   moments_pt_tau2,
			   moments_pv_tau,
			   moments_pv_tau2,

			   moments_pv_buy,
			   moments_pv_sell,
			   moments_pt_buy,
			   moments_pt_sell,

			   vol,
			   np,
			   tsize>0?tb[tsize-1]:bref,
			   trade_last_buy!=null?trade_last_buy:(ref!=null?ref.trade_last_buy:null),
			   trade_last_sell!=null?trade_last_sell:(ref!=null?ref.trade_last_sell:null));
}


} // END OF class TradeIPData

static class DIPTradeIP {
    static double [] getPCoefs(final double [] pta,final double [] ptb,final double [] pva,final double [] pvb,final int np)
    {
	final double [][] matr=new double [np][np];
	final double [] rpart=new double [np];
	final double pta0=pta[0],ptb0=ptb[0],pva0=pva[0],pvb0=pvb[0];
	for(int k=0;k<np;k++)
	{
	    final int k1=k+1;
	    for(int l=0;l<np;l++)
	    {
		final int l1=l+1;
		matr[k][l]=pta[l1+k1]*ptb0-pta[l1]*ptb[k1]+pta0*ptb[l1+k1]-pta[k1]*ptb[l1];
	    }
	    rpart[k]=pva[k1]*ptb0-pva0*ptb[k1]+pta0*pvb[k1]-pta[k1]*pvb0;
	}
	TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
	for(int k=0;k<np;k++)
	{
	    rpart[k]*=k+1;
	}
	return rpart;
    }

    final DPTradeIP kdiff,diff,diff3,buy,sell;
    private final TradeIPData odata;
    DIPTradeIP(final TradeIPData odata)
    {
	this.odata=odata;
	this.kdiff=new DPTradeIP(odata) {
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np][np];
		    final double [] rpart=new double [np];
		    final double [] pt=odata.moments_pt,pv=odata.moments_pv;
		    for(int k=0;k<np;k++)
		    {
			final int k1=k+1;
			for(int l=0;l<np;l++)
			{
			    final int l1=l+1;
			    matr[k][l]=pt[0]*pt[l1+k1]-pt[l1]*pt[k1];
			}
			rpart[k]=pv[k1]*pt[0]-pv[0]*pt[k1];
		    }
		    TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
		    for(int k=0;k<np;k++)
		    {
			rpart[k]*=k+1;
		    }
		    return rpart;
		}
	    };
	this.diff=new DPTradeIP(odata) {
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    return DIPTradeIP.getPCoefs(odata.moments_pt,odata.moments_pt,odata.moments_pv,odata.moments_pv,np);
		}
	    };
	this.diff3=new DPTradeIP(odata) {
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    return DIPTradeIP.getPCoefs(odata.moments_pt_sell,odata.moments_pt_buy,odata.moments_pv_sell,odata.moments_pv_buy,np);
		}
	    };
	this.buy=new DPTradeIP(odata) {
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    return DIPTradeIP.getPCoefs(odata.moments_pt_buy,odata.moments_pt_buy,odata.moments_pv_buy,odata.moments_pv_buy,np);
		}
	    };
	this.sell=new DPTradeIP(odata) {
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    return DIPTradeIP.getPCoefs(odata.moments_pt_sell,odata.moments_pt_sell,odata.moments_pv_sell,odata.moments_pv_sell,np);
		}
	    };
    }

    static DIPTradeIP calculateDIPTradeIP(final int N_EM,
					  final long currenttime,
					  final long tau,
					  final BookItem [] tb,
					  final int tsize,
					  final DIPTradeIP ref)
    {
	return new DIPTradeIP(TradeIPData.calculateTradeIPData(N_EM+1,
							       currenttime,
							       tau,
							       tb,
							       tsize,
							       ref!=null?ref.odata:null,
							       true));

    }


} // END OF class DIPTradeIP






static class PTVV implements TradingEventsCheck {

    private static final boolean FORCE_FAST_EXIT=false;

    final double pm,pm_l,pm_h,pmo,d0,pe,pe_l,pe_h,pmm,p_offset,sigma;
    final double [] p_imin,i_imin,p_imax,i_imax,p_i_eq_imax_low,p_i_eq_imax_high,p_i_eq_imin_low,p_i_eq_imin_high;
    private final int tdir;

    PTVV(final double d0,
	 final double p_offset,
	 final double sigma,
	 final double [] p_imin,
	 final double [] i_imin,
	 final double [] p_i_eq_imin_low,
	 final double [] p_i_eq_imin_high,

	 final double [] p_imax,
	 final double [] i_imax,
	 final double [] p_i_eq_imax_low,
	 final double [] p_i_eq_imax_high)
    {
	this.d0=d0;
	this.p_offset=p_offset;
	this.sigma=sigma;
	this.p_imin=p_imin;
	this.i_imin=i_imin;
	this.p_i_eq_imin_low=p_i_eq_imin_low;
	this.p_i_eq_imin_high=p_i_eq_imin_high;
	this.p_imax=p_imax;
	this.i_imax=i_imax;
	this.p_i_eq_imax_low=p_i_eq_imax_low;
	this.p_i_eq_imax_high=p_i_eq_imax_high;


	if(p_imax.length>0)
	{
	    this.tdir=p_imax.length==p_imin.length?(p_imax[0]<p_imin[0]?1:-1):0;
	}
	else if(p_imin.length==0 && p_imax.length==0)
	{
	    this.tdir=d0>0?1:(d0<0?-1:0);
	}
	else
	{
	    this.tdir=0;
	}

	int ptr_pmm=-1;
	for(int k=p_imax.length;--k>=0;)
	{
	    if(ptr_pmm<0||i_imax[k]<i_imax[ptr_pmm])
	    {
		ptr_pmm=k;
	    }
	}
	this.pmm=ptr_pmm<0?Double.NaN:p_imax[ptr_pmm];

	int ptr_pe=-1;
	for(int k=p_imin.length;--k>=0;)
	{
	    if(ptr_pe<0||i_imin[k]>i_imin[ptr_pe])
	    {
		ptr_pe=k;
	    }
	}
	if(ptr_pe<0)
	{
	    this.pm=this.pm_l=this.pm_h=this.pmo=this.pe=this.pe_l=this.pe_h=Double.NaN;

	}
	else
	{
	    this.pe=p_imin[ptr_pe];
	    this.pe_l=p_i_eq_imin_low[ptr_pe];
	    this.pe_h=p_i_eq_imin_high[ptr_pe];

	    int ptr_pm=-1,ptr_pmo=-1;
	    for(int k=0;k<=p_imax.length;k++)
	    {
		if((k<=0||p_imax[k-1]<pe) && (k>=p_imax.length||pe<=p_imax[k]))
		{
		    if(k<=0)
		    {
			ptr_pm=p_imax.length>0?0:-1;
		    }
		    else if(k>=p_imax.length)
		    {
			ptr_pm=p_imax.length-1;
		    }
		    else
		    {
			if(i_imax[k-1]<i_imax[k])
			{
			    ptr_pm=k-1;
			    ptr_pmo=k;
			}
			else
			{
			    ptr_pm=k;
			    ptr_pmo=k-1;
			}
		    }
		    break;
		}
	    }
	    if(ptr_pm<0)
	    {
		this.pm=this.pm_l=this.pm_h=Double.NaN;
	    }
	    else
	    {
		this.pm=p_imax[ptr_pm];
		this.pm_l=p_i_eq_imax_low[ptr_pm];
		this.pm_h=p_i_eq_imax_high[ptr_pm];
	    }
	    this.pmo=ptr_pmo<0?Double.NaN:p_imax[ptr_pmo];
	}
    }

    int getTID()
    {
	final int min=this.i_imin.length<=1?0:(this.i_imin[0]<this.i_imin[1]?1:-1),
	    max=this.i_imax.length<=1?0:(this.i_imax[0]<this.i_imax[1]?1:-1),
	    nex=this.i_imin.length+this.i_imax.length;

	return
	    ((this.tdir*(3*this.tdir+1))/2)*1000+
	    ((min*(3*min+1))/2)*100+
	    ((max*(3*max+1))/2)*10+
	    nex;
    }

    public String dumpData()
    {
	final StringBuilder buf=new StringBuilder();
	buf.append("PTVV: getTID()="+getTID()+" pm="+pm+" pm_l="+pm_l+" pm_h="+pm_h+" d0="+d0+" tdir="+tdir+"\n");
	for(int k=0;k<p_imin.length;k++)
	{
	    buf.append("min_"+k+": "+" p_imin="+p_imin[k]+" i_imin="+i_imin[k]+"\n");
	}
	for(int k=0;k<p_imax.length;k++)
	{
	    buf.append("max_"+k+": "+" p_imax="+p_imax[k]+" i_imax="+i_imax[k]+
		       " p_i_eq_imax_low="+p_i_eq_imax_low[k]+
		       " p_i_eq_imax_high="+p_i_eq_imax_high[k]+"\n");
	}
	return buf.toString();
    }


    public boolean checkEnterBuy(final double p)
    {
	if(!(tdir>0) ||p_imin.length!=1)
	{
	    return false;
	}
	return p<p_imin[0];
    }

    public boolean checkExitBuy(final double p)
    {
	return p_imin.length==1 && p>p_imin[0];
    }


    public boolean checkEnterSell(final double p)
    {
	if(!(tdir>0)||p_imin.length!=1)
	{
	    return false;
	}
	return p>p_imin[0];
    }


    public boolean checkExitSell(final double p)
    {
	return p_imin.length==1 && p<p_imin[0];
    }



} // END OF class PTVV





static class PTopology implements TradingEventsCheck {
    private static final boolean REQUIRE_EXIT_AVAILABLE_ON_ENTER=true;

    final double pm,pm_l,pm_h,d0;
    private final double [] p_imin,i_imin,p_imax,i_imax,p_i_eq_imax_low,p_i_eq_imax_high;
    private final int tdir;

    PTopology(final double d0,
	      final double [] p_imin,
	      final double [] i_imin,
	      final double [] p_imax,
	      final double [] i_imax,

	      final double [] p_i_eq_imax_low,
	      final double [] p_i_eq_imax_high)
    {
	this.d0=d0;
	this.p_imin=p_imin;
	this.i_imin=i_imin;
	this.p_imax=p_imax;
	this.i_imax=i_imax;
	this.p_i_eq_imax_low=p_i_eq_imax_low;
	this.p_i_eq_imax_high=p_i_eq_imax_high;

	if(!(p_imin.length==1 && p_imax.length==1) &&
	   !(p_imin.length==2 && p_imax.length==2))
	{
	    this.pm=this.pm_l=this.pm_h=Double.NaN;
	    if(p_imin.length==0 && p_imax.length==0)
	    {
		this.tdir=d0>0?1:(d0<0?-1:0);
	    }
	    else
	    {
		this.tdir=0;
	    }
	}
	else
	{
	    final int i;
	    if(p_imax[0]<p_imin[0])
	    {
		i=p_imax.length-1;
		this.tdir=1;
	    }
	    else
	    {
		i=0;
		this.tdir=-1;
	    }
	    this.pm=p_imax[i];
	    this.pm_l=p_i_eq_imax_low[i];
	    this.pm_h=p_i_eq_imax_high[i];
	}
    }
    public boolean checkExitSell(final double p)
    {
	// buy to cover short
	return p<this.pm_l || (this.tdir<0 && this.i_imin.length<=0);
    }
    public boolean checkExitBuy(final double p)
    {
	// sell existing position
	return p>this.pm_h || (this.tdir>0 && this.i_imin.length<=0);
    }

    public boolean checkEnterBuy(final double p)
    {
	return
	    (p<this.pm_h||!REQUIRE_EXIT_AVAILABLE_ON_ENTER) && this.i_imin.length>0 &&
	    this.pm<p && p<this.p_imin[this.p_imin.length-1] &&
	    (this.i_imin.length<=1 || this.i_imin[0]<this.i_imin[1]);
    }

    public boolean checkEnterSell(final double p)
    {
	return
	    (p>this.pm_l||!REQUIRE_EXIT_AVAILABLE_ON_ENTER) && this.i_imin.length>0 &&
	    this.p_imin[0]<p && p<this.pm &&
	    (this.i_imin.length<=1 || this.i_imin[0]>this.i_imin[1]);
    }

    int getTID()
    {
	if(this.tdir==0)
	{
	    return 0;
	}

	final int min=this.i_imin.length<=1?0:(this.i_imin[0]<this.i_imin[1]?1:-1),
	    max=this.i_imax.length<=1?0:(this.i_imax[0]<this.i_imax[1]?1:-1),
	    nex=this.i_imin.length+this.i_imax.length;

	return
	    ((this.tdir*(3*this.tdir+1))/2)*1000+
	    ((min*(3*min+1))/2)*100+
	    ((max*(3*max+1))/2)*10+
	    nex;
    }

    public String dumpData()
    {
	final StringBuilder buf=new StringBuilder();
	buf.append("PTopology: getTID()="+getTID()+" pm="+pm+" pm_l="+pm_l+" pm_h="+pm_h+" d0="+d0+" tdir="+tdir+"\n");
	for(int k=0;k<p_imin.length;k++)
	{
	    buf.append("min_"+k+": "+" p_imin="+p_imin[k]+" i_imin="+i_imin[k]+"\n");
	}
	for(int k=0;k<p_imax.length;k++)
	{
	    buf.append("max_"+k+": "+" p_imax="+p_imax[k]+" i_imax="+i_imax[k]+
		       " p_i_eq_imax_low="+p_i_eq_imax_low[k]+
		       " p_i_eq_imax_high="+p_i_eq_imax_high[k]+"\n");
	}
	return buf.toString();
    }

} // END OF class PTopology



static class TCode {
    final double p,pm,pm_l,pm_h,d0;
    final int cmd,code;
    final boolean flag_lh;
    TCode(final double p,
	  final double d0,
	  final double pm,
	  final double pm_l,
	  final double pm_h,
	  final int cmd,
	  final boolean flag_lh,
	  final int code)
    {
	this.p=p;
	this.d0=d0;
	this.pm=pm;
	this.pm_l=pm_l;
	this.pm_h=pm_h;
	this.cmd=cmd;
	this.flag_lh=flag_lh;
	this.code=code;
    }

static final int S_BUY=5891,S_BUY_EXIT=5892,S_SELL=5893,S_SELL_EXIT=5894;
static TCode getSTCode(final int cmd,
		       final double p,
		       final double d0,
		       final double [] p_imin,
		       final double [] i_imin,
		       final double [] p_imax,
		       final double [] i_imax,

		       final double [] p_i_eq_imax_low,
		       final double [] p_i_eq_imax_high)
{
    if(cmd!=S_BUY&&cmd!=S_BUY_EXIT&&cmd!=S_SELL&&cmd!=S_SELL_EXIT)
    {
	throw new IllegalArgumentException();
    }
    int code=0;

    if(p_imin.length==0 && p_imax.length==0)
    {
	if(d0>0)
	{
	    if(cmd==S_BUY_EXIT)
	    {
		code=75140;
	    }
	    return new TCode(p,d0,Double.NaN,Double.NaN,Double.NaN,cmd,false,code);
	}
	if(d0<0)
	{
	    if(cmd==S_SELL_EXIT)
	    {
		code=85140;
	    }
	    return new TCode(p,d0,Double.NaN,Double.NaN,Double.NaN,cmd,true,code);
	}
	return null;
    }

    if(p_imin.length==1 && p_imax.length==1)
    {
	if(p_imax[0]<p_imin[0])
	{
	    if(cmd==S_BUY)
	    {
		if(p_imax[0]<p && p<p_imin[0]) code=70200;
	    }
	    else if(cmd==S_BUY_EXIT)
	    {
		if(p>p_i_eq_imax_high[0]) code=70360;
	    }
	    return new TCode(p,d0,p_imax[0],p_i_eq_imax_low[0],p_i_eq_imax_high[0],cmd,false,code);
	}
	else
	{
	    if(cmd==S_SELL)
	    {
		if(p_imin[0]<p && p<p_imax[0]) code=80200;
	    }
	    else if(cmd==S_SELL_EXIT)
	    {
		if(p<p_i_eq_imax_low[0]) code=80150;
	    }
	    return new TCode(p,d0,p_imax[0],p_i_eq_imax_low[0],p_i_eq_imax_high[0],cmd,true,code);
	}

    }

    if(p_imin.length!=2 || p_imax.length!=2)
    {
	return null;
    }
    final double L0=i_imin[0],L1=i_imin[1],H0=i_imax[0],H1=i_imax[1];
    if(p_imax[1]<p_imin[1])
    {
	if(H0<H1 && L0<L1)
	{
	  if(cmd==S_BUY)
	  {
	      if(p_imax[1]<p && p<p_imin[1]) code=71400;
	  }
	  else if(cmd==S_BUY_EXIT)
	  {
	      if(p>p_i_eq_imax_high[1]) code=71561;
	  }
	}
	else if(H0<H1 && L0>L1)
	{
	    if(cmd==S_SELL)
	    {
		if(!PTopology.REQUIRE_EXIT_AVAILABLE_ON_ENTER)
		{
		    if(p_imin[0]<p && p<p_imax[1]) code=72300;
		}
	    }
	    else if(cmd==S_BUY_EXIT)
	    {
		if(p>p_i_eq_imax_high[1]) code=72561;
	    }
	}
	else if(H0>H1 && L0>L1)
	{
	    if(cmd==S_SELL)
	    {
		if(p_imin[0]<p && p<p_imax[1]) code=73300;
	    }
	    else if(cmd==S_SELL_EXIT)
	    {
		if(p<p_i_eq_imax_low[1]) code=73251;
	    }
	    else if(cmd==S_BUY_EXIT)
	    {
		if(p>p_i_eq_imax_high[1]) code=73561;
	    }
	}
	else if(H0>H1 && L0<L1)
	{
	    if(cmd==S_BUY)
	    {
		if(p_imax[1]<p && p<p_imin[1]) code=74400;
	    }
	    else if(cmd==S_BUY_EXIT)
	    {
		if(p>p_i_eq_imax_high[1]) code=74561;
	    }
	    else if(cmd==S_SELL_EXIT)
	    {
		if(p<p_i_eq_imax_low[1]) code=74251;
	    }
	}
	return new TCode(p,d0,p_imax[1],p_i_eq_imax_low[1],p_i_eq_imax_high[1],cmd,false,code);
    }
    else
    {
	if(H0>H1 && L0>L1)
	{
	    if(cmd==S_SELL)
	    {
		if(p_imin[0]<p && p<p_imax[0]) code=81200;
	    }
	    else if(cmd==S_SELL_EXIT)
	    {
		if(p<p_i_eq_imax_low[0]) code=81150;
	    }
	}
	else if(H0>H1 && L0<L1)
	{
	    if(cmd==S_BUY)
	    {
		if(!PTopology.REQUIRE_EXIT_AVAILABLE_ON_ENTER)
		{
		    if(p_imax[0]<p && p<p_imin[1]) code=82300;
		}
	    }
	    else if(cmd==S_SELL_EXIT)
	    {
		if(p<p_i_eq_imax_low[0]) code=82150;
	    }
	}
	else if(H0<H1 && L0<L1)
	{
	    if(cmd==S_BUY)
	    {
		if(p_imax[0]<p && p<p_imin[1]) code=83300;
	    }
	    else if(cmd==S_BUY_EXIT)
	    {
		if(p>p_i_eq_imax_high[0]) code=83460;
	    }
	    else if(cmd==S_SELL_EXIT)
	    {
		if(p<p_i_eq_imax_low[0]) code=83150;
	    }
	}
	else if(H0<H1 && L0>L1)
	{
	    if(cmd==S_SELL)
	    {
		if(p_imin[0]<p && p<p_imax[0]) code=84200;
	    }
	    else if(cmd==S_SELL_EXIT)
	    {
		if(p<p_i_eq_imax_low[0]) code=84150;
	    }
	    else if(cmd==S_BUY_EXIT)
	    {
		if(p>p_i_eq_imax_high[0]) code=84460;
	    }
	}
	return new TCode(p,d0,p_imax[0],p_i_eq_imax_low[0],p_i_eq_imax_high[0],cmd,true,code);
    }
}


private static void checkCMD(final TCode r,
			     final PTopology pt,
			     final DPTradeIP dip,
			     final double eps)
{

    final double eps2=eps*50; // a kludge for unit test
    if(pt.p_imax.length>0 && pt.p_imin.length>0 && (pt.p_imax[0]<pt.p_imin[0] != !r.flag_lh))
    {
	throw new RuntimeException("pt="+pt.dumpData()+" r.flag_lh="+r.flag_lh);
    }

    UnitTests.checkDoubleRel(r.d0,"r.d0",pt.d0,"pt.d0",eps);
    UnitTests.checkDoubleRel(r.pm,"r.pm",pt.pm,"pt.pm",eps);
    UnitTests.checkDoubleRel(r.pm_l,"r.pm_l",pt.pm_l,"pt.pm_l",eps);
    UnitTests.checkDoubleRel(r.pm_h,"r.pm_h",pt.pm_h,"pt.pm_h",eps);
    for(int k=0;k<pt.p_imin.length;k++)
    {
	final double px=pt.p_imin[k];
	final double im=dip.getI1(px),imscale=Math.abs(im)>1?Math.abs(im):1;
	UnitTests.checkDoubleRel(im,"dip.getI1(p_imin[k])",pt.i_imin[k],"pt.i_imin[k]",eps);
	UnitTests.checkDoubleRel(dip.getDI1(px)/imscale,"dip.getDI1(p_imin[k])/imscale",0,"0",eps);
	UnitTests.checkDoubleRel(Math.signum(dip.getD2I1(px)),"Math.signum(dip.getD2I1(p_imin[k]))",1,"1",eps);
    }
    for(int k=0;k<pt.p_imax.length;k++)
    {
	final double px=pt.p_imax[k];
	final double im=dip.getI1(px),imscale=Math.abs(im)>1?Math.abs(im):1;
	UnitTests.checkDoubleRel(im,"dip.getI1(p_imax[k])",pt.i_imax[k],"pt.i_imax[k]",eps);
	UnitTests.checkDoubleRel(dip.getDI1(px)/imscale,"dip.getDI1(p_imax[k])/imscale",0,"0",eps);
	UnitTests.checkDoubleRel(Math.signum(dip.getD2I1(px)),"Math.signum(dip.getD2I1(p_imax[k]))",-1,"-1",eps);
	if(!Double.isNaN(pt.p_i_eq_imax_low[k]))
	{
	    UnitTests.checkDoubleRel(im,"dip.getI1(p_imax[k])",dip.getI1(pt.p_i_eq_imax_low[k]),"dip.getI1(pt.p_i_eq_imax_low[k])",eps2);
	    UnitTests.checkDoubleRel(Math.signum(dip.getDI1(pt.p_i_eq_imax_low[k])),"Math.signum(dip.getDI1(pt.p_i_eq_imax_low[k]))",-1,"-1",eps);
	}
	if(!Double.isNaN(pt.p_i_eq_imax_high[k]))
	{
	    UnitTests.checkDoubleRel(im,"dip.getI1(p_imax[k])",dip.getI1(pt.p_i_eq_imax_high[k]),"dip.getI1(pt.p_i_eq_imax_high[k])",eps2);
	    UnitTests.checkDoubleRel(Math.signum(dip.getDI1(pt.p_i_eq_imax_high[k])),"Math.signum(dip.getDI1(pt.p_i_eq_imax_high[k]))",1,"1",eps);
	}
    }

    final double ipm=dip.getI1(r.pm),imscale=Math.abs(ipm)>1?Math.abs(ipm):1;
    if(!Double.isNaN(r.pm))
    {
	UnitTests.checkDoubleRel(dip.getDI1(r.pm)/imscale,"dip.getDI1(r.pm)/imscale",0,"0",eps);
	UnitTests.checkDoubleRel(Math.signum(dip.getD2I1(r.pm)),"Math.signum(dip.getD2I1(r.pm))",-1,"-1",eps);
    }
    if(!Double.isNaN(r.pm_l))
    {
	UnitTests.checkDoubleRel(dip.getI1(r.pm_l),"dip.getI1(r.pm_l)",ipm,"ipm",eps2);
	UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.pm_l)),"Math.signum(dip.getDI1(r.pm_l))",-1,"-1",eps);

    }
    if(!Double.isNaN(r.pm_h))
    {
	UnitTests.checkDoubleRel(dip.getI1(r.pm_h),"dip.getI1(r.pm_h)",ipm,"ipm",eps2);
	UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.pm_h)),"Math.signum(dip.getDI1(r.pm_h))",1,"1",eps);
    }


    final int code=r.code,dir=code/10000,topology=(code/1000)%10,
	interval=(code/100)%10,LH_exit=(code/10)%10,interval_exit=code%10;
    final int buy_sell=(r.cmd==S_BUY?1:(r.cmd==S_SELL?-1:0)),
	buy_sell_exit=(r.cmd==S_BUY_EXIT?-1:(r.cmd==S_SELL_EXIT?1:0)),
	n_intervals=pt.p_imin.length+pt.p_imax.length+1;

    if(code!=0 && !( (dir==7 && r.flag_lh==false)||(dir==8 && r.flag_lh==true)))
    {
	throw new RuntimeException("code="+code+" r.flag_lh="+r.flag_lh);
    }
    if(code!=0 && buy_sell!=0)
    {
	UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",-buy_sell,String.valueOf(-buy_sell),eps);
	if(pt.i_imin.length==2)
	{
	    if(buy_sell==1 && !(pt.i_imin[0]<pt.i_imin[1]))
	    {
		throw new RuntimeException("wrong buy topology");
	    }
	    if(buy_sell==-1 && !(pt.i_imin[0]>pt.i_imin[1]))
	    {
		throw new RuntimeException("wrong sell topology");
	    }
	}
    }

    if(buy_sell_exit==1)
    {
	// buy to cover short
	if(interval==1 || (r.flag_lh && code!=0))
	{
	    UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",-1,"-1",eps);
	}
	if(!r.flag_lh && code!=0)
	{
	    if(r.p>pt.p_imax[0])
	    {
		UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",-1,"-1",eps);
	    }
	    else
	    {
		UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",1,"1",eps);
	    }
	}
    }

    if(buy_sell_exit==-1)
    {
	// sell existing position
	if(interval==n_intervals || (!r.flag_lh && code!=0))
	{
	    UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",1,"1",eps);
	}
	if(r.flag_lh && code!=0)
	{
	    if(r.p<pt.p_imax[pt.p_imax.length-1])
	    {
		UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",1,"1",eps);
	    }
	    else
	    {
		UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",-1,"-1",eps);
	    }
	}
    }

    if(pt.p_imin.length>0)
    {
	if(r.flag_lh==false)
	{
	    if(r.p>pt.p_imin[pt.p_imin.length-1])
	    {
		UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",1,"1",eps);
	    }
	}
	else
	{
	    if(r.p<pt.p_imin[0])
	    {
		UnitTests.checkDoubleRel(Math.signum(dip.getDI1(r.p)),"Math.signum(dip.getDI1(r.p))",-1,"-1",eps);
	    }
	}
    }

    if(r.cmd==S_BUY_EXIT && ((code!=0) != pt.checkExitBuy(r.p)))
    {
	throw new RuntimeException("buy exit mismatch r.code="+code+
				   " cmd="+r.cmd+" p="+r.p+
				   " pt=\n"+pt.dumpData());
    }
    if(r.cmd==S_SELL_EXIT && ((code!=0) != pt.checkExitSell(r.p)))
    {
	throw new RuntimeException("sell exit mismatch r.code="+code+
				   " cmd="+r.cmd+" p="+r.p+
				   " pt=\n"+pt.dumpData());
    }
    if(r.cmd==S_BUY && ((code!=0) != pt.checkEnterBuy(r.p)))
    {
	throw new RuntimeException("buy enter mismatch r.code="+code+
				   " cmd="+r.cmd+" p="+r.p+
				   " pt=\n"+pt.dumpData());
    }
    if(r.cmd==S_SELL && ((code!=0) != pt.checkEnterSell(r.p)))
    {
	throw new RuntimeException("sell enter mismatch r.code="+code+
				   " cmd="+r.cmd+" p="+r.p+
				   " pt=\n"+pt.dumpData());
    }
}





    private static void matchWithTest(final String description,
				      final DPTradeIP dip,
				      final java.util.Random r,
				      final double eps)
    {
	System.err.println("TCode:matchWithTest():======================"+description);
	final int nprices=1000;
	final PTopology pt=dip.ptopology;

	int n_tests=0;
	for(int cmd:new int []{S_BUY,S_BUY_EXIT,S_SELL,S_SELL_EXIT})
	{
	    for(int nn=0;nn<nprices;nn++)
	    {
		final double p=TradeEMoments.getRandomPrice(r);
		final TCode tc=dip.getTCode(cmd,p);
		if(tc!=null)
		{
		    n_tests++;
		    checkCMD(tc,pt,dip,eps);
		}
		else
		{
		    if(pt.tdir!=0)
		    {
			throw new RuntimeException("tc=null should correspond to pt.tdir=0\n"+pt.dumpData());
		    }
		}
	    }
	}
	System.err.println("TCode:matchWithTest: matched OK nprices="+nprices+" n_tests="+n_tests);
    }


    private static void checkMatching(final PTopology pt1,final PTopology pt2,final double p)
    {
	if(pt1.checkEnterBuy(p)!=pt2.checkEnterBuy(p))
	{
	    throw new RuntimeException("buy enter mismatch: p="+p+
				       " pt1=\n"+pt1.dumpData()+
				       " pt2=\n"+pt2.dumpData());
	}
	if(pt1.checkExitBuy(p)!=pt2.checkExitBuy(p))
	{
	    throw new RuntimeException("buy exit mismatch: p="+p+
				       " pt1=\n"+pt1.dumpData()+
				       " pt2=\n"+pt2.dumpData());
	}
	if(pt1.checkEnterSell(p)!=pt2.checkEnterSell(p))
	{
	    throw new RuntimeException("sell enter mismatch: p="+p+
				       " pt1=\n"+pt1.dumpData()+
				       " pt2=\n"+pt2.dumpData());
	}
	if(pt1.checkExitSell(p)!=pt2.checkExitSell(p))
	{
	    throw new RuntimeException("sell exit mismatch: p="+p+
				       " pt1=\n"+pt1.dumpData()+
				       " pt2=\n"+pt2.dumpData());
	}
    }

    private static void matchPTopologies(final PTopology pt1,
					 final PTopology pt2,
					 final java.util.Random r,
					 final double iscale,
					 final double eps)
    {
	final int nprices=20000;
	for(int nn=0;nn<nprices;nn++)
	{
	    final double p=TradeEMoments.getRandomPrice(r);
	    checkMatching(pt1,pt2,p);
	}
	matchPTopologiesLevels(pt1,pt2,iscale,eps);
	System.err.println("TCode:matchPTopologies: matched OK nprices="+nprices);
    }
    private static void matchPTopologiesLevels(final PTopology pt1,
					       final PTopology pt2,
					       final double iscale,
					       final double eps)
    {
	UnitTests.checkDoubleRel(pt1.d0,"pt1.d0",pt2.d0*iscale,"pt2.d0*iscale",eps);
	UnitTests.checkDoubleRel(pt1.pm,"pt1.pm",pt2.pm,"pt2.pm",eps);
	UnitTests.checkDoubleRel(pt1.pm_l,"pt1.pm_l",pt2.pm_l,"pt2.pm_l",eps);
	UnitTests.checkDoubleRel(pt1.pm_h,"pt1.pm_h",pt2.pm_h,"pt2.pm_h",eps);
	UnitTests.checkDoubleArrayRel(pt1.p_imin,"pt1.p_imin",pt2.p_imin,"pt2.p_imin",eps);
	UnitTests.checkDoubleArrayRel(pt1.p_imax,"pt1.p_imax",pt2.p_imax,"pt2.p_imax",eps);
	UnitTests.checkDoubleArrayRel(pt1.p_i_eq_imax_low,"pt1.p_i_eq_imax_low",pt2.p_i_eq_imax_low,"pt2.p_i_eq_imax_low",eps);
	UnitTests.checkDoubleArrayRel(pt1.p_i_eq_imax_high,"pt1.p_i_eq_imax_high",pt2.p_i_eq_imax_high,"pt2.p_i_eq_imax_high",eps);
	for(int k=pt2.i_imin.length;--k>=0;)
	{
	    UnitTests.checkDoubleRel(pt1.i_imin[k],"pt1.i_imin[k]",pt2.i_imin[k]*iscale,"pt2.i_imin[k]*iscale",eps);
	}
	for(int k=pt2.i_imax.length;--k>=0;)
	{
	    UnitTests.checkDoubleRel(pt1.i_imax[k],"pt1.i_imax[k]",pt2.i_imax[k]*iscale,"pt2.i_imax[k]*iscale",eps);
	}
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");

	System.err.println("TCode:checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);


	final DIPTradeIP ro=DIPTradeIP.calculateDIPTradeIP(N_EM,
							   currenttime,
							   tau,
							   tb,
							   tsize,
							   null);

	matchWithTest("INITIAL CALCULATION",ro.kdiff,r,eps);

	/* // Commented out after PTVV migration
	final QTradeIP rq_kdiff=QTradeIP.calculateQTradeIP(N_EM,
							   currenttime,
							   tau,
							   tb,
							   tsize,
							   null);
	matchPTopologies(ro.kdiff.ptopology,rq_kdiff.ptopology,r,rq_kdiff.odata.moments_pv[0]/rq_kdiff.odata.moments_pt[0],eps);
	*/

	System.err.println("TCode:checkCalculations(): OK with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(17903);
	for(int N_EM :new int []{3,4,N_EMOMENTS,3,4,5})
	{
	    for(int i=0;i<330;i++)
	    {
		checkCalculations(1000+r.nextInt(900),r,N_EM);
		checkCalculations(23010+r.nextInt(23010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TCode::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

} // END OF class TCode


static class OTradeIP {
    final TradeIPData tipdata;
    private final double [] normformoments_pv; // typically [m]=<(p-offset)^2>_t^{m/2}

    private final double [] vpcoefs,vreal_roots,p_balance;
    final double sigma,skewness,kurtosis;
    final double [] vattractors,vrepellers,p_up,p_down;
    final double min_A,max_A,ex_A,i0_A,min_R,max_R,ex_R,i0_R;

    OTradeIP(final TradeIPData tipdata)
    {
	this.tipdata=tipdata;
	final int np=this.tipdata.np;

	this.normformoments_pv=TradeIPData.getPowersOfStdev(this.tipdata.moments_pv[2]/this.tipdata.v_sum,np);

	final double [] norms=this.normformoments_pv;
	final double [][] msol=new double [np][np];
	this.vpcoefs=new double [np];
	TradeIPData.setPriceMatrixAndRightPart(msol,this.vpcoefs,this.tipdata.moments_pv,norms,this.tipdata.moments_pt,np);

	TradeDPVR.solveLinsystem(msol,this.vpcoefs,np,new int [np]);
	this.sigma=norms[1];
	this.skewness=this.tipdata.moments_pv[3]/this.tipdata.v_sum;
	this.kurtosis=this.tipdata.moments_pv[4]/this.tipdata.v_sum;

	final double p_offset=this.tipdata.p_offset;

	//  dt/d|v|->min equation
	if(true)
	{
	    final double [] pcoe=new double [np-1]; // temporary array
	    // use scale=1 expected by PolynomialRoot
	    for(int k=1;k<np;k++)
	    {
		pcoe[k-1]=k*this.vpcoefs[k]/norms[k];
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    final double [] root_re=u.re_root,root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-2,root_re,root_im);
	    final double [] reroots=pcoe, // re-use as temporary arrays
		save_attractors=root_re,save_repellers=root_im;
	    int n_reroots=0,n_attractors=0,n_repellers=0;
	    double rr=Double.NaN;
	    for(int m=0;m<np-2;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{ // real root
		    if(getD2WithScale(this.vpcoefs,rr,norms)<0)
		    { // repeller
			save_repellers[n_repellers++]=rr;
		    }
		    else
		    { // attractor
			save_attractors[n_attractors++]=rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.vattractors=java.util.Arrays.copyOf(save_attractors,n_attractors);
	    this.vrepellers=java.util.Arrays.copyOf(save_repellers,n_repellers);
	    this.vreal_roots=java.util.Arrays.copyOf(reroots,n_reroots);


	    if(true)
	    {
		final double [] sar=TradeIP.getmmAR(this.vpcoefs,norms,this.vattractors,false);
		// shift to 0
		this.min_A=sar[0]+p_offset;
		this.max_A=sar[1]+p_offset;
		this.ex_A=sar[2]+p_offset;
		this.i0_A=sar[3];

		final double [] srr=TradeIP.getmmAR(this.vpcoefs,norms,this.vrepellers,true);
		// shift to 0
		this.min_R=srr[0]+p_offset;
		this.max_R=srr[1]+p_offset;
		this.ex_R=srr[2]+p_offset;
		this.i0_R=srr[3];
	    }

	    // shift attractor & repeller offset to 0
	    for(int m=this.vattractors.length;--m>=0;this.vattractors[m]+=p_offset){}
	    for(int m=this.vrepellers.length;--m>=0;this.vrepellers[m]+=p_offset){}
	}

	// dt/d|v|=tau0/v0 , find real roots
	if(true)
	{
	    final double [] pcoe=new double [np]; // temporary array
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0;k<np;k++)
	    {
		pcoe[k]=this.vpcoefs[k]/norms[k];
	    }
	    pcoe[0]-=1; // equation 1=sum_k pcoe_k * (p-p_offset)^k
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    final double [] root_re=u.re_root,root_im=u.im_root;
	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-1,root_re,root_im);

	    int n_reroots=0,n_p_down=0,n_p_up=0;
	    double rr=Double.NaN;
	    final double [] reroots=pcoe, // use pcoe as a temporary array
		save_p_down=new double [np],save_p_up=new double [np];
	    for(int m=0;m<root_re.length;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{
		    if(getD1WithScale(this.vpcoefs,rr,norms)<0)
		    { // up
			save_p_up[n_p_up++]=p_offset+rr;
		    }
		    else
		    { // down
			save_p_down[n_p_down++]=p_offset+rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.p_up=java.util.Arrays.copyOf(save_p_up,n_p_up);
	    this.p_down=java.util.Arrays.copyOf(save_p_down,n_p_down);
	    this.p_balance=java.util.Arrays.copyOf(reroots,n_reroots);
	}
    }

    double getD0T(final double p)
    {
	return getD0WithScale(this.vpcoefs,p-this.tipdata.p_offset,this.normformoments_pv);
    }


    private static OTradeIP calculateTestOTradeIP(final int N_EM,
						  final long currenttime,
						  final long tau,
						  final BookItem [] tb,
						  final int tsize)
    {
	final boolean CALCULATE_VT=false;
	return new OTradeIP(TradeIPData.calculateTestTradeIPData(N_EM,
								 currenttime,
								 tau,
								 tb,
								 tsize,
								 CALCULATE_VT));
    }


    private static void checkRoots(final OTradeIP r,final double eps)
    {
	final double [] norm=r.normformoments_pv;
	final double p_offset=r.tipdata.p_offset;
	for(int k=0;k<r.vattractors.length;k++)
	{
	    UnitTests.checkDoubleRel(getD1WithScale(r.vpcoefs,r.vattractors[k]-p_offset,norm),
			   "attractor:d oI/dp(root)",0,"0",eps);
	    UnitTests.checkDouble(Math.signum(getD2WithScale(r.vpcoefs,r.vattractors[k]-p_offset,norm)),
			"sign(d^2 oI/d p^2(root))",1,"1");
	}
	for(int k=0;k<r.vrepellers.length;k++)
	{
	    UnitTests.checkDoubleRel(getD1WithScale(r.vpcoefs,r.vrepellers[k]-p_offset,norm),
			   "repeller:d oI/dp(root)",0,"0",eps);
	    UnitTests.checkDouble(Math.signum(getD2WithScale(r.vpcoefs,r.vrepellers[k]-p_offset,norm)),
			"sign(d^2 oI/d p^2(root))",-1,"-1");
	}
	for(int k=0;k<r.vreal_roots.length;k++)
	{
	    UnitTests.checkDoubleRel(getD1WithScale(r.vpcoefs,r.vreal_roots[k],norm),
			   "d oI/dp(root)",0,"0",eps);
	}
	for(int k=0;k<r.p_up.length+r.p_down.length;k++)
	{
	    final double pI=k<r.p_up.length?r.p_up[k]:r.p_down[k-r.p_up.length];
	    UnitTests.checkDoubleRel(getD0WithScale(r.vpcoefs,pI-p_offset,norm),"oI(1)",1,"1",eps);
	}
    }


    private static void checkMatching(final OTradeIP r,final OTradeIP ref)
    {
	TradeIPData.checkMatching(r.tipdata,ref.tipdata);

	final double eps=1e-5;

	UnitTests.checkDoubleRel(r.tipdata.p_offset,"r.tipdata.p_offset",r.tipdata.pv_sum/r.tipdata.v_sum,"r.tipdata.pv_sum/r.tipdata.v_sum",eps);
	UnitTests.checkDoubleRel(r.sigma,"r.sigma",Math.sqrt(r.tipdata.moments_pv[2]/r.tipdata.moments_pv[0]),"sqrt(r.tipdata.moments_pv[2]/r.tipdata.moments_pv[0])",eps);
	UnitTests.checkDoubleRel(r.skewness,"r.skewness",r.tipdata.moments_pv[3]/r.tipdata.moments_pv[0],"r.tipdata.moments_pv[3]/r.tipdata.moments_pv[0]",eps);
	UnitTests.checkDoubleRel(r.kurtosis,"r.kurtosis",r.tipdata.moments_pv[4]/r.tipdata.moments_pv[0],"r.tipdata.moments_pv[4]/r.tipdata.moments_pv[0]",eps);

	TradeIPData.checkPCoefs(r.vpcoefs,r.tipdata.moments_pv,r.normformoments_pv,r.tipdata.moments_pt,eps);
	checkRoots(r,eps);
	System.err.println("OTradeIP:checkMatching(): matching OK");
    }


    private static void matchWithTest(final OTradeIP r,
				      final Trade [] tb,
				      final int tsize)
    {
	for(byte btype:new byte []{BookItem.BOOK_BUY,BookItem.BOOK_SELL})
	{
	    final double eps=1e-6;
	    final Trade [] ts=TradeEMoments.selectTradesOfSpecificType(tsize,tb,btype);
	    final OTradeIP rt=calculateTestOTradeIP(r.vpcoefs.length,
						    r.tipdata.currenttime,
						    r.tipdata.tau,
						    ts,
						    ts.length);
	    TradeIPData.checkMatchingSubtype(r.tipdata,rt.tipdata,btype,eps);
	}

	final OTradeIP rt=calculateTestOTradeIP(r.vpcoefs.length,
						r.tipdata.currenttime,
						r.tipdata.tau,
						tb,
						tsize);
	checkMatching(r,rt);
	final TradeEMoments ref=calculateEMoments(r.tipdata.currenttime,
						  r.tipdata.tau,
						  tb,
						  tsize,
						  null);
	TradeEMoments.checkMatchingTradeIPData(ref,r.tipdata);
    }



    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println("checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);
	final OTradeIP ro=calculateOTradeIP(N_EM,
					    currenttime,
					    tau,
					    tb,
					    tsize,
					    null);
	matchWithTest(ro,tb,tsize);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final OTradeIP em_ref=calculateOTradeIP(N_EM,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest(em_ref,tb,n_split);

	final OTradeIP em_fromref=calculateOTradeIP(N_EM,currenttime,tau,tb,tsize,em_ref);
	checkMatching(ro,em_fromref);
	matchWithTest(em_fromref,tb,tsize);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final OTradeIP emc_cut=calculateOTradeIP(N_EM,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut);

	System.err.println("OTradeIP::checkCalculations(): test random values OK N points="+tsize+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{3,4,N_EMOMENTS,2,3,4,5})
	{
	    for(int i=0;i<100;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("OTradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }

} // END OF class OTradeIP


static OTradeIP calculateOTradeIP(final int N_EM,
				  final long currenttime,
				  final long tau,
				  final BookItem [] tb,
				  final int tsize,
				  final OTradeIP ref)
{
    final boolean CALCULATE_VT=false;
    return new OTradeIP(TradeIPData.calculateTradeIPData(N_EM,
							 currenttime,
							 tau,
							 tb,
							 tsize,
							 ref!=null?ref.tipdata:null,
							 CALCULATE_VT));
}

























static class PTradeIP {
    private final double [] normformoments; // typically [m]=<(p-offset)^2>_t^{m/2}

    private final double [] pcoefs;
    final double p_offset;

    final double [] vattractors,vrepellers;
    final double min_A,max_A,ex_A,i0_A,min_R,max_R,ex_R,i0_R;
    final double [] p_up,p_down;
    private final double [] p_real_roots;

    PTradeIP(final double [] moments_left,final double [] moments_right,final TradeIPRData odata)
    {
	final int np=getNP(odata.np);
	if(np<3)
	{
	    throw new RuntimeException("np="+np+" is not supported");
	}
	final double [] norms=this.normformoments=getNorms(moments_left,np);
	final double p_offset=this.p_offset=odata.p_offset;

	this.pcoefs=getPCoefs(moments_left,norms,moments_right,np);
	if(this.pcoefs.length!=np)
	{
	    throw new RuntimeException("np="+np+" pcoefs.length="+this.pcoefs.length);
	}

	//  dt/d|v|->min,max equation
	if(true)
	{
	    final double [] pcoe=new double [np-1]; // temporary arrays
	    // use scale=1 expected by PolynomialRoot
	    for(int k=1;k<np;k++)
	    {
		pcoe[k-1]=k*this.pcoefs[k]/norms[k];
	    }
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    final double [] root_re=u.re_root,root_im=u.im_root;

	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-2,root_re,root_im);
	    final double [] reroots=pcoe, // re-use as temporary arrays
		save_attractors=root_re,save_repellers=root_im;
	    int n_reroots=0,n_attractors=0,n_repellers=0;
	    double rr=Double.NaN;
	    for(int m=0;m<np-2;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{ // real root
		    if(getD2WithScale(this.pcoefs,rr,norms)>0)
		    { // repeller
			save_repellers[n_repellers++]=rr;
		    }
		    else
		    { // attractor
			save_attractors[n_attractors++]=rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.vattractors=java.util.Arrays.copyOf(save_attractors,n_attractors);
	    this.vrepellers=java.util.Arrays.copyOf(save_repellers,n_repellers);


	    if(true)
	    {
		final double [] sar=TradeIP.getmmAR(this.pcoefs,norms,this.vattractors,true);
		// shift to 0
		this.min_A=sar[0]+p_offset;
		this.max_A=sar[1]+p_offset;
		this.ex_A=sar[2]+p_offset;
		this.i0_A=sar[3];

		final double [] srr=TradeIP.getmmAR(this.pcoefs,norms,this.vrepellers,false);
		// shift to 0
		this.min_R=srr[0]+p_offset;
		this.max_R=srr[1]+p_offset;
		this.ex_R=srr[2]+p_offset;
		this.i0_R=srr[3];
	    }

	    // shift attractor & repeller offset to 0
	    for(int m=this.vattractors.length;--m>=0;this.vattractors[m]+=p_offset){}
	    for(int m=this.vrepellers.length;--m>=0;this.vrepellers[m]+=p_offset){}
	}


	if(true)
	{
	    final double [] pcoe=new double [np]; // temporary arrays
	    // use scale=1 expected by PolynomialRoot
	    for(int k=0;k<np;k++)
	    {
		pcoe[k]=this.pcoefs[k]/norms[k];
	    }
	    // equation 1=sum_k pcoe_k * (p-p_offset)^k
	    final PolynomialRootsFinder.PRoots u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoe);
	    final double [] root_re=u.re_root,root_im=u.im_root;
	    // insertion sort by real part with extra load
	    SortArrays.sortArrayWithLoad(np-1,root_re,root_im);

	    int n_reroots=0,n_p_down=0,n_p_up=0;
	    double rr=Double.NaN;
	    final double [] reroots=pcoe, // re-use as temporary arrays
		save_p_up=root_re,save_p_down=root_im;
	    for(int m=0;m<np-1;m++)
	    {
		if(rr>root_re[m])
		{
		    throw new RuntimeException("Not sorted");
		}
		rr=root_re[m];
		if(Math.abs(root_im[m])<=EPS_IM)
		{
		    if(getD1WithScale(this.pcoefs,rr,norms)<=0)
		    { // down
			save_p_down[n_p_down++]=p_offset+rr;
		    }
		    else
		    { // up
			save_p_up[n_p_up++]=p_offset+rr;
		    }
		    reroots[n_reroots++]=rr;
		}
	    }
	    this.p_up=java.util.Arrays.copyOf(save_p_up,n_p_up);
	    this.p_down=java.util.Arrays.copyOf(save_p_down,n_p_down);
	    this.p_real_roots=java.util.Arrays.copyOf(reroots,n_reroots);
	}
    }

    double getPDownClose(final double p)
    {
	final double dp_init=p-this.p_offset;
	int ptr_close_to_init=-1;
	double drr=Double.POSITIVE_INFINITY;
	for(int m=this.p_real_roots.length;--m>=0;)
	{
	    final double rr=this.p_real_roots[m],drr1=Math.abs(dp_init-rr);
	    if(drr1<drr)
	    {
		drr=drr1;
		ptr_close_to_init=m;
	    }
	}
	if(ptr_close_to_init<0)
	{
	    return Double.NaN;
	}
	final double rr=this.p_real_roots[ptr_close_to_init];
	if(getD1WithScale(this.pcoefs,rr,this.normformoments)<0)
	{
	    return rr+this.p_offset;
	}
	if(rr>dp_init)
	{
	    return ptr_close_to_init>0?this.p_real_roots[ptr_close_to_init-1]+this.p_offset:Double.NaN;
	}
	return ptr_close_to_init<this.p_real_roots.length-1?this.p_real_roots[ptr_close_to_init+1]+this.p_offset:Double.NaN;
    }

    int getNP(final int np0)
    {
	return np0;
    }

    double [] getNorms(final double [] moments_left,final int np)
    {
	return TradeIPData.getPowersOfStdev(moments_left[2]/moments_left[0],np);
    }
    double getIEq()
    {
	return 1;
    }

    double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
    {
	final double [][] msol=new double [np][np];
	final double [] pcoefs=new double [np];
	TradeIPData.setPriceMatrixAndRightPart(msol,pcoefs,moments_left,norms,moments_right,np);
	TradeDPVR.solveLinsystem(msol,pcoefs,np,new int [np]);
	return pcoefs;
    }

    double getD0T(final double p)
    {
	return getD0WithScale(this.pcoefs,p-this.p_offset,this.normformoments);
    }

    double getD1T(final double p)
    {
	return getD1WithScale(this.pcoefs,p-this.p_offset,this.normformoments);
    }

    static double [] getArray1Filled(final int np)
    {
	final double [] norms=new double [np];
	java.util.Arrays.fill(norms,1);
	return norms;
    }

} // END OF class PTradeIP


static class DBTradeIP {
    final PTradeIP buy,sell,total,total2,diff,difft;
    private final TradeIPData odata;
    private DBTradeIP(final TradeIPData odata)
    {
	this.odata=odata;
	this.total=new PTradeIP(odata.moments_pt,odata.moments_pv,odata);

	this.buy=new PTradeIP(odata.moments_pt_buy,odata.moments_pv_buy,odata);
	this.sell=new PTradeIP(odata.moments_pt_sell,odata.moments_pv_sell,odata);
	this.diff=new PTradeIP(null,null,odata){
		double getIEq(){return 0;}
		double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double
			nbuy=odata.moments_pv_buy[0]/odata.moments_pt_buy[0],
			nsell=odata.moments_pv_sell[0]/odata.moments_pt_sell[0],
			s=nbuy+nsell;
		    final double [] pcoefs=new double [np];
		    for(int k=0;k<np;k++)
		    {
			//pcoefs[k]=sell.pcoefs[k]/sell.normformoments[k]; // for testing
			//pcoefs[k]=buy.pcoefs[k]/buy.normformoments[k]; // for testing
			pcoefs[k]=(nsell*(sell.pcoefs[k]/sell.normformoments[k])-nbuy*(buy.pcoefs[k]/buy.normformoments[k]))/s;
		    }
		    return pcoefs;
		}
	    };
	this.total2=new PTradeIP(null,null,odata){
		double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double
			nbuy=odata.moments_pv_buy[0]/odata.moments_pt_buy[0],
			nsell=odata.moments_pv_sell[0]/odata.moments_pt_sell[0],
			s=nbuy+nsell;
		    final double [] pcoefs=new double [np];
		    for(int k=0;k<np;k++)
		    {
			//pcoefs[k]=sell.pcoefs[k]/sell.normformoments[k]; // for testing
			//pcoefs[k]=buy.pcoefs[k]/buy.normformoments[k]; // for testing
			pcoefs[k]=(nsell*(sell.pcoefs[k]/sell.normformoments[k])+nbuy*(buy.pcoefs[k]/buy.normformoments[k]))/s;
		    }
		    return pcoefs;
		}
	    };
	this.difft=new PTradeIP(null,null,odata){
		double getIEq(){return 0;}
		double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] msol=new double [np][np];
		    final double [] rpart=new double [np];
		    for(int l=0;l<np;l++)
		    {
			for(int k=0;k<np;k++)
			{
			    msol[l][k]=odata.moments_pt[l+k];
			}
			rpart[l]=odata.moments_pv_sell[l]-odata.moments_pv_buy[l];
		    }
		    TradeDPVR.solveLinsystem(msol,rpart,np,new int [np]);
		    return rpart;
		}
	    };
    }



    private static final boolean CALCULATE_VT=true;
    private static DBTradeIP calculateTestDBTradeIP(final int N_EM,
						    final long currenttime,
						    final long tau,
						    final BookItem [] tb,
						    final int tsize)
    {
	return new DBTradeIP(TradeIPData.calculateTestTradeIPData(N_EM,
								  currenttime,
								  tau,
								  tb,
								  tsize,
								  CALCULATE_VT));
    }

    static DBTradeIP calculateDBTradeIP(final int N_EM,
					final long currenttime,
					final long tau,
					final BookItem [] tb,
					final int tsize,
					final DBTradeIP ref)
    {
	return new DBTradeIP(TradeIPData.calculateTradeIPData(N_EM,
							      currenttime,
							      tau,
							      tb,
							      tsize,
							      ref!=null?ref.odata:null,
							      CALCULATE_VT));
    }

    static void checkDiffCoefs(final DBTradeIP r,final double eps)
    {
	final TradeIPData odata=r.odata;
	final int np=odata.np;
	final double [][] matr=new double [2*np][2*np];
	final double [] rpart=new double [2*np];
	for(int l=0;l<np;l++)
	{
	    final int lpnp=l+np;
	    for(int k=0;k<np;k++)
	    {
		final int lpk=l+k;
		matr[l][k]=matr[lpnp][k+np]=odata.moments_pt_sell[lpk]+odata.moments_pt_buy[lpk];
		matr[l][k+np]=matr[lpnp][k]=odata.moments_pt_sell[lpk]-odata.moments_pt_buy[lpk];
	    }
	    rpart[l]=odata.moments_pv_sell[l]+odata.moments_pv_buy[l];
	    rpart[lpnp]=odata.moments_pv_sell[l]-odata.moments_pv_buy[l];
	}
	TradeDPVR.solveLinsystem(matr,rpart,2*np,new int [2*np]);
	final double [] pcoefs=new double [np];
	System.arraycopy(rpart,np,pcoefs,0,np);
	final double s=odata.moments_pv_buy[0]/odata.moments_pt_buy[0]+
	               odata.moments_pv_sell[0]/odata.moments_pt_sell[0];
	for(int k=0;k<np;k++)
	{
	    pcoefs[k]/=0.5*s; // bring to the same scale
	}

	// check matching
	UnitTests.checkDoubleArrayRel(pcoefs,"pcoefs",r.diff.pcoefs,"r.diff.pcoefs",eps);
    }



    private static void checkRoots(final PTradeIP r,final double eps)
    {
	final double [] norm=r.normformoments;
	final double p_offset=r.p_offset;
	for(int k=0;k<r.vattractors.length;k++)
	{
	    UnitTests.checkDoubleRel(getD1WithScale(r.pcoefs,r.vattractors[k]-p_offset,norm),
			   "attractor:d I/dp(root)",0,"0",eps);
	    UnitTests.checkDouble(Math.signum(getD2WithScale(r.pcoefs,r.vattractors[k]-p_offset,norm)),
			"sign(d^2 I/d p^2(root))",-1,"-1");
	}
	for(int k=0;k<r.vrepellers.length;k++)
	{
	    UnitTests.checkDoubleRel(getD1WithScale(r.pcoefs,r.vrepellers[k]-p_offset,norm),
			   "repeller:d I/dp(root)",0,"0",eps);
	    UnitTests.checkDouble(Math.signum(getD2WithScale(r.pcoefs,r.vrepellers[k]-p_offset,norm)),
			"sign(d^2 I/d p^2(root))",1,"1");
	}
    }

    private static void checkMatching(final DBTradeIP r,final DBTradeIP ref)
    {
	TradeIPData.checkMatching(r.odata,ref.odata);

	final double eps=1e-5;

	UnitTests.checkDoubleRel(r.odata.p_offset,"r.odata.p_offset",r.odata.pt_sum/r.odata.t_sum,"r.odata.pt_sum/r.odata.t_sum",eps);
	UnitTests.checkDoubleArrayRel(r.buy.normformoments,"r.buy.normformoments",
			    TradeIPData.getPowersOfStdev(r.odata.moments_pt_buy[2]/r.odata.moments_pt_buy[0],r.odata.np),"powers buy",eps);
	UnitTests.checkDoubleArrayRel(r.sell.normformoments,"r.sell.normformoments",
			    TradeIPData.getPowersOfStdev(r.odata.moments_pt_sell[2]/r.odata.moments_pt_sell[0],r.odata.np),"powers sell",eps);
	UnitTests.checkDoubleArrayRel(r.total.normformoments,"r.total.normformoments",
			    TradeIPData.getPowersOfStdev(r.odata.moments_pt[2]/r.odata.moments_pt[0],r.odata.np),"powers total",eps);


	TradeIPData.checkPCoefs(r.buy.pcoefs,r.odata.moments_pt_buy,r.buy.normformoments,r.odata.moments_pv_buy,eps);
	TradeIPData.checkPCoefs(r.sell.pcoefs,r.odata.moments_pt_sell,r.sell.normformoments,r.odata.moments_pv_sell,eps);
	checkDiffCoefs(r,eps);


	checkRoots(r.buy,eps);
	checkRoots(r.sell,eps);
	System.err.println("DBTradeIP:checkMatching(): matching OK");
    }

    private static void matchWithTest(final String description,
				      final DBTradeIP r,
				      final Trade [] tb,
				      final int tsize)
    {
	System.err.println("DBTradeIP:matchWithTest():======================"+description);
	final double eps=1e-6;
	for(byte btype:new byte []{BookItem.BOOK_BUY,BookItem.BOOK_SELL})
	{
	    final Trade [] ts=TradeEMoments.selectTradesOfSpecificType(tsize,tb,btype);
	    final DBTradeIP rt=calculateTestDBTradeIP(r.odata.np,
						      r.odata.currenttime,
						      r.odata.tau,
						      ts,
						      ts.length);
	    TradeIPData.checkMatchingSubtype(r.odata,rt.odata,btype,eps);
	}

	final DBTradeIP rt=calculateTestDBTradeIP(r.odata.np,
						  r.odata.currenttime,
						  r.odata.tau,
						  tb,
						  tsize);
	checkMatching(r,rt);

	final TradeEMoments ref=calculateEMoments(r.odata.currenttime,
						  r.odata.tau,
						  tb,
						  tsize,
						  null);
	TradeEMoments.checkMatchingTradeIPData(ref,r.odata);
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println("DBTradeIP:checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau);
	final DBTradeIP ro=calculateDBTradeIP(N_EM,
					      currenttime,
					      tau,
					      tb,
					      tsize,
					      null);
	matchWithTest("INITIAL CALCULATION",ro,tb,tsize);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final DBTradeIP em_ref=calculateDBTradeIP(N_EM,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest("GET REFERENCE FOR RECURRENT",em_ref,tb,n_split);

	final DBTradeIP em_fromref=calculateDBTradeIP(N_EM,currenttime,tau,tb,tsize,em_ref);
	checkMatching(ro,em_fromref);
	matchWithTest("RECURRENT CALCULATIONS",em_fromref,tb,tsize);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final DBTradeIP emc_cut=calculateDBTradeIP(N_EM,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut);

	System.err.println("DBTradeIP::checkCalculations(): test random values OK N points="+tsize+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{3,4,N_EMOMENTS,3,4,5})
	{
	    for(int i=0;i<100;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("DBTradeIP::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }



} // END OF class DBTradeIP




static class TradeIPBSData extends TradeIPRData {

    final double pt_sum,t_sum;
    private final double []
	moments_p_buy_v_buy,
	moments_p_buy_v_sell,
	moments_p_sell_v_buy,
	moments_p_sell_v_sell
	;
    private final double [][]
	moments_p_buy_pt_buy,
	moments_p_buy_x_p_sell,
	moments_p_sell_pt_sell,

	moments_p_buy_x_p_buy, // similar to moments_p_buy_pt_buy, but calculated on the same grid as moments_p_buy_x_p_sell
	moments_p_sell_x_p_sell // similar to moments_p_sell_pt_sell, but calculated on the same grid as moments_p_buy_x_p_sell
	;


    // references to previos calculations
    private final BookItem trade_last,trade_last_buy,trade_last_sell;


    TradeIPBSData(final long currenttime,
		  final long tau,

		  final double pt_sum,
		  final double t_sum,
		  final double p_offset,

		  final double [] moments_p_buy_v_buy,
		  final double [] moments_p_buy_v_sell,
		  final double [] moments_p_sell_v_buy,
		  final double [] moments_p_sell_v_sell,
		  final double [][] moments_p_buy_pt_buy,
		  final double [][] moments_p_buy_x_p_sell,
		  final double [][] moments_p_sell_pt_sell,

		  final double [][] moments_p_buy_x_p_buy,
		  final double [][] moments_p_sell_x_p_sell,

		  final long vol,
		  final int np,

		  // references
		  final BookItem trade_last,
		  final BookItem trade_last_buy,
		  final BookItem trade_last_sell)
    {
	super(currenttime,tau,p_offset,vol,np);

	this.pt_sum=pt_sum;
	this.t_sum=t_sum;

	this.moments_p_buy_v_buy=moments_p_buy_v_buy;
	this.moments_p_buy_v_sell=moments_p_buy_v_sell;
	this.moments_p_sell_v_buy=moments_p_sell_v_buy;
	this.moments_p_sell_v_sell=moments_p_sell_v_sell;

	this.moments_p_buy_pt_buy=moments_p_buy_pt_buy;
	this.moments_p_buy_x_p_sell=moments_p_buy_x_p_sell;
	this.moments_p_sell_pt_sell=moments_p_sell_pt_sell;

	this.moments_p_buy_x_p_buy=moments_p_buy_x_p_buy;
	this.moments_p_sell_x_p_sell=moments_p_sell_x_p_sell;

	this.trade_last=trade_last;
	this.trade_last_buy=trade_last_buy;
	this.trade_last_sell=trade_last_sell;
    }

    static TradeIPBSData calculateTestTradeIPBSData(final int N_EM,
						    final long currenttime,
						    final long tau,
						    final BookItem [] tb,
						    final int tsize)
    {
	final int np=N_EM;
	final double atau=1.0/tau;

	// calculate 0th & 1st moments
	double pt_sum=0,t_sum=0;
	long prevt=currenttime,tinterval=currenttime,vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    //System.err.println("i="+i+" b.time="+b.time);
	    prevt=b.time;
	    vol+=b.shares;
	    final double twmw=Math.exp(-(currenttime-prevt)*atau)*((tinterval-prevt)*atau);

	    pt_sum+=b.price*twmw;
	    t_sum+=twmw;

	    tinterval=prevt;
	}

	final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);


	final double [][][] m=checkBxS(np,currenttime,tau,tb,tsize,p_offset);
	return new TradeIPBSData(currenttime,
				 tau,

				 pt_sum,
				 t_sum,
				 p_offset,

				 checkTV(BookItem.BOOK_BUY,BookItem.BOOK_BUY,np,currenttime,tau,tb,tsize,p_offset),
				 checkTV(BookItem.BOOK_BUY,BookItem.BOOK_SELL,np,currenttime,tau,tb,tsize,p_offset),
				 checkTV(BookItem.BOOK_SELL,BookItem.BOOK_BUY,np,currenttime,tau,tb,tsize,p_offset),
				 checkTV(BookItem.BOOK_SELL,BookItem.BOOK_SELL,np,currenttime,tau,tb,tsize,p_offset),
				 checkT(BookItem.BOOK_BUY,np,currenttime,tau,tb,tsize,p_offset),
				 m[0],
				 checkT(BookItem.BOOK_SELL,np,currenttime,tau,tb,tsize,p_offset),
				 m[1],
				 m[2],

				 vol,
				 np,
				 null,
				 null,
				 null);
    }


    private static double [][][] checkBxS(final int np,
					  final long currenttime,
					  final long tau,
					  final BookItem [] tb,
					  final int tsize,
					  final double p_offset)
    {
	final double atau=1.0/tau;
	final double [][] moments_p_buy_x_psell=new double [np][np],
	    moments_p_buy_x_p_buy=new double [np][np],
	    moments_p_sell_x_p_sell=new double [np][np];
	BookItem processed=null,typeB=null,typeS=null;
	long prevt=-1;
	for(int i=0;i<=tsize;i++)
	{
	    final BookItem b;
	    if(i<tsize)
	    {
		b=tb[i];
		prevt=b.time;
	    }
	    else
	    {
		b=null;
		prevt=currenttime;
	    }
	    if(processed!=null && typeB!=null && typeS!=null)
	    {
		final double ex=Math.exp(-(currenttime-processed.time)*atau),
		    twmw=((prevt-processed.time)*atau)*ex,
		    dp_buy=typeB.price-p_offset,dp_sell=typeS.price-p_offset;
		double wbuy=1,wsell=1;
		for(int m=0;m<np;m++)
		{
		    double wbs=wbuy,wb=wbuy,ws=wsell;
		    for(int k=0;k<np;k++)
		    {
			moments_p_buy_x_psell[m][k]+=twmw*wbs;
			moments_p_buy_x_p_buy[m][k]+=twmw*wb;
			moments_p_sell_x_p_sell[m][k]+=twmw*ws;
			wbs*=dp_sell;
			wb*=dp_buy;
			ws*=dp_sell;
		    }
		    wbuy*=dp_buy;
		    wsell*=dp_sell;
		}
	    }
	    if(b==null)
	    {
		continue;
	    }
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		typeB=b;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		typeS=b;
	    }
	    processed=b;
	}
	return new double [][][]{moments_p_buy_x_psell,moments_p_buy_x_p_buy,moments_p_sell_x_p_sell};
    }

    private static double [][] checkT(final byte type,
				      final int np,
				      final long currenttime,
				      final long tau,
				      final BookItem [] tb,
				      final int tsize,
				      final double p_offset)
    {
	final double atau=1.0/tau;
	final double [][] matr=new double [np][np];
	BookItem processed=null;
	long prevt=-1;
	for(int i=0;i<=tsize;i++)
	{
	    final BookItem b;
	    if(i<tsize)
	    {
		b=tb[i];
		prevt=b.time;
		if(b.buy_sell_indicator!=type)
		{
		    continue;
		}
	    }
	    else
	    {
		b=null;
		prevt=currenttime;
	    }
	    if(processed!=null)
	    {
		final double ex=Math.exp(-(currenttime-processed.time)*atau),
		    twmw=((prevt-processed.time)*atau)*ex,
		    dp=processed.price-p_offset;
		double w=1;
		for(int m=0;m<np;m++)
		{
		    double wbs=w;
		    for(int k=0;k<np;k++)
		    {
			matr[m][k]+=twmw*wbs;
			wbs*=dp;
		    }
		    w*=dp;
		}
	    }
	    processed=b;
	}
	return matr;
    }

    private static double [] checkTV(final byte type_p,
				     final byte type_v,
				     final int np,
				     final long currenttime,
				     final long tau,
				     final BookItem [] tb,
				     final int tsize,
				     final double p_offset)
    {
	final double atau=1.0/tau;
	final double [] matr=new double [np];
	BookItem item_p=null,processed=null;
	long prevt=-1;
	for(int i=0;i<tsize;i++)
	{
	    final BookItem b=tb[i];
	    prevt=b.time;
	    if(b.buy_sell_indicator==type_p)
	    {
		item_p=b;
	    }
	    if(b.buy_sell_indicator!=type_v)
	    {
		continue;
	    }
	    processed=b;

	    if(processed!=null && item_p!=null)
	    {
		final double ex=Math.exp(-(currenttime-processed.time)*atau),
		    bwmw=processed.shares*ex,
		    dp=item_p.price-p_offset;
		double w=1;
		for(int m=0;m<np;m++)
		{
		    matr[m]+=bwmw*w;
		    w*=dp;
		}
	    }
	}
	return matr;
    }

    private static void checkMatching(final TradeIPBSData r,final TradeIPBSData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.np!=ref.np)
	{
	    throw new RuntimeException("r.np="+r.np+" ref.np="+ref.np);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	final int np=r.np;
	for(int m=0;m<np;m++)
	{
	    UnitTests.checkDoubleRel(r.moments_p_buy_v_buy[m],"r.moments_p_buy_v_buy["+m+"]",ref.moments_p_buy_v_buy[m],"ref.moments_p_buy_v_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_p_buy_v_sell[m],"r.moments_p_buy_v_sell["+m+"]",ref.moments_p_buy_v_sell[m],"ref.moments_p_buy_v_sell["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_p_sell_v_buy[m],"r.moments_p_sell_v_buy["+m+"]",ref.moments_p_sell_v_buy[m],"ref.moments_p_sell_v_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_p_sell_v_sell[m],"r.moments_p_sell_v_sell["+m+"]",ref.moments_p_sell_v_sell[m],"ref.moments_p_sell_v_sell["+m+"]",eps);
	    for(int k=0;k<np;k++)
	    {
		UnitTests.checkDoubleRel(r.moments_p_buy_pt_buy[m][k],"r.moments_p_buy_pt_buy["+m+"]["+k+"]",ref.moments_p_buy_pt_buy[m][k],"ref.moments_p_buy_pt_buy["+m+"]["+k+"]",eps);
		UnitTests.checkDoubleRel(r.moments_p_sell_pt_sell[m][k],"r.moments_p_sell_pt_sell["+m+"]["+k+"]",ref.moments_p_sell_pt_sell[m][k],"ref.moments_p_sell_pt_sell["+m+"]["+k+"]",eps);
		UnitTests.checkDoubleRel(r.moments_p_buy_x_p_sell[m][k],"r.moments_p_buy_x_p_sell["+m+"]["+k+"]",ref.moments_p_buy_x_p_sell[m][k],"ref.moments_p_buy_x_p_sell["+m+"]["+k+"]",eps);
		UnitTests.checkDoubleRel(r.moments_p_buy_x_p_buy[m][k],"r.moments_p_buy_x_p_buy["+m+"]["+k+"]",ref.moments_p_buy_x_p_buy[m][k],"ref.moments_p_buy_x_p_buy["+m+"]["+k+"]",eps);
		UnitTests.checkDoubleRel(r.moments_p_sell_x_p_sell[m][k],"r.moments_p_sell_x_p_sell["+m+"]["+k+"]",ref.moments_p_sell_x_p_sell[m][k],"ref.moments_p_sell_x_p_sell["+m+"]["+k+"]",eps);

		if(false)
		{
		    // matches approximately
		    UnitTests.checkDoubleRel(r.moments_p_sell_pt_sell[m][k],"r.moments_p_sell_pt_sell["+m+"]["+k+"]",r.moments_p_sell_x_p_sell[m][k],"r.moments_p_sell_x_p_sell["+m+"]["+k+"]",eps);
		    UnitTests.checkDoubleRel(r.moments_p_buy_pt_buy[m][k],"r.moments_p_buy_pt_buy["+m+"]["+k+"]",r.moments_p_buy_x_p_buy[m][k],"r.moments_p_buy_x_p_buy["+m+"]["+k+"]",eps);
		}

	    }
	}
	System.err.println("TradeIPBSData:checkMatching(): matching OK");
    }

    static void checkMatchingTradeIPS(final TradeIPSData r,final TradeIPBSData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	if(r.S_FACTOR_POWER==0)
	{
	    final int np=r.np;
	    for(int m=0;m<np;m++)
	    {
		UnitTests.checkDoubleRel(r.moments_pvS_p_buy_v_buy[m],"r.moments_pvS_p_buy_v_buy["+m+"]",ref.moments_p_buy_v_buy[m],"ref.moments_p_buy_v_buy["+m+"]",eps);
		UnitTests.checkDoubleRel(r.moments_pvS_p_buy_v_sell[m],"r.moments_pvS_p_buy_v_sell["+m+"]",ref.moments_p_buy_v_sell[m],"ref.moments_p_buy_v_sell["+m+"]",eps);
		UnitTests.checkDoubleRel(r.moments_pvS_p_sell_v_buy[m],"r.moments_pvS_p_sell_v_buy["+m+"]",ref.moments_p_sell_v_buy[m],"ref.moments_p_sell_v_buy["+m+"]",eps);
		UnitTests.checkDoubleRel(r.moments_pvS_p_sell_v_sell[m],"r.moments_pvS_p_sell_v_sell["+m+"]",ref.moments_p_sell_v_sell[m],"ref.moments_p_sell_v_sell["+m+"]",eps);

		for(int k=0;k<np;k++)
		{
		    UnitTests.checkDoubleRel(r.moments_tSS_p_buy_x_p_sell[m][k],"r.moments_tSS_p_buy_x_p_sell["+m+"]["+k+"]",ref.moments_p_buy_x_p_sell[m][k],"ref.moments_p_buy_x_p_sell["+m+"]["+k+"]",eps);

		    UnitTests.checkDoubleRel(r.moments_ptSS_buy[m+k],"r.moments_ptSS_buy["+(m+k)+"]",ref.moments_p_buy_x_p_buy[m][k],"ref.moments_p_buy_x_p_buy["+m+"]["+k+"]",eps);
		    UnitTests.checkDoubleRel(r.moments_ptSS_sell[m+k],"r.moments_ptSS_sell["+(m+k)+"]",ref.moments_p_sell_x_p_sell[m][k],"ref.moments_p_sell_x_p_sell["+m+"]["+k+"]",eps);
		}
	    }
	    System.err.println("TradeIPBSData:checkMatchingTradeIPS(): t-moments matched OK");
	}
	else
	{
	    System.err.println("TradeIPBSData:checkMatchingTradeIPS(): r.S_FACTOR_POWER="+r.S_FACTOR_POWER+" SKIPPING MOMENTS MATCH");
	}

	System.err.println("TradeIPBSData:checkMatchingTradeIPS(): matching OK");
    }

    private static void matchWithTest(final String description,
				      final TradeIPBSData r,
				      final Trade [] tb,
				      final int tsize,
				      final double eps)
    {
	System.err.println("TradeIPBSData:matchWithTest():======================"+description);
	final int np=r.np;

	final TradeIPBSData rt=calculateTestTradeIPBSData(np,
							  r.currenttime,
							  r.tau,
							  tb,
							  tsize);
	checkMatching(r,rt,eps);

	final TradeIPData rtipd=TradeIPData.calculateTestTradeIPData(np,
								     r.currenttime,
								     r.tau,
								     tb,
								     tsize,
								     true);
	TradeIPData.checkMatchingTradeIPBS(r,rtipd,eps);
	System.err.println("TradeIPBSData:matchWithTest(): OK");
    }

     private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("TradeIPBSData:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);

	final TradeIPBSData ro=calculateTradeIPBSData(N_EM,
						      currenttime,
						      tau,
						      tb,
						      tsize,
						      null);
	matchWithTest("INITIAL CALCULATION",ro,tb,tsize,eps);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeIPBSData em_ref=calculateTradeIPBSData(N_EM,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest("GET REFERENCE FOR RECURRENT",em_ref,tb,n_split,eps);

	final TradeIPBSData em_fromref=calculateTradeIPBSData(N_EM,currenttime,tau,tb,tsize,em_ref);
	checkMatching(ro,em_fromref,eps);
	matchWithTest("RECURRENT CALCULATIONS",em_fromref,tb,tsize,eps);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeIPBSData emc_cut=calculateTradeIPBSData(N_EM,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut,eps);

	System.err.println("TradeIPBSData:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{3,4,N_EMOMENTS,3,4,5})
	{
	    for(int i=0;i<100;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeIPBSData::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }


static TradeIPBSData calculateTradeIPBSData(final int N_EM,
					    final long currenttime,
					    final long tau,
					    final BookItem [] tb,
					    final int tsize,
					    final TradeIPBSData ref)
{
    final int np=N_EM;
    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime||np!=ref.np)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " np="+np+" ref.np="+ref.np+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }

    // calculate 0th & 1st moments
    double pt_sum=0,t_sum=0;
    long prevt=currenttime,tinterval=currenttime,vol=0;
    int ptr_start=tsize;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	vol+=b.shares;
	final double twmw=Math.exp(-(currenttime-prevt)*atau)*((tinterval-prevt)*atau);
	pt_sum+=b.price*twmw;
	t_sum+=twmw;
	tinterval=prevt;
	ptr_start=i;
    }
    BookItem trade_last,trade_last_buy,trade_last_sell;
    long tinterval_buy,tinterval_sell;
    final double ref_exp_delay;
    double exp_buy,exp_sell,exp_last;
    if(bref!=null)
    { // optimization
	// add missed half slice
	exp_last=Math.exp(-(currenttime-bref.time)*atau);
	final double ref_dtr_exp=exp_last*((tinterval-ref.currenttime)*atau);
	pt_sum+=ref_dtr_exp*bref.price;
	t_sum+=ref_dtr_exp;

	// previous interval
	ref_exp_delay=Math.exp(-(currenttime-ref.currenttime)*atau);
	pt_sum+=ref_exp_delay*ref.pt_sum;
	t_sum+=ref_exp_delay*ref.t_sum;
	vol+=ref.vol;

	// init values for second pass
	trade_last=bref;
	trade_last_buy=ref.trade_last_buy;
	exp_buy=trade_last_buy!=null?Math.exp(-(currenttime-trade_last_buy.time)*atau):0;
	trade_last_sell=ref.trade_last_sell;
	exp_sell=trade_last_sell!=null?Math.exp(-(currenttime-trade_last_sell.time)*atau):0;
	tinterval=tinterval_buy=tinterval_sell=ref.currenttime;
    }
    else
    {
	ref_exp_delay=exp_buy=exp_sell=exp_last=0;
	trade_last=trade_last_buy=trade_last_sell=null;
	tinterval=tinterval_buy=tinterval_sell=currenttime;
    }
    final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);


    final double []
	moments_p_buy_v_buy=new double [np],
	moments_p_buy_v_sell=new double [np],
	moments_p_sell_v_buy=new double [np],
	moments_p_sell_v_sell=new double [np];
    final double [][]
	moments_p_buy_pt_buy=new double [np][np],
	moments_p_buy_x_p_sell=new double [np][np],
	moments_p_sell_pt_sell=new double [np][np],
	moments_p_buy_x_p_buy=new double [np][np],
	moments_p_sell_x_p_sell=new double [np][np];
    prevt=Long.MIN_VALUE;

    // calculate with given offset. using average as offset reduces rounding errors & degeneracy.
    for(int i=ptr_start;i<=tsize;i++)
    {
	final double bw_pbuy_vbuy,bw_pbuy_vsell,bw_psell_vbuy,bw_psell_vsell,tw_buy,tw_sell,tw_buy_x_sell,dpl_buy,dpl_sell,dp_buy,dp_sell;
	if(i>=tsize)
	{ // last slice up to currenttime
	    prevt=currenttime;
	    bw_pbuy_vbuy=bw_pbuy_vsell=bw_psell_vbuy=bw_psell_vsell=dp_buy=dp_sell=0;
	    tw_buy=exp_buy*((prevt-tinterval_buy)*atau);
	    tw_sell=exp_sell*((prevt-tinterval_sell)*atau);
	    tw_buy_x_sell=exp_last*((prevt-tinterval)*atau);
	    dpl_buy=trade_last_buy!=null?trade_last_buy.price-p_offset:0;
	    dpl_sell=trade_last_sell!=null?trade_last_sell.price-p_offset:0;
	}
	else
	{
	    final BookItem b=tb[i];
	    prevt=b.time;
	    final int tp=b.buy_sell_indicator;
	    final double ex=Math.exp(-(currenttime-prevt)*atau),dp=b.price-p_offset;

	    if(trade_last_buy!=null)
	    {
		tw_buy=tp==BookItem.BOOK_BUY?exp_buy*((prevt-tinterval_buy)*atau):0;
		dpl_buy=trade_last_buy.price-p_offset;
	    }
	    else
	    {
		tw_buy=dpl_buy=0;
	    }
	    if(trade_last_sell!=null)
	    {
		tw_sell=tp==BookItem.BOOK_SELL?exp_sell*((prevt-tinterval_sell)*atau):0;
		dpl_sell=trade_last_sell.price-p_offset;
		if(trade_last!=null && trade_last_buy!=null)
		{
		    tw_buy_x_sell=exp_last*((prevt-tinterval)*atau);
		}
		else
		{
		    tw_buy_x_sell=0;
		}
	    }
	    else
	    {
		tw_sell=dpl_sell=tw_buy_x_sell=0;
	    }
	    if(tp==BookItem.BOOK_BUY)
	    {
		trade_last_buy=b;
		exp_buy=ex;
		tinterval_buy=prevt;
		dp_buy=dp;
		dp_sell=dpl_sell;
		bw_pbuy_vbuy=ex*b.shares;
		bw_psell_vbuy=trade_last_sell!=null?bw_pbuy_vbuy:0;
		bw_pbuy_vsell=bw_psell_vsell=0;
	    }
	    else if(tp==BookItem.BOOK_SELL)
	    {
		trade_last_sell=b;
		exp_sell=ex;
		tinterval_sell=prevt;
		dp_buy=dpl_buy;
		dp_sell=dp;
		bw_pbuy_vbuy=bw_psell_vbuy=0;
		bw_psell_vsell=ex*b.shares;
		bw_pbuy_vsell=trade_last_buy!=null?bw_psell_vsell:0;
	    }
	    else
	    {
		bw_pbuy_vbuy=bw_pbuy_vsell=bw_psell_vbuy=bw_psell_vsell=dp_buy=dp_sell=0;
	    }
	    trade_last=b;
	    exp_last=ex;
	    tinterval=prevt;
	}
	double wdp_buy=1,wdp_sell=1,wdpl_buy=1,wdpl_sell=1;
	for(int m=0;m<np;m++)
	{
	    moments_p_buy_v_buy[m]+=wdp_buy*bw_pbuy_vbuy;
	    moments_p_buy_v_sell[m]+=wdp_buy*bw_pbuy_vsell;
	    moments_p_sell_v_buy[m]+=wdp_sell*bw_psell_vbuy;
	    moments_p_sell_v_sell[m]+=wdp_sell*bw_psell_vsell;
	    wdp_buy*=dp_buy;
	    wdp_sell*=dp_sell;


	    final double [] x=moments_p_buy_pt_buy[m],y=moments_p_buy_x_p_sell[m],z=moments_p_sell_pt_sell[m],
		xx=moments_p_buy_x_p_buy[m],zz=moments_p_sell_x_p_sell[m];
	    double buy_buy=wdpl_buy,sell_sell=wdpl_sell,wdpl_y=wdpl_buy*tw_buy_x_sell;
	    for(int k=0;k<np;k++)
	    {
		x[k]+=tw_buy*buy_buy;
		xx[k]+=tw_buy_x_sell*buy_buy;
		y[k]+=wdpl_y;
		z[k]+=tw_sell*sell_sell;
		zz[k]+=tw_buy_x_sell*sell_sell;
		buy_buy*=dpl_buy;
		wdpl_y*=dpl_sell;
		sell_sell*=dpl_sell;
	    }
	    wdpl_buy*=dpl_buy;
	    wdpl_sell*=dpl_sell;
	}
    }

    if(bref!=null)
    { // optimization

	// convert old interval to new p_offest basis
	final double dpa=ref.p_offset-p_offset;

	for(int l=0;l<np;l++)
	{
	    // P(np)D1 polynomial group transform
	    double wv=1,s_p_buy_v_buy=0,s_p_buy_v_sell=0,s_p_sell_v_buy=0,s_p_sell_v_sell=0;
	    long fv=1;
	    for(int ls=l;ls>=0;ls--)
	    {
		final double a=fv*wv;
		s_p_buy_v_buy+=a*ref.moments_p_buy_v_buy[ls];
		s_p_buy_v_sell+=a*ref.moments_p_buy_v_sell[ls];
		s_p_sell_v_buy+=a*ref.moments_p_sell_v_buy[ls];
		s_p_sell_v_sell+=a*ref.moments_p_sell_v_sell[ls];
		wv*=dpa;
		fv*=ls;
		fv/=1+l-ls;
	    }
	    moments_p_buy_v_buy[l]+=ref_exp_delay*s_p_buy_v_buy;
	    moments_p_buy_v_sell[l]+=ref_exp_delay*s_p_buy_v_sell;
	    moments_p_sell_v_buy[l]+=ref_exp_delay*s_p_sell_v_buy;
	    moments_p_sell_v_sell[l]+=ref_exp_delay*s_p_sell_v_sell;

	    for(int m=0;m<np;m++)
	    {
		double wl=1,s_p_buy_pt_buy=0,s_p_buy_x_p_sell=0,s_p_sell_pt_sell=0,s_p_buy_x_p_buy=0,s_p_sell_x_p_sell=0;
		// P(np)D2 polynomial group transform
		long fl=1;
		for(int ls=l;ls>=0;ls--)
		{
		    double wm=fl*wl;
		    long fm=1;
		    for(int ms=m;ms>=0;ms--)
		    {
			final double a=fm*wm;
			s_p_buy_pt_buy+=a*ref.moments_p_buy_pt_buy[ls][ms];
			s_p_buy_x_p_sell+=a*ref.moments_p_buy_x_p_sell[ls][ms];
			s_p_sell_pt_sell+=a*ref.moments_p_sell_pt_sell[ls][ms];
			s_p_buy_x_p_buy+=a*ref.moments_p_buy_x_p_buy[ls][ms];
			s_p_sell_x_p_sell+=a*ref.moments_p_sell_x_p_sell[ls][ms];
			wm*=dpa;
			fm*=ms;
			fm/=1+m-ms;
		    }
		    wl*=dpa;
		    fl*=ls;
		    fl/=1+l-ls;
		}
		moments_p_buy_pt_buy[l][m]+=ref_exp_delay*s_p_buy_pt_buy;
		moments_p_buy_x_p_sell[l][m]+=ref_exp_delay*s_p_buy_x_p_sell;
		moments_p_sell_pt_sell[l][m]+=ref_exp_delay*s_p_sell_pt_sell;
		moments_p_buy_x_p_buy[l][m]+=ref_exp_delay*s_p_buy_x_p_buy;
		moments_p_sell_x_p_sell[l][m]+=ref_exp_delay*s_p_sell_x_p_sell;
	    }
	}
    }

    return new TradeIPBSData(currenttime,
			     tau,

			     pt_sum,
			     t_sum,
			     p_offset,

			     moments_p_buy_v_buy,
			     moments_p_buy_v_sell,
			     moments_p_sell_v_buy,
			     moments_p_sell_v_sell,
			     moments_p_buy_pt_buy,
			     moments_p_buy_x_p_sell,
			     moments_p_sell_pt_sell,

			     moments_p_buy_x_p_buy,
			     moments_p_sell_x_p_sell,

			     vol,
			     np,
			     trade_last,
			     trade_last_buy,
			     trade_last_sell);

}

    static DBSTradeIP<TradeIPBSData> calculateDBSTradeIP(final int N_EM,
							 final long currenttime,
							 final long tau,
							 final BookItem [] tb,
							 final int tsize,
							 final DBSTradeIP<TradeIPBSData> ref)
    {
	final TradeIPBSData o=TradeIPBSData.calculateTradeIPBSData(N_EM+1,
								   currenttime,
								   tau,
								   tb,
								   tsize,
								   ref!=null?ref.odata:null);
	return new DBSTradeIP<TradeIPBSData>(o.moments_p_buy_x_p_buy,
					     o.moments_p_sell_x_p_sell,
					     o.moments_p_buy_x_p_sell,
					     o.moments_p_buy_v_buy,
					     o.moments_p_buy_v_sell,
					     o.moments_p_sell_v_buy,
					     o.moments_p_sell_v_sell,
					     o);
    }

} // END OF class TradeIPBSData



static class DPTradeIP extends PTradeIP {
    final double pa_Ihigh,pa_Ilow,pa_m,pr_m,min_IA,max_IA,ex_IA,min_IR,max_IR,ex_IR;
    final double [] Iplow_p_up,Iphigh_p_up,Iplow_p_down,Iphigh_p_down;
    final PTopology ptopology;
    DPTradeIP(final TradeIPRData odata)
    {
	super(null,null,odata);
	final double p_offset=super.p_offset;
	final double [] pcoefs=super.pcoefs,norms=super.normformoments,reroots=super.p_real_roots;
	if(true)
	{
	    if(this.p_down.length<=1||this.p_down.length>2)
	    {
		this.pa_Ilow=Double.NaN;
		this.pa_Ihigh=this.p_down.length==1?this.p_down[0]:Double.NaN;
	    }
	    else
	    {
		final double i0=getI1Integrated(pcoefs,this.p_down[0]-p_offset,norms),
		    i1=getI1Integrated(pcoefs,this.p_down[1]-p_offset,norms);
		if(i0>i1)
		{
		    this.pa_Ihigh=this.p_down[0];
		    this.pa_Ilow=this.p_down[1];
		}
		else
		{
		    this.pa_Ihigh=this.p_down[1];
		    this.pa_Ilow=this.p_down[0];
		}
	    }
	}

	if(true)
	{
	    double pa=Double.NaN,pr=Double.NaN,ia=Double.NEGATIVE_INFINITY;
	    final int ne=reroots.length-1;
	    for(int k=0;k<=ne;k++)
	    {
		final double rr=reroots[k]; // no offset for reroots
		if(getD1WithScale(pcoefs,rr,norms)<0)
		{ // p_down
		    final double i=getI1Integrated(pcoefs,rr,norms);
		    if(i>ia)
		    {
			ia=i;
			pa=rr;
			if(k<=0)
			{
			    pr=k>=ne?Double.NaN:reroots[k+1];
			}
			else if(k>=ne)
			{
			    pr=k<=0?Double.NaN:reroots[k-1];
			}
			else
			{
			    final double im1=getI1Integrated(pcoefs,reroots[k-1],norms),
				ip1=getI1Integrated(pcoefs,reroots[k+1],norms);
			    if(im1>ip1)
			    {
				pr=reroots[k-1];
			    }
			    else
			    {
				pr=reroots[k+1];
			    }
			}
		    }
		}
	    }
	    this.pa_m=pa+p_offset;
	    this.pr_m=pr+p_offset;
	}

	if(true)
	{
	    final int np=pcoefs.length;
	    final double [] ipcoefs=new double [np+1],inorm=PTradeIP.getArray1Filled(np+1);
	    ipcoefs[0]=0;
	    for(int k=1;k<=np;k++)
	    {
		ipcoefs[k]=pcoefs[k-1]/k;
	    }
	    final double [] xp_down=new double [this.p_down.length];
	    for(int k=xp_down.length;--k>=0;xp_down[k]=this.p_down[k]-p_offset){}
	    final double [] sar=TradeIP.getmmAR(ipcoefs,inorm,xp_down,true);
	    this.min_IA=sar[0]+p_offset;
	    this.max_IA=sar[1]+p_offset;
	    this.ex_IA=sar[2]+p_offset;
	    final double [] xp_up=xp_down.length==this.p_up.length?xp_down:new double [this.p_up.length];
	    for(int k=xp_up.length;--k>=0;xp_up[k]=this.p_up[k]-p_offset){}
	    final double [] srr=TradeIP.getmmAR(ipcoefs,inorm,xp_up,false);
	    this.min_IR=srr[0]+p_offset;
	    this.max_IR=srr[1]+p_offset;
	    this.ex_IR=srr[2]+p_offset;


	    final double [] reslh=new double [2];
	    this.Iplow_p_up=new double [this.p_up.length];
	    this.Iphigh_p_up=new double [this.p_up.length];
	    for(int k=this.p_up.length;--k>=0;)
	    {
		// inorm=1
		TradeIP.getPointsCloseToEXmmAR(reslh,ipcoefs,inorm,this.p_up[k]-p_offset);
		this.Iplow_p_up[k]=reslh[0]+p_offset;
		this.Iphigh_p_up[k]=reslh[1]+p_offset;
	    }
	    this.Iplow_p_down=new double [this.p_down.length];
	    this.Iphigh_p_down=new double [this.p_down.length];
	    for(int k=this.p_down.length;--k>=0;)
	    {
		TradeIP.getPointsCloseToEXmmAR(reslh,ipcoefs,inorm,this.p_down[k]-p_offset);
		this.Iplow_p_down[k]=reslh[0]+p_offset;
		this.Iphigh_p_down[k]=reslh[1]+p_offset;
	    }

	    this.ptopology=getPTopology();
	}

    }
    int getNP(final int np0){return np0-1;}
    double getIEq(){return 0;}
    double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
    double getI1(final double p)
    {
	return getI1Integrated(super.pcoefs,p-super.p_offset,super.normformoments);
    }
    double getDI1(final double p)
    {
	return getD0T(p);
    }
    double getD2I1(final double p)
    {
	return getD1T(p);
    }

    TCode getTCode(final int cmd,final double p)
    {
	final double [] i_p_up=new double [this.p_up.length],i_p_down=new double [this.p_down.length];
	for(int i=i_p_up.length;--i>=0;)
	{
	    i_p_up[i]=getI1(this.p_up[i]);
	}
	for(int i=i_p_down.length;--i>=0;)
	{
	    i_p_down[i]=getI1(this.p_down[i]);
	}

	return TCode.getSTCode(cmd,p,getD0WithScale(super.pcoefs,0,super.normformoments),
			       this.p_up,
			       i_p_up,
			       this.p_down,
			       i_p_down,
			       this.Iplow_p_down,
			       this.Iphigh_p_down);
    }

    private PTopology getPTopology()
    {
	final double [] i_p_up=new double [this.p_up.length],i_p_down=new double [this.p_down.length];
	for(int i=i_p_up.length;--i>=0;)
	{
	    i_p_up[i]=getI1(this.p_up[i]);
	}
	for(int i=i_p_down.length;--i>=0;)
	{
	    i_p_down[i]=getI1(this.p_down[i]);
	}

	return new PTopology(getD0WithScale(super.pcoefs,0,super.normformoments),
			     this.p_up,
			     i_p_up,
			     this.p_down,
			     i_p_down,
			     this.Iplow_p_down,
			     this.Iphigh_p_down);
    }

}


static class DBSTradeIP<T extends TradeIPRData> {
    final DPTradeIP total,diff,diff2,kdiff;
    private final T odata;
    DBSTradeIP(final double [][] bb,final double [][] ss,final double [][] bs,
	       final double [] pbvb,final double [] pbvs,final double [] psvb,final double [] psvs,
	       final T odata)
    {
	this.odata=odata;
	this.total=new DPTradeIP(odata){
		double getIEq(){return 1;}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np][np];
		    final double [] rpart=new double [np];
		    final double normm=bb[0][0]+ss[0][0]+2*bs[0][0],normr=psvs[0]+psvb[0]+pbvb[0]+pbvs[0];
		    for(int k=0;k<np;k++)
		    {
			for(int l=0;l<np;l++)
			{
			    matr[k][l]=(bb[k][l]+ss[k][l]+bs[k][l]+bs[l][k])/normm;
			}
			rpart[k]=(psvs[k]+psvb[k]+pbvb[k]+pbvs[k])/normr;
		    }
		    TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
		    return rpart;
		}
	    };
	this.diff=new DPTradeIP(odata){
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np][np];
		    final double [] rpart=new double [np];
		    for(int k=0;k<np;k++)
		    {
			final int k1=k+1;
			for(int l=0;l<np;l++)
			{
			    final int l1=l+1;
			    matr[k][l]=bb[k1][l1]-bs[k1][l1]+ss[k1][l1]-bs[l1][k1];
			}
			rpart[k]=psvs[k1]-psvb[k1]+pbvb[k1]-pbvs[k1];
		    }
		    TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
		    for(int k=0;k<np;k++)
		    {
			rpart[k]*=k+1;
		    }
		    return rpart;
		}
	    };
	this.kdiff=new DPTradeIP(odata){
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np][np];
		    final double [] rpart=new double [np];
		    final double [] bt=bb[0],st=ss[0];
		    final double t0=bt[0],vb0=pbvb[0],vs0=psvs[0];
		    for(int k=0;k<np;k++)
		    {
			final int k1=k+1;
			for(int l=0;l<np;l++)
			{
			    final int l1=l+1;
			    matr[k][l]=(t0*bb[k1][l1]-bt[k1]*bt[l1])
				      -(t0*bs[k1][l1]-bt[k1]*st[l1])
				      +(t0*ss[k1][l1]-st[k1]*st[l1])
				      -(t0*bs[l1][k1]-bt[l1]*st[k1]);
			}
			rpart[k]=(t0*psvs[k1]-st[k1]*vs0)
			        -(t0*psvb[k1]-st[k1]*vb0)
			        +(t0*pbvb[k1]-bt[k1]*vb0)
			        -(t0*pbvs[k1]-bt[k1]*vs0);
		    }
		    TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
		    for(int k=0;k<np;k++)
		    {
			rpart[k]*=k+1;
		    }
		    return rpart;
		}
	    };
	this.diff2=new DPTradeIP(odata){
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np+1][np+1];
		    final double [] rpart=new double [np+1];
		    for(int k=0;k<np;k++)
		    {
			final int k1=k+1;
			for(int l=0;l<np;l++)
			{
			    final int l1=l+1;
			    matr[k][l]=bb[k1][l1]-bs[k1][l1]+ss[k1][l1]-bs[l1][k1];
			}
			matr[k][np]=matr[np][k]=ss[0][k1]-bb[0][k1];
			rpart[k]=psvs[k1]-psvb[k1]+pbvb[k1]-pbvs[k1];
		    }
		    rpart[np]=psvs[0]-pbvb[0];
		    TradeDPVR.solveLinsystem(matr,rpart,np+1,new int [np+1]);
		    final double [] pcoefs=new double [np];
		    for(int k=0;k<np;k++)
		    {
			pcoefs[k]=rpart[k]*(k+1);
		    }
		    return pcoefs;
		}
	    };
    }

} // END OF class DBSTradeIP











static class TradeIPTBSData extends TradeIPRData {

    private static final int S_FACTOR_POWER=0;

    final double pt_sum,t_sum;
    private final double []
	moments_p_buy_v_buy,
	moments_p_buy_v_sell,
	moments_p_sell_v_buy,
	moments_p_sell_v_sell,
	moments_pv
	;
    private final double [][]
	moments_p_buy_x_p_sell,
	moments_p_buy_x_p_buy,
	moments_p_sell_x_p_sell,
	moments_pt;


    // references to previos calculations
    private final Trade trade_last;


    TradeIPTBSData(final long currenttime,
		   final long tau,

		   final double pt_sum,
		   final double t_sum,
		   final double p_offset,

		   final double [] moments_p_buy_v_buy,
		   final double [] moments_p_buy_v_sell,
		   final double [] moments_p_sell_v_buy,
		   final double [] moments_p_sell_v_sell,
		   final double [] moments_pv,

		   final double [][] moments_p_buy_x_p_sell,
		   final double [][] moments_p_buy_x_p_buy,
		   final double [][] moments_p_sell_x_p_sell,
		   final double [][] moments_pt,

		   final long vol,
		   final int np,

		   // references
		   final Trade trade_last)
    {
	super(currenttime,tau,p_offset,vol,np);

	this.pt_sum=pt_sum;
	this.t_sum=t_sum;

	this.moments_p_buy_v_buy=moments_p_buy_v_buy;
	this.moments_p_buy_v_sell=moments_p_buy_v_sell;
	this.moments_p_sell_v_buy=moments_p_sell_v_buy;
	this.moments_p_sell_v_sell=moments_p_sell_v_sell;
	this.moments_pv=moments_pv;

	this.moments_p_buy_x_p_sell=moments_p_buy_x_p_sell;
	this.moments_p_buy_x_p_buy=moments_p_buy_x_p_buy;
	this.moments_p_sell_x_p_sell=moments_p_sell_x_p_sell;
	this.moments_pt=moments_pt;

	this.trade_last=trade_last;
    }

   static TradeIPTBSData calculateTestTradeIPTBSData(final int N_EM,
						     final long currenttime,
						     final long tau,
						     final Trade [] tb,
						     final int tsize)
    {
	final int np=N_EM;
	final double atau=1.0/tau;

	// calculate 0th & 1st moments
	double pt_sum=0,t_sum=0;
	long prevt=currenttime,tinterval=currenttime;
	for(int i=tsize;--i>=0;)
	{
	    final Trade b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    //System.err.println("i="+i+" b.time="+b.time);
	    prevt=b.time;
	    final double twmw=Math.exp(-(currenttime-prevt)*atau)*((tinterval-prevt)*atau);
	    pt_sum+=b.price*twmw;
	    t_sum+=twmw;
	    tinterval=prevt;
	}
	final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);

	final double []
	    moments_p_buy_v_buy=new double [np],
	    moments_p_buy_v_sell=new double [np],
	    moments_p_sell_v_buy=new double [np],
	    moments_p_sell_v_sell=new double [np],
	    moments_pv=new double [np];

	final double [][]
	    moments_p_buy_x_p_sell=new double [np][np],
	    moments_p_buy_x_p_buy=new double [np][np],
	    moments_p_sell_x_p_sell=new double [np][np],
	    moments_pt=new double [np][np];

	prevt=tinterval=currenttime;
	long vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final Trade b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    prevt=b.time;
	    final double ex=Math.exp(-(currenttime-prevt)*atau);
	    final double dp_buy,dp_sell,bwmw_buy,bwmw_sell,S;
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		if(b.best_sell==null)
		{
		    continue;
		}
		dp_buy=b.price-p_offset;
		dp_sell=b.best_sell.price-p_offset;
		S=b.best_sell.price-b.price;
		bwmw_buy=ex*b.shares*getSFactor(S);
		bwmw_sell=0;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		if(b.best_buy==null)
		{
		    continue;
		}
		dp_buy=b.best_buy.price-p_offset;
		dp_sell=b.price-p_offset;
		S=b.price-b.best_buy.price;
		bwmw_buy=0;
		bwmw_sell=ex*b.shares*getSFactor(S);
	    }
	    else
	    {
		continue;
	    }
	    final double twmw=ex*((tinterval-prevt)*atau)*(getSFactor(S)*getSFactor(S)),dp_l=b.price-p_offset,bwmw=ex*b.shares*getSFactor(S);
	    double wdp_buy=1,wdp_sell=1,wdp_l=1;
	    for(int m=0;m<np;m++)
	    {
		moments_p_buy_v_buy[m]+=wdp_buy*bwmw_buy;
		moments_p_buy_v_sell[m]+=wdp_buy*bwmw_sell;
		moments_p_sell_v_buy[m]+=wdp_sell*bwmw_buy;
		moments_p_sell_v_sell[m]+=wdp_sell*bwmw_sell;
		moments_pv[m]+=wdp_l*bwmw;

		final double [] x=moments_p_buy_x_p_buy[m],y=moments_p_buy_x_p_sell[m],z=moments_p_sell_x_p_sell[m],t=moments_pt[m];
		double buy_buy=wdp_buy*twmw,sell_sell=wdp_sell*twmw,buy_sell=buy_buy,l_l=wdp_l*twmw;
		for(int k=0;k<np;k++)
		{
		    x[k]+=buy_buy;
		    y[k]+=buy_sell;
		    z[k]+=sell_sell;
		    t[k]+=l_l;
		    buy_buy*=dp_buy;
		    buy_sell*=dp_sell;
		    sell_sell*=dp_sell;
		    l_l*=dp_l;
		}
		wdp_buy*=dp_buy;
		wdp_sell*=dp_sell;
		wdp_l*=dp_l;
	    }
	    tinterval=prevt;
	    vol+=b.shares;
	}

	return new TradeIPTBSData(currenttime,
				  tau,

				  pt_sum,
				  t_sum,
				  p_offset,

				  moments_p_buy_v_buy,
				  moments_p_buy_v_sell,
				  moments_p_sell_v_buy,
				  moments_p_sell_v_sell,
				  moments_pv,

				  moments_p_buy_x_p_sell,
				  moments_p_buy_x_p_buy,
				  moments_p_sell_x_p_sell,
				  moments_pt,

				  vol,
				  np,
				  null);
    }


    private static double getSFactor(final double S)
    {
	if(S_FACTOR_POWER==0)
	{
	    return 1;
	}
	if(S_FACTOR_POWER==1)
	{
	    return Math.abs(S);
	}
	if(S_FACTOR_POWER==2)
	{
	    return S*S;
	}
	throw new RuntimeException("S_FACTOR_POWER="+S_FACTOR_POWER);
    }

static TradeIPTBSData calculateTradeIPTBSData(final int N_EM,
					      final long currenttime,
					      final long tau,
					      final Trade [] tb,
					      final int tsize,
					      final TradeIPTBSData ref)
{
    final int np=N_EM;
    final double atau=1.0/tau;
    final Trade bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime||np!=ref.np)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " np="+np+" ref.np="+ref.np+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }

    // calculate 0th & 1st moments
    double pt_sum=0,t_sum=0;
    long prevt=currenttime,tinterval=currenttime;
    for(int i=tsize;--i>=0;)
    {
	final Trade b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	final double twmw=Math.exp(-(currenttime-prevt)*atau)*((tinterval-prevt)*atau);
	pt_sum+=b.price*twmw;
	t_sum+=twmw;
	tinterval=prevt;
    }

    final double exp_last,ref_exp_delay;
    if(bref!=null)
    { // optimization
	// add missed half slice
	exp_last=Math.exp(-(currenttime-bref.time)*atau);
	final double ref_dtr_exp=exp_last*((tinterval-ref.currenttime)*atau);
	pt_sum+=ref_dtr_exp*bref.price;
	t_sum+=ref_dtr_exp;

	// previous interval
	ref_exp_delay=Math.exp(-(currenttime-ref.currenttime)*atau);
	pt_sum+=ref_exp_delay*ref.pt_sum;
	t_sum+=ref_exp_delay*ref.t_sum;
    }
    else
    {
	exp_last=ref_exp_delay=0;
    }
    final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);


    final double []
	moments_p_buy_v_buy=new double [np],
	moments_p_buy_v_sell=new double [np],
	moments_p_sell_v_buy=new double [np],
	moments_p_sell_v_sell=new double [np],
	moments_pv=new double [np];
    final double [][]
	moments_p_buy_x_p_sell=new double [np][np],
	moments_p_buy_x_p_buy=new double [np][np],
	moments_p_sell_x_p_sell=new double [np][np],
	moments_pt=new double [np][np];

    Trade trade_last=null;
    prevt=tinterval=currenttime;
    long vol=0;
    // calculate with given offset. using average as offset reduces rounding errors & degeneracy.
    for(int i=tsize;--i>=0;)
    {
	final Trade b=tb[i];
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	final int tp=b.buy_sell_indicator;
	final double ex=Math.exp(-(currenttime-prevt)*atau),dp_buy,dp_sell,bwmw_buy,bwmw_sell,bwmw,dp_l,S;
	if(tp==BookItem.BOOK_BUY)
	{
	    if(b.best_sell==null)
	    {
		continue;
	    }
	    dp_l=dp_buy=b.price-p_offset;
	    dp_sell=b.best_sell.price-p_offset;
	    S=b.best_sell.price-b.price;
	    bwmw=bwmw_buy=ex*b.shares*getSFactor(S);
	    bwmw_sell=0;
	}
	else if(tp==BookItem.BOOK_SELL)
	{
	    if(b.best_buy==null)
	    {
		continue;
	    }
	    dp_buy=b.best_buy.price-p_offset;
	    dp_l=dp_sell=b.price-p_offset;
	    S=b.price-b.best_buy.price;
	    bwmw_buy=0;
	    bwmw=bwmw_sell=ex*b.shares*getSFactor(S);
	}
	else
	{
	    continue;
	}
	final double twmw=ex*((tinterval-prevt)*atau)*(getSFactor(S)*getSFactor(S));
	double wdp_buy=1,wdp_sell=1,wdp_l=1;
	for(int m=0;m<np;m++)
	{
	    moments_p_buy_v_buy[m]+=wdp_buy*bwmw_buy;
	    moments_p_buy_v_sell[m]+=wdp_buy*bwmw_sell;
	    moments_p_sell_v_buy[m]+=wdp_sell*bwmw_buy;
	    moments_p_sell_v_sell[m]+=wdp_sell*bwmw_sell;
	    moments_pv[m]+=wdp_l*bwmw;

	    final double [] x=moments_p_buy_x_p_buy[m],y=moments_p_buy_x_p_sell[m],z=moments_p_sell_x_p_sell[m],t=moments_pt[m];
	    double buy_buy=wdp_buy*twmw,sell_sell=wdp_sell*twmw,buy_sell=buy_buy,l_l=wdp_l*twmw;
	    for(int k=0;k<np;k++)
	    {
		x[k]+=buy_buy;
		y[k]+=buy_sell;
		z[k]+=sell_sell;
		t[k]+=l_l;
		buy_buy*=dp_buy;
		buy_sell*=dp_sell;
		sell_sell*=dp_sell;
		l_l*=dp_l;
	    }
	    wdp_buy*=dp_buy;
	    wdp_sell*=dp_sell;
	    wdp_l*=dp_l;
	}
	tinterval=prevt;
	if(trade_last==null)
	{
	    trade_last=b;
	}
	vol+=b.shares;
    }


    if(bref!=null)
    { // optimization
	vol+=ref.vol;

	// convert old interval to new p_offest basis
	final double dpa=ref.p_offset-p_offset,bref_dp_buy,bref_dp_sell,bref_dp_l,S;
	if(bref.buy_sell_indicator==BookItem.BOOK_BUY)
	{
	    bref_dp_l=bref_dp_buy=bref.price-p_offset;
	    bref_dp_sell=bref.best_sell.price-p_offset;
	    S=bref.best_sell.price-bref.price;
	}
	else if(bref.buy_sell_indicator==BookItem.BOOK_SELL)
	{
	    bref_dp_buy=bref.best_buy.price-p_offset;
	    bref_dp_l=bref_dp_sell=bref.price-p_offset;
	    S=bref.price-bref.best_buy.price;
	}
	else
	{
	    throw new RuntimeException();
	}
	final double twmw=exp_last*((tinterval-ref.currenttime)*atau)*(getSFactor(S)*getSFactor(S));

	double refwdp_buy=1,refwdp_sell=1,refwdp_l=1;
	for(int l=0;l<np;l++)
	{
	    // P(np-1)D1 polynomial group transform
	    double wv=1,s_p_buy_v_buy=0,s_p_buy_v_sell=0,s_p_sell_v_buy=0,s_p_sell_v_sell=0,s_pv=0;
	    long fv=1;
	    for(int ls=l;ls>=0;ls--)
	    {
		final double a=fv*wv;
		s_p_buy_v_buy+=a*ref.moments_p_buy_v_buy[ls];
		s_p_buy_v_sell+=a*ref.moments_p_buy_v_sell[ls];
		s_p_sell_v_buy+=a*ref.moments_p_sell_v_buy[ls];
		s_p_sell_v_sell+=a*ref.moments_p_sell_v_sell[ls];
		s_pv+=a*ref.moments_pv[ls];

		wv*=dpa;
		fv*=ls;
		fv/=1+l-ls;
	    }
	    moments_p_buy_v_buy[l]+=ref_exp_delay*s_p_buy_v_buy;
	    moments_p_buy_v_sell[l]+=ref_exp_delay*s_p_buy_v_sell;
	    moments_p_sell_v_buy[l]+=ref_exp_delay*s_p_sell_v_buy;
	    moments_p_sell_v_sell[l]+=ref_exp_delay*s_p_sell_v_sell;
	    moments_pv[l]+=ref_exp_delay*s_pv;

	    double refwbb=twmw*refwdp_buy,refwbs=refwbb,refwss=twmw*refwdp_sell,refwll=twmw*refwdp_l;
	    for(int m=0;m<np;m++)
	    {
		double wl=1,s_p_buy_x_p_sell=0,s_p_buy_x_p_buy=0,s_p_sell_x_p_sell=0,s_pt=0;
		// P(np-1)D2 polynomial group transform
		long fl=1;
		for(int ls=l;ls>=0;ls--)
		{
		    double wm=fl*wl;
		    long fm=1;
		    for(int ms=m;ms>=0;ms--)
		    {
			final double a=fm*wm;
			s_p_buy_x_p_sell+=a*ref.moments_p_buy_x_p_sell[ls][ms];
			s_p_buy_x_p_buy+=a*ref.moments_p_buy_x_p_buy[ls][ms];
			s_p_sell_x_p_sell+=a*ref.moments_p_sell_x_p_sell[ls][ms];
			s_pt+=a*ref.moments_pt[ls][ms];
			wm*=dpa;
			fm*=ms;
			fm/=1+m-ms;
		    }
		    wl*=dpa;
		    fl*=ls;
		    fl/=1+l-ls;
		}
		moments_p_buy_x_p_sell[l][m]+=ref_exp_delay*s_p_buy_x_p_sell+refwbs;
		moments_p_buy_x_p_buy[l][m]+=ref_exp_delay*s_p_buy_x_p_buy+refwbb;
		moments_p_sell_x_p_sell[l][m]+=ref_exp_delay*s_p_sell_x_p_sell+refwss;
		moments_pt[l][m]+=ref_exp_delay*s_pt+refwll;

		refwbb*=bref_dp_buy;
		refwbs*=bref_dp_sell;
		refwss*=bref_dp_sell;
		refwll*=bref_dp_l;
	    }
	    refwdp_buy*=bref_dp_buy;
	    refwdp_sell*=bref_dp_sell;
	    refwdp_l*=bref_dp_l;
	}
    }

    return new TradeIPTBSData(currenttime,
			      tau,

			      pt_sum,
			      t_sum,
			      p_offset,

			      moments_p_buy_v_buy,
			      moments_p_buy_v_sell,
			      moments_p_sell_v_buy,
			      moments_p_sell_v_sell,
			      moments_pv,

			      moments_p_buy_x_p_sell,
			      moments_p_buy_x_p_buy,
			      moments_p_sell_x_p_sell,
			      moments_pt,

			      vol,
			      np,
			      trade_last==null?bref:trade_last);

}


    private static void checkMatching(final TradeIPTBSData r,final TradeIPTBSData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.np!=ref.np)
	{
	    throw new RuntimeException("r.np="+r.np+" ref.np="+ref.np);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	final int np=r.np;
	for(int m=0;m<np;m++)
	{
	    UnitTests.checkDoubleRel(r.moments_p_buy_v_buy[m],"r.moments_p_buy_v_buy["+m+"]",ref.moments_p_buy_v_buy[m],"ref.moments_p_buy_v_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_p_buy_v_sell[m],"r.moments_p_buy_v_sell["+m+"]",ref.moments_p_buy_v_sell[m],"ref.moments_p_buy_v_sell["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_p_sell_v_buy[m],"r.moments_p_sell_v_buy["+m+"]",ref.moments_p_sell_v_buy[m],"ref.moments_p_sell_v_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_p_sell_v_sell[m],"r.moments_p_sell_v_sell["+m+"]",ref.moments_p_sell_v_sell[m],"ref.moments_p_sell_v_sell["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pv[m],"r.moments_pv["+m+"]",ref.moments_pv[m],"ref.moments_pv["+m+"]",eps);
	    for(int k=0;k<np;k++)
	    {
		UnitTests.checkDoubleRel(r.moments_p_buy_x_p_sell[m][k],"r.moments_p_buy_x_p_sell["+m+"]["+k+"]",ref.moments_p_buy_x_p_sell[m][k],"ref.moments_p_buy_x_p_sell["+m+"]["+k+"]",eps);
		UnitTests.checkDoubleRel(r.moments_p_buy_x_p_buy[m][k],"r.moments_p_buy_x_p_buy["+m+"]["+k+"]",ref.moments_p_buy_x_p_buy[m][k],"ref.moments_p_buy_x_p_buy["+m+"]["+k+"]",eps);
		UnitTests.checkDoubleRel(r.moments_p_sell_x_p_sell[m][k],"r.moments_p_sell_x_p_sell["+m+"]["+k+"]",ref.moments_p_sell_x_p_sell[m][k],"ref.moments_p_sell_x_p_sell["+m+"]["+k+"]",eps);
		if(r.S_FACTOR_POWER==0 && m+k==1)
		{
		    // a kludge to avoid test failure
		    UnitTests.checkDouble(r.moments_pt[m][k],"r.moments_pt["+m+"]["+k+"]",ref.moments_pt[m][k],"ref.moments_pt["+m+"]["+k+"]",eps);
		}
		else
		{
		    UnitTests.checkDoubleRel(r.moments_pt[m][k],"r.moments_pt["+m+"]["+k+"]",ref.moments_pt[m][k],"ref.moments_pt["+m+"]["+k+"]",eps);
		}
	    }
	}
	System.err.println("TradeIPTBSData:checkMatching(): matching OK");
    }


    private static void matchWithTest(final String description,
				      final TradeIPTBSData r,
				      final Trade [] tb,
				      final int tsize,
				      final double eps)
    {
	System.err.println("TradeIPTBSData:matchWithTest():======================"+description);
	final int np=r.np;

	final TradeIPTBSData rt=calculateTestTradeIPTBSData(np,
							    r.currenttime,
							    r.tau,
							    tb,
							    tsize);
	checkMatching(r,rt,eps);

	final TradeIPData rtipd=TradeIPData.calculateTestTradeIPData(np,
								     r.currenttime,
								     r.tau,
								     tb,
								     tsize,
								     true);
	TradeIPData.checkMatchingTradeIPTBSData(r,rtipd,eps);
	System.err.println("TradeIPTBSData:matchWithTest(): OK");
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("TradeIPTBSData:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);

	final TradeIPTBSData ro=calculateTradeIPTBSData(N_EM,
							currenttime,
							tau,
							tb,
							tsize,
							null);
	matchWithTest("INITIAL CALCULATION",ro,tb,tsize,eps);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeIPTBSData em_ref=calculateTradeIPTBSData(N_EM,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest("GET REFERENCE FOR RECURRENT",em_ref,tb,n_split,eps);

	final TradeIPTBSData em_fromref=calculateTradeIPTBSData(N_EM,currenttime,tau,tb,tsize,em_ref);
	//System.err.println("em_ref.vol="+em_ref.vol+" ro.vol="+ro.vol);
	checkMatching(ro,em_fromref,eps);
	matchWithTest("RECURRENT CALCULATIONS",em_fromref,tb,tsize,eps);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeIPTBSData emc_cut=calculateTradeIPTBSData(N_EM,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut,eps);

	System.err.println("TradeIPTBSData:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{3,4,N_EMOMENTS,3,4,5})
	{
	    for(int i=0;i<100;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeIPTBSData::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }


    static DBSTradeIP<TradeIPTBSData> calculateDBSTradeIP(final int N_EM,
							  final long currenttime,
							  final long tau,
							  final Trade [] tb,
							  final int tsize,
							  final DBSTradeIP<TradeIPTBSData> ref)
    {
	final TradeIPTBSData o=TradeIPTBSData.calculateTradeIPTBSData(N_EM+1,
								      currenttime,
								      tau,
								      tb,
								      tsize,
								      ref!=null?ref.odata:null);
	return new DBSTradeIP<TradeIPTBSData>(o.moments_p_buy_x_p_buy,
					      o.moments_p_sell_x_p_sell,
					      o.moments_p_buy_x_p_sell,
					      o.moments_p_buy_v_buy,
					      o.moments_p_buy_v_sell,
					      o.moments_p_sell_v_buy,
					      o.moments_p_sell_v_sell,
					      o);
    }

} // END of class TradeIPTBSData






static class DTBSTradeIP {
    final PTradeIP diff,diff2,buy,sell;
    private final TradeIPTBSData odata;
    private DTBSTradeIP(final TradeIPTBSData odata)
    {
	this.odata=odata;

	this.diff=new PTradeIP(null,null,odata){
		double getIEq(){return 0;}
		double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np][np];
		    final double [] rpart=new double [np];
		    final double [][] bb=odata.moments_p_buy_x_p_buy,ss=odata.moments_p_sell_x_p_sell,bs=odata.moments_p_buy_x_p_sell;
		    for(int k=0;k<np;k++)
		    {
			for(int l=0;l<np;l++)
			{
			    matr[k][l]=bb[k][l]+ss[k][l]+bs[k][l]+bs[l][k];
			}
			rpart[k]=odata.moments_p_sell_v_sell[k]-odata.moments_p_sell_v_buy[k]
			         -odata.moments_p_buy_v_buy[k]+odata.moments_p_buy_v_sell[k];
		    }
		    TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
		    return rpart;
		}
	    };
	this.diff2=new PTradeIP(null,null,odata){
		double getIEq(){return 0;}
		double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np][np];
		    final double [] rpart_sell=new double [np];
		    final int [] ipvt=new int [np];
		    final double [][] ss=odata.moments_p_sell_x_p_sell,bb=odata.moments_p_buy_x_p_buy;
		    for(int k=0;k<np;k++)
		    {
			for(int l=0;l<np;l++)
			{
			    matr[k][l]=ss[k][l];
			}
			rpart_sell[k]=odata.moments_p_sell_v_sell[k];
		    }
		    TradeDPVR.solveLinsystem(matr,rpart_sell,np,ipvt);
		    final double [] rpart_buy=new double [np];
		    for(int k=0;k<np;k++)
		    {
			for(int l=0;l<np;l++)
			{
			    matr[k][l]=bb[k][l];
			}
			rpart_buy[k]=odata.moments_p_buy_v_buy[k];
		    }
		    TradeDPVR.solveLinsystem(matr,rpart_buy,np,ipvt);
		    for(int k=0;k<np;k++)
		    {
			rpart_sell[k]-=rpart_buy[k];
		    }
		    return rpart_sell;
		}
	    };
	final int np=odata.np;
	final double [] moments_buy=new double [2*np-1],moments_sell=new double [2*np-1];
	for(int k=0;k<np;k++)
	{
	    for(int l=0;l<np;l++)
	    {
		moments_buy[k+l]=odata.moments_p_buy_x_p_buy[k][l];
		moments_sell[k+l]=odata.moments_p_sell_x_p_sell[k][l];
	    }
	}
	this.buy=new PTradeIP(moments_buy,odata.moments_p_buy_v_buy,odata);
	this.sell=new PTradeIP(moments_sell,odata.moments_p_sell_v_sell,odata);
    }


    static DTBSTradeIP calculateDTBSTradeIP(final int N_EM,
					final long currenttime,
					final long tau,
					final Trade [] tb,
					final int tsize,
					final DTBSTradeIP ref)
    {
	return new DTBSTradeIP(TradeIPTBSData.calculateTradeIPTBSData(N_EM,
								      currenttime,
								      tau,
								      tb,
								      tsize,
								      ref!=null?ref.odata:null));
    }

} // END OF class DTBSTradeIP








static class TradeIPSData extends TradeIPRData {

    private static final int S_FACTOR_POWER=0;

    final double pt_sum,t_sum;
    private final double []
	moments_pvS_p_buy_v_buy,
	moments_pvS_p_buy_v_sell,
	moments_pvS_p_sell_v_buy,
	moments_pvS_p_sell_v_sell,
	moments_pv,

	moments_ptSS,
	moments_ptSS_buy,
	moments_ptSS_sell;

    private final double [][] moments_tSS_p_buy_x_p_sell;

    // references to previos calculations
    private final BookItem trade_last,trade_last_buy,trade_last_sell;


    TradeIPSData(final long currenttime,
		 final long tau,
		 final double pt_sum,
		 final double t_sum,
		 final double p_offset,

		 final double [] moments_pvS_p_buy_v_buy,
		 final double [] moments_pvS_p_buy_v_sell,
		 final double [] moments_pvS_p_sell_v_buy,
		 final double [] moments_pvS_p_sell_v_sell,
		 final double [] moments_pv,

		 final double [] moments_ptSS,
		 final double [] moments_ptSS_buy,
		 final double [] moments_ptSS_sell,

		 final double [][] moments_tSS_p_buy_x_p_sell,

		 final long vol,
		 final int np,

		 // references
		 final BookItem trade_last,
		 final BookItem trade_last_buy,
		 final BookItem trade_last_sell)
    {
	super(currenttime,tau,p_offset,vol,np);

	this.pt_sum=pt_sum;
	this.t_sum=t_sum;

	this.moments_pvS_p_buy_v_buy=moments_pvS_p_buy_v_buy;
	this.moments_pvS_p_buy_v_sell=moments_pvS_p_buy_v_sell;
	this.moments_pvS_p_sell_v_buy=moments_pvS_p_sell_v_buy;
	this.moments_pvS_p_sell_v_sell=moments_pvS_p_sell_v_sell;
	this.moments_pv=moments_pv;

	this.moments_ptSS=moments_ptSS;
	this.moments_ptSS_buy=moments_ptSS_buy;
	this.moments_ptSS_sell=moments_ptSS_sell;
	this.moments_tSS_p_buy_x_p_sell=moments_tSS_p_buy_x_p_sell;

	this.trade_last=trade_last;
	this.trade_last_buy=trade_last_buy;
	this.trade_last_sell=trade_last_sell;
    }

   static TradeIPSData calculateTestTradeIPSData(final int N_EM,
						 final long currenttime,
						 final long tau,
						 final BookItem [] tb,
						 final int tsize)
    {
	final int np=N_EM,nvs=1+2*np;
	final double atau=1.0/tau;

	// calculate 0th & 1st moments
	double pt_sum=0,t_sum=0;
	long prevt=currenttime,tinterval=currenttime,vol=0;
	for(int i=tsize;--i>=0;)
	{
	    final BookItem b=tb[i];
	    if(b.time>prevt)
	    {
		throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	    }
	    //System.err.println("i="+i+" b.time="+b.time);
	    prevt=b.time;
	    final double twmw=Math.exp(-(currenttime-prevt)*atau)*((tinterval-prevt)*atau);
	    pt_sum+=b.price*twmw;
	    t_sum+=twmw;
	    tinterval=prevt;
	    vol+=b.shares;
	}
	final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);

	return new TradeIPSData(currenttime,
				tau,

				pt_sum,
				t_sum,
				p_offset,

				checkSPV(BookItem.BOOK_BUY,BookItem.BOOK_BUY,nvs,currenttime,tau,tb,tsize,p_offset),
				checkSPV(BookItem.BOOK_BUY,BookItem.BOOK_SELL,nvs,currenttime,tau,tb,tsize,p_offset),
				checkSPV(BookItem.BOOK_SELL,BookItem.BOOK_BUY,nvs,currenttime,tau,tb,tsize,p_offset),
				checkSPV(BookItem.BOOK_SELL,BookItem.BOOK_SELL,nvs,currenttime,tau,tb,tsize,p_offset),
				checkSPV(BookItem.BOOK_UNKNOWN,BookItem.BOOK_UNKNOWN,nvs,currenttime,tau,tb,tsize,p_offset),

				checkS2PTtype(BookItem.BOOK_UNKNOWN,nvs,currenttime,tau,tb,tsize,p_offset),
				checkS2PTtype(BookItem.BOOK_BUY,nvs,currenttime,tau,tb,tsize,p_offset),
				checkS2PTtype(BookItem.BOOK_SELL,nvs,currenttime,tau,tb,tsize,p_offset),
				checkS2T_buy_x_sell(np,currenttime,tau,tb,tsize,p_offset),

				vol,
				np,
				null,
				null,
				null);
    }

    private static double getSFactor(final double S)
    {
	if(S_FACTOR_POWER==0)
	{
	    return 1;
	}
	if(S_FACTOR_POWER==1)
	{
	    return Math.abs(S);
	}
	if(S_FACTOR_POWER==2)
	{
	    return S*S;
	}
	throw new RuntimeException("S_FACTOR_POWER="+S_FACTOR_POWER);
    }

    private static double [] checkSPV(final byte type_p,
				      final byte type_v,
				      final int nvs,
				      final long currenttime,
				      final long tau,
				      final BookItem [] tb,
				      final int tsize,
				      final double p_offset)
    {
	final double atau=1.0/tau;
	final double [] matr=new double [nvs];
	BookItem trade_last_buy=null,trade_last_sell=null,trade_type_p=null;
	long prevt=-1;
	for(int i=0;i<tsize;i++)
	{
	    final BookItem b=tb[i];
	    prevt=b.time;
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		trade_last_buy=b;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		trade_last_sell=b;
	    }
	    else
	    {
		continue;
	    }

	    if(b.buy_sell_indicator==type_p || type_p==BookItem.BOOK_UNKNOWN)
	    {
		trade_type_p=b;
	    }
	    if((type_v==b.buy_sell_indicator || type_v==BookItem.BOOK_UNKNOWN) && trade_type_p!=null)
	    {
		final double ex=Math.exp(-(currenttime-b.time)*atau),bwmw=b.shares*ex,dp=trade_type_p.price-p_offset;
		double w=S_FACTOR_POWER==0?1:(trade_last_buy!=null && trade_last_sell!=null?getSFactor(trade_last_sell.price-trade_last_buy.price):0);
		for(int m=0;m<nvs;m++)
		{
		    matr[m]+=bwmw*w;
		    w*=dp;
		}
	    }
	}
	return matr;
    }


    private static double [] checkS2PTtype(final byte type,
					   final int nvs,
					   final long currenttime,
					   final long tau,
					   final BookItem [] tb,
					   final int tsize,
					   final double p_offset)
    {
	final double atau=1.0/tau;
	final double [] matr=new double [nvs];
	BookItem processed=null,typeB=null,typeS=null;
	long prevt=-1;
	double ex_type=0,dp_type=0;
	for(int i=0;i<=tsize;i++)
	{
	    final BookItem b;
	    if(i<tsize)
	    {
		b=tb[i];
		prevt=b.time;
	    }
	    else
	    {
		b=null;
		prevt=currenttime;
	    }
	    if(processed!=null && typeB!=null && typeS!=null)
	    {
		final double twmw=((prevt-processed.time)*atau)*ex_type,
		                S=typeS.price-typeB.price;
		double w=getSFactor(S)*getSFactor(S);
		for(int m=0;m<nvs;m++)
		{
		    matr[m]+=twmw*w;
		    w*=dp_type;
		}
	    }
	    if(b==null)
	    {
		continue;
	    }
	    final double ex=Math.exp(-(currenttime-prevt)*atau),dp=b.price-p_offset;
	    if(b.buy_sell_indicator==type||type==BookItem.BOOK_UNKNOWN)
	    {
		dp_type=dp;
	    }
	    ex_type=ex;
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		typeB=b;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		typeS=b;
	    }
	    processed=b;
	}
	return matr;
    }

    private static double [][] checkS2T_buy_x_sell(final int np,
						   final long currenttime,
						   final long tau,
						   final BookItem [] tb,
						   final int tsize,
						   final double p_offset)
    {
	final double atau=1.0/tau;
	final double [][] matr=new double [np][np];
	BookItem processed=null,typeB=null,typeS=null;
	long prevt=-1;
	double ex_type=0;
	for(int i=0;i<=tsize;i++)
	{
	    final BookItem b;
	    if(i<tsize)
	    {
		b=tb[i];
		prevt=b.time;
	    }
	    else
	    {
		b=null;
		prevt=currenttime;
	    }
	    if(processed!=null && typeB!=null && typeS!=null)
	    {
		final double twmw=((prevt-processed.time)*atau)*ex_type,
		    S=typeS.price-typeB.price,
		    dp_buy=typeB.price-p_offset,dp_sell=typeS.price-p_offset;
		double w=getSFactor(S)*getSFactor(S);
		for(int m=0;m<np;m++)
		{
		    double wa=w;
		    for(int l=0;l<np;l++)
		    {
			matr[m][l]+=twmw*wa;
			wa*=dp_sell;
		    }
		    w*=dp_buy;
		}
	    }
	    if(b==null)
	    {
		continue;
	    }
	    ex_type=Math.exp(-(currenttime-prevt)*atau);
	    if(b.buy_sell_indicator==BookItem.BOOK_BUY)
	    {
		typeB=b;
	    }
	    else if(b.buy_sell_indicator==BookItem.BOOK_SELL)
	    {
		typeS=b;
	    }
	    processed=b;
	}
	return matr;
    }


    private static void checkMatching(final TradeIPSData r,final TradeIPSData ref,final double eps)
    {
	if(r.currenttime!=ref.currenttime)
	{
	    throw new RuntimeException("r.currenttime="+r.currenttime+
				       " ref.currenttime="+ref.currenttime);
	}
	if(r.tau!=ref.tau)
	{
	    throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
	}
	if(r.np!=ref.np)
	{
	    throw new RuntimeException("r.np="+r.np+" ref.np="+ref.np);
	}
	if(r.vol!=ref.vol)
	{
	    throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
	}

	UnitTests.checkDoubleRel(r.pt_sum,"r.pt_sum",ref.pt_sum,"ref.pt_sum",eps);
	UnitTests.checkDoubleRel(r.t_sum,"r.t_sum",ref.t_sum,"ref.t_sum",eps);
	UnitTests.checkDoubleRel(r.p_offset,"r.p_offset",ref.p_offset,"ref.p_offset",eps);

	final int np=r.np,nvs=1+2*np;
	for(int m=0;m<nvs;m++)
	{
	    UnitTests.checkDoubleRel(r.moments_pvS_p_buy_v_buy[m],"r.moments_pvS_p_buy_v_buy["+m+"]",ref.moments_pvS_p_buy_v_buy[m],"ref.moments_pvS_p_buy_v_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pvS_p_buy_v_sell[m],"r.moments_pvS_p_buy_v_sell["+m+"]",ref.moments_pvS_p_buy_v_sell[m],"ref.moments_pvS_p_buy_v_sell["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pvS_p_sell_v_buy[m],"r.moments_pvS_p_sell_v_buy["+m+"]",ref.moments_pvS_p_sell_v_buy[m],"ref.moments_pvS_p_sell_v_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pvS_p_sell_v_sell[m],"r.moments_pvS_p_sell_v_sell["+m+"]",ref.moments_pvS_p_sell_v_sell[m],"ref.moments_pvS_p_sell_v_sell["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_pv[m],"r.moments_pv["+m+"]",ref.moments_pv[m],"ref.moments_pv["+m+"]",eps);

	    if(m==1 && r.S_FACTOR_POWER==0)
	    {
		// a kludge to avoid test failure
		UnitTests.checkDouble(r.moments_ptSS[m],"r.moments_ptSS["+m+"]",ref.moments_ptSS[m],"ref.moments_ptSS["+m+"]",eps);
	    }
	    else
	    {
		UnitTests.checkDoubleRel(r.moments_ptSS[m],"r.moments_ptSS["+m+"]",ref.moments_ptSS[m],"ref.moments_ptSS["+m+"]",eps);
	    }
	    UnitTests.checkDoubleRel(r.moments_ptSS_buy[m],"r.moments_ptSS_buy["+m+"]",ref.moments_ptSS_buy[m],"ref.moments_ptSS_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_ptSS_sell[m],"r.moments_ptSS_sell["+m+"]",ref.moments_ptSS_sell[m],"ref.moments_ptSS_sell["+m+"]",eps);
	}
	for(int m=0;m<np;m++)
	{
	    UnitTests.checkDoubleRel(r.moments_tSS_p_buy_x_p_sell[m][0],"r.moments_tSS_p_buy_x_p_sell["+m+"][0]",r.moments_ptSS_buy[m],"r.moments_ptSS_buy["+m+"]",eps);
	    UnitTests.checkDoubleRel(r.moments_tSS_p_buy_x_p_sell[0][m],"r.moments_tSS_p_buy_x_p_sell[0]["+m+"]",r.moments_ptSS_sell[m],"r.moments_ptSS_sell["+m+"]",eps);
	    for(int l=0;l<np;l++)
	    {
		UnitTests.checkDoubleRel(r.moments_tSS_p_buy_x_p_sell[m][l],"r.moments_tSS_p_buy_x_p_sell["+m+"]["+l+"]",ref.moments_tSS_p_buy_x_p_sell[m][l],"ref.moments_tSS_p_buy_x_p_sell["+m+"]["+l+"]",eps);
	    }
	}
	System.err.println("TradeIPSData:checkMatching(): matching OK");
    }

    private static void matchWithTest(final String description,
				      final TradeIPSData r,
				      final BookItem [] tb,
				      final int tsize,
				      final double eps)
    {
	System.err.println("TradeIPSData:matchWithTest():======================"+description);
	final int np=r.np;

	final TradeIPSData rt=calculateTestTradeIPSData(np,
							r.currenttime,
							r.tau,
							tb,
							tsize);
	checkMatching(r,rt,eps);

	final TradeIPBSData rtbspd=TradeIPBSData.calculateTestTradeIPBSData(np,
									    r.currenttime,
									    r.tau,
									    tb,
									    tsize);
	TradeIPBSData.checkMatchingTradeIPS(r,rtbspd,eps);

	final TradeIPData rtipd=TradeIPData.calculateTestTradeIPData(np,
								     r.currenttime,
								     r.tau,
								     tb,
								     tsize,
								     true);
	TradeIPData.checkMatchingTradeIPS(r,rtipd,eps);
    }

    private static void checkCalculations(final int tsize,final java.util.Random r,final int N_EM)
    {
	final double eps=1e-6;
	final Trade [] tb=TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");

	System.err.println("TradeIPSData:checkCalculations(): test with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);

	final TradeIPSData ro=calculateTradeIPSData(N_EM,
						    currenttime,
						    tau,
						    tb,
						    tsize,
						    null);
	matchWithTest("INITIAL CALCULATION",ro,tb,tsize,eps);

	// recurrent test
	final int n_split=8*(tb.length/10);
	final TradeIPSData em_ref=calculateTradeIPSData(N_EM,(tb[n_split-1].time+tb[n_split].time)/2,tau,tb,n_split,null);
	matchWithTest("GET REFERENCE FOR RECURRENT",em_ref,tb,n_split,eps);

	final TradeIPSData em_fromref=calculateTradeIPSData(N_EM,currenttime,tau,tb,tsize,em_ref);
	checkMatching(ro,em_fromref,eps);
	matchWithTest("RECURRENT CALCULATIONS",em_fromref,tb,tsize,eps);

	// check with data cut
	final int ncut=tb.length-n_split;
	System.arraycopy(tb,n_split,tb,0,ncut);
	final TradeIPSData emc_cut=calculateTradeIPSData(N_EM,currenttime,tau,tb,ncut,em_ref);
	checkMatching(ro,emc_cut,eps);

	System.err.println("TradeIPSData:checkCalculations(): OK with N points="+tb.length+
			   " currenttime="+currenttime+" tau="+tau+" N_EM="+N_EM);
    }

    public static void main(String [] args)
    {
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int N_EM :new int []{3,4,N_EMOMENTS,3,4,5})
	{
	    for(int i=0;i<37;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,N_EM);
		checkCalculations(10010+r.nextInt(100010),r,N_EM);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("TradeIPSData::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
    }




static TradeIPSData calculateTradeIPSData(final int N_EM,
					    final long currenttime,
					    final long tau,
					    final BookItem [] tb,
					    final int tsize,
					    final TradeIPSData ref)
{
    final int np=N_EM,nvs=1+2*np;
    final double atau=1.0/tau;
    final BookItem bref;
    if(ref!=null)
    {
	bref=ref.trade_last;
	if(ref.tau!=tau||ref.currenttime>currenttime||np!=ref.np)
	{
	    throw new RuntimeException("Invalid ref; ref.tau="+ref.tau+" tau="+tau+
				       " np="+np+" ref.np="+ref.np+
				       " ref.currenttime="+ref.currenttime+" currenttime="+currenttime);

	}
    }
    else
    {
	bref=null;
    }

    // calculate 0th & 1st moments
    double pt_sum=0,t_sum=0;
    long prevt=currenttime,tinterval=currenttime,vol=0;
    int ptr_start=tsize;
    for(int i=tsize;--i>=0;)
    {
	final BookItem b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	if(b==bref)
	{ // optimization
	    break;
	}
	prevt=b.time;
	vol+=b.shares;
	final double twmw=Math.exp(-(currenttime-prevt)*atau)*((tinterval-prevt)*atau);
	pt_sum+=b.price*twmw;
	t_sum+=twmw;
	tinterval=prevt;
	ptr_start=i;
    }
    BookItem trade_last,trade_last_buy,trade_last_sell;
    final double ref_exp_delay;
    double exp_last;
    if(bref!=null)
    { // optimization
	// add missed half slice
	exp_last=Math.exp(-(currenttime-bref.time)*atau);
	final double ref_dtr_exp=exp_last*((tinterval-ref.currenttime)*atau);
	pt_sum+=ref_dtr_exp*bref.price;
	t_sum+=ref_dtr_exp;

	// previous interval
	ref_exp_delay=Math.exp(-(currenttime-ref.currenttime)*atau);
	pt_sum+=ref_exp_delay*ref.pt_sum;
	t_sum+=ref_exp_delay*ref.t_sum;
	vol+=ref.vol;

	// init values for second pass
	trade_last=bref;
	trade_last_buy=ref.trade_last_buy;
	trade_last_sell=ref.trade_last_sell;
	tinterval=ref.currenttime;
    }
    else
    {
	ref_exp_delay=exp_last=0;
	trade_last=trade_last_buy=trade_last_sell=null;
	tinterval=currenttime;
    }
    final double p_offset=t_sum>0?pt_sum/t_sum:(tsize>0?tb[tsize-1].price:0);

    final double []
	moments_pvS_p_buy_v_buy=new double [nvs],
	moments_pvS_p_buy_v_sell=new double [nvs],
	moments_pvS_p_sell_v_buy=new double [nvs],
	moments_pvS_p_sell_v_sell=new double [nvs],
	moments_pv=new double [nvs],

	moments_ptSS=new double [nvs],
	moments_ptSS_buy=new double [nvs],
	moments_ptSS_sell=new double [nvs];
    final double [][] moments_tSS_p_buy_x_p_sell=new double [np][np];

    prevt=Long.MIN_VALUE;

    // calculate with given offset. using average as offset reduces rounding errors & degeneracy.
    for(int i=ptr_start;i<=tsize;i++)
    {
	final double S_twmw,dp_twmw_buy,dp_twmw_sell;
	if(trade_last_buy!=null && trade_last_sell!=null)
	{
	    S_twmw=trade_last_sell.price-trade_last_buy.price;
	    dp_twmw_buy=trade_last_buy.price-p_offset;
	    dp_twmw_sell=trade_last_sell.price-p_offset;
	}
	else
	{
	    exp_last=S_twmw=dp_twmw_buy=dp_twmw_sell=0;
	}
	final double bwmw,bwmw_pbuy_vbuy,bwmw_pbuy_vsell,bwmw_psell_vbuy,bwmw_psell_vsell,
	    dp_bwmw,dp_bwmw_buy,dp_bwmw_sell,S_bwmw,twmw,
	    dp_twmw=trade_last!=null?trade_last.price-p_offset:0;

	if(i>=tsize)
	{ // last slice up to currenttime
	    prevt=currenttime;
	    dp_bwmw=dp_bwmw_buy=dp_bwmw_sell=bwmw=bwmw_pbuy_vbuy=bwmw_pbuy_vsell=bwmw_psell_vbuy=bwmw_psell_vsell=S_bwmw=0;
	    twmw=exp_last*((prevt-tinterval)*atau);
	}
	else
	{
	    final BookItem b=tb[i];
	    prevt=b.time;
	    final int tp=b.buy_sell_indicator;
	    final double ex=Math.exp(-(currenttime-prevt)*atau),dp=b.price-p_offset;
	    twmw=exp_last*((prevt-tinterval)*atau);
	    bwmw=ex*b.shares;
	    dp_bwmw=dp;

	    if(tp==BookItem.BOOK_BUY)
	    {
		trade_last_buy=b;
		bwmw_pbuy_vsell=bwmw_psell_vsell=0;
		dp_bwmw_buy=dp;
		if(trade_last_sell!=null)
		{
		    dp_bwmw_sell=trade_last_sell.price-p_offset;
		    bwmw_pbuy_vbuy=bwmw_psell_vbuy=bwmw;
		    S_bwmw=trade_last_sell.price-trade_last_buy.price;
		}
		else
		{
		    dp_bwmw_sell=bwmw_psell_vbuy=S_bwmw=0;
		    bwmw_pbuy_vbuy=S_FACTOR_POWER==0?bwmw:0;
		}
	    }
	    else if(tp==BookItem.BOOK_SELL)
	    {
		trade_last_sell=b;
		bwmw_pbuy_vbuy=bwmw_psell_vbuy=0;
		dp_bwmw_sell=dp;
		if(trade_last_buy!=null)
		{
		    dp_bwmw_buy=trade_last_buy.price-p_offset;
		    bwmw_pbuy_vsell=bwmw_psell_vsell=bwmw;
		    S_bwmw=trade_last_sell.price-trade_last_buy.price;
		}
		else
		{
		    dp_bwmw_buy=bwmw_pbuy_vsell=S_bwmw=0;
		    bwmw_psell_vsell=S_FACTOR_POWER==0?bwmw:0;
		}
	    }
	    else
	    {
		dp_bwmw_buy=dp_bwmw_sell=bwmw_pbuy_vbuy=bwmw_pbuy_vsell=bwmw_psell_vbuy=bwmw_psell_vsell=S_bwmw=0;
	    }
	    trade_last=b;
	    exp_last=ex;
	    tinterval=prevt;
	}
	double w_bwdp=getSFactor(S_bwmw),w_bwdp_buy=w_bwdp,w_bwdp_sell=w_bwdp,w_twdp=getSFactor(S_twmw)*getSFactor(S_twmw),w_twdp_buy=w_twdp,w_twdp_sell=w_twdp;
	for(int m=0;m<nvs;m++)
	{
	    moments_pvS_p_buy_v_buy[m]+=w_bwdp_buy*bwmw_pbuy_vbuy;
	    moments_pvS_p_buy_v_sell[m]+=w_bwdp_buy*bwmw_pbuy_vsell;
	    moments_pvS_p_sell_v_buy[m]+=w_bwdp_sell*bwmw_psell_vbuy;
	    moments_pvS_p_sell_v_sell[m]+=w_bwdp_sell*bwmw_psell_vsell;
	    moments_pv[m]+=w_bwdp*bwmw;

	    moments_ptSS[m]+=w_twdp*twmw;
	    moments_ptSS_buy[m]+=w_twdp_buy*twmw;
	    moments_ptSS_sell[m]+=w_twdp_sell*twmw;
	    if(m<np)
	    {
		final double [] mm=moments_tSS_p_buy_x_p_sell[m];
		double w_buy_x_sell=w_twdp_buy;
		for(int l=0;l<np;l++)
		{
		    mm[l]+=w_buy_x_sell*twmw;
		    w_buy_x_sell*=dp_twmw_sell;
		}
	    }

	    w_bwdp_buy*=dp_bwmw_buy;
	    w_bwdp_sell*=dp_bwmw_sell;
	    w_bwdp*=dp_bwmw;

	    w_twdp*=dp_twmw;
	    w_twdp_buy*=dp_twmw_buy;
	    w_twdp_sell*=dp_twmw_sell;
	}
    }

    if(bref!=null)
    { // optimization

	// convert old interval to new p_offest basis
	final double dpa=ref.p_offset-p_offset;

	for(int l=0;l<nvs;l++)
	{
	    // P(nvs-1)D1 polynomial group transform
	    double wv=1,s_pvS_p_buy_v_buy=0,s_pvS_p_buy_v_sell=0,s_pvS_p_sell_v_buy=0,s_pvS_p_sell_v_sell=0,s_pv=0,s_ptSS=0,s_ptSS_buy=0,s_ptSS_sell=0;
	    long fv=1;
	    for(int ls=l;ls>=0;ls--)
	    {
		final double a=fv*wv;
		s_pvS_p_buy_v_buy+=a*ref.moments_pvS_p_buy_v_buy[ls];
		s_pvS_p_buy_v_sell+=a*ref.moments_pvS_p_buy_v_sell[ls];
		s_pvS_p_sell_v_buy+=a*ref.moments_pvS_p_sell_v_buy[ls];
		s_pvS_p_sell_v_sell+=a*ref.moments_pvS_p_sell_v_sell[ls];
		s_pv+=a*ref.moments_pv[ls];

		s_ptSS+=a*ref.moments_ptSS[ls];
		s_ptSS_buy+=a*ref.moments_ptSS_buy[ls];
		s_ptSS_sell+=a*ref.moments_ptSS_sell[ls];
		wv*=dpa;
		fv*=ls;
		fv/=1+l-ls;
	    }
	    moments_pvS_p_buy_v_buy[l]+=ref_exp_delay*s_pvS_p_buy_v_buy;
	    moments_pvS_p_buy_v_sell[l]+=ref_exp_delay*s_pvS_p_buy_v_sell;
	    moments_pvS_p_sell_v_buy[l]+=ref_exp_delay*s_pvS_p_sell_v_buy;
	    moments_pvS_p_sell_v_sell[l]+=ref_exp_delay*s_pvS_p_sell_v_sell;
	    moments_pv[l]+=ref_exp_delay*s_pv;

	    moments_ptSS[l]+=ref_exp_delay*s_ptSS;
	    moments_ptSS_buy[l]+=ref_exp_delay*s_ptSS_buy;
	    moments_ptSS_sell[l]+=ref_exp_delay*s_ptSS_sell;
	}
	for(int l=0;l<np;l++)
	{
	    for(int m=0;m<np;m++)
	    {
		double wl=1,s_p_buy_x_p_sell=0;
		// P(np-1)D2 polynomial group transform
		long fl=1;
		for(int ls=l;ls>=0;ls--)
		{
		    double wm=fl*wl;
		    long fm=1;
		    for(int ms=m;ms>=0;ms--)
		    {
			final double a=fm*wm;
			s_p_buy_x_p_sell+=a*ref.moments_tSS_p_buy_x_p_sell[ls][ms];
			wm*=dpa;
			fm*=ms;
			fm/=1+m-ms;
		    }
		    wl*=dpa;
		    fl*=ls;
		    fl/=1+l-ls;
		}
		moments_tSS_p_buy_x_p_sell[l][m]+=ref_exp_delay*s_p_buy_x_p_sell;
	    }
	}
    }

    return new TradeIPSData(currenttime,
			    tau,

			    pt_sum,
			    t_sum,
			    p_offset,

			    moments_pvS_p_buy_v_buy,
			    moments_pvS_p_buy_v_sell,
			    moments_pvS_p_sell_v_buy,
			    moments_pvS_p_sell_v_sell,
			    moments_pv,

			    moments_ptSS,
			    moments_ptSS_buy,
			    moments_ptSS_sell,

			    moments_tSS_p_buy_x_p_sell,

			    vol,
			    np,
			    trade_last,
			    trade_last_buy,
			    trade_last_sell);

}


} // END OF class TradeIPSData



static class DTradeIPS {
    final PTradeIP diff,diff2,buy,sell;
    private final TradeIPSData odata;
    private DTradeIPS(final TradeIPSData odata)
    {
	this.odata=odata;
	this.buy=new PTradeIP(odata.moments_ptSS_buy,odata.moments_pvS_p_buy_v_buy,odata);
	this.sell=new PTradeIP(odata.moments_ptSS_sell,odata.moments_pvS_p_sell_v_sell,odata);
	this.diff=new PTradeIP(null,null,odata){
		double getIEq(){return 0;}
		double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double
			nbuy=odata.moments_pvS_p_buy_v_buy[0]/odata.moments_ptSS_buy[0],
			nsell=odata.moments_pvS_p_sell_v_sell[0]/odata.moments_ptSS_sell[0],
			s=nbuy+nsell;
		    final double [] pcoefs=new double [np];
		    for(int k=0;k<np;k++)
		    {
			pcoefs[k]=(nsell*(sell.pcoefs[k]/sell.normformoments[k])-nbuy*(buy.pcoefs[k]/buy.normformoments[k]))/s;
		    }
		    return pcoefs;
		}
	    };
	this.diff2=new PTradeIP(null,null,odata){
		double getIEq(){return 0;}
		double [] getNorms(final double [] moments_left,final int np){return PTradeIP.getArray1Filled(np);}
		double [] getPCoefs(final double [] moments_left,final double [] norms,final double [] moments_right,final int np)
		{
		    final double [][] matr=new double [np][np];
		    final double [] rpart=new double [np];
		    for(int l=0;l<np;l++)
		    {
			for(int k=0;k<np;k++)
			{
			    matr[l][k]=odata.moments_ptSS[l+k];
			}
			rpart[l]=odata.moments_pvS_p_sell_v_sell[l]-odata.moments_pvS_p_buy_v_buy[l];
		    }
		    TradeDPVR.solveLinsystem(matr,rpart,np,new int [np]);
		    return rpart;
		}
	    };
    }

    static DTradeIPS calculateDTradeIPS(final int N_EM,
					final long currenttime,
					final long tau,
					final BookItem [] tb,
					final int tsize,
					final DTradeIPS ref)
    {
	return new DTradeIPS(TradeIPSData.calculateTradeIPSData(N_EM,
								currenttime,
								tau,
								tb,
								tsize,
								ref!=null?ref.odata:null));
    }
} // END OF class DTradeIPS



public static void main(String [] args)
{
 System.err.println("CalculateKendall::main(): starting ALL tests; ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
 final long t0=System.currentTimeMillis();

 System.err.println("Version");
 System.err.println("Started="+new java.util.Date(t0));
 System.getProperties().list(System.err);
 System.err.println("END Version");

 // old version moved here
 EMoments.main(args);
 FMoments.main(args);
 ExponentialInterpolation.main(args);
 Vandermonde2Interpolation.main(args);
 VandermondeInterpolation.main(args);


 CalculateKendall.TradeEMoments.main(args);
 CalculateKendall.TradePTMoments.main(args);
 CalculateKendall.IntgrTIFun.main(args);
 CalculateKendall.TestPolynomialT.main(args);
 CalculateKendall.PrintRoots.main(args);
 CalculateKendall.TradeVX.main(args);
 CalculateKendall.TradePV.main(args);
 CalculateKendall.TradeDPV.main(args);
 CalculateKendall.TradeDPV.TestPlowPhigh.main(args);
 CalculateKendall.TradeDPVR.main(args);
 CalculateKendall.TradeDPVR.TestPlowPhighR.main(args);
 CalculateKendall.TestPlowPhigh2.main(args);
 CalculateKendall.TradeDPVRN.main(args);
 CalculateKendall.TradeIP.main(args);
 CalculateKendall.QQTradeIP.main(args);
 CalculateKendall.QATradeIP.main(args);
 CalculateKendall.QVTradeIP.main(args);
 CalculateKendall.TradeIPMData.main(args);
 CalculateKendall.QMTradeIP.main(args);
 CalculateKendall.TradeIPVMData.main(args);
 CalculateKendall.QVMTradeIP.main(args);
 CalculateKendall.TCode.main(args);
 CalculateKendall.OTradeIP.main(args);
 CalculateKendall.DBTradeIP.main(args);
 CalculateKendall.TradeIPBSData.main(args);
 CalculateKendall.TradeIPTBSData.main(args);
 CalculateKendall.TradeIPSData.main(args);
 CalculateDCoefs.main(args);
 CalculateBookCriticalExponent.main(args);
 CalculateLiquidityDeficit3.main(args);

 final long t1=System.currentTimeMillis();
 System.err.println("Version");
 System.err.println("Finished="+new java.util.Date(t1)+" Started="+new java.util.Date(t0));
 System.getProperties().list(System.err);
 System.err.println("END Version");
 System.err.println("CalculateKendall::main(): ALL tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: CalculateKendall.java,v 1.2550 2022/07/28 22:30:48 mal Exp $");
}


}
