package com.polytechnik.trading;
/*
 * (C) Vladislav Malyshkin 2011
 * This file is under GPL version 3.
 *
 */

import com.polytechnik.realtimebook.*;
import com.polytechnik.utils.UnitTests;
import com.polytechnik.utils.Polynomials;
import com.polytechnik.utils.Laguerre;
import com.polytechnik.utils.LegendreShifted;
import com.polytechnik.utils.TMatr;
import com.polytechnik.utils.SortArrays;
import com.polytechnik.utils.PolynomialRootsFinder;

/** Quantitative Volume Moments(QVM) model.
 *  @version $Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $
 *  @author Vladislav Malyshkin mal@gromco.com
 */
public class QVM {


/** The data for QVM coefficients & projection. */
abstract static class QVMCoefs {

    private static final boolean LOOK_FORWARD=true;

    private final double [] pcoefs;
    final double sigma,pa_offset;
    final double [] p_imin,p_imax;

    final int neq_v,neq_t,neq_p;

    private final double f_dv,f_dt,i0;
    private final double [] p_real_roots;
    /*private*/ final QVMData odata;


private QVMCoefs(final com.polytechnik.utils.PolynomialRootsFinder rf,
		 final int neq_v,final int neq_t,final int neq_p,
		 final QVMData odata)
    {
	this.odata=odata;
	this.neq_v=neq_v;
	this.neq_t=neq_t;
	this.neq_p=neq_p;

	final double [] abc=odata.tmoments_vol_x_tau_x_p;
	final double pt0=abc[0];

	this.pa_offset=this.odata.p_offset;
	this.f_dv=!LOOK_FORWARD?0:-1*(odata.nvm<2?0:abc[1*odata.ntm*odata.npm]/pt0);
	this.f_dt=!LOOK_FORWARD?0:-1*(odata.ntm<2?0:abc[1*odata.npm]/pt0);
	this.sigma=odata.npm<3?Double.NaN:Math.sqrt(abc[2]/pt0-Math.pow(abc[1]/pt0,2));
	this.pcoefs=getPCoefs(this.f_dv,this.f_dt,odata.trade_last==null?Double.NaN:odata.trade_last.price-odata.p_offset);

	double si0=0;
	for(int m=this.pcoefs.length;--m>=0;){si0+=this.pcoefs[m]*(abc[m]/pt0);}
	this.i0=si0;

	final double [][] xx=CalculateKendall.QATradeIP.getPolynomialMinMax(this.pcoefs.length-1,this.pcoefs,this.odata.p_offset,rf);
	this.p_imin=xx[0];
	this.p_imax=xx[1];
	this.p_real_roots=xx[2];
    }

    abstract double [] getPCoefs(final double dv,final double dt,final double dp);

    double getDIDPDT_TOT(final double p)
    {
	return Polynomials.getD1(this.pcoefs,p-this.odata.p_offset);
    }
    double getDIDT(final double p)
    {
	return Polynomials.getD0(this.pcoefs,p-this.odata.p_offset);
    }

    double getPGMIN(final double p)
    {
	return CalculateKendall.QATradeIP.getPGMIN(p,this.p_imin,this.pcoefs,this.odata.p_offset);
    }

    double getPLMIN(final double p)
    {
	return CalculateKendall.QATradeIP.getPLMIN(p,this.p_real_roots,this.odata.p_offset);
    }

    double getIF(final double p)
    {
	return (Polynomials.getD0(this.pcoefs,p-this.odata.p_offset)-this.i0)/this.i0;
    }

    double getIDP(final double p)
    {
	final double dp=p-this.odata.p_offset;
	return Polynomials.getD0(this.pcoefs,dp)/Polynomials.getD1(this.pcoefs,dp);
    }


private static void checkQVMTradeIP(final QVMCoefs ro,final Trade [] tb, final int tsize,final double eps,
				    final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{
    final CalculateKendall.QVMTradeIP ref=
	CalculateKendall.QVMTradeIP.calculateQVMTradeIP(ro.neq_v,
							ro.neq_t,
							ro.neq_p,
							ro.odata.currenttime,
							ro.odata.tau,
							tb,
							tsize,
							rf,
							null);
    checkMatchingQVMTradeIP(ro,ref,eps);
    System.err.println("checkQVMTradeIP: tests OK with tsize="+tsize+
		       " neq_v="+ro.neq_v+" neq_t="+ro.neq_t+" neq_p="+ro.neq_p);
}

private static void checkMatchingQVMTradeIP(final QVMCoefs ro,CalculateKendall.QVMTradeIP ref,final double eps)
{
    UnitTests.checkDoubleRel(ro.sigma,"ro.sigma",ref.sigma,"ref.sigma",eps);
    UnitTests.checkDoubleRel(ro.odata.p_offset,"ro.odata.p_offset",ref.p_offset,"ref.p_offset",eps);
    UnitTests.checkDoubleRel(ro.f_dv,"ro.f_dv",ref.f_dv,"ref.f_dv",eps);
    UnitTests.checkDoubleRel(ro.f_dt,"ro.f_dt",ref.f_dt,"ref.f_dt",eps);
    if(false)
    {
	// matches only for old getPCoefs() versions
	UnitTests.checkDoubleArrayRel(ro.pcoefs,"ro.pcoefs",ref.pcoefs,"ref.pcoefs",eps);
	UnitTests.checkDoubleRel(ro.i0,"ro.i0",ref.i0,"ref.i0",eps);
	//UnitTests.checkDoubleArrayRel(ro.p_imin,"ro.p_imin",ref.p_imin,"ref.p_imin",eps);
	//UnitTests.checkDoubleArrayRel(ro.p_imax,"ro.p_imax",ref.p_imax,"ref.p_imax",eps);
    }
    else
    {
	System.err.println("checkMatchingQVMTradeIP: pcoefs match is skipped");
    }

    // match calculated moments
    final double [][][] abc=QVMData.get3ArrayFromArrayForTesting(ro.odata.tmoments_vol_x_tau_x_p,ro.odata.nvm,ro.odata.ntm,ro.odata.npm);
    for(int k=0;k<ro.odata.nvm;k++)
    for(int l=0;l<ro.odata.ntm;l++)
    {
	 UnitTests.checkDoubleArrayRel(abc[k][l],"ro.odata.tmoments_vol_x_tau_x_p[k][l][*]",
					      ref.odata.moments_vol_x_tau_x_p[k][l],"ref.odata.moments_vol_x_tau_x_p[k][l]",eps);
    }
    UnitTests.checkDoubleArrayRel(ro.odata.tmoments_vol_x_tau_x_p,"ro.odata.tmoments_vol_x_tau_x_p",
					 QVMData.getArrayFrom3ArrayForTesting(ref.odata.moments_vol_x_tau_x_p,ref.odata.nvm,ref.odata.ntm,ref.odata.np),"A1(ref.odata.moments_vol_x_tau_x_p[][][])",eps);

    System.err.println("checkMatchingQVMTradeIP: tests OK");
}


} // END of class QVMCoefs



/** QVMAverage version of QVMCoefs. */
static class QVMAverage extends QVMCoefs {

final int neq_nf;

private QVMAverage(final com.polytechnik.utils.PolynomialRootsFinder rf,
		   final int neq_v,final int neq_t,final int neq_p,
		   final QVMData odata)
{
	super(rf,neq_v,neq_t,neq_p,odata);
	this.neq_nf=0; // neq_nf != 0 is conceptually wrong
}

    double [] getPCoefs(final double dv,final double dt,final double dp)
    {
	return getAPCoefs(dv,dt);
    }

    private double [] getPCoefs(final double f_dv,final double f_dt)
    {
	final QVMData o=super.odata;
	final double [] d0=new double []{0,0,o.trade_last==null?Double.NaN:o.trade_last.price-o.p_offset};
	return getPLCondititionCoefsProjected(
	    o.tmoments_vol_x_tau_x_p,
	    new int []{o.nvm,o.ntm,o.npm},
	    d0,d0,
	    new int []{this.neq_v,this.neq_t,this.neq_p},
	    2,
	    new double []{f_dv,f_dt,Double.NaN},
	    0);
    }


    private double [] getAPCoefs(final double f_dv,final double f_dt)
    {
	final QVMData o=super.odata;
	return getAverageProjected(o.tmoments_vol_x_tau_x_p,
				   new int []{o.nvm,o.ntm,o.npm},
				   new int []{this.neq_v,this.neq_t,this.neq_p},
				   2,
				   new double []{f_dv,f_dt,Double.NaN},
				   this.neq_nf);
    }


    private double [] getPCoefsVT(final int neq_d)
    {
	final QVMData o=super.odata;
	return getAverageProjected(o.projecToVT(1+2*neq_d),
				   new int []{1+2*neq_d,o.npm},
				   new int []{neq_d,this.neq_p},
				   1,
				   new double []{0,Double.NaN},
				   0);
    }

private static double [] getPLCondititionCoefsProjected(final double [] mom,
							final int [] mom_dims,
							final double [] d0m,
							final double [] d0g,
							final int [] neq,
							final int proj_ptr,
							final double [] dproj,
							final int nf)
{
    final int nd=neq.length;
    if(proj_ptr<0||proj_ptr>=nd||mom_dims.length!=nd)
    {
	throw new IllegalArgumentException("proj_ptr="+proj_ptr+
					   " or mom_dims.length="+mom_dims.length+
					   " does not correspond to neq.length="+neq.length);
    }

    int mom_nels=1,neqs=0;
    for(int k=nd;--k>=0;)
    {
	final int kdim=mom_dims[k];
	if(kdim<1)
	{
	    throw new IllegalArgumentException("mom_dims["+k+"]="+mom_dims[k]);
	}
	mom_nels*=kdim;
	if(mom_nels<=0)
	{
	    throw new IllegalArgumentException("mom_nels("+k+")="+mom_nels);
	}
	final int kneq=neq[k];
	if(kneq<0 || (1+2*kneq)>kdim)
	{
	    throw new IllegalArgumentException("neq["+k+"]="+neq[k]+
					       " does not correspond to"+
					       " mom_dims["+k+"]="+mom_dims[k]);
	}
	neqs+=kneq;
    }
    if(mom.length!=mom_nels)
    {
	throw new IllegalArgumentException("mom.length="+mom.length+
					   " does not correspond the size calculated from mom_dims;"+
					   " size="+mom_nels);
    }
    final int ne=neqs,momoffs=mom_nels,neq_proj=neq[proj_ptr];
    if(nf<0 || 1+2*(neq_proj+nf)>mom_dims[proj_ptr])
    {
	throw new IllegalArgumentException();
    }

    final double pt0=mom[0];
    final double [] norms=new double [ne];
    double sf=Double.NaN;

    for(int ek=0,offsk=momoffs,ik=0;ek<nd;ek++)
    {
	offsk/=mom_dims[ek];
	final int nk=neq[ek];
	final double stdevk=nk<=0?Double.NaN:Math.sqrt(mom[2*offsk]/pt0);
	double w=1;
	for(int k=1;k<=nk;k++,ik++)
	{
	    norms[ik]=(w*=(stdevk*k));
	    //System.err.println("QVMAverage: norms["+ik+"]="+norms[ik]);
	}
	if(ek==proj_ptr)
	{
	    final int nf2=(nf/2)*2;
	    sf=nf==nf2?(mom[nf*offsk]/pt0):(mom[nf2*offsk]/pt0)*stdevk;
	}
    }

    final boolean aprojm=d0m==null;
    if(!aprojm && nf>0)
    {
	throw new IllegalArgumentException();
    }
    final double [] matr=new double [ne*ne];
    for(int ek=0,offsk=momoffs,ik=0,j=0;ek<nd;ek++)
    {
	offsk/=mom_dims[ek];
	final int onfk;final double sfk;
	if(ek==proj_ptr){onfk=nf*offsk;sfk=sf;}
	else{onfk=0;sfk=1;}
	double wk=1;
	for(int k=1,nk=neq[ek],koffsk=offsk+onfk;k<=nk;k++,koffsk+=offsk,ik++)
	{
	    final double normsk=norms[ik],pk=mom[koffsk]/pt0;
	    wk=aprojm?pk:(wk*d0m[ek]);
	    for(int el=0,offsl=momoffs,il=0;el<nd;el++)
	    {
		offsl/=mom_dims[el];
		final int onfl;final double sfl;
		if(el==proj_ptr){onfl=nf*offsl;sfl=sf;}
		else{onfl=0;sfl=1;}
		double wl=1;
		for(int l=1,nl=neq[el],loffsl=offsl+onfl;l<=nl;l++,loffsl+=offsl,il++,j++)
		{
		    final double pl=mom[loffsl]/pt0,pkl=mom[koffsk+loffsl]/pt0;
		    wl=aprojm?pl:(wl*d0m[el]);
		    matr[j]=((pkl-wl*pk+wk*(wl-pl))/sfl/sfk)/normsk/norms[il];
		}
	    }
	}
    }

    final double [] pproj=new double [1+2*neq_proj];
    final int [] ipvt=new int [ne];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,ne,ne,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(pproj,Double.NaN);
	return pproj;
    }

    final boolean aprojg=d0g==null;
    if(!aprojm && nf>0)
    {
	throw new IllegalArgumentException();
    }
    final double [] rpart=new double [ne];
    for(int ek=0,offsk=momoffs,ik=0;ek<nd;ek++)
    {
	final double drwk=dproj[ek];
	double wk=1,rwk=1;
	for(int k=1,nk=neq[ek],koffsk=(offsk/=mom_dims[ek])+(ek==proj_ptr?nf*offsk:0);k<=nk;k++,koffsk+=offsk,ik++)
	{
	    wk=aprojg?(mom[koffsk]/pt0):(wk*d0g[ek]);
	    rwk*=drwk;
	    rpart[ik]=1;
	    com.polytechnik.utils.Linsystems.dgesl_m1(matr,ne,ne,ipvt,rpart,1);
	    final double normsk=norms[ik];
	    for(int el=0,offsl=momoffs,il=0;el<nd;el++)
	    {
		final double drwl=dproj[el];
		double wl=1,rwl=1;
		for(int l=1,nl=neq[el],loffsl=(offsl/=mom_dims[el])+(el==proj_ptr?nf*offsl:0);l<=nl;l++,loffsl+=offsl,il++)
		{
		    wl=aprojg?(mom[loffsl]/pt0):(wl*d0g[el]);
		    rwl*=drwl;
		    final double c=rpart[il]/normsk/norms[il];
		    rpart[il]=0;

		    // project
		    if(el==proj_ptr)
		    {
			if(ek==proj_ptr)
			{
			    final double cwl=c*(wl/sf),wkf=wk/sf;
			    pproj[0]+=cwl*wkf;
			    pproj[k]-=cwl;
			    pproj[l]-=c*wkf;
			    pproj[k+l]+=c;
			}
			else
			{
			    final double cdw=c*(rwk-wk);
			    pproj[0]-=cdw*(wl/sf);
			    pproj[l]+=cdw;
			}
		    }
		    else
		    {
			if(ek==proj_ptr)
			{
			    final double cdw=c*(rwl-wl);
			    pproj[0]-=cdw*(wk/sf);
			    pproj[k]+=cdw;
			}
			else
			{
			    pproj[0]+=c*(rwl-wl)*(rwk-wk);
			}
		    }
		}
	    }
	}
    }
    return pproj;
}



private static double [] getAverageProjected(final double [] mom,
					     final int [] mom_dims,
					     final int [] neq,
					     final int proj_ptr,
					     final double [] dproj,
					     final int nf)
{
    return getPLCondititionCoefsProjected(mom,mom_dims,null,null,neq,proj_ptr,dproj,nf);
}



    public static QVMAverage calculateQVMAverage(final int neq_v,
						 final int neq_t,
						 final int neq_p,
						 final long currenttime,
						 final long tau,
						 final Trade [] tb,
						 final int tsize,
						 final com.polytechnik.utils.PolynomialRootsFinder rf,
						 final QVMAverage ref)
    {
	final QVMData o=QVMData.calculateQVMData(2*neq_v+1,
						 2*neq_t+1,
						 2*neq_p+1,
						 currenttime,
						 tau,
						 tb,
						 tsize,
						 ref!=null?ref.odata:null);
	return new QVMAverage(rf,neq_v,neq_t,neq_p,o);
    }


    private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_v,final int neq_t,final int neq_p)
    {
	final double eps=1e-5;
	final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("QVMAverage:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);

	final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();

	final QVMAverage ro=QVMAverage.calculateQVMAverage(neq_v,
							   neq_t,
							   neq_p,
							   currenttime,
							   tau,
							   tb,
							   tsize,
							   rf,
							   null);
	// check QVMTradeIP match
	QVMCoefs.checkQVMTradeIP(ro,tb,tsize,eps,rf);

	// regular check
	checkMatchingPLCondititionProjections(r,ro,eps,tb,tsize);

	// average projections
	checkMatchingAverageProjections(neq_v,neq_t,neq_p,r,ro.odata,eps);

	// NF check
	checkMatchingNF(neq_v,neq_t,neq_p,r,ro.odata,eps);


	System.err.println("QVMAverage:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
    }


private static void checkMatchingPLCondititionProjections(final java.util.Random r,final QVMCoefs ro,final double eps,final Trade [] tb,final int tsize)
    {
	final double [] abc=ro.odata.tmoments_vol_x_tau_x_p;
	final double [][][] abc3=QVMData.get3ArrayFromArrayForTesting(abc,ro.odata.nvm,ro.odata.ntm,ro.odata.npm);

	final double pt0=abc[0],stdevv=1.1,stdevt=1.03;
	final int [] mom_dims=new int []{ro.odata.nvm,ro.odata.ntm,ro.odata.npm},
	             neq=new int []{ro.neq_v,ro.neq_t,ro.neq_p};
	final double [] d0=new double [3],dproj=new double [3];

	for(int iv=0;iv<9;iv++)
	{
	    double dv=r.nextDouble()*stdevv;
	    for(int it=0;it<10;it++)
	    {
		double dt=r.nextDouble()*stdevt;
		for(int ip=0;ip<11;ip++)
		{
		    double dp=CalculateKendall.TradeEMoments.getRandomPrice(r)-ro.odata.p_offset;
		    if(iv==0 && it==0 && ip==0)
		    {
			dv=0;
			dt=0;
			dp=ro.odata.trade_last.price-ro.odata.p_offset;
		    }
		    d0[0]=dv;
		    d0[1]=dt;
		    d0[2]=dp;

		    UnitTests.checkDoubleRel(
			Polynomials.getD0(getPLCondititionCoefsProjected(abc,mom_dims,d0,d0,neq,2,d0,0),dp),"I0(dp)",
			0,"0",eps);

		    // check projection
		    for(int iprv=0;iprv<7;iprv++)
		    {
			double dprv=r.nextDouble()*stdevv;
			for(int iprt=0;iprt<7;iprt++)
			{
			    double dprt=r.nextDouble()*stdevt;
			    if(iprv==0 && iprt==0)
			    {
				dprv=ro.f_dv;
				dprt=ro.f_dt;
			    }
			    dproj[0]=dprv;
			    dproj[1]=dprt;
			    dproj[2]=Double.NaN;

			    final double [] pcoefs=
				getPLCondititionCoefsProjected(abc,
							       mom_dims,
							       d0,
							       d0,
							       neq,
							       2,
							       dproj,
							       0);
			    // match to QVMTradeIP.
			    final double [] proefsQVMTradeIP=
				CalculateKendall.QVMTradeIP.getPProjected(
				    CalculateKendall.QVMTradeIP.getPLCondititionCoefs(
					abc3,
					dv,
					dt,
					dp,
					ro.neq_v,
					ro.neq_t,
					ro.neq_p),
				    dprv,dprt);
			    UnitTests.checkDoubleArrayRel1(pcoefs,"pcoefs",proefsQVMTradeIP,"proefsQVMTradeIP",eps);
			}
		    }
		}
	    }
	}
	System.err.println("checkMatchingPLCondititionProjections: tests OK");
    }

private static void checkMatchingAverageProjections(final int neq_v,
						    final int neq_t,
						    final int neq_p,
						    final java.util.Random r,
						    final QVMData odata,
						    final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final double [][][] abc3=QVMData.get3ArrayFromArrayForTesting(abc,odata.nvm,odata.ntm,odata.npm);
    final double pt0=abc3[0][0][0];

    final int [] mom_dims=new int []{odata.nvm,odata.ntm,odata.npm},
	         neq=new int []{neq_v,neq_t,neq_p};
    if(true)
    {
	final double [] p0=getAverageProjected(abc,mom_dims,new int []{0,0,neq_p},2,new double[]{Double.NaN,Double.NaN,Double.NaN},0);
	UnitTests.checkDoubleRel(getAverage(p0,abc3[0][0],pt0),"s(0)",neq_p,"neq_p",eps);

	final double [] qap0=CalculateKendall.QATradeIP.getPCoefs(abc3[0][0],abc3[0][0],neq_p)[1];
	UnitTests.checkDoubleArrayRel(p0,"p0",qap0,"qap0",eps);
	final double [] qvmx=QVM.getPDPCoefs(abc3[0][0],neq_p)[0];
	qvmx[0]-=1;
	UnitTests.checkDoubleArrayRel(p0,"p0",qvmx,"QVM.getPDPCoefs",eps);
	System.err.println("checkMatchingAverageProjections:[0,0] moment matched OK QATradeIP & QVM.getPDPCoefs for neq_p="+neq_p);

	final double p_offset2=odata.p_offset-2*(0.5-r.nextDouble());
	final double [] p0ds=getAverageProjected(odata.shiftOffset(0,0,odata.p_offset-p_offset2,QVMData.MOM_dT),mom_dims,new int []{0,0,neq_p},2,new double[]{Double.NaN,Double.NaN,Double.NaN},0);
	UnitTests.checkDoubleArrayRel1(p0,"p0",Polynomials.getPolynomialShifted(p0ds,odata.p_offset-p_offset2),"p0 after shift",eps);

	for(int ip=0;ip<11;ip++)
	{
	    final double p=CalculateKendall.TradeEMoments.getRandomPrice(r);
	    UnitTests.checkDoubleRel(
		Polynomials.getD0(p0,p-odata.p_offset),"I with p_offset1",
		Polynomials.getD0(p0ds,p-p_offset2),"I with p_offset2",eps);
	}
	System.err.println("checkMatchingAverageProjections:[0,0] p_offset shift matched for neq_p="+neq_p);
    }

    if(true)
    {
	final double [][][] P=new double [][][]{CalculateKendall.QMTradeIP.getPNoCondititionCoefs(abc3[0],neq_t,neq_p)};
	for(int iprt=0;iprt<100;iprt++)
	{
	    double dprt=r.nextDouble();
	    final double [] p0=getAverageProjected(abc,mom_dims,new int []{0,neq_t,neq_p},2,new double[]{Double.NaN,dprt,Double.NaN},0),
		qmp0=CalculateKendall.QVMTradeIP.getPProjected(P,Double.NaN,dprt);
	    UnitTests.checkDoubleArrayRel1(p0,"p0",qmp0,"qmp0",eps);

	    final double p_offset2=odata.p_offset-2*(0.5-r.nextDouble()),dto=2*(0.5-r.nextDouble());
	    final double [] p0ds=getAverageProjected(odata.shiftOffset(r.nextDouble(),dto,odata.p_offset-p_offset2,QVMData.MOM_dT),mom_dims,new int []{0,neq_t,neq_p},2,new double[]{Double.NaN,dprt+dto,Double.NaN},0);
	    UnitTests.checkDoubleArrayRel1(p0,"p0",Polynomials.getPolynomialShifted(p0ds,odata.p_offset-p_offset2),"p0 after shift",eps);
	    for(int ip=0;ip<11;ip++)
	    {
		final double p=CalculateKendall.TradeEMoments.getRandomPrice(r);
		UnitTests.checkDoubleRel(
		    Polynomials.getD0(p0,p-odata.p_offset),"I with p_offset1",
		    Polynomials.getD0(p0ds,p-p_offset2),"I with p_offset2",eps);
	    }
	}
	System.err.println("checkMatchingAverageProjections:[0] moment matched OK for neq_t="+neq_t+" neq_p="+neq_p);
	System.err.println("checkMatchingAverageProjections:[0] p_offset shift matched for neq_t="+neq_t+" neq_p="+neq_p);
    }

    for(int j=1;j<=2;j++)
    {
	final int [] neqs;
	final double [] d0,d1,dm1;
	final int iv,it;
	if(j==1 )
	{
	    if(neq_v<=0) continue;
	    neqs=new int []{1,0,neq_p};
	    d0=new double[]{0,Double.NaN,Double.NaN};
	    d1=new double[]{1,Double.NaN,Double.NaN};
	    dm1=new double[]{-1,Double.NaN,Double.NaN};
	    iv=1;
	    it=0;
	    System.err.println("checkMatchingAverageProjections:v-test");
	}
	else
	{
	    if(neq_t<=0) continue;
	    neqs=new int []{0,1,neq_p};
	    d0=new double[]{Double.NaN,0,Double.NaN};
	    d1=new double[]{Double.NaN,1,Double.NaN};
	    dm1=new double[]{Double.NaN,-1,Double.NaN};
	    iv=0;
	    it=1;
	    System.err.println("checkMatchingAverageProjections:t-test");
	}
	final double [] p0=getAverageProjected(abc,mom_dims,neqs,2,d0,0);
	final double [] pp=getAverageProjected(abc,mom_dims,neqs,2,d1,0);
	final double [] pm=getAverageProjected(abc,mom_dims,neqs,2,dm1,0);
	final double [] p1=new double [p0.length],p2=new double [p0.length];
	for(int k=p0.length;--k>=0;)
	{
	    p1[k]=0.5*(pp[k]-pm[k]);
	    p2[k]=0.5*(pp[k]+pm[k])-p0[k];
	}
	double s=0;
	for(int k=p0.length;--k>=0;)
	{
	    s+=p0[k]*(abc3[0*iv][0*it][k]/pt0);
	    s+=p1[k]*(abc3[1*iv][1*it][k]/pt0);
	    s+=p2[k]*(abc3[2*iv][2*it][k]/pt0);
	}
	UnitTests.checkDoubleRel(s,"s(0)",neq_p+1,"neq_p+1",eps);
    }


    System.err.println("checkMatchingAverageProjections: tests OK neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
}


private static void checkMatchingNF(final int neq_v,
				    final int neq_t,
				    final int neq_p0,
				    final java.util.Random r,
				    final QVMData odata,
				    final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final double [][][] abc3=QVMData.get3ArrayFromArrayForTesting(abc,odata.nvm,odata.ntm,odata.npm);
    final int [] mom_dims=new int []{odata.nvm,odata.ntm,odata.npm};

    for(int nf=0;nf<neq_p0;nf++)
    {
	final int neq_p=neq_p0-nf;
	if(true)
	{
	    // match to QATradeIP
	    final double []
		p0nf=getAverageProjected(abc,mom_dims,new int []{0,0,neq_p},2,new double[]{Double.NaN,Double.NaN,Double.NaN},nf),
		qap0nf=CalculateKendall.QATradeIP.getPNoCondititionCoefs(abc3[0][0],abc3[0][0],neq_p,nf)[1];
	    UnitTests.checkDoubleArrayRel1(p0nf,"p0nf",qap0nf,"qap0nf",eps);
	    System.err.println("checkMatchingNF:[0,0] moment matched QATradeIP OK for neq_p="+neq_p+" nf="+nf);

	    // check [0][0] invariance with respect to linear transforms conserving order
	    final double p_offset2=odata.p_offset-2*(0.5-r.nextDouble());
	    final double [] pt_shifted=odata.shiftOffset(0,0,odata.p_offset-p_offset2,QVMData.MOM_dT);
	    final double []
		qap0nf_offsetp=CalculateKendall.QATradeIP.getPNoCondititionCoefs(pt_shifted,pt_shifted,neq_p,nf)[1],
		p0nf_offsetp=getAverageProjected(pt_shifted,mom_dims,new int []{0,0,neq_p},2,new double[]{Double.NaN,Double.NaN,Double.NaN},nf);
	    UnitTests.checkDoubleArrayRel1(p0nf_offsetp,"p0nf_offsetp",qap0nf_offsetp,"qap0nf_offsetp",eps);

	    if(nf<=0)
	    {
		for(int ip=0;ip<11;ip++)
		{
		    final double p=CalculateKendall.TradeEMoments.getRandomPrice(r);
		    UnitTests.checkDoubleRel(
			Polynomials.getD0(qap0nf,p-odata.p_offset),"(QA: I with p_offset1 & nf)",
			Polynomials.getD0(qap0nf_offsetp,p-p_offset2),"(QA: I with p_offset2 && nf="+nf+")",eps);
		    UnitTests.checkDoubleRel(
			Polynomials.getD0(p0nf,p-odata.p_offset),"(QVM: I with p_offset1 & nf)",
			Polynomials.getD0(p0nf_offsetp,p-p_offset2),"(QVM: I with p_offset2 && nf="+nf+")",eps);
		}
		System.err.println("checkMatchingNF:[0,0] p0nf & p0nf_offsetp matched for neq_p="+neq_p+" nf="+nf);
	    }
	    else
	    {
		System.err.println("checkMatchingNF:[0,0] p0nf & p0nf_offsetp match is skipped for neq_p="+neq_p+" nf="+nf+
				   " because of still unknown invariant group & equations decoupling procedure.");
	    }
	}

	if(neq_t>0)
	{ // match to QQTradeIP
	    final double [][][] P=new double [][][]{
		CalculateKendall.QQTradeIP.getPNoCondititionCoefs(abc3[0][0],abc3[0][1],abc3[0][2],neq_p,nf)};
	    for(int iprt=0;iprt<100;iprt++)
	    {
		final double dprt=r.nextDouble();
		final double []
		    qqp0nf=CalculateKendall.QVMTradeIP.getPProjected(P,Double.NaN,dprt),
		    qvmp0nf=getAverageProjected(abc,mom_dims,new int []{0,1,neq_p},2,new double[]{Double.NaN,dprt,Double.NaN},nf);
		UnitTests.checkDoubleArrayRel1(qvmp0nf,"qvmp0nf",qqp0nf,"qqp0nf",eps);
	    }
	    System.err.println("checkMatchingNF:[0] moment matched QQTradeIP OK for neq_p="+neq_p+" neq_t="+1+" nf="+nf);
	}

    }
}

    public static void main(String [] args)
    {
	System.err.println("QVMAverage::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int neq_v=2;neq_v<=6;neq_v++)
	{
	    for(int neq_t=1;neq_t<=6;neq_t++)
	    {
		for(int neq_p=3;neq_p<=8;neq_p++)
		{
		    for(int i=0;i<3;i++)
		    {
			checkCalculations(100+r.nextInt(100),r,neq_v,neq_t,neq_p);
			checkCalculations(10010+r.nextInt(100010),r,neq_v,neq_t,neq_p);
		    }
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QVMAverage::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    }

} // END of class QVMAverage


private static final int RES_SUM=89,RES_MAX=90,RES_MAXF=91,RES_FLAST=92,RES_FDIRECT=93;

/** QVMEV version of QVMCoefs. */
static class QVMEV extends QVMCoefs {


private QVMEV(final com.polytechnik.utils.PolynomialRootsFinder rf,
	      final int neq_v,final int neq_t,final int neq_p,
	      final QVMData odata)
{
	super(rf,neq_v,neq_t,neq_p,odata);
}

    double [] getPCoefs(final double dv,final double dt,final double dp)
    {
	final QVMData o=super.odata;
	if(false)
	{
	    return getEVProjected(o.tmoments_vol_x_tau_x_p,
				  new int []{o.nvm,o.ntm,o.npm},
				  new int []{this.neq_v,this.neq_t,this.neq_p},
				  2,
				  new double []{dv,dt,dp},
				  RES_SUM);
	}
	else
	{
	    return getEVProjected(o.tmoments_vol_x_tau_x_p,
				  new int []{o.nvm,o.ntm,o.npm},
				  new int []{this.neq_v,this.neq_t,this.neq_p},
				  2,
				  new double []{0,0,dp},
				  RES_MAX);
	}
    }

private static double [] getEVProjected(final double [] mom,
					final int [] mom_dims,
					final int [] neq,
					final int proj_ptr,
					final double [] dproj,
					final int res)
{
    final int nd=neq.length;
    if(proj_ptr<0||proj_ptr>=nd||mom_dims.length!=nd)
    {
	throw new IllegalArgumentException("proj_ptr="+proj_ptr+
					   " or mom_dims.length="+mom_dims.length+
					   " does not correspond to neq.length="+neq.length);
    }

    int mom_nels=1,neqs=0;
    for(int k=nd;--k>=0;)
    {
	final int kdim=mom_dims[k];
	if(kdim<1)
	{
	    throw new IllegalArgumentException("mom_dims["+k+"]="+mom_dims[k]);
	}
	mom_nels*=kdim;
	final int kneq=neq[k];
	if(kneq<0 || (1+2*kneq)>kdim)
	{
	    throw new IllegalArgumentException("neq["+k+"]="+neq[k]+
					       " does not correspond to"+
					       " mom_dims["+k+"]="+mom_dims[k]);
	}
	neqs+=kneq;
    }
    if(mom.length!=mom_nels)
    {
	throw new IllegalArgumentException("mom.length="+mom.length+
					   " does not correspond the size calculated from mom_dims;"+
					   " size="+mom_nels);
    }
    final int ne=neqs,momoffs=mom_nels,neq_proj1=1+neq[proj_ptr];

    final double pt0=mom[0];
    final double [] norms=new double [ne];

    for(int ek=0,offsk=momoffs,ik=0;ek<nd;ek++)
    {
	for(int k=1,nk=neq[ek],koffsk=(offsk/=mom_dims[ek]);k<=nk;k++,koffsk+=offsk,ik++)
	{
	    norms[ik]=Math.sqrt(mom[2*koffsk]/pt0-Math.pow(mom[koffsk]/pt0,2));
	}
    }

    final double [] matr=new double [ne*ne];
    for(int ek=0,offsk=momoffs,ik=0,j=0;ek<nd;ek++)
    {
	for(int k=1,nk=neq[ek],koffsk=(offsk/=mom_dims[ek]);k<=nk;k++,koffsk+=offsk,ik++)
	{
	    final double normsk=norms[ik],pk=mom[koffsk]/pt0;
	    for(int el=0,offsl=momoffs,il=0;el<nd;el++)
	    {
		for(int l=1,nl=neq[el],loffsl=(offsl/=mom_dims[el]);l<=nl;l++,loffsl+=offsl,il++,j++)
		{
		    final double pl=mom[loffsl]/pt0,pkl=mom[koffsk+loffsl]/pt0;
		    matr[j]=(pkl-pl*pk)/normsk/norms[il];
		}
	    }
	}
    }


    final double [] pproj;
    if(res==RES_SUM || res==RES_MAX){pproj=new double [2*neq_proj1-1];}
    else if(res==RES_MAXF){pproj=new double [neq_proj1];}
    else{throw new IllegalArgumentException("res="+res);}


    final double [] lambdas=new double [ne],work=new double [34*ne+5];
    /*
    final org.netlib.util.intW info = new org.netlib.util.intW(2);
    // on output the eigenvectors are in matr
    org.netlib.lapack.Dsyev.dsyev("V","U",ne,matr,0,ne,lambdas,0,work,0,work.length,info);
    final int pinfo=info.val;
    */
    final int pinfo=-1;
    if(pinfo<0) throw new RuntimeException("Temporary disabled to avoid jlapack dependency");

    if(pinfo!=0)
    {
	java.util.Arrays.fill(pproj,Double.NaN);
	return pproj;
    }


    double maxdk=Double.NEGATIVE_INFINITY;
    final double [] pevnormalized=new double [neq_proj1];
    for(int ev=ne,oev=ne*(ne-1);--ev>=0;oev-=ne)
    {
	final double sqrtlambda=Math.sqrt(lambdas[ev]);
	pevnormalized[0]=0;
	for(int ek=0,offsk=momoffs,ik=0;ek<nd;ek++)
	{
	    final double drwk=dproj[ek];
	    double rwk=1;
	    for(int k=1,nk=neq[ek],koffsk=(offsk/=mom_dims[ek]);k<=nk;k++,koffsk+=offsk,ik++)
	    {
		final double wk=mom[koffsk]/pt0,c=matr[ik+oev]/sqrtlambda/norms[ik];
		rwk*=drwk;
		// project
		if(ek==proj_ptr)
		{
		    pevnormalized[k]=c;
		    pevnormalized[0]-=wk*c;
		}
		else
		{
		    pevnormalized[0]+=(rwk-wk)*c;
		}
	    }
	}
	if(res==RES_SUM)
	{
	    DeprecatedHelperTMatr.sumP2(pproj,pevnormalized);
	}
	else
	{
	    if(neq_proj1!=pevnormalized.length) throw new RuntimeException();
	    final double dk=Math.abs(Polynomials.getD0(pevnormalized,dproj[proj_ptr]));
	    if(dk>=maxdk)
	    {
		maxdk=dk;
		if(res==RES_MAXF)
		{
		    System.arraycopy(pevnormalized,0,pproj,0,neq_proj1);
		}
		else if(res==RES_MAX)
		{
		    java.util.Arrays.fill(pproj,0);
		    DeprecatedHelperTMatr.sumP2(pproj,pevnormalized);
		}
		else
		{
		    throw new RuntimeException();
		}
	    }
	}
    }
    return pproj;
}


   public static QVMEV calculateQVMEV(final int neq_v,
				      final int neq_t,
				      final int neq_p,
				      final long currenttime,
				      final long tau,
				      final Trade [] tb,
				      final int tsize,
				      final com.polytechnik.utils.PolynomialRootsFinder rf,
				      final QVMEV ref)
   {
	final QVMData o=QVMData.calculateQVMData(2*neq_v+1,
						 2*neq_t+1,
						 2*neq_p+1,
						 currenttime,
						 tau,
						 tb,
						 tsize,
						 ref!=null?ref.odata:null);
	return new QVMEV(rf,neq_v,neq_t,neq_p,o);
   }





private static void checkMatchingAverageSum(final int neq_v,
					    final int neq_t,
					    final int neq_p,
					    final java.util.Random r,
					    final QVMData odata,
					    final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final double [][][] abc3=QVMData.get3ArrayFromArrayForTesting(abc,odata.nvm,odata.ntm,odata.npm);
    final double pt0=abc3[0][0][0];

    final int [] mom_dims=new int []{odata.nvm,odata.ntm,odata.npm},
	         neq=new int []{neq_v,neq_t,neq_p};
    if(true)
    {
	final double [] dproj=new double[]{Double.NaN,Double.NaN,Double.NaN};
 	final double [] p0=getEVProjected(abc,mom_dims,new int []{0,0,neq_p},2,dproj,RES_SUM);
	UnitTests.checkDoubleRel(getAverage(p0,abc3[0][0],pt0),"s(0)",neq_p,"neq_p",eps);

	final double [] qap0=CalculateKendall.QATradeIP.getPCoefs(abc3[0][0],abc3[0][0],neq_p)[1];
	UnitTests.checkDoubleArrayRel(p0,"p0",qap0,"qap0",eps);
	final double [] qvmx=QVM.getPDPCoefs(abc3[0][0],neq_p)[0];
	qvmx[0]-=1;
	UnitTests.checkDoubleArrayRel(p0,"p0",qvmx,"QVM.getPDPCoefs",eps);
	System.err.println("checkMatchingAverageSum:[0,0] moment matched OK for neq_p="+neq_p);

	dproj[2]=CalculateKendall.TradeEMoments.getRandomPrice(r)-odata.p_offset;
	final double [] p0m=getEVProjected(abc,mom_dims,new int []{0,0,neq_p},2,dproj,RES_MAX);
	UnitTests.checkDoubleRel(getAverage(p0m,abc3[0][0],pt0),"sev(0)",1,"1",eps);


	final double [] p_RES_SUM=getEVProjected(abc,mom_dims,new int []{0,0,1},2,dproj,RES_SUM),
	    p_RES_MAX=getEVProjected(abc,mom_dims,new int []{0,0,1},2,dproj,RES_MAX);
	UnitTests.checkDoubleArrayRel(p_RES_SUM,"p_RES_SUM",p_RES_MAX,"p_RES_MAX",eps);
	System.err.println("checkMatchingAverageSum: p_RES_SUM & RES_MAX matched for d=1");

	final double [] pev=new double [3];
	DeprecatedHelperTMatr.sumP2(pev,getEVProjected(abc,mom_dims,new int []{0,0,1},2,dproj,RES_MAXF));
	UnitTests.checkDoubleArrayRel(p_RES_SUM,"p_RES_SUM",pev,"pev",eps);
	System.err.println("checkMatchingAverageSum: p_RES_SUM & pev matched for d=1");
    }

    for(int iv=0;iv<15;iv++)
    {
	double dv=r.nextDouble();
	for(int it=0;it<17;it++)
	{
	    double dt=r.nextDouble();
	    final double [] dproj=new double []{dv,dt,Double.NaN};
	    final double [] paverage=QVMAverage.getAverageProjected(abc,mom_dims,neq,2,dproj,0),
		pevaverage=getEVProjected(abc,mom_dims,neq,2,dproj,RES_SUM);
	    UnitTests.checkDoubleArrayRel(paverage,"paverage form QVMAverage",pevaverage,"pevaverage from QEMEV",eps);

	    // check basis invariance
	    final double p_offset2=odata.p_offset-2*(0.5-r.nextDouble()),dto=0.3*2*(0.5-r.nextDouble()),dvo=0.3*2*(0.5-r.nextDouble());
	    final double [] abc_shifted=odata.shiftOffset(dvo,dto,odata.p_offset-p_offset2,QVMData.MOM_dT);
	    final double [] pevaverage_offset2=getEVProjected(abc_shifted,mom_dims,neq,2,new double []{dv+dvo,dt+dto,Double.NaN},RES_SUM);
	    for(int ip=0;ip<12;ip++)
	    {
		final double p=CalculateKendall.TradeEMoments.getRandomPrice(r);
		UnitTests.checkDoubleRel(
		    Polynomials.getD0(pevaverage,p-odata.p_offset),"I with offset",
		    Polynomials.getD0(pevaverage_offset2,p-p_offset2),"I with offset2",eps);
	    }

	    if(false)
	    {
		dproj[2]=CalculateKendall.TradeEMoments.getRandomPrice(r)-odata.p_offset;
		final double [] ev_off=getEVProjected(abc,mom_dims,neq,2,dproj,RES_MAXF),
		    ev_off2=getEVProjected(abc_shifted,mom_dims,neq,2,new double []{dproj[0]+dvo,dproj[1]+dto,dproj[2]+(odata.p_offset-p_offset2)},RES_MAXF);
		for(int ip=0;ip<12;ip++)
		{
		    final double p=CalculateKendall.TradeEMoments.getRandomPrice(r);
		    UnitTests.checkDoubleRel(
			Polynomials.getD0(ev_off,p-odata.p_offset),"ev_off",
			Polynomials.getD0(ev_off2,p-p_offset2),"ev_off2",eps);
		}
	    }
	    else
	    {
		//System.err.println("QVMEV:checkMatchingAverageSum:EV invariance test is skipped as it is not basis invariant");
	    }
	}
    }
    System.err.println("QVMEV:checkMatchingAverageSum:EV invariance test is skipped as it is not basis invariant");
    System.err.println("QVMEV:checkMatchingAverageSum: matched QVMAverage OK for neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
    System.err.println("QVMEV:checkMatchingAverageSum: basis invariance OK for neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);



    System.err.println("QVMEV:checkMatchingAverageSum: tests OK neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
}



    private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_v,final int neq_t,final int neq_p)
    {
	final double eps=1e-5;
	final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("QVMEV:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);

	final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();

	final QVMEV ro=QVMEV.calculateQVMEV(neq_v,
					    neq_t,
					    neq_p,
					    currenttime,
					    tau,
					    tb,
					    tsize,
					    rf,
					    null);
	// check QVMTradeIP match
	QVMCoefs.checkQVMTradeIP(ro,tb,tsize,eps,rf);

	// average projections
	checkMatchingAverageSum(neq_v,neq_t,neq_p,r,ro.odata,eps);


	System.err.println("QVMEV:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
    }


    public static void main(String [] args)
    {
	System.err.println("QVMEV::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79093);
	for(int neq_v=0;neq_v<=5;neq_v++)
	{
	    for(int neq_t=0;neq_t<=5;neq_t++)
	    {
		for(int neq_p=2;neq_p<=5;neq_p++)
		{
		    for(int i=0;i<110;i++)
		    {
			checkCalculations(100+r.nextInt(100),r,neq_v,neq_t,neq_p);
			checkCalculations(10010+r.nextInt(100010),r,neq_v,neq_t,neq_p);
		    }
		}
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QVMEV::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    }




} // end of class QVMEV




/** Like getLinsysSol but with 2 basises: first(A) -polynomials, second(B) -orthogonal.
 *  Right part is assumed to have the same offset as mom.
 */
private static double [][] getLinsys2(final int nA,final int nB,
				      final double [] mom,final int doffset,
				      final double [] rpart)
{
    if(nA<1||nB<1) throw new IllegalArgumentException();
    final int neq=nA+nB-1,neqa=nA-1; // assume const is in B basis.
    final double pt0=mom[0];

    final double [] r=new double [neq],matr=new double [neq*neq],norma=new double [neqa];
    for(int m=0;m<neq;m++)
    {
	if(m<neqa)
	{
	    r[m]=rpart[m+1]/(norma[m]=Math.sqrt(mom[2*(m+1)]));
	}
	else
	{
	    r[m]=rpart[(m-neqa)*doffset];
	}

	for(int k=0;k<=m;k++)
	{
	    final double v;
	    if(m<neqa)
	    {
		if(k<neqa)
		{
		    // A basis is assumed to be polynomials
		    v=mom[k+1+m+1]/norma[m]/norma[k];
		}
		else
		{
		    v=mom[m+1+(k-neqa)*doffset]/norma[m];
		}
	    }
	    else
	    {
		if(k<neqa)
		{
		    v=mom[k+1+(m-neqa)*doffset]/norma[k];
		}
		else
		{
		    // B basis is assumed to be orthogonal
		    v=(k==m?pt0:0);
		}
	    }
	    matr[m+k*neq]=matr[k+m*neq]=v;
	}
    }

    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(r,Double.NaN);
    }
    else
    {
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,r,1);
    }
    final double [] ra=new double [nA],rb=new double [nB];
    for(int k=neqa;--k>=0;){ra[k+1]=r[k]/norma[k];} // ra[0] is always =0
    System.arraycopy(r,neqa,rb,0,nB);
    return new double [][]{ra,rb};
}


/** Like getLinsys2 but input is in polynomial basis, output is the same as getLinsys2. Used for testing.
 */
private static double [][] testing_polynomialInput_getLinsys2(final int nA,final int nB,
							      final double [] mom,final int doffset,
							      final double [] rpart,
							      final double [][] BASIS)
{
    if(nA<1||nB<1) throw new IllegalArgumentException();
    final int neq=nA+nB-1,neqa=nA-1; // assume const is in B basis.
    final double [] matr=new double [neq*neq],r=new double [neq];
    final double scale=mom[0]/rpart[0];
    for(int m=0;m<neq;m++)
    {
	if(m<neqa)
	{
	    r[m]=rpart[m+1]*scale;
	}
	else
	{
	    r[m]=rpart[(m-neqa)*doffset]*scale;
	}
	for(int k=0;k<neq;k++)
	{
	    final double v;
	    if(m<neqa)
	    {
		if(k<neqa)
		{
		    v=mom[m+1+k+1];
		}
		else
		{
		    v=mom[m+1+(k-neqa)*doffset];
		}
	    }
	    else
	    {
		if(k<neqa)
		{
		    v=mom[k+1+(m-neqa)*doffset];
		}
		else
		{
		    v=mom[(m-neqa+k-neqa)*doffset];
		}
	    }
	    matr[k+m*neq]=v;
	}
    }
    final int [] ipvt=new int [neq];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq,neq,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(r,Double.NaN);
    }
    else
    {
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq,neq,ipvt,r,1);
    }
    final double [] pvcoefs=new double [nA];
    System.arraycopy(r,0,pvcoefs,1,neqa);
    System.arraycopy(r,neqa,matr,0,nB);
    return new double [][]{pvcoefs,convertToPBasis(matr,nB,BASIS)};

}

private static double [] convertToPBasis(final double [] pol,final int nels,final double [][] BASIS)
{
    return TMatr.expandOverLowerDiagBasis(java.util.Arrays.copyOf(pol,nels),BASIS);
}


/*private*/ static double [] getPolynomialAIntegrated(final double [] pcoefs,final double [] mom,final double pt0)
{
    return getPolynomialAIntegrated(pcoefs,pcoefs.length,mom,pt0);
}
/*private*/ static double [] getPolynomialAIntegrated(final double [] pcoefs,final int nels,final double [] mom,final double pt0)
{
    final double [] pol=new double [1+nels];
    setPolynomialAIntegrated(pol,pcoefs,nels,mom,pt0);
    return pol;
}
private static void setPolynomialAIntegrated(final double [] pol,final double [] pcoefs,final int nels,final double [] mom,final double pt0)
{
    double s=0;
    for(int k=0;k<nels;)
    {
	final double p=pcoefs[k++];
	s+=(pol[k]=p/k)*(mom[k]/pt0);
    }
    pol[0]=-s;
}


static double getAverage(final double [] pcoefs,final double [] mom,final double pt0)
{
    return getAverage(pcoefs,pcoefs.length,mom,0,1,pt0);
}
static double getAverage(final double [] pcoefs,final int nels,final double [] mom,final double pt0)
{
    return getAverage(pcoefs,nels,mom,0,1,pt0);
}
static double getAverage(final double [] pcoefs,final int nels,final double [] mom,final int mom_offset,final double pt0)
{
    return getAverage(pcoefs,nels,mom,mom_offset,1,pt0);
}
static double getAverage(final double [] pcoefs,final int nels,final double [] mom,final int mom_offset,final int mom_increment,final double pt0)
{
    double s=0;
    for(int k=0,o=mom_offset;k<nels;k++,o+=mom_increment)
    {
	s+=pcoefs[k]*(mom[o]/pt0);
    }
    return s;
}
private static double sumElementse(final double [] pcoefs,final int nels)
{
    double s=0;
    for(int k=0;k<nels;k++)
    {
	s+=pcoefs[k];
    }
    return s;
}


private static double [] getP1xP2(final double [] p1,final int n1,final double [] p2,final int n2)
{
    if(n1!=n2) throw new IllegalArgumentException();
    final double [] r=new double [n1];
    for(int i=n1;--i>=0;)
    {
	r[i]=p1[i]*p2[i];
    }
    return r;
}



private static int getIndAMAX(final int i_low,final int i_high,final double [] amax)
{
    if(i_low<0) return i_high;
    if(i_high<0) return i_low;
    if(amax[i_high]<amax[i_low]) return i_low;
    return i_high;
}


static class QVMVT {
    /*private*/ final QVMData odata;
    final QVMORT vm,tm;
    final double pvpa_low,ptpa_low,pvpa_high,ptpa_high;
    final double pvya_low,ptya_low,pvya_high,ptya_high,pvya,ptya;
    final double pv_low,ptiv_low,pv_high,ptiv_high,wv_low,wtiv_low,wv_high,wtiv_high;
    final int enter_buy,exit_buy,enter_sell,exit_sell;
    final double pav_low,pav_high,pat_low,pat_high;
    final double wav_low,wav_high,wat_low,wat_high;
    final double sd_low,sd_high,xsd_low,xsd_high;

    final double ptiv_close,pvit_close;
    final double p2v_close,p2tiv_close;

    QVMVT(final QVMORT vm,
	  final QVMORT tm,
	  final QVMData odata)
    {
	this.vm=vm;
	this.tm=tm;
	this.odata=odata;

	final double p_last=this.odata.trade_last==null?Double.NaN:this.odata.trade_last.price;
	final int [] in=getPRangeIndex(p_last,vm.p_imin);
	final int i_low=in[0],i_high=in[1];

	if(vm.p_imin.length!=tm.p_imin.length)
	{
	    this.ptiv_close=this.pvit_close=Double.NaN;
	    this.p2v_close=this.p2tiv_close=Double.NaN;
	}
	else
	{
	    this.ptiv_close=vm.ind_close<0?Double.NaN:tm.p_imin[vm.ind_close];
	    this.pvit_close=tm.ind_close<0?Double.NaN:vm.p_imin[tm.ind_close];

	    final int ind_v_close=-1;//getIndAMAX(i_low,i_high,vm.dtx);
	    this.p2v_close=ind_v_close<0?Double.NaN:vm.p_imin[ind_v_close];
	    this.p2tiv_close=ind_v_close<0?Double.NaN:tm.p_imin[ind_v_close];
	}


	if(vm.p_imin.length!=tm.p_imin.length||i_low<0||i_high<0)
	{
	    this.pvpa_low=this.ptpa_low=this.pvpa_high=this.ptpa_high=Double.NaN;
	    this.pvya_low=this.ptya_low=this.pvya_high=this.ptya_high=this.pvya=this.ptya=Double.NaN;
	}
	else
	{
	    final double wwv=vm.icoefs[i_low]+vm.icoefs[i_high],wwt=tm.icoefs[i_low]+tm.icoefs[i_high];
	    this.pvpa_low=(vm.p_imin[i_low]-p_last)*(vm.icoefs[i_low]/wwv);
	    this.pvpa_high=(vm.p_imin[i_high]-p_last)*(vm.icoefs[i_high]/wwv);
	    this.ptpa_low=(tm.p_imin[i_low]-p_last)*(tm.icoefs[i_low]/wwt);
	    this.ptpa_high=(tm.p_imin[i_high]-p_last)*(tm.icoefs[i_high]/wwt);


	    {
	    final double wvl=vm.icoefs[i_low],wvh=vm.icoefs[i_high],dpv=vm.p_imin[i_high]-vm.p_imin[i_low],
		wvla=((vm.p_imin[i_high]-p_last)/dpv)*wvl,wvha=((p_last-vm.p_imin[i_low])/dpv)*wvl;
	    this.pvya_low=(vm.p_imin[i_low]*wvl+p_last*wvha)/(wvl+wvha);
	    this.pvya_high=(p_last*wvla+vm.p_imin[i_high]*wvh)/(wvla+wvh);
	    this.pvya=(vm.p_imin[i_low]*wvla+vm.p_imin[i_high]*wvha)/(wvla+wvha);
	    }

	    {
	    final double wtl=tm.icoefs[i_low],wth=tm.icoefs[i_high],dpt=tm.p_imin[i_high]-tm.p_imin[i_low],
		wtla=((tm.p_imin[i_high]-p_last)/dpt)*wtl,wtha=((p_last-tm.p_imin[i_low])/dpt)*wtl;
	    this.ptya_low=(tm.p_imin[i_low]*wtl+p_last*wtha)/(wtl+wtha);
	    this.ptya_high=(p_last*wtla+tm.p_imin[i_high]*wth)/(wtla+wth);
	    this.ptya=(tm.p_imin[i_low]*wtla+tm.p_imin[i_high]*wtha)/(wtla+wtha);
	    }

	}


	if(i_low<0||vm.p_imin.length!=tm.p_imin.length)
	{
	    this.pv_low=this.ptiv_low=this.wv_low=this.wtiv_low=Double.NaN;
	}
	else
	{
	    this.pv_low=vm.p_imin[i_low];
	    this.ptiv_low=tm.p_imin[i_low];
	    this.wv_low=vm.icoefs[i_low];
	    this.wtiv_low=tm.icoefs[i_low];
	}
	if(i_high<0||vm.p_imin.length!=tm.p_imin.length)
	{
	    this.pv_high=this.ptiv_high=this.wv_high=this.wtiv_high=Double.NaN;
	}
	else
	{
	    this.pv_high=vm.p_imin[i_high];
	    this.ptiv_high=tm.p_imin[i_high];
	    this.wv_high=vm.icoefs[i_high];
	    this.wtiv_high=tm.icoefs[i_high];
	}

	if(vm.p_imin.length!=tm.p_imin.length)
	{
	    this.sd_low=this.sd_high=this.xsd_low=this.xsd_high=Double.NaN;
	    this.pav_low=this.pav_high=this.pat_low=this.pat_high=Double.NaN;
	    this.wav_low=this.wav_high=this.wat_low=this.wat_high=Double.NaN;
	}
	else
	{
	    this.xsd_low=i_low<0?Double.NaN:(vm.p_imin[i_low]-p_last)*vm.icoefs[i_low]-(tm.p_imin[i_low]-p_last)*tm.icoefs[i_low];
	    this.xsd_high=i_high<0?Double.NaN:(vm.p_imin[i_high]-p_last)*vm.icoefs[i_high]-(tm.p_imin[i_high]-p_last)*tm.icoefs[i_high];

	    double wv_low=0,wv_high=0,wt_low=0,wt_high=0,sv_low=0,sv_high=0,st_low=0,st_high=0;
	    double sd_high=0,sd_low=0;
	    for(int i=vm.p_imin.length;--i>=0;)
	    {
		final double d=(vm.p_imin[i]-p_last)*vm.icoefs[i]-(tm.p_imin[i]-p_last)*tm.icoefs[i];
		if(i_high>=0 && i>=i_high)
		{
		    sd_high+=d;

		    wv_high+=vm.icoefs[i];
		    sv_high+=(vm.p_imin[i]-p_last)*vm.icoefs[i];
		    wt_high+=tm.icoefs[i];
		    st_high+=(tm.p_imin[i]-p_last)*tm.icoefs[i];
		}
		if(i_low>=0 && i<=i_low)
		{
		    sd_low+=d;

		    wv_low+=vm.icoefs[i];
		    sv_low+=(vm.p_imin[i]-p_last)*vm.icoefs[i];
		    wt_low+=tm.icoefs[i];
		    st_low+=(tm.p_imin[i]-p_last)*tm.icoefs[i];
		}
	    }
	    this.sd_low=sd_low;
	    this.sd_high=sd_high;

	    this.pav_low=i_low>=0?sv_low/wv_low:Double.NaN;
	    this.pav_high=i_high>=0?sv_high/wv_high:Double.NaN;
	    this.pat_low=i_low>=0?st_low/wt_low:Double.NaN;
	    this.pat_high=i_high>=0?st_high/wt_high:Double.NaN;
	    this.wav_low=wv_low;
	    this.wav_high=wv_high;
	    this.wat_low=wt_low;
	    this.wat_high=wt_high;
	}


        final double eps_enter=0,eps_exit=0;
	if(vm.p_imin.length!=tm.p_imin.length)
	{
	    this.enter_buy=this.exit_buy=this.enter_sell=this.exit_sell=0;
	}
	else
	{
	    final int nr=vm.p_imin.length;
	    final double dl=i_low >=0? tm.p_imin[i_low] -vm.p_imin[i_low] :Double.NaN,
		         dh=i_high>=0? tm.p_imin[i_high]-vm.p_imin[i_high]:Double.NaN;

	    final int i_sh=i_high<0 || i_high>=nr-1?-1:i_high+1;
	    this.enter_buy=
		!(dl < -eps_enter) || !(dh < -eps_enter) ? 0:
		((i_sh>=0   && tm.p_imin[i_sh]-vm.p_imin[i_sh]<-eps_enter) ? 1: (p_last<tm.p_imin[i_high]?2:0));
	    this.exit_buy=i_high>=0?(dh > eps_exit?1:0):(dl > eps_exit?2:0);

	    final int i_sl=i_low<=0?-1:i_low-1;
	    this.enter_sell=
		!(dl > eps_enter) || !(dh > eps_enter) ? 0:
		((i_sl>=0 && tm.p_imin[i_sl]-vm.p_imin[i_sl]>eps_enter) ? 1: (p_last>tm.p_imin[i_low]?2:0));
	    this.exit_sell=i_low>=0?(dl < -eps_exit?1:0):(dh < -eps_exit?2:0);
	}

    }
    public static QVMVT calculateQVMVT(final int neq_p,
				       final long currenttime,
				       final long tau,
				       final Trade [] tb,
				       final int tsize,
				       final com.polytechnik.utils.PolynomialRootsFinder rf,
				       final QVMVT ref)
{
    final QVMData o=QVMData.calculateQVMData(1+1,
					     1+1,
					     1+2*neq_p,
					     currenttime,
					     tau,
					     tb,
					     tsize,
					     ref!=null?ref.odata:null);
    final QVMORT v=new QVMORT(rf,neq_p,o,o.vmoments_vol_x_tau_x_p,QVMData.MOM_dT),
	t=new QVMORT(rf,neq_p,o,o.tmoments_vol_x_tau_x_p,QVMData.MOM_dT);
    return new QVMVT(v,t,o);
}

} // END of class QVMVT


/** QVMORT version of QVMCoefs. */
static class QVMORT {

final double sigma,p_average;
final double [] p_imin,icoefs;
final int neq_p;
/*private*/ final QVMData odata;

final double deltaS,dmudp_pl,mu_pl,dvldp_pl,vl_pl,dtdp_pl,t_pl;
final int ind_close;
final double [] dtx;
final double [] p_real_roots;
final double [] QQ;

private QVMORT(final com.polytechnik.utils.PolynomialRootsFinder rf,
	       final int neq_p,
	       final QVMData odata,
	       final double [] odata_abc,
	       final int mom_type)
{
    this.odata=odata;
    this.neq_p=neq_p;

    final double [] abc=odata_abc;//odata.moments_vol_x_tau_x_p;
    this.QQ=getQQMatrFromMonomialsMoments(neq_p,abc);

    final double pt0=abc[0],pa=odata.npm<2?Double.NaN:abc[1]/pt0;
    this.p_average=pa+odata.p_offset;
    this.sigma=odata.npm<3?Double.NaN:Math.sqrt((abc[2]/pt0)-(pa*pa));

    final double [][] x=getPCoefs(abc,mom_type);
    final double [] pcoefs=x[0],wcoefs=x[1],dpcoefs=x[2],tpcoefs=x[3];
    if(true)
    {
	final double p_last=this.odata.trade_last==null?Double.NaN:this.odata.trade_last.price;
	// to remove?
	final double dp_last=p_last-odata.p_offset;
	this.mu_pl=Polynomials.getD0(wcoefs,dp_last);
	this.dmudp_pl=0.5*Polynomials.getD1(wcoefs,dp_last);
	this.vl_pl=Polynomials.getD0(dpcoefs,dp_last);
	this.dvldp_pl=0.5*Polynomials.getD1(dpcoefs,dp_last);
	this.deltaS=this.dvldp_pl/this.vl_pl-this.dmudp_pl/this.mu_pl;
	this.dtdp_pl=tpcoefs==null?Double.NaN:Polynomials.getD1(tpcoefs,dp_last);
	this.t_pl=tpcoefs==null?Double.NaN:Polynomials.getD0(tpcoefs,dp_last);
    }

    final PolynomialRootsFinder.PRoots u;
    if(neq_p>4){
	if(rf==null) {
	    u=new PolynomialRootsFinder.PRoots(0,pcoefs,new double [neq_p],new double [neq_p]);
	} else {
	    u=rf.findRoots(pcoefs);
	}
    } else {
	u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoefs);
    }
    final int nroots=u.nroots;
    final double [] root_re=u.re_root,root_im=u.im_root;

    // insertion sort by real part with extra load
    SortArrays.sortArrayWithLoad(nroots,root_re,root_im);
    this.p_real_roots=root_re.clone();

    this.p_imin=java.util.Arrays.copyOf(root_re,
	CalculateKendall.selectRealRoots(root_re,nroots,root_re,root_im,CalculateKendall.EPS_IM));

    this.icoefs=this.p_imin.length==root_im.length?root_im:new double [this.p_imin.length]; // try to re-use an array
    int ind_close=-1;double distm=Double.NEGATIVE_INFINITY;
    this.dtx=new double [this.p_imin.length]; // FIXME to remove
    final double ta=mom_type>0?-abc[getOOffset(odata,mom_type)]/pt0:Double.NaN;
    for(int k=this.p_imin.length;--k>=0;)
    {
	final double pk=this.p_imin[k];
	if(false)
	{
	    this.icoefs[k]=1/Polynomials.getD0(wcoefs,pk);
	}
	else
	{ // different way to calculate icoefs, less prone to numerical instability
	    final double [] tmp=Polynomials.sdiv1(pcoefs,pk).q;
	    if(neq_p!=tmp.length) throw new RuntimeException();
	    this.icoefs[k]=getAverage(tmp,neq_p,abc,pt0)/Polynomials.getD0(tmp,pk);
	}
	if(tpcoefs!=null)
	{
	    final double dist=Polynomials.getD0(tpcoefs,pk); // /this.icoefs[k];
	    if(dist<=ta*this.icoefs[k] && dist>=distm){distm=dist;ind_close=k;}
	    this.dtx[k]=dist; // FIXME to remove
	}
	this.p_imin[k]+=odata.p_offset; // shift to 0 offset
    }
    this.ind_close=ind_close;

}


private double [][] getPCoefs(final double [] abc,final int mom_type)
{

    final int o;final double [] rp;
    if(mom_type>0)
    {
	o=getOOffset(odata,mom_type);
	rp=abc;
    }
    else
    {
	o=-1;
	rp=null;
    }
    return getFLast(abc,0,odata.npm,this.neq_p,rp,o); // use <tp^k> moments if available
}

public static double [] getQQMatrFromMonomialsMoments(final int n,final double [] moms)
{
    final double [] QQ=new double [n*n];
    for(int k=0;k<n;k++)
	for(int l=0;l<n;l++)
	    QQ[k*n+l]=moms[k+l];
    return QQ;
}
/*
public static double [] getQudraticFormPolynomial(final int n,final double [] QQ)
{
    final double [] p=new double [2*n-1];
    for(int k=0;k<n;k++)
	for(int l=0;l<n;l++)
	    p[k+l]+=QQ[k*n+l];
    return p;
}
*/


private static double [] getRPEM(final QVMData odata,final int neq)
{
    final double [] v=odata.vmoments_vol_x_tau_x_p,t=odata.tmoments_vol_x_tau_x_p;
    final int ot=getOOffset(odata,QVMData.MOM_dT),ov=getOOffset(odata,QVMData.MOM_dV);
    final double [] r=new double [neq];
    for(int k=r.length;--k>=0;)
    {
	r[k]=v[k+ot]/odata.v_scale-t[k+ov];
    }
    return r;
}

private static int getOOffset(final QVMData odata,final int mom_type)
{
    if(mom_type==QVMData.MOM_dV) return (1*odata.ntm+0)*odata.npm;
    else if(mom_type==QVMData.MOM_dT) return (0*odata.ntm+1)*odata.npm;
    throw new IllegalArgumentException("mom_type="+mom_type);
}

private static double [] getORTBasis(final double [] mom,
				     final int mom_dim,
				     final int neq)
{
    if(mom_dim<2*neq+1||neq<1)
    {
	throw new IllegalArgumentException();
    }

    final double pt0=mom[0];
    final double [] basis=new double [neq*neq];
    for(int i=0,oi=0;i<neq;i++,oi+=neq)
    {
	basis[oi+i]=1;
	final double wi=mom[i+1]/pt0;
	for(int k=0,ok=0;k<i;k++,ok+=neq)
	{
	    double scal=0;
	    for(int j=0,pi=i+2;j<=k;j++,pi++)
	    {
		scal+=((mom[pi]/pt0)-wi*(mom[j+1]/pt0))*basis[ok+j];
	    }
	    for(int j=0;j<=k;j++)
	    {
		basis[oi+j]-=scal*basis[ok+j];
	    }
	}

	normalizebasisvector(basis,oi,i+1,mom,0,pt0);
    }
    return basis;
}


private static void converttopolynomial(final double [] res,final double [] basis,final int basis_offset,final int nels,final double [] mom,final int mom_offset,final double pt0)
{
    double s=0;
    for(int k=1,o=basis_offset;k<=nels;++k,++o)
    {
	s-=(res[k]=basis[o])*(mom[k+mom_offset]/pt0);
    }
    res[0]=s;
}


private static void normalizebasisvector(final double [] basis,final int basis_offset,final int nels,final double [] mom,final int mom_offset,final double pt0)
{
    // normalize
    double anorm=0;
    for(int nels1=nels-1,mp=2*nels1,mo1=1+mom_offset,p=0;p<=mp;p++)
    {
	final double momp=mom[p+1+mo1]/pt0;
	double s=0;
	if(p<nels)
	{
	    for(int j=0,pmj=p;pmj>=0;++j,--pmj)
	    {
		s+=basis[basis_offset+j]*basis[basis_offset+pmj]*(momp-(mom[j+mo1]/pt0)*(mom[pmj+mo1]/pt0));
	    }
	}
	else
	{
	    for(int j=p-nels1,pmj=nels1;j<nels;++j,--pmj)
	    {
		s+=basis[basis_offset+j]*basis[basis_offset+pmj]*(momp-(mom[j+mo1]/pt0)*(mom[pmj+mo1]/pt0));
	    }
	}
	anorm+=s;
    }
    final double nr=Math.sqrt(anorm);
    for(int j=nels,o=basis_offset;--j>=0;basis[o++]/=nr);
}



static double [][] getFLast(final double [] mom,
				    final int mom_dim,
				    final int neq)
{
    return getFLast(mom,0,mom_dim,neq,null,-1);
}

/** Returns: orthogonal polynomial, 1/Christoffel function, volatility
 * and result corresponding mom_extra.
 */
/*private*/ static double [][] getFLast(final double [] mom,
				    final int mom_offset,
				    final int mom_dim,
				    final int neq,
				    final double [] mom_extra,
				    final int mom_extra_offset)
{
    if(neq==0)
    {
	return new double [][]{
	    new double []{1},
	    new double []{},
	    new double []{0},
	    mom_extra==null?null:new double []{}
	};
    }
    if(mom_dim<2*neq+1||neq<1)
    {
	throw new IllegalArgumentException("mom_dim="+mom_dim+" neq="+neq);
    }

    final int np=neq-1;
    final double [] norms=new double [np];
    final double pt0=mom[0+mom_offset],pa=mom[1+mom_offset]/pt0,pneq=mom[neq+mom_offset]/pt0;
    if(true)
    { // diagonal moments norm; with this norm diagonal elements are maximal in row.
	for(int k=1,kk=k+mom_offset;k<neq;++k,++kk)
	{
	    norms[k-1]=Math.sqrt((mom[k+kk]/pt0)-Math.pow(mom[kk]/pt0,2));
	    //System.err.println("norms["+(k-1)+"]="+norms[k-1]+" np="+np);
	}
    }
    else
    { // factorial approximation norm
	final double stdevk=Math.sqrt((mom[2+mom_offset]/pt0)-(pa*pa)); double w=1;
	for(int k=1;k<neq;k++)
	{
	    norms[k-1]=(w*=(stdevk*k));
	}
    }
    final double [] matr=new double [np*np],rpart=new double [neq];
    for(int k=0,j=0;k<np;k++)
    {
	final int kk=k+1+mom_offset;
	final double pk=mom[kk]/pt0,normsk=norms[k];
	for(int l=0;l<np;l++,j++)
	{
	    final int ll=l+1;
	    matr[j]=((mom[ll+kk]/pt0)-pk*(mom[ll+mom_offset]/pt0))/normsk/norms[l];
	    //System.err.println("matr["+k+"]["+l+"]="+matr[j]+" np="+np);
	}
	rpart[k]=(pk*pneq-(mom[kk+neq]/pt0))/normsk;
    }
    final double [] pcoefs=new double [neq+1],wcoefs=new double [2*np+1],dpcoefs=new double [2*np+3];
    final int [] ipvt=new int [np];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,np,np,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(pcoefs,Double.NaN);
	java.util.Arrays.fill(wcoefs,Double.NaN);
	java.util.Arrays.fill(dpcoefs,Double.NaN);
	return new double [][]{pcoefs,wcoefs,dpcoefs,mom_extra==null?null:new double [neq]};
    }
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,np,np,ipvt,rpart,1);
    for(int k=np;--k>=0;rpart[k]/=norms[k]);
    rpart[np]=1;
    normalizebasisvector(rpart,0,neq,mom,mom_offset,pt0);
    converttopolynomial(pcoefs,rpart,0,neq,mom,mom_offset,pt0);

    final double [] pcoefs_mom_extra;
    if(mom_extra!=null)
    {
	pcoefs_mom_extra=new double [neq];
	final double me0=mom_extra[mom_extra_offset]/pt0;
	for(int k=np;--k>=0;)
	{
	    final int kk=k+1;
	    rpart[k]=(mom_extra[mom_extra_offset+kk]/pt0-me0*(mom[kk+mom_offset]/pt0))/norms[k];
	}
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,np,np,ipvt,rpart,1);
	for(int k=np;--k>=0;rpart[k]/=norms[k]);
	converttopolynomial(pcoefs_mom_extra,rpart,0,np,mom,mom_offset,pt0);
	pcoefs_mom_extra[0]+=me0; // missed 0-th element
    }
    else
    {
	pcoefs_mom_extra=null;
    }

    java.util.Arrays.fill(rpart,0,np,0);
    for(int k=0;k<np;k++)
    {
	final int kk=k+1,kk1=kk+1;
	final double pk=mom[kk+mom_offset]/pt0,pkk1=((mom[kk1+mom_offset]/pt0)/kk1)-(pk*pa),normsk=norms[k];
	rpart[k]=1;
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,np,np,ipvt,rpart,1);
	for(int l=0;l<np;l++)
	{
	    final int ll=l+1,ll1=ll+1;
	    final double c=rpart[l]/norms[l]/normsk,pl=mom[ll+mom_offset]/pt0,cpl=c*pl,
		         pll1=((mom[ll1+mom_offset]/pt0)/ll1)-(pl*pa),cpll1=c*pll1;
	    rpart[l]=0;
	    wcoefs[0]+=cpl*pk;
	    wcoefs[kk]-=cpl;
	    wcoefs[ll]-=pk*c;
	    wcoefs[kk+ll]+=c;
	    dpcoefs[0]+=cpll1*pkk1;
	    dpcoefs[1]+=c*(pl*pkk1+pk*pll1);
	    dpcoefs[2]+=cpl*pk;
	    dpcoefs[kk1]-=cpll1/kk1;
	    dpcoefs[kk1+1]-=cpl/kk1;
	    dpcoefs[ll1]-=(c*pkk1)/ll1;
	    dpcoefs[ll1+1]-=(c*pk)/ll1;
	    dpcoefs[kk1+ll1]+=c/(ll1*kk1);
	}
    }
    wcoefs[0]+=1; // to match Christoffel function exactly
    dpcoefs[0]+=pa*pa;
    dpcoefs[1]-=2*pa;
    dpcoefs[2]+=1;

    return new double [][]{pcoefs,wcoefs,dpcoefs,pcoefs_mom_extra};
}

private static double [] getORTProjected(final double [] mom,
					 final int mom_dim,
					 final int neq,
					 final double dpr,
					 final int res)
{

    if(res==RES_FDIRECT)
    {
	return getFLast(mom,mom_dim,neq)[0];
    }

    final double [] basis=getORTBasis(mom,mom_dim,neq);

    final double [] pproj;
    if(res==RES_SUM || res==RES_MAX){pproj=new double [2*neq+1];}
    else if(res==RES_MAXF||res==RES_FLAST){pproj=new double [neq+1];}
    else{throw new IllegalArgumentException("res="+res);}

    final double pt0=mom[0];
    if(res==RES_FLAST)
    {
	converttopolynomial(pproj,basis,neq*(neq-1),neq,mom,0,pt0);
	return pproj;
    }

    double maxdk=Double.NEGATIVE_INFINITY;
    int nelpmax=-1;
    for(int ev=0,oev=0;ev<neq;ev++,oev+=neq)
    {
	final int npoly=2+ev;
	final double [] poly=new double [npoly];
	converttopolynomial(poly,basis,oev,ev+1,mom,0,pt0);

	if(res==RES_SUM)
	{
	    DeprecatedHelperTMatr.sumP2(pproj,poly);
	}
	else
	{
	    if(npoly!=poly.length) throw new RuntimeException();
	    final double dk=Math.abs(Polynomials.getD0(poly,dpr));
	    if(dk>=maxdk)
	    {
		maxdk=dk;
		if(res==RES_MAXF)
		{
		    System.arraycopy(poly,0,pproj,0,npoly);
		    nelpmax=npoly;
		}
		else if(res==RES_MAX)
		{
		    java.util.Arrays.fill(pproj,0);
		    DeprecatedHelperTMatr.sumP2(pproj,poly);
		    nelpmax=npoly*2+1;
		}
		else
		{
		    throw new RuntimeException();
		}
	    }
	}
    }
    if(nelpmax>=0 && nelpmax<pproj.length)
    {
	return java.util.Arrays.copyOf(pproj,nelpmax);
    }
    return pproj;
}




private static void checkMatchingAverageSum(final int neq_p,
					    final java.util.Random r,
					    final QVMData odata,
					    final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final double [][][] abc3=QVMData.get3ArrayFromArrayForTesting(abc,odata.nvm,odata.ntm,odata.npm);
    final double pt0=abc3[0][0][0];

    final int [] mom_dims=new int []{odata.nvm,odata.ntm,odata.npm},
	         neq=new int []{0,0,neq_p};
    final double [] dproj=new double[]{Double.NaN,Double.NaN,Double.NaN};

    final double [] p0qvmaverage=QVMAverage.getAverageProjected(abc,mom_dims,neq,2,dproj,0);
    final double [] p0ort=getORTProjected(abc,odata.npm,neq_p,Double.NaN,RES_SUM);
    UnitTests.checkDoubleArrayRel(p0qvmaverage,"p0qvmaverage",p0ort,"p0ort",eps);
    System.err.println("checkMatchingAverageSum:[0,0] moment matched OK for neq_p="+neq_p);


    UnitTests.checkDoubleArrayRel(getORTProjected(abc,odata.npm,neq_p,Double.NaN,RES_FLAST),"RES_FLAST",
					 getFLast(abc,odata.npm,neq_p)[0],"RES_F_DIRECT",eps);

    for(int res:new int []{RES_SUM,RES_MAX,RES_MAXF,RES_FLAST,RES_FDIRECT})
    {
	final double [] p0,p0r=getORTProjected(abc,odata.npm,neq_p,CalculateKendall.TradeEMoments.getRandomPrice(r)-odata.p_offset,res);
	if(res==RES_SUM||res==RES_MAX)
	{
	    p0=p0r;
	}
	else
	{
	    UnitTests.checkDoubleRel(getAverage(p0r,abc3[0][0],pt0),"s(0)",0,"0",eps);
	    p0=new double [2*p0r.length-1];
	    DeprecatedHelperTMatr.sumP2(p0,p0r);
	}
	final int nr=res==RES_SUM?neq_p:1;
	UnitTests.checkDoubleRel(getAverage(p0,abc3[0][0],pt0),"Is(0)",nr,"nr",eps);
	System.err.println("checkMatchingAverageSum:norm OK for res="+res+" neq_p="+neq_p);
    }

    System.err.println("QVMORT:checkMatchingAverageSum: tests OK  neq_p="+neq_p);
}


private static void checkMatchingInvariance(final int neq_p,
					    final java.util.Random r,
					    final QVMData odata,
					    final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final double p_offset2=odata.p_offset-2*(0.5-r.nextDouble());
    final double [] abc_shifted=odata.shiftOffset(r.nextDouble(),r.nextDouble(),odata.p_offset-p_offset2,QVMData.MOM_dT);

    for(int res:new int []{RES_SUM,RES_MAX,RES_MAXF,RES_FLAST,RES_FDIRECT})
    {
	final double ps=CalculateKendall.TradeEMoments.getRandomPrice(r);
	final double [] p0=getORTProjected(abc,odata.npm,neq_p,ps-odata.p_offset,res);
	final double [] p0_shifted=getORTProjected(abc_shifted,odata.npm,neq_p,ps-p_offset2,res);
	// match shifted
	UnitTests.checkDoubleArrayRel1(p0,"p0",Polynomials.getPolynomialShifted(p0_shifted,odata.p_offset-p_offset2),"p0 after shift",eps);
	System.err.println("checkMatchingInvariance:invariance OK for res="+res+" neq_p="+neq_p);
    }

    // check QVM.getPDPCoefs invariance
    final double [][] x=QVM.getPDPCoefs(abc,neq_p),xshifted=QVM.getPDPCoefs(abc_shifted,neq_p);
    UnitTests.checkDoubleArrayRel1(x[0],"K",Polynomials.getPolynomialShifted(xshifted[0],odata.p_offset-p_offset2),"K after shift",eps);
    UnitTests.checkDoubleArrayRel1(x[1],"IK",Polynomials.getPolynomialShifted(xshifted[1],odata.p_offset-p_offset2),"IK after shift",eps);
    System.err.println("checkMatchingInvariance:invariance OK for QVM.getPDPCoefs neq_p="+neq_p);


    // getFLast extra check
    final double [][] k0=getFLast(abc,odata.npm,neq_p),k0_shifted=getFLast(abc_shifted,odata.npm,neq_p);
    UnitTests.checkDoubleArrayRel1(k0[0],"k0[0]",Polynomials.getPolynomialShifted(k0_shifted[0],odata.p_offset-p_offset2),"k0[0] after shift",eps);
    System.err.println("checkMatchingInvariance:invariance OK for pcoefs neq_p="+neq_p);
    UnitTests.checkDoubleArrayRel1(k0[1],"k0[1]",Polynomials.getPolynomialShifted(k0_shifted[1],odata.p_offset-p_offset2),"k0[1] after shift",eps);
    System.err.println("checkMatchingInvariance:invariance OK for wcoefs neq_p="+neq_p);
    UnitTests.checkDoubleArrayRel1(k0[2],"k0[2]",Polynomials.getPolynomialShifted(k0_shifted[2],odata.p_offset-p_offset2),"k0[2] after shift",eps);
    System.err.println("checkMatchingInvariance:invariance OK for dpcoefs neq_p="+neq_p);

    // compare QVM.getPDPCoefs & getFLast
    final double [][] pdp=QVM.getPDPCoefs(abc,neq_p-1),pdp_shifted=QVM.getPDPCoefs(abc_shifted,neq_p-1);
    UnitTests.checkDoubleArrayRel(k0[1],"k0[1]",pdp[0],"pdp[0]",eps);
    UnitTests.checkDoubleArrayRel(k0[2],"k0[2]",pdp[1],"pdp[1]",eps);
    UnitTests.checkDoubleArrayRel(k0_shifted[1],"k0_shifted[1]",pdp_shifted[0],"pdp_shifted[0]",eps);
    UnitTests.checkDoubleArrayRel(k0_shifted[2],"k0_shifted[2]",pdp_shifted[1],"pdp_shifted[1]",eps);

    UnitTests.checkDoubleArrayRel1(k0[1],"k0[1]",Polynomials.getPolynomialShifted(pdp_shifted[0],odata.p_offset-p_offset2),"pdp_shifted[0]",eps);
    UnitTests.checkDoubleArrayRel1(k0[2],"k0[2]",Polynomials.getPolynomialShifted(pdp_shifted[1],odata.p_offset-p_offset2),"pdp_shifted[0]",eps);

    System.err.println("checkMatchingInvariance:wcoefs & dpcoefs matched OK to QVM.getPDPCoefs neq_p="+neq_p);


    System.err.println("QVMORT:checkMatchingInvariance: tests OK  neq_p="+neq_p);
}


private static void checkMatchingFLast(final int neq_p,
				       final java.util.Random r,
				       final QVMData odata,
				       final double eps,
				       final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{
  for(double [] abc:new double [][]{odata.tmoments_vol_x_tau_x_p,odata.vmoments_vol_x_tau_x_p})
  {
    final double [] p0qvmaverage=QVMAverage.getAverageProjected(abc,new int []{odata.nvm,odata.ntm,odata.npm},new int []{0,0,neq_p},2,new double[]{Double.NaN,Double.NaN,Double.NaN},0);
    p0qvmaverage[0]+=1; // to match getPDPCoefs definition


    // check wcoefs & dpcoefs with one extra element in basis
    final double [] wcoefs=new double [2*neq_p+1],dpcoefs=new double [2*neq_p+3];
    for(int n=0;n<=neq_p;n++)
    {
	final double [][] fl=getFLast(abc,odata.npm,n);
	final double [] f=fl[0];
	if(n==neq_p)
	{
	    final double [] wcoefs_f=java.util.Arrays.copyOf(wcoefs,2*(neq_p-1)+1),
		dpcoefs_f=java.util.Arrays.copyOf(dpcoefs,2*(neq_p-1)+3);
	    UnitTests.checkDoubleArrayRel(wcoefs_f,"wcoefs_f",fl[1],"fl[1]",eps);
	    UnitTests.checkDoubleArrayRel(dpcoefs_f,"dpcoefs_f",fl[2],"fl[2]",eps);
	    System.err.println("checkMatchingFLast:getFLast sum rule OK for wcoefs & dpcoefs for neq_p="+neq_p);
	    final double [][] pdp=QVM.getPDPCoefs(abc,neq_p-1);
	    UnitTests.checkDoubleArrayRel(wcoefs_f,"wcoefs_f",pdp[0],"pdp[0]",eps);
	    UnitTests.checkDoubleArrayRel(dpcoefs_f,"dpcoefs_f",pdp[1],"pdp[1]",eps);
	    UnitTests.checkDoubleArrayRel(fl[1],"fl[1]",pdp[0],"pdp[0]",eps);
	    UnitTests.checkDoubleArrayRel(fl[2],"fl[2]",pdp[1],"pdp[1]",eps);
	    System.err.println("checkMatchingFLast:getFLast integrated & QVM.getPDPCoefs matched OK for neq_p="+neq_p);
	}
	DeprecatedHelperTMatr.sumP2(wcoefs,f);
	final double [] iint=getPolynomialAIntegrated(f,abc,abc[0]);
	DeprecatedHelperTMatr.sumP2(dpcoefs,iint);
	// check averages
	UnitTests.checkDoubleRel(getAverage(wcoefs,2*n+1,abc,abc[0]),"norm",n+1,"norm",eps);
	UnitTests.checkDoubleRel(getAverage(iint,abc,abc[0]),"<iint>",0,"0",eps);
	UnitTests.checkDoubleRel(getAverage(f,abc,abc[0]),"<f>",n>0?0:1,n>0?"0":"1",eps);
	UnitTests.checkDoubleArrayRel(f,"f",Polynomials.getPolynomialDifferentiated(iint),"d(iint)/dp",eps);
    }
    // match getAverageProjected
    UnitTests.checkDoubleArrayRel(p0qvmaverage,"p0qvmaverage",wcoefs,"wcoefs",eps);
    System.err.println("checkMatchingFLast:getAverageProjected &  wcoefs matched OK for neq_p="+neq_p);

    // match QVM.getPDPCoefs
    final double [][] pdp0=QVM.getPDPCoefs(abc,neq_p);
    UnitTests.checkDoubleArrayRel(wcoefs,"wcoefs",pdp0[0],"pdp0[0]",eps);
    UnitTests.checkDoubleArrayRel(dpcoefs,"dpcoefs",pdp0[1],"pdp0[1]",eps);
    System.err.println("checkMatchingFLast:icoefs from QVM.getPDPCoefs &  getFLast  matched OK for neq_p="+neq_p);

    // check scale
    final double [] abc_scaled=new double [abc.length];
    TMatr.sumP1(abc_scaled,abc,2+3*r.nextDouble());
    final double [][] x0=getFLast(abc,odata.npm,neq_p),xs0=getFLast(abc_scaled,odata.npm,neq_p);
    UnitTests.checkDoubleArrayRel(x0[0],"pcoefs",xs0[0],"pcoefs scaled",eps);
    UnitTests.checkDoubleArrayRel(x0[1],"wcoefs",xs0[1],"wcoefs scaled",eps);
    UnitTests.checkDoubleArrayRel(x0[2],"dpcoefs",xs0[2],"dpcoefs scaled",eps);
    System.err.println("checkMatchingFLast:scale matched OK for neq_p="+neq_p);
    final int o=10+r.nextInt(100);
    final double [] abc_o=new double [o+odata.npm];
    System.arraycopy(abc,0,abc_o,o,odata.npm);
    final double [][] x0o=getFLast(abc_o,o,odata.npm,neq_p,abc,neq_p);
    UnitTests.checkDoubleArrayRel(x0[0],"x0[0]",x0o[0],"x0o[0]",eps*0);
    UnitTests.checkDoubleArrayRel(x0[1],"x0[1]",x0o[1],"x0o[1]",eps*0);
    UnitTests.checkDoubleArrayRel(x0[2],"x0[2]",x0o[2],"x0o[2]",eps*0);
    final double [] xextra=x0o[3],xw=java.util.Arrays.copyOf(x0[0],neq_p);
    //System.err.println("neq_p="+neq_p+" abc="+java.util.Arrays.toString(abc));
    for(int k=0;k<neq_p;k++) {xw[k]/=-x0[0][neq_p];} // get monic polynomial with -1 coef
    UnitTests.checkDoubleArrayRel(xextra,"xextra",xw,"xw",eps);
    UnitTests.checkDoubleArrayRel(DeprecatedHelperEVXData.getLinsysSol(neq_p,abc_o,o,abc,neq_p),"LS",xextra,"xextra",eps);
    System.err.println("checkMatchingFLast:mom_extra matched OK for neq_p="+neq_p);


    final QVMORT qvmort=new QVMORT(rf,neq_p,odata,abc,-1);
    final double [] icoefs=new double [qvmort.p_imin.length],
	dpf=new double [qvmort.p_imin.length],
	dwp=new double [qvmort.p_imin.length];
    for(int i=0;i<icoefs.length;i++)
    {
	final double pd=qvmort.p_imin[i]-qvmort.odata.p_offset;
	icoefs[i]=1/Polynomials.getD0(wcoefs,pd);
	dpf[i]=Polynomials.getD1(dpcoefs,pd);
	dwp[i]=Polynomials.getD1(wcoefs,pd);
    }
    UnitTests.checkDoubleArrayRel(qvmort.icoefs,"qvmort.icoefs",icoefs,"icoefs",eps);
    //UnitTests.checkDoubleArrayRel(qvmort.dpf,"qvmort.dpf",dpf,"dpf",eps);
    //UnitTests.checkDoubleArrayRel(qvmort.dwp,"qvmort.dwp",dwp,"dwp",eps);
    System.err.println("checkMatchingFLast:icoefs & dpf & dwp matched to basis expansion for neq_p="+neq_p);

    double nsum=0,dpf_a=0,dwp_a=0;
    for(int i=0;i<qvmort.icoefs.length;i++)
    {
	nsum+=qvmort.icoefs[i];
	dpf_a+=qvmort.icoefs[i]*dpf[i];
	dwp_a+=qvmort.icoefs[i]*dwp[i];
	System.err.println("qvmort.p_imin["+i+"]="+qvmort.p_imin[i]+
			   "\tqvmort.icoefs["+i+"]="+qvmort.icoefs[i]+
			   "\tdwp["+i+"]="+dwp[i]+
			   "\tdpf["+i+"]="+dpf[i]);
    }
    UnitTests.checkDoubleRel(nsum,"nsum",1,"1",eps);
    UnitTests.checkDoubleRel(dwp_a,"dwp_a",0,"0",eps);
    UnitTests.checkDoubleRel(getAverage(Polynomials.getPolynomialDifferentiated(getFLast(abc,odata.npm,neq_p)[2]),abc,abc[0]),"<dpf>",dpf_a,"<dpf>",eps);
    UnitTests.checkDoubleRel(getAverage(Polynomials.getPolynomialDifferentiated(getFLast(abc,odata.npm,neq_p)[1]),abc,abc[0]),"<dwp>",0,"<dwp>=0",eps);
    System.err.println("checkMatchingFLast:icoefs: norm nsum="+nsum+" dpf_a="+dpf_a+" dwp_a="+dwp_a+" matched OK for neq_p="+neq_p);

    // test integral calculation
    final int n=2*neq_p-1;
    final double [] tpol=new double [n+1];
    for(int i=n;i>=0;i--) tpol[i]=2*(0.5-r.nextDouble());
    double si=0;
    for(int i=0;i<qvmort.icoefs.length;i++)
    {
	si+=qvmort.icoefs[i]*Polynomials.getD0(tpol,qvmort.p_imin[i]-qvmort.odata.p_offset);
    }
    final double msi=getAverage(tpol,abc,abc[0]);
    UnitTests.checkDoubleRel(si,"si",msi,"msi",eps);
    System.err.println("checkMatchingFLast:integral OK: si="+si+" msi="+msi+" for n="+n+" neq_p="+neq_p);

  }
  System.err.println("QVMORT:checkMatchingFLast: tests OK  neq_p="+neq_p);
}


private static void checkMatchingFLastExtra(final int neq_p,
					    final java.util.Random r,
					    final QVMData odata,
					    final double eps,
					    final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{
    if(odata.nvm<1 || odata.ntm<2) throw new IllegalArgumentException("odata.nvm="+odata.nvm+" odata.ntm="+odata.ntm+" odata.npm="+odata.npm);
  for(int mom_type:new int []{QVMData.MOM_dT,QVMData.MOM_dV})
  for(double [] abc:new double [][]{odata.tmoments_vol_x_tau_x_p,odata.vmoments_vol_x_tau_x_p})
  {
      final double [][][] abc3=QVMData.get3ArrayFromArrayForTesting(abc,odata.nvm,odata.ntm,odata.npm);
      final double pt0=abc3[0][0][0];
      final double [] tinterp=new double [neq_p];
      final int iv=mom_type==QVMData.MOM_dV?1:0,it=1-iv;
      for(int k=0;k<neq_p;k++)
      {
	  final double [] pk=getFLast(abc3[0][0],odata.npm,k)[0];
	  TMatr.sumP1(tinterp,pk,getAverage(pk,abc3[iv][it],pt0));
      }
      final double [] tinterp_direct=getFLast(abc,0,odata.npm,neq_p,abc,getOOffset(odata,mom_type))[3];
      UnitTests.checkDoubleArrayRel(tinterp,"tinterp (pk)",tinterp_direct,"tinterp_direct",eps);
      UnitTests.checkDoubleArrayRel(DeprecatedHelperEVXData.getLinsysSol(neq_p,abc3[0][0],0,abc3[iv][it],0),"LS",tinterp,"tinterp",eps);

      // linear offset check
      final double dtx=r.nextDouble();
      final double [] pts=abc3[iv][it].clone();
      TMatr.sumP1(pts,abc3[0][0],dtx);
      final double [] tinterp_direct_dtx=getFLast(abc,0,odata.npm,neq_p,pts,0)[3];
      tinterp_direct_dtx[0]-=dtx;
      UnitTests.checkDoubleArrayRel(tinterp,"tinterp",tinterp_direct_dtx,"tinterp_direct_dtx",eps);
      UnitTests.checkDoubleArrayRel(tinterp_direct,"tinterp_direct",tinterp_direct_dtx,"tinterp_direct_dtx",eps);

      // regular moments offset check
      final int offx=7+r.nextInt(100),offex=7+r.nextInt(100);
      final double [] abc_offx=new double [offx+odata.npm],mom_ex=new double [offex+odata.npm];
      for(int k=0;k<odata.npm;k++) {abc_offx[k+offx]=-abc[k];mom_ex[k+offex]=-abc[k+getOOffset(odata,mom_type)];}
      final double [][] xoffs=getFLast(abc_offx,offx,odata.npm,neq_p,mom_ex,offex),
	  x=getFLast(abc,0,odata.npm,neq_p,abc,getOOffset(odata,mom_type));
      for(int e=0;e<x.length;e++)
      {
	  // have to match exactly
	  UnitTests.checkDoubleArrayRel(x[e],"x["+e+"]",xoffs[e],"xoffs[e]",eps*0);
      }

      final QVMORT qvmort=new QVMORT(rf,neq_p,odata,abc,mom_type);
      for(int k=0;k<qvmort.p_imin.length;k++)
      {
	  final double pk=qvmort.p_imin[k]-qvmort.odata.p_offset;
	  final double dist=Polynomials.getD0(tinterp,pk);
	  UnitTests.checkDoubleRel(dist,"dist",qvmort.dtx[k],"qvmort.dtx[k]",eps);
	  System.err.println("checkMatchingFLastExtra: qvmort.p_imin["+k+"]="+qvmort.p_imin[k]+
			     "\tqvmort.icoefs["+k+"]="+qvmort.icoefs[k]+
			     "\tdist="+dist);
      }
  }
  System.err.println("QVMORT:checkMatchingFLastExtra: tests OK  neq_p="+neq_p);
}


    private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_p)
    {
	final double eps=1e-5;
	final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
	final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
	System.err.println();
	System.err.println("------------------------------------------------------------------------");
	System.err.println("QVMORT:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_p="+neq_p);

	final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	    new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();
	final QVMData odata=QVMData.calculateQVMData(1,
						     1,
						     1+2*neq_p,
						     currenttime,
						     tau,
						     tb,
						     tsize,
						     null);

	// average projections
	checkMatchingAverageSum(neq_p,r,odata,eps);
	// invariance
	checkMatchingInvariance(neq_p,r,odata,eps);
	// getFLast
	checkMatchingFLast(neq_p,r,odata,eps,rf);
	// getFLastExtra
	final QVMData odatavt=QVMData.calculateQVMData(2+r.nextInt(4),
						       2+r.nextInt(4),
						       1+2*neq_p+r.nextInt(5),
						       currenttime,
						       tau,
						       tb,
						       tsize,
						       null);
	checkMatchingFLastExtra(neq_p,r,odatavt,eps,rf);

	System.err.println("QVMORT:checkCalculations(): OK with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_p="+neq_p);
    }

    public static void main(String [] args)
    {
	System.err.println("QVMORT::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
	final long t0=System.currentTimeMillis();
	final java.util.Random r=new java.util.Random(-79091);
	for(int neq_p=1;neq_p<=8;neq_p++)
	{
	    for(int i=0;i<118;i++)
	    {
		checkCalculations(100+r.nextInt(100),r,neq_p);
		checkCalculations(10010+r.nextInt(100010),r,neq_p);
	    }
	}
	final long t1=System.currentTimeMillis();
	System.err.println("QVMORT::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    }

} // END of class QVMORT


static double [][] getPDPCoefs(final double [] pt,final int np)
{
    if(np<0||pt.length<2*np+1||pt.length<2) throw new IllegalArgumentException();
    final double pt0=pt[0],pa=pt[1]/pt0;
    if(np==0)
    {
	return new double[][]{
	    new double []{1},
	    new double []{pa*pa,-2*pa,1}
	};
    }

    final double [] norms=new double [np];
    final double stdevk=Math.sqrt((pt[2]/pt0)-(pa*pa));
    double w=1;
    for(int k=1;k<=np;k++)
    {
	norms[k-1]=(w*=(stdevk*k));
    }

    final double [] matr=new double [np*np];
    for(int k=0,j=0;k<np;k++)
    {
	final int k1=k+1;
	final double normsk=norms[k];
	for(int l=0;l<np;l++,j++)
	{
	    final int l1=l+1;
	    matr[j]=((pt[l1+k1]/pt0)-(pt[l1]/pt0)*(pt[k1]/pt0))/normsk/norms[l];
	}
    }
    final int [] ipvt=new int [np];
    final double [] K=new double [2*np+1],IK=new double [2*np+3];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,np,np,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(K,Double.NaN);
	java.util.Arrays.fill(IK,Double.NaN);
	return new double [][]{K,IK};
    }

    final double [] rpart=new double [np];
    for(int k=0;k<np;k++)
    {
	final int kk=k+1,k1=kk+1;
	final double pk=pt[kk]/pt0,pk1=((pt[k1]/pt0)/k1)-(pk*pa),normsk=norms[k];
	rpart[k]=1;
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,np,np,ipvt,rpart,1);
	for(int l=0;l<np;l++)
	{
	    final int ll=l+1,l1=ll+1;
	    final double c=rpart[l]/norms[l]/normsk,
		         pl=pt[ll]/pt0,cpl=c*pl,
		         pl1=((pt[l1]/pt0)/l1)-(pl*pa),cpl1=c*pl1;
	    rpart[l]=0;
	    K[0]+=cpl*pk;
	    K[kk]-=cpl;
	    K[ll]-=c*pk;
	    K[kk+ll]+=c;
	    IK[0]+=cpl1*pk1;
	    IK[1]+=c*(pl*pk1+pk*pl1);
	    IK[2]+=cpl*pk;
	    IK[k1]-=cpl1/k1;
	    IK[k1+1]-=cpl/k1;
	    IK[l1]-=(c*pk1)/l1;
	    IK[l1+1]-=(c*pk)/l1;
	    IK[k1+l1]+=c/(l1*k1);
	}
    }
    K[0]+=1; // to match Christoffel function exactly
    IK[0]+=pa*pa;
    IK[1]-=2*pa;
    IK[2]+=1;

    return new double [][] {
	K,
	IK
    };
} // END OF getPDPCoefs (used for testing)



/** QVMMORT. */
static class QVMMORT {

    final int neq_v,neq_t,neq_p;
    private final QVMData odata;
private QVMMORT(final com.polytechnik.utils.PolynomialRootsFinder rf,
		final int neq_v,final int neq_t,final int neq_p,
		final QVMData odata)
{
    this.neq_v=neq_v;
    this.neq_t=neq_t;
    this.neq_p=neq_p;
    this.odata=odata;
}


private static double [][] getEVORTProjected(final double [] mom,
					     final int [] mom_dims,
					     final int [] neq,
					     final int proj_ptr)
{
    final int nd=neq.length;
    if(proj_ptr<0||proj_ptr>=nd||mom_dims.length!=nd)
    {
	throw new IllegalArgumentException("proj_ptr="+proj_ptr+
					   " or mom_dims.length="+mom_dims.length+
					   " does not correspond to neq.length="+neq.length);
    }

    final double [][] pproj=new double [nd][];
    int mom_nels=1,neqs=0;
    for(int k=nd;--k>=0;)
    {
	final int kdim=mom_dims[k];
	if(kdim<1)
	{
	    throw new IllegalArgumentException("mom_dims["+k+"]="+mom_dims[k]);
	}
	mom_nels*=kdim;
	final int kneq=neq[k];
	if(kneq<0 || (1+2*kneq)>kdim)
	{
	    throw new IllegalArgumentException("neq["+k+"]="+neq[k]+
					       " does not correspond to"+
					       " mom_dims["+k+"]="+mom_dims[k]);
	}
	pproj[k]=new double [1+kneq];
	neqs+=kneq;
    }
    if(mom.length!=mom_nels)
    {
	throw new IllegalArgumentException("mom.length="+mom.length+
					   " does not correspond the size calculated from mom_dims;"+
					   " size="+mom_nels);
    }
    final int momoffs=mom_nels,neq_proj=neq[proj_ptr],ne=neqs-1;
    if(neq_proj<1)
    {
	throw new IllegalArgumentException("neq["+proj_ptr+"]="+neq_proj);
    }

    final double pt0=mom[0];
    final double [] norms=new double [ne];

    neq[proj_ptr]=neq_proj-1; // clobbered
    int offs_proj=-1;
    for(int ek=0,offsk=momoffs,ik=0;ek<nd;ek++)
    {
	offsk/=mom_dims[ek];
	if(ek==proj_ptr){offs_proj=offsk;}
	final int nk=neq[ek];
	final double stdevk=nk<=0?Double.NaN:Math.sqrt(mom[2*offsk]/pt0);
	double w=1;
	for(int k=1;k<=nk;k++,ik++)
	{
	    norms[ik]=(w*=(stdevk*k));
	}
    }

    final double [] matr=new double [ne*ne],rpart=new double [neqs];
    final double pneq=mom[offs_proj*=neq_proj]/pt0;
    for(int ek=0,offsk=momoffs,ik=0,j=0;ek<nd;ek++)
    {
	for(int k=1,nk=neq[ek],koffsk=(offsk/=mom_dims[ek]);k<=nk;k++,koffsk+=offsk,ik++)
	{
	    final double normsk=norms[ik],pk=mom[koffsk]/pt0;
	    for(int el=0,offsl=momoffs,il=0;el<nd;el++)
	    {
		for(int l=1,nl=neq[el],loffsl=(offsl/=mom_dims[el]);l<=nl;l++,loffsl+=offsl,il++,j++)
		{
		    final double pl=mom[loffsl]/pt0,pkl=mom[koffsk+loffsl]/pt0;
		    matr[j]=(pkl-pl*pk)/normsk/norms[il];
		}
	    }
	    rpart[ik]=(pk*pneq-(mom[koffsk+offs_proj]/pt0))/normsk;
	}
    }
    neq[proj_ptr]=neq_proj; // restore clobbered value

    final int [] ipvt=new int [ne];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,ne,ne,ipvt);
    if(pinfo!=0)
    {
	for(double [] p:pproj)
	{
	    java.util.Arrays.fill(p,Double.NaN);
	}
	return pproj;
    }
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,ne,ne,ipvt,rpart,1);

    // convert basis
    final double [] basis=rpart; // re-use array
    for(int ek=nd,ik=ne-1,ikn=neqs;--ek>=0;)
    {
	final int nk=neq[ek];
	if(ek!=proj_ptr) {
	    for(int k=nk;k>=1;--k,--ik)
	    {
		basis[--ikn]=rpart[ik]/norms[ik];
	    }
	} else {
	    basis[--ikn]=1;
	    for(int k=nk;--k>=1;--ik)
	    {
		basis[--ikn]=rpart[ik]/norms[ik];
	    }
	}
    }
    // normalize
    double anorm=0;
    for(int ek=0,offsk=momoffs,ik=0;ek<nd;ek++)
    {
	for(int k=1,nk=neq[ek],koffsk=(offsk/=mom_dims[ek]);k<=nk;k++,koffsk+=offsk,ik++)
	{
	    final double pk=mom[koffsk]/pt0;
	    double s=0;
	    for(int el=0,offsl=momoffs,il=0;el<nd;el++)
	    {
		for(int l=1,nl=neq[el],loffsl=(offsl/=mom_dims[el]);l<=nl;l++,loffsl+=offsl,il++)
		{
		    s+=basis[il]*((mom[koffsk+loffsl]/pt0)-pk*(mom[loffsl]/pt0));
		}
	    }
	    anorm+=s*basis[ik];
	}
    }
    final double nr=Math.sqrt(anorm);

    // convert to 1D polynomials by projection
    for(int ek=0,offsk=momoffs,ik=0;ek<nd;ek++)
    {
	final double [] p=pproj[ek];
	for(int k=1,nk=neq[ek],koffsk=(offsk/=mom_dims[ek]);k<=nk;k++,koffsk+=offsk,ik++)
	{
	    final double c=basis[ik]/nr;
	    p[k]=c;
	    p[0]-=(mom[koffsk]/pt0)*c;
	}
    }

    return pproj;
}


private static void checkMatchingQVMORT(final java.util.Random r,
					final QVMData odata,
					final int neq_p,
					final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final double [] flast=QVMORT.getFLast(abc,odata.npm,neq_p)[0];
    final double [] projQVMMORT=getEVORTProjected(abc,
						  new int []{odata.nvm,odata.ntm,odata.npm},
						  new int []{0,0,neq_p},
						  2)[2];
    UnitTests.checkDoubleArrayRel(flast,"flast",projQVMMORT,"projQVMMORT",eps);
    System.err.println("QVMMORT:checkMatchingQVMORT: tests OK  neq_p="+neq_p);
}


private static void checkMatchingAverage(final int neq_v,final int neq_t,final int neq_p,final java.util.Random r,final QVMData odata,final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final int [] mom_dims=new int []{odata.nvm,odata.ntm,odata.npm},
	neq=new int []{neq_v,neq_t,neq_p},neqf=new int [3];
    final double [] dproj=new double [3];
    for(int iv=0;iv<9;iv++)
    {
	double dv=r.nextDouble();
	dproj[0]=dv;
	for(int it=0;it<10;it++)
	{
	    double dt=r.nextDouble();
	    dproj[1]=dt;

	    final double [] projAverage=QVMAverage.getAverageProjected(abc,
								       mom_dims,
								       neq,
								       2,
								       dproj,
								       0);
	    final double [][][] fls=new double [neq_v+neq_t+neq_p][][];
	    for(int k=1;k<=fls.length;k++)
	    {
		int proj_ptr;
		if(k<=neq_t)
		{
		    neqf[0]=0;
		    neqf[1]=k;
		    neqf[2]=0;
		    proj_ptr=1;
		}
		else if(k<=neq_v+neq_t)
		{
		    neqf[0]=k-neq_t;
		    neqf[1]=neq_t;
		    neqf[2]=0;
		    proj_ptr=0;
		}
		else
		{
		    neqf[0]=neq_v;
		    neqf[1]=neq_t;
		    neqf[2]=k-(neq_v+neq_t);
		    proj_ptr=2;
		}
		fls[k-1]=getEVORTProjected(abc,
					   mom_dims,
					   neqf,
					   proj_ptr);
	    }

	    for(int ip=0;ip<11;ip++)
	    {
		double dp=CalculateKendall.TradeEMoments.getRandomPrice(r)-odata.p_offset;
		final double ia=Polynomials.getD0(projAverage,dp);
		double ifl=0;
		for(int k=0;k<fls.length;k++)
		{
		    final double [][] p=fls[k];
		    double s=Polynomials.getD0(p[0],dv)
			    +Polynomials.getD0(p[1],dt)
		            +Polynomials.getD0(p[2],dp);
		    ifl+=s*s;
		}
		UnitTests.checkDoubleRel(ia,"ia",ifl,"ifl",eps);
	    }
	}
    }

    System.err.println("checkMatchingAverage: tests OK neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
}


private static void checkMatchingInvariance(final int neq_v,final int neq_t,final int neq_p,
					    final java.util.Random r,
					    final QVMData odata,
					    final double eps)
{
    final double [] abc=odata.tmoments_vol_x_tau_x_p;
    final double p_offset2=odata.p_offset-0.3*2*(0.5-r.nextDouble());
    final double dv_shift=0.3*r.nextDouble(),dt_shift=0.3*r.nextDouble();
    final double [] abc_shifted=odata.shiftOffset(dv_shift,dt_shift,odata.p_offset-p_offset2,QVMData.MOM_dT);
    final int [] mom_dims=new int []{odata.nvm,odata.ntm,odata.npm},
	neq=new int []{neq_v,neq_t,neq_p};
    final double [][] p0=getEVORTProjected(abc,mom_dims,neq,2);
    final double [][] p0_shifted=getEVORTProjected(abc_shifted,mom_dims,neq,2);
    for(int k=0;k<p0.length;k++)
    {
	final double d=k==0?dv_shift:(k==1?dt_shift:odata.p_offset-p_offset2);
	// shift polynomial back to match directly
	UnitTests.checkDoubleArrayRel1(
	    p0[k],"p0[k]",Polynomials.getPolynomialShifted(p0_shifted[k],d),"p0[k] shifted",eps);
    }

    System.err.println("QVMMORT:checkMatchingInvariance: tests OK neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
}

private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_v,final int neq_t,final int neq_p)
{
    final double eps=1e-5;
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMMORT:checkCalculations(): test with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);

    final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();

    final QVMData odata=QVMData.calculateQVMData(1+2*neq_v,
						 1+2*neq_t,
						 1+2*neq_p,
						 currenttime,
						 tau,
						 tb,
						 tsize,
						 null);
    // regular check
    checkMatchingQVMORT(r,odata,neq_p,eps);

    // average projections
    checkMatchingAverage(neq_v,neq_t,neq_p,r,odata,eps);

    // invariance
    checkMatchingInvariance(neq_v,neq_t,neq_p,r,odata,eps);


    System.err.println("QVMMORT:checkCalculations(): OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " neq_v="+neq_v+" neq_t="+neq_t+" neq_p="+neq_p);
}


public static void main(String [] args)
{
    System.err.println("QVMMORT::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-79091);
    for(int neq_v=0;neq_v<=5;neq_v++)
    {
	for(int neq_t=0;neq_t<=5;neq_t++)
	{
	    for(int neq_p=1;neq_p<=5;neq_p++)
	    {
		for(int i=0;i<4;i++)
		{
		    checkCalculations(100+r.nextInt(100),r,neq_v,neq_t,neq_p);
		    checkCalculations(10010+r.nextInt(100010),r,neq_v,neq_t,neq_p);
		}
	    }
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println("QVMMORT::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
}

} // END OF class QVMMORT


static int [] getPRangeIndex(final double p,final double [] points)
{
    int ind_low=-1,ind_high=-1;
    for(int ne=points.length,i=0;i<=ne;i++)
    {
	if((i<=0||points[i-1]<=p) && (i>=ne||p<points[i]))
	{
	    ind_low=i<=0?-1:i-1;
	    ind_high=i>=ne?-1:i;
	    break;
	}
    }
    return new int []{ind_low,ind_high};
}





/** QVMMU version. */
static class QVMMU {

final double [] p_imin,icoefs,f,dfdp;
final int neq_p;
final double f_pl,dfdp_pl;
private final QVMData odata;

private QVMMU(final com.polytechnik.utils.PolynomialRootsFinder rf,
	      final int neq_p,
	      final QVMData odata,
	      final int mom_type)
{
    this.odata=odata;
    this.neq_p=neq_p;
    final int o=QVMORT.getOOffset(odata,mom_type);
    final double [] abc,rp=QVMORT.getRPEM(odata,neq_p);
    if(mom_type==QVMData.MOM_dV) {abc=odata.vmoments_vol_x_tau_x_p;for(int k=rp.length;--k>=0;rp[k]*=odata.v_scale);}
    else if(mom_type==QVMData.MOM_dT) abc=odata.tmoments_vol_x_tau_x_p;
    else throw new IllegalArgumentException("mom_type="+mom_type);

    final double [][] x=QVMORT.getFLast(abc,o,odata.npm,neq_p,rp,0);
    final double [] pcoefs=x[0],wcoefs=x[1],dvdtc=x[3];
    if(true)
    {
	final double p_last=this.odata.trade_last==null?Double.NaN:this.odata.trade_last.price;
	// to remove?
	final double dp_last=p_last-odata.p_offset;
	this.f_pl=Polynomials.getD0(dvdtc,dp_last);
	this.dfdp_pl=Polynomials.getD1(dvdtc,dp_last);
    }

    final PolynomialRootsFinder.PRoots u;
    if(neq_p>4){
	if(rf==null) {
	    u=new PolynomialRootsFinder.PRoots(0,pcoefs,new double [neq_p],new double [neq_p]);
	} else {
	    u=rf.findRoots(pcoefs);
	}
    } else {
	u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoefs);
    }
    final int nroots=u.nroots;
    final double [] root_re=u.re_root,root_im=u.im_root;

    // insertion sort by real part with extra load
    SortArrays.sortArrayWithLoad(nroots,root_re,root_im);

    this.p_imin=java.util.Arrays.copyOf(root_re,
	CalculateKendall.selectRealRoots(root_re,nroots,root_re,root_im,CalculateKendall.EPS_IM));

    this.icoefs=this.p_imin.length==root_im.length?root_im:new double [this.p_imin.length]; // try to re-use an array
    this.f=new double [this.p_imin.length];
    this.dfdp=new double [this.p_imin.length];
    for(int k=this.p_imin.length;--k>=0;)
    {
	final double pk=this.p_imin[k];
	final double [] tmp=Polynomials.sdiv1(pcoefs,pk).q;
	if(neq_p!=tmp.length) throw new RuntimeException();
	this.icoefs[k]=getAverage(tmp,neq_p,abc,o,abc[o])/Polynomials.getD0(tmp,pk);
	this.f[k]=Polynomials.getD0(dvdtc,pk);
	this.dfdp[k]=Polynomials.getD1(dvdtc,pk);
	this.p_imin[k]+=odata.p_offset; // shift to 0 offset
    }
}


private static double [] getMoment2Shifted(final int neq,final double [] abc,final double dp)
{
    final int nd=2*neq+1;
    if(abc.length<2+nd) throw new IllegalArgumentException();
    final double [] abc2=new double [nd];
    for(int k=0;k<nd;k++)
    {
	abc2[k]=abc[k+2]+dp*(dp*abc[k]-2*abc[k+1]);
    }
    return abc2;
}
private static double [] getRPEM2Shifted(final int neq,final QVMData odata,final double dp)
{
    final double [] v=odata.vmoments_vol_x_tau_x_p,t=odata.tmoments_vol_x_tau_x_p;
    final int ot=QVMORT.getOOffset(odata,QVMData.MOM_dT),ov=QVMORT.getOOffset(odata,QVMData.MOM_dV);
    final double [] r=new double [neq];
    for(int k=r.length;--k>=0;)
    {
	r[k]=(v[k+1+ot]-dp*v[k+ot])/odata.v_scale-(t[k+1+ov]-dp*t[k+ov]);
    }
    return r;
}

private static void checkMatchingInt(final int neq_p,
				     final java.util.Random r,
				     final QVMData odata,
				     final double eps,
				     final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{
    final double [][][] abc3v=QVMData.get3ArrayFromArrayForTesting(odata.vmoments_vol_x_tau_x_p,odata.nvm,odata.ntm,odata.npm),
	abc3t=QVMData.get3ArrayFromArrayForTesting(odata.tmoments_vol_x_tau_x_p,odata.nvm,odata.ntm,odata.npm);
  for(int mom_type:new int []{QVMData.MOM_dT,QVMData.MOM_dV})
  {
      final int iv=mom_type==QVMData.MOM_dV?1:0,it=1-iv;
      final double [] abc=mom_type==QVMData.MOM_dV?odata.vmoments_vol_x_tau_x_p:odata.tmoments_vol_x_tau_x_p;

      final double [] atx=QVMData.get3ArrayFromArrayForTesting(abc,odata.nvm,odata.ntm,odata.npm)[iv][it].clone();
      if(mom_type==QVMData.MOM_dV) for(int k=atx.length;--k>=0;atx[k]/=odata.v_scale);
      final double [] rp=new double [neq_p];
      final double [] wcoefs=new double [2*neq_p+1],vtcoefs=new double [neq_p];
      for(int n=0;n<neq_p;n++)
      {
	  rp[n]=abc3v[0][1][n]/odata.v_scale-abc3t[1][0][n];
	  final double [] f=QVMORT.getFLast(atx,odata.npm,n)[0];
	  DeprecatedHelperTMatr.sumP2(wcoefs,f);
	  TMatr.sumP1(vtcoefs,f,getAverage(f,rp,atx[0]));
      }
      UnitTests.checkDoubleArrayRel(rp,"rp",QVMORT.getRPEM(odata,neq_p),"getRPEM(odata,neq_p)",eps);
      UnitTests.checkDoubleArrayRel(DeprecatedHelperEVXData.getLinsysSol(neq_p,atx,0,rp,0),"LS",vtcoefs,"vtcoefs",eps);

      final QVMMU qmu=new QVMMU(rf,neq_p,odata,mom_type);
      UnitTests.checkDoubleRel(qmu.f_pl,"qmu.f_pl",Polynomials.getD0(vtcoefs,qmu.odata.trade_last.price-qmu.odata.p_offset),"f_pl direct",eps);

      for(int k=0;k<qmu.p_imin.length;k++)
      {
	  final double pk=qmu.p_imin[k]-qmu.odata.p_offset;
	  final double wk=1/Polynomials.getD0(wcoefs,pk);
	  UnitTests.checkDoubleRel(wk,"wk",qmu.icoefs[k],"qmu.icoefs[k]",eps);
	  UnitTests.checkDoubleRel(Polynomials.getD0(vtcoefs,pk),"f direct",qmu.f[k],"qmu.f[k]",eps);
      }


      // test integral calculation
      final int n=2*neq_p-1;
      final double [] tpol=new double [n+1];
      for(int i=n;i>=0;i--) tpol[i]=2*(0.5-r.nextDouble());

      double si=0;
      for(int i=0;i<qmu.icoefs.length;i++)
      {
	  si+=qmu.icoefs[i]*Polynomials.getD0(tpol,qmu.p_imin[i]-qmu.odata.p_offset);
      }
      final double msi=getAverage(tpol,atx,atx[0]);
      UnitTests.checkDoubleRel(si,"si",msi,"msi",eps);
      System.err.println("checkMatchingInt:integral OK: si="+si+" msi="+msi+" for n="+n+" neq_p="+neq_p);
  }
  System.err.println("QVMMU:checkMatchingInt: tests OK  neq_p="+neq_p);
}


private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_p)
{
    final double eps=1e-5;
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMMU:checkCalculations(): test with N points="+tsize+
			   " currenttime="+currenttime+" tau="+tau+
			   " neq_p="+neq_p);

    final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();
    final QVMData odata=QVMData.calculateQVMData(1+1+r.nextInt(2),
						 1+1+r.nextInt(2),
						 1+2*neq_p+2,
						 currenttime,
						 tau,
						 tb,
						 tsize,
						 null);

	// integral
    checkMatchingInt(neq_p,r,odata,eps,rf);

    System.err.println("QVMMU:checkCalculations(): OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " neq_p="+neq_p);
}

public static void main(String [] args)
{
    System.err.println("QVMMU::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-79091);
    for(int neq_p=2;neq_p<=8;neq_p++)
    {
	for(int i=0;i<110;i++)
	{
	    checkCalculations(100+r.nextInt(100),r,neq_p);
	    checkCalculations(10010+r.nextInt(100010),r,neq_p);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println("QVMMU::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
}

} // END OF class QVMMU


static class QVMVTMU {
private final QVMData odata;
final QVMMU vm,tm;

QVMVTMU(final QVMMU vm,
	final QVMMU tm,
	final QVMData odata)
{
    this.vm=vm;
    this.tm=tm;
    this.odata=odata;
}

public static QVMVTMU calculateQVMVTMU(final long currenttime,
				       final int neq_p,
				       final long tau,
				       final Trade [] tb,
				       final int tsize,
				       final com.polytechnik.utils.PolynomialRootsFinder rf,
				       final QVMVTMU ref)

{
    final QVMData o=QVMData.calculateQVMData(1+1,
					     1+1,
					     1+2*neq_p,
					     currenttime,
					     tau,
					     tb,
					     tsize,
					     ref!=null?ref.odata:null);
    final QVMMU v=new QVMMU(rf,neq_p,o,QVMData.MOM_dT),t=new QVMMU(rf,neq_p,o,QVMData.MOM_dV);
    return new QVMVTMU(v,t,o);
}

} // END of class QVMVTMU








/** The data for QVM calculations. */
static class QVMDataMU {

final long currenttime,tau,vol,tol;
final double p_offset,v_scale;

private final int nsm,nvm,ntm,npm;

private final double [] // here offset=p_offset and array dims are: nsm,nvm,ntm,npm.
    moments_s_x_vol_x_tau_x_p, // [k][l][m][n]=sum s^k (vol-v_cur)^l (t-t_cur)^m (p-offset)^n dt_i exp()
    vmoments_s_x_vol_x_tau_x_p // [k][l][m][n]=sum s^k (vol-v_cur)^l (t-t_cur)^m (p-offset)^n dv_i exp()
    ;

// references to previos calculations
private final Trade trade_last;


private QVMDataMU(final long currenttime,
		  final long tau,
		  final long vol,
		  final long tol,
		  final double p_offset,
		  final double v_scale,

		  final double [] moments_s_x_vol_x_tau_x_p,
		  final double [] vmoments_s_x_vol_x_tau_x_p,
		  final int nsm,
		  final int nvm,
		  final int ntm,
		  final int npm,

		  // references
		  final Trade trade_last)
{
    this.currenttime=currenttime;
    this.tau=tau;
    this.vol=vol;
    this.tol=tol;
    this.p_offset=p_offset;
    this.v_scale=v_scale;
    this.moments_s_x_vol_x_tau_x_p=moments_s_x_vol_x_tau_x_p;
    this.vmoments_s_x_vol_x_tau_x_p=vmoments_s_x_vol_x_tau_x_p;
    this.nsm=nsm;
    this.nvm=nvm;
    this.ntm=ntm;
    this.npm=npm;
    this.trade_last=trade_last;
}

private static void checkMatchingQVMData(final QVMDataMU r,final QVMData ref,final double eps)
{
    if(r.currenttime!=ref.currenttime)
    {
	throw new RuntimeException("r.currenttime="+r.currenttime+
				   " ref.currenttime="+ref.currenttime);
    }
    if(r.tau!=ref.tau)
    {
	throw new RuntimeException("r.tau="+r.tau+" ref.tau="+ref.tau);
    }
    if(r.vol!=ref.vol)
    {
	throw new RuntimeException("r.vol="+r.vol+" ref.vol="+ref.vol);
    }
    if(r.nvm!=ref.nvm||r.ntm!=ref.ntm||r.npm!=ref.npm||r.nsm<1)
    {
	throw new RuntimeException("r.nvm="+r.nvm+" ref.nvm="+ref.nvm+
				   "r.ntm="+r.ntm+" ref.ntm="+ref.ntm+
				   "r.npm="+r.npm+" ref.npm="+ref.npm+
				   "r.nsm="+r.nsm);
    }

    final double [] vscale=new double [r.nvm*r.ntm*r.npm];
    final double dsk=ref.v_scale/r.v_scale;
    double s=1;
    for(int k=0;k<r.nvm;k++,s*=dsk) java.util.Arrays.fill(vscale,k*r.ntm*r.npm,(k+1)*r.ntm*r.npm,s);

    UnitTests.checkDoubleArrayRel(java.util.Arrays.copyOf(r.moments_s_x_vol_x_tau_x_p,vscale.length),"r.moments_s_x_vol_x_tau_x_p",
						   getP1xP2(ref.shiftOffset(0,0,ref.p_offset-r.p_offset,QVMData.MOM_dT),vscale.length,vscale,vscale.length),"shifted ref.moments_vol_x_tau_x_p",eps);
    UnitTests.checkDoubleArrayRel(java.util.Arrays.copyOf(r.vmoments_s_x_vol_x_tau_x_p,vscale.length),"r.vmoments_s_x_vol_x_tau_x_p",
						   getP1xP2(ref.shiftOffset(0,0,ref.p_offset-r.p_offset,QVMData.MOM_dV),vscale.length,vscale,vscale.length),"shifted ref.moments_vol_x_tau_x_p",eps);

    System.err.println("QVMDataMU:checkMatchingQVMData: matching OK");
}

static QVMDataMU calculateQVMDataMU(final int nsm,
				    final int nvm,
				    final int ntm,
				    final int npm,
				    final long currenttime,
				    final long tau,
				    final Trade [] tb,
				    final int tsize)
{
    final double atau=tau;
    final double p_offset=tsize>0?tb[tsize-1].price:0,v_scale=1;

    final double []  // all 0 by default
	moments_s_x_vol_x_tau_x_p=new double [nsm*nvm*ntm*npm],
	vmoments_s_x_vol_x_tau_x_p=new double [nsm*nvm*ntm*npm];

    long vx=0,tx=0,txdv_m_vxdt=0,prevt=currenttime,tinterval=currenttime;
    for(int i=tsize;--i>=0;)
    {
	final Trade b=tb[i];
	if(b.time>prevt)
	{
	    throw new RuntimeException("The data is not sorted:"+prevt+" and "+b.time);
	}
	final long dt=tinterval-(prevt=b.time);
	txdv_m_vxdt+=(tx-=dt)*b.shares-(vx-=b.shares)*dt;
	final double s=(tx>=0||vx>=0)?0:((((double)txdv_m_vxdt)/((double)tx))/((double)vx));
	final double etau=tx/atau,ex=Math.exp(etau),evol=vx/v_scale,
	    dp=b.price-p_offset,twmw=ex*(dt/atau),bwmw=ex*b.shares;

	double wse=1;
	for(int k=0,j=0;k<nsm;k++)
	{
	    double wve=wse;
	    for(int l=0;l<nvm;l++)
	    {
		double wte=wve;
		for(int m=0;m<ntm;m++)
		{
		    double wpe=wte;
		    for(int n=0;n<npm;n++,j++)
		    {
			moments_s_x_vol_x_tau_x_p[j]+=wpe*twmw;
			vmoments_s_x_vol_x_tau_x_p[j]+=wpe*bwmw;
			wpe*=dp;
		    }
		    wte*=etau;
		}
		wve*=evol;
	    }
	    wse*=s;
	}
	tinterval=prevt;
    }

    return new QVMDataMU(currenttime,
			 tau,
			 -vx,
			 -tx,
			 p_offset,
			 v_scale,
			 moments_s_x_vol_x_tau_x_p,
			 vmoments_s_x_vol_x_tau_x_p,
			 nsm,
			 nvm,
			 ntm,
			 npm,
			 tsize>0?tb[tsize-1]:null);
}


private static void checkCalculations(final int tsize,
				      final java.util.Random r,
				      final int nsm,
				      final int nvm,
				      final int ntm,
				      final int npm,
				      final double eps)
{
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMDataMU:checkCalculations(): test with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " nsm="+nsm+" nvm="+nvm+" ntm="+ntm+" npm="+npm);

    final QVMData ro=QVMData.calculateQVMData(nvm,ntm,npm,currenttime,tau,tb,tsize,null);

    final QVMDataMU rmu=QVMDataMU.calculateQVMDataMU(nsm,nvm,ntm,npm,currenttime,tau,tb,tsize);

    checkMatchingQVMData(rmu,ro,eps);

    System.err.println("QVMDataMU:checkCalculations(): matched QVMData OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " nsm="+nsm+" nvm="+nvm+" ntm="+ntm+" npm="+npm);
}

public static void main(String [] args)
{
    System.err.println("QVMDataMU::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-79091);
    final double eps=1e-5;

    for(int nsm=2;nsm<=4;nsm++)
    for(int nvm=1;nvm<=7;nvm++)
    for(int ntm=1;ntm<=7;ntm++)
    for(int npm=1;npm<=7;npm++)
    {
	for(int i=0;i<4;i++)
	{
	    checkCalculations(100+r.nextInt(100),r,nsm,nvm,ntm,npm,eps);
	    checkCalculations(10010+r.nextInt(100010),r,nsm,nvm,ntm,npm,eps);
	}
    }

    final long t1=System.currentTimeMillis();
    System.err.println("QVMDataMU::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
}

} // END of class QVMDataMU

/** QVMSORT version of QVMCoefs. */
static class QVMSORT {

final double sigma,p_average;
final double [] p_imin,icoefs;
final int neq_p;
private final QVMDataMU odata;

final double f_pl,dfdp_pl;
final double [] f,dfdp;

private QVMSORT(final com.polytechnik.utils.PolynomialRootsFinder rf,
		final int neq_p,
		final QVMDataMU odata,
		final double [] odata_abc)
{
    this.odata=odata;
    this.neq_p=neq_p;
    if(odata.nsm<=1) throw new IllegalArgumentException();
    final double [] abc=odata_abc;
    final double pt0=abc[0],pa=odata.npm<2?Double.NaN:abc[1]/pt0;
    this.p_average=pa+odata.p_offset;
    this.sigma=odata.npm<3?Double.NaN:Math.sqrt((abc[2]/pt0)-(pa*pa));

    final double [][] x=QVMORT.getFLast(abc,0,odata.npm,this.neq_p,abc,((1*odata.nvm+0)*odata.ntm+0)*odata.npm);
    final double [] pcoefs=x[0],scoefs=x[3];
    if(true)
    {
	final double p_last=this.odata.trade_last==null?Double.NaN:this.odata.trade_last.price;
	// to remove?
	final double dp_last=p_last-odata.p_offset;
	this.f_pl=Polynomials.getD0(scoefs,dp_last);
	this.dfdp_pl=Polynomials.getD1(scoefs,dp_last);
    }

    final PolynomialRootsFinder.PRoots u;
    if(neq_p>4){
	if(rf==null) {
	    u=new PolynomialRootsFinder.PRoots(0,pcoefs,new double [neq_p],new double [neq_p]);
	} else {
	    u=rf.findRoots(pcoefs);
	}
    } else {
	u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoefs);
    }
    final int nroots=u.nroots;
    final double [] root_re=u.re_root,root_im=u.im_root;

    // insertion sort by real part with extra load
    SortArrays.sortArrayWithLoad(nroots,root_re,root_im);

    this.p_imin=java.util.Arrays.copyOf(root_re,
	CalculateKendall.selectRealRoots(root_re,nroots,root_re,root_im,CalculateKendall.EPS_IM));

    this.icoefs=this.p_imin.length==root_im.length?root_im:new double [this.p_imin.length]; // try to re-use an array
    this.f=new double [this.p_imin.length]; // FIXME to remove
    this.dfdp=new double [this.p_imin.length]; // FIXME to remove
    for(int k=this.p_imin.length;--k>=0;)
    {
	final double pk=this.p_imin[k];
	final double [] tmp=Polynomials.sdiv1(pcoefs,pk).q;
	if(neq_p!=tmp.length) throw new RuntimeException();
	this.icoefs[k]=getAverage(tmp,neq_p,abc,pt0)/Polynomials.getD0(tmp,pk);
	this.f[k]=Polynomials.getD0(scoefs,pk); // FIXME to remove
	this.dfdp[k]=Polynomials.getD1(scoefs,pk); // FIXME to remove
	this.p_imin[k]+=odata.p_offset; // shift to 0 offset
    }
}


private static void checkMatchingInt(final int neq_p,
				     final java.util.Random r,
				     final QVMDataMU odata,
				     final double eps,
				     final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{
  for(double [] abc:new double [][]{odata.vmoments_s_x_vol_x_tau_x_p,odata.moments_s_x_vol_x_tau_x_p})
  {
    final double [][][][] abc4=QVMData.get4ArrayFromArrayForTesting(abc,odata.nsm,odata.nvm,odata.ntm,odata.npm);
    final double [] sm=abc4[1][0][0];
    final double [] wcoefs=new double [2*neq_p+1],scoefs=new double [neq_p];
    for(int n=0;n<neq_p;n++)
    {
	final double [] f=QVMORT.getFLast(abc4[0][0][0],odata.npm,n)[0];
	DeprecatedHelperTMatr.sumP2(wcoefs,f);
	TMatr.sumP1(scoefs,f,getAverage(f,sm,abc4[0][0][0][0]));
	final double [] fa=new double [n];
	for(int k=n;--k>=0;fa[k]=-f[k]/f[n]);
	UnitTests.checkDoubleArrayRel(DeprecatedHelperEVXData.getLinsysSol(n,abc,0,abc,n),"LS",fa,"fa",eps);
    }
    UnitTests.checkDoubleArrayRel1(DeprecatedHelperEVXData.getLinsysSol(neq_p,abc,0,sm,0),"LS",scoefs,"scoefs",eps);

    final QVMSORT qmu=new QVMSORT(rf,neq_p,odata,abc);
    UnitTests.checkDoubleRel(qmu.f_pl,"qmu.f_pl",Polynomials.getD0(scoefs,qmu.odata.trade_last.price-qmu.odata.p_offset),"f_pl direct",eps);

    for(int k=0;k<qmu.p_imin.length;k++)
    {
	final double pk=qmu.p_imin[k]-qmu.odata.p_offset;
	final double wk=1/Polynomials.getD0(wcoefs,pk);
	UnitTests.checkDoubleRel(wk,"wk",qmu.icoefs[k],"qmu.icoefs[k]",eps);
	UnitTests.checkDoubleRel(Polynomials.getD0(scoefs,pk),"f direct",qmu.f[k],"qmu.f[k]",eps);
    }

    final QVMORT qo=new QVMORT(rf,neq_p,
			       new QVMData(qmu.odata.currenttime,
					   qmu.odata.tau,
					   qmu.odata.vol,
					   0,
					   qmu.odata.p_offset,
					   Double.NaN,
					   Double.NaN,
					   Double.NaN,
					   qmu.odata.v_scale,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   null,
					   qmu.odata.nvm,
					   qmu.odata.ntm,
					   qmu.odata.npm,
					   qmu.odata.trade_last,null),
			       abc,
			       -1);
    UnitTests.checkDoubleArrayRel(qmu.p_imin,"qmu.p_imin",qo.p_imin,"qo.p_imin",eps);
    UnitTests.checkDoubleRel(qmu.p_average,"qmu.p_average",qo.p_average,"qo.p_average",eps);
    UnitTests.checkDoubleRel(qmu.sigma,"qmu.sigma",qo.sigma,"qo.sigma",eps);

    // test integral calculation
    final int n=2*neq_p-1;
    final double [] tpol=new double [n+1];
    for(int i=n;i>=0;i--) tpol[i]=2*(0.5-r.nextDouble());

    double si=0;
    for(int i=0;i<qmu.icoefs.length;i++)
    {
	si+=qmu.icoefs[i]*Polynomials.getD0(tpol,qmu.p_imin[i]-qmu.odata.p_offset);
    }
    final double msi=getAverage(tpol,abc,abc[0]);
    UnitTests.checkDoubleRel(si,"si",msi,"msi",eps);
    System.err.println("checkMatchingInt:integral OK: si="+si+" msi="+msi+" for n="+n+" neq_p="+neq_p);
  }
  System.err.println("QVMSORT:checkMatchingInt: tests OK  neq_p="+neq_p);
}


private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_p)
{
    final double eps=1e-5;
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/20,currenttime=tb[tb.length-1].time+20000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMSORT:checkCalculations(): test with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
			   " neq_p="+neq_p);

    final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();
    final QVMDataMU odata=QVMDataMU.calculateQVMDataMU(1+1+r.nextInt(2),
						       1+r.nextInt(2),
						       1+r.nextInt(2),
						       1+2*neq_p+r.nextInt(2),
						       currenttime,
						       tau,
						       tb,
						       tsize);

    // integral
    checkMatchingInt(neq_p,r,odata,eps,rf);

    System.err.println("QVMSORT:checkCalculations(): OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " neq_p="+neq_p);
}

public static void main(String [] args)
{
    System.err.println("QVMSORT::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-79091);
    for(int neq_p=2;neq_p<=8;neq_p++)
    {
	for(int i=0;i<110;i++)
	{
	    checkCalculations(100+r.nextInt(100),r,neq_p);
	    checkCalculations(10010+r.nextInt(100010),r,neq_p);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println("QVMSORT::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
}

public static QVMSORT calculateQVMSORT(final int neq_p,
				       final long currenttime,
				       final long tau,
				       final Trade [] tb,
				       final int tsize,
				       final com.polytechnik.utils.PolynomialRootsFinder rf)
{
    final QVMDataMU o=QVMDataMU.calculateQVMDataMU(1+1,
						   1,
						   1,
						   1+2*neq_p,
						   currenttime,
						   tau,
						   tb,
						   tsize);
    return new QVMSORT(rf,neq_p,o,o.vmoments_s_x_vol_x_tau_x_p /* o.moments_s_x_vol_x_tau_x_p */);
}

} // END of class QVMSORT





/** Return a polynomial of p_l */
static double [] getCPLCPIcoefs(final double G11,final double G12,final double G22,
				final double O1,final double O2,
				final double R1,final double R2)
{
    final double [] matr=new double [4],rpart=new double [2];
    matr[0]=G11;
    matr[1]=matr[2]=G12;
    matr[3]=G22;
    final int [] ipvt=new int [2];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,2,2,ipvt);
    if(pinfo!=0)
    {
	return new double []{Double.NaN,Double.NaN,Double.NaN};
    }
    rpart[0]=O1;
    rpart[1]=O2;
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,2,2,ipvt,rpart,1);
    final double cpp=O1*rpart[0]+O2*rpart[1];
    final double cp=-2*(R1*rpart[0]+R2*rpart[1]); // the matrix is symmetric
    rpart[0]=R1;
    rpart[1]=R2;
    com.polytechnik.utils.Linsystems.dgesl_m1(matr,2,2,ipvt,rpart,1);
    final double c=R1*rpart[0]+R2*rpart[1];
    return new double []{c,cp,cpp};
}
/** Same as getCPLCPIcoefs. */
/*
private static double [] getCPLCPI_Dcoefs(final double G11,final double G12,final double G22,
					  final double O1,final double O2,
					  final double R1,final double R2)
{
    final double d=G11*G22-G12*G12,Gi11=(G22/d),Gi12=-(G12/d),Gi22=(G11/d);
    final double x1=Gi11*O1+Gi12*O2,x2=Gi12*O1+Gi22*O2,
	r1=Gi11*R1+Gi12*R2,r2=Gi12*R1+Gi22*R2;
    return new double [] {(R1*r1+R2*r2),-2*(R1*x1+R2*x2),(O1*x1+O2*x2)};
}
*/

/** Similar to getCPLCPIcoefs. G,O,R are clobbered on output. */
static double [] getCPLCPIcoefsM(final int neq_i,final double [] G,final double [] O,final double [] R)
{
    final int [] ipvt=new int [neq_i];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(G,neq_i,neq_i,ipvt);
    if(pinfo!=0)
    {
	return new double []{Double.NaN,Double.NaN,Double.NaN};
    }
    final double [] rpart=new double [neq_i];
    System.arraycopy(O,0,rpart,0,neq_i);
    com.polytechnik.utils.Linsystems.dgesl_m1(G,neq_i,neq_i,ipvt,rpart,1);
    final double cpp=TMatr.mulVV(O,rpart);
    final double cp=-2*TMatr.mulVV(R,rpart); // the G matrix is symmetric
    System.arraycopy(R,0,rpart,0,neq_i);
    com.polytechnik.utils.Linsystems.dgesl_m1(G,neq_i,neq_i,ipvt,rpart,1);
    final double c=TMatr.mulVV(R,rpart);
    return new double []{c,cp,cpp};
}

/** Like getCPLCPIcoefsM but calculate two solutions at once.
 */
static double [][] getCPLCPIcoefsM2Sol(final int neq_i,final double [] G,
				       final double [] Oa,final double [] Ra,
				       final double [] Ob,final double [] Rb)
{
    final int [] ipvt=new int [neq_i];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(G,neq_i,neq_i,ipvt);
    if(pinfo!=0)
    {
	return new double [][]{
	    {Double.NaN,Double.NaN,Double.NaN},
	    {Double.NaN,Double.NaN,Double.NaN}};
    }
    final double [] dissa,dissb,rpart=new double [neq_i];
    {
	System.arraycopy(Oa,0,rpart,0,neq_i);
	com.polytechnik.utils.Linsystems.dgesl_m1(G,neq_i,neq_i,ipvt,rpart,1);
	final double cpp=TMatr.mulVV(Oa,rpart);
	final double cp=-2*TMatr.mulVV(Ra,rpart); // the G matrix is symmetric
	System.arraycopy(Ra,0,rpart,0,neq_i);
	com.polytechnik.utils.Linsystems.dgesl_m1(G,neq_i,neq_i,ipvt,rpart,1);
	final double c=TMatr.mulVV(Ra,rpart);
	dissa=new double []{c,cp,cpp};
    }{
	System.arraycopy(Ob,0,rpart,0,neq_i);
	com.polytechnik.utils.Linsystems.dgesl_m1(G,neq_i,neq_i,ipvt,rpart,1);
	final double cpp=TMatr.mulVV(Ob,rpart);
	final double cp=-2*TMatr.mulVV(Rb,rpart); // the G matrix is symmetric
	System.arraycopy(Rb,0,rpart,0,neq_i);
	com.polytechnik.utils.Linsystems.dgesl_m1(G,neq_i,neq_i,ipvt,rpart,1);
	final double c=TMatr.mulVV(Rb,rpart);
	dissb=new double []{c,cp,cpp};
    }
    return new double [][]{dissa,dissb};
}








/** Like QVMLL with the PTV-moments. */
static class QVMLPV {

static final Laguerre BL=new Laguerre();


final double sigma,p_average;
private final double [] pvcoefs,icoefs,tPcoefs;
final int neq_t,neq_p;
final double I,dIdT,dIdP,dpt;
private final QVMDataL odata;

private QVMLPV(final com.polytechnik.utils.PolynomialRootsFinder rf,
	       final int neq_t,
	       final int neq_p,
	       final QVMDataL odata)
{
    this.odata=odata;
    this.neq_t=neq_t;
    this.neq_p=neq_p;
    if(odata.npm<2||odata.ntm<neq_t||odata.npm<2*neq_p-1) throw new IllegalArgumentException();

    final double [] abc=odata.vmoments_Ltau_x_p,abc_t=odata.tmoments_Ltau_x_p;
    final double pt0=abc[0],pa=odata.npm<2?Double.NaN:abc[1]/pt0;
    this.p_average=pa+odata.p_offset;
    this.sigma=odata.npm<3?Double.NaN:Math.sqrt((abc[2]/pt0)-(pa*pa));

    final double [][] x=getLinsys2(neq_p,neq_t,abc_t,odata.npm,abc);
    final double pt0t=abc_t[0],scale=pt0t/pt0;
    this.icoefs=x[1];
    this.tPcoefs=new double [neq_t];
    for(int n=0,o=0;n<neq_t;++n,o+=odata.npm)
    {
	this.icoefs[n]*=scale;
	this.tPcoefs[n]=abc_t[1+o]/pt0t; // L_n moments, just for testing.
    }
    this.pvcoefs=x[0];
    for(int n=0;n<neq_p;++n)
    {
	this.pvcoefs[n]*=scale;
    }

    if(true)
    {
	final double dp_last=odata.trade_last==null?Double.NaN:(odata.trade_last.price-odata.p_offset);
	if(neq_p!=this.pvcoefs.length) throw new RuntimeException();
	this.I=Polynomials.getD0(this.pvcoefs,dp_last)+BL.getD0(this.icoefs,0);
	this.dIdT=-BL.getD1(this.icoefs,0); // "-" for inverted time
	if(neq_p!=this.pvcoefs.length) throw new RuntimeException();
	this.dIdP=Polynomials.getD1(this.pvcoefs,dp_last);
	this.dpt=-BL.getD1(this.tPcoefs,0); // "-" for inverted time
    }
}



public static QVMLPV calculateQVMLPV(final int neq_t,
				     final int neq_p,
				     final long currenttime,
				     final long tau,
				     final Trade [] tb,
				     final int tsize,
				     final com.polytechnik.utils.PolynomialRootsFinder rf,
				     final QVMLPV ref)
{
    final QVMDataL o=QVMDataL.calculateQVMDataL(neq_t,
						Math.max(2,2*neq_p-1),
						currenttime,
						tau,
						tb,
						tsize,
						ref!=null?ref.odata:null);
    return new QVMLPV(rf,neq_t,neq_p,o);
}

private static void checkMatchingValues(final int neq_t,final int neq_p,final java.util.Random r,final long currenttime,final long tau,final Trade [] tb,final int tsize,final double eps,final double epssum,final com.polytechnik.utils.PolynomialRootsFinder rf)
{
    //final QVMLPV q=QVMLPV.calculateQVMLPV(neq_t,neq_p,currenttime,tau,tb,tsize,rf,null);
    // nothing for now

    System.err.println("QVMLPV:checkMatchingValues: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}

private static void checkMatchingLinsystemValues(final int neq_t,final int neq_p,final java.util.Random r,final long currenttime,final long tau,final Trade [] tb,final int tsize,final double eps,final double epssum,final com.polytechnik.utils.PolynomialRootsFinder rf)
{
    final QVMDataL odata=QVMDataL.calculateQVMDataL(2*neq_t-1+r.nextInt(4),Math.max(3,2*neq_p-1)+r.nextInt(4),currenttime,tau,tb,tsize,null);
    final double [] abc=odata.vmoments_Ltau_x_p,abc_t=odata.tmoments_Ltau_x_p;


    if(true)
    { // icoefs match
	final QVMLPV q=QVMLPV.calculateQVMLPV(neq_t,1,currenttime,tau,tb,tsize,rf,null);

	final double [] ref_icoefs=DeprecatedHelperEVXData.getLinsysSolPBasis(neq_t,abc_t,0,odata.npm,TMatr.getVectorScaled(abc,abc_t[0]/abc[0]),0,odata.npm,QVMDataL.Laguerre_POLS),
	    ref_tPcoefs=DeprecatedHelperEVXData.getLinsysSolPBasis(neq_t,abc_t,0,odata.npm,abc_t,1,odata.npm,QVMDataL.Laguerre_POLS);

	UnitTests.checkDoubleArrayRel1(q.icoefs,"q.icoefs",ref_icoefs,"ref_icoefs",epssum);
	UnitTests.checkDoubleArrayRel1(q.tPcoefs,"q.tPcoefs",ref_tPcoefs,"ref_tPcoefs",epssum);
    }
    if(true)
    { // pvcoefs match
	final QVMLPV q=QVMLPV.calculateQVMLPV(1,neq_p,currenttime,tau,tb,tsize,rf,null);
	final double [] pvcoefs=q.pvcoefs.clone();pvcoefs[0]+=q.icoefs[0];

	final double [] ref_pvcoefs=CalculateKendall.QATradeIP.getPCoefs(abc_t,abc,neq_p-1)[0];
	ref_pvcoefs[0]+=1; // To match our current definition, which is different from one used in QATradeIP.
	UnitTests.checkDoubleArrayRel1(pvcoefs,"pvcoefs",ref_pvcoefs,"ref_pvcoefs",eps);
	if(neq_p>=2)
	{
	    final CalculateKendall.QATradeIP qa=CalculateKendall.QATradeIP.calculateQATradeIP(neq_p-1,currenttime,tau,tb,tsize,rf,null);
	    final double [] ref2_pvcoefs=Polynomials.getPolynomialShifted(qa.pcoefs,q.odata.p_offset-qa.odata.p_offset);ref2_pvcoefs[0]+=1;
	    UnitTests.checkDoubleArrayRel1(pvcoefs,"pvcoefs",ref2_pvcoefs,"ref2_pvcoefs",eps);
	    UnitTests.checkDoubleRel(q.I,"q.I",qa.getAI0(q.odata.trade_last.price)+1,"I:QATradeIP",eps);
	    UnitTests.checkDoubleRel(q.dIdP,"q.dIdP",qa.getDIDP(q.odata.trade_last.price),"DI:QATradeIP",eps);
	}
    }

    System.err.println("QVMLPV:checkMatchingLinsystemValues: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}

private static void checkMatchingQVMLL(final int neq_t,
				       final int neq_p,
				       final java.util.Random r,
				       final long currenttime,
				       final long tau,
				       final Trade [] tb,
				       final int tsize,
				       final double eps,
				       final double epssum,
				       final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{

  final QVMLL ref=QVMLL.calculateQVMLL(neq_t,currenttime,tau,tb,tsize,null,rf);
  final QVMDataL ox=QVMDataL.calculateQVMDataL(neq_t+r.nextInt(4),Math.max(2*neq_p-1,2)+r.nextInt(4),currenttime,tau,tb,tsize,null);

  for(QVMLPV q: new QVMLPV[]{QVMLPV.calculateQVMLPV(neq_t,1,currenttime,tau,tb,tsize,rf,null),new QVMLPV(rf,neq_t,1,ox)})
  {
    UnitTests.checkDoubleRel(q.odata.p_offset,"q.odata.p_offset",ref.odata.p_offset,"ref.odata.p_offset",eps);
    UnitTests.checkDoubleRel(q.p_average,"q.p_average",ref.p_average,"ref.p_average",eps);
    if(q.odata.npm>2 == ref.odata.npm>2)
	UnitTests.checkDoubleRel(q.sigma,"q.sigma",ref.sigma,"ref.sigma",eps);
    UnitTests.checkDoubleRel(q.neq_t,"q.neq_t",ref.neq_t,"ref.neq_t",eps);


    UnitTests.checkDoubleArrayRel1(q.icoefs,"q.icoefs",ref.icoefs,"ref.icoefs",epssum);
    UnitTests.checkDoubleArrayRel1(q.tPcoefs,"q.tPcoefs",ref.tPcoefs,"ref.tPcoefs",eps);
  }

  System.err.println("QVMLPV:checkMatchingQVML: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}


private static void checkMatchingQATradeIP(final int neq_t,
					   final int neq_p,
					   final java.util.Random r,
					   final long currenttime,
					   final long tau,
					   final Trade [] tb,
					   final int tsize,
					   final double eps,
					   final double epssum,
					   final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{

  if(neq_p<2)
  {
      System.err.println("QVMLPV:checkMatchingQATradeIP: neq_t="+neq_t+" neq_p="+neq_p+"<2; skipping test");
      return;
  }
  final CalculateKendall.QATradeIP ref=CalculateKendall.QATradeIP.calculateQATradeIP(neq_p-1,currenttime,tau,tb,tsize,rf,null);

  final QVMDataL ox=QVMDataL.calculateQVMDataL(1+r.nextInt(4),Math.max(2*neq_p-1,2)+r.nextInt(4),currenttime,tau,tb,tsize,null);

  for(QVMLPV q: new QVMLPV[]{QVMLPV.calculateQVMLPV(1,neq_p,currenttime,tau,tb,tsize,rf,null),new QVMLPV(rf,1,neq_p,ox)})
  {

      UnitTests.checkDoubleRel(q.I,"q.I",ref.getAI0(q.odata.trade_last.price)+1,"I:QATradeIP",eps);
      UnitTests.checkDoubleRel(q.dIdP,"q.dIdP",ref.getDIDP(q.odata.trade_last.price),"DI:QATradeIP",eps);
      final double [] ref_pvcoefs=Polynomials.getPolynomialShifted(ref.pcoefs,q.odata.p_offset-ref.odata.p_offset);ref_pvcoefs[0]+=1;
      final double [] pvcoefs=q.pvcoefs.clone();pvcoefs[0]+=q.icoefs[0];
      UnitTests.checkDoubleArrayRel1(pvcoefs,"pvcoefs",ref_pvcoefs,"ref_pvcoefs",eps);
  }

  System.err.println("QVMLPV:checkMatchingQATradeIP: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}



private static void checkMatchingPolynomialP(final int neq_t,
					     final int neq_p,
					     final java.util.Random r,
					     final double eps,
					     final double epssum,
					     final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf,
					     final int tsize,
					     final long tau)
{
    if(false && neq_t<=1)
    {
	System.err.println("QVMLPV:checkMatchingPolynomialP: neq_t="+neq_t+" is too low, skipping test. Also note that trades with 'all prices the same' possibly make new version of QVMData{L,P} extremely slow!");
	return;
    }

    final double [] pcoefs_o=new double [neq_t],pcoefs_o_TradeVX=new double [neq_t];
    for(int k=0;k<neq_t;k++)
    {
	// Set random coefficients. Also make sure high order powers are relatively low.
	pcoefs_o[k]=r.nextDouble()/(2+Math.exp(k)+r.nextDouble());
	// Different normalizing is used by TradeVX.getRandomTradesPolynomialP, sign is the same.
	pcoefs_o_TradeVX[k]=pcoefs_o[k]*Polynomials.fact(k);
    }

    final Trade [] tb=new Trade [tsize];
    final long currenttime=CalculateKendall.TradeVX.getRandomTradesPolynomialP(tb,r,tau,pcoefs_o_TradeVX);
    final QVMLPV q=QVMLPV.calculateQVMLPV(neq_t,1,currenttime,tau,tb,tsize,rf,null);

    // q.tPcoefs has an offset q.odata.p_offset
    final double [] tpcoefs=q.tPcoefs.clone();tpcoefs[0]+=q.odata.p_offset;
    UnitTests.checkDoubleArrayRel1(BL.convertFunctionTo_Monomials_from_Basis(tpcoefs),"tpcoefs",pcoefs_o,"pcoefs_o",epssum);

    System.err.println("QVMLPV:checkMatchingPolynomialP: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}


private static void checkMatching_getLinsys2(final int neq_t,
					     final int neq_p,
					     final java.util.Random r,
					     final long currenttime,
					     final long tau,
					     final Trade [] tb,
					     final int tsize,
					     final double eps,
					     final double epssum,
					     final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{


    if(true)
    { // match t
	final QVMLL ql=QVMLL.calculateQVMLL(neq_t,currenttime,tau,tb,tsize,null,rf);
	final QVMDataL odata=QVMDataL.calculateQVMDataL(neq_t+r.nextInt(4),1+r.nextInt(4),currenttime,tau,tb,tsize,null);
	final double [] abc=odata.vmoments_Ltau_x_p,abc_t=odata.tmoments_Ltau_x_p;
	final double [] abc_scaled=TMatr.getVectorScaled(abc,abc_t[0]/abc[0]);
	final double [] l=getLinsys2(1,neq_t,abc_t,odata.npm,abc_scaled)[1];
	UnitTests.checkDoubleArrayRel1(ql.icoefs,"ql.icoefs",l,"getLinsys2[1]",eps);
    }

    if(neq_p>=2)
    { // match p
	final CalculateKendall.QATradeIP qa=CalculateKendall.QATradeIP.calculateQATradeIP(neq_p-1,currenttime,tau,tb,tsize,rf,null);
	final QVMDataL odata=QVMDataL.calculateQVMDataL(1+r.nextInt(4),2*neq_p-1+r.nextInt(4),currenttime,tau,tb,tsize,null);
	final double [] abc=odata.vmoments_Ltau_x_p,abc_t=odata.tmoments_Ltau_x_p;
	final double [] abc_scaled=TMatr.getVectorScaled(abc,abc_t[0]/abc[0]);
	final double [] qa_pvcoefs=Polynomials.getPolynomialShifted(qa.pcoefs,odata.p_offset-qa.odata.p_offset);qa_pvcoefs[0]+=1;
	final double [][] lsol=getLinsys2(neq_p,1,abc_t,odata.npm,abc_scaled);
	final double [] p=lsol[0];p[0]+=lsol[1][0];
	UnitTests.checkDoubleArrayRel1(qa_pvcoefs,"qa_pvcoefs",p,"p",eps);
    }

    if(true)
    { // check t & p combined, but only for nA=2, it is harder to check for nA>2
	final double [] px=new double [neq_t+1];
	//for(int k=px.length;--k>=0;){px[k]=r.nextDouble();}
	px[neq_t]=1+r.nextDouble();
	final double [] p=BL.convertFunctionTo_Monomials_from_Basis(px),p_o_TradeVX=new double [p.length];
	for(int k=0;k<p.length;k++)
	{
	    // Different normalizing is used by TradeVX.getRandomTradesPolynomialP, sign is the same.
	    p_o_TradeVX[k]=p[k]*Polynomials.fact(k);
	}
	final Trade [] ptb=new Trade [tsize];
	final long pcurrenttime=CalculateKendall.TradeVX.getRandomTradesPolynomialP(ptb,r,tau,p_o_TradeVX);
	final QVMDataL odata=QVMDataL.calculateQVMDataL(neq_t+1+r.nextInt(4),3+r.nextInt(4),pcurrenttime,tau,ptb,tsize,null);
	final double [] abc=odata.vmoments_Ltau_x_p,abc_t=odata.tmoments_Ltau_x_p;
	final double [] abc_scaled=TMatr.getVectorScaled(abc,abc_t[0]/abc[0]);
	final double [] lt=getLinsys2(1,neq_t+1,abc_t,odata.npm,abc_scaled)[1];

	// match lt through lp
	final double [] lp=new double [neq_t+1];
	final double [][] lsol=getLinsys2(2,neq_t,abc_t,odata.npm,abc_scaled);
	for(int k=0;k<neq_t;k++)
	{
	    lp[k]=lsol[1][k]+px[k]*lsol[0][1];
	}
	lp[neq_t]=px[neq_t]*lsol[0][1];
	lp[0]-=lsol[0][1]*odata.p_offset;

	System.err.println("lp="+java.util.Arrays.toString(lp));
	System.err.println("lt="+java.util.Arrays.toString(lt));
	UnitTests.checkDoubleArrayRel1(lt,"lt",lp,"lp",epssum);

    }


    System.err.println("QVMLPV:checkMatching_getLinsys2: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}



private static void checkMatching_getLinsys2Direct(final int neq_t,
						   final int neq_p,
						   final java.util.Random r,
						   final long currenttime,
						   final long tau,
						   final Trade [] tb,
						   final int tsize,
						   final double eps,
						   final double epssum,
						   final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf)
{
    final QVMData qd=QVMData.calculateQVMData(1,2*neq_t-1+r.nextInt(3),2*neq_p-1+r.nextInt(3),currenttime,tau,tb,tsize,null);
    final double [] p_abc=new double [qd.ntm*qd.npm],p_abc_t=new double [qd.ntm*qd.npm];
    for(int k=0,j=0,s=1;k<qd.ntm;k++,s=-s) // to match inverted time
    {
	for(int m=0;m<qd.npm;m++,j++)
	{
	    p_abc[j]=s*qd.vmoments_vol_x_tau_x_p[j];
	    p_abc_t[j]=s*qd.tmoments_vol_x_tau_x_p[j];
	}
    }
    final double [][] xd=testing_polynomialInput_getLinsys2(neq_p,neq_t,p_abc_t,qd.npm,p_abc,QVMDataL.Laguerre_POLS);
    final QVMLPV q=QVMLPV.calculateQVMLPV(neq_t,neq_p,currenttime,tau,tb,tsize,rf,null);
    final double [] pvcoefs=Polynomials.getPolynomialShifted(xd[0],q.odata.p_offset-qd.p_offset),icoefs=xd[1];
    icoefs[0]+=pvcoefs[0];pvcoefs[0]=0;

    System.err.println("1.icoefs="+java.util.Arrays.toString(icoefs));
    System.err.println("2.icoefs="+java.util.Arrays.toString(q.icoefs));
    System.err.println("1.pvcoefs="+java.util.Arrays.toString(pvcoefs));
    System.err.println("2.pvcoefs="+java.util.Arrays.toString(q.pvcoefs));
    UnitTests.checkDoubleArrayRel1(icoefs,"icoefs",q.icoefs,"q.icoefs",epssum);
    UnitTests.checkDoubleArrayRel1(pvcoefs,"pvcoefs",q.pvcoefs,"q.pvcoefs",epssum);


    System.err.println("QVMLPV:checkMatching_getLinsys2Direct: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}


    private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_t,final int neq_p,final double eps,final double epssum)
{
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/50,currenttime=tb[tb.length-1].time+1000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMLPV:checkCalculations(): test with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+" neq_t="+neq_t+" neq_p="+neq_p);

    final com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG rf=
	new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();

    // match to QVMLL
    checkMatchingQVMLL(neq_t,neq_p,r,currenttime,tau,tb,tsize,eps*0,eps,rf);
    // match to QATradeIP
    checkMatchingQATradeIP(neq_t,neq_p,r,currenttime,tau,tb,tsize,eps,epssum,rf);
    // other matches
    checkMatchingValues(neq_t,neq_p,r,currenttime,tau,tb,tsize,eps,epssum,rf);
    checkMatchingLinsystemValues(neq_t>5?5:neq_t,neq_p,r,currenttime,tau,tb,tsize,eps,epssum,rf);
    checkMatchingPolynomialP(neq_t>5?5:neq_t,neq_p,r,eps,epssum*5,rf,tsize,tau);
    checkMatching_getLinsys2(neq_t>6?6:neq_t,neq_p,r,currenttime,tau,tb,tsize,eps,epssum,rf);
    checkMatching_getLinsys2Direct(neq_t>6?6:neq_t,neq_p,r,currenttime,tau,tb,tsize,eps,epssum,rf);

    System.err.println("QVMLPV:checkCalculations(): OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+" neq_t="+neq_t+" neq_p="+neq_p);
}

public static void main(String [] args)
{
    System.err.println("QVMLPV::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-179091);
    final double eps=1e-6,epssum=1e-3;
    for(int neq_p=1;neq_p<=5;neq_p++)
    for(int neq_t=1;neq_t<=11;neq_t++)
    {
	for(int i=0;i<4;i++)
	{
	    checkCalculations(100010+r.nextInt(100010),r,neq_t,neq_p,eps,epssum);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println("QVMLPV::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
}

} // END of class QVMLPV





private static double getSum2(final int neq,
			      final double [] m,final int m_offset,final int m_increment,
			      final double [] n,final int n_offset,final int n_increment,
			      final double scale,final double [] t_norm)
{
    //System.err.println("neq="+neq+" t_norm.length="+t_norm.length);
    double s=0;
    for(int k=0,om=m_offset,on=n_offset;k<neq;k++,om+=m_increment,on+=n_increment)
    {
	s+=((m[om]/scale)*(n[on]/scale))*t_norm[k];
    }
    return s;
}

private static double [][] getPolynomialAndChristoffelFromMomTxP(
    final int neq_t,final int neq_p,
    final double [] abc,final int ntm,final int npm,
    final double [] t_norm)
{
    if(!(neq_t>neq_p) || !(npm>neq_p) || !(ntm>=neq_t))
	throw new IllegalArgumentException("neq_t="+neq_t+" neq_p="+neq_p+" npm="+npm+" ntm="+ntm);
    final double pt0=abc[0];
    final double [] matr=new double [neq_p*neq_p],pcoefs=new double [neq_p+1],norma=new double [neq_p];
    // calculate <Ip^k|Ip^l> moments, using finite sums, preserving positive determiness of the matrix.
    for(int k=0;k<neq_p;k++)
    {
	norma[k]=Math.sqrt(getSum2(neq_t,abc,k,npm,abc,k,npm,pt0,t_norm));
	matr[k+neq_p*k]=1;
	pcoefs[k]=-getSum2(neq_t,abc,neq_p,npm,abc,k,npm,pt0,t_norm)/norma[k];
	for(int l=k;--l>=0;)
	{
	    matr[l+neq_p*k]=matr[k+neq_p*l]=getSum2(neq_t,abc,k,npm,abc,l,npm,pt0,t_norm)/norma[k]/norma[l];
	}
    }

    final double [] wcoefs=new double [Math.max(2*neq_p-1,0)];
    final int [] ipvt=new int [neq_p];
    final int pinfo=com.polytechnik.utils.Linsystems.dgefa_m1(matr,neq_p,neq_p,ipvt);
    if(pinfo!=0)
    {
	java.util.Arrays.fill(pcoefs,Double.NaN);
	java.util.Arrays.fill(wcoefs,Double.NaN);
	return new double [][]{pcoefs,wcoefs};
    }

    com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq_p,neq_p,ipvt,pcoefs,1);
    for(int k=0;k<neq_p;k++){pcoefs[k]/=norma[k];}
    pcoefs[neq_p]=1; // monian polynomial
    final double [] rpart=new double [neq_p];
    for(int k=0;k<neq_p;k++)
    {
	rpart[k]=1;
	com.polytechnik.utils.Linsystems.dgesl_m1(matr,neq_p,neq_p,ipvt,rpart,1);
	for(int l=0;l<neq_p;l++)
	{
	    wcoefs[k+l]+=rpart[l]/norma[l]/norma[k];
	    rpart[l]=0;
	}
    }
    return new double [][]{pcoefs,wcoefs};
}


private static double [][] getGrammSchmidtOrthogonal(final int neq_t,final int neq_p,
						     final double [] abc,final int ntm,final int npm,
						     final double [] t_norm)
{
    if(!(neq_t>neq_p) || !(npm>neq_p) || !(ntm>=neq_t))
	throw new IllegalArgumentException("neq_t="+neq_t+" neq_p="+neq_p+" npm="+npm+" ntm="+ntm);
    final double [][] basis=new double [neq_p+1][];

    final double pt0=abc[0];

    for(int i=0;i<=neq_p;i++)
    {
	final double [] b=basis[i]=new double [i+1];
	b[i]=1;
	for(int k=0;k<i;k++)
	{
	    double scal=0;
	    for(int j=0;j<=k;j++)
	    {
		scal+=basis[k][j]*getSum2(neq_t,abc,j,npm,abc,i,npm,pt0,t_norm);
	    }
	    TMatr.sumP1(b,basis[k],-scal);
	}
	// normalize
	double s=0;
	for(int j=0;j<=i;j++)
	for(int k=0;k<=i;k++)
	{
	    s+=b[j]*b[k]*getSum2(neq_t,abc,j,npm,abc,k,npm,pt0,t_norm);
	}
	final double anorm=Math.sqrt(s);
	for(int j=0;j<=i;j++)
	{
	    b[j]/=anorm;
	}
    }

    return basis;
}




/** Like QVMLL but in price polynomial basis. */
static class QVMLpi {

final int neq_t,neq_p;
private final QVMDataL odata;
final double pvtf_E,ipvtf_E;
final double [] p_imin,icoefs;

private QVMLpi(final int neq_t,final int neq_p,final QVMDataL odata,
	       final com.polytechnik.utils.PolynomialRootsFinder rf)
{
    this.odata=odata;
    this.neq_t=neq_t;
    this.neq_p=neq_p;
    final double [] t_norm=new double [neq_t];
    java.util.Arrays.fill(t_norm,1); // user t_norm=1
    final double [][] x=getPolynomialAndChristoffelFromMomTxP(neq_t,neq_p,odata.vmoments_Ltau_x_p,odata.ntm,odata.npm,t_norm);

    final double [] pcoefs=x[0],wcoefs=x[1];
    System.err.println("pcoefs="+java.util.Arrays.toString(pcoefs)+"\twcoefs="+java.util.Arrays.toString(wcoefs));
    if(true)
    {
	final double [][] xx=getPolynomialAndChristoffelFromMomTxP(neq_t,neq_p,TMatr.getVectorScaled(odata.vmoments_Ltau_x_p,4.6781),odata.ntm,odata.npm,t_norm);
	System.err.println("Apcoefs="+java.util.Arrays.toString(xx[0])+"\twcoefs="+java.util.Arrays.toString(xx[1]));

    }

    final PolynomialRootsFinder.PRoots u;
    if(neq_p>4){
	if(rf==null) {
	    u=new PolynomialRootsFinder.PRoots(0,pcoefs,new double [neq_p],new double [neq_p]);
	} else {
	    u=rf.findRoots(pcoefs);
	}
    } else {
	u=com.polytechnik.utils.PolynomialRootsElementary.ROOTSFINDER.findRoots(pcoefs);
    }
    final int nroots=u.nroots;
    final double [] root_re=u.re_root,root_im=u.im_root;

    // insertion sort by real part with extra load
    SortArrays.sortArrayWithLoad(nroots,root_re,root_im);

    this.p_imin=java.util.Arrays.copyOf(root_re,
	CalculateKendall.selectRealRoots(root_re,nroots,root_re,root_im,CalculateKendall.EPS_IM));

    // FIXME
    if(this.p_imin.length!=neq_p) System.err.println("!!!!!!!!!!! p_imin.length="+p_imin.length+" neq_t="+neq_t+" neq_p="+neq_p);
    //if(this.p_imin.length!=neq_p) throw new RuntimeException("p_imin.length="+p_imin.length+" neq_p="+neq_p);

    this.icoefs=this.p_imin.length==root_im.length?root_im:new double [this.p_imin.length]; // try to re-use an array
    int ind_max=-1;
    double imax=Double.NEGATIVE_INFINITY;
    for(int k=this.p_imin.length;--k>=0;)
    {
	final double pk=this.p_imin[k];
	this.icoefs[k]=1/Polynomials.getD0(wcoefs,pk);
	this.p_imin[k]+=odata.p_offset; // shift to 0 offset
	if(this.icoefs[k]>=imax) imax=this.icoefs[ind_max=k];
    }
    System.err.println("INIT icoefs="+java.util.Arrays.toString(icoefs)+" sum="+sumElementse(icoefs,icoefs.length)+" neq_t="+neq_t+" neq_p="+neq_p);
    if(ind_max>=0)
    {
	this.pvtf_E=this.p_imin[ind_max];
	this.ipvtf_E=this.icoefs[ind_max];
    }
    else
    {
	this.pvtf_E=this.ipvtf_E=Double.NaN;
    }
}



public static QVMLpi calculateQVMLpi(final int neq_t,
				     final int neq_p,
				     final long currenttime,
				     final long tau,
				     final Trade [] tb,
				     final int tsize,
				     final com.polytechnik.utils.PolynomialRootsFinder rf,
				     final QVMLpi ref)
{
    final QVMDataL o=QVMDataL.calculateQVMDataL(neq_t,
						neq_p+1,
						currenttime,
						tau,
						tb,
						tsize,
						ref!=null?ref.odata:null);
    return new QVMLpi(neq_t,neq_p,o,rf);
}




private static void checkMatchingGrammSchmidt(final int neq_t,
					      final int neq_p,
					      final java.util.Random r,
					      final long currenttime,
					      final long tau,
					      final Trade [] tb,
					      final int tsize,
					      final double eps,
					      final double epssum,
					      final com.polytechnik.utils.PolynomialRootsFinder rf)
{
    final double [] t_norm=new double [neq_t];
    java.util.Arrays.fill(t_norm,1); // user t_norm=1
    final QVMDataL odata=QVMDataL.calculateQVMDataL(neq_t+r.nextInt(3),
						    neq_p+1+r.nextInt(3),
						    currenttime,
						    tau,
						    tb,
						    tsize,
						    null);
    final double [][] x=getPolynomialAndChristoffelFromMomTxP(neq_t,neq_p,odata.vmoments_Ltau_x_p,odata.ntm,odata.npm,t_norm);
    final double [] pcoefs=x[0];

    final double [][] basis=getGrammSchmidtOrthogonal(neq_t,neq_p,odata.vmoments_Ltau_x_p,odata.ntm,odata.npm,t_norm);
    final double [] pcoefsgs=basis[neq_p].clone();
    for(int k=0;k<=neq_p;k++) {pcoefsgs[k]/=pcoefsgs[neq_p];}
    UnitTests.checkDoubleArrayRel1(pcoefs,"pcoefs",pcoefsgs,"pcoefsgs",eps);

    final double [] wcoefs=x[1],wcoefsgs=new double [Math.max(2*neq_p-1,0)];
    for(int k=0;k<neq_p;k++) {DeprecatedHelperTMatr.sumP2(wcoefsgs,basis[k]);}
    UnitTests.checkDoubleArrayRel1(wcoefs,"wcoefs",wcoefsgs,"wcoefsgs",eps);
    System.err.println("wcoefs="+java.util.Arrays.toString(wcoefs)+"\nwcoefsgs="+java.util.Arrays.toString(wcoefsgs)+"\nneq_t="+neq_t+" neq_p="+neq_p);

    System.err.println("checkMatchingGrammSchmidt: tests OK  neq_t="+neq_t+" neq_p="+neq_p+" eps="+eps+" epssum="+epssum);
}



private static void checkMatchingQVMLL(final int neq_t,
				       final java.util.Random r,
				       final long currenttime,
				       final long tau,
				       final Trade [] tb,
				       final int tsize,
				       final double eps,
				       final double epssum,
				       final com.polytechnik.utils.PolynomialRootsFinder rf)
{

    final QVMLL ref=QVMLL.calculateQVMLL(neq_t,currenttime,tau,tb,tsize,null,rf);
    final QVMLpi q=calculateQVMLpi(neq_t,1,currenttime,tau,tb,tsize,rf,null);

    for(int m=1;m<neq_t;m++)
    {
	final QVMLpi q2=calculateQVMLpi(neq_t,m,currenttime,tau,tb,tsize,rf,null);
	System.err.println("q2.icoefs="+java.util.Arrays.toString(q2.icoefs));
	/*
	System.err.println("q2.p_imin="+java.util.Arrays.toString(q2.p_imin)+" q.pvtf_E="+q.pvtf_E);
	if(true)
	{
	    // test integral calculation
	    final int n=2*q2.neq_p-1;
	    final double [] tpol=new double [n+1];
	    for(int i=n;i>=0;i--) tpol[i]=2*(0.5-r.nextDouble());
	    double si=0;
	    for(int i=0;i<q2.icoefs.length;i++)
	    {
		si+=q2.icoefs[i]*Polynomials.getD0(tpol,q2.p_imin[i]-q2.odata.p_offset);
	    }
	    final double msi=getAverage(tpol,q2.test_ipm,q2.test_ipm[0]);
	    System.err.println("checkMatchingQVMLL:integral OK: si="+si+" msi="+msi+" for n="+n+" q2.neq_t="+q2.neq_t+" q2.neq_p="+q2.neq_p);
	    UnitTests.checkDoubleRel(si,"si",msi,"msi",eps);
	}
	if(true && 2*m-1>=2)
	{
	    final double [] tpol=new double []{0,0,1};
	    double si=0;
	    for(int k=0;k<q2.p_imin.length;k++)
	    {
		si+=q2.icoefs[k]*Polynomials.getD0(tpol,q2.p_imin[k]-q2.odata.p_offset);
	    }
	    UnitTests.checkDoubleRel(ref.sigma_pvtf*ref.sigma_pvtf+Math.pow(ref.pvtf-ref.odata.p_offset,2),"ref.d2",si,"si",eps);
	    System.err.println("checkMatchingQVMLL:stdev QVMLL OK");
	}
	*/
    }


    System.err.println("QVMLpi:checkMatchingQVMLL: tests OK  neq_t="+neq_t+" eps="+eps+" epssum="+epssum);
}



private static void checkCalculations(final int tsize,final java.util.Random r,final int neq_t,final int neq_p,final double eps,final double epssum)
{
    final Trade [] tb=CalculateKendall.TradeEMoments.getRandomTrades(tsize,r);
    final long tau=10000L*tb.length/30,currenttime=tb[tb.length-1].time+1000;
    System.err.println();
    System.err.println("------------------------------------------------------------------------");
    System.err.println("QVMLpi:checkCalculations(): test with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " neq_t="+neq_t+" neq_p="+neq_p);

    final com.polytechnik.utils.PolynomialRootsFinder rf=
	new com.polytechnik.utils.PolynomialRootsJenkinsTraubORIG();

    //checkMatchingQVMLL(neq_t,r,currenttime,tau,tb,tsize,eps,epssum,null,rf);
    checkMatchingGrammSchmidt(neq_t,neq_p,r,currenttime,tau,tb,tsize,eps,epssum,null);
    System.err.println("QVMLpi:checkCalculations(): OK with N points="+tsize+
		       " currenttime="+currenttime+" tau="+tau+
		       " neq_t="+neq_t+" neq_p="+neq_p);
}


public static void main(String [] args)
{
    System.err.println("QVMLpi::main(): starting tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    final long t0=System.currentTimeMillis();
    final java.util.Random r=new java.util.Random(-79092);
    final double eps=1e-5,epssum=1e-5;
    for(int neq_t=1;neq_t<=5;neq_t++)
    for(int neq_p=0;neq_p<neq_t;neq_p++)
    {
	for(int i=0;i<110;i++)
	{
	    checkCalculations(10010+r.nextInt(10010),r,neq_t,neq_p,eps,epssum);
	}
    }
    final long t1=System.currentTimeMillis();
    System.err.println("QVMLpi::main(): tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
}

} // END of class QVMLpi


static void test_QVM_internal(String [] args)
{
     //QVM.QVMEV.main(args);
    QVM.QVMORT.main(args);
    QVM.QVMMORT.main(args);
    QVM.QVMMU.main(args);
    QVM.QVMSORT.main(args);
    QVM.QVMLPV.main(args);
    QVM.QVMLpi.main(args);
    QVM.QVMAverage.main(args);
    QVM.QVMDataMU.main(args);
}



public static void main(String [] args)
{
    System.err.println("QVM::main(): starting ALL tests; ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
    final long t0=System.currentTimeMillis();
    System.err.println("Version");
    System.err.println("Started="+new java.util.Date(t0));
    System.getProperties().list(System.err);
    System.err.println("END Version");

    Test_DataInterfaceToMoments.main(args);
    SM.main(args);
    Polynomials.main(args);
    QVMData.main(args);
    test_QVM_internal(args);
    Laguerre.main(args);
    QVML.main(args);
    QVMDataL.main(args);
    QVMDataBL.main(args);
    QVMLL.main(args);
    LegendreShifted.main(args);
    QVMDataP.main(args);
    QVMDataBP.main(args);
    QVMDataBMP.main(args);
    UkbQVMBMP.main(args);
    QVMPP.main(args);
    QVMpi.main(args);
    QkQVMpi.main(args);
    QmQVMpi.main(args);
    SPIMinLaguerre.main(args);
    SPIMinLegendreShifted.main(args);
    QVMLMV.main(args);
    QVMPMV.main(args);
    QVMDLaguerre.main(args);
    QVMDLegendreShifted.main(args);
    AQVMLaguerre.main(args);
    AQVMLegendreShifted.main(args);
    EQVMLaguerre.main(args);
    EQVMLegendreShifted.main(args);
    FQVMLaguerre.main(args);
    FQVMLegendreShifted.main(args);
    IQVMLaguerre.main(args);
    IQVMLegendreShifted.main(args);
    RQVMLaguerre.main(args);
    RQVMLegendreShifted.main(args);
    SQVMLaguerre.main(args);
    SQVMLegendreShifted.main(args);
    KQVMLaguerre.main(args);
    KQVMLegendreShifted.main(args);
    LQVMLaguerre.main(args);
    LQVMLegendreShifted.main(args);
    RadauLQVMLaguerre.main(args);
    RadauLQVMLegendreShifted.main(args);
    MQVMLaguerre.main(args);
    MQVMLegendreShifted.main(args);
    NQVMLaguerre.main(args);
    NQVMLegendreShifted.main(args);
    OQVMLaguerre.main(args);
    OQVMLegendreShifted.main(args);
    QQVMLaguerre.main(args);
    QQVMLegendreShifted.main(args);
    UQVMLaguerre.main(args);
    UQVMLegendreShifted.main(args);
    VQVMLaguerre.main(args);
    VQVMLegendreShifted.main(args);
    WQVMLaguerre.main(args);
    WQVMLegendreShifted.main(args);
    XQVMLaguerre.main(args);
    XQVMLegendreShifted.main(args);
    YQVMLaguerre.main(args);
    YQVMLegendreShifted.main(args);
    ZQVMLaguerre.main(args);
    ZQVMLegendreShifted.main(args);
    ZVQVMLaguerre.main(args);
    ZVQVMLegendreShifted.main(args);
    TQVMLaguerre.main(args);
    TQVMLegendreShifted.main(args);
    BQVMLaguerre.main(args);
    BQVMLegendreShifted.main(args);
    CQVMLaguerre.main(args);
    CQVMLegendreShifted.main(args);
    CQVMMonomialsT.main(args);
    DQVMLaguerre.main(args);
    DQVMLegendreShifted.main(args);
    DQVMMonomials.main(args);
    QVMFQVMDataBL.main(args);
    DoQVMLegendreShifted.main(args);
    DoQVMLaguerre.main(args);
    HaarQVM.main(args);
    BiQVMLaguerre.main(args);
    BiQVMLegendreShifted.main(args);
    BjQVMLaguerre.main(args);
    BjQVMLegendreShifted.main(args);
    BkQVMLaguerre.main(args);
    BkQVMLegendreShifted.main(args);
    BkQVMMonomials.main(args);
    CiQVMLegendreShifted.main(args);
    DiQVMLaguerre.main(args);
    DiQVMLegendreShifted.main(args);
    DjQVMLaguerre.main(args);
    DjQVMLegendreShifted.main(args);
    DkQVMLaguerre.main(args);
    DkQVMLegendreShifted.main(args);
    EkQVMLaguerre.main(args);
    EkQVMLegendreShifted.main(args);
    IkQVMMonomials.main(args);
    IkQVMLaguerre.main(args);
    IkQVMLegendreShifted.main(args);
    JkQVMMonomials.main(args);
    JkQVMLaguerre.main(args);
    JkQVMLegendreShifted.main(args);
    KkQVMMonomials.main(args);
    KkQVMLaguerre.main(args);
    KkQVMLegendreShifted.main(args);
    LkQVMLaguerre.main(args);
    LkQVMLegendreShifted.main(args);
    LwQVMMonomials.main(args);
    LwQVMLaguerre.main(args);
    LwQVMLegendreShifted.main(args);
    MkQVMMonomials.main(args);
    MkQVMLaguerre.main(args);
    MkQVMLegendreShifted.main(args);
    NkQVMMonomials.main(args);
    NkQVMLaguerre.main(args);
    NkQVMLegendreShifted.main(args);
    OkQVMMonomials.main(args);
    OkQVMLaguerre.main(args);
    OkQVMLegendreShifted.main(args);
    QkQVMMonomials.main(args);
    QkQVMLaguerre.main(args);
    QkQVMLegendreShifted.main(args);
    RkQVMMonomials.main(args);
    RkQVMLaguerre.main(args);
    RkQVMLegendreShifted.main(args);
    RkiQVMMonomials.main(args);
    RkiQVMLaguerre.main(args);
    RkiQVMLegendreShifted.main(args);
    PkQVMMonomials.main(args);
    PkQVMLaguerre.main(args);
    PkQVMLegendreShifted.main(args);
    SkQVMMonomials.main(args);
    SkQVMLaguerre.main(args);
    SkQVMLegendreShifted.main(args);
    TkQVMLaguerre.main(args);
    TkQVMLegendreShifted.main(args);
    TkQVMMonomials.main(args);
    UkQVMLaguerre.main(args);
    UkQVMLegendreShifted.main(args);
    UkQVMMonomials.main(args);
    UnQVMLaguerre.main(args);
    UnQVMLegendreShifted.main(args);
    UnQVMMonomials.main(args);
    UMnQVMLaguerre.main(args);
    UMnQVMLegendreShifted.main(args);
    UMnQVMMonomials.main(args);
    UMsQVMLaguerre.main(args);
    UMsQVMLegendreShifted.main(args);
    UMsQVMMonomials.main(args);
    VkQVMLaguerre.main(args);
    VkQVMLegendreShifted.main(args);
    VkQVMMonomials.main(args);
    WkQVMLaguerre.main(args);
    WkQVMLegendreShifted.main(args);
    WkQVMMonomials.main(args);
    XkQVMLaguerre.main(args);
    XkQVMLegendreShifted.main(args);
    XkQVMMonomials.main(args);
    YkQVMLaguerre.main(args);
    YkQVMLegendreShifted.main(args);
    YkQVMMonomials.main(args);
    YpkQVMLaguerre.main(args);
    YpkQVMLegendreShifted.main(args);
    YpkQVMMonomials.main(args);
    YpnlQVMLaguerre.main(args);
    YpnlQVMLegendreShifted.main(args);
    YpnlQVMMonomials.main(args);
    YkDpILaguerre.main(args);
    YkDpILegendreShifted.main(args);
    YkDpIMonomials.main(args);
    UMlQVMLaguerre.main(args);
    UMlQVMLegendreShifted.main(args);
    UMlQVMMonomials.main(args);
    UMdQVMLegendreShifted.main(args);
    ImpactQVMLaguerre.main(args);
    ImpactQVMLegendreShifted.main(args);
    ImpactQVMMonomials.main(args);
    ImpactQVMMuseLaguerre.main(args);
    ImpactQVMMuseLegendreShifted.main(args);
    ImpactQVMMuseMonomials.main(args);
    ImpactQVMMuse_pi.main(args);
    QkPnLpi_PBasis.main(args);
    DynMicroImpactLaguerre.main(args);
    DynMicroImpactLegendreShifted.main(args);
    DynMicroImpactMonomials.main(args);
    DynMicroImpact_pi.main(args);
    DynImpactAsymmetryLaguerre.main(args);
    DynImpactAsymmetryLegendreShifted.main(args);
    DynImpactAsymmetryMonomials.main(args);
    DynImpactAsymmetry_pi.main(args);
    IAImpactLaguerre.main(args);
    IAImpactLegendreShifted.main(args);
    IAImpactMonomials.main(args);
    IAImpact_pi.main(args);
    IStateLaguerre.main(args);
    IStateLegendreShifted.main(args);
    IStateMonomials.main(args);
    IState_pi.main(args);
    IHStateLaguerre.IHStateLaguerreTest.main(args);
    IHStateLegendreShifted.IHStateLegendreShiftedTest.main(args);
    IHStateMonomials.IHStateMonomialsTest.main(args);
    IHStateDMLaguerre.main(args);
    IHStateDMLegendreShifted.main(args);
    IHStateDMMonomials.main(args);
    ScalpedMaxIProjectionLaguerre.main(args);
    ScalpedMaxIProjectionLegendreShifted.main(args);
    ScalpedMaxIProjectionMonomials.main(args);
    SupportPriceLaguerre.main(args);
    SupportPriceLegendreShifted.main(args);
    SupportPriceMonomials.main(args);
    TstPnLILowIHighLaguerre.main(args);
    TstPnLILowIHighLegendreShifted.main(args);
    TstPnLILowIHighMonomials.main(args);
    PnLInPsiHstateLaguerre.PnLInPsiHstateLaguerreTest.main(args);
    PnLInPsiHstateLegendreShifted.PnLInPsiHstateLegendreShiftedTest.main(args);
    PnLInPsiHstateMonomials.PnLInPsiHstateMonomialsTest.main(args);

    // slow tests in the end
    CalculateKendall.main(args);
    UnitTests.main(args);

    final long t1=System.currentTimeMillis();
    System.err.println("Version");
    System.err.println("Finished="+new java.util.Date(t1)+" Started="+new java.util.Date(t0));
    System.getProperties().list(System.err);
    System.err.println("END Version");
    System.err.println("QVM::main(): ALL tests OK done in "+(t1-t0)+" milliseconds. ver=$Id: QVM.java,v 1.1752 2022/09/10 13:25:37 mal Exp $");
}


} // END of class QVM
